#!/usr/bin/perl
use strict;
use Getopt::Long;

my $version = "0.2";
my $GL = $ENV{'GLOBUS_LOCATION'};
my (@tests, @gridmaps, %testhash, $interactive, $all, $help, $verbose);
my %cahash;
my %certhash;
my %modulus;

@tests = ("gsi");

GetOptions( 'tests=s' => \@tests,
            'verbose!' => \$verbose,
            'interactive!' => \$interactive,
            'all!' => \$all,
            'help' => \$help,
) or &short_usage();

if ( $help )
{
   &short_usage();
   exit;
}

@tests = split(/,/,join(',',@tests));
for my $test (@tests)
{
    $testhash{$test} = 1;
}

sub banner
{
   my ($banline) = @_;
   print "@" x 80 . "\n";
   print "$banline";
   print "@" x 80 . "\n";
}


sub die_if_bad
{
   if ( $? != 0 )
   {
	die "@_";
   }
}

sub run_cmd_quiet
{
   my @cmd = @_;
   system("@cmd > /dev/null 2>&1");
   return $?;
}

sub check_for_file
{
   my ($file) = @_;

   if ( not -e $file )
   {
	print "$file not found.\n" if $verbose;
	return 1;
   } else
   {
	print "$file found.\n";
        system("ls -l $file");
	return 0;
   }
}

sub check_for_exe
{
   my ($exe) = @_;

   system("which $exe");
   if ( $? != 0 )
   {
	print "$exe not found.\n" if $verbose;
	return 1;
   } else
   {
	print "$exe found.\n";
	return 0;
   }
}
sub check_for_dir
{
   my ($dir) = @_;
   if ( not -d $dir )
   {
	print "$dir not found.\n" if $verbose;
	return 1;
   } else
   {
	print "$dir found.\n";
	return 0;
   }
}

sub check_for_env
{
   my ($ev) = @_;
   if ( $ENV{"$ev"} )
   {
	print "$ev is set to $ENV{$ev}\n";
	return 0;
   } else
   {
	print "$ev unset.\n" if $verbose;
	return 1;
   }
}

sub short_usage()
{
   print "Usage: $0\n";
   print "\nSet your GLOBUS_LOCATION to an existing installation ";
   print "for a report on\n configuration.\n\n";
   #print "Use --all to see all test reports.  The tests are:\n";
   #print "\tcompile: Compile environment (does not require GLOBUS_LOCATION)\n";
   #print "\tgsi: Security information\n";
   #print "\tgram: Resource Management information\n";
   #print "\tgridftp: GridFTP information\n";
   #print "\tmds: Information Services information\n";
   print "\n";
   #print "You can select tests with --tests=.  For instance,\n";
   #print "$0 --tests=gsi,gram\n";
}

sub env_setup()
{
   $ENV{'PATH'} = "$GL/bin:$ENV{'PATH'}";
   $ENV{'LD_LIBRARY_PATH'} = "$GL/lib:$ENV{'LD_LIBRARY_PATH'}";
}

sub gl_check()
{
   my $res = check_for_env("GLOBUS_LOCATION");
   if ( $res != 0 )
   {
        print "ERROR: GLOBUS_LOCATION not set.\n";
	short_usage();
	exit 1;
   }
}

sub prelim_checks()
{
   if ( ! @tests )
   {
        print "ERROR: No tests selected.\n";
	short_usage();
	exit 1;
   }
   print "Whoami says you are: " . `whoami`;
   print "id says: " .  `id`;
   print "This is version $version of the script.\n";
   print;
}

sub compile_checks()
{
  banner "Checking for compile-time prerequisites:\n";
  check_for_exe("ant");
  check_for_exe("java");
  check_for_exe("gcc");

  check_for_env("CC");
  check_for_env("GLOBUS_CC");
  check_for_env("JAVA_HOME");
  check_for_env("ANT_HOME");
  check_for_env("CLASSPATH");
  print "\n";
}

sub file_and_issuer
{
   my ($file) = @_;
   my $res =  check_for_file $file;
   if ( $res eq 0 )
   {
       print "Checking issuer hash for $file.\n";
       my ($issuer, $modulus) = `openssl x509 -in $file -issuer_hash -modulus -noout`;
       chomp $issuer;
       chomp $modulus;
       print "I see issuer hash $issuer\n";
       push @{$certhash{$file}}, $issuer;
       push @{$certhash{$file}}, $modulus;
       if ( $cahash{$issuer} )
       {
          foreach my $dir ( @{$cahash{$issuer}} ) {
             print "There's a CA for this cert in $dir\n";
          }
       } else {
          print "I have not seen a CA that matches this hash.\n";
       }
   }

   return $res;
}

# Store a CA hash value with the directory it was seen
sub load_ca_dir
{
   my ($dir) = @_;
   my $res =  check_for_dir $dir;
   if ( $res eq 0 )
   {
      for my $file (<$dir/????????.?>)
      {
         print "Saw CA $file\n";
         $file =~ s#$dir/(........)\.\d#$1#;
         push @{$cahash{$file}}, $dir;

         if ( ! -f "$dir/$file.signing_policy" )
         {
            print "No signing policy was found for that CA!\n";
         }
      }
   }
}

sub check_for_gridmap
{
   my ($file) = @_;

   my $res = check_for_file $file;
   if ( $res eq 0 )
   {
      push @gridmaps, $file;
   }

   return $res;
}

sub cert_pair
{
   my ($cert, $key) = @_;

   my ($certok, $keyok, $modcert, $modkey);

   $certok = file_and_issuer($cert);
   $keyok = check_for_file($key);

   if ($keyok eq 0)
   {
      if ($interactive) {
         print "You will need to enter your key's password for this check\n";
         $modkey = `openssl rsa -in $key -modulus -noout`;
         chomp $modkey;
      } else {
         print "Skipping modulus test.  Use --interactive to supply a password for your key\n";
      }
   }

   if ( ($certok eq 0) && ($keyok eq 0)  )
   {
      if ( ($certhash{$cert}[1] ne $modkey) && $interactive)
      {
         print "$cert and $key have mismatched modulus\n";
         return 1;
      } else {
         print "$cert and $key have matching moduli\n" unless (! $interactive);
         return 0;
      }
   } else {
      return 1;
   }
}

sub system_gsi_checks()
{
   my $res;

   # General host security
   print "Checking general host security...\n";
   load_ca_dir  "/etc/grid-security/certificates";
   load_ca_dir  "$GL/share/certificates";

   print "Done looking for CA dirs.\n\n";

   print "Looking for /etc/grid-security files.\n";
   cert_pair("/etc/grid-security/hostcert.pem", "/etc/grid-security/hostkey.pem");
   cert_pair("/etc/grid-security/containercert.pem", "/etc/grid-security/containerkey.pem");
   cert_pair( "$GL/etc/hostcert.pem", "$GL/etc/hostkey.pem");
   check_for_gridmap "/etc/grid-security/grid-mapfile";

   print "Checking for fully-qualfied hostname: ";
   system("globus-hostname");
   print "\n";
}

sub gridmap_check
{
   my ($cert) = @_;
   my $found = 1;

   my $subject = `grid-cert-info -subject -file $cert`;
   chomp $subject;
   print "Looking for $subject in gridmap files ...\n";

   for my $gm ( @gridmaps )
   {
      my $res = system("grep \"$subject\" $gm > /dev/null");

      if ( $res eq 0 )
      {
          print "Found that subject in \"$gm\".\n";
	  $found = 0;
      }
   }

   if ( $found eq 1 ) {print "$subject not found in any gridmap file.\n";}
   return $found;
}

sub usercert_pair
{
   my ($cert, $key) = @_;

   my $res = cert_pair( $cert, $key );

   if ( $res eq 0 )
   {
      $res = gridmap_check $cert;
   }

   return $res;
}

sub user_gsi_checks()
{
   my $res;

   print "Checking user security\n";
   check_for_env "GLOBUS_HOSTNAME";

   # Must finish looking for certdirs before
   # checking usercerts
   print "Checking for userland cert dirs.\n";
   $res = check_for_env "X509_CERT_DIR";
   if ( $res eq 0 )
   {
      load_ca_dir "$ENV{'X509_CERT_DIR'}";
   }
   check_for_env "GRID_SECURITY_DIR";
   load_ca_dir "$ENV{'HOME'}/.globus/certificates";
   print "\n";

   # Must finish looking for gridmaps before
   # checking usercerts
   print "Looking for userland gridmaps.\n";
   $res = check_for_env "GRIDMAP";
   if ( $res eq 0 )
   {
      check_for_gridmap "$ENV{'GRIDMAP'}";
   }
   check_for_gridmap "$ENV{'HOME'}/.gridmap";

   print "\nLooking for userland certificates.\n";
   $res = check_for_env "X509_USER_CERT";
   if ( $res eq 0 )
   {
      file_and_issuer $ENV{'X509_USER_CERT'};
      gridmap_check($ENV{'X509_USER_CERT'});
   }

   check_for_env "X509_USER_KEY";
   check_for_env "X509_USER_PROXY";

   cert_pair( "$ENV{'HOME'}/.globus/hostcert.pem",
              "$ENV{'HOME'}/.globus/hostkey.pem" );
   usercert_pair( "$ENV{'HOME'}/.globus/usercert.pem", "$ENV{'HOME'}/.globus/userkey.pem" );
   check_for_file "$ENV{'HOME'}/.globus/usercred.p12";
   check_for_file "$ENV{'HOME'}/.globus/cog.properties";

   $res = run_cmd_quiet("grid-proxy-info -exists");
   if ( $res eq 0 )
   {
      print "grid-proxy-info says you have a proxy.\n";
   } else {
      print "grid-proxy-info says you don't have a proxy right now.\n";
   }
   
   print "\n";
}

sub simple_ca_checks()
{
   print "Checking if SimpleCA has been setup from this location.\n";
   my $glob = <$GL/setup/globus_simple_ca_????????_setup>;
   check_for_dir($glob);
   print "Checking if there is a SimpleCA in this user's directory.\n";
   check_for_dir("$ENV{'HOME'}/.globus/simpleCA");
   file_and_issuer("$ENV{'HOME'}/.globus/simpleCA/cacert.pem");
   print "\n";
}

sub gsi_checks()
{
   banner "GSI checks\n";
   print "For explanation of why we check what, see http://www-unix.globus.org/toolkit/docs/development/4.0-drafts/security/prewsaa/Pre_WS_AA_Interface_Config_Frag.html\n";
   system_gsi_checks;
   user_gsi_checks;
   simple_ca_checks;
}

sub search_services
{
   my ($portno) = @_;

   open(SERV, "</etc/services") or die "Can't open /etc/services\n";
   while( <SERV> )
   {
      if (/^([-_\w]+)\s+(\d+).*/)
      {
         if ( $2 eq "$portno" )
	 {
	     print "Found service $1 on port $portno\n";
	     close SERV;
	     return $1;
         }
      }
   }

   close SERV;
   return "";
}

sub search_xinetd
{
   my ($service) = @_;
   my $found = 0;

   return if ( $service eq "" );

   print "Looking for $service in /etc/xinetd.d/*\n";
   my @xinetds = </etc/xinetd.d/*>;
   for my $entry (@xinetds)
   {
      system("grep $service $entry > /dev/null");
      if ( $? eq 0 )
      {
         $found = 1;
         print "Found likely xinetd entry: $entry\n";
         system("cat $entry");
      }
   }

   print "No xinetd entries found for $service\n" unless $found; 
}

sub pre_ws_gram_checks
{
   my $schedfound = 0;
   my $candidate;

   print "Pre-WS GRAM checks:\n";
   print "Looking for port 2119 in /etc/services\n";

   $candidate = search_services(2119);
   if ( $candidate eq "" ) 
   {
      print "Did not find anything in /etc/services on port 2119\n";
   } else {
      search_xinetd($candidate);
   }

   print "Looking for schedulers\n";
   my @jobman = <$GL/etc/grid-services/*>;
   for my $entry (@jobman)
   {
      print "I see scheduler $entry\n";
      $schedfound = 1;
   }
   print "No schedulers found in $GL/etc/grid-services/\n" unless $schedfound;
   print "\n";
}

sub ws_gram_checks
{
   my $schedfound = 0;

   print "WS GRAM checks:\n";
   check_for_file "/etc/grid-security/grim-port-type.xml";
   check_for_file "$GL/bin/globus-grim";
   check_for_file "$GL/bin/launch_uhe_setuid";
   check_for_file "$GL/bin/launch_uhe.sh";

   my @scheds = <$GL/lib/perl/Globus/GRAM/JobManager/*.pm>;
   for my $entry (@scheds)
   {
      print "I see scheduler $entry\n";
      $schedfound = 1;
   }
   print "I don't see schedulers in $GL/lib/perl/Globus/GRAM/JobManager\n" unless $schedfound;

   print "\n";
}

sub gram_checks()
{
   banner "GRAM checks\n";
   pre_ws_gram_checks;
   ws_gram_checks;
}

sub pre_ws_gridftp_checks()
{
   my $servicefound = 0;
   my $xinetdfound = 0;

   print "Looking for port 2811 in /etc/services\n";

   my $candidate = search_services(2811);
   if ( $candidate eq "" ) 
   {
      print "Did not find anything in /etc/services on port 2811\n";
   } else {
      search_xinetd($candidate);
   }

   print "Checking for GAA callouts\n";
   check_for_file "/etc/grid-security/gsi-gaa.conf";
   check_for_file "/etc/grid-security/gsi-authz.conf";
   check_for_file "$GL/etc/grid-security/gsi-gaa.conf";
   check_for_file "$GL/etc/grid-security/gsi-authz.conf";
   print "\n";
}

sub gridftp_checks()
{
   banner "GridFTP checks\n";
   pre_ws_gridftp_checks;
}

sub pre_ws_mds_checks
{
   my $initdfound = 0;

   my @initds = </etc/init.d/*>;
   for my $entry (@initds)
   {
      system("grep globus-mds $entry > /dev/null ");
      if ( $? eq 0 )
      {
         print "Found likely MDS init.d entry: $entry\n";
         system("cat $entry");
         $initdfound = 1;
      }
   }
   print "No init.d entry found\n" if ( $initdfound eq 0 );

   check_for_file "/etc/grid-security/ldap/ldapcert.pem";
   check_for_file "/etc/grid-security/ldap/ldapkey.pem";
   check_for_file "$GL/etc/grid-security/ldap/ldapcert.pem";
   check_for_file "$GL/etc/grid-security/ldap/ldapkey.pem";

   print "\n";
}

sub mds_checks()
{
   banner "MDS checks\n";
   pre_ws_mds_checks;
}

env_setup;
prelim_checks;
compile_checks if ($all or $testhash{'compile'});

exit if ( @tests eq ("compile") );

gl_check;

gsi_checks if ($all or $testhash{'gsi'});
gram_checks if ($all or $testhash{'gram'});
gridftp_checks if ($all or $testhash{'gridftp'});
mds_checks if ($all or $testhash{'mds'});
