/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.CASPortType;
import org.globus.cas.CommunityAuthorizationServiceLocator;

import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;

import org.globus.axis.gsi.GSIConstants;

import org.globus.cas.impl.CasConstants;

import org.oasis.wsrf.faults.BaseFaultType;

import java.net.URL;
import javax.xml.rpc.Stub;

import org.globus.axis.util.Util;

public class CasClientSetup {

    public static final String casServerURL = "CAS_SERVER_URL";
    public static final String casServerIdentity = "CAS_SERVER_IDENTITY";

    public int argsStartIndex = 0;
    public boolean debug = false;
    String[] args = null;
    String errorMessage = null;
    String instanceURL = null;
    String serverIdentity = null;
    String securityType = null;
    Object protectionType = Constants.SIGNATURE;

    public CasClientSetup() {
    }

    public CasClientSetup(String[] args_, String msg) {
        args = args_;
        errorMessage = msg;
        for (int i=0; i<args.length; i++) {
            if (args[i].equals("-debug")) {
                debug = true;
                argsStartIndex = argsStartIndex + 1;
            }

            if (args[i].equals("-v")) {
                System.out.println("CAS Version number: " 
                                   + CasConstants.versionNumber);
                System.exit(0);
            }

            if (args[i].equals("-help")) {
                System.out.println(errorMessage);
                System.exit(0);
            }

            if (args[i].equals("-c")) {
                if (args.length == i+1) {
                    System.err.println("-c option should have instance URL");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                instanceURL = args[i+1];
                argsStartIndex = argsStartIndex + 2;
            }

            if (args[i].equals("-s")) {
                if (args.length == i+1) {
                    System.err
                        .println("-s option should have server identity");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                serverIdentity = args[i+1];
                argsStartIndex = argsStartIndex + 2;
            }

            if (args[i].equals("-m")) {
                if (args.length == i+1) {
                    System.err.println("-m option should have 'msg' or "
                                       + "'conv' or 'trans'");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                String val = args[i+1];
                if (val.equals("msg")) {
                    securityType = Constants.GSI_SEC_MSG;
                } else if (val.equals("conv")) {
                    securityType = Constants.GSI_SEC_CONV;
                } else if (val.equals("trans")) {
                    securityType = Constants.GSI_TRANSPORT;
                } else {
                    System.err.println("-m option should have 'msg' or "
                                       + "'conv' or 'trans'");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                argsStartIndex = argsStartIndex + 2;
            }

            if (args[i].equals("-p")) {
                if (args.length == i+1) {
                    System.err.println("-p option should have 'sig' or "
                                       + "'enc'");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                String val = args[i+1];
                if (val.equals("sig")) {
                    protectionType = Constants.SIGNATURE;
                } else if (val.equals("enc")) {
                    protectionType = Constants.ENCRYPTION;
                } else {
                    System.err.println("-p option should have 'sig' or "
                                       + "'enc'");
                    System.err.println(errorMessage);
                    System.exit(-1);
                }
                argsStartIndex = argsStartIndex + 2;
            }
        }
    }

    public String getInstanceURL() {

        printMessage("getting instance URL");
        
        // Option not set, look for property CAS_SERVER_URL
        if (instanceURL == null) {
            printMessage("Instance URL not set using -c");
            instanceURL = System.getProperty(casServerURL);
            if ((instanceURL != null) && (instanceURL.trim().equals(""))) {
                instanceURL = null;
            }
        }
        
        if (instanceURL == null) {
            System.err.println("Error: CAS server URL must be set using -c "
                               + " or environment variable CAS_SERVER_URL "
                               + " should be set.");
            System.err.println(errorMessage);
            System.exit(-1);
        }

        printMessage("Instance URL is " + instanceURL);
        return instanceURL;
    }
    
    public String getServerIdentity() {
        
        printMessage("getting server identity");
        
        // Option not set, look for property CAS_SERVER_IDENTITY
        if (serverIdentity == null) {
            printMessage("Server identity not set using -s");
            serverIdentity = System.getProperty(casServerIdentity);
            if ((serverIdentity != null) && 
                (serverIdentity.trim().equals(""))) {
                serverIdentity = null;
            }
        }
        
        if (debug) {
            if (serverIdentity == null) {
                System.err
                    .println("Server identity is null, hence host authz");
            }
        }
        printMessage("Server identity is " + serverIdentity);
        return serverIdentity;
    }
    

    public void checkParameters(int numberOfArgs) {
        
        printMessage("Number optional args " + argsStartIndex);
        numberOfArgs = numberOfArgs + argsStartIndex;
        printMessage("checkParameters " + numberOfArgs);
        printMessage("args.length " + args.length);
        if (args.length < numberOfArgs) {
            System.err.println("Not enough parameters");
            System.err.println(errorMessage);
            System.exit(-1);
        }
    }
    
    // identity is null, host authz is done.
    public CASPortType 
        getCASPort(String instanceURL) {
        return getCASPort(instanceURL, null);
    }

    // if identity is null, then do host auth
    public CASPortType getCASPort(String instanceURL, String identity) {

        if (securityType == null) {
            if (instanceURL.startsWith("https")) {
                securityType = Constants.GSI_TRANSPORT;
            } else {
                securityType = Constants.GSI_SEC_MSG;
            }
        }

        boolean trans = false;     
        if (securityType.equals(Constants.GSI_TRANSPORT)) { 
            // register util.
            Util.registerTransport();
            trans = true;
        }
        String authzObject = getAuthzParamObject(trans);
        Object authzValue = getAuthzParamValue(trans, identity);

        return getCASPort(instanceURL, securityType, protectionType,
                          authzObject, authzValue);
    }

    public String getAuthzParamObject(boolean trans) {
        String authzObject = Constants.AUTHORIZATION;
        if (trans) {
            // set authz
            authzObject = GSIConstants.GSI_AUTHORIZATION;
        }
        return authzObject;
    }

    public Object getAuthzParamValue(boolean trans, String identity) {
        Object authzValue = null;
        if (identity != null) {
            printMessage("Using identity authorization");
            if (trans) {
                authzValue = new org.globus.gsi.gssapi.auth
                    .IdentityAuthorization(identity);
            } else {
                authzValue = new IdentityAuthorization(identity);
            }
        } else {
            printMessage("Using Host authorization");
            if (trans) {
                authzValue = new org.globus.gsi.gssapi.auth
                    .HostAuthorization("cas");
            } else {
                authzValue = new HostAuthorization("cas");
            }
        }
        return authzValue;
    }

    public CASPortType getCASPort(String instanceURL, 
                                  String secType, Object protType,
                                  String authzObj, Object authzVal) {

        printMessage("Getting reference to cas port");

        if ((instanceURL.startsWith("https")) &&
            (!Constants.GSI_TRANSPORT.equals(secType))) {
            System.err.println("Instance URL with https is used, secure "
                               + "transport is required");
            System.exit(-1);
        }

        CASPortType casPort = null;
        try {
            CommunityAuthorizationServiceLocator loc = 
                new CommunityAuthorizationServiceLocator();
            casPort = loc.getCASPortTypePort(new URL(instanceURL));
        } catch (Exception exp) {
            String errorMesg = 
                "Could not get reference to service instance for"
                + " instance URL " + instanceURL;
            if (debug) {
                System.err.println(errorMesg);
                exp.printStackTrace(System.err);
            } else {
                System.out.println(errorMesg);
                printMessage(exp.toString());
            }
            System.exit(-1);
        }
        ((Stub)casPort)._setProperty(secType, protType);
        ((Stub)casPort)._setProperty(authzObj, authzVal);
        return casPort;
    }
    
    // Identity is null, implies host auth
    public CASPortType getCASPort(String instanceURL, int numberOfArgs, 
                                  String identity) {
        checkParameters(numberOfArgs);
        return getCASPort(instanceURL, identity);
    }
    
    private void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }

    public static String getErrorMsg(Exception e) {
        if (e instanceof BaseFaultType) {
            BaseFaultType fault = (BaseFaultType)e;
            return fault.getClass().getName() + ": " 
                + fault.getDescription(0);
        } else {
            return e.getMessage();
        }
    }
}
