/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import java.util.Vector;

import java.io.FileReader;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileNotFoundException;

import org.globus.wsrf.impl.security.authentication.Constants;

import org.globus.cas.impl.CasConstants;

/**
 * Command line client to create a proxy with CAS Assertions embedded
 */
public class CasProxyInit {

    static String mesg = 
        " Usage: cas-proxy-init [<options>] [ -t tag | -p proxyFile]\n"
        + " Options are  -debug Runs with debug trace \n"
        + "              -c <instanceURL> Specify URL for CAS server\n"
        + "              -help Prints this message\n"   
        + "              -v Prints version number\n"
        + "              -l Requested CAS credential lifetime in hours\n"
        + "                 By defualt, lifetime of 24 hours is used\n"
        + "              -f Requested policies that need to be included in\n"
        + "                 CAS credential. By defualt, all applicable \n"
        + "                 policies are included.\n"
        + "              -s Expected server identity. If not specified, \n"
        + "                 a identity with <fqdn>/cas is expected, where\n"
        + "                 <fqdn> is DN of the hostname of CAS server\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation If unspecified\n"
        + "                 If instance url is 'https', then trasport \n"
        + "                 security is used else mechanism defaults to\n"
        + "                 secure message\n"
        + "              -n protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " tag : Generated credential is placed in file with default proxy\n"
        + "       filename with the tag appended to it\n"
        + " proxyFile : File to place generated credential\n"
        + " If neither a tag not a proxyFile is specified, tag with value \n"
        + " \"cas\" is assumed. If both are specifies, -p takes precedence.";

    static String newProxyFileName = null;
    static int assertionLifetime = 24 * 60 * 60;
    static String policyFileName = null;
    static String instanceURL = null;
    static String serverIdentity = null;
    static boolean debug = false;
    static String casTag = CasConstants.CAS_PROXY_TAG;

    static String securityType = null;
    static Object protectionType = null;

    static String resourceSpec = "Resource:";

    public static void main(String args[]) throws Exception {
        
        // parse aguments
        parseArgs(args);
        printMessage("CAS client to generate restrictd proxy with CAS"
                           + " assertions.");

        // Option not set, look for property CAS_SERVER_URL
        if (instanceURL == null) {
            printMessage("Instance URL not set using -c");
            instanceURL = System.getProperty(CasClientSetup.casServerURL);
        }
        
        if ((instanceURL == null) || (instanceURL.trim().equals(""))) {
            System.err.println("Error: CAS server URL must be set using -c "
                               + " or environment variable CAS_SERVER_URL "
                               + " should be set.");
            System.err.println(mesg);
            System.exit(-1);
        }

        // If debug mode, print arguments
        printArgs();

        CasProxyHelper casProxyHelper = 
            new CasProxyHelper(instanceURL, serverIdentity, debug);

        // Always picks up default credential
        ClientParams clientParams = new ClientParams();
        clientParams.setSecurityType(securityType);
        clientParams.setProtectionType(protectionType);

        // Read from a file for policies requested as assertion.
        if (policyFileName != null) {
            // Vector of ResourceActionsMap object
            Vector resActionsVector = null;
            printMessage("filename " + policyFileName);
            BufferedReader reader = null;
            try {
                reader = new BufferedReader(new FileReader(policyFileName));
            } catch (FileNotFoundException fnfe) {
                System.err.println("Error reading policyFileName " 
                                   + policyFileName + "\n" 
                                   + fnfe.getMessage());
                if (debug) {
                    System.err.println(fnfe.toString());
                }
                System.exit(-1);
            }
            String resLine = null;
            String resource = null;
            int index = -1;
            try {

                if (((resLine = reader.readLine()) != null) 
                    && ((index = resLine.indexOf(resourceSpec)) != -1)) {
                    resource = resLine.substring(index 
                                                 + resourceSpec.length());
                    Vector actions = null;
                    index = -1;
                    while ((resLine = reader.readLine()) != null) {
                        printMessage("line: " + resLine);
                        if ((index = resLine.indexOf(resourceSpec)) == -1) {
                            printMessage("Action");
                            if (actions == null) {
                                actions = new Vector();
                            }
                            printMessage("adding action " + resLine);
                            actions.add(resLine);
                        } else {
                            printMessage("Resource " + resource);
                            ResourceActionsMap resActions = 
                                new ResourceActionsMap(resource, actions);
                            if (resActionsVector == null) {
                                resActionsVector = new Vector();
                            }
                            printMessage("adding for resource " + resource);
                            resActionsVector.add(resActions);
                            actions = null;
                            resource = resLine.substring(index + 
                                                     resourceSpec.length());
                            printMessage("Next Resource " + resource);
                        }
                    }

                    ResourceActionsMap resActions = 
                        new ResourceActionsMap(resource, actions);
                    if (resActionsVector == null) {
                        resActionsVector = new Vector();
                    }
                    printMessage("adding for resource " + resource);
                    resActionsVector.add(resActions);
                    
                    // Convert to array
                    ResourceActionsMap[] resActionsArray = 
                        new ResourceActionsMap[resActionsVector.size()];
                    resActionsVector.toArray(resActionsArray);
                    clientParams.setResourceActionsMap(resActionsArray);
                }
                else {
                    printMessage("Policy file has no data. "
                                   + " getMaximalAssertion");
                }
            } catch (IOException ioe) {
                System.err.println("Error reading from policy file " 
                                   + policyFileName + "\n" + ioe.getMessage());
                if (debug) {
                    System.err.println(ioe.toString());
                }
                System.exit(-1);
            } 
        }

        // assertion lifetime
        clientParams.setAssertionLifetime(assertionLifetime);

        // tag
        clientParams.setCasProxyTag(casTag);

        // Name of proxy file with CAS assertion embedded.
        clientParams.setCasProxyFileName(newProxyFileName);

        printMessage("Client params: " + clientParams.toString());

        String casProxyFileName = casProxyHelper.getCasProxy(clientParams);
        System.out.println("Completed successfully. Proxy written to " 
                           + casProxyFileName);
    }
    
    private static void parseArgs(String[] args) {

        for (int i=0; i<args.length; i++) {

            if (args[i].equals("-debug")) {
                debug = true;
            }

            if (args[i].equals("-help")) {
                System.out.println(mesg);
                System.exit(0);
            }

            if (args[i].equals("-v")) {
                System.out.println("CAS Version number: " 
                                   + CasConstants.versionNumber);
                System.exit(0);
            }

            if (args[i].equals("-c")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -c requires an argument.");
                    System.out.println(mesg);
                    System.exit(-1);
                }
                instanceURL = args[i+1];
            }

            if (args[i].equals("-l")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -l requires an argument.");
                    System.out.println(mesg);
                    System.exit(-1);
                }
                try {
                    assertionLifetime = Integer.parseInt(args[i+1]) * 60 * 60;
                } catch (NumberFormatException exp) {
                    System.err.println("-l should have an integer argument.\n "
                                       + exp.getMessage());
                    if (debug) {
                        System.err.println(exp.toString());
                    }
                    System.exit(-1);
                }
            }

            if (args[i].equals("-f")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -f requires an argument.");
                    System.exit(-1);
                }
                policyFileName = args[i+1];
            }

            if (args[i].equals("-t")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -t requires an argument.");
                    System.exit(-1);
                }
                casTag = args[i+1];
            }

            if (args[i].equals("-p")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -p requires an argument.");
                    System.exit(-1);
                }
                newProxyFileName = args[i+1];
            }

            if (args[i].equals("-s")) {
                if (i+1 == args.length) {
                    System.err.println("Error: -s requires an argument.");
                    System.exit(-1);
                }
                serverIdentity = args[i+1];
            }

            if (args[i].equals("-m")) {
                if (args.length == i+1) {
                    System.err.println("-m option should have 'msg' or "
                                       + "'conv'");
                    System.exit(-1);
                }
                String val = args[i+1];
                if (val.equals("msg")) {
                    securityType = Constants.GSI_SEC_MSG;
                } else if (val.equals("conv")) {
                    securityType = Constants.GSI_SEC_CONV;
                } else {
                    System.err.println("-m option should have 'msg' or "
                                       + "'conv'");
                    System.exit(-1);
                }
            }

            if (args[i].equals("-n")) {
                if (args.length == i+1) {
                    System.err.println("-n option should have 'sig' or "
                                       + "'enc'");
                    System.exit(-1);
                }
                String val = args[i+1];
                if (val.equals("sig")) {
                    protectionType = Constants.SIGNATURE;
                } else if (val.equals("enc")) {
                    protectionType = Constants.ENCRYPTION;
                } else {
                    System.err.println("-n option should have 'sig' or "
                                       + "'enc'");
                    System.exit(-1);
                }
            }
        }
    }

    private static void printArgs() {
        if (!debug)
            return;
        System.out.println("instanceURL " + instanceURL);
        System.out.println("lifetime " + assertionLifetime);
        System.out.println("tag " + casTag);
        System.out.println("new proxyFilename " + newProxyFileName);
        System.out.println("policyFileName " + policyFileName);
        System.out.println("Server identity " + serverIdentity);
    }

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
