/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.CASPortType;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.utils.CasStringUtils;

import org.globus.cas.types.TrustAnchorData;
import org.globus.cas.types.NamespaceData;
import org.globus.cas.types.ServiceTypeData;
import org.globus.cas.types.UserData;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.CasObjectData;
import org.globus.cas.types.CasObjectDesc;

import org.globus.wsrf.utils.FaultHelper;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

/**
 * Command client to get a trust anchor, user, object, namespace, service type 
 * object given identifier
 */
public class GetCasObject {
    
    static String msg = 
        " Usage: cas-get-object [<options>] <objectType> <objectIdentifier>\n"
        + " Options are  -debug Runs with debug trace \n"
        + "              -c <instanceURl> Specify URL for CAS server\n"
        + "              -help Prints this message\n"
        + "              -v Prints version number\n"
        + "              -s Identity of CAS server. If not specified, will \n"
        + "                 default to host authorization.\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation or 'trans' for\n"
        + "                 transport security. If unspecified, if instance\n"
        + "                 url is 'https', then trasport security is used \n"
        + "                 else mechanism defaults to secure message\n"
        + "              -p protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " <objectType> : The type of object that needs to be retrived.\n"
        + "                Should be \"trustAnchor\" or \"user\" or \"object\""
        + "\n              \"namespace\" or \"serviceType\" or \"policy\" \n"
        + " <objectIdentifier> : String that uniquely identifies the object \n"
        + "                      that needs to be retrieved. For the above \n"
        + "                      type, should be trust anchor nickname\n"
        + "                      (or) user nickname (or) \n"
        + "                      <objectNamespace> <objectName> where\n"
        + "                      <objectNamespace> is namespace the object \n"
        + "                      belongs to and <objectName> is name of object"
        + "\n                      (or) namespace nickname (or) service type\n"
        + "                      name (or) policy id respectively";

    static boolean debug = false;

    public static void main(String args[]) throws Exception {

        if (args.length<1) {    
            System.out.println(msg);
            System.exit(0);
        }

        CasClientSetup clientSetup = new CasClientSetup(args, msg);
        String instanceURL =  clientSetup.getInstanceURL();     
        String serverIdentity = clientSetup.getServerIdentity();
        int startIndex = clientSetup.argsStartIndex;
        debug = clientSetup.debug;

        printMessage("CAS client to get details about a CAS entity.");

        if (args.length < startIndex + 1) {
            System.err.println("Not enough parameters");
            System.err.println(msg);
            System.exit(-1);
        }

        String type = args[startIndex].trim();
        if ( !(type.equals(CasConstants.TRUSTANCHOR_SPEC) || 
               type.equals(CasConstants.USER_SPEC) || 
               type.equals(CasConstants.OBJECT_SPEC) || 
               type.equals(CasConstants.NAMESPACE_SPEC) ||
               type.equals(CasConstants.POLICY_SPEC) ||
               type.equals(CasConstants.SERVICETYPE_SPEC))) {

            System.err.println("Erroneous type " + type);
            System.err.println(msg);
            System.exit(-1);
        }

        String identifier = null;
        CASPortType casPort = null;
        if (type.equals(CasConstants.OBJECT_SPEC)) {
            casPort = clientSetup.getCASPort(instanceURL, 3, serverIdentity);
            identifier = args[startIndex+1] + CasConstants.OBJECTSPEC_DELIMITER
                + args[startIndex+2];
        } else {
            casPort = clientSetup.getCASPort(instanceURL, 2, serverIdentity);
            identifier = args[startIndex+1];
        }

        printMessage("getCasObject called with " + type + " " + identifier);
        CasObjectData casObject = null;
        try {
            CasObjectDesc tmp = new CasObjectDesc();
            tmp.setNameOfCasObject(identifier);
            tmp.setTypeOfCasObject(type);
            casObject = casPort.getCasObject(tmp);
        } catch (NoPermissionFault noPermFault) {
            System.err.println(noPermFault.getDescription());
            System.exit(-1);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc = 
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            System.exit(-1);
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            System.exit(-1);
        }

        if (casObject == null) {
            System.out.println("Query for CAS Object of type " + type + " \"" 
                               + args[startIndex+1] + " \" returned null. "
                               + "Such an object does not exist");
            System.exit(0);
        } else if (type.equals(CasConstants.TRUSTANCHOR_SPEC)) {
            System.out.println(CasStringUtils.trustAnchorToString(
                              (TrustAnchorData)casObject));
        } else if (type.equals(CasConstants.USER_SPEC)) {
            System.out.println(CasStringUtils.userToString(
                              (UserData)casObject));
        } else if (type.equals(CasConstants.NAMESPACE_SPEC)) {
            System.out.println(CasStringUtils.namespaceToString(
                              (NamespaceData)casObject));
        } else if (type.equals(CasConstants.OBJECT_SPEC)) {
            System.out.println(CasStringUtils.objectToString(
                              (ObjectData)casObject));
        } else if (type.equals(CasConstants.SERVICETYPE_SPEC)) {
            System.out.println(CasStringUtils.serviceTypeToString(
                              (ServiceTypeData)casObject));
        } else if (type.equals(CasConstants.POLICY_SPEC)) {
            System.out.println(CasStringUtils.policyToString(
                              (PolicyData)casObject));
        } else {
            // shld not happen!
            System.err.println("Erroneous type " + type);
            System.err.println(msg);
            System.exit(-1);
        }
        System.out.println("Completed successfully");
    }

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
