/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.CASPortType;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.types.DeleteObject;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.wsrf.utils.FaultHelper;
/**
 * Command client used to unenroll trust anchor, user, namespace, service type 
 */
public class Unenroll {
    
    static String msg = 
        " Usage : cas-remove [<options>] <objectType> <objectDetails>\n"
        + " Options are  -debug Runs with debug trace \n"
        + "              -c <instanceURl> Specify URL for CAS server\n"
        + "              -help Prints this message\n"
        + "              -v Prints version number\n"
        + "              -s Identity of CAS server. If not specified, will \n"
        + "                 defualt to host authorization.\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation or 'trans' for\n"
        + "                 transport security. If unspecified, if instance\n"
        + "                 url is 'https', then trasport security is used \n"
        + "                 else mechanism defaults to secure message\n"
        + "              -p protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " <objectType> : Type of object to unenroll. Should be \n"
        + "                \"trustAnchor\" or \"user\" or \"namespace\"\n"
        + "                \"object\" or \"serviceType\"\n"
        + " <objectDetails> :\n"
        + "    If \"trustAnchor\", <nickname> \n"
        + "       <nickname> : Nickname of trust anchor \n"
        + "    If \"user\", <nickname>\n"
        + "       <nickname> : Nickname of user \n"
        + "    If \"namespace\", <nickname>\n"
        + "       <nickname> : Nickname of namespace \n"
        + "    If \"object\", <objectName> <objectNamespce>\n"
        + "       <objectName> : Name of object\n"
        + "       <objectNamespace> : Namespace the object belongs to\n"
        + "    If \"serviceType\", <serviceType>\n"
        + "       <serviceType> : Name of the service type \n";

    static boolean debug = false;

    public static void main(String args[]) throws Exception {

        if (args.length<1) {    
            System.out.println(msg);
            System.exit(0);
        }

        CasClientSetup clientSetup = new CasClientSetup(args, msg);
        String instanceURL =  clientSetup.getInstanceURL();     
        String serverIdentity = clientSetup.getServerIdentity();
        int startIndex = clientSetup.argsStartIndex;
        debug = clientSetup.debug;

        printMessage("CAS client to unenroll CAS entities.");

        if (args.length < startIndex + 1) {
            System.err.println("Not enough parameters");
            System.err.println(msg);
            System.exit(-1);
        }

        printMessage("Argument starting index is " + startIndex);
        String objectType = args[startIndex];    
        
        try {
            // Trust anchor or user
            if (objectType.equals(CasConstants.TRUSTANCHOR_SPEC)) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 2, serverIdentity);
                printMessage(" Remove trust anchor " + args[startIndex+1]);
                casPort.removeTrustAnchor(args[startIndex+1]);
            } else if (objectType.equals(CasConstants.USER_SPEC)) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 2, serverIdentity);
                printMessage(" Remove user " + args[startIndex+1]);
                casPort.removeUser(args[startIndex+1]);
            } else if (objectType.equals(CasConstants.NAMESPACE_SPEC)) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 2, serverIdentity);
                printMessage("Unenroll Namespace " + args[startIndex+1]);
                casPort.deleteObjectNamespace(args[startIndex+1]);
            } else if (objectType.equals(CasConstants.OBJECT_SPEC)) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 3, serverIdentity);
                printMessage("Object " + args[startIndex+1] + " " 
                             + args[startIndex+2]);
                DeleteObject deleteObject = new DeleteObject();
                deleteObject.setObjectName(args[startIndex+1]);
                deleteObject.setObjectNamespace(args[startIndex+2]);
                casPort.deleteObject(deleteObject);
            } else if (objectType.equals(CasConstants.SERVICETYPE_SPEC)) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 2, serverIdentity);
                printMessage("ServiceType " + args[startIndex+1]);
                casPort.deleteServiceType(args[startIndex+1]);
            } else {
                System.err.println("Erroneous object type " + objectType);
                System.err.println(msg);
                System.exit(-1);
            }
        } catch (NoPermissionFault noPermFault) {
            System.err.println(noPermFault.getDescription());
            System.exit(-1);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc = 
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            System.exit(-1);
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            System.exit(-1);
        }
        System.out.println("Completed successfully");
    }

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
