/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.bootstrap;

import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;

import java.util.Properties;

import org.globus.cas.impl.databaseAccess.CasDBOptions;
import org.globus.cas.impl.databaseAccess.CasDBStorage;
import org.globus.cas.impl.databaseAccess.CasDBException;

public class SetupDatabase  {

    static final String bootstrapProp = "bootstrapFilename";    

    static final String mesg = 
        " Usage: cas-server-bootstrap [<options>] -d <dbPropFile> \n"
        + "      [ -implicit | -b <bootstrapFile> ]"
        + "\n Options are -help Prints this help message\n"
        + "               -debug Runs with debug trace\n"
        + " -d <dbPropFile> : File name with database properties as follows\n"
        + "                \"dbDriver=\"<database driver name \n"
        + "                \"dbConnectionURL=\"<database connection URL\n"
        + "                \"dbUserName=\"<Username to access database\n"
        + "                \"dbPassword=\"<Password for the above username\n"
        + " -b <bootstrapFile> : Populates data with user data. File with \n"
        + "                      data to bootstrap database with.\n"
        + " -implicit : Populates database with CAS server implicit data and\n"
        + "             service/action and namespace relevant to FTP";

    public static void main(String[] args) throws Exception {
        
        boolean debug = false;
        boolean implicit = false;
        String dbPropFilename = null;
        String bootstrapFilename = null;

        for (int i=0; i<args.length; i++) {
            
            if (args[i].equals("-debug")) {
                debug = true;
            }

            if (args[i].equals("-help")) {
                System.out.println(mesg);
                System.exit(0);
            }

            if (args[i].equals("-d")) {
                if (args.length == i+1) {
                    System.err.println("Error: -d option should have database"
                                       + " properties file name");
                    System.err.println(mesg);
                    System.exit(-1);
                }
                dbPropFilename = args[i+1];
            }
            
            if (args[i].equals("-implicit")) {
                implicit = true;
            }

            if (args[i].equals("-b")) {
                if (args.length == i+1) {
                    System.err.println("Error: -b option should have bootstrap"
                                       + "file name");
                    System.err.println(mesg);
                    System.exit(-1);
                }
                bootstrapFilename = args[i+1];
            }
        }
        
        if ((!implicit) && (bootstrapFilename == null)) {
            System.err.println("Error: Neither implicit bootstrap requested "
                               + "nor user data provided");
            System.err.println(mesg);
            System.exit(-1);
        }

        if (dbPropFilename == null) {
            System.err.println("Error: -d option with database properties file"
                               + " needs to be specified");
            System.err.println(mesg);
            System.exit(-1);
        }

        Properties prop = new Properties();
        try {
            prop.load(new FileInputStream(dbPropFilename));
        } catch (FileNotFoundException fnfe) {
            System.err.println("Error: Database properties file " 
                               + dbPropFilename + " not found");
            if (debug) {
                System.err.println(fnfe);
            }
            System.exit(-1);
        } catch (IOException ioe) {
            System.err.println("Error loading file " + dbPropFilename + "\n"
                               + ioe.getMessage());
            if (debug) {
                System.err.println(ioe);
            }
            System.exit(-1);
        }
        CasDBOptions casDbOption = 
            new CasDBOptions(prop.getProperty("dbDriver"), 
                             prop.getProperty("dbConnectionURL"),
                             prop.getProperty("dbUsername"), 
                             prop.getProperty("dbPassword"));
        String val = prop.getProperty("activeConnections");
        if (val != null) {
            if (debug) {
                System.out.println("Active connections " + val);
            }
            casDbOption.setActiveConnections((new Integer(val)).intValue());
        }

        val = prop.getProperty("onExhaustAction");
        if (val != null) {
            if (debug) {
                System.out.println("On exhaust action " + val);
            }
            casDbOption.setOnExhaustAction((new Integer(val)).byteValue());
        }

        val = prop.getProperty("maxWait");
        if (val != null) {
            if (debug) {
                System.out.println("max value " + val);
            }
            casDbOption.setMaxWait((new Integer(val)).longValue());
        }

        val = prop.getProperty("idleConnections");
        if (val != null) {
            if (debug) {
                System.out.println("idle connections " + val);
            }
            casDbOption.setIdleConnections((new Integer(val)).intValue());
        }

        try {
            CasDBStorage.setupDBConnection(casDbOption);
        } catch (CasDBException exp) {
            System.err.println("Erorr initializing database " 
                               + exp.getMessage());
            if (debug) {
                System.err.println(exp);
            }
            System.exit(-1);
        }
        if (debug) {
            System.out.println("Database initialized");
        }

        if (implicit) {
            try {
                PopulateImplicitObjectData.populateDb();
            } catch (Exception exp) {
                System.err.println("Error populating db with CAS server "
                                   + "implicit data \n" + exp.getMessage());
                if (debug) {
                    System.err.println(exp);
                }
                System.exit(-1);
            }
            System.out.println("Database has been populated with CAS server"
                               + " implicit data");
            try {
                PopulateImplicitFTPData.populateDb();
            } catch (Exception exp) {
                System.err.println("Error populating db with FTP "
                                   + "implicit data \n" + exp.getMessage());
                if (debug) {
                    System.err.println(exp);
                }
                System.exit(-1);
            }
            System.out.println("Database has been populated with FTP "
                               + " implicit data");
        }

        if (bootstrapFilename != null) {
            if (debug) {
                System.out.println("Populating database with data from "
                                   + " following bootstrap file " 
                                   + bootstrapFilename);
            }
            try {
                PopulateUserData.populateDb(bootstrapFilename);
            } catch (Exception exp) {
                System.err.println("Error populating CAS server database with "
                                   + "user data from " + bootstrapFilename
                                   + "\n" + exp.getMessage());
                if (debug) {
                    System.err.println(exp);
                }
                System.exit(-1);
            }
            System.out.println("Database has been populated with user data");
        } else {
            if (debug) {
                System.out.println("No bootstrap file name specified");
            }
        }
        System.out.println("Bootstrap completed.");
    }
}
