/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.Vector;

// Classes for database pooling
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.PoolingDriver;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.DriverManagerConnectionFactory;

import org.globus.util.I18n;

/**
 * Class for manipulating connections through database pooling
 */
public class CasDBStorage {

    static Log logger = LogFactory.getLog (CasDBStorage.class.getName() );

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     CasDBStorage.class.getClassLoader());

    private static boolean driverSetup = false;
    private static String driverName;
    
    final public static String baseURI = "jdbc:apache:commons:dbcp:";
    final public static String managedJobURI = "cas";

    // Track number of connections
    static int connNumber;
    // conenction object pool
    static GenericObjectPool connectionPool = null;
    
    /*
      Constructor
    */
    private CasDBStorage() {
    }

    /*
     * Opens a new connection and sets up the pool
     */    
    public synchronized static 
        void setupDBConnection(CasDBOptions casDbOptions_)  
        throws CasDBException {

        logger.debug(" CasDBStorage setup connection called");
        if (!driverSetup) {
            logger.debug("dbOptions: " + casDbOptions_.toString());
            driverName = casDbOptions_.getDriver();
            try {
                setupDriver(casDbOptions_.getConnectionURL (), 
                            casDbOptions_.getUserName (),
                            casDbOptions_.getPassword (), 
                            casDbOptions_.getActiveConnections(), 
                            casDbOptions_.getOnExhaustAction(), 
                            casDbOptions_.getMaxWait(), 
                            casDbOptions_.getIdleConnections());
            }
            catch(Exception e) {
                logger.error(i18n.getMessage("setupErr"), e);
                throw new CasDBException(i18n.getMessage("setupErr")
                                         + e.getMessage(), e);
            }
            driverSetup = true;
        }
        connNumber = 0;
    }

    public static Connection getDBConnection() throws CasDBException {
        if (!driverSetup) {
            throw new CasDBException(i18n.getMessage("driverNotInit"));
        }

        try {
            Class.forName(driverName);
            Connection connection  = 
                DriverManager.getConnection(baseURI + managedJobURI);
            connNumber++;
            if (logger.isDebugEnabled()) {
                logger.debug("Added " + connNumber + " active " 
                            + connectionPool.getNumActive() 
                            + " idle " + connectionPool.getNumIdle());
            }
            return connection;
        }
        catch(SQLException e) {
            logger.error(i18n.getMessage("dbConnErr"));
            throw new CasDBException(i18n.getMessage("dbConnErr") 
                                     + e.getMessage(), e);
        }
        catch(ClassNotFoundException e) {
            logger.error(i18n.getMessage("dbDriverClass"), e);
            throw new CasDBException(i18n.getMessage("dbDriverClass") 
                                     + e.getMessage(), e);
        }
    }


    private static void setupDriver(String connectURI, String username, 
                                    String password, int activeConnections, 
                                    byte onExhaustAction, long maxWait, 
                                    int idleConnections) 
        throws Exception {

        // Object pool which is a pool of conection
        connectionPool = new GenericObjectPool(null, activeConnections, 
                                               onExhaustAction, maxWait, 
                                               idleConnections);
        // ConnectionFactory that pool uses to create connectiosn
        ConnectionFactory connectionFactory = 
            new DriverManagerConnectionFactory(connectURI,username,password);

        // PoolableConnectionFactory used for pooling functionality
        PoolableConnectionFactory poolableConnectionFactory = 
            new PoolableConnectionFactory(connectionFactory,
                                          (ObjectPool)connectionPool,
                                          null,null,false,true);

        // Create and Register PoolingDriver
        PoolingDriver driver = new PoolingDriver();
        driver.registerPool(managedJobURI,connectionPool);
    }


    /*
      Returns the connection to the database pool
    */    
    public static void returnDBConnection(Connection connection) 
        throws CasDBException {

        if (!driverSetup) {
            throw new CasDBException(i18n.getMessage("driverNotInit"));
        }
        
        if (connection == null) {
            return;
        }
            
        try {
            connection.close();
            connNumber--;
            if (logger.isDebugEnabled()) {
                logger.debug("Reduced " + connNumber + " active " 
                            + connectionPool.getNumActive() 
                            + " idle " + connectionPool.getNumIdle());
            }
        }
        catch (SQLException sqlExcep) {
            logger.error(i18n.getMessage("dbConnReturnErr"), sqlExcep);
            throw new CasDBException(i18n.getMessage("dbConnReturnErr")
                                     + sqlExcep.getMessage(), sqlExcep);
        }
    }

    /* Method to run update query  */
    public static int runUpdateQuery(String query) 
        throws CasDBException {
        logger.debug("run update query:  Query is " + query);
        Connection connection = getDBConnection();
        Statement statement = null;
        int result = -1;
        try {
            statement = connection.createStatement();
            result = statement.executeUpdate(query);
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("stmtExecErr", query), exp);
            throw new CasDBException(i18n.getMessage("stmtExecErr", query) 
                                     + exp.getMessage(), exp);
        }
        finally {
            // retuning connection irrespective of whether stmt
            // and result set are closed or not.
            returnDBConnection(connection);
            try {
                if (statement != null) 
                    statement.close();
            } catch (SQLException exp) {
                logger.warn(i18n.getMessage("updateErrClose"), exp);
            }
        }
        return result;
    }

    /* Method to run query for list methods */
    public static String[] runListQuery(String query, String colName) 
        throws CasDBException {
        logger.debug("run list query:  Query is " + query);
        Vector listResult = null ;
        Connection connection = getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                int i = 0;
                listResult = new Vector(resultSet.getFetchSize());
                do {
                    listResult.add(resultSet.getString(colName.trim()));
                    logger.debug(listResult.get(i));
                    i++;
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("stmtExecErr", query), exp);
            throw new CasDBException(i18n.getMessage("stmtExecErr", query)
                                     + exp.getMessage(), exp);
        }
        finally {
            returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("listErrClose"), exp);
            }
        }
        if (listResult != null) {
            String[] returnStrings = new String[listResult.size()];
            listResult.toArray(returnStrings);
            return returnStrings;
        } else {
            return null;
        }
    }
    
    /**
     * Returns an string array of the elements in the vector
     */
    public static String[] vectorToStringArray(Vector vec) {
        if (vec != null) {
            String[] retStrings = new String[vec.size()];
            vec.toArray(retStrings);
            return retStrings;
        } else {
            return null;
        }
    }


    /**
     * Given a table name and column name, it does a query for columnName=value 
     * Constructs query and returns true if the query has any entry in result set
     */
    public static boolean rowExists(String tableName, String columnName, 
                                      String value) 
        throws CasDBException {
        
        String query = "select" + columnName + " from" + tableName + " where" 
            + columnName ;
        if (tableName.equals(CasDBConstants.TABLE_OBJECT))
            query = query + " = " + value.trim();
        else
            query = query + "='" + value + "'";
        return rowExists(query);
    }

    /**
     * Returns true if the query has any entry in result set
     */ 
    public static boolean rowExists(String query) throws CasDBException {
        logger.debug("\n Query is \n" + query + "\n");
        boolean result = false;
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                result = true;
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("dataExistenceErr", query), exp);
            throw new CasDBException(i18n.getMessage("dataExistenceErr", query)
                                     + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("dataExistenceCloseErr"), exp);
            }
        }
        return result;
    }
}
