/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;

import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.cas.utils.CasStringUtils;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.types.UserData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.NamespaceData;
import org.globus.cas.types.CasObjectData;
import org.globus.cas.types.UserGroupData;
import org.globus.cas.types.TrustAnchorData;
import org.globus.cas.types.ServiceTypeData;
import org.globus.cas.types.ObjectGroupData;
import org.globus.cas.types.ArrayOfString;

import org.globus.util.I18n;

/**
 * Used to manipulate PolicyData in the database
 */
public class PolicyDataHandler {
    
    static Log logger = LogFactory.getLog(PolicyDataHandler.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     PolicyDataHandler.class.getClassLoader());

    /**
     * Store the policy data object - (grant of rights)
     * Manipulates only the policy_table in the database
     */
    public static String storeObject(PolicyData policyData) 
        throws CasDBException {

        String baseErrMesg = i18n.getMessage("policyStrErr");

        // Need to ensure that the user/object/action specifed exist 
        // in the database
        String userGroupName = policyData.getUserGroupName();
        if (!CasDBStorage.rowExists(CasDBConstants.TABLE_USERGP, 
                        CasDBConstants.COL_USERGP_NAME, userGroupName)) {
            String err = i18n.getMessage("doesNotExistTable", new Object[] {
                userGroupName, "user group" });
            logger.error(err);
            throw new CasDBException(baseErrMesg + err);
        }
        
        boolean actionValid = false;
        String actionSpec = policyData.getActionSpec();
        String actionSpecDesc = policyData.getActionSpecDesc();
        // action specification can be serviceAction or serviceActionGroup
        if (actionSpecDesc.equals(CasConstants.SERVICEACTION_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants
                                        .TABLE_SERVICETYPE_ACTION, 
                            CasDBConstants.COL_SERVICEACTION_ID, actionSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     actionSpec, 
                                                     "service action" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            actionValid = true;
        }
        
        if (actionSpecDesc.equals(CasConstants.SERVICEACTIONGP_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_SERVICEACTIONGP, 
                            CasDBConstants.COL_SERVICEACTIONGP_NAME, 
                            actionSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     actionSpec, 
                                                     "service action group" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            actionValid = true;
        }
        if (!actionValid) {
            logger.error(i18n.getMessage("invalidActionSpec"));
            throw new CasDBException(baseErrMesg 
                                     + i18n.getMessage("invalidActionSpec"));

        }

        boolean objValid = false;
        String objectSpec = policyData.getObjectSpec();
        String objectSpecDesc = policyData.getObjectSpecDesc();
        // object specification can be object, objectGroup, user, userGroup, 
        // serviceType, namespace, trustAnchor
        // object
        if (objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_OBJECT, 
                            CasDBConstants.COL_OBJECT_ID, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "object" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // objectGp
        if (objectSpecDesc.equals(CasConstants.OBJECTGP_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_OBJECTGP, 
                            CasDBConstants.COL_OBJECTGP_NAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "objectGp" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // user
        if (objectSpecDesc.equals(CasConstants.USER_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_USER, 
                            CasDBConstants.COL_USER_NICKNAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "user" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // userGp
        if (objectSpecDesc.equals(CasConstants.USERGP_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_USERGP, 
                            CasDBConstants.COL_USERGP_NAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "user group" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // serviceType
        if (objectSpecDesc.equals(CasConstants.SERVICETYPE_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_SERVICETYPE, 
                            CasDBConstants.COL_SERVICETYPE_NAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "service type" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // Namespace
        if (objectSpecDesc.equals(CasConstants.NAMESPACE_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_NAMESPACE, 
                            CasDBConstants.COL_NS_NICKNAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "namespace" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // trust anchor
        if (objectSpecDesc.equals(CasConstants.TRUSTANCHOR_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_TRUSTANCHOR, 
                            CasDBConstants.COL_TRUST_NICKNAME, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "trustanchor" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // ServiceActionGp
        if (objectSpecDesc.equals(CasConstants.SERVICEACTIONGP_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_SERVICEACTIONGP, 
                            CasDBConstants.COL_SERVICEACTIONGP_NAME, 
                            objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "service action group" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        // PolicyData
        if (objectSpecDesc.equals(CasConstants.POLICY_SPEC)) {
            if (!CasDBStorage.rowExists(CasDBConstants.TABLE_POLICY, 
                            CasDBConstants.COL_POLICY_ID, objectSpec)) {
                String errMesg = i18n.getMessage("doesNotExistTable", 
                                                 new Object[] {
                                                     objectSpec,
                                                     "policy" });
                logger.error(errMesg);
                throw new CasDBException(baseErrMesg + errMesg);
            }
            objValid = true;
        }
        
        if (!objValid) {
            logger.error(i18n.getMessage("invalidObjSpec"));
            throw new CasDBException(baseErrMesg 
                                     + i18n.getMessage("invalidObjSpec"));
        }

        String columnString = "insert into" + CasDBConstants.TABLE_POLICY 
            + " (" + CasDBConstants.COL_USERGP_NAME + "," 
            + CasDBConstants.COL_OBJECT_SPEC + "," 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "," 
            + CasDBConstants.COL_ACTION_SPEC + "," 
            + CasDBConstants.COL_ACTION_SPEC_DESC + ") values ('";

        StringBuffer query = new StringBuffer(columnString);
        query = query.append(policyData.getUserGroupName()).append("','")
            .append(policyData.getObjectSpec()).append("','")
            .append(policyData.getObjectSpecDesc()).append("','")
            .append(policyData.getActionSpec()).append("','")
            .append(policyData.getActionSpecDesc()).append("')");
        CasDBStorage.runUpdateQuery(query.toString());
    
        String resultString = getPolicyId(policyData);
        if (resultString == null) {
            logger.error(baseErrMesg);
            throw new CasDBException(baseErrMesg);
        }
        return resultString;
    }
    
    // given policy object, returns policy id
    public static String getPolicyId(PolicyData policyObj) 
        throws CasDBException {

        String policyId = null;
        String userGroupName = policyObj.getUserGroupName();
        String actionSpec = policyObj.getActionSpec();
        String actionSpecDesc = policyObj.getActionSpecDesc();
        String objectSpec = policyObj.getObjectSpec();
        String objectSpecDesc = policyObj.getObjectSpecDesc();
        StringBuffer query = 
            new StringBuffer("select" + CasDBConstants.COL_POLICY_ID + " from" 
                             + CasDBConstants.TABLE_POLICY + " where (" 
                             + CasDBConstants.COL_USERGP_NAME + "='");
        query = query.append(userGroupName).append("') and (")
            .append(CasDBConstants.COL_OBJECT_SPEC).append("='")
            .append(objectSpec).append("') and (")
            .append(CasDBConstants.COL_OBJECT_SPEC_DESC).append("='")
            .append(objectSpecDesc).append("') and (")
            .append(CasDBConstants.COL_ACTION_SPEC).append("='")
            .append(actionSpec).append("') and (")
            .append(CasDBConstants.COL_ACTION_SPEC_DESC).append("='")
            .append(actionSpecDesc).append("')");
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query.toString());
            if ((resultSet!=null) && (resultSet.next())) {
                policyId = resultSet.getString(
                                     CasDBConstants.COL_POLICY_ID.trim());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object [] {
                "policy id", query }));
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                                                     new Object [] { 
                                                         "policy id", ""})
                                     + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", "policy id"), 
                            exp);
            }
        }
        return policyId;
    }

    /**
     * List of all Policy Data
     */
    public static String[] list() throws CasDBException {
        
        String query = "select" + CasDBConstants.COL_POLICY_ID + " from" 
            + CasDBConstants.TABLE_POLICY;
        return CasDBStorage.runListQuery(query, CasDBConstants.COL_POLICY_ID);
    }

    /*
     * Retrieves PolicData data 
     * Picks up values from policy_table  
     */
    public static CasObjectData retrieveObject(String policyId) 
        throws CasDBException {
        
        PolicyData returnObject = null;
        String query = "select * from" + CasDBConstants.TABLE_POLICY 
            + " where" + CasDBConstants.COL_POLICY_ID + " = '" 
            + policyId.trim() + "'";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if (resultSet!=null) {
                if (resultSet.next()) {
                    returnObject = new PolicyData();
                    returnObject.setPolicyId(policyId);
                    returnObject.setUserGroupName(
                                 resultSet.getString(
                                 CasDBConstants.COL_USERGP_NAME.trim()));
                    returnObject.setObjectSpec(
                                 resultSet.getString(
                                 CasDBConstants.COL_OBJECT_SPEC.trim()));
                    returnObject.setObjectSpecDesc(
                                  resultSet.getString(
                                  CasDBConstants.COL_OBJECT_SPEC_DESC.trim()));
                    returnObject.setActionSpec(
                                 resultSet.getString(
                                 CasDBConstants.COL_ACTION_SPEC.trim()));
                    returnObject.setActionSpecDesc(
                                 resultSet.getString(
                                 CasDBConstants.COL_ACTION_SPEC_DESC.trim()));
                }
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object[] {
                "policy", query }), exp);
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                                                     new Object[] {
                                                         "policy", "" })
                                     + exp.getMessage() , exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", "policy"),
                            exp);
            }
        }
        return returnObject;
    }
    
    /**
     * Remove policy table entry from database - (revoke rights)
     * Manipulates only policy table
     */
    public static void deleteObject(String policyId) throws CasDBException {
        
        String query = "delete from" + CasDBConstants.TABLE_POLICY 
            + " where" + CasDBConstants.COL_POLICY_ID + " = '" 
            + policyId.trim() + "'";
        CasDBStorage.runUpdateQuery(query);
    }

    /**
     * Returns true if policy object exists in the table
     * This is *not* a permission query and does not have any 
     * semantic ramifications
     */
    public boolean policyObjectExists(PolicyData policyData) 
        throws CasDBException {
        
        String resultString = getPolicyId(policyData);
        if (resultString == null) 
            return false;
        else
            return true;
    }
    
    /**
     * Store the policy represented by the given parameters
     * Construct the policy object and store
     * Does not handle serviceActionGp as spec
     */
    public static PolicyData grantObject(String userGroupName, 
                                         CasObjectData objSpec, 
                                         String objSpecDesc, String actionSpec,
                                         String actionSpecDesc ) 
        throws CasDBException {
        String objSpecString = getObjectSpecString(objSpec, objSpecDesc);
        return grant(userGroupName, objSpecString, objSpecDesc, actionSpec, 
                     actionSpecDesc);
    }
    /**
     * Store the policy represented by the given parameters
     * Construct the policy object and store
     */
    public static PolicyData grant(String userGroupName, String objSpec, 
                                   String objSpecDesc, String actionSpec, 
                                   String actionSpecDesc ) 
        throws CasDBException {
        PolicyData returnPolicy = new PolicyData();
        returnPolicy.setUserGroupName(userGroupName.trim());
        returnPolicy.setObjectSpec(objSpec.trim());
        returnPolicy.setObjectSpecDesc(objSpecDesc.trim());
        returnPolicy.setActionSpec(actionSpec.trim());
        returnPolicy.setActionSpecDesc(actionSpecDesc.trim());
        String policyId =storeObject(returnPolicy);
        returnPolicy.setPolicyId(policyId);
        return returnPolicy;
    }

    public static String getObjectSpecString(CasObjectData objSpec, 
                                             String objectSpecDesc)
        throws CasDBException {

        objectSpecDesc = objectSpecDesc.trim();
        // object
        if (objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            logger.debug("it is object");
            if (((ObjectData)objSpec).getObjectId() == null) {
                logger.debug("object id not set");
                String tempObjName = ((ObjectData)objSpec).getObjectName();
                String tempObjNS = ((ObjectData)objSpec).getObjectNamespace();
                int tempObjectId = 
                    ObjectDataHandler.getObjectId(tempObjName, tempObjNS); 
                return Integer.toString(tempObjectId);
            }
            else
                return ((ObjectData)objSpec).getObjectId();
        }
        // objectGp
        if (objectSpecDesc.equals(CasConstants.OBJECTGP_SPEC)) {
            logger.debug("it is object group");
            return ((ObjectGroupData)objSpec).getGroupName();
        }
        // user
        if (objectSpecDesc.equals(CasConstants.USER_SPEC)) {
            logger.debug("it is user");
            return ((UserData)objSpec).getNickname();
        }
        // userGp
        if (objectSpecDesc.equals(CasConstants.USERGP_SPEC)) {
            logger.debug("it is userGp");
            return ((UserGroupData)objSpec).getGroupName();
        }
        // serviceType
        if (objectSpecDesc.equals(CasConstants.SERVICETYPE_SPEC)) {
            logger.debug("it is serviceType");
            return ((ServiceTypeData)objSpec).getName();
        }
        // Namespace
        if (objectSpecDesc.equals(CasConstants.NAMESPACE_SPEC)) {
            logger.debug("it is Namespace");
            return ((NamespaceData)objSpec).getNickname();
        }
        // trust anchor
        if (objectSpecDesc.equals(CasConstants.TRUSTANCHOR_SPEC)) {
            logger.debug("it is TrustAnchor");
            return ((TrustAnchorData)objSpec).getNickname();
        }
        // policy data
        if (objectSpecDesc.equals(CasConstants.POLICY_SPEC)) {
            logger.debug("it is policy");
            return ((PolicyData)objSpec).getPolicyId();
        }       
        logger.error(i18n.getMessage("invalidObjSpec"));
        throw new CasDBException(i18n.getMessage("invalidObjSpec"));
    }

    public static void deletePolicyForObject(String objectSpec, 
                                             String objectSpecDesc) 
        throws CasDBException {
        
        String query = "delete from" + CasDBConstants.TABLE_POLICY + " where " 
            + CasDBConstants.COL_OBJECT_SPEC + "='" + objectSpec.trim() 
            + "' and " + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + objectSpecDesc.trim() + "'";
        CasDBStorage.runUpdateQuery(query);
    }
    
    // Vector of policyData
    public static Vector constructPolicyData(Connection connection, 
                                             ResultSet resultSet, 
                                             boolean external)
        throws CasDBException {

        logger.debug("Construct Policy data " + external);
        Vector policiesVector = null;
        try {
            if ((resultSet != null) && (resultSet.next())) {
                String selfObjId = 
                    Integer.toString(
                    ObjectDataHandler
                    .getObjectId(connection, CasConstants.OBJECT_SELF,
                                 CasConstants.NAMESPACE_SELF));
                policiesVector = new Vector(resultSet.getFetchSize());
                do {
                    boolean addPolicy = false;
                    logger.debug("ConstuctData :" 
                                 + resultSet.getString(
                                 CasDBConstants.COL_POLICY_ID.trim()));
                    String policyId =
                        resultSet.getString(
                                  CasDBConstants.COL_POLICY_ID.trim());
                    String userGpName = 
                        resultSet.getString(
                                  CasDBConstants.COL_USERGP_NAME.trim());
                    String objSpec = 
                        resultSet.getString(
                                  CasDBConstants.COL_OBJECT_SPEC.trim());
                    String objSpecDesc = 
                        resultSet.getString(
                                  CasDBConstants.COL_OBJECT_SPEC_DESC.trim());
                    String actionSpec = 
                        resultSet.getString(
                                  CasDBConstants.COL_ACTION_SPEC.trim());
                     String actionSpecDesc = resultSet
                         .getString(CasDBConstants.COL_ACTION_SPEC_DESC.trim());

                     if (external) {
                         // object spec can only be objectgp or object
                         if ((objSpecDesc
                              .equals(CasConstants.OBJECTGP_SPEC))) {
                             logger.debug("external and Object group "
                                          + "processing");
                             Vector tempVector = 
                                 getPolicyFromObjGp(connection,
                                                    policyId, userGpName, 
                                                    objSpec, actionSpec,
                                                    actionSpecDesc);
                             if (tempVector != null) {
                                 logger.debug("adding the object gp policies "
                                              + tempVector.size());
                                 policiesVector.addAll(tempVector);
                             }
                         } else {
                             if (objSpecDesc
                                 .equals(CasConstants.OBJECT_SPEC)) {
                                 logger.debug("External and object");
                                 if (!objSpec.equals(selfObjId)) {
                                     if ((actionSpecDesc.equals(
                                         CasConstants.SERVICEACTIONGP_SPEC))) {
                                         logger.debug("not cas, action gp");
                                         Vector tempVector =
                                             getPolicyFromActionGp(connection,
                                                                   policyId, 
                                                                   userGpName,
                                                                   objSpec, 
                                                                   actionSpec);
                                         if (tempVector != null) {
                                             logger.debug("adding action gp "
                                                          + " policies");
                                             policiesVector.addAll(tempVector);
                                         }       
                                     } else 
                                         addPolicy = true;
                                 } else
                                     logger.debug("Cas policy");
                             }
                         }
                     } else {
                         logger.debug("Not external");
                         addPolicy = true;
                     }
                     
                     if (addPolicy) {
                         logger.debug("constructin policy");
                         logger.debug("adding policy " + policyId);
                         PolicyData policyData = new PolicyData();
                         policyData.setPolicyId(policyId);
                         policyData.setUserGroupName(userGpName);
                         policyData.setObjectSpec(objSpec);
                         policyData.setObjectSpecDesc(objSpecDesc);
                         policyData.setActionSpec(actionSpec);
                         policyData.setActionSpecDesc(actionSpecDesc);
                         policiesVector.add(policyData); 
                     }
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("constPolicyErr"), exp);
            throw new CasDBException(i18n.getMessage("constPolicyErr")
                                     + exp.getMessage(), exp);
        }
        finally {
            try {
                if (resultSet != null)
                    resultSet.close();
            }       
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("constPolicyErrClose"), exp);
            }
        }
        if ((policiesVector != null) && (policiesVector.size() >0)) {
            if (logger.isDebugEnabled()) {
                printPolicyVector(policiesVector);
            }
            return policiesVector;
        }
        else
            return null;
    }
    
    // Returns set of policies that are part of the service action group
    // expects ObjectSpec is always object and not cas
    private static Vector getPolicyFromActionGp(Connection connection,
                                                String policyId, 
                                                String userGpName,
                                                String objectSpec,
                                                String actionSpec)
        throws CasDBException {
        logger.debug("getPolicyFromActionGp " + actionSpec);
        String serviceActionId[] =
            ServiceTypeActionHandler.retrieveServiceActionGpEntriesAsId(
                                     connection, actionSpec);
        if (serviceActionId != null) {
            logger.debug("service action gp has members, adding");
            Vector policiesVector = new Vector(serviceActionId.length);
            for (int i=0; i<serviceActionId.length; i++) {
                logger.debug("adding policy id " + policyId);
                PolicyData policyData = new PolicyData();
                policyData.setPolicyId(policyId);
                policyData.setUserGroupName(userGpName);
                policyData.setObjectSpec(objectSpec);
                policyData.setObjectSpecDesc(CasConstants.OBJECT_SPEC);
                policyData.setActionSpec(serviceActionId[i]);
                policyData.setActionSpecDesc(
                           CasConstants.SERVICEACTION_SPEC);
                policiesVector.add(policyData); 
            }      
            return policiesVector;
        }
        return null;
    }

    // Returns set of policies that are part of this object group
    // Only external policies are returned
    private static Vector getPolicyFromObjGp(Connection connection,
                                             String policyId, 
                                             String userGpName, 
                                             String objSpec,
                                             String actionSpec,
                                             String actionSpecDesc)
        throws CasDBException {
        logger.debug("getPolicyFromObjGp " + objSpec);

        int selfObjId = 
            ObjectDataHandler.getObjectId(connection, 
                                          CasConstants.OBJECT_SELF,
                                          CasConstants.NAMESPACE_SELF);
        
        String query = "select * from" + CasDBConstants.TABLE_OBJECTGP_ENTRY
            + " where (" + CasDBConstants.COL_OBJECTGP_NAME + "='"
            + objSpec + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC 
            + "='" + CasConstants.OBJECT_SPEC + "') and not (" 
            + CasDBConstants.COL_OBJECT_SPEC + " ='" 
            + Integer.toString(selfObjId) + "')";

        logger.debug("query " + query);
        Vector policiesVector = null;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet != null) && (resultSet.next())) {
                logger.debug("result size " + resultSet.getFetchSize());
                policiesVector = new Vector(resultSet.getFetchSize());
                do {
                    String objSpecDesc =                    
                        resultSet.getString(
                              CasDBConstants.COL_OBJECT_SPEC_DESC.trim());
                    String objName = 
                        resultSet.getString(
                                  CasDBConstants.COL_OBJECT_SPEC.trim());
                    if (actionSpecDesc.equals(
                                       CasConstants.SERVICEACTIONGP_SPEC)) {
                        logger.debug("Action gp");
                        Vector tempVector =
                            getPolicyFromActionGp(connection, policyId, 
                                                  userGpName,
                                                  objName, actionSpec);
                        if (tempVector != null) {
                            logger.debug("adding action gp policies");
                            policiesVector.addAll(tempVector);
                        }
                    }
                    else {
                        // This policy object will have objectSpec as 
                        // object and actionSpec as action
                        logger.debug("adding policy with obj and action"
                                     + policyId);
                        PolicyData policyData = new PolicyData();
                        policyData.setPolicyId(policyId);
                        policyData.setUserGroupName(userGpName);
                        policyData.setObjectSpec(objName);
                        policyData.setObjectSpecDesc(objSpecDesc);
                        policyData.setActionSpec(actionSpec);
                        policyData.setActionSpecDesc(actionSpecDesc);
                        policiesVector.add(policyData); 
                    }
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrPolicyGpMem", query), exp);
            throw new CasDBException(i18n.getMessage("retrPolicyGpMem", "")
                                     + exp.getMessage(), exp);
        }
        finally {
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrPolicyGpMemClose"), exp);
            }
        }
        return policiesVector;
    }

    public static void printPolicyVector(Vector policies) {
        if (policies != null) {
            logger.debug("Printing policies " + policies.size());
            for (int i=0; i<policies.size(); i++) {
                logger.debug(CasStringUtils.policyToString(
                                            (PolicyData)policies.get(i)));
            }
        }
    }
}
