/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.service;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.cas.types.VoidType;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.CasObjectData;
import org.globus.cas.types.UserGroupData;
import org.globus.cas.types.ObjectGroupData;
import org.globus.cas.types.SAMLAuthzQueryType;
import org.globus.cas.types.CasObjectDesc;
import org.globus.cas.types.ArrayOfCasObjectData;
import org.globus.cas.types.ArrayOfString;
import org.globus.cas.types.GetGroupMembers;
import org.globus.cas.types.GetAssertionParam;
import org.globus.cas.types.ArrayOfSAMLAuthzQueryType;

import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.security.SecurityException;

import javax.security.auth.Subject;

import java.security.cert.X509Certificate;

import org.globus.gsi.jaas.JaasGssUtil;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.impl.databaseAccess.CasDBException;
import org.globus.cas.impl.databaseAccess.UserDataHandler;
import org.globus.cas.impl.databaseAccess.ObjectDataHandler;
import org.globus.cas.impl.databaseAccess.PolicyDataHandler;
import org.globus.cas.impl.databaseAccess.PermissionsHandler;
import org.globus.cas.impl.databaseAccess.PolicyQueryHandler;
import org.globus.cas.impl.databaseAccess.NamespaceDataHandler;
import org.globus.cas.impl.databaseAccess.UserGroupDataHandler;
import org.globus.cas.impl.databaseAccess.ServiceTypeDataHandler;
import org.globus.cas.impl.databaseAccess.TrustAnchorDataHandler;
import org.globus.cas.impl.databaseAccess.ObjectGroupDataHandler;
import org.globus.cas.impl.databaseAccess.ExternalPolicyEvaluator;
import org.globus.cas.impl.databaseAccess.ServiceTypeActionHandler;

import java.util.Vector;
import java.util.Iterator;
import java.util.Calendar;
import java.util.StringTokenizer;

import org.w3c.dom.Element;

import org.apache.axis.message.MessageElement;

import org.opensaml.SAMLAction;
import org.opensaml.SAMLSubject;
import org.opensaml.SAMLException;
import org.opensaml.SAMLAuthorizationDecisionQuery;
import org.opensaml.SAMLAuthorizationDecisionStatement;

import org.apache.xml.security.signature.XMLSignature;

import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.impl.security.SecurityMessageElement;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

/**
 * Implementation of CAS Query interface
 */
public class CasQueryPortImpl {

    static Log logger = LogFactory.getLog(CasQueryPortImpl.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.service.errors",
                     CasQueryPortImpl.class.getClassLoader());

    private int defaultAssertionLifetime = 24 * 60 * 60;
    private int serverAssertionLifetime = 24 * 60 * 60;


    public CasQueryPortImpl() throws Exception {

        ResourceContext context = ResourceContext.getResourceContext();
        CasResource resource = (CasResource)context.getResource();
        if (resource.getMaxAssertionLifetime() != -1) {
            serverAssertionLifetime = resource.getMaxAssertionLifetime();
        } else {
            serverAssertionLifetime = defaultAssertionLifetime;
        }
    }

    /**
     * Method to get nickname of the user invoking the method
     *
     * @return user name of the user invoking the method
     * @exception CasFault if any other error occurs.
     */
    public String whoami(VoidType voidType) throws CasFault {
        logger.debug("whoami");
        return CasService.getCallerNickname();
    }

    /**
     * Method to get all applicable policies
     *
     * @param type
     *        type of object whose policy is needed
     *        trustAnchor/namespace/user/userGroup/object/objectGroup/
     *        serviceType, serviceTypeAction/serviceActionGroup
     * @param name
     *        identifier for the object
     *        trustAnchorNickname/namespaceNickname/userNickname/userGroupName/
     *        object(namespace|name)/objectGroupName/serviceTypeName/
     *        serviceTypeAction(serviceType/action)/serviceActionGroupName
     * @return an array of PolicyData
     * @exception NoPermissionFault if the client does not have
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public ArrayOfCasObjectData findApplicablePolicy(CasObjectDesc desc)
        throws CasFault, NoPermissionFault {

        String type = desc.getTypeOfCasObject();
        String name = desc.getNameOfCasObject();

        String baseErrMsg = i18n.getMessage("retrErr", "applicable policies");
        logger.debug("findAppPolicy" + name + " " + type);
        if ((type == null) || (name == null)) {
            String err = i18n.getMessage("allParamErr");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }
        String userName = CasService.getCallerNickname();
        // policy check: if user has cas/query permissions
        checkQueryPermissions(userName, baseErrMsg);
        try {
            if (type.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("Object, get object Id");
                name = CasService.getObjectId(name);
            } else if (type.trim().equals(CasConstants.SERVICEACTION_SPEC)) {
                logger.debug("ServiceAction, get serviceAction Id");
                name = CasService.getServiceActionId(name);
            }
            return new ArrayOfCasObjectData(PolicyQueryHandler.
                                            getAllApplicablePolicy(type,
                                                                   name));
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method to get a list of objects
     *
     * @param type
     *        type of object
     *        "user" or "userGroup" or "object" or "objectGroup" or
     *        "serviceAction" or "serviceActionGroup" or "serviceType"
     *        or "namespace" or "trustAnchor" or "policy"
     * @return list of objects of the particular type
     * @exception NoPermissionFault if the client does not have
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public ArrayOfString list(String type) throws
        CasFault, NoPermissionFault {
        logger.debug("list of " + type);

        String baseErrMsg = i18n.getMessage("listErr", type);
        String userName = CasService.getCallerNickname();
        // policy check: if user has cas/query permissions
        checkQueryPermissions(userName, baseErrMsg);

        if (type == null) {
            String err = i18n.getMessage("allParamErr");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }
        try {
            if (type.trim().equals(CasConstants.TRUSTANCHOR_SPEC)) {
                logger.debug("List trust achors");
                return new ArrayOfString(TrustAnchorDataHandler.list());
            } else if (type.trim().equals(CasConstants.USER_SPEC)) {
                logger.debug("List user");
                return new ArrayOfString(UserDataHandler.list());
            } else if (type.trim().equals(CasConstants.USERGP_SPEC)) {
                logger.debug("List user group");
                return new ArrayOfString(UserGroupDataHandler.list());
            } else if (type.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("List object");
                return new ArrayOfString(ObjectDataHandler.list());
            } else if (type.trim().equals(CasConstants.OBJECTGP_SPEC)) {
                logger.debug("List object group");
                return new ArrayOfString(ObjectGroupDataHandler.list());
            } else if (type.trim().equals(CasConstants.NAMESPACE_SPEC)) {
                logger.debug("List Namespace");
                return new ArrayOfString(NamespaceDataHandler.list());
            } else if (type.trim().equals(CasConstants.SERVICETYPE_SPEC)) {
                logger.debug("List Service Type");
                return new ArrayOfString(ServiceTypeDataHandler.list());
            } else if (type.trim().equals(
                                   CasConstants.SERVICEACTION_SPEC)) {
                logger.debug("List Service Action ");
                return new ArrayOfString(ServiceTypeActionHandler
                                         .listServiceActionMappings());
            } else if (type.trim().equals(
                                   CasConstants.SERVICEACTIONGP_SPEC)) {
                logger.debug("List Service Action Groups ");
                return new ArrayOfString(ServiceTypeActionHandler
                                         .listServiceActionGroups());
            } else if (type.trim().equals(CasConstants.POLICY_SPEC)) {
                logger.debug("List Policy");
                return new ArrayOfString(PolicyDataHandler.list());
            }else {
                String errMesg =
                    i18n.getMessage("typeErr",
                                    CasConstants.TRUSTANCHOR_SPEC + " or "
                                    + CasConstants.USER_SPEC + " or "
                                    + CasConstants.USERGP_SPEC + " or "
                                    + CasConstants.OBJECT_SPEC + " or "
                                    + CasConstants.OBJECTGP_SPEC + " or "
                                    + CasConstants.NAMESPACE_SPEC + " or "
                                    + CasConstants.SERVICETYPE_SPEC + " or "
                                    + CasConstants.SERVICEACTION_SPEC + " or "
                                    + CasConstants.SERVICEACTIONGP_SPEC
                                    + " or " + CasConstants.POLICY_SPEC);
                logger.error(errMesg);
                throw CasService.makeFault(baseErrMsg + errMesg);
            }
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /** Method to get an object given its identifier
     *
     * @param type
     *        type of object
     *        "user" or "userGroup" or "object" or "objectGroup" or
     *        "serviceType" or "trustAnchor" or "namespace" or "policy"
     * @param name
     *        name that identifies the object
     *        userName, userGroupName, object(objectNamespace|objectName),
     *        objectGroupName, serviceTypeName, trustAnchorNick,
     *        namespaceNick, policyId
     * @return returns the CasObject that represents the object of type
     *         "type" and name. Returns null if such an object does not exist
     * @exception NoPermissionFault if the client does not have
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public CasObjectData getCasObject(CasObjectDesc desc)
        throws CasFault, NoPermissionFault {

        String type = desc.getTypeOfCasObject();
        String name = desc.getNameOfCasObject();

        String baseErrMsg = i18n.getMessage("retrErr", type);
        logger.debug("getCasObject " + type + " " + name);
        String userName = CasService.getCallerNickname();
        // policy check: if user has cas/query permissions
        checkQueryPermissions(userName, baseErrMsg);

        if ((type == null) || (name == null)) {
            String err = i18n.getMessage("allParamErr");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }

        try {
            if (type.trim().equals(CasConstants.TRUSTANCHOR_SPEC)) {
                logger.debug("Get trust anchors");
                return TrustAnchorDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.USER_SPEC)) {
                logger.debug("Get user");
                return UserDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.USERGP_SPEC)) {
                logger.debug("Get user group");
                return UserGroupDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("Get object Id and get object");
                String id = CasService.getObjectId(name);
                return ObjectDataHandler.retrieveObjectForId(id);
            } else if (type.trim().equals(CasConstants.OBJECTGP_SPEC)) {
                logger.debug("Get object group");
                return ObjectGroupDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.NAMESPACE_SPEC)) {
                logger.debug("Get namespace");
                return NamespaceDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.SERVICETYPE_SPEC)) {
                logger.debug("Get service type");
                return ServiceTypeDataHandler.retrieveObject(name);
            } else if (type.trim().equals(CasConstants.POLICY_SPEC)) {
                logger.debug("Get policy");
                return PolicyDataHandler.retrieveObject(name);
            }else {
                String errMesg =
                    i18n.getMessage("typeErr",
                                    CasConstants.TRUSTANCHOR_SPEC + " or "
                                    + CasConstants.USER_SPEC + " or "
                                    + CasConstants.USERGP_SPEC + " or "
                                    + CasConstants.OBJECT_SPEC + " or "
                                    + CasConstants.OBJECTGP_SPEC + " or "
                                    + CasConstants.NAMESPACE_SPEC + " or "
                                    + CasConstants.SERVICETYPE_SPEC + " or "
                                    + CasConstants.POLICY_SPEC);
                logger.error(errMesg);
                throw CasService.makeFault(baseErrMsg + errMesg);
            }
        }
        catch (CasDBException exp) {
            logger.error("getCasObject of " + type + " group failed", exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method that returns the group members as an array of string
     *
     * @param groupType
     *        Type of group. Should be "user" or "object" or "serviceAction"
     * @param groupName
     *        Name of the group whose members are to be retrieved
     * @return Group member identifiers as an array of strings
     * @exception NoPermissionFault if the client does not have
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public ArrayOfString getGroupMembers(GetGroupMembers gpMembers)
        throws CasFault, NoPermissionFault {

        String groupType = gpMembers.getTypeOfGroup();
        String groupName = gpMembers.getNameOfGroup();

        String baseErrMsg = i18n.getMessage("gpMemListErr");
        logger.debug("getGroupMembers " + groupType + " " + groupName);
        String userName = CasService.getCallerNickname();
        // policy check: if user has cas/query permissions
        checkQueryPermissions(userName, baseErrMsg);

        if ((groupType == null) || (groupName == null)) {
            String err = i18n.getMessage("allParamErr");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }

        try {
            if (groupType.trim().equals(CasConstants.USER_SPEC)) {
                logger.debug("User group");
                UserGroupData userGroup =
                    (UserGroupData)UserGroupDataHandler.retrieveObject(
                                                        groupName.trim());
                if (userGroup == null) {
                    String errMesg =
                        i18n.getMessage("doesNotExist", new Object[] {
                                        "User group ", groupName.trim() });
                    logger.error(errMesg);
                    throw CasService.makeFault(baseErrMsg + errMesg);
                }
                return userGroup.getUserNames();
            } else if (groupType.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("Object group");
                ObjectGroupData objGroup =
                    (ObjectGroupData)ObjectGroupDataHandler.retrieveObject(
                                                            groupName.trim());
                if (objGroup == null) {
                    String errMesg =
                        i18n.getMessage("doesNotExist", new Object[]
                                        { "Object group ", groupName.trim()});
                    logger.error(errMesg);
                    throw CasService.makeFault(baseErrMsg + errMesg);
                }
                ArrayOfString returnSpecArray =
                    objGroup.getObjectSpecs();
                if (returnSpecArray == null) {
                    logger.debug("No members in this group");
                    return null;
                } else {
                    String returnSpec[] = returnSpecArray.getStrings();
                    String returnSpecDesc[] =
                        objGroup.getObjectSpecsDesc().getStrings();
                    String[] retString = new String[returnSpec.length];
                    for (int i=0; i<returnSpec.length; i++) {
                        String retSpec = returnSpec[i];
                        if (returnSpecDesc[i].equals(
                                              CasConstants.OBJECT_SPEC)) {
                            ObjectData obj =
                                ObjectDataHandler.retrieveObjectForId(retSpec);
                            retSpec = obj.getObjectNamespace()
                                + CasConstants.OBJECTSPEC_DELIMITER
                                + obj.getObjectName();
                        }
                        retString[i] = retSpec + ", " + returnSpecDesc[i];
                    }
                    return new ArrayOfString(retString);
                }
            } else if (groupType.trim().equals(
                                        CasConstants.SERVICEACTION_SPEC)) {
                logger.debug("Service action group");
                String retValue[] =
                    ServiceTypeActionHandler.retrieveServiceActionGpEntries(
                                                groupName.trim());
                return new ArrayOfString(retValue);
            } else {
                String errMesg =
                    i18n.getMessage("typeErr", CasConstants.USER_SPEC +
                                    " or " + CasConstants.OBJECT_SPEC + " or "
                                    + CasConstants.SERVICEACTION_SPEC);
                logger.error(errMesg);
                throw CasService.makeFault(baseErrMsg + errMesg);
            }
        } catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method that generated an assertion with a list of AuthzDecisionStmt
     * that represent the list of actions that are permitted.
     *
     * @param lifetimeInSeconds
     *        Requested lifetime on the assertion in seconds
     * @param SAMLAthzQueryType
     *        - A SAMLAuthorizationDecisionQuery object (OpenSAML data type is
     *        passed as xsd:any).
     *        - The resource maybe set as wildcard or namspace|name.
     *        - A vector of actions are set as a part of the query -
     *        if the first SAMLAction is set to wildcard, then rest
     *        are ignored. Otherwise action should be serviceType/action
     *        - SAMLSubject must have the user's DN.
     * @return SAMLAsssertion
     *        A SAMLAssertion object with SAMLAuthorizationDecisionStatement(s)
     *        (OpenSAML datatype is passed as xsd:any)
     * @exception NoPermissionFault if the client does not have
     *            permission to perform this operation.
     * @exception CasFault if any error occurs.
     */
    public org.globus.cas.types.SAMLAssertion
        getAssertion(GetAssertionParam getAssertionParam)
        throws CasFault, NoPermissionFault {

        SAMLAuthzQueryType[] queries = null;
        ArrayOfSAMLAuthzQueryType array =
            getAssertionParam.getSamlAuthzQuery();
        if (array != null) {
            queries = array.getSAMLAuthzTypes();
        }
        int lifetimeInSeconds = getAssertionParam.getLifetime();

        logger.debug(" Generate assertion " + lifetimeInSeconds);
        String baseErrMsg = i18n.getMessage("assertionGenErr");

        String userNick = CasService.getCallerNickname();
        logger.debug("User nick is " + userNick);

        boolean casQueryPermChecked = false;
        if (queries != null) {
            Vector samlQueryVector = null;
            Vector queryNickVector = null;
            for (int i=0; i<queries.length; i++) {
                logger.debug("get authzDecStmt Query");
                SAMLAuthorizationDecisionQuery samlQuery =
                    getAuthzDecisionQuery(queries[i]);
                if (samlQuery == null)
                    continue;
                // Get subject name and hence username for which the query is.
                SAMLSubject receivedSubject = samlQuery.getSubject();
                String userDN = receivedSubject.getName();
                String issuerDN = receivedSubject.getNameQualifier();
                Iterator confIterator =
                    receivedSubject.getConfirmationMethods();
                String confMethod = null;
                if (confIterator.hasNext()) {
                    confMethod = (String)confIterator.next();
                } else {
                    String err = i18n.getMessage("noConfMethod");
                    logger.error(err);
                    throw CasService.makeFault(baseErrMsg + err);
                }
                String queryUserNick =
                    getUserNickname(userDN, issuerDN, confMethod);
                // Check permissions
                if (queryUserNick.equals(userNick)) {
                    logger.debug("Assertions for self requested, permission"
                                 + " check not required");
                } else {
                    if (!casQueryPermChecked) {
                        checkQueryPermissions(userNick, baseErrMsg);
                        casQueryPermChecked = true;
                    }
                }
                if (samlQueryVector == null)
                    samlQueryVector = new Vector();
                if (queryNickVector == null)
                    queryNickVector = new Vector();
                samlQueryVector.add(samlQuery);
                queryNickVector.add(queryUserNick);
            }

            if (samlQueryVector == null)
                return null;
            // Permissions for allqueries ascertained. Get stmts now
            Vector samlAuthzDecisionVector = null;
            for (int i=0; i<samlQueryVector.size(); i++) {
                SAMLAuthorizationDecisionQuery samlQuery =
                    (SAMLAuthorizationDecisionQuery)samlQueryVector.get(i);
                Vector authzStmts =
                    getAuthzDecisionStmts(samlQuery,
                                          (String)queryNickVector.get(i));
                if (authzStmts != null) {
                    if (samlAuthzDecisionVector == null) {
                        samlAuthzDecisionVector = new Vector();
                    }
                    logger.debug("add stmts");
                    samlAuthzDecisionVector.addAll(authzStmts);
                }
            }
            // No Authz Decision stmt - return null
            if (samlAuthzDecisionVector ==null) {
                logger.debug("No authz decision vector");
                return null;
            }
            return constructSAMLAssertion(lifetimeInSeconds, baseErrMsg,
                                          samlAuthzDecisionVector);
        }
        else {
            return null;
        }
    }

    private org.globus.cas.types.SAMLAssertion
        constructSAMLAssertion(int lifetimeInSeconds, String baseErrMsg,
                               Vector samlAuthzDecisionVector)
        throws CasFault {

        // get server credential
        GlobusGSSCredentialImpl credential = getServerCredential();
        // get DN of CA
        String dnOfCA = getDNofServerCA(credential);
        // SAML assertion construction
        Calendar notOnOrAfter = Calendar.getInstance();
        Calendar notBefore = Calendar.getInstance();
        notBefore.setTime(notOnOrAfter.getTime());
        // requested lifetime is greater than max server assertion
        // then use max server assertion
        logger.debug("lifitimeInSec " + lifetimeInSeconds + " max server "
                     + serverAssertionLifetime);
        if (lifetimeInSeconds > serverAssertionLifetime) {
            logger.debug("Requested lifetime greater than max server"
                         + " lifetime");
            notOnOrAfter.add(Calendar.SECOND, serverAssertionLifetime);
        } else {
            notOnOrAfter.add(Calendar.SECOND, lifetimeInSeconds);
        }

        org.opensaml.SAMLAssertion samlAssertion = null;
        try {

            samlAssertion =
                new org.opensaml.SAMLAssertion(dnOfCA,
                                               notBefore.getTime(),
                                               notOnOrAfter.getTime(),
                                               null,
                                               null,
                                               samlAuthzDecisionVector);
        }
        catch (SAMLException exp) {
            String errMesg = i18n.getMessage("assertionGenErr");
            logger.error(errMesg, exp);
            throw CasService.makeFault(baseErrMsg + errMesg + exp.getMessage(),
                                       exp);
        }

        // get cert chain
        Vector certs = getCertificates(credential);
        // Sign assertion
        try {
            samlAssertion.sign(XMLSignature.ALGO_ID_SIGNATURE_RSA,
                               credential.getPrivateKey(), certs, false);
        }
        catch (SAMLException exp) {
            String errMesg = i18n.getMessage("assertionSignErr");
            logger.error(errMesg, exp);
            throw CasService.makeFault(baseErrMsg + errMesg + exp.getMessage(),
                                       exp);
        }

        org.globus.cas.types.SAMLAssertion retSAMLAssertion =
            new org.globus.cas.types.SAMLAssertion();
        SecurityMessageElement msgElem =
            new SecurityMessageElement((Element)samlAssertion.toDOM());
        retSAMLAssertion.set_any(new MessageElement[] { msgElem });

        logger.debug("SAML assertion: " + samlAssertion.toString());
        return retSAMLAssertion;
    }

    // Takes a single SAMLQuery and returns SAMLAuthorizationDecisionQuery
    private SAMLAuthorizationDecisionQuery
        getAuthzDecisionQuery(SAMLAuthzQueryType query) throws CasFault {
        logger.debug("get Authz Query");
        if (query == null)
            return null;

        MessageElement[] msgElement = query.get_any();
        SAMLAuthorizationDecisionQuery samlQuery = null;
        try {
            samlQuery =
                new SAMLAuthorizationDecisionQuery(msgElement[0].getAsDOM());
        }
        catch (Exception exp) {
            String err = i18n.getMessage("samlAuthzQueryErr");
            logger.error(err, exp);
            throw CasService.makeFault(err + exp.getMessage(), exp);
        }

        return samlQuery;
    }

    // Takes a single SAMLQuery and returns a Vector of SAMLAuthzDecision
    // Stmts.
    private Vector
        getAuthzDecisionStmts(SAMLAuthorizationDecisionQuery samlQuery,
                              String queryUserNick)
        throws CasFault {

        logger.debug("get Authz Decision Stmt");
        if (samlQuery == null) {
            logger.debug("Query is null");
            return null;
        }

        String baseErrMsg = i18n.getMessage("samlDecisionStmtErr");

        Vector samlDecisionStmtVector = null;

        // wildcard
        boolean resourceWildcard = false;
        // Get resource name (should objectNamespace|objectName or wildcard)
        String resource = samlQuery.getResource();
        if (resource == null) {
            String err = i18n.getMessage("samlResourceNull");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }
        logger.debug("Resource is " + resource);
        if (resource.equals(CasConstants.RESOURCE_WILDCARD)) {
            logger.debug("Resource is wildcard");
            resourceWildcard = true;
        }

        // Vector of actions
        // Get actions for which assertion is requested
        Iterator actionEnum = samlQuery.getActions();
        if (actionEnum.hasNext()) {
            // authz policies
            Vector authorizedPolicies = null;
            // Get action
            SAMLAction samlAction = (SAMLAction)actionEnum.next();
            String actionNamespace = samlAction.getNamespace();
            String actionName = samlAction.getData();
            logger.debug("samlAction name " + actionNamespace + " namespace "
                         + actionName);
            // action can be wildcard or serviceType and action
            String serviceActionId = null;
            try {
                if ((actionNamespace.equals(CasConstants.ACTION_NS_WILDCARD))
                    && (actionName.equals(CasConstants.ACTION_WILDCARD))) {
                    logger.debug("Action is wildcard, all actions");
                    if (resourceWildcard) {
                        logger.debug("resource and action are wildcard");
                        authorizedPolicies =
                            ExternalPolicyEvaluator.getPolicies(queryUserNick);
                    } else {
                        logger.debug("action is wildcard, resource is not");
                        authorizedPolicies =
                            ExternalPolicyEvaluator.getPoliciesForResource(
                                                    queryUserNick,
                                                    resource);
                    }
                } else {
                    logger.debug("Action is not wildcard, serviceType/action");
                    // get a set of action ids FILLME
                    Vector actionIds = new Vector();
                    actionIds.add(CasService.getServiceActionId(
                                  actionNamespace
                                  + CasConstants.SERVICEACTION_DELIMITER
                                  + actionName));
                    while (actionEnum.hasNext()) {
                        SAMLAction temp = (SAMLAction)actionEnum.next();
                        serviceActionId =
                            CasService.getServiceActionId(
                                       temp.getNamespace() +
                                       CasConstants.SERVICEACTION_DELIMITER
                                       + temp.getData());
                        actionIds.add(serviceActionId);
                    }
                    if (actionIds == null) {
                        logger.debug("No action ids found");
                        return null;
                    }
                    if (resourceWildcard) {
                        logger.debug("resource is wildcard, action is not");
                        authorizedPolicies =
                            ExternalPolicyEvaluator.getPoliciesForActions(
                                                    queryUserNick,
                                                    actionIds);
                    } else {
                        logger.debug("resource and action are not wild card");
                        authorizedPolicies =
                            ExternalPolicyEvaluator.getPolicies(
                                                    queryUserNick,
                                                    resource,
                                                    actionIds);
                    }
                }
            } catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }

            // The authorized policies is a Vector of Vector, each
            // is per resource and each requires one AuthzDecision Stmt.
            if ((authorizedPolicies != null) &&
                (authorizedPolicies.size() > 0)) {
                logger.debug("Authorized policies not null "
                             + authorizedPolicies.size());
                for (int i=0; i<authorizedPolicies.size(); i++) {
                    // All these policies are on same resource.
                    Vector policies = (Vector)authorizedPolicies.get(i);
                    int numPolicies = policies.size();
                    logger.debug("On one resurce " + numPolicies);
                    // Set of actions for this resource.
                    Vector authzActions = new Vector(numPolicies);
                    PolicyData policyData = null;
                    for (int j=0; j<numPolicies; j++) {
                        policyData = (PolicyData)policies.get(j);
                        String serviceInfo = null;
                        try {
                            serviceInfo =
                            ServiceTypeActionHandler.getServiceActionMapping(
                            policyData.getActionSpec());
                        } catch (CasDBException exp) {
                            String errMesg =
                                i18n.getMessage("retrErr",
                                                "service mapping");
                            logger.error(errMesg, exp);
                            throw CasService.makeFault(baseErrMsg + errMesg
                                                       + exp.getMessage(),exp);

                        }
                        // FIXME add check for tokenizer failing, just in case!
                        StringTokenizer str =
                            new StringTokenizer
                            (serviceInfo,
                             CasConstants.SERVICEACTION_DELIMITER);
                        try {
                            SAMLAction authzAction =
                                new SAMLAction(str.nextToken(),
                                               str.nextToken());
                            authzActions.add(authzAction);
                        }
                        catch (SAMLException exp) {
                            String errMesg = i18n.getMessage("samlActionErr");
                            logger.error(errMesg, exp);
                            throw CasService.makeFault(baseErrMsg + errMesg
                                                       + exp.getMessage(),exp);
                        }
                    }

                    String currentVectorResource = null;
                    try {
                        if (resourceWildcard) {
                            currentVectorResource = ObjectDataHandler
                                .getObjectDescription(policyData
                                                      .getObjectSpec());
                        } else {
                            // all assertions should only be onr
                            // requested object irrespective of what
                            // policy object it matched on. For wild
                            // card scenarios, that object may not
                            // exist in database, so use the resource
                            // string provided.
                            currentVectorResource = resource;
                        }
                    } catch (CasDBException exp) {
                        String errMesg = i18n.getMessage("retrErr",
                                                         "Object description");
                        logger.error(errMesg, exp);
                        throw CasService.makeFault(baseErrMsg + errMesg
                                                   + exp.getMessage(),exp);
                    }
                    // Construct stmt. One per resource is created.
                    SAMLAuthorizationDecisionStatement samlAuthzDecisionStmt =
                        null;
                    if (samlDecisionStmtVector == null) {
                        samlDecisionStmtVector = new Vector();
                    }
                    SAMLSubject receivedSubject = samlQuery.getSubject();
                    logger.debug("Adding SAML authz with subject "
                                 + receivedSubject);
                    logger.debug("SAML Authz for resource "
                                 + currentVectorResource);
                    try {
                        samlAuthzDecisionStmt =
                            new SAMLAuthorizationDecisionStatement(
                                receivedSubject, currentVectorResource,
                                CasConstants.SAML_DECISION,
                                authzActions, null);
                    }
                    catch (SAMLException exp) {
                        logger.error(baseErrMsg, exp);
                        throw CasService.makeFault(baseErrMsg + "\n"
                                                   + exp.getMessage(), exp);
                    }
                    logger.debug("SAML Authz stmt is "
                                 + samlAuthzDecisionStmt);
                    samlDecisionStmtVector.add(samlAuthzDecisionStmt);
                }
            }
        }
        return samlDecisionStmtVector;
    }

    // Returns credential
    private GlobusGSSCredentialImpl getServerCredential() throws CasFault {

        Subject subject = null;
        try {
            subject = SecurityManager.getManager()
                .getServiceSubject(CasConstants.SERVICE_NAME);
        } catch (SecurityException exp) {
            String err = i18n.getMessage("serverSubErr");
            logger.error(err, exp);
            throw CasService.makeFault(err, exp);
        }
        if (subject == null) {
            logger.debug("Subject is null");
        }

        GlobusGSSCredentialImpl credential =
            (GlobusGSSCredentialImpl)JaasGssUtil.getCredential(subject);

        if (credential == null) {
            String errMesg = i18n.getMessage("noCredFound");
            logger.error(errMesg);
            throw CasService.makeFault(errMesg);
        }
        return credential;
    }

    private String getUserNickname(String userDN, String issuerDN,
                                   String authMethod)
        throws CasFault {
        /* FIXME Uncomment when caller's issuer's DN can be extracted
           so that all calls expects userDN and issuerDN
        try {
            String nickname =
                TrustAnchorDataHandler.getNickname(issuerDN, authMethod);
            if (nickname == null) {
                String err = "Trust anchor with " + issuerDN + " and "
                    + authMethod + " not on CAS server";
                logger.error(err);
                throw (CasFault)FaultHelper.makeFault(CasFault.class, err);
            }
        */
        try {
            String userNick =
                UserDataHandler.getUserNickname(userDN);
            if (userNick == null) {
                String err = i18n.getMessage("retrUserNickErr", userDN);
                logger.error(err);
                throw CasService.makeFault(err);
            }
            return userNick;
        } catch (CasDBException exp) {
            String errMesg = i18n.getMessage("retrErr", "user data");
            logger.error(errMesg + exp.toString());
            throw CasService.makeFault(errMesg + exp.getMessage(), exp);
        }
    }


    // Returns certs if it exists, else null
    private Vector getCertificates(GlobusGSSCredentialImpl credential) {

        X509Certificate[] certArray = credential.getCertificateChain();
        Vector certs = null;
        if (certArray.length > 0) {
            certs = new Vector(certArray.length);
            for (int i=0; i<certArray.length; i++) {
                certs.add(certArray[i]);
            }
        }
        return certs;
    }

    // Returns DN of CA
    private String getDNofServerCA(GlobusGSSCredentialImpl credential)
        throws CasFault {

        Vector certs = getCertificates(credential);
        if (certs == null) {
            String errMesg = i18n.getMessage("certNotFound");
            logger.error(errMesg);
            throw CasService.makeFault(errMesg);
        }
        X509Certificate caCert = (X509Certificate)certs.get(certs.size()-1);
        return caCert.getIssuerDN().getName();
    }

    private void checkQueryPermissions(String userName, String baseErrMsg)
        throws NoPermissionFault, CasFault {
        try {
            if (!PermissionsHandler.canQuery(userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }
}
