/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.service;

import javax.security.auth.Subject;

import org.globus.gsi.jaas.JaasGssUtil;

import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;

import java.security.cert.X509Certificate;

import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;

import java.util.Set;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.globus.wsrf.config.ConfigException;

import org.globus.wsrf.utils.FaultHelper;

import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.security.SecurityException;

import org.globus.wsrf.impl.security.descriptor.ServiceSecurityConfig;

import org.globus.cas.faults.CasFault;

import org.globus.cas.types.ObjectData;

import org.globus.cas.impl.CasConstants;
import org.globus.cas.impl.CasException;

import org.globus.cas.impl.databaseAccess.CasDBStorage;
import org.globus.cas.impl.databaseAccess.CasDBOptions;
import org.globus.cas.impl.databaseAccess.CasDBException;
import org.globus.cas.impl.databaseAccess.UserDataHandler;
import org.globus.cas.impl.databaseAccess.ObjectDataHandler;
import org.globus.cas.impl.databaseAccess.ServiceTypeActionHandler;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;

import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceContextException;

import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.utils.AddressingUtils;

import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.servicegroup.client.MDSConfiguration;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;



public class CasService {
 
    static Log logger = LogFactory.getLog(CasService.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.service.errors",
                     CasService.class.getClassLoader());

    private static final String DB_CONFIGURATION = "databaseConfiguration";

    public CasService() throws CasException {
        logger.debug("Initialize database");

        CasDBOptions casDbOptions = null;
        ResourceContext ctx = null;
        try {
            ctx = ResourceContext.getResourceContext();
        } catch (ResourceContextException exp) {
            throw new CasException(exp);
        }
        
        try {
            Context initialContext = new InitialContext();
            casDbOptions = (CasDBOptions)initialContext
                .lookup(org.globus.wsrf.Constants.JNDI_SERVICES_BASE_NAME
                        + ctx.getService() + "/" + DB_CONFIGURATION);
        } catch (NamingException exp) {
            throw new CasException(exp);
        }
        
        try {
            CasDBStorage.setupDBConnection(casDbOptions);
        }
        catch (CasDBException exp) {
            logger.error(i18n.getMessage("dbConnErr"), exp);
            throw new CasException(i18n.getMessage("dbConnErr")
                                   + exp.getMessage(), exp);
        }
        logger.debug("initialize policies called");
        try {
            CasConstants.initializePolicies();
        } catch (CasDBException exp) {
            logger.error(i18n.getMessage("policyInitErr"));
            throw new CasException(i18n.getMessage("policyInitErr")
                                   + exp.getMessage(), exp);
        }

        // Initialize security properties, since DN is being extracted
        logger.debug("Initializing security in CAS");
        try {
            ServiceSecurityConfig.initialize(CasConstants.SERVICE_NAME);
        } catch (ConfigException exp) {
            String str = i18n.getMessage("secDescInitErr");
            logger.error(str, exp);
            throw new CasException(str, exp);
        } catch (SecurityException exp) {
            String str = i18n.getMessage("secDescInitErr");
            logger.error(str, exp);
            throw new CasException(str, exp);
        }
        
        Subject subject = null;
        try {
             SecurityManager manager = SecurityManager.getManager();
             subject = manager.getServiceSubject(CasConstants.SERVICE_NAME);
        } catch (SecurityException exp) {
            String err = i18n.getMessage("serverSubErr");
            logger.error(err, exp);
            throw new CasException(err, exp);
        }

	GlobusGSSCredentialImpl credential = 
	    (GlobusGSSCredentialImpl)JaasGssUtil.getCredential(subject);
	if (credential == null) {
	    String errMesg = i18n.getMessage("noCredFound");
	    logger.error(errMesg);
	    throw new CasException(errMesg);
	}
	String subjectDN = credential.getGlobusCredential().getIdentity();

        CasResource resource = null;
        try {
            ResourceContext context = ResourceContext.getResourceContext();
            resource = (CasResource)context.getResource();
        } catch (Exception exp) {
            String err = i18n.getMessage("retrResourceErr");
            logger.error(err, exp);
            throw new CasException(err, exp);            
        }
        resource.setServerDN(subjectDN);

        registerToMDS();

    }
    
    // Returns the CAS user nickname of the user making the call
    public static String getCallerNickname() throws CasFault {
        
        Subject callerSubject = 
            (Subject)MessageContext.getCurrentContext()
            .getProperty(Constants.PEER_SUBJECT);
        if (callerSubject != null) {
            Set publicCreds = callerSubject.getPublicCredentials();
            if (publicCreds != null) {
                Iterator iter = publicCreds.iterator();
                if (iter.hasNext()) {
                    X509Certificate[] certs = (X509Certificate[])iter.next();
                    // FIXME - use this info.
                    if (certs != null) {
                        String issuer = certs[0].getIssuerDN().getName();
                        logger.debug("Issuer" + issuer);
                    }
                }
            }
        }

        String subjectDN = SecurityManager.getManager().getCaller();
        return getCasNickname(subjectDN);
    }
    
    // Returns the CAS user nickname for the given subjectDN
    public static String getCasNickname(String subjectDN) throws CasFault {
        try {
            String userNick = UserDataHandler.getUserNickname(subjectDN);
            if (userNick == null) {
                String errMesg = i18n.getMessage("subjectDNMapErr", subjectDN);
                logger.error(errMesg);
                throw makeFault(errMesg);
            }
            return userNick;
        }
        catch (CasDBException exp) {
            String err = i18n.getMessage("retrUserNickErr", subjectDN);
            logger.error(err, exp);
            throw makeFault(err + exp.getMessage(), exp);
        }
    }
    
    // Takes in objectNamespace|object and returns objectId
    public static String getObjectId(String objSpec) throws CasFault {
        
        String baseErrMsg = i18n.getMessage("retrObjDataErr", objSpec);
        StringTokenizer strTok = 
            new StringTokenizer(objSpec.trim(), 
                                CasConstants.OBJECTSPEC_DELIMITER);
        if (strTok.countTokens() != 2) {
            String errMesg = 
                i18n.getMessage("malFormedObjStr", 
                                CasConstants.OBJECTSPEC_DELIMITER); 
            logger.error(errMesg);
            throw makeFault(baseErrMsg + errMesg);
        }
        String objectNamespace = strTok.nextToken();
        String objectName = strTok.nextToken();
        try {
            int objId = ObjectDataHandler.getObjectId(objectName, 
                                                      objectNamespace);
            if (objId <= 0) {
                String errMsg = i18n.getMessage("doesNotExist", new Object[] 
                    { "Object", objSpec });
                logger.error(errMsg);
                throw makeFault(baseErrMsg + errMsg);
            }
            objSpec = Integer.toString(objId);
        }
        catch (CasDBException e) {
            logger.error(i18n.getMessage("objIdErr", objSpec), e);
            throw makeFault(baseErrMsg + e.getMessage(), e);
        }
        return objSpec;
    }
    
    // Take in serviceType/action and returns objectId
    public static String getServiceActionId(String actionSpec) 
        throws CasFault {
        
        String baseErrMsg = "Retrieving service action data failed.\n";
        logger.debug("parsing actionSpec to get serviceActionId");
        StringTokenizer strTok = 
            new StringTokenizer(actionSpec, 
                                CasConstants.SERVICEACTION_DELIMITER);
        if (strTok.countTokens() != 2) {
            String errMesg = 
                i18n.getMessage("malFormedServiceAct", 
                                CasConstants.SERVICEACTION_DELIMITER);
            logger.error(errMesg);
            throw makeFault(baseErrMsg + errMesg);
        }
        try {
            int serviceActionId = 
                ServiceTypeActionHandler
                .getServiceActionId(strTok.nextToken(),
                                    strTok.nextToken());
            if (serviceActionId == -1) {
                String errMesg = i18n.getMessage("doesNotExist", new Object[]
                                                 { "Service/Action" ,
                                                   actionSpec });
                logger.error(errMesg);
                throw makeFault(baseErrMsg + errMesg);
            }
            actionSpec = Integer.toString(serviceActionId);
        } catch (CasDBException exp) {
            String errMesg = i18n.getMessage("serviceActionData", actionSpec);
            logger.error(errMesg, exp);
            throw makeFault(baseErrMsg + errMesg + exp.getMessage(), exp);
        }
        return actionSpec;
    }
    
    // Take in service action id and return serviceType/action mapping
    public static String getServiceActionMapping(String serviceId)
        throws CasFault {
        
        String baseErrMsg = i18n.getMessage("retrErr", 
                                            "service action mapping");
        logger.debug("parsing serviceActionId to serviceType/action");
        String serviceAction = null;
        try {
            serviceAction =
                ServiceTypeActionHandler.getServiceActionMapping(
                                         Integer.parseInt(serviceId));
        }
        catch (CasDBException exp) {
            logger.error(i18n.getMessage("serviceActionMap"), exp);
            throw makeFault(baseErrMsg + exp.getMessage(), exp);
        }
        
        if (serviceAction == null) {
            logger.error(baseErrMsg);
            throw makeFault(baseErrMsg);
        }
        return serviceAction;
    }
    
    // Take in object id and return objNS|objName
    public static String getObjectData(String objectId) throws CasFault {
        
        String baseErrMsg = i18n.getMessage("retrErr", "object data");
        try {
            ObjectData objData = 
                ObjectDataHandler.retrieveObjectForId(objectId);
            if (objData == null) {
                logger.error(baseErrMsg);
                throw makeFault(baseErrMsg);
            }
            return objData.getObjectNamespace() + 
                CasConstants.OBJECTSPEC_DELIMITER + objData.getObjectName();
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    public static CasFault makeFault(String errorMsg) {
        return makeFault(errorMsg, null);
    }
    
    public static CasFault makeFault(String errMsg, Exception exp) {
        
        CasFault fault = new CasFault();
        FaultHelper helper = new FaultHelper(fault);
        if (errMsg != null) {
            helper.setDescription(errMsg);
        }
        if (exp != null) {
            helper.addFaultCause(exp);
        }

        return (CasFault)helper.getFault();
    }

   public void registerToMDS() {

        try {
            // get resource context
            ResourceContext rCtx = ResourceContext.getResourceContext();

            // get mds info
            Context ctx = new InitialContext();
            String config = org.globus.wsrf.Constants.JNDI_SERVICES_BASE_NAME +
                rCtx.getService() + "/mdsConfiguration";
            MDSConfiguration mdsConfiguration =
                (MDSConfiguration) ctx.lookup(config);

            // determine if we should register
            if(mdsConfiguration.getRegAsBoolean()) {
                logger.info(i18n.getMessage("mdsEnabled"));
                // register

                EndpointReferenceType epr = 
                    AddressingUtils.createEndpointReference(null);

                // perform registraion
                ServiceGroupRegistrationClient.register(
                                 rCtx,
                                 epr,
                                 "/etc/globus_cas_service/registration.xml");
            } else {
                logger.debug(i18n.getMessage("mdsSuppressed"));
            }
        } catch(Exception e) {
            logger.error(i18n.getMessage("mdsRegErr"),e);
        }

    }

}
