/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_i_globusrun_ws.h"
#include "ManagedExecutableJobService_client.h"
#include "wsgram_StateChangeNotificationMessageWrapperType.h"
#include "wsgram_stdoutURL.h"
#include "wsgram_stderrURL.h"
#include "wsgram_subJobEndpoint.h"
#include "wsgram_state.h"
#include "wsgram_holding.h"
#include "wsgram_fault.h"
#include "wsgram_exitCode.h"

static globus_xio_driver_t              globus_l_file_driver;

static
void
globus_l_shutdown_monitor(
    globus_i_monitor_info_t *           monitor);

static
void
globus_l_signal_done(
    globus_i_monitor_info_t *           monitor,
    globus_result_t                     result)
{
    globus_mutex_lock(&monitor->lock);
    {
        if(result != GLOBUS_SUCCESS &&
            monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
        {
            monitor->error = globus_error_get(result);
            globus_l_shutdown_monitor(monitor);
        }
    
        if(--monitor->cb_count == 0)
        {
            globus_cond_signal(&monitor->cond);
        }
    }
    globus_mutex_unlock(&monitor->lock);
}

static
void
globus_l_release_cleanup_hold_cb(
    ManagedJobService_client_handle_t   handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsgram_ReleaseOutputType *    releaseResponse,
    ManagedJobPortType_release_fault_t  fault_type,
    const xsd_any *                     fault)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) user_args;
    
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("removing hold state", result);
        goto error;
    }
        
error:
    ManagedJobService_client_destroy(handle);
    globus_l_signal_done(monitor, result);
}
 
/* called locked */
static
void
globus_l_release_cleanup_hold(
    globus_i_monitor_info_t *           monitor)
{
    globus_result_t                     result;
    wsgram_ReleaseInputType             release;
    ManagedJobService_client_handle_t   handle;
    globus_bool_t                       tls = GLOBUS_FALSE;

    if (strncmp(
            monitor->command_info->job_epr->Address.base_value,
            "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    wsgram_ReleaseInputType_init_contents(&release);
    result = ManagedJobService_client_init(
            &handle,
            monitor->command_info->message_attr,
            tls
                ? monitor->command_info->tls_handler_chain
                : monitor->command_info->secmess_handler_chain);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("removing hold state", result);
        goto error_init;
    }
    
    result = ManagedJobPortType_release_epr_register(
        handle,
        monitor->command_info->job_epr,
        &release,
        globus_l_release_cleanup_hold_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("removing hold state", result);
        goto error_register;
    }
    
    monitor->my_state = GLOBUS_I_MY_STATE_REMOVE_HOLD;
    monitor->cb_count++;
    wsgram_ReleaseInputType_destroy_contents(&release);
    
    return;

error_register:
    ManagedJobService_client_destroy(handle);
error_init:
    wsgram_ReleaseInputType_destroy_contents(&release);
    
    if(result != GLOBUS_SUCCESS  &&
        monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
    {
        monitor->error = globus_error_get(result);
        globus_l_shutdown_monitor(monitor);
    }
}

static
void
globus_l_streaming_cb(
    globus_i_streaming_t *              handle,
    const char *                        url,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) user_arg;
    
    if(result != GLOBUS_SUCCESS)
    {
        char *                          tmp;
        
        tmp = globus_error_print_friendly(globus_error_peek(result));
        fprintf(stderr,
            "globusrun-ws: ignoring error while streaming %s:\n%s",
            url, tmp);
        globus_free(tmp);
    }
    
    globus_mutex_lock(&monitor->lock);
    if(--monitor->stream_count == 0)
    {
        if(monitor->my_state == GLOBUS_I_MY_STATE_STREAMING ||
            monitor->my_state == GLOBUS_I_MY_STATE_STREAMING_FINISHING)
        {
            monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHED;
            if(monitor->job_state == GLOBUS_I_JOB_STATE_CLEANUP_HOLD)
            {
                globus_l_release_cleanup_hold(monitor);
            }
        }
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
}

static
void
globus_l_get_stream_urls_cb(
    ManagedExecutableJobService_client_handle_t handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrp_GetMultipleResourcePropertiesResponseType * gmrpr,
    ManagedExecutableJobPortType_GetMultipleResourceProperties_fault_t fault_type,
    const xsd_any *                     fault)
{
    globus_i_monitor_info_t *           monitor;
    int                                 i;
    xsd_anyURI *                        stdout_url = NULL;
    xsd_anyURI *                        stderr_url = NULL;
    
    monitor = (globus_i_monitor_info_t *) user_args;
    
    globus_mutex_lock(&monitor->lock);
    globus_list_remove(&monitor->active_queries,
        globus_list_search(monitor->active_queries, handle));
        
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying stream urls", result);
        goto error;
    }
    
    if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING ||
        monitor->my_state == GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH)
    {
        for(i = 0; i < gmrpr->any.length; i++)
        {
            if(gmrpr->any.elements[i].any_info == 
                &ManagedExecutableJobPortType_stdoutURL_rp_info ||
		gmrpr->any.elements[i].any_info ==
		&wsgram_stdoutURL_info)
            {
                globus_assert(!stdout_url && "Unexpected duplicate entry");
                
                stdout_url = (xsd_anyURI *) gmrpr->any.elements[i].value;
            }
            else if(gmrpr->any.elements[i].any_info == 
                &ManagedExecutableJobPortType_stderrURL_rp_info ||
		gmrpr->any.elements[i].any_info ==
		&wsgram_stderrURL_info)
            {
                globus_assert(!stderr_url && "Unexpected duplicate entry");
                
                stderr_url = (xsd_anyURI *) gmrpr->any.elements[i].value;
            }
            else
            {
                result = GlobusLErrorInvalid("response",
                    "expected only stdoutURL and stderrURL");
                goto error;
            }
        }
        
        if(stdout_url && *stdout_url)
        {
            result = globus_i_streaming_add(
                &monitor->streaming,
                *stdout_url,
                &monitor->command_info->ftp_attr,
                monitor->stdout_handle);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  tmp;
                
                tmp = globus_error_print_friendly(globus_error_peek(result));
                fprintf(stderr,
                    "globusrun-ws: ignoring error while streaming %s\n%s",
                    *stdout_url, tmp);
                globus_free(tmp);
            }
            
            monitor->stream_count++;
            monitor->cb_count++;
        }
        
        if(stderr_url && *stderr_url)
        {
            result = globus_i_streaming_add(
                &monitor->streaming,
                *stderr_url,
                &monitor->command_info->ftp_attr,
                monitor->stderr_handle);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  tmp;
                
                tmp = globus_error_print_friendly(globus_error_peek(result));
                fprintf(stderr,
                    "globusrun-ws: ignoring error while streaming %s\n%s",
                    *stderr_url, tmp);
                globus_free(tmp);
            }
            
            monitor->stream_count++;
            monitor->cb_count++;
        }
        
        result = GLOBUS_SUCCESS;
    }
    
error:
    ManagedExecutableJobService_client_destroy(handle);
    
    /* only warn on errors related to streaming */
    if(result != GLOBUS_SUCCESS)
    {
        char *                          tmp;
        
        tmp = globus_error_print_friendly(globus_error_peek(result));
        fprintf(stderr,
            "globusrun-ws: ignoring error while streaming\n%s", tmp);
        globus_free(tmp);
    }
    
    if(--monitor->stream_query_count == 0)
    {
        if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING ||
            monitor->my_state == GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH)
        {
            if(monitor->stream_count == 0)
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHED;
                if(monitor->job_state == GLOBUS_I_JOB_STATE_CLEANUP_HOLD)
                {
                    globus_l_release_cleanup_hold(monitor);
                }
            }
            else if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING)
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING;
            }
            else /* GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH */
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHING;
                globus_i_streaming_stop(
                    &monitor->streaming, GLOBUS_I_MJG_STREAMING_FINISH, NULL);
            }
        }
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
}

/* call locked */
static
void
globus_l_get_stream_urls(
    globus_i_monitor_info_t *           monitor,
    wsa_EndpointReferenceType *         job_epr)
{
    ManagedExecutableJobService_client_handle_t handle;
    globus_result_t                     result;
    char *                              tmp;
    xsd_QName *                         qname;
    wsrp_GetMultipleResourcePropertiesType gmrp;
    globus_bool_t                       tls = GLOBUS_FALSE;

    if (strncmp(job_epr->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    wsrp_GetMultipleResourcePropertiesType_init_contents(&gmrp);
    
    result = ManagedExecutableJobService_client_init(&handle,
        monitor->command_info->message_attr,
        tls
            ? monitor->command_info->tls_handler_chain
            : monitor->command_info->secmess_handler_chain);

    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying stream urls", result);
        goto error_init;
    }
    
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        result = GlobusLErrorWrapResult(
            "querying stream urls", GlobusLErrorMemory("qname entry"));
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname, &ManagedExecutableJobPortType_stdoutURL_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying stream urls", result);
        goto error;
    }
    
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        result = GlobusLErrorWrapResult(
            "querying stream urls", GlobusLErrorMemory("qname entry"));
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname, &ManagedExecutableJobPortType_stderrURL_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying stream urls", result);
        goto error;
    }
    
    result = ManagedExecutableJobPortType_GetMultipleResourceProperties_epr_register(
        handle,
        job_epr,
        &gmrp,
        globus_l_get_stream_urls_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying stream urls", result);
        goto error;
    }
    
    monitor->cb_count++;
    monitor->stream_query_count++;
    globus_list_insert(&monitor->active_queries, handle);
    
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);
    
    return;

error:
    ManagedExecutableJobService_client_destroy(handle);
error_init:
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);

    tmp = globus_error_print_friendly(globus_error_peek(result));
    fprintf(stderr,
        "globusrun-ws: ignoring error while streaming\n%s", tmp);
    globus_free(tmp);
}

static
void
globus_l_get_multijob_eprs_cb(
    ManagedMultiJobService_client_handle_t handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrp_GetResourcePropertyResponseType * response,
    ManagedMultiJobPortType_GetResourceProperty_fault_t fault_type,
    const xsd_any *                     fault)
{
    globus_i_monitor_info_t *           monitor;
    int                                 i;
    wsa_EndpointReferenceType *         epr;
    
    monitor = (globus_i_monitor_info_t *) user_args;
    
    globus_mutex_lock(&monitor->lock);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying multijob eprs", result);
        goto error;
    }
    
    if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING ||
        monitor->my_state == GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH)
    {
        for(i = 0; i < response->any.length; i++)
        {
            globus_assert(response->any.elements[i].any_info == 
                &ManagedMultiJobPortType_subJobEndpoint_rp_info ||
		response->any.elements[i].any_info ==
		&wsgram_subJobEndpoint_info);
            
            epr = (wsa_EndpointReferenceType *)
                response->any.elements[i].value;
            
            globus_l_get_stream_urls(monitor, epr);
        }
    }
    
error:
    ManagedMultiJobService_client_destroy(monitor->multijob_handle);
    monitor->multijob_handle = NULL;
    
    /* only warn on errors related to streaming */
    if(result != GLOBUS_SUCCESS)
    {
        char *                          tmp;
        
        tmp = globus_error_print_friendly(globus_error_peek(result));
        fprintf(stderr,
            "globusrun-ws: ignoring error while streaming\n%s", tmp);
        globus_free(tmp);
    }
    
    if(--monitor->stream_query_count == 0)
    {
        if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING ||
            monitor->my_state == GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH)
        {
            if(monitor->stream_count == 0)
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHED;
                if(monitor->job_state == GLOBUS_I_JOB_STATE_CLEANUP_HOLD)
                {
                    globus_l_release_cleanup_hold(monitor);
                }
            }
            else if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING)
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING;
            }
            else /* GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH */
            {
                monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHING;
                globus_i_streaming_stop(
                    &monitor->streaming, GLOBUS_I_MJG_STREAMING_FINISH, NULL);
            }
        }
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
}

/* called locked */
static
void
globus_l_get_multijob_eprs(
    globus_i_monitor_info_t *           monitor,
    wsa_EndpointReferenceType *         job_epr)
{
    globus_result_t                     result;
    char *                              tmp;
    globus_bool_t                       tls = GLOBUS_FALSE;

    if (strncmp(job_epr->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    result = ManagedMultiJobService_client_init(&monitor->multijob_handle,
        monitor->command_info->message_attr,
        tls
            ? monitor->command_info->tls_handler_chain
            : monitor->command_info->secmess_handler_chain);

    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying multijob eprs", result);
        goto error_init;
    }
    
    result = ManagedMultiJobPortType_GetResourceProperty_epr_register(
        monitor->multijob_handle,
        job_epr,
        &ManagedMultiJobPortType_subJobEndpoint_rp_qname,
        globus_l_get_multijob_eprs_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying multijob eprs", result);
        goto error;
    }
    
    monitor->cb_count++;
    monitor->stream_query_count++;
    
    return;

error:
    ManagedMultiJobService_client_destroy(monitor->multijob_handle);
    monitor->multijob_handle = NULL;
error_init:
    tmp = globus_error_print_friendly(globus_error_peek(result));
    fprintf(stderr,
        "globusrun-ws: ignoring error while streaming\n%s", tmp);
    globus_free(tmp);
}

/* called locked */
static
void
globus_l_start_streaming(
    globus_i_monitor_info_t *           monitor)
{
    if(monitor->command_info->streaming)
    {
        if(strstr(
            monitor->command_info->job_epr->Address.base_value,
            MANAGEDEXECUTABLEJOBSERVICE_BASE_PATH))
        {
            globus_l_get_stream_urls(monitor, monitor->command_info->job_epr);
        }
        else
        {
            globus_l_get_multijob_eprs(
                monitor, monitor->command_info->job_epr);
        }
    }
    
    if(monitor->stream_query_count > 0)
    {
        monitor->my_state = GLOBUS_I_MY_STATE_QUERYING;
    }
    else
    {
        monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHED;
    }
}

static
void
globus_l_service_engine_done_cb(
    globus_result_t                     result,
    globus_service_engine_t             engine,
    void *                              args)
{
    globus_l_signal_done((globus_i_monitor_info_t *) args, result);
}

static
globus_i_monitor_job_state_t
globus_l_state_to_enum(
    const char *                        state_str)
{
    if(strcmp(state_str, wsgram_StateEnumeration_Unsubmitted) == 0)
        return GLOBUS_I_JOB_STATE_UNSUBMITTED;
    else if(strcmp(state_str, wsgram_StateEnumeration_StageIn) == 0)
        return GLOBUS_I_JOB_STATE_STAGEIN;
    else if(strcmp(state_str, wsgram_StateEnumeration_Pending) == 0)
        return GLOBUS_I_JOB_STATE_PENDING;
    else if(strcmp(state_str, wsgram_StateEnumeration_Active) == 0)
        return GLOBUS_I_JOB_STATE_ACTIVE;
    else if(strcmp(state_str, wsgram_StateEnumeration_Suspended) == 0)
        return GLOBUS_I_JOB_STATE_SUSPENDED;
    else if(strcmp(state_str, wsgram_StateEnumeration_StageOut) == 0)
        return GLOBUS_I_JOB_STATE_STAGEOUT;
    else if(strcmp(state_str, wsgram_StateEnumeration_CleanUp) == 0)
        return GLOBUS_I_JOB_STATE_CLEANUP;
    else if(strcmp(state_str, wsgram_StateEnumeration_Done) == 0)
        return GLOBUS_I_JOB_STATE_DONE;
    else if(strcmp(state_str, wsgram_StateEnumeration_Failed) == 0)
        return GLOBUS_I_JOB_STATE_FAILED;
    else
    {
        fprintf(stderr, "warning: Unexpected state %s\n", state_str);
        return GLOBUS_I_JOB_STATE_UNKNOWN;
    }
}

/* call locked */
static
void
globus_l_state_change(
    globus_i_monitor_info_t *           monitor,
    const char *                        state_str,
    globus_bool_t                       holding,
    wsgram_FaultResourcePropertyType *  gram_fault,
    xsd_int *                           exit_code)
{
    globus_i_monitor_job_state_t        job_state;
    
    monitor->is_idle = GLOBUS_FALSE;
    
    job_state = globus_l_state_to_enum(state_str);
    
    if(holding)
    {
        if(job_state == GLOBUS_I_JOB_STATE_CLEANUP)
        {
            job_state = GLOBUS_I_JOB_STATE_CLEANUP_HOLD;
        }
    }
    
    /* since i fire off a subscribe and get status at the same time,
     * need to make sure timing differences dont cause me to misorder the
     * states
     */
    if(job_state <= monitor->job_state &&
        !(monitor->job_state == GLOBUS_I_JOB_STATE_SUSPENDED &&
            job_state == GLOBUS_I_JOB_STATE_ACTIVE))
    {
        return;
    }
    
    if(job_state == GLOBUS_I_JOB_STATE_DONE ||
        job_state == GLOBUS_I_JOB_STATE_FAILED)
    {
        if(exit_code)
        {
            monitor->command_info->job_exit_code = *exit_code;
        }
        
        if(gram_fault && !monitor->command_info->job_fault)
        {
            if (monitor->command_info->print_fault_type)
            {
                char * tns = WSGRAM_FAULTRESOURCEPROPERTYTYPE_NS;
                char * local = NULL;

                switch (gram_fault->choice_value.type)
                {
                    case wsgram_FaultResourcePropertyType_fault:
                        local = "FaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_credentialSerializationFault:
                        local = "CredentialSerializationFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_insufficientCredentialsFault:
                        local = "InsufficientCredentialsFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_internalFault:
                        local = "InternalFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_invalidCredentialsFault:
                        local = "InvalidCredentialsFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_serviceLevelAgreementFault:
                        local = "ServiceLevelAgreementFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_executionFailedFault:
                        local = "ExecutionFailedFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_filePermissionsFault:
                        local = "FilePermissionsFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_invalidPathFault:
                        local = "InvalidPathFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_stagingFault:
                        local = "StagingFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_unsupportedFeatureFault:
                        local = "UnsupportedFeatureFaultType";
                        break;
                    case wsgram_FaultResourcePropertyType_undefined:
                        ;
                }

                if (local != NULL)
                {
                    fprintf(stderr, "Fault Type: {%s}%s\n", tns, local);
                }
                else
                {
                    fprintf(stderr, "Fault Type: UNKNOWN\n");
                }
            }
            monitor->command_info->job_fault =
                globus_i_gram_fault_to_error(gram_fault);
        }
    }
    
    if(monitor->job_state != job_state && !monitor->command_info->quiet)
    {
        if(!monitor->command_info->streaming ||
            monitor->command_info->stream_err_file ||
            !monitor->command_info->stream_out_file)
        {
            fprintf(stderr, "Current job state: %s%s\n",
                state_str, holding ? "-Hold" : "");
        }
        else if(monitor->command_info->stream_out_file)
        {
            fprintf(stdout, "Current job state: %s%s\n",
                state_str, holding ? "-Hold" : "");
        }
    }
    
    monitor->job_state = job_state;
    
    switch(job_state)
    {
      case GLOBUS_I_JOB_STATE_STAGEIN:
      case GLOBUS_I_JOB_STATE_PENDING:
        if(holding)
        {
            globus_l_release_cleanup_hold(monitor);
        }
        break;
      case GLOBUS_I_JOB_STATE_ACTIVE:
      case GLOBUS_I_JOB_STATE_SUSPENDED:
        if(monitor->my_state == GLOBUS_I_MY_STATE_NONE)
        {
            globus_l_start_streaming(monitor);
        }
        break;
      case GLOBUS_I_JOB_STATE_STAGEOUT:
      case GLOBUS_I_JOB_STATE_CLEANUP_HOLD:
      case GLOBUS_I_JOB_STATE_CLEANUP:
        if(monitor->my_state == GLOBUS_I_MY_STATE_NONE)
        {
            globus_l_start_streaming(monitor);
        }
        
        if(monitor->my_state == GLOBUS_I_MY_STATE_QUERYING)
        {
            monitor->my_state = GLOBUS_I_MY_STATE_QUERYING_NEED_FINISH;
        }
        else if(monitor->my_state == GLOBUS_I_MY_STATE_STREAMING)
        {
            monitor->my_state = GLOBUS_I_MY_STATE_STREAMING_FINISHING;
            globus_i_streaming_stop(
                &monitor->streaming, GLOBUS_I_MJG_STREAMING_FINISH, NULL);
        }
        
        if((job_state == GLOBUS_I_JOB_STATE_CLEANUP_HOLD &&
            monitor->my_state == GLOBUS_I_MY_STATE_STREAMING_FINISHED) ||
           (job_state == GLOBUS_I_JOB_STATE_STAGEOUT && holding))
        {
            globus_l_release_cleanup_hold(monitor);
        }
        break;
      case GLOBUS_I_JOB_STATE_DONE:
      case GLOBUS_I_JOB_STATE_FAILED:
        if(monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
        {
            globus_l_shutdown_monitor(monitor);
        }
      default:
        break;
    }
}

static
void
globus_l_notify_cb(
    void *                              arg,
    wsnt_TopicExpressionType *          topic,
    wsa_EndpointReferenceType *         producer,
    xsd_anyType *                       message)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) arg;

    if(message && message->any_info == 
        &wsgram_StateChangeNotificationMessageWrapperType_info)
    {
        wsgram_StateChangeNotificationMessageWrapperType * scnmw;
        wsgram_StateChangeNotificationMessageType * scnm;
        
        scnmw = (wsgram_StateChangeNotificationMessageWrapperType *)
            message->value;
        scnm = &scnmw->stateChangeNotificationMessage;
        
        /* XXX should verify the source of the notification, first */
        globus_mutex_lock(&monitor->lock);
        {
            globus_l_state_change(monitor,
                scnm->state, scnm->holding, scnm->fault, &scnm->exitCode);
        }
        globus_mutex_unlock(&monitor->lock);
    }
    else if(message && message->any_info && message->any_info->type &&
        !monitor->command_info->quiet)
    {
        fprintf(stderr, "warning: Unexpected notification type %s:%s\n",
            message->any_info->type->Namespace
                ? message->any_info->type->Namespace : "<no namespace>",
            message->any_info->type->local
                ? message->any_info->type->local : "<no name>");
    }
}

static
globus_result_t
globus_l_setup_notification_consumer(
    globus_i_monitor_info_t *           monitor)
{
    globus_result_t                     result;
    wsnt_SubscribeType *                subscribe;
    
    subscribe = &monitor->subscribe;
    
    wsnt_SubscribeType_init_contents(subscribe);
    result = xsd_any_init(&subscribe->TopicExpression.any);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_topic_init;
    }
    
    subscribe->TopicExpression.any->any_info = &xsd_QName_contents_info;
    result = xsd_QName_copy(
        (xsd_QName **) &subscribe->TopicExpression.any->value,
        &ManagedJobPortType_state_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_topic_init;
    }
    
    result = xsd_anyURI_copy_cstr(
        &subscribe->TopicExpression._Dialect,
        "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple");
    if(result != GLOBUS_SUCCESS)
    {
        goto error_topic_init;
    }
    
    result = xsd_boolean_init(&subscribe->UseNotify);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_topic_init;
    }
    
    *subscribe->UseNotify = GLOBUS_TRUE;

    result = globus_service_engine_init(&monitor->engine,
        monitor->command_info->message_attr, NULL,
        monitor->use_https
            ? monitor->command_info->service_handler_chain
            : monitor->command_info->secmess_service_handler_chain, 
        monitor->use_https);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_engine;
    }
    
    result = globus_notification_create_consumer(
        &subscribe->ConsumerReference,
        monitor->engine, globus_l_notify_cb, monitor);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_consumer;
    }
    
    return GLOBUS_SUCCESS;

error_consumer:
    globus_service_engine_destroy(monitor->engine);
error_engine:
error_topic_init:
    wsnt_SubscribeType_destroy_contents(subscribe);
    return result;
}

static
void
globus_l_cleanup_notification_consumer(
    globus_i_monitor_info_t *           monitor)
{
    globus_notification_destroy_consumer(
        &monitor->subscribe.ConsumerReference);
    wsnt_SubscribeType_destroy_contents(&monitor->subscribe);
    globus_service_engine_destroy(monitor->engine);
}

static
void
globus_l_subscription_destroy_cb(
    SubscriptionManagerService_client_handle_t   handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrl_DestroyResponseType *    DestroyResponse,
    SubscriptionManager_Destroy_fault_t fault_type,
    const xsd_any *                     fault)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) user_args;
    SubscriptionManagerService_client_destroy(handle);
    globus_l_signal_done(monitor, GLOBUS_SUCCESS);
}
 
/* called locked */
static
void
globus_l_destroy_subscription(
    globus_i_monitor_info_t *           monitor)
{
    globus_result_t                     result;
    wsrl_DestroyType                    destroy;
    SubscriptionManagerService_client_handle_t handle;
    globus_bool_t                       tls = GLOBUS_FALSE;

    if (strncmp(monitor->subscribe_epr->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    wsrl_DestroyType_init_contents(&destroy);
    result = SubscriptionManagerService_client_init(&handle,
        monitor->command_info->message_attr,
        tls 
            ? monitor->command_info->tls_handler_chain
            : monitor->command_info->secmess_handler_chain);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_init;
    }
    
    result = SubscriptionManager_Destroy_epr_register(
        handle,
        monitor->subscribe_epr,
        &destroy,
        globus_l_subscription_destroy_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_register;
    }
    
    wsrl_DestroyType_destroy_contents(&destroy);
    monitor->cb_count++;
    
    return;

error_register:
    SubscriptionManagerService_client_destroy(handle);
error_init:
    wsrl_DestroyType_destroy_contents(&destroy);
    return;
}

static
void
globus_l_update_poll_unreg_cb(
    void *                              arg)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) arg;
    
    globus_l_signal_done(monitor, GLOBUS_SUCCESS);
}

/* call locked */
static
void
globus_l_shutdown_monitor(
    globus_i_monitor_info_t *           monitor)
{
    globus_list_t *                     tmp;
    
    if(monitor->stream_count > 0 &&
        (monitor->my_state == GLOBUS_I_MY_STATE_STREAMING ||
            (monitor->my_state == GLOBUS_I_MY_STATE_STREAMING_FINISHING &&
                monitor->error)))
    {
        globus_i_streaming_stop(
            &monitor->streaming,
            monitor->error
                ? GLOBUS_I_MJG_STREAMING_FINISH
                : GLOBUS_I_MJG_STREAMING_ABORT, NULL);
    }
    
    if(monitor->job_handle)
    {
        ManagedJobService_client_operation_cancel(monitor->job_handle);
    }
    
    if(monitor->multijob_handle)
    {
        ManagedMultiJobService_client_operation_cancel(
            monitor->multijob_handle);
    }
    
    for(tmp = monitor->active_queries;
        !globus_list_empty(tmp);
        tmp = globus_list_rest(tmp))
    {
        ManagedExecutableJobService_client_handle_t job_handle;
        
        job_handle = (ManagedExecutableJobService_client_handle_t)
            globus_list_first(tmp);
        
        ManagedExecutableJobService_client_operation_cancel(job_handle);
    }
    
    if(monitor->subscribe_epr)
    {
        globus_l_destroy_subscription(monitor);
    }
    
    if(monitor->poll_handle)
    {
        globus_callback_unregister(
            monitor->poll_handle,
            globus_l_update_poll_unreg_cb,
            monitor,
            NULL);
    }
    
    globus_service_engine_register_stop(monitor->engine);
    
    monitor->my_state = GLOBUS_I_MY_STATE_CLEANUP;
}

static
void
globus_l_get_status_cb(
    ManagedJobService_client_handle_t   handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrp_GetMultipleResourcePropertiesResponseType * gmrpr,
    ManagedJobPortType_GetMultipleResourceProperties_fault_t fault_type,
    const xsd_any *                     fault)
{
    globus_i_monitor_info_t *           monitor;
    wsgram_StateEnumeration *           state = NULL;
    xsd_boolean *                       holding = NULL;
    xsd_int *                           exit_code = NULL;
    wsgram_FaultResourcePropertyType *  gram_fault = NULL;
    int                                 i;
    
    monitor = (globus_i_monitor_info_t *) user_args;
    
    globus_mutex_lock(&monitor->lock);
    monitor->update_pending = GLOBUS_FALSE;
    
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying job state", result);
        goto error;
    }
    
    for(i = 0; i < gmrpr->any.length; i++)
    {
        if(gmrpr->any.elements[i].any_info == 
            &ManagedJobPortType_state_rp_info ||
	    gmrpr->any.elements[i].any_info ==
	    &wsgram_state_info)
        {
            globus_assert(!state && "Unexpected duplicate entry");
            
            state = (wsgram_StateEnumeration *) gmrpr->any.elements[i].value;
        }
        else if(gmrpr->any.elements[i].any_info == 
            &ManagedJobPortType_holding_rp_info ||
	    gmrpr->any.elements[i].any_info ==
	    &wsgram_holding_info)
        {
            globus_assert(!holding && "Unexpected duplicate entry");
            
            holding = (xsd_boolean *) gmrpr->any.elements[i].value;
        }
        else if(gmrpr->any.elements[i].any_info == 
            &ManagedJobPortType_fault_rp_info ||
	    gmrpr->any.elements[i].any_info ==
	    &wsgram_fault_info)
        {
            globus_assert(!gram_fault && "Unexpected duplicate entry");
            
            gram_fault = (wsgram_FaultResourcePropertyType *)
                gmrpr->any.elements[i].value;
        }
        else if(gmrpr->any.elements[i].any_info == 
            &ManagedExecutableJobPortType_exitCode_rp_info ||
	    gmrpr->any.elements[i].any_info ==
	    &wsgram_exitCode_info)
        {
            globus_assert(!exit_code && "Unexpected duplicate entry");
            
            exit_code = (xsd_int *) gmrpr->any.elements[i].value;
        }
        else
        {
            result = GlobusLErrorInvalid("response",
                "expected only state, holding, fault and exitcode");
            goto error;
        }
    }
    
    if(!state || !holding)
    {
        result = GlobusLErrorInvalid("response", "expected state and holding");
        goto error;
    }
    
    globus_l_state_change(monitor, *state, *holding, gram_fault, exit_code);

error:
    if(result != GLOBUS_SUCCESS &&
        monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
    {
        monitor->error = globus_error_get(result);
        globus_l_shutdown_monitor(monitor);
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
}

/* call locked */
globus_result_t
globus_l_get_status(
    globus_i_monitor_info_t *           monitor)
{
    globus_result_t                     result;
    xsd_QName *                         qname;
    wsrp_GetMultipleResourcePropertiesType gmrp;
    
    wsrp_GetMultipleResourcePropertiesType_init_contents(&gmrp);
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        result = GlobusLErrorWrapResult(
            "querying job state", GlobusLErrorMemory("qname entry"));
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname, &ManagedJobPortType_state_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying job state", result);
        goto error;
    }
    
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        result = GlobusLErrorWrapResult(
            "querying job state", GlobusLErrorMemory("qname entry"));
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname, &ManagedJobPortType_holding_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying job state", result);
        goto error;
    }
    
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        result = GlobusLErrorWrapResult(
            "querying job state", GlobusLErrorMemory("qname entry"));
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname, &ManagedJobPortType_fault_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying job state", result);
        goto error;
    }
    
    if(strstr(
        monitor->command_info->job_epr->Address.base_value,
        MANAGEDEXECUTABLEJOBSERVICE_BASE_PATH))
    {
        /* this probably isnt allowed within the MJS context, but it works */
        qname = xsd_QName_array_push(&gmrp.ResourceProperty);
        if(!qname)
        {
            result = GlobusLErrorWrapResult(
                "querying job state", GlobusLErrorMemory("qname entry"));
            goto error;
        }
        
        result = xsd_QName_copy_contents(
            qname, &ManagedExecutableJobPortType_exitCode_rp_qname);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult("querying job state", result);
            goto error;
        }
    }

    result = ManagedJobPortType_GetMultipleResourceProperties_epr_register(
        monitor->job_handle,
        monitor->command_info->job_epr,
        &gmrp,
        globus_l_get_status_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("querying job state", result);
        goto error;
    }
    
    monitor->cb_count++;
    monitor->update_pending = GLOBUS_TRUE;
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);
    
    return GLOBUS_SUCCESS;

error:
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);
    return result;
}

static
globus_result_t
globus_l_file_attr_stack_init(
    globus_xio_stack_t *                stack,
    globus_xio_attr_t *                 attr)
{
    globus_result_t                     result;
    
    result = globus_xio_stack_init(stack, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_stack;
    }
    
    result = globus_xio_stack_push_driver(*stack, globus_l_file_driver);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_push;
    }
    
    result = globus_xio_attr_init(attr);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_attr;
    }
    
    result = globus_xio_attr_cntl(
        *attr,
        globus_l_file_driver,
        GLOBUS_XIO_FILE_SET_FLAGS,
        GLOBUS_XIO_FILE_CREAT|GLOBUS_XIO_FILE_WRONLY|GLOBUS_XIO_FILE_APPEND);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_setup;
    }

    return GLOBUS_SUCCESS;

error_setup:
    globus_xio_attr_destroy(*attr);
error_attr:
error_push:
    globus_xio_stack_destroy(*stack);
error_stack:
    return result;
}

static
void
globus_l_file_attr_stack_destroy(
    globus_xio_stack_t                  stack,
    globus_xio_attr_t                   attr)
{
    globus_xio_attr_destroy(attr);
    globus_xio_stack_destroy(stack);
}

static
globus_result_t
globus_l_open_file_handle(
    const char *                        file,
    globus_xio_handle_t *               handle,
    globus_xio_stack_t                  stack,
    globus_xio_attr_t                   attr)
{
    globus_result_t                     result;
    
    result = globus_xio_handle_create(handle, stack);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_create;
    }
    
    result = globus_xio_open(*handle, file, attr);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_open;
    }
    
    return GLOBUS_SUCCESS;

error_open:
    globus_xio_close(*handle, NULL);
error_create:
    return result;
}

static
globus_result_t
globus_l_initialize_streaming(
    globus_i_monitor_info_t *           monitor)
{
    globus_result_t                     result;
    globus_xio_stack_t                  xio_stack;
    globus_xio_attr_t                   xio_attr;
    globus_ftp_client_handleattr_t      ftp_handleattr;
    globus_ftp_client_plugin_t          debug_plugin;
    
    result = globus_xio_driver_load("file", &globus_l_file_driver);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("loading file driver", result);
        goto error_load;
    }
    
    result = globus_ftp_client_handleattr_init(&ftp_handleattr);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_attr_init;
    }
    
    if(monitor->command_info->debug)
    {
        result = globus_ftp_client_debug_plugin_init(
            &debug_plugin, stderr, "debug");
        if(result != GLOBUS_SUCCESS)
        {
            goto error_debug_init;
        }
        
        result = globus_ftp_client_handleattr_add_plugin(
            &ftp_handleattr, &debug_plugin);
        globus_ftp_client_debug_plugin_destroy(&debug_plugin);
        if(result != GLOBUS_SUCCESS)
        {
            goto error_debug_init;
        }
    }
    
    result = globus_i_streaming_init(&monitor->streaming, &ftp_handleattr,
        NULL, globus_l_streaming_cb, monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("initializing streaming", result);
        goto error_streaming_init;
    }
    
    result = globus_l_file_attr_stack_init(&xio_stack, &xio_attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("initializing stack/attr", result);
        goto error_stack_init;
    }
    
    result = globus_l_open_file_handle(
        monitor->command_info->stream_out_file 
            ? monitor->command_info->stream_out_file : "stdout://",
        &monitor->stdout_handle, xio_stack, xio_attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("opening stdout", result);
        goto error_open_out;
    }
    
    result = globus_l_open_file_handle(
        monitor->command_info->stream_err_file 
            ? monitor->command_info->stream_err_file : "stderr://",
        &monitor->stderr_handle, xio_stack, xio_attr);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("opening stderr", result);
        goto error_open_err;
    }
    
    globus_l_file_attr_stack_destroy(xio_stack, xio_attr);
    globus_ftp_client_handleattr_destroy(&ftp_handleattr);
     return GLOBUS_SUCCESS;
    
error_open_err:
    globus_xio_close(monitor->stdout_handle, NULL);
error_open_out:
    globus_l_file_attr_stack_destroy(xio_stack, xio_attr);
error_stack_init:
    globus_i_streaming_destroy(&monitor->streaming);
error_streaming_init:
error_debug_init:
    globus_ftp_client_handleattr_destroy(&ftp_handleattr);
error_attr_init:
    globus_xio_driver_unload(globus_l_file_driver);
error_load:
    return result;
}

static
void
globus_l_destroy_streaming(
    globus_i_monitor_info_t *           monitor)
{
    globus_xio_close(monitor->stderr_handle, NULL);
    globus_xio_close(monitor->stdout_handle, NULL);
    globus_i_streaming_destroy(&monitor->streaming);
    globus_xio_driver_unload(globus_l_file_driver);
}

globus_result_t
globus_i_monitor_init(
    globus_i_monitor_info_t *           monitor,
    globus_i_info_t *                   info)
{
    globus_result_t                     result;
    
    memset(monitor, 0, sizeof(*monitor));
    monitor->command_info = info;
    globus_mutex_init(&monitor->lock, NULL);
    globus_cond_init(&monitor->cond, NULL);
    
    if(info->streaming)
    {
        result = globus_l_initialize_streaming(monitor);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult("initializing streaming", result);
            goto error_streaming_init;
        }
    }

    /* Default to TLS */
    monitor->use_https = GLOBUS_TRUE;

    if (info->factory != NULL)
    {
        /* But change our mind if the factory URI is set */
        monitor->use_https = (strncmp(info->factory, "https:", 6) == 0);
    }
    else if (info->job_epr != NULL)
    {
        /* Or if the job epr is set (for -monitor) */
        monitor->use_https =
            (strncmp(info->job_epr->Address.base_value, "https:", 6) == 0);
    }
    
    result = globus_l_setup_notification_consumer(monitor);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "setting up notification consumer", result);
        goto error_notification;
    }
    
    return GLOBUS_SUCCESS;
    
error_notification:
    globus_l_destroy_streaming(monitor);
error_streaming_init:
    return result;
}

void
globus_i_monitor_destroy(
    globus_i_monitor_info_t *           monitor)
{
    if(monitor->subscribe_epr)
    {
        wsa_EndpointReferenceType_destroy(monitor->subscribe_epr);
    }
    globus_l_cleanup_notification_consumer(monitor);
    if(monitor->command_info->streaming)
    {
        globus_l_destroy_streaming(monitor);
    }
    globus_cond_destroy(&monitor->cond);
    globus_mutex_destroy(&monitor->lock);
}

static
void
globus_l_cancel_cb(
    void *                              arg)
{
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) arg;
    
    globus_mutex_lock(&monitor->lock);
    {
        if(monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
        {
            monitor->error = GlobusLErrorCanceledObj();
            globus_l_shutdown_monitor(monitor);
        }
        else if(monitor->stream_count > 0)
        {
            globus_i_streaming_stop(
                &monitor->streaming, GLOBUS_I_MJG_STREAMING_ABORT, NULL);
        }
    }
    globus_mutex_unlock(&monitor->lock);
}

static
void
globus_l_update_poll_cb(
    void *                              arg)
{
    globus_result_t                     result;
    globus_i_monitor_info_t *           monitor;
    
    monitor = (globus_i_monitor_info_t *) arg;
    
    globus_mutex_lock(&monitor->lock);
    {
        if(monitor->is_idle && !monitor->update_pending &&
            monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
        {
            result = globus_l_get_status(monitor);
            if(result != GLOBUS_SUCCESS)
            {
                monitor->error = globus_error_get(result);
                globus_l_shutdown_monitor(monitor);
            }
        }
        else
        {
            monitor->is_idle = GLOBUS_TRUE;
        }
    }
    globus_mutex_unlock(&monitor->lock);
}

globus_result_t
globus_i_monitor(
    globus_i_monitor_info_t *           monitor)
{
    ManagedJobPortType_Subscribe_fault_t fault_type;
    xsd_any *                           fault;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_bool_t                       update_state = GLOBUS_FALSE;
    globus_bool_t                       tls = GLOBUS_FALSE;

    if (strncmp(monitor->command_info->job_epr->Address.base_value,
                "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    result = ManagedJobService_client_init(&monitor->job_handle,
        monitor->command_info->message_attr,
        tls
            ? monitor->command_info->tls_handler_chain
            : monitor->command_info->secmess_handler_chain);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("initializing MJS client", result);
        goto error_client;
    }
    
    if(globus_i_set_cancel_notify(globus_l_cancel_cb, monitor))
    {
        result = GlobusLErrorCanceled();
        goto error_canceled;
    }
    
    if(!monitor->subscribe_epr)
    {
        wsnt_SubscribeResponseType *    subscribe_response;
        
        result = ManagedJobPortType_Subscribe_epr(
            monitor->job_handle,
            monitor->command_info->job_epr,
            &monitor->subscribe,
            &subscribe_response,
            &fault_type,
            &fault);
        if(result != GLOBUS_SUCCESS)
        {
            if(fault_type != GLOBUS_SUCCESS)
            {
                xsd_any_destroy(fault);
            }
            result = GlobusLErrorWrapResult("subscribing", result);
            goto error_subscribe;
        }
    
        globus_mutex_lock(&monitor->lock);
        {
            monitor->subscribe_epr = subscribe_response->SubscriptionReference;
            subscribe_response->SubscriptionReference = NULL;
            
            /* if we were canceled, the shutdown wouldnt have had a chance to
             * destroy the subscription... do it now
             */
            if(monitor->my_state == GLOBUS_I_MY_STATE_CLEANUP)
            {
                globus_l_destroy_subscription(monitor);
            }
        }
        globus_mutex_unlock(&monitor->lock);
        
        wsnt_SubscribeResponseType_destroy(subscribe_response);
        update_state = GLOBUS_TRUE;
    }
    
    globus_mutex_lock(&monitor->lock);
    {
        result = globus_service_engine_register_start(
            monitor->engine, globus_l_service_engine_done_cb, monitor);
        if(result == GLOBUS_SUCCESS)
        {
            monitor->cb_count++;
        }
        else
        {
            result = GlobusLErrorWrapResult("starting service engine", result);
            globus_l_destroy_subscription(monitor);
        }

        if(result == GLOBUS_SUCCESS && update_state)
        {
            /* get current state just in case I missed any notifications */
            result = globus_l_get_status(monitor);
        }
        
        if(result == GLOBUS_SUCCESS)
        {
            globus_reltime_t            period;
            
            GlobusTimeReltimeSet(period, GLOBUS_I_STATE_UPDATE_POLL, 0);
            result = globus_callback_register_periodic(
                &monitor->poll_handle,
                &period,
                &period,
                globus_l_update_poll_cb,
                monitor);
            if(result == GLOBUS_SUCCESS)
            {
                monitor->cb_count++;
            }
        }
        
        if(result != GLOBUS_SUCCESS &&
            monitor->my_state != GLOBUS_I_MY_STATE_CLEANUP)
        {
            monitor->error = globus_error_get(result);
            globus_l_shutdown_monitor(monitor);
        }
        
        while(monitor->cb_count > 0)
        {
            globus_cond_wait(&monitor->cond, &monitor->lock);
        }
        
        if(monitor->error)
        {
            result = globus_error_put(monitor->error);
        }
    }
    globus_mutex_unlock(&monitor->lock);
    
error_subscribe:
    globus_i_set_cancel_notify(NULL, NULL);
error_canceled:
    ManagedJobService_client_destroy(monitor->job_handle);
error_client:
    return result;
}
