/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.test.throughput;

import java.net.URL;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.exec.client.GramJob;
import org.globus.exec.client.GramJobListener;
import org.globus.exec.utils.FaultUtils;
import org.globus.exec.utils.client.ManagedJobFactoryClientHelper;

import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.DeleteType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;

import org.globus.gram.GramException;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.PerformanceLog;

import org.ietf.jgss.GSSException;

/**
 *  Description of the Class
 *
 *@author     lane
 *@author     madduri
 *@created    July 6, 2004
 */
public class ClientThread
         implements Runnable,
        GramJobListener {

    static Log logger = LogFactory.getLog( ClientThread.class.getName() );
    static int instanceCount = 0;
    static Object InstanceCountMonitor = new Object();
    static UUIDGen uuidgen = UUIDGenFactory.getUUIDGen();

    ThroughputTester harness = null;
    int clientIndex = -1;
    boolean started = false;
    boolean completed = false;
    Vector jobs = null;
    Hashtable jobPollDelays = null;
    Hashtable jobPollCounters = null;
    int createdCount = 0;
    int completedCount = 0;

    PerformanceLog perfLog = new PerformanceLog(
            ClientThread.class.getName() );


    /**
     *  Constructor for the ClientThread object
     *
     *@param  harness  Description of the Parameter
     */
    public ClientThread( ThroughputTester harness ) {
        synchronized ( ClientThread.InstanceCountMonitor ) {
            this.clientIndex = ClientThread.instanceCount;
            ClientThread.instanceCount++;
        }

        this.harness = harness;
    }


    /**
     *  Gets the index attribute of the ClientThread object
     *
     *@return    The index value
     */
    public int getIndex() {
        return this.clientIndex;
    }


    // This method is invoked once each time the stress target is run
    // with this class
    /**
     *  Main processing method for the ClientThread object
     */
    public synchronized void run() {
        if ( logger.isDebugEnabled() ) {
            logger.debug( "running client thread #" + this.clientIndex );
        }

        int load = this.harness.getLoad();
        logger.debug("Accessing jobs vector.");
        this.jobs = new Vector(load);
        this.jobPollDelays = new Hashtable(load);
        this.jobPollCounters = new Hashtable(load);

        //start timing acutal duration
        if ( logger.isDebugEnabled() ) {
            logger.debug( "perf log start [finished]" );
        }
        this.perfLog.start();

        boolean polling = this.harness.isPolling();

        //maintain load
        boolean poll = false;
        while ( !this.harness.isDurationElapsed() )
        {
            if ( logger.isDebugEnabled() ) {
                logger.debug("# created (outer loop): " + this.createdCount);
                logger.debug("# completed (outer loop): " +this.completedCount);
            }
            //use cached created count so we don't get stuck in this next loop
            int tmpCompletedCount = this.completedCount;
            while ( ( this.createdCount - tmpCompletedCount ) < load )
            {
                if ( logger.isDebugEnabled() ) {
                    logger.debug( "# created: " + this.createdCount );
                    logger.debug( "# completed: " + this.completedCount );
                }
                try {
                    addJob();
                } catch ( Exception e ) {
                    logger.error( "failed to add job", e );
                    if (this.createdCount > 0)
                    {
                        this.createdCount--;
                    }
                }
            }

            try {
                if (polling)
                {
                    poll = true;
                    wait(2000);
                }
                else
                {
                    long startTime = System.currentTimeMillis();
                    wait(15000);
                    if ((System.currentTimeMillis()-startTime) < 15000)
                    {
                        //don't poll if we were signalled by a state change
                        poll = false;
                    }
                    else
                    {
                        //poll since we haven't been signalled in 15 seconds
                        poll = true;
                    }
                }
            } catch ( InterruptedException ie ) {}

            if (poll)
            {
                logger.debug("Accessing jobs vector.");
                logger.debug("Refreshing " + this.jobs.size() + " jobs.");
                logger.debug("Accessing jobs vector.");
                synchronized (this.jobs)
                {
                    //Not using iterator because it complains of concurrent
                    //modifications
                    for (int index=0; index<this.jobs.size(); index++)
                    {
                        GramJob job = (GramJob) this.jobs.elementAt(index);
                        Integer count = (Integer) this.jobPollCounters.get(job);
                        Integer delay = (Integer) this.jobPollDelays.get(job);
                        if (count.equals(delay))
                        {
                            if (polling)
                            {
                                //reset counter
                                this.jobPollCounters.put(job, new Integer(0));
                                logger.info(
                                    "\nResetting poll counter to "
                                    +(Integer) this.jobPollCounters.get(job));
                                //exponentially backoff the poll delay
                                this.jobPollDelays.put(
                                    job, new Integer(delay.intValue() * 2));
                                logger.info(
                                    "\nBacking off poll delay to "
                                    +(Integer) this.jobPollDelays.get(job));
                            }
                            try
                            {
                                job.refreshStatus();
                            } catch (Exception e) {
                                logger.error("Unable to refresh job.", e);
                            }
                        }
                        else if (polling)
                        {
                            this.jobPollCounters.put(
                                job, new Integer(count.intValue() + 1));
                            logger.info(
                                "\nIncrementing poll counter to "
                                +(Integer) this.jobPollCounters.get(job));
                        }
                    }
                }
            }
        }

        logger.debug( "waiting for any outstanding jobs to finish..." );

        //wait for jobs to finish
        int pollCounter = 0;
        while ( this.completedCount < this.createdCount ) {
            logger.debug( "Created: " + this.createdCount );
            logger.debug( "Completed: " + this.completedCount );
            try {
                wait( 2000 );
            } catch ( InterruptedException ie ) {}

            if ((pollCounter % 8) == 0)
            {
                logger.debug("Accessing jobs vector.");
                logger.debug("Refreshing " + this.jobs.size() + " jobs.");
                logger.debug("Accessing jobs vector.");
                synchronized (this.jobs)
                {
                    //Not using iterator because it complains of concurrent
                    //modifications
                    for (int index=0; index<this.jobs.size(); index++)
                    {
                        GramJob job = (GramJob) this.jobs.elementAt(index);
                        try
                        {
                            job.refreshStatus();
                        } catch (Exception e) {
                            logger.warn("Unable to refresh job.", e);
                        }
                    }
                }
                logger.debug("Done refreshing jobs.");
            }

            pollCounter++;
        }

        //stop timing actual duration
        this.perfLog.stop( "finished" );

        //pass total number of jobs completed by this client back to harness
        this.harness.notifyCompleted( this.clientIndex, this.completedCount );
    }


    protected void resolveJobIdVariable(
        JobDescriptionType                  jobDescription,
        String                              jobId)
    {
        //resolve the ${THROUGHPUT_TESTER_JOB_ID} variable

        String directory = jobDescription.getDirectory();
        logger.debug("Original directory string: " + directory);
        if (directory != null)
        {
            directory = directory.replaceAll(
                "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                jobId);
            jobDescription.setDirectory(directory);

            logger.debug("New directory string: " + directory);
        }

        String stdout = jobDescription.getStdout();
        logger.debug("Original stdout string: " + stdout);
        if (stdout != null)
        {
            stdout = stdout.replaceAll(
                "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                jobId);
            jobDescription.setStdout(stdout);

            logger.debug("New stdout string: " + stdout);
        }

        String stderr = jobDescription.getStderr();
        logger.debug("Original stderr string: " + stderr);
        if (stderr != null)
        {
            stderr = stderr.replaceAll(
                "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                jobId);
            jobDescription.setStderr(stderr);

            logger.debug("New stderr string: " + stderr);
        }

        String stdin = jobDescription.getStdin();
        logger.debug("Original stdin string: " + stdin);
        if (stdin != null)
        {
            stdin = stdin.replaceAll(
                "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                jobId);
            jobDescription.setStdin(stdin);

            logger.debug("New stdin string: " + stdin);
        }

        TransferRequestType fileStageIn = jobDescription.getFileStageIn();
        if (fileStageIn != null)
        {
            TransferType[] fileStageInTransfers = fileStageIn.getTransfer();
            for (int index=0; index<fileStageInTransfers.length; index++)
            {
                String sourceUrl
                    = fileStageInTransfers[index].getSourceUrl();
                sourceUrl = sourceUrl.replaceAll(
                    "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                    jobId);
                fileStageInTransfers[index].setSourceUrl(sourceUrl);

                logger.debug("New sourceUrl string: " + sourceUrl);

                String destinationUrl
                    = fileStageInTransfers[index].getDestinationUrl();
                destinationUrl = destinationUrl.replaceAll(
                    "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                    jobId);
                fileStageInTransfers[index].setDestinationUrl(destinationUrl);

                logger.debug("New destinationUrl string: " + destinationUrl);
            }
        }

        TransferRequestType fileStageOut = jobDescription.getFileStageOut();
        if (fileStageOut != null)
        {
            TransferType[] fileStageOutTransfers = fileStageOut.getTransfer();
            for (int index=0; index<fileStageOutTransfers.length; index++)
            {
                String sourceUrl
                    = fileStageOutTransfers[index].getSourceUrl();
                sourceUrl = sourceUrl.replaceAll(
                    "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                    jobId);
                fileStageOutTransfers[index].setSourceUrl(sourceUrl);

                logger.debug("New sourceUrl string: " + sourceUrl);

                String destinationUrl
                    = fileStageOutTransfers[index].getDestinationUrl();
                destinationUrl = destinationUrl.replaceAll(
                    "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                    jobId);
                fileStageOutTransfers[index].setDestinationUrl(destinationUrl);

                logger.debug("New destinationUrl string: " + destinationUrl);
            }
        }

        DeleteRequestType fileCleanUp = jobDescription.getFileCleanUp();
        if (fileCleanUp != null)
        {
            DeleteType[] deletions = fileCleanUp.getDeletion();
            for (int index=0; index<deletions.length; index++)
            {
                String file
                    = deletions[index].getFile();
                file = file.replaceAll(
                    "\\$\\{THROUGHPUT_TESTER_JOB_ID\\}",
                    jobId);
                deletions[index].setFile(file);

                logger.debug("New file string: " + file);
            }
        }
    }


    /**
     *  Adds a feature to the Job attribute of the ClientThread object
     *
     *@exception  GramException  Description of the Exception
     *@exception  GSSException   Description of the Exception
     *@exception  Exception      Description of the Exception
     */
    protected void addJob()
        throws                              GramException,
                                            GSSException,
                                            Exception
    {
        String uuid =  null;
        synchronized (uuidgen)
        {
            uuid =  uuidgen.nextUUID();
        }
        String jobId = "uuid:" + uuid;

        logger.warn("Next UUID: " + uuid);

        JobDescriptionType jobDescription
            = (JobDescriptionType) ObjectSerializer.clone(
                harness.getJobDescription());
        resolveJobIdVariable(jobDescription, uuid);

        //create job
        GramJob job = new GramJob(jobDescription);

        job.setDelegationEnabled(this.harness.isDelegationEnabled());

        job.addListener( this );

        URL factoryUrl = ManagedJobFactoryClientHelper.getServiceURL(
            this.harness.getContactString()).getURL();

        boolean polling = this.harness.isPolling();
        boolean batch = polling; //if polling then batch
        boolean limitedDelegation = true; //not used

        job.submit(ManagedJobFactoryClientHelper.getFactoryEndpoint(
            factoryUrl,
            this.harness.getFactoryType()),
            batch,
            limitedDelegation,
            jobId);

        logger.debug("Accessing jobs vector.");
        this.jobs.add(job);
        if (polling)
        {
            this.jobPollDelays.put(job, new Integer(1));
            this.jobPollCounters.put(job, new Integer(0));
        }
        else
        {
            this.jobPollDelays.put(job, new Integer(0));
            this.jobPollCounters.put(job, new Integer(0));
        }

        this.createdCount++;

        logger.info("\nAdded job to client #" + this.clientIndex + ". "
                   +"Total created for client: " + this.createdCount);

    }


    /**
     *  Description of the Method
     *
     *@param  job  Description of the Parameter
     */
    public void stateChanged(GramJob job)
    {
        StateEnumeration status = job.getState();
        if (logger.isInfoEnabled())
        {
            logger.info(
                "\nState changed for a job " + job.getID()
                + " in client" + this.clientIndex +": " + status.getValue());
        }

        if (status.equals( StateEnumeration.Done) )
        {
            logger.debug("Accessing jobs vector.");
            synchronized (this.jobs)
            {
                if (this.jobs.remove(job))
                {
                    this.jobPollDelays.remove(job);
                    this.jobPollCounters.remove(job);
                    try {
                        if (logger.isInfoEnabled())
                        {
                            logger.info("destroying Done job " + job.getID());
                        }

                        job.destroy();
                    }
                    catch (Exception e) {
                        logger.error( "unable to destroy job", e );
                    }
                    this.completedCount++;
                    logger.info("\nRemoving job from client #"
                               + this.clientIndex + ". "
                               +"Total completed for client: "
                               + this.completedCount);
                }
            }

            synchronized (this)
            {
                logger.trace("notifying client of Done state for job");
                notify();
            }
            logger.trace("notified client of Done state for job");
        }
        else if (status.equals(StateEnumeration.Failed))
        {
            logger.error("a job failed with ID " + job.getID() + ":\n"
                        + FaultUtils.faultToString(job.getFault()));
            logger.debug("Accessing jobs vector.");
            synchronized (this.jobs)
            {
                if (this.jobs.remove(job))
                {
                    this.jobPollDelays.remove(job);
                    this.jobPollCounters.remove(job);
                    try {
                        if (logger.isInfoEnabled())
                        {
                            logger.info("destroying Failed job " + job.getID());
                        }

                        job.destroy();
                    }
                    catch (Exception e) {
                        logger.error( "unable to destroy job", e );
                    }
                    if (this.createdCount > 0)
                    {
                        this.createdCount--;
                    }
                    logger.info("\nA job failed in client #" + this.clientIndex
                               + ". " +"Total created for client: "
                               + this.completedCount);
                }
            }

            synchronized (this)
            {
                notify();
            }
        }
        logger.debug("Done processing state change");
    }
}

