/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.usage;

import java.net.Inet6Address;
import java.nio.ReadOnlyBufferException;
import java.sql.Timestamp;
import java.util.Date;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.generated.JobTypeEnumeration;

import org.globus.usage.packets.CustomByteBuffer;
import org.globus.usage.packets.IPTimeMonitorPacket;

/*GRAM usage monitor packets, in addition to the fields in IPTimeMonitorPacket,
include the following:
    - job created timestamp
    - scheduler type (Fork, PBS, LSF, Condor, SGE, etc...)
    - jobCredentialEndpoint present flag (i.e. do they use server-side user
     proxies)
    - fileStageIn present flag
    - fileStageOut present flag
    - fileCleanUp present flag
    - CleanUpHold flag
    - job type (Single, Multiple, MPI, Condor)
    - gt2 error code if present and Failed
    - fault class name or identifier if Failed
*/
public class GramUsageMonitorPacket
    extends                                 IPTimeMonitorPacket
{
    static Log logger = LogFactory.getLog(GramUsageMonitorPacket.class);

    private static short COMPONENT_CODE = 1;
    private static short PACKET_VERSION = 1;

    private static byte FALSE   = 0;
    private static byte TRUE    = 1;

    //job created timestamp
    private Date creationTime;

    //scheduler type
    private final static int MAX_SCHEDULER_TYPE_SIZE = 20;
    private String localResourceManager;

    //jobCredentialEndpoint present flag
    private boolean jobCredentialEndpointUsed;

    //fileStageIn present flag
    private boolean fileStageInUsed;

    //fileStageOut present flag
    private boolean fileStageOutUsed;

    //fileCleanUp present flag
    private boolean fileCleanUpUsed;

    //CleanUpHold flag
    private boolean cleanUpHoldUsed;

    //job type
    private static final byte JOB_TYPE_UNKNOWN   = 0;
    private static final byte JOB_TYPE_SINGLE    = 1;
    private static final byte JOB_TYPE_MULTIPLE  = 2;
    private static final byte JOB_TYPE_MPI       = 3;
    private static final byte JOB_TYPE_CONDOR    = 4;
    private JobTypeEnumeration jobType;

    //gt2 error
    private int gt2ErrorCode;

    //fault class name
    private static final byte FAULT_CLASS_UNKNOWN                   = 0;
    private static final byte FAULT_CLASS_CREDENTIAL_SERIALIZATION  = 1;
    private static final byte FAULT_CLASS_EXECUTION_FAILED          = 2;
    private static final byte FAULT_CLASS_FAULT                     = 3;
    private static final byte FAULT_CLASS_FILE_PERMISSIONS          = 4;
    private static final byte FAULT_CLASS_INSUFFICIENT_CREDENTIALS  = 5;
    private static final byte FAULT_CLASS_INTERNAL                  = 6;
    private static final byte FAULT_CLASS_INVALID_CREDENTIALS       = 7;
    private static final byte FAULT_CLASS_INVALID_PATH              = 8;
    private static final byte FAULT_CLASS_SERVICE_LEVEL_AGREEMENT   = 9;
    private static final byte FAULT_CLASS_STAGING                   = 10;
    private static final byte FAULT_CLASS_UNSUPPORTED_FEATURE       = 11;
    private Class faultClass;

    public GramUsageMonitorPacket()
    {
        setComponentCode(COMPONENT_CODE);
        setPacketVersion(PACKET_VERSION);
    }

    public void setCreationTime(Date creationTime)
    {
        this.creationTime = creationTime;
    }

    public Date getCreationTime() {
        return this.creationTime;
    }

    public void setLocalResourceManager(String localResourceManager)
    {
        this.localResourceManager = localResourceManager;
    }

    public String getLocalResourceManager() {
        return this.localResourceManager;
    }

    public void setJobCredentialEndpointUsed(boolean jobCredentialEndpointUsed)
    {
        this.jobCredentialEndpointUsed = jobCredentialEndpointUsed;
    }

    public boolean getJobCredentialEndpointUsed() {
        return this.jobCredentialEndpointUsed;
    }

    public void setFileStageInUsed(boolean FileStageInUsed)
    {
        this.fileStageInUsed = FileStageInUsed;
    }

    public boolean isFileStageInUsed() {
        return this.fileStageInUsed;
    }

    public void setFileStageOutUsed(boolean FileStageOutUsed)
    {
        this.fileStageOutUsed = FileStageOutUsed;
    }

    public boolean isFileStageOutUsed() {
        return this.fileStageOutUsed;
    }

    public void setFileCleanUpUsed(boolean FileCleanUpUsed)
    {
        this.fileCleanUpUsed = FileCleanUpUsed;
    }

    public boolean isFileCleanUpUsed() {
        return this.fileCleanUpUsed;
    }

    public void setCleanUpHoldUsed(boolean CleanUpHoldUsed)
    {
        this.cleanUpHoldUsed = CleanUpHoldUsed;
    }

    public boolean isCleanUpHoldUsed() {
        return this.cleanUpHoldUsed;
    }

    public void setJobType(JobTypeEnumeration jobType)
    {
        this.jobType = jobType;
    }

    public JobTypeEnumeration getJobType() {
        return this.jobType;
    }

    public void setGt2ErrorCode(int gt2ErrorCode)
    {
        this.gt2ErrorCode = gt2ErrorCode;
    }

    public int getGt2ErrorCode() {
        return this.gt2ErrorCode;
    }

    public void setFaultClass(Class faultClass)
    {
        this.faultClass = faultClass;
    }

    public Class getFaultClass() {
        return this.faultClass;
    }

    private short getIPVersion()
    {
        if (this.senderAddress instanceof Inet6Address)
            return 6;
        else
            return 4;
    }

    //TODO update everything bellow

    public void packCustomFields(CustomByteBuffer buf)
    {
        super.packCustomFields(buf);

        //creationTime
        buf.putLong(creationTime.getTime());

        //localResourceManager
        int localResourceManagerActualLength = localResourceManager.length();
        String localResourceManagerFixedLength = null;
        if (localResourceManagerActualLength > MAX_SCHEDULER_TYPE_SIZE)
        {
            //truncate localResourceManager string
            localResourceManagerFixedLength
                = localResourceManager.substring(0, MAX_SCHEDULER_TYPE_SIZE);
        }
        else if (localResourceManagerActualLength < MAX_SCHEDULER_TYPE_SIZE)
        {
            //pad localResourceManager string
            localResourceManagerFixedLength
                = localResourceManager
                + new char[MAX_SCHEDULER_TYPE_SIZE - localResourceManagerActualLength];
        }
        else
        {
            //do nothing to localResourceManager string
            localResourceManagerFixedLength = localResourceManager;
        }
        byte[] localResourceManagerFixedBytes = localResourceManagerFixedLength.getBytes();
        buf.put(localResourceManagerFixedBytes);

        //jobCredentialEndpointUsed
        buf.put(this.jobCredentialEndpointUsed?TRUE:FALSE);

        //FileStageInUsed
        buf.put(this.fileStageInUsed?TRUE:FALSE);

        //FileStageOutUsed
        buf.put(this.fileStageOutUsed?TRUE:FALSE);

        //FileCleanUpUsed
        buf.put(this.fileCleanUpUsed?TRUE:FALSE);

        //CleanUpHoldUsed
        buf.put(this.cleanUpHoldUsed?TRUE:FALSE);

        //jobType
        buf.put(jobTypeEnumerationToByte(this.jobType));

        //gt2ErrorCode
        buf.putInt(this.gt2ErrorCode);

        //faultClass
        buf.put(GramUsageMonitorPacket.faultClassToByte(this.faultClass));
    }
   
    public void unpackCustomFields(CustomByteBuffer buf)
    {
        super.unpackCustomFields(buf);

        //creationTime
        this.creationTime = new Date(buf.getLong());

        //localResourceManager
        byte[] localResourceManagerBytes = new byte[MAX_SCHEDULER_TYPE_SIZE];
        buf.get(localResourceManagerBytes);
        this.localResourceManager = new String(localResourceManagerBytes);

        //jobCredentialEndpointUsed
        this.jobCredentialEndpointUsed = (buf.get()==1?true:false);

        //FileStageInUsed
        this.fileStageInUsed = (buf.get()==1?true:false);

        //FileStageOutUsed
        this.fileStageOutUsed = (buf.get()==1?true:false);

        //FileCleanUpUsed
        this.fileCleanUpUsed = (buf.get()==1?true:false);

        //CleanUpHoldUsed
        this.cleanUpHoldUsed = (buf.get()==1?true:false);

        //jobType
        this.jobType = byteToJobTypeEnumeration(buf.get());

        //gt2ErrorCode
        this.gt2ErrorCode = buf.getInt();

        //faultClass
        this.faultClass = GramUsageMonitorPacket.byteToFaultClass(buf.get());
    }

    /**
     * Converts a GRAM job type to it's corresponding byte code.
     *
     * @returns The byte code associated with the given GRAM job type, or
     *          JOB_TYPE_UNKNOWN if the job type string isn't recognized.
     */
    private static byte jobTypeEnumerationToByte(JobTypeEnumeration jobType)
    {
        if (jobType == null)
        {
            return JOB_TYPE_UNKNOWN;
        }

        if (jobType.equals(JobTypeEnumeration.mpi))
        {
            return JOB_TYPE_MPI;
        } else
        if (jobType.equals(JobTypeEnumeration.single))
        {
            return JOB_TYPE_SINGLE;
        } else
        if (jobType.equals(JobTypeEnumeration.multiple))
        {
            return JOB_TYPE_MULTIPLE;
        } else
        if (jobType.equals(JobTypeEnumeration.condor))
        {
            return JOB_TYPE_CONDOR;
        } else
        {
            return JOB_TYPE_UNKNOWN;
        }
    }

    /**
     * Converts the byte code associated with the given GRAM job type
     * to the job type String representation.
     *
     * @returns The String representation of the job type assocated with the
     *          given byte code, or null if the byte code is JOB_TYPE_UNKONWN or
     *          unrecognized.
     */
    private static JobTypeEnumeration byteToJobTypeEnumeration(byte jobTypeByte)
    {
        switch (jobTypeByte)
        {
         case JOB_TYPE_UNKNOWN:
         {
            return null;
         }
         case JOB_TYPE_MPI:
         {
            return JobTypeEnumeration.mpi;
         }
         case JOB_TYPE_SINGLE:
         {
            return JobTypeEnumeration.single;
         }
         case JOB_TYPE_MULTIPLE:
         {
            return JobTypeEnumeration.multiple;
         }
         case JOB_TYPE_CONDOR:
         {
            return JobTypeEnumeration.condor;
         }
         default:
         {
            return null;
         }
        }
    }

    /**
     * Converts a GRAM fault Class object to it's associated byte code.
     *
     * @returns The byte code associated with the given GRAM fault class, or
     *          FAULT_CLASS_UNKOWN if the given Class object is not a
     *          recognized GRAM fault class.
     */
    private static byte faultClassToByte(Class faultClass)
    {
        if (faultClass == null)
        {
            return FAULT_CLASS_UNKNOWN;
        }

        if (faultClass.getName().equals(
                "org.globus.exec.generated.CredentialSerializationFaultType"))
        {
            return FAULT_CLASS_CREDENTIAL_SERIALIZATION;
        } else
        if (faultClass.getName().equals(
            "org.globus.exec.generated.ExecutionFailedFaultType"))
        {
            return FAULT_CLASS_EXECUTION_FAILED;
        } else
        if (faultClass.getName().equals(
            "org.globus.exec.generated.FaultType"))
        {
            return FAULT_CLASS_FAULT;
        } else
        if (faultClass.getName().equals(
            "org.globus.exec.generated.FilePermissionsFaultType"))
        {
            return FAULT_CLASS_FILE_PERMISSIONS;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.InsufficientCredentialsFaultType"))
        {
            return FAULT_CLASS_INSUFFICIENT_CREDENTIALS;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.InternalFaultType"))
        {
            return FAULT_CLASS_INTERNAL;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.InvalidCredentialsFaultType"))
        {
            return FAULT_CLASS_INVALID_CREDENTIALS;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.InvalidPathFaultType"))
        {
            return FAULT_CLASS_INVALID_PATH;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.ServiceLevelAgreementFaultType"))
        {
            return FAULT_CLASS_SERVICE_LEVEL_AGREEMENT;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.StagingFaultType"))
        {
            return FAULT_CLASS_STAGING;
        } else
        if (faultClass.getName().equals(
                "org.globus.exec.generated.UnsupportedFeatureFaultType"))
        {
            return FAULT_CLASS_UNSUPPORTED_FEATURE;
        } else
        {
            return FAULT_CLASS_UNKNOWN;
        }
    }

    /**
     * Converts the byte code associated with a GRAM fault class to a
     * class instance.
     *
     * @returns a GRAM fault Class object or null if a FAULT_CLASS_UNKNOWN
     *          or unrecognized byte was received.
     */
    private static Class byteToFaultClass(byte faultClassByte)
    {
        try {
            switch (faultClassByte)
            {
            case FAULT_CLASS_UNKNOWN:
            {
                return null;
            }
            case FAULT_CLASS_CREDENTIAL_SERIALIZATION:
            {
                return Class.forName(
                "org.globus.exec.generated.CredentialSerializationFaultType");
            }
            case FAULT_CLASS_EXECUTION_FAILED:
            {
                return Class.forName(
                "org.globus.exec.generated.ExecutionFailedFaultType");
            }
            case FAULT_CLASS_FAULT:
            {
                return Class.forName(
                "org.globus.exec.generated.FaultType");
            }
            case FAULT_CLASS_FILE_PERMISSIONS:
            {
                return Class.forName(
                "org.globus.exec.generated.FilePermissionsFaultType");
            }
            case FAULT_CLASS_INSUFFICIENT_CREDENTIALS:
            {
                return Class.forName(
                "org.globus.exec.generated.InsufficientCredentialsFaultType");
            }
            case FAULT_CLASS_INTERNAL:
            {
                return Class.forName(
                "org.globus.exec.generated.InternalFaultType");
            }
            case FAULT_CLASS_INVALID_CREDENTIALS:
            {
                return Class.forName(
                "org.globus.exec.generated.InvalidCredentialsFaultType");
            }
            case FAULT_CLASS_INVALID_PATH:
            {
                return Class.forName(
                "org.globus.exec.generated.InvalidPathFaultType");
            }
            case FAULT_CLASS_SERVICE_LEVEL_AGREEMENT:
            {
                return Class.forName(
                "org.globus.exec.generated.ServiceLevelAgreementFaultType");
            }
            case FAULT_CLASS_STAGING:
            {
                return Class.forName(
                "org.globus.exec.generated.StagingFaultType");
            }
            case FAULT_CLASS_UNSUPPORTED_FEATURE:
            {
                return Class.forName(
                "org.globus.exec.generated.UnsupportedFeatureFaultType");
            }
            default:
            {
                return null;
            }
            }
        } catch (Exception e)
        {
            return null;
        }
    }

    public void display()
    {
        logger.info(super.toString());

        logger.info("creationTime = "+this.creationTime);
        logger.info("localResourceManager = "+this.localResourceManager);
        logger.info("jobCredentialEndpointUsed = "
                    +this.jobCredentialEndpointUsed);
        logger.info("FileStageInUsed = "+fileStageInUsed);
        logger.info("FileStageOutUsed = "+fileStageOutUsed);
        logger.info("FileCleanUpUsed = "+fileCleanUpUsed);
        logger.info("CleanUpHoldUsed = "+cleanUpHoldUsed);
        logger.info("jobType = "+jobType);
        logger.info("gt2ErrorCode = "+gt2ErrorCode);
        logger.info("faultClass = "+faultClass);
    }

    public String toSQL()
    {
        StringBuffer sqlCommand = new StringBuffer("(");
        sqlCommand.append("creation_time,");
        sqlCommand.append("scheduler_type,");
        sqlCommand.append("job_credential_endpoint_used,");
        sqlCommand.append("stage_in_used,");
        sqlCommand.append("file_stage_out_used,");
        sqlCommand.append("file_clean_up_used,");
        sqlCommand.append("clean_up_hold_used,");
        sqlCommand.append("job_type,");
        sqlCommand.append("gt2_error_code,");
        sqlCommand.append("fault_class");
        sqlCommand.append(") VALUES ('");
        sqlCommand.append("'").append(this.creationTime).append("'");
        sqlCommand.append("'").append(this.localResourceManager).append("'");
        sqlCommand.append("'").append(this.jobCredentialEndpointUsed)
                  .append("'");
        sqlCommand.append("'").append(this.fileStageInUsed).append("'");
        sqlCommand.append("'").append(this.fileStageOutUsed).append("'");
        sqlCommand.append("'").append(this.fileCleanUpUsed).append("'");
        sqlCommand.append("'").append(this.cleanUpHoldUsed).append("'");
        sqlCommand.append("'").append(this.jobType).append("'");
        sqlCommand.append("'").append(this.gt2ErrorCode).append("'");
        sqlCommand.append("'").append(this.faultClass).append("'");
        sqlCommand.append("')");

        return sqlCommand.toString();
    }
}
