/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.index.impl;

import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.net.UnknownHostException;

import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.addressing.ReferencePropertiesType;

import javax.xml.namespace.QName;

import org.apache.axis.utils.XMLUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;
import org.globus.mds.aggregator.impl.AggregatorUtils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.InputStream;
import java.io.FileInputStream;
import java.net.InetAddress;

import org.globus.wsrf.encoding.ObjectDeserializer;

import org.globus.mds.index.SimpleRegConfig;


/**
 * Default index service implementation. 
*/

public class DefaultIndexService extends IndexService
{

    static final Log logger = LogFactory.getLog(DefaultIndexService.class);

    private Hashtable upstreamTable = new Hashtable();
    private Hashtable downstreamTable = new Hashtable();
    
    private String configDirPath;
    private String registrationFilePath;

    // these are paths to XML serialized registration 
    // parameter blocks (with servicegroup and registrant EPRs
    // left to be populated by the below code)
    private String skeletonUpstreamRegistrationPath;
    private String skeletonDownstreamRegistrationPath;
    
    private EndpointReferenceType localEpr = null;
    
    public DefaultIndexService()
    {
        String baseDir = getBaseDirectory();
        this.configDirPath = baseDir +  "/etc/globus_wsrf_mds_index/";
        this.registrationFilePath = this.configDirPath + "hierarchy.xml";
        this.skeletonUpstreamRegistrationPath = this.configDirPath + 
            "upstream.xml";
        this.skeletonDownstreamRegistrationPath = this.configDirPath + 
            "downstream.xml";    

        // process the config file
        processConfigFile();
        // perform default upstream and downstream registrations
        performDefaultRegistrations();

    }

    private static String getBaseDirectory() 
    {
        return ContainerConfig.getBaseDirectory();
    }

    /* wait a little while after making registrations so that it is
       more likely that everything will have started up. */
    static final int REG_DELAY = 8000;

    void processConfigFile() 
    {
        try {
          // get the local EPR
          localEpr = AddressingUtils.createEndpointReference(null);             
        } catch(Exception e) {
          logger.error("Fatal error when getting local service EPR: "+e);
          return;
        }  
  
        logger.info("Reading default registration configuration from file: " +
                    registrationFilePath);

            
        SimpleRegConfig config;
        try {
            config = readConfig(this.registrationFilePath);
        } catch(Exception e) {
            logger.warn("Could not read default index config: "+e);
            return;
        }
        
        String[] downstream = config.getDownstream();
        for(int i=0; downstream != null && i<downstream.length; i++)
        {            
            String entry = downstream[i].trim();
            downstreamTable.put(entry, entry);
        }

        String[] upstream = config.getUpstream();
        for(int i=0; upstream!=null && i<upstream.length; i++)
        {
            String entry = upstream[i].trim();
            upstreamTable.put(entry, entry);
        }
    }
    
    void performDefaultRegistrations()
    { 
         
        // perform the registrations
        // perform the downstream registrations first, and delay longer
        // when making the upstream registrations so that the local index
        // has a greater chance of being populated meaningfully when the
        // upstream indexes probe.

        ServiceGroupRegistrationClient client = 
            ServiceGroupRegistrationClient.getContainerClient();
        
        //     registrant EPR = from config file
        //     servicegroup EPR = leave null so inherit from container client
        String downstreamEntry = null;
        Enumeration downstreamEntries = downstreamTable.elements();
        while (downstreamEntries.hasMoreElements())
        {
            downstreamEntry = (String)downstreamEntries.nextElement();
            logger.info("Processing downstream registration to "+ downstreamEntry);

            try {
                ServiceGroupRegistrationParameters params = 
                    ServiceGroupRegistrationClient.readParams(this.skeletonDownstreamRegistrationPath);

                // form an EPR from the supplied URL
                EndpointReferenceType epr = new EndpointReferenceType(); 
                epr.setAddress(new Address(downstreamEntry));
                
                // do some basic sanity checks
                String dup = (String)upstreamTable.get(downstreamEntry);
                if (dup != null) {
                    logger.warn("Check configuration - a matching upstream registration was found for a downstream registration for this address: " +
                                downstreamEntry);
                }
                if (AggregatorUtils.detectLoopback(epr,localEpr)) { 
                    logger.warn("Check configuration - loopback condition detected in downstream registration - check configuration: " + 
                                "Registrant EPR was: [" + epr.getAddress() +
                                "] Local EPR was: [" + localEpr.getAddress() + "]"); 
                } 
                
                // set into reg block
                params.setRegistrantEPR(epr);


                client.register(params, REG_DELAY);
            } catch(Exception e) {
                logger.error("When registering a downstream service: "+e);
            }

        }

        //     registrant EPR = this index
        //     servicegroup EPR = from config file
        String upstreamEntry = null;
        Enumeration upstreamEntries = upstreamTable.elements();
        while (upstreamEntries.hasMoreElements())
        {
            upstreamEntry = (String)upstreamEntries.nextElement();
            logger.info("Processing upstream registration to "+ upstreamEntry);
            
            try {
                ServiceGroupRegistrationParameters params = 
                    ServiceGroupRegistrationClient.readParams(this.skeletonUpstreamRegistrationPath);

                params.setRegistrantEPR(null);

                EndpointReferenceType epr = new EndpointReferenceType(); 
                epr.setAddress(new Address(upstreamEntry));
                
                // do some basic sanity checks                
                String dup = (String)downstreamTable.get(upstreamEntry);
                if (dup != null) {
                    logger.warn("Check configuration - a matching downstream registration was found for a upstream registration for this address: " +
                                upstreamEntry);
                }                
                if (AggregatorUtils.detectLoopback(epr,localEpr)) { 
                    logger.warn("Check configuration - loopback condition detected in upstream registration - check configuration: " + 
                                "Registrant EPR was: [" + epr.getAddress() +
                                "] Local EPR was: [" + localEpr.getAddress() + "]"); 
                }        
                params.setServiceGroupEPR(epr);

                client.register(params, 2*REG_DELAY);
            } catch(Exception e) {
                logger.error("When registering to an upstream index: "+e);
            }
        }
    }

    SimpleRegConfig readConfig(String filename) 
        throws Exception
    {
        InputStream inputStream = null;
        try {
            inputStream = new FileInputStream(filename);
            Document doc = XMLUtils.newDocument(inputStream);

            SimpleRegConfig config = 
                (SimpleRegConfig) 
                ObjectDeserializer.toObject(doc.getDocumentElement(),
                                            SimpleRegConfig.class);
            return config;
        } finally {
            if (inputStream != null) {
                inputStream.close(); 
            }
        }
    }

}
