/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/

package org.globus.mds.usefulrp;

import java.util.Date;
import java.util.Calendar;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;
import org.globus.wsrf.ResourceProperty;


import org.globus.wsrf.impl.ReflectionResourceProperty;

import org.globus.mds.usefulrp.types.Accumulator;
import org.globus.mds.usefulrp.types.ExponentialRate;

/** An MDS ResourceProperty that will allow services to easily publish
    metrics about their load. <br>
    This RP is specfically designed for measuring cumulative quantities
    (such as number of hits, number of bytes transferred), with the
    TODO ability to publish derived statistics such as average rate of 
    change of the accumulator (which will provide, for example, 
    hits/second or bytes/second).
    <br>
    Note that this RP implements no persistence (in order to have minimal 
    dependencies on other configuration) but will TODO indicate in its
    output when it began making the present set of measurements.
        
*/

public class AccumulatorResourceProperty extends ReflectionResourceProperty {

    private static Log logger =
            LogFactory.getLog(AccumulatorResourceProperty.class.getName());


    /** The stub type that will be serialised to go over the wire. */
    Accumulator publishedValue = new Accumulator(); 

    /**  
    @throws java.lang.Exception passes along any exception encountered
        during construction of underlying ReflectionResourceProperty.
    */
    public AccumulatorResourceProperty(QName name) throws Exception {
        super(new SimpleResourcePropertyMetaData(name));
        setObject(this);
        setPropertyName("DynamicValue");
        initialize();

        // set the start time to now.
        publishedValue.setStartTime(Calendar.getInstance());
        publishedValue.setLastChange(Calendar.getInstance());

        // configure the exponential rate elements
        // 1 min
        ExponentialRate[] rates = new ExponentialRate[] {
            new ExponentialRate(0,60), // 1min
            new ExponentialRate(0,3600), // 1hour
            new ExponentialRate(0,86400) // 1day
        };
        publishedValue.setExpRate(rates);

    }

    public Object getDynamicValue() {
        logger.debug("Performing decay calc. to zero.");
        performDecays(0);
        return publishedValue;
    }

    public void accumulate(long amount)
    {
        logger.debug("Accumulating "+amount);
        long total = publishedValue.getTotal();
        total+=amount;
        publishedValue.setTotal(total);
        publishedValue.setLastChange(Calendar.getInstance());
        performDecays(amount);
    }

    /** keeps track of when we last performed decays. This is different
        from the output schema lastChange value. */
    Date lastCalculation = new Date(); 

    void performDecays(double q)
    {
        ExponentialRate[] rates = publishedValue.getExpRate();
        Date now = new Date();
        for(int i=0; i<rates.length; i++) {
            ExponentialRate rate = rates[i];

            double decayPeriod=(double) rate.getDecay();
            double decayConstant = 1.0 - Math.exp(-1.0/(decayPeriod));

            // TODO finish this

            double decayPerUnitTime = (1.0-decayConstant);


            double deltaTms = (now.getTime() - lastCalculation.getTime());

            // change to seconds as our time unit.
            double deltaTs = deltaTms / 1000.0;

            double decayForThisReading = Math.pow(1.0-decayConstant,deltaTs);

            // decayForThisReading + decayConstant <= 1
            if(decayForThisReading + decayConstant > 1.0)
            {
                logger.debug("Decay constants sum to "+(decayForThisReading+decayConstant)+" when they should sum to <= 1.0");
            }

            rate.setRate(rate.getRate()*decayForThisReading 
                                   + q * decayConstant);

        }
        lastCalculation = now;
    }
}


