/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.mds.webmds;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;
import java.util.Enumeration;
import java.util.Set;
import java.util.StringTokenizer;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletConfig;
import org.globus.mds.webmds.WebmdsXmlSource;
import org.globus.mds.webmds.WebmdsUtil;
import org.globus.mds.webmds.ParameterMap;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;
import org.xml.sax.InputSource;
import org.apache.xalan.transformer.TransformerIdentityImpl;
import org.globus.mds.webmdsConfig.WebmdsConfigType;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.DeserializationException;


/**
 * Webmds servlet -- acquires some XML and applies some XSL to it.
 * Several servlet configuration parameters control where the XML and
 * XSL come from.
 * <dl>
 * <dt><b>xmlSources</b></dt>
 * <dd>a space-separated list of <i>sourceNames</i> -- short
 * names of sources of XML.  These names will appear in other configuration
 * parameters.</dd>
 * <dt><b>info</b></dt>
 * <dd>A <i>sourceName</i> specifying where the information
 * to be displayed should come from.</dd>
 * <dt><b>xsl</b></dt>
 * <dd>A <i>sourceName</i> specifying where the xml transform
 * (that will be applied to the information) will come from.</dd>
 * <dt><b>xmlSource.</b><i>sourceName</i><b>.class</b></dt>
 * <dd> the name a class that implements the WebmdsXmlSource interface.
 * This class will be to be used to acquire XML if <i>sourceName</i>
 * is specified as an <b>info</b> or <b>xsl</b> parameter</dd>
 * <dt><b>xmlSource.</b><i>sourceName</i><b>.param.</b><i>class_specific_params</i></dt>
 * <dd>additional parameters to be passed to the xmlSource.<i>sourceName</i>.class.
 * When these parameters are passed, the <i>class_specific_params</i> names
 * will be used (i.e., the "xmlSource.<i>sourceName</i>.param." prefix
 * will be stripped off).</dd>
 * </dl>
 * @see WebmdsXmlSource
 */

public final class WebmdsServlet extends HttpServlet {
    private static final String XML_CONFIG_DIRECTORY_NAME="configDirectory";
    public static final String INFO_SOURCE_ARG_NAME="info";
    public static final String XSL_SOURCE_ARG_NAME="xsl";
    public static final String REQUEST_XSL_PARAM_PREFIX="xslParam";
    private static final String URI_XSL_PARAM="baseURI";
    private static final String INFO_SOURCE_INFO_XSL_PARAM="requestInfo";
    public static final String BASE_DIRECTORY_NAME="baseDirectory";
    private ParameterMap initParameters = null;
    private WebmdsConfigEntry[] configEntryArray = null;
    private WebmdsUtil webmdsUtil = null;

    /**
     * Respond to a GET request for the content produced by
     * this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doGet(HttpServletRequest request,
                      HttpServletResponse response)
	throws IOException, ServletException {

	String contentType = new String("text/html");
	
 	PrintWriter writer = response.getWriter();

	ParameterMap requestParameters = getRequestParameters(request);
	if (requestParameters == null) {
	    throw new WebmdsException("No " + this.INFO_SOURCE_ARG_NAME +
				      " argument was specified.");
	}

	Parameter infoSourceParam =
	    (Parameter)requestParameters.get(this.INFO_SOURCE_ARG_NAME);
	if (infoSourceParam == null) {
	    throw new WebmdsException("No " + this.INFO_SOURCE_ARG_NAME +
				       " argument was specified.");
	}
	String infoSourceName = infoSourceParam.getSingleValue();
	WebmdsXmlSource infoXmlSource = getXmlSource(infoSourceName, false);
	Source infoSource = getSource(infoSourceName,
				      infoXmlSource,
				      requestParameters);

	String xslSourceName = null;
	Parameter xslSourceParam =
	    (Parameter)requestParameters.get(this.XSL_SOURCE_ARG_NAME);
	if (xslSourceParam != null) {
	    xslSourceName = xslSourceParam.getSingleValue();
	}
	Transformer transformer;
	if (xslSourceName == null) {
	    transformer = new TransformerIdentityImpl();
	    contentType = new String("text/xml");
	} else {
	    WebmdsXmlSource xslXmlSource = getXmlSource(xslSourceName, true);
	    Source xslSource = getSource(xslSourceName,
					 xslXmlSource,
					 requestParameters);

	    Parameter contentTypeParam =
		(Parameter)this.initParameters.get(
		    WebmdsConfigEntry.getContentTypeParamName(xslSourceName));

	    if (contentTypeParam != null) {
		contentType = contentTypeParam.getSingleValue();
	    }

	    TransformerFactory transformerFactory =
		TransformerFactory.newInstance();

	    try {
		transformer = 
		    transformerFactory.newTransformer(xslSource);
	    } catch (TransformerConfigurationException e) {
		throw new WebmdsException("Error initializing xsl transformation",
					  e);
	    }

	    setXslParameters(transformer, requestParameters);
	}

	transformer.setParameter(this.INFO_SOURCE_INFO_XSL_PARAM,
				 infoXmlSource.getRequestInfo());

	String uriParam = nonXslURIString(request, requestParameters);

	transformer.setParameter(URI_XSL_PARAM,
				 nonXslURIString(request, requestParameters));

	response.setContentType(contentType);

 	try {
 	    transformer.transform(infoSource, 
 				  new StreamResult(writer));
 	} catch (TransformerException e) {
 	    throw new WebmdsException("Error performing xsl transformation",
 				       e);
	}

    }

    /**
     * Initialize the servlet (read configuration).
     */
    public void init(ServletConfig config) throws ServletException
    {
	String configDirectoryNameString =
	    config.getInitParameter(this.XML_CONFIG_DIRECTORY_NAME);

	if (configDirectoryNameString == null) {
	    throw new WebmdsException("No " + this.XML_CONFIG_DIRECTORY_NAME +
				       " parameter has been configured.");
	}
	this.initParameters = new ParameterMap();

	// Initialize util object
	this.webmdsUtil = new WebmdsUtil();

	ServletContext context = config.getServletContext();
	String baseDirectory = context.getRealPath("/");
	if (baseDirectory != null) {
	    this.webmdsUtil.setDirectoryName(baseDirectory);
	}

	// Add config parameters
	Enumeration paramNames = config.getInitParameterNames();
	while (paramNames.hasMoreElements()) {
	    String name = (String)paramNames.nextElement();
	    if (name != null) {
		Parameter parameter =
		    Parameter.createSingleValuedParameter(
			name,
			config.getInitParameter(name));
		this.initParameters.putParameter(parameter);
	    }
	}

	File configDirectory =
	    this.webmdsUtil.newFile(configDirectoryNameString);
	if (configDirectory == null) {
	    throw new ServletException("Bad configuration directory '" +
				       configDirectoryNameString + "'");
	}
	File[] configFiles = configDirectory.listFiles();
	if (configFiles == null) {
	    throw new WebmdsException("Bad configuration directory '" +
				      configDirectoryNameString + "'");
	}
	this.configEntryArray = new WebmdsConfigEntry[configFiles.length];
	for (int i = 0; i < configFiles.length; i++) {
	    try {
		configEntryArray[i] = initWebmdsConfigEntry(configFiles[i]);
	    } catch (IOException e) {
		throw new WebmdsException(e);
	    }
	}
    }

    private WebmdsConfigEntry initWebmdsConfigEntry(File configFile)
	throws ServletException, IOException {
	WebmdsConfigType configType = new WebmdsConfigType();
	FileInputStream inFile = new FileInputStream(configFile);
	InputSource source = new InputSource(inFile);
	WebmdsConfigType config;
	try {
	    config = (WebmdsConfigType)ObjectDeserializer.deserialize(
		source, WebmdsConfigType.class);
	} catch (DeserializationException e) {
	    throw new WebmdsException(e);
	}
	inFile.close();
	if ((config == null) || (config.getClassName() == null)) {
	    throw new ServletException("Error in config file " +
				       configFile.getAbsolutePath());
	}
	WebmdsConfigEntry configEntry =
	    new WebmdsConfigEntry(configFile.getName(), config);
	configEntry.init();
	return(configEntry);
    }

    private WebmdsXmlSource getXmlSource(String sourceName, boolean checkXSL)
    throws WebmdsException {
	WebmdsConfigEntry config = this.findConfigEntry(sourceName);
	if (config == null) {
	    throw new WebmdsException("No configuration information for '" +
				      sourceName + "'");
	}
	if (checkXSL && (! config.isSuitableForXSL())) {
		throw new WebmdsException("The '" + sourceName + "' XML source has not been configured to be used as a source of XSL.  To change this, set the " + WebmdsConfigEntry.SUITABLE_FOR_XSL_NAME + " parameter to 'true' in the config entry for '" + sourceName + "' on the WebMDS server.");
	}
	WebmdsXmlSource webmdsXmlSource;
	try {
	    webmdsXmlSource =
		(WebmdsXmlSource)config.getSourceClass().newInstance();
	} catch (java.lang.InstantiationException e) {
	    throw new WebmdsException("Error instantiating xml source class for " +
				      sourceName,
				      e);
	} catch (java.lang.IllegalAccessException e) {
	    throw new WebmdsException("Error instantiating xml source class for " +
				      sourceName,
				      e);
	}
	webmdsXmlSource.init(this.webmdsUtil);
	return(webmdsXmlSource);
    }

    private Source getSource(String sourceName,
			     WebmdsXmlSource webmdsXmlSource,
			     ParameterMap requestParams)
	throws ServletException, IOException {
	WebmdsConfigEntry config = this.findConfigEntry(sourceName);
	if (config == null) {
	    throw new WebmdsException("No configuration information for " +
				      sourceName);
	}

	ParameterMap submappedRequestParams =
	    requestParams.prefixedSubMap(
		WebmdsConfigEntry.getCustomParamPrefix(sourceName));
	return(webmdsXmlSource.getXmlSource(sourceName,
					    config.getParams(),
					    submappedRequestParams));

    }

    private WebmdsConfigEntry findConfigEntry(String sourceName) {
	for (int i = 0; i < this.configEntryArray.length; i++) {
	    if (configEntryArray[i].getName().equals(sourceName)) {
		return(configEntryArray[i]);
	    }
	}
	return(null);
    }

    private ParameterMap getRequestParameters(HttpServletRequest request)
	throws ParameterMapException {
	Enumeration paramNames = request.getParameterNames();
	if (paramNames ==  null) {
	    return(null);
	}
	ParameterMap paramMap = new ParameterMap();
	while (paramNames.hasMoreElements()) {
	    String key = (String)paramNames.nextElement();
	    if (key != null) {
		Parameter parameter =
		    new Parameter(key, request.getParameterValues(key));
		paramMap.putParameter(parameter);
	    }
	}
	if (paramMap.size() == 0) {
	    return(null);
	}
	return(paramMap);
    }

    private void setXslParameters(Transformer transformer,
				  ParameterMap paramMap)
	throws WebmdsException
    {
	if (paramMap == null) {
	    return;
	}

	ParameterMap xslParamMap =
	    paramMap.prefixedSubMap(this.REQUEST_XSL_PARAM_PREFIX);

	if (xslParamMap == null) {
	    return;
	}
	Set keySet = xslParamMap.keySet();
	if (keySet == null) {
	    return;
	}
	String[] keys = (String[])keySet.toArray(new String[0]);
	for (int i = 0; i < keys.length; i++) {
	    Parameter xslParam = (Parameter)xslParamMap.get(keys[i]);
	    transformer.setParameter(xslParam.getName(),
				     xslParam.getSingleValue());
	}
    }

    /**
     * nonXslURIString -- returns a URI that consists of the base URI
     * without the xsl-specific parameters ("xsl", "xslParam", etc.);
     */
    private String nonXslURIString(HttpServletRequest request,
				   ParameterMap paramMap)
    {
	String separator = "?";
	StringBuffer uriBuf = new StringBuffer(request.getRequestURI());
	if (paramMap == null) {
	    return(uriBuf.toString());
	}
	Set keySet = paramMap.keySet();
	if (keySet == null) {
	    return(uriBuf.toString());
	}
	String[] keys = (String[])keySet.toArray(new String[0]);
	for (int i = 0; i < keys.length; i++) {
	    Parameter param = (Parameter)paramMap.get(keys[i]);
	    if (! (param.getName().equals(XSL_SOURCE_ARG_NAME) ||
		   (param.getName().startsWith(REQUEST_XSL_PARAM_PREFIX)))) {
		String[] values = param.getValues();
		if (values != null) {
		    for (int j = 0; j < values.length; j++) {
			if (values[j] != null) {
			    uriBuf.append(separator);
			    uriBuf.append(param.getName());
			    uriBuf.append("=");
			    uriBuf.append(values[j]);
			    separator = "&";
			}
		    }
		}
	    }
	}
	return(uriBuf.toString());
    }
}

