/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.mds.webmds.test;

/**
 * This is NOT a test of WebMDS; it's a bunch of utility methods that are
 * used to build unit tests.  For a test of the WebMDS servlet's functionality,
 * run the tests in the file plugin test directory; these test both the
 * WebMDS servlet and the file plugin.
 */

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import com.meterware.httpunit.WebClient;
import com.meterware.httpunit.WebConversation;
import com.meterware.httpunit.WebRequest;
import com.meterware.httpunit.WebResponse;
import com.meterware.httpunit.GetMethodWebRequest;
import com.meterware.servletunit.ServletRunner;
import com.meterware.servletunit.ServletUnitClient;
import com.meterware.httpunit.WebTable;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import org.xml.sax.SAXException;
import java.net.MalformedURLException;
import java.util.Properties;

import java.net.URL;

import org.globus.mds.webmds.WebmdsServlet;

public class ServletTestFramework extends TestCase {
    private File webXmlFile = null;
    private String servletPath = null;
    public static String WEBXML_DIR_PROPERTY_NAME = new String("webmds.test.webXmlDir");
    public static String TEST_FILE_DIR_PROPERTY_NAME = new String("webmds.test.testFileDir");
    public static String SERVLET_URL_PROPERTY_NAME = new String("webmds.test.servletURL");
    public static String PROPERTY_FILE_PROPERTY_NAME = new String("webmds.test.propFile");
    public static String DEFAULT_WEBXML_FILE_NAME = new String("web.xml");
    public static String DEFAULT_SERVLET_PATH = new String("/webmds");
    public static String SERVLET_NAME = new String("webmds");
    public static int RESPONSE_CODE_SUCCESS = 200;
    public static String CONTENT_TYPE_XML = new String("text/xml");
    public static String CONTENT_TYPE_HTML = new String("text/html");

    private Properties prop;
    private ServletUnitClient defaultSimClient = null;
    private WebConversation defaultHttpClient = null;
    private String servletURL = null;

    public ServletTestFramework (String s) {
	super(s);
    }

    protected void setUp() throws Exception {
	String propFileName = System.getProperty(PROPERTY_FILE_PROPERTY_NAME);
	/*
	 * Open properties file.  If it doesn't exist, assume all properties
	 * will be specified as system properties, but if it exists
	 * and can't be read, assume something is wrong.
	 */
	if (propFileName != null) {
	    this.prop = new Properties();
	    try {
		prop.load(new FileInputStream(propFileName));
	    } catch (FileNotFoundException e) {
		this.prop = null;
	    }
	}

	this.defaultSimClient = this.defaultServletUnitClient();
	this.servletURL = this.findPropertyValue(SERVLET_URL_PROPERTY_NAME);
	if (this.servletURL != null) {
	    this.defaultHttpClient = new WebConversation();
	}

    }

    public static Test suite() 
    {
        return new TestSuite(ServletTestFramework.class);
    }

    /**
     * Gets the name of the directory where the web.xml file will be looked
     * for by default (i.e., setDefaults() is called and setWebXmlFile isn't).
     * The web.xml file is used by newServletUnitClient() and defaultServletUnitClient().
     *
     * @return the path name of the directory containing the web.xml file.
     */
    public String defaultWebXmlDir() {
	return(this.findPropertyValue(WEBXML_DIR_PROPERTY_NAME));
    }

    /**
     *
     * @return the path name of the directory containing the web.xml file.
     */
    public String testFileDir() {
	return(this.findPropertyValue(TEST_FILE_DIR_PROPERTY_NAME));
    }

    /**
     *
     * @return the base url of the servlet.
     */
    public String getServletURL() {
	return(this.servletURL);
    }

    /**
     *
     * @return the default servlet runner client.
     */
    public ServletUnitClient getDefaultSimClient() {
	return(this.defaultSimClient);
    }

    /**
     *
     * @return the default http client.
     */
    public WebConversation getDefaultHttpClient() {
	return(this.defaultHttpClient);
    }

    /**
     * Sets the names of the web.xml file and the directory containing it.
     * The web.xml file is used by newServletUnitClient() and defaultServletUnitClient().
     *
     * @param dirname	the name of the directory containing the web.xml file
     * @param filename	the name of the file
     */

    public void setWebXmlFile(String dirname, String filename) {
	this.webXmlFile = new File(dirname, filename);
    }

    /**
     * Sets default values of the web.xml file and the servlet path.
     */
    public void setDefaults() {
	setWebXmlFile(defaultWebXmlDir(), DEFAULT_WEBXML_FILE_NAME);
	setServletPath(DEFAULT_SERVLET_PATH);
    }

    /**
     * Creates a ServletRunner using default values for the web.xml file
     * and servlet path, and returns a client to that ServletRunner.
     */
    public ServletUnitClient defaultServletUnitClient()
    throws IOException, SAXException {
	setDefaults();
	return (newServletUnitClient());
    }

    /**
     * Sets the value of the servlet path; this will be used in subsequent
     * calls to newServletUnitClient.
     *
     * @param path	the servlet path (e.g., "/webmds");
     */
    public void setServletPath(String path) {
	this.servletPath = path;
    }

    /**
     * Creates a ServletRunner using the current values for the web.xml
     * directory and file name and servlet path, and returns a client for
     * that servlet runer.
     *
     * @param path	the servlet path (e.g., "/webmds");
     */
    ServletUnitClient newServletUnitClient()
    throws IOException, SAXException {
        return(new ServletRunner(this.webXmlFile, this.servletPath).newClient());
    }


    /**
     * Returns a string corresponding to the pseudo-url that corresponds
     * to calling a servlet (with ServletRunner) using the webmds servlet.
     * The servlet path must be set (with setServletPath or setDefaults)
     * prior to calling this method.
     *
     * @return	the url string.
     */
    public String servletRunnerUrlBase() {
	return(new String("http://localhost" + this.servletPath));
    }

    public WebResponse runFormTest(WebClient client,
				   GetMethodWebRequest request,
				   String expectedContentType) throws Exception {
	return(runFormTest(client, request, expectedContentType,
			   RESPONSE_CODE_SUCCESS));
    }


    /**
     * Utility function to be called by the test runner -- makes a
     * (pseudo) request and checks the return code and content-type
     * of the response.
     *
     * @param client	A ServletUnitClient as produced by newServletUnitClient.
     * @param request	A request as produced by newGetMethodWebRequest.
     * @param expectedContentType	The expected content-type of the response.
     * @param expectedResponseCode	The expected response code.
     *
     * @return	the response.
     */
    public WebResponse runFormTest(WebClient client,
				   GetMethodWebRequest request,
				   String expectedContentType,
				   int expectedResponseCode) throws Exception {
	WebResponse response = client.getResponse(request);
	assertEquals(expectedResponseCode, response.getResponseCode());
	assertEquals(expectedContentType, response.getContentType());

	return(response);
    }

    /**
     * Creates a GetMethodWebRequest that uses the appropriate URL for
     * a ServletRunner and includes an info parameter.
     *
     * @param infoName	the value of the "info" form parameter for this request.
     */
    public GetMethodWebRequest makeSRGetMethodWebRequest(String infoName)
    throws MalformedURLException {
	return(makeSRGetMethodWebRequest(infoName, null));
    }

    /**
     * Creates a GetMethodWebRequest that uses the appropriate URL for
     * a ServletRunner and includes info and xsl parameters.
     *
     * @param infoParam	the value of the "info" form parameter for this request.
     * @param xslParam	the value of the "xsl" form parameter for this request.
     */
    public GetMethodWebRequest makeSRGetMethodWebRequest(String infoParam,
							 String xslParam) {
	return(makeGetMethodWebRequest(this.servletRunnerUrlBase() +
				       "/" + SERVLET_NAME,
				       infoParam,
				       xslParam));
    }

    /**
     * Creates a GetMethodWebRequest that and includes info and xsl parameters.
     *
     * @param servletUrlString	the servlet URL (usually something like "http[s]://<host>:<port>/webmds/webmds")
     * @param infoParam	the value of the "info" form parameter for this request.
     * @param xslParam	the value of the "xsl" form parameter for this request.
     */
    public GetMethodWebRequest makeGetMethodWebRequest(String servletUrlString,
						       String infoParam,
						       String xslParam) {
	GetMethodWebRequest gr = new GetMethodWebRequest(servletUrlString);
	setWebRequestParameters(gr, infoParam, xslParam);
	return(gr);
    }

    /**
     * Creates a GetMethodWebRequest that and includes info and xsl parameters.
     *
     * @param servletURLString	the servlet URL (usually something like "http[s]://<host>:<port>/webmds/webmds")
     * @param infoParam	the value of the "info" form parameter for this request.
     */
    public GetMethodWebRequest makeGetMethodWebRequest(String servletURLString,
						       String infoParam) {
	return(makeGetMethodWebRequest(servletURLString, infoParam, null));
    }


    /**
     * Sets "info" and "xsl" form parameters in a web request.
     *
     * @param infoParam	the value of the "info" form parameter for this request.
     * @param xslParam	the value of the "xsl" form parameter for this request.
     */
    public void setWebRequestParameters(WebRequest request,
					String infoParam, String xslParam) {
	request.setParameter(WebmdsServlet.INFO_SOURCE_ARG_NAME, infoParam);
	if (xslParam != null) {
	    request.setParameter(WebmdsServlet.XSL_SOURCE_ARG_NAME, xslParam);
	}
    }	


    /**
     * Utility routine for use by junit tests -- asserts that the text
     * of the web response is the same as the contents of a file.
     *
     * @param sourceFile	File containing expected results
     * @param response	Web response (as returned by runFormTest).
     */
    public void assertResultsMatchFile(File sourceFile, WebResponse response)
    throws Exception {
	FileInputStream inputStream = new FileInputStream(sourceFile);
	byte[] expectedBytes = new byte[(int)sourceFile.length()];
	inputStream.read(expectedBytes);

	String responseText = response.getText();
	assertNotNull(responseText);

	/*
	 * The following section compares the response to an expected
	 * XML file.  Unfortunately, a straight text comparison won't
	 * work, because equivalent xml documents can have different
	 * textual representations.  So we just test that the length
	 * is close.
	 */
	char[] responseChars = responseText.toCharArray();
	float lengthRatio = (float)expectedBytes.length / (float)responseChars.length;
	assertTrue(lengthRatio > 0.9);
	assertTrue(lengthRatio < 1.1);
    }

    public String findPropertyValue(String propName) {
	String value = System.getProperty(propName);
	if ((value == null) && (prop != null)) {
	    value = prop.getProperty(propName);
	}
	if (value != null) {
	    if (value.equals(new String(""))) {
		value = null;
	    }
	}
	return(value);
    }

    public void testPrintMessage() throws Exception {
	if (this.getClass() == ServletTestFramework.class) {
	    System.out.println(this.getClass().getName() +
			       " isn't a real test; it's a base class from which tests can be constructed.");
	}
    }

    public void compareRow(WebTable table,
			   int rowNumber,
			   String[] expectedColumns)
	throws Exception {
	for (int col = 0; col < expectedColumns.length; col++) {
	    assertEquals(expectedColumns[col],
			 table.getCellAsText(rowNumber, col).trim());
	}
    }
}
