/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.mds.webmds.xmlSources.xmlDomNode;
import java.util.Vector;
import org.globus.mds.webmds.Parameter;
import org.globus.mds.webmds.ParameterException;
import org.globus.mds.webmds.WebmdsXmlSource;
import org.globus.mds.webmds.WebmdsXmlSourceException;
import org.globus.mds.webmds.ParameterMap;
import org.globus.mds.webmds.xmlSources.xmlDomNode.WebmdsNodeSource;
import org.globus.mds.webmds.WebmdsConfigException;
import org.globus.mds.webmds.WebmdsUtil;
import javax.servlet.ServletException;
import java.io.IOException;
import java.io.FileReader;
import java.io.File;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import org.w3c.dom.Node;


/**
 * NodeXmlSource -- calls a WebmdsNode class to get a DOM node object, and
 * creates a Source object from it.
 * This uses some configuration parameters from the servlet config file:
 * <dl>
 * <dt><b>class</b>
 * (specified as <b>xmlSource.</b><i>sourceName</i>.<b>param.class</b> in
 * the servlet configuration)
 * <dd>The name of a class that implements the WebmdsNodeSource interface.
 * This class will be called to get an XML document</dd>
 * <dt><i>class_specific_parameters</i>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.</b><i>class_specific_parameters</i> in
 * the servlet configuration)
 * <dd>Additional parameters to be passed to 
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.class</b>.
 * </dl>
 *
 * @see WebmdsNodeSource
 */

public class NodeXmlSource implements WebmdsXmlSource
{
    private ParameterMap initParameters = null;

    private final String NODE_CLASS_PARAM_NAME = "class";

    private String requestInfoString = null;

    private WebmdsUtil webmdsUtil = null;

    /**
     * Get an xml Source object -- call a WebmdsNodeSource class to get a Node
     * object, make a Source object out of it, and return it.
     * @param class WebmdsNodeSource class to call
     * @param initParams initialization parameters.  This must include a
     * "class" parameter with the path name to the file
     * @param requestParams parameters from the http request.  Passed
     * on to the WebmdsNodeSource class.
     */
    public Source getXmlSource(String name,
			       ParameterMap initParams,
			       ParameterMap requestParams)
	throws WebmdsXmlSourceException,
	       WebmdsConfigException,
	       IOException {
	Parameter classParam = null;
	classParam = (Parameter)initParams.get(NODE_CLASS_PARAM_NAME);
	if (classParam == null) {
	    throw new WebmdsXmlSourceException(
		"No WebmdsNodeSource class configured for '" +
				      name + "'");
	}

	String className = null;

	try {
	    className = classParam.getSingleValue();
	} catch (ParameterException e) {
	    throw new WebmdsXmlSourceException("Error getting " +
					       NODE_CLASS_PARAM_NAME +
					       " parameter value",
					       e);
	}

	Class nodeClass = null;
	try {
	    nodeClass = Class.forName(className);
	} catch (java.lang.ClassNotFoundException e) {
	    throw new WebmdsXmlSourceException ("Node source class '" +
						className
						+ "' not found",
						e);
	}

	WebmdsNodeSource nodeSource = null;
	try {
	    nodeSource = (WebmdsNodeSource)nodeClass.newInstance(); 
	} catch (java.lang.InstantiationException e) {
	    throw new WebmdsXmlSourceException(
		"Error creating node source object ('" +
		className + "')",
		e);
	} catch (java.lang.IllegalAccessException e) {
	    throw new WebmdsXmlSourceException(
		"Error creating node source object ('" +
		className + "')",
		e);
	} catch (java.lang.ClassCastException e) {
	    throw new WebmdsXmlSourceException (
		"Configuration error: class '" +
		className +
		"' does not implement WebmdsNodeSource " +
		e);
	}

	nodeSource.init(this.webmdsUtil);

	Node node = nodeSource.getNode(initParams, requestParams);
	if (node.getNodeType() != Node.DOCUMENT_NODE) {
	    throw new WebmdsXmlSourceException(
		nodeClass.getName() +
		".getNode returned a Node that is not a document");
	}
	requestInfoString = nodeSource.getRequestInfo();
	return(new DOMSource(node));
    }

    /**
     * Queries the WebmdsNodeSource for a descriptive string and returns it.
     */
    public String getRequestInfo() {
	return(this.requestInfoString);
    }

    public void init(WebmdsUtil webmdsUtil) {
	this.webmdsUtil = webmdsUtil;
    }

}
