/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.client;

import java.util.List;
import java.util.Iterator;
import java.util.Properties;
import java.io.FileInputStream;
import java.io.Writer;

import javax.xml.rpc.Stub;
import javax.xml.namespace.QName;
import org.xml.sax.InputSource;

import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.axis.util.Util;

import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.replica.replicator.generated.service.ReplicationServiceAddressingLocator;
import org.globus.replica.replicator.generated.ReplicatorPortType;
import org.globus.replica.replicator.generated.CreateReplicator;
import org.globus.replica.replicator.generated.CreateReplicatorResponse;
import org.globus.replica.replicator.generated.RequestFileReplicationRequestType;
import org.globus.replica.replicator.generated.RequestFileFormatEnumerationType;
import org.globus.replica.replicator.generated.ReplicationOptionsType;
import org.globus.replica.replicator.generated.TransferOptionsType;
import org.globus.replica.replicator.generated.Start;
import org.globus.replica.replicator.generated.StartResponse;

/**
 * The client used to create a Replicator resource.
 */
public class Create extends org.globus.wsrf.client.BaseClient {

    // CONSTANTS //

    private static final QName NAME =
        new QName("http://www.globus.org/namespaces/2005/05/replica/replicator", "ReplicatorEndpoint");

    private static final int MODE_FILE = 1;
    private static final int MODE_STDIN = 2;

    private static final String CUSTOM_USAGE =
        "request-file";

    private static final String FOOTER =
        "Where:\n" +
        "  request-file - URL of a file describing the data replication request.\n" +
        "           The file must be of the form [source <tab> destination <crlf>]+.\n" +
        "           Spaces are permitted but not required.\n" +
        "\n" +
        "  source - An identifier used to identify the source of the replication.\n" +
        "           This should typically be a LFN name.\n" +
        "\n" +
        "  destination - A URL used to specify the destination of the replication.\n";

    private static final Integer INT_TYPE = new Integer(0);
    private static final Boolean BOOL_TYPE = new Boolean(false);

    protected static final String OPT_DELEGCRED_EPR_FILE = "C";
    protected static final Option DELEGCRED_EPR_FILE =
        OptionBuilder.withLongOpt("delegatedCredential")
        .hasArg()
        .withArgName("file")
        .withDescription("Loads Delegated Credential EPR from file")
        .create(OPT_DELEGCRED_EPR_FILE);

    protected static final String OPT_START = "S";
    protected static final Option START =
        OptionBuilder.withLongOpt("start")
        .withDescription("Starts the Replicator resource immediately")
        .create(OPT_START);

    protected static final String OPT_SAVEEPR = "V";
    protected static final Option SAVEEPR =
        OptionBuilder.withLongOpt("saveEpr")
        .hasArg()
        .withArgName("file")
        .withDescription("Save EPR of newly created Replicator to file")
        .create(OPT_SAVEEPR);

    protected static final Option CONCURRENCY =
        OptionBuilder.withLongOpt("concurrency")
        .hasArg()
        .withArgName("int")
        .withType(INT_TYPE)
        .withDescription("Concurrency of data transfer")
        .create();

    protected static final Option BINARY =
        OptionBuilder.withLongOpt("binary")
        .hasArg()
        .withArgName("boolean")
        .withType(BOOL_TYPE)
        .withDescription("Specifies binary data transfer")
        .create();

    protected static final Option BLOCKSIZE =
        OptionBuilder.withLongOpt("blockSize")
        .hasArg()
        .withArgName("int")
        .withType(INT_TYPE)
        .withDescription("Block size for data transfer")
        .create();

    protected static final Option TCPBUFFERSIZE =
        OptionBuilder.withLongOpt("tcpBufferSize")
        .hasArg()
        .withArgName("int")
        .withType(INT_TYPE)
        .withDescription("TCP buffer size for data transfer")
        .create();

    protected static final Option PARALLELSTREAMS =
        OptionBuilder.withLongOpt("parallelStreams")
        .hasArg()
        .withArgName("int")
        .withType(INT_TYPE)
        .withDescription("Parallel streams for data transfer")
        .create();

    protected static final Option NOTPT =
        OptionBuilder.withLongOpt("noThirdPartyTransfers")
        .hasArg()
        .withArgName("boolean")
        .withType(BOOL_TYPE)
        .withDescription("Disables third-party transfers")
        .create();

    protected static final Option DCAU =
        OptionBuilder.withLongOpt("dataChannelAuth")
        .hasArg()
        .withArgName("boolean")
        .withType(BOOL_TYPE)
        .withDescription("Data channel authentication for transfers")
        .create();

    protected static final Option SUBJ =
        OptionBuilder.withLongOpt("subject")
        .hasArg()
        .withArgName("name")
        .withDescription("Subject name for data transfer")
        .create();

    protected static final Option DESTSUBJ =
        OptionBuilder.withLongOpt("destinationSubject")
        .hasArg()
        .withArgName("name")
        .withDescription("Destination subject name for data transfer")
        .create();

    protected static final Option SRCSUBJ =
        OptionBuilder.withLongOpt("sourceSubject")
        .hasArg()
        .withArgName("name")
        .withDescription("Source subject name for data transfer")
        .create();

    protected static final Option USERNAME =
        OptionBuilder.withLongOpt("userName")
        .hasArg()
        .withArgName("name")
        .withDescription("User name for data transfer")
        .create();


    // FIELDS //

    protected EndpointReferenceType delegatedCredentialEndpoint;
    protected boolean start = false;
    protected String saveEprFile = null;
    protected Integer concurrency = null;
    protected Boolean binary = null;
    protected Integer blockSize = null;
    protected Integer tcpBufferSize = null;
    protected Integer parallel = null;
    protected Boolean noTPT  = null;
    protected Boolean dcau  = null;
    protected String subject = null;
    protected String destinationSubject = null;
    protected String sourceSubject = null;
    protected String username = null;


    // CONSTRUCTORS //

    public Create() {
        this.setCustomUsage(CUSTOM_USAGE);
        this.setHelpFooter(FOOTER);
        this.options.addOption(DELEGCRED_EPR_FILE);
        this.options.addOption(START);
        this.options.addOption(SAVEEPR);
        this.options.addOption(CONCURRENCY);
        this.options.addOption(BINARY);
        this.options.addOption(BLOCKSIZE);
        this.options.addOption(TCPBUFFERSIZE);
        this.options.addOption(PARALLELSTREAMS);
        this.options.addOption(NOTPT);
        this.options.addOption(DCAU);
        this.options.addOption(SUBJ);
        this.options.addOption(DESTSUBJ);
        this.options.addOption(SRCSUBJ);
        this.options.addOption(USERNAME);
    }


    // METHODS //

    /**
     * Main method of the <code>Create</code> client.
     */
    public static void main(String[] args) {
        Create client = new Create();

        // Parse command line arguments and get request file URL
        int mode = 0;
        String requestFileURL = null;
        List requestPairs = null;
        try {
            CommandLine line = client.parse(args);

            List options = line.getArgList();
            if (options == null || options.isEmpty() ||
                    options.size() != 1) {
                // TODO: at a later time, I will get to this...
//                    (options.size() > 1 && (options.size()%2) != 0)) {
                // Illegal arguments
                client.displayUsage();
                System.exit(COMMAND_LINE_ERROR);
            } else if (options.size() == 1) {
                requestFileURL = (String) options.get(0);
                mode = MODE_FILE;
            } else {
                throw new ParseException("Request file URL required.");
//              TODO: at a later time, I will get to this...
//                requestPairs = options;
//                mode = MODE_STDIN;
            }
        } catch(ParseException e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(COMMAND_LINE_ERROR);
        } catch (Exception e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(COMMAND_LINE_ERROR);
        }

        // Locate the replicator service and create instance
        ReplicationServiceAddressingLocator locator =
            new ReplicationServiceAddressingLocator();
        try {
            if (client.isDebugMode()) {
                System.out.println("Service endpoint: " +
                        client.getEPR().toString());
            }
            ReplicatorPortType port = locator.getReplicatorPortTypePort(
                    client.getEPR());
            client.setOptions((Stub)port);

            // Assume request file
            RequestFileReplicationRequestType request
                = new RequestFileReplicationRequestType();
            request.setRequestFileUri(requestFileURL);
            request.setFormat(RequestFileFormatEnumerationType.Table);

            // Create replication options
            TransferOptionsType xferOpts = new TransferOptionsType();
            client.initializeTransferOptions(xferOpts);
            ReplicationOptionsType options = new ReplicationOptionsType();
            options.setTransferOptions(xferOpts);
            request.setOptions(options);

            // Create delegated credential
            request.setCredentialEPR(client.delegatedCredentialEndpoint);

            // Create replicator
            if (client.isDebugMode()) {
                System.out.println("Creating Replicator...");
            }
            CreateReplicator create = new CreateReplicator();
            create.setRequestFileRequest(request);
            CreateReplicatorResponse createResponse =
                port.createReplicator(create);

            EndpointReferenceType epr =
                createResponse.getEndpointReference();
            if (client.isDebugMode()) {
                System.out.println("Replicator created.");
            }

            // Save EPR (Optionally)
            if (client.saveEprFile != null) {
                client.saveEndpoint(epr, client.saveEprFile);
            }

            // Start replicator (Optionally)
            if (client.start) {
                // lookup newly created replicator
                port = locator.getReplicatorPortTypePort(epr);
                client.setOptions((Stub)port);

                if (client.isDebugMode()) {
                    System.out.println("Starting Replicator...");
                }
                port.start(new Start());
                if (client.isDebugMode()) {
                    System.out.println("Replicator started.");
                }
            }

        } catch(Exception e) {
            System.err.println("Error: " + e.getMessage());
            if (client.isDebugMode()) {
                e.printStackTrace(System.err);
            }
            System.exit(APPLICATION_ERROR);
        }
    }

    /**
     * Parses the command-line arguments specific to the <code>Create</code>
     * client.
     */
    protected CommandLine parse(String [] args, Properties defaultOptions)
        throws Exception {

        CommandLine line = super.parse(args, defaultOptions);

        // Check required option for delegated credential EPR
        if (line.hasOption(OPT_DELEGCRED_EPR_FILE)) {
            FileInputStream in = null;
            try {
                in = new FileInputStream(
                        line.getOptionValue(OPT_DELEGCRED_EPR_FILE));
                this.delegatedCredentialEndpoint =
                (EndpointReferenceType)ObjectDeserializer.deserialize(
                        new InputSource(in),
                        EndpointReferenceType.class);
            } finally {
                if (in != null) {
                    try { in.close(); } catch (Exception e) {}
                }
            }
        } else {
            throw new ParseException(OPT_DELEGCRED_EPR_FILE + " is required");
        }

        // Check start option
        if (line.hasOption(OPT_START)) {
            this.start = true;
        }

        // Check save epr option
        if (line.hasOption(OPT_SAVEEPR)) {
            this.saveEprFile = line.getOptionValue(OPT_SAVEEPR);
        }

        // Check optional transfer options
        Iterator iter = line.iterator();
        while (iter.hasNext()) {
            Option opt = (Option) iter.next();

            if (this.isDebugMode()) {
                System.out.println("Argument:");
                System.out.println("    Option: " + opt.getLongOpt());
                System.out.println("    Value:  " + opt.getValue());
            }

            if (opt == CONCURRENCY) {
                this.concurrency = new Integer(opt.getValue());
            }
            else if (opt == BINARY) {
                this.binary = new Boolean(opt.getValue());
            }
            else if (opt == BLOCKSIZE) {
                this.blockSize = new Integer(opt.getValue());
            }
            else if (opt == TCPBUFFERSIZE) {
                this.tcpBufferSize = new Integer(opt.getValue());
            }
            else if (opt == PARALLELSTREAMS) {
                this.parallel = new Integer(opt.getValue());
            }
            else if (opt == NOTPT) {
                this.noTPT = new Boolean(opt.getValue());
            }
            else if (opt == DCAU) {
                this.dcau = new Boolean(opt.getValue());
            }
            else if (opt == SUBJ) {
                this.subject = opt.getValue();
            }
            else if (opt == DESTSUBJ) {
                this.destinationSubject = opt.getValue();
            }
            else if (opt == SRCSUBJ) {
                this.sourceSubject = opt.getValue();
            }
            else if (opt == USERNAME) {
                this.username = opt.getValue();
            }

        }

        return line;
    }

    /**
     * Initializes the transfer options according to the values parsed from the
     * command-line arguments.
     */
    protected void initializeTransferOptions(TransferOptionsType opts) {
        opts.setConcurrency(this.concurrency);
        opts.setBinary(this.binary);
        opts.setBlockSize(this.blockSize);
        opts.setTcpBufferSize(this.tcpBufferSize);
        opts.setParallelStreams(this.parallel);
        opts.setNoTPT(this.noTPT);
        opts.setDcau(this.dcau);
        opts.setSubjectName(this.subject);
        opts.setDestinationSubjectName(this.destinationSubject);
        opts.setSourceSubjectName(this.sourceSubject);
        opts.setUsername(this.username);
    }

    /**
     * Save endpoint to file.
     */
    protected static void saveEndpoint(EndpointReferenceType epr,
            String filename) throws Exception {
        Writer writer = null;
        try {
            writer = new java.io.BufferedWriter(new java.io.FileWriter(filename));
            ObjectSerializer.serialize(writer, epr, NAME);
        } finally {
            if (writer != null) try { writer.close(); } catch (Exception ignored) { }
        }
    }
}
