/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.client;

import java.util.Properties;
import java.io.Writer;
import javax.xml.rpc.Stub;
import javax.xml.namespace.QName;

import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;

import org.globus.wsrf.utils.FaultHelper;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.replica.replicator.generated.service.ReplicationServiceAddressingLocator;
import org.globus.replica.replicator.generated.ReplicatorPortType;
import org.globus.replica.replicator.generated.FindItemsResponse;
import org.globus.replica.replicator.generated.ReplicationItemStatusEnumerationType;

public class FindItems extends org.globus.wsrf.client.BaseClient {

    // CONSTANTS //

    private static final QName RESPONSE_NAME =
        new QName("http://www.globus.org/namespaces/2005/05/replica/replicator", "FindItemsResponse");

    private static final String CUSTOM_USAGE =
        "{-N name | -S status}";

    protected static final String OPT_BY_NAME = "N";
    protected static final Option BY_NAME =
        OptionBuilder.withLongOpt("byName")
        .hasArg()
        .withArgName("name")
        .withDescription("Finds item by the Logical Filename (LFN) name.")
        .create(OPT_BY_NAME);

    protected static final String OPT_BY_STATUS = "S";
    protected static final Option BY_STATUS =
        OptionBuilder.withLongOpt("byStatus")
        .hasArg()
        .withArgName("status")
        .withDescription("Finds item(s) by status. Valid status values include" +
                " \"Pending\", \"Finished\", \"Failed\", and \"Terminated\".")
        .create(OPT_BY_STATUS);

    protected static final String OPT_OFFSET = "O";
    protected static final Option OFFSET =
        OptionBuilder.withLongOpt("offset")
        .hasArg()
        .withArgName("num")
        .withDescription("Offset into the results set. Indexed by 0.")
        .create(OPT_OFFSET);

    protected static final String OPT_LIMIT = "L";
    protected static final Option LIMIT =
        OptionBuilder.withLongOpt("limit")
        .hasArg()
        .withArgName("num")
        .withDescription("Limit on the size of the result set.")
        .create(OPT_LIMIT);

    // FIELDS //

    String name = null;
    ReplicationItemStatusEnumerationType status = null;
    Integer limit = null;
    Integer offset = null;


    // CONSTRUCTORS //

    public FindItems() {
        this.setCustomUsage(CUSTOM_USAGE);
        this.options.addOption(BY_NAME);
        this.options.addOption(BY_STATUS);
        this.options.addOption(OFFSET);
        this.options.addOption(LIMIT);
    }


    // METHODS //

    public static void main(String[] args) {
        FindItems client = new FindItems();
        try {
            client.parse(args);
        } catch (ParseException e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(COMMAND_LINE_ERROR);
        } catch (Exception e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(COMMAND_LINE_ERROR);
        }

        ReplicationServiceAddressingLocator locator =
            new ReplicationServiceAddressingLocator();

        try {
            ReplicatorPortType port =
                locator.getReplicatorPortTypePort(client.getEPR());
            client.setOptions((Stub)port);

            // Initialize the parameters
            org.globus.replica.replicator.generated.FindItems params =
                new org.globus.replica.replicator.generated.FindItems();
            params.setByUri(client.name);
            params.setByStatus(client.status);
            params.setOffset(client.offset);
            params.setLimit(client.limit);

            // Call the Replicator
            FindItemsResponse response = port.findItems(params);

            // Display the response
            client.displayResponse(response);
        } catch (Exception e) {
            if (client.isDebugMode()) {
                FaultHelper helper = new FaultHelper(FaultHelper.toBaseFault(e));
                helper.printStackTrace(System.err);
            } else {
                System.err.println("Error: " + FaultHelper.getMessage(e));
            }
            System.exit(APPLICATION_ERROR);
        }
    }


    /**
     * Parses the command-line arguments specific to the <code>FindItem</code>
     * client.
     */
    protected CommandLine parse(String [] args, Properties defaultOptions)
        throws Exception {

        CommandLine line = super.parse(args, defaultOptions);

        // Check byName/byStatus options
        if (line.hasOption(OPT_BY_NAME)) {
            this.name = line.getOptionValue(OPT_BY_NAME);
            if (line.hasOption(OPT_BY_STATUS))
                throw new ParseException(OPT_BY_NAME + " and " + OPT_BY_STATUS +
                        " are mutually exclusive.");
        }
        else if (line.hasOption(OPT_BY_STATUS)) {
            try {
                this.status = ReplicationItemStatusEnumerationType.fromValue(
                        line.getOptionValue(OPT_BY_STATUS));
            }
            catch (IllegalArgumentException iaex) {
                throw new ParseException("Illegal value specified for option "
                        + OPT_BY_STATUS);
            }
        }
        else {
            throw new ParseException(OPT_BY_NAME + " or " + OPT_BY_STATUS +
                    " must be specified.");
        }

        // Check for offset
        if (line.hasOption(OPT_OFFSET))
            this.offset = new Integer(line.getOptionValue(OPT_OFFSET));

        // Check for limit
        if (line.hasOption(OPT_LIMIT))
            this.limit = new Integer(line.getOptionValue(OPT_LIMIT));

        return line;
    }

    /**
     * Display response.
     */
    protected static void displayResponse(FindItemsResponse response)
            throws Exception {
        Writer writer = null;
        try {
            writer = new java.io.OutputStreamWriter(System.out);
            ObjectSerializer.serialize(writer, response, RESPONSE_NAME);
        } finally {
            if (writer != null) try { writer.close(); } catch (Exception ignored) { }
        }
    }
}
