/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.replica.rls.RLSClient;
import org.globus.replica.rls.RLSException;

/**
 * INTERNAL USE ONLY.
 * Provides utility method(s) to simplify RLS usage.
 */
class RLSUtil {

    // CONSTANTS //

    private static Log logger =
        LogFactory.getLog(RLSUtil.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
                RLSUtil.class.getClassLoader());

    /**
     * INTERNAL USE ONLY.
     * Attempts connection to RLS using retry logic.
     * @param url The url for this connection.
     * @param proxyFileName The filename of the user proxy.
     * @param retryMax The max number of times to retry the connection.
     * @param retryWait The seconds to wait between connection retries.
     * @param retryBackoff Flag to indicate whether to exponentially backoff
     *     by exponentially increasing the wait time per retry.
     * @throws RLSException Indicates that the call failed due to: (a)
     *     the max number of retries was reached, or (b) the connection error
     *     type indicates that no retries will succeed.
     * @throws InterruptedExcpetion Indicates that the thread was interrupted
     *     while waiting inbetween connection retries.
     */
    static RLSClient connect(
            String url,
            String proxyFilename,
            int retryMax,
            int retryWait,
            boolean retryBackoff)
        throws RLSException,
               InterruptedException {
        logger.trace("Attempting connection to RLS, url: " + url);
        Object objWait = new Object();
        RLSException lastex = null;
        RLSClient rls = null;
        for (int i = 0; (rls==null) && (i<retryMax); i++) {
            // If this is a retry, Wait before trying connection
            if (i>0) {
                logger.debug("Retry, i: " + i);
                long waitPeriod = retryWait * 1000;
                if (retryBackoff)
                    waitPeriod *= i;
                objWait.wait(waitPeriod);
            }

            // Attempt connection
            try {
                rls = new RLSClient(url, proxyFilename);
            }
            catch (RLSException rlsex) {
                logger.warn(rlsex.getMessage());
                if (rlsex.GetRC() != RLSClient.RLS_TOO_MANY_CONNECTIONS) {
                    // Can't retry after this type of error
                    logger.debug("RLS connection error, rc: " + rlsex.GetRC());
                    throw rlsex;
                }
                else {
                    lastex = rlsex;
                }
            }
        }

        // If failed to connect, throw last RLS Exception
        if (rls==null) {
            logger.debug("After retry, still failed to connect to RLS");
            throw lastex;
        }

        return rls;
    }

    /**
     * INTERNAL USE ONLY.
     * Attempts connection to RLS using retry logic.
     * @param rls A valid connection to RLS.
     * @throws RLSException Indicates that the call failed.
     */
    static void close(RLSClient rls) throws RLSException {
        logger.trace("Attempting to close connection to RLS");
        rls.Close();
    }

    /**
     * INTERNAL USE ONLY.
     * Performs a bulk operation by dividing the parameters into subsets,
     * performing multiple bulk operations, and collecting all results.
     * @param rls The rls object representing a valid client connection.
     * @param operation The operation to be performed.
     * @param params The operations parameters.
     * @param results The operations collected results.
     * @param maxPerBulkOperation The number of parameters to use in each
     *     bulk operation performed by <code>operation</code>.
     * @param timeout Optional timeout value which will be used to 
     *     <code>rls.SetTimeout(timeout)</code>. If <code>-1</code>, no
     *     timeout value will be set. If a timeout value is set, then the
     *     original timeout value will be reset at the end of the call.
     * @throws RLSException Indicates that a <code>RLSException</code> was
     *     thrown by the <code>operation</code>.
     */
    static void bulkOperation(
            RLSClient rls,
            RLSBulkOperation operation,
            ArrayList params,
            ArrayList results,
            int maxPerBulkOperation,
            int timeout
            )
        throws
            RLSException {

        // Validate parameters
        if (rls == null
                || operation == null
                || params == null
                || results == null
                || maxPerBulkOperation <= 0)
            throw new IllegalArgumentException();

        // Set timeout if specified
        int originalTimeout = -1;
        if (timeout >= 0) {
            originalTimeout = rls.GetTimeout();
            rls.SetTimeout(timeout);
        }

        // Get RLS Url for debug purposes
        String url = null;
        if (logger.isDebugEnabled())
            url = rls.GetURL();

        // Divide parameters into subqueries
        List subqueries = new java.util.LinkedList();
        int size = params.size();
        for (int i=0; (i*maxPerBulkOperation)<size; i++) {
            logger.debug("Creating subquery, i: " + i);
            int start = i*maxPerBulkOperation;
            int end = ((i+1)*maxPerBulkOperation<size) ?
                (i+1)*maxPerBulkOperation : size;
            List subquery = params.subList(start,end);
            subqueries.add(new java.util.ArrayList(subquery));
        }

        // Perform the bulk operations
        Iterator iter = subqueries.iterator();
        for (int i=0; iter.hasNext(); i++) {
            ArrayList subquery = (ArrayList) iter.next();
            if (logger.isDebugEnabled())
                logger.debug("Performing operation, url: " + url + ", i: " + i + ", size: " + subquery.size());
            results.addAll(operation.invoke(rls, subquery));
            if (logger.isDebugEnabled())
                logger.debug("Finished operation, url: " + url + ", i: " + i);
            subquery.clear();
        }
        subqueries.clear();

        // Reset timeout, if one was specified
        if (originalTimeout >= 0)
            rls.SetTimeout(originalTimeout);
    }
}
