/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.wsrf.utils.PerformanceLog;

import org.globus.replica.rls.RLSClient;
import org.globus.replica.rls.RLSClient.LRC;
import org.globus.replica.rls.RLSException;
import org.globus.replica.rls.RLSString2;
import org.globus.replica.rls.RLSString2Bulk;

import org.globus.replica.replicator.ReplicationStatus;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.Destination;
import org.globus.replica.replicator.util.DelegatedCredential;

/**
 * Implements the <code>Work</code> required to process replication
 * requests.
 */
class RegisterWork implements commonj.work.Work {

    // STATIC //

    private static RlsLrcCreateBulk createOperation = new RlsLrcCreateBulk();
    private static RlsLrcAddBulk addOperation = new RlsLrcAddBulk();

    private static Log logger =
        LogFactory.getLog(RegisterWork.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
                RegisterWork.class.getClassLoader());
    private static PerformanceLog perfLogger =
        new PerformanceLog(RegisterWork.class.getName());

    // FIELDS //

    private ReplicationManagerImpl manager;
    private ReplicationImpl replication;
    private RLSClient.LRC lrc;

    int retryMax;
    int retryWait;
    boolean retryBackoff;
    int maxPerUpdate;
    int updateTimeout;

    // CONSTRUCTORS //

    RegisterWork(ReplicationManagerImpl m, ReplicationImpl r) {
        this.manager = m;
        this.replication = r;

        this.retryMax = this.manager.getRegisterRetryMax();
        this.retryWait = this.manager.getRegisterRetryWait();
        this.retryBackoff = this.manager.getRegisterRetryBackoff();
        this.maxPerUpdate = this.manager.getRegisterMaxPerUpdate();
        this.updateTimeout = this.manager.getRegisterUpdateTimeout();
    }


    // METHODS //

    public void release() {
        this.manager = null;
        this.replication = null;
        this.lrc = null;
    }

    public boolean isDaemon() {
        return false;
    }

    public void run() {
        perfLogger.start();
        RLSClient rls = null;
        try {
            // Connect to RLI
            rls = RLSUtil.connect(this.replication.registrationUrl,
                    this.replication.credentials.getProxyFilename(),
                    this.retryMax, this.retryWait, this.retryBackoff);

            // Create bulk create params
            List creates = new java.util.LinkedList();
            List adds = new java.util.LinkedList();
            _populateInitialBulkParams(creates, adds);
            logger.trace("Initial registration params, creates: "
                    + creates.size() + ", adds: " + adds.size());

            if (creates.size() > 0) {
                // Attemp registration by create operation
                List tempCreates = new java.util.ArrayList(creates);
                List createResults = new java.util.ArrayList(0);
                RLSUtil.bulkOperation(rls, RegisterWork.createOperation,
                        (ArrayList)tempCreates, (ArrayList)createResults,
                        this.maxPerUpdate, this.updateTimeout);

                // Update bulk add param
                _processCreateResults(adds, createResults);
                tempCreates.clear();
                createResults.clear();
                creates.clear();
                logger.trace("Updated param, add: " + adds.size());
            }

            if (adds.size() > 0) {
                // Attempt registration by add operation
                List tempAdds = new java.util.ArrayList(adds);
                List addResults = new java.util.ArrayList(0);
                RLSUtil.bulkOperation(rls, RegisterWork.addOperation,
                        (ArrayList)tempAdds, (ArrayList)addResults,
                        this.maxPerUpdate, this.updateTimeout);

                // Update replication items and destinations
                _processAddResults(addResults);
                tempAdds.clear();
                addResults.clear();
                adds.clear();
            }

            // Update replication
            _updateReplication();

        } catch (RLSException rlsex) {
            logger.error("Failed registration, exception: " + rlsex.getMessage());
            logger.debug(rlsex.getMessage(),rlsex);
            // General rls failures (i.e., those encountered after initial connect)
            // are considered fatal errors.
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "Failed registration, exception: " + rlsex.getMessage()); // TODO I18n
        } catch (Exception ex) {
            logger.error("Internal error, exception: " + ex.getMessage());
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(ex.getMessage(),ex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    ex.getMessage());
        } finally {
            // Close RLS connection
            if (rls!=null)
                try { RLSUtil.close(rls); } catch (Exception ignored) { }

            // Persist the replication
            _persistReplication();

            // Log performance
            perfLogger.stop(ReplicationConstants.PERFLOG_REGISTER);

            // Notify status observers
            this.replication.notifyStatusObservers();
        }
    }

    /**
     * INTERNAL USE ONLY.
     * Populates the <code>creates</code> and <code>adds</code> lists
     * with the lfn-pfn mappings to be created or added. The <code>adds</code>
     * list is used in cases where more than one destination exists for
     * a given item.
     */
    private void _populateInitialBulkParams(List creates, List adds) {
        logger.trace("_populateInitialBulkParams");

        Iterator iter = this.replication.items().iterator();
        while (iter.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
            // Skip failed items
            if (item.status == ReplicationItem.FAILED)
                continue;

            // Add lfn-pfn mappings to the appropriate list by keeping
            // track of whether the destination was the "first" to be
            // added to a given list. Non-"first" destinations of course
            // belong in the adds list.
            boolean first = true;
            for (int i=0; i<item.destinations.length; i++) {
                DestinationImpl dest = item.destinations[i];
                if (dest.status == Destination.TRANSFERED) {
                    RLSString2 s2 = new RLSString2(item.uri, dest.uri);
                    if (first) {
                        creates.add(s2);
                        first = false;
                    }
                    else {
                        adds.add(s2);
                    }
                }
            }
        }
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _processCreateResults(List adds, List createResults) {
        logger.trace("_processCreateResults");

        Iterator iter = createResults.iterator();
        while (iter.hasNext()) {
            RLSString2Bulk s2b = (RLSString2Bulk) iter.next();
            String lfn = s2b.s1;
            String pfn = s2b.s2;
            int rc = s2b.rc;

            if (rc == RLSClient.RLS_LFN_EXIST) {
                // LFN exists so we try during add operation
                adds.add(new RLSString2(lfn,pfn));
            }
            else {
                // RC indicates success or falure, which requires an
                // update of the replication item's destination.

                // Lookup item and destination
                ReplicationItemImpl item = this.replication.findItemByUri(lfn);
                if (item == null) {
                    logger.warn("Invalid LFN name returned from RLS, lfn: " + lfn);
                    continue;
                }
                DestinationImpl dest = item.findDestinationByUri(pfn);
                if (dest == null) {
                    logger.warn("Invalid PFN name returned from RLS, pfn: " + lfn);
                    continue;
                }

                // Update destination status
                if (rc == RLSClient.RLS_SUCCESS) {
                    dest.status = Destination.REGISTERED;
                }
                else {
                    if (logger.isDebugEnabled())
                        logger.debug("Registration failed, pfn: " + pfn + ", rc: " + rc);
                    dest.changeStatus(Destination.FAILED,
                            "Registration failed, rc: " + rc);
                }

                // Set dirty
                item.setDirty();
            }
        }
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _processAddResults(List addResults) {
        logger.trace("_processAddResults");

        Iterator iter = addResults.iterator();
        while (iter.hasNext()) {
            RLSString2Bulk s2b = (RLSString2Bulk) iter.next();
            String lfn = s2b.s1;
            String pfn = s2b.s2;
            int rc = s2b.rc;

            // Lookup item and destination
            ReplicationItemImpl item = this.replication.findItemByUri(lfn);
            if (item == null) {
                logger.warn("Invalid LFN name returned from RLS, lfn: " + lfn);
                continue;
            }
            DestinationImpl dest = item.findDestinationByUri(pfn);
            if (dest == null) {
                logger.warn("Invalid PFN name returned from RLS, pfn: " + lfn);
                continue;
            }

            // Update destination status
            if (rc == RLSClient.RLS_SUCCESS ||
                    rc == RLSClient.RLS_MAPPING_EXIST) {
                // NOTE: we can consider _mapping_exist_ a success case because
                // for whatever reason the lfn-pfn mapping exists and it isn't
                // detrimental that it existed before our attempt to add it.
                dest.status = Destination.REGISTERED;
            }
            else {
                if (logger.isDebugEnabled())
                    logger.debug("Registration failed, pfn: " + pfn + ", rc: " + rc);
                dest.changeStatus(Destination.FAILED,
                        "Registration failed, rc: " + rc);
            }

            // Set dirty
            item.setDirty();
        }
    }

    /**
     * INTERNAL USE ONLY.
     * Update the replication's cound and status by iterating through the
     * entire collection of replication items and each item's destinations.
     */
    private void _updateReplication() {
        logger.trace("_updateReplication");
        // Update replication count
        int finished = 0;
        int registerFailed = 0;
        Iterator iter = this.replication.items().iterator();
        while (iter.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
            if (item.status == ReplicationItem.FAILED)
                continue;

            // Check if all destinations failed
            boolean failed = true;
            for (int i=0; i<item.destinations.length; i++) {
                int status = item.destinations[i].status;
                if (status == Destination.REGISTERED) {
                    // NOTE: At this point, if at least one destination was
                    // registered we count this as a success.
                    failed = false;
                }
                else if (status == Destination.TRANSFERED) {
                    // RLS does not return a result for items which succeed.
                    // Therefore the previous processing of results would not
                    // be able to update a destination's status if it had
                    // been registered. Thus at this point we update the
                    // status to REGISTERED and set the flag accordingly.
                    item.destinations[i].status = Destination.REGISTERED;
                    failed = false;
                }
                else if (status != Destination.FAILED) {
                    // If for some reason a destination is neither finished nor
                    // transfered, we mark as failed at this point.
                    item.destinations[i].changeStatus(
                            Destination.FAILED,
                            "Unknown final state");
                }
            }


            // Update count so far
            if (failed)
                registerFailed++;
            else {
                finished++;
                item.status = ReplicationItem.FINISHED;
            }

            // Set dirty
            item.setDirty();
        }

        // Update replication count
        this.replication.failed += registerFailed;
        this.replication.finished = finished;
        logger.debug("Replication, total: " + this.replication.total +
                ", finished: " + this.replication.finished +
                ", failed: " + this.replication.failed);

        // Check for total failure
        if (this.replication.total == this.replication.failed) {
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "All items have failed to be located, transfered, or registered.");
        }
        else {
            // Reset stage and update status as finished
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FINISHED,
                    "Finished " + new java.util.Date().toString());
            this.replication.stage = ReplicationImpl.NONE;
        }

        // Set dirty flag
        this.replication.setDirty();

        // Notify observers
        this.replication.notifyCountObservers();
        this.replication.notifyStatusObservers();
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _persistReplication() {
        logger.trace("_persistReplication");
    }
}

