/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Iterator;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.replica.replicator.Replication;
import org.globus.replica.replicator.ReplicationStatus;
import org.globus.replica.replicator.ReplicationCount;
import org.globus.replica.replicator.ReplicationOptions;
import org.globus.replica.replicator.TransferOptions;
import org.globus.replica.replicator.ReplicationCredentials;
import org.globus.replica.replicator.ReplicationStatusObserver;
import org.globus.replica.replicator.ReplicationCountObserver;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.NotFoundException;

import org.globus.replica.replicator.util.DelegatedCredential;
import org.globus.replica.replicator.util.DirtyFlag;

/**
 * INTERNAL USE ONLY.
 */
class ReplicationImpl implements Replication,
    ReplicationStatus, 
    ReplicationCount,
    ReplicationOptions,
    DirtyFlag
{

    // CONSTANTS //

    static final int INACTIVE_FLAGS = ReplicationStatus.SUSPENDED
                                        | ReplicationStatus.TERMINATED
                                        | ReplicationStatus.DESTROYED;


    // STATIC //
    private static Log logger =
        LogFactory.getLog(ReplicationImpl.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
        ReplicationImpl.class.getClassLoader());


    // FIELDS //

    // per Replication interface
    Object key;
    private Map itemsMap;
    private List items;

    // per DirtyFlag interface
    boolean dirty;

    // per ReplicationStatus interface
    Object statusLock;
    int status;
    int result;
    private String errMsg;
    private List statusObservers;
    int stage;

    // per ReplicationCount interface
    long total;
    long finished;
    long failed;
    long terminated;
    private List countObservers;

    // per ReplicationOptions interface
    String indexUrl;
    String reliableTransferUrl;
    String registrationUrl;
    TransferOptionsImpl transferOptions;

    // per ReplicationCredentials interface
    DelegatedCredential credentials;


    // CONSTRUCTORS //

    ReplicationImpl() {
        this.key = null;
        this.itemsMap = null;
        this.items = null;
        this.dirty = false;
        this.statusLock = new Object();
        this.status = ReplicationStatus.PENDING;
        this.result = ReplicationStatus.NONE;
        this.total = 0;
        this.finished = 0;
        this.failed = 0;
        this.terminated = 0;
        statusObservers = new java.util.LinkedList();
        countObservers = new java.util.LinkedList();
        this.stage = ReplicationImpl.NONE;
    }

    /**
     * Destroys the <code>ReplicationImpl</code> state. This call does not
     * release the resources held by the object.
     */
    void destroy() {
        if (this.itemsMap != null) {
            this.itemsMap.clear();
        }
        if (this.items != null) {
            // TODO: could iterate the collection and call item.destroy()
            this.items.clear();
        }
        if (this.credentials != null) {
            this.credentials.release();
            this.credentials = null;
        }
    }

    /**
     * Release resources held by this object.
     */
    void release() {
        this.key = null;
        if (this.itemsMap != null) {
            this.itemsMap.clear();
            this.itemsMap = null;
        }
        if (this.items != null) {
            // TODO: could iterate the collection and call item.release()
            this.items.clear();
            this.items = null;
        }
        this.dirty = false;
        this.statusLock = null;
        this.errMsg = null;
        if (this.statusObservers != null) {
            this.statusObservers.clear();
            this.statusObservers = null;
        }
        if (this.countObservers != null) {
            this.countObservers.clear();
            this.countObservers = null;
        }
        this.indexUrl = null;
        this.reliableTransferUrl = null;
        this.registrationUrl = null;
        this.transferOptions = null;
        if (this.credentials != null) {
            this.credentials.release();
            this.credentials = null;
        }
    }

    /**
     * The intention is for <code>release</code> to be called by the application
     * logic prior to this method. Only as a catchall, this method also calls
     * <code>release</code>.
     */
    protected void finalize() throws Throwable {
        this.release();
    }

    // METHODS //

    // per DirtyFlag interface
    public boolean isDirty() { return this.dirty; }
    public void setDirty() { this.dirty = true; }
    public void clearDirty() { this.dirty = false; }

    // per Replication interface
    public Object getKey() { return this.key; }
    public ReplicationStatus getReplicationStatus() { return this; }
    public ReplicationCount getReplicationCount() { return this; }
    public ReplicationCredentials getReplicationCredentials() { return this.credentials; }
    public ReplicationOptions getReplicationOptions() { return this; }
    public Collection items() { return this.items; }

    public Collection findItemsByUri(String uri, int offset, int limit)
        throws NotFoundException, IndexOutOfBoundsException {
        ReplicationItem item = this.findItemByUri(uri);
        if (item == null)
            throw new NotFoundException();
        Collection results = new java.util.ArrayList(1);
        results.add(item);
        return results;
    }

    public Collection findItemsByStatus(int status, int offset, int limit)
        throws NotFoundException, IndexOutOfBoundsException {
        Collection results = null;
        if (limit>0)
            results = new java.util.ArrayList(limit);
        else
            results = new java.util.LinkedList();

        // Search items
        int found = 0;
        Iterator iter = this.items.iterator();
        while (iter.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
            if (item.status == status) {
                // Increment count of found items
                found++;
                // Add to results only if passed the offset
                if (found>offset)
                    results.add(item);
                // Break, stop searching if limit met
                if (limit>0 && limit<=(found-offset))
                    break;
            }
        }

        if (results.isEmpty())
            throw new NotFoundException();
        return results;
    }

    ReplicationItemImpl findItemByUri(String uri) {
        return (ReplicationItemImpl) this.itemsMap.get(uri);
    }

    java.util.Collection uris() {
        return this.itemsMap.keySet();
    }

    void addItems(List items) {

        // TODO this check is not needed since this is non public
        if (items == null || items.isEmpty())
            throw new IllegalArgumentException("Items null or empty");

        // Create or extend list collection of items
        if (this.items == null)
            this.items = new java.util.ArrayList(items);
        else
            this.items.addAll(items);

        // Create or extend map collection of items
        if (this.itemsMap == null)
            this.itemsMap = new java.util.HashMap(items.size());

        // Add items to map
        Iterator iter = items.iterator();
        while (iter.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
            if (item.uri==null || item.uri.length()==0)
                throw new IllegalArgumentException("Null URI in ReplicationItem");
            this.itemsMap.put(item.uri, item);
        }
    }

    // per ReplicationStatus interface
    public int getStatus() {
        synchronized (this.statusLock) {
            return this.status;
        }
    }
    public int getResult() {
        synchronized (this.statusLock) {
            return this.result;
        }
    }
    public String getErrorMsg() {
        synchronized (this.statusLock) {
            return this.errMsg;
        }
    }
    void changeStatus(int status) {
        synchronized (this.statusLock) {
            if (!_isStatusChangeValid(status))
                return;
            this.status = status;
            this.dirty = true;
        }
    }
    void changeStatus(int status, int result, String errMsg) {
        synchronized (this.statusLock) {
            if (!_isStatusChangeValid(status))
                return;
            this.status = status;
            this.result = result;
            this.errMsg = this.errMsg;
            this.dirty = true;
        }
    }
    /**
     * INTERNAL USE ONLY.
     * This method is used to prevent an invalid status change. For instance,
     * the background workers are likely to attempt a status change at an
     * inappropriate time.
     */
    private boolean _isStatusChangeValid(int status) {
        if (status == ReplicationStatus.TERMINATED)
            return ((this.status & (ReplicationStatus.TERMINATED
                        | ReplicationStatus.DESTROYED)) == 0);
        if (status == ReplicationStatus.SUSPENDED)
            return ((this.status & (ReplicationStatus.TERMINATED
                        | ReplicationStatus.SUSPENDED
                        | ReplicationStatus.DESTROYED)) == 0);
        return true;
    }
    public void addObserver(ReplicationStatusObserver o) { 
        synchronized (statusObservers) {
            statusObservers.add(o);
        }
    }
    public void removeObserver(ReplicationStatusObserver o) {
        synchronized (statusObservers) {
            statusObservers.remove(o);
        }
    }
    void notifyStatusObservers() {
        synchronized (statusObservers) {
            Iterator iter = statusObservers.iterator();
            while (iter.hasNext()) {
                ReplicationStatusObserver o = (ReplicationStatusObserver) iter.next();
                o.statusUpdate(this);
            }
        }
    }
    public int getStage() { return this.stage; }
    void setStage(int stage) {
        this.stage = stage;
        this.dirty = true;
    }

    // per ReplicationCount interface
    public long getTotal() { return this.total; }
    public long getFinished() { return this.finished; }
    public long getFailed() { return this.failed; }
    public long getTerminated() { return this.terminated; }
    public void addObserver( ReplicationCountObserver o ) {
        synchronized (countObservers) {
            countObservers.add(o);
        }
    }
    public void removeObserver(ReplicationCountObserver o) {
        synchronized (countObservers) {
            countObservers.remove(o);
        }
    }
    void notifyCountObservers() {
        synchronized (countObservers) {
            Iterator iter = countObservers.iterator();
            while (iter.hasNext()) {
                ReplicationCountObserver o = (ReplicationCountObserver) iter.next();
                o.countUpdate(this);
            }
        }
    }

    // per ReplicationOptions interface
    public String getIndexServiceUrl() { return this.indexUrl; }
    public String getTransferServiceUrl() { return this.reliableTransferUrl; }
    public String getRegistrationServiceUrl() { return this.registrationUrl; }
    public TransferOptions getTransferOptions() { return this.transferOptions; }
}
