/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.List;
import java.util.Iterator;

import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.ReplicationOptions;
import org.globus.replica.replicator.Source;
import org.globus.replica.replicator.Destination;

import org.globus.replica.replicator.util.DirtyFlag;

/**
 * INTERNAL USE ONLY.
 */
class ReplicationItemImpl implements ReplicationItem, DirtyFlag {

    // FIELDS //

    Object statusLock;
    boolean dirty;

    // per ReplicationItem interface
    String uri;
    int status;
    String errMsg;
    int priority;
    SourceImpl[] sources;
    DestinationImpl[] destinations;


    // CONSTRUCTORS //

    ReplicationItemImpl( String uri, Integer priority, String[] destinationUris ) {
        // Validate destinations
        if (destinationUris == null || destinationUris.length == 0)
            throw new IllegalArgumentException();
        int len = destinationUris.length;

        this.statusLock = new Object();

        this.uri = uri;
        this.status = ReplicationItem.PENDING;
        this.errMsg = null;
        this.priority = (priority != null) ? priority.intValue()
                            : ReplicationItem.MEDIUM;
        this.sources = new SourceImpl[0];
        this.destinations = new DestinationImpl[len];
        for (int i=0; i<len; i++) {
            this.destinations[i] = new DestinationImpl(destinationUris[i]);
        }
        this.dirty = true;
    }


    // METHODS //

    // per DirtyFlag interface
    public boolean isDirty() { return this.dirty; }
    public void setDirty() { this.dirty = true; }
    public void clearDirty() { this.dirty = false; }

    // per ReplicationItem interface
    public String getUri() { return uri; }
    public int getStatus() {
        synchronized (this.statusLock) {
            return status;
        }
    }
    public String getErrorMsg() {
        synchronized (this.statusLock) {
            return errMsg;
        }
    }
    public int getPriority() { return priority; }
    public java.util.Collection sources() {
        return java.util.Arrays.asList(this.sources);
    }
    void addSources(SourceImpl[] sources) {
        this.sources = sources;
    }
    public java.util.Collection destinations() {
        return java.util.Arrays.asList(this.destinations);
    }
    DestinationImpl findDestinationByUri(String uri) {
        for (int i=0; i<this.destinations.length; i++) {
            if (uri.equals(this.destinations[i].uri))
                return this.destinations[i];
        }
        return null;
    }

    // non-interface related

    /** Changes status and error message atomically. */
    void changeStatus(int status, String errMsg) {
        synchronized (this.statusLock) {
            this.status = status;
            this.errMsg = errMsg;
        }
    }
    void changeStatus(int status) {
        this.status = status;
    }

    /**
     * INTERNAL USE ONLY.
     * Need to carefully consider synchronization issues with this implementation.
     */
    private class SrcDestIterator implements java.util.Iterator {

        int curr;
        SrcDestImpl[] srcdests;

        SrcDestIterator(SrcDestImpl[] srcdests) {
            this.curr = 0;
            this.srcdests = srcdests;
        }

        public boolean hasNext() {
            return (this.srcdests != null && curr < this.srcdests.length);
        }

        public Object next() {
            if (this.srcdests == null || this.curr >= this.srcdests.length)
                throw new java.util.NoSuchElementException();
            return this.srcdests[this.curr++];
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
