/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Map;
import java.util.Iterator;
import java.net.URL;
import javax.xml.rpc.Stub;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.MessageElement;
import org.ietf.jgss.GSSCredential;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;

import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.wsrf.utils.PerformanceLog;
import org.globus.axis.util.Util;
import org.globus.util.I18n;
import org.globus.wsrf.utils.PerformanceLog;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.security.Constants;
import org.globus.wsrf.impl.security.util.AuthUtil;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.Authorization;
import org.globus.wsrf.impl.security.authorization.SelfAuthorization;

import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.rft.generated.service.ReliableFileTransferFactoryServiceAddressingLocator;
import org.globus.rft.generated.service.ReliableFileTransferServiceAddressingLocator;
import org.globus.rft.generated.CreateReliableFileTransferInputType;
import org.globus.rft.generated.CreateReliableFileTransferOutputType;
import org.globus.rft.generated.Start;
import org.globus.rft.generated.ReliableFileTransferFactoryPortType;
import org.globus.rft.generated.ReliableFileTransferPortType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.TransferStatusType;
import org.globus.rft.generated.TransferStatusTypeEnumeration;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;

import org.globus.replica.replicator.util.DelegatedCredential;

/**
 * Implements the <code>Work</code> required to process replication
 * requests.
 */
class TransferRequestWork implements commonj.work.Work {

    // CONSTANTS //

    private static Log logger =
        LogFactory.getLog(TransferRequestWork.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
                TransferRequestWork.class.getClassLoader());
    private static PerformanceLog perfLogger =
        new PerformanceLog(TransferRequestWork.class.getName());


    // STATICS //

    public static Object authType = Constants.GSI_TRANSPORT;
    public static Object authVal = Constants.SIGNATURE;
    public static Authorization authzVal = AuthUtil.getClientAuthorization("host");

    // FIELDS //

    private String rftUrl;
    private TransferRequestType request;
    private Map results;
    private DelegatedCredential delegatedCredential;
    private int transferWaitInt;
    private int transferWaitMax;
    private boolean transferWaitBackoff;
    private ReliableFileTransferFactoryPortType rftFactory = null;
    private ReliableFileTransferPortType rft = null;
    private EndpointReferenceType eprRft = null;
    private TransferStatusType[] transferStatusSet = null;

    boolean failed = false;
    String errmsg = null;


    // CONSTRUCTORS //

    TransferRequestWork(
            String rftUrl,
            TransferRequestType request,
            Map results,
            DelegatedCredential delegatedCredential,
            int transferWaitInt,
            int transferWaitMax,
            boolean transferWaitBackoff) {
        this.rftUrl = rftUrl;
        this.request = request;
        this.results = results;
        this.delegatedCredential = delegatedCredential;
        this.transferWaitInt = transferWaitInt;
        this.transferWaitMax = transferWaitMax;
        this.transferWaitBackoff = transferWaitBackoff;
    }


    // METHODS //

    public void release() {
        _clear();
        this.rftUrl = null;
        this.request = null; // TODO should try to clear this
        this.delegatedCredential = null;
        this.rftFactory = null;
        this.rft = null;
        this.eprRft = null;
        this.results = null; // DO NOT CLEAR! It's a shared data structure
        this.transferStatusSet = null;
    }

    /**
     * INTERNAL USE ONLY.
     * Clears the internal data structures.
     */
    private void _clear() {
        if (this.transferStatusSet == null)
            return;

        for (int i=0; i<this.transferStatusSet.length; i++)
            this.transferStatusSet[i] = null;
    }

 
    public boolean isDaemon() {
        return true;
    }

    /**
     * INTERNAL USE ONLY.
     * Main body of the work object.
     */
    public void run() {
        try {
            _locateTransferFactory();

            _createTransfer();

            _startTransfer();

            boolean done = _waitForTransfer();

            if (done) {
                _getTransferStatusSet();

                _updateResults();
            }
            else {
                this.failed = true;
                // TODO get rft fault message
            }
        }
        catch (Exception ex) {
            logger.warn("Failed, errmsg: " + ex.getMessage());
            logger.debug(ex.getMessage(),ex);
            this.failed = true;
            this.errmsg = ex.getMessage();
        }
    }

   /**
     * INTERNAL USE ONLY.
     * Locates the trasfer factory.
     */
    private void _locateTransferFactory() throws Exception {
        logger.trace("_locateTransferFactory");
        // create rft factory instance
        ReliableFileTransferFactoryServiceAddressingLocator locator =
            new ReliableFileTransferFactoryServiceAddressingLocator();
        URL endpoint = new URL(this.rftUrl);
        this.rftFactory = locator
                .getReliableFileTransferFactoryPortTypePort(endpoint);
        _initStubSecurity((Stub)this.rftFactory);
    }

    /**
     * INTERNAL USE ONLY.
     * Creates the transfer resource.
     */
    private void _createTransfer() throws Exception {
        logger.trace("_createTransfer");
        perfLogger.start();
        // Create input message
        CreateReliableFileTransferInputType input =
            new CreateReliableFileTransferInputType();
        input.setTransferRequest(this.request);

        // Create RFT Resource
        CreateReliableFileTransferOutputType response = this.rftFactory
            .createReliableFileTransfer(input);
        this.eprRft = response.getReliableTransferEPR();
        perfLogger.stop(ReplicationConstants.PERFLOG_CREATETRANSFERREQ);
        logger.debug("Created RFT Resource, eprRft: " + this.eprRft.toString());
    }

    /**
     * INTERNAL USE ONLY.
     * Starts the transfer.
     */
    private void _startTransfer() throws Exception {
        logger.trace("_startTransfer");
        // Start the transfer
        ReliableFileTransferServiceAddressingLocator rftLocator =
            new ReliableFileTransferServiceAddressingLocator();
        this.rft = rftLocator.getReliableFileTransferPortTypePort(this.eprRft);
        _initStubSecurity((Stub)this.rft);
        this.rft.start(new Start());
    }

    /**
     * INTERNAL USE ONLY.
     * Waits for transfer request to finish, checks status of finished transfer, and
     * returns boolean indicator of done/failed. Waits according to wait internval,
     * max, and backoff settings.
     */
    private boolean _waitForTransfer() throws Exception {
        logger.trace("_waitForTransfer");

        // Check and wait for transfer completion
        int totalWait = 0;
        boolean done = false;
        boolean cancelled = false;
        try {
            int i = 0;
            do {
                // Wait if not first iteration
                if (i>0) {
                    // Account for exponential backoff
                    int currentWait;
                    if (this.transferWaitBackoff)
                        currentWait = i*i*this.transferWaitInt;
                    else
                        currentWait = this.transferWaitInt;

                    // Trim to max wait time
                    if (this.transferWaitMax != 0 &&
                            totalWait+currentWait > this.transferWaitMax)
                        currentWait = (this.transferWaitMax > totalWait) ?
                            this.transferWaitMax - totalWait : 0;

                    // Convert to millis, Wait, and update totalWait time
                    logger.trace("Waiting for RFT transfer to finish, wait: "
                            + currentWait);
                    synchronized (this) {
                        this.wait(currentWait*1000);
                    }
                    totalWait += currentWait;
                }

                // Get RFT RP for RequestStatus
                GetResourcePropertyResponse getRPResponse = this.rft.getResourceProperty(
                        RFTConstants.REQUEST_STATUS_RESOURCE);
                MessageElement msgElem[] = getRPResponse.get_any();
                RequestStatusType status = (RequestStatusType) ObjectDeserializer.toObject(
                        msgElem[0], RequestStatusType.class);

                // Check if Done or Failed
                RequestStatusTypeEnumeration statusEnum = status.getRequestStatus();
                logger.debug("RFT RequestStatus, statusEnum: " + statusEnum.toString());
                if ((statusEnum == RequestStatusTypeEnumeration.Done)
                        || (statusEnum == RequestStatusTypeEnumeration.Failed)) {
                    done = true;
                } else if (statusEnum == RequestStatusTypeEnumeration.Cancelled) {
                    cancelled = true;
                } else {
                    // Else -- status is either Active or Pending
                    // Increment loop counter
                    i++;
                }
            } while (!done && !cancelled && (this.transferWaitMax==0 ||
                         totalWait<this.transferWaitMax));

        } catch (Exception ex) {
            // TODO can consider checking again or considering this a transient error
            //      For now, just throw it and count it as a persistent error
            throw ex;
        }

        // Warn user if maximum wait time exceeded.
        if (!done && !cancelled) {
            String msg = "Transfer status could not be determined. Maximum wait time reached. Waited: "
                                    + totalWait;
            // TODO ideally, should add transfer id to above message
            logger.warn(msg);
            throw new Exception(msg);
        }

        return done;
    }

    /**
     * INTERNAL USE ONLY.
     * Starts the transfer.
     */
    private void _getTransferStatusSet() throws Exception {
        logger.trace("_getTransferStatusSet");
        org.globus.rft.generated.GetStatusSetResponse response =
            this.rft.getStatusSet(new org.globus.rft.generated.GetStatusSet());
        this.transferStatusSet = response.getTransferStatusSet();
    }


    /**
     * INTERNAL USE ONLY.
     * Updates the shared results data structure.
     */
    private void _updateResults() throws Exception {
        logger.trace("_updateResults");

        // Create local results structure mapping
        int length = this.transferStatusSet.length;
        Map localResults = new java.util.HashMap(length);
        for (int i=0; i<length; i++) {
            localResults.put(this.transferStatusSet[i].getDestinationUrl(),
                    this.transferStatusSet[i]);
        }

        // Put local results into shared results structure
        synchronized (this.results) {
            this.results.putAll(localResults);
        }

        // Clear local results
        localResults.clear();
    }

    // HELPERS //

    private void _initStubSecurity(Stub stub) throws Exception {
        stub._setProperty(Constants.CLIENT_DESCRIPTOR,
                          _createClientSecurityDescriptor());
    }

    private ClientSecurityDescriptor _createClientSecurityDescriptor() throws Exception {
        ClientSecurityDescriptor desc = new ClientSecurityDescriptor();

        // set authentication
        if (authType.equals(Constants.GSI_SEC_MSG)) {
            desc.setGSISecureMsg((Integer)authVal);
        } else if (authType.equals(Constants.GSI_SEC_CONV)) {
            desc.setGSISecureConv((Integer)authVal);
        } else if (authType.equals(Constants.GSI_TRANSPORT)) {
            desc.setGSITransport((Integer)authVal);
            Util.registerTransport();
        }

        // set authorization
        desc.setAuthz(authzVal);

        // set credential
        desc.setGSSCredential(
                   new GlobusGSSCredentialImpl(
                         this.delegatedCredential.getCredential(),
                         GSSCredential.INITIATE_AND_ACCEPT));

        return desc;
    }
}

