/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.service;

import java.rmi.RemoteException;
import java.util.Calendar;
import javax.naming.Context;
import javax.naming.InitialContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;
import org.globus.wsrf.utils.PerformanceLog;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.oasis.wsrf.lifetime.UnableToSetTerminationTimeFaultType;

import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.utils.AddressingUtils;
import org.globus.wsrf.security.SecurityManager;

import org.globus.replica.replicator.generated.CreateReplicator;
import org.globus.replica.replicator.generated.CreateReplicatorResponse;
import org.globus.replica.replicator.generated.Start;
import org.globus.replica.replicator.generated.StartResponse;
import org.globus.replica.replicator.generated.Stop;
import org.globus.replica.replicator.generated.StopResponse;
import org.globus.replica.replicator.generated.Suspend;
import org.globus.replica.replicator.generated.SuspendResponse;
import org.globus.replica.replicator.generated.Resume;
import org.globus.replica.replicator.generated.ResumeResponse;
import org.globus.replica.replicator.generated.BaseRequestType;
import org.globus.replica.replicator.generated.FindItemsResponse;
import org.globus.replica.replicator.generated.FindItems;
import org.globus.replica.replicator.generated.CreateReplicatorFaultType;
import org.globus.replica.replicator.generated.InvalidStateFaultType;
import org.globus.replica.replicator.generated.RequestBodyMissingFaultType;
import org.globus.replica.replicator.generated.CredentialEprMissingFaultType;


/**
 * The Replicator Service implementation.
 */
public class ReplicationServiceImpl {

    private static Log logger = LogFactory
            .getLog(ReplicationServiceImpl.class.getName());
    private static I18n i18n = 
        I18n.getI18n("org.globus.replica.replicator.errors", 
                ReplicationServiceImpl.class.getClassLoader());
    private static PerformanceLog perfLog = 
        new PerformanceLog(ReplicationServiceImpl.class.getName());

    /**
     * Implementation of the <code>createReplicator</code> operation.
     *
     * @param parameters <code>CreateReplicator</code> message input.
     * @return Returns an EPR to the newly created Replicator resource
     * @throws RemoteException Indicates that in internal error was encountered.
     * @throws CreateReplicatorFaultType May indicate that the request input
     *      is invalid, such as a missing delegated credential EPR or missing
     *      request body.
     */
    public CreateReplicatorResponse createReplicator(CreateReplicator parameters)
        throws java.rmi.RemoteException, CreateReplicatorFaultType {

        logger.debug("createReplicator");
        perfLog.start();

        // Get termination time from request and ensure that it is valid
        Calendar currentTime = Calendar.getInstance();
        Calendar initialTerminationTime = parameters.getInitialTerminationTime();
        if (initialTerminationTime != null &&
                initialTerminationTime.getTime().before(currentTime.getTime())) {
            String msg = "Termination time set earlier than current time, term time: "
                + initialTerminationTime.toString();
            logger.error(msg);
            throw new UnableToSetTerminationTimeFaultType();
        }

        // Get Request
        BaseRequestType request = null;
        if ( ((request = parameters.getRequestFileRequest()) == null) &&
             ((request = parameters.getItemSequenceRequest()) == null) )
            throw new RequestBodyMissingFaultType();

        // Check for delegated credential EPR
        if (request.getCredentialEPR() == null )
            throw new CredentialEprMissingFaultType();

        // Get the caller's user subject and local username
        SecurityManager securityManager = SecurityManager.getManager();
        String userSubject = securityManager.getCaller();
        String localUsername = securityManager.getLocalUsernames()[0];

        // Create the Replicator resource
        ResourceContext ctx;
        ResourceKey key;
        try {
            ctx = ResourceContext.getResourceContext();
            ReplicatorHome home = (ReplicatorHome) ctx.getResourceHome();
            key = home.create(currentTime, initialTerminationTime, request,
                    userSubject, localUsername);
        }
        catch (Exception e) {
            logger.error(i18n.getMessage("resourceCreateErr"), e);
            throw new RemoteException(i18n.getMessage("resourceCreateErr"), e);
        }

        // Create resultant EPR
        EndpointReferenceType epr = null;
        try {
            epr = AddressingUtils.createEndpointReference(ctx, key);
        }
        catch(Exception e) {
            logger.error(i18n.getMessage("resourceCreateErr"), e);
            throw new RemoteException(i18n.getMessage("resourceCreateErr"), e);
        }

        perfLog.stop(ReplicatorConstants.PERFLOG_CREATEREPLICATOR);
        return new CreateReplicatorResponse(epr);
    }

    /**
     * Implements the <code>start</code> operation.
     *
     * @param parameters Unused.
     * @return Unused 
     * @throws RemoteException Indicates that an internal error occurred.
     *      Users should consult log file for more information.
     * @throws InvalidStateFaultType Indicates that the operation cannot
     *      be performed, because the Replicator resource is not in a valid
     *      state to permit the operation.
     */
    public StartResponse start(Start parameters) throws java.rmi.RemoteException,
        org.globus.replica.replicator.generated.InvalidStateFaultType {

        logger.debug("start");
        perfLog.start();

        ReplicatorResource replicator = null;
        try {
            replicator = (ReplicatorResource)
                ResourceContext.getResourceContext().getResource();
        }
        catch(RemoteException e) {
            logger.error(e);
            throw e;
        }
        catch(Exception e) {
            logger.error(e);
            throw new RemoteException(e.getMessage(), e);
        }

        synchronized (replicator) {
            replicator.start();
        }

        perfLog.stop(ReplicatorConstants.PERFLOG_START);
        return new StartResponse();
    }

    /**
     * Implements the <code>stop</code> operation.
     *
     * @param parameters Unused.
     * @return Unused 
     * @throws RemoteException Indicates that an internal error occurred.
     *      Users should consult log file for more information.
     * @throws InvalidStateFaultType Indicates that the operation cannot
     *      be performed, because the Replicator resource is not in a valid
     *      state to permit the operation.
     */
    public StopResponse stop(Stop parameters) throws java.rmi.RemoteException,
        org.globus.replica.replicator.generated.InvalidStateFaultType {

        logger.debug("stop");
        perfLog.start();

        ReplicatorResource replicator = null;
        try {
            replicator = (ReplicatorResource)
                ResourceContext.getResourceContext().getResource();
        }
        catch(RemoteException e) {
            logger.error(e);
            throw e;
        }
        catch(Exception e) {
            logger.error(e);
            throw new RemoteException(e.getMessage(), e);
        }

        synchronized (replicator) {
            replicator.stop();
        }

        perfLog.stop(ReplicatorConstants.PERFLOG_STOP);
        return new StopResponse();
    }

    /**
     * Implements the <code>suspend</code> operation.
     *
     * @param parameters Unused.
     * @return Unused 
     * @throws RemoteException Indicates that an internal error occurred.
     *      Users should consult log file for more information.
     * @throws InvalidStateFaultType Indicates that the operation cannot
     *      be performed, because the Replicator resource is not in a valid
     *      state to permit the operation.
     */
    public SuspendResponse suspend(Suspend parameters) throws java.rmi.RemoteException,
        org.globus.replica.replicator.generated.InvalidStateFaultType {

        logger.debug("suspend");
        perfLog.start();

        ReplicatorResource replicator = null;
        try {
            replicator = (ReplicatorResource)
                ResourceContext.getResourceContext().getResource();
        }
        catch(RemoteException e) {
            logger.error(e);
            throw e;
        }
        catch(Exception e) {
            logger.error(e);
            throw new RemoteException(e.getMessage(), e);
        }

        synchronized (replicator) {
            replicator.suspend();
        }

        perfLog.stop(ReplicatorConstants.PERFLOG_SUSPEND);
        return new SuspendResponse();
    }

    /**
     * Implements the <code>resume</code> operation.
     *
     * @param parameters Unused.
     * @return Unused 
     * @throws RemoteException Indicates that an internal error occurred.
     *      Users should consult log file for more information.
     * @throws InvalidStateFaultType Indicates that the operation cannot
     *      be performed, because the Replicator resource is not in a valid
     *      state to permit the operation.
     */
    public ResumeResponse resume(Resume parameters) throws java.rmi.RemoteException,
        org.globus.replica.replicator.generated.InvalidStateFaultType {

        logger.debug("resume");
        perfLog.start();

        ReplicatorResource replicator = null;
        try {
            replicator = (ReplicatorResource)
                ResourceContext.getResourceContext().getResource();
        }
        catch(RemoteException e) {
            logger.error(e);
            throw e;
        }
        catch(Exception e) {
            logger.error(e);
            throw new RemoteException(e.getMessage(), e);
        }

        synchronized (replicator) {
            replicator.resume();
        }

        perfLog.stop(ReplicatorConstants.PERFLOG_RESUME);
        return new ResumeResponse();
    }


    public FindItemsResponse findItems(FindItems parameters) throws 
        java.rmi.RemoteException, 
        org.globus.replica.replicator.generated.InternalErrorFaultType {

        logger.debug("findItems");
        ReplicatorResource replicator = null;
        try {
            replicator = (ReplicatorResource)
                ResourceContext.getResourceContext().getResource();
        }
        catch(RemoteException e) {
            logger.error(e);
            throw e;
        }
        catch(Exception e) {
            logger.error(e);
            throw new RemoteException(e.getMessage(), e);
        }

        synchronized (replicator) {
            return replicator.findItems(parameters);
        }
    }




    // TO BE IMPLEMENTED
    // Examples taken from RFT.

    /*
     * To query the status of a transfer request
     * 
     * @param params
     * @return GetStatusResponse
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     */
     /*
    public GetStatusResponse getStatus(GetStatus params)
            throws RemoteException, ResourceUnknownFaultType,
            RFTDatabaseFaultType {
        String sourceUrl = params.getSourceUrl();
        TransferStatusType statusType = null;
        logger.debug("Finding status of " + sourceUrl);
        int requestId = -1;
        try {
            requestId = this.getRequestId();
        } catch (Exception e) {
            throw new RemoteException(
                    i18n.getMessage("getStatusErr"), e);
        }
        if (this.resourceKey == null || this.resource == null
                || requestId == -1) {
            logger.error(i18n.getMessage("resourceNotFoundErr"));
            throw new ResourceUnknownFaultType();
        }
        logger.debug("for the resource " + this.resourceKey);
        try {
            if (requestId != -1) {
                statusType = 
                    this.dbAdapter.getStatus(requestId, sourceUrl);
            } else {
                throw new ResourceUnknownFaultType();
            }
        } catch (RftDBException e) {
            throw (RFTDatabaseFaultType)
            FaultUtil.makeFault(RFTDatabaseFaultType.class,requestId, 
                    i18n.getMessage("dbRetrieveErr"),e);
        }

        if (statusType == null) {
            throw new RemoteException(i18n.getMessage("getStatusErr"));
        }
        return new GetStatusResponse(statusType);
    }
    */

    /*
     * Get an array of status of transfers starting from initial
     * 
     * @param params
     * @return GetStatusSetResponse
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     */
     /*
    public GetStatusSetResponse getStatusSet(GetStatusSet params)
            throws RemoteException, ResourceUnknownFaultType,
            RFTDatabaseFaultType {
        GetStatusSetResponse response = null;
        TransferStatusType statusSet[] = null;
        Vector statusSetVector = null;
        int initial, offset, requestId = -1;
        initial = params.getFrom();
        offset = params.getOffset();
        logger.debug("GetStatus of " + offset + " transfers starting from "
                + initial);
        try {
            requestId = this.getRequestId();
        } catch (Exception e) {
            throw new RemoteException(
                    i18n.getMessage("getStatusSetErr"), e);
        }
        if (this.resourceKey == null || this.resource == null
                || requestId == -1) {
            logger.error(i18n.getMessage("resourceNotFoundErr"));
            throw new ResourceUnknownFaultType();
        }
        try {
            statusSetVector = this.dbAdapter.getStatusSet(requestId, initial,
                    offset);
            int statusSetSize = statusSetVector.size();
            if (statusSetSize > 0) {
                statusSet = new TransferStatusType[statusSetSize];
                for (int i = 0; i < statusSetSize; i++) {
                    statusSet[i] = (TransferStatusType) 
                    statusSetVector.elementAt(i);
                }
                response = new GetStatusSetResponse();
                response.setTransferStatusSet(statusSet);
            } else {
                logger.error(i18n.getMessage("getStatusSetErr"));
            }
        } catch (RftDBException e) {
            throw (RFTDatabaseFaultType)
            FaultUtil.makeFault(RFTDatabaseFaultType.class,requestId, 
                    i18n.getMessage("dbRetrieveErr"),e);
        }
        if (statusSet == null) {
            throw new RemoteException(i18n.getMessage(
                    "getStatusSetErr"));
        }
        return response;
    }
    */

    /*
     * cancels transfers in a request
     * 
     * @param params
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     * @return CancelResponse
     */
     /*
    public CancelResponse cancel(Cancel params) throws RemoteException,
        ResourceUnknownFaultType, RFTDatabaseFaultType {
        CancelResponse response = null;
        int fromId = -1;
        int toId = -1;
        int requestId = -1;
        int transferId = -1;
        fromId = params.getFromId();
        toId = params.getToId();
        logger.debug("Invoked cancel for transfer from " + fromId + " to "
                + toId);
        try {
            requestId = this.getRequestId();
            if (requestId == -1) {
                throw new ResourceUnknownFaultType();
            }
        } catch (Exception e) {
            throw new RemoteException(
                    i18n.getMessage("unableToCancelErr"), e);
        }
        try {
            
            transferId = this.dbAdapter.getFirstTransferId(requestId);
            if (fromId != -1 && toId != -1) {
                fromId = transferId + fromId - 1;
                toId = transferId + toId - 1;
            }
            logger.debug("Cancelling from:" + fromId + " to " + toId);
            TransferWork.stopActiveTransfers(requestId, fromId, toId);
            this.dbAdapter.updateToCancelled(requestId, fromId, toId);
        } catch (RftDBException e) {
            logger.error(i18n.getMessage("dbRetrieveErr"), e);
            throw (RFTDatabaseFaultType)
            FaultUtil.makeFault(RFTDatabaseFaultType.class,requestId, 
                    i18n.getMessage("dbRetrieveErr"),e);
            
        }

        response = new CancelResponse();
        return response;

    }
    */
    
    /*
     * Util method to get the requestId from the database for a resource
     * 
     * @return requestId
     * @throws Exception
     */
     /*
    private int getRequestId() throws Exception {
        int requestId = -1;
        this.resourceContext = ResourceContext.getResourceContext();
        this.resourceHome = this.resourceContext.getResourceHome();
        this.resourceKey = this.resourceContext.getResourceKey();
        requestId = Integer.parseInt((String) this.resourceKey.getValue());
        logger.debug("ResourceId :" + requestId);
        return requestId;
    }
    */

    /*
    private int getRequestIdFromKey() {
        return Integer.parseInt((String) this.resourceKey.getValue());
    }
    */
    
    /*
    private static RFTDatabaseOptions dbConfiguration;
    {
        try {
            ResourceContext ctx = ResourceContext.getResourceContext();
            Context initialContext = new InitialContext();
            String lookupString = Constants.JNDI_SERVICES_BASE_NAME
            + ctx.getService() + "/dbConfiguration";
            logger.debug(lookupString);
            dbConfiguration = (RFTDatabaseOptions) initialContext
                    .lookup(lookupString);
            logger.debug("Connection url: " 
                    + dbConfiguration.getConnectionString());
        } catch (Exception e) {
            logger.error(i18n.getMessage("dbConfigFileErr") +
                    e.getMessage(), e);
            throw new RuntimeException(
                    i18n.getMessage("dbConfigFileErr"), e);
        }
    }
    
    private static RFTConfiguration serviceConfiguration;
    {
        try {
            ResourceContext ctx = ResourceContext.getResourceContext();
            Context initialContext = new InitialContext();
            String lookupString = Constants.JNDI_SERVICES_BASE_NAME
                + ctx.getService() + "/configuration";
            logger.debug("looking up: " + lookupString);
            serviceConfiguration = (RFTConfiguration) initialContext
                    .lookup(lookupString);
            logger.debug("backOff time: " 
                    + serviceConfiguration.getBackOff());
            logger.debug("maximum active allowed:" 
                    + serviceConfiguration.getMaxActiveAllowed());
        } catch (Exception e) {
            logger.error(i18n.getMessage(
                    "serviceConfigErr") + e.getMessage(), e);
            throw new RuntimeException(
                    i18n.getMessage(
                    "serviceConfigErr") + e.getMessage(), e);
        }
    }
    */
    
    
    
    /*
     * @return the factory configuration object
     */
     /*
    public static RFTDatabaseOptions getDatabaseConfiguration() {
        return dbConfiguration;
    }
    */
    
    /*
     * @return service configuration
     */
     /*
    public static RFTConfiguration getRFTConfiguration() {
        return serviceConfiguration;
    }
    */
}

