/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.service;

import java.util.List;
import java.io.File;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.net.HttpURLConnection;
import javax.naming.InitialContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.replica.replicator.ReplicationFactory;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.CreateException;

import org.globus.replica.replicator.generated.RequestFileReplicationRequestType;

class RequestFileUtil {

    private static Log logger =
        LogFactory.getLog(RequestFileUtil.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicatorConstants.MESSAGE_RESOURCE_BUNDLE,
                RequestFileUtil.class.getClassLoader());

    private static final char DELIM = '\t';
    private static final String COMMENT = "#";
    private static File requestfileDir = null;
    static {
        try {
            // load replicator configuration
            InitialContext ctx = new InitialContext();
            ReplicatorConfiguration config = (ReplicatorConfiguration) ctx.lookup(
                    ReplicatorConstants.CONFIGURATION );

            // lookup request file settings
            String s = config.getRequestfileDir();
            if (s!=null && s.length()>0 && !s.equalsIgnoreCase("null")) {
                try {
                    RequestFileUtil.requestfileDir = new File(s);
                    if (!requestfileDir.exists()) {
                        logger.warn("Path for temp request file storage does not exist: " + s);
                        RequestFileUtil.requestfileDir = null;
                    }
                    else if (!requestfileDir.isDirectory()) {
                        logger.warn("Path for temp request file storage is not a directory: " + s);
                        RequestFileUtil.requestfileDir = null;
                    }
                }
                catch (NullPointerException ignored) {
                }
            }
        }
        catch (javax.naming.NamingException nex) {
            logger.warn("Temp request file dir setting does not exist");
        }
    }



    /**
     * Processes a replication request file.
     */
    static List processRequestFile(
            ReplicationFactory factory,
            RequestFileReplicationRequestType request)
        throws java.net.URISyntaxException,
               java.net.MalformedURLException,
               java.io.FileNotFoundException,
               java.io.IOException,
               CreateException {

        logger.debug("processRequestFile");
        List items = null;

        // Retrieve request file to local file
        String localFilename = _retrieveRequestFile(request.getRequestFileUri());
        File f = new File(localFilename);

        java.io.LineNumberReader reader = null;
        try {
            // Open file
            reader = new java.io.LineNumberReader(new java.io.FileReader(f));

            // Read items
            items = new java.util.LinkedList();
            for (String line = reader.readLine(); line != null; line = reader.readLine()) {

                // Skip comments
                if (line.startsWith(COMMENT))
                    continue;

                // Parse lfn and pfn
                int start = 0;
                int end = line.indexOf(DELIM);
                String lfn = line.substring(start,end).trim();
                start = end+1;
                String pfn = line.substring(start).trim();

                // Create replication item
                ReplicationItem item = factory.createReplicationItem(
                        lfn,
                        null,
                        new String[] {pfn});

                // Add to collection
                items.add(item);
            }
        }
        finally {
            if (reader != null) try { reader.close(); } catch (Exception ignored) { }
        }

        return items;
    }

    private static String _retrieveRequestFile(String requestFileUri)
        throws java.net.URISyntaxException,
               java.net.MalformedURLException,
               java.io.IOException,
               IllegalArgumentException {

        logger.debug("_retrieveRequestFile");

        // Validate URI
        URI uri = new URI(requestFileUri);
        if (!uri.isAbsolute())
            throw new IllegalArgumentException("Request file URI must be absolute.");

        // Check if "file:"
        String uriScheme = uri.getScheme();
        if (uriScheme == null || "file".equals(uriScheme))
            return uri.getRawPath();
        else if ("http".equals(uriScheme) || "https".equals(uriScheme)
                    || "ftp".equals(uriScheme))
            return _moveRequestFileViaUrlConnection(uri.toURL());
        else if ("gsiftp".equals(uriScheme))
            return _moveRequestFileViaGridFtp(uri.toURL());
        else
            throw new IllegalArgumentException("Scheme not supported. Scheme: " + uriScheme);
    }

    private static String _moveRequestFileViaUrlConnection(URL url)
        throws java.io.IOException {

        logger.debug("_moveRequestFileViaUrlConnection");
        String localFilename;

        URLConnection urlconn = null;
        HttpURLConnection http = null;
        InputStream in = null;
        OutputStream out = null;
        try {
            // Open connection
            urlconn = url.openConnection();

            // Set options for connection
            urlconn.setUseCaches(false);
            if (urlconn instanceof HttpURLConnection) {
                http = (HttpURLConnection) urlconn;
                http.setRequestMethod("GET");
                http.setInstanceFollowRedirects(true);
            }

            // Open stream
            urlconn.connect();
            in = new BufferedInputStream(urlconn.getInputStream());

            // Create output file
            File fout = File.createTempFile("requestfile", null,
                    RequestFileUtil.requestfileDir);
            fout.deleteOnExit();
            localFilename = fout.getAbsolutePath();
            out = new BufferedOutputStream(new FileOutputStream(fout));

            // Read from stream
            byte b[] = new byte[1024]; // 1k buffer
            int cb = 0;
            while ((cb = in.read(b)) >= 0) {
                out.write(b, 0, cb);
            }
        }
        finally {
            if (out != null) {
                out.flush();
                out.close();
            }
            if (in != null) {
                in.close();
            }
            if (http != null) {
                http.disconnect();
            }
        }

        return localFilename;
    }

    private static String _moveRequestFileViaGridFtp(URL url) {

        // Not read to support this yet
        throw new IllegalArgumentException("GridFTP protocol not yet supported for request file retrieval.");
    }
}
