/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.util;

import java.io.File;
import javax.security.auth.Subject;
import javax.naming.InitialContext;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

import org.globus.gsi.GlobusCredential;
import org.globus.gsi.jaas.GlobusPrincipal;

import org.globus.delegation.DelegationException;
import org.globus.delegation.DelegationListener;
import org.globus.delegation.DelegationUtil;

import org.globus.replica.replicator.ReplicationCredentials;

import org.globus.replica.replicator.service.ReplicatorConfiguration; // TODO eliminate this dependency???
import org.globus.replica.replicator.service.ReplicatorConstants; // TODO eliminate this dependency???


public class DelegatedCredential implements ReplicationCredentials,
        DelegationListener, java.io.Serializable {

    private static final String PROXYPREFIX = "x509up_";
    private static final String CHANGEPERMCMD = "chmod 600 ";

    private static Log logger = LogFactory.getLog(DelegatedCredential.class.getName());
    private static I18n i18n = I18n.getI18n("org.globus.replica.replicator.errors",
            DelegatedCredential.class.getClassLoader());
    private static File proxyfileDir = null;
    private static String proxyfileChangePermsCmd = CHANGEPERMCMD;
    static {
        try {
            // load replicator configuration
            InitialContext ctx = new InitialContext();
            ReplicatorConfiguration config = (ReplicatorConfiguration) ctx.lookup(
                    ReplicatorConstants.CONFIGURATION );

            // lookup proxyfile settings
            String s = config.getProxyfileDir();
            if (s!=null && s.length() > 0 && !s.equalsIgnoreCase("null")) {
                try {
                    DelegatedCredential.proxyfileDir = new File(s);
                    if (!proxyfileDir.exists()) {
                        logger.warn("Path for temp proxy storage does not exist: " + s);
                        proxyfileDir = null;
                    }
                    else if (!proxyfileDir.isDirectory()) {
                        logger.warn("Path for temp proxy storage is not a directory: " + s);
                        proxyfileDir = null;
                    }
                }
                catch (NullPointerException ignored) {
                }
            }

            s = config.getProxyfileChangePermsCmd();
            if (s!=null && s.length() > 0 && !s.equalsIgnoreCase("null"))
                DelegatedCredential.proxyfileChangePermsCmd = s + " ";
        }
        catch (javax.naming.NamingException nex) {
        }
    }


    private boolean initialized = false;
    private String listenerId = null;
    private String localUsername = null;
    private String userSubject = null;
    transient private GlobusCredential credential = null;
    transient private EndpointReferenceType epr = null;
    transient private String proxyFilename = null;

    public DelegatedCredential(EndpointReferenceType epr,
            String localUsername,
            String userSubject)
        throws DelegationException
    {
        this.localUsername = localUsername;
        this.userSubject = userSubject;
        this.epr = epr;
        Subject subject = new Subject();
        subject.getPrincipals().add(new GlobusPrincipal(userSubject));
        DelegationUtil.registerDelegationListener(epr, this, subject);
        _initialize();
    }

    private void _initialize() throws DelegationException
    {
        synchronized (this) {
            while (!this.initialized) {  // TODO limit this to a MAXWAIT value
                try {
                    wait(); // TODO I really hate this
                } catch (InterruptedException intex) {
                    logger.debug("Interrupted while waiting for credential initializion: " + intex.getMessage());
                }
            }
        }
    }

    /**
     * Releases resources of this object. This object should not be used after
     * this call.
     */
    public void release() {
        try {
            // Remove delegation listener
            DelegationUtil.removeDelegationListener(this.epr, this.listenerId);

            // Delete proxy file
            File f = new File(this.proxyFilename);
            if (!f.delete())
                logger.warn("Failed to delete proxy file, filename: " + this.proxyFilename);

            // Null out references
            this.initialized = false;
            this.listenerId = null;
            this.localUsername = null;
            this.userSubject = null;
            this.credential = null;
            this.epr = null;
            this.proxyFilename = null;
        }
        catch (Exception e) {
            logger.warn(e.getMessage());
            logger.debug(e.getMessage(),e);
        }
    }

    public void setId(String listenerId) { this.listenerId = listenerId; }

    public String getId() { return this.listenerId; }

    public void setCredential(GlobusCredential credential) {
        logger.debug("Credential initialized: " + credential.toString());
        this.credential = credential;
        _saveCredential();
        synchronized (this) {
            this.initialized = true;
            this.notifyAll();
        }
    }

    public GlobusCredential getCredential() {
        return this.credential;
    }

    public void credentialDeleted() {
        logger.debug("Credential deleted");
        this.credential = null;
    }

    public EndpointReferenceType getEPR() {
        return this.epr;
    }

    public String getLocalUsername() {
        return this.localUsername;
    }

    public String getUserSubject() {
        return this.userSubject;
    }

    public String getProxyFilename() {
        return this.proxyFilename;
    }

    private void _saveCredential() {
        java.io.FileOutputStream out = null;
        try {
            File f = File.createTempFile(PROXYPREFIX,null,DelegatedCredential.proxyfileDir);
            String filename = f.getAbsolutePath();
            logger.debug("Saving credential to: " + filename);
            out = new java.io.FileOutputStream(f);
            this.credential.save(out);
            this.proxyFilename = filename;
            _changeFilePerms();
            f.deleteOnExit();
        } catch (java.io.IOException ioex) {
            logger.error("Failed to save proxy -- " + ioex.getMessage());
        } finally {
            if (out != null)
                try { out.close(); } catch (java.io.IOException ignored) {}
        }
    }

    private void _changeFilePerms() {
        try {
            String command = DelegatedCredential.proxyfileChangePermsCmd + this.proxyFilename;
            Runtime r = Runtime.getRuntime();
            Process p = r.exec(command);
            p.waitFor();
            // TODO could check exit code here
        } catch (java.io.IOException ioex) {
            logger.error("Failed to change perms on proxy file -- " + ioex.getMessage());
        } catch (Exception ex) {
            logger.error("Failed to change perms on proxy file -- " + ex.getMessage());
        }
    }
}
