/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.transfer.reliable.service;

import java.io.IOException;
import java.net.ConnectException;
import java.net.MalformedURLException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.ftp.GridFTPClient;
import org.globus.ftp.vanilla.Reply;
import org.globus.ftp.exception.ServerException;
import org.globus.gsi.gssapi.auth.IdentityAuthorization;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.util.GlobusURL;
import org.globus.util.I18n;
import org.ietf.jgss.GSSCredential;

/**
 */
public class DeleteClient {
    private static Log logger = 
        LogFactory.getLog(DeleteClient.class.getName());
    private static I18n i18n =
        I18n.getI18n("org.globus.transfer.reliable.errors", 
                DeleteClient.class.getClassLoader());

    private String filePath = null;

    private GlobusURL fileURL;

    private GridFTPClient gridFtpHost = null;

    private String host = null;

    private int port = 0;

    private int status = -1;

    private GSSCredential credential = null;

    private String subjectName = null;

    private URLExpander expander = null;
    
    private int requestId = 0;

    /**
     * @param url
     * @param subjectName
     * @param proxy
     * @param userName
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     * 
     */
    public DeleteClient(String url, String subjectName, 
            GSSCredential credential,
            String userName) throws RftException, RftTransientException,
            RftDBException {
        try {
            this.fileURL = new GlobusURL(url);
            if(!this.fileURL.getProtocol().equalsIgnoreCase("gsiftp")) {
                throw new RftException(
                    i18n.getMessage(
                        "thirdPartyOnlyErr",
                        new String[] { this.fileURL.getURL() }));
            }
            this.host = this.fileURL.getHost();
            this.port = this.fileURL.getPort();
            this.filePath = "/" + this.fileURL.getPath();
            this.gridFtpHost = new GridFTPClient(this.host,
                    this.port);
            if ((subjectName != null) && 
                    !(subjectName.equalsIgnoreCase("null"))) {
                this.subjectName = subjectName;
                this.gridFtpHost.setAuthorization(new IdentityAuthorization(
                        this.subjectName));
            } else {
                if(logger.isDebugEnabled()) {
                    logger.debug("Subject name is null in DeleteClient");
                }
            }
            setCredential(credential);
            if (userName != null && !userName.equalsIgnoreCase("null")) {
                this.gridFtpHost.authenticate(this.credential, userName);
            } else {
                this.gridFtpHost.authenticate(this.credential);
            }
        } catch (MalformedURLException mue) {
            this.status = RFTConstants.STATUS_FAILED;
            throw new RftException(mue);
        } catch(ConnectException ce) {
            this.status = RFTConstants.STATUS_RETRYING;
            ReliableFileTransferDbAdapter.getDbAdapter().backOff(
                    this.host + ":"
                            + +this.port);
            throw new RftTransientException(i18n.getMessage(
                    "unableToConnectErr",
                    new Object[] {
                    this.host,
                    String.valueOf(this.port)
                    }));
        }catch (ServerException se) {
            this.status = RFTConstants.STATUS_FAILED;
            throw new RftException(se);
        } catch (IOException ie) {
            this.status = RFTConstants.STATUS_FAILED;
            throw new RftException(ie);
        }
    }

    /**
     * @param filePath
     */
    public void setFileURL(String filePath) {
        this.filePath = "/" + filePath;
    }

    /**
     * @return urlExpander
     */
    public URLExpander getUrlExpander() {
        return this.expander;
    }

    /**
     * @throws RftException
     */
    public synchronized void delete() throws RftException {
        try {
            if (this.filePath.endsWith("/")) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Deleting a directory " + this.filePath);
                }
                this.status = RFTConstants.STATUS_EXPANDING;
                RFTOptionsType rftOptions = new RFTOptionsType();
                rftOptions.setSourceSubjectName(this.subjectName);
                // Try to send SITE rdel <filepath> 
                String recursiveDeleteArgs = "rdel " + this.filePath; 
                try {
                    //Check if file exists
                    if (this.gridFtpHost.exists(this.filePath)) {
                        Reply reply = 
                            this.gridFtpHost.site(recursiveDeleteArgs);
                        if (logger.isDebugEnabled()) {
                            logger.debug("Recursive dir delete complete " 
                                + this.filePath);
                        }
                        this.status=0;
                    } else {
                        // Dir already deleted
                        logger.debug("This dir: " + this.filePath 
                                + " is deleted");
                        this.status=0;
                    }
                } catch (Exception urce) {
                    // Try this instead
                    if (logger.isDebugEnabled()) {
                        logger.debug("site rdel is not supported" 
                                + urce.getMessage());
                    }
                    this.expander = new URLExpander(this.gridFtpHost, 
                            this.fileURL,
                            rftOptions);
                    this.expander.setRequestId(this.requestId);
                    this.expander.start();
                    return;
                }
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug("Deleting " + this.filePath);
                }
                this.gridFtpHost.deleteFile(this.filePath);
                this.status = 0;
            }
        } catch (Exception e) {
            this.status = RFTConstants.STATUS_RETRYING;
            throw new RftException(
                    i18n.getMessage("deleteFileErr", 
                        this.filePath), e);
        }
    }

    /**
     * @return status of the client
     */
    public int getStatus() {
        return this.status;
    }

    public void setCredential(GSSCredential credential) {
        this.credential = credential;
    }
    
    /**
     * @param url
     * @param requestId
     * @return
     * @throws RftException
     */
    public boolean 
        checkForReuse(String url, int requestId) throws RftException {
        //TODO: check if same resource
        GlobusURL newUrl = null;
        try {
            newUrl = new GlobusURL(url);
        } catch (MalformedURLException mue) {
            throw new RftException(mue);
        }
        String newHost = newUrl.getHost();
        int newPort = newUrl.getPort();
        if (newHost.equals(this.host) && newPort == this.port 
                && this.requestId == requestId) {
            return true;
        }
        return false;
    }
    /**
     * Closes the socket
     */
    public void close() {
        try {
            this.gridFtpHost.close();
        } catch(Exception e) {
            logger.error(i18n.getMessage("closeClientErr"),e);
        }
    }
    /**
     * @return requestId
     */
    public int getRequestId() {
        return this.requestId;
    }
    /**
     * @param requestId
     */
    public void setRequestId(int requestId) {
        this.requestId = requestId;
    }
}
