/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.naming.InitialContext;
import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.rft.generated.BaseRequestType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.OverallStatus;
import org.globus.rft.generated.TotalBytesTransferredResource;
import org.globus.rft.generated.TotalTimeTakenResource;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.rft.generated.ReliableTransferProperties;
import org.globus.wsrf.PersistenceCallback;
import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceLifetime;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.Topic;
import org.globus.wsrf.TopicList;
import org.globus.wsrf.TopicListAccessor;
import org.globus.wsrf.Constants;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.container.UsageConfig;
import org.globus.wsrf.impl.ReflectionResource;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.ResourcePropertyTopic;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleTopic;
import org.globus.wsrf.impl.SimpleTopicList;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;
import org.globus.wsrf.security.SecureResource;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.utils.FilePersistenceHelper;
import org.globus.wsrf.utils.SubscriptionPersistenceUtils;

import org.globus.util.I18n;

import commonj.work.WorkManager;

import org.globus.gsi.GlobusCredential;

import org.apache.axis.message.addressing.EndpointReferenceType;

import org.globus.delegation.DelegationUtil;

import org.globus.security.gridmap.GridMap;

import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.exception.RftException;

import org.globus.transfer.reliable.service.usage.RFTUsageMonitorPacket;

/**
 * Resource Implementation
 *  
 */
public class ReliableFileTransferResource extends ReflectionResource implements
        TopicListAccessor, ResourceProperties, ResourceLifetime, Resource,
        SecureResource, RemoveCallback, PersistenceCallback {

    String HOME_LOCATION = Constants.JNDI_SERVICES_BASE_NAME
            + RFTConstants.SERVICE_NAME + Constants.HOME_NAME;

    private static Log logger = LogFactory
            .getLog(ReliableFileTransferResource.class.getName());

    private static I18n i18n = I18n.getI18n(
            "org.globus.transfer.reliable.errors",
            ReliableFileTransferResource.class.getClassLoader());

    private ResourcePropertySet propSet;

    private TopicList topicList;

    private ResourceSecurityDescriptor desc = null;

    private Calendar terminationTime = null;

    private Calendar resourceCreationTime = null;

    private ResourceKey key;

    private int requestId = 0;

    private ResourceProperty overallStatusRP;

    private ResourceProperty totalBytesTransferredRP;

    private ResourceProperty requestStatusRP;

    private ResourceProperty totalTimeTakenResourceRP;

    private TotalBytesTransferredResource totalBytesTransferred;

    private OverallStatus overallStatus;

    private RequestStatusType requestStatusType;

    private RequestStatusTypeEnumeration requestStatusTypeEnumeration;

    private TotalTimeTakenResource totalTimeTakenType;

    private int size = 0;

    private ReliableFileTransferDbAdapter dbAdapter = ReliableFileTransferDbAdapter
            .getDbAdapter();

    private TransferRequestType transferRequest;

    private DeleteRequestType deleteRequest;

    private int concurrency = 1;

    private WorkManager workManager;

    private boolean started = false;

    private int maxAttempts = 5;

    private FilePersistenceHelper persistenceHelper;

    private static String SUBSCRIPTION_FILE = System
            .getProperty("file.separator")
            + "rft-subscriptions-";

    private boolean delete = false;

    private RFTResourceManager resourceManager;

    private GlobusCredential credential;

    //Dummy constructor used in recovery
    /**
     * Constructor for the ReliableFileTransferResource object
     */
    public ReliableFileTransferResource() {
        // Empty constructor used in recovery
    }

    //Construtor to create the resource

    /**
     * Constructor for the ReliableFileTransferResource object
     * 
     * @param req
     *            <code> BaseRequestType </code>
     * @param userSubject
     * @param localUserId
     * @param initialTerminationTime
     * @throws Exception
     */
    public ReliableFileTransferResource(BaseRequestType req,
            String userSubject, String localUserId,
            Calendar initialTerminationTime) throws Exception {

        if (req instanceof TransferRequestType) {
            // transfers
            this.transferRequest = (TransferRequestType) req;
            this.requestId = this.dbAdapter.storeTransferRequest(
                    this.transferRequest, initialTerminationTime, localUserId,
                    userSubject);
            this.size = this.transferRequest.getTransfer().length;
        } else if (req instanceof DeleteRequestType) {
            // delete list
            this.delete = true;
            this.deleteRequest = (DeleteRequestType) req;
            this.requestId = this.dbAdapter.storeDeleteRequest(
                    this.deleteRequest, initialTerminationTime, localUserId,
                    userSubject);
            if (this.deleteRequest.getDeletion() != null) {
                this.size = this.deleteRequest.getDeletion().length;
            } else {
                logger.error(i18n.getMessage("nullRequestErr"));
                throw new RftException(i18n.getMessage("nullRequestErr"));

            }

        }
        Integer tempConcurrency = req.getConcurrency();
        if (tempConcurrency != null) {
            if (req.getConcurrency().intValue() > 0) {
                this.concurrency = req.getConcurrency().intValue();
            }
        }
        Integer tempMaxAttempts = req.getMaxAttempts();
        if (tempMaxAttempts != null) {
            if (req.getMaxAttempts().intValue() > 0) {
                this.maxAttempts = req.getMaxAttempts().intValue();
            }
        }

	this.resourceManager = 
	    new RFTResourceManager(this.requestId, this.concurrency);

       // Set security properties
        setSecurityDescriptor(userSubject, localUserId);
    }

    /**
     * @return type of request
     */
    public boolean getRequestType() {
        return this.delete;
    }

    /**
     * @return concurrency
     */
    public int getConcurrency() {
        return this.concurrency;
    }
    
    /**
     * @param credential
     * @throws RftException
     */
    public void setCredential(GlobusCredential credential) 
        throws RftException {
            this.credential = credential;
            try {
                if (this.resourceManager != null) {
                    this.resourceManager.setCredential(this.credential);
                }
            } catch (Exception e) {
                throw new 
                    RftException(i18n.getMessage("delegationCredErr"), e);
            }
    }

    private synchronized void storeSubscriptions() throws Exception {
        FileOutputStream fos = null;
        File tmpFile = null;
        try {
            String fileName = SUBSCRIPTION_FILE + this.requestId + ".tmp";
            tmpFile = new File(getPersistenceHelper().getStorageDirectory(),
                    fileName);
            fos = new FileOutputStream(tmpFile);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            SubscriptionPersistenceUtils.storeSubscriptionListeners(this
                        .topicList, oos);
        } catch (Exception e) {
            tmpFile.delete();
            throw new Exception(i18n.getMessage("storeSubscriptionsErr")
                    + e.getMessage(), e);
        } finally {
            if (fos != null) {
                try {
                    fos.close();
                } catch (Exception ee) {
                    logger.error(i18n.getMessage("storeSubscriptionsErr"), ee);
                }
            }
        }
    }

    private void loadSecurityContext(RFTRecoveryData recoveryData)
            throws Exception {
        //TODO Re-register with persisted Delegation resource
        String userSubject = recoveryData.getUserSubject();
        String localUserId = recoveryData.getUserName();
        setSecurityDescriptor(userSubject, localUserId);
        loadSubscribers();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.globus.wsrf.PersistenceCallback#store()
     */
    public void store() {
        try {
            storeSubscriptions();
        } catch (Exception e) {
            logger.error(i18n.getMessage("storeSubscriptionsErr"), e);
        }
    }

    public void loadSubscribers() {
        FileInputStream fis = null;
        try {
            String fileName = getPersistenceHelper().getStorageDirectory()
                + SUBSCRIPTION_FILE + this.requestId + ".tmp";
            if (logger.isDebugEnabled()) {
                logger.debug("Loading rft subscribers from : " + fileName);
            }
            File file = new File(fileName);
            if (file.exists()) {
                fis = new FileInputStream(file);
                ObjectInputStream ois = new ObjectInputStream(fis);
                SubscriptionPersistenceUtils.loadSubscriptionListeners(this
                        .getTopicList(), ois);
            }
        } catch (Exception e) {
            logger.error("Error loading subscribers: ",e);
        //    throw new Exception(i18n.getMessage("loadSubscriptionsErr"), e);
        } finally {
            if (fis != null) {
                try {
                    fis.close();
                } catch (Exception e) {
                    logger.error(i18n.getMessage("loadSubscriptionsErr"
                                + e.getMessage()));
                }
            }
        }
    }

    protected synchronized FilePersistenceHelper getPersistenceHelper() {
        if (this.persistenceHelper == null) {
            try {
                this.persistenceHelper = new FilePersistenceHelper(getClass(),
                        ".ser");
            } catch (Exception e) {
                throw new RuntimeException(e.getMessage());
            }
        }
        return this.persistenceHelper;
    }

    protected synchronized void
        processDelegatedCredential() throws Exception {
	processDelegatedCredential(this.requestId);
    }

    private synchronized void 
        processDelegatedCredential(int requestId) throws Exception {
        try {
            EndpointReferenceType epr = null;
            // get credential and store it in file.
            if (this.transferRequest != null) {
                epr = this.transferRequest.getTransferCredentialEndpoint();
            } else if (this.deleteRequest != null) {
                epr = this.deleteRequest.getTransferCredentialEndpoint();
            }
            if (epr == null) {
                throw new RftException(i18n.getMessage("delegationCredErr"));
            }
            this.key = ReliableFileTransferHome.getResourceKey(requestId);
            if (this.key != null) {
                CredentialRefreshListener listener = 
                    new CredentialRefreshListener(
                        this.HOME_LOCATION, this.key);
                DelegationUtil.registerDelegationListener(epr, listener);
                
                if(logger.isDebugEnabled()) {
                    logger.debug("Registered a new listener with id:" 
                            + listener.getId());
                }
                this.dbAdapter.storeDelegationListenerId(
                        this.requestId, listener.getId());
            } else {
                throw new RftException(i18n.getMessage("delegationCredErr"));
            }
        } catch (Exception e) {
            logger.error(i18n.getMessage("delegationCredErr") + e.getMessage(),
                    e);
            throw new RftException(i18n.getMessage("delegationCredErr")
                    + e.getMessage(), e);
        }
    }

    private void setSecurityDescriptor(String userSubject, String localUserId) {
        try {
            this.desc = new ResourceSecurityDescriptor();
            GridMap gridMap = new GridMap();
            gridMap.map(userSubject, localUserId);
            this.desc.setGridMap(gridMap);
        } catch (Exception exp) {
            throw new RuntimeException(exp.getMessage());
        }
    }

    /**
     * Initialize the rft resource
     * 
     * @param resourceBean
     * @param resourceElementQName
     * @param key
     * @exception ResourceException
     */
    public synchronized 
        void initialize(Object resourceBean, QName resourceElementQName,
            Object key) throws ResourceException {
        super.initialize(resourceBean, resourceElementQName, key);
        logger.debug("Enter initialize()");
        this.key = (ResourceKey) key;
        SimpleResourceKey simpleResourceKey = (SimpleResourceKey) this.key;
        String value = (String) simpleResourceKey.getValue();
        this.requestId = Integer.parseInt(value);
        this.propSet = new SimpleResourcePropertySet(RFTConstants.RFT_RP_SET);
        this.topicList = new SimpleTopicList(this);
        this.totalBytesTransferredRP = new ResourcePropertyTopic(
                new SimpleResourceProperty(RFTConstants.TOTAL_BYTES_RESOURCE));
        this.totalBytesTransferred = new TotalBytesTransferredResource(0);
        this.totalTimeTakenResourceRP = new ResourcePropertyTopic(
                new SimpleResourceProperty(RFTConstants.TOTAL_TIME_RESOURCE));
        this.totalTimeTakenType = new TotalTimeTakenResource(0);
        this.resourceCreationTime = Calendar.getInstance();

        try {
            this.initializeOverallStatusRP();
            this.intializeRequestStatusRP();
            this.setResourceProperties();
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage());
        }
        logger.debug("Exit initialize()");
    }

    private void initializeOverallStatusRP() throws RftDBException {
        //      OverallStatus initialize
        this.overallStatus = this.dbAdapter.getOverallStatus(this.requestId);
        this.overallStatusRP = new ResourcePropertyTopic(
            new SimpleResourceProperty(RFTConstants.OVERALL_STATUS_RESOURCE));

    }

    private void intializeRequestStatusRP() {
        this.requestStatusType = new RequestStatusType();
        this.requestStatusRP = new ResourcePropertyTopic(
                new SimpleResourceProperty(RFTConstants.REQUEST_STATUS_RESOURCE));
        this.requestStatusType.setRequestStatus(RequestStatusTypeEnumeration.Pending);
        // FIXME:this is from TransferWork.java
        int totalTransfers = 0;
        int activeTransfers = this.overallStatus.getTransfersActive();
        int failedTransfers = this.overallStatus.getTransfersFailed();
        int pendingTransfers = this.overallStatus.getTransfersPending();
        int finishedTransfers = this.overallStatus.getTransfersFinished();
        int restartedTransfers = this.overallStatus.getTransfersRestarted();
        totalTransfers = activeTransfers + failedTransfers + pendingTransfers
            + finishedTransfers + restartedTransfers;
        if (activeTransfers > 0 ) {
            this.requestStatusType
                .setRequestStatus(RequestStatusTypeEnumeration.Active);
        }
        if ((failedTransfers + finishedTransfers == totalTransfers) &&
                (failedTransfers != 0)) {
            this.requestStatusType
                .setRequestStatus(RequestStatusTypeEnumeration.Failed);
        }
        if (totalTransfers == finishedTransfers) {
            this.requestStatusType
                .setRequestStatus(RequestStatusTypeEnumeration.Done);
        }
    }

    private void setResourceProperties() throws Exception {
        this.propSet.add(this.overallStatusRP);
        this.propSet.add(this.totalBytesTransferredRP);
        this.propSet.add(this.requestStatusRP);
        this.propSet.add(this.totalTimeTakenResourceRP);

        this.topicList.addTopic((Topic) this.overallStatusRP);
        this.topicList.addTopic((Topic) this.totalBytesTransferredRP);
        this.topicList.addTopic((Topic) this.requestStatusRP);
        this.topicList.addTopic((Topic) this.totalTimeTakenResourceRP);

        this.overallStatusRP.add(this.overallStatus);
        this.totalBytesTransferredRP.add(this.totalBytesTransferred);
        this.requestStatusRP.add(this.requestStatusType);
        this.totalTimeTakenResourceRP.add(this.totalTimeTakenType);

        ResourceProperty prop = new ReflectionResourceProperty(
                SimpleResourcePropertyMetaData.TERMINATION_TIME, this);
        this.propSet.add(prop);
        this.topicList
                .addTopic(new SimpleTopic(WSRFConstants.TERMINATION_TOPIC));

        prop = new ReflectionResourceProperty(
                SimpleResourcePropertyMetaData.CURRENT_TIME, this);
        this.propSet.add(prop);
    }

    /**
     * @return this.workManager
     */
    public WorkManager getWorkManager() {
        return this.workManager;
    }

    public RFTResourceManager getResourceManager() {
        if (this.resourceManager == null) {
            this.resourceManager = 
                new RFTResourceManager(this.requestId, this.concurrency);
        }
        return this.resourceManager;
    }

    /**
     * Gets the securityDescriptor attribute of the ReliableFileTransferResource
     * object
     * 
     * @return The securityDescriptor value
     */
    public ResourceSecurityDescriptor getSecurityDescriptor() {
        return this.desc;
    }

    /**
     * Gets the storedResourceKeyValues attribute of the
     * ReliableFileTransferResource object
     * 
     * @return The storedResourceKeyValues value
     * @throws RftDBException
     */
    public List getStoredResourceKeyValues() throws RftDBException {
        List activeTransfers = this.dbAdapter.getActiveRequestIds();
        return activeTransfers;
    }

    /**
     * Gets the resourcePropertySet attribute of the
     * ReliableFileTransferResource object
     * 
     * @return The resourcePropertySet value
     */
    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    /**
     * Gets the topicList attribute of the ReliableFileTransferResource object
     * 
     * @return The topicList value
     */
    public TopicList getTopicList() {
        return this.topicList;
    }

    /**
     * Gets the iD attribute of the ReliableFileTransferResource object
     * 
     * @return The iD value
     */
    public Object getID() {
        return this.key;
    }

    /**
     * Sets the terminationTime attribute of the ReliableFileTransferResource
     * object
     * 
     * @param time
     *            The new terminationTime value
     */
    public void setTerminationTime(Calendar time) {
        this.terminationTime = time;
        try {
            this.dbAdapter
                    .updateRequest(this.requestId, time.getTimeInMillis());
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage());
        }
    }

    /**
     * @return requestId of this resource
     */
    public int getRequestIdFromKey() {
        SimpleResourceKey simpleResourceKey = (SimpleResourceKey) this.key;
        String value = (String) simpleResourceKey.getValue();
        return Integer.parseInt(value);
    }

    /**
     * @return requestId of this resource
     */
    public int getRequestId() {
        return this.requestId;
    }

    /**
     * Gets the terminationTime attribute of the ReliableFileTransferResource
     * object
     * 
     * @return The terminationTime value
     */
    public Calendar getTerminationTime() {
        return this.terminationTime;
    }

    /**
     * Gets the currentTime attribute of the ReliableFileTransferResource object
     * 
     * @return The currentTime value
     */
    public Calendar getCurrentTime() {
        return Calendar.getInstance();
    }

    /**
     * Gets the resourceBeanClass attribute of the ReliableFileTransferResource
     * object
     * 
     * @return The resourceBeanClass value
     */
    protected Class getResourceBeanClass() {
        return ReliableTransferProperties.class;
    }

    /**
     * Is started
     * 
     * @return if resource is already started
     */
    public boolean isStarted() {
        return this.started;
    }

    /**
     * Called when activating a Transfer resource by ResourceHomeImpl
     * 
     * @param key
     *            resource key
     * @exception ResourceException
     */
    public void load(ResourceKey key) throws ResourceException {
        this.requestId = Integer.parseInt((String) key.getValue());
        if (logger.isDebugEnabled()) {
            logger.debug("Attempting to reload RFT resource with id: " 
                    + this.requestId);
        }
        Boolean isStarted = null;
        try {
            isStarted = this.dbAdapter.isStarted(this.requestId);
        } catch (RftDBException rftdbe) {
            logger.error(i18n.getMessage("resourceLoadErr")
                    + rftdbe.getMessage(), rftdbe);
            throw new ResourceException(i18n.getMessage("resourceLoadErr")
                    + rftdbe.getMessage(), rftdbe);
        }
        if (isStarted == null) {
            throw new NoSuchResourceException((String) key.getValue());
        }
        this.started = isStarted.booleanValue();
        if (this.started) {
            try {
                RFTRecoveryData recoveryData = this.dbAdapter
                    .getRecoveryData(this.requestId);

                InitialContext initialCtx = new InitialContext();
                this.workManager = (WorkManager) initialCtx
                    .lookup(RFTConstants.WORKMANAGER);
                Collection activeTransferJobs = recoveryData
                    .getRecoveredTransfers();
                Iterator it = activeTransferJobs.iterator();

                this.initialize(ReliableFileTransferResource
                        .createResourceBean(null, this.terminationTime),
                        RFTConstants.RFT_KEY_QNAME, key);
                this.resourceManager = 
                    new RFTResourceManager(this.requestId, 
                            activeTransferJobs.size());
                loadSecurityContext(recoveryData);
                while (it.hasNext()) {
                    TransferType transfer = (TransferType) it.next();
                    TransferWork transferWork = new TransferWork(
                            this.requestId, this.propSet, transfer);
                    transferWork.setWorkManager(this.workManager);
                    transferWork.setResourceManager(resourceManager);
                    this.workManager.schedule(transferWork);
                }

            } catch (Exception e) {
                logger.error(i18n.getMessage("resourceLoadErr")
                        + e.getMessage(), e);
                throw new ResourceException(i18n
                        .getMessage("resourceLoadErr")
                        + e.getMessage(), e);
            }
        } else {
            try {
                this.initialize(ReliableFileTransferResource
                        .createResourceBean(null, this.terminationTime),
                        RFTConstants.RFT_KEY_QNAME, key);
                RFTRecoveryData recoveryData = this.dbAdapter
                    .getSecurityRecoveryData(this.requestId);
                loadSecurityContext(recoveryData);
            } catch (Exception e ) {
                logger.error("Error initializing",e);
            }
        }
    }

    public synchronized void setStartedTrue() {
        this.started = true;
    }

    /**
     * Return the max retries for this resource
     * 
     * @return maxAttempts
     */
    public int getMaxAttempts() {
        return this.maxAttempts;
    }

    /**
     * Removes the resources from the container.
     * 
     * @exception ResourceException
     */
    public void remove() throws ResourceException {
        if (logger.isDebugEnabled()) {
            logger.debug("Removing the resource with id: " + this.requestId);
        }
        try {
            String fileName = SUBSCRIPTION_FILE + this.requestId + ".tmp";
            File subscriptionsFile = new File(getPersistenceHelper()
                    .getStorageDirectory(), fileName);
            if (subscriptionsFile.exists()) {
                if (logger.isDebugEnabled()) {
                    logger.debug("deleting subscriptions file");
                }
                subscriptionsFile.delete();
            }
            //Delete delegated cred listeners 
            this.resourceManager.removeDelegatedCredListeners();
            this.dbAdapter.deleteAll(this.requestId);
            TransferWork.closeTransferClients(this.requestId);
            if(logger.isDebugEnabled()) {
                logger.debug("PerformanceLog RFT Resource with key: "+
                        this.requestId + " Destroyed");
            }

        } catch (Exception e) {
            logger.error(i18n.getMessage("resourceRemoveErr") + e.getMessage(),
                    e);
            throw new ResourceException(i18n.getMessage("resourceRemoveErr")
                    + e.getMessage(), e);
        }
        sendUsagePacket();
    }

    /**
     * This should be moved to resource
     * 
     * @param initialTerminationTime
     * @param currentTime
     *            current time
     * @return <code> ReliableTransferProperties </code>
     */
    public static ReliableTransferProperties createResourceBean(
            Calendar initialTerminationTime, Calendar currentTime) {
        ReliableTransferProperties reliableTransferProperties = new ReliableTransferProperties();
        reliableTransferProperties
                .setTotalBytes(new TotalBytesTransferredResource());
        reliableTransferProperties.setTotalTime(new TotalTimeTakenResource());
        reliableTransferProperties
                .setOverallStatusProperty(new OverallStatus());
        reliableTransferProperties.setCurrentTime(currentTime);
        reliableTransferProperties.setTerminationTime(initialTerminationTime);
        return reliableTransferProperties;
    }

    private void sendUsagePacket() {
        try {
            RFTUsageMonitorPacket usagePacket = new RFTUsageMonitorPacket();
            if(this.delete) {
                usagePacket.setRequestType(this.delete);
            }
            usagePacket.setTotalNumberOfFiles
                (this.dbAdapter.getTotalNumberOfTransfers());
            usagePacket.setTotalNumberOfBytes(
                    this.dbAdapter.getTotalNumberOfBytes());
            usagePacket.setNumberOfFiles(this.size);
            usagePacket.setNumberOfBytes(
                    this.totalBytesTransferred.getTotalBytes());
            usagePacket.setResourceCreationTime(this.resourceCreationTime.getTime());
            usagePacket.setFactoryStartTime(this.getCurrentTime().getTime());

            UsageConfig usageConfig = UsageConfig.getUsageConfig();
            if (usageConfig != null) {
                List targets = usageConfig.getTargets();
                usagePacket.setHostIP(usageConfig.getContainerHost()); 
                if (targets != null) {
                    usagePacket.sendPacket(targets);
                }
            }
        } catch (Exception e) {
            logger.warn("Could not send usage stat "+e.getMessage());
        }
    }
}

