/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import java.net.MalformedURLException;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.OverallStatus;
import org.globus.rft.generated.RFTFaultResourcePropertyType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.util.GlobusURL;
import org.globus.util.I18n;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import commonj.work.Work;
import commonj.work.WorkException;
import commonj.work.WorkManager;

import org.globus.transfer.reliable.service.database.RFTDatabasePoller;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RFTDatabaseSetup;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;

/**
 * Does the actual transfer
 *  
 */
public class TransferWork implements Work {
    private TransferType transferJob;

    private int status = -1;

    private int attempts;

    private int requestId = 0;

    private ReliableFileTransferDbAdapter dbAdapter = ReliableFileTransferDbAdapter
            .getDbAdapter();

    private static Log logger = LogFactory.getLog(TransferWork.class.getName());

    private static I18n i18n = I18n.getI18n("org.globus.transfer.reliable.errors",
            TransferWork.class.getClassLoader());
    
    /**
     * List of transfer clients to be used across resources for same user.
     */
    private static Vector transferClients = new Vector();

    private static Vector deleteClients = new Vector();

    private OverallStatus overallStatus;

    private ResourceProperty overallStatusRP;

    private RequestStatusType requestStatusType;

    private ResourceProperty requestStatusTypeRP;

    private ResourcePropertySet rpSet;

    private int maxAttempts = 0;

    private WorkManager workManager;

    private RFTFaultResourcePropertyType fault;

    private TransferClient transferClient = null;

    private DeleteClient deleteClient = null;
    
    private RFTResourceManager resourceManager = null;
    /**
     * Constructor for the TransferWork object
     * 
     * @param requestId
     * @param rpSet
     * @param transferJob
     */
    public TransferWork(int requestId, ResourcePropertySet rpSet,
            TransferType transferJob) {
        this.requestId = requestId;
        this.rpSet = rpSet;
        this.transferJob = transferJob;
        this.status = -1;
        this.attempts = transferJob.getAttempts().intValue();
        this.overallStatusRP = 
            this.rpSet.get(RFTConstants.OVERALL_STATUS_RESOURCE);
        this.overallStatus = (OverallStatus) this.overallStatusRP.get(0);
        this.requestStatusTypeRP = 
            this.rpSet.get(RFTConstants.REQUEST_STATUS_RESOURCE);
        this.requestStatusType = (RequestStatusType) 
        this.requestStatusTypeRP.get(0);
    }

    /**
     * @param transferJob
     */
    public void setTransferJob(TransferType transferJob) {
        this.transferJob = transferJob;
    }
    /**
     * @return The daemon value
     */
    public boolean isDaemon() {
        return true;
    }

    /**
     * releases the thread resources
     */
    public void release() {
        //Part of interface Work
    }

    /**
     * Gets the status attribute of the TransferWork object
     * 
     * @return The status value
     */
    public int getStatus() {
        return this.status;
    }

    public void setResourceManager(RFTResourceManager resourceManager) 
        throws RftDBException {
        this.resourceManager = resourceManager;
        this.maxAttempts = this.resourceManager.getMaxAttempts();  
    }
    /**
     * generates notification of status changes of a request
     * 
     * @param status
     */
    private synchronized void statusChanged(int status) {
        try {
            this.status = status;
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "status changed called with status: " 
                             + this.status);
            }
            this.transferJob.setStatus(RFTUtil.mapStatus(this.status));
            this.resourceManager.updateTransferJob(
                    this.transferJob.getTransferId().intValue(), 
                    this.status, 
                    this.attempts);
            //this.dbAdapter.update(this.requestId, this.transferJob);
            // A Active notification is already sent in this case
            // But it should be updated in the database
            if (this.status != RFTConstants.STATUS_EXPANDING) {
                this.overallStatus = 
                    this.dbAdapter.getOverallStatus(this.requestId);
                if (this.fault != null) {
                    this.overallStatus.setFault(this.fault);
                }

                if (logger.isDebugEnabled()) {
                    logger.debug(this.getTransferIdentifiers()
                            + "setting overall rp to "
                            + "(finished: "
                            + this.overallStatus.getTransfersFinished()
                            + ", active: "
                            + this.overallStatus.getTransfersActive() 
                            + ", failed: "
                            + this.overallStatus.getTransfersFailed()
                            + ", restarted: "
                            + this.overallStatus.getTransfersRestarted()
                            + ", pending: "
                            + this.overallStatus.getTransfersPending()
                            + ", cancelled: "
                            + this.overallStatus.getTransfersCancelled()
                            + ")");
                }

                this.overallStatusRP.set(0, this.overallStatus);
                setRequestStatus();
            }
        } catch (RftDBException e) {
            logger.error(i18n.getMessage("dbUpdateErr") + e.getMessage(), e);
        }
    }

    /**
     * Sets the request status resource property
     */
    private void setRequestStatus() {

        RequestStatusTypeEnumeration reqStatus = this.requestStatusType
                .getRequestStatus();
        int totalTransfers = 0;
        int activeTransfers = this.overallStatus.getTransfersActive();
        int failedTransfers = this.overallStatus.getTransfersFailed();
        int pendingTransfers = this.overallStatus.getTransfersPending();
        int finishedTransfers = this.overallStatus.getTransfersFinished();
        int restartedTransfers = this.overallStatus.getTransfersRestarted();
        totalTransfers = activeTransfers + failedTransfers + pendingTransfers
                + finishedTransfers + restartedTransfers;
        if (activeTransfers > 0 && !reqStatus.getValue().equals("Active")) {
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Active);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
        }
        if ((failedTransfers > 0) && !reqStatus.getValue().equals("Failed")) {
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Failed);
            this.requestStatusType.setFault(this.fault);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
            return;
        }
        if (totalTransfers == finishedTransfers) {
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "setting status RP to Done");
            }
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Done);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
            return;
        }
    }

    /**
     * @param workManager
     */
    public void setWorkManager(WorkManager workManager) {
        this.workManager = workManager;
    }

    /**
     * used to stop the active transfers.
     * 
     * @param requestId
     * @param toId
     * @param fromId
     */
    public static void stopActiveTransfers(int requestId, int fromId, int toId) {
        synchronized(transferClients) {
            for (int i = 0; i < transferClients.size(); i++) {
                TransferClient transferClient = (TransferClient) transferClients
                    .elementAt(i);
                int transferRequestId = transferClient.getRequestId();
                int transferClientStatus = transferClient.getStatus();
                if (transferRequestId == requestId
                        && transferClientStatus == RFTConstants.STATUS_ACTIVE) {
                    int transferId = transferClient.getTransferID();
                    if (transferId >= fromId && transferId <= toId) {
                        //check to see if this is invoked from service impl
                        // and if range is valid
                        transferClient.close();
                        transferClient.setStatus(RFTConstants.STATUS_CANCELLED);
                    } else {
                        //stop all the active transfers in that req
                        transferClient.close();
                        transferClient.setStatus(RFTConstants.STATUS_CANCELLED);
                    }
                }
            }
        }
    }

    /**
     * Closes the active transfer clients when a resource is removed
     * 
     * @param requestId
     */
    public static void closeTransferClients(int requestId) {
        synchronized(transferClients) {
            for (int i = 0; i < transferClients.size(); i++) {
                TransferClient closeMe = (TransferClient) transferClients
                    .elementAt(i);
                if (closeMe.getRequestId() == requestId) {
                    transferClients.remove(i);
                    if (i > 0 ) {
                        i--;
                    }
                    closeMe.close();
                    logger.debug("Closing all clients for request: " + requestId);
                }
            }
        }
        synchronized(deleteClients) {
            for (int i = 0; i < deleteClients.size(); i++) {
                DeleteClient closeMe = (DeleteClient) deleteClients.elementAt(i);
                if (closeMe.getRequestId() == requestId) {
                    deleteClients.remove(i);
                    if (i > 0) {
                        i--;
                    }
                    closeMe.close();
                }
            }
        }

    }

    /**
     * Gets the transferClient attribute of the TransferWork object
     * 
     * @return The transferClient value
     * @throws RftException
     */
    private TransferClient getTransferClient() 
        throws RftException, RftTransientException, RftDBException {
        TransferClient temp = null;
        GlobusURL sourceURL = null;
        GlobusURL destURL = null;
        boolean reuse = false;
        int i = 0;
        try {
            sourceURL = new GlobusURL(this.transferJob.getSourceUrl());
            if (!sourceURL.getProtocol().equalsIgnoreCase("gsiftp")) {
                throw new RftException("Only 3rd party transfers supported");
            }
            destURL = new GlobusURL(this.transferJob.getDestinationUrl());
            if (!destURL.getProtocol().equalsIgnoreCase("gsiftp")) {
                throw new RftException("Only 3rd party transfers supported");
            }
        } catch (MalformedURLException mue) {
            throw new RftException(mue);
        }
        logger.debug(this.getTransferIdentifiers()
                    + "transferring "
                    + this.transferJob.getSourceUrl()
                    + "  ->  "
                    + this.transferJob.getDestinationUrl());
        synchronized(transferClients) {
            for (; i < transferClients.size(); i++) {
                temp = (TransferClient) transferClients.elementAt(i);
                if (temp.checkForReuse(this.requestId, sourceURL, destURL)) {
                    reuse = true;
                    break;
                }
            }
            if (reuse && transferClients.size() > 0 ) {
                temp = (TransferClient) transferClients.remove(i);

                if (logger.isDebugEnabled()) {
                    logger.debug(this.getTransferIdentifiers()
                            + "reusing transfer client " 
                            + transferClients.size());
                }
                temp.setSourcePath(sourceURL.getPath());
                temp.setDestinationPath("/" + destURL.getPath());
                temp.setSourceURL(sourceURL);
                temp.setDestinationURL(destURL);
                if (temp.getStatus() != RFTConstants.STATUS_FAILED) {
                    temp.setStatus(RFTConstants.STATUS_ACTIVE);
                } 
            }
        }
        if(!reuse) {
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                        + "no client to reuse");
            }
            temp = getNewClient();
        }
        return temp;
        }

    private DeleteClient getDeleteClient() throws RftDBException,
            RftException, RftTransientException {
        DeleteClient temp = null;
        GlobusURL sourceURL = null;
        boolean reuse = false;
        int i = 0;
        try {
            sourceURL = new GlobusURL(this.transferJob.getSourceUrl());
        } catch (MalformedURLException mue) {
            throw new RftException(mue);
        }
        logger.debug(this.getTransferIdentifiers()
                    + "deleting "
                    + this.transferJob.getSourceUrl());
        synchronized(deleteClients) {
            for (; i < deleteClients.size(); i++) {
                temp = (DeleteClient) deleteClients.elementAt(i);
                if (temp.checkForReuse(this.transferJob.getSourceUrl(),
                            this.requestId)) {
                    reuse = true;
                    break;
                }
            }
            if (reuse) {
                temp = (DeleteClient) deleteClients.remove(i);
                temp.setFileURL(sourceURL.getPath());
                if (temp.getStatus() != RFTConstants.STATUS_FAILED) {
                    this.status = temp.getStatus();
                }
            }
        }
        if (!reuse) {
            temp = new DeleteClient(this.transferJob.getSourceUrl(),
                    this.transferJob.getRftOptions().getSourceSubjectName(),
                    this.resourceManager.getCredential(),
                    this.transferJob.getRftOptions().getUserName());

        }
        temp.setRequestId(this.requestId);
        return temp;
    }

    private TransferClient getNewClient() throws RftException,
            RftTransientException, RftDBException {
        logger.debug("max active allowed: " + RFTDatabaseSetup.maxActive );
        TransferClient temp = new TransferClient();
        String userName = this.dbAdapter.getUserName(this.transferJob
                .getTransferId().intValue());
        if (logger.isDebugEnabled()) {
            logger.debug("Transfer " + this.transferJob.getTransferId() + ": "
                         + "new transfer client for user: " + userName);
        }
        temp.setUserName(userName);
        temp.setSourceHost(this.transferJob.getSourceUrl());
        temp.setDestinationHost(this.transferJob.getDestinationUrl());
        temp.setOptions(this.transferJob.getRftOptions());
        temp.setCredential(this.resourceManager.getCredential());
        temp.authorize();
        temp.authenticate();
        temp.setRFTOptions();
        temp.setRequestId(this.requestId);
        temp.setModeE();
        temp.setStriping();
        temp.setStatus(RFTConstants.STATUS_ACTIVE);
        /*
         * synchronized(this) { // To be able to stop active transfers
         * logger.debug("request id " + this.requestId);
         * transferClients.add(temp); }
         */
         // A new client is added
        return temp;
    }

    /*
     * Sets Faults on resource properties
     */
    private void setFault(Exception e) {
        if (e instanceof RftTransientException) {
            this.fault = FaultUtil.createTransferTransientFault(this.requestId,
                    e.getMessage(), e);
            if (logger.isDebugEnabled()) {
                logger.debug("setting transient fault");
            }
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
        if (e instanceof RftException) {
            this.fault = FaultUtil.createTransferFault(this.requestId,
                    e.getMessage(), e);
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
        if (e instanceof RftDBException) {
            this.fault = FaultUtil.createRFTDatabaseFault(this.requestId, e);
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
    }

    /*
     * For Processing states
     */
    private void processStates() {
        int transferId = this.transferJob.getTransferId().intValue();
       
        if (logger.isDebugEnabled()) {
            logger.debug(this.getTransferIdentifiers()
                         + "processing state for transfer of "
                         + this.transferJob.getSourceUrl() + "  ->  "
                         + this.transferJob.getDestinationUrl());
        }

        switch (this.status) {
            
        case RFTConstants.STATUS_FINISHED: {
            //Transfer done
            this.resourceManager.decrementActive();
            logger.debug(this.getTransferIdentifiers()
                        + "transfer done");
            if (this.transferClient != null) {
                synchronized(transferClients) {
                    transferClients.add(this.transferClient);
                }
            }
            if (this.deleteClient != null) {
                synchronized(deleteClients) {
                    deleteClients.add(this.deleteClient);
                }
            }
            statusChanged(RFTConstants.STATUS_FINISHED);
            break;
        }
        case RFTConstants.STATUS_FAILED: {
            // Transfer Failed
            this.resourceManager.decrementActive();
            logger.debug(this.getTransferIdentifiers()
                        + "transfer failed");
            statusChanged(RFTConstants.STATUS_FAILED);
            try {
                if (this.dbAdapter.getAllOrNone(this.requestId)) {
                    stopActiveTransfers(this.requestId, -1, -1);
                    this.dbAdapter.updateToDirty(this.requestId);
                }
            } catch (RftDBException rftdbe) {
                logger.error(i18n.getMessage("dbRetrieveErr"), rftdbe);
            }
            if (this.transferClient != null) {
                this.transferClient.close();
            }
            break;
        }
        case RFTConstants.STATUS_RETRYING: {
            this.resourceManager.decrementActive();
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "retry attempt "
                             + this.transferJob.getAttempts()
                             + " of "
                             + this.maxAttempts);
            }
            if (this.transferJob.getAttempts().intValue() >= this.maxAttempts) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Transfer " + transferId + ": "
                            + "transfer failed (retried "
                            + this.transferJob.getAttempts()
                            + " times)");
                }
                statusChanged(RFTConstants.STATUS_FAILED);
            } else {
                // Retrying
                statusChanged(RFTConstants.STATUS_RETRYING);
            }
            if(this.transferClient != null) {
                this.transferClient.close();
            }
            break;
        }
        case RFTConstants.STATUS_EXPANDING: {
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "expanding the url: " 
                             + this.transferJob.getSourceUrl());
            }
        
            statusChanged(RFTConstants.STATUS_EXPANDING);
            URLExpander expander = null;
            if (this.transferClient != null) {
                expander = this.transferClient.getUrlExpander();
            } else if (this.deleteClient != null) {
                expander = this.deleteClient.getUrlExpander();
            }
            if (expander != null) {
                boolean expStatus = expander.getStatus();
                while (expStatus == false) {
                    try {
                        expander.join();
                        expStatus = expander.getStatus();
                    } catch (InterruptedException ee) {
                        logger.error(i18n.getMessage("urlExpansionErr"), ee);
                    }
                }
            }
            this.resourceManager.setNumberOfTransfers(
                expander.getNumberOfTransfers());
            statusChanged(RFTConstants.STATUS_EXPANDING_DONE);
            this.resourceManager.decrementActive();
            expander.close();
            //startNewTransfer();
            startDirTransfer();
            break;
        }
        }
    }

    private void startDirTransfer() {
        int concurrency = this.resourceManager.getConcurrency();
        int activeTransfers = this.resourceManager.getActive();
        while (activeTransfers < concurrency) {
            startNewTransfer();
            activeTransfers++;
        }
    }


    private synchronized void startNewTransfer() {
        try {
            TransferType newTransferJob = this.resourceManager.getNext();
            if (newTransferJob != null) {
                TransferWork newTransferWork 
                    = new TransferWork( this.requestId, 
                            this.rpSet, newTransferJob);
                newTransferWork.setWorkManager(this.workManager);
                newTransferWork.setResourceManager(this.resourceManager);
                this.workManager.schedule(newTransferWork);
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug("PerformanceLog All transfers done for "
                            + this.requestId);
                }
                long retryTime = this.dbAdapter.getMinimumPollTime();
                if (retryTime != 0) {
                    RFTDatabasePoller dbPoller = new RFTDatabasePoller(
                            retryTime, this.rpSet, this.requestId,
                            this.resourceManager);
                    this.workManager.schedule(dbPoller);
                }
                //Check if all done
                if (this.dbAdapter.checkAllDone(this.requestId)) {
                    if(logger.isDebugEnabled()) {
                        logger.debug(this.getTransferIdentifiers()
                                + "all transfers of " 
                                + this.requestId + " are done");
                    }
                    TransferWork.closeTransferClients(this.requestId);
                }
            }
        } catch (RftDBException dbe) {
            logger.error(i18n.getMessage("dbRetrieveErr") + dbe.getMessage(),
                    dbe);
            setFault(dbe);
            statusChanged(RFTConstants.STATUS_FAILED);
        } catch (WorkException we) {
            logger.error(i18n.getMessage("unableToScheduleErr"), we);
            setFault(we);
            statusChanged(RFTConstants.STATUS_FAILED);
        }
    }

    private boolean checkDelete() {
        boolean returnValue = false;
        String transferType = this.transferJob.getStatus().getValue();
        String destUrl = this.transferJob.getDestinationUrl();
        if (destUrl == null) {
            returnValue = true ;
        }
        if (transferType.equals("Dirty")) {
            if (destUrl != null) { 
            this.transferJob.setSourceUrl(this.transferJob.getDestinationUrl());
            returnValue = true;
            }
        }
        return returnValue;
    }
 

    /**
     * Main processing method for the TransferWork object 
     */
    public void run() {
        try {
            int transferId = this.transferJob.getTransferId().intValue();
            this.attempts = this.transferJob.getAttempts().intValue();
            this.attempts++;
            this.transferJob.setAttempts(new Integer(this.attempts));
            if (checkDelete()) {
                if (logger.isDebugEnabled()) {
                    logger.debug("This is a delete:" + transferId);
                }
                this.deleteClient = getDeleteClient();
                statusChanged(RFTConstants.STATUS_ACTIVE);
                this.deleteClient.delete();
                this.status = this.deleteClient.getStatus();
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug("This is a transfer:" + transferId);
                }
                this.transferClient = getTransferClient();
                if ( this.transferClient == null) {
                    return;
                }
                statusChanged(RFTConstants.STATUS_ACTIVE);
                this.transferJob.setAttempts(new Integer(this.attempts));
                boolean restart = false;
                String restartMarker = this.dbAdapter
                        .getRestartMarker(transferId);
                this.transferClient.setTransferId(transferId);
                if (restartMarker != null) {
                    if(logger.isDebugEnabled()) {
                        logger.debug(this.getTransferIdentifiers()
                                     + "setting restart marker: " 
                                     + restartMarker);
                    }
                    this.transferClient.setRestartMarker(restartMarker);
                    restart = true;
                }
                this.transferClient.setRequestId(this.requestId);
                this.transferClient
                    .setResourcePropertySet(this.rpSet);
                this.transferClient.transfer(restart);
                this.status = this.transferClient.getStatus();
            }
        } catch (RftDBException dbe) {
            logger.error(i18n.getMessage("dbRetrieveErr"), dbe);
            this.status =  RFTConstants.STATUS_FAILED;
            setFault(dbe);
        } catch (RftException rfe) {
            logger.error(i18n.getMessage("terminalTransferErr")
                    + rfe.getMessage(), rfe);
            this.status =  RFTConstants.STATUS_FAILED;
            setFault(rfe);
        } catch (RftTransientException rfte) {
            logger.error(i18n.getMessage("transientTransferErr")
                    + rfte.getMessage(), rfte);
            this.status = RFTConstants.STATUS_RETRYING;
            setFault(rfte);
        }
        processStates();
        startNewTransfer();
    }

    private String getTransferIdentifiers() {
        return "[Request " + this.requestId
               + ", Transfer " + this.transferJob.getTransferId().intValue()
               + "] ";         
    }
}

