/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "wsnt_NotifyType.h"
#include "globus_i_notification_consumer.h"
#include "NotificationConsumerService.h"
#include "NotificationConsumerService_client.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/* Prototypes */
static
int
globus_l_notification_consumer_activate(void);

static
int
globus_l_notification_consumer_deactivate(void);

static
globus_result_t
globus_l_notification_create_resource_id(
    xsd_any * *                         property,
    char *                              resource_id);

/* Local Variables */
globus_mutex_t globus_i_notification_lock;
globus_cond_t globus_i_notification_cond;

globus_module_descriptor_t globus_i_notification_consumer_module =
{
    "globus_notification_consumer",
    globus_l_notification_consumer_activate,
    globus_l_notification_consumer_deactivate
};
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

/* Client Helper API */
/**
 * Create a NotificationConsumerService resource.
 * @ingroup notification_consumer
 *
 * Create a notification consumer resource, which binds an resource endpoint
 * to a user-defined callback. The @a consumer_reference passed to this 
 * function will be modified to contain an EPR to this resource which can
 * be used as the ConsumerReference element in a Subscribe operation call
 * when using the client stubs for a service which implements the
 * NotificationProducer API.
 * 
 * @param consumer_reference
 *     EPR to be populated to point to a new NotificationConsumer service
 *     resource instance.
 * @param engine
 *     Service engine which will be used to process requests for this new
 *     new service.
 * @param callback
 *     Pointer to the application callback to be called when a notification
 *     message is received for this resouce.
 * @param callback_arg
 *     Application-specific parameter to the callback function.
 *
 * @retval GLOBUS_SUCCESS
 *     Consumer resource created successfully.
 * @retval GLOBUS_NOTIFICATION_CONSUMER_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function.
 * @retval GLOBUS_NOTIFICATION_CONSUMER_ERROR_TYPE_UNABLE_TO_CREATE_CONSUMER
 *     Unable to create consumer. Underlying cause will be chained to this
 *     error, but is typically an out-of-memory situation.
 */
extern
globus_result_t
globus_notification_create_consumer(
    wsa_EndpointReferenceType *         consumer_reference,
    globus_service_engine_t             engine,
    globus_notification_consumer_func_t callback,
    void *                              callback_arg)
{
    globus_result_t                     result;
    globus_uuid_t                       resource_uuid;
    globus_resource_t                   resource;
    globus_i_notification_consumer_t *  consumer;
    xsd_any *                           resource_reference;

    if (consumer_reference == NULL || callback == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_NULL_PARAM(NULL);

        goto out;
    }

    result = globus_uuid_create(&resource_uuid);
    if (result != GLOBUS_SUCCESS)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));

        goto out;
    }

    /* Create a resource which will represent this consumer. We'll associate
     * the callback/arg with this resource
     */
    result = globus_resource_create(resource_uuid.text, &resource);
    if (result != GLOBUS_SUCCESS)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));
        goto out;
    }

    consumer = globus_libc_malloc(sizeof(globus_i_notification_consumer_t));

    if (consumer == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));

        goto destroy_resource_out;
    }

    consumer->callback = callback;
    consumer->arg = callback_arg;
    consumer->state = CONSUMER_OK;

    result = globus_resource_set_resource_specific(
            resource,
            consumer,
            globus_libc_free);

    if (result != GLOBUS_SUCCESS)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));

        goto free_consumer_out;
    }

    result = globus_l_notification_create_resource_id(
        &resource_reference,
        resource_uuid.text);

    if (result != GLOBUS_SUCCESS)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));

        goto destroy_resource_out;
    }

    result = globus_wsrf_core_create_endpoint_reference(
            engine,
            NOTIFICATIONCONSUMERSERVICE_BASE_PATH,
            &resource_reference,
            consumer_reference);

    if (result != GLOBUS_SUCCESS)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNABLE_TO_CREATE_CONSUMER(
                globus_error_get(result));

        goto destroy_resource_out;
    }

    return result;

free_consumer_out:
    globus_libc_free(consumer);
destroy_resource_out:
    globus_resource_destroy(resource);
out:
    return result;
}
/* globus_notification_create_consumer() */

/**
 * Destroy a consumer resource.
 * @ingroup notification_consumer
 *
 * The resource associated with the EPR passed to this function will be
 * destroyed. After this function returns, the callback associated with this
 * resource will no longer be called. 
 *
 * @param consumer_reference
 *     EPR of the resource to destroy.
 *
 * @retval GLOBUS_SUCCESS
 *     Consumer resource destroyed successfully.
 * @retval GLOBUS_NOTIFICATION_CONSUMER_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function.
 * @retval GLOBUS_NOTIFICATION_CONSUMER_ERROR_TYPE_UNKNOWN_RESOURCE
 *     The EPR does not refer to a valid resource.
 */
extern
globus_result_t
globus_notification_destroy_consumer(
    wsa_EndpointReferenceType *         consumer_reference)
{
    globus_result_t                     result;
    globus_resource_t                   resource;
    globus_i_notification_consumer_t *  consumer;
    globus_bool_t                       finish_it = GLOBUS_FALSE;

    if (consumer_reference == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_NULL_PARAM(NULL);

        goto out;
    }

    globus_mutex_lock(&globus_i_notification_lock);

    result = globus_wsrf_core_get_resource_from_epr(
            consumer_reference,
            &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNKNOWN_RESOURCE(
                result ? globus_error_get(result) : NULL);

        goto unlock_out;
    }

    result = globus_resource_get_resource_specific(
            resource,
            (void **)&consumer);

    if (result != GLOBUS_SUCCESS || consumer == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNKNOWN_RESOURCE(
                result ? globus_error_get(result) : NULL);

        goto finish_out;
    }

    while (consumer->state != CONSUMER_DESTROYED)
    {
        switch (consumer->state)
        {
            case CONSUMER_OK:
                consumer->state = CONSUMER_DESTROYED;
                globus_cond_broadcast(&globus_i_notification_cond);
                finish_it = GLOBUS_TRUE;
                break;

            case CONSUMER_ACTIVE:
                finish_it = GLOBUS_TRUE;
                consumer->state = CONSUMER_ACTIVE_DESTROY_CALLED;
                /* FALLSTHROUGH */
            case CONSUMER_ACTIVE_DESTROY_CALLED:
                globus_cond_wait(
                        &globus_i_notification_cond,
                        &globus_i_notification_lock);
                break;

            case CONSUMER_DESTROYED:
                break;
        }
    }

finish_out:

    if (finish_it)
    {
        /* Destroy resource---consumer is freed by the destroy call */
        result = globus_resource_finish_and_destroy(resource);
    }
    else
    {
        /* Free reference from this function call */
        result = globus_resource_finish(resource);
    }

unlock_out:
    globus_mutex_unlock(&globus_i_notification_lock);
out:
    return result;
}
/* globus_notification_destroy_consumer() */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
static
int
globus_l_notification_consumer_activate(void)
{
    globus_module_activate(GLOBUS_COMMON_MODULE);
    globus_module_activate(GLOBUS_SERVICE_ENGINE_MODULE);
    globus_module_activate(GLOBUS_WSRF_RESOURCE_MODULE);
    globus_module_activate(NOTIFICATIONCONSUMERSERVICE_MODULE);

    globus_mutex_init(&globus_i_notification_lock, NULL);
    globus_cond_init(&globus_i_notification_cond, NULL);

    return GLOBUS_SUCCESS;
}
/* globus_l_notification_consumer_activate() */

static
int
globus_l_notification_consumer_deactivate(void)
{
    globus_cond_destroy(&globus_i_notification_cond);
    globus_mutex_destroy(&globus_i_notification_lock);

    globus_module_deactivate(NOTIFICATIONCONSUMERSERVICE_MODULE);
    globus_module_deactivate(GLOBUS_WSRF_RESOURCE_MODULE);
    globus_module_deactivate(GLOBUS_SERVICE_ENGINE_MODULE);
    globus_module_deactivate(GLOBUS_COMMON_MODULE);

    return 0;
}
/* globus_l_notification_consumer_deactivate() */

/**
 * Wrap the resource ID string in an XML any element with the QName
 * http://www.ibm.com/xmlns/stdwip/web-services/WS-BaseNotification#ResourceId
 *
 * @param property
 *     Pointer to a xsd_any to allocate and populate.
 * @param resource_id
 *     String to be used as the ResourceID
 */
static
globus_result_t
globus_l_notification_create_resource_id(
    xsd_any * *                         property,
    char *                              resource_id)
{
    xsd_string * val;
    xsd_any_init(property);

    (*property)->registry = NULL;
    (*property)->any_info = &xsd_string_info;
    (*property)->element = malloc(sizeof(xsd_QName));
    (*property)->element->Namespace =
            globus_libc_strdup(NOTIFICATION_NS);
    (*property)->element->local = globus_libc_strdup("ResourceID");
    val = (xsd_string *)malloc(sizeof(xsd_string));
    *val = globus_libc_strdup(resource_id);
    (*property)->value = val;

    return GLOBUS_SUCCESS;
}
/* globus_l_notification_create_resource_id() */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
