/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "wsnt_NotifyType.h"
#include "globus_i_notification_consumer.h"
#include "NotificationConsumerService.h"
#include "globus_xml_buffer.h"
#include "wsnt_NotificationMessageHolderType.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * Topic expressions we know to be string valued. Anything else we
 * leave alone.
 */
/* @{ */
#define SIMPLE_TOPIC_EXPRESSION \
    "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
#define CONCRETE_TOPIC_EXPRESSION \
    "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Concrete"
#define FULL_TOPIC_EXPRESSION \
    "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Full"
/* @} */

static
globus_result_t
globus_i_notification_consumer_service_deserialize_topic(
    wsnt_TopicExpressionType *          topic);

/* Implementation of ConsumerService skeleton functions */
extern
globus_result_t
NotificationConsumerService_init(
    globus_service_descriptor_t *       service_desc)
{
    globus_module_activate(GLOBUS_NOTIFICATION_CONSUMER_MODULE);

    return GLOBUS_SUCCESS;
}
/* NotificationConsumerService_init() */

extern
globus_result_t
NotificationConsumerService_finalize(
    globus_service_descriptor_t *       service_desc)
{
    globus_module_deactivate(GLOBUS_NOTIFICATION_CONSUMER_MODULE);

    return GLOBUS_SUCCESS;
}
/* NotificationConsumerService_finalize() */

globus_result_t
Consumer_Notify_init(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        message,
    wsnt_NotifyType *                   Notify)
{
    /* NOOP */
    return GLOBUS_SUCCESS;
}
/* Consumer_Notify_init() */

globus_result_t
Consumer_Notify_impl(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        message,
    globus_service_descriptor_t *       service,
    wsnt_NotifyType *                   Notify)
{
    globus_result_t                     result;
    globus_resource_t                   resource;
    globus_i_notification_consumer_t *  consumer;
    int                                 i;

    globus_mutex_lock(&globus_i_notification_lock);

    result = globus_wsrf_core_get_resource(message, service, &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNKNOWN_RESOURCE(
            result ? globus_error_get(result) : NULL);

        goto unlock_out;
    }

    result = globus_resource_get_resource_specific(
            resource,
            (void **)&consumer);

    if (result != GLOBUS_SUCCESS || consumer == NULL)
    {
        result = GLOBUS_NOTIFICATION_CONSUMER_ERROR_UNKNOWN_RESOURCE(
            result ? globus_error_get(result) : NULL);

        goto finish_out;
    }

    for (i = 0; i < Notify->NotificationMessage.length; i++)
    {
        globus_bool_t                   called = GLOBUS_FALSE;
        wsnt_NotificationMessageHolderType *
                input = &Notify->NotificationMessage.elements[i];

        result = globus_i_notification_consumer_service_deserialize_topic(
                            &input->Topic);

        if (result != GLOBUS_SUCCESS)
        {
            continue;
        }


        while (!called)
        {
            switch (consumer->state)
            {
                case CONSUMER_OK:
                    consumer->state = CONSUMER_ACTIVE;
                    globus_mutex_unlock(&globus_i_notification_lock);


                    consumer->callback(
                            consumer->arg,
                            &input->Topic,
                            input->ProducerReference,
                            &input->Message);

                    globus_mutex_lock(&globus_i_notification_lock);

                    if (consumer->state == CONSUMER_ACTIVE)
                    {
                        consumer->state = CONSUMER_OK;
                    }
                    else if (consumer->state == CONSUMER_ACTIVE_DESTROY_CALLED)
                    {
                        consumer->state = CONSUMER_DESTROYED;
                    }
                    called = GLOBUS_TRUE;
                    globus_cond_broadcast(&globus_i_notification_cond);

                    break;
                case CONSUMER_ACTIVE:
                    globus_cond_wait(
                            &globus_i_notification_cond,
                            &globus_i_notification_lock);
                    break;
                case CONSUMER_ACTIVE_DESTROY_CALLED:
                case CONSUMER_DESTROYED:
                    /* Can't call it now */
                    called = GLOBUS_TRUE;
                    break;
            }
        }
    }

finish_out:
    result = globus_resource_finish(resource);

unlock_out:
    globus_mutex_unlock(&globus_i_notification_lock);
    return result;
}
/* Consumer_Notify_impl() */

/**
 * Deserialize the topic expression into an appropriate xsd type if the
 * dialect is known to us. Otherwise, leave it alone.
 *
 * @param topic
 *     TopicExpression which will be potentially modified with a new
 *     value of any if the Dialect attribute refers to one that this
 *     function knows of.
 */
static
globus_result_t
globus_i_notification_consumer_service_deserialize_topic(
    wsnt_TopicExpressionType *          topic)
{
    xsd_string *                        topic_string = NULL;
    globus_xml_buffer *                 buffer;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_soap_message_handle_t        message_handle;

    /* Deserialize the notification topic to a string if it is one of the
     * known string-based topic dialects.
     */
    if (topic->_Dialect &&
        topic->any && topic->any->any_info == &globus_xml_buffer_info &&
        (strcmp(*topic->_Dialect, SIMPLE_TOPIC_EXPRESSION) == 0 ||
        strcmp(*topic->_Dialect, CONCRETE_TOPIC_EXPRESSION) == 0 ||
        strcmp(*topic->_Dialect, FULL_TOPIC_EXPRESSION) == 0))
    {
        buffer = topic->any->value;

        result = globus_soap_message_handle_init_from_memory(
            &message_handle,
            buffer->buffer,
            buffer->length);

        if (result != GLOBUS_SUCCESS)
        {
            goto out;
        }

        result = xsd_string_init(&topic_string);

        if (result != GLOBUS_SUCCESS || topic_string == NULL)
        {
            goto destroy_handle_out;
        }

        result = xsd_string_deserialize_contents(
            NULL,
            topic_string,
            message_handle,
            0);

        if (result != GLOBUS_SUCCESS)
        {
            goto destroy_topic_out;
        }

        globus_xml_buffer_destroy(buffer);

        topic->any->any_info = &xsd_string_contents_info;
        topic->any->value = topic_string;

        topic_string = NULL;
    }
    else
    {
        /* Don't know this dialect, leave it as an xml buffer */
        goto out;
    }


destroy_topic_out:
    if (topic_string != NULL)
    {
        xsd_string_destroy(topic_string);
    }
destroy_handle_out:
    globus_soap_message_handle_destroy(message_handle);
out:
    if (result != GLOBUS_SUCCESS)
    {
        return GLOBUS_NOTIFICATION_CONSUMER_ERROR_DESERIALIZING_TOPIC(
                globus_error_get(result));
    }
    else
    {
        return result;
    }
}
/* globus_i_notification_consumer_service_deserialize_topic() */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
