/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_operation_provider.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "version.h"

#include "wsrl_DestroyType.h"
#include "wsrl_DestroyResponseType.h"

#define WSRL_IMMEDIATE_RESOURCE_TERMINATION 1

#include "wsrl_i_faults.h"

/**
 * @page wsrl_ImmediateResourceTermination_provider ImmediateResourceTermination Provider
 *
 * <h2>Provider Usage Overview</h2>
 * This operation provider implements the ImmediateResourceTermination portType
 * defined in
 * http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.wsdl .
 * This portType consists of the @a Destroy operation and
 * related faults.
 *
 * The WSRF service engine will automatically register this provider to handle
 * the Destroy operation for services which use the port type.
 */

GlobusExtensionDeclareModule(wsrl_ImmediateResourceTermination);

/* Prototypes */
static
globus_result_t
wsrl_Destroy_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrl_DestroyType *                  input);

static
globus_result_t
wsrl_Destroy_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrl_DestroyType *                  input,
    wsrl_DestroyResponseType *          output,
    char **                             fault_name,
    void **                             fault);

static
int
wsrl_ImmediateResourceTermination_activate(void);

static
int
wsrl_ImmediateResourceTermination_deactivate(void);

/* Local Variables */
static xsd_QName wsrl_Destroy_qname =
{
    "http://www.globus.org/docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.wsdl",
    "Destroy"
};

static
globus_operation_provider_descriptor_t
wsrl_Destroy_descriptor =
{
    &wsrl_Destroy_qname,
    "Destroy",
    (void *)wsrl_Destroy_init_provider,
    (void *)wsrl_Destroy_provider,
};

GlobusExtensionDefineModule(wsrl_ImmediateResourceTermination) =
{
    "wsrl_ImmediateResourceTermination",
    wsrl_ImmediateResourceTermination_activate,
    wsrl_ImmediateResourceTermination_deactivate,
    NULL,
    NULL,
    &local_version
};

/* Implementation */

static
globus_result_t
wsrl_Destroy_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrl_DestroyType *                  input)
{
    return GLOBUS_SUCCESS;
}
/* wsrp_GetResourceProperty_init_provider() */

static
globus_result_t
wsrl_Destroy_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrl_DestroyType *                  input,
    wsrl_DestroyResponseType *          output,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result;
    globus_resource_t                   resource;
    globus_object_t *                   error;

    result = globus_wsrf_core_get_resource(handle, service, &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = wsrl_l_ResourceUnknownFaultType_create(
                fault_name,
                (wsrl_ResourceUnknownFaultType **) fault);

        goto out;
    }

    result = globus_resource_finish_and_destroy(resource);

    if (result != GLOBUS_SUCCESS)
    {
        error = globus_error_get(result);

        if (globus_error_match(error,
                    GLOBUS_WSRF_RESOURCE_MODULE,
                    GLOBUS_WSRF_RESOURCE_ERROR_TYPE_ALREADY_DESTROYED))
        {
            result = wsrl_l_ResourceNotDestroyedFaultType_create(
                    fault_name,
                    (wsrl_ResourceNotDestroyedFaultType **) fault);
        }
        else
        {
            result = wsrl_l_ResourceUnknownFaultType_create(
                    fault_name,
                    (wsrl_ResourceUnknownFaultType **) fault);
        }
        globus_object_free(error);

        goto out;
    }

    return GLOBUS_SUCCESS;

out:
    return result;
}
/* wsrp_GetResourceProperty_provider() */


static
int
wsrl_ImmediateResourceTermination_activate(void)
{
    int                                 res = 0;
    GlobusFuncName(wsrl_ImmediateResourceTermination_activate);
    
    res = globus_extension_registry_add(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrl_Destroy_qname,
        GlobusExtensionMyModule(wsrl_ImmediateResourceTermination),
        &wsrl_Destroy_descriptor);

    return res;
}
/* wsrl_ImmediateResourceTermination_activate() */

static
int
wsrl_ImmediateResourceTermination_deactivate(void)
{
    GlobusFuncName(wsrl_ImmediateResourceTermination_deactivate);

    globus_extension_registry_remove(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrl_Destroy_qname);

    return 0;
}
/* wsrl_ImmediateResourceTermination_deactivate() */
