/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_operation_provider.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"

#include "xsd_dateTime.h"

#include "wsrl_SetTerminationTimeType.h"
#include "wsrl_SetTerminationTimeResponseType.h"

#include "version.h"

/*
 * This define causes the wsrl_i_faults.h to define faults relevant to this
 * port type.
 */
#define WSRL_SCHEDULED_RESOURCE_TERMINATION 1
/* We don't implement any sort of change policy, so we don't use this fault
 * type.
 */
#define NEED_TERMINATION_TIME_CHANGE_REJECTED_FAULT 0

#include "wsrl_i_faults.h"

/**
 * @page wsrl_ScheduledResourceTermination_provider ScheduledResourceTermination Provider
 *
 * <h2>Provider Usage Overview</h2>
 * This operation provider implements the ScheduledResourceTermination portType
 * defined in
 * http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.wsdl .
 * This portType consists of the @a SetTerminationTime operation, @a
 * TerminationTime and @a CurrentTime resource properties,  and
 * related faults.
 *
 * The WSRF service engine will automatically register this provider to handle
 * the SetTerminationTime operation for services which use the port type.
 *
 * <h2>Notes</h2>
 * - The CurrentTime and TerminationTime resource properties are stored in GMT.
 *
 * <h2>Limitations</h2>
 * - This provider does not implement any policies regarding what is an
 * acceptable change of the TerminationTime. The
 * TerminationTimeChangeRejectedFaultType is never issued by this provider.
 */
GlobusExtensionDeclareModule(wsrl_ScheduledResourceTermination);

/* Prototypes */
static
globus_result_t
wsrl_ScheduledResourceTermination_init_resource(
    const char *                        resource_id);

static
globus_result_t
wsrl_SetTerminationTime_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrl_SetTerminationTimeType *       SetTerminationTime);

static
globus_result_t
wsrl_SetTerminationTime_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrl_SetTerminationTimeType *       SetTerminationTime,
    wsrl_SetTerminationTimeResponseType *
                                        SetTerminationTimeResponse,
    char **                             fault_name,
    void **                             fault);

static
void
wsrl_CurrentTime_callback(
    void *                              arg,
    const xsd_QName *                   qname,
    void **                             property);

static
time_t
wsrl_l_timegm(
    const struct tm *                   tm);

static
int
wsrl_ScheduledResourceTermination_activate(void);

static
int
wsrl_ScheduledResourceTermination_deactivate(void);

/* Local Variables */
static xsd_QName wsrl_SetTerminationTime_qname =
{
    "http://www.globus.org/docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.wsdl",
    "SetTerminationTime"
};

static xsd_QName wsrl_CurrentTime_qname =
{
    "http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd",
    "CurrentTime"
};

static xsd_QName wsrl_TerminationTime_qname =
{
    "http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd",
    "TerminationTime"
};

static
globus_operation_provider_descriptor_t
wsrl_SetTerminationTime_descriptor =
{
    &wsrl_SetTerminationTime_qname,
    "SetTerminationTime",
    (void *)wsrl_SetTerminationTime_init_provider,
    (void *)wsrl_SetTerminationTime_provider,
    wsrl_ScheduledResourceTermination_init_resource
};

GlobusExtensionDefineModule(wsrl_ScheduledResourceTermination) =
{
    "wsrl_ScheduledResourceTermination",
    wsrl_ScheduledResourceTermination_activate,
    wsrl_ScheduledResourceTermination_deactivate,
    NULL,
    NULL,
    &local_version
};

/* Implementation */
static
globus_result_t
wsrl_ScheduledResourceTermination_init_resource(
    const char *                        resource_id)
{
    globus_resource_t                   resource;
    globus_result_t                     result;

    result = globus_resource_find(resource_id, &resource);

    if (result != GLOBUS_SUCCESS)
    {
        goto out;
    }

    result = globus_resource_create_property(
        resource,
        &wsrl_TerminationTime_qname,
        &xsd_dateTime_info,
        NULL);

    if (result != GLOBUS_SUCCESS)
    {
        goto finish_out;
    }

    result = globus_resource_create_property_callback(
        resource,
        &wsrl_CurrentTime_qname,
        &xsd_dateTime_info,
        wsrl_CurrentTime_callback,
        NULL);

finish_out:
    globus_resource_finish(resource);
out:
    return result;
}
/* wsrl_ScheduledResourceTermination_init_resource() */


static
globus_result_t
wsrl_SetTerminationTime_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrl_SetTerminationTimeType *       SetTerminationTime)
{
    /* NOOP */
    return GLOBUS_SUCCESS;
}

static
globus_result_t
wsrl_SetTerminationTime_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrl_SetTerminationTimeType *       SetTerminationTime,
    wsrl_SetTerminationTimeResponseType *
                                        SetTerminationTimeResponse,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result;
    globus_resource_t                   resource;
    globus_abstime_t                    termination_time;
    time_t                              now;
    xsd_dateTime *                      termination_time_rp;

    result = globus_wsrf_core_get_resource(handle, service, &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = wsrl_l_ResourceUnknownFaultType_create(
                fault_name,
                (wsrl_ResourceUnknownFaultType **) fault);

        goto out;
    }

    termination_time.tv_sec = wsrl_l_timegm(
            &SetTerminationTime->RequestedTerminationTime);

    termination_time.tv_nsec = 0;

    result = globus_resource_set_destroy_time(resource, &termination_time);

    if (result != GLOBUS_SUCCESS)
    {
        result = wsrl_l_UnableToSetTerminationTimeFaultType_create(
                fault_name,
                (wsrl_UnableToSetTerminationTimeFaultType **) fault);


        goto finish_out;
    }

    now = time(NULL);

    memcpy(&SetTerminationTimeResponse->NewTerminationTime,
           &SetTerminationTime->RequestedTerminationTime,
           sizeof(xsd_dateTime));

    globus_libc_gmtime_r(&now, &SetTerminationTimeResponse->CurrentTime);

    xsd_dateTime_init(&termination_time_rp);

    memcpy(termination_time_rp,
           &SetTerminationTime->RequestedTerminationTime,
           sizeof(xsd_dateTime));

    result = globus_resource_set_property(
            resource,
            &wsrl_TerminationTime_qname,
            termination_time_rp);

    if (result != GLOBUS_SUCCESS)
    {
        goto finish_out;
    }

    result = globus_resource_finish(resource);

    if (result != GLOBUS_SUCCESS)
    {
        goto out;
    }
    return GLOBUS_SUCCESS;

finish_out:
    globus_resource_finish(resource);
out:
    return result;
}
/* wsrl_SetTerminationTime_provider() */


static
void
wsrl_CurrentTime_callback(
    void *                              arg,
    const xsd_QName *                   qname,
    void **                             property)
{
    time_t                              now_time;
    xsd_dateTime *                      now;

    globus_assert(
            strcmp(qname->Namespace,
                    wsrl_CurrentTime_qname.Namespace) == 0);
    globus_assert(
            strcmp(qname->local,
                    wsrl_CurrentTime_qname.local) == 0);

    now_time = time(NULL);

    xsd_dateTime_init((xsd_dateTime **) property);
    now = *property;

    globus_libc_gmtime_r(&now_time, now);

    return;
}
/* wsrl_CurrentTime_callback() */

/**
 * Convert a struct tm to a time_t without accounting for the local timezone
 * or modifying the struct tm contents.
 *
 * This may get a little kooky when the TZ shift occurs near the DST change.
 */
static
time_t
wsrl_l_timegm(
    const struct tm *                   tm)
{
    struct tm                           tmp;
    struct tm                           local_tm;
    time_t                              local_time;
    time_t                              shifted_time;

    memcpy(&tmp, tm, sizeof(struct tm));

    /* convert the tm to time_t in the local timezone */
    local_time = mktime(&tmp);

    /* create a new tm with this time (but without the TZ-related parts of the
     * tm set) */
    globus_libc_gmtime_r(&local_time, &local_tm);

    /* convert the time to local time again, causing it to be shifted by
     * time zone twice
     */
    shifted_time = mktime(&local_tm);

    /*
     * The difference between shifted_time and local_time is the TZ correction.
     */
    return local_time + (local_time - shifted_time);

    /*
     * This snippet is based on the timegm manpage, but is not thread safe
     * 
     * tz = globus_libc_getenv("TZ");
     * setenv("TZ", "", 1);
     * tzset();
     * termination_time.tv_sec = mktime(
     *         &SetTerminationTime->RequestedTerminationTime);
     * if (tz != NULL)
     * {
     *     globus_libc_setenv("TZ", tz, 1);
     * }
     * else
     * {
     *     globus_libc_unsetenv("TZ");
     * }
     * tzset();
     */
}
/* wsrl_l_timegm() */

static
int
wsrl_ScheduledResourceTermination_activate(void)
{
    int                                 res = 0;
    GlobusFuncName(wsrl_ScheduledResourceTermination_activate);
    
    res = globus_extension_registry_add(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrl_SetTerminationTime_qname,
        GlobusExtensionMyModule(wsrl_ScheduledResourceTermination),
        &wsrl_SetTerminationTime_descriptor);

    return res;
}
/* wsrl_ScheduledResourceTermination_activate() */

static
int
wsrl_ScheduledResourceTermination_deactivate(void)
{
    GlobusFuncName(wsrl_ScheduledResourceTermination_deactivate);

    globus_extension_registry_remove(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrl_SetTerminationTime_qname);

    return 0;
}
/* wsrl_ScheduledResourceTermination_deactivate() */
