/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_operation_provider.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "version.h"

#include "wsrp_GetResourcePropertyResponseType.h"

#define WSRP_GET_RESOURCE_PROPERTY 1
#include "wsrp_i_faults.h"

/**
 * @page wsrp_GetResourceProperty_provider GetResourceProperty Provider
 *
 * <h2>Provider Usage Overview</h2>
 * This operation provider implements the GetResourceProperty portType defined
 * in
 * http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl .
 * This portType consists of the @a GetResourceProperty operation and
 * related faults.
 *
 * The WSRF service engine will automatically register this provider to handle
 * the GetResourceProperties operation for services which use the port type.
 */
GlobusExtensionDeclareModule(wsrp_GetResourceProperty);

/* Prototypes */
static
globus_result_t
wsrp_GetResourceProperty_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    xsd_QName *                         input);

static
globus_result_t
wsrp_GetResourceProperty_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    xsd_QName *                         input,
    wsrp_GetResourcePropertyResponseType *
                                        output,
    char **                             fault_name,
    void **                             fault);

static
int
wsrp_GetResourceProperty_activate(void);

static
int
wsrp_GetResourceProperty_deactivate(void);

/* Local Variables */
static xsd_QName wsrp_GetResourceProperty_qname =
{
    "http://www.globus.org/docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl",
    "GetResourceProperty"
};

static
globus_operation_provider_descriptor_t
wsrp_GetResourceProperty_descriptor =
{
    &wsrp_GetResourceProperty_qname,
    "GetResourceProperty",
    (void *)wsrp_GetResourceProperty_init_provider,
    (void *)wsrp_GetResourceProperty_provider
};

GlobusExtensionDefineModule(wsrp_GetResourceProperty) =
{
    "wsrp_GetResourceProperty",
    wsrp_GetResourceProperty_activate,
    wsrp_GetResourceProperty_deactivate,
    NULL,
    NULL,
    &local_version
};

/* Implementation */

static
globus_result_t
wsrp_GetResourceProperty_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    xsd_QName *                         input)
{
    return GLOBUS_SUCCESS;
}
/* wsrp_GetResourceProperty_init_provider() */

static
globus_result_t
wsrp_GetResourceProperty_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    xsd_QName *                         input,
    wsrp_GetResourcePropertyResponseType *
                                        output,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result;
    globus_resource_t                   resource;
    xsd_any *                           value;
    globus_xsd_type_info_t              info;
    void *                              property;
    xsd_any *                           prop_any = NULL;

    result = globus_wsrf_core_get_resource(handle, service, &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = wsrp_l_ResourceUnknownFaultType_create(
                fault_name,
                (wsrp_ResourceUnknownFaultType **) fault);

        goto out;
    }

    result = globus_resource_get_property(
            resource,
            input,
            &property,
            &info);

    if (result != GLOBUS_SUCCESS)
    {
        globus_object_t *               error;

        error = globus_error_get(result);

        if (globus_error_match(error,
                    GLOBUS_WSRF_RESOURCE_MODULE,
                    GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
        {
            result = wsrp_l_InvalidResourcePropertyQNameFaultType_create(
                    fault_name,
                    (wsrp_InvalidResourcePropertyQNameFaultType **) fault);
        }
        else
        {
            result = wsrp_l_ResourceUnknownFaultType_create(
                    fault_name,
                    (wsrp_ResourceUnknownFaultType **) fault);
        }
        globus_object_free(error);

        goto finish_out;
    }

    if (info == &xsd_any_info)
    {
        prop_any = property;
    }
    if ((prop_any != NULL &&
         prop_any->any_info != NULL &&
         prop_any->value != NULL) || 
        (prop_any == NULL && property != NULL))
    {
        value = xsd_any_array_push(&output->any);

        value->any_info = info;
        result = xsd_QName_copy(&value->element, input);

        if (result != GLOBUS_SUCCESS)
        {
            goto finish_out;
        }

        result = info->copy(&value->value, property);
        if (result != GLOBUS_SUCCESS)
        {
            goto finish_out;
        }
    }

    globus_resource_finish(resource);

    return GLOBUS_SUCCESS;

finish_out:
    globus_resource_finish(resource);
out:
    return result;
}
/* wsrp_GetResourceProperty_provider() */

static
int
wsrp_GetResourceProperty_activate(void)
{
    int                                 res = 0;
    GlobusFuncName(wsrp_GetResourcePropertyActivate);
    
    res = globus_extension_registry_add(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_GetResourceProperty_qname,
        GlobusExtensionMyModule(wsrp_GetResourceProperty),
        &wsrp_GetResourceProperty_descriptor);

    return res;
}
/* wsrp_GetResourceProperty_activate() */

static
int
wsrp_GetResourceProperty_deactivate(void)
{
    GlobusFuncName(wsrp_GetResourceProperty_deactivate);

    globus_extension_registry_remove(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_GetResourceProperty_qname);

    return 0;
}
/* wsrp_GetResourceProperty_deactivate() */
