/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_ws_security.h"
#include "globus_i_ws_security.h"

globus_result_t
globus_ws_security_create_pkipath(
    globus_gsi_cred_handle_t            cred,
    char **                             encoded_buffer,
    int *                               encoded_length)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    STACK_OF(X509) *                    chain = NULL;
    X509 *                              cert = NULL;
    unsigned char *                     pack_buff = NULL;
    int                                 pack_buff_length = 0;
    char *                              pkipath_buff = NULL;
    int                                 pkipath_length = 0;
    BIO *                               b64;
    BIO *                               mem;
    BIO *                               bio;
    BUF_MEM *                           bm;
    STACK_OF(X509) *                    pkipath_stack = NULL;
    GlobusFuncName(globus_ws_security_create_pkipath);
    GlobusWSSecurityDebugEnter();

    result = globus_gsi_cred_get_cert_chain(cred, &chain);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusWSSecurityErrorCreatingPKIPath(
            result, "Failed to get cert chain from credential");
        goto exit;
    }

    pkipath_stack = sk_X509_new_null();
    
    while(sk_X509_num(chain) > 0)
    {
        sk_X509_push(pkipath_stack, sk_X509_pop(chain));
    }

    sk_X509_free(chain);

    result = globus_gsi_cred_get_cert(
        cred, &cert);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusWSSecurityErrorCreatingPKIPath(
            result, "Failed to get certificate from credential");
        goto free_stack;
    }

    sk_X509_push(pkipath_stack, (char *)cert);
    
    pack_buff = ASN1_seq_pack(
        pkipath_stack, i2d_X509, &pack_buff, &pack_buff_length);
    if(!pack_buff)
    {
        result = GlobusWSSecurityErrorCreatingPKIPath(
            result, 
            "Failed to pack certificates"
            " into DER encoded ASN1 sequence");
        goto free_stack;
    }

    b64 = BIO_new(BIO_f_base64());
    mem = BIO_new(BIO_s_mem());
    bio = BIO_push(b64, mem);

    if(BIO_write(bio, pack_buff, pack_buff_length) <= 0)
    {
        result = GlobusWSSecurityErrorOpenSSL(
            "Failed to base64 encode DER sequence");
        goto free_bio;
    }
    
    if(BIO_flush(bio) <= 0)
    {
        result = GlobusWSSecurityErrorOpenSSL(
            "Failed to flush base64 encoding BIO");
        goto free_bio;
    }

    BIO_get_mem_ptr(bio, &bm);
    
    pkipath_buff = globus_malloc(bm->length + 1);
    pkipath_length = bm->length;
    memcpy(pkipath_buff, bm->data, pkipath_length);
    pkipath_buff[pkipath_length] = '\0';

    GlobusWSSecurityDebugPrintf(
        GLOBUS_L_WS_SECURITY_DEBUG_PKIPATH, 
        ("============== ENCODED PKI PATH BUFFER ============= (size: %d)"
         "\n%*s\n"
         "----------------------------------------------------\n", 
         pkipath_length, pkipath_length, pkipath_buff));
    
    *encoded_buffer = pkipath_buff;
    *encoded_length = pkipath_length;
    
 free_bio:
    
    BIO_free_all(bio);

/*  free_pack_buff: */

    OPENSSL_free(pack_buff);

 free_stack:

    sk_X509_pop_free(pkipath_stack, X509_free);

/*  free_cert: */
/*     X509_free(cert); */

 exit:
    
    GlobusWSSecurityDebugExit();
    return result;
}

globus_result_t
globus_ws_security_get_credential(
    const unsigned char *               encoded_buff,
    int                                 encoded_length,
    globus_gsi_cred_handle_t *          new_cred)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    unsigned char *                     dec;
    int                                 length;
    BIO *                               b64;
    BIO *                               mem;
    BIO *                               bio;
    STACK_OF(X509) *                    pkipath_stack;
    STACK_OF(X509) *                    chain;
    X509 *                              user_cert;
    globus_gsi_cred_handle_t            cred;
    GlobusFuncName(globus_ws_security_get_credential);
    GlobusWSSecurityDebugEnter();

    b64 = BIO_new(BIO_f_base64());
    mem = BIO_new_mem_buf((unsigned char *)encoded_buff, encoded_length);
    bio = BIO_push(b64, mem);

    length = BIO_pending(bio);
    dec = globus_malloc(length);
    if(!dec)
    {
        result = GlobusWSSecurityErrorOutOfMemory(length);
        goto free_bio;
    }

    length = BIO_read(bio, dec, length);
    pkipath_stack = ASN1_seq_unpack(dec, length,
                                    (char *(*)())d2i_X509,
                                    (void (*)(void *))X509_free);
    if(!pkipath_stack)
    {
        result = GlobusWSSecurityErrorOpenSSL(
            "Could not unpack the DER encoded PKI path"
            "buffer into a stack of X509 certificates");
        goto free_dec;
    }
    
    user_cert = sk_X509_pop(pkipath_stack);

    result = globus_gsi_cred_handle_init(&cred, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusWSSecurityErrorImportingCredential(
            result, "can't create handle for credential");
        goto free_stack;
    }

    result = globus_gsi_cred_set_cert(cred, user_cert);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusWSSecurityErrorImportingCredential(
            result, "can't set certificate in credential handle");
        goto free_cred;
    }

    /* reverse stack for chain */
    chain = sk_X509_new_null();
    while(sk_X509_num(pkipath_stack) > 0)
    {
        sk_X509_push(chain, sk_X509_pop(pkipath_stack));
    }
        
    result = globus_gsi_cred_set_cert_chain(cred, chain);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusWSSecurityErrorImportingCredential(
            result, "can't set certificate chain in credential handle");
        goto free_cred;
    }

    *new_cred = cred;

    goto free_stack;

 free_cred:

    globus_gsi_cred_handle_destroy(cred);

 free_stack:

    sk_X509_pop_free(pkipath_stack, X509_free);

/* free chain */

    sk_X509_pop_free(chain, X509_free);

/*  free_cert: */

    X509_free(user_cert);

 free_dec:

    globus_free(dec);

 free_bio:

    BIO_free_all(bio);

/* exit:*/

    GlobusWSSecurityDebugExit();
    return result;
}
