/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_i_handler_wsse_cred.h"
#include "globus_ws_security.h"
#include "openssl/x509.h"

#define WSSE_CACHED_USER_SUBJECT_HASH "WSSE_CACHED_USER_SUBJECT_HASH"

static unsigned long int                       cached_hash = 0;
static globus_i_handler_wsse_cache_t           pkipath_table = NULL;
static globus_mutex_t                          pkipath_table_mutex;

globus_result_t
globus_i_handler_wsse_cred_pkipath_table_init()
{
    globus_reltime_t                    expiration;
    globus_mutex_init(&pkipath_table_mutex, NULL);
    globus_mutex_lock(&pkipath_table_mutex);

    cached_hash = 0;
    pkipath_table = NULL;

    /* expiration = 5 mins */
    GlobusTimeReltimeSet(expiration, (5 * 60), 0);
    globus_i_handler_wsse_cache_init(
	&pkipath_table,
	20, /* cache size for creds */
	expiration);
    globus_mutex_unlock(&pkipath_table_mutex);

    return GLOBUS_SUCCESS;
}

void
globus_i_handler_wsse_cred_pkipath_table_destroy()
{
    globus_mutex_lock(&pkipath_table_mutex);
    globus_i_handler_wsse_cache_destroy(pkipath_table);
    globus_mutex_unlock(&pkipath_table_mutex);

    globus_mutex_destroy(&pkipath_table_mutex);

    cached_hash = 0;
    pkipath_table = NULL;
}

unsigned long
globus_i_handler_wsse_cred_subject_hash(
    globus_gsi_cred_handle_t            credential)
{
    unsigned long                       hash;
    X509_NAME *                         name;

    globus_gsi_cred_get_X509_subject_name(credential, &name);
    hash = X509_NAME_hash(name);
    X509_NAME_free(name);
    return hash;
}

void 
globus_i_handler_wsse_cache_init (
    globus_i_handler_wsse_cache_t *     new_cache,
    int                                 capacity,
    globus_reltime_t                    reltime_limit)
{
    globus_i_handler_wsse_cache_t       cache;

    if ( new_cache == NULL ) return;


    cache = globus_malloc(sizeof(struct globus_i_handler_wsse_cache_s));
    if(!cache)
    {
        *new_cache = NULL;
        return;
    }

    globus_hashtable_init (&(cache->handlemap),
                           capacity,
                           globus_hashtable_ulong_hash,
                           globus_hashtable_ulong_keyeq);
    globus_fifo_init (&(cache->handles));

    cache->capacity_limit = capacity;
    cache->entry_count = 0;
    GlobusTimeReltimeCopy(cache->time_limit, reltime_limit);

    *new_cache = cache;
}

void
globus_i_handler_wsse_cache_destroy (
    globus_i_handler_wsse_cache_t       cache)
{
    if(cache == NULL) return;

    while(!globus_fifo_empty(&(cache->handles)))
    {
        globus_gsi_cred_handle_t        cred;
        char *                          pkipath_buff;
        int                             pkipath_length;
        globus_i_handler_wsse_cache_remove(
            cache,
            (int)globus_fifo_dequeue(&(cache->handles)),
            &cred, &pkipath_buff, &pkipath_length);
        
        if(cred)
        {
            globus_gsi_cred_handle_destroy(cred);
        }

        if(pkipath_buff)
        {
            free(pkipath_buff);
        }
    }
    globus_hashtable_destroy_all(
	&(cache->handlemap),
	globus_i_handler_wsse_pkipath_entry_destroy);
    globus_fifo_destroy (&(cache->handles));

    globus_free(cache);
}

globus_result_t
globus_i_handler_wsse_cache_insert(
    globus_i_handler_wsse_cache_t       cache,
    globus_gsi_cred_handle_t            credential,
    unsigned long *                     hash)
{
    globus_i_handler_wsse_pkipath_entry_t * entry;
    globus_result_t                     result = GLOBUS_SUCCESS;
    unsigned long                       subject_hash;

    if(credential == NULL) return GLOBUS_FAILURE;

    subject_hash = globus_i_handler_wsse_cred_subject_hash(
        credential);
    
    *hash = subject_hash;

    if (cache == NULL) return GLOBUS_FAILURE;

    if ( cache->entry_count > cache->capacity_limit ) 
    {
        globus_gsi_cred_handle_t        cred;
        char *                          pkipath_buff;
        globus_i_handler_wsse_cache_remove ( 
            cache,
            (int)globus_fifo_dequeue ( &(cache->handles) ),
            &cred, &pkipath_buff, NULL);
        
        if(cred)
        {
            globus_gsi_cred_handle_destroy(cred);
        }

        if(pkipath_buff)
        {
            globus_free(pkipath_buff);
        }
    }

    entry = globus_malloc(sizeof(globus_i_handler_wsse_pkipath_entry_t));
    if(!entry)
    {
        result = GlobusWSSEErrorOutOfMemory(
            sizeof(globus_i_handler_wsse_pkipath_entry_t));
        return result;
    }
    memset(entry, 0, sizeof(globus_i_handler_wsse_pkipath_entry_t));

    entry->subject = subject_hash;
    result = globus_gsi_cred_handle_copy(credential, &entry->credential);
    if(result != GLOBUS_SUCCESS)
    {
        goto free_entry;
    }

    result = globus_ws_security_create_pkipath(
        entry->credential, &entry->pkipath_buffer, &entry->pkipath_length);
    if(result != GLOBUS_SUCCESS)
    {
        goto free_entry;
    }

    GlobusTimeAbstimeGetCurrent(entry->entry_time);

    globus_hashtable_insert(
        &(cache->handlemap),
        (void *)entry->subject, 
        (void *)entry);
    globus_fifo_enqueue (&(cache->handles), (void *)entry->subject);

    cache->entry_count += 1;

    goto exit;

 free_entry:

    globus_i_handler_wsse_pkipath_entry_destroy(entry);

 exit:

    return result;
}

globus_result_t
globus_i_handler_wsse_cache_lookup (
    globus_i_handler_wsse_cache_t       cache,
    unsigned long                       subject_hash,
    globus_gsi_cred_handle_t *          credential,
    char **                             pkipath_buffer,
    int *                               pkipath_length)
{
    globus_i_handler_wsse_pkipath_entry_t * entry;
    globus_result_t                     result = GLOBUS_SUCCESS;

    if ( cache == NULL ) return GLOBUS_FAILURE;

    entry = globus_hashtable_lookup (&(cache->handlemap), 
                                     (void *)subject_hash);
    if(entry)
    {
        if(!globus_time_reltime_is_infinity(&cache->time_limit))
        {
            char *                      tmp_buff;
            int                         tmp_length;
            globus_gsi_cred_handle_t    cred;
            globus_reltime_t            diff;
            globus_abstime_t            current;
            GlobusTimeAbstimeGetCurrent(current);
            GlobusTimeAbstimeDiff(diff, entry->entry_time, current);
            if(globus_reltime_cmp(&diff, &cache->time_limit) >= 0)
            {
                /* expired cache entry.  must remove */
                globus_i_handler_wsse_cache_remove(cache, subject_hash, 
                                                   &cred,
                                                   &tmp_buff, &tmp_length);
                if(tmp_buff)
                {
                    globus_free(tmp_buff);
                }

                if(cred)
                {
                    globus_gsi_cred_handle_destroy(cred);
                }

                result = GLOBUS_FAILURE;
                goto exit;
            }
        }

        if(credential)
        {
            *credential = entry->credential;
        }
        
        if(pkipath_buffer)
        {
            *pkipath_buffer = entry->pkipath_buffer;
            *pkipath_length = entry->pkipath_length;
        }
    }
    else
    {
        result = GLOBUS_FAILURE;
    }

 exit:

    return result;
}

globus_result_t
globus_i_handler_wsse_cache_remove (
    globus_i_handler_wsse_cache_t       cache,
    unsigned long                       subject_hash,
    globus_gsi_cred_handle_t *          credential,
    char **                             pkipath_buffer,
    int *                               pkipath_length)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_i_handler_wsse_pkipath_entry_t * entry;

    if ( cache == NULL ) return GLOBUS_FAILURE;

    entry = globus_hashtable_remove (&(cache->handlemap), 
                                     (void *)subject_hash);
    globus_fifo_remove (&(cache->handles), (void *)subject_hash);
    if ( entry != NULL ) cache->entry_count -= 1;

    if(entry)
    {
        if(pkipath_buffer)
        {
            *pkipath_buffer = entry->pkipath_buffer;
            *pkipath_length = entry->pkipath_length;
        }
        else
        {
            globus_free(entry->pkipath_buffer);
        }

        if(credential)
        {
            *credential = entry->credential;
        }
        else
        {
            globus_gsi_cred_handle_destroy(entry->credential);
        }

        globus_free(entry);
    }

    return result;
}

void
globus_i_handler_wsse_pkipath_entry_destroy(
    globus_i_handler_wsse_pkipath_entry_t * entry)
{
    if(entry)
    {
        if(entry->pkipath_buffer)
        {
            globus_free(entry->pkipath_buffer);
        }

        if(entry->credential)
        {
            globus_gsi_cred_handle_destroy(
                entry->credential);
        }

        globus_free(entry);
    }
}

globus_result_t
globus_i_handler_wsse_load_local_cred(
    globus_soap_message_handle_t        message_handle,
    globus_gsi_cred_handle_t *          credential,
    char **                             pkipath,
    int *                               pkipath_length)
{
    globus_gsi_cred_handle_t            user_cred = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    unsigned long                       hash;
    int                                 user_cred_initialized = 0;
    X509_NAME *                         x509n = NULL;


    user_cred = globus_soap_message_handle_get_attr(
        message_handle,
        GLOBUS_HANDLER_WSSE_USER_CREDENTIAL_KEY);
    if(!user_cred)
    {
        char *                          desired_subject;
        
        desired_subject =
        globus_soap_message_handle_get_attr(
            message_handle,
            GLOBUS_HANDLER_WSSE_DESIRED_SUBJECT_KEY);

        if(desired_subject)
        {
            x509n = X509_NAME_new();
            result = globus_gsi_cert_utils_get_x509_name(
                desired_subject, strlen(desired_subject), x509n);
            if(result != GLOBUS_SUCCESS)
            {
                result = GlobusWSSEErrorPKIPath(
                    result,
                    "Failed to get PKIPath from local user credential");
                goto free_x509n;
            }

            globus_mutex_lock(&pkipath_table_mutex);
            result = globus_i_handler_wsse_cache_lookup(
                pkipath_table, X509_NAME_hash(x509n),
                credential, pkipath, pkipath_length);
            if(result == GLOBUS_SUCCESS)
            {
                globus_mutex_unlock(&pkipath_table_mutex);
                goto free_x509n;
            }
            globus_mutex_unlock(&pkipath_table_mutex);
        }
           
	if(cached_hash != 0)
	{
	    result = globus_i_handler_wsse_cache_lookup(
		pkipath_table, cached_hash,
		credential, pkipath, pkipath_length);
	    if(result == GLOBUS_SUCCESS)
	    {
		globus_mutex_unlock(&pkipath_table_mutex);
		goto exit;
	    }
	    globus_mutex_unlock(&pkipath_table_mutex);
	}

  	user_cred_initialized = 1; 
        result = globus_gsi_cred_handle_init(&user_cred, NULL);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusWSSEErrorPKIPath(
                result, "Failed to load local user credential");
            goto free_x509n;
        }
        
        result = globus_gsi_cred_read(user_cred, x509n);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusWSSEErrorPKIPath(
                result, "Failed to load local user credential");
            goto free_handle;
        }

        globus_mutex_lock(&pkipath_table_mutex);
        result = globus_i_handler_wsse_cache_insert(
            pkipath_table, 
            user_cred,
            &hash);
        if(result != GLOBUS_SUCCESS)
        {
            globus_mutex_unlock(&pkipath_table_mutex);
            result = GlobusWSSEErrorPKIPath(
                result, "Failed to insert user credential into cache");
            goto free_handle;
        }

        result = globus_i_handler_wsse_cache_lookup(
            pkipath_table,
            hash,
            credential,
            pkipath,
            pkipath_length);
        if(result == GLOBUS_SUCCESS)
        {
	    globus_mutex_unlock(&pkipath_table_mutex);
	    cached_hash = hash;
	    goto free_handle;
        }
	globus_mutex_unlock(&pkipath_table_mutex);
    }
    else
    {
        result = globus_gsi_cred_get_X509_subject_name(user_cred, &x509n);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusWSSEErrorPKIPath(
                result, "Failed to load user cred");
            goto free_handle;
        }
        
        globus_mutex_lock(&pkipath_table_mutex);
        result = globus_i_handler_wsse_cache_lookup(
            pkipath_table, 
            X509_NAME_hash(x509n),
            credential,
            pkipath, 
            pkipath_length);
        if(result == GLOBUS_SUCCESS)
        {
            globus_mutex_unlock(&pkipath_table_mutex);
            goto free_handle;
        }
        globus_mutex_unlock(&pkipath_table_mutex);
    }

    globus_mutex_lock(&pkipath_table_mutex);
    result = globus_i_handler_wsse_cache_insert(
        pkipath_table,
        user_cred,
        &hash);
    if(result != GLOBUS_SUCCESS)
    {
        globus_mutex_unlock(&pkipath_table_mutex);
        result = GlobusWSSEErrorPKIPath(
            result, "Failed to insert credential into cache");
        goto free_handle;
    }

    result = globus_i_handler_wsse_cache_lookup(
        pkipath_table,
        hash,
        credential,
        pkipath,
        pkipath_length);
    if(result != GLOBUS_SUCCESS)
    {
        globus_mutex_unlock(&pkipath_table_mutex);
        result = GlobusWSSEErrorPKIPath(
            result, "Failed to get credential from cache");
        goto free_handle;
    }

    globus_mutex_unlock(&pkipath_table_mutex);

 free_handle:

    if(user_cred_initialized && user_cred)
    {
        globus_gsi_cred_handle_destroy(user_cred);
    }

 free_x509n:

    if(x509n)
    {
        X509_NAME_free(x509n);
    }

 exit:

    return result;
}
