/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "globus_i_handler_chain.h"
#include "globus_extension.h"
#include "globus_soap_message.h"

globus_extension_registry_t             globus_handler_registry =
{
    NULL,
    GLOBUS_FALSE,
    GLOBUS_FALSE
};

GlobusDebugDefine(GLOBUS_HANDLER);

static globus_list_t *                  globus_l_handler_extensions = NULL;
static globus_mutex_t                   globus_l_handler_extensions_mutex;

enum globus_l_handler_debug_levels
{
    GLOBUS_L_HANDLER_DEBUG_TRACE      = 1,
    GLOBUS_L_HANDLER_DEBUG_ERROR      = 2
};

#define GlobusHandlerDebugPrintf(level, message)                            \
    GlobusDebugPrintf(GLOBUS_HANDLER, level, message)

#define GlobusHandlerDebugEnter()                                           \
    GlobusHandlerDebugPrintf(                                               \
        GLOBUS_L_HANDLER_DEBUG_TRACE,                                       \
        ("[%s] Entering\n", _globus_func_name))

#define GlobusHandlerDebugExit()                                            \
    GlobusHandlerDebugPrintf(                                               \
        GLOBUS_L_HANDLER_DEBUG_TRACE,                                       \
        ("[%s] Exiting\n", _globus_func_name))

#define GlobusHandlerErrorHeaderElementNotFound() \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_HANDLER_MODULE, \
            NULL, \
            GLOBUS_HANDLER_ERROR_HEADER_ELEMENT_NOT_FOUND, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "No Header element found"))

#define GlobusHandlerErrorRegistryLookupFailed(CAUSE, ENDPOINT) \
    globus_error_put( \
                globus_error_construct_error( \
                    GLOBUS_HANDLER_MODULE, \
                    (CAUSE ? globus_error_get(CAUSE) : NULL), \
                    GLOBUS_HANDLER_ERROR_REGISTRY_LOOKUP_FAILED, \
                    __FILE__, \
                    _globus_func_name, \
                    __LINE__, \
                    "No handler found at endpoint: %s", ENDPOINT))

#define GlobusHandlerErrorDefaultTriggerAlreadyDefined()                  \
    globus_error_put(                                                     \
                globus_error_construct_error(                             \
                    GLOBUS_HANDLER_MODULE,                                \
                    NULL,                                                 \
                    GLOBUS_HANDLER_ERROR_DEFAULT_TRIGGER_ALREADY_DEFINED, \
                    __FILE__,                                             \
                    _globus_func_name,                                    \
                    __LINE__,                                             \
                    "A default trigger handler has already been defined"))

#define GlobusHandlerErrorNoTriggersDefined()                   \
    globus_error_put(                                           \
                globus_error_construct_error(                   \
                    GLOBUS_HANDLER_MODULE,                      \
                    NULL,                                       \
                    GLOBUS_HANDLER_ERROR_NO_TRIGGERS_DEFINED,   \
                    __FILE__,                                   \
                    _globus_func_name,                          \
                    __LINE__,                                   \
                    "No triggers defined"))

static int
globus_l_handler_activate(void);

static int
globus_l_handler_deactivate(void);

#include "version.h"

globus_module_descriptor_t              globus_i_handler_module =
{
    "globus_handler",
    globus_l_handler_activate,
    globus_l_handler_deactivate,
    GLOBUS_NULL,
    GLOBUS_NULL,
    &local_version
};

static int
globus_l_handler_activate(void)
{
    int rc;

    rc = globus_module_activate(GLOBUS_COMMON_MODULE);
    if(rc != GLOBUS_SUCCESS)
    {
        return rc;
    }

    GlobusHandlerDebugEnter();

    rc = globus_module_activate(GLOBUS_SOAP_MESSAGE_MODULE);
    if(rc != GLOBUS_SUCCESS)
    {
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return rc;
    }

    globus_mutex_init(&globus_l_handler_extensions_mutex, NULL);

    GlobusHandlerDebugExit();
    return GLOBUS_SUCCESS;
}

static
void
globus_l_handler_extension_destroy(
    void *                              ep)
{
    char *                              symbol = (char *) ep;

    globus_extension_deactivate(symbol);
    globus_free(symbol);
}

static int
globus_l_handler_deactivate(void)
{
    GlobusFuncName(globus_l_handler_deactivate);
    GlobusHandlerDebugEnter();

    globus_mutex_lock(&globus_l_handler_extensions_mutex); 
    if(globus_l_handler_extensions) 
    { 
        globus_list_destroy_all(globus_l_handler_extensions, 
                                globus_l_handler_extension_destroy); 
        globus_l_handler_extensions = NULL; 
    } 
    globus_mutex_unlock(&globus_l_handler_extensions_mutex); 
    globus_mutex_destroy(&globus_l_handler_extensions_mutex); 
    
    globus_module_deactivate(GLOBUS_SOAP_MESSAGE_MODULE);

    GlobusHandlerDebugExit();
    GlobusDebugDestroy(GLOBUS_HANDLER);

    globus_module_deactivate(GLOBUS_COMMON_MODULE);

    return GLOBUS_SUCCESS;
}

typedef struct globus_l_handler_s
{
    globus_extension_handle_t           extension_handle;
    globus_handler_invoke_func_t        invoke;
    globus_handler_trigger_t *          trigger;
} globus_l_handler_t;

globus_result_t
globus_handler_chain_init(
    globus_handler_chain_t *            chain)
{
    globus_handler_chain_t              new_chain;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusFuncName(globus_handler_chain_init);
    GlobusHandlerDebugEnter();

    new_chain = globus_malloc(sizeof(globus_i_handler_chain_t));
    if(!new_chain)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        GlobusHandlerDebugExit();
        return result;
    }
    memset(new_chain, 0, sizeof(globus_i_handler_chain_t));
    
    *chain = new_chain;

    GlobusHandlerDebugExit();
    return result;
}

void
globus_l_handler_destroy(
    void *                              h)
{
    globus_l_handler_t *                entry;

    entry = (globus_l_handler_t *) h;

    if(entry)
    {
        free(entry);
    }
}

void
globus_handler_chain_destroy(
    globus_handler_chain_t              chain)
{
    globus_extension_handle_t           ext;
    globus_list_t *                     extensions;
    GlobusFuncName(globus_handler_chain_init);
    GlobusHandlerDebugEnter();

    globus_list_destroy_all(chain->request_handlers, 
                            globus_l_handler_destroy);
    globus_list_destroy_all(chain->response_handlers, 
                            globus_l_handler_destroy);
    globus_list_destroy_all(chain->request_init_handlers,
                            globus_l_handler_destroy);
    globus_list_destroy_all(chain->request_destroy_handlers,
                            globus_l_handler_destroy);
    globus_list_destroy_all(chain->response_init_handlers,
                            globus_l_handler_destroy);
    globus_list_destroy_all(chain->response_destroy_handlers,
                            globus_l_handler_destroy);
    if(chain->triggers)
    {
        globus_hashtable_destroy_all(
	    &chain->triggers,
	    globus_l_handler_destroy);
    }

    extensions = chain->extensions;
    while(extensions)
    {
	ext = globus_list_first(extensions);
	globus_extension_release(ext);
	extensions = globus_list_rest(extensions);
    }

    globus_list_free(chain->extensions);

    globus_free(chain);

    GlobusHandlerDebugExit();
}

static
void
globus_l_handler_hash_entry_copy(
    void **                             dest_key,
    void **                             dest_datum,
    void *                              src_key,
    void *                              src_datum)
{
    globus_l_handler_t *                dest_handler;
    globus_l_handler_t *                src_handler;

    src_handler = (globus_l_handler_t *)src_datum;
    dest_handler = globus_malloc(sizeof(globus_l_handler_t));
    if(!dest_handler)
    {
        *dest_key = NULL;
        *dest_datum = NULL;
    }
    
    memset(dest_handler, 0, sizeof(globus_l_handler_t));
    
    dest_handler->invoke = src_handler->invoke;
    dest_handler->trigger = src_handler->trigger;

    *dest_datum = (void *)dest_handler;
    *dest_key = (void *)dest_handler->trigger->element;
}

static
globus_result_t
globus_l_handle_chain_copy(
    globus_list_t **                    dest,
    globus_list_t *                     src)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_handler_t *                src_handler;
    globus_l_handler_t *                new_handler;
    GlobusFuncName(globus_l_handler_chain_copy);
    GlobusHandlerDebugEnter();
    
    for(; !globus_list_empty(src); src = globus_list_rest(src))
    {
        src_handler = globus_list_first(src);
        
        new_handler = globus_malloc(sizeof(globus_l_handler_t));
        if(!new_handler)
        {
            result = GlobusSoapMessageErrorOutOfMemory;
            GlobusHandlerDebugExit();
            return result;
        }

        memset(new_handler, 0, sizeof(globus_l_handler_t));
        
        new_handler->invoke = src_handler->invoke;
        globus_list_insert(dest, new_handler);
        dest = globus_list_rest_ref(*dest);
    }

    GlobusHandlerDebugExit();
    return result;
}

globus_result_t
globus_handler_chain_copy(
    globus_handler_chain_t *            dest,
    globus_handler_chain_t              src)
{
    globus_extension_handle_t           ext;
    globus_list_t *                     extensions;
    globus_handler_chain_t              new_chain;
    globus_result_t                     result = GLOBUS_SUCCESS;
    
    GlobusFuncName(globus_handler_chain_copy);
    GlobusHandlerDebugEnter();

    new_chain = globus_malloc(sizeof(globus_i_handler_chain_t));
    if(!new_chain)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }

    memset(new_chain, 0, sizeof(globus_i_handler_chain_t));
    
    result = globus_l_handle_chain_copy(
        &new_chain->request_handlers, src->request_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }
    
    result = globus_l_handle_chain_copy(
        &new_chain->response_handlers, src->response_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto destroy_request_handlers;
    }

    result = globus_l_handle_chain_copy(
        &new_chain->request_init_handlers, src->request_init_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto destroy_response_handlers;
    }

    result = globus_l_handle_chain_copy(
        &new_chain->request_destroy_handlers, src->request_destroy_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto destroy_request_init_handlers;
    }

    result = globus_l_handle_chain_copy(
        &new_chain->response_init_handlers, src->response_init_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto destroy_request_destroy_handlers;
    }

    result = globus_l_handle_chain_copy(
        &new_chain->response_destroy_handlers, src->response_destroy_handlers);
    if(result != GLOBUS_SUCCESS)
    {
        goto destroy_response_init_handlers;
    }

    if(src->triggers)
    {
        globus_hashtable_copy(
            &new_chain->triggers,
            &src->triggers,
            globus_l_handler_hash_entry_copy);
    }

    new_chain->default_trigger = src->default_trigger;

    extensions = src->extensions;
    while(extensions)
    {
	ext = globus_list_first(extensions);
	globus_extension_reference(ext);
	globus_list_insert(&new_chain->extensions, (void *)ext);
	extensions = globus_list_rest(extensions);
    }
    
    *dest = new_chain;

    GlobusHandlerDebugExit();
    return result;
 
 destroy_response_init_handlers:
    globus_list_destroy_all(
	new_chain->response_init_handlers, globus_l_handler_destroy);

 destroy_request_destroy_handlers:
    globus_list_destroy_all(
	new_chain->request_destroy_handlers, globus_l_handler_destroy);
 
 destroy_request_init_handlers:
    globus_list_destroy_all(
        new_chain->request_init_handlers, globus_l_handler_destroy);

 destroy_response_handlers:
    globus_list_destroy_all(
        new_chain->response_handlers, globus_l_handler_destroy);
    
 destroy_request_handlers:
    globus_list_destroy_all(
        new_chain->request_handlers, globus_l_handler_destroy);
    
 error:
    globus_free(new_chain);

 exit:
    GlobusHandlerDebugExit();
    return result;
}

static
int
globus_l_handler_chain_list_insert_last(
     globus_list_t * volatile *         headp,
     void *                             datum)
{
    globus_list_t *                     entry = NULL;
    globus_list_t *                     tmp_list;
    
    if(!*headp)
    {   
        globus_list_insert(headp, datum);   
    }   
    else   
    {   
        tmp_list = *headp;   
    
        while(tmp_list->next)
        {
            tmp_list = globus_list_rest(tmp_list);
        }
        globus_list_insert(&entry, datum);
        tmp_list->next = entry;   
    }
    
    return 0;
}

static
globus_result_t
globus_l_handler_chain_push(
    globus_list_t **                    chain,
    globus_extension_handle_t           extension_handle,
    globus_handler_invoke_func_t        invoke,
    globus_bool_t                       last)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_handler_t *                handler;
    GlobusFuncName(globus_handler_chain_push);
    GlobusHandlerDebugEnter();
    
    handler = globus_malloc(sizeof(globus_l_handler_t));
    if(!handler)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }

    memset(handler, 0, sizeof(globus_l_handler_t));
    handler->invoke = invoke;
    
    if(last)
    {
        globus_l_handler_chain_list_insert_last(chain, handler);
    }
    else
    {
        globus_list_insert(chain, handler);
    }

 exit:
    GlobusHandlerDebugExit();
    return result;
}

globus_result_t
globus_l_handler_triggers_insert(
    globus_hashtable_t *                triggers,
    globus_extension_handle_t           ext,
    globus_list_t *                     trigger_list)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_list_t *                     trigger_iter;
    globus_handler_trigger_t *          trigger;
    globus_l_handler_t *                handler;
    GlobusFuncName(globus_handler_chain_push);
    GlobusHandlerDebugEnter();
    
    if(!*triggers)
    {
        globus_hashtable_init(
            triggers, 20, xsd_QName_hash, xsd_QName_keyeq);
    }
    
    trigger_iter = trigger_list;
    while(trigger_iter)
    {
        trigger = globus_list_first(trigger_iter);

        handler = globus_malloc(sizeof(globus_l_handler_t));
        if(!handler)
        {
            result = GlobusSoapMessageErrorOutOfMemory;
            goto exit;
        }

        memset(handler, 0, sizeof(globus_l_handler_t));
        handler->trigger = trigger;

        globus_hashtable_insert(
            triggers, 
            handler->trigger->element,
            handler);
        trigger_iter = globus_list_rest(trigger_iter);
    }

 exit:
    GlobusHandlerDebugExit();
    return result;
}

globus_result_t
globus_handler_chain_push(
    globus_handler_chain_t              chain,
    globus_handler_type_t               type,
    const char *                        endpoint)
{
    globus_handler_descriptor_t *       handler_desc;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_extension_handle_t           extension_handle;
    GlobusFuncName(globus_handler_chain_push);
    GlobusHandlerDebugEnter();

    handler_desc = globus_extension_lookup(
        &extension_handle, GLOBUS_HANDLER_REGISTRY, (void *)endpoint);
    if(!handler_desc)
    {
        globus_result_t                 tmp_result;

        tmp_result = globus_extension_activate(endpoint);
        if(tmp_result != GLOBUS_SUCCESS)
        {
            result = GlobusHandlerErrorRegistryLookupFailed(
                tmp_result, endpoint);            
            goto exit;
        }

        globus_mutex_lock(&globus_l_handler_extensions_mutex);
        globus_list_insert(
            &globus_l_handler_extensions, strdup(endpoint));
        globus_mutex_unlock(&globus_l_handler_extensions_mutex);

        handler_desc = globus_extension_lookup(
            &extension_handle, GLOBUS_HANDLER_REGISTRY, (void *)endpoint);
        if(!handler_desc)
        {
            result = GlobusHandlerErrorRegistryLookupFailed(
                GLOBUS_FAILURE, endpoint);
            goto exit;
        }
    }

    globus_list_insert(&chain->extensions, (void *)extension_handle);

    if((type & GLOBUS_HANDLER_TYPE_REQUEST) && handler_desc->request)
    {
        globus_l_handler_chain_push(&chain->request_handlers,
            extension_handle, handler_desc->request, GLOBUS_FALSE);
    }
    
    if((type & GLOBUS_HANDLER_TYPE_RESPONSE) && handler_desc->response)
    {
        globus_l_handler_chain_push(&chain->response_handlers,
            extension_handle, handler_desc->response, GLOBUS_TRUE);
    }

    if((type & GLOBUS_HANDLER_TYPE_REQUEST_INIT) && handler_desc->request_init)
    {
        globus_l_handler_chain_push(
            &chain->request_init_handlers,
            extension_handle, handler_desc->request_init, GLOBUS_FALSE);
    }

    if((type & GLOBUS_HANDLER_TYPE_RESPONSE_INIT) && handler_desc->response_init)
    {
        globus_l_handler_chain_push(
            &chain->response_init_handlers,
            extension_handle, handler_desc->response_init, GLOBUS_FALSE);
    }

    if((type & GLOBUS_HANDLER_TYPE_REQUEST_DESTROY) && handler_desc->request_destroy)
    {
        globus_l_handler_chain_push(
            &chain->request_destroy_handlers,
            extension_handle, handler_desc->request_destroy, GLOBUS_FALSE);
    }

    if((type & GLOBUS_HANDLER_TYPE_RESPONSE_DESTROY) && 
       handler_desc->response_destroy)
    {
        globus_l_handler_chain_push(
            &chain->response_destroy_handlers,
            extension_handle, handler_desc->response_destroy, GLOBUS_FALSE);
    }


    if(handler_desc->default_trigger)
    {
        if(chain->default_trigger && 
           (chain->default_trigger != handler_desc->default_trigger))
        {
            result = GlobusHandlerErrorDefaultTriggerAlreadyDefined();
            goto exit;
        }

        chain->default_trigger = handler_desc->default_trigger;
    }

    if(handler_desc->trigger_list)
    {
        globus_l_handler_triggers_insert(
            &chain->triggers, extension_handle,
            handler_desc->trigger_list);
    }

exit:
    
    GlobusHandlerDebugExit();
    return result;
}

static
void
globus_l_handler_callback_call_next(
    void *                              args)
{
    globus_l_handler_t *                next_handler;
    globus_handler_session_descriptor_t session;
    globus_handler_invoke_func_t        trigger;
    globus_assert(args);
    session = (globus_handler_session_descriptor_t) args;

    if(session->trigger)
    {
        trigger = session->trigger;
        session->trigger = NULL;
        trigger(session, session->message, session->result);
        return;
    }

    if(globus_list_empty(session->remaining_handlers))
    {
        session->callback_func(session->result, session->callback_args);
        globus_free(session);
        return;
    }

    next_handler = globus_list_first(session->remaining_handlers);
    session->remaining_handlers = globus_list_rest(
        session->remaining_handlers);
    next_handler->invoke(session, session->message, session->result);
}

void
globus_handler_finished(
    globus_handler_session_descriptor_t session,
    globus_result_t                     result)
{
    globus_result_t                     local_result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_handler_finished);
    GlobusHandlerDebugEnter();

    session->result = result;

    local_result = globus_callback_register_oneshot(
        NULL, NULL,
        globus_l_handler_callback_call_next,
        session);
    if(local_result != GLOBUS_SUCCESS)
    {
        globus_panic(GLOBUS_HANDLER_MODULE,
                     local_result,
                     "[%s:%d] Couldn't register callback",
                     _globus_func_name,
                     __LINE__);
    }

    GlobusHandlerDebugExit();
}

globus_bool_t
globus_handler_chain_has_trigger(
    globus_handler_chain_t              chain,
    const xsd_QName *                   element)
{
    if(chain->triggers && globus_hashtable_lookup(
	    &chain->triggers, 
	    (void *)element))
    {
	return GLOBUS_TRUE;
    }
    return GLOBUS_FALSE;
}

globus_result_t
globus_handler_chain_register_trigger(
    globus_handler_chain_t              chain,
    globus_soap_message_handle_t        message,
    const xsd_QName *                   element,
    globus_handler_callback_func_t      callback,
    void *                              user_args)
{
    globus_handler_session_descriptor_t desc;
    globus_l_handler_t *                handler;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_handler_chain_register_trigger);
    GlobusHandlerDebugEnter();
    
    if(!chain->triggers)
    {
        result = GlobusHandlerErrorNoTriggersDefined();
        goto exit;
    }

    desc = (globus_handler_session_descriptor_t) globus_malloc(
        sizeof(globus_i_handler_session_descriptor_t));
    if(!desc)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(desc, 0, sizeof(globus_i_handler_session_descriptor_t));
    
    desc->message = message;
    desc->callback_func = callback;
    desc->callback_args = user_args;
    
    handler = globus_hashtable_lookup(
        &chain->triggers, 
        (void *)element);
    if(handler)
    {
        desc->trigger = handler->trigger->invoke;
    }
    else if(chain->default_trigger)
    {
        desc->trigger = chain->default_trigger;
    }
    else
    {
        result = GLOBUS_HANDLER_STATUS_NOT_TRIGGERED;
        goto exit;
    }
    
    globus_handler_finished(desc, GLOBUS_SUCCESS);
        
 exit:

    GlobusHandlerDebugExit();
    return result;
}

void
globus_handler_chain_register_invoke(
    globus_handler_chain_t              chain,
    globus_handler_type_t               type,
    globus_soap_message_handle_t        message,
    globus_handler_callback_func_t      callback,
    void *                              user_args)
{
    globus_handler_session_descriptor_t desc;
    GlobusFuncName(globus_handler_chain_register_invoke);
    GlobusHandlerDebugEnter();
    
    desc = (globus_handler_session_descriptor_t) globus_malloc(
        sizeof(globus_i_handler_session_descriptor_t));
    if(!desc)
    {
        globus_handler_finished(NULL, GlobusSoapMessageErrorOutOfMemory);
    }

    memset(desc, 0, sizeof(globus_i_handler_session_descriptor_t));

    desc->message = message;
    desc->callback_func = callback;
    desc->callback_args = user_args;

    switch(type)
    {
    case GLOBUS_HANDLER_TYPE_REQUEST_INIT:
        desc->remaining_handlers = chain->request_init_handlers;
        break;
    case GLOBUS_HANDLER_TYPE_REQUEST_DESTROY:
        desc->remaining_handlers = chain->request_destroy_handlers;
        break;
    case GLOBUS_HANDLER_TYPE_REQUEST:
        desc->remaining_handlers = chain->request_handlers;
        break;
    case GLOBUS_HANDLER_TYPE_RESPONSE_INIT:
        desc->remaining_handlers = chain->response_init_handlers;
        break;
    case GLOBUS_HANDLER_TYPE_RESPONSE_DESTROY:
        desc->remaining_handlers = chain->response_destroy_handlers;
        break;
    case GLOBUS_HANDLER_TYPE_RESPONSE:
        desc->remaining_handlers = chain->response_handlers;
        break;
    default:
        desc->remaining_handlers = NULL;
    }

    globus_handler_finished(desc, GLOBUS_SUCCESS);

    GlobusHandlerDebugExit();
}
