/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_service_engine.h"
#include <signal.h>

#define SHORT_USAGE_FORMAT \
"\nSyntax: %s [-help][-max <max sessions>] ...\n"

static char *  LONG_USAGE = 
"\n"
"    Options\n"
"    -help, -usage             Displays usage\n"
"    -version                  Displays version\n"
"    -max      <max sessions>  Max sessions that can be started in parallel\n"
"    -port <port>              Set the port of the container\n"
"    -pidfile <path>           Write PID of container to this file\n"
"    -bg                       Run container in the background\n"
"    -nosec                    Don't use https\n"
"\n" ;

#   define args_show_version() \
    { \
        char buf[64]; \
        sprintf( buf, \
                 "%s-%s", \
                 PACKAGE, \
                 VERSION); \
        fprintf(stderr, "%s\n", buf); \
        globus_module_deactivate_all(); \
        exit(0); \
    }

#   define args_show_short_help() \
    { \
        fprintf(stderr, \
                SHORT_USAGE_FORMAT \
                "\nOption -help will display usage.\n", \
                program); \
        globus_module_deactivate_all(); \
    }

#   define args_show_full_usage() \
    { \
        fprintf(stderr, SHORT_USAGE_FORMAT \
                "%s", \
                program, \
                LONG_USAGE); \
        globus_module_deactivate_all(); \
        exit(0); \
    }

#   define args_error_message(errmsg) \
    { \
        fprintf(stderr, "ERROR: %s\n", errmsg); \
        args_show_short_help(); \
        globus_module_deactivate_all(); \
        exit(1); \
    }

#   define args_error(argnum, argval, errmsg) \
    { \
        char buf[1024]; \
        sprintf(buf, "argument #%d (%s) : %s", argnum, argval, errmsg); \
        args_error_message(buf); \
    }

#   define args_verify_next(argnum, argval, errmsg) \
    { \
        if ((argnum+1 >= argc) || (argv[argnum+1][0] == '-')) \
            args_error(argnum,argval,errmsg); \
    }

/* sighandler_t defined in the GNU libc is not portable */
typedef void (* signal_handler_t) (int);

int begin_shutdown = 0;

void load_signal_handlers();
void cleanup_container(int);

static
void
globus_l_container_session_started_callback(
    globus_result_t                     result,
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        message,
    void *                              user_args);

static
void
globus_l_container_session_complete_callback(
    globus_result_t                     result,
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        message,
    void *                              user_args);

globus_mutex_t                          sessions_mutex;
int                                     sessions_started = 0;
int                                     max_sessions = 100;


int main(int argc, char * argv[])
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    char *                              program = NULL;
    int                                 arg_index;
    char *                              argp = NULL;
    char *                              contact = NULL;
    globus_service_engine_t             engine = NULL;
    globus_abstime_t                    timeout;
    char *                              port = NULL;
    char *                              pidfile_path = NULL;
    FILE *                              pidfile = NULL;
    globus_bool_t                       background = GLOBUS_FALSE;
    pid_t                               pid;
    int                                 pipes[2];
    int                                 rc;
    char                                portbuf[81];
    globus_bool_t                       do_secure = GLOBUS_TRUE;
    globus_soap_message_attr_t          attrs = NULL;

    load_signal_handlers();

    /* get the program name */
    if (strrchr(argv[0], '/'))
    {
        program = strrchr(argv[0], '/') + 1;
    }
    else
    {
        program = argv[0];
    }

    /* parse the arguments */
    for(arg_index = 1; arg_index < argc; ++arg_index)
    {
        argp = argv[arg_index];

        if((strcmp(argp, "-help") == 0) ||
           (strcmp(argp, "-usage") == 0))
        {
            args_show_full_usage();
        }
        else if(strcmp(argp, "-version") == 0)
        {
            args_show_version();
        }
        else if(strcmp(argp, "-max") == 0)
        {
            args_verify_next(arg_index, argp, "integer argument missing");
            max_sessions = atoi(argv[arg_index + 1]);
            arg_index++;
        }
        else if(strcmp(argp, "-port") == 0)
        {
            args_verify_next(arg_index, argp, "integer argument missing");
            port = argv[arg_index + 1];
            arg_index++;
        }
        else if (strcmp(argp, "-pidfile") == 0)
        {
            args_verify_next(arg_index, argp, "path argument missing");
            pidfile_path = argv[arg_index + 1];
            arg_index++;
        }
        else if (strcmp(argp, "-bg") == 0)
        {
            background = GLOBUS_TRUE;
        }
        else if (strcmp(argp, "-nosec") == 0)
        {
            do_secure = GLOBUS_FALSE;
        }
        else
        {
            args_error(arg_index, argp, "unrecognized option");
        }
    }

    if (background)
    {
        rc = pipe(pipes);
        if (rc < 0)
        {
            perror("Error opening pipe");

            exit(1);
        }
        pid = fork();

        if (pid < 0)
        {
            fprintf(stderr, "Error forking: %s", strerror(errno));

            exit(1);
        }
        else if (pid > 0)
        {
            close(pipes[1]);
            /* replace stdin fd with pipe from child */
            dup2(pipes[0], 0);

            rc = fscanf(stdin, "Contact: %80s", portbuf);

            if (rc < 0)
            {
                fprintf(stderr, "Error getting port\n");

                kill(pid, SIGTERM);

                exit(1);
            }

            printf("Contact: %s\n", portbuf);
            fflush(stdout);

            exit(0);
        }
        else
        {
            close(0);
            close(1);
            close(2);

            open("/dev/null", O_RDONLY);
            dup2(pipes[1], 1);
            open("/dev/null", O_WRONLY);
        }
    }

    if (pidfile_path != NULL)
    {
        pidfile = fopen(pidfile_path, "w");
        fprintf(pidfile, "%lu\n", (unsigned long) getpid());
        fclose(pidfile);
    }


    result = globus_module_activate(GLOBUS_SERVICE_ENGINE_MODULE);
    if(result != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Failed Engine Activation: %s:%d\n", 
                __FILE__, __LINE__);
        goto exit;
    }        

    result = globus_service_engine_init(&engine, attrs, port, NULL, do_secure);
    if(result != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Failed Engine Initialize: %s:%d\n", 
                __FILE__, __LINE__);
        goto exit;
    }

    result = globus_service_engine_get_contact(engine, &contact);
    if(result != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Failed to get container contact string: %s:%d\n", 
                __FILE__, __LINE__);
        goto exit;
    }

    fprintf(stdout, "Contact: %s\n\n", contact);
    fflush(stdout);

    globus_libc_free(contact);

    globus_mutex_init(&sessions_mutex, NULL);

    result = globus_service_engine_register_session(
        engine,
        globus_l_container_session_started_callback,
        NULL);
    if(result != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "ERROR: Initial Register of Session Failed.\n");
        goto exit;
    }

    while(!begin_shutdown)
    {

        GlobusTimeAbstimeGetCurrent(timeout);
        GlobusTimeAbstimeSet(timeout, 1, 0);
        
        globus_callback_poll(&timeout);
    }

 exit:

    if(engine)
    {
        globus_service_engine_destroy(engine);
        engine = NULL;
    }

    if(result != GLOBUS_SUCCESS)
    {
        const char *                    error_string;
        globus_object_t *               error_obj;

        error_obj = globus_error_get(result);
        if(!error_obj)
        {
            fprintf(stderr, 
                    "Failed to create error object: %s:%d\n", 
                    __FILE__, __LINE__);
            exit(1);
        }

        error_string = globus_error_print_chain(error_obj);
        if(!error_string)
        {
            fprintf(stderr, 
                    "Failed to create error string: %s:%d\n", 
                    __FILE__, __LINE__);
            exit(1);
        }

        fprintf(stderr, "%s\n\n", error_string);
        exit(1);
    }

    if (pidfile_path != NULL)
    {
        remove(pidfile_path);
    }
 
    result = globus_module_deactivate(GLOBUS_SERVICE_ENGINE_MODULE);
    /*globus_module_deactivate_all();*/

    return 0;
}

void
globus_l_container_session_started_callback(
    globus_result_t                     result,
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        session,
    void *                              args)
{
    globus_result_t                     fault_result = result;

    if(begin_shutdown)
    {
        fprintf(stderr, "Execution cancelled, cleaning up.\n");
        return;
    }

    globus_mutex_lock(&sessions_mutex);
    sessions_started++;
    globus_mutex_unlock(&sessions_mutex);

    if(sessions_started < max_sessions)
    {
        result = globus_service_engine_register_session(
            engine,
            globus_l_container_session_started_callback,
            NULL);
        if(result != GLOBUS_SUCCESS)
        {
            globus_object_t *               error_obj;
            char *                          error_str;
            
            error_obj = globus_error_get(result);
            error_str = globus_error_print_chain(error_obj);
            fprintf(stderr, 
                    "ERROR: Failed to Register Session\n%s\n\n",
                    error_str);
            globus_free(error_str);
            globus_object_free(error_obj);
        }
    }

    result = globus_service_engine_register_process(
        engine,
        session,
        fault_result,
        globus_l_container_session_complete_callback,
        args);

    if(result != GLOBUS_SUCCESS)
    {
        globus_object_t *               error_obj;
        char *                          error_str;

        error_obj = globus_error_get(result);
        error_str = globus_error_print_chain(error_obj);
        fprintf(stderr, "ERROR: Failed to Process Session\n");
        globus_free(error_str);
        globus_object_free(error_obj);

        /* Couldn't start processing this session, do the cleanup for it */
        globus_l_container_session_complete_callback(
            result, engine, session, args);
    }
}

void
globus_l_container_session_complete_callback(
    globus_result_t                     result,
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        session,
    void *                              user_args)
{
    if(result != GLOBUS_SUCCESS)
    {
        globus_soap_message_fault_t     soap_fault;
        char *                          error_string;
        globus_object_t *               errobj = globus_error_get(result);

        soap_fault = globus_soap_message_fault_get(errobj);
        if(soap_fault)
        {
            fprintf(stderr, 
                    "[SERVICE FAULT]:\nfaultcode = %s\n"
                    "faultstring = %s\n\n",
                    soap_fault->faultcode,
                    soap_fault->faultstring);
        }
        else
        {
            error_string = globus_error_print_chain(errobj);
            fprintf(stderr, "[SERVICE ERROR]: %s\n", error_string);
            free(error_string);
        }
    }

    globus_soap_message_handle_destroy(
        session);
    
    globus_mutex_lock(&sessions_mutex);
    sessions_started--;
    globus_mutex_unlock(&sessions_mutex);
}

void cleanup_container(int              signal)
{
    begin_shutdown = 1;
}    

void load_signal_handlers()
{
    signal_handler_t                    sigres;

    sigres = signal(1, cleanup_container); /* SIGHUP */
    if(sigres == SIG_ERR)
    {
        fprintf(stderr, "Failed to load signal handler for SIGHUP\n");
        exit(1);
    }

    sigres = signal(2, cleanup_container); /* SIGINT */
    if(sigres == SIG_ERR)
    {
        fprintf(stderr, "Failed to load signal handler for SIGINT\n");
        exit(1);
    }

    sigres = signal(15, cleanup_container); /* SIGTERM */
    if(sigres == SIG_ERR)
    {
        fprintf(stderr, "Failed to load signal handler for SIGTERM\n");
        exit(1);
    }
}
