/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_soap_message_attrs.h"
#include "globus_soap_message_utils.h"
#include "globus_i_soap_message.h"

typedef struct globus_l_message_attr_s
{
    char *                                   name;
    void *                                   value;
    globus_soap_message_attr_copy_func_t     copy;
    globus_soap_message_attr_destroy_func_t  destroy;
} globus_l_message_attr_t;

void
globus_l_message_attr_destroy(
    void *                              attr)
{
    globus_l_message_attr_t *           attribute;
    GlobusFuncName(globus_l_message_attr_destroy);
    GlobusSoapMessageDebugEnter();


    attribute = (globus_l_message_attr_t *) attr;

    if(attribute)
    {
        if(attribute->name)
        {
            free(attribute->name);
        }
        
        if(attribute->destroy)
        {
            attribute->destroy(attribute->value);
        }

        free(attribute);
    }

    GlobusSoapMessageDebugExit();
}

globus_result_t
globus_soap_message_attr_init(
    globus_soap_message_attr_t *        attrs)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_attr_init);
    GlobusSoapMessageDebugEnter();
    
    result = globus_hashtable_init(attrs, 20,
                                   globus_hashtable_string_hash,
                                   globus_hashtable_string_keyeq);
    
    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_soap_message_attr_copy(
    globus_soap_message_attr_t *        copied_attrs,
    globus_soap_message_attr_t          attrs)
{
    globus_soap_message_attr_t          new_attrs;
    globus_l_message_attr_t *           attr;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_attr_copy);
    GlobusSoapMessageDebugEnter();

    globus_soap_message_attr_init(&new_attrs);

    attr = globus_hashtable_first(&attrs);
    while(attr)
    {
        /* _set will copy the attr value */
        result = globus_soap_message_attr_set(
            new_attrs,
            attr->name,
            attr->copy,
            attr->destroy,
            attr->value);
        if(result != GLOBUS_SUCCESS)
        {
            goto exit;
        }

        attr = globus_hashtable_next(&attrs);
    }

    *copied_attrs = new_attrs;

 exit:
    GlobusSoapMessageDebugExit();
    return result;
}

void
globus_soap_message_attr_destroy(
    globus_soap_message_attr_t                  attrs)
{
    GlobusFuncName(globus_soap_message_attr_destroy);
    GlobusSoapMessageDebugEnter();

    globus_hashtable_destroy_all(
        &attrs, globus_l_message_attr_destroy);

    GlobusSoapMessageDebugExit();
}

globus_result_t
globus_soap_message_attr_set(
    globus_soap_message_attr_t               attrs,
    const char *                             name,
    globus_soap_message_attr_copy_func_t     copy,
    globus_soap_message_attr_destroy_func_t  destroy,
    void *                                   attrvalue)
{
    void *                              newvalue = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_message_attr_t *           attr;
    GlobusFuncName(globus_soap_message_attr_set);
    GlobusSoapMessageDebugEnter();

    globus_assert_string(attrs, "NULL attrs in handle");

    globus_assert_string(name, 
                         "attr_name passed to function is NULL");
    
    attr = globus_hashtable_lookup(&attrs, (char *)name);
    if(!attr)
    {
        attr = globus_malloc(sizeof(globus_l_message_attr_t));
        if(!attr)
        {
            result = GlobusSoapMessageErrorOutOfMemory;
            goto exit;
        }
        memset(attr, 0, sizeof(globus_l_message_attr_t));

        attr->name = globus_libc_strdup(name);
        attr->destroy = destroy;
        attr->copy = copy;

        result = globus_hashtable_insert(&attrs, 
                                         attr->name, attr);
        if(result != GLOBUS_SUCCESS)
        {
            goto exit;
        }
    }
    else
    {
        if(attr->destroy)
        {
            attr->destroy(attr->value);
        }

        attr->destroy = destroy;
        attr->copy = copy;
    }

    if(copy)
    {
        copy(&newvalue, attrvalue);
    }
    else
    {
        newvalue = attrvalue;
    }

    attr->value = newvalue;

 exit:
    
    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_soap_message_handle_set_attr(
    globus_soap_message_handle_t             handle,
    const char *                             attr_name,
    globus_soap_message_attr_copy_func_t     copy,
    globus_soap_message_attr_destroy_func_t  destroy,
    void *                                   attrvalue)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_handle_set_attr);
    GlobusSoapMessageDebugEnter();

    globus_assert_string(handle, "NULL handle passed to function");
    globus_assert_string(handle->attrs, "NULL attrs in handle");
    globus_assert_string(attr_name, 
                         "attr_name passed to function is NULL");

    result = globus_soap_message_attr_set(
        handle->attrs,
        attr_name,
        copy,
        destroy,
        attrvalue);

    /* Hack: make the verbose attribute part of the handle struct for
     * performance as its queried during all serialization and deserialization
     * errors.
     */
    if (strcmp(attr_name, GLOBUS_SOAP_MESSAGE_VERBOSE_ERRORS_KEY) == 0)
    {
        handle->verbose = (globus_bool_t) attrvalue;
    }
    
    GlobusSoapMessageDebugExit();
    return result;
}

void *
globus_soap_message_attr_remove(
    globus_soap_message_attr_t          attrs,
    const char *                        attr_name)
{
    void *                              value;
    globus_l_message_attr_t *           attr;
    GlobusFuncName(globus_soap_message_attr_remove);
    GlobusSoapMessageDebugEnter();
    
    attr = globus_hashtable_remove(&attrs, (void *)attr_name);
    if(!attr)
    {
        value = NULL;
        goto exit;
    }

    value = attr->value;

    globus_l_message_attr_destroy(attr);

 exit:

    GlobusSoapMessageDebugExit();
    return value;
}

void *
globus_soap_message_handle_remove_attr(
    globus_soap_message_handle_t        handle,
    const char *                        attr_name)
{
    globus_assert_string(handle, "NULL handle passed to function");
    globus_assert_string(handle->attrs, "NULL attrs in handle");
    globus_assert_string(attr_name, 
                         "attr_name passed to function is NULL");

    return globus_soap_message_attr_remove(
        handle->attrs, attr_name);
}

void *
globus_soap_message_attr_get(
    globus_soap_message_attr_t          attrs,
    const char *                        attr_name)
{
    void *                              value;
    globus_l_message_attr_t *           attr;
    GlobusFuncName(globus_soap_message_attr_get);
    GlobusSoapMessageDebugEnter();

    attr = globus_hashtable_lookup(&attrs, (char *)attr_name);
    if(!attr)
    {
        value = NULL;
        goto exit;
    }

    value = attr->value;

 exit:
    
    GlobusSoapMessageDebugExit();
    return value;
}
    
void *
globus_soap_message_handle_get_attr(
    globus_soap_message_handle_t        handle,
    char *                              attr_name)
{
    void *                              value;
    GlobusFuncName(globus_soap_message_handle_get_attr);
    GlobusSoapMessageDebugEnter();

    value = globus_soap_message_attr_get(
        handle->attrs,
        attr_name);

    GlobusSoapMessageDebugExit();
    return value;
}

globus_result_t
globus_soap_message_handle_set_attrs(
    globus_soap_message_handle_t        handle,
    globus_soap_message_attr_t          attrs)
{
    GlobusFuncName(globus_soap_message_handle_set_attr);
    GlobusSoapMessageDebugEnter();

    if(handle->attrs)
    {
        globus_soap_message_attr_destroy(handle->attrs);
    }

    globus_soap_message_attr_copy(
        &handle->attrs, attrs);

    /* Hack: make the verbose attribute part of the handle struct for
     * performance as its queried during all serialization and deserialization
     * errors.
     */
    handle->verbose = (globus_bool_t) globus_soap_message_attr_get(
                attrs,
                GLOBUS_SOAP_MESSAGE_VERBOSE_ERRORS_KEY);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_soap_message_attr_copy_string(
    void **                             new_str,
    void *                              str)
{
    *new_str = (void *) globus_libc_strdup((char *)str);
    return GLOBUS_SUCCESS;
}

