/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_common.h"
#include "globus_xio_buffer.h"
#include "globus_xio_tcp_driver.h"
#include "globus_xio_http.h"
#include "globus_soap_message_handle.h"
#include "globus_i_soap_message.h"
#include "globus_soap_message.h"
#include "globus_soap_message_attrs.h"
#include "globus_xio_gsi.h"

static
void
globus_l_soap_message_response_ready_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes,
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg);

typedef struct
{
    struct globus_soap_message_handle_s * message;
    globus_soap_message_callback_func_t   callback;
    void *                                args;
    enum 
    {
        GLOBUS_SOAP_MESSAGE_REQUEST_CALLBACK,
        GLOBUS_SOAP_MESSAGE_RESPONSE_CALLBACK
    } type;
} globus_l_soap_message_callback_handle_t;


void
globus_l_soap_message_error_callback(
    void *                              arg,
    const char *                        msg,
    xmlParserSeverities                 severity,
    xmlTextReaderLocatorPtr             locator)
{
    globus_soap_message_handle_t        handle;
    int                                 lno;
    xmlChar *                           baseuri;
    GlobusFuncName(globus_i_soap_message_error_callback);
    GlobusSoapMessageDebugEnter();

    handle = (globus_soap_message_handle_t) arg;
    if(handle->result != GLOBUS_SUCCESS)
    {
        return;
    }

    if(handle->result != GLOBUS_SUCCESS)
    {
        return;
    }

    if(locator)
    {
        lno = xmlTextReaderLocatorLineNumber(locator);
        baseuri = xmlTextReaderLocatorBaseURI(locator);
    }
    
    switch(severity)
    {
    case XML_PARSER_SEVERITY_ERROR:
    case XML_PARSER_SEVERITY_VALIDITY_ERROR:
    
        handle->result = GlobusSoapMessageErrorParserFailed(msg, lno, baseuri);
        break;

    case XML_PARSER_SEVERITY_WARNING:
    case XML_PARSER_SEVERITY_VALIDITY_WARNING:

        GlobusSoapMessageDebugPrintf(GLOBUS_SOAP_MESSAGE_DEBUG_WARN,
                                     ("%s:%d: %s", baseuri, lno, msg));

    default:
        break;
    }

    GlobusSoapMessageDebugExit();
}

void
globus_l_soap_message_structured_error_callback(
    void *                              userdata,
    xmlErrorPtr                         error)
{
    globus_soap_message_handle_t        handle;
    GlobusFuncName(globus_i_soap_message_structured_error_callback);
    GlobusSoapMessageDebugEnter();

    handle = (globus_soap_message_handle_t) userdata;
   
    if(handle->result != GLOBUS_SUCCESS)
    {
        return;
    }
 
    handle->result = GlobusSoapMessageErrorParserFailed(
        error->message, error->line, error->file);

    GlobusSoapMessageDebugExit();
}

static
int
globus_l_soap_message_handle_libxml_io_read_callback(
    void *                              context,
    char *                              buffer,
    int                                 length)
{
    globus_size_t                       read_length = 0;
    globus_soap_message_handle_t        handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(
        globus_l_soap_message_handle_libxml_io_read_callback);
    GlobusSoapMessageDebugEnter();

    handle = (globus_soap_message_handle_t) context;
    globus_assert_string(handle, "NULL handle passed to function");

    if(handle->buffer_eof)
    {
        goto exit;
    }

    result = globus_xio_read(
        handle->xio_handle, (globus_byte_t *)buffer,
        length, 1, &read_length, NULL);
    if(result != GLOBUS_SUCCESS && 
       globus_xio_driver_error_match(
           globus_i_soap_message_buffer_driver, 
           globus_error_peek(result),
           GLOBUS_XIO_BUFFER_ERROR_EOF))
    {
        handle->result = GLOBUS_SUCCESS;
        handle->buffer_eof = 1;
    }
    else if(result != GLOBUS_SUCCESS)
    {
        handle->result = result;
    }

  exit:
    
    GlobusSoapMessageDebugExit();
    return read_length;
}

static
int
globus_l_soap_message_handle_libxml_io_write_callback(
    void *                              context,
    const char *                        buffer,
    int                                 length)
{
    globus_size_t                       bytes_written = 0;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_soap_message_handle_t        handle;
    GlobusFuncName(
        globus_l_soap_message_handle_libxml_io_write_callback);
    GlobusSoapMessageDebugEnter();

    handle = (globus_soap_message_handle_t) context;
    globus_assert_string(handle, "NULL handle passed to function");

    result = globus_xio_write(
        handle->xio_handle, (globus_byte_t *) buffer, 
        length, length, &bytes_written, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        handle->result = result;
    }

    GlobusSoapMessageDebugExit();
    return bytes_written;
}

static
int
globus_l_soap_message_handle_libxml_io_close_callback(
    void *                              context)
{
    globus_soap_message_handle_t             handle;
    GlobusFuncName(
        globus_l_soap_message_handle_libxml_io_close_callback);
    GlobusSoapMessageDebugEnter();

    handle = (globus_soap_message_handle_t) context;
    globus_assert_string(handle, "NULL handle passed to function");

    /* do nothing.  close should be done async. */
    
    GlobusSoapMessageDebugExit();
    return 0;
}

static
void
globus_l_soap_message_open_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              user_args)
{
    globus_l_soap_message_callback_handle_t * callback_handle;

    callback_handle = (globus_l_soap_message_callback_handle_t *) user_args;

    if(result == GLOBUS_SUCCESS)
    {
        result = globus_soap_message_setup_writer(
            callback_handle->message);
        if(result != GLOBUS_SUCCESS)
        {
            goto exit;
        }
    }
    else
    {
        callback_handle->message->result = result;
    }
    

 exit:
    
    callback_handle->callback(result, callback_handle->args);
    free(callback_handle);
}

globus_result_t
globus_i_soap_message_setup_unbuffered_writer(
    globus_soap_message_handle_t        message)
{
    xmlOutputBufferPtr                  output_buffer = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_i_soap_message_setup_unbuffered_writer);
    GlobusSoapMessageDebugEnter();

    output_buffer = xmlOutputBufferCreateIO(
        globus_l_soap_message_handle_libxml_io_write_callback,
        globus_l_soap_message_handle_libxml_io_close_callback,
        message,
        0);
    if(!output_buffer)
    {
        result = GlobusSoapMessageErrorInitFailed(
            "Failed to initialize output buffer");
        goto exit;
    }
    
    message->writer = xmlNewTextWriter(output_buffer);
    if(!message->writer)
    {
        result = GlobusSoapMessageErrorInitFailed(
            "Failed to initialize xmlWriter");
        goto free_output;
    }

    goto exit;

 free_output:

    xmlOutputBufferClose(output_buffer);

 exit:

    GlobusSoapMessageDebugExit();
    return result;
}
    

globus_result_t
globus_soap_message_setup_writer(
    globus_soap_message_handle_t        message)
{
    xmlOutputBufferPtr                  output_buffer = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_setup_writer);
    GlobusSoapMessageDebugEnter();

    globus_assert(message);

    result = globus_xio_handle_cntl(
        message->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_CLEAR);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to enable buffering for message write");
        goto exit;
    }

    result = globus_xio_handle_cntl(
        message->xio_handle, 
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_ENABLE_WRITE_BUFFERING);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to enable buffering for message write");
        goto exit;
    }
    
    output_buffer = xmlOutputBufferCreateIO(
        globus_l_soap_message_handle_libxml_io_write_callback,
        globus_l_soap_message_handle_libxml_io_close_callback,
        message,
        0);
    if(!output_buffer)
    {
        result = GlobusSoapMessageErrorInitFailed(
            "Failed to initialize output buffer");
        goto exit;
    }
    
    message->writer = xmlNewTextWriter(output_buffer);
    if(!message->writer)
    {
        result = GlobusSoapMessageErrorInitFailed(
            "Failed to initialize xmlWriter");
        goto free_output;
    }

    goto exit;

 free_output:

    xmlOutputBufferClose(output_buffer);

 exit:

    GlobusSoapMessageDebugExit();
    return result;
}

static
void
globus_l_soap_message_flush_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    globus_xio_iovec_t *                iovec,
    int                                 count,
    globus_size_t                       nbytes,
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_args)
{
    globus_result_t                     internal_result = GLOBUS_SUCCESS;
    static globus_byte_t                buffer[1];
    globus_l_soap_message_callback_handle_t * callback_handle;

    callback_handle = (globus_l_soap_message_callback_handle_t *) user_args;
    callback_handle->message->result = result;

    if(count > 0 && (iovec->iov_len > 0))
    {
        internal_result = globus_xio_handle_cntl(
            callback_handle->message->xio_handle,
            globus_i_soap_message_http_driver,
            GLOBUS_XIO_HTTP_HANDLE_SET_END_OF_ENTITY);
        if(internal_result != GLOBUS_SUCCESS)
        {
            internal_result = GlobusSoapMessageErrorFailedTransport(
                internal_result, 
                "Failed to flush message");
            goto error;
        }
    }

    internal_result = globus_xio_handle_cntl(
        callback_handle->message->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_CLEAR);
    if(internal_result != GLOBUS_SUCCESS)
    {
        internal_result = GlobusSoapMessageErrorFailedTransport(
            internal_result, 
            "Failed to flush message");
        goto error;
    }

    callback_handle->callback(result, callback_handle->args);

    if (callback_handle->type == GLOBUS_SOAP_MESSAGE_REQUEST_CALLBACK)
    {
        result = globus_xio_register_read(handle,
            buffer,
            0,
            0,
            NULL,
            globus_l_soap_message_response_ready_callback,
            callback_handle->message);

        if(result != GLOBUS_SUCCESS)
        {
            globus_l_soap_message_response_ready_callback(
                    handle,
                    result,
                    buffer,
                    0,
                    0,
                    NULL,
                    callback_handle->message);
        }
    }
    else
    {
        globus_assert(
            callback_handle->type == GLOBUS_SOAP_MESSAGE_RESPONSE_CALLBACK);
    }

    goto exit;
    
error:
    callback_handle->callback(internal_result, callback_handle->args);
    
exit:
    free(callback_handle);
}

static
void
globus_l_soap_message_read_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes,
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_l_soap_message_callback_handle_t * callback_handle;
    globus_result_t                     local_result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_l_soap_message_read_callback);
    GlobusSoapMessageDebugEnter();

    callback_handle = (globus_l_soap_message_callback_handle_t *) user_arg; 
    callback_handle->message->result = result;
   
    if(result != GLOBUS_SUCCESS)
    {
        if(globus_xio_error_is_eof(result) ||
           globus_xio_driver_error_match(
               globus_i_soap_message_http_driver,
               globus_error_peek(result), GLOBUS_XIO_HTTP_ERROR_EOF) ||
           globus_xio_driver_error_match(
               globus_i_soap_message_buffer_driver,
               globus_error_peek(result), GLOBUS_XIO_BUFFER_ERROR_EOF))
        {
            result = GLOBUS_SUCCESS;

            local_result = globus_xio_handle_cntl(
                callback_handle->message->xio_handle,
                globus_i_soap_message_buffer_driver,
                GLOBUS_XIO_BUFFER_DONE_BUFFERING);
            if(local_result != GLOBUS_SUCCESS)
            {
                local_result = GlobusSoapMessageErrorFailedTransport(
                    local_result, 
                    "Failed to disable buffering for buffer driver");
                goto exit;
            }

            if(GlobusSoapMessageDebug(GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES))
            {
                char *                          buffer;
                int                             read_length;
                
                globus_xio_handle_cntl(
                    callback_handle->message->xio_handle,
                    globus_i_soap_message_buffer_driver,
                    GLOBUS_XIO_BUFFER_GET_MARKED_BUFFERS,
                    NULL, NULL, &buffer, &read_length);

                GlobusSoapMessageDebugPrintf(
                    GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES,
                    ("----------------READING--------------\n%.*s\n"
                     "=====================================\n\n",
                     read_length, buffer));

                free(buffer);
            }

            local_result = globus_xio_handle_cntl(
                callback_handle->message->xio_handle,
                globus_i_soap_message_buffer_driver,
                GLOBUS_XIO_BUFFER_RESET_READ_POSITION);
            if(local_result != GLOBUS_SUCCESS)
            {
                local_result = GlobusSoapMessageErrorFailedTransport(
                    local_result,
                    "failed to reset buffer read position to start");
                goto exit;
            }

            if (callback_handle->message->reader != NULL)
            {
                xmlFreeTextReader(callback_handle->message->reader);
            }
            callback_handle->message->reader = xmlReaderForIO(
                globus_l_soap_message_handle_libxml_io_read_callback,
                globus_l_soap_message_handle_libxml_io_close_callback,
                callback_handle->message,
                NULL,
                NULL,
                0);
            if(!callback_handle->message->reader)
            {
                result = GlobusSoapMessageErrorInitFailed(
                    "Failed to initialize xmlReader");
                goto exit;
            }
            
            xmlTextReaderSetErrorHandler(callback_handle->message->reader, 
                                         globus_l_soap_message_error_callback, 
                                         callback_handle->message);
            
            xmlTextReaderSetStructuredErrorHandler(
                callback_handle->message->reader,
                globus_l_soap_message_structured_error_callback,
                callback_handle->message);
            
            callback_handle->callback(result, callback_handle->args);
            free(callback_handle);
        }
        else
        {
            local_result = result;
        }
        
        goto exit;
    }
    
    local_result = globus_xio_register_read(
        callback_handle->message->xio_handle,
        (globus_byte_t *)callback_handle->message->readbuf,
        GLOBUS_I_SOAP_MESSAGE_READBUF_SIZE,
        1,
        NULL,
        globus_l_soap_message_read_callback,
        callback_handle);
    if(local_result != GLOBUS_SUCCESS)
    {
        local_result = GlobusSoapMessageErrorFailedTransport(
            local_result, "Failed to begin reading SOAP message");
        goto exit;
    }
    
 exit:

    if(local_result != GLOBUS_SUCCESS)
    {
        callback_handle->message->result = local_result;
        callback_handle->callback(result, callback_handle->args);
    }

    GlobusSoapMessageDebugExit();
    return;
}

static
void
globus_l_soap_message_response_callback(
    void *                              args)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_soap_message_callback_handle_t * callback_handle;
    GlobusFuncName(globus_l_soap_message_response_callback);
    GlobusSoapMessageDebugEnter();

    callback_handle = (globus_l_soap_message_callback_handle_t *) args;

    if(callback_handle->message->response_semaphore >= 2)
    {
        globus_mutex_lock(&callback_handle->message->mutex);
        callback_handle->message->response_semaphore = 0;
        globus_mutex_unlock(&callback_handle->message->mutex);

        if(callback_handle->message->result != GLOBUS_SUCCESS)
        {
            if (globus_error_match
                        (globus_error_peek(
                                callback_handle->message->result),
                        GLOBUS_SOAP_MESSAGE_MODULE,
                        GLOBUS_SOAP_MESSAGE_ERROR_TYPE_FAILED_TRANSPORT))
            {
                result = callback_handle->message->result;
                goto exit;
            }
        }
        result = globus_xio_handle_cntl(
            callback_handle->message->xio_handle,
            globus_i_soap_message_buffer_driver,
            GLOBUS_XIO_BUFFER_ENABLE_READ_BUFFERING);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to enable read buffering in buffer driver");
            goto exit;
        }

        result = globus_xio_register_read(
            callback_handle->message->xio_handle,
            (globus_byte_t *)callback_handle->message->readbuf,
            GLOBUS_I_SOAP_MESSAGE_READBUF_SIZE,
            1,
            NULL,
            globus_l_soap_message_read_callback,
            callback_handle);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to begin reading SOAP message");
            goto exit;
        }
    }

 exit:

    if(result != GLOBUS_SUCCESS)
    {
            
        callback_handle->callback(result, callback_handle->args);
        free(callback_handle);
    }

    GlobusSoapMessageDebugExit();
    return;
}

static
void
globus_l_soap_message_response_ready_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes,
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_soap_message_handle_t        message_handle;
    int                                 response_code;
    const char *                        message;
    globus_xio_http_version_t           version;
    globus_hashtable_t                  headers;
    GlobusFuncName(globus_l_soap_message_response_ready_callback);
    GlobusSoapMessageDebugEnter();

    message_handle = (globus_soap_message_handle_t) user_arg;

    if (result == GLOBUS_SUCCESS)
    {
        result = globus_xio_data_descriptor_cntl(
            data_desc,
            globus_i_soap_message_http_driver,
            GLOBUS_XIO_HTTP_GET_RESPONSE,
            &response_code,
            &message,
            &version,
            &headers);
    }

    if(result != GLOBUS_SUCCESS)
    {
        message_handle->result = GlobusSoapMessageErrorFailedTransport(
            result, "Error in HTTP response");
        response_code = -1;
        GlobusSoapMessageDebugError(message_handle->result);
    }
    GlobusSoapMessageDebugPrintf(GLOBUS_SOAP_MESSAGE_DEBUG_DEBUG,
                                 ("response code: %d\n", response_code));
    

    if(response_code > 299)
    {
        if(response_code > 399 && response_code < 500)
        {
            message_handle->result = GlobusSoapMessageErrorBadRequest(
                response_code, message);
        }

        if(response_code > 499)
        {
            message_handle->result = GlobusSoapMessageErrorFailedResponse(
                response_code, message);
        }
    }

    globus_mutex_lock(&message_handle->mutex);
    message_handle->response_semaphore++;

    if(message_handle->response_semaphore > 1 &&
       message_handle->response_handle)
    {
        globus_l_soap_message_callback_handle_t * response_handle =
        (globus_l_soap_message_callback_handle_t *)
        message_handle->response_handle;

        globus_mutex_unlock(&message_handle->mutex);

        message_handle->response_handle = NULL;

        globus_l_soap_message_response_callback(
            response_handle);
    }
    else
    {
        globus_mutex_unlock(&message_handle->mutex);
    }

    GlobusSoapMessageDebugExit();
    return;
}    

static
globus_bool_t
globus_l_soap_message_timeout_callback(
    globus_xio_handle_t                 handle,
    globus_xio_operation_type_t         type,
    void *                              user_arg)
{
    return GLOBUS_TRUE;
}


globus_result_t
globus_soap_message_register_open(
    globus_soap_message_handle_t        handle,
    char *                              endpoint,
    globus_soap_message_callback_func_t callback,
    void *                              args)
{
    globus_soap_message_auth_method_t   sectype;
    globus_xio_attr_t                   attr = NULL;
    globus_l_soap_message_callback_handle_t * callback_handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    char *                              soap_action;
    char *                              http_version = NULL;
    char *                              connection = NULL;
    void *                              timeout = NULL;
    char *                              quoted_soap_action = NULL;
    GlobusFuncName(globus_soap_message_register_open);
    GlobusSoapMessageDebugEnter();

    globus_mutex_lock(&handle->cancel_mutex);
    handle->cancelled = 0;
    globus_mutex_unlock(&handle->cancel_mutex);

    callback_handle = (globus_l_soap_message_callback_handle_t *) 
    globus_malloc(sizeof(globus_l_soap_message_callback_handle_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(callback_handle, 0, 
           (sizeof(globus_l_soap_message_callback_handle_t)));

    if(handle->xio_handle)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Session already opened");
        goto exit;
    }

    callback_handle->callback = callback;
    callback_handle->args = args;
    callback_handle->message = handle;

    if(strncmp(endpoint, "https://", 8) == 0)
    {
        result = globus_xio_handle_create(
            &handle->xio_handle,
            globus_i_soap_message_xio_https_stack);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to create xio handle with https stack");
            goto exit;
        }

        result = globus_xio_attr_init(&attr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to initialize xio attr");
            goto exit;
        }

        result = globus_i_soap_message_setup_transport_security(
            handle->attrs,
            attr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to initialize xio attributes for https");
            goto exit;
        }
    }
    else
    {
        sectype = (globus_soap_message_auth_method_t)
            globus_soap_message_handle_get_attr(
                handle,
                GLOBUS_SOAP_MESSAGE_AUTHENTICATION_METHOD_KEY);
        
        if(sectype && (sectype == GLOBUS_SOAP_MESSAGE_AUTH_SECURE))
        {
            /* XXX may not be a good idea if users can 'set attrs' on this
             * handle while it is open
             */
            globus_soap_message_handle_set_attr(
                handle,
                GLOBUS_SOAP_MESSAGE_AUTHENTICATION_METHOD_KEY,
                NULL,
                NULL,
                (void *) GLOBUS_SOAP_MESSAGE_AUTH_SECURE_MESSAGE);
        }
        
        result = globus_xio_handle_create(&handle->xio_handle, 
                                          globus_i_soap_message_xio_stack);

        result = globus_xio_attr_init(&attr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to initialize xio attr");
            goto exit;
        }
    }

    timeout = globus_soap_message_handle_get_attr(
        handle, GLOBUS_SOAP_MESSAGE_TIMEOUT_KEY);
    if(timeout)
    {
        result = globus_xio_attr_cntl(
            attr,
            GLOBUS_NULL,
            GLOBUS_XIO_ATTR_SET_TIMEOUT_ALL,
            globus_l_soap_message_timeout_callback,
            timeout,
            GLOBUS_NULL);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set timeout for xio handle");
            goto exit;
        }
    }

    http_version = globus_soap_message_handle_get_attr(
        handle, GLOBUS_SOAP_MESSAGE_HTTP_VERSION_KEY);
    if(http_version && !strcmp(http_version, "1.0"))
    {
        result = globus_xio_attr_cntl(
            attr, 
            globus_i_soap_message_http_driver, 
            GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_HTTP_VERSION, 
            GLOBUS_XIO_HTTP_VERSION_1_0); 
        if(result != GLOBUS_SUCCESS) 
        { 
            result = GlobusSoapMessageErrorFailedTransport( 
                result, "Failed to set HTTP version to 1.0"); 
            goto exit; 
        } 

        result = globus_xio_attr_cntl( 
            attr, 
            globus_i_soap_message_http_driver, 
            GLOBUS_XIO_HTTP_ATTR_DELAY_WRITE_HEADER); 
        if(result != GLOBUS_SUCCESS) 
        { 
            result = GlobusSoapMessageErrorFailedTransport( 
                result, "Failed to set delay write of HTTP Header"); 
            goto exit; 
        } 
    }

    result = globus_xio_attr_cntl(
        attr, globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_HEADER,
        "Content-Type",
        "text/xml; charset=utf-8");
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set HTTP request header: Content-Type");
        goto exit;
    }

    result = globus_xio_attr_cntl(
        attr, globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_HEADER,
        "Accept",
        "text/xml");
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set HTTP request header: Accept");
        goto exit;
    }

    connection = (char *) globus_soap_message_handle_get_attr(
                handle,
                GLOBUS_SOAP_MESSAGE_CONNECTION_KEY);
    if (connection && !strcmp(connection, "close"))
    {
        result = globus_xio_attr_cntl(
            attr, globus_i_soap_message_http_driver,
            GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_HEADER,
            "Connection",
            "close");
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set HTTP request header: Connection");
            goto exit;
        }
    }

    soap_action = globus_soap_message_handle_get_attr(
        handle, GLOBUS_SOAP_MESSAGE_SOAP_ACTION_KEY);
    if(!soap_action)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "SOAPAction not set in message attributes");
        goto exit;
    }

    if (soap_action[0] != '"')
    {
        /* WS-I BP requires that SOAPAction be quoted */
        size_t                          soap_action_len;

        soap_action_len = strlen(soap_action);

        quoted_soap_action = malloc(soap_action_len + 3);

        memcpy(quoted_soap_action + 1, soap_action, soap_action_len);
        quoted_soap_action[0] = '"';
        quoted_soap_action[soap_action_len+1] = '"';
        quoted_soap_action[soap_action_len+2] = '\0';

        soap_action = quoted_soap_action;
    }

    result = globus_xio_attr_cntl(
        attr,
        globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_HEADER,
        "SOAPAction",
        soap_action);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set request method in http driver");
        goto exit;
    }
    
    result = globus_xio_attr_cntl(
        attr, globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_ATTR_SET_REQUEST_METHOD,
        "POST");
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set request method in http driver");
        goto exit;
    }

    result = globus_xio_attr_cntl(attr,
                                  globus_i_soap_message_tcp_driver,
                                  GLOBUS_XIO_TCP_SET_NODELAY,
                                  GLOBUS_TRUE);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set tcp nodelay.");
        goto exit;
    }


    result = globus_xio_register_open(
        handle->xio_handle,
        endpoint,
        attr,
        globus_l_soap_message_open_callback,
        callback_handle);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to connect to server");
        goto exit;
    }

    if(attr)
    {
        globus_xio_attr_destroy(attr);
        attr = NULL;
    }

 exit:

    if(result != GLOBUS_SUCCESS)
    {
        free(callback_handle);
        if(attr)
        {
            globus_xio_attr_destroy(attr);
        }
    }
    if (quoted_soap_action != NULL)
    {
        free(quoted_soap_action);
    }

    GlobusSoapMessageDebugExit();
    return result;
}

globus_bool_t
globus_soap_message_is_cancelled(
    globus_soap_message_handle_t        handle)
{
    globus_bool_t                       cancelled;

    globus_mutex_lock(&handle->cancel_mutex);
    cancelled = handle->cancelled;
    globus_mutex_unlock(&handle->cancel_mutex);

    return cancelled;
}

globus_result_t
globus_soap_message_cancel_message(
    globus_soap_message_handle_t        handle)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_operations_cancel);
    GlobusSoapMessageDebugEnter();

    globus_xio_handle_cancel_operations(
        handle->xio_handle, 
        GLOBUS_XIO_CANCEL_OPEN|
        GLOBUS_XIO_CANCEL_CLOSE|
        GLOBUS_XIO_CANCEL_READ|
        GLOBUS_XIO_CANCEL_WRITE);
    
    globus_mutex_lock(&handle->cancel_mutex);
    handle->cancelled = 1;
    globus_mutex_unlock(&handle->cancel_mutex);
    
    return result;
}

globus_result_t
globus_soap_message_register_write_request(
    globus_soap_message_handle_t        handle,
    globus_soap_message_callback_func_t callback,
    void *                              args,
    globus_bool_t                       response)
{
    char                                content_length_buffer[100];
    int                                 buffer_length;
    globus_l_soap_message_callback_handle_t * callback_handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_register_write_request);
    GlobusSoapMessageDebugEnter();

    globus_xio_handle_cntl(
        handle->xio_handle, globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_DONE_BUFFERING);

    callback_handle = (globus_l_soap_message_callback_handle_t *) 
    globus_malloc(sizeof(globus_l_soap_message_callback_handle_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(callback_handle, 0, 
           (sizeof(globus_l_soap_message_callback_handle_t)));

    callback_handle->callback = callback;
    callback_handle->args = args;
    callback_handle->message = handle;
    callback_handle->type = (response ? GLOBUS_SOAP_MESSAGE_REQUEST_CALLBACK:
			     		GLOBUS_SOAP_MESSAGE_RESPONSE_CALLBACK);

    if(callback_handle->message->writer)
    {
        xmlFreeTextWriter(callback_handle->message->writer);
        callback_handle->message->writer = NULL;
    }

    result = globus_xio_handle_cntl(
        handle->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_GET_LENGTH,
        &buffer_length);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set get length of message");
        goto exit;
    }

    memset(content_length_buffer, 0, 100);
    sprintf(content_length_buffer, "%d", buffer_length);

    result = globus_xio_handle_cntl(
        handle->xio_handle,
        globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_HANDLE_SET_REQUEST_HEADER,
        "Content-Length",
        content_length_buffer);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set Content-Length in HTTP header");
        goto exit;
    }

    if(GlobusSoapMessageDebug(GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES))
    {
        char *                          buffer;
        int                             read_length;
        
        globus_xio_handle_cntl(
            callback_handle->message->xio_handle,
            globus_i_soap_message_buffer_driver,
            GLOBUS_XIO_BUFFER_GET_MARKED_BUFFERS,
            NULL, NULL, &buffer, &read_length);
        
        GlobusSoapMessageDebugPrintf(
            GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES,
            ("----------------WRITING--------------\n%.*s\n"
             "=====================================\n\n",
             read_length, buffer));
        
        free(buffer);
    }

    result = globus_xio_handle_cntl(
        handle->xio_handle, 
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_FLUSH,
        handle->xio_handle,
        globus_l_soap_message_flush_callback,
        callback_handle);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to write message");
        goto exit;
    }        

 exit:

    if(result != GLOBUS_SUCCESS)
    {
        free(callback_handle);
    }

    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_soap_message_register_write_response(
    globus_soap_message_handle_t        handle,
    globus_soap_message_callback_func_t callback,
    void *                              args)
{
    char                                content_length_buffer[100];
    char *                              connection;
    int                                 buffer_length;
    globus_l_soap_message_callback_handle_t * callback_handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_register_write_response);
    GlobusSoapMessageDebugEnter();

    globus_xio_handle_cntl(
        handle->xio_handle, globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_DONE_BUFFERING);

    callback_handle = (globus_l_soap_message_callback_handle_t *) 
    globus_malloc(sizeof(globus_l_soap_message_callback_handle_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(callback_handle, 0, 
           (sizeof(globus_l_soap_message_callback_handle_t)));

    callback_handle->callback = callback;
    callback_handle->args = args;
    callback_handle->message = handle;
    callback_handle->type = GLOBUS_SOAP_MESSAGE_RESPONSE_CALLBACK;

    if(callback_handle->message->writer)
    {
        xmlFreeTextWriter(callback_handle->message->writer);
        callback_handle->message->writer = NULL;
    }

    result = globus_xio_handle_cntl(
        handle->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_GET_LENGTH,
        &buffer_length);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set get length of message");
        goto exit;
    }

    memset(content_length_buffer, 0, 100);
    sprintf(content_length_buffer, "%d", buffer_length);

    result = globus_xio_handle_cntl(
        handle->xio_handle,
        globus_i_soap_message_http_driver,
        GLOBUS_XIO_HTTP_HANDLE_SET_RESPONSE_HEADER,
        "Content-Length",
        content_length_buffer);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set Content-Length in HTTP header");
        goto exit;
    }

    connection = globus_soap_message_handle_get_attr(
        handle,
        GLOBUS_SOAP_MESSAGE_CONNECTION_KEY);

    if (connection && !strcmp(connection, "close"))
    {
        result = globus_xio_handle_cntl(
            handle->xio_handle,
            globus_i_soap_message_http_driver,
            GLOBUS_XIO_HTTP_HANDLE_SET_RESPONSE_HEADER,
            "Connection",
            "close");

        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set Connection in HTTP header");
            goto exit;
        }
    }

    if(GlobusSoapMessageDebug(GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES))
    {
        char *                          buffer;
        int                             read_length;
        
        globus_xio_handle_cntl(
            callback_handle->message->xio_handle,
            globus_i_soap_message_buffer_driver,
            GLOBUS_XIO_BUFFER_GET_MARKED_BUFFERS,
            NULL, NULL, &buffer, &read_length);
        
        GlobusSoapMessageDebugPrintf(
            GLOBUS_SOAP_MESSAGE_DEBUG_MESSAGES,
            ("----------------WRITING--------------\n%.*s\n"
             "=====================================\n\n",
             read_length, buffer));
        
        free(buffer);
    }

    if(buffer_length > 0)
    {
        result = globus_xio_handle_cntl(
            handle->xio_handle, 
            globus_i_soap_message_buffer_driver,
            GLOBUS_XIO_BUFFER_FLUSH,
            handle->xio_handle,
            globus_l_soap_message_flush_callback,
            callback_handle);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to write message");
            goto exit;
        }        
    }
    else
    {
        globus_xio_handle_cntl(
            handle->xio_handle,
            globus_i_soap_message_http_driver,
            GLOBUS_XIO_HTTP_HANDLE_SET_END_OF_ENTITY);

        callback_handle->callback(result, callback_handle->args);

        free(callback_handle);
        callback_handle = NULL;
    }        


    result = globus_xio_handle_cntl(
        handle->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_DONE_BUFFERING);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to write message");
        goto exit;
    }

 exit:

    if(result != GLOBUS_SUCCESS && callback_handle != NULL)
    {
        free(callback_handle);
    }

    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_soap_message_register_read_response(
    globus_soap_message_handle_t        handle,
    globus_soap_message_callback_func_t callback,
    void *                              args)
{
    globus_l_soap_message_callback_handle_t *  callback_handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_register_read_response);
    GlobusSoapMessageDebugEnter();


    callback_handle = globus_malloc(
        sizeof(globus_l_soap_message_callback_handle_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(callback_handle, 0, 
           (sizeof(globus_l_soap_message_callback_handle_t)));

    callback_handle->callback = callback;
    callback_handle->args = args;
    callback_handle->message = handle;

    globus_mutex_lock(&handle->mutex);
    handle->response_semaphore++;

    if(handle->response_semaphore > 1)
    {
        globus_mutex_unlock(&handle->mutex);
        result = globus_callback_register_oneshot(
            NULL,
            &globus_i_reltime_zero,
            globus_l_soap_message_response_callback,
            callback_handle);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to register periodic callbcak for response");
            goto exit;
        }
    }
    else
    {
        globus_mutex_unlock(&handle->mutex);
        handle->response_handle = (void *)callback_handle;
    }

 exit:

    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_soap_message_register_read_request(
    globus_soap_message_handle_t        handle,
    globus_soap_message_callback_func_t callback,
    void *                              args)
{
    globus_l_soap_message_callback_handle_t *  callback_handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_soap_message_register_read_request);
    GlobusSoapMessageDebugEnter();

    callback_handle = globus_malloc(
        sizeof(globus_l_soap_message_callback_handle_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    memset(callback_handle, 0, 
           (sizeof(globus_l_soap_message_callback_handle_t)));

    callback_handle->callback = callback;
    callback_handle->args = args;
    callback_handle->message = handle;
            
    result = globus_xio_handle_cntl(
        callback_handle->message->xio_handle,
        globus_i_soap_message_buffer_driver,
        GLOBUS_XIO_BUFFER_ENABLE_READ_BUFFERING);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to enable read buffering in buffer driver");
        goto exit;
    }

    result = globus_xio_register_read(
        handle->xio_handle,
        (globus_byte_t *)handle->readbuf,
        GLOBUS_I_SOAP_MESSAGE_READBUF_SIZE,
        1,
        NULL,
        globus_l_soap_message_read_callback,
        callback_handle);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to begin reading SOAP message");
        goto exit;
    }

 exit:

    GlobusSoapMessageDebugExit();
    return result;
}

typedef struct
{
    globus_soap_message_callback_func_t callback;
    void *                              args;
} globus_l_soap_message_close_callback_t;

void
globus_l_soap_message_register_close_callback(
    globus_xio_handle_t                 handle,
    globus_result_t                     result,
    void *                              args)
{
    globus_l_soap_message_close_callback_t * callback_handle;

    callback_handle = (globus_l_soap_message_close_callback_t *) args;

    callback_handle->callback(result, callback_handle->args);

    free(callback_handle);
}

globus_result_t
globus_soap_message_register_close(
    globus_soap_message_handle_t        handle,
    globus_soap_message_callback_func_t callback,
    void *                              args)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_soap_message_close_callback_t * callback_handle;
    GlobusFuncName(globus_soap_message_register_close);
    GlobusSoapMessageDebugEnter();

    if(handle->reader)
    {
        xmlFreeTextReader(handle->reader);
    }
    handle->reader = NULL;

    callback_handle = globus_malloc(
        sizeof(globus_l_soap_message_close_callback_t));
    if(!callback_handle)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto exit;
    }
    callback_handle->callback = callback;
    callback_handle->args = args;

    globus_xio_register_close(
        handle->xio_handle,
        NULL,
        globus_l_soap_message_register_close_callback,
        callback_handle);

    handle->xio_handle = NULL;

 exit:
    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_i_soap_message_reset_reader(
    globus_soap_message_handle_t        handle)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_i_soap_message_reset_reader);
    GlobusSoapMessageDebugEnter();

    if(handle->reader)
    {
        xmlFreeTextReader(handle->reader);
    }

    handle->buffer_eof = 0;
    
    handle->reader = xmlReaderForIO(
        globus_l_soap_message_handle_libxml_io_read_callback,
        globus_l_soap_message_handle_libxml_io_close_callback,
        handle,
        NULL,
        NULL,
        0);
    if(!handle->reader)
    {
        result = GlobusSoapMessageErrorInitFailed(
            "Failed to initialize xmlReader");
        goto exit;
    }

    xmlTextReaderSetErrorHandler(handle->reader, 
                                 globus_l_soap_message_error_callback, 
                                 handle);
    
    xmlTextReaderSetStructuredErrorHandler(
        handle->reader,
        globus_l_soap_message_structured_error_callback,
        handle);
    
 exit:

    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_i_soap_message_setup_transport_security(
    globus_soap_message_attr_t          attrs,
    globus_xio_attr_t                   xio_attr)
{
    globus_xio_gsi_authorization_mode_t         gsi_authz;
    globus_soap_message_auth_protection_level_t protect;
    globus_soap_message_authz_method_t          authz;
    void *                                      credential;
    void *                                      target_name;
    globus_result_t                             result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_l_soap_message_setup_transport_security);
    GlobusSoapMessageDebugEnter();

    credential = globus_soap_message_attr_get(
        attrs,
        GLOBUS_SOAP_MESSAGE_USER_CREDENTIAL_KEY);
    if(credential)
    {
        result = globus_xio_attr_cntl(
            xio_attr,
            globus_i_soap_message_gsi_driver,
            GLOBUS_XIO_GSI_SET_CREDENTIAL,
            credential);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set credential for gsi driver");
            goto exit;
        }
    }

    result = globus_xio_attr_cntl(
        xio_attr,
        globus_i_soap_message_gsi_driver,
        GLOBUS_XIO_GSI_SET_DELEGATION_MODE,
        GLOBUS_XIO_GSI_DELEGATION_MODE_NONE);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set delegation mode for gsi driver");
        goto exit;
    }        

    result = globus_xio_attr_cntl(
        xio_attr,
        globus_i_soap_message_gsi_driver,
        GLOBUS_XIO_GSI_SET_SSL_COMPATIBLE,
        1);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set delegation mode for gsi driver");
        goto exit;
    }        

    protect = 
    (globus_soap_message_auth_protection_level_t)
    globus_soap_message_attr_get(
        attrs,
        GLOBUS_SOAP_MESSAGE_AUTH_PROTECTION_KEY);
    if(protect != GLOBUS_SOAP_MESSAGE_AUTH_PROTECTION_NOT_SET)
    {
        result = globus_xio_attr_cntl(
            xio_attr,
            globus_i_soap_message_gsi_driver,
            GLOBUS_XIO_GSI_SET_PROTECTION_LEVEL,
            ((protect == GLOBUS_SOAP_MESSAGE_AUTH_PROTECTION_INTEGRITY) ?
             GLOBUS_XIO_GSI_PROTECTION_LEVEL_INTEGRITY :
             GLOBUS_XIO_GSI_PROTECTION_LEVEL_PRIVACY));
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set delegation mode for gsi driver");
            goto exit;
        }        
    }

    authz = 
    (globus_soap_message_authz_method_t)
    globus_soap_message_attr_get(
        attrs,
        GLOBUS_SOAP_MESSAGE_AUTHZ_METHOD_KEY);
    
    switch(authz)
    {
    case GLOBUS_SOAP_MESSAGE_AUTHZ_SELF:
        gsi_authz = GLOBUS_XIO_GSI_SELF_AUTHORIZATION;
        break;
    case GLOBUS_SOAP_MESSAGE_AUTHZ_IDENTITY:
        gsi_authz = GLOBUS_XIO_GSI_IDENTITY_AUTHORIZATION;
        break;
    case GLOBUS_SOAP_MESSAGE_AUTHZ_HOST:
        gsi_authz = GLOBUS_XIO_GSI_HOST_AUTHORIZATION;
        break;
    case GLOBUS_SOAP_MESSAGE_AUTHZ_NONE:
        gsi_authz = GLOBUS_XIO_GSI_NO_AUTHORIZATION;
        break;
    default:
        gsi_authz = GLOBUS_XIO_GSI_SELF_AUTHORIZATION;
    }

    result = globus_xio_attr_cntl(
        xio_attr,
        globus_i_soap_message_gsi_driver,
        GLOBUS_XIO_GSI_SET_AUTHORIZATION_MODE,
        gsi_authz);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorFailedTransport(
            result, "Failed to set delegation mode for gsi driver");
        goto exit;
    }        

    target_name = globus_soap_message_attr_get(
        attrs,
        GLOBUS_SOAP_MESSAGE_AUTHZ_TARGET_NAME_KEY);
    if(target_name)
    {
        result = globus_xio_attr_cntl(
            xio_attr,
            globus_i_soap_message_gsi_driver,
            GLOBUS_XIO_GSI_SET_TARGET_NAME,
            target_name);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorFailedTransport(
                result, "Failed to set target name on gsi driver");
            goto exit;
        }
    }

exit:

    GlobusSoapMessageDebugExit();
    return result;
}
    
