/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_soap_message.h"
#include "globus_i_soap_message.h"
#include "xsd_anyAttributes.h"

typedef struct
{
    xsd_QName                           qname;
    char *                              value;
} xsd_l_anyAttributes_element_t;

int
xsd_anyAttributes_size(
    xsd_anyAttributes_t                 attrs)
{
    return globus_hashtable_size(&attrs);
}

const char *
xsd_anyAttributes_lookup(
    xsd_anyAttributes_t                 attrs,
    xsd_QName *                         name)
{
    xsd_l_anyAttributes_element_t *     element;

    element = (xsd_l_anyAttributes_element_t *)
        globus_hashtable_lookup(&attrs, name);
    if(!element)
    {
        return NULL;
    }

    return element->value;
}

int
xsd_anyAttributes_insert(
    xsd_anyAttributes_t                 attrs,
    xsd_QName *                         name,
    const char *                        value)
{
    xsd_l_anyAttributes_element_t *     element;

    element = (xsd_l_anyAttributes_element_t *)
        globus_malloc(sizeof(xsd_l_anyAttributes_element_t));
    xsd_QName_copy_contents(&element->qname, name);
    element->value = globus_libc_strdup(value);

    return globus_hashtable_insert(&attrs, &element->qname, element);
}

static
void
xsd_anyAttributes_element_delete(
    void *                              value)
{
    xsd_l_anyAttributes_element_t *     element;
    
    element = (xsd_l_anyAttributes_element_t *) value;
    
    if(element)
    {
        xsd_QName_destroy_contents(&element->qname);
        if(element->value)
        {
            globus_free(element->value);
        }
        globus_free(element);
    }
}

int
xsd_anyAttributes_remove(
    xsd_anyAttributes_t                 attrs,
    xsd_QName *                         name)
{
    void *                              element;

    element = globus_hashtable_remove(&attrs, name);
    if(!element)
    {
        return GLOBUS_FAILURE;
    }
    
    xsd_anyAttributes_element_delete(element);
    return GLOBUS_SUCCESS;
}

void
xsd_anyAttributes_destroy(
    xsd_anyAttributes_t                 attrs)
{
    if(attrs)
    {
        globus_hashtable_destroy_all(&attrs, xsd_anyAttributes_element_delete);
    }
}

globus_result_t
xsd_anyAttributes_init(
    xsd_anyAttributes_t *               attrs)
{
    return globus_hashtable_init(attrs, 10, xsd_QName_hash, xsd_QName_keyeq);
}

static
int
xsd_QName_compare(
    void *                              a,
    void *                              b,
    void *                              args)
{
    int                                 res;
    xsd_QName *                         qa;
    xsd_QName *                         qb;

    
    if(!a)
    {
        return 1;
    }

    if(!b)
    {
        return 0;
    }

    qa = (xsd_QName *)a;
    qb = (xsd_QName *)b;

    if(!qa->Namespace && !qb->Namespace)
    {
        globus_assert_string(qa->local && qb->local, 
                             "QName structs must have valid local values");
        
        if(strcmp(qa->local, qb->local) <= 0)
        {
            return 1;
        }
        
        return 0;
    }

    if(qa->Namespace && qb->Namespace)
    {
        res = strcmp(qa->Namespace, qb->Namespace);
        if(res == 0)
        {
            if(strcmp(qa->local, qb->local) <= 0)
            {
                return 1;
            }

            return 0;
        }
        
        if(res < 0)
        {
            return 1;
        }

        return 0;
    }
         
    if((!qa->Namespace) && qb->Namespace)
    {
        return 1;
    }

    if((!qb->Namespace) && qa->Namespace)
    {
        return 0;
    }

    return 1;
}

globus_list_t *
xsd_anyAttributes_lex_sort(
    xsd_anyAttributes_t                 attrs)
{
    xsd_l_anyAttributes_element_t *     element;
    globus_list_t *                     keys = NULL;
    
    element = (xsd_l_anyAttributes_element_t *) globus_hashtable_first(&attrs);
    while(element)
    {
        globus_list_insert(&keys, &element->qname);
        element = (xsd_l_anyAttributes_element_t *) 
            globus_hashtable_next(&attrs);
    }

    keys = globus_list_sort(keys, xsd_QName_compare, NULL);

    return keys;
}

globus_result_t
xsd_anyAttributes_namespace_set(
    globus_soap_message_handle_t        message,
    xsd_anyAttributes_t                 attrs,
    globus_xsd_element_options_t *      options)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_l_anyAttributes_element_t *     element;
    if(attrs)
    {
        element = globus_hashtable_first(&attrs);
        while(element)
        {
            result = globus_soap_message_attribute_ns_set(
                message,
                element->qname.Namespace,
                options);
            if(result != GLOBUS_SUCCESS)
            {
                xsd_QName *             eqn = &element->qname;
                result = GlobusSoapMessageErrorSerializeFailed(
                    result, NULL, eqn);
                goto exit;
            }

            element = globus_hashtable_next(&attrs);
        }
    }

exit:
    
    return result;
}
globus_result_t
xsd_anyAttributes_namespace_add(
    globus_soap_message_handle_t        message,
    xsd_anyAttributes_t                 attrs,
    globus_xsd_element_options_t *      options)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_l_anyAttributes_element_t *     element;
    if(attrs)
    {
        element = globus_hashtable_first(&attrs);
        while(element)
        {
            result = globus_soap_message_attribute_ns_add(
                message,
                element->qname.Namespace,
                options);
            if(result != GLOBUS_SUCCESS)
            {
                xsd_QName *             eqn = &element->qname;
                result = GlobusSoapMessageErrorSerializeFailed(
                    result, NULL, eqn);
                goto exit;
            }

            element = globus_hashtable_next(&attrs);
        }
    }

exit:
    
    return result;
}

globus_result_t
xsd_anyAttributes_namespace_remove(
    globus_soap_message_handle_t        message,
    xsd_anyAttributes_t                 attrs,
    globus_xsd_element_options_t *      options)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_l_anyAttributes_element_t *     element;
    if(attrs)
    {
        element = globus_hashtable_first(&attrs);
        while(element)
        {
            globus_soap_message_attribute_ns_remove(
                message,
                element->qname.Namespace,
                options);
            if(result != GLOBUS_SUCCESS)
            {
                xsd_QName *             eqn = &element->qname;
                result = GlobusSoapMessageErrorSerializeFailed(
                    result, NULL, eqn);
                goto exit;
            }

            element = globus_hashtable_next(&attrs);
        }
    }

exit:
    
    return result;
}

globus_result_t
xsd_anyAttributes_serialize(
    globus_soap_message_handle_t        message,
    xsd_anyAttributes_t                 attrs)
{
    globus_list_t *                     lex_sorted_keys;
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_l_anyAttributes_element_t *     element;
    xsd_QName *                         key;

    lex_sorted_keys = xsd_anyAttributes_lex_sort(attrs);

    key = globus_list_first(lex_sorted_keys);
    while(key)
    {
        lex_sorted_keys = globus_list_rest(lex_sorted_keys);
        element = (xsd_l_anyAttributes_element_t *) 
            globus_hashtable_lookup(&attrs, key);
        if(!element)
        {
            result = GlobusSoapMessageErrorSerializeAnyAttrsFailed(
                result, "Null element in attributes list");
            goto exit;
        }
        
        result = xsd_string_serialize_attribute(
            &element->qname, &element->value, message, 0);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorSerializeAnyAttrsFailed(
                result, "Failed to serialize any attribute");
            goto exit;
        }
     
        key = globus_list_first(lex_sorted_keys);
    }

 exit:

    return result;
}

globus_result_t
xsd_anyAttributes_serialize_lex_before(
    globus_soap_message_handle_t        message,
    globus_list_t **                    lex_sorted_keys,
    xsd_anyAttributes_t                 attrs,
    xsd_QName *                         comparable_value)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_l_anyAttributes_element_t *     element;
    xsd_QName *                         key;

    while(*lex_sorted_keys)
    {
        key = globus_list_first(*lex_sorted_keys);
        
        if(!xsd_QName_compare(comparable_value, key, NULL))
        {
	    break;
        } 
         
        element = (xsd_l_anyAttributes_element_t *) 
            globus_hashtable_lookup(&attrs, key);
        if(!element)
        {
            result = GlobusSoapMessageErrorSerializeAnyAttrsFailed(
                result, "Null element in attributes list");
            goto exit;
        }
        
        result = xsd_string_serialize_attribute(
            &element->qname, &element->value, message, 0);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusSoapMessageErrorSerializeAnyAttrsFailed(
                result, "Failed to serialize any attribute");
            goto exit;
        }
     
        *lex_sorted_keys = globus_list_rest(*lex_sorted_keys);
    }

 exit:

    return result;
}

static
void
xsd_anyAttributes_QName_copy_callback(
    void **                             dest_key,
    void **                             dest_datum,
    void *                              src_key,
    void *                              src_datum)
{
    xsd_l_anyAttributes_element_t *     dest_element;
    xsd_l_anyAttributes_element_t *     src_element;

    src_element = (xsd_l_anyAttributes_element_t *) src_datum;

    dest_element = (xsd_l_anyAttributes_element_t *)
        globus_malloc(sizeof(xsd_l_anyAttributes_element_t));

    xsd_QName_copy_contents(&dest_element->qname, (const xsd_QName *) src_key);
    dest_element->value = globus_libc_strdup(src_element->value);

    *dest_datum = dest_element;
    *dest_key = &dest_element->qname;
}

globus_result_t
xsd_anyAttributes_copy(
    xsd_anyAttributes_t *               dest_attrs,
    xsd_anyAttributes_t                 src_attrs)
{
    globus_result_t                     result = GLOBUS_SUCCESS;

    result = globus_hashtable_copy(dest_attrs, &src_attrs, 
                                   xsd_anyAttributes_QName_copy_callback);
    
    return result;
}

globus_result_t
xsd_anyAttributes_deserialize(
    globus_soap_message_handle_t        message,
    xsd_anyAttributes_t                 attrs)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    int                                 res = 0;

    res = xmlTextReaderMoveToFirstAttribute(message->reader);
    if(res < 0)
    {
        result = GlobusSoapMessageErrorDeserializeAnyAttrsFailed(
            message->result, NULL);

        goto exit;
    }

    while(res)
    {
        xmlNodePtr                      attrNode;
        xsd_l_anyAttributes_element_t * element;
        
        attrNode = xmlTextReaderCurrentNode(message->reader);
        if(attrNode->type == XML_ATTRIBUTE_NODE)
        {
            element = (xsd_l_anyAttributes_element_t *)
                globus_malloc(sizeof(xsd_l_anyAttributes_element_t));
            element->qname.local = xmlTextReaderLocalName(message->reader);
            element->qname.Namespace =
                xmlTextReaderNamespaceUri(message->reader);
            element->value = xmlTextReaderValue(message->reader);
            
            globus_hashtable_insert(&attrs, &element->qname, element);
        }

        res = xmlTextReaderMoveToNextAttribute(message->reader);
        if(res == -1)
        {
            result = GlobusSoapMessageErrorDeserializeAnyAttrsFailed(
                message->result, NULL);
            goto exit;
        }
    }

    res = xmlTextReaderMoveToElement(message->reader);
    if(res < 0)
    {
        result = GlobusSoapMessageErrorDeserializeAnyAttrsFailed(
            message->result, NULL);

        goto exit;
    }

 exit:

    return result;
}

