/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_common.h"
#include "version.h"
#include "globus_i_wsdl_parser.h"

#include "libxml/xmlschemas.h"
#include "libxml/schemasInternals.h"

#include "wsdl.h"

static int globus_l_wsdl_parser_activate();
static int globus_l_wsdl_parser_deactivate();

globus_module_descriptor_t              globus_i_wsdl_parser_module =
{
    "globus_wsdl_parser",
    globus_l_wsdl_parser_activate,
    globus_l_wsdl_parser_deactivate,
    GLOBUS_NULL,
    GLOBUS_NULL,
    &local_version
};

GlobusDebugDefine(GLOBUS_WSDL_PARSER);

static
int
globus_l_wsdl_parser_activate()
{
    int                                 res = (int) GLOBUS_SUCCESS;

    res = (int) globus_module_activate(GLOBUS_COMMON_MODULE);

    GlobusDebugInit(GLOBUS_WSDL_PARSER, ERROR WARN TRACE INFO TEMPLATE XSDDUMP JSDEBUG);

    return res;
}

static
int
globus_l_wsdl_parser_deactivate()
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_module_deactivate(GLOBUS_COMMON_MODULE);

    GlobusDebugDestroy(GLOBUS_WSDL_PARSER);

    return (int) result;
}

static void
internal_error_func(
    void *                              context,
    const char *                        message,
    ...)
{
    char *                              message_string;
    va_list                             ap;
    va_start(ap, message);
    message_string = globus_common_v_create_string(message, ap);
    va_end(ap);

    if(GlobusWSDLDebug(GLOBUS_WSDL_DEBUG_WARN))
    {
        GlobusWSDLDebugPrintf(GLOBUS_WSDL_DEBUG_WARN, 
                              (message_string ? 
                               message_string : "Unknown Error"));
    }
    
    *((globus_result_t *)context) = GlobusWSDLErrorParsingSchema(
        (*((globus_result_t *)context)),
        (message_string ? message_string : "Unknown Error"));
    
    if(message_string)
    {
        globus_free(message_string);
    }
}

static void
internal_warn_func(
    void *                              context,
    const char *                        message,
    ...)
{
    if(GlobusWSDLDebug(GLOBUS_WSDL_DEBUG_WARN))
    {
        va_list                             ap;
        va_start(ap, message);
        vfprintf(GlobusWSDLDebugStream, message, ap);
        va_end(ap);

        *((globus_result_t *)context) = GLOBUS_SUCCESS;
    }
}

void
xmlSchemaImportDump(void * payload, void * data, xmlChar * name)
{
    FILE *                              output = (FILE *) data;
    xmlSchemaImportPtr                  import = (xmlSchemaImportPtr) payload;

    if(import == NULL)
    {
        fprintf(output, "Import: NULL\n");
    }

    xmlSchemaDump(output, import->schema);
}

void
wsdlImportDump(void * payload, void * data, xmlChar * name)
{
    wsdlSchemaPtr                       schema;
    wsdlImportPtr                       import = (wsdlImportPtr) payload;
    FILE *                              output = (FILE *) data;

    if(import == NULL)
    {
        fprintf(output, "WSDL Import: NULL\n");
    }
    
    schema = import->schema;
 
    if(GlobusWSDLDebug(GLOBUS_WSDL_DEBUG_XSD_DUMP))
    {
        xmlSchemaPtr                    xsd;
        xsd = schema->schemaTypes;
        while(xsd)
        {
            xmlHashScan(xsd->schemasImports,
                        xmlSchemaImportDump,
                        GlobusWSDLDebugStream);
            xmlSchemaDump(GlobusWSDLDebugStream, xsd);
        
            xsd = (xmlSchemaPtr) xsd->_private;
        }

        xmlHashScan(schema->importsDecl, 
                    wsdlImportDump, GlobusWSDLDebugStream);
    }
}

globus_result_t
globus_i_wsdl_parse_schema(
    const char *                        schema_filename, 
    wsdlSchemaPtr *                     schema)
{
    xmlErrorPtr                         error = NULL;
    wsdlParserCtxtPtr                   parser = NULL;
    globus_result_t                     wsdl_result = GLOBUS_SUCCESS;
    globus_result_t                     xsd_result = GLOBUS_SUCCESS;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_i_wsdl_parse_schema);
    GlobusWSDLDebugEnter();

    parser = wsdlNewParserCtxt(schema_filename);
    if(!parser)
    {
        error = xmlGetLastError();
        result = GlobusWSDLErrorParsingSchema(
            GLOBUS_SUCCESS,
            (error && error->message) ? error->message : "Parser Error");
        goto exit;
    }

    wsdlSetParserErrors(parser, 
                        internal_error_func, 
                        internal_warn_func,
                        (void *)&wsdl_result);

    *schema = wsdlParse(parser);
    if(!*schema)
    {
        if(wsdl_result != GLOBUS_SUCCESS)
        {
            if(GlobusWSDLDebug(GLOBUS_WSDL_DEBUG_WARN))
            {
                GlobusWSDLDebugPrintf(
                    GLOBUS_WSDL_DEBUG_WARN,
                    (globus_error_print_chain(globus_error_peek(wsdl_result))));
            }
        }

        wsdlSetParserErrors(parser, 
                            internal_error_func, 
                            internal_warn_func,
                            (void *)&xsd_result);
        *schema = wsdlParseXmlSchema(parser);
        if(xsd_result != GLOBUS_SUCCESS)
        {
            if(wsdl_result != GLOBUS_SUCCESS)
            {
                result = GlobusWSDLErrorParsingSchema(
                    wsdl_result, "Parser Error");
                goto exit;
            }
            
            result = GlobusWSDLErrorParsingSchema(
                xsd_result, "Parser Error");
            goto exit;
        }
        
        if(!*schema)
        {
            result = GlobusWSDLErrorParsingSchema(
                result, "Parser Error");
            goto exit;
        }

        result = GLOBUS_SUCCESS;
    }

    if(GlobusWSDLDebug(GLOBUS_WSDL_DEBUG_XSD_DUMP))
    {
        xmlSchemaPtr                    xsd;
        xsd = (*schema)->schemaTypes;
        while(xsd)
        {
            xmlHashScan(xsd->schemasImports,
                        xmlSchemaImportDump,
                        GlobusWSDLDebugStream);
            xmlSchemaDump(GlobusWSDLDebugStream, xsd);
        
            xsd = (xmlSchemaPtr) xsd->_private;
        }

        xmlHashScan((*schema)->importsDecl, 
                    wsdlImportDump, GlobusWSDLDebugStream);
    }

  exit:

    if(parser)
    {
        wsdlFreeParserCtxt(parser);
    }

    GlobusWSDLDebugExit();
    return result;
}
