/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "libxml/xmlerror.h"
#include "globus_i_wsdl_parser.h"

static
int
flatten_wsdl(
    const char *                        input_schema_filename,
    const char **                       suffixes);

static
void
flatten_import(
    void *                              imp,
    void *                              data,
    xmlChar *                           Namespace);

static
void
flatten(
    void *                              port_type,
    void *                              data,
    xmlChar *                           name);

static
int
get_operations(
    wsdlParserCtxtPtr                   parser,
    wsdlSchemaPtr                       schema,
    wsdlPortTypePtr                     portType,
    xsdQNamePtr                         implemented,
    xmlHashTablePtr                     operations);

static
int
get_resource_properties(
    wsdlParserCtxtPtr                   parser,
    wsdlSchemaPtr                       schema,
    xsdQNamePtr                         implemented,
    xmlListPtr                          props);

static
int
flatten_wsdl(
    const char *                        input_schema_filename,
    const char **                       suffixes)
{
    xmlErrorPtr                         error;
    wsdlParserCtxtPtr                   parser;
    wsdlSchemaPtr                       schema = NULL;
    char *                              start_suff;
    char *                              outfn;
    char *                              infn;
    globus_result_t                     result = GLOBUS_SUCCESS;
    int                                 res = 0;

    parser = wsdlNewParserCtxt(input_schema_filename);
    if(!parser)
    {
        error = xmlGetLastError();
        result = GlobusWSDLErrorParsingSchema(
            GLOBUS_NULL,
            (error && error->message) ? error->message : "Unknown Error");
    }

    schema = wsdlParse(parser);
    if(!schema)
    {
        if(result == GLOBUS_SUCCESS)
        {
            result = GlobusWSDLErrorParsingSchema(
                GLOBUS_SUCCESS, "Unknown Error");
        }
    }

/*     if(schema && schema->importsDecl) */
/*     { */
/*         xmlHashScan(schema->importsDecl, flatten_import, suffixes); */
/*     } */

    if(schema && schema->portTypesDecl)
    {
        xmlHashScan(schema->portTypesDecl, flatten, parser);
    }

    infn = globus_libc_strdup(input_schema_filename);

    start_suff = strstr(infn, suffixes[0]);
    if(!start_suff)
    {
        /* error */
    }

    *start_suff = '\0';

    outfn = malloc((start_suff - input_schema_filename) + 
                   strlen(suffixes[1]) + 1);
    if(!outfn)
    {
        /* error */
    }

    sprintf(outfn, "%s%s", 
            infn, 
            suffixes[1]);
            
    res = xmlSaveFormatFile(outfn, parser->doc, 1);
    if(res < 0)
    {
        error = xmlGetLastError();
        fprintf(stderr, "ERROR: %s\n", error->message);
        exit(1);
    }

    return 0;
}

static
void
flatten_import(
    void *                              imp,
    void *                              data,
    xmlChar *                           Namespace)
{
    wsdlImportPtr                       import = (wsdlImportPtr) imp;
    const char **                       suffixes = (const char **) data;

    flatten_wsdl(import->wsdlLocation, suffixes);
}

static
void
flatten(
    void *                              port_type,
    void *                              data,
    xmlChar *                           name)
{
    int                                 res = 0;
    xmlNodePtr                          current_node = NULL;
    xmlSchemaElementPtr                 rp_doc = NULL;
    xmlSchemaElementPtr                 rp_element = NULL;
    xmlSchemaElementPtr                 any_element = NULL;
    xmlSchemaTypePtr                    tmp_type = NULL;
    wsdlPortTypePtr                     portType = NULL;
    wsdlParserCtxtPtr                   parser = (wsdlParserCtxtPtr) data;
    xmlHashTablePtr                     props_hash;
    xmlListPtr                          props;

    portType = (wsdlPortTypePtr) port_type;

    if(portType->implements)
    {
        props = xmlListCreate(NULL, NULL);
        res = get_resource_properties(
            parser,
            parser->schema,
            portType->implements,
            props);
        
        if(portType->ResourceProperties)
        {
            rp_doc = wsdlFindElement(
                parser, parser->schema,
                portType->ResourceProperties->name,
                portType->ResourceProperties->Namespace);
            if(!rp_doc)
            {
                /* error */
            }

            if(!rp_doc->subtypes ||
               rp_doc->subtypes->type != XML_SCHEMA_TYPE_COMPLEX)
            {
                /* error */
            }

            tmp_type = rp_doc->subtypes->subtypes;
            while(tmp_type &&
                  ((tmp_type->type != XML_SCHEMA_TYPE_SEQUENCE) &&
                   (tmp_type->type != XML_SCHEMA_TYPE_ALL)))
            {
                tmp_type = tmp_type->subtypes;
            }

            if(!tmp_type)
            {
                /* error */
            }

            if(!tmp_type->subtypes)
            {
                /* error */
            }
            
            props_hash = xmlHashCreate(10);
            
            rp_element = xmlLinkGetData(xmlListFront(props));
            while(rp_element)
            {
                if(rp_element->type == XML_SCHEMA_TYPE_ANY)
                {
                    any_element = rp_element;
                }
                else if(!xmlHashLookup2(props_hash, 
                                  rp_element->refNs, 
                                  rp_element->ref))
                {
                    if(!tmp_type->subtypes->node)
                    {
                        xmlAddChild(tmp_type->node,
                                    xmlCopyNode(rp_element->node, 1));
                    }
                    else
                    {
                        xmlAddPrevSibling(tmp_type->subtypes->node,
                                          xmlCopyNode(rp_element->node, 1));
                    }
                    
                    xmlHashAddEntry2(props_hash, 
                                     rp_element->refNs, 
                                     rp_element->ref,
                                     rp_element);
                }

                xmlListPopFront(props);
                rp_element = xmlLinkGetData(xmlListFront(props));
            }

            if(any_element)
            {
                if(!tmp_type->subtypes->node)
                {
                    xmlAddChild(tmp_type->node,
                                xmlCopyNode(any_element->node, 1));
                }
                else
                {
                    xmlAddPrevSibling(tmp_type->subtypes->node,
                                      xmlCopyNode(any_element->node, 1));
                }
            }
        }
        else
        {
            /* create GED RP element in doc and add RP attr to PortType */
        }
            
        if(portType->operations)
        {
            current_node = portType->operations->node;
            while(current_node->next)
            {
                current_node = current_node->next;
            }
        }

        if(portType->implements)
        {
            xmlHashTablePtr                 ops;
            xmlAttrPtr                      impl_attr;
        
            impl_attr = xmlHasNsProp(portType->node, "implements", GTWSDL_NS);
            if(impl_attr)
            {
                xmlRemoveProp(impl_attr);
            }

            ops = xmlHashCreate(10);

            res = get_operations(
                parser, 
                parser->schema, 
                portType, 
                portType->implements,
                ops);
            xmlHashFree(ops, NULL);
        }
    }
}

static
int
get_resource_properties(
    wsdlParserCtxtPtr                   parser,
    wsdlSchemaPtr                       schema,
    xsdQNamePtr                         implements,
    xmlListPtr                          props)
{
    wsdlPortTypePtr                     pt;
    xmlSchemaElementPtr                 rp_doc;
    xmlSchemaTypePtr                    tmp_type;
    xsdQNamePtr                         impl = implements;
     
    while(impl)
    {
        pt = wsdlFindPortType(
            parser, parser->schema, 
            impl->name, 
            impl->Namespace);
        if(!pt)
        {
            return -1;
        }
        
        if(pt->implements)
        {
            get_resource_properties(parser, schema, pt->implements, props);
        }

        if(pt->ResourceProperties)
        {
            rp_doc = wsdlFindElement(
                parser, parser->schema,
                pt->ResourceProperties->name,
                pt->ResourceProperties->Namespace);
            if(!rp_doc)
            {
                /* error */
            }
            
            tmp_type = rp_doc->subtypes;
            while(tmp_type && (tmp_type->type != XML_SCHEMA_TYPE_ELEMENT))
            {
                tmp_type = tmp_type->subtypes;
            }

            if(!tmp_type)
            {
                /* error */
            }

            while(tmp_type)
            {
                xmlListAppend(props, tmp_type);
                tmp_type = tmp_type->next;
            }
        }

        impl = impl->next;
    }

    return 0;
}
     
static
int
get_operations(
    wsdlParserCtxtPtr                   parser,
    wsdlSchemaPtr                       schema,
    wsdlPortTypePtr                     portType,
    xsdQNamePtr                         implemented,
    xmlHashTablePtr                     operations)
{
    xmlNodePtr                          current_node = NULL;
    xsdQNamePtr                         impl = implemented;
    wsdlPortTypePtr                     pt;
    wsdlOperationPtr                    op;
    int                                 res = 0;
        
    while(impl)
    {
        pt = wsdlFindPortType(
            parser, parser->schema, 
            impl->name, 
            impl->Namespace);
        if(!pt)
        {
            return -1;
        }
        
        if(pt->implements)
        {
            res = get_operations(
                parser, schema, portType, pt->implements, operations);
            if(res < 0)
            {
                return res;
            }
        }

        if(pt->operations)
        {
            op = pt->operations;
            while(op)
            {
                if(!xmlHashLookup(operations, op->name))
                {
                    if(current_node)
                    {
                        xmlAddNextSibling(current_node,
                                          xmlCopyNode(op->node, 1));
                        current_node = current_node->next;
                    }
                    else
                    {
                        current_node = xmlAddChild(portType->node,
                                    xmlCopyNode(op->node, 1));
                    }

                    xmlHashAddEntry(operations, op->name, (void *)op->name);
                }
                
                op = op->next;
            }
        }
        
        impl = impl->next;
    }

    return 0;
}

int
main(int argc, char * argv[])
{
    const char *                        suffixes[2];

    suffixes[0] = "_compact.wsdl";
    suffixes[1] = "_flattened.wsdl";

    flatten_wsdl(argv[1], suffixes);

    return 0;
}
