/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_i_xio_buffer.h"
#include "version.h"
#include "globus_xio_http.h"

#define GROW_SIZE                       10

static int
globus_l_xio_buffer_activate();

static int
globus_l_xio_buffer_deactivate();


GlobusXIODefineModule(buffer) =
{
    "globus_xio_buffer",
    globus_l_xio_buffer_activate,
    globus_l_xio_buffer_deactivate,
    GLOBUS_NULL,
    GLOBUS_NULL,
    &local_version
};

GlobusDebugDefine(GLOBUS_XIO_BUFFER);

static
void
globus_l_buffer_write_marker_destroy(
    void *                              element)
{
    globus_l_buffer_write_marker_t *     write_marker;

    write_marker = (globus_l_buffer_write_marker_t *) element;

    free(write_marker->name);
    free(write_marker);
}

static
void
globus_l_buffer_read_marker_destroy(
    void *                              element)
{
    globus_l_buffer_read_marker_t *    read_marker;

    read_marker = (globus_l_buffer_read_marker_t *) element;

    free(read_marker->name);
    free(read_marker);
}

static
void
globus_l_buffer_handle_destroy(
    globus_l_buffer_handle_t *          handle)
{
    GlobusXIOName(globus_l_buffer_handle_destroy);
    GlobusXIOBufferDebugInternalEnter();

    if(handle->write_buffers)
    {
        int i;
        for(i = 0; i < handle->buffer_count; ++i)
        {
            free(handle->write_buffers[i].iov_base);
        }

        free(handle->write_buffers);
    }

    if(handle->read_buffers)
    {
        int i = 0;
        for(i = 0; i < handle->buffer_count; ++i)
        {
            globus_assert(handle->read_buffers[i].read_element);
            handle->read_buffers[i].read_element->ref_count--;
            if(handle->read_buffers[i].read_element->ref_count == 0)
            {
                free(handle->read_buffers[i].read_element->buffer);
                free(handle->read_buffers[i].read_element);
            }
        }

        free(handle->read_buffers);
    }

    if(handle->buffer_map)
    {
        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            globus_hashtable_destroy_all(
                &handle->buffer_map, globus_l_buffer_write_marker_destroy);
        }
        else
        {
            globus_hashtable_destroy_all(
                &handle->buffer_map, globus_l_buffer_read_marker_destroy);
            handle->current_read_marker = NULL;
        }
    }

    free(handle);
}

static
void
globus_l_xio_buffer_handle_clear(
    globus_l_buffer_handle_t *          handle)
{
    int                                 jj = 0;

    if(handle->write_buffers)
    {
        for(jj = 0; jj < handle->buffer_count; ++jj)
        {
            if(handle->write_buffers[jj].iov_len)
            {
                free(handle->write_buffers[jj].iov_base);
            }
        }

        free(handle->write_buffers);
        handle->write_buffers = NULL;
        handle->buffer_count = 0;
        handle->buffers_length = 0;
    }

    if(handle->read_buffers)
    {
        int i = 0;
        for(i = 0; i < handle->buffer_count; ++i)
        {
            handle->read_buffers[i].read_element->ref_count--;
            if(handle->read_buffers[i].read_element->ref_count == 0)
            {
                free(handle->read_buffers[i].read_element->buffer);
                free(handle->read_buffers[i].read_element);
            }
        }

        free(handle->read_buffers);
        handle->read_buffers = NULL;
        handle->buffer_count = 0;
        handle->buffers_length = 0;
    }

    if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
    {
        globus_hashtable_destroy_all(
            &handle->buffer_map, globus_l_buffer_write_marker_destroy);
    }
    else
    {
        globus_hashtable_destroy_all(
            &handle->buffer_map, globus_l_buffer_read_marker_destroy);
            handle->current_read_marker = NULL;
    }

    globus_hashtable_init(&handle->buffer_map, 10,
                          globus_hashtable_string_hash,
                          globus_hashtable_string_keyeq);
}

static
void
globus_l_xio_buffer_close_callback(
    struct globus_i_xio_op_s *          op,
    globus_result_t                     result,
    globus_size_t                       nbytes,
    void *                              user_args)
{
    globus_l_buffer_handle_t *          handle;
    GlobusXIOName(globus_l_xio_buffer_close_callback);
    GlobusXIOBufferDebugInternalEnter();

    handle = (globus_l_buffer_handle_t *) user_args;

    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }

    globus_l_buffer_handle_destroy(handle);

    globus_xio_driver_pass_close(op, NULL, NULL);

    GlobusXIOBufferDebugInternalExit();

 error:

    GlobusXIOBufferDebugInternalExitWithError();
}

static
int
globus_l_xio_buffer_get_marked_length(
    globus_l_buffer_handle_t *          handle,
    globus_l_buffer_read_marker_t *     begin_marker, 
    globus_l_buffer_read_marker_t *     end_marker)
{
    int len = 0;
    int buff_ind;
    
    globus_assert_string(handle, "NULL handle passed to function");

    buff_ind = (begin_marker ? begin_marker->buffer_index : 0);
    
    if(end_marker)
    {
        if(begin_marker->buffer_index == end_marker->buffer_index)
        {
            len += end_marker->offset - begin_marker->offset;
        }
        else
        {
            len += handle->read_buffers[buff_ind].length -
            begin_marker->offset;
            
            buff_ind++;
            while(buff_ind < end_marker->buffer_index)
            {
                len += handle->read_buffers[buff_ind].length;
                buff_ind++;
            }
            
            len += end_marker->offset;
        }
    }
    else
    {
        len += handle->read_buffers[buff_ind].length - 
               (begin_marker ? begin_marker->offset : 0);
        buff_ind++;
        while(buff_ind < handle->buffer_count)
        {
            len += handle->read_buffers[buff_ind].length;
            buff_ind++;
        }
    }

    return len;
}

static
globus_result_t
globus_l_xio_buffer_get_marked_buffers(
    globus_l_buffer_handle_t *          handle,
    globus_l_buffer_read_marker_t *     begin_marker, 
    globus_l_buffer_read_marker_t *     end_marker,
    char **                             out_buffer,
    size_t *                            out_size)
{
    int                                 len = 0;
    int                                 buff_ind;
    int                                 bytes_copied = 0;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusXIOName(globus_l_xio_buffer_get_marked_buffers);
    GlobusXIOBufferDebugEnter();
    
    globus_assert_string(handle, "NULL handle passed to function");
    buff_ind = (begin_marker ? begin_marker->buffer_index : 0);

    len = globus_l_xio_buffer_get_marked_length(
        handle, begin_marker, end_marker);
    *out_buffer = globus_malloc(len);
    if(!*out_buffer)
    {
        *out_size = 0;
        result = GlobusXIOBufferErrorOutOfMemory;
        goto exit;
    }
    *out_size = len;

    GlobusXIOBufferDebugPrintf(GLOBUS_XIO_BUFFER_DEBUG_BUFFERS,
			       ("MARKED BUFFER LENGTH: %d\n", len));

    if(end_marker)
    {
        if(begin_marker->buffer_index == end_marker->buffer_index)
        {
            memcpy(*out_buffer,
                   handle->read_buffers[buff_ind].read_element->buffer +
                   handle->read_buffers[buff_ind].start_pos +
                   begin_marker->offset,
                   len);
            bytes_copied = len;
        }
        else
        {
            memcpy(*out_buffer,
                   handle->read_buffers[buff_ind].read_element->buffer +
                   handle->read_buffers[buff_ind].start_pos +
                   begin_marker->offset,
                   handle->read_buffers[buff_ind].length -
                   begin_marker->offset);
            
            bytes_copied = handle->read_buffers[buff_ind].length -
            begin_marker->offset;
            buff_ind++;
            while(buff_ind < end_marker->buffer_index)
            {
                memcpy(*out_buffer + bytes_copied,
                       handle->read_buffers[buff_ind].read_element->buffer +
                       handle->read_buffers[buff_ind].start_pos,
                       handle->read_buffers[buff_ind].length);
                bytes_copied += handle->read_buffers[buff_ind].length;
                *out_size += bytes_copied;
                buff_ind++;
            }
            
            memcpy(*out_buffer + bytes_copied,
                   handle->read_buffers[buff_ind].read_element->buffer +
                   handle->read_buffers[buff_ind].start_pos,
                   end_marker->offset);
            bytes_copied += end_marker->offset;
        }
    }
    else
    {
        int                             begin_offset =
        (begin_marker ? begin_marker->offset : 0);

        memcpy(*out_buffer,
               handle->read_buffers[buff_ind].read_element->buffer +
               handle->read_buffers[buff_ind].start_pos +
               begin_offset,
               handle->read_buffers[buff_ind].length - begin_offset);
        bytes_copied += handle->read_buffers[buff_ind].length - 
        begin_offset;

        buff_ind++;
        while(buff_ind < handle->buffer_count)
        {
            memcpy(*out_buffer + bytes_copied,
                   handle->read_buffers[buff_ind].read_element->buffer +
                   handle->read_buffers[buff_ind].start_pos,
                   handle->read_buffers[buff_ind].length);

            bytes_copied += handle->read_buffers[buff_ind].length;
            buff_ind++;
        }
    }

    GlobusXIOBufferDebugPrintf(GLOBUS_XIO_BUFFER_DEBUG_BUFFERS,
			       ("MARKED BUFFER BYTES COPIED: %d\n",
				bytes_copied));

    *out_size = bytes_copied;

exit:
    GlobusXIOBufferDebugExit();
    return result;
}

static
globus_result_t
globus_l_xio_buffer_replace_marked_buffers(
    globus_l_buffer_handle_t *          handle,
    globus_l_buffer_read_marker_t *     begin_marker, 
    globus_l_buffer_read_marker_t *     end_marker,
    char  *                             in_buffer,
    size_t                              in_length)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    int                                 buff_ind;
    globus_l_read_buffer_element_t *    re = NULL;
    GlobusXIOName(globus_l_xio_buffer_replace_marked_buffers);

    globus_assert_string(handle, "NULL handle passed to function");
    globus_assert_string(begin_marker, "NULL marker passed to function");

    handle->buffers_length += in_length - 
    globus_l_xio_buffer_get_marked_length(
        handle, begin_marker, end_marker);

    buff_ind = begin_marker->buffer_index;

    if(end_marker)
    {
        int                             move_length;
        
        buff_ind = begin_marker->buffer_index + 1;
        while(buff_ind < end_marker->buffer_index)
        {
            handle->read_buffers[buff_ind].read_element->ref_count--;
            if(handle->read_buffers[buff_ind].read_element->ref_count == 0)
            {
                free(handle->read_buffers[buff_ind].read_element->buffer);
                free(handle->read_buffers[buff_ind].read_element);
            }
        }

        move_length = begin_marker->buffer_index - 
            end_marker->buffer_index + 2;

        if(move_length > 0)
        {
            handle->read_buffers = realloc(
                handle->read_buffers,
                sizeof(globus_l_read_buffer_t) *
                (move_length + handle->buffer_count));
        }

        memmove(handle->read_buffers + begin_marker->buffer_index + 2,
                handle->read_buffers + end_marker->buffer_index,
                (handle->buffer_count - end_marker->buffer_index) *
                sizeof(globus_l_read_buffer_t));

        handle->buffer_count += move_length;
        
        end_marker->buffer_index += move_length;
        handle->read_buffers[end_marker->buffer_index].
            read_element->ref_count++;
        
        handle->read_buffers[end_marker->buffer_index].start_pos +=
            end_marker->offset;
        handle->read_buffers[end_marker->buffer_index].length -= 
            end_marker->offset;
        end_marker->offset = 0;

        handle->read_buffers[begin_marker->buffer_index].length = 
            begin_marker->offset + 1;

        re = globus_malloc(sizeof(globus_l_read_buffer_element_t));
        if(!re)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            goto exit;
        }
        
        re->buffer = globus_malloc(in_length);
        if(!re->buffer)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            goto exit;
        }

        re->ref_count++;
        
        memcpy(re->buffer, in_buffer, in_length);

        handle->read_buffers[end_marker->buffer_index - 1].read_element = re;
        handle->read_buffers[end_marker->buffer_index - 1].length = in_length;
        handle->read_buffers[end_marker->buffer_index - 1].start_pos = 0;

        while(begin_marker->next && begin_marker->next != end_marker)
        {
            globus_l_buffer_read_marker_t * tmp_marker = NULL;
            tmp_marker = begin_marker->next;
            begin_marker->next = begin_marker->next->next;
            tmp_marker = globus_hashtable_remove(
                &handle->buffer_map, tmp_marker->name);
            free(tmp_marker->name);
            free(tmp_marker);
        }

        end_marker->prev = begin_marker;
    }
    else
    {
        int                             new_re_index = 0;

        buff_ind = begin_marker->buffer_index + 1;
        while(buff_ind < handle->buffer_count)
        {
            handle->read_buffers[buff_ind].read_element->ref_count--;
            if(handle->read_buffers[buff_ind].read_element->ref_count == 0)
            {
                free(handle->read_buffers[buff_ind].read_element->buffer);
                free(handle->read_buffers[buff_ind].read_element);
            }
        }

        if(begin_marker->buffer_index + 1 == handle->buffer_count)
        {
            handle->read_buffers = realloc(handle->read_buffers,
                                           (handle->buffer_count + 1) *
                                           sizeof(globus_l_read_buffer_t));
        }

        handle->buffer_count = begin_marker->buffer_index + 2;
        re = malloc(sizeof(globus_l_read_buffer_element_t));
        if(!re)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            goto exit;
        }
        
        re->buffer = malloc(in_length);
        if(!re->buffer)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            goto exit;
        }
        re->ref_count++;
        
        memcpy(re->buffer, in_buffer, in_length);

        new_re_index = begin_marker->buffer_index + 1;

        handle->read_buffers[new_re_index].read_element = re;
        handle->read_buffers[new_re_index].length = in_length;
        handle->read_buffers[new_re_index].start_pos = 0;

        while(begin_marker->next)
        {
            globus_l_buffer_read_marker_t * tmp_marker;
            tmp_marker = begin_marker->next;
            begin_marker->next = begin_marker->next->next;
            tmp_marker = globus_hashtable_remove(
                &handle->buffer_map, tmp_marker->name);
            free(tmp_marker->name);
            free(tmp_marker);
        }
    }

 exit:

    return result;
}

static
void
globus_l_xio_buffer_remove_marked_buffers(
    globus_l_buffer_handle_t *          handle,
    globus_l_buffer_read_marker_t *     begin_marker, 
    globus_l_buffer_read_marker_t *     end_marker)
{
    int                                 buff_ind;

    globus_assert_string(handle, "NULL handle passed to function");
    globus_assert_string(begin_marker, "NULL marker passed to function");

    handle->buffers_length -= 
        globus_l_xio_buffer_get_marked_length(
            handle, begin_marker, end_marker);

    buff_ind = begin_marker->buffer_index;

    if(end_marker)
    {
        int                             move_length;
        
        buff_ind = begin_marker->buffer_index;
        while(buff_ind < end_marker->buffer_index)
        {
            handle->read_buffers[buff_ind].read_element->ref_count--;
            if(handle->read_buffers[buff_ind].read_element->ref_count == 0)
            {
                free(handle->read_buffers[buff_ind].read_element->buffer);
                free(handle->read_buffers[buff_ind].read_element);
            }
        }

        move_length = begin_marker->buffer_index - 
            end_marker->buffer_index + 1;

        if(move_length > 0)
        {
            handle->read_buffers = realloc(
                handle->read_buffers,
                sizeof(globus_l_read_buffer_t) *
                (move_length + handle->buffer_count));
        }

        memmove(handle->read_buffers + begin_marker->buffer_index + 2,
                handle->read_buffers + end_marker->buffer_index,
                (handle->buffer_count - end_marker->buffer_index) *
                sizeof(globus_l_read_buffer_t));

        handle->buffer_count += move_length;
        end_marker->buffer_index += move_length;
        handle->read_buffers[end_marker->buffer_index].
            read_element->ref_count++;
        
        handle->read_buffers[end_marker->buffer_index].start_pos +=
            end_marker->offset;
        handle->read_buffers[end_marker->buffer_index].length -= 
            end_marker->offset;
        end_marker->offset = 0;

        handle->read_buffers[begin_marker->buffer_index].length = 
            begin_marker->offset + 1;

        end_marker->prev = begin_marker->prev;
        
        begin_marker = begin_marker->prev;
        while(begin_marker->next && begin_marker->next != end_marker)
        {
            globus_l_buffer_read_marker_t * tmp_marker;

            tmp_marker = begin_marker->next;
            begin_marker->next = begin_marker->next->next;
            tmp_marker = globus_hashtable_remove(
                &handle->buffer_map, tmp_marker->name);
            free(tmp_marker->name);
            free(tmp_marker);
        }
    }
    else
    {
        buff_ind = begin_marker->buffer_index + 1;
        while(buff_ind < handle->buffer_count)
        {
            handle->read_buffers[buff_ind].read_element->ref_count--;
            if(handle->read_buffers[buff_ind].read_element->ref_count == 0)
            {
                free(handle->read_buffers[buff_ind].read_element->buffer);
                free(handle->read_buffers[buff_ind].read_element);
            }
        }

        handle->buffer_count = begin_marker->buffer_index + 1;

        begin_marker->prev->next = NULL;
        while(begin_marker)
        {
            globus_l_buffer_read_marker_t * tmp_marker;

            tmp_marker = begin_marker;
            begin_marker = begin_marker->next;
            tmp_marker = globus_hashtable_remove(
                &handle->buffer_map, begin_marker->name);
            free(tmp_marker->name);
            free(tmp_marker);
        }
    }
}

/*
 * write interface function
 */
static
globus_result_t
globus_l_xio_buffer_write(
    void *                              driver_specific_handle,
    const globus_xio_iovec_t *          iovec,
    int                                 iovec_count,
    globus_xio_operation_t              op)
{
    int                                 wait_for;
    int                                 jj;
    int                                 append_length = 0;
    int                                 grow_length = 0;
    globus_l_buffer_handle_t *          handle = NULL;
    globus_xio_iovec_t *                append_iovec;
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusXIOName(globus_l_xio_buffer_write);
    GlobusXIOBufferDebugEnter();

    if(!driver_specific_handle)
    {
        result = GlobusXIOErrorParameter("driver_specific_handle");
        goto error;
    }

    handle = (globus_l_buffer_handle_t *) driver_specific_handle;
    
    wait_for = globus_xio_operation_get_wait_for(op);

    if(iovec_count < 1)
    {
        if(wait_for > 0)
        {
            result = GlobusXIOErrorParameter("iovec_count");
        }
        goto error;
    }

    if(handle->io_mode != GLOBUS_L_XIO_BUFFER_WRITING_MODE)
    {
        globus_l_xio_buffer_handle_clear(handle);
        handle->io_mode = GLOBUS_L_XIO_BUFFER_WRITING_MODE;
    }

    if(handle->mode == GLOBUS_L_XIO_BUFFERING_DISABLED)
    {
        result = globus_xio_driver_pass_write(
            op, (globus_xio_iovec_t *)iovec, iovec_count, wait_for,
            NULL, NULL);
        GlobusXIOBufferDebugExit();
        return result;
    }
    
    if(!handle->write_buffers)
    {
        globus_l_buffer_write_marker_t *       marker;

        handle->write_buffers = globus_malloc(sizeof(globus_xio_iovec_t));
        if(!handle->write_buffers)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            GlobusXIOBufferDebugExit();
            return result;
        }
        handle->write_buffers[0].iov_base = NULL;
        handle->write_buffers[0].iov_len = 0;

        marker = globus_malloc(sizeof(globus_l_buffer_write_marker_t));
        if(!marker)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            GlobusXIOBufferDebugExit();
            return result;
        }
        marker->name = strdup("DEFAULT");
        marker->index = 0;
        globus_hashtable_insert(&handle->buffer_map, marker->name, marker);
        handle->buffer_count++;
        handle->current_write_buffer = 0;
    }

    append_iovec = &handle->write_buffers[handle->current_write_buffer];

    for(jj = 0; jj < iovec_count; ++jj)
    {
        append_length += iovec[jj].iov_len;
    }

    if(append_iovec->iov_len + append_length > 0)
    {
        append_iovec->iov_base = realloc(
            append_iovec->iov_base, append_iovec->iov_len + append_length);
        if(!append_iovec->iov_base)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            GlobusXIOBufferDebugExit();
            return result;
        }
        
        grow_length = append_iovec->iov_len;
        for(jj = 0; jj < iovec_count; ++jj)
        {
            memcpy((char *)(append_iovec->iov_base) + grow_length, 
                   iovec[jj].iov_base, iovec[jj].iov_len);
            grow_length += iovec[jj].iov_len;
        }
        
        append_iovec->iov_len += append_length;
        handle->buffers_length += append_length;        
    }

    globus_xio_driver_finished_write(
        op, GLOBUS_SUCCESS, append_length);

    GlobusXIOBufferDebugExit();
    return result;

 error:

    GlobusXIOBufferDebugExitWithError();
    return result;
}

static
globus_result_t
globus_l_xio_buffer_cntl(
    void *                              driver_specific_handle,
    int                                 cmd,
    va_list                             ap)
{
    globus_l_buffer_handle_t *          handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_buffer_write_marker_t *    write_marker;
    globus_l_buffer_write_marker_t *    begin_write_marker = NULL;
    globus_l_buffer_write_marker_t *    end_write_marker = NULL;
    globus_l_buffer_read_marker_t *     begin_read_marker = NULL;
    globus_l_buffer_read_marker_t *     end_read_marker = NULL;
    globus_l_buffer_read_marker_t *     read_marker = NULL;

    char *                              tag_name;
    char **                             out_buffer;
    size_t *                            out_length;
    char *                              begin_tag_name;
    char *                              end_tag_name;
    char *                              in_buffer;
    size_t                              in_length;
    int                                 jj;
    globus_xio_handle_t                 xio_handle;
    globus_xio_iovec_callback_t         callback;
    void *                              args;
    int *                               buffer_length;
    int                                 index, bufind, buflen;
    
    GlobusXIOName(globus_l_xio_buffer_cntl);
    GlobusXIOBufferDebugEnter();

    if(!driver_specific_handle)
    {
        GlobusXIOBufferDebugExitWithError();
        return GlobusXIOErrorParameter("driver_specific_handle");
    }

    handle = (globus_l_buffer_handle_t *) driver_specific_handle;

    switch(cmd)
    {
    case GLOBUS_XIO_BUFFER_GET_LENGTH:
        buffer_length = va_arg(ap, int *);
        *buffer_length = handle->buffers_length;
        break;

    case GLOBUS_XIO_BUFFER_CLEAR:

        handle->mode = GLOBUS_L_XIO_BUFFERING_DISABLED;
        globus_l_xio_buffer_handle_clear(handle);
        break;

    case GLOBUS_XIO_BUFFER_ENABLE_READ_BUFFERING:

        handle->io_mode = GLOBUS_L_XIO_BUFFER_READING_MODE;
        handle->mode = GLOBUS_L_XIO_BUFFERING_ENABLED;
        break;

    case GLOBUS_XIO_BUFFER_ENABLE_WRITE_BUFFERING:

        handle->io_mode = GLOBUS_L_XIO_BUFFER_WRITING_MODE;
        handle->mode = GLOBUS_L_XIO_BUFFERING_ENABLED;
        break;

    case GLOBUS_XIO_BUFFER_DISABLE_BUFFERING:

        handle->mode = GLOBUS_L_XIO_BUFFERING_DISABLED;
        break;

    case GLOBUS_XIO_BUFFER_DONE_BUFFERING:
        
        handle->mode = GLOBUS_L_XIO_BUFFERING_DONE;
        break;

    case GLOBUS_XIO_BUFFER_FLUSH:

        xio_handle = va_arg(ap, globus_xio_handle_t);
        callback = va_arg(ap, globus_xio_iovec_callback_t);
        args = va_arg(ap, void *);
        handle->mode = GLOBUS_L_XIO_BUFFERING_DISABLED;
        result = globus_xio_register_writev(
            xio_handle,
            handle->write_buffers,
            handle->buffer_count,
            handle->buffers_length,
            NULL,
            callback,
            args);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusXIOErrorBufferFlushFailed(result);
            goto flush_failed;
        }

        break;

    case GLOBUS_XIO_BUFFER_SET_MARKER:

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            if(handle->buffer_count == 0)
            {
                handle->current_write_buffer = -1;
            }

            handle->write_buffers = realloc(
                handle->write_buffers, 
                sizeof(globus_xio_iovec_t) * (handle->buffer_count + 1));
            if((handle->current_write_buffer + 1) < handle->buffer_count)
            {
                memmove(handle->write_buffers + 
                        handle->current_write_buffer + 2,
                        handle->write_buffers +
                        handle->current_write_buffer + 1,
                        sizeof(globus_xio_iovec_t) * 
                        (handle->buffer_count - 
                         (handle->current_write_buffer + 1)));
                    
                write_marker = globus_hashtable_first(
                    &handle->buffer_map);
                while(write_marker)
                {
                    if(write_marker->index > handle->current_write_buffer)
                    {
                        write_marker->index++;
                    }
                        
                    write_marker = globus_hashtable_next(
                        &handle->buffer_map);
                }
            }

            handle->current_write_buffer++;
            handle->write_buffers[handle->current_write_buffer].iov_base
            = NULL;
            handle->write_buffers[handle->current_write_buffer].iov_len 
            = 0;

            handle->buffer_count++;

            write_marker = globus_malloc(
                sizeof(globus_l_buffer_write_marker_t));
            if(!write_marker)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                GlobusXIOBufferDebugExit();
                return result;
            }
            write_marker->name = strdup(va_arg(ap, char *));
            write_marker->index = handle->current_write_buffer;
            globus_hashtable_insert(&handle->buffer_map, 
                                    write_marker->name, 
                                    write_marker);
        }
        else
        {
	    tag_name = va_arg(ap, char *);

	    if(globus_hashtable_lookup(&handle->buffer_map, tag_name))
	    {
		break;
	    }
	    
            read_marker = globus_malloc(sizeof(globus_l_buffer_read_marker_t));
            if(!read_marker)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                GlobusXIOBufferDebugExit();
                return result;
            }

            read_marker->name = strdup(tag_name);
            read_marker->offset = handle->traverse_position;
            read_marker->buffer_index = handle->traverse_buffer_index;
            globus_hashtable_insert(
                &handle->buffer_map, read_marker->name, read_marker);
            if(handle->current_read_marker)
            {
                handle->current_read_marker->next = read_marker;
            }

            read_marker->prev = handle->current_read_marker;
            handle->current_read_marker = read_marker;
        }

        break;

    case GLOBUS_XIO_BUFFER_SET_READ_MARKER_AT_INDEX:

        tag_name = va_arg(ap, char *);
	if(globus_hashtable_lookup(&handle->buffer_map, tag_name))
	{
	    break;
	}
	
        read_marker = globus_malloc(sizeof(globus_l_buffer_read_marker_t));
        if(!read_marker)
        {
            result = GlobusXIOBufferErrorOutOfMemory;
            goto failed_set_marker;
        }
        read_marker->name = strdup(tag_name);
        index = va_arg(ap, int);

        if(index > handle->buffers_length)
        {
            result = GlobusXIOErrorBufferSetMarkerFailed(read_marker->name);
            goto failed_set_marker;
        }

        buflen = 0;
        for(bufind = 0; bufind < handle->buffer_count; ++bufind)
        {
            if(index < (buflen + handle->read_buffers[bufind].length))
            {
                read_marker->offset = index - buflen;
                read_marker->buffer_index = bufind;
                globus_hashtable_insert(
                    &handle->buffer_map, read_marker->name, read_marker);
                if(handle->current_read_marker)
                {
                    handle->current_read_marker->next = read_marker;
                }
                    
                read_marker->prev = handle->current_read_marker;
                handle->current_read_marker = read_marker;
                break;
            }
            else
            {
                buflen += handle->read_buffers[bufind].length;
            }
        }

        break;

    case GLOBUS_XIO_BUFFER_GET_MARKED_BUFFER:

        tag_name = va_arg(ap, char *);
        out_buffer = va_arg(ap, char **);
        out_length = va_arg(ap, size_t *);
        
        *out_buffer = NULL;
        *out_length = 0;

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {

            write_marker = globus_hashtable_lookup(
                &handle->buffer_map, tag_name);
            if(!write_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }

            if(handle->write_buffers[write_marker->index].iov_len)
            {
                *out_buffer = globus_malloc(
                    handle->write_buffers[write_marker->index].iov_len);
                memcpy(*out_buffer, 
                       handle->write_buffers[write_marker->index].iov_base,
                       handle->write_buffers[write_marker->index].iov_len);
                
                *out_length = 
                  handle->write_buffers[write_marker->index].iov_len;
            }
        }
        else
        {
            if(!handle->read_buffers)
            {
                result = GlobusXIOErrorBufferEmpty();
                goto buffer_empty;
            }

            read_marker = globus_hashtable_lookup(
                &handle->buffer_map, tag_name);
            if(!read_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }

            
            result = globus_l_xio_buffer_get_marked_buffers(
                handle, read_marker, read_marker->next,
                out_buffer, out_length);
	    if(result != GLOBUS_SUCCESS)
	    {
		goto error;
	    }
        }
        
        break;
        
    case GLOBUS_XIO_BUFFER_GET_MARKED_BUFFERS:

        begin_tag_name = va_arg(ap, char *);
        end_tag_name = va_arg(ap, char *);
        out_buffer = va_arg(ap, char **);
        out_length = va_arg(ap, size_t *);

        *out_buffer = NULL;
        *out_length = 0;

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            int                         begin_index = 0;
            int                         end_index = handle->buffer_count;

            if(begin_tag_name)
            {
                begin_write_marker = globus_hashtable_lookup(
                    &handle->buffer_map, begin_tag_name);
                if(begin_write_marker)
                {
                    begin_index = begin_write_marker->index;
                }
            }
            
            if(end_tag_name)
            {
                end_write_marker = globus_hashtable_lookup(
                    &handle->buffer_map, end_tag_name);
                if(end_write_marker)
                {
                    end_index = end_write_marker->index;
                }
            }
            
            for(jj = begin_index; 
                jj < end_index; ++jj)
            {
                *out_buffer = realloc(
                    *out_buffer,
                    handle->write_buffers[jj].iov_len + *out_length);
                memcpy(*out_buffer + *out_length,
                       handle->write_buffers[jj].iov_base,
                       handle->write_buffers[jj].iov_len);
                *out_length += handle->write_buffers[jj].iov_len;
            }
        }
        else
        {
            if(!handle->read_buffers)
            {
                result = GlobusXIOErrorBufferEmpty();
                goto buffer_empty;
            }

            if(begin_tag_name)
            {
                begin_read_marker = globus_hashtable_lookup(
                    &handle->buffer_map, begin_tag_name);
            }

            if(end_tag_name)
            {
                end_read_marker = globus_hashtable_lookup(
                    &handle->buffer_map, end_tag_name);
            }

            result = globus_l_xio_buffer_get_marked_buffers(
                handle, begin_read_marker, end_read_marker,
                out_buffer, out_length);
	    if(result != GLOBUS_SUCCESS)
	    {
		goto error;
	    }
        }

        break;
        
    case GLOBUS_XIO_BUFFER_REPLACE_MARKED_BUFFER:

        tag_name = va_arg(ap, char *);
        in_buffer = va_arg(ap, char *);
        in_length = va_arg(ap, size_t);

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            write_marker = globus_hashtable_lookup(
                &handle->buffer_map, tag_name);
            if(!write_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }
            
            free(handle->write_buffers[write_marker->index].iov_base);
            handle->buffers_length -= 
                handle->write_buffers[write_marker->index].iov_len;
            handle->buffers_length += in_length;
            handle->write_buffers[write_marker->index].iov_base = 
            malloc(in_length);
            memcpy(handle->write_buffers[write_marker->index].iov_base, 
                   in_buffer, in_length);
            handle->write_buffers[write_marker->index].iov_len = in_length;
        }
        else
        {
            if(!handle->read_buffers)
            {
                result = GlobusXIOErrorBufferEmpty();
                goto buffer_empty;
            }

            read_marker = globus_hashtable_lookup(
                &handle->buffer_map, tag_name);
            if(!read_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }

            result = globus_l_xio_buffer_replace_marked_buffers(
                handle, read_marker, read_marker->next, in_buffer, in_length);
            if(result != GLOBUS_SUCCESS)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                goto tag_not_found;
            }
        }
                
        break;

    case GLOBUS_XIO_BUFFER_REPLACE_MARKED_BUFFERS:

        begin_tag_name = va_arg(ap, char *);
        end_tag_name = va_arg(ap, char *);
        in_buffer = va_arg(ap, char *);
        in_length = va_arg(ap, size_t);

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            begin_write_marker = globus_hashtable_lookup(
                &handle->buffer_map, begin_tag_name);
            if(!begin_write_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(begin_tag_name);
                goto tag_not_found;
            }

            end_write_marker = globus_hashtable_lookup(
                &handle->buffer_map, end_tag_name);
            if(!end_write_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(end_tag_name);
                goto tag_not_found;
            }
            
            for(jj = begin_write_marker->index; 
                jj < end_write_marker->index; ++jj)
            {
                free(handle->write_buffers[jj].iov_base);
                handle->buffers_length -= handle->write_buffers[jj].iov_len;
                handle->write_buffers[jj].iov_len = 0;
            }

            if(begin_write_marker->index == end_write_marker->index)
            {
                handle->write_buffers = realloc(
                    handle->write_buffers,
                    (handle->buffer_count + 1) * sizeof(globus_xio_iovec_t));
                memmove(handle->write_buffers + end_write_marker->index + 1,
                        handle->write_buffers + end_write_marker->index,
                        sizeof(globus_xio_iovec_t) * 
                        (handle->buffer_count - end_write_marker->index));
                handle->buffer_count++;
            }

            handle->buffers_length += in_length;
            handle->write_buffers[begin_write_marker->index].iov_base = 
            malloc(in_length);
            memcpy(handle->write_buffers[begin_write_marker->index].iov_base,
                   in_buffer, in_length);
            handle->write_buffers[begin_write_marker->index].iov_len = 
                in_length;
        }
        else
        {
            if(!handle->read_buffers)
            {
                result = GlobusXIOErrorBufferEmpty();
                goto buffer_empty;
            }

            begin_read_marker = globus_hashtable_lookup(
                &handle->buffer_map, begin_tag_name);
            if(!begin_read_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(begin_tag_name);
                goto tag_not_found;
            }

            end_read_marker = globus_hashtable_lookup(
                &handle->buffer_map, end_tag_name);
            if(!end_read_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(end_tag_name);
                goto tag_not_found;
            }

            globus_l_xio_buffer_replace_marked_buffers(
                handle, begin_read_marker, end_read_marker, 
                in_buffer, in_length);
        }
                
        break;
            
    case GLOBUS_XIO_BUFFER_REMOVE_MARKED_BUFFER:

        tag_name = va_arg(ap, char *);

        if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE)
        {
            write_marker = globus_hashtable_remove(
                &handle->buffer_map, tag_name);
            if(!write_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }

            free(handle->write_buffers[write_marker->index].iov_base);
            handle->write_buffers[write_marker->index].iov_base = NULL;
            handle->buffers_length -= 
                handle->write_buffers[write_marker->index].iov_len;
            handle->write_buffers[write_marker->index].iov_len = 0;

            free(write_marker->name);
            free(write_marker);
        }
        else
        {
            read_marker = globus_hashtable_lookup(
                &handle->buffer_map, tag_name);
            if(!read_marker)
            {
                result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
                goto tag_not_found;
            }

            globus_l_xio_buffer_remove_marked_buffers(
                handle, read_marker, read_marker->next);
        }
                
        break;

    case GLOBUS_XIO_BUFFER_RESET_READ_POSITION:

        handle->traverse_position = 0;
        handle->traverse_buffer_index = 0;
        break;
        
    case GLOBUS_XIO_BUFFER_SET_READ_POSITION_TO_MARKER:
      
        tag_name = va_arg(ap, char *);
        
        read_marker = globus_hashtable_lookup(&handle->buffer_map, tag_name);
        if(!read_marker)
        {
            result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
            goto tag_not_found;
        }

        handle->traverse_position = read_marker->offset;
        handle->traverse_buffer_index = read_marker->buffer_index;
        break;

    case GLOBUS_XIO_BUFFER_SET_WRITE_POSITION_TO_MARKER:

        tag_name = va_arg(ap, char *);
        
        write_marker = globus_hashtable_lookup(&handle->buffer_map, tag_name);
        if(!write_marker)
        {
            result = GlobusXIOErrorBufferMarkerNotFound(tag_name);
            goto tag_not_found;
        }

        if(handle->write_buffers[write_marker->index].iov_base != NULL)
        {
            handle->write_buffers = realloc(
                handle->write_buffers,
                (handle->buffer_count + 1) * sizeof(globus_xio_iovec_t));
            memmove(handle->write_buffers + write_marker->index + 1,
                    handle->write_buffers + write_marker->index,
                    sizeof(globus_xio_iovec_t) * 
                    (handle->buffer_count - write_marker->index));
            handle->write_buffers[write_marker->index].iov_base = NULL;
            handle->write_buffers[write_marker->index].iov_len = 0;
            handle->buffer_count++;

            handle->current_write_buffer = write_marker->index;

            write_marker = globus_hashtable_first(&handle->buffer_map);
            while(write_marker)
            {
                if(write_marker->index > handle->current_write_buffer)
                {
                    write_marker->index++;
                }
                
                write_marker = globus_hashtable_next(&handle->buffer_map);
            }
        }
        else
        {
            handle->current_write_buffer = write_marker->index;
        }

        break;

    default:

        result = GlobusXIOErrorInvalidCommand(cmd);
        GlobusXIOBufferDebugExitWithError();
        return result;
    }

    GlobusXIOBufferDebugExit();
    return result;

 tag_not_found:

 buffer_empty:

 flush_failed:

 failed_set_marker:

 error:
    
    GlobusXIOBufferDebugExitWithError();
    return result;
}

static
void
globus_l_xio_buffer_read_callback(
    globus_xio_operation_t              op,
    globus_result_t                     result,
    globus_size_t                       nbytes,
    void *                              user_arg)
{
    globus_l_buffer_handle_t *          handle;
    int                                 i = 0;
    int                                 read_len = 0;
    int                                 cpylen = 0;
    int                                 added_buffer = 0;
    GlobusXIOName(globus_l_xio_buffer_read_callback);
    GlobusXIOBufferDebugInternalEnter();

    handle = (globus_l_buffer_handle_t *) user_arg;

    if(handle->mode == GLOBUS_L_XIO_BUFFERING_ENABLED)
    {
        if(globus_xio_error_is_eof(result))
        {
            result = GlobusXIOErrorBufferEOF();
        }

        if(!handle->read_buffers)
        {
            handle->read_buffers = globus_malloc(
                sizeof(globus_l_read_buffer_t) * GROW_SIZE);
            if(!handle->read_buffers)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                globus_xio_driver_finished_read(op, result, nbytes);
                goto error;
            }

            memset(handle->read_buffers, 0, 
                   GROW_SIZE * sizeof(globus_l_read_buffer_t));
            handle->read_buffers_available = GROW_SIZE;
        }
        else if(handle->read_buffers_available == 0)
        {
            handle->read_buffers = realloc(
                handle->read_buffers,
                sizeof(globus_l_read_buffer_t) * 
                (handle->buffer_count + GROW_SIZE));
            if(!handle->read_buffers)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                globus_xio_driver_finished_read(op, result, nbytes);
                goto error;
            }
            memset(handle->read_buffers + 
                   handle->buffer_count,
                   0, GROW_SIZE * sizeof(globus_l_read_buffer_t));
            handle->read_buffers_available += GROW_SIZE;
        }

        if(nbytes > 0)
        {
            handle->read_buffers[handle->buffer_count].read_element = 
                malloc(sizeof(globus_l_read_buffer_element_t));
            if(!handle->read_buffers[handle->buffer_count].read_element)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                globus_xio_driver_finished_read(op, result, nbytes);
                goto error;
            }
            memset(handle->read_buffers[handle->buffer_count].read_element,
                   0, sizeof(globus_l_read_buffer_element_t));
    
            handle->read_buffers[handle->buffer_count].read_element->buffer = 
                malloc(nbytes);
            if(!handle->read_buffers[handle->buffer_count].read_element->buffer)
            {
                result = GlobusXIOBufferErrorOutOfMemory;
                globus_xio_driver_finished_read(op, result, nbytes);
                goto error;
            }
            handle->read_buffers[handle->buffer_count].
                read_element->ref_count++;
        }
        
        cpylen = nbytes;
        for(i = 0; read_len < nbytes && i < handle->user_iovec_count; ++i)
        {
            cpylen = nbytes - read_len;
            cpylen = (cpylen < handle->user_iovec[i].iov_len) ? cpylen :
            handle->user_iovec[i].iov_len;
            added_buffer = 1;

            if(handle->user_iovec[i].iov_len > 0)
            {
                memcpy(
                    handle->read_buffers[handle->buffer_count].
                    read_element->buffer +
                    handle->read_buffers[handle->buffer_count].length,
                    handle->user_iovec[i].iov_base,
                    cpylen);
                handle->read_buffers[handle->buffer_count].length += cpylen;
                read_len += cpylen;
            }
        }
        
        if(added_buffer)
        {
            handle->buffer_count++;
            handle->buffers_length += nbytes;
            handle->read_buffers_available--;
        }

        globus_xio_driver_finished_read(op, result, nbytes);
    }        
    else
    {
        goto error;
    }

 error:

    GlobusXIOBufferDebugInternalExit();
}
        
static
globus_result_t
globus_l_xio_buffer_read(
    void *                              driver_specific_handle,
    const globus_xio_iovec_t *          iovec,
    int                                 iovec_count,
    globus_xio_operation_t              op)
{
    globus_l_buffer_handle_t *          handle;
    globus_result_t                     result = GLOBUS_SUCCESS;
    size_t                              wait_for;
    int                                 i = 0;
    int                                 copy_length = 0;
    int                                 total_copy = 0;

    GlobusXIOName(globus_l_xio_buffer_read);
    GlobusXIOBufferDebugEnter();

    if(!driver_specific_handle)
    {
        result = GlobusXIOErrorParameter("driver_specific_handle");
        goto error;
    }

    handle = (globus_l_buffer_handle_t *) driver_specific_handle;

    wait_for = globus_xio_operation_get_wait_for(op);

    if(handle->io_mode != GLOBUS_L_XIO_BUFFER_READING_MODE)
    {
        globus_l_xio_buffer_handle_clear(handle);
        handle->io_mode = GLOBUS_L_XIO_BUFFER_READING_MODE;
    }

    if(handle->mode == GLOBUS_L_XIO_BUFFERING_ENABLED)
    {
        handle->user_iovec = (globus_xio_iovec_t *) iovec;
        handle->user_iovec_count = iovec_count;
        
        result = globus_xio_driver_pass_read(
            op, (globus_xio_iovec_t *) iovec, iovec_count, wait_for, 
            globus_l_xio_buffer_read_callback, handle);
        if(result != GLOBUS_SUCCESS)
        {
            goto error;
        }
    }
    else if(handle->mode == GLOBUS_L_XIO_BUFFERING_DISABLED)
    {
        result = globus_xio_driver_pass_read(
            op, (globus_xio_iovec_t *) iovec, iovec_count, wait_for,
            NULL, NULL);
        if(result != GLOBUS_SUCCESS)
        {
            goto error;
        }
    }
    else if(handle->mode == GLOBUS_L_XIO_BUFFERING_DONE)
    {
        if(!handle->read_buffers)
        {
            goto error;
        }

        for(i = 0; i < iovec_count; ++i)
        {
            if(iovec[i].iov_len > 0)
            {
                int                     buff_ind; 
                int                     cbuff_unread_length;
                int                     current_iovec_offset = 0;

                /* For each non-empty iovec, we'll copy data up to the
                 * iovec limit from the buffers.
                 */
                for (buff_ind = handle->traverse_buffer_index;
                     buff_ind < handle->buffer_count;
                     buff_ind++)
                {
                    cbuff_unread_length = 
                        handle->read_buffers[buff_ind].length - 
                        handle->traverse_position;

                    /* Copy at smaller of remaining space in the iovec
                     * and cbuff_unread_length
                     */
                    if(iovec[i].iov_len - current_iovec_offset
                            > cbuff_unread_length)
                    {
                        copy_length = cbuff_unread_length;
                    }
                    else
                    {
                        copy_length = iovec[i].iov_len - current_iovec_offset;
                    }

                    memcpy((char *) iovec[i].iov_base + current_iovec_offset,
                        handle->read_buffers[buff_ind].read_element->buffer +
                        handle->read_buffers[buff_ind].start_pos +
                        handle->traverse_position,
                        copy_length);

                    /* Track how much is copied, offset into our destination,
                     * and traverse position
                     */
                    total_copy += copy_length;
                    current_iovec_offset += copy_length;
                    handle->traverse_position += copy_length;

                    /* If we copied to the end of the current buffer, we'll advance to the
                     * next one (resetting the traverse position), unless this
                     * is the last buffer.
                     */
                    if (handle->traverse_position == handle->read_buffers[buff_ind].length)
                    {
                        if (handle->traverse_buffer_index + 1 < handle->buffer_count)
                        {
                            handle->traverse_position = 0;
                            handle->traverse_buffer_index++;
                        }
                        else
                        {
                            /* Nothing more in our buffers. Trigger EOF Condition */
                            result = GlobusXIOErrorBufferEOF();
                            break;
                        }
                    }

                    if (current_iovec_offset == iovec[i].iov_len)
                    {
                        /* Can't put any more data into the iovec, so we break
                         */
                        break;
                    }
                }
            }
        }

        globus_xio_driver_finished_read(op, result, total_copy);
        result = GLOBUS_SUCCESS;
    }

 error:

    if(result != GLOBUS_SUCCESS)
    {
        GlobusXIOBufferDebugExitWithError();
        return result;
    }

    GlobusXIOBufferDebugExit();
    return result;
}

static
void
globus_l_xio_buffer_open_callback(
    globus_xio_operation_t              op,
    globus_result_t                     result,
    void *                              user_arg)
{
    globus_l_buffer_handle_t *          handle;
    
    GlobusXIOName(globus_l_xio_buffer_open_callback);
    GlobusXIOBufferDebugEnter();

    handle = (globus_l_buffer_handle_t *) user_arg;

    if(result != GLOBUS_SUCCESS)
    {
        globus_l_buffer_handle_destroy(handle);
        handle = NULL;
    }

    globus_xio_driver_finished_open(handle, op, result);

    GlobusXIOBufferDebugExit();
}

    
    
static
globus_result_t
globus_l_xio_buffer_open(
    const globus_xio_contact_t *        contact_info,
    void *                              driver_link,
    void *                              driver_attr,
    globus_xio_operation_t              op)
{
    globus_result_t                     result;
    globus_l_buffer_handle_t *          handle;
    int                                 rc;

    GlobusXIOName(globus_l_xio_buffer_open);
    GlobusXIOBufferDebugEnter();

    handle = malloc(sizeof(globus_l_buffer_handle_t));
    if(!handle)
    {
        result = GlobusXIOErrorMemory("handle");
        goto error;
    }
    memset(handle, 0, sizeof(globus_l_buffer_handle_t));
    
    rc = globus_hashtable_init(&handle->buffer_map, 10,
                          globus_hashtable_string_hash,
                          globus_hashtable_string_keyeq);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusXIOErrorMemory("handle.buffer_map");
        goto free_handle_error;
    }

    result = globus_xio_driver_pass_open(
        op, contact_info, 
        globus_l_xio_buffer_open_callback, handle);
    if (result != GLOBUS_SUCCESS)
    {
        goto destroy_handle_error;
    }

    GlobusXIOBufferDebugExit();
    return GLOBUS_SUCCESS;

destroy_handle_error:
    globus_hashtable_destroy(&handle->buffer_map);
free_handle_error:
    free(handle);
 error:

    GlobusXIOBufferDebugExitWithError();
    return result;
}
    
static
globus_result_t
globus_l_xio_buffer_close(
    void *                              driver_specific_handle,
    void *                              attr,
    globus_xio_operation_t              op)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_buffer_handle_t *          handle = NULL;

    GlobusXIOName(globus_l_xio_buffer_close);
    GlobusXIOBufferDebugEnter();
    
    if(!driver_specific_handle)
    {
        GlobusXIOBufferDebugExitWithError();
        return GlobusXIOErrorParameter("driver_specific_handle");
    }

    handle = (globus_l_buffer_handle_t *) driver_specific_handle;

    if(handle->io_mode == GLOBUS_L_XIO_BUFFER_WRITING_MODE &&
       handle->mode == GLOBUS_L_XIO_BUFFERING_ENABLED)
    {
        handle->io_mode = GLOBUS_L_XIO_BUFFERING_DONE;
        result = globus_xio_driver_pass_write(
            op, handle->write_buffers, handle->buffer_count, 
            handle->buffers_length, globus_l_xio_buffer_close_callback, 
            handle);
        if(result != GLOBUS_SUCCESS)
        {
            goto error;
        }

        goto exit;
    }

 error:

    if(handle)
    {
        globus_l_buffer_handle_destroy(handle);
    }

    result = globus_xio_driver_pass_close(
        op, 
        NULL, NULL);

 exit:

    GlobusXIOBufferDebugExit();
    return result;
}

globus_result_t
globus_l_xio_buffer_accept(
    void *                              driver_server,
    globus_xio_operation_t              accept_op)
{
    return globus_xio_driver_pass_accept(accept_op,
            NULL, NULL);
}

/*
 * Driver load function
 */
static globus_result_t
globus_l_xio_buffer_init(
    globus_xio_driver_t *               out_driver)
{
    globus_xio_driver_t                 driver;
    globus_result_t                     result;
    GlobusXIOName(globus_l_xio_buffer_init);
    GlobusXIOBufferDebugEnter();

    result = globus_xio_driver_init(&driver, "buffer", NULL);
    if(result != GLOBUS_SUCCESS)
    {
        GlobusXIOBufferDebugExitWithError();
        return result;
    }

    globus_xio_driver_set_transform(
        driver,
        globus_l_xio_buffer_open,
        globus_l_xio_buffer_close,
        globus_l_xio_buffer_read,
        globus_l_xio_buffer_write,
        globus_l_xio_buffer_cntl,
        NULL);

    globus_xio_driver_set_server(
        driver,
        NULL,
        globus_l_xio_buffer_accept,
        NULL,
        NULL,
        NULL,
        NULL);
    
    *out_driver = driver;

    GlobusXIOBufferDebugExit();
    return GLOBUS_SUCCESS;
}

/*
 * Driver unload function
 */
static void
globus_l_xio_buffer_destroy(
    globus_xio_driver_t                 driver)
{
    GlobusXIOName(globus_l_xio_buffer_destroy);
    GlobusXIOBufferDebugEnter();
    globus_xio_driver_destroy(driver);
    GlobusXIOBufferDebugExit();
    return;
}


/*
 * Driver definition
 */
GlobusXIODefineDriver(
    buffer,
    globus_l_xio_buffer_init,
    globus_l_xio_buffer_destroy);

/*
 * Module activation
 */
static
int
globus_l_xio_buffer_activate(void)
{
    int                                 rc;

    GlobusXIOName(globus_l_xio_buffer_activate);
    GlobusDebugInit(GLOBUS_XIO_BUFFER, TRACE INTERNAL_TRACE MARKERS BUFFERS);
    GlobusXIOBufferDebugEnter();
    rc = globus_module_activate(GLOBUS_COMMON_MODULE);
    if(rc == GLOBUS_SUCCESS)
    {
        GlobusXIORegisterDriver(buffer);
    }
    GlobusXIOBufferDebugExit();
    return rc;
}

/*
 * Module deactivation
 */
static
int
globus_l_xio_buffer_deactivate(void)
{
    int                                 rc;
    GlobusXIOName(globus_l_xio_buffer_deactivate);
    GlobusXIOBufferDebugEnter();
    GlobusXIOUnRegisterDriver(buffer);
    rc = globus_module_deactivate(GLOBUS_COMMON_MODULE);
    GlobusXIOBufferDebugExit();
    GlobusDebugDestroy(GLOBUS_XIO_BUFFER);
    return rc;
}
