/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.security.GeneralSecurityException;
import java.security.cert.X509Certificate;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.axis.message.MessageElement;

import org.opensaml.SAMLAction;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAuthorizationDecisionQuery;
import org.opensaml.SAMLException;
import org.opensaml.SAMLSubject;
import org.w3c.dom.Element;

import org.globus.axis.util.Util;
import org.globus.cas.CASPortType;
import org.globus.cas.faults.CasFault;
import org.globus.cas.impl.CasConstants;
import org.globus.wsrf.impl.security.SecurityMessageElement;
import org.globus.cas.types.ArrayOfSAMLAuthzQueryType;
import org.globus.cas.types.GetAssertionParam;
import org.globus.cas.types.SAMLAuthzQueryType;
import org.globus.common.CoGProperties;
import org.globus.gsi.GSIConstants;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.GlobusCredentialException;
import org.globus.gsi.X509Extension;
import org.globus.gsi.X509ExtensionSet;
import org.globus.gsi.bc.BouncyCastleCertProcessingFactory;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.SecurityMessageElement;
import org.globus.wsrf.utils.FaultHelper;

/**
 * Helper class to get a proxy embedded with CAS (SAML) assertions.
 */
public class CasProxyHelper {

    boolean debug;
    String instanceURL;
    String serverIdentity;

    /**
     * @param instanceURL
     *        Used to contact CAS service.
     * <br>Host authz is used and no debug information is printed.
     */
    public CasProxyHelper(String instanceURL) throws CasClientException {
        this(instanceURL, null, false);
    }

    /**
     * @param instanceURL
     *        Used to contact CAS service.
     * @param debug
     *        trace is printed is debug is set to true.
     * <br>Host authz is used.
     */
    public CasProxyHelper(String instanceURL, boolean debug)
        throws CasClientException {
        this(instanceURL, null, debug);
    }

    /**
     * @param instanceURL
     *        Used to contact CAS service.
     * @param serverIdentity
     *        Identity authz is done.
     * <br>No debug information is printed.
     */
    public CasProxyHelper(String instanceURL, String serverIdentity)
        throws CasClientException {
        this(instanceURL, serverIdentity, false);
    }

    /**
     * @param instanceURL
     *        Used to contact CAS service.
     * @param serverIdentity
     *        Identity authz is done.
     * @param debug
     *        trace is printed is debug is set to true.
     */
    public CasProxyHelper(String instanceURL, String serverIdentity,
                          boolean debug)
        throws CasClientException {
        if ((instanceURL == null) || (instanceURL.trim().equals("")))
            throw new CasClientException("Invalid instance URL "
                                         + instanceURL);
        this.instanceURL = instanceURL;
        this.serverIdentity = serverIdentity;
        this.debug = debug;
    }

    /**
     * @param debug
     *        trace is printed if debug is set
     */
    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    /**
     * @param params
     *        Parameters for generating CAS assertion.
     * @return Path to the proxy file with embedded CAS assertions
     */
    public String getCasProxy(ClientParams params) throws CasClientException {

        String proxyFilename = params.getProxyFileName();
        // get credential
        GlobusCredential credential = getCredential(proxyFilename);

        // Set subject dn and trust anchor DN from proxy used.
        String clientDN = credential.getIdentity();
        printMessage(System.out, "Client DN is " + clientDN);
        String clientTADN = credential.getIssuer();
        printMessage(System.out, "Client cert TA DN is " + clientTADN);

        // Confirmation method
        Vector confMethods = new Vector(1);
        confMethods.add(CasConstants.X509_CONFIRMATION_METHOD);

        // SAML Subject
        SAMLSubject samlSubject = null;
        try {
            samlSubject =
                new SAMLSubject(clientDN, clientTADN, CasConstants.X509_FORMAT,
                                confMethods, null, null);
        } catch (SAMLException exp) {
            String err = "Cannot constrct SAML Subject";
            printMessage(System.err, err + exp.toString());
        }

        ResourceActionsMap[] resActions = params.getResourceActionsMap();
        // Construct SAML query
        SAMLAuthzQueryType[] samlAnyQueries =
            getSAMLAuthzQueries(samlSubject,  resActions);

        CasClientSetup clientSetup = new CasClientSetup();
        // Get casPort
        boolean trans = false;
        String securityType = params.getSecurityType();
        if (securityType == null) {
            if (instanceURL.startsWith("https")) {
                securityType = Constants.GSI_TRANSPORT;
                // register util.
                Util.registerTransport();
                trans = true;
            } else {
                securityType = Constants.GSI_SEC_MSG;
            }
        }
        String authzObject = clientSetup.getAuthzParamObject(trans);
        Object authzValue = clientSetup.getAuthzParamValue(trans,
                                                           serverIdentity);
        CASPortType casPort =
            clientSetup.getCASPort(instanceURL, securityType,
                                   params.getProtectionType(), authzObject,
                                   authzValue);

        org.globus.cas.types.SAMLAssertion assertion = null;
        try {
            ArrayOfSAMLAuthzQueryType arrayOfSamlAuthz =
                new ArrayOfSAMLAuthzQueryType(samlAnyQueries);

            GetAssertionParam getAssertion = new GetAssertionParam();
            getAssertion.setLifetime(params.getAssertionLifetime());
            getAssertion.setSamlAuthzQuery(arrayOfSamlAuthz);

            assertion = casPort.getAssertion(getAssertion);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc =
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            throw new CasClientException((new FaultHelper(casFault))
                                         .getDescriptionAsString());
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            throw new CasClientException(exp);
        }

        // Get assertion.
        SAMLAssertion samlAssertion = parseAndVerifyAssertion(assertion);
        // Embed assertion
        return embedAssertionInCredential(samlAssertion, credential,
                                          proxyFilename,
                                          params.getCasProxyFileName(),
                                          params.getCasProxyTag());
    }

    // Embed assertion in the proxy.
    private String embedAssertionInCredential(SAMLAssertion samlAssertion,
                                              GlobusCredential credential,
                                              String proxyFilename,
                                              String casProxyFilename,
                                              String casProxyTag)
        throws CasClientException {
        if (samlAssertion == null) {
            System.out.println("Assertion is null.");
            // FIXME
            return null;
        }

        boolean critical = false;
        X509ExtensionSet extensionSet = new X509ExtensionSet();
        X509Extension extension =
            new X509Extension(CasConstants.OID, critical,
                              samlAssertion.toString().getBytes());
        extensionSet.add(extension);

        BouncyCastleCertProcessingFactory certProcessingFactory =
            BouncyCastleCertProcessingFactory.getDefault();

        // Life time is set to the old proxy's life time
        Long lifetime = new Long(credential.getTimeLeft());
        printMessage(System.out, "lifetime " + lifetime);
        GlobusCredential newCredential = null;
        try {
            newCredential =
                certProcessingFactory.createCredential(
                                      credential.getCertificateChain(),
                                      credential.getPrivateKey(), 512 ,
                                      Integer.parseInt(lifetime.toString()),
                                      GSIConstants.GSI_3_IMPERSONATION_PROXY,
                                      extensionSet, null);
        } catch (GeneralSecurityException exp) {
            String err = "Error creating new credentials\n ";
            printMessage(System.err, err + exp.toString());
        }

        if (casProxyFilename == null) {
            if (casProxyTag == null)
                casProxyTag = CasConstants.CAS_PROXY_TAG;
            if (proxyFilename == null) {
                proxyFilename = CoGProperties.getDefault().getProxyFile();
            }
            casProxyFilename = proxyFilename + "." + casProxyTag;
        }

        printMessage(System.out, "New proxy filename " + casProxyFilename);
        try {
            File casProxyFile = 
                org.globus.util.Util.createFile(casProxyFilename);
            if (!org.globus.util.Util
                .setFilePermissions(casProxyFilename, 600)) {
                System.out.println("Check permissions on proxy file, "
                                   + casProxyFilename);
            }
            
            newCredential.save(new FileOutputStream(casProxyFile));
        } catch (Exception exp) {
            String err = "Error writing out CAS credential to "
                + casProxyFilename;
            printMessage(System.err, err + "\n" + exp.toString());
        }
        
        printMessage(System.out, "New Proxy " + newCredential.toString());

        if (debug) {
            try {
                GlobusCredential newCred =
                    new GlobusCredential(casProxyFilename);
                X509Certificate array[] = newCred.getCertificateChain();
                if ((array != null) && (array.length > 1)) {
                    byte[] extensionVal =
                        X509Extension.getExtensionValue(array[0],
                                                        CasConstants.OID);
                    if (extensionVal == null)
                        System.out.println("No such OID found");
                    else {
                        String casPolicy  = new String(extensionVal);
                        System.out.println("CAS Policy " + casPolicy);
                    }
                }
                else {
                    System.err.println("Certificate array is null or "
                                       + "length < 1");
                }
            } catch (Exception exp) {
                System.err.println(exp);
            }
        }
        return casProxyFilename;
    }

    // Verify returned assertion signature.
    private SAMLAssertion
        parseAndVerifyAssertion(org.globus.cas.types.SAMLAssertion
                                assertion) throws CasClientException {
        SAMLAssertion samlAssertion = null;
        if (assertion != null) {
            org.apache.xml.security.Init.init();
            try {
                MessageElement[] msgElement = assertion.get_any();
                samlAssertion = new SAMLAssertion(msgElement[0].getAsDOM());
            } catch (Exception exp) {
                String err = "Error extracting SAML Assertion object ";
                printMessage(System.err, err + "\n" + exp.toString());
                throw new CasClientException(err, exp);
            }

            printMessage(System.out, "SAML " + samlAssertion);

            // signature needs to be verified.
            try {
                samlAssertion.verify(false);
            } catch (SAMLException samlExp) {
                String err = "Could not verify signature on assertion. \n ";
                printMessage(System.err, err + samlExp.toString());
                throw new CasClientException(err, samlExp);
            }
        }
        return samlAssertion;
    }

    // Construct SAML Authz Query
    private SAMLAuthzQueryType[]
        getSAMLAuthzQueries(SAMLSubject samlSubject,
                            ResourceActionsMap[] resActions)
        throws CasClientException {

        SAMLAuthzQueryType[] samlAnyQueries = null;

        if (resActions != null) {
            samlAnyQueries = new SAMLAuthzQueryType[resActions.length];
            for (int i=0; i<resActions.length; i++) {
                String resource = resActions[i].getResource();
                String[] actionStrs = resActions[i].getActions();
                Vector samlActions = new Vector(actionStrs.length);
                for (int j=0; j<actionStrs.length; j++) {
                    samlActions.add(getSAMLActionObject(actionStrs[j]));
                }
                SAMLAuthorizationDecisionQuery samlAuthQuery = null;
                try {
                    samlAuthQuery =
                        new SAMLAuthorizationDecisionQuery(samlSubject,
                                                           resource,
                                                           samlActions,
                                                           null);
                } catch (SAMLException samlExp) {
                    String err = "Error constucting SAMLAuthzQuery \n";
                    printMessage(System.err, err + "\n" + samlExp.toString());
                    throw new CasClientException(err, samlExp);
                }
                samlAnyQueries[i] = new SAMLAuthzQueryType();
                samlAnyQueries[i].set_any(new MessageElement[] {
                    new SecurityMessageElement((Element)samlAuthQuery.toDOM())});
            }
            return samlAnyQueries;
        } else {
            // set of all assertions
            String resourceWildcard = CasConstants.RESOURCE_WILDCARD;
            Vector wildCardActions = new Vector(1);
            SAMLAction wildCardAction = null;
            try {
                wildCardAction =
                    new SAMLAction(CasConstants.ACTION_NS_WILDCARD,
                                   CasConstants.ACTION_WILDCARD);
            } catch (SAMLException samlExp) {
                String err = "Error constucting wildcard SAMLAction \n";
                printMessage(System.err, err + "\n" + samlExp.toString());
                throw new CasClientException(err, samlExp);
            }
            wildCardActions.add(wildCardAction);
            SAMLAuthorizationDecisionQuery samlAuthQuery = null;
            try {
                samlAuthQuery =
                    new SAMLAuthorizationDecisionQuery(samlSubject,
                                                       resourceWildcard,
                                                       wildCardActions,
                                                       null);
            } catch (SAMLException samlExp) {
                String err = "Error constucting SAML Authz decision query \n";
                printMessage(System.err, err + "\n" + samlExp.toString());
                throw new CasClientException(err, samlExp);
            }
            samlAnyQueries = new SAMLAuthzQueryType[1];
            samlAnyQueries[0] = new SAMLAuthzQueryType();
            samlAnyQueries[0].set_any(new MessageElement[] {
                new SecurityMessageElement((Element)samlAuthQuery.toDOM())});
        }
        return samlAnyQueries;
    }

    // Construct SAML Action
    private SAMLAction getSAMLActionObject(String actionStr)
        throws CasClientException {

        printMessage(System.out, "Action: " + actionStr);
        StringTokenizer strTok = new StringTokenizer(actionStr);
        try {
            if (strTok.countTokens() == 2) {
                return new SAMLAction(strTok.nextToken(),
                                      strTok.nextToken());
            }
        } catch (SAMLException samlExp) {
            String err = "Error constucting SAMLAction \n";
            printMessage(System.err, err + "\n" + samlExp.toString());
            throw new CasClientException(err, samlExp);
        }
        printMessage(System.err, "Bad action string: " + actionStr);
        throw new CasClientException("Bad action string: " + actionStr);
    }

    private GlobusCredential getCredential(String proxyFilename)
        throws CasClientException {

        GlobusCredential credential = null;
        try {
            if (proxyFilename == null) {
                credential = GlobusCredential.getDefaultCredential();
            } else {
                credential = new GlobusCredential(proxyFilename);
            }
        } catch (GlobusCredentialException exp) {
            String err = "Error loading default credential";
            printMessage(System.err, err + "/n" + exp.toString());
            throw new CasClientException(err, exp);
        }
        return credential;
    }

    private void printMessage(PrintStream stream, String msg) {
        if (debug) {
            stream.println(msg);
        }
    }
}
