/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.types.PolicyData;

import org.globus.cas.CASPortType;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.cas.types.PolicyDetails;
import org.globus.cas.types.PolicyResponse;

import org.globus.cas.impl.CasConstants;

import org.globus.wsrf.utils.FaultHelper;

/**
 * Command line client that can be used to grant or revoke rights
 * on CAS server
 */
public class CasRightsHandler {
    
    static String msg = 
        " Usage: cas-rights-admin [<options>] <operationType> "
        + "<operationDetails>\n"
        + " Options are -debug Runs with debug trace \n"
        + "              -c <instanceURl> Specify URL for CAS server\n"
        + "              -help Prints this message\n"
        + "              -v Prints version number\n"
        + "              -s Identity of CAS server. If not specified, will \n"
        + "                 defualt to host authorization.\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation or 'trans' for\n"
        + "                 transport security. If unspecified, if instance\n"
        + "                 url is 'https', then trasport security is used \n"
        + "                 else mechanism defaults to secure message\n"
        + "              -p protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " <operationType> : Indicates type of operation. Should be \n"
        + "                   \"grant\" or \"revoke\"\n"
        + " <operationDetails> : \n"
        + "   <userGroupName> <objectSpecDesc> <objectSpec> <actionSpecDesc>\n"
        + "    <actionSpec>\n"
        + "        <userGroupName> : Name of the user group to grant \n"
        + "                         permission for\n" 
        + "        <objectSpecDesc> : Type of object. Should be \"object\" \n"
        + "                           or \"user\" or \"userGroup\" or \n"
        + "                           \"serviceType\" or \"namespace\" or\n"
        + "                           \"trustAnchor\" \n"
        + "        <objectSpec> : Data to uniquely identify the CAS Object. \n"
        + "                       <objectNamespace> <objectName>, where \n"
        + "                       <objectNamespace> is Namespace the object\n"
        + "                       belongs to and <objectName> is name of the\n"
        + "                       object (or) nickname of the user (or) name\n"
        + "                       of user group (or) service type name (or)\n"
        + "                       nickname of the namespace or) nickname of\n"
        + "                       trust anchor respectively\n"
        + "        <actionSpecDesc> : Should be \"serviceAction\" or \n"
        + "                            \"serviceActionGroup\"\n"
        + "        <actionSpec> : <serviceTypeName> <actionName>, \n"
        + "                       where <serviceTypeName> is name of the\n"
        + "                       service type and <actionName> is name of\n"
        + "                       action (or) name of the service group\n";

    static boolean debug = false;

    public static void main(String args[]) throws Exception {
        
        if (args.length<1) {    
            System.out.println(msg);
            System.exit(0);
        }

        CasClientSetup clientSetup = new CasClientSetup(args, msg);
        String instanceURL =  clientSetup.getInstanceURL();     
        String serverIdentity = clientSetup.getServerIdentity();
        int startIndex = clientSetup.argsStartIndex;
        debug = clientSetup.debug;
        printMessage("CAS client to grant or revoke rights.");

        if (args.length < startIndex + 5) {
            System.err.println("Not enough parameters");
            System.err.println(msg);
            System.exit(-1);
        }

        String userGpName = args[startIndex+1];
        String objSpecDesc = args[startIndex+2];
        String objSpec = args[startIndex+3];
        int actionIndex = 4;
        // If objectSpecDesc is object, construct <objectNamespace>|<objectName>
        if (objSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            printMessage("objectSpecDesc is object");
            objSpec = args[startIndex+3] 
                        + CasConstants.OBJECTSPEC_DELIMITER + args[startIndex+4];
            actionIndex = 5;
        }       
        
        clientSetup.checkParameters(actionIndex + 2);
        
        String actionSpecDesc = args[startIndex + actionIndex];
        String actionSpec = args[startIndex + actionIndex + 1];
        // If actionSpecDesc is serviceAction, construct 
        // <serviceName>/<actionName>
        if (actionSpecDesc.equals(CasConstants.SERVICEACTION_SPEC)) {
            printMessage("actionSpecDesc is serviceAction");
            clientSetup.checkParameters(actionIndex + 3);
            actionSpec = args[startIndex + actionIndex + 1 ] 
                + CasConstants.SERVICEACTION_DELIMITER 
                + args[startIndex + actionIndex + 2];
        }       

        printMessage("Parameters " + userGpName + " " + objSpec + " " 
                     + objSpecDesc + " " + actionSpec + " " + actionSpecDesc);

        try {
            if (args[startIndex].equalsIgnoreCase("grant")) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, serverIdentity);
                // swapped for object and action spec and desc
                printMessage("Grant " + userGpName + " " + objSpec + " " 
                             + objSpecDesc + " " + actionSpec + " " 
                             + actionSpecDesc);

                
                PolicyDetails policyDetails = new PolicyDetails();
                policyDetails.setUserGpName(userGpName);
                policyDetails.setObjectSpec(objSpec);
                policyDetails.setObjectSpecDesc(objSpecDesc);
                policyDetails.setActionSpec(actionSpec);
                policyDetails.setActionSpecDesc(actionSpecDesc);
                PolicyResponse policyData = casPort.grant(policyDetails);
                System.out.println("PolicyId is " + ((PolicyData)policyData
                                   .getPolicy()).getPolicyId());
            }
            else if (args[startIndex].equalsIgnoreCase("revoke")) {
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, serverIdentity);
                printMessage("Revoke " + userGpName + " " + objSpec + " " 
                             + objSpecDesc + " " + actionSpec + " " 
                             + actionSpecDesc);

                PolicyDetails policyDetails = new PolicyDetails();
                policyDetails.setUserGpName(userGpName);
                policyDetails.setObjectSpec(objSpec);
                policyDetails.setObjectSpecDesc(objSpecDesc);
                policyDetails.setActionSpec(actionSpec);
                policyDetails.setActionSpecDesc(actionSpecDesc);
                casPort.revoke(policyDetails);
            } else {
                System.err.println("Erroneous operation type " 
                                   + args[startIndex]);
                System.err.println(msg);
                System.exit(-1);
            }
        } catch (NoPermissionFault noPermFault) {
            System.err.println(noPermFault.getDescription());
            System.exit(-1);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc = 
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            System.exit(-1);
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            System.exit(-1);
        }
        System.out.println("Completed successfully");
    } 

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
