/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.CASPortType;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.types.AddUser;
import org.globus.cas.types.AddTrustAnchor;
import org.globus.cas.types.CreateServiceType;
import org.globus.cas.types.CreateObject;
import org.globus.cas.types.CreateObjectNamespace;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.wsrf.utils.FaultHelper;

/**
 * Command client used to enroll trust anchor, user, namespace, service type 
 */
public class Enroll {
    
    static String msg = 
        " Usage: cas-enroll [<options>] <objectType> <userGroup> "
        + "      <objectDetails>\n"
        + " Options are  -debug Runs with debug trace \n"
        + "              -c <instanceURL> Specify URL for CAS server\n"
        + "              -help Prints this message\n"
        + "              -v Prints version number\n"
        + "              -s Identity of CAS server. If not specified, will \n"
        + "                 defualt to host authorization.\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation or 'trans' for\n"
        + "                 transport security. If unspecified, if instance\n"
        + "                 url is 'https', then trasport security is used \n"
        + "                 else mechanism defaults to secure message\n"
        + "              -p protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " <objectType> : Type of data that needs to be enrolled. Should be\n"
        + "                 \"trustAnchor\" or \"user\" or \"object\" or \n"
        + "                 \"namespace\" or \"serviceType\" \n"
        + " <userGroup> : User group name to which all rights on this object\n"
        + "                should be granted \n"
        + " <objectDetails> : \n" 
        + "    If \"trustAnchor\", <nickname> <authMethod> <authData>\n"
        + "       <nickname> : The trust anchor nickname\n"
        + "       <authMethod> : Authorization method used by the trust anchor"
        + "\n       <authData> : Authorization data \n"
        + "    If \"user\", <nickname> <subjectName> <trustAnchorNick> \n"
        + "       <nickname> : Nickname of the user \n"
        + "       <subjectName> : DN of the user\n"
        + "       <trustAnchorNick> : Nickname of the trust anchor used to \n"
        + "                           ratify user\n "
        + "    If \"object\", <objectName> <objectNamespace> \n"
        + "       <objectName> Name of the object that needs to be added \n"
        + "       <objectNamespace> Namespace the object belongs to\n"
        + "    If \"namespace\", <nickname> <basename> <comparisonAlg>\n"
        + "       <nickname> : Nickname for the object namespace \n"
        + "       <basename> : basename for the namesape\n"
        + "       <comparisonAlg> : Comparison algorithm used for objects \n"
        + "                         that belongs to this namespace \n"
        + "    If \"serviceType\", <serviceTypeName> \n"
        + "       <serviceTypeName> : Name of the service type that needs to \n"
        + "                           be added \n";

    static boolean debug = false;    

    public static void main(String args[]) throws Exception {

        System.out.println("Called");
        if (args.length<1) {    
            System.out.println(msg);
            System.exit(0);
        }

        CasClientSetup clientSetup = new CasClientSetup(args, msg);
        String instanceURL =  clientSetup.getInstanceURL();     
        String serverIdentity = clientSetup.getServerIdentity();
        int startIndex = clientSetup.argsStartIndex;
        debug = clientSetup.debug;
        printMessage("CAS client to enroll CAS entities");

        if (args.length < startIndex + 1) {
            System.err.println("Not enough parameters");
            System.err.println(msg);
            System.exit(-1);
        }

        String casObject = args[startIndex];    

        try {
            // Trust anchor or user
            if (casObject.equals(CasConstants.TRUSTANCHOR_SPEC)) {
                printMessage("Enroll Trust anchor");
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 5, serverIdentity);
                AddTrustAnchor addTrust = new AddTrustAnchor();
                addTrust.setUserGpName(args[startIndex+1]);
                addTrust.setNickname(args[startIndex+2]);
                addTrust.setAuthMethod(args[startIndex+3]);
                addTrust.setAuthData(args[startIndex+4]);
                casPort.addTrustAnchor(addTrust);
            } else if (casObject.equals(CasConstants.USER_SPEC)) {
                printMessage("Enroll User");
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 5, serverIdentity);
                AddUser addUser = new AddUser();
                addUser.setUserGpName(args[startIndex+1]);
                addUser.setNickname(args[startIndex+2]);
                addUser.setSubjectDN(args[startIndex+3]);
                addUser.setTrustAnchorNickname(args[startIndex+4]);
                casPort.addUser(addUser);
            } else if (casObject.equals(CasConstants.NAMESPACE_SPEC)) {
                printMessage("Enroll namespace");
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 5, serverIdentity);
                CreateObjectNamespace createObjNS = 
                    new CreateObjectNamespace();
                createObjNS.setUserGpName(args[startIndex+1]);
                createObjNS.setNickname(args[startIndex+2]);
                createObjNS.setBaseName(args[startIndex+3]);
                createObjNS.setComparisonAlg(args[startIndex+4]);
                casPort.createObjectNamespace(createObjNS);
            } else if (casObject.equals(CasConstants.OBJECT_SPEC)) {
                printMessage("Enroll object");
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 4, serverIdentity);
                CreateObject createObj = new CreateObject();
                createObj.setUserGpName(args[startIndex+1]);
                createObj.setObjectName(args[startIndex+2]);
                createObj.setObjectNamespace(args[startIndex+3]);
                casPort.createObject(createObj);
            } else if (casObject.equals(CasConstants.SERVICETYPE_SPEC)) {
                printMessage("Enroll service type");
                CASPortType casPort = 
                    clientSetup.getCASPort(instanceURL, 3, serverIdentity);
                CreateServiceType createServiceType = 
                    new CreateServiceType();
                createServiceType.setUserGpName(args[startIndex+1]);
                createServiceType.setServiceTypeName(args[startIndex+2]);
                casPort.createServiceType(createServiceType);
            } else {
                System.err.println("Erroneous object type " + casObject);
                System.err.println(msg);
                System.exit(-1);
            }
        } catch (NoPermissionFault noPermFault) {
            System.err.println(noPermFault.getDescription());
            System.exit(-1);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc = 
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            System.exit(-1);
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            System.exit(-1);
        }
        System.out.println("Completed successfully");
    } 

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
