/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.client;

import org.globus.cas.CASPortType;

import org.globus.cas.utils.CasStringUtils;

import org.globus.cas.types.CasObjectDesc;
import org.globus.cas.types.CasObjectData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.ArrayOfCasObjectData;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.wsrf.utils.FaultHelper;

import org.globus.cas.impl.CasConstants;

/**
 * Command line client to get all applicable policy for trustAnchor/namespace/
 * user/userGroup/object/objectGroup/serviceType/serviceTypeAction/
 * serviceActionGroup. 
 */
public class FindApplicablePolicy {
    
    static final String msg = 
        " Usage : cas-find-policy [<options>] <objectType> <objectDetails>\n"
        + " Options are -debug Runs with debug trace \n"
        + "             -c <instanceURl> Specify URL for CAS server\n"
        + "             -help Prints this message\n"
        + "              -v Prints version number\n"
        + "             -s Identity of CAS server. If not specified, will \n"
        + "                default to host authorization.\n"
        + "              -m security mechanism. 'msg' for secure message\n"
        + "                 or 'conv' for secure conversation or 'trans' for\n"
        + "                 transport security. If unspecified, if instance\n"
        + "                 url is 'https', then trasport security is used \n"
        + "                 else mechanism defaults to secure message\n"
        + "              -p protection type, 'sig' signature and 'enc' \n"
        + "                 encryption, defaults to signature\n"
        + " <objectType> : Type of object to retrieve policies for. Should \n"
        + "                be \"trustAnchor\" or \"user\" or \"object\" or \n"
        + "                \"userGroup\" or \"objectGroup\" or \"namespace\" "
        + "\n                or \"serviceType\" or \"serviceAction\" or \n"
        + "                \"serviceActionGroup\" \n"
        + " <objectDetails> : \n"
        + "    If \"trustAnchor\", <nickname>\n"
        + "       <nickname> : Nickname of trust anchor\n"
        + "    If \"user\", <nickname>\n"
        + "       <nickname> : Nickname of user\n"
        + "    If \"object\", <objectNamespace> <object>\n"
        + "       <objectNamespace> : Namespace of the object\n"
        + "       <objectName> : Name of object\n"
        + "    If \"userGroup\", <groupName>\n"
        + "       <groupName> : Name of the user group\n"
        + "    If \"objectGroup\", <groupName>\n"
        + "       <groupName> : Name of the object group\n"
        + "    If \"namespace\", <nickanme>\n"
        + "       <nickname> : Nickname of namespace\n"
        + "    If \"serviceType\", <serviceType>\n"
        + "       <serviceType> : name of the service type\n"
        + "    If \"serviceActionGroup\", <groupName>\n"
        + "       <groupName> : Name of service action group\n";

    static boolean debug =false;
    
    public static void main(String args[]) {

        if (args.length<1) {    
            System.out.println(msg);
            System.exit(0);
        }

        CasClientSetup clientSetup = new CasClientSetup(args, msg);
        String instanceURL =  clientSetup.getInstanceURL();     
        String serverIdentity = clientSetup.getServerIdentity();
        int startIndex = clientSetup.argsStartIndex;
        debug = clientSetup.debug;

        printMessage("CAS to list all applicable policy.");

        if (args.length < startIndex + 1) {
            System.err.println("Not enough parameters");
            System.err.println(msg);
            System.exit(-1);
        }

        String type = args[startIndex].trim();
        if ( !(type.equals(CasConstants.TRUSTANCHOR_SPEC) || 
               type.equals(CasConstants.USER_SPEC) || 
               type.equals(CasConstants.OBJECT_SPEC) || 
               type.equals(CasConstants.USERGP_SPEC) ||
               type.equals(CasConstants.OBJECTGP_SPEC) ||
               type.equals(CasConstants.NAMESPACE_SPEC) ||
               type.equals(CasConstants.SERVICETYPE_SPEC) ||
               type.equals(CasConstants.SERVICEACTION_SPEC) ||
               type.equals(CasConstants.SERVICEACTIONGP_SPEC))) {
            System.err.println("Erroneous type " + type);
            System.err.println(msg);
            System.exit(-1);
        }

        String identifier = null;
        CASPortType casPort = null;
        if (type.equals(CasConstants.OBJECT_SPEC)) {
            casPort = clientSetup.getCASPort(instanceURL, 3, serverIdentity);
            identifier = args[startIndex+1] + CasConstants.OBJECTSPEC_DELIMITER
                + args[startIndex+2];
        } else {
            casPort = clientSetup.getCASPort(instanceURL, 2, serverIdentity);
            identifier = args[startIndex+1];
        }

        CasObjectData[] casObj = null;

        printMessage("Parameters are " + type + " " + identifier);
        try {
            CasObjectDesc tmp = new CasObjectDesc();
            tmp.setNameOfCasObject(identifier);
            tmp.setTypeOfCasObject(type);
            ArrayOfCasObjectData retArray = 
                casPort.findApplicablePolicy(tmp);
            if (retArray != null) {
                casObj = retArray.getCasObjects();
            }
        } catch (NoPermissionFault noPermFault) {
            System.err.println(noPermFault.getDescription());
            System.exit(-1);
        } catch (CasFault casFault) {
            if (!debug) {
                String desc = 
                    (new FaultHelper(casFault)).getDescriptionAsString();
                System.err.println(desc);
            } else {
                System.err.println((new FaultHelper(casFault))
                                   .getStackTrace());
            }
            System.exit(-1);
        } catch (Exception exp) {
            if (debug) {
                System.err.println(exp.toString());
            } else {
                System.err.println(CasClientSetup.getErrorMsg(exp));
            }
            System.exit(-1);
        }

        if (casObj == null) {
            System.out.println("No applicable policies found");
        }
        else {
            System.out.println("Applicable policies:");
            for (int i=0; i<casObj.length; i++) {
                System.out.println(CasStringUtils.policyToString(
                                                  (PolicyData)casObj[i]));
            }
        }
        System.out.println("Completed successfully");
    }

    private static void printMessage(String msg) {
        if (debug) {
            System.out.println(msg);
        }
    }
}
