/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import org.globus.cas.types.UserData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.ArrayOfString;

import org.globus.cas.impl.CasConstants;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;

import java.util.Vector;
import java.util.HashMap;
import java.util.StringTokenizer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

/**
 * Evaluates and returns relevant external policies i.e policies on objects
 * other than the CAS server itself. All methods in this class deal only with 
 * external objects.
 */
public class ExternalPolicyEvaluator {

    static Log logger = 
        LogFactory.getLog(ExternalPolicyEvaluator.class.getName() );

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     ExternalPolicyEvaluator.class.getClassLoader());
    
    // Returns all policies for a given user (any resource, any action)
    public static Vector getPolicies(String userNick) 
        throws CasDBException {
        logger.debug("Method to get policies given user: " + userNick);
        String userGpQuery = getUserGroupNameQuery(userNick);
        if (userGpQuery == null) 
            return null;
        String query = getBaseQuery() + " and " + userGpQuery + " and " 
            +  getBasicObjectQuery();
        logger.debug("Query is: " + query);
        Vector policiesVector = getPoliciesForQuery(query);
        if (policiesVector == null) {
            logger.debug("Policies vector for query is null");
            return null;
        } else {
            logger.debug("Policies vector for query " + policiesVector.size());
            return sortPolicies(policiesVector);
        }
    }

    // Returns all polcies for a given user with a particular resource.
    public static Vector getPoliciesForResource(String userNick, 
                                                String resource) 
        throws CasDBException {
        logger.debug("Method to get policies give user: " + userNick 
                     + " resource is " + resource);
        String userGpQuery = getUserGroupNameQuery(userNick);
        if (userGpQuery == null) 
            return null;
        String query = getBaseQuery() + " and " + userGpQuery;
        // Get query for resource
        String queryForObjects = getResourceQueries(resource);
        if (!queryForObjects.equals("")) 
            query = query + " and " + queryForObjects;
        logger.debug("Query is: " + query);
        Vector policiesVector = getPoliciesForQuery(query);
        Vector toSortPoliciesVector = null;
        if (policiesVector == null) {
            logger.debug("Policies vector for query is null");
            return null;
        } else {
            logger.debug("Policies vector for query " + policiesVector.size());
            // overwrite object group
            for (int i=0; i<policiesVector.size(); i++) {
                PolicyData policyData = (PolicyData)policiesVector.get(i);
                policyData.setObjectSpec(resource);
                policiesVector.setElementAt(policyData, i);
            }
            if (toSortPoliciesVector == null)
                toSortPoliciesVector = new Vector();
            toSortPoliciesVector.addAll(policiesVector);
            return sortPolicies(toSortPoliciesVector);
        }
    }
    
    // Returns all polcies for a given user with a particular service/action
    public static Vector getPoliciesForActions(String userNick, 
                                               Vector actionIds)
        throws CasDBException {
        logger.debug("Method to get policies given user " + userNick + 
                     " and action id ");
        String userGpQuery = getUserGroupNameQuery(userNick);
        if (userGpQuery == null)
            return null;
        Vector toSortPoliciesVector = null;
        String preQuery = getBaseQuery() + " and " + userGpQuery + " and (";
        for (int i=0; i<actionIds.size(); i++) {
            String actionId = (String)actionIds.get(i);
            String query = preQuery + getActionQuery(actionId) + ") and " 
                + getBasicObjectQuery();
            logger.debug("ActionId " + actionId + "Query is " + query);
            Vector policiesVector = getPoliciesForQuery(query);
            if (policiesVector == null) {
                logger.debug("Policies vector for query is null");
            } else {
                logger.debug("Policies vector for query " 
                             + policiesVector.size());
                // overwrite action group
                for (int j=0; j<policiesVector.size(); j++) {
                    PolicyData policyData = (PolicyData)policiesVector.get(j);
                    policyData.setActionSpec(actionId);
                    policyData.setActionSpecDesc(
                               CasConstants.SERVICEACTION_SPEC);
                    policiesVector.setElementAt(policyData, j);
                }
                if (toSortPoliciesVector == null)
                    toSortPoliciesVector = new Vector();
                toSortPoliciesVector.addAll(policiesVector);
            }
        }
        if (toSortPoliciesVector == null)
            return null;
        else 
            return sortPolicies(toSortPoliciesVector);
    }
    
    public static Vector getPolicies(String userNick, String resource,
                                     Vector actionIds) 
        throws CasDBException {
        logger.debug("Method to get policies give user: " + userNick 
                     + " resource is " + resource + " actionIds");
        String userGpQuery = getUserGroupNameQuery(userNick);
        if (userGpQuery == null) 
            return null;
        String query = getBaseQuery() + " and " + userGpQuery;
        // Get query for resource
        String queryForObjects = getResourceQueries(resource);
        if (!queryForObjects.equals("")) 
            query = query + " and " + queryForObjects;
        Vector toSortPoliciesVector = null;
        for (int i=0; i<actionIds.size(); i++) {
            String actionId = (String)actionIds.get(i);
            String completeQuery = query + " and (" + getActionQuery(actionId) 
                + ")";
            Vector policiesVector = getPoliciesForQuery(completeQuery);
            if (policiesVector == null) {
                logger.debug("Policies vector for query is null");
            } else {
                logger.debug("Policies vector for query " 
                             + policiesVector.size());
                // overwrite action group and object
                for (int j=0; j<policiesVector.size(); j++) {
                    PolicyData policyData = (PolicyData)policiesVector.get(j);
                    policyData.setActionSpec(actionId);
                    policyData.setActionSpecDesc(
                               CasConstants.SERVICEACTION_SPEC);
                    String objSpecs[] = parseResourceString(resource);
                    String objId = 
                        Integer.toString(ObjectDataHandler.getObjectId(
                                                            objSpecs[1], 
                                                            objSpecs[0]));
                    policyData.setObjectSpec(objId);
                    policiesVector.setElementAt(policyData, j);
                }
                if (toSortPoliciesVector == null)
                    toSortPoliciesVector = new Vector();
                toSortPoliciesVector.addAll(policiesVector);
            }
        }
        if (toSortPoliciesVector == null)
            return null;
        else 
            return sortPolicies(toSortPoliciesVector);
    }

    private static String getResourceQueries(String resource) 
        throws CasDBException {

        String objSpecs[] = parseResourceString(resource);

        // All objects that match (using comp algorithm)
        Vector matchingObjIds = 
            ObjectDataHandler.retrieveMatchingObjects(objSpecs[0], 
                                                      objSpecs[1]);
        StringBuffer objQueryBuf = null;
        StringBuffer objGroupQueryBuf = null;
        if (matchingObjIds != null) {
            logger.debug("macthing ids found");
            // find all groups for those objects
            for (int i=0; i<matchingObjIds.size(); i++) {
                String objId = (String)matchingObjIds.get(i);
                if (objQueryBuf == null) {
                    objQueryBuf = new StringBuffer();
                    objQueryBuf.append("'");
                } else {
                    objQueryBuf.append(", '");
                }
                objQueryBuf.append(objId);
                objQueryBuf.append("'");
                String[] gpNames = 
                    ObjectGroupDataHandler.getObjectGroupsForMember(
                                           objId, CasConstants.OBJECT_SPEC);
                if (gpNames != null) {
                    for (int j=0; j<gpNames.length; j++) {
                        if (objGroupQueryBuf == null) {
                            objGroupQueryBuf = new StringBuffer();
                            objGroupQueryBuf.append("'");
                        } else {
                            objGroupQueryBuf.append(", '");
                        }
                        objGroupQueryBuf.append(gpNames[j]);
                        objGroupQueryBuf.append("'");
                    }
                }
            }
        }

        // If object exists, then it would match itself nad hence will
        // be automatically included in the above search (including the
        // groups the object belongs to)
        String queryForObjects = "";
        if (objQueryBuf != null) {
            String query = objQueryBuf.toString();
            logger.debug("Matching ids, object query: " + query);
            queryForObjects = "((" + CasDBConstants.COL_OBJECT_SPEC + " in ("
                + query + ")) and (" 
                + CasDBConstants.COL_OBJECT_SPEC_DESC
                + "='" + CasConstants.OBJECT_SPEC + "'))";
        }
        if (objGroupQueryBuf != null) {
            String query = objGroupQueryBuf.toString();
            logger.debug("Matching ids, objGp qurey: " + query);
            queryForObjects = "(" + queryForObjects + "or ((" 
                + CasDBConstants.COL_OBJECT_SPEC + " in (" + query
                + ")) and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
                + "='" + CasConstants.OBJECTGP_SPEC + "')))";
        }

        return queryForObjects;
    }

    private static String[] parseResourceString(String resource) 
        throws CasDBException {

        StringTokenizer strTok = 
            new StringTokenizer(resource, CasConstants.OBJECTSPEC_DELIMITER);
        if (strTok.countTokens() != 2) {
            String errMesg = 
                i18n.getMessage("malformedObjStr", 
                                CasConstants.OBJECTSPEC_DELIMITER);
            logger.error(errMesg);
            throw new CasDBException(errMesg);
        }
        String obj[] = new String[2];
        obj[0] = strTok.nextToken();
        obj[1] = strTok.nextToken();
        return obj;
    }

    // Query to return only objects or objectGroups (since external objects
    // can be stored in only one of the above)
    private static String getBasicObjectQuery() {
        return "((" + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" + 
            CasConstants.OBJECT_SPEC + "') or (" 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.OBJECTGP_SPEC + "'))";
    }

    //Query selects from policy table, excludes cas server object
    private static String getBaseQuery() throws CasDBException {
        int selfObjId = 
            ObjectDataHandler.getObjectId(CasConstants.OBJECT_SELF,
                                          CasConstants.NAMESPACE_SELF);
        return "select * from" + CasDBConstants.TABLE_POLICY 
            + " where (not ((" 
            + CasDBConstants.COL_OBJECT_SPEC + " = '" 
            + Integer.toString(selfObjId) + "') and (" 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.OBJECT_SPEC + "')))";
    }

    // Returns the query that checks for all groups that a user belongs to.
    private static String getUserGroupNameQuery(String userNick) 
        throws CasDBException {
        UserData userData = (UserData)UserDataHandler.retrieveObject(userNick);
        if (userData == null) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "User with nickname", userNick });
            logger.error(err);
            throw new CasDBException(err);
        }
        ArrayOfString arrayOfGpNames = userData.getGroupNames();
        if (arrayOfGpNames == null) {
            logger.debug("The user does not belong to any group and hence"
                         + " cannot have policies");
            return null;
        } else {
            String gpNames[] = arrayOfGpNames.getStrings();
            String gpNamesList = returnGroupMemberList(gpNames);
            return "(" + CasDBConstants.COL_USERGP_NAME + " in ( " 
                + gpNamesList + " ))";      
        }
    }

    // Returns action query given service/action id.
    private static String getActionQuery(String actionId) 
        throws CasDBException {

        String returnString = "((" + CasDBConstants.COL_ACTION_SPEC + "='"
            + actionId + "') and (" + CasDBConstants.COL_ACTION_SPEC_DESC +
            "='" + CasConstants.SERVICEACTION_SPEC + "'))";

        String groups[] = 
            ServiceTypeActionHandler.retrieveServiceActionGroups(actionId);
        String groupList = returnGroupMemberList(groups);
        if (!groupList.equals("")) {
            returnString = returnString + " or ((" 
                + CasDBConstants.COL_ACTION_SPEC + " in ( " + groupList
                + " )) and (" + CasDBConstants.COL_ACTION_SPEC_DESC +
                "='" + CasConstants.SERVICEACTIONGP_SPEC + "'))";
        }
        return returnString;
    }

    private static String returnGroupMemberList(String gpNames[])
        throws CasDBException {
        if ((gpNames == null) || (gpNames.length < 1)) {
            return "";
        } else {
            StringBuffer buf = new StringBuffer("'");
            buf.append(gpNames[0]);
            buf.append("'");
            
            for (int i=1; i<gpNames.length; i++) {
                buf.append(", '");
                buf.append(gpNames[i]);
                buf.append("'");
            }
            return buf.toString();
        }
    }

    // Returns all policies relevant to the query passed. Invokes
    // PolicyData.constructPolicyData, which expands all the action and object
    // groups; and in the latter case also ensure that only external object 
    // members of the group are returned.
    private static Vector getPoliciesForQuery(String query)
        throws CasDBException {

        logger.debug("Query is "+ query);
        
        Vector policiesVector = null;
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            policiesVector = 
                PolicyDataHandler.constructPolicyData(connection, 
                                                      resultSet, true);
            if (logger.isDebugEnabled()) {
                PolicyDataHandler.printPolicyVector(policiesVector);
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("nonCASRetrieveErr", query), exp);
            throw new CasDBException(i18n.getMessage("nonCASRetrieveErr", 
                                                     query) + exp.getMessage(),
                                     exp);
        }
        catch (CasDBException exp) {
            logger.error(i18n.getMessage("nonCASCheckErr"), exp);
            throw new CasDBException(i18n.getMessage("nonCASCheckErr") 
                                     + exp.getMessage(),
                                     exp);
        }
        finally {
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                String err = "Error  checking for external policies. "
                    + " Could not close SQL statement.";
                logger.warn(err, exp);
            }
        }
        if (policiesVector == null) {
            logger.debug("Null");
        } else {
            logger.debug("Not null");
        }
        return policiesVector;
    }

   
    // Sorts policies bases on resources, returns a vector of vector;
    // a vector per resource (object).
    public static Vector sortPolicies(Vector policyVector)
        throws CasDBException {

        if (policyVector == null) {
            logger.debug("Policies null");
            return null;
        }
        logger.debug("Policies not null");

        // Sort policies. Returned is Vector of vectors, each internal
        // vector having policy for one objectSpec
        
        //Get a Vector of obejct specification
        Vector objSpec = new Vector(policyVector.size());
        for (int i=0; i<policyVector.size(); i++) {
            objSpec.add(((PolicyData)policyVector.get(i)).getObjectSpec());
        }

        // Policies at start 
        if (logger.isDebugEnabled()) {
            logger.debug("Policies before sort per object");
            PolicyDataHandler.printPolicyVector(policyVector);
        }

        logger.debug("objSpec size at start" + objSpec.size());
        // Polices vector - vector of vectors!
        Vector policiesVector = new Vector();
        while (objSpec.size() >0) {
            int i=0;
            HashMap newObjMap = new HashMap();
            String objSpecString = (String)objSpec.remove(i);
            logger.debug("Object in consideration is " + objSpecString);
            PolicyData remPolicyData = (PolicyData)policyVector.remove(i);
            newObjMap.put(remPolicyData.getActionSpec(), remPolicyData);
            logger.debug("objSpec size afert a remove " + objSpec.size());
            while (i < objSpec.size()) {
                logger.debug("i is " + i + " objSpec size " + objSpec.size());
                if (objSpecString.equals(objSpec.get(i))) {
                    logger.debug("Equals");
                    remPolicyData = (PolicyData)policyVector.remove(i);
                    if (!newObjMap.containsKey(remPolicyData.getActionSpec())) {
                        newObjMap.put(remPolicyData.getActionSpec(), 
                                      remPolicyData);
                    }
                    objSpec.remove(i);
                }
                else {
                    logger.debug("not equals, move to next");
                    i++;
                }
            }
            policiesVector.add(new Vector(newObjMap.values()));
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Final Policies Vector " + policiesVector.size());
            for (int i=0; i<policiesVector.size(); i++) {
                logger.debug(" ObjVector " + i);
                Vector objVec = (Vector)policiesVector.get(i);
                for (int j=0; j<objVec.size(); j++) {
                    logger.debug("New policy vector");
                    PolicyData policyData = (PolicyData)objVec.get(j);
                    logger.debug("policy is " + policyData.getObjectSpec() + " "
                                 + policyData.getPolicyId());
                }
            }
        }
        return policiesVector;
    }
}
