/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import java.util.Vector;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;

import org.globus.cas.utils.CasStringUtils;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.types.PolicyData;
import org.globus.cas.types.UserData;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.ArrayOfString;

import org.globus.cas.impl.service.ObjectComparison;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

/**
 * Checks for presence of policy in CAS database based on
 * super user, group presence of user, action or object.
 * These are only internal permissions and not used while assessing
 * external policies.
 */
public class PermissionsEvaluator {

    static Log logger = 
        LogFactory.getLog(PermissionsEvaluator.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     PermissionsEvaluator.class.getClassLoader());

    static final String baseQuery = ("select" + CasDBConstants.COL_POLICY_ID 
                                     + " from" + CasDBConstants.TABLE_POLICY 
                                     + " where ");
 
    /**
     * Method returns true if there is a policy in the CAS database that
     * sematically is equivalent to the policy data. The super user check
     * is done and if the user is a superuser then the policy is accepted
     */
    public static boolean userHasPermissions(String userNick, 
                                             PolicyData policyData)
        throws CasDBException {
        logger.debug("userHasPermissions " + userNick + " does not ignore "
                     + "super user");
        return userHasPermissions(userNick, policyData, false);
    }

    /** 
     * Method returns true if there is a policy in the CAS databse that
     * sematically is equivalent to the policy data. The super user check
     * is done only if the boolean parameter is false. If the user is a 
     * superuser then the policy is accepted
     */
    public static boolean userHasPermissions(String userNick, 
                                             PolicyData policyData,
                                             boolean ignoreSuperUser) 
        throws CasDBException {
        logger.debug("userHasPermissions " + userNick + " ignore super user "
                     + ignoreSuperUser);

        UserData userData = (UserData)UserDataHandler.retrieveObject(userNick);
        ArrayOfString arrayOfGpnames = userData.getGroupNames();
        if (arrayOfGpnames != null) {
            logger.debug("gpNames is not null");
            String[] gpNames = arrayOfGpnames.getStrings();            
            for (int i=0; i<gpNames.length; i++) {
                logger.debug("gpNames " + gpNames[i]);
                policyData.setUserGroupName(gpNames[i]);
                if (hasPermissions(policyData, ignoreSuperUser)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Returns true if the userGroup can perform actionSpec on objectSpec
     * Algorithm
     * If ignoreSuperUser is false
     *   - Check if the given userGroup has cas/superuser on casServer. 
     *     Return true if yes.
     * Check if the community has permisions on objectSpec to do actionSpec
     *     Return true if yes
     * Check if the actual userGroup itslef has permission on objectSpec 
     * for actionSpec
     */
    public static boolean hasPermissions(PolicyData policyData, 
                                         boolean ignoreSuperUser) 
        throws CasDBException {

        logger.debug("hasPermissions called");
        logger.debug("Sent in policy Data " + 
                     CasStringUtils.policyToString(policyData));
        
        String userGroupName = policyData.getUserGroupName();
        logger.debug("userGroupName is " + userGroupName);

        // superuser - say yes always.
        String userQuery = queryStringForUser(userGroupName);
        logger.debug("User query " + userQuery);

        if (!ignoreSuperUser) {
            if (isSuperUser(userQuery))
                return true;
            logger.debug("not a superUser");
        }
        else
            logger.debug("Super user ignored");

        // grantAll - say yes always
        if (hasGrantAllPermissions(userQuery, policyData))
            return true;
        logger.debug("not grant all");
        
        // check if community permissions cover the action/object
        String actionQuery = 
            queryStringForAction(policyData.getActionSpec(), 
                                 policyData.getActionSpecDesc()); 
        logger.debug("action query " + actionQuery);
        if (hasCommunityPermissions(policyData, actionQuery))
            return true;
        logger.debug("no community permissions");
        
        // check if permission on the actual policy data covers it
        return userGroupHasPermissions(policyData, userQuery, actionQuery);
    }

    /**
     * Checks of the userGroupName given has cas/superuser permissions on 
     * casServer. The object comparison has to be exact here and so no 
     * wildcard issues.
     */
    private static boolean isSuperUser(String queryForUser) 
        throws CasDBException {

        StringBuffer suBuf = new StringBuffer(baseQuery);
        suBuf.append(queryForUser);

        // FIXME - can cas/superuser be in an action group 
        // and object "casServer" 
        // be as a part of a group
        // get action id for "cas/superuser"
        int suActionId = ServiceTypeActionHandler.getServiceActionId(
                                                  CasConstants.SERVICETYPE_CAS, 
                                                  CasConstants.ACTION_SU);
        if (suActionId == -1) {
            logger.debug(CasConstants.SERVICETYPE_CAS + "/" 
                         + CasConstants.ACTION_SU + " is not in the database");
            return false;
        }
        int suObjectId = 
            ObjectDataHandler.getObjectId(
                                          CasConstants.OBJECT_SELF,
                                          CasConstants.NAMESPACE_SELF);
        if (suObjectId == -1) {
            logger.debug(CasConstants.OBJECT_SELF + "/" 
                         + CasConstants.NAMESPACE_SELF + "is not in database");
            return false;
        }
        
        String superuserQuery = " and ((" + CasDBConstants.COL_ACTION_SPEC 
            + "='" + suActionId + "') and (" 
            + CasDBConstants.COL_ACTION_SPEC_DESC
            + "='" + CasConstants.SERVICEACTION_SPEC + "')) and ((" 
            + CasDBConstants.COL_OBJECT_SPEC + "='" + suObjectId + "') and (" 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.OBJECT_SPEC + "'))";
        
        suBuf.append(superuserQuery);
        logger.debug("superUserQuery " + suBuf.toString());
        return (CasDBStorage.rowExists(suBuf.toString()));
    }

    /**
     * Returns true if the community or user has grantAll permissions on the
     * the object. Query needs to deal with "objects" differently since their
     * comparison is not trivial
     */
    private static boolean hasGrantAllPermissions(String userQuery, 
                                                  PolicyData policyData) 
        throws CasDBException {

        StringBuffer grantAllBuf = new StringBuffer(baseQuery);
        
        int grantAllId = ServiceTypeActionHandler.getServiceActionId(
                                                  CasConstants.SERVICETYPE_CAS, 
                                                  CasConstants.ACTION_GRANTALL);
        if (grantAllId == -1) {
            logger.debug(CasConstants.SERVICETYPE_CAS + "/" 
                         + CasConstants.ACTION_GRANTALL + " is not in the "
                         + "database");
            return false;
        }       
        // action query
        String actionQuery = "((" + CasDBConstants.COL_ACTION_SPEC + "='" 
            + grantAllId + "') and (" + CasDBConstants.COL_ACTION_SPEC_DESC 
            + "='" + CasConstants.SERVICEACTION_SPEC + "'))";
        grantAllBuf.append(userQuery).append(" and ");
        grantAllBuf.append(actionQuery);
        logger.debug("Grant all query before object " + grantAllBuf.toString());
        String objectSpecDesc = policyData.getObjectSpecDesc();
        String objectSpec = policyData.getObjectSpec();
        // check of object has permissions because of its group membership
        if (qualifiesOnGroupPermissions(userQuery, actionQuery, objectSpec,
                                        objectSpecDesc))
            return true;
        logger.debug("Does not qualify on group permissions");
        // if it is not object
        if (!objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            grantAllBuf.append(" and ").append(queryStringForNonObject(objectSpec,objectSpecDesc));
            logger.debug("If not object: grantAll query " 
                         + grantAllBuf.toString());
            return (CasDBStorage.rowExists(grantAllBuf.toString()));
        }
        else { // object
            logger.debug("Check permission on object");
            // check if object as such qualifies
            if (qualifiesOnObjectPermissions(userQuery, actionQuery, 
                                             objectSpec))
                return true;
        }
        return false;
    }

    /**
     * Returns true if the community has permissions on objectSpec to do 
     * actionSpec. Query needs to deal with "objects" differently since 
     * their comparison is not trivial
     */
    private static boolean hasCommunityPermissions(PolicyData policyData, 
                                                   String actionQuery) 
        throws CasDBException {

        StringBuffer communityBuf = new StringBuffer(baseQuery);
        
        // community user query
        String communityUserQuery = "(" + CasDBConstants.COL_USERGP_NAME + "='" 
            + CasConstants.USER_COMMUNITY + "')";
        communityBuf.append(communityUserQuery).append(" and ");
        communityBuf.append(actionQuery);
        logger.debug("Community query before object " 
                     + communityBuf.toString());
        String objectSpec = policyData.getObjectSpec();
        String objectSpecDesc = policyData.getObjectSpecDesc();
        // check of object has permissions because of its group membership
        if (qualifiesOnGroupPermissions(communityUserQuery, actionQuery, 
                                        objectSpec, objectSpecDesc))
            return true;
        logger.debug("Does not qualify on group permissions");
        // if it is not object
        if (!objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            communityBuf.append(" and ").append(queryStringForNonObject(objectSpec, objectSpecDesc));
            logger.debug("If not object: Community query " 
                         + communityBuf.toString());
            return (CasDBStorage.rowExists(communityBuf.toString()));
        }
        else { // object
            logger.debug("Check object permissions");
            // check if object as such qualifies
            if (qualifiesOnObjectPermissions(communityUserQuery, actionQuery, 
                                             objectSpec))
                return true;
        }
        return false;
    }
    
    /** 
     * Returns true if the actin userGroup in policyData has permission for 
     * actionSpec on objectSpec. Query needs to deal with "objects" differently 
     * since their comparison is not trivial
     */
    private static boolean userGroupHasPermissions(PolicyData policyData, 
                                                   String userQuery, 
                                                   String actionQuery) 
        throws CasDBException {

        String objectSpec = policyData.getObjectSpec();
        String objectSpecDesc = policyData.getObjectSpecDesc();
        // check of object has permissions because of its group membership
        if (qualifiesOnGroupPermissions(userQuery, actionQuery, objectSpec,
                                        objectSpecDesc))
            return true;

        // objectSpec is not object 
        if (!objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            StringBuffer permBuf = new StringBuffer(baseQuery);
            permBuf.append(userQuery).append(" and ").append(actionQuery);
            permBuf.append(" and ").append(queryStringForNonObject(objectSpec,objectSpecDesc));
            logger.debug("Perm query " + permBuf.toString());
            return (CasDBStorage.rowExists(permBuf.toString()));
        }
        else { // object
            logger.debug("Check for permissions on the object");
            // check if object as such qualifies
            if (qualifiesOnObjectPermissions(userQuery, actionQuery, 
                                             objectSpec)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks if this object has permission becuase a group it belongs to has 
     * permissions. 
     */
    private static boolean qualifiesOnGroupPermissions(String userQuery, 
                                                       String actionQuery, 
                                                       String objectSpec,
                                                       String objectSpecDesc) 
        throws CasDBException {

        String objectQuery = "((" + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.OBJECTGP_SPEC + "') and (" 
            + CasDBConstants.COL_OBJECT_SPEC + " in (select" 
            + CasDBConstants.COL_OBJECTGP_NAME + " from" 
            + CasDBConstants.TABLE_OBJECTGP_ENTRY + " where" 
            + CasDBConstants.COL_OBJECT_SPEC 
            + "='" + objectSpec + "' and " 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + objectSpecDesc + "')))";
        StringBuffer query = new StringBuffer(baseQuery);
        query.append(userQuery).append(" and " ).append(actionQuery).append(" and ").append(objectQuery);
        logger.debug("Qualifies on groupPermission query " + objectQuery);
        return CasDBStorage.rowExists(query.toString());
    }

    // FIXME - split this as exact and non-exact comparison methods.
    /**
     * Checks if the object itself has permissions. 
     * Needs to run through comparison using the algorithm specified
     */
    private static boolean qualifiesOnObjectPermissions(String userQuery, 
                                                        String actionQuery, 
                                                        String objectSpec) 
        throws CasDBException {

        // get a list of all policyIds that satisfy user, action and are on 
        // objects
        StringBuffer qualifiedObjectQuery = 
            new StringBuffer("select * from" + CasDBConstants.TABLE_POLICY 
                             + " where ");
        String objectQuery = "(" + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.OBJECT_SPEC + "')";
        qualifiedObjectQuery.append(userQuery).append(" and ")
            .append(actionQuery).append(" and ").append(objectQuery);
        logger.debug("Qualifies object permissions: " 
                     + qualifiedObjectQuery.toString());
        // FIXME policy id is not need - left for debugging
        Vector policyIds = null;
        Vector objectIds = null;
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(qualifiedObjectQuery.toString());
            if ((resultSet!=null) && (resultSet.next())) {
                // policy id is not needed
                policyIds = new Vector(resultSet.getFetchSize());
                objectIds = new Vector(resultSet.getFetchSize());
                int index = 0;
                do {
                    policyIds.add(resultSet.getString(
                                  CasDBConstants.COL_POLICY_ID.trim()));
                    objectIds.add(resultSet.getString(
                                  CasDBConstants.COL_OBJECT_SPEC.trim()));
                    index++;
                } while (resultSet.next());
            }
            else {
                logger.debug("There are no policy Ids matching user/action "
                             + " spec and are objects");
                return false;
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("policyErrObj", 
                                         qualifiedObjectQuery.toString()), 
                         exp);
            throw new CasDBException(i18n.getMessage("policyErrObj",
                                                     exp.getMessage()), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("policyErrObjClose"), exp);
            }
        }
        
        // run comparison on objects in PolicyData that satisfy user, 
        // action and objectSpecDesc="object"
        ObjectData policyObjectData = 
            ObjectDataHandler.retrieveObjectForId(objectSpec);
        String compAlgClassName = 
            ObjectDataHandler.getComparisonAlg(
                              policyObjectData.getObjectNamespace());
        if (compAlgClassName == null) {
            logger.error(i18n.getMessage("compAlgUnspec"));
            throw new CasDBException(i18n.getMessage("compAlgUnspec"));
        }
        logger.debug("Comparison alg is " + compAlgClassName);
        // Returned value is used as the name for class for comparison
        ObjectComparison objectComp =
            ObjectDataHandler.getComparisonClass(compAlgClassName);
        for (int i=0; i<objectIds.size(); i++) {
            if (objectComp.objectMatches(
                           ObjectDataHandler.retrieveObjectForId(
                                             (String)objectIds.get(i)), 
                                             policyObjectData)) {
                logger.debug("True because of policy Id "+ policyIds.get(i));
                return true;
            }
        }
        return false;
    }  

    // Constructs query for user spec comparison
    private static String queryStringForUser(String userGroupName) {
        return "("+ CasDBConstants.COL_USERGP_NAME + "='" + userGroupName + "')";
    }

    // Constructs query for action spec comparison
    private static String queryStringForAction(String actionSpec, 
                                               String actionSpecDesc) {

        // extra open bracket to be closed when appending
        String commonActionQuery = "(((" + CasDBConstants.COL_ACTION_SPEC 
            + "='" + actionSpec + "') and (" 
            + CasDBConstants.COL_ACTION_SPEC_DESC + "='" + actionSpecDesc 
            + "'))";
        logger.debug("commonActionQuery " + commonActionQuery);

        // actionSpec
        // if its a serviceAction, check serviceAction and serviceActionGroup
        String actionQueryService = null;
        if (actionSpecDesc.equals(CasConstants.SERVICEACTION_SPEC)) {
            actionQueryService = " or (("+ CasDBConstants.COL_ACTION_SPEC_DESC 
                + "='" + CasConstants.SERVICEACTIONGP_SPEC + "') and (" 
                + CasDBConstants.COL_ACTION_SPEC + " in ( select " 
                + CasDBConstants.COL_SERVICEACTIONGP_NAME + " from" 
                + CasDBConstants.TABLE_SERVICEACTIONGP_ENTRY + " where" 
                + CasDBConstants.COL_SERVICEACTION_ID + "=" + actionSpec + ")))";
            logger.debug("action actionQuery " + actionQueryService);
        }

        StringBuffer actionQuery = new StringBuffer(commonActionQuery);
        if (actionQueryService != null) 
            actionQuery.append(actionQueryService).append(")");
        else
            actionQuery.append(")");
        return actionQuery.toString();
    }

    // Constructs query for objectSpec comparison provid objectSpecDesc is 
    // not object
    private static String queryStringForNonObject(String objectSpec, 
                                                  String objectSpecDesc) {
        
        // Paranoid check: return null incase this method is called with 
        // objectSpecDesc as object
        if (objectSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            logger.error("Cannot invoke this method with object spec");
            return null;
        }
            
        // extra bracket to be closed when appending
        String commonObjectQuery = "(((" + CasDBConstants.COL_OBJECT_SPEC 
            + "='" + objectSpec + "') and (" 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" + objectSpecDesc 
            + "'))";
        logger.debug("commonObjectQuery " + commonObjectQuery);

        // if objectSpecDesc is user then include user groups
        String objectQueryUser = null;
        if (objectSpecDesc.equals(CasConstants.USER_SPEC)) {
            objectQueryUser = " or ((" + CasDBConstants.COL_OBJECT_SPEC_DESC
                + "='" + CasConstants.USERGP_SPEC + "') and (" 
                + CasDBConstants.COL_OBJECT_SPEC + " in ( select " 
                + CasDBConstants.COL_USERGP_NAME + " from" 
                + CasDBConstants.TABLE_USERGP_ENTRY + " where" 
                + CasDBConstants.COL_USER_NICKNAME + "='" + objectSpec + "')))";
            logger.debug("object userQuery " + objectQueryUser);
        }
        
        StringBuffer objectQuery = new StringBuffer(commonObjectQuery);
        if (objectQueryUser != null) 
            objectQuery.append(objectQueryUser).append(")");
        else
            objectQuery.append(")");
        return objectQuery.toString();
    }
}
