/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import org.globus.cas.types.UserData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.ArrayOfString;

import org.globus.cas.impl.CasConstants;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.Statement;

import java.util.Vector;

import org.globus.util.I18n;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Used to retrieve all policies (on external objects and CAS server)
 * on a specific object or all policies for a user.
 */
public class PolicyQueryHandler {
    
    static Log logger = 
        LogFactory.getLog(PolicyQueryHandler.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     PolicyQueryHandler.class.getClassLoader());

    /**
     * Returns all polcies for a given trust anchor. 
     * a) part of objectSpec and objectSpecDesc = TrustAnchor
     * b) part of "ObjectGroup" (of which it is a member)
     */
    private static Vector getAllPoliciesForTrustAnchor(String nickName) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + nickName);
        String errMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "trust anchor", nickName });
        String existsQuery = "select * from" + CasDBConstants.TABLE_TRUSTANCHOR
            + " where" + CasDBConstants.COL_TRUST_NICKNAME + "='" 
            + nickName + "'";
        if (!CasDBStorage.rowExists(existsQuery)) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "trust anchor", nickName });
            logger.error(err);
            throw new CasDBException(err);
        }
        Vector policiesVector = new Vector();
        // TrustAnchor data could be in object specification of a policy 
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_OBJECT_SPEC + "='" 
            + nickName + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
            + "='" + CasConstants.TRUSTANCHOR_SPEC + "')";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement  = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet, 
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(errMesg + "\n" + query, exp);
            throw new CasDBException(errMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(errMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(errMesg, exp);
            }
        }

        // add obj group data
        String[] groupNames = 
            ObjectGroupDataHandler.getObjectGroupsForMember(
                                   nickName, 
                                   CasConstants.TRUSTANCHOR_SPEC);
        if (groupNames != null) {
            for (int i=0; i<groupNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                           groupNames[i]);
                if (temp != null) {
                    logger.debug("adding policies");
                    policiesVector.addAll(temp);
                }
            }
        }
        if ((policiesVector != null) && (policiesVector.size() > 0)) {
            logger.debug("policies size " + policiesVector.size());
            PolicyData[] allPolicies = new PolicyData[policiesVector.size()];
            for (int i=0; i<policiesVector.size(); i++) {
                logger.debug("polices " + i);
                allPolicies[i] = (PolicyData)policiesVector.get(i);
            }
            return policiesVector;
        }
        else
            return null;
    }

    /**
     * Returns all polcies for a given namespace
     * a) part of objectSpec and objectSpecDesc = Namespace
     * b) part of "ObjectGroup" (of which it is a member)        
     */
    private static Vector getAllPoliciesForNamespace(String nickName) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + nickName);

        String errMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "namespace", nickName });

        String existsQuery = "select * from" + CasDBConstants.TABLE_NAMESPACE
            + " where" + CasDBConstants.COL_NS_NICKNAME + "='" 
            + nickName + "'";
        if (!CasDBStorage.rowExists(existsQuery)) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "Namespace ", nickName });
            logger.error(err);
            throw new CasDBException(err);
        }
        Vector policiesVector = new Vector();
        // Namespace data could be in object specification of a policy 
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_OBJECT_SPEC + "='" 
            + nickName + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
            + "='" + CasConstants.NAMESPACE_SPEC + "')";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet, 
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(errMesg + "\n" + query, exp);
            throw new CasDBException(errMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(errMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(errMesg, exp);
            }
        }
        // add obj group data
        String[] groupNames = 
            ObjectGroupDataHandler.getObjectGroupsForMember(
                                   nickName, 
                                   CasConstants.NAMESPACE_SPEC);
        if (groupNames != null) {
            for (int i=0; i<groupNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                           groupNames[i]);
                if (temp != null) {
                    logger.debug("adding policies");
                    policiesVector.addAll(temp);
                }
            }
        }
        if ((policiesVector != null) && (policiesVector.size() > 0))
            return policiesVector;
        else
            return null;
    }
    
    /**
     * Returns all polcies for a given user. Both CAS and External policies 
     * are retuned.
     * b) part of objectSpec and objectSpecDesc = user
     * c) part "UserGroup" (of which it is a member), userSpec or objectSpec
     * d) part of "ObjectGroup" (of which it is a member)
     */
    private static Vector getAllPoliciesForUser(String userNick) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + userNick);
        String baseErrMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "user", userNick });
        UserData userData = 
            (UserData)UserDataHandler.retrieveObject(userNick);
        if (userData == null) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "user ", userNick });
            logger.error(err);
            throw new CasDBException(err);
        }
        ArrayOfString arrayOfGpnames = userData.getGroupNames();
        Vector policiesVector = new Vector();
        if (arrayOfGpnames != null) {
            String gpNames[] = arrayOfGpnames.getStrings();
            logger.debug("User: gpNames is not null " + gpNames.length);
            for (int i=0; i<gpNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.USERGP_SPEC,
                                           gpNames[i]);
                if (temp != null) {
                    logger.debug("adding policies for " 
                                 + gpNames[i]);
                    policiesVector.addAll(temp);
                }
            }
        }
        // User data could be in object specification of a policy 
        // Pick up policy only if it has already not been picked up
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_OBJECT_SPEC + "='" 
            + userNick + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
            + "='" + CasConstants.USER_SPEC + "')" ;
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet,
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(baseErrMesg + "\n" + query, exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(baseErrMesg, exp);
            }
        }
        // user could be member of a object group
        String[] groupNames = 
            ObjectGroupDataHandler.getObjectGroupsForMember(
                                   userNick, 
                                   CasConstants.USER_SPEC);
        if (groupNames != null) {
            for (int i=0; i<groupNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                           groupNames[i]);
                if (temp != null) {
                    logger.debug("adding policies");
                    policiesVector.addAll(temp);
                }
            }
        }
        if (policiesVector.size() > 0) 
            return policiesVector;
        else
            return null;
    }
    
    /**
     * Returns all polcies for a given object. Both CAS and External policies 
     * are returned.
     * a) part of objectSpec and objectSpecDesc = object
     * b) part of "ObjectGroup" (of which it is a member)
     */
    private static Vector getAllPoliciesForObject(String objectId) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + objectId);
        String baseErrMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "object", objectId });
        Vector policiesVector = new Vector();
        // policies on object group
        ObjectData objectData = 
            (ObjectData)ObjectDataHandler.retrieveObjectForId(objectId);
        if (objectData == null) {
            String err = i18n.getMessage("doesNotExist", new Object[] { 
                "object ", objectId});
            logger.error(err);
            throw new CasDBException(err);
        }
        ArrayOfString arrayOfGpnames = objectData.getGroupNames();
        if (arrayOfGpnames != null) {
            String gpNames[] = arrayOfGpnames.getStrings();            
            logger.debug("gpNames is not null");
            for (int i=0; i<gpNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                           gpNames[i]);
                if (temp != null) {
                    logger.debug("adding policies for " + gpNames[i]);
                    policiesVector.addAll(temp);
                }
            }
        }

        // policies on object
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_OBJECT_SPEC + "='" 
            + objectId + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
            + "='" + CasConstants.OBJECT_SPEC + "')";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet,
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(baseErrMesg + "\n" + query, exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(baseErrMesg, exp);
            }
        }
        if (policiesVector.size() > 0) 
            return policiesVector;
        else
            return null;
    }

    /**
     * Returns all polcies for a given service. Both CAS and External policies 
     * are returned.
     * a) objectSpec
     * b) part of "ObjectGroup" (of which it is a member)
     */
    private static Vector getAllPoliciesForServiceType(String serviceName) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + serviceName);
        String baseErrMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "service type", serviceName });
        // Ensure service type exists
        String existsQuery = "select * from" + CasDBConstants.TABLE_SERVICETYPE
            + " where" + CasDBConstants.COL_SERVICETYPE_NAME + "='"
            + serviceName + "'";
        if (!CasDBStorage.rowExists(existsQuery)) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "service type", serviceName });
            logger.error(err);
            throw new CasDBException(err);
        }
        Vector policiesVector = new Vector();
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_OBJECT_SPEC + "='" 
            + serviceName + "') and (" 
            + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
            + CasConstants.SERVICETYPE_SPEC + "')";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet, 
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(baseErrMesg + "\n" + query, exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(baseErrMesg, exp);
            }
        }

        // serviceType could be member of a object group
        String[] groupNames = 
            ObjectGroupDataHandler.getObjectGroupsForMember(
                                   serviceName, 
                                   CasConstants.SERVICETYPE_SPEC);
        if (groupNames != null) {
            for (int i=0; i<groupNames.length; i++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                           groupNames[i]);
                if (temp != null) {
                    logger.debug("adding policies");
                    policiesVector.addAll(temp);
                }
            }
        }
        if (policiesVector.size() > 0) 
            return policiesVector;
        else
            return null;
    }
    
    /**
     * Returns all polcies for a given service/action mapping.
     * a) part of actionSpec and actionSpecDesc = serviceTypeAction
     * b) part of "ServiceActionGroup" (of which it is a member)
     */
    private static Vector 
        getAllPoliciesForServiceAction(String serviceActionId) 
        throws CasDBException {
        logger.debug("getAllPolicies for " + serviceActionId);

        String baseErrMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "service type/action", serviceActionId });
        // Ensure service type exists
        String existsQuery = "select * from" 
            + CasDBConstants.TABLE_SERVICETYPE_ACTION
            + " where" + CasDBConstants.COL_SERVICEACTION_ID + "='"
            + serviceActionId + "'";
        if (!CasDBStorage.rowExists(existsQuery)) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "Service Action id ", serviceActionId }); 
            logger.error(err);
            throw new CasDBException(err);
        }
        Vector policiesVector = new Vector();
        String query = "select * from" + CasDBConstants.TABLE_POLICY
            + " where (" + CasDBConstants.COL_ACTION_SPEC_DESC + "='" 
            + CasConstants.SERVICEACTION_SPEC + "') and (" 
            + CasDBConstants.COL_ACTION_SPEC + "='" + serviceActionId + "')";
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            Vector temp = 
                PolicyDataHandler.constructPolicyData(connection, resultSet, 
                                                      false);
            if (temp != null) {
                logger.debug("adding policies");
                policiesVector.addAll(temp);
            }
        }
        catch (SQLException exp) {
            logger.error(baseErrMesg + "\n" + query, exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(baseErrMesg, exp);
            }
        }
        
        // Retrieve groups for these
        String gpNames[] = ServiceTypeActionHandler
            .retrieveServiceActionGroups(serviceActionId);
        if (gpNames != null) {
            for (int j=0; j<gpNames.length; j++) {
                Vector temp = 
                    getAllPoliciesForGroup(CasConstants.SERVICEACTIONGP_SPEC,
                                           gpNames[j]);
                if (temp != null) {
                    logger.debug("adding action group policies");
                    policiesVector.addAll(temp);
                }
            }
        }
        if (policiesVector.size() > 0) 
            return policiesVector;
        else
            return null;
    }

    /**
     * Returns all polcies for a given userGroup/obejctGroup/serviceActionGroup
     * Both CAS and External policies are returned
     */
    private static Vector getAllPoliciesForGroup(String type, String groupName)
        throws CasDBException {

        logger.debug("type " + type + "groupName " + groupName);
        String baseErrMesg = i18n.getMessage("retrPolicyFor", new Object[] {
            "group", "" });
        String queryInit = "select * from" + CasDBConstants.TABLE_POLICY
            + " where ";
        String queryTail = null;
        String existsQuery = null;
        if (type.equals(CasConstants.USERGP_SPEC)) {
            logger.debug("usergroup processing");
            queryTail = "(" + CasDBConstants.COL_USERGP_NAME + "='" 
                + groupName + "') or ((" + CasDBConstants.COL_OBJECT_SPEC 
                + "='" + groupName + "') and (" 
                + CasDBConstants.COL_OBJECT_SPEC_DESC + "='" 
                + CasConstants.USERGP_SPEC + "'))";
            existsQuery = "select * from" + CasDBConstants.TABLE_USERGP
                + " where" + CasDBConstants.COL_USERGP_NAME + "='"
                + groupName + "'";
        }
        else if (type.equals(CasConstants.OBJECTGP_SPEC)) {
            logger.debug("obejct group");
            queryTail = "(" + CasDBConstants.COL_OBJECT_SPEC + "='" 
                + groupName + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC
                + "='" + CasConstants.OBJECTGP_SPEC + "')";
            existsQuery = "select * from" + CasDBConstants.TABLE_OBJECTGP
                + " where" + CasDBConstants.COL_OBJECTGP_NAME + "='"
                + groupName + "'";
        } else if (type.equals(CasConstants.SERVICEACTIONGP_SPEC)) {
            logger.debug("service action group");
            queryTail = "((" + CasDBConstants.COL_ACTION_SPEC + "='" 
                + groupName + "') and (" + CasDBConstants.COL_ACTION_SPEC_DESC
                + "='" + CasConstants.SERVICEACTIONGP_SPEC + "')) or (("
                + CasDBConstants.COL_OBJECT_SPEC + "='" + groupName 
                + "') and (" + CasDBConstants.COL_OBJECT_SPEC_DESC  + "='" 
                + CasConstants.SERVICEACTIONGP_SPEC + "'))";
            existsQuery = "select * from" 
                + CasDBConstants.TABLE_SERVICEACTIONGP + " where" 
                + CasDBConstants.COL_SERVICEACTIONGP_NAME + "='"
                + groupName + "'";
        } else {
            String errMesg = "Type must be " + CasConstants.USERGP_SPEC
                + " " + CasConstants.OBJECTGP_SPEC + " "
                + CasConstants.SERVICEACTIONGP_SPEC;
            logger.error(errMesg);
            throw new CasDBException(baseErrMesg + errMesg);
        }

        if (!CasDBStorage.rowExists(existsQuery)) {
            String errMesg = type + "group " + groupName 
                + " does not exist";
            logger.error(errMesg);
            throw new CasDBException(baseErrMesg + errMesg);
        }

        String query = queryInit + queryTail;
        logger.debug("Query is "+ query);
        Connection connection = CasDBStorage.getDBConnection();
        Vector policies = null;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            policies = 
                PolicyDataHandler.constructPolicyData(connection, resultSet, 
                                                      false);
        }
        catch (SQLException exp) {
            logger.error(baseErrMesg + "\n" + query, exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        catch (CasDBException exp) {
            logger.error(exp);
            throw new CasDBException(baseErrMesg + exp.getMessage(), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(baseErrMesg, exp);
            }
        }

        // UserGp can be a part of objectGp
        if (type.equals(CasConstants.USERGP_SPEC)) {
            logger.debug("User group can be part of object group");
            String[] groupNames = 
                ObjectGroupDataHandler.getObjectGroupsForMember(
                                       groupName, 
                                       CasConstants.USERGP_SPEC);
            if (groupNames != null) {
                for (int i=0; i<groupNames.length; i++) {
                    Vector temp = 
                        getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                               groupName);
                    if (temp != null) {
                        logger.debug("adding policies");
                        policies.addAll(temp);
                    }
                    
                }
            }
        }
        return policies;
    }

    /**
     * Returns all applicable policies for a given type and name
     */
    public static PolicyData[] getAllApplicablePolicy(String type,
                                                      String name)
        throws CasDBException {
        Vector policyVector = null;
        if (type.trim().equals(CasConstants.USER_SPEC)) {
            logger.debug("Policy for user " + name);
            policyVector = getAllPoliciesForUser(name.trim());
        } else if (type.trim().equals(CasConstants.USERGP_SPEC)) {
            logger.debug("Policy for user group" + name);
            policyVector = getAllPoliciesForGroup(CasConstants.USERGP_SPEC,
                                                  name.trim());
        } else if (type.trim().equals(CasConstants.OBJECT_SPEC)) {
            logger.debug("Policy for object group" + name);
            policyVector = getAllPoliciesForObject(name.trim());
        } else if (type.trim().equals(CasConstants.OBJECTGP_SPEC)) {
            logger.debug("Policy for object group" + name);
            policyVector = getAllPoliciesForGroup(CasConstants.OBJECTGP_SPEC,
                                                  name.trim());
        } else if (type.trim().equals(CasConstants.SERVICETYPE_SPEC)) {
            logger.debug("Policy for service type" + name);
            policyVector = getAllPoliciesForServiceType(name.trim());
        } else if (type.trim().equals(CasConstants.SERVICEACTIONGP_SPEC)) {
            logger.debug("Policy for service/action gp" + name);
            policyVector = 
                getAllPoliciesForGroup(CasConstants.SERVICEACTIONGP_SPEC,
                                       name.trim());
        } else if (type.trim().equals(CasConstants.TRUSTANCHOR_SPEC)) {
            logger.debug("Policy for trust anchor" + name);
            policyVector = getAllPoliciesForTrustAnchor(name.trim());
        } else if (type.trim().equals(CasConstants.NAMESPACE_SPEC)) {
            logger.debug("Policy for namespace" + name);
            policyVector = getAllPoliciesForNamespace(name.trim());
        } else if (type.trim().equals(CasConstants.SERVICEACTION_SPEC)) {
            logger.debug("Policy for serviceAction" + name);
            policyVector = getAllPoliciesForServiceAction(name.trim());
        } else {
            String err = i18n.getMessage("retrPolicyFor", new Object[]
                { "erroneous type", type.trim() });
            logger.error(err);
            throw new CasDBException(err);
        }
        if (policyVector != null) {
            PolicyData[] allPolicies = new PolicyData[policyVector.size()];
            logger.debug("Policies vector size " + policyVector.size());
            for (int i=0; i<policyVector.size(); i++) {
                logger.debug("polices " + i);
                PolicyData externalPolicy = 
                    replaceInternalIds((PolicyData)policyVector.get(i));
                allPolicies[i] = externalPolicy;
            }
            return allPolicies;
        }
        return null;
    }
    
    private static PolicyData replaceInternalIds(PolicyData policyData) 
        throws CasDBException {
        
        if (policyData == null)
            return null;

        if (policyData.getActionSpecDesc().equals(
                                           CasConstants.SERVICEACTION_SPEC)) {
            String actionSpec = policyData.getActionSpec();
            String serviceAction = 
                ServiceTypeActionHandler.getServiceActionMapping(actionSpec);
            policyData.setActionSpec(serviceAction);
        } 

        if (policyData.getObjectSpecDesc().equals(CasConstants.OBJECT_SPEC)) {
            String objSpec = policyData.getObjectSpec();
            String objNsName =                  
                ObjectDataHandler.getObjectDescription(objSpec);
            policyData.setObjectSpec(objNsName);
        }
        return policyData;
    }
}
