/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.SQLException;

import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.cas.impl.CasConstants;

import org.globus.util.I18n;

/**
 * Used to manipulate the serviceType/action mappings in the database
 */
public class ServiceTypeActionHandler {
    
    static Log logger = 
        LogFactory.getLog(ServiceTypeActionHandler.class.getName() );

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     ServiceTypeActionHandler.class.getClassLoader());

    /**
     * Adds a new action mapping to an existing service type.
     * Inserts data into service action table.
     */
    public static String addServiceActionMapping(String serviceName, 
                                                 String actionName) 
        throws CasDBException {

        logger.debug("addServiceActionMapping " + serviceName + " " 
                     + actionName);
        String result = null;
        String query = "insert into" + CasDBConstants.TABLE_SERVICETYPE_ACTION 
            + " (" + CasDBConstants.COL_SERVICETYPE_NAME + "," 
            + CasDBConstants.COL_ACTION_NAME + ") values ('"
            + serviceName.trim() + "','" + actionName.trim() + "')";
        CasDBStorage.runUpdateQuery(query);
        int id = getServiceActionId(serviceName, actionName);
        if (id == -1) {
            String err = i18n.getMessage("serviceActionRetrId");
            logger.error(err);
            throw new CasDBException(err);
        }
        else 
            result = Integer.toString(id);
        return result;
    }

    /** 
     * List of service action mappings
     */
    public static String[] listServiceActionMappings() throws CasDBException {
        
        String query = "select * from" + 
            CasDBConstants.TABLE_SERVICETYPE_ACTION;
        Connection connection = CasDBStorage.getDBConnection();
        Vector serviceAction = null;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                serviceAction = new Vector(resultSet.getFetchSize());
                do {
                    String service = 
                        resultSet.getString(
                                  CasDBConstants.COL_SERVICETYPE_NAME.trim());
                    String action = 
                        resultSet.getString(
                                  CasDBConstants.COL_ACTION_NAME.trim());
                    serviceAction.add(service 
                                      + CasConstants.SERVICEACTION_DELIMITER 
                                      + action);
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("serviceActionList", query),
                         exp);
            throw new CasDBException(i18n.getMessage("serviceActionList",
                                                     exp.getMessage()), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("serviceActionListClose"), exp);
            }
        }
        return CasDBStorage.vectorToStringArray(serviceAction);
    }
    
    /**
     * Retrieves all the actions mapped to a particular service type
     * Picks up data from from service action table
     * Returns null if no mappings exist
     */
    public static String[] retrieveServiceActionMappings(String serviceName) 
        throws CasDBException {
        String query = "select * from" + CasDBConstants.TABLE_SERVICETYPE_ACTION 
            + " where" + CasDBConstants.COL_SERVICETYPE_NAME + "='" 
            + serviceName.trim() + "'";
        logger.debug("Query to retrieve service/action mappings is " + query);
        Connection connection = CasDBStorage.getDBConnection();
        Vector actionNames = null;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                actionNames = new Vector(resultSet.getFetchSize());
                do {
                    actionNames.add(resultSet.getString(
                                CasDBConstants.COL_ACTION_NAME.trim()));
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object[] {
                "service/action mappings", query }), exp);
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                      new Object[] { "service/action mappings", 
                                     exp.getMessage() }), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose",
                                            "service/action mapping"), exp);
            }
        }
        return CasDBStorage.vectorToStringArray(actionNames);
    }

    /**
     * Retrieves all the sercice/action ids on a service type
     * Picks up data from from service action table
     * Returns null if no mappings exist
     */
    public static String[] retrieveServiceActionIds(String serviceName) 
        throws CasDBException {
        String query = "select * from" 
            + CasDBConstants.TABLE_SERVICETYPE_ACTION 
            + " where" + CasDBConstants.COL_SERVICETYPE_NAME + "='" 
            + serviceName.trim() + "'";
        logger.debug("Query to retrieve service/action mappings is " + query);
        Connection connection = CasDBStorage.getDBConnection();
        String serviceActionId[] = null;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                int i =0;
                serviceActionId = new String[resultSet.getFetchSize()];
                do {
                    serviceActionId[i] = 
                        resultSet.getString(
                                  CasDBConstants.COL_SERVICEACTION_ID.trim());
                    i++;
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object [] {
                "service action ids", query }), exp);
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                                                     new Object [] {
                                                         "service action ids",
                                                         exp.getMessage() }), 
                                     exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", 
                                            "service action id"), exp);
            }
        }
        return serviceActionId;
    }

    /**
     * Removes the action mapped to a particular service type
     * Deletes from service action table
     */
    public static void removeServiceActionMapping(String serviceName, 
                                                  String actionName) 
        throws CasDBException {
        
        logger.debug("removeServiceAction " + serviceName + " " + actionName);
        String query = "delete from" + CasDBConstants.TABLE_SERVICETYPE_ACTION 
            + " where (" + CasDBConstants.COL_SERVICETYPE_NAME + "='" 
            + serviceName.trim() + "') and (" + CasDBConstants.COL_ACTION_NAME 
            + "='" + actionName.trim() + "')";
        int result = CasDBStorage.runUpdateQuery(query);
        if (result <=0) {
            String err = i18n.getMessage("serviceActionDelError", new Object[] 
                { serviceName, actionName });
            logger.error(err);
            throw new CasDBException(err);
        }
    }

    /**
     * Removes the action mapped to a particular service type
     * Deletes from service action table
     */
    public static void removeServiceActionMapping(String serviceActionId) 
        throws CasDBException {
        
        String query = "delete from" + CasDBConstants.TABLE_SERVICETYPE_ACTION 
            + " where (" + CasDBConstants.COL_SERVICEACTION_ID + "=" 
            + Integer.parseInt(serviceActionId.trim()) + ")";
        int result = CasDBStorage.runUpdateQuery(query);
        if (result <=0) {
            String err = i18n.getMessage("serviceActionIdDelError", 
                                         serviceActionId);
            logger.error(err);
            throw new CasDBException(err);
        }
    }

    /** 
     * Retrieves the serviceaction groups this service/action belongs to
     */
    public static String[] retrieveServiceActionGroups(String serviceActionId)
        throws CasDBException {
        
        String query = "select" + CasDBConstants.COL_SERVICEACTIONGP_NAME +
            " from " + CasDBConstants.TABLE_SERVICEACTIONGP_ENTRY + " where"
            + CasDBConstants.COL_SERVICEACTION_ID + "='" + serviceActionId
            + "'";
        Vector gpNames = null;
        Connection connection = CasDBStorage.getDBConnection();
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                gpNames = new Vector(resultSet.getFetchSize());
                do {
                    gpNames.add(resultSet.getString(
                            CasDBConstants.COL_SERVICEACTIONGP_NAME.trim()));
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object[] {
                "service action groups", query }),  exp);
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                   new Object[] { "service action groups", exp.getMessage()}),
                                     exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", 
                                            "service action groups"), exp);
            }
        }
        return CasDBStorage.vectorToStringArray(gpNames);
    }

    /**
     * Creates a new service/action group
     * Inserts data into service action group
     **/
    public static void createServiceActionGroup(String serviceGroupName) 
        throws CasDBException { 
        
        String query = "insert into" + CasDBConstants.TABLE_SERVICEACTIONGP 
            + " (" + CasDBConstants.COL_SERVICEACTIONGP_NAME + ") values ('" 
            + serviceGroupName.trim() + "')"; 
        CasDBStorage.runUpdateQuery(query);
    }

    /** 
     * Adds entry to service/action group 
     * (get service_id from service_type_action table and then insert entry)
     */
    public static void addServiceActionGpEntry(String serviceGroupName, 
                                               String serviceName, 
                                               String actionName) 
        throws CasDBException { 
        
        logger.debug("addServiceActionGpEntry " + serviceGroupName + " "
                    + serviceName + " " + actionName);
        Connection connection = CasDBStorage.getDBConnection();
        /*
        try {
            connection.setAutoCommit(false);
        }
        catch (SQLException exp) {
            logger.error("Could not set auto commit property " 
                         + MessageUtils.toString(exp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Could not set auto commit property " 
                                     + exp.getMessage(), exp);
                                     }
        */

        int serviceActionId = getServiceActionId(connection, serviceName, 
                                                 actionName);
        if (serviceActionId == -1) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "service/action mapping ", serviceName + "/" + actionName });
            logger.error(err);
            throw new CasDBException(err);
        }

        String query = "insert into " 
            + CasDBConstants.TABLE_SERVICEACTIONGP_ENTRY + " (" 
            + CasDBConstants.COL_SERVICEACTIONGP_NAME + "," 
            + CasDBConstants.COL_SERVICEACTION_ID + " ) values ('" 
            + serviceGroupName.trim() + "','" 
            + Integer.toString(serviceActionId) + "')";
        runUpdateQuery(connection, query);
        /*
        try {
            connection.commit();
        }
        catch (SQLException sqlExp) {
            logger.error("Commit failed. Database may not be consistent" 
                         + MessageUtils.toString(sqlExp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Commit failed. Database may not be "
                                     + " consistent. " + sqlExp.getMessage(), 
                                     sqlExp);
        }
        try {
            connection.setAutoCommit(true);
        }
        catch (SQLException exp) {
            logger.error("Could not set auto commit property to true" 
                         + MessageUtils.toString(exp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Could not set auto commit property "
                                     + exp.getMessage(), exp);
        }
        */
        CasDBStorage.returnDBConnection(connection);
    }    

    /**
     * List of service action group
     */
    public static String[] listServiceActionGroups() throws CasDBException { 
        
        String query = "select" + CasDBConstants.COL_SERVICEACTIONGP_NAME 
            + " from" + CasDBConstants.TABLE_SERVICEACTIONGP;
        return CasDBStorage.runListQuery(query, 
                            CasDBConstants.COL_SERVICEACTIONGP_NAME);
    }

    /**
     * Retrives service type/action id in a serviceActionGroup. 
     * Returns an array of service/action ids.
     */
    public static String[] 
        retrieveServiceActionGpEntriesAsId(String serviceGroupName) 
        throws CasDBException {

        Connection connection = CasDBStorage.getDBConnection();
        try {
            return retrieveServiceActionGpEntriesAsId(connection, 
                                                      serviceGroupName);
        } finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(connection);
        }
    }

    public static String[] 
        retrieveServiceActionGpEntriesAsId(Connection connection,
                                           String serviceGroupName) 
        throws CasDBException {

        Vector serviceAction = null;
        String query = "select * from" 
            + CasDBConstants.TABLE_SERVICEACTIONGP_ENTRY + " where" 
            + CasDBConstants.COL_SERVICEACTIONGP_NAME + "='" 
            + serviceGroupName.trim() + "'";

        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = connection.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                serviceAction = new Vector(resultSet.getFetchSize());
                do {
                    serviceAction.add(resultSet.getString(
                                  CasDBConstants.COL_SERVICEACTION_ID.trim()));
                } while (resultSet.next());
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object[] { 
                "action group ids", query }), exp);
            throw new CasDBException(i18n.getMessage("retrieveErr", 
                                                     new Object[] { 
                                                         "action group ids",
                                                         exp.getMessage() }),
                                     exp);
        }
        finally {
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", 
                                            "service action groups id"), exp);
            }
        }
        return CasDBStorage.vectorToStringArray(serviceAction);
    }
    
    /**
     * Retrives service type/actions in a serviceActionGroup. 
     * Returned as "service_type_name/action"
     */
    public static String[] 
        retrieveServiceActionGpEntries(String serviceGroupName) 
        throws CasDBException {
        logger.debug("retrieve service action gp entries " + serviceGroupName);
        String serviceActionId[] = 
            retrieveServiceActionGpEntriesAsId(serviceGroupName);
        if (serviceActionId != null) {
            String serviceAction[] = new String[serviceActionId.length];
            for (int i=0; i<serviceActionId.length; i++) {
                serviceAction[i] = 
                    getServiceActionMapping(new Integer(
                                                serviceActionId[i]).intValue());
            }
            return serviceAction;
        }
        return null;
    }

    /**
     * Deletes entry from service/action group 
     * (get service_id from service_type_action table and then delete entry)
     */
    public static void removeServiceActionGpEntry(String serviceGroupName, 
                                                  String serviceName, 
                                                  String actionName) throws 
                                                      CasDBException { 

        Connection connection = CasDBStorage.getDBConnection();
        /*
        try {
            connection.setAutoCommit(false);
        }
        catch (SQLException exp) {
            logger.error("Could not set auto commit property " 
                         + MessageUtils.toString(exp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Could not auto commit property. "
                                     + exp.getMessage(), exp);
        }
        */
        int serviceActionId = getServiceActionId(connection, serviceName, 
                                                 actionName);
        if (serviceActionId == -1) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "service/action mapping", serviceName + "/" + actionName });
            logger.error(err);
            throw new CasDBException(err);
        }

        String query = "delete from " + 
            CasDBConstants.TABLE_SERVICEACTIONGP_ENTRY + " where (" 
            + CasDBConstants.COL_SERVICEACTIONGP_NAME + "='" + 
            serviceGroupName.trim() + "') and (" 
            + CasDBConstants.COL_SERVICEACTION_ID + "='" 
            + (Integer.toString(serviceActionId)) + "')";
        int result = runUpdateQuery(connection, query);
        if (result <= 0) {
            String err = i18n.getMessage("serviceActionMemDelError", 
                                         new Object[] { serviceGroupName, 
                                                        serviceName,
                                                        actionName });
            logger.error(err);
            throw new CasDBException(err);
        }

        /*
        try {
            connection.commit();
        }
        catch (SQLException sqlExp) {
            logger.error("Commit failed. Database may not be consistent" 
                         + MessageUtils.toString(sqlExp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Commit failed. Database may not be "
                                     + "consistent. " + sqlExp.getMessage(), 
                                     sqlExp);
        }
        try {
            connection.setAutoCommit(true);
        }
        catch (SQLException exp) {
            logger.error("Could not set auto commit property to true " 
                         + MessageUtils.toString(exp));
            CasDBStorage.returnDBConnection(connection);
            throw new CasDBException("Could not auto commit property. "
                                     + exp.getMessage(), exp);
        }
        */
        CasDBStorage.returnDBConnection(connection);
    }

    /**
     * Removes service/action group 
     */
    public static void deleteServiceActionGroup(String serviceGroupName) 
        throws CasDBException { 
        
        Connection connection = CasDBStorage.getDBConnection();
        String query = "delete from" + CasDBConstants.TABLE_SERVICEACTIONGP 
            + " where" + CasDBConstants.COL_SERVICEACTIONGP_NAME + "='" 
            + serviceGroupName + "'";
        int result = runUpdateQuery(connection, query);
        if (result <=0) {
            String err = i18n.getMessage("serviceActionGpDelError", 
                                         serviceGroupName);
            logger.error(err);
            throw new CasDBException(err);
        }
        CasDBStorage.returnDBConnection(connection);
    }

    public static String getServiceActionMapping(String serviceActionId)
        throws CasDBException {
        return getServiceActionMapping(Integer.parseInt(serviceActionId));
    }

    //Retrieves service_type/action mapping given serviceActionId
    public static String getServiceActionMapping(int serviceActionId)   
        throws CasDBException {
        
        String serviceTypeAction = null;
        Connection conn = CasDBStorage.getDBConnection();
        // To get service_action_id
        String query = "select * from" 
            + CasDBConstants.TABLE_SERVICETYPE_ACTION + " where" 
            + CasDBConstants.COL_SERVICEACTION_ID + "=" + serviceActionId;
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = conn.createStatement();
            resultSet = statement.executeQuery(query);
            if (resultSet==null) {
                String err = 
                    i18n.getMessage("doesNotExistTable", 
                                    new Object[] { "Service action id", 
                                                   "serviceAction" });
                logger.error(err);
                throw new CasDBException(err);
            }
            else {
                if (resultSet.next()) {
                    String temp = resultSet.getString(
                                  CasDBConstants.COL_SERVICETYPE_NAME.trim());
                    String temp1 = resultSet.getString(
                                   CasDBConstants.COL_ACTION_NAME.trim());
                    serviceTypeAction = temp 
                        + CasConstants.SERVICEACTION_DELIMITER + temp1;
                }
            }
        }
        catch (SQLException exp) {
            logger.error(i18n.getMessage("retrieveErr", new Object[] {
                "service action mapping", query }), exp);
            throw new CasDBException(i18n
                      .getMessage("retrieveErr", new Object[] { 
                          "service action mapping", exp.getMessage()}), exp);
        }
        finally {
            // returning connection irrespective of whether stmt
            // and result set are closed or not.
            CasDBStorage.returnDBConnection(conn);
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", 
                                            "service action mapping"), exp);
            }
        }
        return serviceTypeAction;
    }

    /**
     * Retieves service action id given service name and action name
     * returns -1 if not found
     */
    public static int getServiceActionId(String serviceName, String actionName) 
        throws CasDBException {
        Connection conn = CasDBStorage.getDBConnection();
        int id = getServiceActionId(conn, serviceName, actionName);
        CasDBStorage.returnDBConnection(conn);
        return id;
    }

    // Retrieves service_action_id given service_type and action_name
    private static int getServiceActionId(Connection conn, String serviceName, 
                                          String actionName)
        throws CasDBException {
        int serviceId = -1;
        // To get service_action_id
        String query = "select" + CasDBConstants.COL_SERVICEACTION_ID + " from" 
            + CasDBConstants.TABLE_SERVICETYPE_ACTION + " where" 
            + CasDBConstants.COL_SERVICETYPE_NAME + "='" + serviceName.trim() 
            + "' and " + CasDBConstants.COL_ACTION_NAME + "='" 
            + actionName.trim() + "'";
        logger.debug("getServiceActionId " + query);
        Statement statement = null;
        ResultSet resultSet = null;
        try {
            statement = conn.createStatement();
            resultSet = statement.executeQuery(query);
            if ((resultSet!=null) && (resultSet.next())) {
                serviceId = resultSet.getInt(
                            CasDBConstants.COL_SERVICEACTION_ID.trim());
            }
        }
        catch (SQLException exp) {
            CasDBStorage.returnDBConnection(conn);
            logger.error(i18n.getMessage("retrieveErr", new Object[] {
                "service action id", query }), exp);
            throw new CasDBException("Error getting service action id " + 
                                     exp.getMessage(), exp);
        }
        finally {
            try {
                if (resultSet != null)
                    resultSet.close();
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("retrieveErrClose", 
                                            "service action id "), exp);
            }
        }
        return serviceId;
    }

    // runs a query string passed as parameter
    private static int runUpdateQuery(Connection conn, String query) 
        throws CasDBException {
        Statement statement = null;
        int result = -1;
        try {
            statement = conn.createStatement();
            result = statement.executeUpdate(query);
        }
        catch (SQLException exp) {
            CasDBStorage.returnDBConnection(conn);
            logger.error(i18n.getMessage("stmtExecErr",query), exp);
            throw new CasDBException(i18n.getMessage("stmtExecErr", 
                                                     exp.getMessage()), exp);
        }
        finally {
            try {
                if (statement != null)
                    statement.close();
            }
            catch (SQLException exp) {
                logger.warn(i18n.getMessage("stmtExecErrClose"), exp);
            }
        }
        return result;
    }
}
