/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.databaseAccess;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.SQLException;
import java.sql.ResultSet;

import java.util.Vector;
import java.util.StringTokenizer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.cas.types.ObjectData;

import org.globus.cas.impl.service.ObjectComparison;

import org.globus.util.I18n;

/**
 * Comparison class for objects that are within the realm of 
 * "CAS Simple Policy Language"
 * Wild card character is "*" and file delimiter is "/" 
 */
public class WildCardComparison extends ObjectComparison {

    static Log logger = LogFactory.getLog(WildCardComparison.class.getName());

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.databaseAccess.errors",
                     WildCardComparison.class.getClassLoader());

    final static String wildCardChar = "*";
    final static String fileDelimiter = "/";

    // Set of all objects that are a subset    
    public ObjectData[] matchingSubset(ObjectData objData) 
	throws CasDBException {

	logger.debug("matching subset " + objData.getObjectNamespace());
	Vector objects = getObjectsInNamespace(objData.getObjectNamespace());
	return matchForSubset(objects, objData);
    }

    public ObjectData[] matchForSubset(Vector objects, ObjectData objData) {
       
	if ((objects == null) || (objects.size() <= 0)) {
	    logger.debug("No namespace match");
	    return null;
	}
	String reqName = objData.getObjectName();
	logger.debug("Subset, Request name " + reqName);
	String toMatch = reqName;
	Vector matchedObjects = new Vector();
	if (toMatch.endsWith(fileDelimiter + wildCardChar)) {
	    // it needs to match till /
	    toMatch = reqName.substring(0, reqName.indexOf(wildCardChar));
	    logger.debug("toMatch " + toMatch);

	    for (int i=0; i<objects.size(); i++) {
		ObjectData obj = (ObjectData)objects.get(i);
		String objInDb = obj.getObjectName();
		if ((!objInDb.equals(reqName)) && 
		    (objInDb.startsWith(toMatch))) {
		    logger.debug("adding " + obj.getObjectName());
		    matchedObjects.add(obj);
		}
	    }
	}

	ObjectData[] retData = null;
	if (matchedObjects.size() > 0) {
	    retData = new ObjectData[matchedObjects.size()];
	    matchedObjects.toArray(retData);
	}
	return retData;
    }

    // Set of all objects that are a super set
    public ObjectData[] matchingSuperset(ObjectData objData)
	throws CasDBException {
	logger.debug("Matching super set");
	Vector objects = getObjectsInNamespace(objData.getObjectNamespace());
	return matchForSuperset(objects, objData);
    }

    public ObjectData[] matchForSuperset(Vector objects, ObjectData objData) {

	if ((objects == null) || (objects.size() <= 0)) {
	    logger.debug("No namespace match");
	    return null;
	}
	String reqName = objData.getObjectName();
	logger.debug("Superset, Request name " + reqName);
	if (reqName.endsWith(fileDelimiter + wildCardChar))
	    reqName = reqName.substring(0, reqName.indexOf(fileDelimiter 
							   + wildCardChar));
	logger.debug("Altered reqName " + reqName);
	StringTokenizer strTok = new StringTokenizer(reqName, fileDelimiter);
	int numberOfRelevantToks = strTok.countTokens()-1;
	logger.debug("Number of relevant tokens " + numberOfRelevantToks);
	String[] tokensToMatch = null;
	if (numberOfRelevantToks == 0) {
	    tokensToMatch = new String[1];
	    tokensToMatch[0] = "*";
	} else {
	    tokensToMatch = new String[numberOfRelevantToks];
	    for (int i=0; i<numberOfRelevantToks; i++) {
		if (i == 0) {
		tokensToMatch[i] = strTok.nextToken();
		} else {
		    tokensToMatch[i] = tokensToMatch[i-1] + fileDelimiter 
			+ strTok.nextToken();
		}
		logger.debug("TOKEN: " + tokensToMatch[i]);
	    }
	}
	int lengthToCompare = reqName.length();
	Vector matchedObjects = new Vector();
	for (int i=0; i<objects.size(); i++) {
	    // if the object has length greater than reqObj length, not a match
	    ObjectData obj = (ObjectData)objects.get(i);
	    String objInDb = obj.getObjectName();
	    if (objInDb.length() <= lengthToCompare) {
		logger.debug("Shorter string, chance its a super set " 
			     + objInDb);
		for (int j=0; j<tokensToMatch.length; j++) {
		    logger.debug("Token to match" + tokensToMatch[j] 
				 + fileDelimiter + wildCardChar);
		    if (objInDb.equals(tokensToMatch[j] + fileDelimiter 
				       + wildCardChar)) {
			logger.debug("Added obj " + objInDb);
			matchedObjects.add(obj);
			break;
		    }
		}
	    }
	}

	ObjectData[] retData = null;
	if (matchedObjects.size() > 0) {
	    retData = new ObjectData[matchedObjects.size()];
	    matchedObjects.toArray(retData);
	}
	return retData;
    }

    // Assuming namespace compAlg is wildcard when this method is invoked.
    public boolean objectMatches(ObjectData policyObj, ObjectData requestObj) 
	throws CasDBException {

	// if not same namespace return false
	if (!policyObj.getObjectNamespace().equals(
					    requestObj.getObjectNamespace()))
	    return false;
	return matchObjectNames(policyObj.getObjectName(), 
				requestObj.getObjectName());
    }

    private boolean matchObjectNames(String policyObjectName, 
				     String requestObjectName) {
	
	logger.debug("matchObjectNames called with policy " + policyObjectName 
		     + " request " + requestObjectName);

	if (requestObjectName.length() < policyObjectName.length()) {
	    logger.debug("Less chars");
	    return false;
	}

	int index = policyObjectName.length();
	if (policyObjectName.endsWith(fileDelimiter + wildCardChar)) {
	    logger.debug("policy object name ends with /*");
	    index = 
		policyObjectName.lastIndexOf(fileDelimiter + wildCardChar);
	    logger.debug("index is " + index);
	    requestObjectName = requestObjectName.substring(0, index);
	    logger.debug("altered request obj Name " + requestObjectName);
	}
	
	if (!requestObjectName.equals(policyObjectName.substring(0, index))) {
	    logger.debug("leading part is not equal");
	    return false;
	}
	return true;
    }

    public  boolean objectExists(ObjectData reqObj) 
	throws CasDBException {

	logger.debug("object exists " + reqObj.getObjectName());

	// get all objects that match the name space of the reqObj
	String query = "select" + CasDBConstants.COL_OBJECT_NAME + " from" 
	    + CasDBConstants.TABLE_OBJECT + " where" 
	    + CasDBConstants.COL_NS_NICKNAME + "='"
	    + reqObj.getObjectNamespace() + "'";
	Connection connection = CasDBStorage.getDBConnection();
	Vector objectName = null;
	Statement statement = null;
	ResultSet resultSet = null;
	try {
	    statement = connection.createStatement();
	    resultSet = statement.executeQuery(query);
	    if ((resultSet!=null) && (resultSet.next())) {
		objectName = new Vector(resultSet.getFetchSize());
		do {
		    objectName.add(resultSet.getString(
				   CasDBConstants.COL_OBJECT_NAME.trim()));
		} while (resultSet.next());
	    }
	}
	catch (SQLException exp) {
            logger.error(i18n.getMessage("existenceErr", new Object[]
                                         { "object", query} ), 
                         exp);
            throw new CasDBException(i18n.getMessage("objExistenceErr", 
                                                     new Object[] {
                                                         "object", "" })
                                     + exp.getMessage(), exp);
	}
	finally {
	    // returning connection irrespective of whether stmt
	    // and result set are closed or not.
	    CasDBStorage.returnDBConnection(connection);
	    try {
		if (resultSet != null)
		    resultSet.close();
		if (statement != null)
		    statement.close();
	    }
	    catch (SQLException exp) {
                logger.warn(i18n.getMessage("existenceCloseErr", "object"),
                            exp);
	    }
	}

	if (objectName == null) {
	    logger.debug("No object in this namesapce");
	    return false;
	}

	String reqObjName = reqObj.getObjectName();
	for (int i=0; i<objectName.size(); i++) {
	    // if a match is found return true
	    if (matchObjectNames( reqObjName, (String)objectName.get(i))) {
		logger.debug("requestName, objName matched");
		return true;
	    }
	    if (matchObjectNames((String)objectName.get(i), reqObjName)) {
		logger.debug("objName, requestName matched");
		return true;
	    }
	}
	return false;
    }

    private Vector getObjectsInNamespace(String namespace) 
	throws CasDBException {

	// get all objects that match the name space of the objData
	String query = "select * from" + CasDBConstants.TABLE_OBJECT 
            + " where" + CasDBConstants.COL_NS_NICKNAME + "='" + namespace 
            + "'";
	logger.debug("Query is " + query);
	Connection connection = CasDBStorage.getDBConnection();
	Vector objects = null;
	Statement statement = null;
	ResultSet resultSet = null;
	try {
	    statement = connection.createStatement();
	    resultSet = statement.executeQuery(query);
	    if ((resultSet!=null) && (resultSet.next())) {
		objects = new Vector(resultSet.getFetchSize());
		do {
		    ObjectData objData = new ObjectData();
		    objData.setObjectId(resultSet.getString(
					CasDBConstants.COL_OBJECT_ID.trim()));
		    objData.setObjectName(resultSet.getString(
					 CasDBConstants.COL_OBJECT_NAME.trim()));
		    objData.setObjectNamespace(resultSet.getString(
					 CasDBConstants.COL_NS_NICKNAME.trim()));
		    objects.add(objData);
		} while (resultSet.next());
	    }
	}
	catch (SQLException exp) {
            logger.error(i18n.getMessage("existenceErr", new Object[]
                                         { "object", query} ), 
                         exp);
            throw new CasDBException(i18n.getMessage("objExistenceErr", 
                                                     new Object[] {
                                                         "object", "" })
                                     + exp.getMessage(), exp);
	}
	finally {
	    // returning connection irrespective of whether stmt
	    // and result set are closed or not.
	    CasDBStorage.returnDBConnection(connection);
	    try {
		if (resultSet != null)
		    resultSet.close();
		if (statement != null)
		    statement.close();
	    }
	    catch (SQLException exp) {
                logger.warn(i18n.getMessage("existenceCloseErr", "object"),
                            exp);
	    }
	}
	return objects;
    }
}
