/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.cas.impl.service;

import org.globus.cas.faults.CasFault;
import org.globus.cas.faults.NoPermissionFault;

import org.globus.cas.types.UserData;
import org.globus.cas.types.ObjectData;
import org.globus.cas.types.PolicyData;
import org.globus.cas.types.UserGroupData;
import org.globus.cas.types.ObjectGroupData;
import org.globus.cas.types.TrustAnchorData;
import org.globus.cas.types.NamespaceData;
import org.globus.cas.types.ServiceTypeData;
import org.globus.cas.types.CasObjectData;

import org.globus.cas.types.AddUser;
import org.globus.cas.types.CreateGroup;
import org.globus.cas.types.DeleteGroup;
import org.globus.cas.types.CreateObject;
import org.globus.cas.types.DeleteObject;
import org.globus.cas.types.AddTrustAnchor;
import org.globus.cas.types.PolicyDetails;
import org.globus.cas.types.PolicyResponse;
import org.globus.cas.types.ManageUserGroups;
import org.globus.cas.types.ManageObjectGroups;
import org.globus.cas.types.ManageServiceAction;
import org.globus.cas.types.CreateServiceType;
import org.globus.cas.types.CreateObjectNamespace;
import org.globus.cas.types.ManageServiceActionGroups;

import org.globus.cas.impl.CasConstants;

import org.globus.cas.impl.databaseAccess.CasDBException;
import org.globus.cas.impl.databaseAccess.UserDataHandler;
import org.globus.cas.impl.databaseAccess.PolicyDataHandler;
import org.globus.cas.impl.databaseAccess.ObjectDataHandler;
import org.globus.cas.impl.databaseAccess.PermissionsHandler;
import org.globus.cas.impl.databaseAccess.NamespaceDataHandler;
import org.globus.cas.impl.databaseAccess.UserGroupDataHandler;
import org.globus.cas.impl.databaseAccess.ServiceTypeDataHandler;
import org.globus.cas.impl.databaseAccess.TrustAnchorDataHandler;
import org.globus.cas.impl.databaseAccess.ObjectGroupDataHandler;
import org.globus.cas.impl.databaseAccess.ServiceTypeActionHandler;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.util.I18n;

/**
 * Implementation of CAS Adminstrator interface
 */
public class CasAdminPortImpl {
    
    static Log logger = LogFactory.getLog(CasAdminPortImpl.class.getName() );

    private static I18n i18n =
        I18n.getI18n("org.globus.cas.impl.service.errors",
                     CasAdminPortImpl.class.getClassLoader());

    /**
     * Method to add user on this CAS Server <p>
     * To enroll user cas/enroll_User permission is needed.
     *
     * @param userGpName
     *        user group to grant all rights for this object
     * @param nickname
     *        userNickname
     * @param subjectDN
     *        subjectName
     * @param trustAnchorNickname
     *         trust anchor nick name
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void addUser(AddUser addUser)

        throws CasFault, NoPermissionFault {

        String userGpName = addUser.getUserGpName();
        String nickname = addUser.getNickname();
        String subjectDN = addUser.getSubjectDN();
        String trustAnchorNickname = addUser.getTrustAnchorNickname();
        
        String baseErrMsg = i18n.getMessage("addUser");
        logger.debug("addUser " + userGpName + " " + nickname 
                     + " " + subjectDN + " " + trustAnchorNickname);
        if ((userGpName == null) || (nickname == null) || 
            (subjectDN == null) || (trustAnchorNickname == null)) {
            String err = i18n.getMessage("allParamErr");
            logger.error(err);
            throw CasService.makeFault(baseErrMsg + err);
        }

        // policy check: if userGroup has cas/enroll_user
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canEnrollUser(userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }

            // check if the group specified for grant all is valid.
            checkUserGroupExists(userGpName, baseErrMsg);
            
            UserData userData = new UserData();
            userData.setNickname(nickname.trim());
            userData.setSubjectName(subjectDN.trim());
            userData.setTrustAnchorName(trustAnchorNickname.trim());
            UserDataHandler.storeObject(userData);
            
            PermissionsHandler.grantAllPermission(userGpName, nickname, 
                                                  CasConstants.USER_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }
    
    /**
     * Method to add trust anchor on this CAS Server <p>
     * To enroll trust anchor cas/enroll_TrustAnchor permission is needed.
     *
     * @param userGpName
     *        user group to grant all rights for this object
     * @param nickname
     *        trust anchor nickname
     * @param authMethod
     *        authentication method
     * @param authData
     *         authentication data
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void addTrustAnchor(AddTrustAnchor addTrustAnchor) 
        throws CasFault, NoPermissionFault {
        
        String userGpName = addTrustAnchor.getUserGpName();
        String nickname = addTrustAnchor.getNickname();
        String authMethod = addTrustAnchor.getAuthMethod();
        String authData = addTrustAnchor.getAuthData();
        
        String baseErrMsg = i18n.getMessage("addTrustAnchor");
        logger.debug("addTrustAnchor " + userGpName + " " + nickname 
                     + " " + authMethod + " " + authData);
        if ((userGpName == null) || (nickname == null) || 
            (authMethod == null) || (authData == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }

        // policy check: if userGroup has cas/enroll_TrustAnchor
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canEnrollTrustAnchor(userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            // check if the group specified for grant all is valid.
            checkUserGroupExists(userGpName, baseErrMsg);
            
            TrustAnchorData trustAnchorData = new TrustAnchorData();
            trustAnchorData.setNickname(nickname.trim());
            trustAnchorData.setAuthMethod(authMethod.trim());
            trustAnchorData.setAuthData(authData.trim());               
            TrustAnchorDataHandler.storeObject(trustAnchorData);
            
            PermissionsHandler.grantAllPermission(
                               userGpName, nickname, 
                               CasConstants.TRUSTANCHOR_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method to remove user on this CAS Server <p>
     * This operation is permitted only if cas/unenroll permission 
     * on nickname has been granted 
     *
     * @param nickname
     *         user nickname
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void removeUser(String nickname) 
        throws CasFault, NoPermissionFault {
        
        String baseErrMsg = i18n.getMessage("removeUser");
        logger.debug("removeUser " + nickname);
        
        if (nickname == null) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }

        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canUnenrollUser(userName, nickname)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            UserDataHandler.deleteObject(nickname);
            PolicyDataHandler.deletePolicyForObject(nickname, 
                                                    CasConstants.USER_SPEC);
        }
        catch (CasDBException ex) {
            logger.error(baseErrMsg, ex);
            throw CasService.makeFault(baseErrMsg + ex.getMessage(), ex);
        }
    }
    
    /**
     * Method to remove trust anchor on this CAS Server <p>
     * This operation is permitted only if cas/unenroll permission 
     * on nickname has been granted 
     *
     * @param nickname
     *        trust anchor nickname
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void removeTrustAnchor(String nickname) 
        throws CasFault, NoPermissionFault {
        
        String baseErrMsg = i18n.getMessage("removeTrustAnchor");
        logger.debug("removeTrustAnchor " + nickname);
        
        if (nickname == null) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }

        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canUnenrollTrustAnchor(userName, 
                                                           nickname)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            TrustAnchorDataHandler.deleteObject(nickname);
            PolicyDataHandler.deletePolicyForObject(nickname, 
                              CasConstants.TRUSTANCHOR_SPEC);
        }
        catch (CasDBException ex) {
            logger.error(baseErrMsg, ex);
            throw CasService.makeFault(baseErrMsg + ex.getMessage(), ex);
        }
    }

    /**
     * Method to create new user, object or serviceAction group<p>
     * This operation is permitted only if the requestor has 
     * cas/create_user_group or cas/create_object_group or 
     * create_serviceAction_group on casServer
     *
     * @param groupType
     *        Type of group: "user" or "object" or "serviceAction"
     * @param userGpName
     *        user group to grant all rights for this object
     * @param groupName
     *        name of the group to be created.
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void createGroup(CreateGroup createGroup)
        throws CasFault, NoPermissionFault {
        
        String groupType = createGroup.getGroupType();
        String userGpName = createGroup.getUserGpName();
        String groupName = createGroup.getGroupName();

        logger.debug("createGroup " + groupType + groupName);
        String baseErrMsg = i18n.getMessage("createGroup", groupType);
        if ((groupType == null) || (userGpName == null) || 
            (groupName == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canCreateGps(groupType,userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            // check if the group specified for grant all is valid.
            //throw error only if this is false and the group specified
            // for grant all is not the new user group being created.
            if (!((groupType.trim().equals(CasConstants.USER_SPEC)) 
                  && (groupName.trim().equals(userGpName.trim())))) {
                checkUserGroupExists(userGpName, baseErrMsg);
            }

            if (groupType.trim().equals(CasConstants.USER_SPEC)) {
                logger.debug("create user group");
                UserGroupData userGpData = new UserGroupData();
                userGpData.setGroupName(groupName.trim());
                UserGroupDataHandler.storeObject(userGpData);
                // If userGpName is the same as the new group, 
                // add the user to the new group
                if (groupName.trim().equals(userGpName.trim())) {
                    logger.debug("grantAll on new user group, hence add user "
                                 + " to that group");
                    UserGroupDataHandler.addGroupMember(groupName, userName);
                }
                PermissionsHandler.grantAllPermission(userGpName, groupName, 
                                   CasConstants.USERGP_SPEC);
            }
            else if (groupType.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("create object group");
                ObjectGroupData objGpData = new ObjectGroupData();
                objGpData.setGroupName(groupName.trim());
                ObjectGroupDataHandler.storeObject(objGpData);      
                PermissionsHandler.grantAllPermission(userGpName, groupName, 
                                   CasConstants.OBJECTGP_SPEC);
            } 
            else if (groupType.trim().equals(
                                      CasConstants.SERVICEACTION_SPEC)) {
                logger.debug("create service Action group");
                ServiceTypeActionHandler.createServiceActionGroup(
                                         groupName.trim());
                PermissionsHandler.grantAllPermission(userGpName, groupName, 
                                   CasConstants.SERVICEACTIONGP_SPEC);
            } else {
                String er = i18n.getMessage("errSpecType", new Object[] {
                                            CasConstants.USER_SPEC,
                                            CasConstants.OBJECT_SPEC,
                                            CasConstants.SERVICEACTION_SPEC});
                logger.error(er);
                throw CasService.makeFault(baseErrMsg + er);
            }
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method to remove a user, object or serviceAction group<p>
     * This operation is permitted only if the requestor has 
     * cas/delete_user_group or cas/delete_object_group or 
     * delete_serviceAction_group on casServer
     *
     * @param groupType 
     *        Type of group: "user" or "object" or "serviceAction"
     * @param  groupName
     *         name of group to be deleted.
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void deleteGroup(DeleteGroup deleteGroup)
        throws CasFault, NoPermissionFault {
        
        String groupType = deleteGroup.getGroupType();
        String groupName = deleteGroup.getGroupName();
        
        String baseMethErr = i18n.getMessage("deleteGroup", groupName);
        logger.debug("deleteGroup " + groupType + " " + groupName);
        
        if ((groupType == null) || (groupName == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseMethErr + er);
        }

        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canDeleteGps(groupType,userName, 
                                                 groupName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseMethErr + er);
            }   
            String groupTypeForDel = null;
            if (groupType.trim().equals(CasConstants.USER_SPEC)) {
                logger.debug("delete user group");
                UserGroupDataHandler.deleteObject(groupName);
                groupTypeForDel = CasConstants.USERGP_SPEC;
            } else if (groupType.trim().equals(CasConstants.OBJECT_SPEC)) {
                logger.debug("delete object group");
                ObjectGroupDataHandler.deleteObject(groupName);     
                groupTypeForDel = CasConstants.OBJECTGP_SPEC;
            } else if (groupType.trim().equals(
                                        CasConstants.SERVICEACTION_SPEC)) {
                logger.debug("delete service Action group");
                ServiceTypeActionHandler.deleteServiceActionGroup(groupName);
                groupTypeForDel = CasConstants.SERVICEACTIONGP_SPEC;
            } else {
                String er = i18n.getMessage("errSpecType", new Object[] {
                    CasConstants.USER_SPEC , CasConstants.OBJECT_SPEC,
                    CasConstants.SERVICEACTION_SPEC });
                logger.error(er);
                throw CasService.makeFault(baseMethErr + er);
            }
            PolicyDataHandler.deletePolicyForObject(groupName, 
                                                    groupTypeForDel);
        }
        catch (CasDBException exp) {
            logger.error(baseMethErr, exp);
            throw CasService.makeFault(baseMethErr + exp.getMessage(), exp);
        }
    }
    
    /**
     * Method to add namespace<p>
     * This operation is permitted only if the requestor has 
     * cas/enroll_namespace permissions on casServer
     *
     * @param userGpName
     *        user group to grant all rights for this object
     * @param nickName
     *        namespace nickname
     * @param baseName
     *        the base url for objects in namespace
     * @param compAlg
     *        comparison algorithm class that needs to be used for 
     *        objects in this namespace
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void createObjectNamespace(CreateObjectNamespace createObjNS)
        throws CasFault, NoPermissionFault {

        String userGpName = createObjNS.getUserGpName();
        String nickname = createObjNS.getNickname();
        String baseName = createObjNS.getBaseName();
        String compAlg = createObjNS.getComparisonAlg();
        
        String baseErrMsg = i18n.getMessage("nsEnroll");
        logger.debug("addNamespace " + userGpName + " " + nickname + " "
                     + baseName + " " + compAlg);
        
        if ((userGpName == null) || (nickname == null) || (baseName == null) 
            || (compAlg == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        
        // policy check: if userGroup has cas/enroll_namespace
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canCreateNamespace(userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            checkUserGroupExists(userGpName, baseErrMsg);

            NamespaceData nsData = new NamespaceData();
            nsData.setNickname(nickname.trim());
            nsData.setBasename(baseName.trim());
            nsData.setComparisonAlg(compAlg.trim());
            NamespaceDataHandler.storeObject(nsData);
            PermissionsHandler.grantAllPermission(userGpName, nickname, 
                               CasConstants.NAMESPACE_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }
    
    /**
     * Method to remove namespace<p>
     * This method is permitted only if the requestor has 
     * cas/unenroll_user permission on the namespace nickname
     *
     * @param nickname
     *        namespace nickname
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void deleteObjectNamespace(String nickname) 
        throws CasFault, NoPermissionFault {

        String baseErrMsg = i18n.getMessage("nsUnenroll");
        logger.debug("removeNamespace " + nickname);
        if (nickname == null) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
        try {
            // policy check: if userGroup has cas/unenroll on nickname
            String userName = CasService.getCallerNickname();
            if (!PermissionsHandler.canDeleteNamespace(userName, nickname)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            NamespaceDataHandler.deleteObject(nickname);
            PolicyDataHandler.deletePolicyForObject(nickname, 
                              CasConstants.NAMESPACE_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }
    
    /**
     * Method to add an object<p>
     * This operation is permitted only if the requestor has cas/enroll_object 
     * permissions on casServer
     *
     * @param userGpName
     *        user group to grant all rights for this object
     * @param objectName
     *        object name 
     * @param objectNamespace
     *        object namespace
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void createObject(CreateObject createObj)
        throws CasFault, NoPermissionFault {

        String userGpName = createObj.getUserGpName();
        String objectName = createObj.getObjectName();
        String objectNamespace = createObj.getObjectNamespace();

        String baseErrMsg = i18n.getMessage("createObj");
        logger.debug("createObject " + userGpName + objectName 
                     + objectNamespace);
        
        if ((userGpName == null) || (objectName == null) || 
            (objectNamespace == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        String userName = CasService.getCallerNickname();
        // policy check: if userGroup has cas/enroll_object
        try {
            if (!PermissionsHandler.canCreateObject(userName, objectNamespace, 
                                                     objectName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }

            // user group existence check
            checkUserGroupExists(userGpName, baseErrMsg);

            ObjectData objData = new ObjectData();
            objData.setObjectName(objectName.trim());
            objData.setObjectNamespace(objectNamespace.trim());
            ObjectDataHandler.storeObject(objData);
            String objId = Integer.toString(
                                   ObjectDataHandler.getObjectId(
                                                     objectName.trim(),
                                                     objectNamespace.trim()));
            PermissionsHandler.grantAllPermission(userGpName, objId, 
                                                  CasConstants.OBJECT_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }
    
    /**
     * Method to remove an object<p>
     * This operation is permitted only if the requestor has cas/unenroll 
     * permissions on objectId
     *
     * @param objectName
     *        object name
     * @param objectNamespace
     *        object namespace
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void deleteObject(DeleteObject deleteObj)
        throws CasFault, NoPermissionFault {

        String objectName = deleteObj.getObjectName();
        String objectNamespace = deleteObj.getObjectNamespace();
        
        String baseErrMsg = i18n.getMessage("deleteObj");
        logger.debug("deleteObject " + objectName + " " + objectNamespace);
        
        if ((objectName == null) || (objectNamespace == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
        
        String userName = CasService.getCallerNickname();
        try {
            // policy check: if userGroup has cas/unenroll on objectId
            if (!PermissionsHandler.canRemoveObject(userName, objectName, 
                                                    objectNamespace)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            int objectId = ObjectDataHandler.getObjectId(objectName, 
                                                         objectNamespace);
            ObjectDataHandler.deleteObject(objectName, objectNamespace);
            PolicyDataHandler.deletePolicyForObject(Integer.toString(objectId),
                                                    CasConstants.OBJECT_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method to add or remove an object from object group<p>
     * Add operation is permitted only if requestor has cas/create_group_entry 
     * permission on the groupName<br>
     * Remove operation is permitted only if requestor has 
     * cas/delete_group_entry permission on the groupName
     *
     * @param requestType
     *        "add" or "remove"
     * @param groupName
     *        object group name
     * @param objectSpec
     *        object name (could be a username, userGroupname, 
     *        objectNamespace|objectName, serviceType name, namespace nickName,
     *        trustAnchor nickname)
     * @param objectDesc
     *         "user" or "userGroup" or "object" or "serviceType" or 
     *         "namespace" or "trustAnchor"
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void manageObjectGroups(ManageObjectGroups manageObjGps)
        throws CasFault, NoPermissionFault {

        String requestType = manageObjGps.getRequestType();
        String groupName = manageObjGps.getGroupName();
        String objectSpec = manageObjGps.getObjectSpec();
        String objectDesc = manageObjGps.getObjectSpecDesc();

        String baseErrMsg = i18n.getMessage("gpOperationErr", requestType);
        logger.debug("objectGroupHandling " + requestType + " " + groupName 
            + " " + objectSpec + " " + objectDesc);
        
        if ((requestType == null) || (groupName == null) || 
            (objectSpec == null) || (objectDesc == null)) {

            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canManipObjectGps(requestType, userName, 
                                                      groupName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
        }
        catch (CasDBException e) {
            String er = i18n.getMessage("permAscErr");
            logger.error(er, e);
            throw CasService.makeFault(baseErrMsg + er + e.getMessage(), e);
        }
        if (objectDesc.equals(CasConstants.OBJECT_SPEC)) {
            logger.debug("object - get object id");
            objectSpec = CasService.getObjectId(objectSpec);
        }

        if (requestType.trim().equals(CasConstants.ADD_OPERATION)) {
            logger.debug("add object group entry ");
            try {
                ObjectGroupDataHandler.addGroupMember(groupName, objectSpec, 
                                                      objectDesc);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        }
        else if (requestType.trim().equals(CasConstants.REMOVE_OPERATION)) {
            logger.debug("remove object group entry ");
            try {
                ObjectGroupDataHandler.removeGroupMember(groupName, objectSpec,
                                                         objectDesc);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        } else {
            String er = i18n.getMessage("opTypeErr", new Object[] {
                CasConstants.ADD_OPERATION, CasConstants.REMOVE_OPERATION});
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
    }

    /**
     * Method to add or remove a user from user group<p>
     * Add operation is permitted only if reuqestor has cas/create_group_entry 
     * permission on the groupName<br>
     * Remove operation is permitted only if requestor has 
     * cas/delete_group_entry permission on the groupName
     * 
     * @param requestType
     *        "add" or "remove"
     * @param groupName
     *        user group name
     * @param userNickname
     *        nickname of the user that needs to be removed from the group
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void manageUserGroups(ManageUserGroups manageUserGps)
        throws CasFault, NoPermissionFault {

        String requestType = manageUserGps.getRequestType();
        String groupName = manageUserGps.getGroupName();
        String userNickname = manageUserGps.getUserNickname();

        String baseErrMsg = i18n.getMessage("userGpErr", requestType);
        logger.debug("userGroupHandling " + requestType + " " + groupName 
                     + " " + userNickname);

        if ((requestType == null) || (groupName == null) || 
            (userNickname == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canManipUserGps(requestType, userName, 
                                                    groupName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
        }
        catch (CasDBException e) {
            logger.error(baseErrMsg, e);
            throw CasService.makeFault(baseErrMsg + e.getMessage(), e);
        }

        if (requestType.trim().equals(CasConstants.ADD_OPERATION)) {
            logger.debug("add user group entry ");
            try {
                UserGroupDataHandler.addGroupMember(groupName, userNickname);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        }
        else if (requestType.trim().equals(CasConstants.REMOVE_OPERATION)) {
            logger.debug("remove user group entry ");
            try {
                UserGroupDataHandler.removeGroupMember(groupName, 
                                                       userNickname);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        } else {
            String er = i18n.getMessage("userGpErr", new Object[] {
                CasConstants.ADD_OPERATION, CasConstants.REMOVE_OPERATION});
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
    }

    /**
     * Method to create service type<p>
     * This operation is permitted only if the requestor has 
     * cas/create_serviceType permissions on the casServer.
     *
     * @param userGpName
     *        user group to grant all rights for this object
     * @param serviceTypeName
     *        name of service type
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void createServiceType(CreateServiceType createServiceType)
        throws CasFault, NoPermissionFault {

        String userGpName = createServiceType.getUserGpName();
        String serviceTypeName = createServiceType.getServiceTypeName();
        
        String baseMethErr = i18n.getMessage("serviceEnroll");
        logger.debug("serviceTypeHandling " + userGpName + " " 
                     + serviceTypeName);

        if ((userGpName == null) || (serviceTypeName == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseMethErr + er);
        }   
        
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canCreateServiceType(userName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseMethErr + er);
            }

            checkUserGroupExists(userGpName, baseMethErr);

            ServiceTypeData serviceTypeData = new ServiceTypeData();
            serviceTypeData.setName(serviceTypeName.trim());
            ServiceTypeDataHandler.storeObject(serviceTypeData);
            PermissionsHandler.grantAllPermission(
                               userGpName, serviceTypeName, 
                               CasConstants.SERVICETYPE_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseMethErr, exp);
            throw CasService.makeFault(baseMethErr + exp.getMessage(), exp);
        }
    }

    /**
     * Method to delete service type<p>
     * This operation is permitted only if the requestor has cas/unenroll 
     * permissions on the serviceTypeName
     *
     * @param serviceTypeName
     *        name of service type
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void deleteServiceType(String serviceTypeName) 
        throws CasFault, NoPermissionFault {

        String baseErrMsg = i18n.getMessage("serviceUnenroll");
        logger.debug("delete service type " + serviceTypeName);
        if (serviceTypeName == null) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canDeleteServiceType(userName, 
                                                         serviceTypeName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            ServiceTypeDataHandler.deleteObject(serviceTypeName);
            PolicyDataHandler.deletePolicyForObject(
                              serviceTypeName, 
                              CasConstants.SERVICETYPE_SPEC);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }    
    
    /**
     * Method to add or remove service type/action mapping<p>
     * Add operation is permitted only if requestor has cas/create_group_entry 
     * permission on the serviceTypeName<br>
     * Remove operation is permitted only if requestor has 
     * cas/delete_group_entry permission on the serviceTypeName
     *
     * @param requestType
     *        "add" or "remove"
     * @param serviceType
     *        service type name
     * @param actionName
     *        action name
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void manageServiceAction(ManageServiceAction manageServiceAction)
        throws CasFault, NoPermissionFault {

        String requestType = manageServiceAction.getRequestType();
        String serviceType = manageServiceAction.getServiceTypeName();
        String actionName = manageServiceAction.getActionName();
        
        String baseErrMsg = i18n.getMessage("serviceActionErr", requestType);
        logger.debug("Service type action mapping " + requestType + " " 
                     + serviceType + " "  + actionName);

        if ((requestType == null) || (serviceType == null) || 
            (actionName == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
        
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler
                .canManipServiceActionMapping(requestType, 
                                              userName,
                                              serviceType)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
        }
        catch (CasDBException exp) {
            logger.error("Permission check for service action mapping failed" 
                         , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
        
        if (requestType.trim().equals(CasConstants.ADD_OPERATION)) {
            logger.debug("add service type action mapping ");
            try {
                ServiceTypeActionHandler.addServiceActionMapping(serviceType, 
                                                                 actionName);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        }
        else if (requestType.trim().equals(CasConstants.REMOVE_OPERATION)) {
            logger.debug("remove service type action mapping ");
            try {
                ServiceTypeActionHandler
                    .removeServiceActionMapping(serviceType,
                                                actionName);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg, exp);
                CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        } else {
            String er = i18n.getMessage("opTypeErr", new Object[] {
                CasConstants.ADD_OPERATION, CasConstants.REMOVE_OPERATION});
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
    }

    /**
     * Method to add or remove serviceAction group entry<p>
     * Add operation is permitted only if requestor has cas/create_group_entry 
     * permission on the groupName<br>
     * Remove operation is permitted only if requestor has 
     * cas/delete_group_entry permission on the groupName
     *
     * @param requestType
     *        add or remove
     * @param groupName
     *        service action group name
     * @param serviceTypeName
     *        service type name
     * @param actionName
     *        action name
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void manageServiceActionGroups(ManageServiceActionGroups managerGp)
        throws CasFault, NoPermissionFault {
        
        String requestType = managerGp.getRequestType();
        String groupName = managerGp.getGroupName();
        String serviceTypeName = managerGp.getServiceTypeName();
        String actionName = managerGp.getActionName();

        String baseErrMsg = i18n.getMessage("serviceActionGpErr", requestType);
        logger.debug("serviceActionGpHandling " + requestType + " " + groupName
                     + " " + serviceTypeName + " " + actionName);
        
        if ((requestType == null) || (groupName == null) || 
            (serviceTypeName == null) || (actionName == null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   
        
        String userName = CasService.getCallerNickname();
        try {
            if (!PermissionsHandler.canManipServiceActionGps(requestType, 
                                                             userName, 
                                                             groupName)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
        }
        catch (CasDBException e) {
            logger.error(baseErrMsg, e);
            throw CasService.makeFault(baseErrMsg + e.getMessage(), e);
        }
        
        if (requestType.trim().equals(CasConstants.ADD_OPERATION)) {
            logger.debug("add service type action group entry ");
            try {
                ServiceTypeActionHandler.addServiceActionGpEntry(
                                         groupName, serviceTypeName, 
                                         actionName);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg , exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        }
        else if (requestType.trim().equals(CasConstants.REMOVE_OPERATION)) {
            logger.debug("remove service type action mapping ");
            try {
                ServiceTypeActionHandler.removeServiceActionGpEntry(
                                         groupName, serviceTypeName, 
                                         actionName);
            }
            catch (CasDBException exp) {
                logger.error(baseErrMsg , exp);
                throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
            }
        } else {
            String er = i18n.getMessage("opTypeErr", new Object[] {
                CasConstants.ADD_OPERATION, CasConstants.REMOVE_OPERATION});
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }
    }

    /**
     * Method to grant permissions 
     *
     * @param userGroupName
     *        name of the user group that this poicy applies to
     * @param objSpec
     *        String that represents the object on which the policy applies
     *        objectNamespace|object or userNickname or userGroupName or
     *        serviceTypeName or namespaceNickname or trustAnchorNickname
     *        or serviceActionGroupName
     * @param objectSpecDesc
     *        "object" or "user" or "userGroup" or "serviceType" or
     * "Namespace" or "TrustAnchor" or "serviceActionGroup"
     * @param actionSpec
     *        String  represents action which the policy allows. 
     *        <serviceName>/<actionName> or serviceActionGroupName
     * @param actionSpecDesc
     *        "serviceAction" or "serviceActionGroup"
     * @return CasObject that represents the policy
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public PolicyResponse grant(PolicyDetails policyDetails)
        throws CasFault, NoPermissionFault {
        
        String userGroupName = policyDetails.getUserGpName();
        String objSpec = policyDetails.getObjectSpec();
        String objSpecDesc = policyDetails.getObjectSpecDesc();
        String actionSpec = policyDetails.getActionSpec();
        String actionSpecDesc = policyDetails.getActionSpecDesc();

        String baseErrMsg = i18n.getMessage("grantErr");

        logger.debug("grant called with " + userGroupName);
        if ((userGroupName == null) || (objSpec == null) || 
            (objSpecDesc == null) || (actionSpec == null) || 
            (actionSpecDesc ==null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   

        if (objSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            logger.debug("object - get object id");
            objSpec = CasService.getObjectId(objSpec);
        }

        if (actionSpecDesc.equals(CasConstants.SERVICEACTION_SPEC)) {
            logger.debug("service action - get service action id");
            actionSpec = CasService.getServiceActionId(actionSpec);
        }

        try {
            String userName = CasService.getCallerNickname();
            if (!PermissionsHandler.canGrant(userName, userGroupName, objSpec, 
                                             objSpecDesc, actionSpec, 
                                             actionSpecDesc)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            
            CasObjectData obj = 
                PolicyDataHandler.grant(userGroupName, objSpec, objSpecDesc,
                                        actionSpec, actionSpecDesc);
            return new PolicyResponse(obj);
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    /**
     * Method to revoke permissions
     *
     * @param userGroupName
     *        name of the user group that this poicy applies to
     * @param objSpec
     *        String that represents the object on which the policy applies
     *        objectNamespace|object or userNickname or userGroupName or
     *        serviceTypeName or namespaceNickname or trustAnchorNickname
     *        or serviceActionGroupName
     * @param objectSpecDesc
     *        "object" or "user" or "userGroup" or "serviceType" or
     * "Namespace" or "TrustAnchor" or "serviceActionGroup"
     * @param actionSpec
     *        String  represents action which the policy allows. 
     *        <serviceName>/<actionName> or serviceActionGroupName
     * @param actionSpecDesc
     *        "serviceAction" or "serviceActionGroup"
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void revoke(PolicyDetails policyDetails)
        throws CasFault, NoPermissionFault {

        String userGroupName = policyDetails.getUserGpName();
        String objSpec = policyDetails.getObjectSpec();
        String objSpecDesc = policyDetails.getObjectSpecDesc();
        String actionSpec = policyDetails.getActionSpec();
        String actionSpecDesc = policyDetails.getActionSpecDesc();

        String baseErrMsg = i18n.getMessage("revokeErr");
        logger.debug("revoke called");
        
        if ((userGroupName == null) || (objSpec == null) || 
            (objSpecDesc == null) || (actionSpec == null) || 
            (actionSpecDesc ==null)) {
            String er = i18n.getMessage("allParamErr");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }   

        if (objSpecDesc.equals(CasConstants.OBJECT_SPEC)) {
            logger.debug("object - get object id");
            objSpec = CasService.getObjectId(objSpec);
        } else if 
            (!((objSpecDesc.equals(CasConstants.USER_SPEC)) || 
               (objSpecDesc.equals(CasConstants.OBJECTGP_SPEC)) || 
               (objSpecDesc.equals(CasConstants.SERVICETYPE_SPEC)) || 
               (objSpecDesc.equals(CasConstants.NAMESPACE_SPEC)) || 
               (objSpecDesc.equals(CasConstants.TRUSTANCHOR_SPEC)) || 
               (objSpecDesc.equals(CasConstants.SERVICEACTIONGP_SPEC)))) {
            String er = i18n.getMessage("badObjSpec");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }

        if (actionSpecDesc.equals(CasConstants.SERVICEACTION_SPEC)) {
            logger.debug("service action - get service action id");
            actionSpec = CasService.getServiceActionId(actionSpec);
        } else if 
            (! (actionSpecDesc.equals(CasConstants.SERVICEACTIONGP_SPEC))) {
            String er = i18n.getMessage("badActionSpec");
            logger.error(er);
            throw CasService.makeFault(baseErrMsg + er);
        }

        PolicyData policyData = new PolicyData();
        policyData.setUserGroupName(userGroupName);
        policyData.setObjectSpec(objSpec);
        policyData.setObjectSpecDesc(objSpecDesc);
        policyData.setActionSpec(actionSpec);
        policyData.setActionSpecDesc(actionSpecDesc);

        String policyId = null;
        try {
            policyId = PolicyDataHandler.getPolicyId(policyData);
            if (policyId == null) {
                String errMsg = i18n.getMessage("retrErr", "policy data");
                logger.error(errMsg);
                throw CasService.makeFault(baseErrMsg + errMsg);
            }
        }
        catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
        logger.debug("revokeForId is called with " + policyId);
        revokeForId(policyId);
    }

    /**
     * Method to revoke permissions
     *
     * @param policyId
     *        policyId that needs to be revoked
     * @exception NoPermissionFault if the client does not have 
     *            permission to perform this operation.
     * @exception CasFault if any other error occurs.
     */
    public void revokeForId(String policyId) throws CasFault, 
                                                    NoPermissionFault {
        String baseErrMsg = "Revoke rights failed.\n";
        logger.debug("revoke called with " + policyId);
        if (policyId == null) {
            logger.error("All parameters were not supplied");
            throw CasService.makeFault(baseErrMsg + "All parameters were not "
                                       + "supplied");
        }
        
        String userName = CasService.getCallerNickname();
        try {
            logger.debug("Policy id is " + policyId);
            if (!PermissionsHandler.canRevoke(userName, policyId)) {
                String er = i18n.getMessage("noPermErr", userName);
                logger.error(er);
                throw new NoPermissionFault(baseErrMsg + er);
            }
            PolicyDataHandler.deleteObject(policyId);
        }
        catch (CasDBException exp) {
            logger.error("revoke rights failed" , exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
    }

    // Checks of user group existence
    private void checkUserGroupExists(String userGpName, String baseErr) 
        throws CasFault {
        boolean exists;
        String baseErrMsg = i18n.getMessage("existenceChkFailed");
        try {
          exists = UserGroupDataHandler.groupExists(userGpName);
        } catch (CasDBException exp) {
            logger.error(baseErrMsg, exp);
            throw CasService.makeFault(baseErrMsg + exp.getMessage(), exp);
        }
        if (!exists) {
            String err = i18n.getMessage("doesNotExist", new Object[] {
                "UserGroup", userGpName });
            logger.error(err);
            throw CasService.makeFault(baseErr + err);
        }
    }
}
