/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.delegation.service;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.net.URL;
import java.security.cert.X509Certificate;

import javax.xml.namespace.QName;
import javax.xml.rpc.Stub;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.xml.sax.InputSource;

import org.globus.delegation.DelegationUtil;
import org.globus.delegationService.DelegationPortType;
import org.globus.delegationService.DelegationServiceAddressingLocator;
import org.globus.delegationService.test.DelegationTestPortType;
import org.globus.delegationService.test.DelegationTestServiceAddressingLocator;
import org.globus.delegationService.test.DelegationTestServiceLocator;
import org.globus.delegationService.test.VoidType;
import org.globus.gsi.GlobusCredential;
import org.globus.ws.trust.RequestSecurityTokenType;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.globus.wsrf.test.GridTestCase;
import org.globus.wsrf.utils.AddressingUtils;

public class DelegationServiceTest extends GridTestCase {

    static Log logger =
        LogFactory.getLog(DelegationServiceTest.class.getName());

    public static String DELEG_EPR_FILE_NAME = "delegationTestEpr";
    public static String TEST_EPR_FILE_NAME = "testServiceEpr";

    String SERVICE_NAME= "DelegationService";
    String FACTORY_NAME = "DelegationFactoryService";
    String TEST_SERVICE_NAME = "DelegationTestService";

    private static boolean postPackage = false;

    String factoryUrl;
    String serviceUrl;
    String testServiceUrl;
    int lifetime = 30000;
    GlobusCredential credential;
    X509Certificate[] certsToDelegateOn;
    X509Certificate certToSign;

    static String mechanism = null;
    static Object protection = null;

    public static void setPostPackage(boolean _postPackage) {
        postPackage = _postPackage;
    }

    public static void setParameters(String authObj, Object authVal) {
        mechanism = authObj;
        protection = authVal;
    }

    public DelegationServiceTest(String name) {
        super(name);
    }

    public void test() throws Exception {

        assertTrue(TEST_CONTAINER != null);
	factoryUrl = TEST_CONTAINER.getBaseURL() + FACTORY_NAME;
	serviceUrl = TEST_CONTAINER.getBaseURL() + SERVICE_NAME;
        testServiceUrl = TEST_CONTAINER.getBaseURL() + TEST_SERVICE_NAME;

        System.out.println("Test service " + testServiceUrl);

        // Credential to sign with
        credential = GlobusCredential.getDefaultCredential();

        // Get the public key to delegate on.
        logger.debug("Factory: "  + factoryUrl);
        EndpointReferenceType delegEpr =
            AddressingUtils.createEndpointReference(factoryUrl, null);
        logger.debug("EPR " + delegEpr);
        certsToDelegateOn =
            DelegationUtil.getCertificateChainRP(delegEpr, getClientSecDesc());

        logger.debug("Certs from server");
        for (int i=0; i<certsToDelegateOn.length; i++) {
            logger.debug("Certs " + certsToDelegateOn[i].getSubjectDN());
        }

        certToSign = certsToDelegateOn[0];

        if (!postPackage) {
            logger.debug("First test");
            EndpointReferenceType ref = delegate();
            testDelegationService(ref);
            // Store ref in a file.
            saveEprToFile(ref, DELEG_EPR_FILE_NAME);
        } else {
            logger.debug("Second test");
            EndpointReferenceType delegatedRef =
                readEprFromFile(DELEG_EPR_FILE_NAME);
            AttributedURI uri = delegatedRef.getAddress();
            uri.setPort(TEST_CONTAINER.getPort());

            delegatedRef.setAddress(uri);
            testDelegationService(delegatedRef);

            File file = new File(DELEG_EPR_FILE_NAME);
            file.delete();
            file = new File(TEST_EPR_FILE_NAME);
            file.delete();
        }
    }

    private EndpointReferenceType delegate() throws Exception {

        // Send it on the wire.
        ClientSecurityDescriptor desc = getClientSecDesc();
        EndpointReferenceType ref =
            DelegationUtil.delegate(factoryUrl, credential, certToSign,
                                    lifetime, false, desc);
        return ref;
    }

    private void testDelegationService(EndpointReferenceType delegatedRef)
        throws Exception {

        // Send EPR over to the test service

        DelegationTestServiceLocator testLocator =
            new DelegationTestServiceLocator();
        DelegationTestPortType testPort =
            testLocator.getDelegationTestPortTypePort(new URL(testServiceUrl));
        ((Stub)testPort)._setProperty(Constants.GSI_SEC_MSG,
                                      Constants.SIGNATURE);
        ((Stub)testPort)._setProperty(Constants.AUTHORIZATION,
                                      new NoAuthorization());
        // Registers listener for delegatedRef
        EndpointReferenceType testResource = testPort.create(delegatedRef);
        if (!postPackage) {
            saveEprToFile(testResource, TEST_EPR_FILE_NAME);
        }
        // Get refresh callback count
        DelegationTestServiceAddressingLocator testLocator1 =
            new DelegationTestServiceAddressingLocator();
        DelegationTestPortType testPort1 =
            testLocator1.getDelegationTestPortTypePort(testResource);
        ((Stub)testPort1)._setProperty(Constants.GSI_SEC_MSG,
                                       Constants.SIGNATURE);
        ((Stub)testPort1)._setProperty(Constants.AUTHORIZATION,
                                       new NoAuthorization());

        // On registering, the count is set to one
        assertTrue(testPort1.getCallbackCount(new VoidType()) == 1);

        // Attempt a refresh call on delegatedRef
        DelegationServiceAddressingLocator locator1 =
            new DelegationServiceAddressingLocator();

        DelegationPortType delegationPort1 = locator1
            .getDelegationPortTypePort((EndpointReferenceType)delegatedRef);

        RequestSecurityTokenType requestToken =
            DelegationUtil.getTokenToDelegate(credential, certToSign,
                                              lifetime, true);

        boolean exception = false;
        try {
            delegationPort1.refresh(requestToken);
        } catch (Exception exp) {
            exception = true;
        }
        assertTrue(exception);

        // Failed refresh, so call back on test resource remains one.
        assertTrue(testPort1.getCallbackCount(new VoidType()) == 1);

        // Refresh call with security properties
        DelegationUtil.refresh(credential, certToSign, lifetime, true,
                               getClientSecDesc(),
                               (EndpointReferenceType)delegatedRef);

        // Create and refresh on two separate test resource.
        assertTrue(testPort1.getCallbackCount(new VoidType()) == 2);

        if (postPackage) {

            EndpointReferenceType testRef =
                readEprFromFile(TEST_EPR_FILE_NAME);
            AttributedURI uri = testRef.getAddress();
            uri.setPort(TEST_CONTAINER.getPort());
            testRef.setAddress(uri);
            System.out.println("test ref " + testRef);

            DelegationTestServiceAddressingLocator testServiceLoc =
                new DelegationTestServiceAddressingLocator();
            testPort = testServiceLoc.getDelegationTestPortTypePort(testRef);
            ((Stub)testPort)._setProperty(Constants.GSI_SEC_MSG,
                                          Constants.SIGNATURE);
            ((Stub)testPort)._setProperty(Constants.AUTHORIZATION,
                                          new NoAuthorization());
            // Last state of test resource prior to persistence is 2
            // callbacks. But because of refresh call in the
            // postPackage case, increases callback to 3.
            assertTrue(testPort.getCallbackCount(new VoidType()) == 3);

            DelegationUtil.refresh(credential, certToSign, lifetime, true,
                                   getClientSecDesc(),
                                   delegatedRef);

            // register new listener
            DelegationTestServiceLocator newTestLocator =
                new DelegationTestServiceLocator();
            DelegationTestPortType newTestPort = newTestLocator
                .getDelegationTestPortTypePort(new URL(testServiceUrl));
            ((Stub)newTestPort)._setProperty(Constants.GSI_SEC_MSG,
                                             Constants.SIGNATURE);
            ((Stub)newTestPort)._setProperty(Constants.AUTHORIZATION,
                                             new NoAuthorization());
            // resource home has counter at 2, so next two will fail.
            testFailure(newTestPort, delegatedRef);

            EndpointReferenceType newTestResource =
                newTestPort.create(delegatedRef);

            // Get refresh callback count
            DelegationTestServiceAddressingLocator newTestLocator1 =
                new DelegationTestServiceAddressingLocator();
            DelegationTestPortType newTestPort1 =
                newTestLocator1.getDelegationTestPortTypePort(newTestResource);
            ((Stub)newTestPort1)._setProperty(Constants.GSI_SEC_MSG,
                                              Constants.SIGNATURE);
            ((Stub)newTestPort1)._setProperty(Constants.AUTHORIZATION,
                                              new NoAuthorization());

            assertTrue(newTestPort1.getCallbackCount(new VoidType()) == 1);

            // this should unregister listener
            assertTrue(testPort.getCallbackCount(new VoidType()) == 4);

            DelegationUtil.refresh(credential, certToSign, lifetime, true,
                                   getClientSecDesc(),
                                   delegatedRef);

            assertTrue(newTestPort1.getCallbackCount(new VoidType()) == 2);

            DelegationUtil.refresh(credential, certToSign, lifetime, true,
                                   getClientSecDesc(),
                                   delegatedRef);

            // This call will try to unregister a listener id that
            // does not exist. Just a warning should be printed            
            assertTrue(testPort.getCallbackCount(new VoidType()) == 4);
            assertTrue(newTestPort1.getCallbackCount(new VoidType()) == 3);
        } else {
            testFailure(testPort, delegatedRef);
        }
    }

    private void testFailure(DelegationTestPortType testPort,
                             EndpointReferenceType delegatedRef)
        throws Exception {

        // second create of test resource, that attempts a register  -
        // should fail - subject null
        boolean exp = false;
        try {
            testPort.create(delegatedRef);
            fail("Did not throw exception as expected");
        } catch (Exception e) {
            if (e.getMessage().indexOf("Unable to retrieve caller") != -1) {
                exp = true;
            }
            else {
                e.printStackTrace();
            }
        }
        assertTrue(exp);
        
        // third create of test resource, that attempts a register  -
        // should fail - subject not allowed
        exp = false;
        try {
            testPort.create(delegatedRef);
            fail("Did not throw exception as expected");
        } catch (Exception e) {
            if (e.getMessage()
                .indexOf("is not authorized to access this resource")
                != -1) {
                exp = true;
            }
            else {
                e.printStackTrace();
            }
        }
        assertTrue(exp);

        // Call should go thro'
        testPort.create(delegatedRef);
    }

    private void saveEprToFile(EndpointReferenceType ref, String fileName)
        throws Exception {
        String eprStr = ObjectSerializer.toString(ref,
                                                  new QName("", "EPR"));
        FileWriter writer = null;
        try {
            writer = new FileWriter(fileName);
            writer.write(eprStr);
        } finally {
            if (writer != null) {
                try { writer.close(); } catch (Exception e) {}
            }
        }
    }

    private EndpointReferenceType readEprFromFile(String fileName)
        throws Exception {
        FileInputStream in = null;
        EndpointReferenceType ref = null;
        try {
            in = new FileInputStream(fileName);
            ref = (EndpointReferenceType)ObjectDeserializer
                .deserialize(new InputSource(in),
                             EndpointReferenceType.class);
        } finally {
            if (in != null) {
                try { in.close(); } catch (Exception e) {}
            }
        }
        return ref;
    }

    private ClientSecurityDescriptor getClientSecDesc() {
        ClientSecurityDescriptor desc = new ClientSecurityDescriptor();
        if (mechanism.equals(Constants.GSI_SEC_MSG)) {
            desc.setGSISecureMsg((Integer)protection);
        } else if (mechanism.equals(Constants.GSI_SEC_CONV)) {
            desc.setGSISecureConv((Integer)protection);
        } else if (mechanism.equals(Constants.GSI_TRANSPORT)) {
            desc.setGSITransport((Integer)protection);
        }
        System.out.println("Desc " + mechanism);
        desc.setAuthz(new NoAuthorization());
        return desc;
    }
}
