/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_i_globusrun_ws.h"
#include "DelegationFactoryService_client.h"
#include "DelegationService_client.h"
#include "globus_delegation_client_util.h"
#include "globusrunwsAnnotationType.h"
#include "wsgram_delegationFactoryEndpoint.h"
#include "wsgram_stagingDelegationFactoryEndpoint.h"

static xsd_QName globus_l_globusrunAnnotation_qname = 
{
    GLOBUSRUNWSANNOTATIONTYPE_NS,
    "globusrunAnnotation"
};

typedef enum
{
    GLOBUS_L_JOB_CREDENTIAL = 1,
    GLOBUS_L_STAGING_CREDENTIAL,
    GLOBUS_L_STAGEIN_CREDENTIAL,
    GLOBUS_L_STAGEOUT_CREDENTIAL,
    GLOBUS_L_CLEANUP_CREDENTIAL
} globus_l_delegation_type_t;

typedef struct
{
    globus_i_info_t *                   info;
    globus_mutex_t                      lock;
    globus_cond_t                       cond;
    int                                 cb_count;
    globus_object_t *                   error;
    globus_list_t *                     active_queries;
    globus_list_t *                     active_delegations;
    globus_list_t *                     active_cleanups;
    globus_hashtable_t                  factory_epr_cache;
    globus_hashtable_t                  delegated_epr_cache;
} globus_l_monitor_t;

typedef struct
{
    globus_bool_t                       full_proxy;
    globus_l_delegation_type_t          type;
    wsa_EndpointReferenceType **        cred_epr;
    wsgram_ExtensionsType **            extensions;
} globus_l_delegation_query_operation_t;

typedef struct
{
    globus_l_delegation_type_t          type;
    wsa_EndpointReferenceType **        cred_epr;
    wsgram_ExtensionsType **            extensions;
} globus_l_delegation_operation_t;

typedef struct
{
    char *                              query_endpoint;
    wsa_EndpointReferenceType           job_factory_epr;
    wsa_EndpointReferenceType           staging_factory_epr;
    globus_bool_t                       ready;
    globus_list_t *                     target_list;
    globus_l_monitor_t *                monitor;
} globus_l_factory_epr_cache_entry_t;

typedef struct
{
    char *                              factory_endpoint;
    globus_bool_t                       full;
} globus_l_delegated_epr_cache_key_t;

typedef struct
{
    globus_l_delegated_epr_cache_key_t  key;
    globus_bool_t                       ready;
    wsa_EndpointReferenceType           delegated_epr;
    globus_list_t *                     target_list;
    globus_l_monitor_t *                monitor;
} globus_l_delegated_epr_cache_entry_t;

static
int
globus_l_delegated_epr_cache_hash(
    void *                              inkey,
    int                                 limit)
{
    globus_l_delegated_epr_cache_key_t *key;
    
    key = (globus_l_delegated_epr_cache_key_t *) inkey;
    
    return globus_hashtable_string_hash(key->factory_endpoint, limit);
}

static
int
globus_l_delegated_epr_cache_keyeq(
    void *                              inkey1,
    void *                              inkey2)
{
    globus_l_delegated_epr_cache_key_t *key1;
    globus_l_delegated_epr_cache_key_t *key2;
    
    key1 = (globus_l_delegated_epr_cache_key_t *) inkey1;
    key2 = (globus_l_delegated_epr_cache_key_t *) inkey2;
    
    if(key1->full == key2->full)
    {
        return globus_hashtable_string_keyeq(
            key1->factory_endpoint, key2->factory_endpoint);
    }
    
    return 0;
}

/* called locked */
static
void
globus_l_cancel_all(
    globus_l_monitor_t *                monitor)
{
    globus_list_t *                     tmp;
    
    for(tmp = monitor->active_queries;
        !globus_list_empty(tmp);
        tmp = globus_list_rest(tmp))
    {
        ManagedJobFactoryService_client_handle_t factory_handle;
        
        factory_handle = (ManagedJobFactoryService_client_handle_t)
            globus_list_first(tmp);
        
        ManagedJobFactoryService_client_operation_cancel(factory_handle);
    }
    
    for(tmp = monitor->active_delegations;
        !globus_list_empty(tmp);
        tmp = globus_list_rest(tmp))
    {
        globus_delegation_client_util_handle_t delegation_handle;
        
        delegation_handle = (globus_delegation_client_util_handle_t)
            globus_list_first(tmp);
        
        globus_delegation_client_util_cancel(delegation_handle);
    }
    
    for(tmp = monitor->active_cleanups;
        !globus_list_empty(tmp);
        tmp = globus_list_rest(tmp))
    {
        DelegationService_client_handle_t service_handle;
        
        service_handle = (DelegationService_client_handle_t)
            globus_list_first(tmp);
        
        DelegationService_client_operation_cancel(service_handle);
    }
}

static
globusrunwsAnnotationType *
globus_l_find_annotation(
    xsd_any_array *                     extensions)
{
    int                                 i;
    
    for(i = 0; i < extensions->length; i++)
    {
        if(extensions->elements[i].any_info == &globusrunwsAnnotationType_info)
        {
            return (globusrunwsAnnotationType *) extensions->elements[i].value;
        }
    }
    
    return NULL;
}

static
void
globus_l_delegate_cb(
    globus_delegation_client_util_handle_t handle,
    globus_result_t                     result,
    const wsa_EndpointReferenceType *   delegated_epr,
    void *                              user_arg)
{
    globus_l_delegation_operation_t *   operation;
    globus_l_delegated_epr_cache_entry_t *cache_entry;
    globus_l_monitor_t *                monitor;
    globus_bool_t                       first = GLOBUS_TRUE;
    
    cache_entry = (globus_l_delegated_epr_cache_entry_t *) user_arg;
    monitor = cache_entry->monitor;
    
    globus_mutex_lock(&monitor->lock);
    globus_list_remove(&monitor->active_delegations,
        globus_list_search(monitor->active_delegations, handle));
        
    if(result != GLOBUS_SUCCESS || monitor->error)
    {
        goto error;
    }
    
    /* XXX any errors happen here and i need to destroy the credential */
    result = wsa_EndpointReferenceType_copy_contents(
        &cache_entry->delegated_epr, delegated_epr);
    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }
    
    cache_entry->ready = GLOBUS_TRUE;
    
    while(!globus_list_empty(cache_entry->target_list))
    {
        operation = (globus_l_delegation_operation_t *)
            globus_list_remove(
                &cache_entry->target_list, cache_entry->target_list);
        
        result = wsa_EndpointReferenceType_copy(
            operation->cred_epr, &cache_entry->delegated_epr);
        if(result != GLOBUS_SUCCESS)
        {
            globus_free(operation);
            result = GlobusLErrorWrapResult("copying epr", result);
            goto error;
        }
        
        if(first)
        {
            xsd_any *               any;
            globusrunwsAnnotationType * annotation;
            
            if(!*operation->extensions)
            {
                result = wsgram_ExtensionsType_init(operation->extensions);
                if(result != GLOBUS_SUCCESS)
                {
                    globus_free(operation);
                    goto error;
                }
            }
            
            annotation = globus_l_find_annotation(
                &(*operation->extensions)->any);
            if(!annotation)
            {
                any = xsd_any_array_push(&(*operation->extensions)->any);
                if(!any)
                {
                    globus_free(operation);
                    result = GlobusLErrorMemory("any entry");
                    goto error;
                }
                
                result = xsd_QName_copy(
                    &any->element, &globus_l_globusrunAnnotation_qname);
                if(result != GLOBUS_SUCCESS)
                {
                    globus_free(operation);
                    goto error;
                }

                any->any_info = &globusrunwsAnnotationType_info;

                result = globusrunwsAnnotationType_init(&annotation);
                if(result != GLOBUS_SUCCESS)
                {
                    globus_free(operation);
                    goto error;
                }

                any->value = annotation;
            }

            switch(operation->type)
            {
              case GLOBUS_L_JOB_CREDENTIAL:
                annotation->automaticJobDelegation = GLOBUS_TRUE;
                break;
              case GLOBUS_L_STAGING_CREDENTIAL:
                annotation->automaticStagingDelegation = GLOBUS_TRUE;
                break;
              case GLOBUS_L_STAGEIN_CREDENTIAL:
                annotation->automaticStageInDelegation = GLOBUS_TRUE;
                break;
              case GLOBUS_L_STAGEOUT_CREDENTIAL:
                annotation->automaticStageOutDelegation = GLOBUS_TRUE;
                break;
              case GLOBUS_L_CLEANUP_CREDENTIAL:
                annotation->automaticCleanUpDelegation = GLOBUS_TRUE;
                break;
              default:
                globus_assert(0 && "Unexpected delegation type");
            }

            first = GLOBUS_FALSE;
        }
        
        globus_free(operation);
    }
    
error:
    if(result != GLOBUS_SUCCESS && !monitor->error)
    {
        monitor->error = globus_error_get(result);
        globus_l_cancel_all(monitor);
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
    
    globus_delegation_client_util_handle_destroy(handle);
}

/* called locked */
static
globus_result_t
globus_l_get_delegated_epr(
    globus_l_delegated_epr_cache_entry_t *cache_entry,
    wsa_EndpointReferenceType *         factory_epr)
{
    globus_delegation_client_util_handle_t handle;
    globus_gsi_cred_handle_t            cred_handle;
    globus_gsi_cert_utils_cert_type_t   proxy_type;
    globus_result_t                     result;
    globus_bool_t                       tls = GLOBUS_FALSE;;

    result = globus_gsi_cred_handle_init(&cred_handle, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "globus_gsi_cred_handle_init", result);
        goto error_cred_init;
    }
    
    result = globus_gsi_cred_read(cred_handle, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "globus_gsi_cred_read", result);
        goto error_cred_read;
    }
    
    result = globus_gsi_cred_get_cert_type(cred_handle, &proxy_type);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "globus_gsi_cred_get_cert_type", result);
        goto error_cred_read;
    }

    if(GLOBUS_GSI_CERT_UTILS_IS_RFC_PROXY(proxy_type))
    {
        if(cache_entry->key.full)
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_RFC_IMPERSONATION_PROXY;
        }
        else
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_RFC_LIMITED_PROXY;
        }
    }
    else if(GLOBUS_GSI_CERT_UTILS_IS_GSI_2_PROXY(proxy_type))
    {
        if(cache_entry->key.full)
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_PROXY;
        }
        else
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_LIMITED_PROXY;
        }
    }
    else
    {
        if(cache_entry->key.full)
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_IMPERSONATION_PROXY;
        }
        else
        {
            proxy_type = GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_LIMITED_PROXY;
        }
    }

    if (strncmp(factory_epr->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    result = globus_delegation_client_util_handle_init(
        &handle,
        cache_entry->monitor->info->message_attr,
        tls 
            ? cache_entry->monitor->info->tls_handler_chain
            : cache_entry->monitor->info->secmess_handler_chain,
        cred_handle, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_cred_read;
    }

    result = globus_delegation_client_util_delegate_epr_register(
        handle,
        factory_epr,
        proxy_type,
        globus_l_delegate_cb,
        cache_entry);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("delegating", result);
        goto error;
    }

    globus_gsi_cred_handle_destroy(cred_handle);
    
    globus_list_insert(&cache_entry->monitor->active_delegations, handle);
    
    cache_entry->monitor->cb_count++;
    return GLOBUS_SUCCESS;
    
error:
    globus_delegation_client_util_handle_destroy(handle);
error_cred_read:
    globus_gsi_cred_handle_destroy(cred_handle);
error_cred_init:
    return result;
}

/* called locked */
static
globus_result_t
globus_l_delegation_operation_add(
    globus_l_delegated_epr_cache_entry_t *cache_entry,
    globus_l_delegation_type_t          type,
    wsa_EndpointReferenceType **        cred_epr,
    wsgram_ExtensionsType **            extensions)
{
    globus_result_t                     result;
    
    if(cache_entry->ready)
    {
        result = wsa_EndpointReferenceType_copy(
            cred_epr, &cache_entry->delegated_epr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult("copying epr", result);
            goto error;
        }
    }
    else
    {    
        globus_l_delegation_operation_t * operation;
    
        operation = (globus_l_delegation_operation_t *)
            globus_malloc(sizeof(*operation));
        if(!operation)
        {
            result = GlobusLErrorMemory("operation");
            goto error;
        }
        
        operation->type         = type;
        operation->cred_epr     = cred_epr;
        operation->extensions   = extensions;
        
        globus_list_insert(&cache_entry->target_list, operation);
    }
    
    return GLOBUS_SUCCESS;

error:
    return result;
}

/* called locked */
static
globus_result_t
globus_l_start_delegate_credential(
    globus_l_factory_epr_cache_entry_t *factory_cache_entry,
    globus_l_delegation_query_operation_t * operation)
{
    globus_result_t                     result;
    globus_l_delegated_epr_cache_key_t  key;
    globus_l_delegated_epr_cache_entry_t *cache_entry;
    wsa_EndpointReferenceType *         factory_epr;
    
    switch(operation->type)
    {
      case GLOBUS_L_JOB_CREDENTIAL:
      case GLOBUS_L_STAGING_CREDENTIAL:
        factory_epr = &factory_cache_entry->job_factory_epr;
        break;
      case GLOBUS_L_STAGEIN_CREDENTIAL:
      case GLOBUS_L_STAGEOUT_CREDENTIAL:
      case GLOBUS_L_CLEANUP_CREDENTIAL:
        factory_epr = &factory_cache_entry->staging_factory_epr;
        break;
      default:
        globus_assert(0 && "Unexpected delegation type");
    }
    
    key.full = operation->full_proxy;
    key.factory_endpoint = factory_epr->Address.base_value;
    
    cache_entry = (globus_l_delegated_epr_cache_entry_t *)
        globus_hashtable_lookup(
            &factory_cache_entry->monitor->delegated_epr_cache, &key);
    if(!cache_entry)
    {
        cache_entry = (globus_l_delegated_epr_cache_entry_t *)
            globus_malloc(sizeof(*cache_entry));
        if(!cache_entry)
        {
            result = GlobusLErrorMemory("cache_entry");
            goto error_cache_entry;
        }
        
        cache_entry->key.factory_endpoint = globus_libc_strdup(
            key.factory_endpoint);
        if(!cache_entry->key.factory_endpoint)
        {
            result = GlobusLErrorMemory("cache_entry->key.factory_endpoint");
            goto error_cache_entry_dup;
        }
        cache_entry->key.full = key.full;
        cache_entry->target_list = NULL;
        cache_entry->ready = GLOBUS_FALSE;
        cache_entry->monitor = factory_cache_entry->monitor;
        
        result = globus_l_get_delegated_epr(cache_entry, factory_epr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult(
                "querying delegation factories", result);
            goto error_delegate;
        }
    
        globus_hashtable_insert(
            &factory_cache_entry->monitor->delegated_epr_cache,
            &cache_entry->key, cache_entry);
    }
    
    result = globus_l_delegation_operation_add(cache_entry,
        operation->type, operation->cred_epr, operation->extensions);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("initializing operation", result);
        goto error_init;
    }
    
    return GLOBUS_SUCCESS;

error_init:
    return result;
    
error_delegate:
    globus_free(cache_entry->key.factory_endpoint);
error_cache_entry_dup:
    globus_free(cache_entry);
error_cache_entry:
    return result;
}

/* called locked */
static
globus_result_t
globus_l_delegation_query_operation_add(
    globus_l_factory_epr_cache_entry_t *cache_entry,
    globus_bool_t                       full_proxy,
    globus_l_delegation_type_t          type,
    wsa_EndpointReferenceType **        cred_epr,
    wsgram_ExtensionsType **            extensions)
{
    globus_result_t                     result;
    
    globus_l_delegation_query_operation_t * operation;
    
    operation = (globus_l_delegation_query_operation_t *)
        globus_malloc(sizeof(*operation));
    if(!operation)
    {
        result = GlobusLErrorMemory("operation");
        goto error_alloc;
    }
    
    operation->full_proxy   = full_proxy;
    operation->type         = type;
    operation->cred_epr     = cred_epr;
    operation->extensions   = extensions;
    
    if(cache_entry->ready)
    {
        /* start delegation here (currently not possible to get here
         * since all operations are added with lock held
         */
        result = globus_l_start_delegate_credential(cache_entry, operation);
        globus_free(operation);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult(
                "getting delegated credential epr", result);
            goto error_delegate;
        }
    }
    else
    {    
        globus_list_insert(&cache_entry->target_list, operation);
    }
    
    return GLOBUS_SUCCESS;

error_delegate:
error_alloc:
    return result;
}

static
void
globus_l_get_delegation_factories_cb(
    ManagedJobFactoryService_client_handle_t handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrp_GetMultipleResourcePropertiesResponseType * gmrpr,
    ManagedJobFactoryPortType_GetMultipleResourceProperties_fault_t fault_type,
    const xsd_any *                     fault)
{
    globus_l_delegation_query_operation_t * operation;
    globus_l_factory_epr_cache_entry_t *cache_entry;
    globus_l_monitor_t *                monitor;
    int                                 i;
    wsa_EndpointReferenceType *         delegation_factory = NULL;
    wsa_EndpointReferenceType *         staging_factory = NULL;
    
    cache_entry = (globus_l_factory_epr_cache_entry_t *) user_args;
    monitor = cache_entry->monitor;
    
    globus_mutex_lock(&monitor->lock);
    globus_list_remove(&monitor->active_queries,
        globus_list_search(monitor->active_queries, handle));
        
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "querying delegation factories", result);
        goto error;
    }
    
    if(monitor->error)
    {
        goto error;
    }
    
    for(i = 0; i < gmrpr->any.length; i++)
    {
        if(gmrpr->any.elements[i].any_info == 
            &ManagedJobFactoryPortType_delegationFactoryEndpoint_rp_info ||
	    gmrpr->any.elements[i].any_info ==
	    &wsgram_delegationFactoryEndpoint_info)
        {
            globus_assert(!delegation_factory && "Unexpected duplicate entry");
            
            delegation_factory =
                (wsa_EndpointReferenceType *) gmrpr->any.elements[i].value;
        }
        else if(gmrpr->any.elements[i].any_info == 
        &ManagedJobFactoryPortType_stagingDelegationFactoryEndpoint_rp_info || 
	gmrpr->any.elements[i].any_info == 
	&wsgram_stagingDelegationFactoryEndpoint_info)
        {
            globus_assert(!staging_factory && "Unexpected duplicate entry");
            
            staging_factory =
                (wsa_EndpointReferenceType *) gmrpr->any.elements[i].value;
        }
        else
        {
            result = GlobusLErrorInvalid("response",
                "expected only delegationFactoryEndpoint and "
                "stagingDelegationFactoryEndpoint");
            goto error;
        }
    }
    
    if(!delegation_factory || !staging_factory)
    {
        result = GlobusLErrorInvalid("response",
    "expected delegationFactoryEndpoint and stagingDelegationFactoryEndpoint");
        goto error;
    }
    
    result = wsa_EndpointReferenceType_copy_contents(
        &cache_entry->job_factory_epr, delegation_factory);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "copying delegation factories", result);
        goto error;
    }
    
    result = wsa_EndpointReferenceType_copy_contents(
        &cache_entry->staging_factory_epr, staging_factory);
    if(result != GLOBUS_SUCCESS)
    {
        wsa_EndpointReferenceType_destroy_contents(
            &cache_entry->job_factory_epr);
        result = GlobusLErrorWrapResult(
            "copying delegation factories", result);
        goto error;
    }
    
    cache_entry->ready = GLOBUS_TRUE;
    
    while(!globus_list_empty(cache_entry->target_list))
    {
        operation = (globus_l_delegation_query_operation_t *)
            globus_list_remove(
                &cache_entry->target_list, cache_entry->target_list);
            
        result = globus_l_start_delegate_credential(cache_entry, operation);
        globus_free(operation);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult(
                "getting delegated credential epr", result);
            goto error;
        }
    } 
    
error:
    if(result != GLOBUS_SUCCESS && !monitor->error)
    {
        monitor->error = globus_error_get(result);
        globus_l_cancel_all(monitor);
    }
    
    if(--monitor->cb_count == 0)
    {
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->lock);
    
    ManagedJobFactoryService_client_destroy(handle);
}

static
globus_result_t
globus_l_get_delegation_factories(
    globus_l_factory_epr_cache_entry_t *cache_entry,
    wsa_EndpointReferenceType *         factory_epr)
{
    globus_result_t                     result;
    xsd_QName *                         qname;
    wsrp_GetMultipleResourcePropertiesType gmrp;
    ManagedJobFactoryService_client_handle_t factory_handle;
    globus_bool_t                       tls = GLOBUS_FALSE;
    
    if (strncmp(factory_epr->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    wsrp_GetMultipleResourcePropertiesType_init_contents(&gmrp);
    
    result = ManagedJobFactoryService_client_init(&factory_handle,
        cache_entry->monitor->info->message_attr,
        tls 
            ? cache_entry->monitor->info->tls_handler_chain
            : cache_entry->monitor->info->secmess_handler_chain);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_mjfs;
    }
    
    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname,
        &ManagedJobFactoryPortType_delegationFactoryEndpoint_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }

    qname = xsd_QName_array_push(&gmrp.ResourceProperty);
    if(!qname)
    {
        goto error;
    }
    
    result = xsd_QName_copy_contents(
        qname,
        &ManagedJobFactoryPortType_stagingDelegationFactoryEndpoint_rp_qname);
    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }
    
    result = ManagedJobFactoryPortType_GetMultipleResourceProperties_epr_register(
        factory_handle,
        factory_epr,
        &gmrp,
        globus_l_get_delegation_factories_cb,
        cache_entry);
    if(result != GLOBUS_SUCCESS)
    {
        goto error;
    }
    
    globus_list_insert(&cache_entry->monitor->active_queries, factory_handle);
    cache_entry->monitor->cb_count++;
    
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);
    
    return GLOBUS_SUCCESS;

error:
    ManagedJobFactoryService_client_destroy(factory_handle);
error_mjfs:
    wsrp_GetMultipleResourcePropertiesType_destroy_contents(&gmrp);
    return result;
}

/* called locked */
static
void
globus_l_start_delegation(
    globus_l_monitor_t *                monitor,
    wsa_EndpointReferenceType *         job_factory_epr,
    globus_bool_t                       full_proxy,
    globus_l_delegation_type_t          type,
    wsa_EndpointReferenceType **        cred_epr,
    wsgram_ExtensionsType **            extensions)
{
    globus_result_t                     result;
    globus_l_delegation_query_operation_t * operation;
    globus_l_factory_epr_cache_entry_t *cache_entry;
    
    cache_entry = (globus_l_factory_epr_cache_entry_t *)
        globus_hashtable_lookup(&monitor->factory_epr_cache,
            job_factory_epr->Address.base_value);
    if(!cache_entry)
    {
        cache_entry = (globus_l_factory_epr_cache_entry_t *)
            globus_malloc(sizeof(*cache_entry));
        if(!cache_entry)
        {
            result = GlobusLErrorMemory("cache_entry");
            goto error_cache_entry;
        }
        
        cache_entry->query_endpoint = globus_libc_strdup(
            job_factory_epr->Address.base_value);
        if(!cache_entry->query_endpoint)
        {
            result = GlobusLErrorMemory("cache_entry->query_endpoint");
            goto error_cache_entry_dup;
        }
        cache_entry->target_list = NULL;
        cache_entry->ready = GLOBUS_FALSE;
        cache_entry->monitor = monitor;
        
        result = globus_l_get_delegation_factories(
            cache_entry, job_factory_epr);
        if(result != GLOBUS_SUCCESS)
        {
            result = GlobusLErrorWrapResult(
                "querying delegation factories", result);
            goto error_query;
        }
    
        globus_hashtable_insert(&monitor->factory_epr_cache,
            cache_entry->query_endpoint, cache_entry);
    }
    
    result = globus_l_delegation_query_operation_add(
        cache_entry, full_proxy, type, cred_epr, extensions);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("initializing operation", result);
        goto error_init;
    }
    
    return;

error_init:
    monitor->error = globus_error_get(result);
    globus_l_cancel_all(monitor);
    return;
    
error_query:
    globus_free(cache_entry->query_endpoint);
error_cache_entry_dup:
    globus_free(cache_entry);
error_cache_entry:
    monitor->error = globus_error_get(result);
    globus_l_cancel_all(monitor);
    return;
}

static
void
globus_l_cancel_cb(
    void *                              arg)
{
    globus_l_monitor_t *                monitor;
    
    monitor = (globus_l_monitor_t *) arg;
    
    globus_mutex_lock(&monitor->lock);
    {
        if(!monitor->error)
        {
            monitor->error = GlobusLErrorCanceledObj();
            globus_l_cancel_all(monitor);
        }
    }
    globus_mutex_unlock(&monitor->lock);
}

static
void
globus_l_factory_epr_cache_entry_destroy(
    void *                              datum)
{
    globus_l_factory_epr_cache_entry_t *entry;
    
    entry = (globus_l_factory_epr_cache_entry_t *) datum;
    
    globus_free(entry->query_endpoint);
    if(entry->ready)
    {
        wsa_EndpointReferenceType_destroy_contents(&entry->job_factory_epr);
        wsa_EndpointReferenceType_destroy_contents(
            &entry->staging_factory_epr);
    }
    globus_list_destroy_all(entry->target_list, free);
    globus_free(entry);
}

static
void
globus_l_delegated_epr_cache_entry_destroy(
    void *                              datum)
{
    globus_l_delegated_epr_cache_entry_t * entry;
    
    entry = (globus_l_delegated_epr_cache_entry_t *) datum;
    globus_free(entry->key.factory_endpoint);
    if(entry->ready)
    {
        wsa_EndpointReferenceType_destroy_contents(&entry->delegated_epr);
    }
    globus_list_destroy_all(entry->target_list, free);
    globus_free(entry);
}

globus_result_t
globus_i_delegate(
    globus_i_info_t *                   info,
    wsgram_CreateManagedJobInputType *  input)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_monitor_t                  monitor;
    wsgram_MultiJobDescriptionType *    mjd = NULL;
    wsgram_JobDescriptionType *         jd;
    int                                 jobcount;
    int                                 i;
    globus_bool_t                       waiting = GLOBUS_FALSE;
    
    if(input->choice_value.type == wsgram_CreateManagedJobInputType_job)
    {
        jd = &input->choice_value.value.job;
        jobcount = 1;
    }
    else
    {
        mjd = &input->choice_value.value.multiJob;
        jd = mjd->job.elements;
        jobcount = mjd->job.length;
    }
    
    memset(&monitor, 0, sizeof(monitor));
    globus_mutex_init(&monitor.lock, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.cb_count = 0;
    monitor.error = NULL;
    monitor.info = info;
    globus_hashtable_init(&monitor.factory_epr_cache, 8,
        globus_hashtable_string_hash, globus_hashtable_string_keyeq);
    globus_hashtable_init(&monitor.delegated_epr_cache, 8,
        globus_l_delegated_epr_cache_hash, globus_l_delegated_epr_cache_keyeq);
    
    if(globus_i_set_cancel_notify(globus_l_cancel_cb, &monitor))
    {
        result = GlobusLErrorCanceled();
        goto error;
    }
    
    globus_mutex_lock(&monitor.lock);
    {
        if(mjd && !mjd->jobCredentialEndpoint)
        {
            globus_l_start_delegation(
                &monitor,
                mjd->factoryEndpoint,
                GLOBUS_TRUE,
                GLOBUS_L_JOB_CREDENTIAL,
                &mjd->jobCredentialEndpoint,
                &mjd->extensions);
        }
        
        for(i = 0; !monitor.error && i < jobcount; i++)
        {
            if(info->delegate_job && !jd->jobCredentialEndpoint)
            {
                globus_l_start_delegation(
                    &monitor,
                    jd->factoryEndpoint,
                    GLOBUS_FALSE,
                    GLOBUS_L_JOB_CREDENTIAL,
                    &jd->jobCredentialEndpoint,
                    &jd->extensions);
            }
            
            if(info->delegate_staging &&
                (jd->fileStageIn || jd->fileStageOut || jd->fileCleanUp))
            {
                if(!monitor.error && !jd->stagingCredentialEndpoint)
                {
                    globus_l_start_delegation(
                        &monitor,
                        jd->factoryEndpoint,
                        GLOBUS_FALSE,
                        GLOBUS_L_STAGING_CREDENTIAL,
                        &jd->stagingCredentialEndpoint,
                        &jd->extensions);
                }
                
                if(!monitor.error && jd->fileStageIn &&
                    !jd->fileStageIn->transferCredentialEndpoint)
                {
                    globus_l_start_delegation(
                        &monitor,
                        jd->factoryEndpoint,
                        GLOBUS_FALSE,
                        GLOBUS_L_STAGEIN_CREDENTIAL,
                        &jd->fileStageIn->transferCredentialEndpoint,
                        &jd->extensions);
                }
                
                if(!monitor.error && jd->fileStageOut &&
                    !jd->fileStageOut->transferCredentialEndpoint)
                {
                    globus_l_start_delegation(
                        &monitor,
                        jd->factoryEndpoint,
                        GLOBUS_FALSE,
                        GLOBUS_L_STAGEOUT_CREDENTIAL,
                        &jd->fileStageOut->transferCredentialEndpoint,
                        &jd->extensions);
                }
                
                if(!monitor.error && jd->fileCleanUp &&
                    !jd->fileCleanUp->transferCredentialEndpoint)
                {
                    globus_l_start_delegation(
                        &monitor,
                        jd->factoryEndpoint,
                        GLOBUS_FALSE,
                        GLOBUS_L_CLEANUP_CREDENTIAL,
                        &jd->fileCleanUp->transferCredentialEndpoint,
                        &jd->extensions);
                }
            }
            
            jd++;
        }
        
        if(monitor.cb_count > 0 && !info->quiet)
        {
            fprintf(stderr, "Delegating user credentials...");
            waiting = GLOBUS_TRUE;
        }
        
        while(monitor.cb_count > 0)
        {
            globus_cond_wait(&monitor.cond, &monitor.lock);
        }
        
        if(monitor.error)
        {
            if(!info->quiet)
            {
                fprintf(stderr, "Failed.\n");
            }
            result = globus_error_put(monitor.error);
        }
        else if(waiting)
        {
            fprintf(stderr, "Done.\n");
        }
    }
    globus_mutex_unlock(&monitor.lock);
    
    globus_i_set_cancel_notify(NULL, NULL);
    
    if(result != GLOBUS_SUCCESS)
    {
        globus_i_delegate_cleanup(info, input);
        result = GlobusLErrorWrapResult("trying to delegate", result);
    }
error:
    globus_hashtable_destroy_all(&monitor.delegated_epr_cache,
        globus_l_delegated_epr_cache_entry_destroy);
    globus_hashtable_destroy_all(&monitor.factory_epr_cache,
        globus_l_factory_epr_cache_entry_destroy);
    globus_cond_destroy(&monitor.cond);
    globus_mutex_destroy(&monitor.lock);
    return result;
}

static
void
globus_l_delegate_destroy_cb(
    DelegationService_client_handle_t   handle,
    void *                              user_args,
    globus_result_t                     result,
    const wsrl_DestroyResponseType *    DestroyResponse,
    DelegationPortType_Destroy_fault_t  fault_type,
    const xsd_any *                     fault)
{
    globus_l_monitor_t *                monitor;
    
    monitor = (globus_l_monitor_t *) user_args;
    
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "cleaning up delegated credentials", result);
    }
    
    globus_mutex_lock(&monitor->lock);
    {
        globus_list_remove(&monitor->active_cleanups,
            globus_list_search(monitor->active_cleanups, handle));
                       
        if(result != GLOBUS_SUCCESS && !monitor->error)
        {
            monitor->error = globus_error_get(result);
        }
    
        if(--monitor->cb_count == 0)
        {
            globus_cond_signal(&monitor->cond);
        }
    }
    globus_mutex_unlock(&monitor->lock);
    
    DelegationService_client_destroy(handle);
}

/* called locked */
static
void
globus_l_start_delegation_cleanup(
    globus_l_monitor_t *                monitor,
    wsa_EndpointReferenceType *         cred_endpoint)
{
    globus_result_t                     result;
    wsrl_DestroyType                    destroy;
    DelegationService_client_handle_t   handle;
    globus_bool_t                       tls = GLOBUS_FALSE;
    
    if (strncmp(cred_endpoint->Address.base_value, "https:", 6) == 0)
    {
        tls = GLOBUS_TRUE;
    }
    
    wsrl_DestroyType_init_contents(&destroy);
    result = DelegationService_client_init(
        &handle,
        monitor->info->message_attr,
        tls 
            ? monitor->info->tls_handler_chain
            : monitor->info->secmess_handler_chain);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_init;
    }
    
    result = DelegationPortType_Destroy_epr_register(
        handle,
        cred_endpoint,
        &destroy,
        globus_l_delegate_destroy_cb,
        monitor);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_register;
    }
    
    globus_list_insert(&monitor->active_cleanups, handle);
    wsrl_DestroyType_destroy_contents(&destroy);
    monitor->cb_count++;
    
    return;

error_register:
    DelegationService_client_destroy(handle);
error_init:
    wsrl_DestroyType_destroy_contents(&destroy);
    monitor->error = globus_error_get(result);
    return;
}

globus_result_t
globus_i_delegate_cleanup(
    globus_i_info_t *                   info,
    wsgram_CreateManagedJobInputType *  input)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_l_monitor_t                  monitor;
    wsgram_MultiJobDescriptionType *    mjd = NULL;
    wsgram_JobDescriptionType *         jd;
    int                                 jobcount;
    int                                 i;
    globusrunwsAnnotationType *         annotation;
    globus_bool_t                       waiting = GLOBUS_FALSE;
    
    if(input->choice_value.type == wsgram_CreateManagedJobInputType_job)
    {
        jd = &input->choice_value.value.job;
        jobcount = 1;
    }
    else
    {
        mjd = &input->choice_value.value.multiJob;
        jd = mjd->job.elements;
        jobcount = mjd->job.length;
    }
    
    memset(&monitor, 0, sizeof(monitor));
    globus_mutex_init(&monitor.lock, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.cb_count = 0;
    monitor.error = NULL;
    monitor.info = info;
    
    if(globus_i_set_cancel_notify(globus_l_cancel_cb, &monitor))
    {
        result = GlobusLErrorCanceled();
        goto error;
    }
    
    globus_mutex_lock(&monitor.lock);
    {
        if(mjd && mjd->jobCredentialEndpoint && mjd->extensions &&
            (annotation = globus_l_find_annotation(&mjd->extensions->any)) &&
            annotation->automaticJobDelegation)
        {
            globus_l_start_delegation_cleanup(
                &monitor, mjd->jobCredentialEndpoint);
        }
        
        for(i = 0; i < jobcount; i++)
        {
            if(jd->extensions &&
                (annotation = globus_l_find_annotation(&jd->extensions->any)))
            {
                if(jd->jobCredentialEndpoint &&
                    annotation->automaticJobDelegation)
                {
                    globus_l_start_delegation_cleanup(
                        &monitor, jd->jobCredentialEndpoint);
                }
                
                if(jd->stagingCredentialEndpoint &&
                    annotation->automaticStagingDelegation)
                {
                    globus_l_start_delegation_cleanup(
                        &monitor, jd->stagingCredentialEndpoint);
                }
                
                if(jd->fileStageIn &&
                    jd->fileStageIn->transferCredentialEndpoint &&
                    annotation->automaticStageInDelegation)
                {
                    globus_l_start_delegation_cleanup(
                        &monitor, jd->fileStageIn->transferCredentialEndpoint);
                }
                
                if(jd->fileStageOut &&
                    jd->fileStageOut->transferCredentialEndpoint &&
                    annotation->automaticStageOutDelegation)
                {
                    globus_l_start_delegation_cleanup(
                       &monitor, jd->fileStageOut->transferCredentialEndpoint);
                }
                
                if(jd->fileCleanUp &&
                    jd->fileCleanUp->transferCredentialEndpoint &&
                    annotation->automaticCleanUpDelegation)
                {
                    globus_l_start_delegation_cleanup(
                        &monitor, jd->fileCleanUp->transferCredentialEndpoint);
                }
            }
            
            jd++;
        }
        
        if(monitor.cb_count > 0 && !info->quiet)
        {
            fprintf(stderr, "Cleaning up any delegated credentials...");
            waiting = GLOBUS_TRUE;
        }
        
        while(monitor.cb_count > 0)
        {
            globus_cond_wait(&monitor.cond, &monitor.lock);
        }
        
        if(monitor.error)
        {
            if(!info->quiet)
            {
                fprintf(stderr, "Failed.\n");
            }
            
            result = globus_error_put(monitor.error);
        }
        else if(waiting)
        {
            fprintf(stderr, "Done.\n");
        }
    }
    globus_mutex_unlock(&monitor.lock);
    
    globus_i_set_cancel_notify(NULL, NULL);
    
    globus_cond_destroy(&monitor.cond);
    globus_mutex_destroy(&monitor.lock);
    
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult(
            "cleaning up automatically delegated credentials", result);
    }
error:
    return result;
}
