/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_xio.h"
#include "globus_i_streaming.h"

#define GLOBUS_I_MJG_STREAMING_NONE 0

typedef struct globus_l_streaming_entry_s
{
    globus_i_streaming_t *              handle;
    char *                              url;
    globus_ftp_client_operationattr_t   ftp_attr;
    globus_callback_handle_t            delay_handle;
    globus_off_t                        current_offset;
    globus_xio_handle_t                 xio_handle;
    int                                 cb_count;
    globus_object_t *                   error;
    globus_bool_t                       eof;
} globus_l_streaming_entry_t;

static
globus_result_t
globus_l_streaming_read(
    globus_i_streaming_t *              handle,
    globus_bool_t                       return_error);
    
static
void
globus_l_streaming_read_done(
    globus_l_streaming_entry_t *        entry);

/* called locked */
static
void
globus_l_streaming_entry_destroy(
    globus_l_streaming_entry_t *        entry)
{
    globus_ftp_client_operationattr_destroy(&entry->ftp_attr);
    globus_free(entry->url);
    globus_free(entry);
}

/* called locked */
static
void
globus_l_streaming_register_oneshot(
    globus_i_streaming_t *              handle,
    globus_callback_handle_t *          cb_handle,
    globus_reltime_t *                  delay,
    globus_callback_func_t              callback,
    void *                              user_arg)
{
    globus_callback_register_oneshot(
        cb_handle,
        delay,
        callback,
        user_arg);
    globus_list_insert(
        &handle->oneshots, (void *) *cb_handle);
}

/* called locked */
static
void
globus_l_streaming_oneshot_done(
    globus_i_streaming_t *              handle,
    globus_callback_handle_t *          cb_handle)
{
    globus_list_remove(
        &handle->oneshots,
        globus_list_search(
            handle->oneshots, (void *) *cb_handle));
    globus_callback_unregister(
        *cb_handle, NULL, NULL, NULL);
    *cb_handle = 0;
}

static
void
globus_l_streaming_write_cb(
    globus_xio_handle_t                 xio_handle, 
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       len,
    globus_size_t                       nbytes, 
    globus_xio_data_descriptor_t        data_desc,
    void *                              user_arg)
{
    globus_l_streaming_entry_t *        entry;
    globus_bool_t                       read_done = GLOBUS_FALSE;
    
    entry = (globus_l_streaming_entry_t *) user_arg;
    
    globus_free(buffer);
    
    globus_mutex_lock(&entry->handle->lock);
    {
        if(result != GLOBUS_SUCCESS && !entry->error)
        {
            entry->error = globus_error_get(result);
        }
        
        if(--entry->cb_count == 0)
        {
            read_done = GLOBUS_TRUE;
        }
    }
    globus_mutex_unlock(&entry->handle->lock);
    
    if(read_done)
    {
        globus_l_streaming_read_done(entry);
    }
}

static 
void
globus_l_streaming_data_cb(
    void *                              user_arg,
    globus_ftp_client_handle_t *        handle,
    globus_object_t *                   error,
    globus_byte_t *                     buffer,
    globus_size_t                       length,
    globus_off_t                        offset,
    globus_bool_t                       eof)
{
    globus_l_streaming_entry_t *        entry;
    globus_bool_t                       read_done = GLOBUS_FALSE;
    
    entry = (globus_l_streaming_entry_t *) user_arg;
    
    globus_mutex_lock(&entry->handle->lock);
    {
        entry->current_offset += length;
        entry->eof = length < GLOBUS_I_MJG_STREAMING_BUFFER_SIZE
            ? eof : GLOBUS_FALSE;
        
        if(error && !entry->error)
        {
            entry->error = globus_object_copy(error);
        }
        
        if(length > 0)
        {
            globus_result_t             result;
            
            result = globus_xio_register_write(
                entry->xio_handle,
                buffer,
                length,
                length,
                NULL,
                globus_l_streaming_write_cb,
                entry);
            if(result != GLOBUS_SUCCESS)
            {
                if(!entry->error)
                {
                    entry->error = globus_error_get(result);
                }
                
                if(--entry->cb_count == 0)
                {
                    read_done = GLOBUS_TRUE;
                }
            }
        }
        else
        {
            if(--entry->cb_count == 0)
            {
                read_done = GLOBUS_TRUE;
            }
        
            globus_free(buffer);
        }
    }
    globus_mutex_unlock(&entry->handle->lock);
    
    if(read_done)
    {
        globus_l_streaming_read_done(entry);
    }
}

static
void
globus_l_streaming_done_cb(
    void *                              user_arg,
    globus_ftp_client_handle_t *        handle,
    globus_object_t *                   error)
{
    globus_l_streaming_entry_t *        entry;
    globus_bool_t                       read_done = GLOBUS_FALSE;
    
    entry = (globus_l_streaming_entry_t *) user_arg;
    
    globus_mutex_lock(&entry->handle->lock);
    {
        if(error && !entry->error)
        {
            entry->error = globus_object_copy(error);
        }
        
        if(--entry->cb_count == 0)
        {
            read_done = GLOBUS_TRUE;
        }
    }
    globus_mutex_unlock(&entry->handle->lock);
    
    if(read_done)
    {
        globus_l_streaming_read_done(entry);
    }
}

static
void
globus_l_streaming_purge_cb(
    void *                              user_arg)
{
    globus_l_streaming_entry_t *        entry;
    
    entry = (globus_l_streaming_entry_t *) user_arg;
    
    entry->handle->callback(
        entry->handle,
        entry->url,
        entry->error ? globus_error_put(entry->error) : GLOBUS_SUCCESS,
        entry->handle->user_arg);
        
    globus_l_streaming_entry_destroy(entry);
}

static
void
globus_l_streaming_idle_delay_cb(
    void *                              user_arg)
{
    globus_l_streaming_entry_t *        entry;
    
    entry = (globus_l_streaming_entry_t *) user_arg;
    
    globus_mutex_lock(&entry->handle->lock);
    {
        globus_l_streaming_oneshot_done(
            entry->handle, &entry->delay_handle);
    
        globus_fifo_enqueue(&entry->handle->ready_q, entry);
    }
    globus_mutex_unlock(&entry->handle->lock);
    
    globus_l_streaming_read(entry->handle, GLOBUS_FALSE);
}

static
globus_result_t
globus_l_streaming_read(
    globus_i_streaming_t *              handle,
    globus_bool_t                       return_error)
{
    globus_l_streaming_entry_t *        entry;
    globus_byte_t *                     buffer;
    globus_result_t                     result;
    
    globus_mutex_lock(&handle->lock);
    {
        if(handle->state == GLOBUS_I_MJG_STREAMING_ABORT)
        {
            while(!globus_fifo_empty(&handle->ready_q))
            {
                entry = (globus_l_streaming_entry_t *)
                    globus_fifo_dequeue(&handle->ready_q);
                
                globus_callback_register_oneshot(
                    NULL,
                    NULL,
                    globus_l_streaming_purge_cb,
                    entry);
            }
        }
        else if(!handle->active && !globus_fifo_empty(&handle->ready_q))
        {
            entry = (globus_l_streaming_entry_t *)
                    globus_fifo_dequeue(&handle->ready_q);
                    
            buffer = (globus_byte_t *)
                globus_malloc(GLOBUS_I_MJG_STREAMING_BUFFER_SIZE);
            if(!buffer)
            {
                result = globus_error_put(
                    globus_error_construct_errno_error(NULL, NULL, errno));
                goto error_buffer;
            }
            
            result = globus_ftp_client_partial_get(
                &entry->handle->ftp_handle,
                entry->url,
                &entry->ftp_attr,
                NULL,
                entry->current_offset,
                entry->current_offset + GLOBUS_I_MJG_STREAMING_BUFFER_SIZE,
                globus_l_streaming_done_cb,
                entry);
            if(result != GLOBUS_SUCCESS)
            {
                goto error_get;
            }
            
            entry->eof = GLOBUS_FALSE;
            entry->cb_count = 1;
            handle->active = entry;
            
            result = globus_ftp_client_register_read(
                &entry->handle->ftp_handle,
                buffer,
                GLOBUS_I_MJG_STREAMING_BUFFER_SIZE,
                globus_l_streaming_data_cb,
                entry);
            if(result != GLOBUS_SUCCESS)
            {
                /* tell the user success,
                 * we'll let the done cb relay the error */
                entry->error = globus_error_get(result);
                globus_ftp_client_abort(&handle->ftp_handle);
                globus_free(buffer);
            }
            else
            {
                entry->cb_count++;
            }
        }
    }
    globus_mutex_unlock(&handle->lock);
    
    return GLOBUS_SUCCESS;

error_get:
    globus_free(buffer);
error_buffer:
    globus_mutex_unlock(&handle->lock);
    if(!return_error)
    {
        entry->error = globus_error_get(result);
        result = GLOBUS_SUCCESS;
        globus_callback_register_oneshot(
            NULL,
            NULL,
            globus_l_streaming_purge_cb,
            entry);
    }
    return result;
}

static
void
globus_l_streaming_read_done(
    globus_l_streaming_entry_t *        entry)
{
    globus_i_streaming_t *              handle;
    globus_bool_t                       kickout = GLOBUS_FALSE;
    
    handle = entry->handle;
    
    globus_mutex_lock(&handle->lock);
    {
        globus_assert(handle->active == entry);
        
        handle->active = NULL;
        if(entry->error ||
            handle->state == GLOBUS_I_MJG_STREAMING_ABORT ||
            (handle->state == GLOBUS_I_MJG_STREAMING_FINISH && entry->eof))
        {
            kickout = GLOBUS_TRUE;
        }
        else
        {
            if(entry->eof)
            {
                globus_assert(handle->state == GLOBUS_I_MJG_STREAMING_NONE);
                
                globus_l_streaming_register_oneshot(
                    handle,
                    &entry->delay_handle,
                    &handle->idle_delay,
                    globus_l_streaming_idle_delay_cb,
                    entry);
            }
            else
            {
                globus_fifo_enqueue(&handle->ready_q, entry);
            }
        }
    }
    globus_mutex_unlock(&handle->lock);
    
    globus_l_streaming_read(handle, GLOBUS_FALSE);
    
    if(kickout)
    {
        handle->callback(
            handle,
            entry->url,
            entry->error ? globus_error_put(entry->error) : GLOBUS_SUCCESS,
            handle->user_arg);
        
        globus_l_streaming_entry_destroy(entry);
    }
}
    
globus_result_t
globus_i_streaming_init(
    globus_i_streaming_t *              handle,
    globus_ftp_client_handleattr_t *    user_attr,
    globus_reltime_t *                  idle_delay,
    globus_i_streaming_cb_t             url_cb,
    void *                              user_arg)
{
    globus_result_t                     result;
    globus_ftp_client_handleattr_t      ftp_attr;
    
    memset(handle, 0, sizeof(*handle));
    globus_mutex_init(&handle->lock, NULL);
    globus_fifo_init(&handle->ready_q);
    handle->callback = url_cb;
    handle->user_arg = user_arg;
    
    if(idle_delay)
    {
        handle->idle_delay = *idle_delay;
    }
    else
    {
        GlobusTimeReltimeSet(
            handle->idle_delay, GLOBUS_I_MJG_STREAMING_IDLE_DELAY, 0);
    }
    
    if(user_attr)
    {
        result = globus_ftp_client_handleattr_copy(&ftp_attr, user_attr);
    }
    else
    {
        result = globus_ftp_client_handleattr_init(&ftp_attr);
    }
    if(result != GLOBUS_SUCCESS)
    {
        goto error_attr;
    }
    
    result = globus_ftp_client_handleattr_set_cache_all(
        &ftp_attr, GLOBUS_TRUE);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_setup;
    }
    
    result = globus_ftp_client_handle_init(&handle->ftp_handle, &ftp_attr);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_init;
    }
    
    handle->ref = 1;
    globus_ftp_client_handleattr_destroy(&ftp_attr);

    return GLOBUS_SUCCESS;

error_init:
error_setup:
    globus_ftp_client_handleattr_destroy(&ftp_attr);
error_attr:
    globus_fifo_destroy(&handle->ready_q);
    globus_mutex_destroy(&handle->lock);
    return result;
}

/* note: keep track of how many urls are added */
globus_result_t
globus_i_streaming_add(
    globus_i_streaming_t *              handle,
    const char *                        url,
    globus_ftp_client_operationattr_t * ftp_attr,
    globus_xio_handle_t                 dest_handle)
{
    globus_result_t                     result;
    globus_l_streaming_entry_t *        entry;
    
    globus_assert(handle->state == GLOBUS_I_MJG_STREAMING_NONE &&
        "Can't add more urls after stopping");
    
    entry = (globus_l_streaming_entry_t *)
        globus_calloc(1, sizeof(*entry));
    if(!entry)
    {
        result = globus_error_put(
            globus_error_construct_errno_error(NULL, NULL, errno));
        goto error_malloc;
    }
    
    entry->handle = handle;
    entry->xio_handle = dest_handle;
    entry->url = globus_libc_strdup(url);
    if(!entry->url)
    {
        result = globus_error_put(
            globus_error_construct_errno_error(NULL, NULL, errno));
        goto error_dup;
    }
    
    if(ftp_attr)
    {
        result = globus_ftp_client_operationattr_copy(  
            &entry->ftp_attr, ftp_attr);
    }
    else
    {
        result = globus_ftp_client_operationattr_init(&entry->ftp_attr);
    }
    if(result != GLOBUS_SUCCESS)
    {
        goto error_attr;
    }
    
    result = globus_ftp_client_operationattr_set_read_all(
        &entry->ftp_attr, GLOBUS_TRUE, NULL, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_setup;
    }
    
    globus_mutex_lock(&handle->lock);
    {
        globus_fifo_enqueue(&handle->ready_q, entry);
    }
    globus_mutex_unlock(&handle->lock);
    
    result = globus_l_streaming_read(handle, GLOBUS_TRUE);
    if(result != GLOBUS_SUCCESS)
    {
        goto error_read;
    }
    
    return GLOBUS_SUCCESS;

error_read:
error_setup:
    globus_ftp_client_operationattr_destroy(&entry->ftp_attr);
error_attr:
    globus_free(entry->url);
error_dup:
    globus_free(entry);
error_malloc:
    return result;
}

static
void
globus_l_streaming_delay_finish_cb(
    void *                              user_arg)
{
    globus_i_streaming_t *              handle;
    globus_list_t *                     list;
    
    handle = (globus_i_streaming_t *) user_arg;
    
    globus_mutex_lock(&handle->lock);
    {
        globus_l_streaming_oneshot_done(
            handle, &handle->finish_delay_handle);
    
        if(handle->state == GLOBUS_I_MJG_STREAMING_NONE)
        {
            handle->state = GLOBUS_I_MJG_STREAMING_FINISH;
            
            for(list = handle->oneshots;
                !globus_list_empty(list);
                list = globus_list_rest(list))
            {
                globus_callback_adjust_oneshot(
                    (globus_callback_handle_t) globus_list_first(list), NULL);
            }
        }
    }
    globus_mutex_unlock(&handle->lock);
    
    /* remove reference and possibly destroy */
    globus_i_streaming_destroy(handle);
}

void
globus_i_streaming_stop(
    globus_i_streaming_t *              handle,
    int                                 how,
    globus_reltime_t *                  delay)
{
    globus_list_t *                     list;

    globus_assert((how == GLOBUS_I_MJG_STREAMING_FINISH ||
        how == GLOBUS_I_MJG_STREAMING_ABORT) && "Invalid how");
    
    globus_mutex_lock(&handle->lock);
    {
        if(handle->state != GLOBUS_I_MJG_STREAMING_ABORT)
        {
            if(how == GLOBUS_I_MJG_STREAMING_FINISH &&
                delay &&
                !handle->finish_delay_handle &&
                (handle->active || handle->oneshots))
            {
                handle->ref++;
                globus_l_streaming_register_oneshot(
                    handle,
                    &handle->finish_delay_handle,
                    delay,
                    globus_l_streaming_delay_finish_cb,
                    handle);
            }
            else
            {
                handle->state = how;
                
                if(handle->active && how == GLOBUS_I_MJG_STREAMING_ABORT)
                {
                    globus_ftp_client_abort(&handle->ftp_handle);
                }
                
                for(list = handle->oneshots;
                    !globus_list_empty(list);
                    list = globus_list_rest(list))
                {
                    globus_callback_adjust_oneshot(
                        (globus_callback_handle_t) globus_list_first(list),
                        NULL);
                }
            }
        }
    }
    globus_mutex_unlock(&handle->lock);
    
    /* will handle purging enqueued entries if aborting */
    globus_l_streaming_read(handle, GLOBUS_FALSE);
}

void
globus_i_streaming_destroy(
    globus_i_streaming_t *              handle)
{
    globus_bool_t                       destroy = GLOBUS_FALSE;
    
    globus_assert(!handle->active && !handle->oneshots &&
        "Can not destroy with outstanding operations");
    
    globus_mutex_lock(&handle->lock);
    {
        if(--handle->ref == 0)
        { 
            destroy = GLOBUS_TRUE;
        }
    }
    globus_mutex_unlock(&handle->lock);
    
    if(destroy)
    {
        globus_ftp_client_handle_destroy(&handle->ftp_handle);
        globus_fifo_destroy(&handle->ready_q);
        globus_mutex_destroy(&handle->lock);
    }
}
