/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_i_globusrun_ws.h"
#include "globus_delegation_client_util.h"
#include "DelegationService_client.h"
#include "ManagedExecutableJobService_client.h"
#include "globus_ws_addressing.h"
#include "globus_ws_security.h"
#include "globus_error_gssapi.h"

static globus_bool_t                    globus_l_canceled;
static globus_bool_t                    globus_l_cancel_count;
static globus_i_cancel_notify_t         globus_l_cancel_cb;
static void *                           globus_l_cancel_arg;
static globus_mutex_t                   globus_l_cancel_lock;

#define globus_l_null_check_free(v__)                                       \
{                                                                           \
    if((v__))                                                               \
    {                                                                       \
        globus_free((v__));                                                 \
    }                                                                       \
}

globus_bool_t
globus_i_set_cancel_notify(
    globus_i_cancel_notify_t            cancel_cb,
    void *                              cancel_arg)
{
    globus_bool_t                       canceled;

    globus_mutex_lock(&globus_l_cancel_lock);
    {
        if(!globus_l_canceled || !cancel_cb)
        {
            globus_l_cancel_cb = cancel_cb;
            globus_l_cancel_arg = cancel_arg;
        }
        canceled = globus_l_canceled;
    }
    globus_mutex_unlock(&globus_l_cancel_lock);

    return canceled;
}

globus_bool_t
globus_i_check_canceled(void)
{
    return globus_l_canceled;
}

void
globus_i_clear_canceled(void)
{
    globus_mutex_lock(&globus_l_cancel_lock);
    {
        globus_l_cancel_count = 0;
        globus_l_canceled = GLOBUS_FALSE;
    }
    globus_mutex_unlock(&globus_l_cancel_lock);
}

static
void
globus_l_interrupt_cb(
    void *                              user_arg)
{
    fprintf(stderr, "Canceling...");

    globus_mutex_lock(&globus_l_cancel_lock);
    {
        globus_l_cancel_count++;
        if(globus_l_cancel_count >= 2)
        {
            globus_callback_unregister_signal_handler(
                GLOBUS_SIGNAL_INTERRUPT, NULL, NULL);
        }

        globus_l_canceled = GLOBUS_TRUE;
        if(globus_l_cancel_cb)
        {
            globus_l_cancel_cb(globus_l_cancel_arg);
            globus_l_cancel_cb = NULL;
        }
    }
    globus_mutex_unlock(&globus_l_cancel_lock);
}

static
void
globus_l_info_init(
    globus_i_info_t *                   info)
{
    memset(info, 0, sizeof(*info));

    globus_i_test_result("initializing message attr",
        globus_soap_message_attr_init(&info->message_attr));

    globus_i_test_result("initializing secmess handler chain",
        globus_handler_chain_init(&info->secmess_handler_chain));
    globus_i_test_result("initializing tls handler chain",
        globus_handler_chain_init(&info->tls_handler_chain));
    globus_i_test_result("initializing service handler chain",
        globus_handler_chain_init(&info->service_handler_chain));
    globus_i_test_result("initializing secmess service handler chain",
        globus_handler_chain_init(&info->secmess_service_handler_chain));

    globus_i_test_result("initializing ftp attr",
        globus_ftp_client_operationattr_init(&info->ftp_attr));
}

static
void
globus_l_info_destroy(
    globus_i_info_t *                   info)
{
    if(info->job_fault)
    {
        globus_object_free(info->job_fault);
    }
    
    if(info->job_epr)
    {
        wsa_EndpointReferenceType_destroy(info->job_epr);
    }
    
    if(info->job_cred_epr)
    {
        wsa_EndpointReferenceType_destroy(info->job_cred_epr);
    }
    
    if(info->staging_cred_epr)
    {
        wsa_EndpointReferenceType_destroy(info->staging_cred_epr);
    }
    
    if(info->transfer_cred_epr)
    {
        wsa_EndpointReferenceType_destroy(info->transfer_cred_epr);
    }
    
    globus_ftp_client_operationattr_destroy(&info->ftp_attr);
    globus_handler_chain_destroy(info->secmess_service_handler_chain);
    globus_handler_chain_destroy(info->service_handler_chain);
    globus_handler_chain_destroy(info->tls_handler_chain);
    globus_handler_chain_destroy(info->secmess_handler_chain);
    globus_soap_message_attr_destroy(info->message_attr);

    globus_extension_deactivate("globus_handler_ws_addressing");
    if(info->debug)
    {
        globus_extension_deactivate("globus_handler_debug");
    }

    globus_l_null_check_free(info->factory);
    globus_l_null_check_free(info->factory_type);
    globus_l_null_check_free(info->job_description_file);
    globus_l_null_check_free(info->submission_id_output_file);
    globus_l_null_check_free(info->job_epr_output_file);
    globus_l_null_check_free(info->stream_out_file);
    globus_l_null_check_free(info->stream_err_file);
    globus_l_null_check_free(info->subject);
    
    if(info->identity != GSS_C_NO_NAME)
    {
        OM_uint32                       minor_status;

        gss_release_name(&minor_status, &info->identity);
    }
}

static
void
globus_l_setup_attributes(
    globus_i_info_t *                   info)
{
    int                                 value;

    globus_i_test_rc("loading ws-secure message handler",
        globus_extension_activate(GLOBUS_HANDLER_WS_SECMESS_LIB));

    globus_i_test_result("pushing ws-secure message handler",
        globus_handler_chain_push(
            info->secmess_handler_chain,
            GLOBUS_HANDLER_TYPE_ALL,
            GLOBUS_HANDLER_WS_SECMESS_CLIENT));

    globus_i_test_rc("loading ws-addressing handler",
        globus_extension_activate("globus_handler_ws_addressing"));

    if(info->debug)
    {
        globus_i_test_rc("loading debug handler",
            globus_extension_activate("globus_handler_debug"));

        globus_i_test_result("pushing debug handler",
            globus_handler_chain_push(
                info->secmess_handler_chain,
                GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
                "globus_handler_debug"));
        globus_i_test_result("pushing debug handler",
            globus_handler_chain_push(
                info->tls_handler_chain,
                GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
                "globus_handler_debug"));
        globus_i_test_result("pushing debug handler",
            globus_handler_chain_push(
                info->service_handler_chain,
                GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
                "globus_handler_debug"));
        globus_i_test_result("pushing debug handler",
            globus_handler_chain_push(
                info->secmess_service_handler_chain,
                GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
                "globus_handler_debug"));
    }
    globus_i_test_result("pushing ws-addressing handler",
        globus_handler_chain_push(
            info->secmess_handler_chain,
            GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
            GLOBUS_HANDLER_WS_ADDRESSING_CLIENT));
    globus_i_test_result("pushing ws-addressing handler",
        globus_handler_chain_push(
            info->tls_handler_chain,
            GLOBUS_HANDLER_TYPE_REQUEST|GLOBUS_HANDLER_TYPE_RESPONSE,
            GLOBUS_HANDLER_WS_ADDRESSING_CLIENT));
    globus_i_test_result("pushing ws-addressing handler",
        globus_handler_chain_push(
            info->service_handler_chain,
            GLOBUS_HANDLER_TYPE_ALL,
            GLOBUS_HANDLER_WS_ADDRESSING_SERVER));
    globus_i_test_result("pushing ws-addressing handler",
        globus_handler_chain_push(
            info->secmess_service_handler_chain,
            GLOBUS_HANDLER_TYPE_ALL,
            GLOBUS_HANDLER_WS_ADDRESSING_SERVER));

    globus_i_test_result("pushing ws-secure message handler",
        globus_handler_chain_push(
            info->secmess_service_handler_chain,
            GLOBUS_HANDLER_TYPE_ALL,
            GLOBUS_HANDLER_WS_SECMESS_SERVER));

    switch(info->auth_mode)
    {
      case GLOBUS_I_AUTH_SELF:
        value = GLOBUS_SOAP_MESSAGE_AUTHZ_SELF;
        {
            OM_uint32                   minor_status;
            OM_uint32                   major_status;
            gss_cred_id_t               cred_handle;
            gss_name_t                  identity;
            gss_buffer_desc             name_buffer;
  
            major_status = gss_acquire_cred(
                &minor_status,
                GSS_C_NO_NAME,
                GSS_C_INDEFINITE,
                GSS_C_NO_OID_SET,
                GSS_C_INITIATE,
                &cred_handle,
                NULL,
                NULL);
            if(!GSS_ERROR(major_status))
            {
                major_status = gss_inquire_cred(
                    &minor_status,
                    cred_handle,
                    &identity,
                    NULL,
                    NULL,
                    NULL);
                if(!GSS_ERROR(major_status))
                {
                    major_status = gss_display_name(
                        &minor_status,
                        identity,
                        &name_buffer,
                        NULL);
                    if(!GSS_ERROR(major_status))
                    {
                        info->subject = (char *) name_buffer.value;
                        
                        globus_i_test_result("setting ftp identity",
                            globus_ftp_client_operationattr_set_authorization(
                                &info->ftp_attr,
                                GSS_C_NO_CREDENTIAL,
                                NULL,
                                NULL,
                                NULL,
                                (char *) name_buffer.value));
                    }

                    gss_release_name(&minor_status, &identity);
                }

                gss_release_cred(&minor_status, &cred_handle);
            }

            if(GSS_ERROR(major_status))
            {
                globus_i_test_result("determining user identity",
                    globus_error_put(
                        globus_error_construct_gssapi_error(
                            NULL, NULL, major_status, minor_status)));
            }
        }
        break;

      case GLOBUS_I_AUTH_SUBJECT:
        value = GLOBUS_SOAP_MESSAGE_AUTHZ_IDENTITY;
        globus_i_test_result("setting authentication identity",
            globus_soap_message_attr_set(
                info->message_attr,
                GLOBUS_SOAP_MESSAGE_AUTHZ_TARGET_NAME_KEY,
                NULL,
                NULL,
                info->identity));

        if(info->streaming)
        {
            globus_i_test_result("setting ftp identity",
                globus_ftp_client_operationattr_set_authorization(
                    &info->ftp_attr,
                    GSS_C_NO_CREDENTIAL,
                    NULL,
                    NULL,
                    NULL,
                    info->subject));
        }
        break;

      case GLOBUS_I_AUTH_HOST:
      default:
        value = GLOBUS_SOAP_MESSAGE_AUTHZ_HOST;
        break;
    }

    globus_i_test_result("setting authentication mode",
        globus_soap_message_attr_set(
            info->message_attr,
            GLOBUS_SOAP_MESSAGE_AUTHZ_METHOD_KEY,
            NULL,
            NULL,
            (void *) value));

    globus_i_test_result("setting security mode",
        globus_soap_message_attr_set(
            info->message_attr,
            GLOBUS_SOAP_MESSAGE_AUTHENTICATION_METHOD_KEY,
            NULL,
            NULL,
            (void *) GLOBUS_SOAP_MESSAGE_AUTH_SECURE));

    if(info->encrypt)
    {
        globus_i_test_result("setting protection level",
            globus_soap_message_attr_set(
                info->message_attr,
                GLOBUS_SOAP_MESSAGE_AUTH_PROTECTION_KEY,
                NULL,
                NULL,
                (void *) GLOBUS_SOAP_MESSAGE_AUTH_PROTECTION_PRIVACY));

        globus_i_test_result("setting protection level",
            globus_ftp_client_operationattr_set_data_protection(
                &info->ftp_attr,
                GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE));
    }

    globus_i_test_result("setting timeout",
        globus_soap_message_attr_set(
            info->message_attr,
            GLOBUS_SOAP_MESSAGE_TIMEOUT_KEY,
            NULL,
            NULL,
            &info->http_timeout));

    if(info->streaming)
    {
        if(!info->use_mode_s)
        {
            globus_i_test_result("setting mode e",
                globus_ftp_client_operationattr_set_mode(
                    &info->ftp_attr, GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK));
        }

        if(info->nodcau)
        {
            globus_ftp_control_dcau_t       dcau;

            dcau.mode = GLOBUS_FTP_CONTROL_DCAU_NONE;
            globus_i_test_result("setting no dcau",
                globus_ftp_client_operationattr_set_dcau(
                    &info->ftp_attr, &dcau));
        }

        if(info->allow_ipv6)
        {
            globus_i_test_result("setting ipv6",
                globus_ftp_client_operationattr_set_allow_ipv6(
                    &info->ftp_attr, GLOBUS_TRUE));
        }
    }
}

static
void
globus_l_module_activate(
    globus_module_descriptor_t *        module)
{
    int                                 rc;

    rc = globus_module_activate(module);
    if(rc != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "globusrun-ws: error while activating %s: %d\n",
            module->module_name, rc);
        exit(1);
    }
}

globus_result_t
globus_l_output_submission_id(
    const char *                        filename,
    const char *                        value)
{
    FILE *                              file;
    int                                 len;

    file = fopen(filename, "w");
    if(!file)
    {
        goto error;
    }
    len = strlen(value);
    if(len != fwrite(value, 1, len, file))
    {
        goto error;
    }
    fclose(file);

    return GLOBUS_SUCCESS;

error:
    return GlobusLSystemError("writing submission id", errno);
}

int
main(
    int                                 argc,
    char **                             argv)
{
    globus_i_info_t                     info;
    globus_result_t                     result;
    globus_result_t                     cleanup_result = GLOBUS_SUCCESS;
    wsgram_CreateManagedJobInputType    input;
    globus_i_monitor_info_t             monitor;
    int                                 rc = 0;

    globus_l_module_activate(GLOBUS_COMMON_MODULE);
    globus_l_module_activate(GLOBUS_GSI_CERT_UTILS_MODULE);
    globus_l_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE);
    globus_l_module_activate(GLOBUS_XIO_MODULE);
    globus_l_module_activate(GLOBUS_FTP_CLIENT_MODULE);
    globus_l_module_activate(GLOBUS_FTP_CLIENT_DEBUG_PLUGIN_MODULE);
    globus_l_module_activate(GLOBUS_SOAP_MESSAGE_MODULE);
    globus_l_module_activate(GLOBUS_NOTIFICATION_CONSUMER_MODULE);
    globus_l_module_activate(GLOBUS_DELEGATION_CLIENT_UTIL_MODULE);
    globus_l_module_activate(MANAGEDJOBFACTORYSERVICE_MODULE);
    globus_l_module_activate(MANAGEDJOBSERVICE_MODULE);
    globus_l_module_activate(MANAGEDEXECUTABLEJOBSERVICE_MODULE);
    globus_l_module_activate(MANAGEDMULTIJOBSERVICE_MODULE);
    globus_l_module_activate(DELEGATIONSERVICE_MODULE);
    globus_l_module_activate(SUBSCRIPTIONMANAGERSERVICE_MODULE);

    globus_l_info_init(&info);
    globus_i_parse_arguments(argc, argv, &info);
    globus_l_setup_attributes(&info);

    result = globus_callback_register_signal_handler(
        GLOBUS_SIGNAL_INTERRUPT, GLOBUS_TRUE, globus_l_interrupt_cb, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        result = GlobusLErrorWrapResult("registering signal handler", result);
        goto wrapup;
    }

    wsgram_CreateManagedJobInputType_init_contents(&input);

    if((info.mode == GLOBUS_I_SUBMIT && !info.batch) ||
        info.mode == GLOBUS_I_MONITOR)
    {
        result = globus_i_monitor_init(&monitor, &info);
        if(result != GLOBUS_SUCCESS)
        {
            goto destroy_input;
        }
    }

    if(info.mode == GLOBUS_I_VALIDATE || info.mode == GLOBUS_I_SUBMIT)
    {
        result = globus_i_load_and_validate(&info, &input);
        if(result != GLOBUS_SUCCESS || info.mode == GLOBUS_I_VALIDATE)
        {
            goto destroy_monitor;
        }

        result = globus_i_prepare_job(&info, &input);
        if(result != GLOBUS_SUCCESS)
        {
            goto destroy_monitor;
        }

        result = globus_i_delegate(&info, &input);
        if(result != GLOBUS_SUCCESS)
        {
            goto destroy_monitor;
        }

        if(!info.quiet)
        {
            fprintf(stderr, "Submitting job...");
        }

        result = globus_i_submit(&info,
            info.batch ? NULL : &monitor.subscribe,
            &monitor.subscribe_epr, &input, &info.job_epr);
        if(result != GLOBUS_SUCCESS)
        {
            if(!info.quiet)
            {
                fprintf(stderr, "Failed.\n");
            }
            globus_i_clear_canceled();
            goto destroy_delegation;
        }

        if(!info.quiet)
        {
            char                        tmp[1024];

            strftime(tmp, sizeof(tmp), "%m/%d/%Y %H:%M GMT",
                input.InitialTerminationTime);
            fprintf(stderr, "Done.\nJob ID: %s\nTermination time: %s\n",
                input.JobID->base_value, tmp);
        }

        /* only output to stdout in batch mode if not quiet */
        if(info.job_epr_output_file || (info.batch && !info.quiet))
        {
            result = globus_i_output_epr(
                info.job_epr_output_file
                    ? info.job_epr_output_file : "stdout://",
                info.job_epr);
            if(result != GLOBUS_SUCCESS)
            {
                /* XXX not sure how else I should indicate this failure
                 * may want to actually fail and clean up job in batch mode
                 */
                globus_i_print_warning(result);
                result = GLOBUS_SUCCESS;
            }
            else if(!info.job_epr_output_file)
            {
                fprintf(stdout, "\n");
            }
        }

        if(info.submission_id_output_file)
        {
            result = globus_l_output_submission_id(
                info.submission_id_output_file,
                input.JobID->base_value);
            if(result != GLOBUS_SUCCESS)
            {
                if(!info.quiet)
                {
                    globus_i_print_warning(result);
                }
                result = GLOBUS_SUCCESS;
            }
        }

        if(info.batch)
        {
            goto destroy_monitor;
        }
    }

    if(info.mode == GLOBUS_I_STATUS)
    {
        result = globus_i_status(&info);
        goto destroy_monitor;
    }

    if(info.mode == GLOBUS_I_SUBMIT || info.mode == GLOBUS_I_MONITOR)
    {
        result = globus_i_monitor(&monitor);
    }

    /* only kill the job if this is a kill request or job was successfully
     * monitored.
     */
    if(info.mode == GLOBUS_I_KILL ||
        ((info.mode == GLOBUS_I_SUBMIT || info.mode == GLOBUS_I_MONITOR) &&
        (result == GLOBUS_SUCCESS ||
            (globus_i_check_canceled() && !info.no_cleanup_on_interrupt))))

    {
        if(info.mode != GLOBUS_I_SUBMIT)
        {
            /* any other mode would not have the job input which contains
             * hints on which delegated creds need to be destroyed
             */
            if(globus_i_check_canceled() && !info.no_cleanup_on_interrupt)
            {
                if(!info.quiet)
                {
                    fprintf(stderr,
                        "Canceled.\nRequesting original job description...");
                }

                globus_i_clear_canceled();
            }
            else if(!info.quiet)
            {
                fprintf(stderr, "Requesting original job description...");
            }
            cleanup_result = globus_i_query_job_description(&info, &input);
            if(cleanup_result != GLOBUS_SUCCESS)
            {
                if(!info.quiet)
                {
                    fprintf(stderr, "Failed.\n");
                }
                goto destroy_monitor;
            }
            else if(!info.quiet)
            {
                fprintf(stderr, "Done.\n");
            }
        }

/* destroy_job: */
        if(globus_i_check_canceled() && !info.no_cleanup_on_interrupt)
        {
            if(!info.quiet)
            {
                fprintf(stderr, "Canceled.\nDestroying job...");
            }

            globus_i_clear_canceled();
        }
        else if(!info.quiet)
        {
            fprintf(stderr, "Destroying job...");
        }

        cleanup_result = globus_i_kill(&info);
        if(cleanup_result != GLOBUS_SUCCESS)
        {
            if(!info.quiet)
            {
                fprintf(stderr, "Failed.\n");
            }
            
            goto destroy_monitor;
        }
        else if(!info.quiet)
        {
            fprintf(stderr, "Done.\n");
        }

destroy_delegation:
        cleanup_result = globus_i_delegate_cleanup(&info, &input);
    }

destroy_monitor:
    if((info.mode == GLOBUS_I_SUBMIT && !info.batch) ||
        info.mode == GLOBUS_I_MONITOR)
    {
        globus_i_monitor_destroy(&monitor);
    }
destroy_input:
    wsgram_CreateManagedJobInputType_destroy_contents(&input);

    if(globus_l_cancel_count < 2)
    {
        globus_callback_unregister_signal_handler(
            GLOBUS_SIGNAL_INTERRUPT, NULL, NULL);
    }

wrapup:
    /**
     * return code logic...
     * 
     * if result != GLOBUS_SUCCESS, then there was an application fault.
     * and the job or request was not submitted successfully.  Errors that
     * occur after a submitted job is Failed or Done are ignored or saved to
     * cleanup_result.  Errors dealing with streaming mode are ignored.
     * 
     * exit status is determined in the following order:
     * 
     * A successful -status on a job that has not reached Failed or Done
     *    will exit with 0
     * 
     * If a job makes it to the Failed state and there was a job fault,
     *    the exit status will be -1 (-status, -submit, -monitor)
     * 
     * If a job makes it to the Failed or Done state and there was no job
     *    fault, then the exit status will be that returned by GRAM
     *    (likely 0 for a multijob) (-status, -submit, -monitor)
     * 
     * Any error received before the Failed or Done state for -submit or
     *    -monitor mode or any error received for the -status or -validate
     *    requests will cause an exit status of -3
     * 
     * Any error with -kill will cause an exit status of -2
     */
    if(result != GLOBUS_SUCCESS)
    {
        rc = -3;
    }
    else if(cleanup_result != GLOBUS_SUCCESS)
    {
        rc = -2;
    }
    
    if(result != GLOBUS_SUCCESS)
    {
        char *                          tmp;

        tmp = globus_error_print_friendly(globus_error_peek(result));
        fprintf(stderr, "globusrun-ws: %s", tmp);
        globus_free(tmp);
    }
    else if((info.mode == GLOBUS_I_SUBMIT && !info.batch) ||
        info.mode == GLOBUS_I_MONITOR || info.mode == GLOBUS_I_STATUS)
    {
        if(info.job_fault)
        {
            char *                          tmp;
            
            rc = -1;
            tmp = globus_error_print_friendly(info.job_fault);
            fprintf(stderr, "globusrun-ws: Job failed: %s", tmp);
            globus_free(tmp);
        }
        else
        {
            rc = info.job_exit_code;
        }
    }
    
    if(cleanup_result != GLOBUS_SUCCESS)
    {
        char *                          tmp;

        tmp = globus_error_print_friendly(globus_error_peek(cleanup_result));
        fprintf(stderr, "globusrun-ws: Unable to destroy job: %s", tmp);
        globus_free(tmp);
    }
    
    globus_l_info_destroy(&info);
    globus_module_deactivate_all();
    return rc;
}

void
globus_i_print_error(
    globus_result_t                     result)
{
    char *                              tmp;

    tmp = globus_error_print_friendly(globus_error_peek(result));
    fprintf(stderr, "globusrun-ws: %s", tmp);
    globus_free(tmp);
}

void
globus_i_print_warning(
    globus_result_t                     result)
{
    char *                              tmp;

    tmp = globus_error_print_friendly(globus_error_peek(result));
    fprintf(stderr, "globusrun-ws: Warning: %s", tmp);
    globus_free(tmp);
}
