/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.monitoring;

import org.globus.exec.generated.StateEnumeration;

/**
 * Class used internally by the JobStateMonitor to store events emmited
 * by the Scheduler Event Generator.
 */
class SchedulerEvent
{
    private java.util.Date timestamp;
    private String localId;
    private StateEnumeration state;
    private int exitCode;
    static private Comparator comparator = new Comparator();

    /**
     * Create a new SchedulerEvent.
     *
     * Event contents cannot change after being created.
     */
    public SchedulerEvent(
            java.util.Date              timestamp,
            String                      localId,
            StateEnumeration            state,
            int                         exitCode)
    {
        this.timestamp = timestamp;
        this.localId = localId;
        this.state = state;
        this.exitCode = exitCode;
    }

    /** Retreive the value of the event's time stamp */
    public java.util.Date getTimeStamp() { return timestamp; }
    /** Retrieve the value of the event's local ID */
    public String getLocalId() { return localId; }
    /** Return the value of the event's state */
    public StateEnumeration getState() { return state; }
    /** Return the value of the event's exit code */
    public int getExitCode() { return exitCode; }

    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append(localId);
        sb.append(" [");
        sb.append(timestamp.toString());
        sb.append("] ");

        sb.append(state.toString());

        return sb.toString();
    }

    /**
     * Return a reference to the comparator used to order events.
     * Used by the JobStateMonitor's event cache.
     */
    static public java.util.Comparator getComparator() {
        return comparator;
    }
    private static class Comparator implements java.util.Comparator
    {
        /**
         * Compare two SchedulerEvents.
         * An event e1 is less than another e2 if (in order of importance)
         * it
         * <ul>
         *   <li>occurs before the other</li>
         *   <li>has a lower-value local id</li>
         *   <li> has a lower-value state</li>
         * </ul>
         */
        public int compare(Object o1, Object o2)
                throws ClassCastException
        {
            SchedulerEvent e1, e2;
            int rc;
            StateEnumeration s1, s2;
            int ec1, ec2;

            e1 = (SchedulerEvent) o1;
            e2 = (SchedulerEvent) o2;

            rc = e1.timestamp.compareTo(e2.timestamp);

            if (rc != 0) {
                return rc;
            }

            rc = e1.localId.compareTo(e2.localId);

            if (rc != 0) {
                return rc;
            }

            s1 = e1.state;
            s2 = e2.state;

            if (! s1.equals(s2)) {
                if (s1.equals(StateEnumeration.Done)) {
                    return 1;
                } else if (s2.equals(StateEnumeration.Done)) {
                    return -1;
                } else if (s1.equals(StateEnumeration.Failed)) {
                    return 1;
                } else if (s2.equals(StateEnumeration.Failed)) {
                    return -1;
                } else if (s1.equals(StateEnumeration.StageOut)) {
                    return 1;
                } else if (s2.equals(StateEnumeration.StageOut)) {
                    return -1;
                } else if (s1.equals(StateEnumeration.Active)) {
                    return 1;
                } else if (s2.equals(StateEnumeration.Active)) {
                    return -1;
                } else if (s1.equals(StateEnumeration.Pending)) {
                    return -1;
                } else if (s2.equals(StateEnumeration.Pending)) {
                    return 1;
                } else if (s1.equals(StateEnumeration.StageIn)) {
                    return 1;
                } else if (s2.equals(StateEnumeration.StageIn)) {
                    return -1;
                }
            }

            ec1 = e1.exitCode;
            ec2 = e2.exitCode;

            if (ec1 < ec2)
            {
                return -1;
            }
            else if (ec1 > ec2)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }
    }
}
