/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.exec.service.exec.utils;

import java.util.LinkedList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.exec.monitoring.AlreadyRegisteredException;
import org.globus.exec.utils.ManagedExecutableJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.exec.service.exec.ManagedExecutableJobResource;
import org.globus.exec.service.exec.StateMachine;

import org.globus.util.I18n;
import org.globus.wsrf.ResourceKey;

public class JobStateMonitorSubscriptionManager
    extends                                 Thread
{
    /** Log4J logger */
    private static Log logger
        = LogFactory.getLog(JobStateMonitorSubscriptionManager.class);
    static I18n i18n = I18n.getI18n(Resources.class.getName());

    LinkedList readyQueue = new LinkedList();
    boolean done = false;

    private static int DEFAULT_SUBSCRIPTION_MANAGER_THREAD_COUNT = 1;
    private static int subscriptionManagerIndex;
    private static JobStateMonitorSubscriptionManager[] INSTANCES;

    public static int SUBSCRIBE_REQUEST     = 0;
    public static int UNSUBSCRIBE_REQUEST   = 1;

    private static int MAX_REGISTER_ATTEMPTS = 5;

    static
    {
        //create the run queue instance array
        int subscriptionManagerCount
            = DEFAULT_SUBSCRIPTION_MANAGER_THREAD_COUNT;
        JobStateMonitorSubscriptionManager.INSTANCES
            = new JobStateMonitorSubscriptionManager[subscriptionManagerCount];

        //start all run queue instances
        for (int index=0; index<INSTANCES.length; index++)
        {
            JobStateMonitorSubscriptionManager.INSTANCES[
                index]
                = new JobStateMonitorSubscriptionManager();
            JobStateMonitorSubscriptionManager.INSTANCES[
                index].setDaemon(true);
            JobStateMonitorSubscriptionManager.INSTANCES[
                index].start();
        }

        //set initial run queue index
        JobStateMonitorSubscriptionManager.subscriptionManagerIndex = 0;
    }

    private JobStateMonitorSubscriptionManager() { }

    public static synchronized JobStateMonitorSubscriptionManager getInstance()
    {
        JobStateMonitorSubscriptionManager subscriptionManager
            = JobStateMonitorSubscriptionManager.INSTANCES[
                JobStateMonitorSubscriptionManager.subscriptionManagerIndex];

        if (   JobStateMonitorSubscriptionManager.subscriptionManagerIndex
            == (JobStateMonitorSubscriptionManager.INSTANCES.length-1))
        {
            JobStateMonitorSubscriptionManager.subscriptionManagerIndex = 0;
        } else
        {
            JobStateMonitorSubscriptionManager.subscriptionManagerIndex++;
        }

        return subscriptionManager;
    }

    public void add(
        ManagedExecutableJobResource        resource,
        int                                 action)
    {
        if (logger.isTraceEnabled())
        {
            logger.debug("Entering add()");
        }

        synchronized (this)
        {
            if (logger.isDebugEnabled()) {
                logger.debug("adding " + resource.getID().toString()
                        + " to queue.");
            }
            readyQueue.add(new SubscriptionEvent(resource, action));
            notify();
        }

        if (logger.isTraceEnabled())
        {
            logger.trace("Leaving add()");
        }
    }

    public synchronized void terminate()
    {
        done = true;
    }

    public void run()
    {
        while (true)
        {
            SubscriptionEvent event;

            synchronized (this)
            {
                while ((!done) && (readyQueue.size() == 0))
                {
                    try
                    {
                        wait();
                    } catch (InterruptedException e) { }
                }
            }

            while (! readyQueue.isEmpty())
            {
                synchronized (this)
                {
                    event = (SubscriptionEvent) readyQueue.removeFirst();
                }

                if (event.getAction() == SUBSCRIBE_REQUEST)
                {
                    ManagedExecutableJobResource resource = event.getResource();
                    try
                    {
                        subscribe(resource);
                    }
                    catch (AlreadyRegisteredException e)
                    {
                        if (event.isMaxRegisterAttemptsExceeded())
                        {
                            //max register attempts exceeded, so fail the job
                            String errorMessage = i18n.getMessage(
                                Resources.JOB_ID_REGISTRATION_ERROR);

                            StateMachine.systemCancel(
                                resource, errorMessage, e);
                        }
                        else
                        {
                            //send event to the back of the queue
                            if (logger.isDebugEnabled())
                            {
                                logger.warn(
                                    "Register attempt failed for job "
                                    + ((ResourceKey) resource.getID())
                                      .getValue().toString()
                                    + ". Assuming repeat local job ID and "
                                    + "retrying.");
                            }
                            readyQueue.add(event);
                            //Don't notify(). Wait for next event to try;

                        }
                    }
                }
                else
                {
                    unsubscribe(event.getResource());
                }

                if (logger.isDebugEnabled())
                {
                    logger.debug("Done processing state for "
                            + event.getResource().getID().toString());
                }
                synchronized (this) {
                    if (done) {
                        break;
                    }
                }
            }
        }
    }

    private void subscribe(
        ManagedExecutableJobResource        resource)
        throws                              AlreadyRegisteredException
    {
        logger.trace("subscribing resource for JSM notifications");
        String[] localJobIds = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
        if (localJobIds != null)
        {
            int size = localJobIds.length;
            for (int i = 0; i < size; i++) {
                if (logger.isDebugEnabled()) {
                    logger.trace("Registering job ID " + localJobIds[i]
                                +" for JSM subscriber " + resource.getID());
                }
                resource.getJobStateMonitor().registerJobID(
                    localJobIds[i],
                    (ResourceKey) resource.getID());
            }
        }

        if (logger.isDebugEnabled()) {
            logger.trace("Exiting subscribeForNotifications()");
        }
    }

    private void unsubscribe(
        ManagedExecutableJobResource        resource)
    {
        if (logger.isDebugEnabled())
        {
            logger.trace("unsubscribing resource for JSM notifications");
        }

        String[] localJobIds = (String[]) resource.getResourceDatum(
            ManagedExecutableJobConstants.RD_LOCAL_JOB_IDS);
        if (localJobIds != null)
        {
            int size = localJobIds.length;

            for (int i = 0; i < size; i++)
            {
                try
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug(
                            "Unregistering job ID from JSM subscribers");
                    }

                    resource.getJobStateMonitor().unregisterJobID(
                        localJobIds[i]);

                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Job ID Unregistered.");
                    }
                }
                catch (Exception e)
                {
                    String errorMessage = i18n.getMessage(
                        Resources.JOB_ID_UNREGISTRATION_ERROR);
                    logger.debug(errorMessage, e); //debug not error...
                }
            }
        }
    }

    class SubscriptionEvent
    {
        private ManagedExecutableJobResource resource;
        private int action;
        private int registerAttemptCount = 0;

        public SubscriptionEvent(
            ManagedExecutableJobResource    resource,
            int                             action)
        {
            this.resource = resource;
            this.action = action;
        }

        public ManagedExecutableJobResource getResource()
        {
            return this.resource;
        }

        public int getAction()
        {
            return this.action;
        }

        public boolean isMaxRegisterAttemptsExceeded()
        {
            if (this.registerAttemptCount == MAX_REGISTER_ATTEMPTS)
            {
                return true;
            }

            this.registerAttemptCount++;
            return false;
        }
    }
}
