/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.aggregator.impl;

import org.globus.mds.aggregator.types.AggregatorContent;
import org.globus.mds.aggregator.types.AggregatorConfig;
import org.globus.mds.aggregator.types.AggregatorData;
import org.globus.mds.aggregator.types.PairedKeyType;

import org.globus.mds.usefulrp.ServiceMetaDataResourceProperty;

import org.apache.axis.encoding.AnyContentType;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.ResourceException;
import org.globus.wsrf.RemoveCallback;

import org.globus.wsrf.ResourceLifetime;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import org.globus.wsrf.impl.servicegroup.ServiceGroupEntryResource;
import org.globus.wsrf.impl.servicegroup.ServiceGroupConstants;

import java.text.DateFormat;
import java.util.Calendar;
import java.util.Date;
import javax.xml.namespace.QName;

/** 
 * Implements an aggregating service group entry resource.
 */

public class AggregatorServiceGroupEntryResource 
    implements ServiceGroupEntryResource,
               ResourceProperties,
               ResourceLifetime,
               RemoveCallback
{
    private static Log logger =
            LogFactory.getLog(AggregatorServiceGroupEntryResource.class.getName());

    public static final QName RP_SET =
        new QName("http://mds.globus.org/aggregator", 
                  "AggregatorServiceGroupEntryRP");

    private ResourceKey key = null;
    private ResourcePropertySet propSet = null;
    private EndpointReferenceType entryEPR = null;
    private EndpointReferenceType memberEPR = null;
    private EndpointReferenceType serviceGroupEPR = null;

    // Resource properties defined in ServiceGroupEntry
    private ResourceProperty contentRP = null;
    private ResourceProperty memberRP = null;
    private ResourceProperty serviceGroupEPRRP = null;
    private ServiceMetaDataResourceProperty metadata = null;
    
    private AggregatorContent content = null;
    private AggregatorSource source = null;
    private AggregatorServiceGroupResource aggrResource = null;
    
    public AggregatorServiceGroupEntryResource(AggregatorServiceGroupResource aggrResource,
                                               EndpointReferenceType memberEPR,
                                               EndpointReferenceType serviceGroupEPR,
                                               AggregatorContent content) 
    {
        this.aggrResource = aggrResource;
        this.memberEPR = memberEPR;
        this.serviceGroupEPR = serviceGroupEPR;
        this.content = content;
        
        try 
        {           
            this.propSet = new SimpleResourcePropertySet(RP_SET);

            PairedKeyType pk = new PairedKeyType();
            pk.setGroupKey((String)this.aggrResource.getKey().getValue());
            pk.setEntryKey(String.valueOf(this.hashCode()));            
            this.key = 
                new SimpleResourceKey(ServiceGroupConstants.ENTRY_KEY, pk);          

            initializeServiceGroupRPs();
            initializeLifetimeRPs();
            
            // add a metadata RP
            metadata=new ServiceMetaDataResourceProperty();
            this.propSet.add(metadata);         

        }
        catch(Exception e) 
        {
            logger.error("Exception constructing service group entry: "+e);
        }   

    }

    /** called by constructor to set up WS-Resource Lifetime RPs */
    private void initializeLifetimeRPs() throws Exception {
        ResourceProperty prop;

        prop = new ReflectionResourceProperty(
            SimpleResourcePropertyMetaData.TERMINATION_TIME, this);
        this.propSet.add(prop);

        prop = new ReflectionResourceProperty(
            SimpleResourcePropertyMetaData.CURRENT_TIME, this);
        this.propSet.add(prop);
    }

    /** called by constructor to set up WS-ServiceGroup RPs */
    private void initializeServiceGroupRPs() 
        throws Exception {
        // WS-ServiceGroup RPs

        this.contentRP = 
            new SimpleResourceProperty(ServiceGroupConstants.CONTENT); 
        this.propSet.add(contentRP);
        this.setContent(content);

        this.memberRP = 
            new SimpleResourceProperty(ServiceGroupConstants.MEMBER_EPR); 
        this.memberRP.add(memberEPR);
        this.propSet.add(memberRP);            

        this.serviceGroupEPRRP = 
            new SimpleResourceProperty(ServiceGroupConstants.SERVICE_GROUP_EPR); 
        this.serviceGroupEPRRP.add(serviceGroupEPR);
        this.propSet.add(serviceGroupEPRRP);

    }

    public void setVersion(String version) {
        this.metadata.setVersion(version);
    }

    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    public ResourceKey getKey() {
        return this.key;
    }

    public EndpointReferenceType getEntryEPR() {
        return this.entryEPR;
    }

    public void setEntryEPR(EndpointReferenceType epr) {
        this.entryEPR = epr;
    }
    public EndpointReferenceType getMemberEPR() {
        return this.memberEPR;
    }

    public void setAggregatorSource(AggregatorSource source) {
        this.source = source;
    }
    
    public AggregatorSource getAggregatorSource() {
        return this.source;
    }
    
    public AggregatorContent getContent() {
        return this.content;
    }

    public void setContent(AggregatorContent content) 
    {              
        if (content == null) {
            return;
        }

        this.content = content;

        this.contentRP.clear();
        this.contentRP.add(this.content);
        
        if (this.aggrResource.getConfiguration().getNotifyOnEntryContentChangeValue()) {
            this.aggrResource.notifyChange();        
        }
    } 

    // lifetime management stuff

    Calendar terminationTime = null;
    DateFormat dateFormat = DateFormat.getTimeInstance(DateFormat.FULL);

    public Calendar getCurrentTime() {
        Calendar now=Calendar.getInstance();
        logger.debug("getCurrentTime()="+dateFormat.format(now.getTime()));
        return now;
    }

    public Calendar getTerminationTime() {
        if(terminationTime!=null) {
            logger.debug("getTerminationTime()="+dateFormat.format(terminationTime.getTime()));
        } else {
            logger.debug("getTerminationTime() returns null");
        }
        return terminationTime;
    }

    public void setTerminationTime(Calendar t) {
        Calendar now = Calendar.getInstance();
        logger.debug("setTerminationTime()="
                    +dateFormat.format(t.getTime())
                    + ", invoked at "
                    +dateFormat.format(now.getTime()));


        if(now.after(t)) {
            logger.warn("New termination time is in past!");
        }

        terminationTime = t;
        
        if (this.aggrResource.getConfiguration().getNotifyOnEntryLifetimeExtensionValue()) {
            this.aggrResource.notifyChange();
        }
    }
    
    public void remove() throws ResourceException 
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Removed a service group entry resource: " + this.toString());
        }        
        
        AggregatorServiceGroupEntryRemovedCallback callback = 
            aggrResource.getConfiguration().getEntryRemovedCallbackInstance();
        
        if (callback != null) {
            if (logger.isDebugEnabled()) {
                logger.debug("Invoking remove method for AggregatorServiceGroupEntryRemoveCallback class: " +
                             callback.getClass().getName());
            }
            try {
                callback.remove(this);
            } catch (Exception e) {
                throw new ResourceException("Unhandled Exception during remove callback", e);
            }
        }
    }
    
}

