/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.aggregator.impl;

import org.globus.mds.aggregator.types.AggregatorContent;
import org.globus.mds.aggregator.types.AggregatorConfig;
import org.globus.mds.aggregator.types.AggregatorData;

import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.usefulrp.AccumulatorResourceProperty;
import org.globus.mds.usefulrp.ServiceMetaDataResourceProperty;

import org.globus.wsrf.Constants;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.utils.AddressingUtils;

import org.oasis.wsrf.servicegroup.EntryType;
import org.oasis.wsrf.servicegroup.Add;
import org.oasis.wsrf.servicegroup.AddRefusedFaultType;
import org.oasis.wsrf.servicegroup.ContentCreationFailedFaultType;
import org.oasis.wsrf.servicegroup.UnsupportedMemberInterfaceFaultType;

import org.globus.wsrf.impl.servicegroup.EntryResourceProperty;
import org.globus.wsrf.impl.servicegroup.EntryResourcePropertyTopic;
import org.globus.wsrf.impl.servicegroup.ServiceGroupConstants;
import org.globus.wsrf.impl.servicegroup.ServiceGroupRegistrationResource;

import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.FaultHelper;
import org.globus.wsrf.utils.XmlUtils;

import java.rmi.RemoteException;
import java.text.DateFormat;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.lang.reflect.Constructor;

import java.net.InetAddress;
import java.net.UnknownHostException;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

/** Implements an aggregating in-memory service group resource.
 *  <br>
 *  Entries will be stored in a hash table under each service group resource
 *  so that we can easily find them; the key for a service group entry
 *  must first contain the key to the appropriate service group and then
 *  the key within that to the appropriate entry. <br />
 * 
 */

public class AggregatorServiceGroupResource 
    implements ResourceProperties, 
               ServiceGroupRegistrationResource 
{

    /** The default entry sweeper, in milliseconds. <br />*/
    public static final long DEFAULT_SWEEP_PERIOD=300000;

    private static Log logger =
            LogFactory.getLog(AggregatorServiceGroupResource.class.getName());

    public static final QName RP_SET =
        new QName("http://mds.globus.org/aggregator",
                  "AggregatorServiceGroupRP");

    public static final QName REGISTRATION_COUNT_RP =
        new QName("http://mds.globus.org/aggregator",
                  "RegistrationCount");

    ResourcePropertySet propSet;

    /** Maps entry keys to entry resource objects. */
    Hashtable entries = new Hashtable();

    /** Maps entry keys to session keys. */
    Hashtable sessionKeys = new Hashtable();

    /** A cache to return concurrency exception safe iterators from */
    Hashtable entryCache = null;
    boolean entryCacheDirty = false;
    
    EntrySweeper entrySweeper = new EntrySweeper();
    private long sweeperInterval = DEFAULT_SWEEP_PERIOD;
    
    /** the AggregationSource(s) used to populate content for this SG */
    ArrayList sources = new ArrayList();

    AccumulatorResourceProperty registrationCountRP;
    ServiceMetaDataResourceProperty metadata;
    AggregatorConfiguration configuration = null;
    
    public AggregatorServiceGroupResource()  {        
        
    }
    
    protected void init(QName rpSetName)
    {
        this.propSet = new SimpleResourcePropertySet(rpSetName);

        // initialise the entry RP; this will need to be generated
        // dynamically from all of the entries within this ServiceGroup
        try {
            ResourceProperty prop;
            prop=new EntryResourceProperty(this);
            this.propSet.add(prop);
        } catch(Exception e) {
            logger.error("Couldn't set up Entry RP");
        }

        // initialise the regcount RP; this will need to be generated
        // dynamically from all of the entries within this ServiceGroup
        try {
            registrationCountRP=new AccumulatorResourceProperty(REGISTRATION_COUNT_RP);
            this.propSet.add(registrationCountRP);
        } catch(Exception e) {
            logger.error("Couldn't set up registration count RP");
        }

        // add a metadata RP
        try {
            metadata=new ServiceMetaDataResourceProperty();
            this.propSet.add(metadata);
        } catch(Exception e) {
            logger.error("Couldn't set up metadata RP");
        }
        
        this.sweeperInterval = getConfiguration().getEntrySweeperInterval();
        entrySweeper.setDaemon(true);
        entrySweeper.start();
    }

    public void addAggregatorSource(AggregatorSource source) {
        this.sources.add(source);
    }
        
    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    public void setVersion(String version) {
        this.metadata.setVersion(version);
    }

    private ResourceKey key =
        new SimpleResourceKey(ServiceGroupConstants.KEY, String.valueOf(hashCode()));

    public ResourceKey getKey() {
        return key;
    }

    // to be a servicegroup

    public synchronized Iterator iterateEntryKeys() 
    {    
        if (this.entryCache == null || this.isEntryCacheDirty()) {
            // TODO: time this to see the performance impact
            this.entryCache =
                (Hashtable) this.entries.clone();
            this.setEntryCacheDirty(false);
        }
        return entryCache.keySet().iterator();     
    }
    
    public synchronized Iterator iterateEntryValues() {
        if (this.entryCache == null || this.isEntryCacheDirty()) {
            // TODO: time this to see the performance impact
            this.entryCache =
                (Hashtable) this.entries.clone();
            this.setEntryCacheDirty(false);
        }        
        return entryCache.entrySet().iterator();
    }
    
    private boolean isEntryCacheDirty() {
        return this.entryCacheDirty;
    }

    private synchronized void setEntryCacheDirty(boolean dirty) {
        this.entryCacheDirty = dirty;
    }

    public AggregatorConfiguration getConfiguration() 
    {
        if (this.configuration != null) {
            return this.configuration;
        }
        MessageContext ctx = MessageContext.getCurrentContext();

        String servicePath = ctx.getTargetService();

        String jndiName = Constants.JNDI_SERVICES_BASE_NAME + servicePath +
               "/configuration";
        logger.debug("Will read configuration from jndi name: "+jndiName);
        try {
            Context initialContext = new InitialContext();
            this.configuration =
                (AggregatorConfiguration) initialContext.lookup(jndiName);

        } catch (NamingException e) {
            logger.error("when performing JNDI lookup for "+jndiName+": "+e );
        } catch (Exception e) {
            logger.error("when performing JNDI lookup for "+jndiName+": "+e );
        }
        return this.configuration;
    }
    
    public void loadCompatibleSources(AggregatorSink sink) throws Exception
    {

        // at this point, we have the config options read in and
        // can construct the AggregatorSource instances
        
        String aggregatorClassName = "";
        List sources = getConfiguration().getAggregatorSourceList();
        
        for (int i=0;i<sources.size();i++)
        {            
            try 
            {
                aggregatorClassName = (String)sources.get(i);
                
                Constructor sourceConstructor = 
                    Class.forName(aggregatorClassName).getConstructor(
                        new Class[] { AggregatorSink.class } );

                AggregatorSource src = 
                    (AggregatorSource) sourceConstructor.newInstance(new Object[] {
                        sink } );

                src.initialize(configuration);
                this.addAggregatorSource(src);    
                
            } 
            catch (Exception e) 
            {
                if (e instanceof AggregatorSinkTypeNotSupportedException) 
                {
                    logger.error("The Aggregator source " + 
                                 aggregatorClassName + 
                                 " is not compatible with the specified Aggregator sink " + 
                                 sink.getClass().getName());
                }  else {
                    logger.error("Unknown exception during Aggregator initialization: " + e);
                }
                throw e;
            }
        }
    }    
    
    /** Adds a registration to the AggregatingServiceGroup. This has 
        two main effects:
        <ol>
          <li> it creates a service group entry for the registration 
               (normal service group behaviour) </li>
          <li> it calls addAggregation on the specified source, passing
               in the newly created entry. TODO: should we pass in a key
               here rather than a reference to the service object in 
               order to cope with activation/deactivation? </li>
        </ol>
    */
       
    public EndpointReferenceType serviceGroupAdd(Add add) 
        throws RemoteException, 
               AddRefusedFaultType,
               ContentCreationFailedFaultType,
               UnsupportedMemberInterfaceFaultType
    {
        // keep count
        registrationCountRP.accumulate(1);

        // First, process the content: check type, deserialize if needed. 
        // TODO: implement membershipContentRule checks
        AggregatorContent content = null;           
        Object obj = add.getContent();        
        
        if (obj instanceof SOAPElement)
        {
            try {
                content = (AggregatorContent) 
                    ObjectDeserializer.toObject((SOAPElement)obj,
                                                AggregatorContent.class);
            } catch (Exception e) {
                logger.warn("Exception deserializing Content parameter: " + e.toString());            
            }
            
        } else if (obj instanceof AggregatorContent) {
            content = (AggregatorContent) obj;
        }        
        
        /* 
         * 1. create an entry resource 
         *
         * Question: 
         *  Do we delete this on addAggregation fail, or should we just do 
         *  nothing and let the sweeper take care of it?
         */
       
        
        EndpointReferenceType entryEPR = null;    
        EndpointReferenceType serviceGroupEPR = null;
                    
        try
        {
            serviceGroupEPR = AddressingUtils.createEndpointReference(null);
        }
        catch(Exception e1)
        {
            throw new RemoteException("", e1);
        }
        
        if (AggregatorUtils.detectLoopback(add.getMemberEPR(),serviceGroupEPR)) {
            AddRefusedFaultType fault = new AddRefusedFaultType();
            FaultHelper faultHelper = new FaultHelper(fault);
            faultHelper.setDescription("Loopback or duplicate registrant address submitted");
            throw fault;            
        }
        
        AggregatorServiceGroupEntryResource entry = 
            new AggregatorServiceGroupEntryResource(this,
                                                    add.getMemberEPR(),
                                                    serviceGroupEPR,
                                                    content);

        ResourceKey key = entry.getKey();
        if (logger.isDebugEnabled()) {
            logger.debug("Created new resource key: "+key);
        }

        // construct an EPR to entry through the entry service.

        MessageContext msgContext = MessageContext.getCurrentContext();
        String transportURL =
            (String) msgContext.getProperty(MessageContext.TRANS_URL);
        // TODO: is there a better way to get the URL to use?
        transportURL = transportURL + "Entry" ;
        if (logger.isDebugEnabled()) {
            logger.debug("Will use transport URL in entry EPRs: "+transportURL);
        }
        
        try
        {
            entryEPR = AddressingUtils.createEndpointReference(transportURL, key);
        }
        catch(Exception e1)
        {
            throw new RemoteException("", e1);
        }

        entry.setEntryEPR(entryEPR);

        entry.setTerminationTime(add.getInitialTerminationTime());
        
        /*
         * 2. Try to perform the data aggregation.
         *
         * Since multiple aggregator sources are supported,
         * we will loop, passing the parameters to each one in turn, 
         * in order to see if it the parameter type is supported. The first
         * source found that performs a successful aggregation using the 
         * specified parameters will cause the loop to terminate. If no 
         * compatible source is found OR another error occurs while invoking
         * the addAggregation method, the operation will throw a remote fault 
         * with a description to that effect.
         */
              
        try 
        {
            Object parameters = content.getAggregatorConfig();
            
            if (parameters == null) {
                throw new AggregatorException("Parameters argument is null!");
            } else if (!(parameters instanceof AggregatorConfig)) {
                 throw new AggregatorException(
                    "Invalid Parameter Type (expected AggregatorConfig, but got " + 
                    parameters.getClass()+")");
            }
            
            AggregatorSource source = null;
            boolean found = false;
            for (int i=0;i<sources.size();i++) 
            {
                try {
                    source = (AggregatorSource)sources.get(i);
                    
                    String sessionKey = 
                        source.addAggregation(parameters, entry);
                    
                    entry.setAggregatorSource(source);
                    
                    sessionKeys.put(key,sessionKey);
                    
                    found = true;
                    break;
                }
                catch (Exception e) 
                {
                    if (e instanceof AggregatorParamTypeNotSupportedException) 
                    {
                        if (logger.isDebugEnabled()) 
                        {
                            AggregatorConfig config = 
                                (AggregatorConfig) parameters;
                        
                            logger.debug("The AggregatorSource " + 
                                         source.getClass().getName() + 
                                         " is unable to process the specified parameters: " + 
                                         XmlUtils.toString(
                                            ObjectSerializer.toElement(config,
                                                config.getTypeDesc().getXmlType())));
                        }
                        continue;
                    }
                    else 
                    {
                        throw e;
                    }                    
                }
            }
            if (!found) {
                throw new AggregatorException("addAggregation Operation Failed: Could not locate an Aggregator source that was compatible with the specified parameters");
            }
        } 
        catch (Exception e) 
        {
            AddRefusedFaultType fault = new AddRefusedFaultType();
            FaultHelper faultHelper = new FaultHelper(fault);
            faultHelper.setDescription("Exception while trying to process a data aggregation instruction");
            faultHelper.addFaultCause(e);
            throw fault;
        }
        
        // put it in the entry map
        entries.put(key,entry);
        this.setEntryCacheDirty(true);

        this.notifyChange();
        
        return entryEPR;
    } 

    public void notifyChange() 
    {
        try {
            ResourceProperty rpTopic =  
                this.propSet.get(ServiceGroupConstants.ENTRY);
            if (rpTopic instanceof EntryResourcePropertyTopic) {
                ((EntryResourcePropertyTopic)rpTopic).notify(null);
            }   
        } catch (Exception e) {
            logger.warn("Exception during ResourcePropertyTopic.notify: " + e);
        }
    }
    
    /** Gets rid of dead registrations */
    class EntrySweeper extends Thread {
        public void run() {
            if (logger.isDebugEnabled()) {
                logger.debug("Starting EntrySweeper thread...");
            }

            while(true) 
            {
                // statistics for this sweep run
                int totalSwept=0;
                int totalKept=0;
                int totalRemoved=0;


                try {
                    Thread.sleep(sweeperInterval);
                } catch(InterruptedException e) {
                    logger.warn("EntrySweeper interrupted. Terminating.");
                    return;
                }
                if (logger.isDebugEnabled()) {
                    logger.debug("Starting sweep");
                }
 
                synchronized(entries) 
                {    
                    // for every entry in the hash
                    Iterator i = entries.entrySet().iterator();
                    DateFormat dateFormat = 
                        DateFormat.getTimeInstance(DateFormat.FULL);

                    while(i.hasNext()) 
                    {
                        totalSwept++;
                        Map.Entry mapEntry = (Map.Entry)i.next();

                        AggregatorServiceGroupEntryResource entryResource = 
                        (AggregatorServiceGroupEntryResource) mapEntry.getValue();
                        ResourceKey entryResourceKey = 
                            (ResourceKey)mapEntry.getKey();

                        Calendar now = entryResource.getCurrentTime();
                        Calendar termination = 
                            entryResource.getTerminationTime();

                        if(termination == null) 
                        {
                            if (logger.isDebugEnabled()) {
                                logger.debug("Entry has infinite lifetime: " + 
                                             entryResource.toString());
                            }
                        } 
                        else 
                        {
                            if (logger.isDebugEnabled()) {
                                logger.debug("Checking entry: now="
                                    +dateFormat.format(now.getTime())
                                    +", termination="
                                    +dateFormat.format(termination.getTime()));
                            }

                            // if termination time is in past
                            if(now.after(termination)) 
                            {
                                if (logger.isDebugEnabled()) {                                
                                    logger.debug("Terminating entry");
                                }
                                String sessionKey = 
                                    (String)sessionKeys.get(entryResourceKey);

                                try {
                                    AggregatorSource source = 
                                        entryResource.getAggregatorSource();
                                    if (source != null) {
                                        source.removeAggregation(sessionKey);
                                    }
                                } catch(Exception e) {
                                    logger.warn("Error while removing aggregation" + e);
                                }
                                try {
                                    entryResource.remove();
                                } catch (Exception e) {
                                    logger.warn("Error while invoking resource remove method: " + e);
                                }
                                i.remove();
                                sessionKeys.remove(entryResourceKey);
                                totalRemoved++;
                            } else {
                                if (logger.isDebugEnabled()) {
                                    logger.debug("Keeping entry");
                                }
                                totalKept++;                                    
                            }
                        }
                    }
                }
                if (totalRemoved > 0) {
                    setEntryCacheDirty(true);
                    notifyChange();
                }
                if (logger.isDebugEnabled()) {
                        logger.debug("Finished sweep, with " + 
                                      totalKept+" entries kept, " +
                                      totalRemoved+" entries removed, out of " +
                                      totalSwept+" entries.");
                }
            }
        }
    }      
}

