/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/

/*
 * AggregatorSource.java
 */

package org.globus.mds.aggregator.impl;

/** 
    The AggregatorSource interface handles session management for establishing 
    data aggregation relationships between services using source-sink paradigms. 
    The exact data generation and consumption methods are abstracted in order to 
    provide an extensible framework. For example, the underlying session may 
    represented by an asynchronous notification channel (e.g. subscription), a 
    synchronous polling mechanism (e.g. repeated queries) or some other 
    mechanism, such as a command buffer or FIFO, etc.
*/

public interface AggregatorSource
{ 
    
    /** 
        Called to initialize any necessary state. This method should also attempt
        to initialize the specified sink and should always throw an exception if 
        the sink is not set. The underlying implementation should throw an 
        AggregatorSinkTypeNotSupportedException if it is not compatible with 
        the supplied AggregationSink.  
        
        The implementation may use the optional parameters Object to read 
        implementation specific data from or just ignore the processing of it
        altogether. 
            
    */
    public void initialize(Object parameters) 
        throws AggregatorSinkTypeNotSupportedException,
               AggregatorException,
               Exception; 

    /** 
        Called to free resources used by the source.
    */
    public void terminate();

    /**
     *  Gets the AggregatorSink instance where data from the underlying 
     *  AggregatorSource will be routed, or NULL if not set. 
     */      
    public AggregatorSink getAggregatorSink();
    
    /**
     *  Sets the AggregatorSink instance where data from the Aggregator 
     *  Source will be routed. 
     */      
    public void setAggregatorSink(AggregatorSink sink, Object sinkParameters);

    /** 
        Establishes an implementation-specific data aggregation session.
        The method of data collection and consumption is defined by the 
        implementation; the contract of the interface is that any data generated
        should be capable of being consumed by a class implementing the 
        AggregationSink interface, which will in turn be bound an AggregationSource.

        @param parameters Implementation-specific parameters of the data to be
        collected.

        @param entry The AggregatorServiceGroupEntryResource instance to 
        associate with this aggregation session.  This instance will be passed 
        unmodified to any connected AggregationSink.
     
        @return A session key, which is a String which must be guaranteed 
        by the underlying AggregatorSource implementation to be a unique 
        token with respect to all other invocations of the addAggregation 
        method with the same parameters and entry instance on the same 
        AggregationSource instance.  A null return value is not permitted, 
        though the empty string "" is. <br />
        An exception MUST be thrown if the aggregation could 
        not be created by the implementation.
    */
    public String addAggregation(Object parameters, 
                                 AggregatorServiceGroupEntryResource entry)
        throws AggregatorParamTypeNotSupportedException,
               AggregatorException,
               Exception;

    /** 
     *  Terminates the data aggregation represented by sessionKey.
     *
     *  @param sessionKey The session key of the aggregation to terminate.
     */
    public void removeAggregation(String sessionKey) 
        throws AggregatorException, Exception;
    
    /** 
     * Retrieve AggregatorServiceGroupEntryResource entry context.
     * 
     * @param sessionKey The session key of the aggregation.
     * @return A valid object, or NULL. The type of object returned is dependent
     * on the underlying implementation.
     */    
    public AggregatorServiceGroupEntryResource getSessionContext(String sessionKey);
    
    
    /**
     *  Get a sessionKey from a given entry context.
     *  
     *  @return A string representing the sessionKey for the given entry. 
     */    
    public String getSessionKey(AggregatorServiceGroupEntryResource entry)
        throws AggregatorException;
    
    /**
     *  Determines if a specified sessionKey is valid (active).
     *  
     *  @return A boolean value indicating whether the specified sessionKey is valid. 
     */    
    public boolean isValidSession(String sessionKey)
        throws AggregatorException;       

}

