/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/

/*
 * PollingAggregatorSource.java
 */

package org.globus.mds.aggregator.impl;

import org.globus.mds.aggregator.types.AggregatorContent;
import org.globus.mds.aggregator.types.AggregatorConfig;
import org.globus.mds.aggregator.types.AggregatorData;

import org.globus.mds.aggregator.types.AggregatorPollType;
import org.globus.mds.aggregator.types.GetResourcePropertyPollType;
import org.globus.mds.aggregator.types.GetMultipleResourcePropertiesPollType;
import org.globus.mds.aggregator.types.QueryResourcePropertiesPollType;

import org.globus.wsrf.Constants;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;

import org.globus.wsrf.impl.timer.TimerManagerImpl;

import org.apache.axis.encoding.AnyContentType;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import commonj.timers.Timer;
import commonj.timers.TimerListener;

import java.util.Iterator;
import java.util.Hashtable;

import javax.xml.namespace.QName;

/**
 *  An AggregatorSource which generates data using synchronous polling
 *  in a background thread. The class is abstract, and requires subclassing
 *  to provide the implementation of the poll action.
 */
public abstract class PollingAggregatorSource implements AggregatorSource 
{
    public static final int DEFAULT_POLL_INTERVAL = 60000;
    
    private AggregatorSink sink = null;
    private Object sinkParameters = null;

    /** This Map maps from session keys (String instances) to session context
        (QueryTimerListener instances) */
    private Hashtable sessions = new Hashtable();

    private TimerManagerImpl timerManager = new TimerManagerImpl();
    private boolean initialized = false;
    
    private static Log logger =
            LogFactory.getLog(PollingAggregatorSource.class.getName());
    
    /** Creates a new instance of PollingAggregatorSource */    
    
    public PollingAggregatorSource(AggregatorSink sink, Object sinkParameters) {
        this.setAggregatorSink(sink, sinkParameters);
    }
    
    public PollingAggregatorSource(AggregatorSink sink) {
        this.setAggregatorSink(sink, null);
    }
    
    public void initialize(Object parameters) 
        throws AggregatorException, Exception 
    {
        if (this.initialized) {
            if (logger.isDebugEnabled()) {
                logger.debug("Reinitializing");
            }
            this.terminate();
        }
        
        if (this.getAggregatorSink() == null) {
            throw new AggregatorException("Aggregator Sink not set!");
        } else {
            this.sink.initialize(this.sinkParameters);
        }

        this.initialized = true;        
    }
    
    public AggregatorServiceGroupEntryResource getSessionContext(String sessionKey) 
    {
        QueryTimerListener q = 
            (QueryTimerListener) this.sessions.get(sessionKey);
        return q.getEntry();
    }    
    
    public String getSessionKey(AggregatorServiceGroupEntryResource entry) 
        throws AggregatorException 
    {
        if (entry == null) {
            throw new AggregatorException("Null Argument");
        }
        
        String key = "";
        for (Iterator keys = this.sessions.keySet().iterator(); keys.hasNext();) 
        {
            key = (String) keys.next();
            QueryTimerListener timer = (QueryTimerListener) this.sessions.get(key);
            if (entry.equals(timer.getEntry())) {
                break;
            }
        }
        return key;
    }
    
    public boolean isValidSession(String sessionKey) 
        throws AggregatorException 
    {
        return this.sessions.containsKey(sessionKey);
    }
    
    public AggregatorSink getAggregatorSink() 
    {
        return this.sink;
    }
    
    public void setAggregatorSink(AggregatorSink sink, Object sinkParameters) 
    {
        this.sink = sink;
        if (this.sink != null) {
            this.sink.setSource(this);
            this.sinkParameters = sinkParameters;
        }
    }
    
    public String addAggregation(Object parameters,
                                 AggregatorServiceGroupEntryResource entry) 
        throws AggregatorParamTypeNotSupportedException,
               AggregatorException,
               Exception
    {
        if (!initialized) {
            throw new AggregatorException("Not Initialized");
        }
        
        AggregatorConfig config = (AggregatorConfig) parameters;
        MessageElement[] any = config.get_any();
        
        // not sure this will ever happen, but check just in case
        if (any == null) {
           throw new AggregatorException("Got null any from AggregatorConfig");
        }
        
        AggregatorPollType aggrParams = null;        

        for (int i=0; i<any.length && aggrParams == null; i++) 
        {
            try 
            {
                aggrParams = deserializeParameter(any[i]);
            } catch (Exception e) {
                logger.error("Exception deserializing parameters: ", e);
            }            
        }         
        
        if (aggrParams == null) {
            throw new AggregatorParamTypeNotSupportedException();
        }
        
        QueryTimerListener queryTimerListener = 
            new QueryTimerListener(aggrParams, entry, config);
      
        Timer timer = this.timerManager.schedule(queryTimerListener,
                                   0,
                                   aggrParams.getPollIntervalMillis()); 
       
        queryTimerListener.setTimer(timer);
 
        String sessionKey = queryTimerListener.getSessionKey();
        this.sessions.put(sessionKey, queryTimerListener);
        
        return sessionKey;
    }
   

    public void removeAggregation(String sessionKey) 
        throws AggregatorException, Exception 
    {
        QueryTimerListener queryTimerListener = 
            (QueryTimerListener) this.sessions.remove(sessionKey);
        
        if (queryTimerListener == null) {
            logger.debug(sessionKey+": Timer not found");
            return;
        }
        
        queryTimerListener.getTimer().cancel();
    }
   
    public void terminate() 
    {
        if (!this.initialized) {
            return;
        }       
        
        for (Iterator keys = this.sessions.keySet().iterator(); keys.hasNext();) 
        {
            try {
                this.removeAggregation((String) keys.next());
            } catch (Exception e) {
                logger.error("Exception during removeAggregation: " + e);
            }
        }
        
        this.getAggregatorSink().terminate();
        
        this.initialized = false;
    }
   
    /** One instance of this class will be created for each aggregation */
 
    public class QueryTimerListener implements TimerListener
    {
        private final AggregatorPollType parameters;
        private final AggregatorServiceGroupEntryResource entry;
        private final AggregatorConfig config;
        private Timer timer;

        /** The session. This will be the string QTL followed by the 
            hash code of this object. 'QTL' is prepended in order to
            provide some disambiguation in debugging output */
        public final String sessionKey = "PollAggr-"+this.hashCode();  
 
        public QueryTimerListener(AggregatorPollType parameters,
                          AggregatorServiceGroupEntryResource entry, 
                          AggregatorConfig config) 
        {
            this.parameters = parameters;
            this.entry = entry;            
            this.config = config;
        }
       
        public String getSessionKey()
        {
            return sessionKey;
        }
 
        public void setTimer(Timer t)
        {
            this.timer = t;
        }

        public Timer getTimer() {
            return this.timer;
        }

        public AggregatorPollType getParameters() {
            return this.parameters;
        }

        public AggregatorServiceGroupEntryResource getEntry() {
            return this.entry;
        }
        
        public void timerExpired(Timer timer) 
        {
            AggregatorSink sink = 
                PollingAggregatorSource.this.getAggregatorSink();
        
            // the sink should not be null at this point, but if it
            // is, cancel the event and return.
            if (sink == null) {
                logger.debug("AggregatorSink is null, query event cancelled");
                try {
                    PollingAggregatorSource.this.removeAggregation(sessionKey);
                } catch (Exception e) {
                    logger.warn(e.toString());
                }
                return;
            }    
            
            EndpointReferenceType epr = this.entry.getMemberEPR();
            AnyContentType response = null;
                response = poll(parameters, epr, config);

            if(response==null)
            {
                logger.debug("Poll operation returned null - will not deliver null response to AggregatorSink");
            } else {
                try {
                    sink.deliver(response, this.entry);
                } catch(Exception e) {
                    logger.error("Delivery failure: "+e);
                }
            }
        }
    }


    /** This method must be implemented by subclasses to provide
        the specific specialised poll implementation. It will be 
        called by the PollingAggregatorSource implementation every time
        it is time to poll for data. 

        TODO: this method signature is getting a little messy... tidy up */
    public abstract AnyContentType poll(AggregatorPollType parameters,
        EndpointReferenceType epr, AggregatorConfig config);

    /** If the supplied message element is a supported parameter block,
        the deserialised form should be returned; otherwise null
        should be returned. This method will be invoked on all config
        parameters supplied to an aggregation until one is found 
        that matches. */
    public abstract AggregatorPollType deserializeParameter(MessageElement m)
                throws DeserializationException;

}
