/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/

package org.globus.mds.aggregator.test.junit;

import junit.framework.TestCase;

import java.util.Calendar;
import java.util.Random;
import java.net.URL;
import java.rmi.RemoteException;
import javax.xml.rpc.Stub;
import javax.xml.namespace.QName;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.addressing.ReferencePropertiesType;
import org.globus.axis.util.Util;
import org.apache.axis.utils.XMLUtils;

import com.widgets.WidgetPortType;
import com.widgets.service.WidgetServiceAddressingLocator;
import com.widgets.CreateWidgetResponse;
import com.widgets.CreateWidget;

import org.globus.axis.gsi.GSIConstants;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;

import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.ImmediateResourceTermination;
import org.oasis.wsrf.lifetime.WSResourceLifetimeServiceAddressingLocator;
import org.oasis.wsrf.properties.InsertType;
import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.oasis.wsrf.properties.GetResourceProperty;
import org.oasis.wsrf.properties.SetResourceProperties_Element;
import org.oasis.wsrf.properties.SetResourceProperties_PortType;
import org.oasis.wsrf.properties.WSResourcePropertiesServiceAddressingLocator;

import org.globus.interop.widget.Widget;
import org.oasis.wsrf.servicegroup.EntryType;

import org.globus.mds.aggregator.types.AggregatorContent;
import org.globus.mds.aggregator.types.AggregatorConfig;
import org.globus.mds.aggregator.types.AggregatorData;
import org.globus.mds.aggregator.impl.AggregatorServiceGroupResource;

import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClientCallback;
import org.globus.wsrf.impl.servicegroup.ServiceGroupConstants;


/** Abstract test case for AggregatingServiceGroups.

TODO: this looks like it actually ends up generic enough to be a
      pure servicegroup test client
*/

public abstract class AggregatorTestCase 
    extends TestCase 
    implements ServiceGroupRegistrationClientCallback
{

    static {
        Util.registerTransport();
    }
    
     int numWidgets = 3;

    /** the number of entries we expect the servicegroup to have
        at a particular time. periodically, the code will assert
        that this number matches with what is returned by the
        servicegroup. */
    int numExpectedEntries = 0;

    /** 30 iterations at about 10s per iteration gives about
        5 mins per test, by default */
    int numChangeIterations = 20;

     private final String widgetServiceAddress =
        "https://127.0.0.1:8443/wsrf/services/WidgetService";
     
     private final String widgetServiceAddressNoSecurity =
        "http://127.0.0.1:8080/wsrf/services/WidgetService";
     
     protected boolean isSecure = true;
     
     private WidgetServiceAddressingLocator widgetLocator =
        new WidgetServiceAddressingLocator();

     private WSResourcePropertiesServiceAddressingLocator rpLocator =
        new WSResourcePropertiesServiceAddressingLocator();

     private Random rnd = new Random();
          
     /** construct this here and make protected, so that we can play 
        with its settings in the various test subclasses */
     protected ServiceGroupRegistrationClient regClient = 
        new ServiceGroupRegistrationClient();


    /** Sets the number of widgets that will be created and registered */
     public void setWidgetCount(int numWidgets) {
        if (numWidgets > 0)
            this.numWidgets = numWidgets;
        else throw new RuntimeException("must have positive number of widgets");
     }         

    /** Sets the number of iterations that will be performed to modify
        the widget resource properties. */
     public void setIterationCount(int numIterations) {
        if (numIterations > 0)
            this.numChangeIterations = numIterations;
        else throw new RuntimeException("must have positive number of iterations");
     }         
     
     private int durationMillis = 0;  // 0 means infinite duration for this

     /** set duration of test
         @param millis number of milliseconds that the test will run for */
     public void setTestDuration(int millis) {
        if (millis > 0) {
            this.durationMillis = millis;
        }
     }
     
    /**
     *  Should return an EPR to the Aggregator to be tested. In the case
     *  of factory-based aggregators it should call the appropriate create 
     *  operation or findSingleton in the case of singletons.
     */
    public abstract EndpointReferenceType getAggregator();

    /** Adds appropriate content to the supplied parameter block, which
        will already have other fields, including member and 
        servicegroup EPRs set. */
    
    public abstract void decorateRegistrationParams(
          ServiceGroupRegistrationParameters params) throws Exception;

    public void dumpRegistrationParams(ServiceGroupRegistrationParameters params)
    {
        try 
        {
            System.out.println("Using the following registration parameters:"); 
            System.out.println(
                XMLUtils.ElementToString(
                    ObjectSerializer.toElement(
                        params,
                        params.getTypeDesc().getXmlType())));
        } catch (Exception e) {
            System.err.println("Error serializing ServiceGroupRegistrationParameters");
        }
    }
     

    public void setAnonymous(Stub stub) {
        stub._setProperty(Constants.GSI_ANONYMOUS, Boolean.TRUE);
        stub._setProperty(Constants.AUTHORIZATION,
            NoAuthorization.getInstance());
        stub._setProperty(GSIConstants.GSI_AUTHORIZATION,
            org.globus.gsi.gssapi.auth.NoAuthorization.getInstance());
    }

    public void testAggregatingServiceGroup() throws Exception
    {
        System.out.println("\n\nExecuting test: " + this.getClass().getName());
        if (Boolean.getBoolean("test.container.nosec")) {
            this.isSecure = false;
        }
        System.out.println("Secure Transport enabled: " + Boolean.toString(this.isSecure));
        EndpointReferenceType aggregatorEPR = getAggregator();

        GetResourceProperty serviceGroupRPStub = 
           rpLocator.getGetResourcePropertyPort(aggregatorEPR);

        setAnonymous((Stub)serviceGroupRPStub);

        assertRPexists(serviceGroupRPStub,ServiceGroupConstants.ENTRY);
        assertRPexists(serviceGroupRPStub,AggregatorServiceGroupResource.REGISTRATION_COUNT_RP);
        // TODO Membership content rule should also be asserted to exist


        EndpointReferenceType[] widgetEPRs = 
            new EndpointReferenceType[numWidgets];
        
        System.out.println("Creating "+numWidgets+" widgets");
        for(int i=0; i<numWidgets; i++)
        {
            System.out.println("Creating widget "+i);
            EndpointReferenceType widgetEPR = createWidget();
            System.out.println("epr="+widgetEPR);
            widgetEPRs[i]=widgetEPR;

            ServiceGroupRegistrationParameters params = 
                    new ServiceGroupRegistrationParameters();

            params.setServiceGroupEPR(aggregatorEPR);
            params.setRegistrantEPR(widgetEPR);

            Calendar cal = Calendar.getInstance();
            cal.add(Calendar.HOUR, 24);
            params.setInitialTerminationTime(cal);
            params.setRefreshIntervalSecs(20);
            //params.setSecurityDescriptorFile("some/sec/file.xml");

            decorateRegistrationParams(params);

            dumpRegistrationParams(params);
            
            this.regClient.setClientCallback(this);
            this.regClient.register(params,0);
            
            try {
                synchronized(params) {
                    params.wait();
                }
            } catch (InterruptedException e) {
                System.out.println(e);
            }
        
            numExpectedEntries++;    
            System.out.println("Widget registered.");                       
        }

        System.out.println("Finished create/register phase... starting update phase...");

        // TODO iteration counter settable?
        for(int iterationCounter=0; iterationCounter<numChangeIterations; iterationCounter++)
        {
            System.out.print("Sleeping before widget RP change... ");
            Thread.sleep(5000);
            System.out.println("Slept.");

            int j = rnd.nextInt(numWidgets);
            EndpointReferenceType widgetEPR = widgetEPRs[j];
            String newValue = "Widget number "+j+" set in iteration "
                              +iterationCounter;
            System.out.print("Setting "+newValue+"... ");
            setWidget(widgetEPR, newValue);
            System.out.println("Set."); 

            System.out.print("Sleeping before validity checks... ");
            Thread.sleep(5000);
            System.out.println("Slept.");
           
            assertRPexists(serviceGroupRPStub,ServiceGroupConstants.ENTRY);
            assertRPexists(serviceGroupRPStub,AggregatorServiceGroupResource.REGISTRATION_COUNT_RP);

            assertEntriesOK(serviceGroupRPStub);
        }        
        System.out.println("Finished update phase... starting shutdown phase...");


        // TODO: stop the registration client code


//        destroyAllServiceGroupEntries(serviceGroupRPStub);


        System.out.println("Finished shutdown phase.");


    }

    /** check that we can communicate with each of the entry services
        published by the ServiceGroup, by getting the entry list and then
        trying out a getRP for each one. <br />

        TODO this should have an 
        overridable checkIndivdualEntry(EPR) method to allow subclasses
        to check for things such as presence of lifetime RPs 
    */

    private void assertEntriesOK(GetResourceProperty getRP)
        throws Exception
    {
        // get the entry list from the ServiceGroup
        GetResourcePropertyResponse propResponse = 
            getRP.getResourceProperty(ServiceGroupConstants.ENTRY);

        // iterate over each, checking that we can communicate.
        MessageElement[] entries = propResponse.get_any();

        // the entries list might be empty -- this is ok in itself, as
        // the servicegroup may not have any entries

        // however, in some cases entries seems to be null --
        // that doesn't seem right

        if(entries == null) {
            fail("Entries MessageElement array was null");
        }

        System.out.println("Checking "+entries.length+" individual entries: ");

        if(entries.length != numExpectedEntries) {
            fail("Expecting "+numExpectedEntries+" entries, but found "+entries.length);
        }

        // TODO assert style
        for(int i=0;i<entries.length; i++) {
            System.out.print("Checking entry "+i+": ");

            // extract the entry EPR from the top level entry
            MessageElement any = entries[i];
            EntryType entry = (EntryType) ObjectDeserializer.toObject(any, 
                EntryType.class);

            System.out.print(".");
            EndpointReferenceType entryEPR = entry.getServiceGroupEntryEPR();

            System.out.print(".");
            assertValidAggregatingServiceGroupEntry(entryEPR);
            
           System.out.println(" PASS");
        }
    }


    /** Asserts that the supplied reference refers to an 
        Aggregating ServiceGroupEntry. */
    protected void assertValidAggregatingServiceGroupEntry(
        EndpointReferenceType entryEPR) {

            GetResourceProperty entryRPStub;
            try {
                // get an RP
                entryRPStub = 
                    rpLocator.getGetResourcePropertyPort(entryEPR);

            } catch(Exception e) {
                e.printStackTrace();
                fail("This doesn't seem to be a valid EPR");
                return;
            }
            setAnonymous((Stub)entryRPStub);

            assertRPexists(entryRPStub, ServiceGroupConstants.MEMBER_EPR, 
                          EndpointReferenceType.class);

            assertRPexists(entryRPStub, ServiceGroupConstants.SERVICE_GROUP_EPR, 
                          EndpointReferenceType.class);

            assertRPexists(entryRPStub, ServiceGroupConstants.CONTENT,
                           AggregatorContent.class);

            assertHasLifetimeRPs(entryRPStub);
    }


    /** asserts that the service has WS-Resource Lifetime RPs */
    protected void assertHasLifetimeRPs(GetResourceProperty entryRPStub)
    {

        // current time
        assertRPexists(entryRPStub, WSRFConstants.CURRENT_TIME,
                       Calendar.class);

        // termination time -- can be nillable
        assertRPexists(entryRPStub, WSRFConstants.TERMINATION_TIME,
                       Calendar.class);
    }
   

    /** asserts that the named RP exists and that any values, if they exist,
        can be deserialized to the specified class. 

        @param cl Class to attempt to deserialize as. 
    */
    protected void assertRPexists(GetResourceProperty getRP, QName rpName, 
                               Class cl) 
    {
            GetResourcePropertyResponse propResponse = null;
            try {
                propResponse = getRP.getResourceProperty(rpName);
            } catch(RemoteException e) {
                e.printStackTrace();
                fail("Claimed that non-existent Resource Property "+rpName+" exists.");
            }

            MessageElement [] any = propResponse.get_any();
            try {
                for(int i=0; i<any.length; i++) {
                    ObjectDeserializer.toObject(any[i], cl);
                }
            } catch(DeserializationException de) {
                de.printStackTrace();
                fail("Claimed that values of Resource Property "+rpName
                    +" can be deserialized as instances of "+cl);
            }
    }

    /** asserts that the named RP exists (does not assert that it 
        has value, though) */
    private void assertRPexists(GetResourceProperty getRP, QName rpName) 
    {

            if(getRP==null) {
                fail("null GetResourceProperty stub passed into assertRPexists.");
            }

            if(rpName==null) {
                fail("null RP QName passed into assertRPexists.");
            }

            try {
                GetResourcePropertyResponse propResponse =
                    getRP.getResourceProperty(rpName);
            } catch(RemoteException e) {
                e.printStackTrace();
                fail("Claimed that non-existent Resource Property "+rpName+" exists.");
            }

    }

    /** Creates a widget resource, which will be registered into the
        service group. This method and <code>setWidget</code> may be
        overridden in subclasses if different-looking widgets are 
        needed; for example, in order to test the subscription 
        aggregator source, a widget that can deliver notifications
        is required. */
     protected EndpointReferenceType createWidget()
            throws Exception
     {
            String serviceAddr = 
                (this.isSecure) ? 
                this.widgetServiceAddress : 
                this.widgetServiceAddressNoSecurity;
                    
            URL endpoint = new URL(serviceAddr);
            WidgetPortType port = widgetLocator.getWidgetPortTypePort(endpoint);

            setAnonymous((Stub)port);
            CreateWidgetResponse response =
                port.createWidget(new CreateWidget());

            return response.getEndpointReference();
     }


    /** see documentation for <code>createWidget</code>. */
    protected void setWidget(EndpointReferenceType epr, String newvalue)
            throws Exception 
    {
            SetResourceProperties_PortType port = 
                rpLocator.getSetResourcePropertiesPort(epr);

            setAnonymous((Stub)port);

            InsertType insert = new InsertType();
            MessageElement [] elements = new MessageElement[1];
            elements[0] =
                (MessageElement)ObjectSerializer.toSOAPElement(newvalue, Widget.FOO);
            insert.set_any(elements);

            SetResourceProperties_Element request = new SetResourceProperties_Element();
            request.setInsert(insert);
            port.setResourceProperties(request);
    } 




    /** for each service group entry: destroy it, and assert that it 
        no longer exists */

    private void destroyAllServiceGroupEntries(GetResourceProperty getRP)
        throws Exception
    {
        // get the entry list from the ServiceGroup
        GetResourcePropertyResponse propResponse = 
            getRP.getResourceProperty(ServiceGroupConstants.ENTRY);

        // iterate over each, checking that we can communicate.
        MessageElement[] entries = propResponse.get_any();

        // the entries list might be empty -- this is ok in itself, as
        // the servicegroup may not have any entries

        // however, in some cases entries seems to be null --
        // that doesn't seem right

        if(entries == null) {
            fail("Entries MessageElement array was null");
        }

        System.out.println("Destroying "+entries.length+" individual entries: ");

        for(int i=0;i<entries.length; i++) {
            System.out.print("Checking entry "+i+": ");

            // extract the entry EPR from the top level entry
            MessageElement any = entries[i];
            EntryType entry = (EntryType) ObjectDeserializer.toObject(any, 
                EntryType.class);

            System.out.print(".");
            EndpointReferenceType entryEPR = entry.getServiceGroupEntryEPR();

            System.out.print(".");
            destroyServiceGroupEntry(entryEPR);
                 
            System.out.println(" DESTROYED");
        }
    }


    WSResourceLifetimeServiceAddressingLocator lifetimeAddressingLocator = 
        new WSResourceLifetimeServiceAddressingLocator();


    /** TODO:  make this use scheduled term time in past; is it possible
        to easily detect which lifetime mechanism is supported? */
    protected void destroyServiceGroupEntry(
        EndpointReferenceType entryEPR) {

        ImmediateResourceTermination immediateTermPort;
        try {
        // get a lifetime management stub
            immediateTermPort = 
                lifetimeAddressingLocator.getImmediateResourceTerminationPort(entryEPR);
        } catch(Exception e) {
            e.printStackTrace();
            fail("This doesn't seem to be a valid EPR: " + 
                  entryEPR.getAddress());
            return;
        }

        // call destroy

        try { 
            immediateTermPort.destroy(new Destroy());
        } catch(java.rmi.RemoteException re) {
            re.printStackTrace();
            fail("Exception when calling destroy for EPR: " + 
                 entryEPR.getAddress());
            return;
        }
    }
    
    public boolean setRegistrationStatus(ServiceGroupRegistrationParameters regParams,
                                         boolean wasSuccessful,
                                         boolean wasRenewalAttempt,
                                         Exception exception) 
    {
        try {
            synchronized(regParams) {
                regParams.notify();
            }
        } catch (Exception e) {
            System.out.println(e);
        }
        
        if (wasRenewalAttempt) {
            if (wasSuccessful) {
                return true;
            } else {
                fail("Exception while trying to renew lifetime for entry EPR: " +
                     regParams.getRegistrantEPR().getAddress());
                return false;
            }
        } else {
            if (wasSuccessful) {
                return true;
            } else {
                fail("Exception while trying to register entry EPR: " +
                     regParams.getRegistrantEPR().getAddress() + 
                     " to ServiceGroupEPR: " + 
                     regParams.getServiceGroupEPR().getAddress());
                return false;
            }          
        }
    }
    
}
