/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.wsrf.impl.servicegroup;

import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.Constants;

import org.globus.wsrf.ResourceException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.InvalidResourceKeyException;

import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.ResourceContextImpl;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import java.util.Vector;
import java.util.Iterator;
import java.util.Map;

import javax.naming.Context;
import javax.naming.InitialContext;

import org.oasis.wsrf.servicegroup.EntryType;

/** 
A resource property that can be added onto a resource that is a service
group to provide the appropriate entry resource property.

*/

public class EntryResourceProperty extends ReflectionResourceProperty {

    private static Log logger =
        LogFactory.getLog(EntryResourceProperty.class.getName());

    ServiceGroupResource serviceGroupResource;
    ResourceHome entryHome = null;
    
    /** must pass in the containing servicegroupresource so that
    relevent servicegroup information can be extracted. */
    
    public EntryResourceProperty(ServiceGroupResource sgr) throws Exception {
        // FIXME: pass right meta data info
        super(new SimpleResourcePropertyMetaData(ServiceGroupConstants.ENTRY));

        serviceGroupResource = sgr;

        // ReflectionResourceProperty impl.
        setObject(this);
        setPropertyName("DynamicValue");
        initialize();                

        try {
            entryHome = getEntryHome();
        } catch(Exception e) {
            logger.error("Exception getting entry home", e);
        }
                
    }

    /** This method will generate the array of entries that the
    service group publishes. */
    public Object[] getDynamicValue() 
    {
        Vector results = new Vector();
        Object entryResource;
        ResourcePropertySet rpSet;
        ResourceProperty contentProperty;
        ResourceProperty memberEPRProperty;
        Object contentValue;        
        Object memberEPR;
        
        if (logger.isDebugEnabled()) {
            logger.debug("Generating value of Entry RP.");
        }

        // retrieve a snapshot of the current entry table
        // depending on the implmentation, the thread might block here
        Iterator entryIterator = serviceGroupResource.iterateEntryValues();        
        while(entryIterator.hasNext())
        {
            entryResource = null;
            rpSet = null;
            contentProperty = null;
            memberEPRProperty = null;
            contentValue = null;        
            memberEPR = null; 
            
            if (logger.isDebugEnabled()) {
                logger.debug("Assembling an entry RP ("+results.size()+" so far)");
            }
            
            try 
            {
                Map.Entry mapEntry = (Map.Entry)entryIterator.next();                
                entryResource = mapEntry.getValue();
                if (logger.isDebugEnabled()) {
                    logger.debug("Assembling entry RP to represent "+ entryResource);
                }
               
                if(entryResource!=null) 
                {
                    EntryType entry = new EntryType();

                    rpSet = 
                        ((ResourceProperties)entryResource).getResourcePropertySet();

                    // 1. retrieve content from resource 
                    // there must be exactly one Content element, according
                    // to schema.
                    if (rpSet != null) {
                        contentProperty = 
                            rpSet.get(ServiceGroupConstants.CONTENT);
                    }
                    
                    if (contentProperty == null) {
                        logger.warn("Could not get Content from ServiceGroupEntry -- RP does not exist");
                    } else {
                        try {
                            contentValue = contentProperty.get(0);
                        } catch (Exception e) { }
                        
                        if (contentValue == null) {
                            logger.warn("Could not get Content from ServiceGroupEntry resource -- RP exists but has no value.");
                        } else {
                            entry.setContent(new Object[] {contentValue});
                        }
                    }

                    // 2. retrieve the member EPR from the resource 
                    if (rpSet != null) {
                        memberEPRProperty = 
                            rpSet.get(ServiceGroupConstants.MEMBER_EPR);
                    }

                    if (memberEPRProperty == null) {
                        logger.warn("Could not get Member EPR from ServiceGroupEntry -- RP does not exist");                        
                    } else {
                        try {
                            memberEPR = memberEPRProperty.get(0);
                        } catch (Exception e) { }
                        if (memberEPR instanceof EndpointReferenceType) {
                            entry.setMemberServiceEPR((EndpointReferenceType)memberEPR);
                        } else if (memberEPR == null)  {
                            logger.warn("Could not get Member EPR from ServiceGroupEntry resource -- RP exists but has no value.");                                             
                        } else {
                            logger.warn("Could not get Member EPR from ServiceGroupEntry resource -- RP value is incorrect type.");
                        }
                    }
                    
                    // 3. get and set SGE epr direct from the entryResource
                    entry.setServiceGroupEntryEPR(( (ServiceGroupEntryResource)entryResource).getEntryEPR());

                    results.add(entry); 

                } else {
                    logger.warn("Found a null service group entry value in ServiceGroupResource - skipping");
                }
            }
            catch(Exception e) 
            {
                // we hope we never get here
                logger.warn("Truncated Entry RP due to an unhandled exception during assembly",e);
                break; // just keep what we have so far
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("Total number of Entry RP values is "+results.size());
        }

        return results.toArray();
    }


    public static ResourceHome getEntryHome()
        throws Exception {

        MessageContext ctx = MessageContext.getCurrentContext();

        String homeLocation = getEntryHomeLocation(ctx);
        Context initialContext = new InitialContext();

        return (ResourceHome) initialContext.lookup(homeLocation);
    }

    public static String getEntryHomeLocation(MessageContext ctx)
        throws Exception 
    {
        if (ctx == null) {
            throw new IllegalArgumentException("ctx == null");
        }
        if (!(ctx instanceof MessageContext)) {
            throw new Exception("ctx must be Axis instance");
        }

        org.apache.axis.MessageContext context =
            (org.apache.axis.MessageContext)ctx;

        String servicePath = context.getTargetService();
        return Constants.JNDI_SERVICES_BASE_NAME + servicePath +
               "/entryHome";
    }
}

