/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/

/*
 * ServiceGroupRegistrationClient.java
 *
 * Created on September 13, 2004, 5:52 PM
 */

package org.globus.wsrf.impl.servicegroup.client;

import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;

import org.oasis.wsrf.lifetime.SetTerminationTime;
import org.oasis.wsrf.lifetime.ScheduledResourceTermination;
import org.oasis.wsrf.lifetime.WSResourceLifetimeServiceAddressingLocator;

import org.oasis.wsrf.properties.GetResourcePropertyResponse;
import org.oasis.wsrf.properties.GetResourceProperty;
import org.oasis.wsrf.properties.InvalidResourcePropertyQNameFaultType;
import org.oasis.wsrf.properties.WSResourcePropertiesServiceAddressingLocator;

import org.oasis.wsrf.servicegroup.Add;
import org.oasis.wsrf.servicegroup.EntryType;
import org.oasis.wsrf.servicegroup.ServiceGroupServiceAddressingLocator;
import org.oasis.wsrf.servicegroup.ServiceGroupRegistration;

import org.globus.wsrf.Constants;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.client.BaseClient;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.FaultHelper;
import org.globus.wsrf.utils.XmlUtils;

import org.globus.wsrf.impl.timer.TimerManagerImpl;
import org.globus.wsrf.impl.timer.TimerListenerWrapper;
import commonj.timers.Timer;
import commonj.timers.TimerListener;

import java.io.InputStream;
import java.io.FileInputStream;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Calendar;

import javax.xml.namespace.QName;
import javax.xml.rpc.Stub;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.URI;
import org.apache.axis.utils.XMLUtils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.message.addressing.Address;

import org.globus.wsrf.container.ServiceHost;
import java.net.URL;

/**
This class is a client library to manage registrations to WS-ServiceGroups. The
term 'managed registration' is used to refer to a registration managed by
this client.<br />
Managed registrations come into existence in two ways:<br />
<ul>

  <li> Through a configuration file. A configuration file is
       specified and read by invoking the processConfigFile()
       method.  </li>

  <li> Supplied programmatically, by invoking the 
       setDefaultEPR() and register() methods.</li>

</ul>

The list of managed registrations may be specified through a combination
of the above two methods.<br />

Once a managed registration has been created in the client, the client 
will attempt to ensure that a corresponding ServiceGroupEntry always
exists, as follows:<br />

The client will create a ServiceGroupEntry in the appropriate 
WS-ServiceGroup. It will then periodically attempt WS-ResourceLifetime 
lifetime extension on the ServiceGroupEntry.  If the client detects 
that the ServiceGroupEntry is no longer available, it will create a new 
one.<br/>


When operating inside a container, there is a distinguished instance of this
client that is configured to register to the default MDS index service
running in that container. This is intended to allow easy registration
of interesting (from an MDS perspective) resources into MDS hierarchies.

*/
public class ServiceGroupRegistrationClient 
{

    static final int INITIAL_DELAY_CONTAINER = 30000;
    private int initialDelay = 0 ;

    private static ServiceGroupRegistrationClient containerClient = null;

    private List registrations = Collections.synchronizedList(new ArrayList());    
    private TimerManagerImpl timerManager = new TimerManagerImpl();
    private BaseClient baseClient = null;
    private EndpointReferenceType defaultServiceGroupEPR = null;
    private EndpointReferenceType defaultRegistrantEPR = null;
    private String defaultSecDescFile = null;
    private ServiceGroupRegistrationClientCallback clientCallback = null;
    
    private static Log logger =
        LogFactory.getLog(ServiceGroupRegistrationClient.class.getName());

    static final private ServiceGroupServiceAddressingLocator sgloc =
        new ServiceGroupServiceAddressingLocator();

    static final private WSResourceLifetimeServiceAddressingLocator 
        lifetimeloc = new WSResourceLifetimeServiceAddressingLocator();

    /** When renewing the lifetime of an entry, the refresh frequency will be
        multiplied by this constant to give the duration (from now) of the
        new lifetime. This should be greater than 1 to cope with the fact
        that renewals do not happen exactly on the refresh period. */
    static final private int LIFETIMECONST = 2;
    
    public boolean isDebug = false;
    public boolean outputToConsole = false;
    private boolean outputToLog = logger.isErrorEnabled();

    public static final String TAG_SG_EPR = "defaultServiceGroupEPR";
    public static final String TAG_REG_EPR = "defaultRegistrantEPR";
    public static final String TAG_SEC_DESC = "defaultSecurityDecriptorFile";
    public static final String TAG_PARAMS = "ServiceGroupRegistrationParameters";
    public static final String TAG_NS = "http://mds.globus.org/servicegroup/client";
    

    /** Creates a new instance of ServiceGroupRegistrationClient */
    public ServiceGroupRegistrationClient() {
        this.isDebug = logger.isDebugEnabled();
    }   

    /**
     *  Creates an instance of the client library backed by an instance of 
     *  a BaseClient derived command line class, inheriting its command line
     *  parameters and redirecting output to the console accordingly.
     *                  
     */    
    public ServiceGroupRegistrationClient(BaseClient baseClient) 
    {
        this.baseClient = baseClient;
        this.defaultServiceGroupEPR = baseClient.getEPR();
        if (this.baseClient != null) {
           this.outputToConsole = true;              
           this.isDebug = this.baseClient.isDebugMode();          
        }
    }

    /** Sets the instance of a ServiceGroupRegistrationClientCallback
        that is to receive registration status messages.
        @since GT4.0.0
    */    
    public void setClientCallback(ServiceGroupRegistrationClientCallback callback) {
        this.clientCallback = callback;
    }
    
    /** Gets the instance of the ServiceGroupRegistrationClientCallback
        that is to receive registration status messages.
        @since GT4.0.0
    */    
    public ServiceGroupRegistrationClientCallback getClientCallback() {
        return this.clientCallback;
    }
    
    /** @deprecated use setDefaultServiceGroupEPR instead */ 
    public void setDefaultEPR(EndpointReferenceType defaultEPR) {
        setDefaultServiceGroupEPR(defaultEPR);
    }

    /** Sets the service group EPR for registrations that have a null
        servicegroup field.
        @since GT3.9.4
    */
    public void setDefaultServiceGroupEPR(EndpointReferenceType epr) {
        this.defaultServiceGroupEPR = epr;        
    }

    /** @since GT3.9.5 */
    public EndpointReferenceType getDefaultServiceGroupEPR() {
        return this.defaultServiceGroupEPR;
    }

    /** Sets the default registrant EPR for registrations that have a 
        null registrant field.
        @since GT3.9.4
    */
    public void setRegistrantEPR(EndpointReferenceType epr) {
        this.defaultRegistrantEPR = epr;        
    }

    /** @since GT3.9.5 */
    public EndpointReferenceType getRegistrantEPR() {
        return this.defaultRegistrantEPR;
    }

    /** Sets the default security descriptor filename for registrations
        that have a null security descriptor filename.
        @since GT3.9.4
    */
    public void setDefaultSecDescFile(String f) {
        this.defaultSecDescFile=f;
    }

    public void setInitialDelay(int d) {
        this.initialDelay = d;
    }
   
    /** Parses the config file for registration entries to process.
    The config file entry format is a series of configuration elements:
       <ul>
         <li>zero or one <code>defaultServiceGroupEPR</code> specifying the
             default service group.</li>
         <li>zero or one <code>defaultRegistrantEPR</code> specifying the
             default service group.</li>
         <li>zero or one <code>defaultSecurityDescriptorFile</code> 
             specifying the defauly security descriptor file.</li>
         <li>zero or more <code>ServiceGroupRegistrationParameters</code>.
             Each <code>ServiceGroupRegistrationParameters</code> element
             defines a managed registration. ServiceGroup, registrant and  
             security fields may be left blank, in which case they will
             be populated with defaults specified in the three 
             default* configuration options.</li>
        </ul>
     
      <br />

      TODO: Define this config file format completely in XSD 
            rather than performing the deserialisation partly ourselves.  <br />
      TODO: shorter name for <code>ServiceGroupRegistrationParameters</code>, 
            eg <code>registration</code>
       
      @param fileName the name of the configuration file to process
      @throws Exception if there is an error during processing of the file
                       
    */       
    public void processConfigFile(String fileName) 
        throws Exception
    {
       
        InputStream input = null;
        Document doc = null;
        Element root = null;
        
        try 
        {
            this.status(LOG_D, "Trying to locate config file: " + fileName);
            
            input = new FileInputStream(fileName);           
            doc = XmlUtils.newDocument(input); 
            root = doc.getDocumentElement();
        } 
        catch (Exception e) 
        {
            throw e;
        }
        finally 
        {
            if (input != null) {
                try {
                    input.close();
                } catch (Exception e) {
                    throw e;
                }
            }
        }
        
        NodeList nodes = root.getChildNodes();
        int length = nodes.getLength();
        if (length == 0) {
            throw new Exception ("Configuration File Error: No valid elements to process");
        }
        
        int numEntries = 0;
        for (int i = 0; i < length; i++) 
        {
            Node node = nodes.item(i);
            if (this.TAG_SG_EPR.equalsIgnoreCase(node.getLocalName()))
            {
                EndpointReferenceType epr = (EndpointReferenceType) 
                ObjectDeserializer.toObject((Element)node, 
                                             EndpointReferenceType.class);
                if (epr != null) {
                    this.defaultServiceGroupEPR = epr;
                }                
            } 
            else if (this.TAG_REG_EPR.equalsIgnoreCase(node.getLocalName()))
            {                 
                EndpointReferenceType epr = (EndpointReferenceType) 
                ObjectDeserializer.toObject((Element)node, 
                                             EndpointReferenceType.class);
                if (epr != null) {
                    this.defaultRegistrantEPR = epr;
                }                
            }
            else if (this.TAG_SEC_DESC.equalsIgnoreCase(node.getLocalName()))
            {              
                String file = node.getNodeValue();
                if (file != null) {
                    this.defaultSecDescFile = file;
                }          
            }
            else if (this.TAG_PARAMS.equalsIgnoreCase(node.getLocalName()))
            {         
                /* process an entry */
                ServiceGroupRegistrationParameters params = null;
                try 
                {
                    this.status(LOG_D, "Deserializing Registration entry...");
                    
                    params = (ServiceGroupRegistrationParameters) 
                        ObjectDeserializer.toObject(
                            (Element)node,
                            ServiceGroupRegistrationParameters.class);
                }
                catch (Exception e) 
                {
                    this.status(LOG_E, "Configuration File Error: " + e.toString());
                    if (this.isDebug) {
                        //FaultHelper.printStackTrace(e);
                        this.status(LOG_D, "Deserialization Failed for: " + 
                                    XMLUtils.ElementToString((Element)node));  

                    }                
                    continue;
                }
                numEntries++;
                this.register(params, 0); 
            }
        }
        
        /* for now this is fatal */
        if (numEntries == 0) {
            throw new Exception("Configuration File Error: No valid entries to register");
        }
                
        this.status("Processed " + 
                    Integer.toString(numEntries) + 
                    " registration entries");
    }
    
    /**
     * Create new managed registration using the supplied registration
     * parameters.
     *
     * @param params registration parameters
     * @param delayMillis the delay in milliseconds before executing the 
     *        request.
     */     
    public Timer register(ServiceGroupRegistrationParameters params, 
                         long delayMillis) 
    {
        Timer timer = this.timerManager.schedule(
            new RegTimer(params),
            delayMillis,
            params.getRefreshIntervalSecs()*1000); // check this for overflow possibilities?

        this.registrations.add(timer); 
        return timer;
    }
    
    /**
     * Create new managed registration using the supplied registration
     * parameters. The registration will be made immediately (but
     * asynchronously.
     *
     * @param params registration parameters
     */     
    public Timer register(ServiceGroupRegistrationParameters params)
    {
        return this.register(params,initialDelay);
    }

    /**
     *
     * Use caution when calling the terminate method, as this method 
     * will cause ALL registration timer events to be canceled.  
     * In particular, excercise caution when calling terminate on the 
     * ServiceGroupRegistrationClient instance returned from the 
     * static getContainerClient method.
     *
     */     
    public void terminate() 
    {
        this.reset(); 
        this.status("ServiceGroupRegistrationClient shut down");
    }
            
    /* 
     * Cancels all timer events and removes registrant lists
     */
    private void reset() 
    {
        
        this.status(LOG_D, "Reset registrant list");
        
        int i = 0;
        while(!this.registrations.isEmpty()) {
            Timer timer = (Timer) this.registrations.remove(0);
            
            if (timer == null) {
                this.status(LOG_W, "Null Timer found during reset");
            } else {
                timer.cancel();          
                i++;
            }
        }
        
        this.status(i + " registration event(s) cancelled");
        
    }
    
    private static final int LOG_I = 0;
    private static final int LOG_E = 1;
    private static final int LOG_W = 2;
    private static final int LOG_D = 3;
   
    /** status output at default level of Info */
    private void status(Object obj) {
        this.status(LOG_I, obj);
    }
    
    /** Internal method for handling console or log4j based output.*/
    private void status(int msgType, Object obj) 
    {       
        if (this.outputToConsole) {
            if ((msgType == LOG_E) || (msgType == LOG_W)) {
                System.err.println(obj);
            } else if (msgType == LOG_D) {
                // skip outputting debug messages to console unless
                // in debug mode
                if(this.isDebug) System.out.println(obj);
            }
            else {
                System.out.println(obj);
            }
        } else if (this.outputToLog) {
            switch (msgType)
            {
                case LOG_E:
                    logger.error(obj);
                    break;
                case LOG_W:
                    logger.warn(obj);
                    break;                    
                case LOG_D:
                    logger.debug(obj);
                    break;
                case LOG_I:
                default:
                    logger.info(obj);
                    break;
            }
        }
    }     
    
    /** 
     * Internal class for registration timer event. Each managed 
     * registration that will have a corresponding instance
     * of this class.
     */    
    private class RegTimer implements TimerListener
    {

        /** if we have decided that the remote entry has no resource 
            lifetime management operations/RPs, then we will set this
            to false so that we will not try them again. Care needs to
            be taken to ensure that it is only set to false by 
            exceptions that indicate the appropriate RPs/ops do not exist,
            rather than (for example) network / resource non-existence 
            errors. */
        private boolean maybeHasLifetime = true;

        private ServiceGroupRegistrationParameters parameters = null;

        /** If we know about an existing ServiceGroupEntry EPR for 
            this managed registration, it
            will be stored here so that we can attempt to extend
            its lifetime rather than adding a new entry. */
        private EndpointReferenceType entryEPR = null;
        
 
        public RegTimer(ServiceGroupRegistrationParameters parameters) 
        {
            this.parameters = parameters;        
        }
        
        public ServiceGroupRegistrationParameters getParameters() 
        {
            return this.parameters;
        }
        
        public void cancel(Timer timer)
        {
            try { timer.cancel(); } catch (Exception e){ }
            ServiceGroupRegistrationClient.this.registrations.remove(timer);
        }
       
        /** Sets security as appropriate for this registration */
        public void setSecurity(Stub port) throws Exception {

            /* use sec options from base client */
            if (baseClient != null) 
            {
                baseClient.setOptions((Stub)port);
            }                
            else /* check for sec descriptor files */ 
            {
                /* use sec option from file for the specified reg params */                    
                String secDescFile = null;
                secDescFile = this.parameters.getSecurityDescriptorFile();    
                /* or fall back on the default, if it is even set */ 
                if (secDescFile == null) {
                    secDescFile = defaultSecDescFile;
                }
                if (secDescFile != null) {
                    ((Stub)port)._setProperty(
                        org.globus.wsrf.impl.security.authentication.Constants.CLIENT_DESCRIPTOR_FILE,
                        secDescFile);
                } else { // default to anonymous
                    port._setProperty(org.globus.wsrf.impl.security.authentication.Constants.GSI_ANONYMOUS, Boolean.TRUE);
                    port._setProperty(org.globus.wsrf.impl.security.authentication.Constants.AUTHORIZATION,
                        org.globus.wsrf.impl.security.authorization.NoAuthorization.getInstance());
                    port._setProperty( org.globus.axis.gsi.GSIConstants.GSI_AUTHORIZATION,
                        org.globus.gsi.gssapi.auth.NoAuthorization.getInstance());

                }
            }
        }
 
        public void timerExpired(Timer timer) 
        {
            if (this.parameters == null) {
                String msg = 
                    "Error: Registration event got null parameters (Registration event will be canceled)";
                ServiceGroupRegistrationClient.this.status(LOG_E, msg); 
                this.cancel(timer);
                return;
            }
            
            ServiceGroupRegistrationClientCallback callback = 
                ServiceGroupRegistrationClient.this.getClientCallback();
                
            /*
             * Check the target Service Group EPR.
             * First, check to see if a target ServiceGroupEPR exists in the supplied
             * parameters object.  If not, fall back on the default value which 
             * should either initially be set automatically from the BaseClient 
             * -e or -s arguments (if BaseClient used) OR set on this object 
             * programatically as a global default.
             */
            EndpointReferenceType epr = 
                (parameters.getServiceGroupEPR() != null) ?
                parameters.getServiceGroupEPR() :
                ServiceGroupRegistrationClient.this.defaultServiceGroupEPR;


            // if we have no servicegroup EPR specified, generate the 
            // local index EPR, but do *not* cache
            if (epr == null) 
            {
                try {
                    epr=getIndexEPR();
                } catch(Exception e) {
                    logger.error("When forming local target index EPR: "+e);
                }
            }


            EndpointReferenceType registrantEPR = 
                (parameters.getRegistrantEPR() != null) ?
                parameters.getRegistrantEPR() :
                ServiceGroupRegistrationClient.this.defaultRegistrantEPR;

            // if have no registratant EPR, generate the local index EPR
            // but do *not* cache
            if (registrantEPR == null) 
            {
                try {
                    registrantEPR=getIndexEPR();
                } catch(Exception e) {
                    logger.error("When forming local registrant index EPR: "+e);
                }
            }
                
            try
            {                      
                ServiceGroupRegistration port =
                    sgloc.getServiceGroupRegistrationPort(epr);
                setSecurity((Stub)port);
                if (ServiceGroupRegistrationClient.this.isDebug) {
                    ServiceGroupRegistrationClient.this.status(LOG_D,
                        "Renewing/Adding: " + 
                         ObjectSerializer.toString(
                            parameters, 
                            ServiceGroupRegistrationParameters.
                                getTypeDesc().getXmlType()
                         )
                    );
                }
               
                /* Get current time (preferably from resource, but 
                   local otherwise) */
                Calendar term = getCurrentTimePreferringRemote(entryEPR);

                /* bump it up to the desired lifetime */
                term.add(Calendar.SECOND, 
                         LIFETIMECONST * parameters.getRefreshIntervalSecs());

               
                /* store termTime back into the parameters block */
                parameters.setInitialTerminationTime(term);
              

                /* TODO at this interaction, should also perhaps 
                    perform clock skew for sanity checking that 
                    local and remote clocks are in sync? The WS-Lifetime
                    spec also suggests that clock skew could be 
                    determined and compensated for, which might be
                    desirable here too. */


                /* First attempt to extend the lifetime of existing
                   entry, if we know about one. If we cannot do that,
                   (entry may have disappeared, entry may not exist)
                   add a new entry and cache that. 

                   TODO: if we cannot extend lifetime because there is
                   no setlifetime operation, but the entry resource
                   still exists, then we should not do an add; instead
                   just leave the entry to exist. If it subsequently
                   disappears, we can catch it and add it next time round. */


                boolean successfullyRenewed=false;
                if(entryEPR!=null && maybeHasLifetime) {

                    try {
                        // TODO: append hostname of EPR here?
                        ServiceGroupRegistrationClient.this.status(LOG_D, 
                            "Attempting lifetime extension of entry");                
                    

                        ScheduledResourceTermination lifetimePort = 
                            lifetimeloc.getScheduledResourceTerminationPort(
                            entryEPR);
                        setSecurity((Stub)lifetimePort);
                        SetTerminationTime setTermTimeReq = new
                            SetTerminationTime(); 
                        setTermTimeReq.setRequestedTerminationTime(
                            parameters.getInitialTerminationTime());
                        lifetimePort.setTerminationTime(setTermTimeReq);
                        
                        /* if we get this far without exception, then
                           we have successfully renewed lifetime. */
                        successfullyRenewed = true;
                    
                        if (callback != null) {
                            if (!callback.setRegistrationStatus(parameters,
                                                                true,
                                                                true,
                                                                null)) 
                            {
                                this.cancel(timer);
                                if (ServiceGroupRegistrationClient.this.isDebug) {
                                    ServiceGroupRegistrationClient.this.status(
                                        LOG_D,
                                        "Canceled registration event for: " + 
                                        entryEPR.getAddress());
                                }
                                return;
                            }
                        }                        
                    }
                    catch(Exception e) 
                    {
                        ServiceGroupRegistrationClient.this.status(LOG_W, 
                            "Exception renewing entry lifetime of a registration for "
                            +entryEPR.getAddress() + " - " + e);
                        
                        if (callback != null) {
                            if (!callback.setRegistrationStatus(parameters,
                                                                false,
                                                                true,
                                                                e)) 
                            {
                                this.cancel(timer);
                                if (ServiceGroupRegistrationClient.this.isDebug) {
                                    ServiceGroupRegistrationClient.this.status(
                                        LOG_D,
                                        "Canceled registration event for: " + 
                                        entryEPR.getAddress());
                                }                                
                                return;
                            }
                        }
                       /* so now we will continue into the add code */
                    }
                }

                if (!successfullyRenewed) 
                { 
                    /*  
                     *  Next, if we couldn't renew then attempt an add instead 
                     *
                     *  We don't cancel the timer here on any sort of 
                     *  failure,  as we need to keep trying to make this 
                     *  registration in the future. Note, the only exception to 
                     *  this is if we receive a false return value from 
                     *  a setRegistrationStatus method call to a previously set
                     *  ServiceGroupRegistrationClientCallback object.
                     * 
                     *  (TODO: we should perhaps perform some kind of backoff?)
                     */
                    ServiceGroupRegistrationClient.this.status(LOG_D, 
                        "Attempting add");

                    Add request = new Add();
                    request.setMemberEPR(registrantEPR);
                    request.setContent(parameters.getContent()); 

                    if(parameters.getInitialTerminationTime() == null) {
                        ServiceGroupRegistrationClient.this.status(LOG_E,
                          "No termination time computed for new registation.");
                    }
                    request.setInitialTerminationTime(
                        parameters.getInitialTerminationTime());
                
                    entryEPR = port.add(request);
                    if (ServiceGroupRegistrationClient.this.isDebug) 
                    {
                        ServiceGroupRegistrationClient.this.status(LOG_D, 
                            "Add response: " + 
                            AnyHelper.toSingleString(entryEPR));                
                    }
                    if (callback != null) {
                        if (!callback.setRegistrationStatus(parameters,
                                                            true,
                                                            false,
                                                            null)) 
                        {
                            this.cancel(timer);
                            if (ServiceGroupRegistrationClient.this.isDebug) {
                                ServiceGroupRegistrationClient.this.status(
                                    LOG_D,
                                    "Canceled registration event for: " + 
                                    entryEPR.getAddress());
                            }                                
                            return;
                        }
                    }  
                    if ((ServiceGroupRegistrationClient.this.outputToConsole) ||
                        (ServiceGroupRegistrationClient.this.isDebug))
                    {
                        String msg;

                        msg = "Successfully registered "
                            + registrantEPR.getAddress()
                            + " to servicegroup at "
                            + epr.getAddress();
                        ServiceGroupRegistrationClient.this.status(msg);
                    }
                }
            }
            catch (Exception e) 
            {
                String msg;

                msg = "Warning: Could not register "
                    + registrantEPR.getAddress()
                    + " to servicegroup at "
                    + epr.getAddress()
                    + " -- check the URL and that the remote service is up. "
                    + " Remote exception was "
                    + FaultHelper.getMessage(e);
                ServiceGroupRegistrationClient.this.status(LOG_W, msg);
                
                if (callback != null) {
                    if (!callback.setRegistrationStatus(parameters,
                                                        false,
                                                        false,
                                                        e)) 
                    {
                        this.cancel(timer);
                        if (ServiceGroupRegistrationClient.this.isDebug) {
                            ServiceGroupRegistrationClient.this.status(
                                LOG_D,
                                "Canceled registration event for: " + 
                                entryEPR.getAddress());
                        }                                
                        return;
                    }
                }                       
            }              
        }

        private final WSResourcePropertiesServiceAddressingLocator rpLocator =
            new WSResourcePropertiesServiceAddressingLocator();

        /** Gets the current time, preferably from the remote resource
            specified, and failing that from local system. */
        private Calendar getCurrentTimePreferringRemote(
            EndpointReferenceType epr)  
        {
            ServiceGroupRegistrationClient.this.status(LOG_D,
                "Attempting to get current time, preferring remote.");

            Calendar now = null;

            if(epr!=null && maybeHasLifetime) {

                try {
                GetResourceProperty getRPPort = 
                    rpLocator.getGetResourcePropertyPort(epr);
                setSecurity((Stub)getRPPort);

                GetResourcePropertyResponse resp = 
                    getRPPort.getResourceProperty(WSRFConstants.CURRENT_TIME);
                } catch(InvalidResourcePropertyQNameFaultType e) {
                    /* this fault means that we can talk to the 
                       resource but it does not have current time,
                       which implies that it does not support 
                       WS-ResourceLifetime mechanisms */
                    maybeHasLifetime=false; 
                    /* now fall through */
                } catch(Exception e) {
                    /* some other exception occurred. return null, but
                    don't change maybeHasLifetime */
                    ;
                }
                
            }

            // if the above hasn't produced a time, then pull time from
            // local clock
            if(now == null) {
                ServiceGroupRegistrationClient.this.status(LOG_D,
                    "No remote current time available. Getting from local clock instead.");
                now = Calendar.getInstance();
            }

            return now;
        }
    }   


    /** This should only be called when running within a container */
    public static synchronized ServiceGroupRegistrationClient getContainerClient() {
        if (containerClient != null) {
            return containerClient;
        }
        // otherwise, initialise and return

        containerClient = new ServiceGroupRegistrationClient();
        containerClient.setInitialDelay(INITIAL_DELAY_CONTAINER);

        try {
            // so that we'll generate at runtime
            containerClient.setDefaultServiceGroupEPR(null); 
        } catch(Exception e) {
            logger.error("Exception when setting default index service: "+e);
        }
        
        return containerClient;
    }


    static public EndpointReferenceType getIndexEPR() 
        throws java.net.MalformedURLException, 
               java.io.IOException, 
               org.apache.axis.types.URI.MalformedURIException {

        EndpointReferenceType indexEPR = new EndpointReferenceType();

        URL baseURL = ServiceHost.getBaseURL();
        URL indexURL = new URL(baseURL,"DefaultIndexService");
        if (logger.isDebugEnabled()) {
            logger.debug("Container registration client will register to local index at "+indexURL);
        }

        indexEPR.setAddress(new Address(indexURL.toString()));

        return indexEPR;
    }


    /** A helper method for reading a params block from a file. This is 
        intended to be used to pull in a skeleton registration block by
        a service prior to populating any run-time defined fields.
        @since GT3.9.5
    */

    static public ServiceGroupRegistrationParameters readParams(String filename)
        throws Exception
    {
        InputStream inputStream = null;
        
        try {
            inputStream = new FileInputStream(filename);
            Document doc = XMLUtils.newDocument(inputStream);

            ServiceGroupRegistrationParameters params =
                (ServiceGroupRegistrationParameters)
                ObjectDeserializer.toObject(doc.getDocumentElement(),
                          ServiceGroupRegistrationParameters.class);
            return params;
        } finally {
            if (inputStream != null) {
                inputStream.close();
            }
        }
    }

    static public Timer register(EndpointReferenceType epr, 
                                String regFile) 
        throws Exception 
    {
        return register(ResourceContext.getResourceContext(), epr, regFile);
    }

    static public Timer register(ResourceContext ctx,
                                EndpointReferenceType epr, 
                                String regFile) 
        throws Exception 
    {
        String regPath = 
            ctx.getProperty(org.apache.axis.Constants.MC_CONFIGPATH) + regFile;

        ServiceGroupRegistrationParameters params =
            ServiceGroupRegistrationClient.readParams(regPath);

        params.setRegistrantEPR(epr);

        ServiceGroupRegistrationClient client =
            ServiceGroupRegistrationClient.getContainerClient();

        return client.register(params);
    }
    
}
