/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.utils.XMLUtils;

import org.globus.mds.glue.GLUECERPType;
import org.globus.mds.glue.ClusterType;
import org.globus.mds.glue.ComputingElementType;
import org.globus.mds.glue.InfoType;
import org.globus.mds.glue.StateType;
import org.globus.mds.glue.PolicyType;
import org.globus.mds.glue.batchprovider.SchedulerType;
import org.globus.mds.glue.batchprovider.QueueType;
import org.globus.mds.glue.batchprovider.ClusterCollectionType;
import org.globus.mds.glue.config.GLUERPConfiguration;

import org.globus.mds.usefulrp.glue.GLUEElementProducer;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import org.globus.wsrf.encoding.DeserializationException;

import org.w3c.dom.Element;
import org.w3c.dom.Document;



/** An MDS ResourceProperty that will allow services to easily publish
    queue and host status information in the GLUE schema. <br />

    This Resource property requires cluster and queue status provider scripts
    to generate the actual data based on the batch system and cluster
    monitoring system in place. <br />

    The output of the queue status provider should be an XML document
    with root element a Scheduler element, as defined in the MDS 
    usefulrp schema package. <br />

    The output of the cluster status provider should be an XML document
    with root element a GLUE Cluster element, as defined in the GLUE
    schema XML mapping. <br />

    The provider scripts will be executed approximately every 5 minutes.
    TODO this could become configurable, or even perhaps become
    non-timebased, for example, by receiving push information.

*/

public class GLUEResourceProperty extends ReflectionResourceProperty {

   private static Log logger =
            LogFactory.getLog(GLUEResourceProperty.class.getName());

    // Config options

    /** The name of the cluster provider executable, null if none is 
        available. 

        <br> 
        TODO: needs to be set in config, keyed by the 
        scheduler name, or scheduler name and queue name. A multilevel
        heirarchy is probably the best: default, then default per 
        scheduler, then per scheduler+queuename */

    private String clusterProviderName = "none"; 
    private String baseDir;
    /** config file <br>
        TODO: pull this from JNDI? */
    private String configFile;

    // Cached information

    /** Holds the cached configuration read from
        <code>configFile</code>. */
    private GLUERPConfiguration config = null;

    /** Computing element cache, from the batch provider. */
    private ComputingElementType[] computingElementCache;

    /** Cluster element cache, from the cluster provider. This will be
        populated by a ClusterListener */
    private ClusterType[] clusterCache;

    private String schedulerName = null;

    private PeriodicExecutor clusterExecuter;
    private PeriodicExecutor schedulerExecuter;

    /**  
    @throws java.lang.Exception passes along any exception encountered
        during construction of underlying ReflectionResourceProperty.
    */
    public GLUEResourceProperty(QName name, String scheduler) throws Exception {
        super(new SimpleResourcePropertyMetaData(name));
        setObject(this);
        setPropertyName("DynamicValue");
        initialize();

        this.baseDir = getBaseDirectory();
        this.configFile = this.baseDir + 
            "/etc/globus_wsrf_mds_usefulrp/gluerp.xml";

        loadConfig();

        schedulerName = scheduler;

        startClusterExecutor();
        startSchedulerExecutor();

    }

    private static String getBaseDirectory() {
        return ContainerConfig.getBaseDirectory();
    }
    
    private void startClusterExecutor() {

        clusterExecuter = new PeriodicExecutor(
            clusterProviderName, ClusterCollectionType.class,
            new ClusterListener(), 300);

    }

    private void startSchedulerExecutor() {

        // special case out multijob scheduler, which has
        // no meaningful GLUE representation
        if (schedulerName.equals("multi")) {
            return;
        }

        // form the batch provider name based on the
        // name of the scheduler
        String batchProviderName = "shell " +  this.baseDir +
            "/libexec/globus-scheduler-provider-" +
            schedulerName;

        schedulerExecuter = new PeriodicExecutor(batchProviderName,
            SchedulerType.class, new SchedulerListener(), 300 );

    }

    /** This method assembles cached data into a GLUE RP value. It should
        return fast, with other threads populating the cache. */
    public GLUECERPType getDynamicValue() {
        GLUECERPType glue = new GLUECERPType();
        glue.setCluster(clusterCache);
        glue.setComputingElement(computingElementCache);
        return glue;
    }



    void loadConfig() {
        FileInputStream input = null;
        try { 
            input = new FileInputStream(this.configFile); 
            Document doc = XMLUtils.newDocument(input);
            Element el = doc.getDocumentElement();
            config = (GLUERPConfiguration)
                ObjectDeserializer.toObject(el, 
                                            GLUERPConfiguration.class);

            clusterProviderName = config.getDefaultProvider();
        } catch(Exception e) {
            logger.error("When loading configuration", e);
        } finally {
            if (input != null) {
                try { input.close(); } catch (Exception e) {}
            }
        }
    }

    /** This is the callback interface that PeriodicExecutors will call
        to deliver collected objects. */
    interface Listener {
        public void executionPerformed(Object o);
    }

    /** This class will periodically execute a specified script.
        TODO: Periodic execution will be started automatically by 
        the constructor. */
    class PeriodicExecutor implements Runnable {

        /** Defines the type and name of the script to run.
           The type is determined by the first word of the parameter,
           which make take the following values:
           <ul> 
             <li><code>java</code> - the remainder of the 
                 paramater will be taken as a java class name to
                 instantiate and execute (TODO cache object?)</li>
             <li><code>shell</code> - the remainder of the 
                  parameter will be taken as a system commandline
                  to execute.
           </ul> */
        String scriptName;

        /** the output of the producer will be deserialised to an
           instance of this class. */
        Class cl;

        /** callback to be made to deliver */
        Listener callback;

        /** number of seconds to wait after one execution has 
            finished before starting the subsequent execution. Note
            that this means that the period between the start of
            one execution and the start of the next will be 
            greater than this period. */
        int period;

        Thread thread;

        /** counter to use for numbering in the names of the
            refresher threads */
        private int refresherThreadNumber = 0;

        PeriodicExecutor(String s, Class c, Listener l, int p) { 
            logger.debug("Creating periodic executor for "+s);
            scriptName = s;
            cl = c;
            callback = l;
            period = p;

            thread = new Thread(this, "GLUE refresher "+(refresherThreadNumber++));
            logger.debug("setting GLUE refresher daemon status to true");
            thread.setDaemon(true);
            thread.start();
        }


        public void run() {
            logger.debug("Starting periodic thread for "+scriptName);
            while(true) {
                logger.debug("Running script for "+scriptName);
                runScript();
                logger.debug("Sleeping for "+scriptName);
                try { 
                    Thread.sleep(1000 * period);
                } catch(InterruptedException ie) {
                    logger.warn("Sleep interrupted for "+scriptName, ie);
                }
            }
        }


        /** Executes the configured provider. Tf the execution is
            successful, a deserialised object of class <code>cl</code>
            will be passed to the listener <code>callback</code>.

        */
        private void runScript()
        {
            logger.debug("executing script "+scriptName);

            /** will eventually be populated with the correctly deserialised
                return value. */
            Object retVal;

            // switch here based on the first word of the scriptName

            if(scriptName.startsWith("none")) {
                logger.debug("Processing none by returning null");
                return;
            } else if(scriptName.startsWith("shell ")) {
                logger.debug("Processing shell.");
                Runtime runtime = Runtime.getRuntime();
                Process process;
                InputStream in = null;
                InputStream err = null;
                OutputStream out = null;
                try
                {
                    String execName = scriptName.substring(6);
                    process = runtime.exec(execName);
                    in = process.getInputStream();
                    out = process.getOutputStream();
                    err = process.getErrorStream();
                    Document doc = XMLUtils.newDocument(in);
                    Element el = doc.getDocumentElement();
                    try {
                        retVal = ObjectDeserializer.toObject(el, cl);
                    } catch(DeserializationException e) {
                        logger.error("Could not deserialise output of "+scriptName+" into class "+cl,e);
                        logger.debug("Undeserializable element is "+XMLUtils.ElementToString(el));
                        return;
                    }
                }
                catch (Exception e)
                {
                    logger.warn("Script Execution error when executing "+scriptName,e);
                    return;
                }
                finally 
                {
                    if (in != null)
                    {
                        try { in.close(); } catch (Exception ee) {}
                    }
                    if (out != null)
                    {
                        try { out.close(); } catch (Exception ee) {}
                    }
                    if (err != null)
                    {
                        try { err.close(); } catch (Exception ee) {}
                    }
                }
            } else if(scriptName.startsWith("java ")) {
                logger.debug("Processing java.");
                // pull out class name
                String classname = scriptName.substring(5); 

                // construct object
                Class c;
                try {
                    c = Class.forName(classname);
                } catch(ClassNotFoundException cnfEx) {
                    logger.error("Could not find class "+classname+" to populate GLUE data"); 
                    return;
                }
                GLUEElementProducer pr;
                try {
                    pr = (GLUEElementProducer)c.newInstance();
                } catch(InstantiationException ie) {
                    logger.error("Could not instantiate "+classname,ie);
                    return;
                } catch(IllegalAccessException ie) {
                    logger.error("Could not instantiate "+classname,ie);
                    return;
                } catch(ClassCastException cce) {
                    logger.error("Could not cast "+classname+" to a GLUEElementProducer");
                    return;
                }

                Element el; // to hold the output from the producer
                try {
                    el = pr.getElement();
                } catch(Exception e) {
                    logger.error("When getting element from an instance of "+classname, e);
                    return;
                }
                try {
                    retVal = ObjectDeserializer.toObject(el, cl);
                } catch(DeserializationException dse) {
                    // output an error always with summary, and then
                    // complete dump of XML at debug level
                    logger.error("Could not deserialize output of producer "+classname+" to an instance of "+cl);
                    logger.debug("Undeserializable element is "+XMLUtils.ElementToString(el));
                    return;
                }

            } else {
                logger.error("Attempting to gather GLUE information using unknown or misconfigured parameter: "+scriptName);
                return;
            }

            callback.executionPerformed(retVal);
        }
    }

    /** This Listener performs simple caching of a single Cluster.
        TODO perform recognition of shared filesystems, and move
        those up to subcluster level.
        perhaps perform other host-to-cluster level moves (for 
        example, for clusters very homogenous nodes) */
    class ClusterListener implements Listener {

        /** Updates the cached cluster value.
            @param c should be a ClusterType */
        public void executionPerformed(Object o) {
            logger.debug("Updating cached cluster information.");
            ClusterCollectionType collection = (ClusterCollectionType)o;
            clusterCache = collection.getCluster();
        }
    }


    /** This performs some post-processing on the output of a 
        scheduler provider, to convert from the internal 
        scheduler schema to GLUE schema. */
    class SchedulerListener implements Listener {

        public void executionPerformed(Object o) {
            logger.debug("Updating cached batch system information");
            ComputingElementType[] ces = null;
            try {

                SchedulerType schedulerOutput = (SchedulerType)o;

                if (schedulerOutput == null)
                {
                    logger.warn("Unable to get scheduler information. Scheduler "+
                                "bundle may not be installed correctly.");
                    return;
                }

                if (schedulerOutput.getQueue() == null)
                {
                    logger.warn("Batch provider generated no useful information.");
                    return;
                }

                logger.debug("Converting batch system information to GLUE schema.");

                ces = 
                    new ComputingElementType[schedulerOutput.getQueue().length];

                logger.debug("There are "+schedulerOutput.getQueue().length+" queues.");

                for(int i=0; i<schedulerOutput.getQueue().length;i++)
                {
                    ComputingElementType ce = new ComputingElementType();
                    InfoType info = schedulerOutput.getInfo();
                    QueueType queue = schedulerOutput.getQueue()[i];

                    String queueName = queue.getName();

                    ce.setName(queueName);
                    ce.setUniqueID(queueName);

                    if (info != null)
                    {
                        try {
                            info.setTotalCPUs("" +
                                              queue.getTotalnodes().getIntValue());
                        }
                        catch(Exception e)
                        {
                            logger.debug("Failed to set InfoType " +
                                         "Parameters: " + e);
                        }
                        ce.setInfo(info);
                    }

                    StateType state = new StateType();
                    state.setStatus(queue.getStatus());

/*
  TODO: not sure how to map these fields from the provider schema:
  <maxCount>143</maxCount>
  <maxReqNodes>unlimited</maxReqNodes>
  <maxTotalMemory>unlimited</maxTotalMemory>
  <maxSingleMemory>unlimited</maxSingleMemory>
  <whenActive>0</whenActive>
  <dispatchType>batch</dispatchType>

  Also, no mappings for these state fields?
  private int worstResponseTime;  // attribute
  private int estimatedResponseTime;  // attribute
*/

                    try {
                        state.setFreeCPUs(queue.getFreenodes().getIntValue());
                    } catch(Exception e) {
                        logger.debug("Unable to parse freeNodes value from queue information" +e);
                    }

                    int totalJobs = 0;
                    try {
                        totalJobs = queue.getTotalJobs().getIntValue();                    
                        state.setTotalJobs(totalJobs);
                    } catch(Exception e) {
                        logger.debug("Unable to parse totalJobs value from queue information" +e);
                    }

                    int runningJobs = 0;
                    try {
                        runningJobs = queue.getRunningJobs().getIntValue();                    
                        state.setRunningJobs(runningJobs);
                    } catch(Exception e) {
                        logger.debug("Unable to parse runningJobs value from queue information" +e);
                    }

                    int waitingJobs = (totalJobs - runningJobs);
                    state.setWaitingJobs(waitingJobs);

                    ce.setState(state);

                    PolicyType policy = new PolicyType();

                    try {
                        policy.setMaxWallClockTime(queue.getMaxtime().getIntValue());
                    } catch(Exception e) {
                        logger.debug("Unable to parse maxTime value from queue information" +e);
                        policy.setMaxWallClockTime(-1);
                    }

                    try {
                        policy.setMaxCPUTime(queue.getMaxCPUtime().getIntValue());
                    } catch(Exception e) {
                        logger.debug("Unable to parse maxCPUtime value from queue information" +e);
                        policy.setMaxCPUTime(-1);
                    }

                    try {
                        policy.setMaxRunningJobs(queue.getMaxRunningJobs().getIntValue());
                    } catch(Exception e) {
                        logger.debug("Unable to parse maxRunningJobs value from queue information" +e);                    
                        policy.setMaxRunningJobs(-1);
                    }

                    try {
                        policy.setMaxTotalJobs(queue.getMaxJobsInQueue().getIntValue());
                    } catch(Exception e) {
                        logger.debug("Unable to parse maxJobsInQueue value from queue information" +e);                    
                        policy.setMaxTotalJobs(-1);
                    }
/*
  No mappings for these policy attributes:
  private int priority;  // attribute
*/
                    ce.setPolicy(policy);            

                    ces[i] = ce;
                } 
            } catch(Exception e) {
                logger.warn("Exception refreshing compute element",e);
                ces=null;
            }
            computingElementCache = ces;
        }
    }
}

