/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.utils.AnyHelper;
import org.globus.wsrf.utils.XmlUtils;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import java.io.File;
import java.io.OutputStream;
import java.net.InetAddress;
import java.util.List;
import java.util.ArrayList;

import org.apache.axis.encoding.AnyContentType;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;

import org.w3c.dom.Element;
import org.w3c.dom.Document;



/** An MDS ResourceProperty that will allow services to easily publish
    host status information (limited to JDK system calls only) in the GLUE schema. <br />
*/

public class SimpleGLUEResourceProperty extends ReflectionResourceProperty {

   private static Log logger =
            LogFactory.getLog(SimpleGLUEResourceProperty.class.getName());

    private static String NS = 
        "http://mds.globus.org/glue/ce/1.1";
    
    private static String PREFIX =
        "mds";   

    boolean isDebug;
    
    private Element docRoot;
    private Element osInfo;
    private Element cpuInfo;        
    private Element memInfo;
    private List drives;

    private String errorString;    

    private Document doc;
    /**  
    @throws java.lang.Exception passes along any exception encountered
        during construction of underlying ReflectionResourceProperty.
    */
    public SimpleGLUEResourceProperty(QName name) throws Exception {
        super(new SimpleResourcePropertyMetaData(name));
        setObject(this);
        setPropertyName("DynamicValue");
        initialize();

        this.doc = XmlUtils.newDocument();
        init();        

    }
    

    /** For now process the RP synchronously.  If we encounter problems
        with this, it will get moved to its own thread */
    public Object getDynamicValue() {
        try {
            updateSystemInfo();
            updateFileSystemInfo();
        } catch (Exception e) {
            logger.error("Exception while forming SimpleGLUEResourceProperty: " + e);
        }
        return this.docRoot;
    }


           
    public void setNamespace(String namespacePrefix, 
                             String namespaceURI) 
        throws Exception 
    {
        PREFIX = namespacePrefix;
        NS = namespaceURI;
    }
    
    
    public String toString()
    {
        return XmlUtils.toString(this.doc);
    }
    
        
    protected void updateSystemInfo() throws Exception
    {
        Runtime runtime = Runtime.getRuntime();
        long totalMem = runtime.totalMemory();
        long freeMem = runtime.freeMemory();
        long maxMem = runtime.maxMemory();
        long usedMem = totalMem - freeMem;

	this.memInfo.setAttributeNS(NS, 
                                    PREFIX + ":" + "RAMSize", 
                                    Long.toString(totalMem));

	this.memInfo.setAttributeNS(NS,
                                    PREFIX + ":" + "RAMAvailable", 
                                    Long.toString(freeMem));

	this.memInfo.setAttributeNS(NS,
                                    PREFIX + ":" + "VirtualSize", 
                                    Long.toString(maxMem));

	this.memInfo.setAttributeNS(NS,
                                    PREFIX + ":" + "VirtualAvailable", 
                                    Long.toString(maxMem - totalMem));
       
	this.osInfo.setAttributeNS(NS,
                                   PREFIX + ":" + "Name", 
                                   System.getProperty("os.name"));

	this.osInfo.setAttributeNS(NS,
                                   PREFIX + ":" + "Version",
                                   System.getProperty("os.version"));

	this.osInfo.setAttributeNS(NS,
                                   PREFIX + ":" + "Architecture", 
                                   System.getProperty("os.arch"));
    }
    
    protected void updateFileSystemInfo() throws Exception
    {
        Element fileSystem;
        
        if (drives == null) {
            drives = new ArrayList();
        }

        File fileSystems[] = File.listRoots();
        for(int i=0;i<fileSystems.length;i++)
        {
            if (i<drives.size()) {
                // reuse
                fileSystem = (Element)drives.get(i);
            } else {
                // create new
                fileSystem =
                    this.doc.createElementNS(NS,  PREFIX + ":" + "FileSystem");
                this.docRoot.appendChild(fileSystem);
                drives.add(fileSystem);
            }
            // set the drive name
            fileSystem.setAttributeNS(NS,
                                      PREFIX + ":" + "Name",
                                      fileSystems[i].getAbsolutePath());

        }

        // trim the list
        for (int i=drives.size();i>fileSystems.length;i--) {
            Element el = (Element)drives.remove(i-1);
            el.getParentNode().removeChild(el);
        }
    }
    
    private void init () throws Exception
    {
        // internal objects
        this.isDebug = logger.isDebugEnabled();
        
        // init document structure
        this.docRoot = this.doc.createElementNS(NS,
                                                PREFIX + ":" + "Host");
        this.doc.appendChild(this.docRoot);

        this.osInfo = 
            this.doc.createElementNS(NS, PREFIX + ":" + "OperatingSystem");
        this.docRoot.appendChild(this.osInfo);
        
        this.cpuInfo = 
            this.doc.createElementNS(NS, PREFIX + ":" + "Processor");
        this.docRoot.appendChild(this.cpuInfo);
        
        this.memInfo = 
            this.doc.createElementNS(NS, PREFIX + ":" + "MainMemory");
        this.docRoot.appendChild(this.memInfo);

	this.docRoot.setAttributeNS(NS,
                                    PREFIX + ":" + "Name",
                                    InetAddress.getLocalHost().getHostName());
        
	this.docRoot.setAttributeNS(NS,
                                    PREFIX + ":" + "UniqueID",
                                    InetAddress.getLocalHost().getHostAddress());
        
	this.docRoot.setAttributeNS("http://www.w3.org/2001/XMLSchema-instance",
                                    "xsi:type",
                                    "mds:HostType");
    }            
}

