/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.mds.webmds.xmlSources.resourceProperties;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.mds.webmds.Parameter;
import org.globus.mds.webmds.ParameterMap;
import org.globus.mds.webmds.WebmdsConfigException;
import org.globus.mds.webmds.xmlSources.resourceProperties.ResourcePropertySourceException;
import java.io.IOException;
import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.types.URI.MalformedURIException;
import org.globus.mds.webmds.WebmdsUtil;
import java.io.File;
import java.io.FileInputStream;
import org.xml.sax.InputSource;
import org.apache.axis.message.addressing.ReferencePropertiesType;
import javax.xml.namespace.QName;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.encoding.SerializationException;
import java.io.FileNotFoundException;
import org.globus.wsrf.encoding.DeserializationException;

public class ResourcePropertyConfigHelper {
    static private final String ENDPOINT_PARAM = "endpoint";
    static private final String ENDPOINT_FILE_PARAM = "endpointFile";
    static private final String ENDPOINT_KEY_NAME_PARAM = "endpointKeyName";
    static private final String ENDPOINT_KEY_VALUE_PARAM = "endpointKeyValue";
    static private final String ALLOW_USER_ENDPOINT_PARAM = "allowUserEndpoints";
   
    static EndpointReferenceType getEndpointReference(
	WebmdsUtil webmdsUtil,
	ParameterMap configParams,
	ParameterMap requestParams)
	throws ResourcePropertySourceException,
	WebmdsConfigException,
	IOException {

	EndpointReferenceType endpoint = null;
	boolean allowUserEndpoints = false;

	Parameter param = (Parameter)configParams.get(ALLOW_USER_ENDPOINT_PARAM);
	if (param != null) {
	    allowUserEndpoints =
		Boolean.valueOf(param.getSingleValue()).booleanValue();
	}
	
	// Get endpoint name -- from the user request (if it was specified
	// and configuration allows it) or from the configuration.

	String endpointURL = findNamedParameter(ENDPOINT_PARAM,
						configParams,
						requestParams,
						allowUserEndpoints);

	// Never let the end user specify a file on the server.
	String endpointFileName = findNamedParameter(ENDPOINT_FILE_PARAM,
						     configParams,
						     requestParams,
						     false);

	String endpointKeyName = findNamedParameter(ENDPOINT_KEY_NAME_PARAM,
						    configParams,
						    requestParams,
						    allowUserEndpoints);
	
	String endpointKeyValue = findNamedParameter(ENDPOINT_KEY_VALUE_PARAM,
						     configParams,
						     requestParams,
						     allowUserEndpoints);


	if ((endpointURL) != null && (endpointFileName == null)) {
	    endpoint = getEndpointFromURLString(endpointURL);
	    if (endpointKeyValue != null) {
		if (endpointKeyName == null) {
		    throw new WebmdsConfigException(
			ENDPOINT_KEY_VALUE_PARAM +
			" may not be specified without " +
			ENDPOINT_KEY_NAME_PARAM);
		}
		QName keyName = QName.valueOf(endpointKeyName);
		ReferencePropertiesType props =
		    new ReferencePropertiesType();
		SimpleResourceKey key =
		    new SimpleResourceKey(keyName, endpointKeyValue);
		try {
		    props.add(key.toSOAPElement());
		} catch (SerializationException e) {
		    throw new ResourcePropertySourceException(e);
		}
		endpoint.setProperties(props);
	    } else if (endpointKeyName != null) {
		throw new WebmdsConfigException(
		    ENDPOINT_KEY_NAME_PARAM +
		    " may not be specified without " +
		    ENDPOINT_KEY_VALUE_PARAM);
	    }
	} else if ((endpointURL) == null && (endpointFileName != null)) {
	    endpoint = getEndpointFromFile(webmdsUtil, endpointFileName);
	    if (endpointKeyName != null) {
		throw new WebmdsConfigException(ENDPOINT_FILE_PARAM +
						" and " +
						ENDPOINT_KEY_NAME_PARAM +
						" may not both be specified");
	    }
	    if (endpointKeyValue != null) {
		throw new WebmdsConfigException(ENDPOINT_FILE_PARAM +
						" and " +
						ENDPOINT_KEY_VALUE_PARAM +
						" may not both be specified");
	    }
	} else if ((endpointURL == null) && (endpointFileName == null)) {
	    throw new WebmdsConfigException("No endpoint specified");
	} else {
	    throw new WebmdsConfigException("Only one of " +
					    ENDPOINT_PARAM +
					    " or " +
					    ENDPOINT_FILE_PARAM +
					    " may be specified.");
	}

	return(endpoint);
    }

    public static EndpointReferenceType getEndpointFromURLString(String url)
	throws ResourcePropertySourceException {
	EndpointReferenceType endpoint = new EndpointReferenceType();
	try {
	    endpoint.setAddress(new AttributedURI(url));
	} catch (MalformedURIException e) {
	    throw new ResourcePropertySourceException(e);
	}
	return(endpoint);
    }

    public static EndpointReferenceType getEndpointFromFile(WebmdsUtil util,
							    String fileName)
	throws ResourcePropertySourceException, IOException {
	File file = util.newFile(fileName);
	FileInputStream stream;
	InputSource source;

	try {
	    stream = new FileInputStream(file);
	} catch (FileNotFoundException e) {
	    throw new ResourcePropertySourceException(
		"Error opening resource property file " +
		file.getAbsolutePath(), e);
	}
	source = new InputSource(stream);
	EndpointReferenceType endpoint = null;
	try {
	    endpoint =
		(EndpointReferenceType)ObjectDeserializer.deserialize(
		    source, EndpointReferenceType.class);
	} catch (DeserializationException e) {
	    throw new ResourcePropertySourceException(e);
	}
	stream.close();
	return(endpoint);
    }

    public static String findNamedParameter(String parameterName,
					    ParameterMap configParams,
					    ParameterMap requestParams,
					    boolean allowRequestOverrides)
    throws ResourcePropertySourceException,
	WebmdsConfigException {
	// Get endpoint name -- from the user request (if it was specified
	// and configuration allows it) or from the configuration.

	String result = null;
	if (requestParams != null) {
	    Parameter param = (Parameter)requestParams.get(parameterName);
	    if (param != null) {
		if (allowRequestOverrides) {
		    result = param.getSingleValue();
		} else {
		    throw new ResourcePropertySourceException(
			"Overriding the configured endpoint is not allowed");
		}
	    }
	}

	if (result == null) {
	    Parameter param = (Parameter)configParams.get(parameterName);
	    if (param != null) {
		result = param.getSingleValue();
	    }
	}

	return(result);
    }
}
