/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.mds.webmds.xmlSources.resourceProperties;

import org.globus.mds.webmds.xmlSources.resourceProperties.ResourcePropertyHelper;
import org.globus.mds.webmds.Parameter;
import org.globus.mds.webmds.WebmdsUtil;
import org.globus.mds.webmds.ParameterMap;
import org.globus.mds.webmds.WebmdsXmlSourceException;
import org.globus.mds.webmds.WebmdsConfigException;
import org.globus.mds.webmds.xmlSources.xmlDomNode.WebmdsNodeSource;
import org.globus.mds.webmds.xmlSources.resourceProperties.ResourcePropertySourceException;
import org.apache.axis.message.addressing.EndpointReferenceType;
import java.lang.Boolean;
import javax.xml.namespace.QName;
import org.w3c.dom.Node;
import java.io.IOException;

/**
 * Class to get a resource property and return it as a DOM node.
 * This uses some configuration parameters from the servlet config file:
 * <dl>
 * <dt><b>endpoint</b>
 * (specified as <b>xmlSource.</b><i>sourceName</i>.<b>param.endpoint</b> in
 * the servlet configuration)
 * <dd>The endpoint name to be used in a getResourceProperty request.</dd>
 * <dt><b>endpointKeyName</b> and <b>endpointKeyValue</b> (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpointKeyName</b> and
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpointKeyValue</b> in
 * the servlet configuration).
 * <dd>An optional key-value pair that can be used to supply reference
 * properties to an endpoint specified by an
 * <b>endpoint</b> parameter.
 * <dt><b>allowUserEndpoints</b>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.allowUserEndpoints</b> in
 * the servlet configuration)
 * <dd>If <b>true</b>, values of
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpoint</b>,
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpointKeyName</b>, and
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpointKeyValue</b>
 * specified in the request will override the configured endpoint value.
 * <dt><b>endpointFile</b>
 * (specified as <b>xmlSource.</b><i>sourceName</i>.<b>param.endpointFile</b> in
 * the servlet configuration)
 * <dd>The name of a file containing an XML representation of an endpoint.
 * The <b>endpointFile</b> parameter can never be overridden by request 
 * arguments.
 * <dt><b>rpNamespace</b>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.rpNamespace</b> in
 * the servlet configuration)
 * <dd>The namespace part of the QName of the resource property to be
 * queried for.</dd>
 * <dt><b>rpName</b>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.rpName</b> in
 * the servlet configuration)
 * <dd>The local name part of the QName of the resource property to be
 * queried for.</dd>
 * <dt><b>allowUserEndpoints</b>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.allowUserEndpoints</b> in
 * the servlet configuration)
 * <dd>If <b>true</b>, a value of
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.endpoint</b> 
 * specified in the request will override the configured endpoint value.
 * <dt><b>allowUserResourceProperties</b>
 * (specified as
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.allowUserResourceProperties</b> in
 * the servlet configuration)
 * <dd>If <b>true</b>, a values of
 * <b>xmlSource.</b><i>sourceName</i>.<b>param.rpNamespace</b>
 * and <b>xmlSource.</b><i>sourceName</i>.<b>param.rpName</b> 
 * specified in the request will override the configured resource
 * property namespace and name.
 * </dl>
 *
 */
public class ResourcePropertyNodeSource implements WebmdsNodeSource {
    static private final String RP_NAMESPACE_PARAM = "rpNamespace";
    static private final String RP_NAME_PARAM = "rpName";
    static private final String ALLOW_USER_RP_PARAM = "allowUserResourceProperties";

    private boolean allowUserEndpoints = false;
    private boolean allowUserResourceProperties = false;

    private String requestInfoString = null;
    private WebmdsUtil webmdsUtil = null;

/**
 * Get a resource property and return it as a node.
 * @param configParams - configured parameters for this XML source
 * @param requestParams - request parameters for this XML source
 */
    public Node getNode(ParameterMap configParams, ParameterMap requestParams)
	throws ResourcePropertySourceException,
	WebmdsConfigException,
	IOException {
	String rpNamespace = null;
	String rpName = null;

	EndpointReferenceType endpoint =
	    ResourcePropertyConfigHelper.getEndpointReference(
		this.webmdsUtil, configParams, requestParams);

	Parameter param = (Parameter)configParams.get(ALLOW_USER_RP_PARAM);
	if (param != null) {
	    this.allowUserResourceProperties =
		Boolean.valueOf(param.getSingleValue()).booleanValue();
	}

	// Get RP namespace -- from the user request (if it was specified
	// and configuration allows it) or from the configuration.

	if (requestParams != null) {
	    param = (Parameter)requestParams.get(RP_NAMESPACE_PARAM);
	    if (param != null) {
		if (this.allowUserResourceProperties) {
		    rpNamespace = param.getSingleValue();
		} else {
		    throw new ResourcePropertySourceException("Overriding the configured resource property is not allowed");
		}
	    }
	}

	if (rpNamespace == null) {
	    param = (Parameter)configParams.get(RP_NAMESPACE_PARAM);
	    if (param != null) {
		rpNamespace = param.getSingleValue();
	    }
	}

	// Get RP name -- from the user request (if it was specified
	// and configuration allows it) or from the configuration.	

	if (requestParams != null) {
	    param = (Parameter)requestParams.get(RP_NAME_PARAM);
	    if (param != null) {
		if (this.allowUserResourceProperties) {
		    rpName = param.getSingleValue();
		} else {
		    throw new ResourcePropertySourceException("Overriding the configured resource property is not allowed");
		}
	    }
	}

	if (rpName == null) {
	    param = (Parameter)configParams.get(RP_NAME_PARAM);
	    if (param != null) {
		rpName = param.getSingleValue();
	    }
	}

	if (endpoint == null) {
	    throw new WebmdsConfigException("No endpoint specified");
	}

	if (rpNamespace == null) {
	    throw new WebmdsConfigException("No resource property namespace specified");
	}
	
	if (rpName == null) {
	    throw new WebmdsConfigException("No resource property name specified");
	}
	
	QName rpQName = new QName(rpNamespace, rpName);
	Node node =
	    ResourcePropertyHelper.getResourceProperty(endpoint, rpQName);

	this.requestInfoString = "Endpoint " +
	    endpoint +
	    " resource property " +
	    rpQName.toString();
	return(node);
    }

    /**
     * Returns a string containing the request endpoint and qname
     */
    public String getRequestInfo() {
	return(this.requestInfoString);
    }

    public void init(WebmdsUtil webmdsUtil) {
	this.webmdsUtil = webmdsUtil;
    }

}
