/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Collection;
import java.util.Map;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import commonj.work.WorkManager;
import commonj.work.WorkItem;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.wsrf.utils.PerformanceLog;

import org.globus.replica.rls.RLSClient;
import org.globus.replica.rls.RLSClient.RLI;
import org.globus.replica.rls.RLSClient.LRC;
import org.globus.replica.rls.RLSException;
import org.globus.replica.rls.RLSString2Bulk;
import org.globus.replica.rls.RLSString2;

import org.globus.replica.replicator.Replication;
import org.globus.replica.replicator.ReplicationStatus;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.util.ReplicaCatalogFilter;

/**
 * Implements the <code>Work</code> required to process replication
 * requests.
 */
class DiscoverWork implements commonj.work.Work {

    // STATIC //

    private static RlsRliGetLRCBulk operation = new RlsRliGetLRCBulk();

    private static Log logger =
        LogFactory.getLog(DiscoverWork.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
                DiscoverWork.class.getClassLoader());
    private static PerformanceLog perfLogger =
        new PerformanceLog(DiscoverWork.class.getName());


    // FIELDS //
    int retryMax;
    int retryWait;
    boolean retryBackoff;
    int rliMaxPerQuery;
    int rliQueryTimeout;
    int lrcMaxPerQuery;
    int lrcQueryTimeout;

    ReplicationManagerImpl manager;
    ReplicationImpl replication;

    // CONSTRUCTORS //

    DiscoverWork(ReplicationManagerImpl m, ReplicationImpl r) {
        this.manager = m;
        this.replication = r;

        this.retryMax = this.manager.getLocateRetryMax();
        this.retryWait = this.manager.getLocateRetryWait();
        this.retryBackoff = this.manager.getLocateRetryBackoff();
        this.rliMaxPerQuery = this.manager.getRliMaxPerQuery();
        this.rliQueryTimeout = this.manager.getRliQueryTimeout();
        this.lrcMaxPerQuery = this.manager.getLrcMaxPerQuery();
        this.lrcQueryTimeout = this.manager.getLrcQueryTimeout();
    }


    // METHODS //

    public void release() {
        this.manager = null;
        this.replication = null;
    }

    public boolean isDaemon() {
        return true;
    }

    public void run() {
        perfLogger.start();
        RLSClient rls = null;
        try {
            // Connect to RLI
            rls = RLSUtil.connect(this.replication.indexUrl,
                    this.replication.credentials.getProxyFilename(),
                    this.retryMax, this.retryWait, this.retryBackoff);

            // Create RLI bulk query param
            ArrayList lfns = _createRliBulkQueryParams();

            // A map so that we can sort out the LFNs as query params on a per
            // LRC-basis
            Map lrcParamsMap = _queryIndex(rls, lfns);

            // Filter out the unwanted LRCs. Not all users may want to get 
            // LFN-PFN mappings from all available LRCs.
            Map filteredLrcParamsMap = _filterLrcs(lrcParamsMap);

            // Query replica catalogs
            Map results = _queryCatalogs(filteredLrcParamsMap);

            // Update replication items
            _updateItemSources(results);

            // Update replicaion
            _updateReplication();

            // Schedule transfer work, only if indicated by replication.stage setting
            synchronized (this.replication.statusLock) {
                if (this.replication.stage == ReplicationImpl.TRANSFER &&
                        (this.replication.status & ReplicationImpl.INACTIVE_FLAGS) == 0)
                    _scheduleTransfer();
            }

        } catch (RLSException rlsex) {
            logger.error("Failed to query RLI, exception: " + rlsex.getMessage());
            logger.debug(rlsex.getMessage(),rlsex);
            // General rls failures (i.e., those encountered after initial connect)
            // are considered fatal errors.
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "Failed to query RLI, exception: " + rlsex.getMessage()); // TODO I18n
        } catch (javax.naming.NamingException nex) {
            logger.error("Failed to lookup jndi property, exception:" + nex.getMessage());
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(nex.getMessage(),nex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    nex.getMessage());
        } catch (commonj.work.WorkException workex) {
            logger.error("Failed to schedule work, exception" + workex.getMessage());
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(workex.getMessage(),workex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    workex.getMessage());
        } catch (Exception ex) {
            logger.error("Internal error, exception" + ex.getMessage());
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(ex.getMessage(),ex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    ex.getMessage());
        } finally {
            // Close RLS connection
            if (rls!=null)
                try { RLSUtil.close(rls); } catch (Exception ignored) { }

            // Persist the replication
            _persistReplication();

            // Log performance
            perfLogger.stop(ReplicationConstants.PERFLOG_DISCOVER);

            // Notify status observers
            this.replication.notifyStatusObservers();
        }
    }

    /**
     * INTERNAL USE ONLY.
     * This method creates the rli's bulk query param from the replication's
     * items' uris.
     */
    private ArrayList _createRliBulkQueryParams() {
        logger.trace("_createRliBulkQueryParams");
        // A little indirection here so that in the future I can change this
        // to better support URIs that aren't assumed to be LFNs
        ArrayList lfns = new ArrayList(this.replication.uris()); // TODO we assume Uris to be LFNs
        return lfns;
    }

    /**
     * INTERNAL USE ONLY.
     * This method performs the RLI Bulk Query and sorts the query results
     * into a map keyed on the lrc's url.
     * @param rls The RLS site to query.
     * @param lfns The list of LFNs to use in the query.
     * @return A Map of LRCurl-LFNlists.
     */
    private Map _queryIndex(RLSClient rls, ArrayList lfns) throws RLSException {
        logger.trace("_queryIndex");
        // Map of key:"lrc" and value:"params for lrc query"
        Map lrcParamsMap = new java.util.HashMap(); // TODO initial capacity?

        // Bulk Query the RLI and get results
        logger.trace("Calling rls.getLRCBulk(lfns)");
//        ArrayList s2blist = rli.getLRCBulk(lfns);
        ArrayList s2blist = new ArrayList(0);
        RLSUtil.bulkOperation(rls, DiscoverWork.operation, lfns, s2blist,
                this.rliMaxPerQuery, this.rliQueryTimeout);
        logger.debug("s2blist, size: " + s2blist.size());

        // Sort the results into Bulk Query params (lists of lfns) for
        // the LRC queries
        Iterator iter = s2blist.iterator();
        while (iter.hasNext()) { // TODO perhaps it's more efficient to use array directly
            // Get result and make it easier to read!!!
            RLSString2Bulk str2b = (RLSString2Bulk) iter.next();
            String lfn = str2b.s1;
            String lrcurl = str2b.s2;
            int rc = str2b.rc;
            // NOTE: not every str2b result is unique wrt the lfn
            //       e.g., for 1 lfn, there may be 1+ lrc's associated

            if (rc == RLSClient.RLS_SUCCESS) {
                // IF the lookup for the LFN was a success, THEN
                //     -- add the LFN to the list of LFNs to be used as
                //        a query param for the matching LRC
                ArrayList lrcParams = (ArrayList) lrcParamsMap.get(lrcurl);
                if (lrcParams == null) {
                    // This is the first time we found a result at this
                    // LRC, so create the LRC parameter for it.
                    lrcParams = new ArrayList(); // TODO initial capabity?
                    lrcParamsMap.put(lrcurl, lrcParams);
                }
                // Add the LFN to the list for the LRC query param
                lrcParams.add(lfn);
            }
            else {
                // ELSE the lookup failed, therefore the LFN does not
                // exist, as far as we can tell. Thus, it must be marked
                // as a permenant failure, for this replication item.
                logger.debug("Item failed, lfn: " + lfn);
                ReplicationItemImpl item = this.replication.findItemByUri(lfn); // TODO again we're assuing lfn means uri
                item.changeStatus(
                        ReplicationItem.FAILED,
                        "No sources for: "+item.uri); // TODO i18n
            }
        }

        // Done
        return lrcParamsMap;
    }

    /**
     * INTERNAL USE ONLY.
     * This method filters the LRC Params Map so that the returned map contains
     * only the params for the LRCs that the user wants to query.
     */
    private Map _filterLrcs(Map unfilteredLrcsMap)
            throws javax.naming.NamingException {
        logger.trace("_filterLrcs");
         // Lookup filter
         ReplicaCatalogFilter filter = this.manager.lookupReplicationCatalogFilter();
         if (filter == null) {
             // User has not specified a filter
             return unfilteredLrcsMap;
         }

         // Filter LRC urls collection
         Collection filteredLrcs = filter.filterCatalogUrls(
                 new java.util.LinkedList(unfilteredLrcsMap.keySet()));

         // Now, filter out the LRC params Map
         Map filteredLrcsMap = new java.util.HashMap(filteredLrcs.size());
         Iterator iter = filteredLrcs.iterator();
         while (iter.hasNext()) {
             String lrc = (String) iter.next();
             filteredLrcsMap.put(lrc, unfilteredLrcsMap.get(lrc));
         }

         // Done
         return filteredLrcsMap;
    }

    /**
     * INTERNAL USE ONLY.
     * This method schedules queries for the various catalogs, waits for
     * all queries to conclude, then merges the results.
     */
    private Map _queryCatalogs(Map lrcParamsMap)
            throws commonj.work.WorkException {
        logger.trace("_queryCatalogs");
        WorkManager wm = this.manager.getWorkManager();

        // Create collection for workitems
        Collection lrcWork = new java.util.ArrayList(lrcParamsMap.size());
        Collection lrcWorkItems = new java.util.ArrayList(lrcParamsMap.size());

        // Schedule work for each LRC, keep collection of Work and WorkItems
        logger.trace("Scheduling LRC Query work");
        Iterator iterEntries = lrcParamsMap.entrySet().iterator();
        while (iterEntries.hasNext()) {
            Map.Entry entry = (Map.Entry) iterEntries.next();
            CatalogQueryWork work = new CatalogQueryWork(
                    (String) entry.getKey(), 
                    (ArrayList) entry.getValue(),
                    this.replication.credentials.getProxyFilename(),
                    this.retryMax,
                    this.retryWait,
                    this.retryBackoff,
                    this.lrcMaxPerQuery,
                    this.lrcQueryTimeout);
            lrcWorkItems.add(wm.schedule(work));
            lrcWork.add(work);
        }

        // Wait for ALL work to complete
        logger.trace("Waiting for Catalog Query work");
        wm.waitForAll(lrcWorkItems, WorkManager.INDEFINITE);
        logger.trace("Catalog Query work finished");

        // Merge results from catalog queries
        Map results = new java.util.HashMap();
        Iterator iterWork = lrcWork.iterator();
        while (iterWork.hasNext()) {
            CatalogQueryWork work = (CatalogQueryWork) iterWork.next();
            if (!work.failed) {
                logger.debug("Merging results, lrcurl: " + work.lrcurl
                        + ", size: " + work.queryResults.size());
                _mergeResults(results, work.queryResults);
            }
            else if (work.failed && logger.isDebugEnabled()) {
                logger.debug("Work failed, lrcurl: " + work.lrcurl + ", msg: "
                        + work.errmsg + ", rc: " + work.rc);
            }
        }

        return results;
    }

    /**
     * INTERNAL USE ONLY.
     * This method merges the results of a single catalog query into a merged
     * results map.
     */
    private void _mergeResults(Map mergedResults, ArrayList s2blist) {
        logger.trace("_mergeResults");
        // Merge the results into a master results map of LFN-PFNlist
        Iterator iter = s2blist.iterator();
        while (iter.hasNext()) { 
            RLSString2Bulk str2b = (RLSString2Bulk) iter.next();
            String lfn = str2b.s1;
            String pfn = str2b.s2;
            int rc = str2b.rc;

            if (rc == RLSClient.RLS_SUCCESS) {
                // IF the lookup for the LFN was a success, THEN
                //     -- add the PFN to the LFN's list of PFNs
                List pfns = (List) mergedResults.get(lfn);
                if (pfns == null) {
                    // This is the first time we found a result at this
                    // LFN, so create the PFN list.
                    pfns = new java.util.LinkedList(); // TODO initial capabity?
                    mergedResults.put(lfn, pfns);
                }
                // Add the PFN to the list for the LFN
                pfns.add(pfn);
            }
            // ELSE the lookup failed, therefore the LFN does not
            // exist, at that LRC site. Nothing more to do.
        }
    }

    /**
     * INTERNAL USE ONLY.
     * This method updates the Replication Item's Sources using the results
     * from the catalog queries.
     */
    private void _updateItemSources(Map results) {
        logger.trace("_updateItemSources");
        // Update replication
        Iterator iterItems = this.replication.items().iterator();
        while (iterItems.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iterItems.next();
            List pfns = (List) results.get(item.uri);
            if (pfns==null || pfns.size()<=0) {
                logger.debug("Lookup failed, item uri: " + item.uri);
                item.changeStatus(
                        ReplicationItem.FAILED,
                        "No sources for item: " + item.uri); // TODO I18n
            }
            else {
                // Create array of pfns size
                SourceImpl sources[] = new SourceImpl[pfns.size()];

                // Populate source array
                Iterator iterPfns = pfns.iterator();
                for (int i=0; iterPfns.hasNext(); i++) {
                    String pfn = (String) iterPfns.next();
                    sources[i] = new SourceImpl(pfn);
                }
                item.addSources(sources);
            }
        }
    }

    /**
     * INTERNAL USE ONLY.
     * This method updates the Replication count and status.
     */
    private void _updateReplication() {
        logger.trace("_updateReplication");
        // Update replication count
        Iterator iterItems = this.replication.items().iterator();
        while (iterItems.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iterItems.next();
            if (item.status == ReplicationItem.FAILED)
                this.replication.failed++;
        }
        logger.debug("Replication, total: " + this.replication.total + 
                ", failed: " + this.replication.failed);

        // Check for total failure
        if (this.replication.total == this.replication.failed) {
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "Failed to locate any items for this replication request");
        }
        else {
            // Update stage now that location has copmleted
            this.replication.stage = ReplicationImpl.TRANSFER;
        }

        // Set dirty flag
        this.replication.setDirty();

        // Notify observers
        this.replication.notifyCountObservers();
        this.replication.notifyStatusObservers();
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _persistReplication() {
        logger.trace("_persistReplication");
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _scheduleTransfer()
            throws commonj.work.WorkException {
        logger.trace("_scheduleTransfer");
        TransferWork work = new TransferWork(
                this.manager,
                this.replication);
        WorkManager wm = this.manager.getWorkManager();
        wm.schedule(work, new ReplicationWorkListener(this.replication));
    }
}
