/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.impl;

import java.util.Collection;
import java.util.Map;
import java.util.Iterator;

import commonj.work.WorkManager;
import commonj.work.WorkItem;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;

import org.globus.wsrf.utils.PerformanceLog;

import org.globus.replica.replicator.ReplicationStatus;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.util.DelegatedCredential;
import org.globus.replica.replicator.Destination;
import org.globus.replica.replicator.Source;

import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.TransferStatusType;
import org.globus.rft.generated.TransferStatusTypeEnumeration;
import org.globus.rft.generated.RequestStatusType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.rft.generated.RFTOptionsType;

import org.globus.replica.replicator.util.SourceSelector;

/**
 * Implements the <code>Work</code> required to process replication
 * requests.
 */
class TransferWork implements commonj.work.Work {

    // CONSTANTS //

    private static Log logger =
        LogFactory.getLog(TransferWork.class.getName());
    private static I18n i18n =
        I18n.getI18n(ReplicationConstants.MESSAGE_RESOURCE_BUNDLE,
                TransferWork.class.getClassLoader());
    private static PerformanceLog perfLogger =
        new PerformanceLog(TransferWork.class.getName());


    // FIELDS //

    private ReplicationManagerImpl manager;
    private ReplicationImpl replication;
    private int transferRequestMax;
    private SourceSelector sourceSelector;


    // CONSTRUCTORS //

    TransferWork(ReplicationManagerImpl m, ReplicationImpl r) {
        this.manager = m;
        this.replication = r;
        this.transferRequestMax = this.manager.getTransferRequestMax();
    }


    // METHODS //

    public void release() {
        this.manager = null;
        this.replication = null;
        this.sourceSelector = null;
    }

    public boolean isDaemon() {
        return true;
    }

    public void run() {
        try {
            this.sourceSelector = this.manager.lookupSourceSelector();

            Collection requests = _createTransferRequests();

            Map results = _scheduleTransferRequests(requests);

            int failed = _updateReplicationItems(results);

            _updateReplication(failed);

            synchronized (this.replication.statusLock) {
                if (this.replication.stage == ReplicationImpl.REGISTER &&
                        (this.replication.status & ReplicationImpl.INACTIVE_FLAGS) == 0)
                    _scheduleRegister();
            }
        }
        catch (javax.naming.NamingException nex) {
            logger.error("Failed to lookup jndi property, exception:" + nex.getMessage());
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(nex.getMessage(),nex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    nex.getMessage());
        }
        catch (commonj.work.WorkException workex) {
            logger.error("Failed to schedule registration work.");
            logger.error("Suspending replication, key: " + this.replication.key);
            logger.debug(workex.getMessage(),workex);
            this.replication.changeStatus(
                    ReplicationStatus.SUSPENDED,
                    ReplicationStatus.EXCEPTION,
                    workex.getMessage());
        }
        catch (Exception ex) {
            logger.error("Transfer failed, exception: " + ex.getMessage());
            logger.debug(ex.getMessage(),ex);
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "Failed to transfer, exception: " + ex.getMessage()); // TODO I18n
        }
        finally {
            // Persist the replication
            _persistReplication();

            // Notify status observers
            this.replication.notifyStatusObservers();
        }
    }

    /**
     * INTERNAL USE ONLY.
     */
    private Collection _createTransferRequests() throws Exception {
        logger.trace("_createTransferRequests");
        Collection requests = new java.util.LinkedList();

        Iterator iter = this.replication.items().iterator();
        while (iter.hasNext()) {

            // Create and populate transfers collection
            Collection transfers = new java.util.LinkedList();
            int numTransfers = 0;
            while (numTransfers < this.transferRequestMax && iter.hasNext()) {
                ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
                if (item.getStatus() == ReplicationItem.FAILED)
                    continue;

                // Select source using selector, or default to source at index 0
                String sourceUrl;
                if (this.sourceSelector != null) {
                    // Use selector to narrow collection
                    // NOTE: For now, we only take the first in the selected list
                    Collection available = java.util.Arrays.asList(item.sources);
                    Collection selected = this.sourceSelector.selectSources(available);
                    sourceUrl = ((Source)selected.iterator().next()).getUri();
                }
                else
                    sourceUrl = item.sources[0].uri;

                // Create a request for each possible destination
                for (int j=0; j<item.destinations.length; j++) {
                    TransferType xfer = new TransferType();
                    xfer.setSourceUrl(sourceUrl);
                    xfer.setDestinationUrl(item.destinations[j].uri);
                    transfers.add(xfer);
                    numTransfers++; // Increment the total count of transfers in this request
                }
            }

            // Create transfer request
            if (logger.isDebugEnabled())
                logger.debug("Creating transfer request, size: " + numTransfers);
            TransferRequestType request = new TransferRequestType();
            request.setTransfer((TransferType[])transfers.toArray(
                        new TransferType[transfers.size()]));

            // Initialize transfer options
            RFTOptionsType options = new RFTOptionsType();
            options.setBinary(this.replication.transferOptions.binary);
            options.setBlockSize(this.replication.transferOptions.blockSize);
            options.setTcpBufferSize(this.replication.transferOptions.tcpBufferSize);
            options.setParallelStreams(this.replication.transferOptions.parallelStreams);
            options.setNotpt(this.replication.transferOptions.noTPT);
            options.setDcau(this.replication.transferOptions.dcau);
            options.setSubjectName(this.replication.transferOptions.subjectName);
            options.setDestinationSubjectName(this.replication.transferOptions.destinationSubjectName);
            options.setSourceSubjectName(this.replication.transferOptions.sourceSubjectName);
            options.setUserName(this.replication.transferOptions.username);
            request.setRftOptions(options);

            // Initialize other settings
            request.setTransferCredentialEndpoint(
                ((DelegatedCredential)
                 this.replication.getReplicationCredentials()).getEPR());
            request.setConcurrency(this.replication.transferOptions.transferConcurrency);

            // Add to requests
            requests.add(request);
            transfers.clear();
        }

        return requests;
    }

    /**
     * INTERNAL USE ONLY.
     */
    private Map _scheduleTransferRequests(Collection requests)
            throws Exception {
        logger.trace("_scheduleTransferRequests");
        perfLogger.start();
        WorkManager wm = this.manager.getWorkManager();

        // Create collections for work and work items
        Collection xferWorkItems = new java.util.ArrayList(requests.size());
        Collection xferWork = new java.util.ArrayList(requests.size());
        Map results = new java.util.HashMap(0);

        // Schedule transfer requests
        Iterator iter = requests.iterator();
        while (iter.hasNext()) {
            TransferRequestWork work = new TransferRequestWork(
                    this.replication.reliableTransferUrl,
                    (TransferRequestType) iter.next(),
                    results,
                    (DelegatedCredential)
                        this.replication.getReplicationCredentials(),
                    this.manager.getTransferWaitInt(),
                    this.manager.getTransferWaitMax(),
                    this.manager.getTransferWaitBackoff());
            xferWorkItems.add(wm.schedule(work));
            xferWork.add(work);
        }

        // Wait for transfers
        logger.trace("Waiting for transfer requests");
        wm.waitForAll(xferWorkItems, WorkManager.INDEFINITE);

        // Transfer completed
        logger.trace("Transfers completed");
        iter = xferWork.iterator();
        while (iter.hasNext()) {
            TransferRequestWork work = (TransferRequestWork) iter.next();
            if (work.failed) {
                logger.warn("One or more transfer requests failed, errmsg: " + work.errmsg);
                // TODO could retry the transfer request again, or could check
                //      if the entire replication should be failed
            }
        }

        // Clear data structure and return results
        xferWorkItems.clear();
        xferWork.clear();
        perfLogger.stop(ReplicationConstants.PERFLOG_TRANSFER);
        return results;
    }

    /**
     * INTERNAL USE ONLY.
     */
    private int _updateReplicationItems(Map xferResults) {
        logger.trace("_updateReplicationItems");
        int failedItems = 0;

        // Update replication items
        Iterator iter = this.replication.items().iterator();
        while (iter.hasNext()) {
            ReplicationItemImpl item = (ReplicationItemImpl) iter.next();
            if (item.getStatus() == ReplicationItem.FAILED)
                continue;

            // Check status of related transfers
            int total = item.destinations.length;
            int failed = 0;
            for (int i=0; i<total; i++) {
                String destUri = item.destinations[i].getUri();
                TransferStatusType status =
                    (TransferStatusType) xferResults.get(destUri);
                if (status == null || status.getStatus()
                        != TransferStatusTypeEnumeration.Finished) {
                    item.destinations[i].changeStatus(Destination.FAILED,
                            "Transfer failed"); // TODO add rft message
                    failed++;
                }
                else {
                    item.destinations[i].status = Destination.TRANSFERED;
                    if (logger.isDebugEnabled()) {
                        logger.debug("Transfer finished, dest uri: " + 
                                item.destinations[i].uri);
                    }
                }
            }

            // Record item failure if all destinations fail
            if (total == failed) {
                logger.debug("All transfer destinations failed, item uri: "
                        + item.uri);
                item.changeStatus(ReplicationItem.FAILED,
                        "All transfer destinations failed.");
                failedItems++;
            }

            // Set dirty
            item.setDirty();
        }

        return failedItems;
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _updateReplication(int xfersFailed) {
        logger.trace("_updateReplication, xfersFailed: " + xfersFailed);
        // Update replication count
        this.replication.failed += xfersFailed;
        logger.debug("Replication, total: " + this.replication.total +
                ", failed: " + this.replication.failed);

        // Check for total failure
        if (this.replication.total == this.replication.failed) {
            this.replication.changeStatus(
                    ReplicationStatus.TERMINATED,
                    ReplicationStatus.FAILED,
                    "All items have failed to be located or transfered.");
        }
        else {
            // Update stage now that location has copmleted
            this.replication.stage = ReplicationImpl.REGISTER;
        }

        // Set dirty flag
        this.replication.setDirty();

        // Notify observers
        this.replication.notifyCountObservers();
        this.replication.notifyStatusObservers();
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _persistReplication() {
        logger.trace("_persistReplication");
    }

    /**
     * INTERNAL USE ONLY.
     */
    private void _scheduleRegister()
            throws commonj.work.WorkException {
        logger.trace("_scheduleRegister");
        RegisterWork work = new RegisterWork(
                this.manager,
                this.replication);

        WorkManager wm = this.manager.getWorkManager();
        wm.schedule(work, new ReplicationWorkListener(this.replication));
    }
}
