/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.replica.replicator.service;

import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.rmi.RemoteException;
import javax.naming.InitialContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.I18n;
import org.globus.wsrf.utils.PerformanceLog;

import org.globus.wsrf.RemoveCallback;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.impl.ReflectionResource;
import org.globus.wsrf.security.SecureResource;
import org.globus.wsrf.impl.security.descriptor.GSISecureMsgAuthMethod;
import org.globus.wsrf.impl.security.descriptor.GSITransportAuthMethod;
import org.globus.wsrf.impl.security.descriptor.GSISecureConvAuthMethod;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.delegation.DelegationException;
import org.globus.security.gridmap.GridMap;

// REPLICATION MANAGER //
import org.globus.replica.replicator.ReplicationManager;
import org.globus.replica.replicator.ReplicationFactory;
import org.globus.replica.replicator.ReplicationController;
import org.globus.replica.replicator.Replication;
import org.globus.replica.replicator.ReplicationItem;
import org.globus.replica.replicator.ReplicationStatus;
import org.globus.replica.replicator.ReplicationStatusObserver;
import org.globus.replica.replicator.ReplicationCount;
import org.globus.replica.replicator.ReplicationCountObserver;
import org.globus.replica.replicator.ReplicationOptions;
import org.globus.replica.replicator.TransferOptions;
import org.globus.replica.replicator.ReplicationCredentials;
import org.globus.replica.replicator.Source;
import org.globus.replica.replicator.Destination;
import org.globus.replica.replicator.CreateException;
import org.globus.replica.replicator.ActivateException;
import org.globus.replica.replicator.NotFoundException;

import org.globus.replica.replicator.util.DelegatedCredential;

// GENERATED //
import org.globus.replica.replicator.generated.BaseRequestType;
import org.globus.replica.replicator.generated.TransferOptionsType;
import org.globus.replica.replicator.generated.RequestFileReplicationRequestType;
import org.globus.replica.replicator.generated.ItemSequenceReplicationRequestType;

import org.globus.replica.replicator.generated.ReplicatorResourceProperties;
import org.globus.replica.replicator.generated.InternalErrorFaultType;
import org.globus.replica.replicator.generated.InvalidStateFaultType;
import org.globus.replica.replicator.generated.CreateReplicatorFaultType;

import org.globus.replica.replicator.generated.ItemSequenceReplicationRequestType;
import org.globus.replica.replicator.generated.ReplicationItemRequestType;

import org.globus.replica.replicator.generated.ReplicationStatusEnumerationType;
import org.globus.replica.replicator.generated.ReplicationStageEnumerationType;
import org.globus.replica.replicator.generated.ReplicationResultEnumerationType;
import org.globus.replica.replicator.generated.ReplicationCountType;

import org.globus.replica.replicator.generated.FindItems;
import org.globus.replica.replicator.generated.FindItemsResponse;
import org.globus.replica.replicator.generated.ReplicationItemType;
import org.globus.replica.replicator.generated.SourceType;
import org.globus.replica.replicator.generated.DestinationType;
import org.globus.replica.replicator.generated.ReplicationItemStatusEnumerationType;
import org.globus.replica.replicator.generated.SourceStatusEnumerationType;
import org.globus.replica.replicator.generated.DestinationStatusEnumerationType;



/**
 * The Replicator resource implementation.
 */
public class ReplicatorResource extends ReflectionResource implements
        SecureResource, RemoveCallback, ReplicationStatusObserver,
        ReplicationCountObserver {

    // STATIC //
    private static Log logger = 
        LogFactory.getLog(ReplicatorResource.class.getName());
    private static I18n i18n =
        I18n.getI18n("org.globus.replica.replicator.errors", 
                ReplicatorResource.class.getClassLoader());
    private static PerformanceLog perfLog = 
        new PerformanceLog(ReplicatorResource.class.getName());

    // FIELDS //
    private ReplicationManager manager = null;
    private Replication replication = null;
    private ResourceSecurityDescriptor securityDesc = null;


    // CONSTRUCTORS //

    /**
     * Constructor for the ReplicatorResource object
     */
    public ReplicatorResource() {
        // Empty constructor 
    }


    // METHODS //

    /**
     * Creates and initializes the internal state of the
     * <code>ReplicatorResource</code> and returns the resource ID.
     *
     * @param currentTime The current time.
     * @param termiantionTime The initial termination time.
     * @param request The body of the create request.
     * @param userSubject The caller's user subject.
     * @param localUsername The caller's local username.
     * @return The resource's ID.
     * @throws Exception Indicates that an internal error occurred.
     * @throws DelegationException Indicates that an internal error occurred.
     */
    public Object create(
            Calendar currentTime,
            Calendar terminationTime,
            BaseRequestType request,
            String userSubject,
            String localUsername) throws ResourceException, DelegationException {

        logger.debug("create");

        // Basic validation
        if (currentTime==null || request==null) {
            logger.error("Bad params");         // TODO i18n
            throw new IllegalArgumentException("Bad params");
        }

        // Create delegated credential helper
        DelegatedCredential credential = new DelegatedCredential(
                request.getCredentialEPR(), localUsername, userSubject);

        try {
            InitialContext ctx = new InitialContext();
            this.manager = (ReplicationManager) 
                    ctx.lookup(ReplicatorConstants.REPLICATIONMANAGER);
            ReplicationFactory factory = this.manager.getFactory();

            // Process request body
            List items = null;
            if (request instanceof RequestFileReplicationRequestType)
                items = RequestFileUtil.processRequestFile(factory,
                        (RequestFileReplicationRequestType) request);
            else if (request instanceof ItemSequenceReplicationRequestType)
                items = _convertRequestItems(factory,
                        (ItemSequenceReplicationRequestType) request);
            else
                throw new Exception("Request not supported");

            // Create transfer options
            TransferOptions transferOptions = null;
            if (request.getOptions() != null &&
                    request.getOptions().getTransferOptions() != null) {
                TransferOptionsType opts =
                    request.getOptions().getTransferOptions();
                transferOptions = factory.createTransferOptions(
                        opts.getConcurrency(),
                        opts.getBinary(),
                        opts.getBlockSize(),
                        opts.getTcpBufferSize(),
                        opts.getParallelStreams(),
                        opts.getNoTPT(),
                        opts.getDcau(),
                        opts.getSubjectName(),
                        opts.getDestinationSubjectName(),
                        opts.getSourceSubjectName(),
                        opts.getUsername());
            }

            // Create replication options
            ReplicationOptions options = factory.createReplicationOptions(
                    null, null, null, transferOptions);

            // Create Replication
            this.replication = factory.createReplication( credential, items,
                                                            options);
        }
        catch (Exception e) {
            String msg = "Failed to create Replication: " + e.getMessage();
            logger.warn(msg);
            throw new ResourceException(msg,e);
        }

        // create a resource bean, set initial values
        ReplicatorResourceProperties bean = _createResourceBean(currentTime, 
                terminationTime, this.replication);

        // initialize security descriptor
        try {
            _initSecurityDescriptor(userSubject, localUsername);
        }
        catch (Exception e) {
            String msg = "Failed to initialize security descriptor: " + e.getMessage();
            logger.warn(msg);
            throw new ResourceException(msg,e);
        }

        // initialize super class
        initialize(bean, NamespaceConstants.RESOURCEPROPERTIES,
                this.replication.getKey());
        logger.debug("ID: " + this.getID());

        // Add observers for count and status
        this.replication.getReplicationStatus().addObserver(this);
        this.replication.getReplicationCount().addObserver(this);

        // return the ID
        return this.getID();
    }

    /**
     * Converts the SOAP Stubs replication items into internal forms.
     */
    private List _convertRequestItems(ReplicationFactory factory,
            ItemSequenceReplicationRequestType request) throws CreateException {
        logger.debug("_convertRequestItems");
        ReplicationItemRequestType[] requestItems = request.getItems();
        List items = new java.util.ArrayList(requestItems.length);
        for (int i=0; i<requestItems.length; i++) {
            ReplicationItemRequestType requestItem = requestItems[i];
            ReplicationItem item = factory.createReplicationItem(
                    requestItem.getUri(),
                    requestItem.getPriority(),
                    new String[] {requestItem.getDestinationUri()});
            items.add(item);
        }
        return items;
    }

    /**
     * Creates a resource bean from a replication object.
     *
     * @param currentTime
     * @param terminationTime
     * @param replication
     */
    private ReplicatorResourceProperties _createResourceBean(
            Calendar currentTime,
            Calendar terminationTime,
            Replication replication) {
        ReplicatorResourceProperties rp = new ReplicatorResourceProperties();
        rp.setCurrentTime(currentTime);
        rp.setTerminationTime(terminationTime);
        rp.setStatus(
                _convertStatus(replication.getReplicationStatus().getStatus()));
        rp.setStage(
                _convertStage(replication.getReplicationStatus().getStage()));
        rp.setResult(
                _convertResult(replication.getReplicationStatus().getResult()));
        rp.setErrorMessage(replication.getReplicationStatus().getErrorMsg());
        rp.setCount(_convertCount(replication.getReplicationCount()));
        return rp;
    }

    /**
     * Accept status updates.
     */
    public void statusUpdate( ReplicationStatus status ) {
        if (logger.isDebugEnabled()) {
            logger.debug("statusUpdate");
        }
        // Get bean and update status fields
        ReplicatorResourceProperties bean =
            (ReplicatorResourceProperties) this.getResourceBean();
        bean.setStatus(_convertStatus(status.getStatus()));
        bean.setStage(_convertStage(status.getStage()));
        bean.setResult(_convertResult(status.getResult()));
        bean.setErrorMessage(status.getErrorMsg());
    }

    /**
     * Accept count updates.
     */
    public void countUpdate( ReplicationCount count ) {
        if (logger.isDebugEnabled()) {
            logger.debug("countUpdate");
        }
        // Get bean and update status fields
        ReplicatorResourceProperties bean =
            (ReplicatorResourceProperties) this.getResourceBean();
        bean.setCount(_convertCount(count));
    }

    /**
     * Convert status.
     */
    private ReplicationStatusEnumerationType _convertStatus(int status) {
        switch (status) {
            case ReplicationStatus.PENDING:
                return ReplicationStatusEnumerationType.Pending;
            case ReplicationStatus.ACTIVE:
                return ReplicationStatusEnumerationType.Active;
            case ReplicationStatus.SUSPENDING:
                return ReplicationStatusEnumerationType.Suspending;
            case ReplicationStatus.SUSPENDED:
                return ReplicationStatusEnumerationType.Suspended;
            case ReplicationStatus.TERMINATING:
                return ReplicationStatusEnumerationType.Terminating;
            case ReplicationStatus.TERMINATED:
                return ReplicationStatusEnumerationType.Terminated;
            case ReplicationStatus.DESTROYED:
                return ReplicationStatusEnumerationType.Destroyed;
            default:
                return null;
        }
    }

    /**
     * Convert stage.
     */
    private ReplicationStageEnumerationType _convertStage(int stage) {
        switch (stage) {
            case ReplicationStatus.NONE:
                return ReplicationStageEnumerationType.None;
            case ReplicationStatus.DISCOVER:
                return ReplicationStageEnumerationType.Discover;
            case ReplicationStatus.TRANSFER:
                return ReplicationStageEnumerationType.Transfer;
            case ReplicationStatus.REGISTER:
                return ReplicationStageEnumerationType.Register;
            default:
                return null;
        }
    }

    /**
     * Convert result.
     */
    private ReplicationResultEnumerationType _convertResult(int result) {
        switch (result) {
            case ReplicationStatus.NONE:
                return ReplicationResultEnumerationType.None;
            case ReplicationStatus.FINISHED:
                return ReplicationResultEnumerationType.Finished;
            case ReplicationStatus.FAILED:
                return ReplicationResultEnumerationType.Failed;
            case ReplicationStatus.EXCEPTION:
                return ReplicationResultEnumerationType.Exception;
            default:
                return null;
        }
    }

    /**
     * Convert count.
     */
    private ReplicationCountType _convertCount(ReplicationCount count) {
        ReplicationCountType countType = new ReplicationCountType();
        countType.setTotal(count.getTotal());
        countType.setFinished(count.getFinished());
        countType.setFailed(count.getFailed());
        countType.setTerminated(count.getTerminated());
        return countType;
    }

    /**
     * Initializes the resource's security descriptor.
     */
    private void _initSecurityDescriptor(String userSubject, String localUsername) 
        throws Exception {

        // Instantiate descriptor
        this.securityDesc = new ResourceSecurityDescriptor();

        // Initialize auth methods
        Vector authMethods = new Vector();
        authMethods.add(GSISecureMsgAuthMethod.BOTH);
        authMethods.add(GSITransportAuthMethod.BOTH);
        authMethods.add(GSISecureConvAuthMethod.BOTH);
        this.securityDesc.setAuthMethods(authMethods);

        // Add gridmap entry
        GridMap gridMap = new GridMap();
        gridMap.map(userSubject, localUsername);
        this.securityDesc.setGridMap(gridMap);
    }

    /**
     * Starts the replication.
     *
     * @exception InvalidStateFaultType The resource is in an
     *      invalid state to perform this operation.
     * @exception ResourceException The resource experienced an internal
     *     error.
     */
    public void start() throws InvalidStateFaultType, ResourceException {
        logger.debug("start");
        try {
            ReplicationController controller =
                this.manager.createController(this.replication);
            controller.activate();
        }
        catch (IllegalStateException e) {
            throw new InvalidStateFaultType();
        }
        catch (ActivateException e) {
            throw new ResourceException(e.getMessage(),e);
        }
        catch (Exception e) {
            throw new ResourceException(e.getMessage(),e);
        }
    }

    /**
     * Stop the replication.
     *
     * @exception InvalidStateFaultType The resource is in an
     *      invalid state to perform this operation.
     * @exception ResourceException The resource experienced an internal
     *     error.
     */
    public void stop() throws InvalidStateFaultType, ResourceException {
        logger.debug("stop");
        try {
            ReplicationController controller =
                this.manager.createController(this.replication);
            controller.terminate();
        }
        catch (IllegalStateException e) {
            throw new InvalidStateFaultType();
        }
        catch (Exception e) {
            throw new ResourceException(e.getMessage(),e);
        }
    }

    /**
     * Suspends the replication.
     *
     * @exception InvalidStateFaultType The resource is in an
     *      invalid state to perform this operation.
     * @exception ResourceException The resource experienced an internal
     *     error.
     */
    public void suspend() throws InvalidStateFaultType, ResourceException {
        logger.debug("suspend");
        try {
            ReplicationController controller =
                this.manager.createController(this.replication);
            controller.suspend();
        }
        catch (IllegalStateException e) {
            throw new InvalidStateFaultType();
        }
        catch (Exception e) {
            throw new ResourceException(e.getMessage(),e);
        }
    }

    /**
     * Resumes the replication.
     *
     * @exception InvalidStateFaultType The resource is in an
     *      invalid state to perform this operation.
     * @exception ResourceException The resource experienced an internal
     *     error.
     */
    public void resume() throws InvalidStateFaultType, ResourceException {
        logger.debug("resume");
        try {
            ReplicationController controller =
                this.manager.createController(this.replication);
            controller.resume();
        }
        catch (IllegalStateException e) {
            throw new InvalidStateFaultType();
        }
        catch (ActivateException e) {
            throw new ResourceException(e.getMessage(),e);
        }
        catch (Exception e) {
            throw new ResourceException(e.getMessage(),e);
        }
    }

    /**
     * Find items.
     */
    public FindItemsResponse findItems(FindItems parameters) throws ResourceException {
        logger.debug("findItems");

        // Find items
        int offset = (parameters.getOffset()==null) ? 0 : parameters.getOffset().intValue();
        int limit = (parameters.getLimit()==null) ? 0 : parameters.getLimit().intValue();
        Collection items = null;
        if (parameters.getByUri() != null) {
            try {
                items = this.replication.findItemsByUri(
                                parameters.getByUri(),
                                offset, limit);
            }
            catch(Exception e) {
                throw new ResourceException(e.getMessage(),e);
            }
        }
        else if (parameters.getByStatus() != null) {
            try {
                items = this.replication.findItemsByStatus(
                                _convertItemStatus(parameters.getByStatus()),
                                offset, limit);
            }
            catch(Exception e) {
                throw new ResourceException(e.getMessage(),e);
            }
        }
        else
            throw new ResourceException("Operation not supported yet");

        // Convert types and add to array
        Iterator iter = items.iterator();
        ReplicationItemType[] itemTypes = new ReplicationItemType[items.size()];
        for (int i=0; i<itemTypes.length; i++) {
            itemTypes[i] = _convertItem((ReplicationItem) iter.next());
        }

        // Create response
        FindItemsResponse response = new FindItemsResponse();
        response.setItems(itemTypes);
        return response;
    }

    /**
     * Convert item.
     */
    private ReplicationItemType _convertItem(ReplicationItem item) {
        ReplicationItemType itemType = new ReplicationItemType();

        // Convert item
        itemType.setUri(item.getUri());
        itemType.setPriority(new Integer(item.getPriority()));
        itemType.setStatus(_convertItemStatus(item.getStatus()));
        itemType.setErrorMessage(item.getErrorMsg());

        // Convert sources
        Collection sources = item.sources();
        SourceType[] sourceTypes = new SourceType[sources.size()];
        Iterator iter = sources.iterator();
        for (int i=0; iter.hasNext(); i++) {
            sourceTypes[i] = _convertSource((Source) iter.next());
        }
        itemType.setSources(sourceTypes);

        // Convert destination
        Collection destinations = item.destinations();
        DestinationType[] destinationTypes = new DestinationType[destinations.size()];
        iter = destinations.iterator();
        for (int i=0; iter.hasNext(); i++) {
            destinationTypes[i] = _convertDestination((Destination) iter.next());
        }
        itemType.setDestinations(destinationTypes);

        return itemType;
    }

    /**
     * Converts item status.
     */
    private ReplicationItemStatusEnumerationType _convertItemStatus(int status) {
        switch (status) {
            case ReplicationItem.PENDING:
                return ReplicationItemStatusEnumerationType.Pending;
            case ReplicationItem.FINISHED:
                return ReplicationItemStatusEnumerationType.Finished;
            case ReplicationItem.FAILED:
                return ReplicationItemStatusEnumerationType.Failed;
            case ReplicationItem.TERMINATED:
                return ReplicationItemStatusEnumerationType.Terminated;
            default:
                return null;
        }
    }

    /**
     * Converts item status.
     */
    private int _convertItemStatus(ReplicationItemStatusEnumerationType status) {
        if (status.equals(ReplicationItemStatusEnumerationType.Pending))
            return ReplicationItem.PENDING;
        if (status.equals( ReplicationItemStatusEnumerationType.Finished))
            return ReplicationItem.FINISHED;
        if (status.equals( ReplicationItemStatusEnumerationType.Failed))
            return ReplicationItem.FAILED;
        if (status.equals( ReplicationItemStatusEnumerationType.Terminated))
            return ReplicationItem.TERMINATED;
        return 0;
    }

    /**
     * Converts source.
     */
    private SourceType _convertSource(Source source) {
        SourceType sourceType = new SourceType();
        sourceType.setUri(source.getUri());
        sourceType.setStatus(_convertSourceStatus(source.getStatus()));
        sourceType.setErrorMessage(source.getErrorMsg());
        return sourceType;
    }

    /**
     * Converts source status.
     */
    private SourceStatusEnumerationType _convertSourceStatus(int status) {
        switch (status) {
            case Source.PENDING:
                return SourceStatusEnumerationType.Pending;
            case Source.FAILED:
                return SourceStatusEnumerationType.Failed;
            case Source.SELECTED:
                return SourceStatusEnumerationType.Selected;
            default:
                return null;
        }
    }

    /**
     * Converts destination.
     */
    private DestinationType _convertDestination(Destination destination) {
        DestinationType destinationType = new DestinationType();
        destinationType.setUri(destination.getUri());
        destinationType.setStatus(_convertDestinationStatus(destination.getStatus()));
        destinationType.setErrorMessage(destination.getErrorMsg());
        return destinationType;
    }

    /**
     * Converts destination status.
     */
    private DestinationStatusEnumerationType _convertDestinationStatus(int status) {
        switch (status) {
            case Destination.PENDING:
                return DestinationStatusEnumerationType.Pending;
            case Destination.FAILED:
                return DestinationStatusEnumerationType.Failed;
            case Destination.TRANSFERED:
                return DestinationStatusEnumerationType.Transfered;
            case Destination.REGISTERED:
                return DestinationStatusEnumerationType.Registered;
            default:
                return null;
        }
    }

    /**
     * Gets the securityDescriptor attribute of the Resource.
     *
     * @return The securityDescriptor value
     */
    public ResourceSecurityDescriptor getSecurityDescriptor() {
        //        return this.securityDesc;
        // return null; // TODO: Should initialize this using helper
        return this.securityDesc;
    }

    /**
     * Removes the resources from the container.
     *
     * @exception ResourceException
     */
    public void remove() throws ResourceException {
        logger.debug("remove, id: " + this.getID());
        ReplicationFactory factory = this.manager.getFactory();
        try {
            factory.destroy(this.replication.getKey());
        } catch (NotFoundException e) {
            throw new ResourceException("Replication not found",e);
        }
        this.manager = null;
        this.replication = null;
        this.securityDesc = null;
    }
}
