/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.util.Calendar;
import java.util.Vector;
import java.io.StringReader;
import java.io.Reader;

import javax.security.auth.Subject;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.OverallStatus;
import org.globus.rft.generated.TransferStatusTypeEnumeration;
import org.globus.rft.generated.TransferType;
import org.globus.transfer.reliable.service.database.RFTDatabaseSetup;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;

import org.globus.delegation.DelegationUtil;
import org.globus.delegation.service.DelegationResource;

import org.globus.wsrf.NoSuchResourceException;

import org.globus.wsrf.encoding.ObjectDeserializer;

import org.apache.axis.message.addressing.EndpointReferenceType;

import org.xml.sax.InputSource;

import org.globus.util.I18n;

import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSException;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.jaas.SimplePrincipal;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
/**
 * Purpose : To cache transfers from the database. One per resource. Also should
 * be able to manage number of transfer threads ?
 */
public class RFTResourceManager {
    private int requestId = 0;

    Vector transferSet;

    private int cacheSize = 0;

    private OverallStatus overallStatus = null;

    int concurrency = 0;

    private static Log logger = LogFactory.getLog(RFTResourceManager.class
            .getName());

    private static I18n i18n = I18n.getI18n(
            "org.globus.transfer.reliable.errors", RFTResourceManager.class
                    .getClassLoader());

    private ReliableFileTransferDbAdapter dbAdapter;

    private int maxAttempts = 0;

    private GSSCredential credential = null;

    private EndpointReferenceType delegationEPR = null;

    private Subject userSubject;

    private String delegationListenerId = null;

    private int pending = 0;

    private static final String getNextQuery = 
        "SELECT * FROM transfer WHERE request_id= ? "
        + " AND (status=4 or status=1 or status=8) AND retry_time < ?"
        + " order by id ";

    private static final String updateTransferQuery =
        "UPDATE transfer set status=?, attempts=? where id=?";

    private static final String updateTransferQuery1 =
        "UPDATE transfer set status=? where id=?";

    private int activeTransfers = 0;

    /**
     * @param requestId
     * @param concurrency
     */
    public RFTResourceManager(int requestId, 
            int concurrency)  {
        this.requestId = requestId;
        this.concurrency = concurrency;
        this.cacheSize = 2 * concurrency;
        this.transferSet = new Vector();
        this.dbAdapter = ReliableFileTransferDbAdapter.getDbAdapter();
        this.overallStatus = new OverallStatus();
    }

    /**
     * @param cacheSize
     */
    public void setCacheSize(int cacheSize) {
        this.cacheSize = cacheSize;
    }

    public void setNumberOfTransfers(int numberOfTransfers) {
        if (numberOfTransfers > 0) {
            this.pending += numberOfTransfers;
        }
    }

    public int getMaxAttempts() throws RftDBException {
        if (this.maxAttempts == 0) {
            this.maxAttempts = this.dbAdapter.getMaxAttempts(this.requestId);
        }
        return this.maxAttempts;
    }

    public void setDelegationListenerId(String delegationListenerId) {
        this.delegationListenerId = delegationListenerId;
    }

    private EndpointReferenceType getDelegationEPR() 
        throws Exception {
        Reader delegationEPRStringReader = new StringReader(
                this.dbAdapter.getDelegationEPR(this.requestId));
        String userDN = this.dbAdapter.getSubject(this.requestId);
        this.userSubject = new Subject();
        SimplePrincipal principal = 
            new SimplePrincipal(userDN);
        this.userSubject.getPrincipals().add(principal);
        InputSource in = new InputSource(delegationEPRStringReader);
        this.delegationEPR =
            (EndpointReferenceType) 
            ObjectDeserializer.deserialize(in,
                    EndpointReferenceType.class);
            return this.delegationEPR;
    }

    public synchronized void incrementActive() {
        this.activeTransfers = this.activeTransfers + 1;
    }

    public synchronized void decrementActive() {
        this.activeTransfers = this.activeTransfers -1 ;
    }

    public synchronized boolean checkForConcurrency() {
        if (this.activeTransfers < this.concurrency) {
            return true;
        } else {
            return false;
        }
    }

    public int getActive() {
        return this.activeTransfers;
    }

    public GSSCredential getCredential() 
    throws RftException {
        if (this.credential != null) {
            return this.credential;
        } else {
            try {
                this.delegationEPR = getDelegationEPR();
                DelegationResource resource 
                    = DelegationUtil.getDelegationResource(this.delegationEPR);
                setCredential(resource.getCredential(this.userSubject));
            } catch (Exception e) {
                throw new RftException(i18n.getMessage("delegationCredErr"), 
                e);
            }
            return this.credential;
        }
    }

    public void removeDelegatedCredListeners() {
        try {
            if (logger.isDebugEnabled()) {
                logger.debug("Removing delegation listeners for id: " 
                        + this.requestId);
            }
            if (this.delegationListenerId == null) {
                this.delegationListenerId = 
                    this.dbAdapter.getDelegationListenerId(this.requestId);
            }
            if (this.delegationEPR == null) {
                this.delegationEPR = getDelegationEPR();
            } 
            if (logger.isDebugEnabled()) {
                    logger.debug("Delegation listener to be destroyed:"
                            + this.delegationListenerId);
            }
            DelegationUtil.removeDelegationListener(
                        this.delegationEPR, this.delegationListenerId);
        } catch (NoSuchResourceException nsre) {
            logger.error("Delegation resource already destroyed");
        } catch (Exception e) {
            logger.error("error deleting a deleg resource",e);
        }
    }
    
    public void setCredential(GlobusCredential credential) 
        throws GSSException {
            this.credential = 
                new GlobusGSSCredentialImpl(credential, 
                        GSSCredential.INITIATE_AND_ACCEPT);
    }

    
    /**
     * Gets the next available transfer from the cache
     * 
     * @return <code> TransferType </code>
     * @throws RftDBException
     */
    public synchronized TransferType getNext() throws RftDBException {
        if (!this.checkForConcurrency()) {
            logger.debug("reached concurrency.no more transfers");
            return null;
        }
        if (this.transferSet.size() > 0) {
            this.incrementActive();
            return (TransferType) this.transferSet.remove(0);
        } else {
            getNextSet();
        } 
        if(this.transferSet.size() > 0 ) {
            this.incrementActive();
            return (TransferType) this.transferSet.remove(0);
        }
        return null;
    }

    public int getConcurrency() {
        return this.concurrency;
    }

    public synchronized void updateTransferJob(int transferId, int status, int attempts)
        throws RftDBException {
            Connection c = null;
            PreparedStatement preparedStatement = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                if (attempts !=- 1) {
                    preparedStatement = c.prepareStatement(updateTransferQuery);
                    preparedStatement.setObject(1,new Integer(status));
                    preparedStatement.setObject(2, new Integer(attempts));
                    preparedStatement.setObject(3, new Integer(transferId));
                    if (logger.isDebugEnabled()) {
                        logger.debug("Updating transfer : " + transferId +
                                " with status: " + status + " and attempts:"+
                                attempts);
                    }
                } else {
                    preparedStatement = c.prepareStatement(updateTransferQuery1);
                    preparedStatement.setObject(1,new Integer(status));
                    preparedStatement.setObject(2, new Integer(transferId));
                    if (logger.isDebugEnabled()) {
                        logger.debug("Updating transfer : " + transferId +
                                " with status: " + status );
                    }
                }
                preparedStatement.executeUpdate();
            } catch (SQLException e) {
                logger.error(i18n.getMessage("dbUpdateErr",
                            "" + requestId), e);
                throw new RftDBException(i18n.getMessage("dbUpdateErr", ""
                            + requestId), e);
            } finally {
                try {
                    if (preparedStatement != null) {
                        preparedStatement.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }
            }
        }

    private synchronized void getNextSet() throws RftDBException {
        this.transferSet = new Vector();
        Connection c = null;
        PreparedStatement statement = null;
        ResultSet rs = null;
        TransferType transfer = null;
        // Check number of active transfers from the cache and not from
        // the database
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.prepareStatement(getNextQuery, 
                    ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_UPDATABLE);
            statement.setObject(1, new Integer(this.requestId));
            statement.setObject(2, new Long(
                        Calendar.getInstance().getTimeInMillis()));
            statement.setMaxRows(this.cacheSize);
            rs = statement.executeQuery();
            for (int i = 0;i < this.cacheSize;i++) {
                while ((rs != null) && (rs.next())) {
                    transfer = new TransferType();
                    int tempTransferId = rs.getInt(1);
                    transfer.setTransferId(new Integer(tempTransferId));
                    transfer.setSourceUrl(rs.getString(3));
                    transfer.setDestinationUrl(rs.getString(4));
                    int status = rs.getInt(5);
                    transfer.setAttempts(new Integer(rs.getInt(6)));
                    RFTOptionsType rftOptions = new RFTOptionsType();
                    rftOptions.setUserName(rs.getString(7));
                    rftOptions.setDcau(new Boolean(rs.getBoolean(8)));
                    rftOptions.setParallelStreams(new Integer(rs.getInt(9)));
                    rftOptions.setTcpBufferSize(new Integer(rs.getInt(10)));
                    rftOptions.setBlockSize(new Integer(rs.getInt(11)));
                    rftOptions.setNotpt(new Boolean(rs.getBoolean(12)));
                    boolean binary = rs.getBoolean(13);
                    if (rs.wasNull()) {
                        binary = true;
                    }
                    rftOptions.setBinary(new Boolean(binary));
                    rftOptions.setSourceSubjectName(rs.getString(14));
                    rftOptions.setDestinationSubjectName(rs.getString(15));
                    rftOptions.setIgnoreFilePermErr(new Boolean(rs
                                .getBoolean(16)));
                    transfer.setRftOptions(rftOptions);
                    transfer.setStatus(TransferStatusTypeEnumeration.Pending);
                    this.updateTransferJob(tempTransferId, 
                            RFTConstants.STATUS_CACHED, -1);
                    if (status == 8) {
                        //This file needs to be deleted
                        transfer.setStatus(TransferStatusTypeEnumeration.Dirty);
                    }
                    transferSet.add(transfer);
                }
            }
        } catch (SQLException e) {
            logger.error(i18n.getMessage("dbRetrieveErr",
                        "" + requestId), e);
            throw new RftDBException(i18n.getMessage("dbRetrieveErr", ""
                        + requestId), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
}
