/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import java.rmi.RemoteException;
import java.util.Vector;

import javax.naming.Context;
import javax.naming.InitialContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.RFTDatabaseFaultType;
import org.globus.rft.generated.RepeatedlyStartedFaultType;
import org.globus.rft.generated.Start;
import org.globus.rft.generated.StartOutputType;
import org.globus.rft.generated.GetStatus;
import org.globus.rft.generated.GetStatusResponse;
import org.globus.rft.generated.GetStatusSet;
import org.globus.rft.generated.GetStatusSetResponse;
import org.globus.rft.generated.Cancel;
import org.globus.rft.generated.CancelResponse;
import org.globus.rft.generated.TransferStatusType;
import org.globus.rft.generated.TransferType;
import org.globus.transfer.reliable.service.database.RFTDatabaseOptions;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.util.I18n;
import org.globus.wsrf.Constants;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceHome;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourcePropertySet;
import org.oasis.wsn.ResourceUnknownFaultType;

import commonj.work.WorkManager;

/**
 * Reliable File Transfer service implementation.
 * 
 */
public class ReliableFileTransferImpl {
    private static Log logger = LogFactory
            .getLog(ReliableFileTransferImpl.class.getName());

    private static I18n i18n = I18n.getI18n(
            "org.globus.transfer.reliable.errors",
            ReliableFileTransferImpl.class.getClassLoader());

    private ReliableFileTransferDbAdapter dbAdapter;

    /**
     * Service constructor
     */
    public ReliableFileTransferImpl() {
        try {
            this.dbAdapter = ReliableFileTransferDbAdapter.setup();
        } catch (Exception db) {
            // Catch everything here
             logger.error(i18n.getMessage("dbSetupErr") + db.getMessage());
            if (logger.isDebugEnabled()) {
                logger.debug(i18n.getMessage("dbSetupErr") 
                        + db.getMessage(), db);
            }
        }
    }

    

    /**
     * Starts the transfers for the request
     * 
     * @param params
     * @return _startResponse
     * @throws RemoteException
     */
    public synchronized StartOutputType start(Start params) 
    throws RemoteException,RepeatedlyStartedFaultType {
        int requestId = 0;
        ReliableFileTransferResource rftResource = null;
        try {
            ResourceContext resourceContext = 
                ResourceContext.getResourceContext();
            ResourceHome resourceHome = resourceContext.getResourceHome();
            ResourceKey resourceKey = resourceContext.getResourceKey();
            rftResource = (ReliableFileTransferResource) 
                resourceHome.find(resourceKey);
            requestId = Integer.parseInt((String) resourceKey.getValue());
            if(logger.isDebugEnabled()) {
                logger.debug("PerformanceLog Enter start() "+ requestId);
            }
            if (requestId == 0 || rftResource == null) {
                throw new ResourceUnknownFaultType();
            }
            if (logger.isDebugEnabled()) {
                logger.debug("Start called for : " + requestId);
            }
            if (!rftResource.isStarted()) {
                rftResource.setStartedTrue();
                this.dbAdapter.updateStarted(requestId, true);
            } else {
                 throw  (RepeatedlyStartedFaultType) FaultUtil.makeFault(
                        RepeatedlyStartedFaultType.class, requestId, 
                        i18n.getMessage("startRepeatedlyCalledErr"), null);
            }
            this.dbAdapter.updateStarted(requestId, true);
            startTransfers(rftResource);
            StartOutputType res = new StartOutputType();
            if(logger.isDebugEnabled()) {
                logger.debug("PerformanceLog Exit start() for " + requestId);
            }
            return res;
        } catch (Exception e) {
            if (e instanceof RepeatedlyStartedFaultType) {
                logger.error(i18n.getMessage("startRepeatedlyCalledErr"), e);
                throw  (RepeatedlyStartedFaultType) FaultUtil.makeFault(
                        RepeatedlyStartedFaultType.class, requestId, 
                        i18n.getMessage("startRepeatedlyCalledErr"), null);
            } else {
                logger.error(i18n.getMessage("startErr"), e);
                throw new RemoteException(i18n.getMessage("startErr"), e);
            }
        }
   }


    /**
     * To query the status of a transfer request
     * 
     * @param params
     * @return GetStatusResponse
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     */
    public GetStatusResponse getStatus(GetStatus params)
            throws RemoteException, ResourceUnknownFaultType,
            RFTDatabaseFaultType {
        String sourceUrl = params.getSourceUrl();
        TransferStatusType statusType = null;
        if (logger.isDebugEnabled()) {
            logger.debug("Finding status of " + sourceUrl);
        }
        int requestId = -1;
        try {
            requestId = this.getRequestId();
        } catch (Exception e) {
            throw new RemoteException(i18n.getMessage("getStatusErr"), e);
        }
        if (requestId == -1) {
            logger.error(i18n.getMessage("resourceNotFoundErr"));
            throw new ResourceUnknownFaultType();
        }
        try {
            statusType = this.dbAdapter.getStatus(requestId, sourceUrl);
        } catch (RftDBException e) {
            throw (RFTDatabaseFaultType) FaultUtil.makeFault(
                    RFTDatabaseFaultType.class, requestId, i18n
                            .getMessage("dbRetrieveErr"), e);
        }

        if (statusType == null) {
            throw new RemoteException(i18n.getMessage("getStatusErr"));
        }
        return new GetStatusResponse(statusType);
    }

    /**
     * Get an array of status of transfers starting from initial
     * 
     * @param params
     * @return GetStatusSetResponse
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     */
    public GetStatusSetResponse getStatusSet(GetStatusSet params)
            throws RemoteException, ResourceUnknownFaultType,
            RFTDatabaseFaultType {
        GetStatusSetResponse response = null;
        TransferStatusType statusSet[] = null;
        Vector statusSetVector = null;
        int initial, offset, requestId = -1;
        initial = params.getFrom();
        offset = params.getOffset();
        if (logger.isDebugEnabled()) {
            logger.debug("GetStatus of " + offset + " transfers starting from "
                    + initial);
        }
        try {
            requestId = this.getRequestId();
        } catch (Exception e) {
            throw new RemoteException(i18n.getMessage("getStatusSetErr"), e);
        }
        if (requestId == -1) {
            logger.error(i18n.getMessage("resourceNotFoundErr"));
            throw new ResourceUnknownFaultType();
        }
        try {
            statusSetVector = this.dbAdapter.getStatusSet(requestId, initial,
                    offset);
            int statusSetSize = statusSetVector.size();
            if (statusSetSize > 0) {
                statusSet = new TransferStatusType[statusSetSize];
                for (int i = 0; i < statusSetSize; i++) {
                    statusSet[i] = (TransferStatusType) statusSetVector
                            .elementAt(i);
                }
                response = new GetStatusSetResponse();
                response.setTransferStatusSet(statusSet);
            } else {
                logger.error(i18n.getMessage("getStatusSetErr"));
            }
        } catch (RftDBException e) {
            throw (RFTDatabaseFaultType) FaultUtil.makeFault(
                    RFTDatabaseFaultType.class, requestId, i18n
                            .getMessage("dbRetrieveErr"), e);
        }
        if (statusSet == null) {
            throw new RemoteException(i18n.getMessage("getStatusSetErr"));
        }
        return response;
    }

    /**
     * cancels transfers in a request
     * 
     * @param params
     * @throws RemoteException
     * @throws ResourceUnknownFaultType
     * @throws RFTDatabaseFaultType
     * @return CancelResponse
     */
    public CancelResponse cancel(Cancel params) throws RemoteException,
            ResourceUnknownFaultType, RFTDatabaseFaultType {
        CancelResponse response = null;
        int fromId = -1;
        int toId = -1;
        int requestId = -1;
        int transferId = -1;
        fromId = params.getFromId();
        toId = params.getToId();
        if (logger.isDebugEnabled()) {
            logger.debug("Invoked cancel for transfer from " + fromId + " to "
                    + toId);
        }
        try {
            requestId = this.getRequestId();
            if (requestId == -1) {
                throw new ResourceUnknownFaultType();
            }
        } catch (Exception e) {
            throw new RemoteException(i18n.getMessage("unableToCancelErr"), e);
        }
        try {

            transferId = this.dbAdapter.getFirstTransferId(requestId);
            if (fromId != -1 && toId != -1) {
                fromId = transferId + fromId - 1;
                toId = transferId + toId - 1;
            }
            if (logger.isDebugEnabled()) {
                logger.debug("Cancelling from:" + fromId + " to " + toId);
            }
            TransferWork.stopActiveTransfers(requestId, fromId, toId);
            this.dbAdapter.updateToCancelled(requestId, fromId, toId);
        } catch (RftDBException e) {
            logger.error(i18n.getMessage("dbRetrieveErr"), e);
            throw (RFTDatabaseFaultType) FaultUtil.makeFault(
                    RFTDatabaseFaultType.class, requestId, i18n
                            .getMessage("dbRetrieveErr"), e);

        }

        response = new CancelResponse();
        return response;

    }
    
    /**
     * Util method to get the requestId from the database for a resource
     * 
     * @return requestId
     * @throws Exception
     */
    private int getRequestId() throws Exception {
        int requestId = -1;
        ResourceContext resourceContext = ResourceContext.getResourceContext();
        ResourceKey resourceKey = resourceContext.getResourceKey();
        requestId = Integer.parseInt((String) resourceKey.getValue());
        return requestId;
    }
    
    private synchronized void 
        startTransfers(ReliableFileTransferResource resource) throws Exception {
        int requestId = resource.getRequestId();
        int concurrency = resource.getConcurrency();
        RFTResourceManager resourceManager = resource.getResourceManager();
        ResourcePropertySet propSet = resource.getResourcePropertySet();
        InitialContext initialCtx = new InitialContext();
        WorkManager workManager = (WorkManager) initialCtx
                .lookup(RFTConstants.WORKMANAGER);
        int countTransfers = 1;
        while (countTransfers <= concurrency) {
            TransferType transferType = resourceManager.getNext(); 
            if( transferType != null ) {
                    TransferWork transferWork = new TransferWork(requestId,
                            propSet, transferType);
                    transferWork.setWorkManager(workManager);
                    transferWork.setResourceManager(
                            resource.getResourceManager());
                    workManager.schedule(transferWork);
            }
            countTransfers++;
        }
    }
    
    private static RFTDatabaseOptions dbConfiguration;
    {
        try {
            ResourceContext ctx = ResourceContext.getResourceContext();
            Context initialContext = new InitialContext();
            String lookupString = Constants.JNDI_SERVICES_BASE_NAME
                    + ctx.getService() + "/dbConfiguration";
            dbConfiguration = (RFTDatabaseOptions) initialContext
                    .lookup(lookupString);
        } catch (Exception e) {
            logger
                    .error(i18n.getMessage("dbConfigFileErr") + e.getMessage(),
                            e);
            throw new RuntimeException(i18n.getMessage("dbConfigFileErr"), e);
        }
    }

    private static RFTConfiguration serviceConfiguration;
    {
        try {
            ResourceContext ctx = ResourceContext.getResourceContext();
            Context initialContext = new InitialContext();
            String lookupString = Constants.JNDI_SERVICES_BASE_NAME
                    + ctx.getService() + "/configuration";
            serviceConfiguration = (RFTConfiguration) initialContext
                    .lookup(lookupString);
            if (logger.isDebugEnabled()) {
                logger.debug("maximum active allowed:"
                        + serviceConfiguration.getMaxActiveAllowed());
            }
        } catch (Exception e) {
            logger.error(i18n.getMessage("serviceConfigErr") + e.getMessage(),
                    e);
            throw new RuntimeException(i18n.getMessage("serviceConfigErr")
                    + e.getMessage(), e);
        }
    }

    /**
     * @return the factory configuration object
     */
    public static RFTDatabaseOptions getDatabaseConfiguration() {
        return dbConfiguration;
    }

    /**
     * @return service configuration
     */
    public static RFTConfiguration getRFTConfiguration() {
        return serviceConfiguration;
    }
}

