/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import java.io.File;

import java.io.IOException;

import java.net.ConnectException;
import java.net.MalformedURLException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.ftp.ByteRangeList;
import org.globus.ftp.DataChannelAuthentication;
import org.globus.ftp.DataSink;
import org.globus.ftp.DataSource;
import org.globus.ftp.FileRandomIO;
import org.globus.ftp.GridFTPClient;
import org.globus.ftp.GridFTPRestartMarker;
import org.globus.ftp.GridFTPSession;
import org.globus.ftp.MlsxEntry;
import org.globus.ftp.RetrieveOptions;
import org.globus.ftp.HostPortList;
import org.globus.ftp.Session;
import org.globus.ftp.exception.ServerException;

import org.globus.util.Util;

import org.globus.gsi.gssapi.auth.IdentityAuthorization;

import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.TotalBytesTransferredResource;

import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.exception.RftException;

import org.globus.util.GlobusURL;
import org.globus.util.I18n;

import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.ResourceProperty;

import org.ietf.jgss.GSSCredential;

/**
 * Transfer Client
 */
public class TransferClient {

    private GridFTPClient sourceHost;

    private GridFTPClient destinationHost;

    private String sourcePath;

    private String destinationPath;

    private int status = -1;

    private int transferid;

    private int parallelism;

    private int tcpBufferSize;

    private GSSCredential credential;

    private GlobusURL sourceGlobusURL;

    private GlobusURL destinationGlobusURL;

    private long size;

    private RFTOptionsType rftOptions;

    private String sourceSubjectName;

    private String destinationSubjectName;

    private MyMarkerListener markerListener;

    private URLExpander urlExpander;

    private static Log logger = LogFactory.getLog(TransferClient.class
            .getName());
    
    private static I18n i18n = 
        I18n.getI18n("org.globus.transfer.reliable.errors", 
                TransferClient.class.getClassLoader());

    private int requestId = 0;

    private ResourcePropertySet rpSet;

    private ResourceProperty totalBytesTransferredRP;

    private TotalBytesTransferredResource totalBytes;

    private String userName = null;
    
    private String permissions = null;

    private int attempts = 1;
    
    /**
     * @param requestId
     */
    public void setRequestId(int requestId) {
        this.requestId = requestId;
    }

    /**
     * @return requestId
     */
    public int getRequestId() {
        return this.requestId;
    }

    /**
     * @param userName
     */
    public void setUserName(String userName) {
        this.userName = userName;
    }

    /**
     * @return userName
     */
    public String getUserName() {
        return this.userName;
    }
    
    /**
     * @param attempts
     */
    public void setAttempts(int attempts) {
        this.attempts = attempts;
    }

    /**
     * @param rpSet
     */
    public void setResourcePropertySet(ResourcePropertySet rpSet) {
        this.rpSet = rpSet;
        this.totalBytesTransferredRP = 
            this.rpSet.get(RFTConstants.TOTAL_BYTES_RESOURCE);
        this.totalBytes = (TotalBytesTransferredResource)
            this.totalBytesTransferredRP.get(0);
    }

    /**
     * Sets the sourceHost attribute of the TransferClient object
     * 
     * @param sourceURL
     *            The new sourceHost value
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public void setSourceHost(String sourceURL) throws RftException,
            RftTransientException, RftDBException {
        try {
            this.sourceGlobusURL = new GlobusURL(sourceURL);
            if(!this.sourceGlobusURL.getProtocol().equalsIgnoreCase("gsiftp")) {
                throw new RftException(i18n.getMessage("thirdPartyOnlyErr"));
            }
        } catch (MalformedURLException mue) {
            setStatus(RFTConstants.STATUS_FAILED);
            logger.error(i18n.getMessage("invalidUrlErr") + sourceURL);
            throw new RftException(i18n.getMessage("invalidUrlErr"), mue);
        }
        try {
            this.sourceHost = new GridFTPClient(this.sourceGlobusURL.getHost(),
                    this.sourceGlobusURL.getPort());
            this.sourcePath = "/" + this.sourceGlobusURL.getPath();
        } catch (ConnectException ce) {
            // This is container level backoff
            logger.error(i18n.getMessage("unableToConnectErr", new Object[] {
                    this.sourceGlobusURL.getHost(),
                    Integer.toString(this.sourceGlobusURL.getPort()) }));
            setStatus(RFTConstants.STATUS_RETRYING);
            ReliableFileTransferDbAdapter.getDbAdapter().backOff(
                    this.sourceGlobusURL.getHost() + ":"
                            + +this.sourceGlobusURL.getPort());
            throw new RftTransientException(i18n.getMessage(
                    "unableToConnectErr",
                    new Object[] {
                    this.sourceGlobusURL.getHost(),
                    Integer.toString(this.sourceGlobusURL.getPort())
                            }));
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            logger.error(i18n.getMessage("unableToConnectErr",
                    new Object[] {
                    this.sourceGlobusURL.getHost(),
                    Integer.toString(this.sourceGlobusURL.getPort())
                            }));
            throw new RftException(i18n.getMessage(
                    "unableToConnectErr",
                    new Object[] {
                    this.sourceGlobusURL.getHost(),
                    Integer.toString(this.sourceGlobusURL.getPort())
                            }));
        }
    }

    /**
     * Sets the destinationHost attribute of the TransferClient object
     * 
     * @param destURL
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public void setDestinationHost(String destURL) throws RftException,
            RftTransientException, RftDBException {
        try {
            this.destinationGlobusURL = new GlobusURL(destURL);
            if(!this.destinationGlobusURL.getProtocol()
                    .equalsIgnoreCase("gsiftp")) {
                throw new RftException(i18n.getMessage("thirdPartyOnlyErr"));
            }
        } catch (MalformedURLException mue) {
            setStatus(RFTConstants.STATUS_FAILED);
            logger.error(i18n.getMessage("invalidUrlErr") + destURL);
            throw new RftException(i18n.getMessage("invalidUrlErr", destURL), mue);
        }
        try {
            this.destinationHost = new GridFTPClient(this.destinationGlobusURL
                    .getHost(), this.destinationGlobusURL.getPort());
            this.destinationPath = "/" + this.destinationGlobusURL.getPath();
        } catch (ConnectException ce) {
            logger.error(i18n.getMessage("unableToConnectErr", new Object[] {
                    this.destinationGlobusURL.getHost(),
                    Integer.toString(this.destinationGlobusURL.getPort()) }));
            setStatus(RFTConstants.STATUS_RETRYING);
            ReliableFileTransferDbAdapter.getDbAdapter().backOff(
                    this.destinationGlobusURL.getHost() + ":"
                            + +this.destinationGlobusURL.getPort());
            throw new RftTransientException(i18n.getMessage(
                    "unableToConnectErr",
                    new Object[] {
                    this.destinationGlobusURL.getHost(),
                    Integer.toString(this.destinationGlobusURL.getPort())
                            }));
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            logger.error(i18n.getMessage("unableToConnectErr",
                    new Object[] {
                    this.destinationGlobusURL.getHost(),
                    Integer.toString(this.destinationGlobusURL.getPort())
                            }));
            throw new RftException(i18n.getMessage(
                    "unableToConnectErr",
                    new Object[] {
                    this.destinationGlobusURL.getHost(),
                    Integer.toString(this.destinationGlobusURL.getPort())
                            }));
        }
    }

    /**
     * Gets the sourceURL attribute of the TransferClient object
     * 
     * @return The sourceURL value
     */
    public GlobusURL getSourceURL() {
        return this.sourceGlobusURL;
    }

    public void setSourceURL(GlobusURL sourceGlobusURL) {
        this.sourceGlobusURL = sourceGlobusURL;
    }
    

    /**
     * Gets the destinationURL attribute of the TransferClient object
     * 
     * @return The destinationURL value
     */
    public GlobusURL getDestinationURL() {
        return this.destinationGlobusURL;
    }

    public void setDestinationURL(GlobusURL destinationGlobusURL) {
        this.destinationGlobusURL = destinationGlobusURL;
    }

    /**
     * Sets the status attribute of the TransferClient object
     * 
     * @param status
     *            The new status value
     */
    public synchronized void setStatus(int status) {
        this.status = status;
    }

   /**
   * Sets the credential to be used for this transfer
   * @param GSSCredential credential
   */
    public void setCredential(GSSCredential credential) {
        this.credential = credential;
    }
            

    /**
     * Sets the transferId attribute of the TransferClient object
     * 
     * @param transferId
     *            The new transferId value
     */
    public void setTransferId(int transferId) {
        this.transferid = transferId;
    }

    /**
     * Gets the size attribute of the TransferClient object
     * 
     * @return The size value
     */
    public long getSize() {
        return this.size;
    }

    /**
     * Gets the urlExpander attribute of the TransferClient object
     * 
     * @return The urlExpander value
     */
    public URLExpander getUrlExpander() {
        return this.urlExpander;
    }

    /**
    * Check for reuse of this transferClient
    * @param requestId
    * @param targetSource
     * @param targetDest
     * @return true if reusable
    */
    public boolean checkForReuse(int requestId, GlobusURL targetSource, 
            GlobusURL targetDest) {
        boolean reuse = false;
        if ((this.requestId == requestId) 
             && ((this.sourceGlobusURL.getHost()).equals(targetSource.getHost()))
             && ((this.destinationGlobusURL.getHost()).equals(targetDest.getHost()))
             && (this.sourceGlobusURL.getPort() == targetSource.getPort())
             && (this.destinationGlobusURL.getPort() == targetDest.getPort()) ) {
                reuse = true;
        }
        return reuse;
    }

    /**
     * Sets the authorization attribute of the TransferClient object
     * 
     * @throws RftException
     */
    public void authorize() throws RftException {
        try {
            String subjectName = this.rftOptions.getSubjectName();
            this.sourceSubjectName = this.rftOptions.getSourceSubjectName();
            this.destinationSubjectName = this.rftOptions
                    .getDestinationSubjectName();
            
            if (subjectName != null) {
                this.sourceSubjectName = subjectName;
                this.destinationSubjectName = subjectName;
            }
            authorizeSource();
           
           authorizeDestination();
        } catch (Exception e) {
            logger.error(
                    i18n.getMessage("authzErr") + e.getMessage(), e);
            setStatus(RFTConstants.STATUS_FAILED);
            throw new RftException(
                    i18n.getMessage("authzErr") + e.getMessage(), e);
        }
    }
    
    private void authorizeSource() {
        if ((this.sourceSubjectName != null)
                && !(this.sourceSubjectName.equals("null"))
                && !(this.sourceSubjectName.trim().equals(""))) {
            this.sourceHost.setAuthorization(new IdentityAuthorization(
                    this.sourceSubjectName));
            
        }
    }
    private void authorizeDestination() {
        if ((this.destinationSubjectName != null)
                && !(this.destinationSubjectName.equals("null"))
                && !(this.destinationSubjectName.trim().equals(""))) {
            this.destinationHost.setAuthorization(new IdentityAuthorization(
                    this.destinationSubjectName));
        }
    }
    
    /**
     * Gets the status attribute of the TransferClient object
     * 
     * @return The status value
     */
    public synchronized int getStatus() {

        return this.status;
    }

    /**
     * Sets the destinationPath attribute of the TransferClient object
     * 
     * @param destPath
     *            The new destinationPath value
     */
    public void setDestinationPath(String destPath) {
        this.destinationPath = destPath;
    }

    /**
     * Gets the transferID attribute of the TransferClient object
     * 
     * @return The transferID value
     */
    public int getTransferID() {

        return this.transferid;
    }

    /**
     * Sets the sourcePath attribute of the TransferClient object
     * 
     * @param sourcePath
     *            The new sourcePath value
     */
    public synchronized void setSourcePath(String sourcePath) { 
        this.sourcePath = "/" + sourcePath;

    }
    
    /**
     * @param rftOptions
     */
    public void setOptions(RFTOptionsType rftOptions) {
        this.rftOptions = rftOptions;
    }
    /**
     * Sets the rFTOptions attribute of the TransferClient object
     * 
     * @throws RftException
     */
    public void setRFTOptions() throws RftException {
        try {
        this.parallelism = this.rftOptions.getParallelStreams().intValue();
        this.tcpBufferSize = this.rftOptions.getTcpBufferSize().intValue();
        if (this.parallelism == 0 ) {
            this.parallelism = 1;
        }
        if (this.tcpBufferSize == 0) {
            this.tcpBufferSize = 16384;
        }
        setSourceOptions();
        setDestinationOptions();
        } catch (Exception e) {
            logger.error(i18n.getMessage("transferOptionsErr") 
                    + e.getMessage(), e);
            throw new RftException(i18n.getMessage("transferOptionsErr") 
                    + e.getMessage(), e);
        }
    }
    
    private void setSourceOptions() throws ServerException, IOException {
        if (!this.sourcePath.endsWith("/")) {
            // This is not a dir transfer
            this.sourceHost.setOptions(new RetrieveOptions(this.parallelism));
            this.sourceHost.setTCPBufferSize(this.tcpBufferSize);
        }
        if (this.rftOptions.getBinary().booleanValue()) {
            this.sourceHost.setType(Session.TYPE_IMAGE);
        } else {
            this.sourceHost.setType(Session.TYPE_ASCII);
        }
        if (this.rftOptions.getDcau().booleanValue()) {
            this.sourceHost.
                setDataChannelAuthentication(DataChannelAuthentication.SELF);
            this.sourceHost.setProtectionBufferSize(16384);
        } else {
            this.sourceHost.
                setDataChannelAuthentication(DataChannelAuthentication.NONE);
        }
    }
    
    private void setDestinationOptions() throws ServerException, IOException {
        this.destinationHost.setOptions(new RetrieveOptions(this.parallelism));
        this.destinationHost.setTCPBufferSize(this.tcpBufferSize);
        if (this.rftOptions.getBinary().booleanValue()) {
            this.destinationHost.setType(Session.TYPE_IMAGE);
        } else {
            this.destinationHost.setType(Session.TYPE_ASCII);
        }
        if (this.rftOptions.getDcau().booleanValue()) {
            this.destinationHost.
                setDataChannelAuthentication(DataChannelAuthentication.SELF);
            this.destinationHost.setProtectionBufferSize(16384);
        } else {
            this.destinationHost.
                setDataChannelAuthentication(DataChannelAuthentication.NONE);
        }
    }

    /**
     * @throws RftException
     */
    public void authenticate() throws RftException {
        authenticateSource();
        authenticateDestination();
    }
    
    private void authenticateSource() throws RftException {
        try {
            String tempUserName = this.rftOptions.getUserName();
            if (tempUserName != null && 
                    !tempUserName.equalsIgnoreCase("null")) {
                this.sourceHost.authenticate(this.credential, tempUserName);
            } else {
                this.sourceHost.authenticate(this.credential);
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            throw new RftException(
                    i18n.getMessage("authErr") + e.getMessage(), 
                    e);
        }
    }
    private void authenticateDestination() 
        throws RftException {
        try {
            String tempUserName = this.rftOptions.getUserName();
            if (tempUserName != null && 
                    !tempUserName.equalsIgnoreCase("null")) {
                this.destinationHost.authenticate(this.credential, tempUserName);
            } else {
                this.destinationHost.authenticate(this.credential);
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            throw new RftException(
                    i18n.getMessage("authErr") + e.getMessage(), 
                    e);
        }
    }
    public void setModeE() throws RftException {
        try {
            if(!this.rftOptions.getNotpt().booleanValue()) {
                this.sourceHost.setMode(GridFTPSession.MODE_EBLOCK);
                this.destinationHost.setMode(GridFTPSession.MODE_EBLOCK);
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            throw new RftException(i18n.getMessage("transferOptionsErr") 
                    + e.getMessage(), e);
        }
    }

    /**
     * Sets the restartMarker attribute of the TransferClient object
     * 
     * @param marker
     *            The new restartMarker value
     */
    public void setRestartMarker(String marker) {

        try {
            marker = "Range Marker " + marker;

            GridFTPRestartMarker restartmarker = new GridFTPRestartMarker(
                    marker);
            ByteRangeList list = new ByteRangeList();
            list.merge(restartmarker.toVector());
            this.sourceHost.setRestartMarker(list);
        } catch (Exception e) {
            logger.error(i18n.getMessage("restartMarkerErr")
                    + e.getMessage(), e);
        }
    }

    /**
     * Closes the sockets to source and destination hosts
     */
    public void close() {
        try {
            if (this.sourceHost != null) {
                this.sourceHost.close(true);
            }
            if (this.destinationHost != null) {
                this.destinationHost.close(true);
            }
        } catch (Exception e) {
            logger.error(i18n.getMessage("closeClientErr")
                    + e.getMessage(), e);
        }
    }
    
    private void updatePermissions() throws RftException {
        try {
            MlsxEntry entry = this.sourceHost.mlst(this.sourcePath);
            this.size = Long.parseLong(entry.get(MlsxEntry.SIZE));
            this.permissions = entry.get("unix.mode");
            ReliableFileTransferDbAdapter.getDbAdapter().
            updateTransfer(this.transferid,this.size,this.permissions);
        } catch (Exception ioe) {
            logger.error(i18n.getMessage("updatePermsErr") + this.sourcePath,
                    ioe);
            if (this.rftOptions.getIgnoreFilePermErr().booleanValue()) {
                logger.error("MLST is not supported, But thats ok");
            } else {
                setStatus(RFTConstants.STATUS_FAILED);
                throw new RftException(i18n.getMessage("updatePermsErr") 
                    + this.sourcePath,
                    ioe);
            }
        }
    }
    /**
     * Performs the actual transfer
     * 
     * @param useExtended
     * @throws RftException
     * @throws RftTransientException
     */
    public void transfer(boolean useExtended) throws RftException,
            RftTransientException {
        if (this.sourcePath.endsWith("/")) {
            setStatus(RFTConstants.STATUS_EXPANDING);
            try {
                this.urlExpander = new URLExpander(this.sourceHost,
                        this.destinationHost, this.sourceGlobusURL,
                        this.destinationGlobusURL, this.rftOptions);
                this.urlExpander.setRequestId(this.requestId);
                this.urlExpander.start();
                return;
            } catch (Exception e) {
                logger.error(i18n.getMessage("terminalTransferErr")
                        + e.getMessage(), e);
                setStatus(RFTConstants.STATUS_FAILED);
                throw new RftException(i18n.getMessage("dirExpansionErr"),
                        e);
            }
        }
        
        boolean noThirdParty = this.rftOptions.getNotpt().booleanValue();
        
        this.markerListener = new MyMarkerListener(this.transferid, this.size);
        if (noThirdParty) {
            updatePermissions();
            if (useExtended) {
                noThirdPartyExtendedTransfer();
            } else {
                noThirdPartyNonExtendedTransfer();
            }
            setPermissions();
        } else {
            updatePermissions();
            if (useExtended) {
                normalExtendedTransfer();
            } else {
                normalNonExtendedTransfer();
            }
            setPermissions();
        }
        if (this.status != RFTConstants.STATUS_RETRYING) {
            setStatus(RFTConstants.STATUS_FINISHED);
            long temp = this.totalBytes.getTotalBytes() + this.size;
            this.totalBytes.setTotalBytes(temp);
            this.totalBytesTransferredRP.set(0, this.totalBytes);
        }
    }

    /**
     */
    public void deleteDestination() {
        String file = this.destinationPath;
        try {
            if (logger.isDebugEnabled()) {
                logger.debug("Deleting " + file);
            }
            this.destinationHost.deleteFile(file);
        } catch (Exception e) {
            logger.error(i18n.getMessage("deleteFileErr",file), e);
        }
    }
    
    private void setPermissions() throws RftException {
        try {
            if (this.permissions != null) {
                String command = "CHMOD " + this.permissions 
                    + " " + this.destinationPath;
                this.destinationHost.site(command);
            } else {
                if (this.rftOptions.getIgnoreFilePermErr().booleanValue()) {
                    logger.error("Not setting file perms");
                } else {
                    throw new RftException("Permissions are null");
                }
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_FAILED);
            throw new RftException(i18n.getMessage("setPermsErr",
                    this.destinationPath) + e.getMessage() ,e);
        }
        
    }
      
   /**
   * Enable Striping if the servers support it
   * @throws IOException
   * @throws ServerException
   */
    public void setStriping() 
        throws RftException {
        try {
            if(!this.rftOptions.getNotpt().booleanValue()) {
                boolean striping = false;
                if (this.sourceHost.isFeatureSupported("SPOR") && 
                        this.destinationHost.isFeatureSupported("SPAS")) {
                    striping = true;
                }
                if(striping) {
                    HostPortList hpl = this.destinationHost.setStripedPassive();
                    this.sourceHost.setStripedActive(hpl);
                }
            }
        } catch (Exception e) {
            throw new RftException("Exception in Striping", e);
        }
    }

    /**
     * Performs notpt non-extended transfer
     * @throws RftTransientException
     */
    private void noThirdPartyNonExtendedTransfer() 
    throws RftTransientException {
        try {
            File fullLocalFile = File.createTempFile("TempRFT", 
                    String.valueOf(this.transferid));
            Util.setOwnerAccessOnly(fullLocalFile.getAbsolutePath());
            fullLocalFile.deleteOnExit();
            DataSink sink = null;
            sink = new FileRandomIO(new java.io.RandomAccessFile(fullLocalFile,
                    "rw"));
            this.sourceHost.get(this.sourcePath, sink, this.markerListener);

            DataSource source = null;
            source = new FileRandomIO(new java.io.RandomAccessFile(
                    fullLocalFile, "r"));
            this.destinationHost.put(this.destinationPath, source,
                    this.markerListener);
            if(fullLocalFile.exists()) {
                fullLocalFile.delete();
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_RETRYING);
            logger.error("Exception in transfer", e);
            //This should be a transfer level backoff
            try {
                ReliableFileTransferDbAdapter.getDbAdapter().
                    backOff(this.transferid, this.attempts);
            } catch (RftDBException ee) {
                logger.error(i18n.getMessage("dbUpdateErr")
                        + ee.getMessage() ,ee);
            }
            throw new RftTransientException(
                    i18n.getMessage("transientTransferErr") 
                    + e.getMessage(), e);
        }
    }
    /**
     * Performs notptExtendedTransfer
     * @throws RftTransientException
     *  
     */
    private void noThirdPartyExtendedTransfer() 
    throws RftTransientException, RftException {
        try {
            File fullLocalFile = File.createTempFile("TempRFT", 
                    String.valueOf(this.transferid));
            Util.setOwnerAccessOnly(fullLocalFile.getAbsolutePath());
            DataSink sink = null;
            sink = new FileRandomIO(new java.io.RandomAccessFile(fullLocalFile,
                    "rw"));
            this.sourceHost.extendedGet(this.sourcePath, this.size, sink,
                    this.markerListener);

            DataSource source = null;
            source = new FileRandomIO(new java.io.RandomAccessFile(
                    fullLocalFile, "r"));
            this.destinationHost.extendedPut(this.destinationPath, source,
                    this.markerListener);
            if(fullLocalFile.exists()) {
                fullLocalFile.delete();
            }
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_RETRYING);
            logger.error("Exception in transfer", e);
            //This should be a transfer level backoff
            try {
                ReliableFileTransferDbAdapter.getDbAdapter().
                    backOff(this.transferid, this.attempts);
            } catch (RftDBException ee) {
                logger.error(i18n.getMessage("dbUpdateErr")
                        + ee.getMessage() ,ee);
            }
            throw new RftTransientException(
                    i18n.getMessage("transientTransferErr") 
                    + e.getMessage(), e);
        }
    }

    /**
     * Performs the normal extended transfer
     * @throws RftTransientException
     */
    private void normalExtendedTransfer() throws RftTransientException {
        try {
            this.sourceHost.extendedTransfer(this.sourcePath,
                    this.destinationHost, this.destinationPath,
                    this.markerListener);
        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_RETRYING);
            logger.error("Exception in transfer", e);
            //This should be a transfer level backoff
            try {
                ReliableFileTransferDbAdapter.getDbAdapter().
                    backOff(this.transferid, this.attempts);
            } catch (RftDBException ee) {
                logger.error(i18n.getMessage("dbUpdateErr")
                        + ee.getMessage() ,ee);
            }
            throw new RftTransientException(
                    i18n.getMessage("transientTransferErr")
                    + e.getMessage(), e);
        }
    }

    /**
     * Performs normal non-extended transfer
     * @throws RftTransientException
     */
    private void normalNonExtendedTransfer() 
        throws RftTransientException {

        try {
            this.sourceHost.transfer(this.sourcePath, this.destinationHost,
                    this.destinationPath, false, this.markerListener);

        } catch (Exception e) {
            setStatus(RFTConstants.STATUS_RETRYING);
            logger.error("Exception in transfer", e);
            //This should be a transfer level backoff
            try {
                ReliableFileTransferDbAdapter.getDbAdapter().
                    backOff(this.transferid, this.attempts);
            } catch (RftDBException ee) {
                logger.error(i18n.getMessage("dbUpdateErr")
                        + ee.getMessage() ,ee);
            }
            throw new RftTransientException(
                    i18n.getMessage("transientTransferErr") 
                    + e.getMessage(), e);
        }
    }
}

