/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service.database;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.DriverManagerConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDriver;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.globus.transfer.reliable.service.RFTConfiguration;
import org.globus.transfer.reliable.service.ReliableFileTransferImpl;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.util.I18n;
/**
 * Sets up database
 */
public class RFTDatabaseSetup {
    private static Log logger = LogFactory.getLog(RFTDatabaseSetup.class
            .getName());
    
    private static I18n i18n = I18n.getI18n("org.globus.transfer.reliable.errors",
            RFTDatabaseSetup.class.getClassLoader()); 

    private static RFTDatabaseOptions rftDatabaseOptions = null;

    private static boolean driverSetup = false;

    final private static String baseURI = "jdbc:apache:commons:dbcp:";

    final private static String rftURI = "rft";

    static GenericObjectPool connectionPool = null;

    static int def_activeConnections = 32;

    static byte def_onExhaustAction = 1;

    static long def_maxWait = -1;

    static int def_idleConnections = 2;

    // Back off when encountered a transient exception
    // in seconds
    static int backOff = 10;

    public static int maxActive = 10;

    /**
     * constructor
     */
    private RFTDatabaseSetup() {
        //Empty constructor
    }

    /**
     * sets up the database
     * 
     * @throws RftDBException
     * @throws RftException
     */
    public static void setup() throws RftDBException, RftException {
        RFTConfiguration rftConfig = 
            ReliableFileTransferImpl.getRFTConfiguration();
        try {
            String tempBackOff = rftConfig.getBackOff();
            if (tempBackOff != null) {
                backOff = Integer.parseInt(tempBackOff);
            }
            String tempMaxActive = rftConfig.getMaxActiveAllowed();
            if (tempMaxActive != null) {
                maxActive = Integer.parseInt(tempMaxActive);
            }
        } catch (Exception e) {
            throw new RftException(i18n.getMessage("dbConfigFileErr")
                    + e.getMessage(), e);
        }
        RFTDatabaseOptions rftDbOption = ReliableFileTransferImpl
                .getDatabaseConfiguration();
        try {
            setupDBConnection(rftDbOption);
        } catch (RftDBException exp) {
            throw new RftDBException(i18n.getMessage("dbSetupErr")
                    + exp.getMessage(), exp);
        }

    }

    /**
     * @param rftDatabaseOptions
     * @param activeConnections
     * @throws RftDBException
     */
    public static void setupDBConnection(RFTDatabaseOptions rftDatabaseOptions,
            int activeConnections) throws RftDBException {
        setupDBConnection(rftDatabaseOptions, activeConnections,
                def_onExhaustAction, def_maxWait, def_idleConnections);
    }

    /**
     * @param rftDatabaseOptions
     * @param activeConnections
     * @param onExhaustAction
     * @throws RftDBException
     */
    public static void setupDBConnection(RFTDatabaseOptions rftDatabaseOptions,
            int activeConnections, byte onExhaustAction) throws RftDBException {
        setupDBConnection(rftDatabaseOptions, activeConnections,
                onExhaustAction, def_maxWait, def_idleConnections);
    }

    /**
     * @param rftDatabaseOptions_
     * @param activeConnections
     * @param onExhaustAction
     * @param maxWait
     * @throws RftDBException
     */
    public static void setupDBConnection(
            RFTDatabaseOptions rftDatabaseOptions_, int activeConnections,
            byte onExhaustAction, long maxWait) throws RftDBException {
        setupDBConnection(rftDatabaseOptions_, activeConnections,
                onExhaustAction, maxWait, def_idleConnections);
    }

    public static void setupDBConnection(RFTDatabaseOptions rftDatabaseOptions)
            throws RftDBException {
        setupDBConnection(rftDatabaseOptions, rftDatabaseOptions.getMaxActive(),
                def_onExhaustAction, rftDatabaseOptions.getMaxWait(),
                rftDatabaseOptions.getMaxIdle());
    }

    /**
     * @param rftDatabaseOptions_
     * @param activeConnections
     * @param onExhaustAction
     * @param maxWait
     * @param idleConnections
     * @throws RftDBException
     */
    public static void setupDBConnection(
            RFTDatabaseOptions rftDatabaseOptions_, int activeConnections,
            byte onExhaustAction, long maxWait, int idleConnections)
            throws RftDBException {
        if (!driverSetup) {
            if(logger.isDebugEnabled()) {
                logger.debug("Setting up db : maximum active connection is "
                        + activeConnections);
                logger.debug("Setting up db : maximum idle connection is "
                        + idleConnections);
                logger.debug("dbOptions"
                        + rftDatabaseOptions_.getConnectionString() + " "
                        + rftDatabaseOptions_.getUserName() + " "
                        + rftDatabaseOptions_.getPassword() + " "
                        + rftDatabaseOptions_.getDriverName() + " " 
                        + activeConnections + " " + onExhaustAction 
                        + " " + maxWait + " " + idleConnections
                        );
            }
            try {
                setupDriver(rftDatabaseOptions_.getConnectionString(),
                        rftDatabaseOptions_.getUserName(), rftDatabaseOptions_
                                .getPassword(), activeConnections,
                        onExhaustAction, maxWait, idleConnections);
               rftDatabaseOptions = rftDatabaseOptions_;
               Class.forName(rftDatabaseOptions.getDriverName());
            } catch (Exception e) {
                logger.error(i18n.getMessage("dbSetupErr"), e);
                throw new RftDBException(i18n.getMessage("dbSetupErr")
                        + e.getMessage(), e);
            }
            driverSetup = true;
        }
    }

    /**
     * @return
     * @throws RftDBException
     */
    public static Connection getDBConnection() throws RftDBException {
        Connection connection = null;
        if (rftDatabaseOptions == null)
            throw new RftDBException(i18n.getMessage("driverNotInitErr"));
        try {
            connection = DriverManager.getConnection(baseURI
                    + rftURI);

            return connection;
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbConnErr") 
                    + e.getMessage(), e);
        }
    }

    
    private static void setupDriver(String connectURI, String username,
            String password, int activeConnections, byte onExhaustAction,
            long maxWait, int idleConnections) throws Exception {

        // Object pool which is a pool of conection
        connectionPool = new GenericObjectPool(null, activeConnections,
                onExhaustAction, maxWait, idleConnections, true, false);
        // ConnectionFactory that pool uses to create connectiosn
        ConnectionFactory connectionFactory = new DriverManagerConnectionFactory(
                connectURI, username, password);

        // PoolableConnectionFactory used for pooling functionality

        String validationQuery = "SELECT max(request_id) from requestid ";
        PoolableConnectionFactory poolableConnectionFactory = 
            new PoolableConnectionFactory(
                connectionFactory, (ObjectPool) connectionPool, null, 
                validationQuery,
                false, true);

        // Create and Register PoolingDriver
        PoolingDriver driver = new PoolingDriver();
        driver.registerPool(rftURI, connectionPool);
    }

    /*
     * Returns the connection to the database pool
     */
    /**
     * @param connection
     * @throws RftDBException
     */
    public static void returnDBConnection(Connection connection)
            throws RftDBException {
        if (rftDatabaseOptions == null)
            throw new RftDBException(i18n.getMessage("dbSetupErr")); 
        try {
            connection.close();
        } catch (SQLException sqlExcep) {
            logger.error(i18n.getMessage("dbConnReturnErr"), sqlExcep);
            throw new RftDBException(i18n.getMessage("dbConnReturnErr")
                    + sqlExcep.getMessage(), sqlExcep);
        }
    }
}

