/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service.database;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.Vector;

import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.encoding.SerializationException;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.DeleteOptionsType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.DeleteType;
import org.globus.rft.generated.OverallStatus;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferStatusType;
import org.globus.rft.generated.TransferType;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.transfer.reliable.service.RFTRecoveryData;
import org.globus.transfer.reliable.service.RFTUtil;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.util.I18n;

/**
 * Singleton to carry out database interaction
 * 
 */
public class ReliableFileTransferDbAdapter {

    private static Log logger = LogFactory
            .getLog(ReliableFileTransferDbAdapter.class.getName());

    private static ReliableFileTransferDbAdapter dbAdapter =
        new ReliableFileTransferDbAdapter();
    private static I18n i18n = 
        I18n.getI18n("org.globus.transfer.reliable.errors",
                ReliableFileTransferDbAdapter.class.getClassLoader());

    private static RFTOptionsType globalRFTOptions = new RFTOptionsType();

    private int queryConstant = 0;

    private static boolean OLD_SCHEMA = false;
    // TODO: READ ALL QUERIES FROM A PROPERTIES FILE.
    private static final String activeRequestIdSQL = "SELECT id from request "
        + "where termination_time > ? and started=? ";
    private static final String requestInsertSQL =  "INSERT INTO request("
        + "id,concurrency,termination_time,all_or_none,maxattempts,"
        + "username,started,delegated_epr,user_subject)" 
        + " VALUES (?,?,?,?,?,?,?,?,?)" ;
    private static final String requestInsertSQL_OLD =  "INSERT INTO request("
        + "concurrency,termination_time,all_or_none,maxattempts,"
        + "username,started,delegated_epr,user_subject)" 
        + " VALUES (?,?,?,?,?,?,?,?)" ;
    private static final String transferInsertSQL = 
        "INSERT INTO transfer(id, request_id,source_url,dest_url,user_name,"
        + "dcau,parallel_streams,tcp_buffer_size,block_size,notpt,binary_mode,"
        + "source_subject,dest_subject,ignore_file_perm_errors) "
        + "VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)";

    private static final String transferInsertSQL_OLD = 
        "INSERT INTO transfer(request_id,source_url,dest_url,user_name,"
        + "dcau,parallel_streams,tcp_buffer_size,block_size,notpt,binary_mode,"
        + "source_subject,dest_subject,ignore_file_perm_errors) "
        + "VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)";

    private static final String updateStartedQuery = 
        "UPDATE request set started = ? where id = ? ";

    private static final String deleteInsertQuery =
        "INSERT INTO transfer(id,request_id,source_subject,source_url," 
        + "status,user_name) VALUES(?,?,?,?,?,?)";

    private static final String deleteInsertQuery_OLD =
        "INSERT INTO transfer(request_id,source_subject,source_url," 
        + "status,user_name) VALUES(?,?,?,?,?)";

    private static final String singleTransferInsertQuery =
        "INSERT INTO transfer(id,request_id,source_url,dest_url,status," 
        + "user_name,dcau,parallel_streams,tcp_buffer_size,block_size,notpt,"
        + "binary_mode,source_subject,dest_subject) VALUES"
        + "(?,?,?,?,?,?,?,?,?,?,?,?,?,?)";

    private static final String singleTransferInsertQuery_OLD =
        "INSERT INTO transfer(request_id,source_url,dest_url,status," 
        + "user_name,dcau,parallel_streams,tcp_buffer_size,block_size,notpt,"
        + "binary_mode,source_subject,dest_subject) VALUES"
        + "(?,?,?,?,?,?,?,?,?,?,?,?,?)";
    /**
     * Constructor for the ReliableFileTransferDbAdapter object
     */
    private ReliableFileTransferDbAdapter() {
        //Empty constructor for dbAdapter
    }


    /**
     * Setup the database
     * 
     * @throws RftDBException
     * @throws RftException
     * @return returns the database adapter singleton
     */ 
    public static ReliableFileTransferDbAdapter setup() 
        throws RftDBException, RftException {
        RFTDatabaseSetup.setup();
        setSchemaVersion();
        return dbAdapter;
    }

    /**
     * Gets the dbAdapter attribute of the ReliableFileTransferDbAdapter class
     * 
     * @return The dbAdapter value
     */
    public static ReliableFileTransferDbAdapter getDbAdapter() {
        return dbAdapter;
    }

    /**
     * Store the transfer request and termination time in the database
     * 
     * @param request
     *            <code> TransferRequestType </code>
     * @param terminationTime
     *           Termination Time 
     * @param username
     * @param userSubject
     * @return requestId
     * @exception RftDBException
     */
    public synchronized int storeTransferRequest(TransferRequestType request,
            Calendar terminationTime, String username, String userSubject) 
    throws RftDBException {
        int requestId = 0;
        int concurrency = 1; //Default
        Integer tempConcurrency = request.getConcurrency();
        if (tempConcurrency != null) {
            if(tempConcurrency.intValue() > 0) {
                concurrency = tempConcurrency.intValue();
            }
        }
        Boolean allOrNone = Boolean.FALSE;
        if(request.getAllOrNone() != null) {
            allOrNone = request.getAllOrNone();
        }
        QName qName = new QName("", "DelegatedEPR");
        Connection c = null;
        PreparedStatement pstmt = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            String delegationEPRString = ObjectSerializer.toString(
                    request.getTransferCredentialEndpoint(), qName);
            c = RFTDatabaseSetup.getDBConnection();
            if(!OLD_SCHEMA) {
                pstmt = c.prepareStatement(requestInsertSQL);
                requestId = getNextRequestId();
                pstmt.setObject(1, new Integer(requestId));
                pstmt.setObject(2, new Integer(concurrency));
                pstmt.setObject(3, new Long(terminationTime.getTimeInMillis()));
                pstmt.setObject(4, allOrNone);
                if (request.getMaxAttempts() == null) {
                    pstmt.setObject(5, new Integer(1));
                } else {
                    pstmt.setObject(5, request.getMaxAttempts());
                }
                pstmt.setObject(6, username);
                pstmt.setObject(7, new Boolean("false"));
                pstmt.setObject(8, delegationEPRString);
                pstmt.setObject(9, userSubject);
                boolean flag = pstmt.execute();

            } else {
                pstmt = c.prepareStatement(requestInsertSQL_OLD);
                pstmt.setObject(1, new Integer(concurrency));
                pstmt.setObject(2, new Long(terminationTime.getTimeInMillis()));
                pstmt.setObject(3, allOrNone);
                pstmt.setObject(4, request.getMaxAttempts());
                pstmt.setObject(5, username);
                pstmt.setObject(6, new Boolean("false"));
                pstmt.setObject(7, delegationEPRString);
                pstmt.setObject(8, userSubject);
                boolean flag = pstmt.execute();
                statement = c.createStatement();
                rs = statement.executeQuery(
                        "select last_value from request_seq");
                while (rs != null && rs.next()) {
                    requestId = rs.getInt(1);
                }
            }
            storeTransfers(requestId, request);
            return requestId;
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbInsertErr"),
                    e);
        } catch(SerializationException se) {
            throw new RftDBException(i18n.getMessage("serializationErr"), se);
        } finally {
            try {
                if (pstmt != null) {
                    pstmt.close();
                }
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * @param request
     * @param terminationTime
     * @param username
     * @param userSubject
     * @return
     * @throws RftDBException
     */
    public synchronized int storeDeleteRequest(DeleteRequestType request,
            Calendar terminationTime, String username, String userSubject) 
    throws RftDBException {
        int requestId = 0;
        int concurrency = 1; //Default
        Integer tempConcurrency = request.getConcurrency();
        if (tempConcurrency != null) {
            if(tempConcurrency.intValue() > 0) {
                concurrency = tempConcurrency.intValue();
            }
        }
        QName qName = new QName("", "DelegatedEPR");
        Connection c = null;
        PreparedStatement pstmt= null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            String delegationEPRString = ObjectSerializer.toString(
            request.getTransferCredentialEndpoint(), qName);
            c = RFTDatabaseSetup.getDBConnection();
            if(!OLD_SCHEMA) {
                pstmt = c.prepareStatement(requestInsertSQL);
                requestId = getNextRequestId();
                pstmt.setObject(1, new Integer(requestId));
                pstmt.setObject(2, new Integer(concurrency));
                pstmt.setObject(3, new Long(terminationTime.getTimeInMillis()));
                pstmt.setObject(4, new Boolean(false));
                pstmt.setObject(5, request.getMaxAttempts());
                pstmt.setObject(6, username);
                pstmt.setObject(7, new Boolean("false"));
                pstmt.setObject(8, delegationEPRString);
                pstmt.setObject(9, userSubject);
                boolean flag = pstmt.execute();
            } else {
                pstmt = c.prepareStatement(requestInsertSQL_OLD);
                pstmt.setObject(1, new Integer(concurrency));
                pstmt.setObject(2, new Long(terminationTime.getTimeInMillis()));
                pstmt.setObject(3, new Boolean(false));
                pstmt.setObject(4, request.getMaxAttempts());
                pstmt.setObject(5, username);
                pstmt.setObject(6, new Boolean("false"));
                pstmt.setObject(7, delegationEPRString);
                pstmt.setObject(8, userSubject);
                boolean flag = pstmt.execute();
                statement = c.createStatement();
                rs = statement.executeQuery(
                        "select last_value from request_seq");
                while (rs != null && rs.next()) {
                    requestId = rs.getInt(1);
                }
            }
            storeDelete(requestId, request);
            return requestId;
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbInsertErr"),
                    e);        
        } catch (SerializationException se) {
            throw new RftDBException(i18n.getMessage("serializationErr"), se);
        } finally {
            try {
                if (pstmt != null) {
                    pstmt.close();
                }
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    public synchronized int getNextRequestId() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        Statement tempStatement = null;
        ResultSet rs = null;
        int nextRequestId = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_UPDATABLE);
            rs = statement.executeQuery("select request_id from requestid");
            if (!rs.next()) {
                // if this is first row in the database
                nextRequestId = nextRequestId + 1;
                tempStatement = c.createStatement();
                tempStatement.executeUpdate("insert into requestid values("
                        + nextRequestId + ")");
                logger.debug("updated row with : " + nextRequestId);
            } else {
                rs.previous(); //Reset the cursor
                while( rs != null && rs.next() ) {
                    nextRequestId = rs.getInt(1);
                    logger.debug("next request id: " + nextRequestId);
                    // Get the req Id and increment it instead of using 
                    // Db Specific sequence
                    nextRequestId = nextRequestId + 1;
                    // Now Update the old value with new value
                    String updateQuery = "update requestid set request_id=";
                    updateQuery = updateQuery + nextRequestId;
                    tempStatement = c.createStatement();
                    tempStatement.executeUpdate(updateQuery);
                    logger.debug("updated row with : " + nextRequestId);
                }
            }
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbInsertErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return nextRequestId;
    }

    private static void setSchemaVersion() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            statement.executeQuery("Select * from requestid");
        } catch(SQLException sql) {
            // Exception here means the table does not exist
            logger.info("Using 4.0 database schema");
            OLD_SCHEMA = true; 
        } finally {
            try {
                if(statement != null) {
                    statement.close();
                } 
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
   /* To be used only in the case of post 4.0 db schema
    */
    public synchronized int getNextTransferId() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        Statement tempStatement = null;
        ResultSet rs = null;
        int nextTransferId = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_UPDATABLE);
            rs = statement.executeQuery("select transfer_id from transferid");
            if (!rs.next())  {
                nextTransferId = nextTransferId + 1;
                StringBuffer query = new StringBuffer();
                query.append("insert into transferid(transfer_id) values (");
                query.append(nextTransferId);
                query.append(")");
                tempStatement = c.createStatement();
                // Now Update the old value with new value
                tempStatement.executeUpdate(query.toString());
            } else {
                rs.previous();
                while( rs != null && rs.next() ) {
                    nextTransferId = rs.getInt(1);
                    // Get the transfer Id and increment it instead of using 
                    // Db Specific sequence
                    nextTransferId = nextTransferId + 1;
                    // Now Update the old value with new value
                    tempStatement = c.createStatement();
                    tempStatement.executeUpdate("update transferid set "+
                            "transfer_id= "+ nextTransferId);
                }
            }
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbInsertErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return nextTransferId;
    }
 
    private void storeDelete(int requestId, DeleteRequestType request)
            throws RftDBException {
        Connection c = null;
        PreparedStatement pstmt = null;
        DeleteOptionsType localDeleteOptions = request.getDeleteOptions();
        DeleteType[] list = request.getDeletion();
        try {
            c = RFTDatabaseSetup.getDBConnection();
            if(OLD_SCHEMA) {
                pstmt = c.prepareStatement(deleteInsertQuery_OLD);
            } else {
                pstmt = c.prepareStatement(deleteInsertQuery);
            }
            for (int i = 0; i < list.length; i++) {
                DeleteOptionsType tempOptions = list[i].getDeleteOptions();
                if(tempOptions!=null) {
                    localDeleteOptions = tempOptions;
                }
                if(localDeleteOptions == null) {
                    localDeleteOptions = new DeleteOptionsType();
                }
                String tempSubjectName = localDeleteOptions.getSubjectName();
                String tempUserName = localDeleteOptions.getUserName();
                if (!OLD_SCHEMA) {
                    int transferId = getNextTransferId();
                    pstmt.setObject(1, new Integer(transferId));
                    pstmt.setObject(2, new Integer(requestId));
                    pstmt.setObject(3, tempSubjectName);
                    pstmt.setObject(4, list[i].getFile());
                    pstmt.setObject(5, new Integer(RFTConstants.STATUS_DIRTY));
                    if (tempUserName == null) {
                        pstmt.setObject(6, "null");
                    } else {
                        pstmt.setObject(6, tempUserName);
                    }
                    pstmt.execute();
                } else {
                    pstmt.setObject(1, new Integer(requestId));
                    pstmt.setObject(2, tempSubjectName);
                    pstmt.setObject(3, list[i].getFile());
                    pstmt.setObject(4, new Integer(RFTConstants.STATUS_DIRTY));
                    if(tempUserName == null) {
                        pstmt.setObject(5, "null");
                    } else {
                        pstmt.setObject(5, tempUserName);
                    }
                    pstmt.execute();
                }
            }
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbInsertErr"),
                    e);
        } finally {
            try {
                if (pstmt!= null) {
                    pstmt.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
    
    /**
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public String getSubject(int requestId) throws RftDBException {
        Connection c = null;
        String subject = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery("SELECT user_subject from " +
            		"request"
                    + " where id=" + requestId);

            while (rs != null && rs.next()) {
                subject = rs.getString(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return subject;
    }

    /**
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public String getDelegationEPR(int requestId) throws RftDBException {
        Connection c = null;
        String delegationURI = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery("SELECT delegated_epr from " +
            		"request"
                    + " where id=" + requestId);

            while (rs != null && rs.next()) {
                delegationURI = rs.getString(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return delegationURI;
    }
    /**
     * Update the termination time of a request
     * 
     * @param requestId
     * @param newTerm
     *            new termination time
     * @throws RftDBException
     */
    public void updateRequest(int requestId, long newTerm)
            throws RftDBException {
        Statement statement = null;
        Connection c = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Update request set termination_time =");
            query.append(newTerm);
            query.append(" where id=");
            query.append(requestId);
            this.queryConstant = statement.executeUpdate(query.toString());
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr") + 
                    e.getMessage(), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * Update the request to started = true after start is called
     * 
     * @param requestId
     * @param started
     * @throws RftDBException
     */
    public void updateStarted(int requestId, boolean started)
            throws RftDBException {
        PreparedStatement pstmt = null;
        Connection c = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            pstmt = c.prepareStatement(updateStartedQuery);
            pstmt.setObject(1, new Boolean(started));
            pstmt.setObject(2, new Integer(requestId));
            pstmt.execute();
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"), e);
        } finally {
            try {
                if (pstmt != null) {
                    pstmt.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public Boolean isStarted(int requestId) throws RftDBException {
        Boolean started = null;
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery("SELECT started from request"
                    + " where id=" + requestId);

            while (rs != null && rs.next()) {
                started = new Boolean( rs.getBoolean(1) );
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr") + e.getMessage() , e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return started;
    }

    /**
     * Gets the activeRequestIds attribute of the ReliableFileTransferDbAdapter
     * object
     * 
     * @return The activeRequestIds value
     * @throws RftDBException
     */
    public List getActiveRequestIds() throws RftDBException {
        Vector temp = new Vector();
        Connection c = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            pstmt = c.prepareStatement(activeRequestIdSQL);
            pstmt.setObject(1, new Long(Calendar.getInstance().
                        getTimeInMillis()));
            pstmt.setObject(2, new Boolean(true));
            rs = pstmt.executeQuery();
            while ((rs != null) && (rs.next())) {
                temp.addElement(new Integer(rs.getInt(1)));
            }
            // Clean up old stuff from the database
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (pstmt != null) {
                    pstmt.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        resetActiveTransfers();
        List list = new ArrayList(temp);
        return list;
    }
    
    private void resetActiveTransfers() throws RftDBException {
        Statement statement = null;
        Connection c = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            //TODO:This query needs to be optimized
            query.append("update transfer set status=4 where (status=3 ");
            query.append(" or status=10)");
            this.queryConstant = statement.executeUpdate(query.toString());
            // Clean up old stuff from the database
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
    
    private TransferType updateDestinationUrl(TransferType transfer) {
        String sourceUrl = transfer.getSourceUrl();
        String fileName = sourceUrl.substring(sourceUrl.lastIndexOf("/") + 1);
        transfer.setDestinationUrl(transfer.getDestinationUrl() + fileName);
        return transfer;
    }

    /** TODO: REFACTOR/SIMPLIFY
     * To store the transfers in the request
     * 
     * @param transferRequest
     *            Transfer request
     * @param requestId
     *            requestId
     * @return id
     * @exception RftDBException
     */
    private int storeTransfers(int requestId,
            TransferRequestType transferRequest) throws RftDBException {
        Connection c = RFTDatabaseSetup.getDBConnection();
        int returnInt = -1;
        PreparedStatement pstmt ;
        try {
            TransferType[] transfers = transferRequest.getTransfer();
            RFTOptionsType localRFTOptions = transferRequest.getRftOptions();
            if (OLD_SCHEMA) {
                pstmt = c.prepareStatement(transferInsertSQL_OLD);
            } else {
                pstmt = c.prepareStatement(transferInsertSQL);
            }
            for (int i = 0; i < transfers.length; i++) {
                TransferType transfer = transfers[i];
                if ((transfer.getDestinationUrl().endsWith("/"))
                        && !(transfer.getSourceUrl().endsWith("/"))) {
                    transfer = updateDestinationUrl(transfer);
                }
                RFTOptionsType rftOptions = transfer.getRftOptions();
                if (rftOptions == null) {
                    if (localRFTOptions != null) {
                        rftOptions = localRFTOptions;
                    } else {
                        rftOptions = globalRFTOptions;
                    }
                }
                String subjectName = rftOptions.getSubjectName();
                if (subjectName != null) {
                    if (!(subjectName.trim().equals(""))) {
                        rftOptions.setSourceSubjectName(subjectName);
                        rftOptions.setDestinationSubjectName(subjectName);
                    }
                }
                if (OLD_SCHEMA) {
                    pstmt.setObject(1, new Integer(requestId));
                    pstmt.setObject(2, transfer.getSourceUrl());
                    pstmt.setObject(3, transfer.getDestinationUrl());
                    pstmt.setObject(4, rftOptions.getUserName());
                    pstmt.setObject(5, rftOptions.getDcau());
                    pstmt.setObject(6, rftOptions.getParallelStreams());
                    pstmt.setObject(7, rftOptions.getTcpBufferSize());
                    pstmt.setObject(8, rftOptions.getBlockSize());
                    pstmt.setObject(9, rftOptions.getNotpt());
                    pstmt.setObject(10, rftOptions.getBinary());
                    pstmt.setObject(11, rftOptions.getSourceSubjectName());
                    pstmt.setObject(12, rftOptions.getDestinationSubjectName());
                    pstmt.setObject(13, rftOptions.getIgnoreFilePermErr());
                } else {
                    int firstTransferId = getNextTransferId();
                    pstmt.setObject(1, new Integer(firstTransferId)); 
                    pstmt.setObject(2, new Integer(requestId));
                    pstmt.setObject(3, transfer.getSourceUrl());
                    pstmt.setObject(4, transfer.getDestinationUrl());
                    if(rftOptions.getUserName() == null) {
                        pstmt.setObject(5, "null");
                    } else {
                        pstmt.setObject(5, rftOptions.getUserName());
                    }
                    if(rftOptions.getDcau() == null) {
                        pstmt.setObject(6, new Boolean(true));
                    } else {
                        pstmt.setObject(6, rftOptions.getDcau());
                    }
                    if(rftOptions.getParallelStreams() == null) {
                        pstmt.setObject(7, new Integer(1));
                    } else {
                        pstmt.setObject(7, rftOptions.getParallelStreams());
                    }
                    if(rftOptions.getTcpBufferSize() == null) {
                        pstmt.setObject(8, new Integer(16384));
                    } else {
                        pstmt.setObject(8, rftOptions.getTcpBufferSize());
                    }
                    if (rftOptions.getBlockSize() == null) {
                        pstmt.setObject(9, new Integer(16384));
                    } else {
                        pstmt.setObject(9, rftOptions.getBlockSize());
                    }
                    if (rftOptions.getNotpt() == null) {
                        pstmt.setObject(10, new Boolean(false));
                    } else {
                        pstmt.setObject(10, rftOptions.getNotpt());
                    }
                    if (rftOptions.getBinary() == null) {
                        pstmt.setObject(11, new Boolean(true));
                    } else {
                        pstmt.setObject(11, rftOptions.getBinary());
                    }
                    if (rftOptions.getSourceSubjectName() == null) {
                        pstmt.setObject(12, "null");
                    } else {
                        pstmt.setObject(12, rftOptions.getSourceSubjectName());
                    }
                    if (rftOptions.getDestinationSubjectName() == null) {
                        pstmt.setObject(13, "null");
                    } else {
                        pstmt.setObject(13, rftOptions.getDestinationSubjectName());
                    }
                    if (rftOptions.getIgnoreFilePermErr() == null) {
                        pstmt.setObject(14, new Boolean(false));
                    } else {
                        pstmt.setObject(14, rftOptions.getIgnoreFilePermErr());
                    }
                }
                boolean flag = pstmt.execute();
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbInsertErr") + 
                    e.getMessage(), e);
        } 
        try {
            if (pstmt != null) {
                pstmt.close();
            }
            if (c != null) {
                RFTDatabaseSetup.returnDBConnection(c);
            }
        } catch (SQLException sql) {
            logger.warn(i18n.getMessage("dbStatementErr"), sql);
        }
        return returnInt;
    }

    private TransferType constructTransferType(ResultSet rs) 
        throws SQLException{
            TransferType transfer = new TransferType();
            transfer.setTransferId(new Integer(rs.getInt(1)));
            //TransferID
            transfer.setSourceUrl(rs.getString(3));
            transfer.setDestinationUrl(rs.getString(4));

            transfer.setStatus(RFTUtil.mapStatus(rs.getInt(5)));
            transfer.setAttempts(new Integer(rs.getInt(6)));
            RFTOptionsType rftOptions = new RFTOptionsType();
            rftOptions.setUserName(rs.getString(7));
            rftOptions.setDcau(new Boolean(rs.getBoolean(8)));
            rftOptions.setParallelStreams(new Integer(rs.getInt(9)));
            rftOptions.setTcpBufferSize(new Integer(rs.getInt(10)));
            rftOptions.setBlockSize(new Integer(rs.getInt(11)));
            rftOptions.setNotpt(new Boolean(rs.getBoolean(12)));
            boolean binary = rs.getBoolean(13);
            if (rs.wasNull()) {
                binary = true;
            }
            rftOptions.setBinary(new Boolean(binary));
            rftOptions.setSourceSubjectName(rs.getString(14));
            rftOptions.setDestinationSubjectName(rs.getString(15));
            rftOptions.setIgnoreFilePermErr(new Boolean(rs.getBoolean(16)));
            transfer.setRftOptions(rftOptions);
            return transfer;
        }


    /**
     * Gets the initialTransfers attribute of the ReliableFileTransferDbAdapter
     * object
     * 
     * @param requestId
     *            requestID
     * @param concurrency
     *            concurrency
     * @return The initialTransfers value
     * @exception RftDBException
     */
    public Vector getInitialTransfers(int requestId, int concurrency)
        throws RftDBException {
            Vector initialList = new Vector();
            Connection c = null;
            Statement statement = null;
            ResultSet rs = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                int activeTransfers = getNumberActive();
                int numberAllowed = RFTDatabaseSetup.maxActive - activeTransfers;
                if(logger.isDebugEnabled()) {
                    logger.debug("Number of active transfers : " + activeTransfers);
                    logger.debug("Number allowed : " + RFTDatabaseSetup.maxActive);
                    logger.debug("concurrency " + concurrency + " number allowed "
                            + numberAllowed);
                }
                int rowLimit = 0;
                //We have calculate number of rows to be returned 
                // so we won't exceed max active allowed
                if (concurrency > numberAllowed) {
                    rowLimit = numberAllowed;
                } else {
                    rowLimit = concurrency;
                }
                StringBuffer query = new StringBuffer();
                query.append("SELECT * FROM transfer where request_id=");
                query.append(requestId);
                query.append(" and status=4 ");
                query.append(" order by id ");
                statement.setMaxRows(rowLimit);
                rs = statement.executeQuery(query.toString());
                while (rs != null && rs.next()) {
                    initialList.add(constructTransferType(rs));
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr",
                            ""+ requestId), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return initialList;
    }

    /**
     * @param requestId
     * @param concurrency
     * @return
     * @throws RftDBException
     */
    public Vector getInitialDeleteTransfers(
            int requestId, int concurrency)
            throws RftDBException {
        Vector initialList = new Vector();
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("SELECT id,source_url,status,attempts,user_name," +
                    "source_subject");
            query.append(" FROM transfer where request_id=");
            query.append(requestId);
            query.append(" order by id");
            query.append(" limit " + concurrency);

            rs = statement.executeQuery(query.toString());
            while (rs != null && rs.next()) {
                TransferType transfer = new TransferType();
                transfer.setTransferId(new Integer(rs.getInt(1)));
                //TransferID
                transfer.setSourceUrl(rs.getString(2));

                transfer.setStatus(RFTUtil.mapStatus(rs.getInt(3)));
                transfer.setAttempts(new Integer(rs.getInt(4)));
                RFTOptionsType rftOptions = new RFTOptionsType();
                rftOptions.setUserName(rs.getString(5));
                rftOptions.setSourceSubjectName(rs.getString(6));
                transfer.setRftOptions(rftOptions);
                initialList.add(transfer);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr",
                            ""+ requestId), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return initialList;
    }

    /**
     * Gets the activeTransfers attribute of the ReliableFileTransferDbAdapter
     * object
     * 
     * @param requestId
     * @return The activeTransfers value
     * @exception RftDBException
     */
    public Collection getActiveTransfers(int requestId) throws RftDBException {
        ArrayList initialList = new ArrayList();
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            int concurrency = getConcurrency(requestId);
            StringBuffer query = new StringBuffer();
            query.append("SELECT * FROM transfer where request_id=");
            query.append(requestId);
            query.append(" AND (status=3 or status=4 or ");
            query.append(" status=1 or status=8 or status=6) ");
            query.append(" order by id ");

            statement.setMaxRows(concurrency);
            rs = statement.executeQuery(query.toString());
            while (rs != null && rs.next()) {
                TransferType transfer = new TransferType();
                transfer.setTransferId(new Integer(rs.getInt(1)));
                //TransferID
                transfer.setSourceUrl(rs.getString(3));
                String tempDest = rs.getString(4);
                if ((tempDest == null) || (tempDest.equals("null"))) {
                    // If it is a  delete 
                    transfer.setStatus(RFTUtil.mapStatus(8));
                } else {
                    transfer.setDestinationUrl(tempDest);
                    int tempStatus = rs.getInt(5);
                    transfer.setStatus(RFTUtil.mapStatus(tempStatus));
                    if (tempStatus == RFTConstants.STATUS_EXPANDING)  {
                        // Cleanup transfer that are already in db
                        recoverDirectoryTransfers(
                        requestId,transfer.getSourceUrl(), 
                        transfer.getDestinationUrl());
                    }
                }
                transfer.setAttempts(new Integer(rs.getInt(6)));
                
                RFTOptionsType rftOptions = new RFTOptionsType();
                rftOptions.setUserName(rs.getString(7));
                rftOptions.setDcau(new Boolean(rs.getBoolean(8)));
                rftOptions.setParallelStreams(new Integer(rs.getInt(9)));
                rftOptions.setTcpBufferSize(new Integer(rs.getInt(10)));
                rftOptions.setBlockSize(new Integer(rs.getInt(11)));
                rftOptions.setNotpt(new Boolean(rs.getBoolean(12)));
                rftOptions.setBinary(new Boolean(rs.getBoolean(13)));
                rftOptions.setSourceSubjectName(rs.getString(14));
                rftOptions.setDestinationSubjectName(rs.getString(15));
                rftOptions.setIgnoreFilePermErr(new Boolean(rs.getBoolean(16)));
                transfer.setRftOptions(rftOptions);

                initialList.add(transfer);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr",
                            ""+ requestId), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return initialList;
    }
    
    public RFTRecoveryData getRecoveryData(int requestId) 
        throws RftDBException {
            RFTRecoveryData recoveryData = getSecurityRecoveryData(requestId);
            recoveryData.setRecoveredTransfers(getActiveTransfers(requestId));
            recoveryData.setOverallStatus(getOverallStatus(requestId));
            return recoveryData;
        }
        
    /** Gets the recovery data from database
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public RFTRecoveryData getSecurityRecoveryData(int requestId) 
        throws RftDBException {
            RFTRecoveryData recoveryData = new RFTRecoveryData();
            Connection c = null;
            Statement statement = null;
            ResultSet rs = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                StringBuffer query = new StringBuffer();
                query.append("Select username,user_subject from request ");
                query.append(" where id=").append(requestId);
                rs = statement.executeQuery(query.toString());
                while(rs !=null && rs.next() ) {
                    recoveryData.setUserName(rs.getString(1));
                    recoveryData.setUserSubject(rs.getString(2));
                }
            } catch(SQLException e) {
                throw new RftDBException(i18n.getMessage("dbRetrieveErr")
                        + e.getMessage(),
                        e);
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                    if (rs != null) {
                        rs.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }
            }
            return recoveryData;
        }

    public synchronized OverallStatus getOverallStatus(int requestId) 
        throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        int status = -1;
        OverallStatus overallStatus = null;
        int finished = 0;
        int active = 0;
        int pending = 0;
        int retrying = 0;
        int cancelled =0;
        int failed = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select status from transfer  ");
            query.append(" where request_id=").append(requestId);
            rs = statement.executeQuery(query.toString());
            while (rs != null && rs.next()) {
                status = rs.getInt(1);
                if (status == 0) {
                    finished++;
                } else if (status == 1) {
                    retrying++;
                } else if (status == 2) {
                    failed++;
                } else if (status == 3) {
                    active++;
                } else if (status == 4 || status == 10 || status == 8) {
                    pending++;
                } else if (status == 5) {
                    cancelled++;
                } else if (status == 6) {
                    active++;
                } else if (status == 7) {
                    finished++;
                }
            }
            overallStatus = new OverallStatus();
            overallStatus.setTransfersFinished(finished);
            overallStatus.setTransfersRestarted(retrying);
            overallStatus.setTransfersFailed(failed);
            overallStatus.setTransfersActive(active);
            overallStatus.setTransfersPending(pending);
            overallStatus.setTransfersCancelled(cancelled);
            if(logger.isDebugEnabled()) {
                logger.debug("OverallStatus for requestId: " + requestId
                        + " Finished:" +
                        finished + " Retrying:" + retrying +
                        " Failed:"+ failed + " Active:" + active +
                        " Pending:" + pending + " Cancelled:" + cancelled);
            }
            } catch(SQLException e) {
            throw new RftDBException(i18n.getMessage("dbRetrieveErr")
                    + e.getMessage(),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return overallStatus;
    }

    private void recoverDirectoryTransfers(int requestId, 
            String sourceUrl, String destinationUrl ) throws RftDBException {
        logger.debug("Recovering directory transfers");
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("DELETE from transfer where request_id=");
            query.append(requestId);
            query.append(" and source_url like '");
            query.append(sourceUrl);
            query.append("%' and dest_url like '");
            query.append(destinationUrl);
            query.append("%'");
            query.append(" and status = 4");
            if (logger.isDebugEnabled()) {
                logger.debug("query to remove dir entries:"+ query.toString());
            }
            statement.executeUpdate(query.toString());

        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    public int getNumberActive(int requestId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        int activeTransfers = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select count(*) from transfer where ");
            query.append(" request_id= ");
            query.append(requestId);
            query.append(" and (status=3 or status=6)");
            rs = statement.executeQuery(query.toString());
            while( rs != null && rs.next()) {
                int status = rs.getInt(1);
                activeTransfers = status;
                logger.debug("status from db: " + status);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return activeTransfers;
    }

    /** Get total number of transfers active in the container
     * @return
     * @throws RftDBException
     */
    public int getNumberActive() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        int activeTransfers = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select count(*) from transfer where ");
            query.append(" status=3");
            rs = statement.executeQuery(query.toString());
            while( rs != null && rs.next()) {
                activeTransfers = rs.getInt(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return activeTransfers;
    }
    /**
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public int getConcurrency(int requestId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        int concurrency = -1;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select concurrency from request ");
            query.append(" where id=");
            query.append(requestId);
            rs = statement.executeQuery(query.toString());
            while (rs.next() && rs != null) {
                concurrency = rs.getInt(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr",
                            ""+ requestId), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return concurrency;
    }

    /**
     * Gets the restartMarker attribute of the ReliableFileTransferDbAdapter
     * object
     * 
     * @param transferJobID
     * @return The restartMarker value
     * @exception RftDBException
     */
    public String getRestartMarker(int transferJobID) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        String marker = null;
        ResultSet rs = null;

        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery("Select marker from restart "
                    + "where transfer_id=" + transferJobID);

            while (rs.next() && rs != null) {
                marker = rs.getString(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr", 
                            ""+ transferJobID), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return marker;
    }

    /**
     * Sets the restartMarker attribute of the ReliableFileTransferDbAdapter
     * object
     * 
     * @param transferJobID
     *            The new restartMarker value
     * @param marker
     *            The new restartMarker value
     * @exception RftDBException
     */
    public void setRestartMarker(int transferJobID, String marker)
            throws RftDBException {
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            int update = statement.executeUpdate("UPDATE restart set "
                    + " marker='" + marker + "' where transfer_id="
                    + transferJobID);
            if (update == 0) {
                this.queryConstant = statement.executeUpdate("INSERT into "
                        + "restart(transfer_id,marker) " + "VALUES("
                        + transferJobID + ",'" + marker + "')");
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"," "+transferJobID), 
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * Updates the status of transferType
     * 
     * @param requestId
     * 
     * @param transferJob
     *            TransferJob
     * @exception RftDBException
     */
    public synchronized void update(int requestId, TransferType transferJob)
            throws RftDBException {
        Connection c = null;
        PreparedStatement updatePS = null;
        int status = RFTUtil.reverseMap(transferJob.getStatus());
        logger.debug("Updating status in database for id: " 
            + transferJob.getTransferId().intValue() + " with status: " 
            + status);
        try {
            c = RFTDatabaseSetup.getDBConnection();
            updatePS = c.prepareStatement(
                    "update transfer set status=?,attempts=?," +
                    "dest_url=?,source_url=? where id=?");
            updatePS.setInt(1, status);
            updatePS.setInt(2, transferJob.getAttempts().intValue());
            updatePS.setString(3, transferJob.getDestinationUrl());
            updatePS.setString(4, transferJob.getSourceUrl());
            updatePS.setInt(5, transferJob.getTransferId().intValue());
            updatePS.executeUpdate();
            logger.debug("Updated status in database for id: " 
            + transferJob.getTransferId().intValue() + " with status: " 
            + status);
        } catch (SQLException e) {
            throw new RftDBException(i18n.getMessage("dbUpdateErr")
                    + e.getMessage(), e);
        } finally {
            try {
                if (updatePS!= null) {
                    updatePS.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    } 

    public void storeDelegationListenerId(int requestId, String delegationId) 
        throws RftDBException {
            Connection c = null;
            Statement statement = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                StringBuffer insertQuery = new StringBuffer();
                insertQuery.append("Update request set proxy_loc='");
                insertQuery.append(delegationId);
                insertQuery.append("' where id=").append(requestId);
                statement.executeUpdate(insertQuery.toString());
            } catch (SQLException e) {
                logger.error(i18n.getMessage("dbUpdateErr"), e);
                throw new RftDBException(i18n.getMessage(
                            "dbUpdateErr"), e);
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }
            }
        }
    
    public String getDelegationListenerId(int requestId) throws RftDBException {
       Connection c = null;
       Statement statement = null;
       ResultSet rs = null;
       String delegationListenerId = null;
       try {
           c = RFTDatabaseSetup.getDBConnection();
           statement = c.createStatement();
           StringBuffer query = new StringBuffer();
           query.append("Select proxy_loc from request where id=");
           query.append(requestId);
           rs = statement.executeQuery(query.toString());
           while( rs !=null && rs.next()) {
               delegationListenerId = rs.getString(1);
           }
       }  catch (SQLException e) {
           logger.error(i18n.getMessage("dbUpdateErr"), e);
           throw new RftDBException(i18n.getMessage(
                       "dbUpdateErr"), e);
       } finally {
           try {
               if (statement != null) {
                   statement.close();
               }
               if (rs != null) {
                   rs.close();
               }
               if (c != null) {
                   RFTDatabaseSetup.returnDBConnection(c);
               }
           } catch (SQLException sql) {
               logger.warn(i18n.getMessage("dbStatementErr"), sql);
           }
       }
       return delegationListenerId;
    }

    private void updateFactoryRPs(long size) throws RftDBException {
        Connection c = null;
        Statement statement = null; 
        ResultSet rs = null;
        long totalTransfers = 0;
        long totalBytes = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select total_transfers,total_bytes ");
            query.append("from factory");
            rs = statement.executeQuery(query.toString());
            while( rs != null && rs.next()) {
                totalTransfers = rs.getLong(1);
                totalBytes = rs.getLong(2);
            }
            StringBuffer query2 = new StringBuffer();
            if (totalTransfers == 0 && totalBytes == 0) {
                totalTransfers = 1;
                totalBytes = size;
                query2.append("insert into factory values(");
                query2.append(totalTransfers);
                query2.append(",");
                query2.append(totalBytes);
                query2.append(")");
            } else {
                totalTransfers = totalTransfers + 1;
                totalBytes = size + totalBytes;
                query2.append("update factory set total_transfers=");
                query2.append(totalTransfers);
                query2.append(",total_bytes=");
                query2.append(totalBytes);
            }
                
            int queryConstant = statement.executeUpdate(query2.toString());
            
        } catch (SQLException e) {
            logger.error(i18n.getMessage("dbUpdateErr"), e);
            throw new RftDBException(i18n.getMessage(
                    "dbUpdateErr"), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
        
    /**
     * Check to see if all transfers in the resource 
     * are done
     * @param requestId
     * @return true - if all transfers in this req are done
     * @throws RftDBException
     */
    public boolean checkAllDone(int requestId) throws RftDBException {
        boolean allDone = false;
        Connection c  = null;
        Statement statement = null;
        ResultSet rs = null;
        int count = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("select count(*) from ");
            query.append(" transfer where status != 0 and status !=7 ");
            query.append(" and status !=2  and request_id=");
            query.append(requestId);
            rs = statement.executeQuery(query.toString());
            while(rs != null && rs.next()) {
                count = rs.getInt(1);
            }
            if (count == 0 ) {
                allDone = true;
            }
        } catch (SQLException e) {
            logger.error(i18n.getMessage("dbRetrieveErr"), e);
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return allDone;
    }
    /**
     * @return minimumPollTime
     * @throws RftDBException
     */
    public long getMinimumPollTime() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        long retryTime = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("select min(retry_time) from ");
            query.append(" transfer where retry_time >= ")
            .append(Calendar.getInstance().getTimeInMillis());
            query.append(" and status=1");
             rs = statement.executeQuery(query.toString());
            while(rs != null && rs.next()) {
                retryTime = rs.getLong(1);
            }
        } catch (SQLException e) {
            logger.error(i18n.getMessage("dbRetrieveErr"), e);
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"), e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return retryTime;
    }
    /**
     * Used to store the transferType from URLExpander.java
     * 
     * @param requestId
     * @param transferJob 
     * @exception RftDBException
     */
    public void storeTransferJob(int requestId, TransferType transferJob)
        throws RftDBException {
        Connection c = null;
        PreparedStatement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            RFTOptionsType localOptions = transferJob.getRftOptions();
            if (localOptions == null) {
                localOptions = globalRFTOptions;
            }
            if (!OLD_SCHEMA) {
                statement = c.prepareStatement(singleTransferInsertQuery);
                int transferId = getNextTransferId();
                statement.setObject(1, new Integer(transferId));
                statement.setObject(2, new Integer(requestId));
                statement.setObject(3, transferJob.getSourceUrl());
                statement.setObject(4, transferJob.getDestinationUrl());
                statement.setObject(5, 
                        new Integer(RFTUtil.reverseMap(
                                transferJob.getStatus())));
                statement.setObject(6, localOptions.getUserName());
                statement.setObject(7, localOptions.getDcau());
                statement.setObject(8, localOptions.getParallelStreams());
                statement.setObject(9, localOptions.getTcpBufferSize());
                statement.setObject(10, localOptions.getBlockSize());
                statement.setObject(11, localOptions.getNotpt());
                statement.setObject(12, localOptions.getBinary());
                statement.setObject(13, 
                        localOptions.getSourceSubjectName());
                statement.setObject(14, 
                        localOptions.getDestinationSubjectName());
            } else {
                statement = c.prepareStatement(singleTransferInsertQuery_OLD);
                statement.setObject(1, new Integer(requestId));
                statement.setObject(2, transferJob.getSourceUrl());
                statement.setObject(3, transferJob.getDestinationUrl());
                statement.setObject(4, 
                        new Integer(RFTUtil.reverseMap(
                                transferJob.getStatus())));
                statement.setObject(5, localOptions.getUserName());
                statement.setObject(6, localOptions.getDcau());
                statement.setObject(7, localOptions.getParallelStreams());
                statement.setObject(8, localOptions.getTcpBufferSize());
                statement.setObject(9, localOptions.getBlockSize());
                statement.setObject(10, localOptions.getNotpt());
                statement.setObject(11, localOptions.getBinary());
                statement.setObject(12, 
                        localOptions.getSourceSubjectName());
                statement.setObject(13, 
                        localOptions.getDestinationSubjectName()); 
            }
            statement.execute();
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbInsertErr"), 
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * @return totalNumberOfTransfers
     * @throws RftDBException
     */
    public long getTotalNumberOfTransfers() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        long totalNumberOfTransfers = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery(
                    "select total_transfers from factory");
            while (rs != null && rs.next()) {
                totalNumberOfTransfers = rs.getLong(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return totalNumberOfTransfers;
    }
    /**
     * @return totalNumberOfBytes
     * @throws RftDBException
     */
    public long getTotalNumberOfBytes() throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        long totalNumberOfBytes = 0;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery(
                    "select total_bytes from factory");
            while (rs != null && rs.next()) {
                totalNumberOfBytes = rs.getLong(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return totalNumberOfBytes;
    }
    
    /**
     * Stores the size of the file and permissions
     * @param transferId
     * @param size
     * @param permissions
     * @throws RftDBException
     */
    public void updateTransfer(int transferId, long size, 
            String permissions) throws RftDBException {

        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            StringBuffer query = new StringBuffer();
            statement = c.createStatement();
            query.append("update transfer set permissions='");
            query.append(permissions);
            query.append("',size=");
            query.append(size);
            query.append(" where id=");
            query.append(transferId);
            this.queryConstant = statement
                .executeUpdate(query.toString());
            updateFactoryRPs(size);
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * Gets the username for a resource
     * 
     * @param transferId
     * @return userName
     * @throws RftDBException
     */
    public String getUserName(int transferId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        String userName = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            rs = statement.executeQuery("SELECT username from request,transfer"
                    + " where request.id = transfer.request_id and "
                    + " transfer.id=" + transferId);

            while (rs != null && rs.next()) {
                userName = rs.getString(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return userName;
    }

    /**
     * update the retry_time of a transfer
     * 
     * @param hostName
     * @throws RftDBException
     */
    public void backOff(String hostName) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        logger.debug("backing off");
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Update transfer set retry_time= ");
            query.append(Calendar.getInstance().getTimeInMillis() 
                    + 2 * RFTDatabaseSetup.backOff);
            query.append(" where (source_url like '%");
            query.append(hostName);
            query.append("%' or dest_url like '%");
            query.append(hostName);
            query.append("%') and (status=1 or status=4 or status=3");
            query.append(" or status=10)");
            this.queryConstant = statement.executeUpdate(
                    query.toString());
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /** Individual transfer backoff
     * @param transferId
     * @param attempts
     * @throws RftDBException
     */
    public void backOff(int transferId, int attempts) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Update transfer set retry_time= ");
            query.append(Calendar.getInstance().getTimeInMillis() 
                    + Math.pow(2,attempts) * RFTDatabaseSetup.backOff);
            query.append(" where id=").append(transferId);
            this.queryConstant = statement.executeUpdate(query.toString());
        } catch(SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);    
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }
    /**
     * @param requestId
     * @param sourceUrl
     * @return
     * @throws RftDBException
     * @throws RftDBException
     */
    public TransferStatusType getStatus(int requestId, String sourceUrl)
        throws RftDBException {
            Connection c = null;
            Statement statement = null;
            TransferStatusType statusType = null;
            ResultSet rs = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                StringBuffer query = new StringBuffer();
                query.append("Select source_url,dest_url,attempts,status ");
                query.append(" from transfer");
                query.append(" where request_id=");
                query.append(requestId);
                query.append(" and source_url='");
                query.append(sourceUrl);
                query.append("' ");
                statement.setMaxRows(1);
                rs = statement.executeQuery(query.toString());
                while (rs.next() && rs != null) {
                    statusType = new TransferStatusType();
                    statusType.setSourceUrl(rs.getString(1));
                    statusType.setDestinationUrl(rs.getString(2));
                    statusType.setAttempts(rs.getInt(3));
                    statusType.setStatus(RFTUtil.mapStatus(rs.getInt(4)));
                }
            } catch (SQLException e) {
                throw new RftDBException(
                        i18n.getMessage("dbRetrieveErr"),
                        e);
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                    if (rs != null) {
                        rs.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }

            }
            return statusType;
        }

    /**
     * @param requestId
     * @param initial
     * @param offset
     * @return set of transfers
     * @throws RftDBException
     */
    public Vector getStatusSet(int requestId, int initial, int offset)
        throws RftDBException {
            Connection c = null;
            Statement statement = null;
            Vector statusTypesVector = new Vector();
            TransferStatusType statusType = null;
            ResultSet rs = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                StringBuffer query = new StringBuffer();
                query.append("Select source_url,dest_url,attempts,status ");
                query.append(" from transfer");
                query.append(" where request_id=");
                query.append(requestId);
                query.append(" and id >= ");
                //FIXME: This needs to be fixed
                query.append(this.getFirstTransferId(requestId) + initial - 1);
                query.append(" order by id");
                statement.setMaxRows(offset);
                rs = statement.executeQuery(query.toString());
                while (rs.next() && rs != null) {
                    statusType = new TransferStatusType();
                    statusType.setSourceUrl(rs.getString(1));
                    statusType.setDestinationUrl(rs.getString(2));
                    statusType.setAttempts(rs.getInt(3));
                    statusType.setStatus(RFTUtil.mapStatus(rs.getInt(4)));
                    statusTypesVector.add(statusType);
                }
            } catch (SQLException e) {
                throw new RftDBException(
                        i18n.getMessage("dbRetrieveErr"),
                        e);
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                    if (rs != null) {
                        rs.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }
            }
            return statusTypesVector;
        }

    /**
     * This returns the first transferId for a request I need this since the
     * transferid that come with request needs to be mapped to id in transfer
     * table
     * 
     * @param requestId
     * @return id of first transfer for this requestId
     * @throws RftDBException
     */
    public int getFirstTransferId(int requestId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        int transferId = -1;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select min(id) from transfer");
            query.append(" where request_id=");
            query.append(requestId);
            statement.setMaxRows(1);
            rs = statement.executeQuery(query.toString());
            while (rs.next() && rs != null) {
                transferId = rs.getInt(1);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return transferId;
    }

    /**
     * @param requestId
     * @return return true if allOrNone is true
     * @throws RftDBException
     */
    public boolean getAllOrNone(int requestId) throws RftDBException {
        boolean allOrNone = false;
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("select all_or_none from request where id=");
            query.append(requestId);
            rs = statement.executeQuery(query.toString());
            while (rs.next() && rs != null) {
                allOrNone = rs.getBoolean(1);
            }
        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return allOrNone;
    }

    /**
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public Vector getFinishedTransfers(int requestId) throws RftDBException {
        Vector transferTypes = null;
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Select * from transfer where request_id=");
            query.append(requestId);
            query.append(" and status=0 order by id");
            rs = statement.executeQuery(query.toString());
            transferTypes = new Vector();
            while (rs.next() && rs != null) {
                TransferType transfer = new TransferType();
                transfer.setTransferId(new Integer(rs.getInt(1)));
                //TransferID
                transfer.setSourceUrl(rs.getString(3));
                transfer.setDestinationUrl(rs.getString(4));

                transfer.setStatus(RFTUtil.mapStatus(rs.getInt(5)));
                transfer.setAttempts(new Integer(rs.getInt(6)));
                RFTOptionsType rftOptions = new RFTOptionsType();
                rftOptions.setDcau(new Boolean(rs.getBoolean(7)));
                rftOptions.setParallelStreams(new Integer(rs.getInt(8)));
                rftOptions.setTcpBufferSize(new Integer(rs.getInt(9)));
                rftOptions.setBlockSize(new Integer(rs.getInt(10)));
                rftOptions.setNotpt(new Boolean(rs.getBoolean(11)));
                rftOptions.setBinary(new Boolean(rs.getBoolean(12)));
                rftOptions.setSourceSubjectName(rs.getString(13));
                rftOptions.setDestinationSubjectName(rs.getString(14));
                rftOptions.setIgnoreFilePermErr(new Boolean(rs.getBoolean(16)));
                transfer.setRftOptions(rftOptions);
                transferTypes.add(transfer);
            }
        } catch (SQLException e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return transferTypes;
    }

    /**
     * @param requestId
     * @throws RftDBException
     */
    public void updateToDirty(int requestId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("Update transfer" +
                    " set status=8 where request_id=");
            query.append(requestId);
            query.append(" and status != 2 ");
            this.queryConstant = statement.executeUpdate(query.toString());
        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }

    /**
     * Updates transfers that were still not started to Cancelled
     * 
     * @param requestId
     * @param fromId
     * @param toId
     * @throws RftDBException
     */
    public void updateToCancelled(int requestId, int fromId, int toId)
        throws RftDBException {
            Connection c = null;
            Statement statement = null;
            try {
                c = RFTDatabaseSetup.getDBConnection();
                statement = c.createStatement();
                StringBuffer query = new StringBuffer();
                query.append("Update transfer set" +
                        " status=5 where request_id=");
                query.append(requestId);
                if (fromId == -1 || toId == -1) {
                    // this means the cancelling the entire request
                    if (logger.isDebugEnabled()) {
                        logger.debug("Cancelling the entire request " + requestId);
                    }
                } else {
                    query.append(" and id>=");
                    query.append(fromId);
                    query.append(" and id<=");
                    query.append(toId);
                }
                this.queryConstant = statement.executeUpdate(query.toString());
            } catch (SQLException e) {
                throw new RftDBException(
                        i18n.getMessage("dbUpdateErr"),
                        e);
            } finally {
                try {
                    if (statement != null) {
                        statement.close();
                    }
                    if (c != null) {
                        RFTDatabaseSetup.returnDBConnection(c);
                    }
                } catch (SQLException sql) {
                    logger.warn(i18n.getMessage("dbStatementErr"), sql);
                }
            }
        }


    /**
     * Deletes all the transfers that correspond to a single resource.
     * 
     * @param requestId
     * @throws RftDBException
     */
    public void deleteAll(int requestId) throws RftDBException {
        Connection c = null;
        Statement statement = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("DELETE from transfer  where request_id=");
            query.append(requestId);
            this.queryConstant = statement.executeUpdate(query.toString());
            StringBuffer query2 = new StringBuffer();
            query2.append("DELETE from request where id=");
            query2.append(requestId);
            this.queryConstant = statement.executeUpdate(query2.toString());
        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbUpdateErr"),
                    e);
        } finally {
            try {
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
                if (statement != null) {
                    statement.close();
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
    }


    /**
     * @param requestId
     * @throws RftDBException
     * @return maxAttempts
     */
    public int getMaxAttempts(int requestId) throws RftDBException {
        int maxAttempts = 0;
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("select maxattempts from request where id=");
            query.append(requestId);
            rs = statement.executeQuery(query.toString());
            while (rs.next() && rs != null) {
                maxAttempts = rs.getInt(1);
            }
        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        return maxAttempts;
    }

    /** This is used to set termination time for recovered instances
     * @param requestId
     * @return
     * @throws RftDBException
     */
    public Calendar getTerminationTime(int requestId) throws RftDBException {
        long terminationTime = 0;
        Connection c = null;
        Statement statement = null;
        ResultSet rs = null;
        try {
            c = RFTDatabaseSetup.getDBConnection();
            statement = c.createStatement();
            StringBuffer query = new StringBuffer();
            query.append("select termination_time from " +
                    "request where id=");
            query.append(requestId);
            rs = statement.executeQuery(query.toString());
            while (rs.next() && rs != null) {
                terminationTime = rs.getLong(1);
            }
        } catch (Exception e) {
            throw new RftDBException(
                    i18n.getMessage("dbRetrieveErr"),
                    e);
        } finally {
            try {
                if (statement != null) {
                    statement.close();
                }
                if (rs != null) {
                    rs.close();
                }
                if (c != null) {
                    RFTDatabaseSetup.returnDBConnection(c);
                }
            } catch (SQLException sql) {
                logger.warn(i18n.getMessage("dbStatementErr"), sql);
            }
        }
        Calendar returnThis = Calendar.getInstance();
        returnThis.setTimeInMillis(terminationTime);
        return returnThis;
    }
}
