/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.transfer.reliable.service.factory;

import java.rmi.RemoteException;
import java.util.Calendar;

import javax.naming.Context;
import javax.naming.InitialContext;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.mds.usefulrp.ServiceMetaDataResourceProperty;
import org.globus.mds.servicegroup.client.ServiceGroupRegistrationParameters;
import org.globus.rft.generated.BaseRequestType;
import org.globus.rft.generated.CreateReliableFileTransferInputType;
import org.globus.rft.generated.CreateReliableFileTransferOutputType;
import org.globus.rft.generated.DelegationEPRMissingFaultType;
import org.globus.rft.generated.DeleteRequestType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.transfer.reliable.service.ReliableFileTransferHome;
import org.globus.util.I18n;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceProperties;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.Constants;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.servicegroup.client.MDSConfiguration;
import org.globus.wsrf.impl.servicegroup.client.ServiceGroupRegistrationClient;
import org.globus.wsrf.utils.AddressingUtils;
import org.oasis.wsrf.faults.BaseFaultType;
import org.oasis.wsrf.lifetime.UnableToSetTerminationTimeFaultType;

import org.globus.wsrf.security.SecurityManager;

/**
 * RFT Factory Service
 */
public class ReliableFileTransferFactoryService 
    implements Resource, ResourceProperties {

    private static Log logger = LogFactory
            .getLog(ReliableFileTransferFactoryService.class.getName());
    private static I18n i18n = 
        I18n.getI18n("org.globus.transfer.reliable.errors",
                ReliableFileTransferFactoryService.class.
                getClassLoader());
    
    private ResourcePropertySet resourcePropertySet;

    private TotalNumberOfTransfersRP totalTransfersRP;

    private ActiveResourceInstancesRP activeResourcesRP;
    
    private DelegationServiceEndpoint delegationEndpoint;
    
    private RFTFactoryStartTime rftFactoryStartTime;
    
    private TotalNumberOfActiveTransfers totalNumberOfActiveTransfers;
    
    private TotalNumberOfBytesTransferred totalNumberOfBytesTransferred;
    
    private ServiceMetaDataResourceProperty metadata;
    
    /**
     * Initialize the factory along with RPs of the factory, and register
     * to the local default index service.
     */
    public ReliableFileTransferFactoryService() {
        this.resourcePropertySet = new SimpleResourcePropertySet(
                RFTConstants.RFT_RP_SET);
        try {
            this.totalTransfersRP = new TotalNumberOfTransfersRP(
                    RFTConstants.TOTAL_TRANSFER_COUNT_FACTORY);
            this.activeResourcesRP = new ActiveResourceInstancesRP(
                    RFTConstants.ACTIVE_RESOURCE_COUNT_FACTORY);
            this.delegationEndpoint =
                DelegationServiceEndpoint.getInstance();
            this.rftFactoryStartTime = new RFTFactoryStartTime(
                    RFTConstants.RFT_FACTORY_START_TIME);
            this.totalNumberOfActiveTransfers = 
                new TotalNumberOfActiveTransfers
                            (RFTConstants.TOTAL_ACTIVE_TRANSFERS_FACTORY);
            this.totalNumberOfBytesTransferred = 
                new TotalNumberOfBytesTransferred(
                        RFTConstants.TOTAL_BYTES_FACTORY);
                    
            this.resourcePropertySet.add(this.totalTransfersRP);
            this.resourcePropertySet.add(this.activeResourcesRP);
            this.resourcePropertySet.add(this.delegationEndpoint);
            this.resourcePropertySet.add(this.rftFactoryStartTime);
            this.resourcePropertySet.add(this.totalNumberOfActiveTransfers);
            this.resourcePropertySet.add(this.totalNumberOfBytesTransferred);
            
            // add a metadata RP
            try {
                metadata=new ServiceMetaDataResourceProperty();
                this.resourcePropertySet.add(metadata);
            } catch(Exception e) {
                logger.error("Unable to add ServiceMetaData RP");
            }              

            // now register to the DefaultIndexService, if so configured

            Context ctx = new InitialContext();
            String config = Constants.JNDI_SERVICES_BASE_NAME +
                RFTConstants.FACTORY_NAME + "/mdsConfiguration";
            MDSConfiguration mdsConfiguration = 
                (MDSConfiguration) ctx.lookup(config);

            if(mdsConfiguration.getRegAsBoolean()) {
                logger.debug(i18n.getMessage("mdsEnabled"));
                registerToMDS();
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug(i18n.getMessage("mdsSuppressed"));
                }
            }

        } catch (Exception e) {
            logger.error(i18n.getMessage("factoryCreateErr"), e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.globus.wsrf.ResourceProperties#getResourcePropertySet()
     */
    public ResourcePropertySet getResourcePropertySet() {
        return this.resourcePropertySet;
    }

    /**
     * @param input
     * @return <code> CreateReliableFileTransferOutputType </code>
     * @throws RemoteException
     * @throws BaseFaultType
     */
    public CreateReliableFileTransferOutputType createReliableFileTransfer(
            CreateReliableFileTransferInputType input) throws RemoteException,
            BaseFaultType {
        if(logger.isDebugEnabled()) {
            logger.debug("PerformanceLog createReliableFileTransfer() enter");
        }
        Calendar requestedTermTime = input.getInitialTerminationTime();
        Calendar currentTime = Calendar.getInstance();
        Calendar initTermTime = currentTime;
        if (requestedTermTime == null) {
            initTermTime = null;
        } else {
            if (!requestedTermTime.getTime().before(currentTime.getTime())) {
                initTermTime = requestedTermTime;
            } else {
                throw new UnableToSetTerminationTimeFaultType();
            }
        }
        CreateReliableFileTransferOutputType response = 
            new CreateReliableFileTransferOutputType();
        response.setCurrentTime(currentTime);
        response.setNewTerminationTime(initTermTime);
        BaseRequestType transferRequestBase = input.getTransferRequest();
        if (transferRequestBase == null) {
            transferRequestBase = input.getDeleteRequest();
        }
        if (transferRequestBase != null) {
            this.activeResourcesRP.addInstance();
        } else {
            throw new RemoteException(
                    i18n.getMessage("nullRequestErr"));
        }
        if(transferRequestBase instanceof DeleteRequestType) {
            if (logger.isDebugEnabled()) {
                logger.debug("Got a delete request");
            }
        } else if(transferRequestBase instanceof TransferRequestType){
            if (logger.isDebugEnabled()) {
                logger.debug("Got a transfer request");
            }
            TransferRequestType tempReq
                = (TransferRequestType) transferRequestBase;
            int size = tempReq.getTransfer().length;
            this.totalTransfersRP.addTransfers(size);
        }
        // Check if EPR to delegated cred is missing
        if (transferRequestBase.getTransferCredentialEndpoint() == 
            null ) {
            logger.error("DelegationEPR is missing from request");
            throw new DelegationEPRMissingFaultType();
        }
        EndpointReferenceType epr = new EndpointReferenceType();

        //get the caller's user subject
        SecurityManager securityManager = SecurityManager.getManager();
        String userSubject = securityManager.getCaller();
        // Cannot be null, authz would have failed. Using the first
        // mapped local name.
        String localUserId = securityManager.getLocalUsernames()[0];
        // Get the user name from the transferRequest
        
        ReliableFileTransferHome reliableHome = null;
        SimpleResourceKey key = null;
        int requestId = 0;
        try {
            Context ctx = new InitialContext();
            String lookupString = Constants.JNDI_SERVICES_BASE_NAME +
                RFTConstants.SERVICE_NAME + "/home";
            reliableHome = (ReliableFileTransferHome) ctx.lookup(lookupString);

            key = reliableHome.create(currentTime, requestedTermTime,
                    transferRequestBase, userSubject, localUserId);
            requestId = Integer.parseInt((String) key.getValue());
            if(logger.isDebugEnabled()) {
                logger.debug("PerformanceLog created RFT" +  
                        "resource with RequestID:" + requestId);
            }
        } catch (Exception e) {
            logger.error(
                    i18n.getMessage("resourceCreateErr"), e);
            throw new RemoteException(
                    i18n.getMessage("resourceCreateErr"), e);
        }
        try {
            epr = AddressingUtils.createEndpointReference(ServiceHost
                    .getBaseURL() + RFTConstants.SERVICE_NAME, key);
        } catch (Exception e) {
            logger.error(i18n.getMessage("resourceCreateErr"), e);
            throw new RemoteException(
                    i18n.getMessage("resourceCreateErr"), e);
        }
        response.setReliableTransferEPR(epr);   
        if(logger.isDebugEnabled()) {
            logger.debug("PerformanceLog exit createReliableFileTransfer() for :" + requestId);
        }
        return response;
    }

    /** registers RFT to local MDS index */

    void registerToMDS() {
        try {
            // form EPR to RFT
            EndpointReferenceType epr = 
                AddressingUtils.createEndpointReference(null);

            // perform registraion
            ServiceGroupRegistrationClient.register(
                                epr, 
                                "/etc/globus_wsrf_rft/registration.xml");
        } catch(Exception e) {
            logger.error(i18n.getMessage("mdsRegErr"), e);
        }
    }
    
}
