/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_operation_provider.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "version.h"

#include "wsrp_GetMultipleResourcePropertiesType.h"
#include "wsrp_GetMultipleResourcePropertiesResponseType.h"

#define WSRP_GET_MULTIPLE_RESOURCE_PROPERTIES 1
#include "wsrp_i_faults.h"

GlobusExtensionDeclareModule(globus_ws_resource_properties);

/**
 * @page wsrp_GetMultipleResourceProperties_provider GetMultipleResourceProperties Provider
 *
 * <h2>Provider Usage Overview</h2>
 * This operation provider implements the GetMultipleResourceProperties
portType defined in
 * http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl .
 * This portType consists of the @a GetMultipleResourceProperties operation and
 * related faults.
 *
 * The WSRF service engine will automatically register this provider to handle
 * the GetMultipleResourceProperties operation for services which use the port
 * type.
 */

/* Prototypes */
static
globus_result_t
wsrp_GetMultipleResourceProperties_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrp_GetMultipleResourcePropertiesType *
                                        input);

static
globus_result_t
wsrp_GetMultipleResourceProperties_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrp_GetMultipleResourcePropertiesType *
                                        input,
    wsrp_GetMultipleResourcePropertiesResponseType * 
                                        output,
    char **                             fault_name,
    void **                             fault);

static
int
wsrp_GetMultipleResourceProperties_activate(void);

static
int
wsrp_GetMultipleResourceProperties_deactivate(void);

/* Local Variables */
static xsd_QName wsrp_GetMultipleResourceProperties_qname =
{
    "http://www.globus.org/docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl",
    "GetMultipleResourceProperties"
};

static
globus_operation_provider_descriptor_t
wsrp_GetMultipleResourceProperties_descriptor =
{
    &wsrp_GetMultipleResourceProperties_qname,
    "GetMultipleResourceProperties",
    (void *)wsrp_GetMultipleResourceProperties_init_provider,
    (void *)wsrp_GetMultipleResourceProperties_provider
};

GlobusExtensionDefineModule(globus_ws_resource_properties) =
{
    "wsrp_GetMultipleResourceProperties",
    wsrp_GetMultipleResourceProperties_activate,
    wsrp_GetMultipleResourceProperties_deactivate,
    NULL,
    NULL,
    &local_version
};

/* Implementation */
static
globus_result_t
wsrp_GetMultipleResourceProperties_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrp_GetMultipleResourcePropertiesType *
                                        input)
{
    return GLOBUS_SUCCESS;
}
/* wsrp_GetMultipleResourceProperties_init_provider() */

static
globus_result_t
wsrp_GetMultipleResourceProperties_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrp_GetMultipleResourcePropertiesType *
                                        input,
    wsrp_GetMultipleResourcePropertiesResponseType * 
                                        output,
    char **                             fault_name,
    void **                             fault)
{
    globus_resource_t                   resource;
    void *                              property;
    xsd_any *                           value;
    globus_xsd_type_info_t              info;
    globus_result_t                     result;
    int                                 i, j;
    xsd_any *                           prop_any = NULL;

    result = globus_wsrf_core_get_resource(
        handle,
        service,
        &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = wsrp_l_ResourceUnknownFaultType_create(
                fault_name,
                (wsrp_ResourceUnknownFaultType **) fault);

        goto out;
    }

    for (i = 0; i < input->ResourceProperty.length; i++)
    {
        globus_object_t *               error;

        result = globus_resource_get_property(
                resource,
                &input->ResourceProperty.elements[i],
                &property,
                &info);

        if (result != GLOBUS_SUCCESS)
        {
            error = globus_error_get(result);

            if (globus_error_match(error,
                        GLOBUS_WSRF_RESOURCE_MODULE,
                        GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
            {
                result = wsrp_l_InvalidResourcePropertyQNameFaultType_create(
                        fault_name,
                        (wsrp_InvalidResourcePropertyQNameFaultType **) fault);
            }
            else
            {
                result = wsrp_l_ResourceUnknownFaultType_create(
                        fault_name,
                        (wsrp_ResourceUnknownFaultType **) fault);

            }
            globus_object_free(error);

            goto free_array_out;
        }

        if (info == &xsd_any_info)
        {
            prop_any = property;
        }
        
        if ((prop_any != NULL &&
             prop_any->any_info != NULL &&
             prop_any->value != NULL) || 
            (prop_any == NULL && property != NULL))
        {
            value = xsd_any_array_push(&output->any);
            if (value == NULL)
            {
                result = globus_error_put(GLOBUS_ERROR_NO_INFO);

                goto free_array_out;
            }
            value->any_info = info;
            result = xsd_QName_copy(&value->element,
                    &input->ResourceProperty.elements[i]);
            if (result != GLOBUS_SUCCESS)
            {
                goto free_array_out;
            }

            result = info->copy(&value->value, property);
            if (result != GLOBUS_SUCCESS)
            {
                goto free_array_out;
            }
        }
    }
    globus_resource_finish(resource);

    return GLOBUS_SUCCESS;

free_array_out:
    for (j = i-1; j >=0; j--)
    {
        info = output->any.elements[j].any_info;
        info->destroy(output->any.elements[j].value);
    }
    free(output->any.elements);
    output->any.length = 0;
    output->any.elements = NULL;

    globus_resource_finish(resource);
out:
    return result;
}
/* wsrp_GetMultipleResourceProperties_provider() */

static
int
wsrp_GetMultipleResourceProperties_activate(void)
{
    int                                 res = 0;
    GlobusFuncName(wsrp_GetMultipleResourceProperties_activate);
    
    res = globus_extension_registry_add(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_GetMultipleResourceProperties_qname,
        GlobusExtensionMyModule(globus_ws_resource_properties),
        &wsrp_GetMultipleResourceProperties_descriptor);


    return res;
}
/* wsrp_GetMultipleResourceProperties_activate() */

static
int
wsrp_GetMultipleResourceProperties_deactivate(void)
{
    GlobusFuncName(wsrp_GetMultipleResourceProperties_deactivate);

    globus_extension_registry_remove(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_GetMultipleResourceProperties_qname);

    return 0;
}
/* wsrp_GetMultipleResourceProperties_deactivate() */
