/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_wsrf_resource.h"
#include "globus_service_engine.h"
#include "globus_operation_provider.h"
#include "globus_service_registry.h"
#include "globus_wsrf_core_tools.h"
#include "globus_service_registry.h"
#include "version.h"

#include "wsrp_SetResourcePropertiesType.h"
#include "wsrp_SetResourcePropertiesResponseType.h"

#define WSRP_SET_RESOURCE_PROPERTIES 1
#define NEED_UNABLE_TO_MODIFY_RESOURCE_PROPERTY_FAULT 0

#include "wsrp_i_faults.h"

/**
 * @page wsrp_SetResourceProperties_provider SetResourceProperties Provider
 *
 * <h2>Provider Usage Overview</h2>
 * This operation provider implements the SetResourceProperties portType defined
 * in
 * http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl .
 * This portType consists of the @a SetResourceProperties operation and
 * related faults.
 *
 * The WSRF service engine will automatically register this provider to handle
 * the SetResourceProperties operation for services which use the port type.
 *
 * <h2>Limitations</h2>
 * - This provider does not provide any transactional behavior when multiple
 * Insert, Update, or Delete elements are present in the
 * SetResourcePropertiesRequest element, nor does it prevent concurrent
 * accesses to resource property values
 * - This provider does not implement read-only resource properties.
 * - The minOccurs and maxOccurs attributes are extremely loosely enforced.
 *   If neither are present, then a resource property may have exactly one
 *   value. If either are present, then the resource property is effectively
 *   unbounded.
 */

static
int
wsrp_SetResourceProperties_activate(void);

static
int
wsrp_SetResourceProperties_deactivate(void);

GlobusExtensionDefineModule(globus_ws_resource_properties) =
{
    "wsrp_SetResourceProperties",
    wsrp_SetResourceProperties_activate,
    wsrp_SetResourceProperties_deactivate,
    NULL,
    NULL,
    &local_version
};

/* Prototypes */
static
globus_result_t
wsrp_SetResourceProperties_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrp_SetResourcePropertiesType *    SetResourceProperties);

static
globus_result_t
wsrp_SetResourceProperties_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrp_SetResourcePropertiesType *    SetResourceProperties,
    wsrp_SetResourcePropertiesResponseType *
                                        SetResourcePropertiesResponse,
    char **                             fault_name,
    void **                             fault);

static
globus_result_t
wsrp_l_InsertResourceProperty(
    globus_resource_t                   resource,
    xsd_any_array *                     any,
    char **                             fault_name,
    void **                             fault);

static
globus_result_t
wsrp_l_UpdateResourceProperty(
    globus_resource_t                   resource,
    xsd_any_array *                     any,
    char **                             fault_name,
    void **                             fault);

static
globus_result_t
wsrp_l_DeleteResourceProperty(
    globus_resource_t                   resource,
    xsd_QName *                         qname,
    char **                             fault_name,
    void **                             fault);

static
globus_result_t
wsrp_l_check_for_uniformity(
    xsd_any_array *                     any_array,
    char **                             fault_name,
    void **                             fault);

/* Local Variables */
static xsd_QName wsrp_SetResourceProperties_qname =
{
    "http://www.globus.org/docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.wsdl",
    "SetResourceProperties"
};

static
globus_operation_provider_descriptor_t
wsrp_SetResourceProperties_descriptor =
{
    &wsrp_SetResourceProperties_qname,
    "SetResourceProperties",
    (void *)wsrp_SetResourceProperties_init_provider,
    (void *)wsrp_SetResourceProperties_provider
};

/* Implementation */

static
globus_result_t
wsrp_SetResourceProperties_init_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    wsrp_SetResourcePropertiesType *    SetResourceProperties)
{
    return GLOBUS_SUCCESS;
}
/* wsrp_SetResourceProperties_init_provider() */

static
globus_result_t
wsrp_SetResourceProperties_provider(
    globus_service_engine_t             engine,
    globus_soap_message_handle_t        handle,
    globus_service_descriptor_t *       service,
    wsrp_SetResourcePropertiesType *    SetResourceProperties,
    wsrp_SetResourcePropertiesResponseType *
                                        SetResourcePropertiesResponse,
    char **                             fault_name,
    void **                             fault)
{
    globus_resource_t                   resource;
    globus_result_t                     result;
    int                                 i;
    xsd_any_array *                     any;

    *fault_name = NULL;
    *fault  = NULL;

    result = globus_wsrf_core_get_resource(
            handle,
            service,
            &resource);

    if (result != GLOBUS_SUCCESS || resource == NULL)
    {
        result = RESOURCE_UNKNOWN(fault_name, fault);

        goto out;
    }

    for (i = 0; i < SetResourceProperties->choice_value.length; i++)
    {
        switch (SetResourceProperties->choice_value.elements[i].type)
        {
            case wsrp_SetResourcePropertiesType_Insert:
                any = &SetResourceProperties->choice_value.elements[i].
                            value.Insert.any,
                result = wsrp_l_check_for_uniformity(any, fault_name, fault);
                if (result != GLOBUS_SUCCESS || *fault != NULL)
                {
                    goto finish_out;
                }
                result = wsrp_l_InsertResourceProperty(
                        resource,
                        any,
                        fault_name,
                        fault);

                if (result != GLOBUS_SUCCESS || *fault != NULL)
                {
                    goto finish_out;
                }

                break;

            case wsrp_SetResourcePropertiesType_Update:
                any = &SetResourceProperties->choice_value.elements[i].
                            value.Update.any,
                result = wsrp_l_check_for_uniformity(any, fault_name, fault);
                if (result != GLOBUS_SUCCESS || *fault != NULL)
                {
                    goto finish_out;
                }
                result = wsrp_l_UpdateResourceProperty(
                        resource,
                        any,
                        fault_name,
                        fault);

                if (result != GLOBUS_SUCCESS || *fault != NULL)
                {
                    goto finish_out;
                }

                break;

            case wsrp_SetResourcePropertiesType_Delete:
                result = wsrp_l_DeleteResourceProperty(
                        resource,
                        &SetResourceProperties->choice_value.elements[i].
                            value.Delete._ResourceProperty,
                        fault_name,
                        fault);

                if (result != GLOBUS_SUCCESS || *fault != NULL)
                {
                    goto finish_out;
                }
                break;

            case wsrp_SetResourcePropertiesType_undefined:
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto finish_out;
        }
    }

finish_out:
    globus_resource_finish(resource);
out:
    return result;
}
/* wsrp_SetResourceProperties_provider() */

static
int
wsrp_SetResourceProperties_activate(void)
{
    int                                 res = 0;
    GlobusFuncName(wsrp_SetResourceProperties_activate);
    
    res = globus_extension_registry_add(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_SetResourceProperties_qname,
        GlobusExtensionMyModule(globus_ws_resource_properties),
        &wsrp_SetResourceProperties_descriptor);

    return res;
}
/* wsrp_SetResourceProperties_activate()) */

static
int
wsrp_SetResourceProperties_deactivate(void)
{
    GlobusFuncName(wsrp_SetResourceProperties_deactivate);

    globus_extension_registry_remove(
        GLOBUS_OPERATION_PROVIDER_REGISTRY,
        &wsrp_SetResourceProperties_qname);

    return 0;
}
/* wsrp_SetResourceProperties_deactivate() */

/** 
 * Insert a new RP element to the RP document.
 */
static
globus_result_t
wsrp_l_InsertResourceProperty(
    globus_resource_t                   resource,
    xsd_any_array *                     any,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_xsd_type_info_t              prop_info;
    globus_xsd_type_info_t              array_info;
    void *                              prop_value;
    xsd_any *                           prop_any;
    globus_bool_t                       use_any = GLOBUS_FALSE;
    int                                 i;
    globus_object_t *                   error;
    void *                              copy_value = NULL;

    result = globus_resource_get_property(
            resource,
            any->elements[0].element,
            &prop_value,
            &prop_info);

    if (result != GLOBUS_SUCCESS)
    {
        /* The property QName doesn't appear to be in our RP document */
        error = globus_error_get(result);
        if (globus_error_match(
                error,
                GLOBUS_WSRF_RESOURCE_MODULE,
                GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
        {
            result = INVALID_RESOURCE_PROPERTY_QNAME(fault_name, fault);
        }
        else
        {
            result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(fault_name, fault);
        }
        globus_object_free(error);
        goto out;
    }

    if (prop_info == &xsd_any_info)
    {
        /* Got an any */
        prop_any = prop_value;

        if (prop_any->any_info == NULL ||
            prop_any->any_info == any->elements[0].any_info)
        {
            /* Single value info */
            if (any->length > 1 || prop_any->value != NULL)
            {
                /* Only a single value will fit in the any */
                result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                        fault_name,
                        fault);
                goto out;
            }
            if (prop_any->element == NULL)
            {
                /* The any is new, we'll need to set element and info */
                prop_any->any_info = any->elements[0].any_info;
                result = xsd_QName_copy(
                        &prop_any->element,
                        any->elements[0].element);

                if (result != GLOBUS_SUCCESS)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);

                    goto out;
                }
            }

            /* Copy the content of the any passed in to this function to the
             * any in the RP document
             */
            result = any->elements[0].any_info->copy(
                    &prop_any->value,
                    any->elements[0].value);
            if (result != GLOBUS_SUCCESS)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto out;
            }
        }
        else if (prop_any->any_info == &xsd_any_array_info ||
            prop_any->any_info == any->elements[0].any_info->array_info)
        {
            if (prop_any->element == NULL)
            {
                /* New any with hint that we can stick as many as we like here
                 */
                result = xsd_QName_copy(
                        &prop_any->element,
                        any->elements[0].element);

                if (result != GLOBUS_SUCCESS)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);

                    goto out;
                }
                /* Make new array for values */
                array_info = any->elements[0].any_info->array_info;
                if (any->elements[0].any_info->array_info == NULL)
                {
                    /* No array info? I guess use the any one */
                    array_info = &xsd_any_array_info;
                    use_any = GLOBUS_TRUE;
                }
                result = array_info->initialize(&prop_any->value);
                if (result != GLOBUS_SUCCESS)
                {
                    xsd_QName_destroy(prop_any->element);
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);

                    goto out;
                }
                prop_any->any_info = array_info;
            }

            /* Add new values to the any array now */
            for (i = 0; i < any->length; i++)
            {
                copy_value = prop_any->any_info->push(prop_any->value);

                if (copy_value == NULL)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);
                    goto out;
                }

                if (!use_any)
                {
                    result = any->elements[i].any_info->copy_contents(
                            copy_value,
                            any->elements[i].value);
                }
                else
                {
                    xsd_any_info.copy_contents(
                            copy_value,
                            &any->elements[i]);
                }
                if (result != GLOBUS_SUCCESS)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);
                    goto out;
                }
            }
        }
    }
    else if (any->elements[0].any_info == prop_info)
    {
        /* The RP element and this element have the same type. We're OK
         * If there's only one element in the any.
         */
        if (any->length > 1 || prop_value != NULL)
        {
            result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                    fault_name,
                    fault);
            goto out;
        }
        prop_info->copy(&copy_value, any->elements[0].value);
        result = globus_resource_set_property(
                resource,
                any->elements[0].element,
                copy_value);
        if (result != GLOBUS_SUCCESS)
        {
            error = globus_error_get(result);

            if (globus_error_match(
                    error,
                    GLOBUS_WSRF_RESOURCE_MODULE,
                    GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
            {
                result = INVALID_RESOURCE_PROPERTY_QNAME(fault_name, fault);
            }
            else
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
            }
            globus_object_free(error);

            goto out;
        }
    }
    else if (any->elements[0].any_info->array_info == prop_info)
    {
        /* The RP element is an array of the type we have in the any, so
         * we'll need to add the new values to the array.
         */
        void * array_value = prop_value;

        if (array_value == NULL)
        {
            result = prop_info->initialize(&array_value);
            if (result != GLOBUS_SUCCESS)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto out;
            }
            result = globus_resource_set_property(
                    resource,
                    any->elements[0].element,
                    array_value);

            if (result != GLOBUS_SUCCESS)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto out;
            }
        }

        for (i = 0; i < any->length; i++)
        {
            copy_value =
                any->elements[i].any_info->array_info->push(array_value);
            any->elements[i].any_info->copy_contents(
                    copy_value,
                    any->elements[i].value);
        }
    }
    else
    {
        /* The types don't match at all, we can't do anything about that */
        result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                fault_name,
                fault);
        goto out;
    }
out:
    return result;
}
/* wsrp_l_InsertResourceProperty() */

static
globus_result_t
wsrp_l_UpdateResourceProperty(
    globus_resource_t                   resource,
    xsd_any_array *                     any,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_xsd_type_info_t              prop_info;
    void *                              prop_value;
    xsd_any *                           prop_any;
    globus_xsd_type_info_t              old_info;
    globus_xsd_type_info_t              array_info;
    globus_bool_t                       use_any = GLOBUS_FALSE;
    xsd_QName *                         old_element;
    void *                              old_value;
    int                                 i;
    globus_object_t *                   error;
    void *                              copy_value = NULL;

    result = globus_resource_get_property(
            resource,
            any->elements[0].element,
            &prop_value,
            &prop_info);

    if (result != GLOBUS_SUCCESS)
    {
        /* The property QName doesn't appear to be in our RP document */
        error = globus_error_get(result);
        if (globus_error_match(
                error,
                GLOBUS_WSRF_RESOURCE_MODULE,
                GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
        {
            result = INVALID_RESOURCE_PROPERTY_QNAME(fault_name, fault);
        }
        else
        {
            result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(fault_name, fault);
        }

        globus_object_free(error);
        goto out;
    }

    if (prop_info == &xsd_any_info)
    {
        /* Got an any */
        prop_any = prop_value;

        if (prop_any->any_info == NULL ||
            prop_any->any_info == any->elements[0].any_info)
        {
            /* Single value info */
            if (any->length > 1)
            {
                /* Only a single value will fit in the any */
                result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                        fault_name,
                        fault);
                goto out;
            }

            old_value = prop_any->value;
            old_info = prop_any->any_info;

            if (prop_any->element == NULL)
            {
                /* The any is new, we'll need to set element and info */
                prop_any->any_info = any->elements[0].any_info;
                result = xsd_QName_copy(
                        &prop_any->element,
                        any->elements[0].element);

                if (result != GLOBUS_SUCCESS)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);

                    goto out;
                }
            }

            /* Copy the content of the any passed in to this function to the
             * any in the RP document
             */
            result = any->elements[0].any_info->copy(
                    &prop_any->value,
                    any->elements[0].value);
            if (result != GLOBUS_SUCCESS)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto out;
            }

            if (old_info && old_value)
            {
                old_info->destroy(old_value);
            }
        }
        else if (prop_any->any_info == &xsd_any_array_info ||
            prop_any->any_info == any->elements[0].any_info->array_info)
        {
            old_info = prop_any->any_info;
            old_value = prop_any->value;
            old_element = prop_any->element;

            prop_any->value = NULL;
            if (prop_any->element == NULL)
            {
                /* New any with hint that we can stick as many as we like here
                 */
                result = xsd_QName_copy(
                        &prop_any->element,
                        any->elements[0].element);

                if (result != GLOBUS_SUCCESS)
                {
                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);

                    goto out;
                }

            }

            /* Make new array for values */
            array_info = any->elements[0].any_info->array_info;
            if (any->elements[0].any_info->array_info == NULL)
            {
                /* No array info? I guess use the any one */
                array_info = &xsd_any_array_info;
                use_any = GLOBUS_TRUE;
            }
            result = array_info->initialize(&prop_any->value);

            if (result != GLOBUS_SUCCESS)
            {
                /* Restore property state */
                xsd_QName_destroy(prop_any->element);
                prop_any->value = old_value;
                prop_any->element = old_element;
                prop_any->any_info = old_info;

                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);

                goto out;
            }
            prop_any->any_info = array_info;

            /* Add new values to the any array now */
            for (i = 0; i < any->length; i++)
            {
                copy_value = prop_any->any_info->push(prop_any->value);

                if (copy_value == NULL)
                {
                    /* Restore property state */
                    xsd_QName_destroy(prop_any->element);
                    prop_any->value = old_value;
                    prop_any->element = old_element;
                    prop_any->any_info = old_info;

                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);
                    goto out;
                }

                if (!use_any)
                {
                    result = any->elements[i].any_info->copy_contents(
                            copy_value,
                            any->elements[i].value);
                }
                else
                {
                    result = xsd_any_info.copy_contents(
                            copy_value,
                            &any->elements[i]);
                }
                if (result != GLOBUS_SUCCESS)
                {
                    prop_any->any_info->destroy(prop_any->value);

                    /* Restore property state */
                    xsd_QName_destroy(prop_any->element);
                    prop_any->value = old_value;
                    prop_any->element = old_element;
                    prop_any->any_info = old_info;


                    result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                            fault_name,
                            fault);
                    goto out;
                }
            }
            if (old_info && old_value)
            {
                old_info->destroy(old_value);
            }
        }
    }
    else if (any->elements[0].any_info == prop_info)
    {
        /* The RP element and this element have the same type. We're OK
         * If there's only one element in the any.
         */
        if (any->length > 1)
        {
            result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                    fault_name,
                    fault);
            goto out;
        }
        prop_info->copy(&copy_value, any->elements[0].value);
        result = globus_resource_set_property(
                resource,
                any->elements[0].element,
                copy_value);
        if (result != GLOBUS_SUCCESS)
        {
            error = globus_error_get(result);

            if (globus_error_match(
                    error,
                    GLOBUS_WSRF_RESOURCE_MODULE,
                    GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
            {
                result = INVALID_RESOURCE_PROPERTY_QNAME(fault_name, fault);
            }
            else
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
            }
            globus_object_free(error);

            goto out;
        }
    }
    else if (any->elements[0].any_info->array_info == prop_info)
    {
        /* The RP element is an array of the type we have in the any, so
         * we'll need to replace the value of the array
         */
        void * array_value = NULL;

        result = prop_info->initialize(&array_value);
        if (result != GLOBUS_SUCCESS)
        {
            result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                    fault_name,
                    fault);
            goto out;
        }
        result = globus_resource_set_property(
                resource,
                any->elements[0].element,
                array_value);

        if (result != GLOBUS_SUCCESS)
        {
            result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                    fault_name,
                    fault);
            goto out;
        }

        for (i = 0; i < any->length; i++)
        {
            copy_value =
                any->elements[i].any_info->array_info->push(array_value);
            if (copy_value == NULL)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);

                goto out;
            }
            result = any->elements[i].any_info->copy_contents(
                    copy_value,
                    any->elements[i].value);
            if (result != GLOBUS_SUCCESS)
            {
                result = SET_RESOURCE_PROPERTY_REQUEST_FAILED(
                        fault_name,
                        fault);
                goto out;
            }
        }
    }
    else
    {
        /* The types don't match at all, we can't do anything about that */
        result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                fault_name,
                fault);
        goto out;
    }
out:
    return result;
}
/* wsrp_l_UpdateResourceProperty() */

static
globus_result_t
wsrp_l_DeleteResourceProperty(
    globus_resource_t                   resource,
    xsd_QName *                         qname,
    char **                             fault_name,
    void **                             fault)
{
    globus_result_t                     result;
    globus_object_t *                   error;

    result = globus_resource_set_property(
        resource,
        qname,
        NULL);

    if (result != GLOBUS_SUCCESS)
    {
        error = globus_error_get(result);

        if (globus_error_match(
            error,
            GLOBUS_WSRF_RESOURCE_MODULE,
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY))
        {
            result = INVALID_RESOURCE_PROPERTY_QNAME(fault_name, fault);
        }
        else
        {
            result = RESOURCE_UNKNOWN(fault_name, fault);
        }
        globus_object_free(error);
    }

    return result;
}
/* wsrp_l_DeleteResourceProperty() */

/**
 * Check that all elements in the any array have the same element name
 * and type info.
 * 
 * If not, return a wsrp_InvalidSetResourcePropertiesRequestContentFaultType
 * fault in the fault_name and fault parameters.
 */
static
globus_result_t
wsrp_l_check_for_uniformity(
    xsd_any_array *                     any_array,
    char **                             fault_name,
    void **                             fault)
{
    int                                 i;
    globus_result_t                     result = GLOBUS_SUCCESS;
    xsd_QName *                         last_element;
    xsd_QName *                         element = NULL;
    globus_xsd_type_info_t              last_type;
    globus_xsd_type_info_t              type = NULL;

    /* Verify that all update elements are the same */
    for (i = 0; i < any_array->length; i++)
    {
        last_element = element;
        last_type = type;

        element = any_array->elements[i].element;
        type = any_array->elements[i].any_info;

        if (i != 0)
        {
            if (! xsd_QName_keyeq(element, last_element))
            {
                result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                        fault_name,
                        fault);

                goto out;
            }
            else if (type != last_type)
            {
                result = INVALID_SET_RESOURCE_PROPERTIES_REQUEST_CONTENT(
                        fault_name,
                        fault);

                goto out;
            }
        }
    }
out:
    return result;
}
/* wsrp_l_check_for_uniformity() */
