/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "WidgetService_client.h"
#include "wsrp_GetMultipleResourcePropertiesType.h"
#include "wsrp_GetMultipleResourcePropertiesResponseType.h"
#include "wsrl_ScheduledResourceTerminationRPType.h"
#include "globus_utp.h"

/*
 * Similar to the XIO throughput tests, but we have a bunch of SOAP overhead,
 * and our minimal GetMultipleResourceProperties message body is going to 
 * look like this:
 *
 * <ns1:GetMultipleResourceProperties>
 * </ns1:GetMultipleResourceProperties>
 *
 * Each message size, we will add another magnitude of names in that body
 *
 * <ns1:GetMultipleResourceProperties>
 * <ns1:ResourceProperty>ns4:TerminationTime</ns1:ResourceProperty>
 * ...
 * </ns1:GetMultipleResourceProperties>
 *
 */
#define USAGE "%s OPTIONS -s service-uri\n" \
 "OPTIONS:\n"  \
 "    -i iterations                       " \
 "number of iterations per message\n" \
 "                                        " \
 "size [1]\n" \
 "    -m maximum                          " \
 "maximum number of message names [100]\n" \
 "    -n                                  " \
 "disable HTTP/1.1 persistent connections\n" \
 "    -l path                             " \
 "log results to path [utp.log]\n"

int
main(int argc, char *argv[])
{
    int                                 rc = 0;
    WidgetService_client_handle_t       client_handle;
    createWidgetType                    createWidget;
    createWidgetResponseType *          createWidgetResponse = NULL;
    wsrl_SetTerminationTimeType         SetTerminationTime;
    wsrl_SetTerminationTimeResponseType *
                                        SetTerminationTimeResponse = NULL;
    wsrp_GetMultipleResourcePropertiesType
                                        GetMultipleResourceProperties;
    wsrp_GetMultipleResourcePropertiesResponseType *
                                        GetMultipleResourcePropertiesResponse;
    wsrl_DestroyType                    Destroy = { NULL };
    wsrl_DestroyResponseType *          DestroyResponse = NULL;
    int                                 fault_type = 0;
    xsd_any *                           fault = NULL;
    globus_result_t                     result;
    xsd_QName *                         qname;
    int                                 maximum_names = 100;
    int                                 i, j;
    char *                              service_uri = NULL;
    int                                 iterations = 1;
    int                                 num_timers;
    int                                 t;
    globus_bool_t                       non_persistent = GLOBUS_FALSE;
    globus_soap_message_attr_t          attrs = NULL;
    char *                              utp_logfile="utp.log";
    time_t                              termination_time;
    struct tm                           termination_tm;
    for (i = 1; i < argc; i++)
    {
        if (argc > i && ! strcmp(argv[i], "-i"))
        {
            iterations = atoi(argv[++i]);
        }
        else if (argc > i && !strcmp(argv[i], "-s"))
        {
            service_uri = argv[++i];
        }
        else if (argc > i && !strcmp(argv[i], "-m"))
        {
            maximum_names = atoi(argv[++i]);
        }
        else if (!strcmp(argv[i], "-n"))
        {
            non_persistent = GLOBUS_TRUE;
        }
        else if (argc > i && !strcmp(argv[i], "-l"))
        {
            utp_logfile = argv[++i];
        }
        else
        {
            fprintf(stderr, "Error: bad parameter: %s\n", argv[i]);
            fprintf(stderr, USAGE,
                    argv[0]);
            rc = 1;

            goto out;
        }
    }
    if (service_uri == NULL)
    {
        fprintf(stderr, "Error: Service URI not set\n");
        fprintf(stderr, USAGE, argv[0]);
        rc = 1;

        goto out;
    }

    rc = globus_module_activate(GLOBUS_SOAP_MESSAGE_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Error activating GLOBUS_SOAP_MESSAGE_MODULE\n");
        rc = 2;

        goto out;
    }
    rc = globus_module_activate(WIDGETSERVICE_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Error activating WIDGETSERVICE_MODULE\n");
        rc = 2;

        goto deactivate_out;
    }
    rc = globus_module_activate(GLOBUS_UTP_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Error activating GLOBUS_UTP_MODULE\n");
        rc = 2;

        goto deactivate_widget_out;
    }

    if (non_persistent)
    {
        result = globus_soap_message_attr_init(&attrs);

        if (result != GLOBUS_SUCCESS)
        {
            fprintf(stderr, "Error initializing SOAP attrs: %s\n",
                    globus_error_print_friendly(globus_error_peek(result)));
            rc = 3;

            goto deactivate_utp_out;
        }

        result = globus_soap_message_attr_set(
            attrs,
            GLOBUS_SOAP_MESSAGE_CONNECTION_KEY,
            NULL,
            NULL,
            "close");
        if (result != GLOBUS_SUCCESS)
        {
            fprintf(stderr, "Error setting SOAP attrs: %s\n",
                    globus_error_print_friendly(globus_error_peek(result)));
            rc = 3;

            goto destroy_attr_out;
        }
    }

    result = WidgetService_client_init(
            &client_handle,
            attrs,
            NULL);
    if (result != GLOBUS_SUCCESS)
    {
        rc = 3;

        goto destroy_attr_out;
    }
    for (i = 1, num_timers=0; i <= maximum_names; i *= 10)
    {
        num_timers++;
    }
    rc = globus_utp_init(num_timers, GLOBUS_UTP_MODE_SHARED);
    if (rc != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "Error initializing utp timers\n");
        rc = 4;

        goto handle_destroy_out;
    }
    globus_utp_set_attribute(
        "http_persistence",
        "",
        non_persistent ? "disabled" : "enabled");
    globus_utp_set_attribute(
        "gsi",
        "",
        strncmp(service_uri, "https", 5) ? "disabled" : "enabled");
    globus_utp_set_attribute(
        "iterations",
        "",
        "%d", iterations);
    
    for (i = 1, j=0; i <= maximum_names; i *= 10)
    {
        globus_utp_name_timer(j, "%d RPs", i);
        j++;
    }
    result = WidgetPortType_createWidget(
        client_handle,
        service_uri,
        &createWidget,
        &createWidgetResponse,
        (WidgetPortType_createWidget_fault_t *) &fault_type,
        &fault);

    if (result != GLOBUS_SUCCESS)
    {
        fprintf(stderr, "%s\n",
                globus_error_print_friendly(globus_error_peek(result)));
        rc = 9;

        goto handle_destroy_out;
    }

    memset(&SetTerminationTime, '\0', sizeof (wsrl_SetTerminationTimeType));
    termination_time = time(NULL) + 60 * 120;
    if (globus_libc_gmtime_r(&termination_time, &termination_tm) == NULL)
    {
        rc = 10;
        fprintf(stderr, "Determining termination time\n");
        goto destroy_attr_out;
    }

    memcpy(&SetTerminationTime.RequestedTerminationTime,
            &termination_tm,
            sizeof (struct tm));

    result = WidgetPortType_SetTerminationTime_epr(
        client_handle,
        &createWidgetResponse->EndpointReference,
        &SetTerminationTime,
        &SetTerminationTimeResponse,
        (WidgetPortType_SetTerminationTime_fault_t *)
                &fault_type,
        &fault);

    if (result != GLOBUS_SUCCESS)
    {
        char * errmsg = globus_error_print_friendly(globus_error_peek(result));
        fprintf(stderr, "Error setting termination time: %s\n", errmsg);
        globus_libc_free(errmsg);

        goto resource_destroy_out;
                
    }
    wsrl_SetTerminationTimeResponseType_destroy(SetTerminationTimeResponse);
    SetTerminationTimeResponse = NULL;
    assert(result == GLOBUS_SUCCESS && fault == NULL);

    for (i = 1, t=0; i <= maximum_names; i *= 10)
    {
        memset(&GetMultipleResourceProperties,
                '\0',
                sizeof(wsrp_GetMultipleResourcePropertiesType));
        for (j = 0; j < i; j++)
        {
            qname = xsd_QName_array_push(
                    &GetMultipleResourceProperties.ResourceProperty);
            qname->Namespace = globus_libc_strdup(
                    WSRL_SCHEDULEDRESOURCETERMINATIONRPTYPE_NS);
            qname->local = globus_libc_strdup(
                    "TerminationTime");
        }

        globus_utp_start_timer(t);
        for (j = 0; j < iterations; j++)
        {
            result = WidgetPortType_GetMultipleResourceProperties_epr(
                client_handle,
                &createWidgetResponse->EndpointReference,
                &GetMultipleResourceProperties,
                &GetMultipleResourcePropertiesResponse,
                (WidgetPortType_GetMultipleResourceProperties_fault_t *)
                        &fault_type,
                &fault);

            if (result != GLOBUS_SUCCESS)
            {
                fprintf(stderr, "Error getting resource properties: %s", 
                        globus_error_print_friendly(globus_error_peek(result)));
            }
            assert(result == GLOBUS_SUCCESS && fault == NULL);

            wsrp_GetMultipleResourcePropertiesResponseType_destroy(
                    GetMultipleResourcePropertiesResponse);
            GetMultipleResourcePropertiesResponse = NULL;
        }
        globus_utp_stop_timer(t);
        t++;
        wsrp_GetMultipleResourcePropertiesType_destroy_contents(
                &GetMultipleResourceProperties);
    }
    globus_utp_write_file(utp_logfile);

resource_destroy_out:
    WidgetPortType_Destroy_epr(
        client_handle,
        &createWidgetResponse->EndpointReference,
        &Destroy,
        &DestroyResponse,
        (WidgetPortType_Destroy_fault_t *) &fault_type,
        &fault);
    if (DestroyResponse != NULL)
    {
        wsrl_DestroyResponseType_destroy(DestroyResponse);
    }
    createWidgetResponseType_destroy(createWidgetResponse);
handle_destroy_out:
    WidgetService_client_destroy(client_handle);
destroy_attr_out:
    if (attrs)
    {
        globus_soap_message_attr_destroy(attrs);
    }
deactivate_utp_out:
    globus_module_deactivate(GLOBUS_UTP_MODULE);
deactivate_widget_out:
    globus_module_deactivate(WIDGETSERVICE_MODULE);
deactivate_out:
    globus_module_deactivate(GLOBUS_SOAP_MESSAGE_MODULE);
out:
    return rc;
}
/* main() */
