/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "globus_i_xsd.h"
#include "globus_xsd_type_info.h"
#include "globus_soap_message_utils.h"

/**
 * Create a new #globus_xsd_type_info_t type for serializing and
 * deserializing an element with the same type as the given template.
 *
 * @param info
 *     Pointer to the info type to be initialized;
 * @param type
 *     Element name.
 * @param source
 *     Original type info.
 *
 * @retval GLOBUS_XSD_ERROR_TYPE_NULL_PARAM
 *     One of the parameters was invalid.
 * @retval GLOBUS_XSD_ERROR_TYPE_OUT_OF_MEMORY
 *     Unable to allocate type-related data structures.
 */
globus_result_t
globus_xsd_type_info_init_from_template(
    globus_xsd_type_info_t *            info,
    const xsd_QName *                   name,
    globus_xsd_type_info_t              source)
{
    struct globus_xsd_type_info_s *     p;
    globus_result_t                     result;
    GlobusFuncName(globus_xsd_type_info_init);
    GlobusSoapMessageDebugEnter();

    if (info == NULL || name == NULL ||
        source == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    *info = NULL;

    p = globus_libc_malloc(sizeof(struct globus_xsd_type_info_s));

    if (p == NULL)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto error;
    }
    memcpy(p, source, sizeof(struct globus_xsd_type_info_s));

    p->type = globus_libc_malloc(sizeof(xsd_QName));
    if (p->type == NULL)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto free_p_error;
    }

    p->type->Namespace = globus_libc_strdup(name->Namespace);
    if (p->type->Namespace == NULL)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        
        goto free_p_type_error;
    }
    p->type->local = globus_libc_strdup(name->local);
    if (p->type->local == NULL)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto free_ns_error;
    }

    *info = p;
    return GLOBUS_SUCCESS;

free_ns_error:
    globus_libc_free(p->type->Namespace);
free_p_type_error:
    globus_libc_free(p->type);
free_p_error:
    globus_libc_free(p);
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_info_init_from_template() */

/**
 * Destroy a previously allocated type info structure.
 *
 * @param info
 *     Info to destroy.
 * @retval GLOBUS_SUCCESS
 *     Info structure destroyed successfully.
 * @retval GLOBUS_XSD_ERROR_TYPE_NULL_PARAM
 *     The @a info parameter was invalid.
 */
void
globus_xsd_type_info_destroy(
    globus_xsd_type_info_t              info)
{
    globus_result_t                     result;
    GlobusFuncName(globus_xsd_type_info_destroy);
    GlobusSoapMessageDebugEnter();

    if (info == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }

    if (info->type->Namespace != NULL)
    {
        globus_libc_free(info->type->Namespace);
    }
    if (info->type->local != NULL)
    {
        globus_libc_free(info->type->local);
    }
    globus_libc_free(info->type);
    globus_libc_free(info);

error:
    GlobusSoapMessageDebugExit();
    return;
}
/* globus_xsd_type_info_destroy() */
