/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "globus_i_xsd.h"
#include "globus_i_xsd_type_registry.h"
#include "globus_i_xsd_type_info.h"
#include "globus_soap_message_utils.h"
#include "globus_soap_message.h"

/**
 * Initialize a new type registry.
 *
 * @param registry
 *     New type registry.
 *
 * @retval GLOBUS_SUCCESS
 *     Registry initialized.
 * @retval GLOBUS_MESSAGE_ERROR_TYPE_NULL_PARAM
 *     The @a registry parameter was NULL.
 * @retval GLOBUS_MESSAGE_ERROR_TYPE_OUT_OF_MEMORY
 *     Unable to allocate registry data.
 *
 * @see globus_xsd_type_registry_destroy()
 */
globus_result_t
globus_xsd_type_registry_init(
    globus_xsd_type_registry_t *        registry)
{
    globus_result_t                     result;
    struct globus_xsd_type_registry_s * p;
    int                                 rc;
    GlobusFuncName(globus_xsd_type_registry_init);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    *registry = NULL;

    p = globus_libc_malloc(sizeof(struct globus_xsd_type_registry_s));
    if (p == NULL)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto error;
    }

    rc = globus_hashtable_init(
            &p->table,
            257,
            xsd_QName_hash,
            xsd_QName_keyeq);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto free_p_error;
    }

    p->default_mapping = &xsd_string_info;

    globus_mutex_init(&p->lock, NULL);
    *registry = p;

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;

free_p_error:
    globus_libc_free(p);
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_init() */

/**
 * Destroy a type registry.
 *
 * @param registry
 *     Type registry to destroy.
 *
 * @retval GLOBUS_SUCCESS
 *     Registry destroyed.
 * @retval GLOBUS_MESSAGE_ERROR_TYPE_NULL_PARAM
 *     The @a registry parameter was NULL.
 *
 * @see globus_xsd_type_registry_init()
 * @see globus_xsd_type_registry_clear()
 */
globus_result_t
globus_xsd_type_registry_destroy(
    globus_xsd_type_registry_t          registry)
{
    globus_result_t                     result;
    GlobusFuncName(globus_xsd_type_registry_destroy);
    GlobusSoapMessageDebugEnter();

    result = globus_xsd_type_registry_clear(registry);

    if (result == GLOBUS_SUCCESS)
    {
        globus_hashtable_destroy(&registry->table);
        globus_mutex_destroy(&registry->lock);
        globus_libc_free(registry);
    }

    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_destroy() */

/**
 * Insert a type into the registry.
 *
 * If an existing type mapping is registered for the specified type, it
 * is replaced with the new type. The old type is returned in value pointed to
 * by the @a old_info  parameter.
 *
 * @param registry
 *     Registry to modify.
 * @param info
 *     Information about the type.
 * @param old_info
 *     Pointer to be set to the value of the previous info for this type.
 *     May be NULL if the caller is not interested in this.
 *
 * @retval GLOBUS_SUCCESS
 *     Type mapping added
 * @retval GLOBUS_MESSAGE_ERROR_TYPE_NULL_PARAM
 *     The @a registry or @a info  parameter was NULL.
 *
 * @see globus_xsd_type_registry_remove()
 * @see globus_xsd_type_registry_get()
 */
globus_result_t
globus_xsd_type_registry_insert(
    globus_xsd_type_registry_t          registry,
    globus_xsd_type_info_t              info,
    globus_xsd_type_info_t *            old_info)
{
    globus_result_t                     result;
    globus_xsd_type_info_t              p;
    int                                 rc;
    GlobusFuncName(globus_xsd_type_registry_insert);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL || info == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }

    rc = globus_mutex_lock(&registry->lock);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto error;
    }
    
    p = globus_hashtable_lookup(&registry->table, info->type);
    if(p == info)
    {
        if(old_info)
        {
            *old_info = p;
        }
        goto exit;
    }
 
    p = globus_hashtable_remove(&registry->table, info->type);

    if (old_info)
    {
        *old_info = p;
    }

    rc = globus_hashtable_insert(&registry->table, info->type, info);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto unlock_error;
    }

 exit:
    globus_mutex_unlock(&registry->lock);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;

unlock_error:
    globus_mutex_unlock(&registry->lock);
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_insert() */

/**
 * Remove a type from the registry.
 *
 * @param registry
 *     Registry to modify.
 * @param name
 *     Name of the type to be unregistered.
 * @param info
 *     Pointer to be set to the previously registred type information. May be
 *     NULL if the caller doesn't want that type info.
 *
 * @retval GLOBUS_SUCCESS
 *     Type mapping removed.
 * @retval GLOBUS_MESSAGE_ERROR_TYPE_NULL_PARAM
 *     The @a registry or @a name parameter was NULL.
 *
 * @see globus_xsd_type_registry_replace()
 * @see globus_xsd_type_registry_remove()
 * @see globus_xsd_type_registry_get()
 */
globus_result_t
globus_xsd_type_registry_remove(
    globus_xsd_type_registry_t          registry,
    const xsd_QName *                   name,
    globus_xsd_type_info_t *            info)
{
    globus_result_t                     result;
    globus_xsd_type_info_t              tmp;
    GlobusFuncName(globus_xsd_type_registry_remove);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL || name == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }

    globus_mutex_lock(&registry->lock);
    tmp = globus_hashtable_remove(&registry->table, (void *) name);

    if (info != NULL)
    {
        *info = tmp;
    }
    globus_mutex_unlock(&registry->lock);
    
    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_remove() */

globus_result_t
globus_xsd_type_registry_clear(
    globus_xsd_type_registry_t          registry)
{
    int                                 rc;
    globus_result_t                     result;
    globus_list_t *                     tmp_list;
    globus_xsd_type_info_t              p;
    GlobusFuncName(globus_xsd_type_registry_clear);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    rc = globus_mutex_lock(&registry->lock);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto error;
    }
    rc = globus_hashtable_to_list(&registry->table, &tmp_list);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto unlock_error;
    }
    while (!globus_list_empty(tmp_list))
    {
        p = globus_list_remove(&tmp_list, tmp_list);

        globus_assert(p);

        globus_hashtable_remove(&registry->table, p->type);
    }

    globus_mutex_unlock(&registry->lock);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;


unlock_error:
    globus_mutex_unlock(&registry->lock);
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_clear() */

globus_result_t
globus_xsd_type_registry_get(
    globus_xsd_type_registry_t          registry,
    const xsd_QName *                   name,
    globus_xsd_type_info_t *            info)
{
    globus_result_t                     result;
    int                                 rc;
    GlobusFuncName(globus_xsd_type_registry_get);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL || name == NULL || info == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    rc = globus_mutex_lock(&registry->lock);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;

        goto error;
    }
    *info = globus_hashtable_lookup(&registry->table, (void *) name);

    globus_mutex_unlock(&registry->lock);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_get() */

globus_result_t
globus_xsd_type_registry_set_default(
    globus_xsd_type_registry_t          registry,
    globus_xsd_type_info_t              info)
{
    globus_result_t                     result;
    GlobusFuncName(globus_xsd_type_registry_set_default);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL || info == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    globus_mutex_lock(&registry->lock);
    registry->default_mapping = info;
    globus_mutex_unlock(&registry->lock);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_set_default() */

globus_result_t
globus_xsd_type_registry_get_default(
    globus_xsd_type_registry_t          registry,
    globus_xsd_type_info_t *            info)
{
    globus_result_t                     result;
    GlobusFuncName(globus_xsd_type_registry_get_default);
    GlobusSoapMessageDebugEnter();

    if (registry == NULL || info == NULL)
    {
        result = GlobusSoapMessageErrorNullParam;

        goto error;
    }
    globus_mutex_lock(&registry->lock);
    *info = registry->default_mapping;
    globus_mutex_unlock(&registry->lock);

    GlobusSoapMessageDebugExit();
    return GLOBUS_SUCCESS;
error:
    GlobusSoapMessageDebugExit();
    return result;
}
/* globus_xsd_type_registry_get_default() */


globus_result_t
globus_xsd_type_registry_add_base_types(
    globus_xsd_type_registry_t          registry)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_xsd_type_registry_add_base_types);
    GlobusSoapMessageDebugEnter();

    result = globus_xsd_type_registry_insert(registry, &xsd_QName_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_anyURI_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_base64Binary_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_boolean_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_byte_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_date_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_dateTime_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_decimal_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_double_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_duration_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_float_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_hexBinary_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_int_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_integer_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_long_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_negativeInteger_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_nonNegativeInteger_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_nonPositiveInteger_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_positiveInteger_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_short_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_string_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_time_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_unsignedByte_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_unsignedInt_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_unsignedLong_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }

    result = globus_xsd_type_registry_insert(registry, &xsd_unsignedShort_info, NULL);
    if(result != GLOBUS_SUCCESS)
    {
        goto exit;
    }
    
 exit:

    GlobusSoapMessageDebugExit();
    return result;
}

globus_result_t
globus_xsd_type_registry_dump(
    globus_xsd_type_registry_t               registry)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_xsd_type_info_t              info;
    int                                 rc = 0;
    GlobusFuncName(globus_xsd_type_registry_dump);
    GlobusSoapMessageDebugEnter();

    rc = globus_mutex_lock(&registry->lock);
    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto error;
    }
    
    info = globus_hashtable_first(&registry->table);

    while(info)
    {
        GlobusSoapMessageDebugPrintf(
            GLOBUS_SOAP_MESSAGE_DEBUG_REGISTRY,
            ("{%s}%s\n",
            info->type->Namespace,
            info->type->local));
        
        info = globus_hashtable_next(&registry->table);
    }

    rc = globus_mutex_unlock(&registry->lock);
    if(rc != GLOBUS_SUCCESS)
    {
        result = GlobusSoapMessageErrorOutOfMemory;
        goto error;
    }

 error:

    GlobusSoapMessageDebugExit();
    return result;
}
