/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_common.h"
#include "libxml/xmlreader.h"
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#define TEMPLATE_NODE                   "tmpl"
#define EVAL_NODE                       "eval"
#define SCRIPT_NODE                     "script"

#define BEGIN_PRINT                     "\nprint(\""
#define BEGIN_PRINT_LENGTH              8
#define END_PRINT                       "\");\n"
#define END_PRINT_LENGTH                4
#define PRINT_LENGTH                    12

#define BEGIN_EPRINT                    "\nprint("
#define BEGIN_EPRINT_LENGTH             7
#define END_EPRINT                      ");\n"
#define END_EPRINT_LENGTH               3
#define EPRINT_LENGTH                   10

#define DELIMITER_COUNT                 2
static const char *                     DELIMITERS = "\n\"";
static const char *                     REPLACEMENTS[DELIMITER_COUNT] = 
                                        {"\\n", "\\\""};

/**
 * Functions for parsing the template file
 */

static
const char *
my_strpbrk(
    const char *                        str,
    size_t                              str_length,
    int *                               locator_index,
    const char *                        delims,
    int *                               delim_index,
    int                                 escape);

static
int
globus_i_template_convert_chars(
    const char *                        old_buffer,
    size_t                              old_length,
    char **                             buffer,
    size_t *                            length,
    int                                 delim_count,
    const char *                        delim_set,
    const char **                       replace_set);

int
globus_wsdl_template_get_jsbuf(
    const char *                        filename,
    char **                             buffer,
    size_t *                            length,
    char **                             error_string,
    char                                script_token,
    char                                eval_token)
{
    size_t                              read_length = 0;
    char                                read_buffer[1000];
    size_t                              new_length = 0;
    char *                              new_buffer = NULL;
    size_t                              tmp_length = 0;
    char *                              tmp_buffer = NULL;
    size_t                              script_length = 0;
    size_t                              eval_length = 0;
    size_t                              tmpl_length = 0;
    const char *                        loc;
    const char *                        prev;
    char                                charset[3];
    size_t                              template_length = 0;
    char *                              template_buffer = NULL;
    int                                 infd, res;
    int                                 in_script = 0;
    int                                 in_eval = 0;
    char                                token;
    int                                 loc_index, loc_length;
    int                                 delim_index;
    int                                 add_loc_index = 0;

    *error_string = NULL;
    charset[0] = script_token;
    charset[1] = eval_token;
    charset[2] = 0;

    infd = open(filename, O_RDONLY, 0);
    if(infd < 0)
    {
/** XXX - this should be changed to use globus errno error calls */
        char *                          error_message;
        error_message = strerror(errno);
        *error_string = globus_libc_strdup(error_message); 
        goto exit;
    }

    memset(read_buffer, 0, 1000);

    while((read_length = read(infd, read_buffer, 1000)) > 0)
    {
        tmp_buffer = realloc(tmp_buffer, tmp_length + read_length);
        if(!tmp_buffer)
        {
/** XXX - this should be fixed to use globus_error_errno calls */
            char *                      error_message;
            error_message = strerror(errno);
            *error_string = globus_libc_strdup(error_message);
            goto exit;
        }
        
        memcpy(tmp_buffer + tmp_length, read_buffer, read_length);
        tmp_length += read_length;
    }

    tmp_buffer = realloc(tmp_buffer, tmp_length + 1);
    tmp_buffer[tmp_length] = 0;

    loc = tmp_buffer;
    loc_length = tmp_length;
    prev = tmp_buffer;

    while(1)
    {
        loc = my_strpbrk(loc, loc_length, &loc_index, charset, &delim_index, 1);
        if(!loc)
        {
            if(in_script)
            {
                *error_string = globus_libc_strdup(
                    "End of file reached inside a script chunk");
                goto exit;
            }

            if(in_eval)
            {
                *error_string = globus_libc_strdup(
                    "End of file reached inside an evalu chunk");
                goto exit;
            }

            tmpl_length = loc_length;
            globus_i_template_convert_chars(
                prev, tmpl_length,
                &template_buffer, &template_length,
                DELIMITER_COUNT,
                DELIMITERS, 
                REPLACEMENTS);
            new_buffer = realloc(
                new_buffer,
                new_length + template_length + PRINT_LENGTH);
            memcpy(new_buffer + new_length, 
                   BEGIN_PRINT, BEGIN_PRINT_LENGTH);
            memcpy(new_buffer + new_length + BEGIN_PRINT_LENGTH,
                   template_buffer, template_length);
            memcpy(new_buffer + new_length + 
                   BEGIN_PRINT_LENGTH + template_length, 
                   END_PRINT, END_PRINT_LENGTH);
            new_length += PRINT_LENGTH + template_length;

            free(template_buffer);
            template_buffer = NULL;

            break;
        }

        if(add_loc_index > 0)
        {
            loc_index += add_loc_index + 1;
            add_loc_index = 0;
        }

        if(loc_index != 0 && (*(loc - 1) == '\\'))
        {
            add_loc_index = loc_index;
            ++loc;
            continue;
        }

        loc_length -= (loc_index + 1);

        token = *loc;
        
        if(token == script_token)
        {
            if(in_eval)
            {
                *error_string = malloc(500);
                sprintf(*error_string,
                        "While parsing %s: Found a script token (%c), "
                        "expected an eval token (%c)",
                        filename, script_token, eval_token);
                goto exit;
            }

            if(in_script)
            {
                in_script = 0;

                script_length = loc_index;
                new_buffer = realloc(new_buffer, new_length + script_length);
                memcpy(new_buffer + new_length, prev, script_length);
                new_length += script_length;
            }
            else
            {
                in_script = 1;
                
                tmpl_length = loc_index;
                globus_i_template_convert_chars(
                    prev, tmpl_length,
                    &template_buffer, &template_length, 
                    DELIMITER_COUNT,
                    DELIMITERS,
                    REPLACEMENTS);
                new_buffer = realloc(
                    new_buffer,
                    new_length + template_length + PRINT_LENGTH);
                memcpy(new_buffer + new_length, 
                       BEGIN_PRINT, BEGIN_PRINT_LENGTH);
                memcpy(new_buffer + new_length + BEGIN_PRINT_LENGTH,
                       template_buffer, template_length);
                memcpy(new_buffer + new_length + 
                       BEGIN_PRINT_LENGTH + template_length, 
                       END_PRINT, END_PRINT_LENGTH);
                new_length += PRINT_LENGTH + template_length;

                free(template_buffer);
                template_buffer = NULL;
            }
        }
        else if(token == eval_token)
        {
            if(in_script)
            {
                *error_string = malloc(500);
                sprintf(
                    *error_string,
                    "While parsing %s: Found an eval token (%c), "
                    "expected a script token (%c)",
                    filename, eval_token, script_token);
                goto exit;
            }

            if(in_eval)
            {
                in_eval = 0;

                eval_length = loc_index;
                new_buffer = realloc(new_buffer, 
                                     new_length + eval_length + EPRINT_LENGTH);
                memcpy(new_buffer + new_length, BEGIN_EPRINT, 
                       BEGIN_EPRINT_LENGTH);
                memcpy(new_buffer + new_length + BEGIN_EPRINT_LENGTH,
                       prev, eval_length);
                memcpy(new_buffer + new_length + 
                       BEGIN_EPRINT_LENGTH + eval_length, 
                       END_EPRINT, END_EPRINT_LENGTH);
                new_length += EPRINT_LENGTH + eval_length;
            }
            else
            {
                in_eval = 1;

                tmpl_length = loc_index;
                globus_i_template_convert_chars(
                    prev, tmpl_length,
                    &template_buffer, &template_length, 
                    DELIMITER_COUNT,
                    DELIMITERS,
                    REPLACEMENTS);
                new_buffer = realloc(
                    new_buffer,
                    new_length + template_length + PRINT_LENGTH);
                memcpy(new_buffer + new_length, 
                       BEGIN_PRINT, BEGIN_PRINT_LENGTH);
                memcpy(new_buffer + new_length + BEGIN_PRINT_LENGTH,
                       template_buffer, template_length);
                memcpy(new_buffer + new_length + 
                       BEGIN_PRINT_LENGTH + template_length, 
                       END_PRINT, END_PRINT_LENGTH);
                new_length += PRINT_LENGTH + template_length;

                free(template_buffer);
                template_buffer = NULL;
            }
        }

        if(loc_length >= 0)
        {
            ++loc;
            prev = loc;
        }
        else
        {
            break;
        }
    }

    new_buffer = realloc(new_buffer, new_length + 2);
    memcpy(new_buffer + new_length, "\r\n", 2);
    new_length += 2;
    
    *buffer = new_buffer;
    *length = new_length;

  exit:

    if(infd > 0)
    {
        close(infd);
    }

    if(*error_string)
    {
        return -1;
    }

    return 0;
}


/* int */
/* globus_wsdl_xml_template_get_jsbuf( */
/*     const char *                        filename, */
/*     char **                             buffer, */
/*     size_t *                            length, */
/*     char **                             error_string) */
/* { */
/*     size_t                              new_length = 0; */
/*     xmlTextReaderPtr                    reader = NULL; */
/*     char *                              new_buffer = NULL; */

/*     reader = xmlNewTextReaderFilename(filename); */
/*     if(!reader) */
/*     { */
/*         xmlErrorPtr                     error; */
/*         char *                          error_message; */

/*         error = xmlGetLastError(); */
/*         if(!error) */
/*         { */
/*             return -1; */
/*         } */

/*         error_message = malloc(strlen(error->message) + 500); */
/*         if(!error_message) */
/*         { */
/*             return -1; */
/*         } */
        
/*         sprintf(error_message, "Could not parse template: %s: %s",  */
/*                 filename, error->message); */
/*         *error_string = error_message; */
/*         return -1; */
/*     } */

/*     while(xmlTextReaderRead(reader) > 0) */
/*     { */
/*         const xmlChar *                 local_name; */

/*         local_name = xmlTextReaderConstLocalName(reader); */
/*         if(xmlStrEqual(local_name, TEMPLATE_NODE)) */
/*         { */
/*             int                         bufflen; */
/*             char *                      template_buffer; */
/*             xmlNodePtr                  node; */

/*             node = xmlTextReaderExpand(reader); */
/*             if(node) */
/*             { */
/*                 node = node->children; */
/*             } */

/*             while(node) */
/*             { */
/*                 if(node->type == XML_TEXT_NODE) */
/*                 { */
/*                     if(node->content) */
/*                     { */
/*                         size_t          template_length; */

/*                         globus_i_template_convert_chars( */
/*                             node->content, strlen(node->content), */
/*                             &template_buffer, &template_length,  */
/*                             '\n', "\\n"); */
/*                         bufflen = template_length + 12; */
/*                         new_buffer = realloc(new_buffer, new_length + bufflen); */
/*                         memcpy(new_buffer + new_length, "\nprint(\"", 8); */
/*                         memcpy(new_buffer + new_length + 8,  */
/*                                template_buffer, template_length); */
/*                         memcpy(new_buffer + new_length + 8 + template_length, */
/*                                "\");\n", 4); */
/*                         new_length += bufflen; */
/*                     } */
/*                 } */
                
/*                 node = node->next; */
/*             } */
/*         } */
/*         else if(xmlStrEqual(local_name, EVAL_NODE)) */
/*         { */
/*             int                         bufflen; */
/*             size_t                      eval_length; */
/*             const xmlChar *             eval_buffer; */
/*             xmlNodePtr                  node; */

/*             node = xmlTextReaderExpand(reader); */
/*             if(node) */
/*             { */
/*                 node = node->children; */
/*             } */

/*             while(node) */
/*             { */
/*                 if(node->type == XML_TEXT_NODE) */
/*                 { */
/*                     eval_buffer = node->content; */
/*                     if(eval_buffer) */
/*                     { */
/*                         eval_length = strlen(eval_buffer); */
/*                         bufflen = eval_length + 10; */
/*                         new_buffer = realloc(new_buffer, new_length + bufflen); */
/*                         memcpy(new_buffer + new_length, "\nprint(", 7); */
/*                         memcpy(new_buffer + new_length + 7,  */
/*                                eval_buffer, eval_length); */
/*                         memcpy(new_buffer + new_length + 7 + eval_length, */
/*                                ");\n", 3); */
/*                         new_length += bufflen; */
/*                     } */
/*                 } */

/*                 node = node->next; */
/*             } */
/*         } */
/*         else if(xmlStrEqual(local_name, SCRIPT_NODE)) */
/*         { */
/*             size_t                      script_length; */
/*             const xmlChar *             script_buffer; */
/*             xmlNodePtr                  node; */

/*             node = xmlTextReaderExpand(reader); */
/*             if(node) */
/*             { */
/*                 node = node->children; */
/*             } */

/*             while(node) */
/*             { */
/*                 if(node->type == XML_TEXT_NODE) */
/*                 { */
/*                     script_buffer = node->content; */
/*                     if(script_buffer) */
/*                     { */
/*                         script_length = strlen(script_buffer); */
/*                         new_buffer = realloc(new_buffer,  */
/*                                              new_length + script_length); */
/*                         memcpy(new_buffer + new_length,  */
/*                                script_buffer, script_length); */
/*                         new_length += script_length; */
/*                     } */
/*                 } */

/*                 node = node->next; */
/*             } */
/*         } */
/*     } */

/*     new_buffer = realloc(new_buffer, new_length + 1); */
/*     memcpy(new_buffer + new_length, "\0", 1); */
/*     new_length += 1; */

/*     *buffer = new_buffer; */
/*     *length = new_length; */
/*     return 0; */
/* }         */

static
const char *
my_strpbrk(
    const char *                        str,
    size_t                              str_length,
    int *                               locator_index,
    const char *                        delims,
    int *                               delim_index,
    int                                 escape)
{
    const char *                        locator;
    int                                 delim_count, ind, i;

    ind = 0;
    i = 0;
    delim_count = strlen(delims);
    locator = str;
    while(*locator && ind < str_length)
    {
        for(i = 0; i < delim_count; ++i)
        {
            if(*locator == delims[i])
            {
                if(escape &&
                   (ind != 0 && (*(locator - 1) == '\\')))
                {
                    continue;
                }

                if(locator_index)
                {
                    *locator_index = ind;
                }

                if(delim_index)
                {
                    *delim_index = i;
                }
                return locator;
            }
        }
        
        locator++;
        ind++;
    }

    if(locator_index)
    {
        *locator_index = -1;
    }
    return NULL;
}

static
int
globus_i_template_convert_chars(
    const char *                        tmp_buffer,
    size_t                              tmp_length,
    char **                             buffer,
    size_t *                            length,
    int                                 delim_count,
    const char *                        delim_set,
    const char **                       replace_set)
{
    const char *                        locator;
    const char *                        prev;
    char *                              new_buffer = NULL;
    size_t                              new_length = 0;
    int                                 upto_length, locator_length; 
    int                                 replace_length;
    int                                 found_index, delim_index;

    locator = tmp_buffer;
    locator_length = tmp_length;
    prev = tmp_buffer;
    while(1)
    {
        if(locator_length <= 0)
        {
            break;
        }

        locator = my_strpbrk(locator, locator_length, &found_index, delim_set, &delim_index, 0);
        if(!locator)
        {
            upto_length = locator_length;
            new_buffer = realloc(new_buffer, new_length + upto_length);
            memcpy(new_buffer + new_length, prev, upto_length);
            new_length += upto_length;
            break;
        }

        upto_length = found_index;
        locator_length -= (found_index + 1);

        replace_length = strlen(replace_set[delim_index]);
        new_buffer = realloc(new_buffer, 
                             new_length + replace_length + upto_length);
        memcpy(new_buffer + new_length, prev, upto_length);
        memcpy(new_buffer + new_length + upto_length, replace_set[delim_index],
               replace_length);
        new_length += upto_length + replace_length;

        locator++;
        prev = locator;
        
    }

    *buffer = new_buffer;
    *length = new_length;

    return 0;
}
