/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */


#include "globus_common.h"
#include "globus_wsdl_parser.h"
#include "globus_wsdl_config.h"

#define SHORT_USAGE_FORMAT \
"\nSyntax: %s [-help][-P <prefix mapping>][-p <prefix file>] \\\n" \
  "           [-g <ignores ns file>][-G <ignore namespace>] \\\n" \
  "           [-n <include ns file>][-N <include namespace>] \\\n" \
  "           [-d <output dir>][-V <variable>=<value>] \\\n" \
  "           [-I <tmpl path>][-v] \\\n" \
  "           <template> <schema file>\n" 

static char * LONG_USAGE = \
"\n" \
"    Options\n" \
"    -help, -usage             Display usage info\n" \
"    -version                  Display version info\n" \
"    -P <prefix mapping>       Specify a Namespace to Prefix mapping\n"
"                              on the command line.  Multiple -P args allowed.\n" \
"                              Formatting: <prefix>=<namespace>\n" \
"    -p <file>                 File with Namespace to Prefix mappings\n" \
"    -g <file>                 File with ignorable Namespaces\n" \
"    -G <namespace>            ignorable namespace\n" \
"    -n <file>                 File with include-only Namespaces\n" \
"    -N <namespace>            include-only namespace\n" \
"    -I <path to templates>    Add to the list of paths to search for template files\n" \
"    -V <variable>=<value>     variable that will get loaded into\n" \
"                              the JS runtime.\n" \
"    -d <output directory>     Directory to output the generated package\n" \
"    -v                        Verbose Output.\n" \
"    <template>                Template file used to generate bindings.\n" \
"    <schema files>            The WSDL schema files to be parsed.\n" \
"\n" ;

#   define args_show_version() \
    { \
        char buf[64]; \
        sprintf( buf, \
                 "%s-%s", \
                 PACKAGE, \
                 VERSION); \
        fprintf(stderr, "%s\n", buf); \
        globus_module_deactivate_all(); \
        exit(0); \
    }

#   define args_show_short_help() \
    { \
        fprintf(stderr, \
                SHORT_USAGE_FORMAT \
                "\nOption -help will display usage.\n", \
                program); \
        globus_module_deactivate_all(); \
    }

#   define args_show_full_usage() \
    { \
        fprintf(stderr, SHORT_USAGE_FORMAT \
                "%s", \
                program, \
                LONG_USAGE); \
        globus_module_deactivate_all(); \
        exit(0); \
    }

#   define args_error_message(errmsg) \
    { \
        fprintf(stderr, "ERROR: %s\n", errmsg); \
        args_show_short_help(); \
        globus_module_deactivate_all(); \
        exit(1); \
    }

#   define args_error(argnum, argval, errmsg) \
    { \
        char buf[1024]; \
        sprintf(buf, "argument #%d (%s) : %s", argnum, argval, errmsg); \
        args_error_message(buf); \
    }

#   define args_verify_next(argnum, argval, errmsg) \
    { \
        if ((argnum+1 >= argc) || (argv[argnum+1][0] == '-')) \
            args_error(argnum,argval,errmsg); \
    }

static
int
get_quoted_variable_def(
    char *                              argv[], 
    int                                 argc,
    int *                               start_arg_index, 
    char *                              quoted_var)
{
    int                                 arg_index = *start_arg_index;
    int                                 len = strlen(argv[arg_index]);
    int                                 new_len = 0;

    if(strchr(argv[arg_index], '"'))
    {
        memcpy(quoted_var, argv[arg_index], len);
        ++arg_index;
        
        while(!index(argv[arg_index], '"') && arg_index < argc)
        {
            new_len = strlen(argv[arg_index]);
            memcpy(quoted_var + len, argv[arg_index], new_len + 1);
            quoted_var[len + new_len] = ' ';
            len += new_len + 1;
            ++arg_index;
        }

        if(arg_index == argc)
        {
            return -1;
        }
        
        new_len = strlen(argv[arg_index]);
        memcpy(quoted_var + len, argv[arg_index], new_len + 1);
        quoted_var[len + new_len] = ' ';
    }
    else
    {
        memcpy(quoted_var, argv[arg_index], len);
    }

    *start_arg_index = arg_index;
    quoted_var[len + new_len] = '\0';
    return len + new_len;
}

int main(int argc, char * argv[])
{
    char *                              program = NULL;
    char *                              argp;
    int                                 arg_index;
    char *                              outd = NULL;
    char *                              rpath = NULL;
    globus_hashtable_t                  props = NULL;
    globus_hashtable_t                  vars = NULL;
    globus_hashtable_t                  ignores = NULL;
    globus_hashtable_t                  includes = NULL;
    globus_list_t *                     template_paths = NULL;
    char *                              template_file = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    int                                 first = 1;
    int                                 second = 1;
    int                                 res = 0;
    int                                 verbose = 0;
    globus_list_t *                     schema_files = NULL;
    globus_module_activate(GLOBUS_WSDL_PARSER_MODULE);

    globus_wsdl_config_properties_init(&vars);
    globus_wsdl_config_properties_init(&props);
    globus_hashtable_init(&ignores, 
                          10,
                          globus_hashtable_string_hash,
                          globus_hashtable_string_keyeq);

    /* get the program name */
    if (strrchr(argv[0], '/'))
    {
        program = strrchr(argv[0], '/') + 1;
    }
    else
    {
        program = argv[0];
    }

    /* parse the arguments */
    for(arg_index = 1; arg_index < argc; ++arg_index)
    {
        argp = argv[arg_index];

        if (strncmp(argp, "--", 2) == 0)
        {
            if (argp[2] != '\0')
            {
                args_error(arg_index, argp, 
                           "double-dashed options are not allowed");
            }
            else
            {
                /* no more parsing */
                arg_index = argc + 1;
                continue;
            }
        }
        if((strcmp(argp, "-help") == 0) ||
           (strcmp(argp, "-usage") == 0))
        {
            args_show_full_usage();
        }
        else if(strcmp(argp, "-version") == 0)
        {
            args_show_version();
        }
        else if(strcmp(argp, "-P") == 0)
        {
            args_verify_next(arg_index, argp, "need a mapping argument");
            result = globus_wsdl_config_set_property(&props, argv[++arg_index]);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  error_str =
                globus_error_print_chain(globus_error_get(result));
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n", 
                        arg_index, error_str);
                exit(1);
            }
        }
        else if(strcmp(argp, "-p") == 0)
        {
            args_verify_next(arg_index, argp, "need a mapping file arg");
            result = globus_wsdl_config_load_properties(argv[++arg_index],
                                                        &props);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  error_str =
                globus_error_print_chain(globus_error_get(result));
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n", 
                        arg_index, error_str);
                exit(1);
            }
        }
        else if(strcmp(argp, "-G") == 0)
        {
            char *                      ignore_ns;
            args_verify_next(arg_index, argp, "need a namespace argument");
            ignore_ns = globus_libc_strdup(argv[++arg_index]);
            result = globus_hashtable_insert(&ignores, 
                                             ignore_ns,
                                             ignore_ns);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  error_str =
                globus_error_print_chain(globus_error_get(result));
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n", 
                        arg_index, error_str);
                exit(1);
            }
        }
        else if(strcmp(argp, "-g") == 0)
        {
            char                        ignore_ns[500];
            char *                      ignore_ns_entry = NULL;
            char *                      res = NULL;
            char *                      tmp_index = NULL;
            FILE *                      ignore_fd;
            char *                      ignore_file;

            args_verify_next(arg_index, argp, "need a ignores file");

            ignore_file = argv[++arg_index];

            if((ignore_fd = fopen(ignore_file, "r")) == NULL)
            {
                fprintf(stderr, "\nERROR failed to open %s\n\n", ignore_file);
                exit(1);
            }
            
            memset(ignore_ns, 0, 500);
            res = fgets(ignore_ns, 500, ignore_fd);
            
            while(res != NULL)
            {
                ignore_ns_entry = globus_libc_strdup(ignore_ns);
                tmp_index = memchr(ignore_ns_entry, '\n', strlen(ignore_ns_entry));
                if(tmp_index != NULL)
                {
                    *tmp_index = 0;
                }

                if(*ignore_ns_entry == '\0')
                {
                    globus_libc_free(ignore_ns_entry);
                    res = fgets(ignore_ns, 500, ignore_fd);
                    
                    continue;
                }

                result = globus_hashtable_insert(&ignores, 
                                                 ignore_ns_entry, 
                                                 ignore_ns_entry);
                if(result != GLOBUS_SUCCESS)
                {
                    fprintf(stderr, 
                            "\nERROR adding ignorable namespaces "
                            "from file %s\n\n", ignore_file);
                    fclose(ignore_fd);
                    exit(1);
                }
                
                memset(ignore_ns, 0, 500);
                res = fgets(ignore_ns, 500, ignore_fd);
            }

            fclose(ignore_fd);
        }
        else if(strcmp(argp, "-N") == 0)
        { 
            char *                      include_ns;
            args_verify_next(arg_index, argp, "need a namespace argument");
            include_ns = globus_libc_strdup(argv[++arg_index]);
            if(!includes)
            {
                globus_hashtable_init(
                    &includes, 
                    30,
                    globus_hashtable_string_hash,
                    globus_hashtable_string_keyeq);
            }

            result = globus_hashtable_insert(&includes, 
                                             include_ns,
                                             include_ns);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  error_str =
                globus_error_print_chain(globus_error_get(result));
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n", 
                        arg_index, error_str);
                exit(1);
            }
        }
        else if(strcmp(argp, "-n") == 0)
        {
            char                        include_ns[500];
            char *                      include_ns_entry = NULL;
            char *                      res = NULL;
            char *                      tmp_index = NULL;
            FILE *                      include_fd;
            char *                      include_file;

            args_verify_next(arg_index, argp, "need a ignores file");

            include_file = argv[++arg_index];

            if((include_fd = fopen(include_file, "r")) == NULL)
            {
                fprintf(stderr, "\nERROR failed to open %s\n\n", include_file);
                exit(1);
            }
            
            memset(include_ns, 0, 500);
            res = fgets(include_ns, 500, include_fd);
            
            if(!includes)
            {
                globus_hashtable_init(&includes,
                                      30,
                                      globus_hashtable_string_hash,
                                      globus_hashtable_string_keyeq);
            }

            while(res != NULL)
            {
                include_ns_entry = globus_libc_strdup(include_ns);
                tmp_index = memchr(include_ns_entry, '\n', strlen(include_ns_entry));
                if(tmp_index != NULL)
                {
                    *tmp_index = 0;
                }

                if(*include_ns_entry == '\0')
                {
                    globus_libc_free(include_ns_entry);
                    res = fgets(include_ns, 500, include_fd);
                    
                    continue;
                }

                result = globus_hashtable_insert(&includes, 
                                                 include_ns_entry, 
                                                 include_ns_entry);
                if(result != GLOBUS_SUCCESS)
                {
                    fprintf(stderr, 
                            "\nERROR adding ignorable namespaces "
                            "from file %s\n\n", include_file);
                    fclose(include_fd);
                    exit(1);
                }
                
                memset(include_ns, 0, 500);
                res = fgets(include_ns, 500, include_fd);
            }

            fclose(include_fd);
        }
        else if(strcmp(argp, "-d") == 0)
        {
            args_verify_next(arg_index, argp, "need a output directory");
            outd = globus_libc_strdup(argv[++arg_index]);
        }
        else if(strcmp(argp, "-V") == 0)
        {
            static char                 quoted_var[5000];
            args_verify_next(arg_index, argp, "need a variable");
            ++arg_index;
            if(get_quoted_variable_def(argv, argc, 
                                       &arg_index, quoted_var) < 0)
            {
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n",
                        arg_index, quoted_var);
                exit(1);
            }

            result = globus_wsdl_config_set_property(&vars, quoted_var);
            if(result != GLOBUS_SUCCESS)
            {
                char *                  error_str =
                globus_error_print_chain(globus_error_get(result));
                fprintf(stderr, "\nERROR with arg %d:\n\n%s\n", 
                        arg_index, error_str);
                exit(1);
            }
        }
        else if(strcmp(argp, "-I") == 0)
        {
            args_verify_next(arg_index, argp, "need a path");
            globus_list_insert(&template_paths,
                globus_libc_strdup(argv[++arg_index]));
        }
        else if(strcmp(argp, "-v") == 0)
        {
            verbose=1;
        }
        else if(first)
        {
            first = 0;
            template_file = globus_libc_strdup(argv[arg_index]);
        }
        else if(second)
        {
            globus_list_insert(&schema_files,
                globus_libc_strdup(argv[arg_index]));
        }
        else
        {
            args_error(arg_index, argp, "unrecognized argument");
        }
    }

    if(first)
    {
        args_error_message("Must specify a template file");
        args_show_short_help();
        exit(1);
    }

    result = globus_wsdl_execute(schema_files, template_file, NULL, outd,
                                 props, vars, template_paths, 
                                 ignores, includes,
                                 1024L * 1024L * 16L * 32L);
    
    if(result != GLOBUS_SUCCESS)
    {
        char *                          error_str;
        globus_object_t *               error_obj;
        error_obj = globus_error_get(result);
        error_str = globus_error_print_friendly(error_obj);
        fprintf(stderr, "\nPARSER ERROR:\n\n%s\n\n", error_str);
        
        free(error_str);
        globus_object_free(error_obj);
        
        res = 1;
        goto exit;
    }
    
    globus_list_destroy_all(template_paths, free);
    globus_wsdl_config_properties_destroy(props);
    globus_wsdl_config_properties_destroy(vars);
    globus_hashtable_destroy_all(&ignores, free);

    globus_module_deactivate_all();

 exit:

    if(outd)
    {
        free(outd);
    }

    if(rpath)
    {
        free(rpath);
    }

    if(schema_files)
    {
        globus_list_destroy_all(schema_files, free);
    }

    if(template_file)
    {
        free(template_file);
    }

    return res;
}
