/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#ifndef GLOBUS_I_WSRF_RESOURCE_H
#define GLOBUS_I_WSRF_RESOURCE_H 1

#include "globus_common.h"
#include "globus_soap_message_handle.h"
#include "globus_wsrf_resource.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * Resource property structure implementation.
 */
typedef struct globus_resource_property_s
{
    /** QName of the ResourceProperty element. */
    xsd_QName                           name;
    /**
     * XSD Type Information for this resource property.
     */
    globus_xsd_type_info_t              type_info;
    /**
     * Resource property value.
     */
    void *                              property;
    /**
     * Resource property callback for dynamic values.
     */
    globus_resource_property_callback_t callback;
    /**
     * User parameter to callback.
     */
    void *                              callback_arg;
}
globus_resource_property_t;

/**
 * Resource handle implementation.
 */
struct globus_resource_s
{
    /** Unique ID for the resource */
    char *                              id;
    /** Reference count for the resource */
    int                                 references;
    /**
     * Hashtable of resource properties (mapping xsd_QName :
     * globus_resource_property_t)
     */
    globus_hashtable_t                  resource_properties;
    /**
     * Service-specific type registry.
     */
    globus_xsd_type_registry_t          registry;
    /**
     * Resource-specific data.
     */
    void *                              data;
    /**
     * Resource-specific data destructor.
     */
    globus_resource_destroy_t           data_destroy_func;
    /**
     * Reference count destruction condition variable.
     */
    globus_cond_t                       cond;
    /**
     * Soft-state lifetime callback handle.
     */
    globus_callback_handle_t            lifetime_callback;
    /**
     * Flag indicating that a destroy call is blocking.
     */
    globus_bool_t                       destroy_blocking;
};

#define GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY() \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_OUT_OF_MEMORY, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "Out of memory"))

#define GLOBUS_RESOURCE_ERROR_NULL_PARAM() \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "NULL parameter"))

#define GLOBUS_RESOURCE_ERROR_DUPLICATE_RESOURCE(id) \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_RESOURCE, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "Resource named \"%s\" already exists", \
            id))

#define GLOBUS_RESOURCE_ERROR_UNKNOWN_RESOURCE(id) \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_RESOURCE, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "Resource named \"%s\" does not exist", \
            id))

#define GLOBUS_RESOURCE_ERROR_DUPLICATE_PROPERTY(qname)         \
    globus_error_put(                                           \
        globus_error_construct_error(                           \
            GLOBUS_WSRF_RESOURCE_MODULE,                        \
            NULL,                                               \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_PROPERTY, \
            __FILE__,                                           \
            _globus_func_name,                                  \
            __LINE__,                                           \
            "Resource Property: {%s}%s already exists",         \
            (qname ?                                            \
                    (qname->Namespace                           \
                        ? qname->Namespace                      \
                        : "Null")                               \
                   : "Null"),                                   \
            (qname ?                                            \
                    (qname->local                               \
                        ? qname->local                          \
                        : "Null")                               \
                   : "Null")))

#define GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname)           \
    globus_error_put(                                           \
        globus_error_construct_error(                           \
            GLOBUS_WSRF_RESOURCE_MODULE,                        \
            NULL,                                               \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY,   \
            __FILE__,                                           \
            _globus_func_name,                                  \
            __LINE__,                                           \
            "Resource Property: {%s}%s does not exist",         \
            (qname ?                                            \
                    (qname->Namespace                           \
                        ? qname->Namespace                      \
                        : "Null")                               \
                   : "Null"),                                   \
            (qname ?                                            \
                    (qname->local                               \
                        ? qname->local                          \
                        : "Null")                               \
                   : "Null")))

#define GLOBUS_RESOURCE_ERROR_ALREADY_DESTROYED() \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_ALREADY_DESTROYED, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "Resource already destroyed"))

#define GLOBUS_RESOURCE_ERROR_NO_REFERENCES(id) \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NO_REFERENCES, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "No references to resource exist"))

#define GLOBUS_RESOURCE_ERROR_INVALID_ANY() \
    globus_error_put( \
        globus_error_construct_error( \
            GLOBUS_WSRF_RESOURCE_MODULE, \
            NULL, \
            GLOBUS_WSRF_RESOURCE_ERROR_TYPE_INVALID_ANY, \
            __FILE__, \
            _globus_func_name, \
            __LINE__, \
            "Invalid property type or callback for any property"))

#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

#endif /* GLOBUS_I_WSRF_RESOURCE_H */
