/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_soap_message.h"
#include "globus_i_wsrf_resource.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file globus_wsrf_resource.c
 *
 * Module implementation and resource creation/destruction functions.
 */
static
int
globus_l_wsrf_resource_activate(void);

static
int
globus_l_wsrf_resource_deactivate(void);

static
void
globus_l_wsrf_resource_hashtable_destroy(
    void *                              datum);

static
void
globus_l_resource_destroy(
    globus_resource_t                   resource);

static
void
globus_l_resource_property_destroy(
    void *                              datum);

static
void
globus_l_resource_lifetime_expired(
    void *                              arg);

/**
 * Table mapping resource id strings to resource instances.
 */
static globus_hashtable_t globus_l_wsrf_resources;
/**
 * Lock for resource table.
 */
static globus_mutex_t globus_l_wsrf_resource_lock;

globus_module_descriptor_t
globus_i_wsrf_resource_module = 
{
    "globus_wsrf_resource",
    globus_l_wsrf_resource_activate,
    globus_l_wsrf_resource_deactivate
};

/**
 * Activate the WSRF Resource module.
 *
 * Activate modules this one uses, create the resource ID : resource hashtable
 * and lock.
 *
 * @retval GLOBUS_SUCCESS
 *     Module successfully activated.
 * @retval 1
 *     Unable to activate module.
 */
static
int
globus_l_wsrf_resource_activate(void)
{
    int rc;

    rc = globus_module_activate(GLOBUS_COMMON_MODULE);
    if (rc != GLOBUS_SUCCESS)
    {
        goto error;
    }
    rc = globus_module_activate(GLOBUS_SOAP_MESSAGE_MODULE);

    if (rc != GLOBUS_SUCCESS)
    {
        goto deactivate_common_error;
    }

    rc = globus_hashtable_init(&globus_l_wsrf_resources,
            16,
            globus_hashtable_string_hash,
            globus_hashtable_string_keyeq);

    if (rc != GLOBUS_SUCCESS)
    {
        goto deactivate_message_error;
    }

    rc = globus_mutex_init(&globus_l_wsrf_resource_lock, NULL);
    if (rc != GLOBUS_SUCCESS)
    {
        goto destroy_hashtable_error;
    }
    
    return GLOBUS_SUCCESS;

destroy_hashtable_error:
    globus_hashtable_destroy(&globus_l_wsrf_resources);
deactivate_message_error:
    globus_module_deactivate(GLOBUS_SOAP_MESSAGE_MODULE);
deactivate_common_error:
    globus_module_deactivate(GLOBUS_COMMON_MODULE);
error:
    return 1;
}
/* globus_l_wsrf_resource_activate() */

static
int
globus_l_wsrf_resource_deactivate(void)
{
    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    globus_hashtable_destroy_all(
        &globus_l_wsrf_resources,
        globus_l_wsrf_resource_hashtable_destroy);

    globus_mutex_unlock(&globus_l_wsrf_resource_lock);
    globus_mutex_destroy(&globus_l_wsrf_resource_lock);
    globus_module_deactivate(GLOBUS_SOAP_MESSAGE_MODULE);
    globus_module_deactivate(GLOBUS_COMMON_MODULE);

    return 0;
}
/* globus_l_wsrf_resource_deactivate() */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

/**
 * Create a new WSRF Resource.
 * @ingroup resource
 *
 * Allocate a new reference-counted WSRF resource. Calls to 
 * globus_resource_find() and globus_resource_finish() manipulate the resource
 * count. Every call to globus_resource_find() must be paired with a call to
 * globus_resource_finish(). The service which
 * calls globus_resource_create() must call globus_resource_destroy() or
 * globus_resource_set_destroy_time() to destroy the resource.
 *
 * @param id
 *     Identifier to use for the new resource.
 * @param resource
 *     Pointer to be set to the newly created resource handle.
 *
 * @retval GLOBUS_SUCCESS
 *     New resource created successfully.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_RESOURCE
 *     Another resource with the specified id already exists, so a new one
 *     could not be created.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_OUT_OF_MEMORY
 *     Insufficient memory to create the new resource.
 *
 * @see globus_resource_find(), globus_resource_finish()
 */
globus_result_t
globus_resource_create(
    const char *                        id,
    globus_resource_t *                 resource)
{
    globus_resource_t                   iresource;
    globus_result_t                     result = GLOBUS_SUCCESS;
    int                                 rc;
    GlobusFuncName(globus_resource_create);

    if (id == NULL || resource == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_NULL_PARAM();
        goto error;
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    iresource = globus_hashtable_lookup(&globus_l_wsrf_resources, (void *) id);

    if (iresource != NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_DUPLICATE_RESOURCE(id);

        goto unlock_error;
    }

    iresource = globus_calloc(1, sizeof(struct globus_resource_s));
    if (iresource == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto unlock_error;
    }

    iresource->id = globus_libc_strdup(id);
    if (iresource->id == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto free_resource_error;
    }

    rc = globus_hashtable_init(&iresource->resource_properties,
            31,
            xsd_QName_hash,
            xsd_QName_keyeq);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto free_id_out;
    }

    globus_cond_init(&iresource->cond, NULL);
    iresource->lifetime_callback = GLOBUS_NULL_HANDLE;
    iresource->destroy_blocking = GLOBUS_FALSE;

    rc = globus_hashtable_insert(
            &globus_l_wsrf_resources,
            iresource->id,
            iresource);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto destroy_resource_error;
    }

    *resource = iresource;
    
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);
    return result;

destroy_resource_error:
    globus_hashtable_destroy(&iresource->resource_properties);
    globus_cond_destroy(&iresource->cond);
free_id_out:
    globus_libc_free(iresource->id);
free_resource_error:
    globus_libc_free(iresource);
unlock_error:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);
error:
    return result;
}
/* globus_resource_create() */

/**
 * Find the resource associated with a resource identifier.
 * @ingroup resource
 *
 * Returns a reference to the resource associated with the id.
 * Destroy this reference by calling #globus_resource_finish().
 *
 * @param id
 *     Resource identifier for the desired resource.
 * @param resource
 *     Pointer to a resource which will modified to point to the
 *     named resource if it is found.
 *
 * @retval GLOBUS_SUCCESS
 *     The resource associated with the specified id was found and a reference
 *     to it was returned in the @a resource parameter.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_RESOURCE
 *     No resource with the specified id exists.
 * @see #globus_resource_create()
 */
globus_result_t
globus_resource_find(
    const char *                        id,
    globus_resource_t *                 resource)
{
    globus_resource_t                   iresource;
    globus_result_t                     result;
    GlobusFuncName(globus_resource_find);

    if (id == NULL || resource == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    iresource = globus_hashtable_lookup(&globus_l_wsrf_resources, (void *) id);

    if (iresource == NULL || iresource->destroy_blocking)
    {
        result = GLOBUS_RESOURCE_ERROR_UNKNOWN_RESOURCE(id);

        goto unlock_error;
    }
    iresource->references++;

    *resource = iresource;
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);

    return GLOBUS_SUCCESS;
unlock_error:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);

    return result;
}
/* globus_resource_find() */

/**
 * Release a reference to a resource.
 * @ingroup resource
 *
 * The caller must no longer access the resource reference once this function
 * returns. This function destroys the resource when all references to the
 * resource have been finished and globus_resource_destroy() or the time
 * passed to globus_resource_set_destroy_time() has elapsed.
 *
 * @param resource
 *     Resource handle which is no longer being used.
 *
 * @retval GLOBUS_SUCCESS
 *     Successfully finished using the resource handle.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null resource handle passed to this function.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NO_REFERENCES
 *     No references to resource exist.
 *
 * @see #globus_resource_find(), #globus_resource_create()
 */
globus_result_t
globus_resource_finish(
    globus_resource_t                   resource)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    GlobusFuncName(globus_resource_finish);

    if (resource == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    if (resource->references <= 0)
    {
        result = GLOBUS_RESOURCE_ERROR_NO_REFERENCES(resource->id);

        goto out;
    }
    resource->references--;

    globus_assert(resource->references >= 0);

    if (resource->references == 0)
    {
        if (resource->destroy_blocking)
        {
            globus_cond_signal(&resource->cond);
        }
    }
out:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);

    return result;
}
/* globus_resource_finish() */

/**
 * Destroy the resource after the final reference has been finished.
 * @ingroup resource
 *
 * This function blocks until all references to the resource have been
 * finished and then destroys the contents of the resource. This destruction
 * will take precedence over a not yet elapsed delayed destruction time set by
 * #globus_resource_set_destroy_time(). It is an
 * error to access the resource once this function returns. All subsequent
 * attempts to find this resource will fail.
 * 
 *
 * @param resource
 *     Resource handle which is no longer being used.
 *
 * @retval GLOBUS_SUCCESS
 *     Successfully destroyed the resource handle.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null resource handle passed to this function.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_ALREADY_DESTROYED
 *     The resource has already been destroyed.
 *
 * @see globus_resource_create(), globus_resource_finish_and_destroy(),
 * #globus_resource_set_destroy_time()
 */
globus_result_t
globus_resource_destroy(
    globus_resource_t                   resource)
{
    globus_result_t                     result;
    globus_bool_t                       active;
    GlobusFuncName(globus_resource_destroy);

    if (resource == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);
    if (resource->destroy_blocking)
    {
        result = GLOBUS_RESOURCE_ERROR_ALREADY_DESTROYED();

        goto out;
    }

    if (resource->lifetime_callback != GLOBUS_NULL_HANDLE)
    {
        /* Try to cancel the lifetime callback */
        result = globus_callback_unregister(
            resource->lifetime_callback,
            NULL,
            NULL,
            &active);

        if (result != GLOBUS_SUCCESS || active)
        {
            /* Callback can't be stopped---we'll rely on it to destroy the
             * service
             */
            goto out;
        }
        resource->lifetime_callback = GLOBUS_NULL_HANDLE;
    }

    resource->destroy_blocking = GLOBUS_TRUE;

    while (resource->references > 0)
    {
        globus_cond_wait(&resource->cond, &globus_l_wsrf_resource_lock);
    }
    (void) globus_hashtable_remove(&globus_l_wsrf_resources,
            resource->id);

    globus_l_resource_destroy(resource);

out:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);
    return GLOBUS_SUCCESS;
}
/* globus_resource_destroy() */

/**
 * Schedule delayed destruction of a resource.
 * @ingroup resource
 *
 * When the time indicated by the @a timestamp parameter has been reached, the
 * resource will be destroyed once all remaining references to the resource
 * have been finished.
 *
 * If globus_resource_destroy() is called before this time has been reached,
 * the timestamp is effectively ignored. 
 * 
 * @param resource
 *     Resource handle which is no longer being used.
 * @param timestamp
 *     Time when the resource should be destroyed.
 *
 * @retval GLOBUS_SUCCESS
 *     Successfully set the destroy timeout for the resource handle.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null resource handle passed to this function.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_ALREADY_DESTROYED
 *     The resource has already been destroyed.
 *
 * @see #globus_resource_create(), #globus_resource_destroy(),
 * #globus_resource_finish_and_destroy()
 */
globus_result_t
globus_resource_set_destroy_time(
    globus_resource_t                   resource,
    const globus_abstime_t *            timestamp)
{
    globus_result_t                     result;
    globus_bool_t                       active;
    globus_abstime_t                    now;
    globus_reltime_t                    delay;
    GlobusFuncName(globus_resource_set_destroy_time);

    if (resource == NULL || timestamp == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    if (resource->destroy_blocking)
    {
        result = GLOBUS_RESOURCE_ERROR_ALREADY_DESTROYED();

        goto out;
    }

    if (resource->lifetime_callback != GLOBUS_NULL_HANDLE)
    {
        /* Try to cancel the previous instance of the lifetime callback */
        result = globus_callback_unregister(
            resource->lifetime_callback,
            NULL,
            NULL,
            &active);

        if (result != GLOBUS_SUCCESS || active)
        {
            /* Callback can't be stopped---assume this resource is already
             * going to be destroyed very soon.
             */
            result = GLOBUS_RESOURCE_ERROR_ALREADY_DESTROYED();

            goto out;
        }
        resource->lifetime_callback = GLOBUS_NULL_HANDLE;
    }

    GlobusTimeAbstimeSet(now, 0, 0);

    if (globus_abstime_cmp(&now, timestamp) >= 0)
    {
        delay.tv_sec = 0;
        delay.tv_usec = 0;
    }
    else
    {
        GlobusTimeAbstimeDiff(delay, *timestamp, now);
    }

    result = globus_callback_register_oneshot(
        &resource->lifetime_callback,
        &delay,
        globus_l_resource_lifetime_expired,
        resource);

out:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);

    return result;
}
/* globus_resource_set_destroy_time() */

/**
 * Finish one reference and then destroy the resource.
 * @ingroup resource
 *
 * This function acts identically to globus_resource_destroy() except it
 * first finishes one reference to the resource. This allows applications to
 * use code like this:
 * @code
 *     globus_resource_find(resource_id, &resource);
 *     ...
 *     globus_resource_finish_and_destroy(resource);
 * @endcode
 * to avoid a race condition between calling globus_resource_finish() and 
 * globus_resource_destroy() during which the reference to the resource
 * would be invalid.
 * 
 * If another thread is currently blocked trying to destroy this resource, then
 * this function acts like #globus_resource_finish().
 *
 * @param resource
 *     Resource handle which is no longer being used.
 *
 * @retval GLOBUS_SUCCESS
 *     Successfully finished using the resource handle.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null resource handle passed to this function.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_ALREADY_DESTROYED
 *     The resource has already been destroyed.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NO_REFERENCES
 *     No references to resource exist.
 *
 * @see #globus_resource_create(), #globus_resource_destroy(),
 * #globus_resource_set_destroy_time()
 */
globus_result_t
globus_resource_finish_and_destroy(
    globus_resource_t                   resource)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_bool_t                       active;
    GlobusFuncName(globus_resource_finish_and_destroy);

    if (resource == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    if (resource->references <= 0)
    {
        result = GLOBUS_RESOURCE_ERROR_NO_REFERENCES(resource->id);

        goto out;
    }

    resource->references--;

    if (resource->destroy_blocking)
    {
        /* If another thread is calling destroy, this acts like
         * globus_resource_finish()
         */
        goto out;
    }

    if (resource->lifetime_callback != GLOBUS_NULL_HANDLE)
    {
        /* Try to cancel the lifetime callback */
        result = globus_callback_unregister(
            resource->lifetime_callback,
            NULL,
            NULL,
            &active);

        if (result != GLOBUS_SUCCESS || active)
        {
            /* Callback can't be stopped---we'll rely on it to destroy the
             * service
             */
            goto out;
        }
        resource->lifetime_callback = GLOBUS_NULL_HANDLE;
    }

    resource->destroy_blocking = GLOBUS_TRUE;

    while (resource->references > 0)
    {
        globus_cond_wait(&resource->cond, &globus_l_wsrf_resource_lock);
    }
    (void) globus_hashtable_remove(&globus_l_wsrf_resources,
            resource->id);

    globus_l_resource_destroy(resource);

out:
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);

    return result;
}
/* globus_resource_finish_and_destroy() */

/**
 * Retrieve resource-specific data from a resource handle.
 * @ingroup resource
 *
 * @param resource
 *     Resource handle to query for data.
 * @param data
 *     Pointer to be set to the value of resource-specific data.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource specific data successfully retrieved.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function. The value of @a data is
 *     undefined when this value is returned.
 *
 * @see #globus_resource_set_resource_specific()
 */
globus_result_t
globus_resource_get_resource_specific(
    const globus_resource_t             resource,
    void **                             data)
{
    GlobusFuncName(globus_resource_get_resource_specific);

    if (resource == NULL || data == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }
    *data = resource->data;

    return GLOBUS_SUCCESS;
}
/* globus_resource_get_resource_specific() */

/**
 * Set resource-specific data for a resource handle.
 * @ingroup resource
 *
 * @param resource
 *     Resource handle to modify.
 * @param data
 *     New value of the resource-specific data. If a non-NULL value was
 *     previously set as resource-specific data, it will be destroyed.
 * @param destroy_func
 *     Function pointer of a function to call to destroy the resource-specific
 *     data. This will be called either when a new instance of
 *     resource-specific data is set by calling this function again, or when
 *     the resource is destroyed because of globus_resource_finish() being
 *     called.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource specific data successfully set.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function. 
 *
 * @see #globus_resource_get_resource_specific(), #globus_resource_finish()
 */
globus_result_t
globus_resource_set_resource_specific(
    globus_resource_t                   resource,
    void *                              data,
    globus_resource_destroy_t           destroy_func)
{
    GlobusFuncName(globus_resource_set_resource_specific);

    if (resource == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    if ((resource->data != NULL) && (resource->data_destroy_func != NULL))
    {
        (resource->data_destroy_func)(resource->data);
    }

    resource->data = data;
    resource->data_destroy_func = destroy_func;

    return GLOBUS_SUCCESS;
}
/* globus_resource_set_resource_specific() */

/**
 * Retrieve resource-specific type registry from a resource handle.
 * @ingroup resource
 *
 * @param resource
 *     Resource handle to query for data.
 * @param registry
 *     Pointer to be set to the value of type registry.
 *
 * @retval GLOBUS_SUCCESS
 *     Type registry successfully retrieved.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function. The value of @a registry is
 *     undefined when this value is returned.
 *
 * @see #globus_resource_set_type_registry()
 */
globus_result_t
globus_resource_get_type_registry(
    globus_resource_t                   resource,
    globus_xsd_type_registry_t *        registry)
{
    GlobusFuncName(globus_resource_get_type_registry);

    if (resource == NULL || registry == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }
    *registry = resource->registry;

    return GLOBUS_SUCCESS;
}
/* globus_resource_get_type_registry() */

/**
 * Set type registry associated with a resource handle.
 * @ingroup resource
 *
 * @param resource
 *     Resource handle to modify.
 * @param registry
 *     New value for the resource's type registry.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource's type registry successfully set.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     Null parameter passed to this function. 
 *
 * @see #globus_resource_get_type_registry()
 */
globus_result_t
globus_resource_set_type_registry(
    globus_resource_t                   resource,
    globus_xsd_type_registry_t          registry)
{
    GlobusFuncName(globus_resource_set_type_registry);

    if (resource == NULL || registry == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    resource->registry = registry;

    return GLOBUS_SUCCESS;
}
/* globus_resource_set_type_registry() */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * Destroy a resource stored in the resource hashtable at deactivation time.
 * Resources could have 
 * - references from finds which were not finished
 * - other threads blocking in globus_resource_destroy()
 * - other threads blocking in a callback from
 *   globus_resource_set_destroy_time()
 * This function is called with the resource mutex locked.
 */
static
void
globus_l_wsrf_resource_hashtable_destroy(
    void *                              datum)
{
    globus_resource_t                   resource = datum;
    globus_bool_t                       active = GLOBUS_FALSE;
    
    /* If a callback is registered, try to remove it; if that fails, then
     * the callback code will try to destroy the resource, so we'll act
     * later as if a destroy call was already blocking.
     */
    if (resource->lifetime_callback != GLOBUS_NULL_HANDLE)
    {
        globus_callback_unregister(
            resource->lifetime_callback,
            NULL,
            NULL,
            &active);

        resource->lifetime_callback = GLOBUS_NULL_HANDLE;
    }

    if (active || resource->destroy_blocking)
    {
        /* If destroy_blocking is true or a callback is waiting for the lock,
         * then another thread is going to try to destroy this, or is waiting
         * for the referece count to reach 0 already.  We'll help that out.
         */
        resource->references = 0;
        globus_cond_signal(&resource->cond);

        return;
    }
    else
    {
        /* Otherwise, no destroy is in progress for this resource, so we must
         * free it ourselves.
         */
        globus_l_resource_destroy(resource);
    }
}
/* globus_l_wsrf_resource_hashtable_destroy() */

/**
 * Free memory associated with a resource handle.
 * When this function is called for a resource, all callbacks and threads
 * blocking on the condition of this resource must be finished, otherwise
 * weird things may happen in those threads.
 *
 * This function is called with the mutex locked.
 *
 * @param datum
 *     Pointer to the resource.
 */
static
void
globus_l_resource_destroy(
    globus_resource_t                   resource)
{

    if (resource->resource_properties != NULL)
    {
        globus_hashtable_destroy_all(
                &resource->resource_properties,
                globus_l_resource_property_destroy);
    }
    globus_libc_free(resource->id);

    if (resource->data && resource->data_destroy_func)
    {
        resource->data_destroy_func(resource->data);
    }
    globus_cond_destroy(&resource->cond);

    if (resource->lifetime_callback != GLOBUS_NULL_HANDLE)
    {
        globus_callback_unregister(
            resource->lifetime_callback,
            NULL,
            NULL,
            NULL);
    }

    globus_libc_free(resource);
}
/* globus_l_resource_destroy() */

/**
 * Free memory associated with a resource property.
 *
 * Function signature uses a void * to match the globus_hashtable_destroy_all()
 * prototype.
 * 
 * @param datum
 *     Pointer to a globus_resource_property_t to destroy.
 */
static
void
globus_l_resource_property_destroy(
    void *                              datum)
{
    globus_resource_property_t *        prop = datum;

    if (prop->property)
    {
        prop->type_info->destroy(prop->property);
    }
    if (prop->name.Namespace)
    {
        globus_libc_free(prop->name.Namespace);
    }
    if (prop->name.local)
    {
        globus_libc_free(prop->name.local);
    }
    globus_libc_free(prop);
}
/* globus_l_resource_property_destroy() */

static
void
globus_l_resource_lifetime_expired(
    void *                              arg)
{
    globus_resource_t                   resource = arg;

    globus_mutex_lock(&globus_l_wsrf_resource_lock);

    resource->destroy_blocking = GLOBUS_TRUE;

    while (resource->references > 0)
    {
        globus_cond_wait(&resource->cond, &globus_l_wsrf_resource_lock);
    }
    (void) globus_hashtable_remove(&globus_l_wsrf_resources,
            resource->id);

    globus_l_resource_destroy(resource);
    globus_mutex_unlock(&globus_l_wsrf_resource_lock);
}
/* globus_l_resource_lifetime_expired() */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
