/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "globus_i_wsrf_resource.h"
#include "xsd_any.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file globus_wsrf_resource_property.c
 * Resource Property Functions
 * @author $author$
 * @date $date$
 * @version $version$
 */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

static
xsd_QName globus_l_resource_properties_any_qname = {
    "any",
    "any"
};

const xsd_QName * globus_resource_properties_any_qname =
    &globus_l_resource_properties_any_qname; 
static
globus_result_t
globus_l_resource_create_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    const globus_xsd_type_info_t        type_info,
    void *                              property,
    globus_resource_property_callback_t callback,
    void *                              callback_arg);

static
globus_result_t
globus_l_resource_get_any_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void **                             property,
    globus_xsd_type_info_t *            info);

globus_result_t
globus_l_resource_set_any_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void *                              property);

/**
 * Get the value of the named Resource Property.
 * @ingroup resource_property
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Name of the Resource Property element.
 * @param property 
 *     Pointer to return the value of the property.
 * @param type_info
 *     Pointer to return the type info for this property. This may be NULL.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource Property found and its value returned in @a property.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
globus_result_t
globus_resource_get_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void **                             property,
    globus_xsd_type_info_t *            type_info)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_resource_property_t *        prop = NULL;
    GlobusFuncName(globus_resource_get_property);

    if (resource == NULL || qname == NULL || property == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    *property = NULL;

    prop = globus_hashtable_lookup(
            &resource->resource_properties,
            (void *) qname);

    if (prop != NULL)
    {
        if (prop->callback != NULL)
        {
            prop->callback(prop->callback_arg, qname, property);

            if (prop->property != NULL)
            {
                prop->type_info->destroy(prop->property);
            }
            prop->property = *property;
        }
        else
        {
            *property = prop->property;
        }
        if (type_info != NULL)
        {
            (*type_info) = prop->type_info;
        }
    }
    else
    {
        result = globus_l_resource_get_any_property(
                resource,
                qname,
                property,
                type_info);

        if (result != GLOBUS_SUCCESS)
        {
            goto out;
        }
    }

out:
    return result;
}
/* globus_resource_get_property() */

/**
 * Set the value of the named Resource Property.
 * @ingroup resource_property
 *
 * Changes the current value of the named Resource Property to be that of
 * the @a property parameter. If the Resource Property already had a value
 * set by a previous call to this function or
 * #globus_resource_create_property(), the old value will be destroyed, using
 * the functions included in the type info associated with the property.
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Name of the Resource Property element.
 * @param property 
 *     New value of the property.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource Property found and its value returned in @a property.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
globus_result_t
globus_resource_set_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void *                              property)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_resource_property_t *        prop = NULL;
    GlobusFuncName(globus_resource_set_property);

    if (resource == NULL || qname == NULL)
    {
        result =  GLOBUS_RESOURCE_ERROR_NULL_PARAM();

        goto out;
    }

    prop = globus_hashtable_lookup(
            &resource->resource_properties,
            (void*) qname);

    if (prop != NULL)
    {
        if (prop->property != NULL && prop->type_info->destroy != NULL)
        {
            prop->type_info->destroy(prop->property);
        }

        prop->property = property;
    }
    else
    {
        result = globus_l_resource_set_any_property(
                resource,
                qname,
                property);
    }

out:
    return result;
}
/* globus_resource_set_property() */

/**
 * Create a new Resource Property.
 * @ingroup resource_property
 *
 * Modifies the resource to contain a new Resource Property with the given
 * element name and type information. The initial value of the property will
 * be set to the @a property parameter's value.
 *
 * @param resource
 *     Resource with which to associated the new resource property.
 * @param qname
 *     Name of the resource property element.
 * @param type_info
 *     XSD type mapping for this resource property type. This mapping
 *     describes how to access the value contained in the @a property
 *     parameter.
 * @param property
 *     Initial value of the property.
 * 
 * @retval GLOBUS_SUCCESS
 *     Property successfully created.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_PROPERTY
 *     A Resource Property with the given QName already exists in this
 *     resource.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_OUT_OF_MEMORY
 *     Not enough memory to allocate the new Resource Property.
 */
globus_result_t
globus_resource_create_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    const globus_xsd_type_info_t        type_info,
    void *                              property)
{
    return globus_l_resource_create_property(
            resource,
            qname,
            type_info,
            property,
            NULL,
            NULL);
}
/* globus_resource_create_property() */

/**
 * Create a new dynamic-valued Resource Property.
 * @ingroup resource_property
 *
 * Modifies the resource to contain a new Resource Property with the given
 * element name and type information. The value of the property is determined
 * by calling the callback function. Subsequent calls to
 * globus_resource_get_property() will delete the previous value generated by
 * the callback.
 *
 * @param resource
 *     Resource with which to associated the new resource property.
 * @param qname
 *     Name of the resource property element.
 * @param type_info
 *     XSD type mapping for this resource property type. This mapping
 *     describes how to access the value contained in the @a property
 *     parameter.
 * @param callback
 *     Callback when creating a dynamic-valued Resource Property.
 * @param callback_arg
 *     User-specific callback argument for dynamic Resource Property
 *     value callbacks.
 * 
 * @retval GLOBUS_SUCCESS
 *     Property successfully created.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_PROPERTY
 *     A Resource Property with the given QName already exists in this
 *     resource.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_OUT_OF_MEMORY
 *     Not enough memory to allocate the new Resource Property.
 */
globus_result_t
globus_resource_create_property_callback(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    const globus_xsd_type_info_t        type_info,
    globus_resource_property_callback_t callback,
    void *                              callback_arg)
{
    GlobusFuncName(globus_resource_create_property_callback);

    if (callback == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    return globus_l_resource_create_property(
            resource,
            qname,
            type_info,
            NULL,
            callback,
            callback_arg);
}
/* globus_resource_create_property_callback() */

/**
 * Remove the named resource property from a resource.
 * @ingroup resource_property
 *
 * The property's current value and type info is returned in the @a property
 * and @a type_info parameters respectively.
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Element name of the resource property.
 * @param property
 *     Pointer to be set to point to the value of the resource property.
 * @param type_info
 *     Pointer to be set to type information about the property value.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource Property successfully deleted.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
globus_result_t
globus_resource_delete_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void **                             property,
    globus_xsd_type_info_t *            type_info)
{
    globus_resource_property_t *        prop = NULL;
    GlobusFuncName(globus_resource_delete_property);

    if (resource == NULL || qname == NULL || property == NULL ||
            type_info == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    if (resource->resource_properties != NULL)
    {
        prop = globus_hashtable_remove(
                &resource->resource_properties,
                (void*) qname);
    }

    if (prop == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);
    }

    *property = prop->property;
    *type_info = prop->type_info;

    globus_free(prop->name.Namespace);
    globus_free(prop->name.local);
    globus_free(prop);

    return GLOBUS_SUCCESS;
}
/* globus_resource_delete_property() */

/**
 * Destroy a named resource property.
 * @ingroup resource_property
 *
 * The current value of the property is destroyed. Subsequent attempts to
 * access the property will fail.
 *
 * and @a type_info parameters respectively.
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Element name of the resource property.
 * @retval GLOBUS_SUCCESS
 *     The property was successfully destroyed
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
globus_result_t
globus_resource_destroy_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname)
{
    globus_resource_property_t *        prop = NULL;
    GlobusFuncName(globus_resource_destroy_property);

    if (resource == NULL || qname == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_NULL_PARAM();
    }

    if (resource->resource_properties != NULL)
    {
        prop = globus_hashtable_remove(
                &resource->resource_properties,
                (void *) qname);
    }

    if (prop == NULL)
    {
        return GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);
    }

    if (prop->property != NULL)
    {
        prop->type_info->destroy(prop->property);
    }

    free(prop->name.Namespace);
    free(prop->name.local);
    free(prop);

    return GLOBUS_SUCCESS;
}
/* globus_resource_destroy_property() */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * Internal implementation of Resource Property creation.
 * 
 * Modifies the resource to contain a new Resource Property with the given
 * element name and type information. The initial value of the property will
 * be set to the @a property parameter's value.
 *
 * @param resource
 *     Resource with which to associated the new resource property.
 * @param qname
 *     Name of the resource property element.
 * @param type_info
 *     XSD type mapping for this resource property type. This mapping
 *     describes how to access the value contained in the @a property
 *     parameter.
 * @param property
 *     Initial value of the property.
 * @param callback
 *     Callback when creating a dynamic-valued Resource Property.
 * @param callback_arg
 *     User-specific callback argument for dynamic Resource Property
 *     value callbacks.
 * 
 * @retval GLOBUS_SUCCESS
 *     Property successfully created.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_DUPLICATE_PROPERTY
 *     A Resource Property with the given QName already exists in this
 *     resource.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_OUT_OF_MEMORY
 *     Not enough memory to allocate the new Resource Property.
 */
static
globus_result_t
globus_l_resource_create_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    const globus_xsd_type_info_t        type_info,
    void *                              property,
    globus_resource_property_callback_t callback,
    void *                              callback_arg)
{
    globus_resource_property_t *        prop;
    globus_result_t                     result;
    int                                 rc;
    GlobusFuncName(globus_l_resource_create_property);

    if (resource == NULL || qname == NULL || type_info == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_NULL_PARAM();

        goto error;
    }

    /* If the globus_resource_properties_any_qname is used for the property
     * name, it must be either an any or any_array and must not have a callback
     */
    if (xsd_QName_keyeq(
                (void *) qname,
                (void *) globus_resource_properties_any_qname) &&
        ((type_info != &xsd_any_info && type_info != &xsd_any_array_info)
            || (callback != NULL)))
    {
        result = GLOBUS_RESOURCE_ERROR_INVALID_ANY();

        goto error;
    }

    prop = globus_hashtable_lookup(
            &resource->resource_properties,
            (void *) qname);

    if (prop != NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_DUPLICATE_PROPERTY(qname);

        goto error;
    }
    prop = malloc(sizeof(struct globus_resource_property_s));

    if (prop == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto error;
    }
    
    result = xsd_QName_copy_contents(&prop->name, qname);
    if(result != GLOBUS_SUCCESS)
    {
        goto free_prop_error;
    }
    prop->type_info = type_info;
    prop->property = property;
    prop->callback = callback;
    prop->callback_arg = callback_arg;

    rc = globus_hashtable_insert(
            &resource->resource_properties,
            &prop->name,
            prop);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GLOBUS_RESOURCE_ERROR_OUT_OF_MEMORY();

        goto free_qname_error;
    }

    if (xsd_QName_keyeq(
                (void *) qname,
                (void *) globus_resource_properties_any_qname)
        && prop->property == NULL)
    {
        result = prop->type_info->initialize(&prop->property);

        if (result != GLOBUS_SUCCESS)
        {
            goto remove_from_hashtable_error;
        }
    }

    return GLOBUS_SUCCESS;

remove_from_hashtable_error:
    globus_hashtable_remove(&resource->resource_properties, &prop->name);
free_qname_error:
    xsd_QName_destroy_contents(&prop->name);
free_prop_error:
    globus_libc_free(prop);
error:
    return result;
}
/* globus_l_resource_create_property() */

/**
 * Check to see if we can match the qname in the resource's any:any property 
 * if it is present.
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Name of the Resource Property element.
 * @param property 
 *     Pointer to return the value of the property.
 * @param type_info
 *     Pointer to return the type info for this property. This may be NULL.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource Property found and its value returned in @a property.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
static
globus_result_t
globus_l_resource_get_any_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void **                             property,
    globus_xsd_type_info_t *            info)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_resource_property_t *        prop;

    prop = globus_hashtable_lookup(
            &resource->resource_properties,
            (void *) globus_resource_properties_any_qname);

    if (prop == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);

        goto out;
    }

    if (prop->type_info == &xsd_any_info)
    {
        xsd_any *                   any = prop->property;

        if (any->element != NULL &&
            !xsd_QName_keyeq((void *) any->element, (void *) qname))
        {
            result = GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);

            goto out;
        }
        *property = any;
    }
    else if (prop->type_info == &xsd_any_array_info)
    {
        xsd_any_array *             any_array = prop->property;
        int i;

        for (i = 0; i < any_array->length; i++)
        {
            if (any_array->elements[i].element == NULL)
            {
                *property = &any_array->elements[i];
                break;
            }
            else if (xsd_QName_keyeq(
                        any_array->elements[i].element,
                        (void *) qname))
            {
                *property = &any_array->elements[i];
                break;
            }
        }
        if (i == any_array->length)
        {
            xsd_any * any = xsd_any_array_push(any_array);
            any->any_info = &xsd_any_array_info;
            *property = any;
        }
    }
    if (info != NULL)
    {
        *info = &xsd_any_info;
    }
out:
    return result;
}
/* globus_l_resource_get_any_property() */

/**
 * Set the value of a named resource property stored within the special
 * "any:any" resource property.
 *
 * Changes the current value of the named Resource Property to be that of
 * the @a property parameter. If the Resource Property already had a value
 * set by a previous call to this function or
 * #globus_resource_create_property(), the old value will be destroyed, using
 * the functions included in the type info associated with the property. If
 * the @a property parameter is NULL and this is in the any resource property,
 * it will be removed.
 *
 * @param resource
 *     Resource containing the property.
 * @param qname
 *     Name of the Resource Property element.
 * @param property 
 *     New value of the property.
 *
 * @retval GLOBUS_SUCCESS
 *     Resource Property found and its value returned in @a property.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_NULL_PARAM
 *     One of the parameters to this function was NULL.
 * @retval GLOBUS_WSRF_RESOURCE_ERROR_TYPE_UNKNOWN_PROPERTY
 *     The resource does not contain a property with the specified name.
 */
globus_result_t
globus_l_resource_set_any_property(
    globus_resource_t                   resource,
    const xsd_QName *                   qname,
    void *                              property)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_resource_property_t *        prop = NULL;
    xsd_any *                           any;
    xsd_any_array *                     any_array;
    int                                 i;
    GlobusFuncName(globus_l_resource_set_any_property);

    prop = globus_hashtable_lookup(
            &resource->resource_properties,
            (void*) globus_resource_properties_any_qname);

    if (prop == NULL)
    {
        result = GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);

        goto out;
    }

    if (prop->type_info == &xsd_any_info)
    {
        /* Single any element */
        any = prop->property;

        if (any->element == NULL ||
            xsd_QName_keyeq(any->element, (void *) qname))
        {
            /* The element name matches, so we'll destroy the old value */
            if (any->value != NULL)
            {
                any->any_info->destroy(any->value);
                any->value = NULL;
            }
            /*
             * Insert the new prop value if we have one; otherwise, we remove
             * the binding of the QName to this any.
             */
            if (property != NULL)
            {
                any->value = property;
            }
            else
            {
                any->value = NULL;
                if (any->element != NULL)
                {
                    xsd_QName_destroy(any->element);
                    any->element = NULL;
                }
                any->any_info = NULL;
            }
        }
        else
        {
            /* element name doesn't match */
            result = GLOBUS_RESOURCE_ERROR_UNKNOWN_PROPERTY(qname);

            goto out;
        }
    }
    else
    {
        assert(prop->type_info == &xsd_any_array_info);
        any_array = prop->property;

        /* The value is an array of anys. We'll look for the one which
         * matches our qname, and replace its contents with the new value.
         * if property is null, we'll remove it from the array altogether
         */
        for (i = 0; i < any_array->length; i++)
        {
            any = &any_array->elements[i];

            if (any->element != NULL &&
                xsd_QName_keyeq(any->element, (void *) qname))
            {
                /* found a match */
                if (any->value != NULL)
                {
                    /* old value -- gone! */
                    any->any_info->destroy(any->value);
                    any->value = NULL;
                }
                if (property != NULL)
                {
                    /* replace with new value */
                    any->value = property;
                }
                else
                {
                    /* delete binding for this QName */
                    xsd_QName_destroy(any->element);
                    any->element = NULL;

                    if (i + 1 < any_array->length)
                    {
                        /* pop this array element out of the array */
                        memmove(
                            &any_array->elements[i],
                            &any_array->elements[i+1],
                            sizeof(any_array->elements[i]) *
                                (any_array->length - (i + 1)));
                    }
                    /* And clear value at the tail of this array */
                    memset(&any_array->elements[any_array->length-1],
                            0,
                            sizeof(any_array->elements[i]));
                    any_array->length--;
                }
                break; /* only one match will be possible */
            }
        }
    }
out:
    return result;
}
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
