//(c) University of Edinburgh 2006.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.authorization;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import uk.org.ogsadai.activity.Activity;
import uk.org.ogsadai.common.SecurityContext;
import uk.org.ogsadai.common.xml.XMLUtilities;
import uk.org.ogsadai.service.authorization.AccessAuthorizer;

/**
 * A simple file-based implementation of the <code>AccessAuthorizer</code>
 * interface which loads an access control list from a file.
 *
 * @author The OGSA-DAI Team.
 */

public class SimpleFileAuthorizer implements AccessAuthorizer {

    /** 
     * Maps a resource ID to a set of users that are authorized to access 
     * the resource.
     */
    protected Map resourceAuthorization = new HashMap();
    
    /** 
     * Maps an activity name to a set of users that are authorized to access 
     * the activity.
     */
    protected Map activityAuthorization = new HashMap();

    /**
     * Constructor. Populates the access control lists.
     * @param config location of the configuration file
     * @throws Exception
     */
    public SimpleFileAuthorizer(String config) throws Exception 
    {
        Document doc = XMLUtilities.xmlFileToDOM(config, false);
        
        // Read the list of authorized users for resource access
        NodeList resources = doc.getElementsByTagName("resource");
        for (int j=0; j<resources.getLength(); j++) 
        {
            Element resource = (Element)resources.item(j);
            String resourceID = resource.getAttribute("id");
            
            Set authorizedUsers = new HashSet();
            NodeList users = resource.getElementsByTagName("user");
            for (int i=0; i<users.getLength(); i++) 
            {
                String userDN = ((Element)users.item(i)).getAttribute("dn");
                authorizedUsers.add(userDN);
            }
            resourceAuthorization.put(resourceID, authorizedUsers);
        }
        
        // Read the list of authorized users for activity access
        NodeList activities = doc.getElementsByTagName("activity");
        for (int j=0; j<activities.getLength(); j++) 
        {
            Element activity = (Element)activities.item(j);
            String activityName = activity.getAttribute("name");
            
            Set authorizedUsers = new HashSet();
            NodeList users = activity.getElementsByTagName("user");
            for (int i=0; i<users.getLength(); i++) 
            {
                String userDN = ((Element)users.item(i)).getAttribute("dn");
                authorizedUsers.add(userDN);
            }
            activityAuthorization.put(activityName, authorizedUsers);
        }
    }

    /**
     * Authorizes access to a resource based on an access control list.
     */
    public boolean authorizeResource(String resourceID, SecurityContext context) 
    {
        // deny access if no security context is available
        if (context == null) {
            return false;
        }
        
        Set users = (Set)resourceAuthorization.get(resourceID);
        if (users == null) {
            return false;
        }
        else {
            return users.contains(context.getDN());
        }
    }
    
    /**
     * Authorizes access to an activity based on an access control list.
     * The resourceID is ignored.
     */
    public boolean authorizeActivity(
            String resourceID, 
            Activity activity, 
            SecurityContext context) 
    {
        if (context == null) {
            return false;
        }
        Set users = (Set)activityAuthorization.get(activity.getActivityName());
        if (users == null) {
            return false;
        }
        else {
            return users.contains(context.getDN());
        }
    }
    
}
