//(c) University of Edinburgh 2006.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.authorization;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import uk.org.ogsadai.activity.Activity;
import uk.org.ogsadai.common.SecurityContext;
import uk.org.ogsadai.common.files.FileUtilities;

/**
 * A simple access authorizer which can be updated by adding or removing users 
 * in the access control lists.
 *
 * @author The OGSA-DAI Team.
 */

public class UpdateableFileAuthorizer extends SimpleFileAuthorizer {
    
    private final String configPath;

    /**
     * Constructor. Reads the configuration file and populates the access
     * control lists.
     * 
     * @param config
     *            path of the configuration file
     * @throws Exception
     *             if a problem occurs
     */
    public UpdateableFileAuthorizer(String config) throws Exception {
        super(config);
        configPath = config;
    }

    synchronized public boolean authorizeResource(
            String resourceID, 
            SecurityContext context) 
    {
        return super.authorizeResource(resourceID, context);
    }
    
    synchronized public boolean authorizeActivity(
            String resourceID,
            Activity activity,
            SecurityContext context) 
    {
        return super.authorizeActivity(resourceID, activity, context);
    }
    
    /**
     * Adds a new user to the access control list for the given resource and
     * rewrites the configuration file.
     * 
     * @param resourceID
     *            resource identifier
     * @param userDN
     *            distinguished name of the user
     * @throws Exception if an error occurred writing the configuration file
     */
    synchronized public void grantResourceAccessToUser (
            String resourceID, 
            String userDN)
    throws Exception
    {
        Set users = (Set)resourceAuthorization.get(resourceID);
        if (users != null) {
            users.add(userDN);
        }
        else {
            users = new HashSet();
            users.add(userDN);
            resourceAuthorization.put(resourceID, users);
        }
        writeConfiguration();
    }
        
    /**
     * Adds a new user to an activity access control list and rewrites the 
     * configuration file.
     * 
     * @param activityName
     *            name of the activity
     * @param resourceID
     *            resource identifier
     * @param userDN
     *            distinguished name of the user to be added
     * @throws Exception if an error occurred writing the configuration file
     */
    synchronized public void grantActivityAccessToUser(
            String activityName, 
            String resourceID, 
            String userDN) 
    throws Exception
    {
        Set users = (Set)activityAuthorization.get(activityName);
        if (users != null) {
            users.add(userDN);
        }
        else {
            users = new HashSet();
            users.add(userDN);
            activityAuthorization.put(activityName, users);
        }
        writeConfiguration();
    }
        
    /**
     * Removes the given user from the resource access list and rewrites the 
     * configuration file.
     * 
     * @param resourceID
     *            resouce identifier
     * @param userDN
     *            distinguished name of the user to be removed
     * @return <code>true</code> if the user name was contained in the access 
     *         control list
     * @throws Exception
     *             if a problem occurs when writing the configuration file
     */
    synchronized public boolean removeUserFromResourceAccess (
            String resourceID, 
            String userDN)
    throws Exception
    {
        boolean removed = false;
        Set users = (Set)resourceAuthorization.get(resourceID);
        if (users != null) {
            removed = users.remove(userDN);
            if (users.isEmpty()) {
                resourceAuthorization.remove(resourceID);
            }
            writeConfiguration();
        }
        return removed;
    }
        
    /**
     * Removes the given user from the activity access list and rewrites the
     * configuration file.
     * 
     * @param activityName
     *            name of the activity
     * @param resourceID
     *            resource identifier
     * @param userDN
     *            distinguished name of the user to be removed
     * @return <code>true</code> if the user name was contained in the access
     *         control list
     * @throws Exception
     *             if a problem occurs when writing the configuration file
     */
    synchronized public boolean removeUserFromActivityAccess( 
            String activityName, 
            String resourceID, 
            String userDN)
    throws Exception
    {
        boolean removed = false;
        Set users = (Set)activityAuthorization.get(activityName);
        if (users != null) {
            removed = users.remove(userDN);
            if (users.isEmpty()) {
                activityAuthorization.remove(activityName);
            }
            writeConfiguration();
        }
        return removed;
    }
    
    /**
     * Rewrites the configuration file.
     * 
     * @throws Exception
     *             if an error occurred writing the configuration file
     */
    private void writeConfiguration() throws Exception
    {
        StringBuffer xml = new StringBuffer();
        xml.append("<?xml version=\"1.0\"?>\n");
        xml.append("<authorization>\n");
        writeUsers(xml, resourceAuthorization, "resource", "id");
        writeUsers(xml, activityAuthorization, "activity", "name");
        xml.append("</authorization>\n");
        FileUtilities.deleteFile(configPath);
        FileUtilities.writeStringToFile(configPath, xml.toString());
    }
    
    /**
     * Writes the distinguished names of authorized users wrapped in the given
     * element.
     * 
     * @param xml
     *            buffer to write to
     * @param map
     *            Mapping of a resource/activity name to the set of authorized 
     *            users
     * @param element
     *            name of the element
     * @param attr
     *            attribute name
     */
    private void writeUsers(
            StringBuffer xml, 
            Map map, 
            String element, 
            String attr) 
    {
        Iterator keys = map.keySet().iterator();
        while (keys.hasNext()) {
            String key = (String)keys.next();
            xml.append("  <");
            xml.append(element);
            xml.append(" ");
            xml.append(attr);
            xml.append("=\"");
            xml.append(key);
            xml.append("\">\n");
            Iterator users = ((Set)map.get(key)).iterator();
            while (users.hasNext()) {
                xml.append("    <user dn=\"");
                xml.append(users.next().toString());
                xml.append("\" />\n");
            }
            xml.append("  </");
            xml.append(element);
            xml.append(">\n");

        }
    }
}
