//(c) International Business Machines Corporation, 2002 - 2006. 
//(c) University of Edinburgh, 2002 - 2006.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.clienttoolkit;

import uk.org.ogsadai.client.toolkit.GenericServiceFetcher;
import uk.org.ogsadai.client.toolkit.activity.ActivityRequest;
import uk.org.ogsadai.client.toolkit.activity.delivery.DTInputStream;
import uk.org.ogsadai.client.toolkit.activity.sql.SQLUpdate;
import uk.org.ogsadai.client.toolkit.properties.RequestStatus;
import uk.org.ogsadai.client.toolkit.service.DataService;
import uk.org.ogsadai.client.toolkit.service.DataTransport;
import uk.org.ogsadai.client.toolkit.service.ServiceFetcher;
import uk.org.ogsadai.client.toolkit.sessions.JoinExistingSession;
import uk.org.ogsadai.client.toolkit.sessions.Session;
import uk.org.ogsadai.client.toolkit.sessions.SessionID;
import uk.org.ogsadai.client.toolkit.sessions.SessionStreamID;

/**
 * This example shows how data transport operations can be used to push data
 * into a data service for an SQL update.
 * 
 * @author The OGSA-DAI Project Team
 */
public class ParameterisedUpdateViaDataTransport 
{
    /** Copyright statement. */
    private static final String COPYRIGHT_NOTICE =
        "(c) IBM Corp. 2002 - 2006. " +
        "(c) The University of Edinburgh 2002 - 2006.";
    
    /** URL for the data service. */
    private static final String SERVICE_URL = 
        "http://localhost:8080/wsrf/services/ogsadai/DataService";
    
    /** Name of the data service resource. */
    private static final String RESOURCE_NAME = "MySQLResource";
    
    /**
     * Runs the program.
     * 
     * @param args
     *            no arguments are used
     * @throws Exception
     *             if anything unexpected goes wrong
     */
    public static void main(String[] args) throws Exception 
    {
        // choose data service resource and fetch data service
        
        final ServiceFetcher fetcher = GenericServiceFetcher.getInstance(); 
        final DataService dataService = 
            fetcher.getDataService(SERVICE_URL, RESOURCE_NAME);

        // create and connect DT input stream and SQL update activities

        DTInputStream input = new DTInputStream();
        String sql = "insert into littleblackbook values (?,?,?,?)";
        SQLUpdate update = new SQLUpdate(sql);
        update.setParameter(1, input.getOutput());
        update.setParameter(2, input.getOutput());
        update.setParameter(3, input.getOutput());
        update.setParameter(4, input.getOutput());

        // assemble a request containing the activities and join it to 
        // the session
        
        final ActivityRequest request = new ActivityRequest();
        request.add(input);
        request.add(update);

        // create a new session for the request

        System.out.println("Opening session");
        Session session = dataService.createSession();
        request.setSessionRequirements(new JoinExistingSession(session));
        
        // spawn a thread to send the request
        
        Thread thread = new Thread()
        {
            public void run()
            {
                try
                {
                    System.out.println("Sending the request");
                    dataService.perform(request);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        };
        thread.start();

        pollDataService(dataService, session, RequestStatus.PROCESSING);

        SessionStreamID sessionStreamID = 
            new SessionStreamID(session, input.getName());
        DataTransport dataTransport = dataService.createDataTransport(
            dataService.getResourceID(), sessionStreamID);
        
        System.out.println("Transporting data into the request");
        dataTransport.putBlock("10010"); 
        dataTransport.putBlock("Albert Einstein"); 
        dataTransport.putBlock("Patent Office, Bern"); 
        dataTransport.putBlock("0123456789"); 
        dataTransport.putBlock("10011"); 
        dataTransport.putBlock("Max Planck"); 
        dataTransport.putBlock("University of Berlin"); 
        dataTransport.putBlock("0123456789"); 
        dataTransport.putClosingBlock();
        System.out.println("Data transportation complete");
        
        System.out.println("Waiting for response");
        thread.join();
        System.out.println("Response received");

        System.out.println(
            "Number of modified rows: " + update.getModifiedRowsCount());

        System.out.println("Closing session");
        dataService.closeSession(session);
        
        System.out.println("Done!");
    }

    /**
     * Polls the session request status at the specified data service until it
     * reaches the specified state.
     * 
     * @param dataService
     *            the data service holding the session
     * @param session
     *            the session that the request is joined to
     * @param state
     *            the session request status to poll until
     * @throws Exception
     *             if an unexpected error occurs
     */
    private static void pollDataService(
        final DataService dataService, 
        final Session session,
        final RequestStatus state) throws Exception
    {
        final SessionID id = session.getSessionID();
        while (!dataService.getStatus(id).equals(state))
        {
            Thread.sleep(100);
        }
    }
}