//(c) University of Edinburgh, 2002 - 2005.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.clienttoolkit;

import uk.org.ogsadai.client.toolkit.GenericServiceFetcher;
import uk.org.ogsadai.client.toolkit.Response;
import uk.org.ogsadai.client.toolkit.activity.ActivityRequest;
import uk.org.ogsadai.client.toolkit.activity.sql.SQLQuery;
import uk.org.ogsadai.client.toolkit.activity.sql.WebRowSet;
import uk.org.ogsadai.client.toolkit.service.DataService;
import uk.org.ogsadai.client.toolkit.sessions.JoinExistingSession;
import uk.org.ogsadai.client.toolkit.sessions.JoinNewSession;
import uk.org.ogsadai.client.toolkit.sessions.Session;
import uk.org.ogsadai.client.toolkit.sessions.SessionRequirements;
import uk.org.ogsadai.client.toolkit.sessions.TerminateSession;

/**
 * This example shows how the client toolkit can specify session requirements,
 * determine what session a request was joined to and access the 
 * session request status property.
 * 
 * @author The OGSA-DAI Project Team
 */
public class UsingSessionsExample 
{
    // Copyright statement
    private static final String COPYRIGHT_NOTICE =
        "(c) IBM Corp. 2002 - 2005. (c) The University of Edinburgh 2002 - 2005.";
    /** Handle used in the client toolkit examples. */
    private static final String SERVICE_HANDLE = 
        "http://localhost:8080/wsrf/services/ogsadai/DataService";

    /** Name of data service resource used for this example. */
    private static final String RESOURCE_NAME = 
        "MySQLResource";
    
    public static void main(String[] args) throws Exception 
    {
        // access data service resource via data service
        
        DataService service = 
            GenericServiceFetcher.getInstance().getDataService(
                SERVICE_HANDLE, RESOURCE_NAME);
        
        // perform various interactions involving sessions

        joinImplictSession(service);
        Session session = createNewSession(service);
        joinExistingSession(service, session);
        terminateSession(service, session);
    }

    /**
     * Sends a request which joins an implicit session then displays the
     * identity of the session from the response.
     * 
     * @param service
     *            a data service configured to access a particular data service
     *            resource
     * @throws Exception
     *             if anything goes wrong
     */
    private static void joinImplictSession(
        DataService service) throws Exception
    {
        ActivityRequest request = new ActivityRequest();
        request.add(new SQLQuery("select * from littleblackbook where id<10"));
        Response response = service.perform(request);
        Session session = response.getSession();
        System.out.println(session.getSessionID());
    }
    
    /**
     * Sends an empty request to a data service resource instructing the 
     * creation of a new session with default settings. Note that activities
     * could also be included in the request if desired.
     * 
     * @param service
     *            a data service configured to access a particular data service
     *            resource
     * @return a client-side <code>Session</code> object
     * @throws Exception
     *             if anything goes wrong
     */
    private static Session createNewSession(DataService service)
        throws Exception
    {
        ActivityRequest request = new ActivityRequest();
        SessionRequirements requirements = new JoinNewSession();
        // A session with a specific timeout can also be created
        // SessionRequirements requirements = new JoinNewSession(300000);
        request.setSessionRequirements(requirements);
        Response response = service.perform(request);
        Session session = response.getSession();
        return session;
    }
    
    /**
     * Sends a request to a data service resource to join a particular session.
     * 
     * @param service
     *            a data service configured to access a particular data service
     *            resource
     * @param session
     *            the session for the request to be joined to
     * @throws Exception
     *             if anything goes wrong
     */
    private static void joinExistingSession(
        DataService service, Session session) throws Exception
    {
        ActivityRequest request = new ActivityRequest();
        SessionRequirements requirements = new JoinExistingSession(session);
        
        // The session to join can also be specified by constructing a SessionID
        // SessionID id = new SessionID("some-session-ID");
        // SessionRequirements requirements = new JoinExistingSession(id);
        
        request.setSessionRequirements(requirements);

        SQLQuery query = 
            new SQLQuery("select * from littleblackbook where id<10");
        WebRowSet webRowSet = new WebRowSet(query.getOutput());
        request.add(query);
        request.add(webRowSet);
        
        service.perform(request);
        
        // If request was asynchronous, the client may wish to wait for 
        // completion using a pole method.
        // RequestStatus status = session.getStatus();
        // service.pollUntilRequestCompleted(session.getSessionID(), 100);
    }
    
    /**
     * Sends a request to a data service resource instructing it to terminate a
     * specific session. Note that activities could also be included in the
     * request if desired.
     * 
     * @param service
     *            a data service configured to access a particular data service
     *            resource
     * @param session
     *            the session for the request to be joined to
     * @throws Exception
     *             if anything goes wrong
     */
    private static void terminateSession(
        DataService service, Session session) throws Exception
    {
        ActivityRequest request = new ActivityRequest();
        SessionRequirements requirements = new TerminateSession(session);
        request.setSessionRequirements(requirements);
        service.perform(request);
    }
}