//(c) University of Edinburgh, 2002 - 2006.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.demoFactory;

import java.io.File;

import javax.xml.namespace.QName;

import uk.org.ogsadai.common.files.FileUtilities;
import uk.org.ogsadai.common.files.DAIFileNotFileException;
import uk.org.ogsadai.common.files.DAIFileNotFoundException;
import uk.org.ogsadai.common.files.DAIFileReadIOException;
import uk.org.ogsadai.common.properties.Property;
import uk.org.ogsadai.common.properties.PropertyValueException;
import uk.org.ogsadai.common.properties.StaticProperty;
import uk.org.ogsadai.dataresource.DataResourceAccessor;
import uk.org.ogsadai.dataresource.DataResourceAccessorConfigException;
import uk.org.ogsadai.dataresource.DataResourceAccessorMetaDataException;
import uk.org.ogsadai.dataresource.DataResourceAccessorSetupException;
import uk.org.ogsadai.dataresource.DataResourcePropertyNames;
import uk.org.ogsadai.dataresource.PersistInFiles;

/**
 * A very simple data resource accessor designed to be used to
 * demonstrate dynamic creation of data service resources.  
 * <p>
 * The dynamically created data service resource will use this
 * accessor. This data resource accessor has only one configuration
 * file called <code>name.txt</code> which contains a string.  The
 * value of this string can be retrieved via either a resource
 * property called 
 * <code>{http://ogsadai.org.uk/namespaces/2005/10/properties}name</code> 
 * or via the <code>getName()</code> method. 
 * 
 * @author The OGSA-DAI Project Team
 */
public class DemoInstanceDataResourceAccessor 
    implements DataResourceAccessor, PersistInFiles, DemoNameProvider
{
    /** Copyright statement */
    private static final String COPYRIGHT_NOTICE = 
        "(c) University of Edinburgh 2002 - 2006.";

    /** Name of data service resource */
    private String mResourceName;
    
    /** A simple name value */
    private String mName; 

    /** Name of resource property holding the name value */
    public static final QName NAME_PROPERTY_QNAME =
        new QName(DataResourcePropertyNames.NAMESPACE, "name");
    
    /* (non-Javadoc)
     * @see uk.org.ogsadai.dataresource.DataResourceAccessor#setResourceName(java.lang.String)
     */
    public void setResourceName(String resourceName)
    {
        mResourceName = resourceName;
    }

    /* 
     * Reads the configuration. For this data resource accessor the
     * configuration consists of one file called <code>name.txt</code>
     * that contains a string that can be accessed via
     * <code>getName</code> or the <code>name</code> resource
     * property.
     */
    public void restoreFromConfig(File directory) 
        throws 
            DataResourceAccessorConfigException, 
            DataResourceAccessorMetaDataException, 
            DataResourceAccessorSetupException, 
            IllegalArgumentException
    {
        if (directory == null)
        {
            throw new IllegalArgumentException("directory must not be null");
        }
        if (! directory.isDirectory())
        {
            throw new IllegalArgumentException("directory " + 
                                               directory.getAbsolutePath() +
                                               " must be a directory");
        }
        File nameFile = new File( directory, "name.txt" );
        try
        {
            mName = FileUtilities.readStringFromFile(nameFile.getAbsolutePath());
        }
        catch (DAIFileNotFoundException e)
        {
            throw new DataResourceAccessorConfigException(mResourceName, e);
        }
        catch (DAIFileNotFileException e)
        {
            throw new DataResourceAccessorConfigException(mResourceName, e);
        }
        catch (DAIFileReadIOException e)
        {
            throw new DataResourceAccessorConfigException(mResourceName, e);
        }
    }

    /* (non-Javadoc)
     *
     * @see uk.org.ogsadai.dataresource.DataResourceAccessor#getProperties()
     */
    public Property[] getProperties() 
    {
        Property prop = new StaticProperty();
        prop.setName(NAME_PROPERTY_QNAME);
        try 
        {
            prop.setValue(mName);
        } 
        catch (PropertyValueException e) 
        {
            // StaticProperty never throws this.
        }
        return new Property[]{prop};
    }
    
    /**
     * Gets the name value.
     * 
     * @return the name value.
     */
    public String getName()
    {
        return mName;
    }
}
