//(c) University of Edinburgh, 2006.
//See OGSA-DAI-Licence.txt for licencing information.

package uk.org.ogsadai.examples.demoTransientFactory;

import org.w3c.dom.Element;

import uk.org.ogsadai.activity.Activity;
import uk.org.ogsadai.activity.ActivityCreationException;
import uk.org.ogsadai.activity.ActivityExecutionException;
import uk.org.ogsadai.activity.ActivitySpecificationException;
import uk.org.ogsadai.activity.ActivityStreamNotFoundException;
import uk.org.ogsadai.dataresource.DataResourceAccessor;
import uk.org.ogsadai.common.BlockWriter;
import uk.org.ogsadai.service.OGSADAIConstants;
import uk.org.ogsadai.service.resource.factory.DataServiceResourceFactory;

/**
 * An activity that creates a new data resource to demonstrate the
 * data service resource creation functionality.
 * 
 * <p>
 * For information on this activity see the OGSA-DAI user doc:
 * <code>
 * OGSA-DAI/doc/interaction/activities/examples/demoCreateTransientInstance.html
 * </code>
 * @author OGSA-DAI team
 */
public class DemoCreateTransientInstanceActivity extends Activity
{
    /** Copyright statement */
    private static final String COPYRIGHT_NOTICE = 
        "(c) University of Edinburgh 2006.";

    /** Activity output. */
    private BlockWriter mOutput;

    /**
     * The name to be returned by activities that access the created
     * data service resource.
     */
    private String mName;

    /**
     * Constructs an demo create transient instance activity using the 
     * specified XML element.
     * 
     * @param element 
     *     <code>Element</code> to configure the activity with - from
     *     a perform document.
     * 
     * @throws ActivityCreationException
     *     If there is a problem constructing the activity due to an
     *     implementation error or OGSA-DAI confguration problem.
     * 
     * @throws ActivitySpecificationException
     *     If there is a problem constructing the activity due to an
     *     invalid setting in <code>element</code>.
     *     mistake such as an invalid setting.
     */
    public DemoCreateTransientInstanceActivity(final Element element)
        throws ActivitySpecificationException, ActivityCreationException 
    {
        super(element);

        // Publish the activity output(s) by setting the inherited
        // mInternalOutputs fields.

        String outputName = ((Element) element.getElementsByTagNameNS(
            OGSADAIConstants.TYPES_NAMESPACE,
            "myOutput").item(0)).getAttribute("name");
        mInternalOutputs = new String[] { outputName };

        // Read the configuration details for other sub-elements of the
        // activity element

        Element configuration = (Element) element.getElementsByTagNameNS(
            OGSADAIConstants.TYPES_NAMESPACE,
            "name").item(0);
        mName = configuration.getAttribute("value");
    }

    /* (non-Javadoc)
     * @see uk.org.ogsadai.activity.Activity#initialise()
     */
    public void initialise()
        throws ActivitySpecificationException, ActivityExecutionException 
    {
        try
        {
            // Set-up references to the activity input and output.
            mOutput = mContext.getOutput(mInternalOutputs[0]);
        }
        catch (ActivityStreamNotFoundException e) 
        {
            throw new ActivitySpecificationException(getName(), e);
        }
    }

    /* (non-Javadoc)
     * @see uk.org.ogsadai.engine.Activity#processBlock()
     */
    public void processBlock()
    {
        try
        {
            // Get the data service resource factory
            DataServiceResourceFactory factory = 
                mContext.getDataResourceFactory();
            
            // Create the data resource accessor
            DataResourceAccessor childDataResourceAccessor = 
                new DemoTransientInstanceDataResourceAccessor(mName);
            
            ChildDataServiceResourceComponentProvider childComponentProvider = 
                (ChildDataServiceResourceComponentProvider) 
                mContext.getDataResourceAccessor();

            // Create the new data service resource
            String resourceId = factory.createResource(
                "test",
                childDataResourceAccessor,
                childComponentProvider.getChildActivitiesConfig(),
                childComponentProvider.getChildSessionComponentFactory());

            mOutput.put(resourceId);
            setCompleted();
        }
        catch (Exception e)
        {
            ActivitySpecificationException error = 
                new ActivitySpecificationException(getName(), e);
            setError(error);
        }
    }
}
