############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides a simple interface to the ftp client module.
"""

from pyGlobus.util import GlobusException
from pyGlobus import utilc
from pyGlobus.ftpClientAttr import HandleAttr, OperationAttr
from pyGlobus.ftpClientPlugin import RestartMarker
import ftpClientc
from pyGlobus.util import Buffer
from pyGlobus.util import Abstime
from pyGlobus.util import Off_t
import threading

#########################################################################
# Module classes.
#########################################################################
class FtpClientException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the ftp client  module. It defines no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class BaseFtpClient:
    """Parent Class for the FtpClient and EasyFtpClient classes"""
    
    def __init__(self):
        raise RuntimeError, "Class can not be initialized"
    
    def mkdir(self, *argslist):
        raise NotImplementedError, "method mkdir has not been implemented in child class"

    def delete(self, *argslist):
        raise NotImplementedError, "delete method has not been implemented  in child class"

    def rmdir(self, *argslist):
        raise NotImplementedError, "rmdir method has not been implemented  in child class"

    def list(self, *argslist):
        raise NotImplementedError, "list method has not been implemented  in child class"

    def verbose_list(self, *argslist):
        raise NotImplementedError, "verbose_list method has not been implemented  in child class"

    def move(self, *argslist):
        raise NotImplementedError, "move method has not been implemented  in child class"

    def get(self, *argslist):
        raise NotImplementedError, "get method has not been implemented  in child class"

    def partial_get(self, *argslist):
        raise NotImplementedError, "partial_get method has not been implemented  in child class"

    def put(self, *argslist):
        raise NotImplementedError, "put method has not been implemented  in child class"

    def partial_put(self, *argslist):
        raise NotImplementedError, "partial_put method has not been implemented  in child class"

    def third_party_transfer(self, *argslist):
        raise NotImplementedError, "third_party_transfer method has not been implemented  in child class"
    
    def partial_third_party_transfer(self, *argslist):
        raise NotImplementedError, "partial_third_party_transfer method has not been implemented  in child class"                           
    def exists(self, *argslist):
        raise NotImplementedError, "exists method has not been implemented  in child class"
    
    def modification_time(self, *argslist):
        raise NotImplementedError," modification_time method has not been implemented  in child class"
    
    def size(self, *argslist):
        raise NotImplementedError,"size method has not been implemented  in child class"


    def cache_url_state(self, url):
        """
        Cache the connection to the FTP server.
              
        When the url is cached, the library will not close the connection after completing
        an operation.

        **Arguments:**
              
        - *url* The URL of the server to cache.

        **Raises:**

        A FtpClientException is thrown if unable to cache the url state.
        """
        
        ret, err = ftpClientc.handle_cache_url_state(self._handle, url)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return

    def flush_url_state(self, url):
        """
        Remove the url from the connection cache.
        
        If an idle connection to the server exists, it will be closed.
        
        **Arguments:**
        
        - *url* The URL to remove from the cache.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to flush the url state.
        """
        
        ret, err = ftpClientc.handle_flush_url_state(self._handle, url)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return

    def set_user_data(self, data):
        """
        Associate some data with the handle.
        
        Access to this data is not synchronized.

        **Arguments:**
        
        - *data* The user data to set.
        
        **Raises:**
        
        An FtpClientException is thrown if unable to set the data.
        """
        
        ret, err = ftpClientc.handle_set_user_pointer(self._handle, data)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return

    def get_user_data(self):
        """
        Returns the user data associated with the handle.
        
        Access to this data is not synchronized.
        
        **Returns:**
        
        The data set with the set_user_data method.
        
        **Raises:**
        
        An FtpClientException is thrown if unable to get the data.
        """
        
        ret, data = ftpClientc.handle_get_user_pointer(self._handle)
        if not ret == 0:
            ex = FtpClientException(data)
            raise ex
        return data

    def abort(self):
        """
        Abort the current operation.
        
        **Raises:**
        
        A FtpClientException is thrown if an error occurs while aborting
        the current operation.
        """
        
        ret, err = ftpClientc.py_abort(self._handle)
        
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return

    def add_plugin(self, plugin):
        """
        Add a plugin to an FTP client handle.
              
        This function adds a plugin to an FTP client handle after it has been
        created. Plugins may be added to an ftp client handle whenever an
        operation is not in progress. The plugin will be appended to the list of
        plugins present in the handle, and will be invoked during any subsequent
        operations processed with this handle.
        
        Only one instance of a particular plugin may be added to a particular handle.
        
        **Arguments**
        
        *plugin* A Plugin class.
        
        **Raises**
        
        A FtpClientException is thrown if unable to add the plugin.
        """
        
        ret, err = ftpClientc.handle_add_plugin(self._handle, plugin._handle)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return

    def remove_plugin(self, plugin):
        """
        Remove a plugin to an FTP client handle.
              
        This function removes a plugin from an FTP client handle after it has been
        created. Plugins may be removed from an ftp client handle whenever an
        operation is not in progress. The plugin will be removed from the list of
        plugins, and will not be used during any subsequent operations processed
        with this handle.
        
        **Arguments**
        
        *plugin* A Plugin class.
        
        **Raises**
        
        A FtpClientException is thrown if unable to remove the plugin.
        """
        
        ret, err = ftpClientc.handle_remove_plugin(self._handle, plugin._handle)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        return
    
    def get_handle(self):
        """
        Return the underlying ftp client handle.
        
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**
              
        A string containing the pointer.
        """
        return self._handle

class FtpClient(BaseFtpClient):
    """
    A class to wrap the ftp client functions.

    It provides blocking and non-blocking transfers and access to the other ftp
    operations.
    """
    
    def __init__(self, handleAttr, handle=None):
        """
        Constructs an instance.
        
        Create a new FtpClient instance. This will init the ftp client module and
        create a new ftp client handle if none is passed in.
        
        **Arguments:**
        
        - *handleAttr* A HandleAttr object with the attributes set for
        this object.
        
        - *handle* If a SWIG'ized pointer to a globus ftp client handle object
        is passed in, create a new instance around this handle. The class
        then owns the underlying pointer, and will delete it when the instance
        is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to init the module or create
        the underlying handle.
        """
        
        self._activated = 0
        self._cbList = []
        ret = utilc.globus_module_activate(ftpClientc.get_module())
        if not ret == 0:
            ex = FtpClientException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = ftpClientc.handle_init(handleAttr.get_handle())
            if not ret == 0:
                ex = FtpClientException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance.
        
        Free's any remaining callbacks, destroyes the underlying handle, and deactivates
        the module.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to destroy the handle or deactivate
        the module.
        """
        
        for cb in self._cbList:
            self.free_callback(cb)
        ret, err = ftpClientc.handle_destroy(self._handle)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        ret = utilc.globus_module_deactivate(ftpClientc.get_module())
        if not ret == 0:
            ex = FtpClientException("Unable to deactivate the module")
            raise ex

        return
    
    def delete(self, url, attr, callback, arg):
        """
        Delete a file from a ftp server.
        
        The callback will be called with the final status of the delete.
        
        **Arguments:**
        
        - *url* The URL to delete.
        
        - *attr* An OperationAttr object containing the attributes for this
        delete.
        
        - *callback* The function to be called when the delete is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.

        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the delete.
        """
        
        ret, cbHandle = ftpClientc.py_delete(self._handle, url,
                                         attr.get_handle(),
                                         callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def mkdir(self, url, attr, callback, arg):
        """
        Make a directory on a ftp server.
              
        The callback will be called with the final status of the mkdir.
              
        **Arguments:**

        - *url* The URL to make.
        
        - *attr* An OperationAttr object containing the attributes for this
        mkdir.
              
        - *callback* The function to be called when the mkdir is completed. It
        has the following signature:
              
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.

        **Returns:**

        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initi ate the mkdir.
        """
        
        ret, cbHandle = ftpClientc.py_mkdir(self._handle, url,
                                           attr.get_handle(),
                                           callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def rmdir(self, url, attr, callback, arg):
        """
        Remove a directory on a ftp server.
        
        The callback will be called with the final status of the rmdir.
        
        **Arguments:**
        
        - *url* The URL to remove.
        
        - *attr* An OperationAttr object containing the attributes for this
        rmdir.
        
        - *callback* The function to be called when the rmdir is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the rmdir.
        """
        
        ret, cbHandle = ftpClientc.py_rmdir(self._handle, url,
                                           attr.get_handle(),
                                           callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def list(self, url, attr, callback, arg):
        """
        Get a file listing from an ftp server.
        
        This method starts a 'NLIST' transfer from an ftp server.  The client may begin
        calling register_read to retrieve the listing, after this method returns. When all
        of the data associated with the listing is returned, and all of the data callbacks
        executed, the callback will be called with the final status of the list.
        
        **Arguments:**
        
        - *url* The url to list.
        
        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        - *callback* The function to be called when the list is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the list.
        """
        
        ret, cbHandle = ftpClientc.py_list(self._handle, url,
                                          attr.get_handle(),
                                          callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def verbose_list(self, url, attr, callback, arg):
        """
        Get a file listing from an ftp server.
        
        This method starts a ' LIST' transfer from an ftp server.  The client may begin
        calling register_read to retrieve the listing, after this method returns. When all
        of the data associated with the listing is returned, and all of the data callbacks
        executed, the callback will be called with the final status of the list.
        
        **Arguments:**
        
        - *url* The url to list.
        
        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        - *callback* The function to be called when the list is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the list.
        """
        
        ret, cbHandle = ftpClientc.verbose_list(self._handle, url,
                                               attr.get_handle(),
                                               callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def move(self, srcUrl, destUrl, attr, callback, arg):
        """
        Move a file on a ftp server.
        
        The complete callback will be called when the move is completed.
        
        **Arguments:**
        
        - *srcUrl* The url to move.
        
        - *destUrl* The url to move to.
        
        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        - *callback* The function to be called when the move is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the move.
        """
        
        ret, cbHandle = ftpClientc.move(self._handle, srcUrl,
                                       destUrl, attr.get_handle(),
                                       callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def get(self, url, callback, arg, attr=None, marker=None):
        """
        Get a file from an ftp server.
        
        After calling this method, register_read can be called to begin retrieving
        the data.  When all of the data associated with this URL is retrieved, and all
        of the data callbacks have completed, or if the get request is aborted, the
        complete callback will be called with the final status of the get.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this get.
        
        - *marker* A RestartMarker object.
        
        - *callback* The function to be called when the get is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the get.
        """
        
        if attr == None:
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        ret, cbHandle = ftpClientc.get(self._handle, url, attr.get_handle(),
                                      marker.get_handle(), callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def partial_get(self, url, attr, marker, offset, endOffset, callback, arg):
        """
        Get part of a file from a ftp server.
        
        After calling this method, register_read can be called to begin retrieving
        the data.  When all of the data associated with this URL is retrieved, and all
        of the data callbacks have completed, or if the get request is aborted, the
        complete callback will be called with the final status of the get.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this get.
        
        - *marker* A RestartMarker object.
        
        - *offset* An int or long offset to begin the transfer at.
        
        - *endOffset* An int or long offset to end the transfer at.
        
        - *callback* The function to be called when the partial get is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the partial get.
        """
        
        ret, cbHandle = ftpClientc.partial_get(self._handle, url, 
                                               attr.get_handle(),
                                               marker.get_handle(), offset,
                                               endOffset, callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def extended_get(self, url, eret_alg_str, callback,
                     arg,attr = None, marker = None):
        """
        Get a file from an ftp server with server-side processing..
        
        After calling this method, register_read can be called to begin retrieving
        the data.  When all of the data associated with this URL is retrieved, and all
        of the data callbacks have completed, or if the get request is aborted, the
        complete callback will be called with the final status of the get.
        
        This method differs from the get method by allowing the user to invoke
        server-side data processing algorithms. GridFTP servers may support
        support algorithms for data reduction or other customized data storage
        requirements. There is no client-side verification done on the algorithm
        string provided by the user. If the server does not understand the
        requested algorithm, the transfer will fail.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this get.
        
        - *marker* A RestartMarker object.
              
        - *eret_alg_str*  The ERET algorithm string. This string contains
        information needed to invoke a server-specific data reduction
        algorithm on the file being retrieved.
        
        - *callback* The function to be called when the get is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the extended_get.
        """

        if attr == None:
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

            ret, cbHandle = ftpClientc.extended_get(self._handle, url,
                                                    attr.get_handle(),
                                                    marker.get_handle(),
                                                    eret_alg_str,
                                                    callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle
    
    def put(self, url, callback, arg, attr = None, marker = None):
        """
        Store a file on an ftp server.
              
        After calling this method, register_write can be called to begin writing
        the data.  When all of the data associated with this URL is sent, and all
        of the data callbacks have completed, or if the put request is aborted, the
        complete callback will be called with the final status of the put.
        
        **Arguments:**
        
        - *url* The url to store the data to.
        
        - *attr* An OperationAttr object containing the attributes for
        this put.
        
        - *marker* A RestartMarker object.
        
        - *callback* The function to be called when the put is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the put.
        """

        if attr == None:               
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()


        ret, cbHandle = ftpClientc.put(self._handle, url, attr.get_handle(),
                                      marker.get_handle(), callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def partial_put(self, url, attr, marker, offset, endOffset, callback, arg):
        """
        Put part of a file onto a ftp server.
        
        After calling this method, register_write can be called to begin writing
        the data.  When all of the data associated with this URL is sent, and all
        of the data callbacks have completed, or if the partial_put request is aborted, the
        complete callback will be called with the final status of the partial_put.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this partial_put.
        
        - *marker* A RestartMarker object.
        
        - *offset* An int or long offset to begin the transfer at.
        
        - *endOffset* An int or long offset to end the transfer at.
        
        - *callback* The function to be called when the partial put is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the partial put.
        """
        
        ret, cbHandle = ftpClientc.partial_put(self._handle, url, 
                                               attr.get_handle(),
                                               marker.get_handle(), offset,
                                               endOffset, callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle
    
    def extended_put(self, url, esto_alg_str, callback,
                     arg, attr = None, marker = None):
        """
        Store a file on an ftp server with server-side processing.
        
        After calling this method, register_write can be called to begin writing
        the data.  When all of the data associated with this URL is sent, and all
        of the data callbacks have completed, or if the put request is aborted, the
        complete callback will be called with the final status of the put.
        
        This method differs from the put method by allowing the user to invoke
        server-side data processing algorithms. GridFTP servers may support
        algorithms for data reduction or other customized data storage requirements.
        There is no client-side verification done on the alogirhtm string provided by
        the user. if the server does not understand * the requested algorithm, the transfer
        will fail.
        
        **Arguments:**
        
        - *url* The url to store the data to.
        
        - *attr* An OperationAttr object containing the attributes for
        this put.
        
        - *marker* A RestartMarker object.
        
        - *esto_alg_str* The algorith string. This string contains
        information needed to invoke a server-specific data reduction
        algorithm on the file being put.
            
        - *callback* The function to be called when the put is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the put.
        """

        if attr == None:               
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()


        ret, cbHandle = ftpClientc.put(self._handle, url, attr.get_handle(),
                                      marker.get_handle(), callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle
    
    def third_party_transfer(self, srcUrl, destUrl, callback, arg,
                             srcAttr = None, destAttr = None, marker = None):
 
        """
        Transfer a file between two ftp servers.
        
        When the transfer is complete, the complete callback will be called with the
        final status of the transfer.
        
        **Arguments:**
        
        - *srcUrl* The url to transfer.  
        
        - *destUrl* The url to store the data to.
              
        - *callback* The function to be called when the transfer is completed. It
        has the following signature:
              
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        - *srcAttr* An OperationAttr object containing the attributes
        for the source.
        
        - *destAttr* An OperationAttr object containing the attributes
        for the destination.
        
        - *marker* A RestartMarker object.  
        
        **Returns:**

        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
            
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the transfer.
        """

        if srcAttr == None:
            srcAttr = OperationAttr()

        if destAttr == None:
            destAttr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        ret, cbHandle = ftpClientc.third_party_transfer(self._handle,
                                                       srcUrl,
                                                       srcAttr.get_handle(),
                                                       destUrl,
                                                       destAttr.get_handle(),
                                                       marker.get_handle(),
                                                       callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def partial_third_party_transfer(self, srcUrl, srcAttr, destUrl, destAttr,
                                     marker, offset, endOffset, callback, arg):
        """
        Transfer part of a file between two ftp servers.
        
        When the transfer is complete, the complete callback will be called with the
        final status of the transfer.
        
        **Arguments:**
        
        - *srcUrl* The url to transfer.

        - *srcAttr* An OperationAttr object containing the attributes
        for the source.
        
        - *destUrl* The url to store the data to.
        
        - *destAttr* An OperationAttr object containing the attributes
        for the destination.
              
        - *marker* A RestartMarker object.
        
        - *offset* An int or long offset to start the transfer from.
        
        - *endOffset* An int or long offset to end the transfer at.
        
        - *callback* The function to be called when the transfer is completed. It
        has the following signature:
    
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object. 
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns:**

        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the transfer.
        """
        
        ret, cbHandle = ftpClientc.partial_third_party_transfer(self._handle,
                                                               srcUrl,
                                                               srcAttr.get_handle(),
                                                               destUrl,
                                                               destAttr.get_handle(),
                                                               marker.get_handle(),
                                                               offset, endOffset,
                                                               callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def extended_third_party_transfer(self, srcUrl, eret_alg_str,
                                      destUrl, esto_alg_str,
                                      callback, arg, srcAttr = None,
                                      destAttr = None, marker = None):
 
        """
        Transfer a file between two ftp servers with server-side processing.
        
        When the transfer is complete, the complete callback will be called with the
        final status of the transfer.
        
        This method differs from the third_party_transfer method by allowing the user
        to invoke server-side data processing algorithms. GridFTP servers may
        support algorithms for data reduction or other customized data storage
        requirements. There is no client-side verification done on the alogirhtm
        string provided by the user. if the server does not understand
        the requested algorithm, the transfer will fail.
        
        **Arguments:**
        
        - *srcUrl* The url to transfer.
        
        - *srcAttr* An OperationAttr object containing the attributes
        for the source.
        
        - *eret_alg_str* The algorithm string for the source server. This
        string contains information needed to invoke a server-specific
        data reduction algorithm on the file being retrieved.
        
        - *destUrl* The url to store the data to.
        
        - *esto_alg_str* The algorithm string for the dest server. This string contains
        information needed to invoke a server-specific data reduction
        algorithm on the file being placed.
            
        - *destAttr* An OperationAttr object containing the attributes
        for the destination.
        
        - *marker* A RestartMarker object.
        
        - *callback* The function to be called when the transfer is completed. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
              
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**

        A FtpClientException is thrown if unable to initiate the transfer.
        """

        if srcAttr == None:
            srcAttr = OperationAttr()

        if destAttr == None:
            destAttr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        ret, cbHandle = ftpClientc.extended_third_party_transfer(self._handle,
                                                                 srcUrl,
                                                                 srcAttr.get_handle(),
                                                                 eret_alg_str,
                                                                 destUrl,
                                                                 destAttr.get_handle(),
                                                                 esto_alg_str,
                                                                 marker.get_handle(),
                                                                 callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle
    
    def register_read(self, buffer, callback, arg):
        """
        Register a data buffer to handle part of the ftp data transfer.
        
        This can only be called after get or partial get has been called.
        
        **Arguments:**

        - *buffer* A pyGlobus.util.Buffer object.. This hides the
        internal globus_byte_t data buffer.
        
        - *callback* The function to be called after the data has been
        read. It has the following signature:
        
        'function(arg, handle, buffer, bufHandle, bufLen, offset, eof, error)'
        where arg is the user argument passed into this call, handle is the
        underlying ftp_client handle, buffer is a Python buffer object, bufHandle
        is a SWIG'ized pointer to the underlying globus_byte_t array, bufLen
        is the length of the underlying globus_byte_t array, offset is the offset
        into the file this data block contains, eof is a bool that is true if
        the end of the data transfer has been reached, and error is a tuple
        (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.

        - *arg* A user argument to be passed through to the callback.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the read.
        """
        
        ret, cbHandle = ftpClientc.register_read(self._handle,
                                                buffer, buffer.get_handle(),
						buffer.get_size(),
                                                callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def register_write(self, buffer, size, offset, eof, callback, arg):
        """
        Register a data buffer to handle part of the ftp data transfer.
        
        This can only be called after put or partial put has been called.
        
        **Arguments:**
        
        - *buffer* A string of the data to be written out.
        
        - *size* The length of the buffer.
        
        - *offset* The offset of the buffer to be written. In extended block
        mode the data doesn't need to be sent in order.
        
        - *eof* True if this buffer contains the end of the data.
        
        - *callback* The function to be called after the data has been
        written. It has the following signature:
        'function(arg, handle, buffer, bufHandle, bufLen, offset, eof, error)'
        where arg is the user argument passed into this call, handle is the
        underlying ftp_client handle, buffer is a Python buffer object, bufHandle
        is a SWIG'ized pointer to the underlying globus_byte_t array, bufLen
        is the length of the underlying globus_byte_t array, offset is the offset
        into the file this dta block contains, eof is a bool that is true if
        the end of the data transfer has been reached, and error is a tuple
        (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be passed through to the callback.
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
            
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the write.
        """
        
        ret, cbHandle = ftpClientc.register_write(self._handle, buffer,
                                                 size, offset, eof,
                                                 callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    
    def free_callback(self, cbHandle):
        """
        Free a callback.
        
        Free's the underlying memory allocated for a callback structure.
        
        **Arguments:**
        
        - *cbHandle* A SWIG'ized pointer to a callback structure.
        """
        
        ftpClientc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return

    
    def size(self, url, callback, arg, attr=None):
        """
        Retrieves the size of a file on a ftp server
        
        When the size is retrieved the complete callback is called
        
        ***XXX*** the object underlying globus_off_t pointer will not be filled with any
        meaningful data until the callback function has been called
        
        **Arguments**
              
        - *url* the location of the file
        
        - *attr* An OperationAttr object, if none is given then
        the default OperationAttr is created and used
        
        - *callback* The function to be called when the file size has been retrieved. It
        has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns**
        An util.Off_t object, to get the size call getLong on this object,
        after the callback has returned. 
        
        **Raises**
        
        An FtpClientException if the size can't be retrieved
        
        """

        if attr == None:
            attr = OperationAttr()

        ret, size, cbHandle = ftpClientc.size(self._handle, url, attr.get_handle(),
                                        callback, arg)

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        
        new_size = Off_t(size)
        self._cbList.append(cbHandle)
        return new_size
        
    def modification_time(self, url, callback, arg, attr=None):
        """
        Retrieves the size of a file on a ftp server
        
        When the modification time is retrieved the complete callback is called
        
        ***XXX*** the Abstime object's underlying timeval struct will not be filled with any
        meaningful data until the callback function has been called
        
        **Arguments**
        
        - * url the location of the file
        
        - * attr An OperationAttr object, if none is given then
        the default OperationAttr is created and used
        
        - *callback* The function to be called when the modification time of the url
        has been retrieved. It has the following signature:
        
        'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        **Returns**
        An util.Abstime object.
        
        **Raises**
        An FtpClientException if the modification time cannot be retrieved
        
        """

        if attr == None:
            attr = OperationAttr()
            
        ret, cbHandle, mtime = ftpClientc.modification_time(self._handle, url, 
                                                            attr.get_handle(),
                                                            callback, arg)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        
        self._cbList.append(cbHandle)
        t = Abstime(mtime)
        return t
        
    def exists(self, url, callback, arg, attr = None):
        """
        Determines whether a file exists on a ftp server
        
        When the existence has been determined  the complete callback is called
        
        **Arguments**
        
        - *url* the location of the file
        
        - *callback* The function to be called when the modification time of the url
        has been retrieved. It has the following signature:
        
        - 'function(arg, handle, error)' where arg is the argument passed into this
        call, handle is a SWIG'ized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
        
        - *arg* A user argument to be returned when the callback is executed.
        
        - *attr* An OperationAttr object, if none is given then
        the default OperationAttr is created and used
        
        **Returns**
        a boolean indicating whether the file exists
        
        **Raises**
        An FtpClientException if the existence cannot be determined
        
        """
        
        if attr == None:
            attr = OperationAttr()

            
        ret, cbHandle = ftpClientc.exists(self._handle, url, attr.get_handle(),
                                          callback, arg)

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        self._cbList.append(cbHandle)
        return
    
class EasyFtpClient(BaseFtpClient):
    """
    A class to wrap the ftp client functions.

    Each operation in this class blocks until the operation has completed

    There is currently no way to abort an operation

    If you must abort an operation use the FtpClient class
    """
    
    def __init__(self, handleAttr = None, handle=None):
        """
              Constructs an instance.
              
              Create a new FtpClient instance. This will init the ftp client module and
              create a new ftp client handle if none is passed in.

              **Arguments:**

              - *handleAttr* A HandleAttr object with the attributes set for
              this object.
            
              - *handle* If a SWIG'ized pointer to a globus ftp client handle object
              is passed in, create a new instance around this handle. The class
              then owns the underlying pointer, and will delete it when the instance
              is destroyed.
              
              **Raises:**

              A FtpClientException is thrown if unable to init the module or create
              the underlying handle.
              """
        
        self._activated = 0
        self._buffer = None
        self._condV = threading.Condition(threading.Lock())
        
        
        ret = utilc.globus_module_activate(ftpClientc.get_module())
        if not ret == 0:
            ex = FtpClientException("Unable to initialize the module")
            raise ex
        self._activated = 1
        
        if handleAttr == None:
            handleAttr = HandleAttr()
            
        
        if handle == None:
            ret, handle = ftpClientc.handle_init(handleAttr.get_handle())
            if not ret == 0:
                ex = FtpClientException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
              Destroy an instance.
              
              Free's any remaining callbacks, destroyes the underlying handle, and deactivates
              the module.
              
              **Raises:**

              A FtpClientException is thrown if unable to destroy the handle or deactivate
              the module.
              """
    
        ret, err = ftpClientc.handle_destroy(self._handle)
        if not ret == 0:
            ex = FtpClientException(err)
            raise ex
        ret = utilc.globus_module_deactivate(ftpClientc.get_module())
        if not ret == 0:
            ex = FtpClientException("Unable to deactivate the module")
            raise ex
        
        del self._buffer 
        return

    def done_cb_exists(self, arglist, handle, error):    
        """
        This callback is for the exists method
        It is called after the existence check has returned from the ftp Server
        """
        
        if not error[0] == 0: 
            arglist[1].append(error)

        cv = arglist[0]
        cv.acquire()
        cv.notify()
        cv.release()
        
        return
        
    def done_cb(self, cv, handle, error):
        """
        Callback that has the signature
        function(arg, handle, error) where arg is the argument passed into this
        call, handle is a SWIGized pointer to a ftp client handle object, and error
        is a tuple (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
             
        This function handles these types of callbacks in the EasyFtpClient class
        """
        
        cv.acquire()
        cv.notify()
        cv.release()
        
        return

    def read_cb(self, fd, handle, buffer, bufHandle, bufLen, offset, eof, error):
        """
        This is a callback function that is for the register_read method
        It is called when a the buffer is full with data that has arrived to the handle
        It is used as a data callback for the *get methods
        """
        
        fd.seek(offset)      
        fd.write(buffer)
        if not eof:
            try:
                ret, cbHandle = ftpClientc.register_read(self._handle,self._buffer,
							self._buffer.get_handle(),
                                                        self._buffer.get_size(),
                                                        self.read_cb, fd)
                if not ret == 0:
                    ex = FtpClientException(cbHandle)
                    raise ex
                
                if self._cb_handle != None:
                    ftpClientc.free_callback(self._cb_handle)    
                self._cb_handle = cbHandle
                
            except Exception, ex:
                fd.close()
                raise ex
         
        return

    def write_cb(self, fd, handle, buffer, bufHandle, bufLen, offset, eof, error):
        """
	This is a callback function for the register_write method.
	It is called when the handle has recieved the buffer
	It is used as a data callback for the *put methods
	"""
        
        if not eof:
            self._buffer = fd.read(bufLen)    
            try:
                ret, cbHandle = ftpClientc.register_write (self._handle, self._buffer,
                                                          len(self._buffer),
                                                          offset + bufLen, self._buffer == "",
                                                          self.write_cb, fd)
                if not ret == 0:
                    ex = FtpClientException(cbHandle)
                    raise ex
            
                if self._cb_handle != None:
                    ftpClientc.free_callback(self._cb_handle)
                self._cb_handle = cbHandle

            except GlobusException, ex:
                fd.close()
                raise ex
       
        return    

    def write_cb_partial(self, args, handle, buffer, bufHandle, bufLen, offset, eof, error):
        """
	This is a callback function for register writes that need to keep track of offsets
	It is used in the partial_put method
	"""
        
        if not eof:
            
            fd, endOffset = args
            if offset + bufLen == endOffset:
                self._buffer = fd.read(0)
            elif endOffset - offset < bufLen:
                self._buffer = fd.read(endOffset - offset)
            else:
                self._buffer = fd.read(bufLen)

            try:
                ret, cbHandle = ftpClientc.register_write (self._handle, 
                                                          self._buffer,
                                                          len(self._buffer),
                                                          offset + bufLen, 
                                                          self._buffer == "",
                                                          self.write_cb_partial,
                                                          (fd, endOffset))

                if not ret == 0:
                    ex = FtpClientException(cbHandle)
                    raise ex
            
                if self._cb_handle != None:
                    ftpClientc.free_callback(self._cb_handle)
                self._cb_handle = cbHandle

            except GlobusException, ex:
                fd.close()
                raise ex
            
    def small_read_cb(self, arg, handle, buffer, bufHandle, bufLen, offset, eof, error):
        """
	This function is for doing small reads such as a verbose_list.
	The result is stored in a string, instead of being written to a file
	"""

        self._result += str(buffer)
    
        if not eof:
            try:
                ret, cbHandle = ftpClientc.register_read(self._handle, self._buffer,
							 self._buffer.get_handle(),
                                                         self._buffer.get_size(),
							 self.small_read_cb, 0)

                if not ret == 0:
                    ex = FtpClientException(cbHandle)
                    raise ex
                    
                if self._cb_handle != None:
                    ftpClientc.free_callback(self._cb_handle)
                self._cb_handle = cbHandle

            except Exception, ex:
                raise ex
        return

    def delete(self, url, attr = None):
        """
        Delete a file from a ftp server.
        
        This method blocks until the file has been deleted
        
        **Arguments:**
        
        - *url* The URL to delete.
        
        - *attr* An OperationAttr object containing the attributes for this
        delete.
        
        **Returns**
        Nothing is returned
              
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the delete.
        """
        
        if attr == None:
            attr = OperationAttr()

        self._condV.acquire()
        ret, cbHandle = ftpClientc.py_delete(self._handle, url,
                                         attr.get_handle(),
                                         self.done_cb,self._condV )

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

    def mkdir(self, url, attr= None):
        """
        Make a directory on a ftp server.
        
        The callback will be called with the final status of the mkdir.
        
        **Arguments:**
        
        - *url* The URL to make.
        
        - *attr* An OperationAttr object containing the attributes for this
        mkdir.
        
        **Returns**
        Nothing is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the mkdir.
        """
        
        if attr == None:
            attr = OperationAttr()

        self._condV.acquire()
        ret, cbHandle = ftpClientc.py_mkdir(self._handle, url,
                                           attr.get_handle(),
                                           self.done_cb, self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

    def rmdir(self, url, attr= None):
        """
        Remove a directory on a ftp server.
        
        The callback will be called with the final status of the rmdir.
        
        **Arguments:**
        
        - *url* The URL to remove.

        - *attr* An OperationAttr object containing the attributes for this
        rmdir.
        
        **Returns**
        Nothing is returned
              
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the rmdir.
        """
        
        if attr == None:
            attr = OperationAttr()
            
        self._condV.acquire()
        ret, cbHandle = ftpClientc.py_rmdir(self._handle, url,
                                           attr.get_handle(),
                                           self.done_cb, self._condV)
        
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

    def list(self, url, attr = None):
        """
        Get a file listing from an ftp server.
        
        This method starts a 'NLIST' transfer from an ftp server.  The client may begin
        calling register_read to retrieve the listing, after this method returns. When all
        of the data associated with the listing is returned, and all of the data callbacks
        executed, the callback will be called with the final status of the list.

        **Arguments:**
        
        - *url* The url to list.

        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        **Returns**
        A string of the 'NLIST' transfer is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the list.
        """
        
        if attr == None:
            attr = OperationAttr()
            
        self._condV.acquire()    
        ret, cbHandle = ftpClientc.py_list(self._handle, url,
                                          attr.get_handle(),
                                          self.done_cb, self._condV)
        
        
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        self._buffer = Buffer(1024 *64)
        self._result = ''
        self._cb_handle = None
        ret, handle = ftpClientc.register_read(self._handle, self._buffer,
					      self._buffer.get_handle(),
                                              self._buffer.get_size(),
					      self.small_read_cb, 0)

        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        return self._result
    
    def verbose_list(self, url, attr = None):
        """
        Get a file listing from an ftp server.
        
        This method blocks until the list of the url is retrieved
        
        This method starts a ' LIST' transfer from an ftp server.  The client may begin
        calling register_read to retrieve the listing, after this method returns. When all
        of the data associated with the listing is returned, and all of the data callbacks
        executed, the callback will be called with the final status of the list.
        
        **Arguments:**

        - *url* The url to list.
              
        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        **Returns**
        A string of the 'LIST' transfer is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the list.
        """
                
        if attr == None:
            attr = OperationAttr()

        self._condV.acquire()
        ret, cbHandle = ftpClientc.verbose_list(self._handle, url,
                                               attr.get_handle(),
                                               self.done_cb, self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        self._buffer = Buffer(1024 *64)
        self._result = ''
        self._cb_handle = None
        ret, handle = ftpClientc.register_read(self._handle, self._buffer,
	 				       self._buffer.get_handle(),
                                               self._buffer.get_size(),
	  				       self.small_read_cb, 0)

        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex
        
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        return self._result

    
    
    
    
    def move(self, srcUrl, destUrl, attr=None):
        """
        Move a file on a ftp server.
        
        The complete callback will be called when the move is completed.
        
        **Arguments:**
        
        - *srcUrl* The url to move.
        
        - *destUrl* The url to move to.
        
        - *attr* An OperationAttr object containing the attributes
        for the listing.
        
        **Returns:**
        Nothing is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the move.
        """
        
        if attr == None:
            attr = OperationAttr()
            
        self._condV.acquire()
        ret, cbHandle = ftpClientc.move(self._handle, srcUrl,
                                        destUrl, attr.get_handle(),
                                        self.done_cb, self._condV )
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

   
    def get(self, dest_file, url, attr=None, marker=None):
        """
        Get a file from an ftp server.
        
        After calling this method, register_read can be called to begin retrieving
        the data.  When all of the data associated with this URL is retrieved, and all
        of the data callbacks have completed, or if the get request is aborted, the
        complete callback will be called with the final status of the get.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this get.
        
        - *marker* A RestartMarker object.
        
        **Returns**
        Nothing is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the get.
        """
       
        if attr == None:
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        self._cb_handle = None
        self._buffer = Buffer(64*1024)
        fd = open(dest_file, "w")

        self._condV.acquire()
        ret, cbHandle = ftpClientc.get(self._handle, url, attr.get_handle(),
                                       marker.get_handle(), self.done_cb, 
                                       self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        ret, handle = ftpClientc.register_read(self._handle, self._buffer,
					       self._buffer.get_handle(),
                                               self._buffer.get_size(),
					       self.read_cb, fd)

        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex
 
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        ftpClientc.free_callback(handle)

    def partial_get(self, dest_file, url, offset, endOffset, attr=None, marker=None):
        """
        Get part of a file from a ftp server.
        
        After calling this method, register_read can be called to begin retrieving
        the data.  When all of the data associated with this URL is retrieved, and all
        of the data callbacks have completed, or if the get request is aborted, the
        complete callback will be called with the final status of the get.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this get.
        
        - *marker* A RestartMarker object.
        
        - *offset* An int or long offset to begin the transfer at.
        
        - *endOffset* An int or long offset to end the transfer at.
        
        **Returns**
        Nothing is returned
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the partial get.
        """
        
        if attr == None:
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        self._cb_handle = None
        self._buffer = Buffer(64*1024)
        fd = open(dest_file, "w")

        self._condV.acquire()
        ret, cbHandle = ftpClientc.partial_get(self._handle, url, attr.get_handle(),
                                              marker.get_handle(), offset,
                                              endOffset, self.done_cb, self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
        
        ret, handle = ftpClientc.register_read(self._handle,self._buffer,
					      self._buffer.get_handle(),
					      self._buffer.get_size(),
                                              self.read_cb, fd)

        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex
        
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        ftpClientc.free_callback(handle)
        
    def put(self, src_file, url, attr=None, marker=None):
        """
        Store a file on an ftp server.
        
        After calling this method, register_write can be called to begin writing
        the data.  When all of the data associated with this URL is sent, and all
        of the data callbacks have completed, or if the put request is aborted, the
        complete callback will be called with the final status of the put.
        
        **Arguments:**
        
        - *url* The url to store the data to.

        - *attr* An OperationAttr object containing the attributes for
        this put.
        
        - *marker* A RestartMarker object.
        
        **Returns**
        Nothing is returned  
        
        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the put.
        """

        if attr == None:               
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        self._cb_handle = None
        fd = open(src_file, "r")
        
        self._condV.acquire()
        ret, cbHandle = ftpClientc.put(self._handle, url, attr.get_handle(),
                                      marker.get_handle(),self.done_cb, 
                                      self._condV)

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        size = 64*1024
        oset = 0
        self._buffer = fd.read(size)
       
        ret, handle = ftpClientc.register_write(self._handle, self._buffer, 
                                                len(self._buffer), oset, 
                                                buffer == "",self.write_cb, fd)
        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex

        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        ftpClientc.free_callback(handle)

    def partial_put(self, src_file, url, offset, endOffset, attr=None, marker=None):
        """
        Put part of a file onto a ftp server.
        
        After calling this method, register_write can be called to begin writing
        the data.  When all of the data associated with this URL is sent, and all
        of the data callbacks have completed, or if the partial_put request is aborted, the
        complete callback will be called with the final status of the partial_put.
        
        **Arguments:**
        
        - *url* The url to download.
        
        - *attr* An OperationAttr object containing the attributes for
        this partial_put.
        
        - *marker* A RestartMarker object.
        
        - *offset* An int representing the byte offset to begin the transfer at.
        
        - *endOffset* An int representing the byte offset to end the transfer at.
        
        **Returns**
        Nothing is returned

        **Raises:**
        
        A FtpClientException is thrown if unable to initiate the partial put.
        """
        
        if attr == None:               
            attr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        self._cb_handle = None
        size = 64 *1024
        self._buffer = Buffer(size)
        fd = open(src_file, "r")

        self._condV.acquire()
        ret, cbHandle = ftpClientc.partial_put(self._handle, url, 
                                               attr.get_handle(),
                                               marker.get_handle(), offset,
                                               endOffset, self.done_cb, 
                                               self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        fd.seek(offset)
        
        if endOffset - offset < size:
            self._buffer = fd.read(endOffset - offset)
        else:
            self._buffer = fd.read(size)
            
            
        ret, handle = ftpClientc.register_write(self._handle, self._buffer, 
                                                len(self._buffer), offset, 
                                                buffer == "", 
                                                self.write_cb_partial, 
                                                (fd, endOffset))

        if not ret == 0:
            ex = FtpClientException(handle)
            raise ex
        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        ftpClientc.free_callback(handle)

    def third_party_transfer(self, srcUrl, destUrl,
                             srcAttr = None, destAttr = None, marker = None):
 
        """
        Transfer a file between two ftp servers.
        
        When the transfer is complete, the complete callback will be called with the
        final status of the transfer.
        
        **Arguments:**

        - *srcUrl* The url to transfer.
        
        - *destUrl* The url to store the data to.
        
        - *srcAttr* An OperationAttr object containing the attributes
        for the source.  
        
        - *destAttr* An OperationAttr object containing the attributes
        for the destination.
        
        - *marker* A RestartMarker object
        
        **Returns**
        Nothing is returned
        
        **Raises:**
              
        A FtpClientException is thrown if unable to initiate the transfer.
        """
        
        if srcAttr == None:
            srcAttr = OperationAttr()

        if destAttr == None:
            destAttr = OperationAttr()

        if marker == None:
            marker = RestartMarker()

        self._condV.acquire()    
        ret, cbHandle = ftpClientc.third_party_transfer(self._handle,
                                                       srcUrl,
                                                       srcAttr.get_handle(),
                                                       destUrl,
                                                       destAttr.get_handle(),
                                                       marker.get_handle(),
                                                       self.done_cb, 
                                                       self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

    def partial_third_party_transfer(self, srcUrl, destUrl, offset, endOffset,
                                     srcAttr = None, destAttr=None, marker=None):
        """
        Transfer part of a file between two ftp servers.
        
        When the transfer is complete, the complete callback will be called with the
        final status of the transfer.
        
        **Arguments:**
        
        - *srcUrl* The url to transfer.

        - *srcAttr* An OperationAttr object containing the attributes
        for the source.
        
        - *destUrl* The url to store the data to.

        - *destAttr* An OperationAttr object containing the attributes
        for the destination.
        
        - *marker* A RestartMarker object.
              
        - *offset* An int or long offset to start the transfer from.
        
        - *endOffset* An int or long offset to end the transfer at.

        **Returns**
        Nothing is returned
        
        **Raises:**
        A FtpClientException is thrown if unable to initiate the transfer.
        """
        
        if srcAttr == None:
            srcAttr = OperationAttr()

        if destAttr == None:
            destAttr = OperationAttr()

        if marker == None:
            marker = RestartMarker()
            
        self._condV.acquire()    
        ret, cbHandle = ftpClientc.partial_third_party_transfer(self._handle,
                                                               srcUrl,
                                                               srcAttr.get_handle(),
                                                               destUrl,
                                                               destAttr.get_handle(),
                                                               marker.get_handle(),
                                                               offset, endOffset,
                                                               self.done_cb, self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex
    
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

    def size(self, url, attr =  None):
        """
        Retrieves the size of a file on a ftp server
        
        When the size is retrieved the complete callback is called
        
        **Arguments**
              
        - *url* the location of the file
        
        - *attr* An OperationAttr object, if none is given then
        the default OperationAttr is created and used
        
        **Returns**
        the size of the file
        
        **Raises**

        An FtpClientException if the size cannont be retrieved
        """
           
        if attr == None:
            attr = OperationAttr()

        self._condV.acquire()
        ret, size, cbHandle = ftpClientc.size(self._handle, url, attr.get_handle(),
                                        self.done_cb, self._condV)

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        
        self._condV.wait()
        self._condV.release()
        new_size = Off_t(size)
        ftpClientc.free_callback(cbHandle)
        return new_size.getLong()
    
    def exists(self, url, attr = None):
        """
        Determines whether a file exists on a ftp server
        
        When the existence has been determined  the complete callback is called
        
        **Arguments**
        
        - *url* the location of the file
        
        - *attr* An OperationAttr object, if none is given then
        the default OperationAttr is created and used

        **Returns**
        a boolean indicating whether the file exists
        
        **Raises**
        An FtpClientException if the existence cannot be determined
        """
        
        if attr == None:
            attr = OperationAttr()

            
        arglist = [self._condV, []]
        self._condV.acquire()
        ret, cbHandle = ftpClientc.exists(self._handle, url, attr.get_handle(),
                                          self.done_cb_exists, arglist)

        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)

        if len(arglist[1]) != 0:
            return 0
        else:
            return 1

    def modification_time(self, url, attr = None):
        """
        Retrieves the size of a file on a ftp server
        
        When the modification time is retrieved the complete callback is called
        
        **Arguments**
        
        - *url* the location of the file
        
        - *attr* An OperationAttr object, if none is given then
        the default OperationAttr is created and used
        
        **Returns**
        the string of the modification time
        
        **Raises**
        An FtpClientException if the modification time cannot be retrieved
        """

        if attr == None:
            attr = OperationAttr()

        self._condV.acquire()
        ret, cbHandle, mtime = ftpClientc.modification_time(self._handle, url, attr.get_handle(),
                                                            self.done_cb, self._condV)
        if not ret == 0:
            ex = FtpClientException(cbHandle)
            raise ex

        
        self._condV.wait()
        self._condV.release()
        ftpClientc.free_callback(cbHandle)
        t = Abstime(mtime)
        return t
        


 
