############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
file transfer protocol client attribute module  
"""

from pyGlobus.ftpControl import Parallelism, TcpBuffer, Layout, DCAU
from pyGlobus import utilc
from pyGlobus.util import GlobusException
from pyGlobus.security import GSSCred
import ftpClientc


class HandleAttrException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the ftp client handle attr module. It defines
    no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class OperationAttrException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the ftp client operation attr module. It defines
    no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class HandleAttr:
    """
    A class to wrap the ftp client handle attr functions.

    It provides get and set methods for all of the various handle
    specific attributes, and hides the underlying memory management.
    """

    def __init__(self, handle = None, own=1):
        """
        Constructs an instance.
        
        Activates the globus ftp client module and creates the ftp handle attr.

        **Arguments**

        - *handle* A SWIG'ized pointer to a globus ftp client handleattr
        pointer. Used to construct an instance from an existing pointer.
        
        - *own* True if the class owns the underlying pointer.
      
        **Raises:**

        A HandleAttr exception is thrown if unable to activate
        the module or create the handle.
        """
        
        self._activatd = 0
        self._own = own
        ret = 1
        ret = utilc.globus_module_activate(ftpClientc.get_module())
        if not ret == 0:
            ex = HandleAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = ftpClientc.handleattr_init()
            if not ret == 0:
                ex = HandleAttrException(handle)
                raise ex
        self._handle = handle

    def __del__(self):
        """
        Destroy an instance.

        Deletes the ftp attr handle and deactivates the underlying globus
        module.

        **Raises:**

        A HandleAttrException is thrown if unable to destroy the
        handle or deactivate the module.
        """
        
        if self._own:
            ret, error = ftpClientc.handleattr_destroy(self._handle)
            if not ret == 0:
                ex = HandleAttrException(error)
                raise ex
        ret = utilc.globus_module_deactivate(ftpClientc.get_module())
        if not ret == 0:
            ex = HandleAttrException("Unable to deactivate the module")
            raise ex

    def copy(self):
        """
        Creates a deep copy of itself.
        
        **Raises:**

        A HandleAttrException is thrown if unable to make the copy.

        **Returns:**

        A new HandleAttr object with the same properties as the
        original is returned.
        """
        
        ret, handle = ftpClientc.handleattr_copy(self._handle)
        if not ret == 0:
            ex = HandleAttrException(handle)
            raise ex
        return HandleAttr(handle)

    def add_cached_url(self, url):
        """
        Enable caching for the specified url.
        
        The handle will keep the connection to the specified url open
        in between ftp operations.

        **Arguments:**

        - *url* The url to cache.

        **Raises:**

        A HandleAttrException is thrown if unable to cache the
        url.
        """
        
        ret, error = ftpClientc.handleattr_add_cached_url(self._handle, url)
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return

    def remove_cached_url(self, url):
        """
        Disable caching for the specified url.
        
        The handle will no longer keep the connection open to the specified
        url between connections.

        **Arguments:**
        
        - *url* The url to remove from the cache.

        **Raises:**

        A HandleAttrException is thrown if unable to remove the url
        from the cache.
        """
        
        ret, error = ftpClientc.handleattr_remove_cached_url(self._handle, url)
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return

    def set_cache_all(self, cacheAll):
        """
        Set the cache all attribute.

        If true, all connections on the handle will be cached in between
        connections. When this is enabled, the user skips the authentication
        handshake and connection establishment overhead for multiple subsequent
        ftp operations to the same server.

        **Arguments:**

        - *cacheAll* A boolean describing whether to cache all connections
        or not.

        **Raises:**

        A HandleAttrException is thrown if unable to set the
        attribute.
        """
        
        ret, error = ftpClientc.handleattr_set_cache_all(self._handle,
                                                         cacheAll)
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return

    def get_cache_all(self):
        """
        Return the cache all attribute.
        
        If true, all connections on the handle will be cached in between
        connections. When this is enabled, the user skips the authentication
        handshake and connection establishment overhead for multiple subsequent
        ftp operations to the same server.
              
        **Raises:**

        A HandleAttrException is thrown if unable to return the
        attribute.
        """
        
        ret, error, cacheAll = ftpClientc.handleattr_get_cache_all(self._handle)
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return cacheAll
    
    def set_netlogger(self, nl_handle):
        """
        Set netlogger handle.

        **Arguments**

        *nl_handle* A netlogger handle.
        
        **Raises**

        An HandleAttrException is thrown if unable to add the plugin.
        """
        
        ret, error = ftpClientc.handleattr_set_cache_all(self._handle,
                                                         nl_handle)
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return
    
    def add_plugin(self, plugin):
        """
        Add a plugin to a handle attribute set.
        
        Each handle attribute set contains a list of plugins associated with it. When a
        handle is created with a particular attribute set, it will be associated with a copy
        of those plugins.
        
        Only one instance of a specific plugin may be added to an attribute set. Each
        plugin must have a different name.

        A copy of the plugin is created via the plugins 'copy' method when it is added to
        an attribute set. Thus, any changes to a particular plugin must be done before the
        plugin is added to an attribute set, and before the attribute set is used to create handles.
        
        **Arguments**

        *plugin* A Plugin class.
        
        **Raises**

        An HandleAttrException is thrown if unable to add the plugin.
        """
        
        ret, err = ftpClientc.handleattr_add_plugin(self._handle, plugin._handle)
        if not ret == 0:
            ex = HandleAttrException(err)
            raise ex
        return

    def remove_plugin(self, plugin):
        """
        Remove a plugin from a handle attribute set.
        
        Each handle attribute set contains a list of plugins associated with it. When a
        handle is created with a particular attribute set, it will be associated with a copy
        of those plugins.

        Only one instance of a specific plugin may be added to an attribute set. Each
        plugin must have a different name.

        A copy of the plugin is created via the plugins 'copy' method when it is added to
        an attribute set. Thus, any changes to a particular plugin must be done before the
        plugin is added to an attribute set, and before the attribute set is used to create handles.
              
        **Arguments**

        *plugin* A Plugin class.
        
        **Raises**

        An HandleAttrException is thrown if unable to remove the plugin.
        """
        
        ret, err = ftpClientc.handleattr_remove_plugin(self._handle, 
                                                       plugin._handle)
        if not ret == 0:
            ex = HandleAttrException(err)
            raise ex
        return
    
    def get_handle(self):
        """
        Return the underlying ftp client handle attr handle.
        
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**

        A string containing the pointer.
        """
        
        return self._handle
    
class OperationAttr:
    """
    A class to wrap the ftp client operation attr functions.

    It provides get and set methods for all of the various operation
    specific attributes, and hides the underlying memory management.
    """
    
    def __init__(self, handle = None):
        """
        Constructs an instance.
        
        Activates the globus ftp client module and creates the underlying handle.

        **Raises:**

        A OperationAttrException is thrown if unable to initialize the
        module or create the handle.
        """
        
        self._activated = 0
        ret = utilc.globus_module_activate(ftpClientc.get_module())
        if not ret == 0:
            ex = OperationAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = ftpClientc.operationattr_init()
            if not ret == 0:
                ex = OperationAttrException(handle)
                raise ex
        self._handle = handle

    def __del__(self):
        """
        Destroy an instance.

        Deletes the underlying handle, and deactivates the underlying module.

        **Raises:**

        A OperationAttrException is thrown if unable to delete the
        handle or destroy the module.
        """
       
        if ftpClientc == None:
            return 
     
        ret, error = ftpClientc.operationattr_destroy(self._handle)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        ret = utilc.globus_module_deactivate(ftpClientc.get_module())
        if not ret == 0:
            ex = OperationAttrException("Unable to deactivate the module")
            raise ex


    def copy(self):
        """
        Creates a deep copy of itself.
        
        **Raises:**

        A OperationAttrException is thrown if unable to make the copy.

        **Returns:**
              
        A new OperationAttr object with the same properties as the
        original is returned.
        """
        
        ret, handle = ftpClientc.operationattr_copy(self._handle)
        if not ret == 0:
            ex = OperationAttrException(handle)
            raise ex
        return OperationAttr(handle)

    def set_append(self, append):
        """
        Set the append mode for the operation.

        Allows a file to be appended to when performing a put to a server instead of
        replacing the file. This attribute is incompatible with the extended block mode,
        the partial file attribute, and the restart attribute.
        
        **Arguments:**

        - *append* A boolean value.

        **Raises:**

        A OperationAttrException is thrown if unable to set the append
        attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_append(self._handle, append)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_append(self):
        """
        Returns the value of the append attribute.

        **Raises:**

        A OperationAttrException is thrown if unable to return the append
        attribute.
            
        **Returns:**

        A boolean value is returned.
        """
        
        ret, val = ftpClientc.operationattr_get_append(self._handle)
        if not ret == 0:
            ex = OperationAttrException(val)
            raise ex
        return val

    def set_authorization(self, credential, user=None, passwd="", account="", subject=""):
        """
        Set the authorization attributes.
              
        **Arguments:**
              
        - *user* The user name to send to the server.

        - *credential* A SWIG'ized pointer to a gss_cred_id_t object. To use the
        default credential, use GSS_C_NO_CREDENTIAL. This can be returned
        from the gsi module.

        - *passwd* A string containing the password to send to the server.
              
        - *account* The account to use for the data transfer. Most ftp servers do not
        require this.

        - *subject* The subject name of the ftp server. This is only needed for
        a gsiftp transfer when the security subject name does not match the subject
        name of the server.

        **Raises:**

        A OperationAttrException is thrown if uanbe to set the
        authorization attributes.
        """
        
        ret, error = ftpClientc.operationattr_set_authorization(self._handle, 
                                                             credential._handle,
                                                             user, passwd, 
                                                             account, subject)
        if not ret == 0:
            pass #print "ret is %i" % ret
            pass #print "error is %s" % error
            ex = OperationAttrException(error)
            raise ex
        return

    def get_authorization(self):
        """
        Returns the authorization attributes.
              
        **Raises:**

        A OperationAttrException is thrown if unable to get the authorization
        attributes.

        **Returns:**
              
        - A GSSCred object

        - A string containg the user name.
              
        - A string containing the password.

        - A string containing the account name.
              
        - A string containing the subject name of the server.
        """
        
        ret, credential, user, passwd, account, subject = ftpClientc.operationattr_get_authorization(self._handle)
         
        if not ret == 0:
            ex = OperationAttrException(credential)
            raise ex
        return GSSCred(credential, 0), user, passwd, account, subject

    def set_dcau(self, dcau):
        """
        Set the data channel authentication instance in the handle.
        
        **Arguments:** 

        - *dcau* A ftpControl.DCAU instance

        **Raises:**
        A OperationAttrException is raised if there is a problem setting the
        DCAU handle
        """
        ret, err = ftpClientc.operationattr_set_dcau(self._handle, 
                                                     dcau.get_handle())
        if not ret == 0:
            ex = OperationAttrException(err)
            raise ex
        return

    def get_dcau(self):
        """
        Get the DCAU handle

        **Returns**
        A ftpControl.DCAU instance

        **Raises:**
        A OperationAttrException is raised if there is a problem setting the
        DCAU handle
        """
        ret, dcau = ftpClientc.operationattr_get_dcau(self._handle)
        if not ret == 0:
            ex = OperationAttrException(dcau)
            raise ex
        return DCAU(dcau)

    def set_data_protection(self, protection):
        """
        Sets the protection attribute for the data channel.
        
        The supported protection attributes are defined in the ftpControl module.
        
        **Arguments:**

        - *protection* One of the possible attributes defined in the ftpControl
        module.

        **Raises:**

        A OperationAttrException is thrown if unable to set the data protection
        attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_data_protection(self._handle, 
                                                                  protection)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_data_protection(self):
        """
        Returns the protection attribute for the data channel.
        
        **Raises:**
              
        A OperationAttrException is thrown if unable to get the attribute.

        **Returns:**

        One of the attributes defined in the ftpControl module is returned.
        """
        
        ret, protection = ftpClientc.operationattr_get_data_protection(self._handle)
        if not ret == 0:
            ex = OperationAttrException(protection)
            raise ex
        return protection

    def set_control_protection(self, protection):
        """
        Sets the protection attribute for the control channel.

        The supported protection attributes are defined in the ftpControl module.
              
        **Arguments:**

        - *protection* One of the possible attributes defined in the ftpControl
        module.

        **Raises:**
              
        A OperationAttrException is thrown if unable to set the control protection
        attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_control_protection(self._handle, protection)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_control_protection(self):
        """
        Returns the protection attribute for the control channel.

        **Raises:**
              
        A OperationAttrException is thrown if unable to get the attribute.

        **Returns:**

        One of the attributes defined in the ftpControl module is returned.
        """
        
        ret, protection = ftpClientc.operationattr_get_control_protection(self._handle)
        if not ret == 0:
            ex = OperationAttrException(protection)
            raise ex
        return protection
    
    def set_layout(self, layout):
        """
        Sets the layout attribute for the transfer.

        This attribute allows the user to control the layout of a file being transfered to a
        striped Grid-FTP server. The striping layout defines what regions of a file will be
        stored on each stripe of a multiple-striped ftp server. 
              
        The layout attribute is used only when the data is being stored on the server
        (on a put or 3rd party transfer). This attribute is ignored for stream mode data
        transfers.
              
        **Arguments:**
              
        - *layout* A ftpControl.Layout object.

        **Raises:**
              
        A OperationAttrException is thrown if unable to set the layout
        attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_layout(self._handle,
                                               layout.get_handle())
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_layout(self):
        """
        Returns the layout attribute for this operation.

        **Raises:**
        
        A OperationAttrException is thrown if unable to get the layout
        attribute.

        **Returns:**

        A ftpControl.Layout object is returned.
        """
        
        ret, layHandle = ftpClientc.operationattr_get_layout(self._handle)
        if not ret == 0:
            ex = OperationAttrException(layHandle)
            raise ex
        return Layout(layHandle)

    def set_mode(self, mode):
        """
        Sets the mode for the transfer.

        The supported modes are defined in the ftpControl module.
              
        **Arguments:**

        - *mode* One of the possible modes defined in the ftpControl
        module.
        
        **Raises:**

        A OperationAttrException is thrown if unable to set the mode.
        """
        
        ret, error = ftpClientc.operationattr_set_mode(self._handle, mode)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_mode(self):
        """
        Returns the mode for this transfer.

        **Raises:**
        
        A OperationAttrException is thrown if unable to get the mode.

        **Returns:**

        One of the modes defined in the ftpControl module is returned.
        """
        
        ret, mode = ftpClientc.operationattr_get_mode(self._handle)
        if not ret == 0:
            ex = OperationAttrException(mode)
            raise ex
        return mode

    def set_parallelism(self, parallelism):
        """
        Set the parallelism attribute.

        Used to control the level of parallelism used for the operation.

        **Arguments:**

        - *parallelism* A ftpContol.Parallelism object.

        **Raises:**

        A OperationAttrException is thrown if unable to set the
        parallelism attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_parallelism(self._handle,
                                                    parallelism.get_handle())
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_parallelism(self):
        """
        Returns the parallelism attribute.

        **Raises:**

        A OperationAttrException is thrown if unable to get the
        parallelism attribute.
              
        **Returns:**
        
        A ftpContol.Parallelism object is returned.
        """
        
        ret, parHandle = ftpClientc.operationattr_get_parallelism(self._handle)
        if not ret == 0:
            ex = OperationAttrException(parHandle)
            raise ex
        return Parallelism(parHandle)

    def set_tcp_buffer(self, tcpbuffer):
        """
        Set the tcp buffer attribute.
        
        Used to control the tcp buffer size for the operation.
              
        **Arguments:**
        
        - *tcpbuffer* A ftpContol.TcpBuffer object with the proper 
        buffer size set.
        
        **Raises:**

        A OperationAttrException is thrown if unable to set the tcp buffer
        attribute.
        """
        
        ret, error = ftpClientc.operationattr_set_tcp_buffer(self._handle,
                                                   tcpbuffer.get_handle())
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_tcp_buffer(self):
        """
        Return the tcp buffer attribute.
        
        **Raises:**
              
        A OperationAttrException is thrown if unable to get the tcp buffer
        attribute.
            
        **Returns:**

        A ftpContol.TcpBuffer object is returned.
        """
        
        ret, bufHandle = ftpClientc.operationattr_get_tcp_buffer(self._handle)
        if not ret == 0:
            ex = OperationAttrException(bufHandle)
            raise ex
        return TcpBuffer(bufHandle)

    def set_type(self, type):
        """
        Sets the type attribute for the operation.

        Controls the type used for the transfer. The types are defined in the 
        ftpControl module. The default type is binary.

        **Arguments:**

        - *type* One of the types defined in the ftpControl module.

        **Raises:**

        A OperationAttrException is thrown if unable to set the type.
        """
        
        ret, error = ftpClientc.operationattr_set_type(self._handle, type)
        if not ret == 0:
            ex = OperationAttrException(error)
            raise ex
        return

    def get_type(self):
        """
        Returns the type attribute for the operation.
              
        **Raises:**
              
        A OperationAttrException is thrown if unable to get the type
        attribute.
              
        **Returns:**
        
        One of the types defined in the ftpControl module is returned.
        """
        
        ret, opType = ftpClientc.operationattr_get_type(self._handle)
        if not ret == 0:
            ex = OperationAttrException(opType)
            raise ex
        return opType

    def get_handle(self):
        """
        Return the underlying ftp client operation attr handle.

        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**

        A string containing the pointer.
        """
  
        return self._handle


