############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
A module that wraps various plugins to monitor gridftp transfers.
"""
from pyGlobus import utilc
from pyGlobus.util import GlobusException
import ftpClientc

class RestartMarkerException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the ftp client restart marker module. It defines
    no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class PluginException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the Plugin module. It defines
    no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class PyGlobusBasePlugin:
    COPY = 0
    DESTROY = 1
    GET = 2
    PUT = 3
    TRANSFER = 4
    ABORT = 5
    CONNECT = 6
    DELETE = 7
    AUTH = 8
    READ = 9
    MOVE = 10
    LIST = 11
    SIZE = 12
    VERBOSELIST = 13
    RMDIR = 14
    MKDIR = 15
    MODIFICATIONTIME = 16
    WRITE = 17
    DATA = 18
    COMMAND = 19
    RESPONSE = 20
    FAULT = 21
    COMPLETE = 22
    
    def __init__(self):
        self._handle = None
        function_list = [
            self.copy,
            self.destroy,
            self.get,
            self.put,
            self.transfer,
            self.abort,
            self.connect,
            self.delete,
            self.auth,
            self.read,
            self.move,
            self.list,
            self.size,
            self.verbose_list,
            self.rmdir,
            self.mkdir,
            self.modification_time,
            self.write,
            self.data,
            self.command,
            self.response,
            self.fault,
            self.complete,
            
            ]
        self._handle = ftpClientc.pyglobus_plugin_init(function_list)
        
    def get_handle(self):
        """
        Return the underlying ftp client plugin handle.
        
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**
        
        A string containing the pointer.
        """
        
        return self._handle

    def copy(self, ftp_client_plugin_handle, user_specific):
        pass
    
    def destroy(self, ftp_client_plugin_handle, user_specific):
        pass
    
    def get(self, ftp_client_plugin_handle, ftp_client_handle_t, url, ftp_client_operationattr, restart, user_specific):
        pass
    
    def put(self, ftp_client_plugin_handle, ftp_client_handle_t, url, ftp_client_operationattr, restart, user_specific):
        pass
    
    def transfer(self, *args):
        pass
    
    def abort(self, *args):
        pass 
    
    def connect(self, *args):
        pass
    
    def delete(self, *args):
        pass

    def auth(self, *args):
        pass
    
    def read(self, *args):
        pass

    def move(self, *args):
        pass 

    def list(self, *args):
        pass
    
    def size(self, *args):
        pass
    
    def verbose_list(self, *args):
        pass
    
    def rmdir(self, *args):
        pass
    
    def mkdir(self, *args):
        pass
    
    def modification_time(self, *args):
        pass

    def write(self, *args):
        pass
    
    def data(self, *args):
        pass
    
    def command(self, *args):
        pass 
    
    def response(self, *args):
        pass 
    
    def fault(self, *args):
        pass

    def complete(self, *args):
        pass

class RestartMarker:
    """
    A class to wrap the ftp client restart marker attributes.

    It provides various get and set methods to control the properties of a transfer
    that is about to restart.
    """
    
    def __init__(self, handle = None):
        """
        Constructs an instance.
        
        Activates the globus ftp client module, and creates the underlying handle.
        
        **Arguments:**
        
        - *handle* An optional argument used to construct an instance from an
        existing handle. The class assumes it then \"owns\" the handle, and
        can free it when it is destroyed.
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to initialize the
        module, or create the handle.
        """
              
        self._activated = 0
        ret = utilc.globus_module_activate(ftpClientc.get_module()) 
        if not ret == 0:
            ex = RestartMarkerException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = ftpClientc.restart_marker_init()
            if not ret == 0:
                ex = RestartMarkerException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance.
        
        Destroys the underlying handle, and deactivates the module.
        
        NOTE: This destructor is sometimes called when the ftpClientc has already
        been destroyed because of somewhat unpleasant python behavior when
        this module is imported
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to deactivate
        the module 
        """
        
        if ftpClientc == None:
            return

        ftpClientc.restart_marker_destroy(self._handle)
        ret = utilc.globus_module_deactivate(ftpClientc.get_module())
        if not ret == 0:
            ex = RestartMarkerException("Unable to deactivate the module")
            raise ex
        return

    def copy(self):
        """
        Creates a deep copy of iteself.
        
        **Raises:**
        
        A RestartMarker is thrown if unable to make the copy.
        
        **Returns:**
        
        A new RestartMarker object with the same attributes as
        the old is returned.
        """
        
        ret, handle = ftpClientc.restart_marker_copy(self._handle)
        if not ret == 0:
            ex = RestartMarkerException(handle)
            raise ex
        return RestartMarker(handle)

    def insert_range(self, offset, endOffset):
        """
        Insert a range into the marker.
        
        This method updates the marker with a new byte range, suitable for using
        to restart an extended block mode transfer. Adjacent ranges within the marker
        will be combined into a single entry in the marker.
        
        A marker can only hold a range list or a stream offset. Calling this method after
        calling set_offset() will result in a marker suitable only for use restarting an
        extended block mode transfer.
              
        **Arguments:**
        
        - *offset* The beginning offset to start at. This may be an int or a python
        long, if your platform supports 64 bit files.
        
        - *endOffset* The offset to end at.
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to insert the
        restart range.
        """
        
        ret, error = ftpClientc.restart_marker_insert_range(self._handle,
                                                           offset,
                                                           endOffset)
        if not ret == 0:
            ex = RestartMarkerException(error)
            raise ex
        return

    def set_offset(self, offset):
        """
        Insert the offset to restart at.
        
        This method modifies the marker to contain a stream offset, suitable for
        using to restart a steam mode transfer.
        
        A marker can only hold a range list or a stream offset. Calling this method after
        calling insert_range() will delete the ranges associated with the marker, and
        replace it with a marker suitable only for use restarting a stream mode transfer. 
              
        When restarting an ASCII type transfer, the offset must take into account the
        additional carriage return characters added to the data stream.
        
        **Arguments:**
        
        - *offset* The offset to begin at. This may be an int, or a python long
        on platforms that support 64 bit files.
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to set the
        offset.
        """
        
        ret, error = ftpClientc.restart_marker_set_offset(self._handle,
                                                         offset)
        if not ret == 0:
            ex = RestartMarkerException(error)
            raise ex
        return

    def set_ascii_offset(self, offset, asciiOffset):
        """
        Set the offset for a restart marker.
        
        This method modifies the marker to contain a stream offset, suitable for
        using to restart a steam mode transfer.
        
        A marker can only hold a range list or a stream offset. Calling this method after
        calling insert_range() will delete the ranges associated with the marker, and
        replace it with a marker suitable only for use restarting a stream mode transfer. 
        
        When restarting an ASCII type transfer, use this method to set both the offset
        used in the local representation of an ACSII file, and the network representation
        of the ASCII file. For UNIX systems, the former includes counts newlines as one
        character towards the file offset, and the latter counts them as 2 characters
        (CRLF).
        
        **Arguments:**
        
        - *offset* The local stream offset.
        
        - *asciiOffset:** The networks ASCII representation of the offset.
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to set the
        ASCII offset.
        """
        
        ret, error = ftpClientc.restart_marker_set_ascii_offset(self._handle,
                                                               offset,
                                                               asciiOffset)
        if not ret == 0:
            ex = RestartMarkerException(error)
            raise ex
        return

    def get_total(self):
        """
        Get total bytes accounted for in restart marker.
        
        This funtion will return the sum of all bytes accounted for in a restart
        marker. If this restart marker contains a stream offset then this value
        is the same as the offset (not the ascii offset) that it was set with. If it
        is a range list, it a sum of all the bytes in the ranges.
        
        **Raises:**

        A RestartMarkerException is thrown if unable to get the
        total bytes.
        
        *Returns:**

        The total bytes in the marker..
        """
        
        ret, total = ftpClientc.restart_marker_get_total(self._handle)
        if not ret == 0:
            ex = RestartMarkerException(total)
            raise ex
        return total
    
    def to_string(self):
        """
        Return a string represenation of the marker.
        
        Returns a string suitable for sending to the FTP REST command, or for a
        later call to from_string().
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to get the
        marker string.
        
        *Returns:**
        
        A string representation of the marker.
        """
        
        ret, error, markerStr = ftpClientc.restart_marker_to_string(self._handle)
        if not ret == 0:
            ex = RestartMarkerException(error)
            raise ex
        return markerStr

    def from_string(self, markerStr):
        """
        Initialize the marker from a string.
        
        **Raises:**
        
        A RestartMarkerException is thrown if unable to initialize
        the marker from the given string.
        """
        
        ret, error = ftpClientc.restart_marker_from_string(self._handle,
                                                          markerStr)
        if not ret == 0:
            ex = RestartMarkerException(error)
            raise ex
        return

    def get_handle(self):
        """
        Return the underlying ftp client restart marker handle.
        
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**
        
        A string containing the pointer.
        """
        
        return self._handle


class Plugin:
    """
       A base class for the ftp client plugins.
       """
    
    def __init__(self):
        self._handle = None
        return
    
    def get_handle(self):
        """
        Return the underlying ftp client plugin handle.
        
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**
        
        A string containing the pointer.
        """
        
        return self._handle

class DebugPlugin(Plugin):
    """
    A wrapper class around the ftp client debug plugin.
    
    The FTP Debugging plugin provides a way for the user to trace FTP
    protocol messages which occur while the GridFTP client library processes
    an FTP operation. This may be useful for debugging FTP configuration problems.
    
    When this plugin is used for a GridFTP Client operation, information will be printed to
    the file stream associated with the plugin when a user begins an operation, for all
    data buffers which pass through while handling a data transfer, and for all protocol
    messages which are sent and received.
    """

    def __init__(self, file, text):
        """
        Initialize an instance of the GridFTP debugging plugin.
        
        This function will initialize the debugging plugin-specific instance data for this
        plugin, and will make the plugin usable for ftp client handle attribute and
        handle creation.
              
        **Arguments**
        
        *file* An open Python file object to write the debug messages to.
        
        *text* A string message to append to the debug info.
        
        **Raises**
        
        A PluginException is thrown if unable to init the plugin.
        """
        
        Plugin.__init__(self)
        ret = utilc.globus_module_activate(ftpClientc.get_debug_plugin_module())
        if not ret == 0:
            ex = PluginException("Unable to activate the module")
            raise ex
        ret, handle = ftpClientc.debug_plugin_init(file, text)
        if not ret == 0:
            ex = PluginException(handle)
            raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance of the GridFTP debugging plugin.
        
        This function will free all debugging plugin-specific instance data from this
        plugin, and will make the plugin unusable for further ftp handle creation.
        
        Existing FTP client handles and handle attributes will not be affected by
        destroying a plugin associated with them, as a local copy of the plugin is
        made upon handle initialization.
        
        **Raises**
        
        A PluginException is thrown if unable to destroy the plugin.
        """
        
        ret, err = ftpClientc.debug_plugin_destroy(self._handle)
        if not ret == 0:
            ex = PluginException(err)
            raise ex
        ret = utilc.globus_module_deactivate(ftpClientc.get_debug_plugin_module())
        if not ret == 0:
            ex = PluginException("unable to deactivate the module")
            raise ex
        return

class RestartMarkerPlugin(Plugin):
    """
    A wrapper class around the ftp client restart marker plugin.
    
    This plugin is intended to allow users to make restart markers persistant.
    During a transfer, every marker received will result in the user's 'marker'
    callback being called with the new restart marker that can be stored. If the
    application were to prematurely terminate (while transferring), the user
    (after restarting the application) could pass this stored marker back to the
    plugin via the 'begin' callback to force the transfer to be restarted from
    the last marked point.
    """

    def __init__(self, beginCB, markerCB, completeCB, arg):
        """
        Initialize an instance of the GridFTP restart marker plugin.
        
        This function will initialize the restart marker plugin-specific instance
        data for this plugin, and will make the plugin usable for ftp client
        handle attribute and handle creation.
        
        **Arguments**
        
        *beginCB* The callback to be called upon the start of the transfer.
        It has the following signature
        
        'function(arg, handle, srcUrl, destUrl, restart_marker)' where arg is the
        argument passed into this call, handle is a SWIG'ized pointer to a ftp
        client handle object, srcUrl is the a string representing the source url,
        destUrl is the a string representing the destination url, a SWIG'ized
        pointer to an uninitialized restart marker,
        
        The function must return an int indicating whether the transfer should
        continue from the restart or start from the beginning.
        
        *markerCB* The callback to be called with each restart marker
        received. It has the following signature
        
        'function(arg, handle, restart_marker)' where arg is the argument passed
        into this call, handle is a SWIG'ized pointer to a ftp client handle object,
        restart_marker is a SWIG'ized pointer to the restart marker that has been
        received
        
        *completeCB* The callback to be called to indicate transfer
        completion. It has the following signature
        
        'function(arg, handle, restart_marker, error, error_url)' where arg is
        the argument passed into this call, handle is a SWIG'ized pointer
        to a ftp client handle object, error is an error string indicating the
        error that occurred, error_url is the url which is the source of the
        above error
        
        *arg* An arbitary argument to be passed to all the callbacks.
        
        **Raises**
        
        A PluginException is thrown if unable to init the plugin.
        """
        
        Plugin.__init__(self)
        self._cbHandle = None
        ret = utilc.globus_module_activate(ftpClientc.get_restart_marker_plugin_module())
        if not ret == 0:
            ex = PluginException("Unable to activate the module")
            raise ex
        ret, handle, cbHandle = ftpClientc.restart_marker_plugin_init(beginCB,
                                                                     markerCB, 
                                                                     completeCB, 
                                                                     arg)
        if not ret == 0:
            ex = PluginException(handle)
            raise ex
        self._handle = handle
        self._cbHandle = cbHandle
        return

    def __del__(self):
        """
        Destroy an instance of the GridFTP restart marker plugin.
        
        This function will free all debugging plugin-specific instance data from this
        plugin, and will make the plugin unusable for further ftp handle creation.
        
        Existing FTP client handles and handle attributes will not be affected by
        destroying a plugin associated with them, as a local copy of the plugin is
        made upon handle initialization.
        
        **Raises**
        
        A PluginException is thrown if unable to destroy the plugin.
        """
        
        ret, err = ftpClientc.restart_marker_plugin_destroy(self._handle)
        if not ret == 0:
            ex = PluginException(err)
            raise ex
        if not self._cbHandle == None:
            ftpClientc.free_callback2(self._cbHandle)
        ret = utilc.globus_module_deactivate(ftpClientc.get_restart_marker_plugin_module())
        if not ret == 0:
            ex = PluginException("unable to deactivate the module")
            raise ex
    

class PerformanceMarkerPlugin(Plugin):
    """
    A wrapper class around the ftp client performance marker plugin.
    
    This plugin provides performance marker information via callbacks on all
    types of transfers except third_party_transfers w/o extended block mode 
    transfer.  During a transfer, every marker received will result in the
    user's 'marker' callback being called with new performance markers that
    can be stored.
    """

    def __init__(self, beginCB, markerCB, completeCB, arg):
        """
        Initialize an instance of the GridFTP performance marker plugin.
        
        This function will initialize the performance marker plugin-specific instance data for this
        plugin, and will make the plugin usable for ftp client handle attribute and
        handle creation.
        
        **Arguments**
        
        *beginCB* The callback to be called upon the start of the transfer. It has
        the following signature
        
        'function(arg, handle, srcUrl, destUrl, restart)' where arg is the argument passed
        into this call, handle is a SWIG'ized pointer to a ftp client handle object,
        srcUrl is the a string representing the source url, destUrl is the a string
        representing the destination url, is an int indicating whether this callback
        is the result of a restart
              
        *markerCB* The callback to be called with each performance marker
        received. It has the following signature
        
        'function(arg, handle, time_stamp, time_stamp_length, stripe_ndx, num_stripes, nbytes)'
        where arg is the argument passed into this call, handle is a SWIG'ized pointer to a ftp
        client handle object, time_stamp is an int indicating the timestamp at which the number
        of bytes is valid, time_stamp_length (not documented and I haven't gotten this to work MKR),
        stripe_ndx an int indicating the stripe index this data refers to, num_stripes an int indicating
        the total number of stripes involved in this transfer, nbytes an int indicating the total bytes
        transfered on this stripe.
        
        *completeCB* The callback to be called to indicate transfer
        completion.  It has the following signature
              
        'function(arg, handle, error)' where arg is the argument passed into this call,
        handle is a SWIG'ized pointer to a ftp client handle object, error is a tuple
        (error_code, error_description). If there is no error, the error_code
        is 0 and the error description is a None object.
              
        *arg* An arbitary argument to be passed to all the callbacks.
        
        **Raises**

        A PluginException is thrown if unable to init the plugin.
        """
        
        Plugin.__init__(self)
        self._cbHandle = None
        ret = utilc.globus_module_activate(ftpClientc.get_performance_marker_plugin_module())
        if not ret == 0:
            ex = PluginException("Unable to activate the module")
            raise ex
        ret, handle, cbHandle = ftpClientc.performance_marker_plugin_init(beginCB,
                                                                          markerCB, 
                                                                          completeCB, 
                                                                          arg)
        if not ret == 0:
            ex = PluginException(handle)
            raise ex
        self._handle = handle
        self._cbHandle = cbHandle
        return

    def __del__(self):
        """
        Destroy an instance of the GridFTP restart marker plugin.
        
        This function will free all debugging plugin-specific instance data from this
        plugin, and will make the plugin unusable for further ftp handle creation.
        
        Existing FTP client handles and handle attributes will not be affected by
        destroying a plugin associated with them, as a local copy of the plugin is
        made upon handle initialization.
              
        **Raises**
        
        A PluginException is thrown if unable to destroy the plugin.
        """
        
        ret, err = ftpClientc.performance_marker_plugin_destroy(self._handle)
        if not ret == 0:
            ex = PluginException(err)
            raise ex
        if not self._cbHandle == None:
            ftpClientc.free_callback2(self._cbHandle)
        ret = utilc.globus_module_deactivate(ftpClientc.get_performance_marker_plugin_module())
        if not ret == 0:
            ex = PluginException("unable to deactivate the module")
            raise ex

class ThroughputPerfPlugin(Plugin):
    """
    A wrapper class around the ftp client throughput performance plugin.
    This plugin is built on top of the Performance Marker Plugin, it is
    not possible to associate both plugins with a handle 
    
    This plugin provides throughput performance information via callbacks on
    all types of transfers except third_party_transfers w/o extended block
    mode transfer.  During a transfer, every marker received will result in
    the user's 'marker' callback being called with new performance markers
    that can be stored.
    """

    def __init__(self, beginCB, stripeCB, totalCB, completeCB, arg):
        """
        Initialize an instance of the GridFTP Throughput plugin.
        
        This function will initialize the Throughput plugin-specific instance data
        for this plugin, and will make the plugin usable for ftp client handle
        attribute and handle creation.
        
        **Arguments**
        
        *beginCB* The callback to be called upon the start of the transfer.
        It has the following signature
        
        'function(arg, handle, srcUrl, destUrl)' where arg is the argument passed
        into this call, handle is a SWIG'ized pointer to a ftp client handle object,
        srcUrl is the a string representing the source url, destUrl is the a string
        representing the destination url.
        
        *stripeCB* The callback to be called with every performance callback
        that is received by the perf plugin. The first callback for each
        stripe_ndx will have an instantaneous_throughput based from the time
        the command was sent. It has the following signature
        
        'function(arg, handle, stripe_ndx, bytes, instantaneous_throughput, avg_throughput)'
        where arg is the argument passed into this call, handle is a SWIG'ized pointer
        to a ftp client handle object, stripe_ndx an int representing which stripe the data
        was received on, bytes the total number of bytes received on this stripe,
        instantaneous_throughput is a float representing instanteous throughput
        on this stripe (bytes / sec), avg_throughput is a float representing average
        throughput on this stripe (bytes / sec)
              
        *totalCB* The callback will be called with every performance callback
        that is received by the perf plugin. The first callback for will have
        an instantaneous_throughput based from the time the command was sent.
        This callback will be called after the per_stripe_cb. It has the following
        signature
        
        'function(arg, handle, bytes, instantaneous_throughput, avg_throughput)'
        where arg is the argument passed into this call, handle is a SWIG'ized pointer
        to a ftp client handle object, bytes the total number of bytes received on this stripe,
        instantaneous_throughput is a float representing instanteous throughput
        on this stripe (bytes / sec), avg_throughput is a float representing average
        throughput on this stripe (bytes / sec)
        
        *completeCB* The callback to be called to indicate transfer
        completion. It has the following signature
        
        'function(arg, handle, success)' where arg is the argument passed into this call,
        handle is a SWIG'ized pointer to a ftp client handle object, success is an int
        representing whether the transfer completed successfully
        
        *arg* An arbitary argument to be passed to all the callbacks.
        
        **Raises**
        
        A PluginException is thrown if unable to init the plugin.
        """
        
        Plugin.__init__(self)
        self._cbHandle = None

        ret = utilc.globus_module_activate(ftpClientc.get_throughput_plugin_module())
        if not ret == 0:
            ex = PluginException("Unable to activate the module")
            raise ex
        ret, handle, cbHandle = ftpClientc.throughput_plugin_init(beginCB, 
                                                                  stripeCB,
                                                                  totalCB, 
                                                                  completeCB, 
                                                                  arg)
        if not ret == 0:
            ex = PluginException(handle)
            raise ex
        self._handle = handle
        self._cbHandle = cbHandle
        return

    def __del__(self):
        """
        Destroy an instance of the GridFTP throughput plugin.
        
        This function will free all debugging plugin-specific instance data from this
        plugin, and will make the plugin unusable for further ftp handle creation.
        
        Existing FTP client handles and handle attributes will not be affected by
        destroying a plugin associated with them, as a local copy of the plugin is
        made upon handle initialization.
        
        **Raises**
        
        A PluginException is thrown if unable to destroy the plugin.
        """
        
        ret, err = ftpClientc.throughput_plugin_destroy(self._handle)
        if not ret == 0:
            ex = PluginException(err)
            raise ex
        if not self._cbHandle == None:
            ftpClientc.free_callback3(self._cbHandle)
        ret = utilc.globus_module_deactivate(ftpClientc.get_throughput_plugin_module())
        if not ret == 0:
            ex = PluginException("unable to deactivate the module")
            raise ex
        
class NetloggerPlugin(Plugin):
    """
    A wrapper class around the netlogger plugin.

    """

    def __init__(self, nl_url, prog_name, opaque_string=None):
        """
        Initialize an instance of the netlogger plugin.
	
        This function will initialize the NetloggerPlugin instance data for this
        plugin, and will make the plugin usable for ftp client handle attribute and
        handle creation.
	
        **Arguments**
        
	*nl_url*  the url to log to. (May be None)
	Valid urls are:  B{file://tmp/netlog.log, x-netlog://host[:port], x-syslog://localhost}
        
	*prog_name* This is used as the prog name in the NetLoggerOpen call
          
	*opaque_string*  this is an opaque string that will be inserted into
	all logged statements. (may be None)

        **Raises**

	A PluginException is thrown if unable to init the plugin.
        """
        Plugin.__init__(self)
        ret = utilc.globus_module_activate(ftpClientc.get_netlogger_plugin_module())
        if ret != 0:
            raise PluginException("Unable to activate the module")
        ret, handle = \
             ftpClientc.netlogger_plugin_init(nl_url, prog_name, opaque_string)
        if ret != 0:
            raise PluginException(handle)
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance of the netlogger plugin.

        This function will free all debugging plugin-specific instance data from this
        plugin, and will make the plugin unusable for further ftp handle creation.

        Existing FTP client handles and handle attributes will not be affected by
        destroying a plugin associated with them, as a local copy of the plugin is
        made upon handle initialization.

        **Raises**

	A PluginException is thrown if unable to destroy the plugin.
        """
        ret, err = ftpClientc.netlogger_plugin_destroy(self._handle)
        if ret != 0:
            raise PluginException(err)
        ret = utilc.globus_module_deactivate(ftpClientc.get_netlogger_plugin_module())
        if ret != 0:
            raise PluginException("unable to deactivate the module")


