############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides support for the lower-level Ftp Control module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
from pyGlobus.util import CallbackQueue
from pyGlobus.util import CallbackDefault
from pyGlobus.util import CallbackData
import ftpControlc

###############################################################################
# Module constants
##############################################################################
MODE_NONE           = 0
MODE_STREAM         = 'S'
MODE_BLOCK          = 'B'
MODE_EXTENDED_BLOCK = 'E'
MODE_COMPRESSED     = 'C'

TYPE_NONE    = 0
TYPE_ASCII   = 'A'
TYPE_EBCDIC  = 'E'
TYPE_IMAGE   = 'I'
TYPE_LOCAL   = 'L'

STRIPING_NONE = 0
STRIPING_PARTITIONED = 1
STRIPING_BLOCKED_ROUND_ROBIN = 2

PARALLELISM_NONE  = 0
PARALLELISM_FIXED = 1

PROTECTION_CLEAR        = 'C'
PROTECTION_SAFE         = 'S'
PROTECTION_CONFIDENTIAL = 'E'
PROTECTION_PRIVATE      = 'P'

DCAU_NONE = 'N'
DCAU_SELF = 'A'
DCAU_SUBJECT = 'S'

#############################################################################
# Module classes
#############################################################################

class FtpControlException(GlobusException):
    pass

class DCAUException(GlobusException):
    pass

class LayoutException(GlobusException):
    pass

class ParallelismException(GlobusException):
    pass

class TcpBufferException(GlobusException):
    pass

class ResponseException(GlobusException):
    pass

class AuthInfoException(GlobusException):
    pass

class CommandException(GlobusException):
    pass

class HostPortException(GlobusException):
    pass

class EofArrayException(GlobusException):
    pass

class FtpControl:
    """
    Class that wraps a ftp control API
    """

    def __init__(self, handle = None):
        """
        Constructs an instance.
        
        Activates the underlying GLOBUS_FTP_CONTROL_MODULE
        
        **Arguments**
        
        - *handle* a SWIG'ized pointer to a globus ftp control handle object
        
        **Returns**
        None
        
        **Raises**
        A FtpControlException is raised if the globus module cannot be activated
        or the handle cannot be initialized.
        """
        
        self._activated = 0
        self._cbQueue = CallbackQueue() 
        ret = utilc.globus_module_activate(ftpControlc.get_module())

        if not ret == 0:
            ex = FtpControlException("Unable to initialize the module")
            raise ex
        self._activated = 1

        if handle == None:
            ret, handle = ftpControlc.handle_init()

        if not ret == 0:
            ex = FtpControlException("unable to create the handle")
            raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Deletes the instance and frees the memory associated with the handle
        
        **Returns:**
        None
        
        **Raises**
        A FtpControlException is raised if the globus module cannot be deactivated
        or the handle cannot be freed.
        """

        if not self._handle == None:
            ret, error = ftpControlc.handle_destroy(self._handle)
            if not ret == 0:
                ex = FtpControlException(error)
                raise ex
 
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = FtpControlException("Unable to deactivate the module")
            raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        """
        Gets a pointer to the handle.
        
        **Returns**
        A SWIG'ized pointer to the handle object
        """
        
        return self._handle


    def data_connect(self, callback, arg):
        """
        Create an outgoing FTP data connection.
        
        This method will register a globus_io_{accept, connect}. Further
        accepts/connects are done by registering a new accept/connect in
        the current accept/connect callback. A call to either local_pasv() or
        local_port() needs to precede this calling this function. This method may
        be followed by a globus_ftp_data_write.
        
        **Arguments**
        
        - *callback* A Python function to be called when the connection occurs. It has
        the following signature:
        
        'function(arg, handle, stripe_ndx, reuse, error)'
        where arg is the argument passed into this call, handle is a SWIG'ized pointer
        to a ftp control handle object, stripe_ndx is an int representing which strip
        the connect is associated with, reuse is a boolean whose value is dependent
        on whether the connnection can be reused and error is a tuple
        (error_code, error_description). If there is no error, the error_code is
        0 and the error description is a None object
        
        - *arg* A user argument to be returned when the callback is executed.
        
        **Returns**
        
        A SWIG'ized pointer to a callback handle
        
        **Raises**
        
        A FtpControlException is raised if there is a problem with the registering
        connection
        """
        
        ret, cbHandle = ftpControlc.data_connect_write( self._handle,
                                                       callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def data_write(self, buffer, offset, eof, callback, arg):
        """
        Writes data from the supplied buffer to data connection(s)
        
        This method writes contained in the buffer to the data
        channel(s).
        
        **Arguments**
        
        - *buffer* A string of the data to be written out.
        
        - *offset* An int which is the offset in the file at which the data in the buffer starts
        
        - *eof*  Indicates that the buffer is that last part of a file. In
        the striped case this will cause a EOF block to be send to
        every data node involved in the transfer.
        
        - *callback* A Python function to be called once the data has been sent. It
        has the following signature:
        
        'func(arg, handle, bufHandle, bufLen, offset, eof)'
        where arg is the user argument passed into this call, handle is the
        underlying ftp_control handle, bufHandle is a SWIG'ized pointer to the
        underlying globus_byte_t array, bufLen is the length of the underlying
        globus_byte_t array, offset is the offset into the file this data block
        contains, and eof is a bool that is true if the end of the data transfer
        has been reached.
        
        - *arg* A user argument to be passed through to the callback.
        
        **Returns**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        
        A FtpControlException is thrown if unable to do the write.
        """
        
        ret, cbHandle = ftpControlc.data_write( self._handle,
                                                buffer, 
                                                offset, eof,
                                                callback, arg)
        
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle,CallbackData(cbHandle))
        return cbHandle

    def local_pasv(self):
        """
        Create a local listening socket, bind it and return the address the
        socket is listening to. If there is a existing data connection it
        is closed.
        
        **Returns**
        A HostPort object is returned
        
        **Raises**
        A FtpControlException is raised if the listening socket cannot be created
        """ 

        ret, handle  = ftpControlc.local_pasv( self._handle)
        if not ret == 0:
            ex = FtpControlException(handle)
            raise ex
        return HostPort(0, 0, handle)
        
    def local_port(self, address):
        """
        Insert the host/port information returned by a PASV on the remote
        host into the local FTP control handle. (close any outstanding data
        con)
        
        **Arguments**

        - *address* A ftpControl.HostPort object
        
        **Returns**
        None
        
        **Raises**
        A FtpControlException is raised if the host/port information cannot be
        put into ftp_control handle
        """

        ret, err  = ftpControlc.local_port( self._handle, address.get_handle())
        if not ret == 0:
            ex = FtpControlException(err)
            raise ex
        return 
    
    def read_commands(self, callback, arg):
        """
	Begin reading GSIFTP commands.  When a command is read the callback
	function is called with its respective arguments. 
	
	**Arguments**

        - *callback*  The Python function that will be called when commands are read.
        It has the following signature:

        'func(arg, handle, error, command)'
        where arg is the user argument passed into this call, handle is a
        SWIG'ized pointer to the underlying ftp_control handle, error is a tuple
        (error_code, error_description). If there is no error, the error_code is 0
        and the error description is a None object, command is a SWIG'ized pointer
        to a ftp_control_command handle.
        
        - *arg*       The user argument passed to the callback.
	
	**Returns**
	Returns a SWIG'ized callback handle.     
	
	**Raises**
	A FtpControlException is raised if the handle argument is NULL, if
	the handle is not connected or another operation is in progress.
	"""

        ret, cbHandle = ftpControlc.read_commands( self._handle,
                                                  callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle, CallbackDefault(cbHandle))
        return cbHandle
    
    def send_response(self, resp, callback, arg):
        """
	Send a response to the GSIFTP client.  This method sends a GSIFTP 
	formatted response to the client.  When a command callback is received 
	the user calls this method to respond to the user's request.
	
	**Arguments**
        
	- *resp*       A formatted string representing the user's response.

        - *callback*   A Python function that will be called when the response
	has been sent. It has the following signature:

        'func(arg, handle, error)'
        where arg is the user argument passed into this call, handle is a
        SWIG'ized pointer to the underlying ftp_control handle, error is a tuple
        (error_code, error_description). If there is no error, the error_code is 0
        and the error description is a None object.

        - *arg*        The user argument passed to the callback.

	**Returns**
        
	Returns a SWIG'ized callback handle.     

	**Raises**
	Raises a FtpControlException if response could not be sent. 	     
	"""

        ret, cbHandle = ftpControlc.send_response(self._handle, resp, 
                                                  callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def raw_command(self, command):
        """ 
	Retrieves a raw command from a globus ftp control command.
	
	**Arguments**

        - *command*  The globus command to be assessed. 
	
	**Returns**
	The raw command (string) corresponding with the command..
	
	**Raises**
	None.	      	      
	"""
	
        ret = ftpControlc.raw_command(command)
        return ret

    def code(self, command):
        """
	Retreives a command code from a globus ftp control command.
	
	**Arguments**
        
	- *command*  The globus command to be assessed.. 
	
	**Returns**
	The code corresponding with the command.
	
	**Raises**
	None	      	      	      
	"""

        ret = ftpControlc.code(command)
        return ret

    def operation(self, command):
        """
	Retrieves the operation associated with a globus ftp control command. 

	**Arguments**

        - *command*   The globus command to be assessed.
	
	**Returns*
        The operation (buffer) associated with the ftp control command.
	"""

        ret = ftpControlc.operation(command)
        return ret

    def create_data_info(self,buffer,offset, eof, callback, arg):
        """
	Creates a write_info data structure,
	
	**Arguments**

        - *buffer*  is a pyGlobus.util Buffer object

        - *offset*  file offset of the data specified

        - *eof*     boolean denoting whether an end of file has been reached

        - *callback* A Python function that is called when all the writes from
        data_write_stripe have occured and release_data_info has been called.
        It has the following signature:

        'func(user_arg, handle, buffer, length, offset, eof, error)'
        user_arg is the arg supplied to this function, handle is a SWIG'ized
        pointer to the ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the error_code is 0
        and the error description is a None object, buffer is a Python Buffer object
        which has the data that was used in the register_write_stripe call,
        length is an int which is the amount of data in the buffer, offset
        is an int which is the file offset of the data in the buffer,
        eof is an int indicating whether all of the data has been transfered
        
        - *arg*      a user supplied argument to the callback function
	
	**Returns**
	Returns a SWIG'ized callback handle, and a SWIG'ized write_info pointer.
	
	**Raises**
	Raises a FtpControlException if the write_info object could not be created.
	"""

        ret, cbHandle, write_info = ftpControlc.create_data_info(self._handle,
								 buffer.get_handle(),
								 buffer.get_size(),
								 offset, eof,
								 callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle, write_info

    def release_data_info(self, write_info):
        """        
	Releases the memory associated with the write_info object
	Must be called after each  create_data_info call
	
	**Arguments**

        - *write_info* a SWIG'ized pointer to a write_info object

	**Returns**
	Nothing
        		   
	**Raises**
	Raises a FtpControlException if the memory to the write_info object cannot be released
	"""
		   
        ret, error = ftpControlc.release_data_info(self._handle, write_info)
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return


    def data_write_stripe(self,buffer,offset,eof,stripe_ndx, callback, arg):
        """
        Writes data to the specified stripe
        
        **Arguments**

        - *buffer* a string of the data to be written out

        - *offset* the offset into file that has been written so far

        - *eof* boolean flag if it is the last data to be written

        - *stripe_ndx* the stripe which the data is written to.

        - *callback* a python function that has the signature
        func(user_arg, handle, error, buffer, length, offset, eof)

        - *arg*      a user supplied arguement to the callback function
        
        **Returns**
        Returns a SWIG'ized callback handle
        
        **Raises**
        Raises a FtpControlException if the data could not be written.
        """

        ret, cbHandle = ftpControlc.data_write_stripe(self._handle,
						      buffer,
						      len(buffer),
						      offset, eof, stripe_ndx,
						      callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackData(cbHandle))
        return cbHandle


    def x_data_write_stripe(self, buffer, offset, eof, stripe_ndx,write_info):
        """
	Writes data to the specified stripe

        **Arguments**

        - *buffer* a pyGlobus.util buffer object

        - *offset* the offset into file that has been written so far

        - *eof* boolean flag if it is the last data to be written

        - *stripe_ndx* the stripe which the data is written to.

        - *write_info* a SWIG'ized pointer to a write_info object
        
        *Returns*
        None
        
        **Raises**
        raises a FtpControlException if the data could not be written
        """

        ret, error = ftpControlc.x_data_write_stripe(self._handle,
						     buffer.get_handle(),
						     buffer.get_size(),
						     offset, eof, stripe_ndx,
						     write_info)
						    
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return

    def control_connect(self, host, port, callback, arg):
        """
        Create a new control connectiong to an FTP server.
        **Arguments**

        - *host*      the hostname on a FTP Server

        - *port*       the TCP port number on a FTP Server

        - *callback* a Python function that has the signature:
        
        'func(user_arg,handle, error, ftp_response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.

        - *arg*        the arg to the callback function
        
        **Returns**
        returns a SWIG'ized callback handle
        
        **Raises**
        raises a FtpControlException if a connection could not be initiated
        """

        ret, cbHandle = ftpControlc.control_connect(self._handle,host, 
                                                    port, callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def authenticate(self, auth_info, use_auth, callback, arg):
        """
        authenticate the user to the FTP server

        **Arguments**

        - *auth_info*    a Python object which wraps a globus_auth_info object

        - *use_auth*    a boolean which if true GSS authentication will be used,
        otherwise username, password authentication will be used

        - *callback*  a Python function that will be called when the client
        has authenticated to the server.
        It has the following signature:

        'func(arg,handle, error, response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.

        - *arg*    the arg to the callback function
        
        **Returns**
        Returns a SWIG'ized callback handle
        
        **Raises**
        raises a FtpControlException if the command could not be sent
        """
        
        ret, cbHandle = ftpControlc.authenticate(self._handle,
                                                 auth_info.get_handle(),
						 use_auth, callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle
    
    def send_command(self,cmdspec, callback, arg):
        """
        send a command to the FTP server and register a response handler

        **Arguments**

        - *cmdspec* a string which has the command to be sent to the FTP Server

        - *callback* A Python function that is called when the authentication
        process completes or an error occurs. It has the following signature:

        'func(user_arg,handle, error, response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.
        
        - *arg*        the user_arg to the callback function
        
        **Returns**
        Returns a SWIG'ized callback handle
        
        **Raises**
        Raises a FtpControlException if the command could not be sent
        """
	      
        ret, cbHandle = ftpControlc.send_command(self._handle, cmdspec, 
                                                 callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def abort(self, callback, arg):
        """
	Send an ABORT to the FTP server and register a response handler
              
	**Arguments**

        - *callback* a Python function that is called when the authentication
        process completes or an error occurs. It has the following signature:

        'func(arg,handle, error, response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.
        
        - *arg*        the user_arg to the callback function
	
	**Returns**
	returns a SWIG'ized callback handle

        **Raises**
        Raises FtpControlException if the ABORT message could not be sent
        """
	      
        ret, cbHandle = ftpControlc.py_abort(self._handle,callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def quit(self, callback, arg):
        """
	send a QUIT to the FTP server and register a response handler

	**Arguments**               

        - *callback* a Python function that is called when the authentication
        process completes or an error occurs. It has the following signature:

        'func(arg,handle, error, response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.
        
        - *arg*      The arg to the callback function
	
	**Returns**
	Returns a SWIG'ized callback handle

	**Raises**
	an FtpControlException is raised if the QUIT message could not be sent
	"""

        ret, cbHandle = ftpControlc.py_quit(self._handle,callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def force_close(self, callback, arg):
        """
	Forces a close of the control connection without sending a QUIT or
        waiting for outstanding commands to complete.
              
	**Arguments**            

        - *callback* a Python function that is called when the authentication
        process completes or an error occurs. It has the following signature:
        
        'func(arg,handle, error, response)'
        where arg is the arg supplied to this method, handle is
        a SWIG'ized pointer to a ftp control handle, error is a tuple
        (error_code, error_description). If there is no error, the
        error_code is 0 and the error description is a None object, and
        response is a ftp control response handle.
        
        - *arg*        the arg to the callback function
	
	**Returns**
	returns a SWIG'ized callback handle
	
	**Raises**
	raises a FtpControlException if the channel could not be closed
	"""

        ret, cbHandle = ftpControlc.force_close(self._handle,callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle


    def data_connect_read(self, callback, arg):
        """
	Create an incoming FTP data connection
	
	**Arguments:**
        
	- *callback* a Python function that is called when the authentication
        process completes or an error occurs. It has the following signature:

        'func(arg, handle, stripe_ndx, reuse, error)'
        where arg is the argument passed into this call, handle is a SWIG'ized pointer
        to a ftp control handle object, stripe_ndx is an int representing which strip
        the connect is associated with, reuse is a boolean whose value is dependent
        on whether the connnection can be reused and error is a tuple
        (error_code, error_description). If there is no error, the error_code is
        0 and the error description is a None object.
           
	- *arg* the arg to the callback function
	
	**Returns:**
	Returns a SWIG'ized callback handle
               
	**Raises**
	Raises a FtpControlException if the data connection could not be created
	"""
	      
        ret, cbHandle = ftpControlc.data_connect_read(self._handle,callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def data_add_channels(self,num_channels, stripe_ndx):
        """
	Opens additional data channels (connections) to the host identified by the stripe parameter

        **Arguments**

        - *num_channels*  The number of additional channels to add

        - *stripe_ndx*  The stripe that the channels will belong to
        
        **Returns**
        None
        
        **Raises**
        raises a FtpControlException if the channels could not be added
        """

        retString = ftpControlc.data_add_channels(self._handle, num_channels, 
                                                  stripe_ndx)
        if not retString == "NONE":
            ex = FtpControlException(retString)
            raise ex
        return

    def data_send_eof(self, eof_array, eof, callback, arg):
        """
	Sends an eof message to each stripe along an open data connection
	
	**Arguments**

        - *eof_array* a python argument which wraps an array of integers.
	The array of integers should contain an integer that will be added
        to the current parallel data connection count on each stripe.
        The order of the integers corresponds to th eeach stripe in the
        same order as what was returned from local_port()

        - *eof* boolean which is the eof message to be sent to all of the
        channels specified in the eof array

        - *callback* A Python function that will be called when the eof message
        has been called. It has the following signature:

        'func(user_arg, handle, error)'
        where arg is the user argument passed into this call, handle is a
        SWIG'ized pointer to the underlying ftp_control handle, error is a tuple
        (error_code, error_description). If there is no error, the error_code is 0
        and the error description is a None object.
        
        - *arg* the user_arg for the callback function
	
	**Returns**
	returns a SWIG'ized callback handle
	
	**Raises**
	raises a FtpControlException if the eofs could not be sent
	"""

        ret, cbHandle =  ftpControlc.data_send_eof(self._handle,
						   eof_array.get_handle(),
						   eof_array.get_size(), eof,
						   callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        self._cbQueue.append_callback(cbHandle,CallbackDefault(cbHandle))
        return cbHandle

    def data_remove_channels(self, num_channels, stripe_ndx):
        """
	Removes data channels, to the host identified by the stripe parameter

        **Arguments**

        - *num_channels*  The number of additional channels to add

        - *stripe_ndx*  The stripe that the channels will belong to

	**Returns**
	None

	**Raises**
        raises a FtpControlException if the channels could not be removed
        """
        
        ret, error = ftpControlc.data_remove_channels(self._handle, 
                                                      num_channels, 
                                                      stripe_ndx) 

        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return

    def data_query_channels(self, stripe_ndx):
        """
	Returns the number of currently open channels for the host identified by the stripe parameter
        **Arguments**

        - *stripe_ndx*  The stripe that you wish to query
        
        **Returns**
        number of channels on that stripe
        
        **Raises**
        raises a FtpControlException if the stripe could not be queried
        """
        ret, num_channels = ftpControlc.data_query_channels(self._handle, 
                                                            stripe_ndx)
        if not ret == 0:
            ex = FtpControlException(num_channels)
            raise ex

        return num_channels

    def get_total_data_channels(self, stripe_ndx):
        """
	Returns the number of data channels used so far on this particular stripe

        **Arguments**

        - *stripe_ndx*  The stripe that you wish to query
	
	**Returns**
	number of channels on that stripe
	
	**Raises**
        raises a FtpControlException if the number of channels could not be determined
        """

        ret, num_channels = ftpControlc.data_get_total_data_channels(self._handle, stripe_ndx)
        if not ret == 0:
            ex = FtpControlException(num_channels)
            raise ex

        return num_channels

    def local_send_eof(self, eof):
        """
	Determines if the library will automatically send an EOF message in extended block mode,
	or if the user will have to explicity do it by calling data_send_eof
	
	**Arguments**

        - *eof* boolean whether to set the handle to have a local eof
	
	**Returns**
	None

	**Raises**
	raises a FtpControlException if the local_eof could not be set
	"""

        ret, error = ftpControlc.local_send_eof(self._handle, eof)
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return
    
    def local_parallelism(self, parallelism):
        """
	Set the parallelism information in a FTP control handle

	**Arguments**

        - *parallelism* a ftpControl.Parallelism object
        
        **Returns**
        None
        
        **Raises**
        a FtpControlException is raised if the parallelism could not be set
        """
        
        ret, error = ftpControlc.local_parallelism(self._handle, 
                                                   parallelism.get_handle())
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return

    def local_spas(self, address_array):
        """
        Striped Passive -- enables parallelism and striping
        
	Create num_addresses local listening sockets,
	bind them and return the addresses the sockets are listening to

        **Arguments**

        - *address_array* a Python object that wraps an array of  host_port_t objects
        
        **Returns**
        the address_array object
        
        **Raises**
        a FtpControlException is raised if the local_spas cannot be set

        NOTE:
        This function is not implemented,
        so an unimplemented exception will always be thrown
        """
	      
        ret, address_handle = ftpControlc.local_spas(self._handle,
						     address_array.get_handle(),
						     address_array.get_size())
        if not ret == 0:
            ex = FtpControlException(address_handle)
            raise ex 

        host_array = HostPortArray(address_array.get_size(),address_handle) 
        return host_array

    def local_spor(self,address_array):
        """
        Striped Port -- enables parallelism and striping
	Insert the host/port addresses returned by a SPOR on the remote host into the local FTP control handle.
        **Arguments**

        - *address_array* a Python object that wraps an array of  host_port_t objects

        **Returns**
        the address_array object
        
        **Raises**
        a FtpControlException is raised if the local_spor cannot be set
        
        NOTE:
        This function is not implemented,
        so an unimplemented exception will always be thrown
        """
        
        ret, address_handle = ftpControlc.local_spor(self._handle,
						     address_array.get_handle(),
						     address_array.get_size())
        if not ret == 0:
            ex = FtpControlException(address_handle)
            raise ex
        
        host_array = HostPortArray(address_array.get_size(), address_handle)
        return host_array

    def local_mode(self, mode):
        """
        Update the ftp handle with the given mode information

        **Arguments**

        - *mode*  a FtpControl Mode type
        
        **Returns**
        None
        
        **Raises**
        a FtpControlException is raised if the local_mode cannot be set
        """
	      
        ret, error = ftpControlc.local_mode(self._handle, mode)
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return

    def local_tcp_buffer(self, tcp_buffer):
        """
        Update the ftp handle with the given socket buffer size information
        **Arguments**

        - *tcp_buffer* a ftpControl.TcpBuffer Object
        
        **Returns**
        None
        
        **Raises**
        a FtpControlException is raised if the local_tcp_buffer cannot be set
        The TcpBuffer.set_fixed call must be made on the tcp_buffer object before
        this call is made
        """
	      
        ret, error = ftpControlc.local_tcp_buffer(self._handle, 
                                                  tcp_buffer.get_handle())
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex

        return

    def local_dcau(self, dcau, credential):
        """
        Channel authentication.
        Update the ftp handle with the given data channel authentication information
        **Arguments**
        
        - *dcau*  a Python DCAU

        - *credential* a Python GSSCred object
        
        **Returns**
        None
        
        **Raises**
        a FtpControlException is raised if the local_dcau cannot be set
        """
	      
        ret,error = ftpControlc.local_dcau(self._handle, dcau.get_handle(),
					   credential.get_handle())
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex
        return

    def local_type(self, dataType):
        """
        Update the FTP control handle with the given type information.

        **Arguments**

        - *dataType* - A Python string with a single character either 'A'
        or 'I', for either ascii or image data, which are the only supported
        types

        **Returns**
        None

        **Raises**
        a FtpControlException is raised if the local_type cannot be set
        """

        ret, error = ftpControlc.local_type(self._handle, dataType)
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex
        
    def local_pbsz(self, size):
        """
        Update the ftp handle with the given protection buffer size information
        **Arguments**
        
        - *size* the size that the buffer will be during a protected transfer
        
        **Returns**
        None
        
        **Raises**
        a FtpControlException is raised if the local_pbsz cannot be set
        """
        
        ret, error = ftpControlc.local_pbsz(self._handle, size)
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex
        return

    def get_pbsz(self):
        """
        Return the size of the protection buffer size information
        
        **Returns**
        the size of the buffer during protected transfers
        
        **Raises**
        a FtpControlException is raised if the pbsz cannot be retrieved
        """
	      
        ret, size = ftpControlc.get_pbsz(self._handle)
        if not ret == 0:
            ex = FtpControlException(size)
            raise ex
        return size
 
    def local_stru(self, structure):
        """
	Update the handle information on the structure of the data being sent on the data channel
	
	**Returns**
	Nothing.
	
	**Raises**
	Raises a FtpControlException if the handle information was not sent.
	"""

        ret, error = ftpControlc.local_stru(self._handle, structure.get_handle())
        if not ret == 0:
            ex = FtpControlException(error)
            raise ex
        return

    def data_read(self, buffer, callback, arg):
        """
	Reads data from data connection(s) and puts it in the specified buffer.
	
	**Arguments**

        - *buffer*   Destination buffer- where the data gets written.

        - *callback* A Python function that is called when the data has been
        read. It has the following signature:

         'func(arg, handle, bufHandle, bufLen, offset, eof)'
        where arg is the user argument passed into this call, handle is the
        underlying ftp_control handle, bufHandle is a SWIG'ized pointer to the
        underlying globus_byte_t array, bufLen is the length of the underlying
        globus_byte_t array, offset is the offset into the file this data block
        contains, and eof is a bool that is true if the end of the data transfer
        has been reached.
        
	- *arg*      a user supplied arguement to the callback function	      
	
	**Returns**
	Returns a SWIG'ized callback handle. 
	
	**Raises**      
	"""

        ret, cbHandle = ftpControlc.data_read(self._handle,
                                              buffer,
                                              buffer.get_handle(),
                                              buffer.get_size(),
                                              callback, arg)
	
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex

        
        self._cbQueue.append_callback(cbHandle, CallbackData(cbHandle))
        return cbHandle

    def free_callbackDefault(self, callbackHandle):
        """
	Frees the memory associated with the callback handle

	**Arguments**
	- *callbackHandle* - a SWIG'ized pointer to a callback handle

	**Returns**
	None

	**Raises**
	None
	"""

        self._cbQueue.remove_callback(callbackHandle)
        
    def free_callbackData(self, callbackHandle):
        """
	Frees the memory associated with the callback handle

	**Arguments**
	- *callbackHandle* - a SWIG'ized pointer to a callback handle

	**Returns**
	None

	**Raises**
	None
	"""

        self._cbQueue.remove_callback(callbackHandle)
        
class HostPortArray:
    """
    Class that wraps an array of host_port
    """

    def __init__(self,size, handle = None):
        self._activated = 0
        self._size = size
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = HostPortException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.host_port_array_init(size)
        self._handle = handle    
        
    def __del__(self):
        ftpControlc.host_port_array_destroy(self._handle)

    def get_handle(self):
        return self._handle

    def get_size(self):
        return self._size

class HostPort:
    """
    Class that wraps the host_port_t struct
    """

    def __init__(self,host, port, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = HostPortException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.host_port_init(host, port)
            
        self._handle = handle

    def __del__(self):
        ftpControlc.host_port_destroy(self._handle)

    def get_handle(self):
        return self._handle

    def get_info(self):
        i,j,k,l, port = ftpControlc.host_port_info(self._handle)
        ipString = str(i) + "." + str(j) + "." + str(k)+"."+str(l)
        return ipString, port
    
class EofArray:
    """This class wraps an array of integers;
    an instance of this class should be created to use FtpControl.data_send_eof()
    """
    def __init__(self, size):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = EofArrayException("Unable to initialize the module")
            raise ex
        self._activated = 1
        self._size = size
        handle = ftpControlc.eof_array_init(size)
        self._handle = handle

    def __del__(self):
        ftpControlc.eof_array_destroy(self._handle)

    def get_handle(self):
        return self._handle
    
class Command:
    
    def __init__(self, raw_command,auth_info, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = CommandException("Unable to initialize the module")
            raise ex
        self._activated = 1 
        if handle == None:
            handle = ftpControlc.command_init(raw_command, 
                                              auth_info.get_handle())
        self._handle = handle
        self._auth_info = auth_info
        self._raw_command = raw_command
        return

    def __del__(self):
        ftpControlc.command_destroy(self._handle)
        ret =  utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = CommandException("Unable to deactivate the module")
            raise ex
        return

    def copy(self):
        ret, new_handle = ftpControlc.command_copy(self._handle)
        if not ret == 0:
            ex = CommandException(new_handle)
            raise ex
        com = Command(self._raw_command, self._auth_info, new_handle)
        return com
    
class DCAU:

    def __init__(self, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = DCAUException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.dcau_init()
        self._handle = handle
        return

    def __del__(self):
        ftpControlc.dcau_destroy(self._handle)
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = DCAUException("Unable to deactivate the module")
            raise ex
        return

    def set_mode(self, mode):
        ret = ftpControlc.dcau_set_mode(self._handle, mode)
        if not ret == 0:
            ex = DCAUException("Unable to set the mode")
            raise ex
        return

    def set_subject(self, subject):
        ret = ftpControlc.dcau_set_subject(self._handle, subject)
        if not ret == 0:
            ex = DCAUException("unable to set the subject")
            raise ex
        return
    
    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle

class Layout:

    def __init__(self, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = LayoutException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.layout_init()
        self._handle = handle
        return

    def __del__(self):
        ftpControlc.layout_destroy(self._handle)
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = LayoutException("Unable to deactivate the module")
            raise ex
        return

    def set_striping_mode(self, mode):
        ret = ftpControlc.layout_set_striping_mode(self._handle, mode)
        if ret == -42:
            ex = LayoutException("unknown striping mode")
            raise ex
        if not ret == 0:
            ex = LayoutException("unable to set the striping mode")
            raise ex
        return

    ######################################################################
    # XXX This call is only useful in round robin mode.  Also on 64bit
    # platoforms size should be a python long.
    #####################################################################
    def set_round_robin_block_size(self, size):
        ftpControlc.layout_set_round_robin_block_size(self._handle,
                                                     size)
        return

    #######################################################################
    # XXX This call is only useful in partitioned mode.  Also the size
    # argument should be a python long on 64 bit machines.
    ######################################################################
    def set_partition_block_size(self, size):
        ftpControlc.layout_set_partition_block_size(self._handle, size)
        return
    
    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle


class Response:
    """
    Proxy class for the globus_ftp_control_response_t object
    """
    def __init__(self, handle):
        """
        The handle is given to you in a callback,
        it can't be initialized any other way
	"""
        self._handle = handle
        

    def __del__(self):
        """
        frees the memory associated with the handle
        The memory associated with the handle is freed by globus
        """
        pass

    def copy(self):
        """
        **Returns**
        A Response object which is a copy of the response

        **Raises**
        A ResponseException is raised if there is a problem
        making a copy of the response
        """
        ret, response_copy = ftpControlc.response_copy(self._handle)
        if not ret == 0:
            ex = ResponseException(response_copy)
            raise ex

        return Response(response_copy)

    def info(self):
        """
        Returns a dictionary of information given in the response

        **Returns**
        A dictionary with 2 keys:
        'code' which maps to the return code
        'respose' which maps to a response string

        If the response handle is NULL then an empty dictionary
        is returned
        """
        responseDict = ftpControlc.response_info(self._handle)
        return responseDict

class AuthInfo:
    """Class that wraps auth_info_t struct"""
    def __init__(self, credential, encrypt, user=None, password=None, account=None, subject=None, handle=None):
        """
        Initializes an AuthInfo, which contains a user's authentication
        information, not all of the arguments you can use your credentials
        or user name and password for authentication.

        **Arguments**

        - *credential* A GSSCred object of the user's credentials

        - *encrypt* A boolean indicating whether encryption should
        be used

        - *user* A string indicating the user or None

        - *password* A string of the password, or None
        Note:
        There should be a better way to do this, because this is
        not secure

        - *account* A string of the user account or None

        - *subject* A string of the subject DN, it must be the DN of the end
        entity that you want to authenticate yourself to. It can also be None
        """
        
        
        if handle ==None:
            ret, handle = ftpControlc.auth_info_init(credential.get_handle(),
                                                     encrypt, user, password, 
                                                     account, subject)
        if not ret == 0:
            ex = AuthInfoException(handle)
            raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Frees the memory allocated in the constructor
        """
        
        ftpControlc.auth_info_destroy(self._handle)
        
    def get_handle(self):
        return self._handle

    def compare(self, authInfo):
        """
        **Arguments**

        - *authInfo*
        an AuthInfo object to compare
        
        **Returns**
        a boolean indictating whether the AuthInfo objects are equivalent
        """

        ret = ftpControlc.auth_info_compare(self._handle, authInfo.get_handle())
        if ret == 0:
            return True
        else:
            return False
        
class Parallelism:

    def __init__(self, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = ParallelismException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.parallelism_init()
        self._handle = handle
        return

    def __del__(self):
        ftpControlc.parallelism_destroy(self._handle)
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = ParallelismException("Unable to deactivate the module")
            raise ex
        return

    def set_size(self, size):
        ret = ftpControlc.parallelism_set_size(self._handle, size)
        if not ret == 0:
            ex = ParallelismException("unable to set the size")
            raise ex
        return

    def set_mode(self, mode):
        ret = ftpControlc.parallelism_set_mode(self._handle, mode)
        if ret == -42:
            ex = ParallelismException("unknown mode")
            raise ex
        if not ret == 0:
            ex = ParallelismException("unable to set the mode")
            raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle

class FtpControlServer:

    def __init__(self, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = FtpControlException("Unable to initialize the module")
            raise ex
        self._cbQueue = CallbackQueue()
        self._activated = 1
        if handle == None:
            ret, handle = ftpControlc.server_handle_init()
        if not ret == 0:
            ex = FtpControlException("unable to create the handle")
            raise ex
        self._handle = handle
        return
    
    def __del__(self):
        if not self._handle == None:
            ret, errString = ftpControlc.server_handle_destroy(self._handle)
        if not ret == 0:
            ex = FtpControlException(errString)
            raise ex
        
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = FtpControlException("Unable to deactivate the module")
            raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle

    def accept(self, handle, callback, arg):
        ret, cbHandle = ftpControlc.server_accept(self._handle, 
                                                  handle.get_handle(), 
                                                  callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle, CallbackDefault(cbHandle))
        return cbHandle



    def authenticate(self, handle, auth_req, callback, arg):
        """
        Authenticate a client connection.
        
        This function is called to authenticate a connection from
        a client.
        
        After a client connection has been accepted (using the
        globus_ftp_control_server_accept call), this function should be called
        to authenticate the client. The caller of this function may specify
        certain authentication requirements using the auth_requirements parameter.

        **Arguments**

        - *handle*  a FtpControl object that has already been used in the accept call

        - *auth_req*  an int specifying the authentication required

        - *callback*  The function to be called when the authentication
        has been accepted or rejected. It has the following signature:
        
        'func(arg,handle,error,authInfo)', where arg is the argument passed
        into this call, handle is a SWIG'ized pointer to the ftp control
        object, error is a tuple (error_code, error_description). If there is
        no error, the error_code is 0 and the error description is a None
        object, and authInfo is a SWIG'ized pointer to a auth info object.

        - *arg* A user argument to be returned when the callback is executed.

        **Returns**
         A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises**
        A FtpControlException is raised if there is a problem authenticating to the
        server
        """
        ret, cbHandle = ftpControlc.server_authenticate(handle.get_handle(), 
                                                        auth_req, callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle, CallbackDefault(cbHandle))
        return cbHandle

	################################################################
	# This is a strange function -- it used to take a pointer to
	# to the port, and modify it to the port that was assigned to
	# the listener. I now just return the assigened port back to the
	# calling funtion 
	#
	################################################################

    def listen(self, port, callback, arg):
        ret, cbHandle, real_port = ftpControlc.server_listen(self._handle, 
                                                             port, 
                                                             callback, 
                                                             arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle, CallbackDefault(cbHandle))
        return cbHandle, port

    def stop(self, callback, arg):
        ret, cbHandle = ftpControlc.server_stop( self._handle, callback, arg)
        if not ret == 0:
            ex = FtpControlException(cbHandle)
            raise ex
        self._cbQueue.append_callback(cbHandle, CallbackDefault(cbHandle))
        return cbHandle

    def free_callback(self, callbackHandle):
        """
	Frees the memory associated with the callback handle

	**Arguments**
	- *callbackHandle* - a SWIG'ized pointer to a callback handle

	**Returns**
	None

	**Raises**
	None
	"""

        self._cbQueue.remove_callback(callbackHandle)
	
class TcpBuffer:

    def __init__(self, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(ftpControlc.get_module())
        if not ret == 0:
            ex = TcpBufferException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            handle = ftpControlc.tcpbuffer_init()
        self._handle = handle
        return

    def __del__(self):
        ftpControlc.tcpbuffer_destroy(self._handle)
        ret = utilc.globus_module_deactivate(ftpControlc.get_module())
        if not ret == 0:
            ex = TcpBufferException("Unable to deactivate the module")
            raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When the
    # object is destroyed, the memory this handle points to is freed.
    ####################################################################### def
    def get_handle(self): 
        return self._handle
    
    def set_automatic(self, initial, minimum, maximum):
        ftpControlc.tcpbuffer_set_automatic(self._handle, initial, 
                                            minimum, maximum)

    def set_fixed(self, size):
        ftpControlc.tcpbuffer_set_fixed(self._handle, size)
        

    def set_default(self):
        ftpControlc.tcpbuffer_set_default(self._handle)
        


