############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides an interface to the gass copy module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
import gassCopyc

###########################################################################
# Module Constants
###########################################################################
# URL modes
UNSUPPORTED_MODE = 0
FTP_MODE         = 1
GASS_MODE        = 2
IO_MODE          = 3

# Copy status
STATUS_FAILURE              = -1
STATUS_NONE                 = 0
STATUS_PENDING              = 1
STATUS_INITIAL              = 2
STATUS_SOURCE_READY         = 3
STATUS_TRANSFER_IN_PROGRESS = 4
STATUS_READ_COMPLETE        = 5
STATUS_WRITE_COMPLETE       = 6
STATUS_DONE                 = 7
STATUS_CANCEL               = 8

########################################################################
# Module classes
########################################################################

class GassCopyException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the gass copy module. It defines no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class HandleAttrException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the gass copy handle attr module. It defines
    no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class AttrException(GlobusException):
    """
    A simple subclass of GlobusException.
    
    Used for errors in the gass copy attr module. It defines no new methods.
    
    **Extends:**
     
         *pyGlobus.util.GlobusException*
    """
    pass

class GassCopy:
    """
    Provides a simple interface to protocol independent file transfers.

    Wraps the globus gass copy module. Supports http, https, ftp, and
    gsiftp based transfers.
    """
    
    def __init__(self, handleAttr, handle = None):
        """
        Constructs an instance.
        
        Activates the gass copy module, and then creates the underlying
        gass copy handle.
        
        **Arguments:**
        
        - *handleAttr* A HandleAttr object.
        
        - *handle* An optional argument that if present will be used as
        the gass copy handle.  The class then owns the handle, and will
        delete it at destruction.
        
        **Raises:**
        
        A GassCopyException is thrown if unable to initialize the module
        or create the handle.
        """

        self._activated = 0
        self._cbList = []
        ret = utilc.globus_module_activate(gassCopyc.get_module())
        if not ret == 0:
            ex = GassCopyException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gassCopyc.handle_init(handleAttr.get_handle())
            if not ret == 0:
                ex = GassCopyException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance.

        Free any outstanding callback handles, destroy the underlying gass
        copy handle, and deactivate the module.

        **Raises:**
        
        A GassCopyException is thrown if unable to destroy the handle
        or deactivate the module.
        """

        for cb in self._cbList:
            self.free_callback(cb)
        ret, err = gassCopyc.handle_destroy(self._handle)
        if not ret == 0:
            ex = GassCopyException(err)
            raise ex
        ret = utilc.globus_module_deactivate(gassCopyc.get_module())
        if not ret == 0:
            ex = GassCopyException("Unable to deactivate the module")
            raise ex
        return
    
    def copy_url_to_url(self, srcUrl, srcAttr, destUrl, destAttr):
        """
        Transfer data from source url to destination url.
        
        The url's may be http, https, ftp, and gsiftp. This method
        blocks until the transfer is complete.
        
        **Arguments:**
        
        - *srcUrl* The URL to transfer the data from.
        
        - *srcAttr* A Attr object with the attributes
        for the source.
        
        - *destUrl* The URL to transfer data to.
        
        - *destAttr* A Attr object with the attributes
        for the destination.
        
        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        attempting to copy the data.
        """

        ret, error = gassCopyc.copy_url_to_url(self._handle, srcUrl,
                                              srcAttr.get_handle(),
                                              destUrl, destAttr.get_handle())
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return

    def copy_url_to_handle(self, srcUrl, srcAttr, destHandle):
        """
        Transfer data from a source URL to a io handle.
        
        Currently the only types of handles implemented are file io handles.
        This method blocks until the transfer is complete.
        
        **Arguments:**
        
        - *srcUrl* The URL to transfer data from.
        
        - *srcAttr* A Attr object with the attributes for the
        source.
        
        - *destHandle* A globus_io_handle_t SWIG pointer.  Currently
        these are returned from the io file methods.
        
        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        attempting to copy the data.
        """

        ret, error = gassCopyc.copy_url_to_handle(self._handle, srcUrl,
                                                 srcAttr.get_handle(),
                                                 destHandle)
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return

    def copy_handle_to_url(self, srcHandle, destUrl, destAttr):
        """
        Transfer data from a io handle to a URL.
        
        Currently the only types of handles implemented are file io handles.
        This method blocks until the transfer is complete.
        
        **Arguments:**
        
        - *srcHandle* A globus_io_handle_t SWIG pointer.  Currently
        these are returned from the io file methods.  The data will
        be read from here.
        
        - *destUrl* The URL to transfer the data to.
        
        - *destAttr* A Attr object with the attributes for the
        destination.
        
        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        attempting to copy the data.
        """

        ret, error = gassCopyc.copy_handle_to_url(self._handle, srcHandle,
                                                 destUrl, destAttr.get_handle())
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return

    def register_copy_url_to_url(self, srcUrl, srcAttr, destUrl,
                                 destAttr, callback, arg):
        """
        Transfer data from source url to destination url.
        
        Asynchronously transfer data between two url's.
        
        **Arguments:**
        
        - *srcUrl* The URL to transfer the data from.
        
        - *srcAttr* A Attr object with the attributes
        for the source.
        
        - *destUrl* The URL to transfer data to.
        
        - *destAttr* A Attr object with the attributes
        for the destination.
        
        - *callback* The function to call when the transfer is
        complete.  It has the following form.
        'function(arg, handle, error)' where arg is the user
        argument passed into this call, handle is the gass
        copy handle, and error is a tuple (error_code, error_description).
        If there is no error, the error_code is 0 and the error
        description is a None object.
        
        - *arg* A user argument to be passed to the callback
        function.
        
        **Raises:**
        
        A GassCopyException is thrown if an error occurs
        while initiating the transfer.
        """
        ret, cbHandle = gassCopyc.register_copy_url_to_url(self._handle, srcUrl,
                                                          srcAttr.get_handle(),
                                                          destUrl,
                                                          destAttr.get_handle(),
                                                          callback, arg)
        if not ret == 0:
            ex = GassCopyException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return

    def register_copy_handle_to_url(self, srcHandle, destUrl, destAttr,
                                    callback, arg):
        """
        Transfer data from a io handle to a URL.

        Asynchronously transfers the data. Currently the only types of handles
        implemented are file io handles.

        **Arguments:**
        
        - *srcHandle* A globus_io_handle_t SWIG pointer.  Currently
        these are returned from the io file methods.  The data will
        be read from here.
        
        - *destUrl* The URL to transfer the data to.
        
        - *destAttr* A Attr object with the attributes for the
        destination.
        
        - *callback* The function to call when the transfer is
        complete.  It has the following form.
        'function(arg, handle, error)' where arg is the user
        argument passed into this call, handle is the gass
        copy handle, and error is a tuple (error_code, error_description).
        If there is no error, the error_code is 0 and the error
        description is a None object.
        
        - *arg* A user argument to be passed to the callback
        function.
        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        initiating the copy.
        """

        ret, cbHandle = gassCopyc.register_copy_handle_to_url(self._handle,
                                                             srcHandle, destUrl,
                                                             destAttr.get_handle(),
                                                             callback, arg)
        if not ret == 0:
            ex = GassCopyException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return
   
    def register_copy_url_to_handle(self, srcUrl, srcAttr, destHandle,
                                    callback, arg):
        """
        Transfer data from a source URL to a io handle.

        Currently the only types of handles implemented are file io handles.
        This method blocks until the transfer is complete.

        **Arguments:**
        
        - *srcUrl* The URL to transfer data from.
        
        - *srcAttr* A Attr object with the attributes for the
        source.
        
        - *destHandle* A globus_io_handle_t SWIG pointer.  Currently
        these are returned from the io file methods.

        - *callback* The function to call when the transfer is
        complete.  It has the following form.
        'function(arg, handle, error)' where arg is the user
        argument passed into this call, handle is the gass
        copy handle, and error is a tuple (error_code, error_description).
        If there is no error, the error_code is 0 and the error
        description is a None object.
        
        - *arg* A user argument to be passed to the callback
        function.
        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        initiating the copy.
        """

        ret, cbHandle = gassCopyc.register_copy_url_to_handle(self._handle, srcUrl,
                                                             srcAttr.get_handle(),
                                                             destHandle, callback, arg)
        if not ret == 0:
            ex = GassCopyException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return

    def get_status(self):
        """
        Return the current status.

        Returns a status code describing the state of the current transfer.

        **Raises:**
        
        A GassCopyException is thrown if unable to return the status.
        
        **Returns:**
        
        A status code. The possible codes are defined above in
        the module constants
        """

        ret, error, status = gassCopyc.get_status(self._handle)
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return status

    def flush_url_state(self, url):
        """
        Remove a cached connection.

        If an idle connection to an FTP server exists, it will be closed.

        **Arguments:**
        
        - *url* The URL to be flushed from the cache.
        
        **Raises:**
        
        A GassCopyException is thrown if an error occurs
        while flushing the url.
        """

        ret, error = gassCopyc.flush_url_state(self._handle, url)
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return

    def cache_url_state(self, url):
        """
        Cache a connection to a server.

        When an URL is cached, the connection to the server will not be closed
        after a file transfer completes.

        **Arguments:**

	- *url* The URL to be cached.

        **Raises:**
        
        A GassCopyException is thrown if an error occurs while
        caching the url.
        """

        ret, error = gassCopyc.cache_url_state(self._handle, url)
        if not ret == 0:
            ex = GassCopyException(error)
            raise ex
        return

    def free_callback(self, cbHandle):
        """
        Free the callbacks resources.

        Free's the memory allocated to hold the callback structure.

        **Arguments:**
        
        - *cbHandle* The SWIG'ized pointer to a cbStruct.
        """

        gassCopyc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return

    def get_handle(self):
        """
        Return the underlying gass copy handle.

        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.

        **Returns:**

            A string containing the pointer.
        """
        return self._handle


def get_url_mode(url):
    """
    Returns the mode of a url.

    Given a URL returns the mode as one of the module constants
    defined above.

    **Arguments:**
    - *url* The url to return the mode of.
    
    **Raises:**
    A GassCopyException is thrown if an error occurs while
    returning the mode.
    
    **Returns:**    
    The mode.  The possible modes are defined above in the
    module constants. 
    """
    
    ret, error, mode = gassCopyc.get_url_mode(url)
    if not ret == 0:
        ex = GassCopyException(error)
        raise ex
    return mode

class Attr:
    """
    A class to wrap the gass copy attr functions.

    It provides get and set methods for all of the various attributes.
    """
    
    def __init__(self, handle = None):
        """
        Constructs an instance.
        
        Activates the gass copy module, and then creates the underlying
        gass copy attr handle.

        **Arguments:**
	- *handle* An optional argument that if present will be used as
	the gass copy attr handle.  The class then owns the handle,
	and will delete it at destruction.
        
        **Raises:**
        
	A AttrException is thrown if unable to initialize the module
	or create the handle.
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gassCopyc.get_module())
        if not ret == 0:
            ex = AttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gassCopyc.attr_init()
            if not ret == 0:
                ex = AttrException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance.

        It destroys the handle, and then deactivates the module.

        **Raises:**
        
        A AttrException is thrown if unable to
        deactivate the module.
        """

        gassCopyc.attr_destroy(self._handle)
        ret = utilc.globus_module_deactivate(gassCopyc.get_module())
        if not ret == 0:
            ex = AttrException("Unable to deactivate the module")
            raise ex

    def get_handle(self):
        """
        Return the underlying gass copy attr handle.

        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        
        **Returns:**
        
        A string containing the pointer.
        """

        return self._handle

    def set_ftp(self, ftpAttr):
        """
        Sets the ftp specific attributes.
        
        **Arguments:**
        
        - *ftpAttr* A FtpClientOperationAttr object with the proper attributes
        set.
        
        **Raises:**
        
        A AttrException is thrown if unable to set the ftp attr.
        """

        ret, error = gassCopyc.attr_set_ftp(self._handle, ftpAttr.get_handle())
        if not ret == 0:
            ex = AttrException(error)
            raise ex
        return

    def set_io(self, ioAttr):
        """
        Sets the file specific attributes.

        Sets the attributes for a file transfer.

        **Arguments:**
        
        - *ioAttr* A FileIOAttr with the proper attributes set.
        
        **Raises:**
        
        A AttrException is thrown if unable to set the io attr.
        """

        ret, error = gassCopyc.attr_set_io(self._handle, ioAttr.get_handle())
        if not ret == 0:
            ex = AttrException(error)
            raise ex
        return

    def set_gass_transfer(self, transferReqAttr):
        """
        Sets the http/https specific attributes.
        
        **Arguments:**
        
        - *transferReqAttr* A TransferReqAttr object with the proper
        attributes set.
        
        **Raises:**

        A AttrException is thrown if unable to set the gass
        transfer attr.
        """

        ret, error = gassCopyc.attr_set_gass(self._handle, 
                                             transferReqAttr.get_handle())
        if not ret == 0:
            ex = AttrException(error)
            raise ex
        return
    
class HandleAttr:
    """
    A class to wrap the gass copy handle attr functions.

    It provides get and set methods for all of the various handle
    specific attributes.
    """
    
    def __init__(self, handle = None):
        """
        Constructs an instance.
        
        Activates the gass copy module, and then creates the underlying
        gass copy handle attr handle.

        **Arguments:**
        
        - *handle* An optional argument that if present will be used as
        the gass copy handle attr handle.  The class then owns the handle,
        and will delete it at destruction.
        
        **Raises:**

        A HandleAttrException is thrown if unable to initialize
        the module or create the handle.
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gassCopyc.get_module())
        if not ret == 0:
            ex = HandleAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gassCopyc.handleattr_init()
            if not ret == 0:
                ex = HandleAttrException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        Destroy an instance.

        It destroys the handle, and then deactivates the module.

        **Raises:**
        
        A HandleAttrException is thrown if unable to
        deactivate the module.
        """

        gassCopyc.handleattr_destroy(self._handle)
        ret = utilc.globus_module_deactivate(gassCopyc.get_module())
        if not ret == 0:
            ex = HandleAttrException("Unable to deactivate the module")
            raise ex

    def get_handle(self):
        """
        Return the underlying gass copy handle attr handle.

        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.

        **Returns:**
        
        A string containing the pointer.
        """

        return self._handle

    def set_ftp(self, ftpAttr):
        """
        Sets the ftp specific attributes.

        **Arguments:**
        
        - *ftpAttr* A FtpClientHandleAttr object with the proper attributes
        set.
        
        **Raises:**
        
        A HandleAttrException is thrown if unable to set the ftp attr.
        """

        ret, error = gassCopyc.handleattr_set_ftp(self._handle, 
                                                  ftpAttr.get_handle())
        if not ret == 0:
            ex = HandleAttrException(error)
            raise ex
        return

