############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
A module wrapping the globus gass server ez module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
import gassServerEZc

class GassServerEZException(GlobusException):
    """
    The module exception class.

    A sub-class of GlobusException used to indicate errors in the module.

    **Extends:**

        *pyGlobus.util.GlobusException*
    """
    pass

##############################################################################
# A set of module constants used for options to the server_init call.  The
# mirror the options in globus_gass_server_ez.h.  They can be bitwise or'd
# together to produce the options for server_init.
################################################################################
LINE_BUFFER            = 1
TILDE_EXPAND           = 2
TILDE_USER_EXPAND      = 4
READ_ENABLE            = 8
WRITE_ENABLE           = 16
STDOUT_ENABLE          = 32
STDERR_ENABLE          = 64
CLIENT_SHUTDOWN_ENABLE = 128

class GassServerEZ:
    """
    Proxy class to the GassServerEZ interfacee
    """
    def __init__(self, opts, listenAttr=None, requestAttr=None):
        """
        Constructs an instance.

        Activates the underlying module and constructs a server with the
        given options.

        **Arguments:**

        - *opts* The bitwise or of the module options defined above.

        - *listenAttr* A gassTransfer.ListenerAttr object.

        - *requestAttr* A gassTransfer.TransferReqAttr object.
            
        **Raises:**

        A GassServerEZExceptions is thrown if unable to initialize the
        module or start the server.
        """
        self.activated = 0
        self.isListening = 0
        self._lAttr = listenAttr
        self._rAttr = requestAttr
        if listenAttr == None:
            self._lAttrHandle = None
        else:
            self._lAttrHandle = listenAttr.get_handle()
        if requestAttr == None:
            self._rAttrHandle = None
        else:
            self._rAttrHandle = requestAttr.get_handle()
        ret = utilc.globus_module_activate(gassServerEZc.get_module())
        if not ret == 0:
            ex = GassServerEZException("Unable to initialize the gass server ez module")
            raise ex
        self.activated = 1
        ret, listener, url = gassServerEZc.pyserver_init(opts,
                                                         self._lAttrHandle,
                                                         self._rAttrHandle)
        if not ret == 0:
            ex = GassServerEZException("Unable to initialize the server")
            raise ex
        self.listener = listener
        self.url = url
        self.isListening = 1
    
    def __del__(self):
        """
        Destroys an instance.

        Shuts down the server and deactivates the underlying module.

        **Raises:**

        A GassServerEZException is thrown if unable to stop the server
        or deactivate the module.
        """
        if self.isListening:
            ret = gassServerEZc.pyserver_shutdown(self.listener)
            if not ret == 0:
                ex = GassServerEZException("Unable to shutdown the server")
                raise ex
        if self.activated:
            ret = utilc.globus_module_deactivate(gassServerEZc.get_module())
            if not ret == 0:
                ex = GassServerEZException("Unable to deactivate the gass server module")
                raise ex

    def getURL(self):
        """
        Returns the url the server is listening on.

        **Returns:**

        A string containing the url the server is listening on.
        """
        return self.url
    
    def server_shutdown(self):
        """
        Shutdown the server.

        This may be called to stop the server from listen and free the underlying
        socket. If not the destructor will free the socket.

        **Raises:**

        A GassServerEZException is thrown if unable to shutdown the server.
        """
        if self.isListening:
            ret = gassServerEZc.pyserver_shutdown(self.listener)
            if not ret == 0:
                ex = GassServerEZException("Unable to shutdown the server")
                raise ex
            self.isListening = 0
        
