############################################################################
#Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides an interface to the Gass Transfer module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
import gassTransferc

class TransferException(GlobusException):
    pass

class TransferServerException(GlobusException):
    pass

class TransferReqAttrException(GlobusException):
    pass

class TransferListenerAttrException(GlobusException):
    pass

class TransferListenerException(GlobusException):
    pass

class GassTransferServerException(GlobusException):
    pass

class TransferRequest:
    """
    Proxy class that wraps the request handle for clients and server
    handle is the request handle pointer
    isClient is true if it is a client-side handle
    """
    
    def __init__(self, handle, isClient):
        self._handle = handle
        self._isClient = isClient
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferException("Unable to initialize the module")
            raise ex
        self._activated = 1 
        return

    def __del__(self):
        """
        frees the memory associated with the request handle
        """
        
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferException("Unable to deactivate the module")
            raise ex
        ret = gassTransferc.request_destroy(self._handle)
        if not ret == 0:
            ex = TransferException("Can't destroy the request_handle")
            raise ex
        return
   
    def get_handle(self):
        return self._handle


    def get_request_type(self):
        """
        Returns a string the request type of the Request handle
        """
        
        reqType = gassTransferc.get_request_type(self._handle)
        if reqType == 0:
            return "INVALID"
        elif reqType == 1:
            return "GET"
        elif reqType == 2:
            return "PUT"
        elif reqType == 3:
            return "APPEND"
        else:
            raise RuntimeError, "Invalid request for TransferData object, aborting!"
    
    
    def get_url(self):
        """
        Returns the url associated with the request
        """

        ret = gassTransferc.get_url(self._handle)
        return ret

    def get_subject(self):
        """
        Returns the subject associated with the request.
        The subject is a string commonly a DN.
        """

        ret = gassTransferc.get_subject(self._handle)
        return ret

    def set_subject(self, subject):
        """
        Sets the subject that will be associated with the
        request.

        **Arguments**
        - *subject* a string, commonly the DN of the end entity.

        **Raises**
        A TransferException is raised if the request is in a
        state where the subject should not be set.
        """
        ret = gassTransferc.set_subject(self._handle, subject)
        if not ret == 0:
            raise TransferException("The subject could not be set on the request.")
        
    
    def get_status(self):
        """
        Return the status of the request handle
        """

        ret = gassTransferc.get_request_status(self._handle)
        if ret == 0:
            return "INVALID"
        elif ret == 1:
            return "STARTING"
        elif ret == 2:
            return "PENDING"
        elif ret == 3:
            return "FAILED"
        elif ret == 4:
            return "REFERED"
        elif ret == 5:
            return "DENIED"
        elif ret == 6:
            return "DONE"
        else:
            raise RuntimeError, "Illegal status for the TransferRequest object" 
        
    
    def get_length(self):
        """
        Returns the amount of data that will be transfered
        to copy the URL. If the length is 0, then the
        length is unknown. 
        """

        ret = gassTransferc.get_length(self._handle)
        return ret
    
    def set_length(self, length=0):
        """
        Sets the amount of data that will be transfered
        to copy the URL. If the length is 0, then the
        length is unknown. 
        """
        gassTransferc.set_length(self._handle, length)

    def get_denial_reason(self):
        """
        Returns the protocol specific denial reason which
        explains why the request was denied. It returns an int.
        """
        ret = gassTransferc.get_denial_reason(self._handle)
        return ret
    
    def get_denial_message(self):
        """
        Returns the protocol specific denial reason which
        explains why the request was denied. It returns a string.
        """
        ret = gassTransferc.get_denial_message(self._handle)
        return ret
    
class TransferListener:
    """
    Proxy class that wraps the listener handle for servers
    handle is the listerner handle pointer
    The TransServer class provides an higher level interface than this.
    """
    
    def __init__(self, attr=None, scheme= "http"):
        """
        **Arguments**
        
        - *attr* is a ListenerAttr object
        
        - *scheme* is a string of the protocol that will be used, ie http
        
        """
        
        if attr == None:
            attr = ListenerAttr(scheme)
            
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        
        
        if not ret == 0:
            ex = TransferReqAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
    
        ret, listener = gassTransferc.create_listener(attr.get_handle(), scheme)
        if not ret == 0:
            ex = TransferServerException(listener + " error code is: " + str(ret))
            raise ex

        self._handle = listener
   
     
     
    def __del__(self):
        """
        frees the request handles memory
        
        *Raises*
        A TransferException is raised if there is a problem freeing the memory
        """

        gassTransferc.listener_destroy(self._handle)
        utilc.globus_module_deactivate(gassTransferc.get_module())
        

    def close_listener(self, callback, arg):
        """
	Callbacks for any outstanding register_accepts are called before the
        listener is closed

        **Arguments**

	- *callback* function that is called when the listener is closed
	
	- *arg* user supplied argument to the callback function
	
	**Throws**
	A TransferListenerException if the listener cannot be closed
	"""
        
        ret, cbHandle, listener = gassTransferc.close_listener(self._handle, 
                                                               callback, arg)
        if not ret == 0:
            ex = TransferListenerException(cbHandle + " error code is: " + str(ret))
            raise ex

        return cbHandle
     
    def get_handle(self):
        return self._handle

    def get_base_url(self):
        """
	Gets the base url that the listener is listening on
	Returns a string of that url
        """

        url = gassTransferc.get_base_url(self._handle)
        return url


class TransferReqAttr:
    """
    A proxy class for gass transfer request attributes
    """
    def __init__(self, urlScheme, handle = None):
        """
        **Arguments**

        - *urlScheme* is a string of the protocol that will be used, ie http

        - *handle* a SWIG'ized pointer to a globus_gass_transfer_requestattr_t
        object
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        
        if not ret == 0:
            ex = TransferReqAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        
        if handle == None:
            ret, handle = gassTransferc.reqattr_init(urlScheme)
        if not ret == 0:
            ex = TransferReqAttrException("unable to init the attr")
            raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        frees the memory associated with the underlying request attribute
        """
        
        ret = gassTransferc.reqattr_destroy(self._handle)
        if not ret == 0:
            ex = TransferReqAttrException("unable to destroy the attr")
            raise ex        
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferReqAttrException("Unable to deactivate the module")
            raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle


class ListenerAttr:
    """
    A proxy class for gass transfer listener attributes
    """ 
    def __init__(self, urlScheme = "http", handle = None):
        """
        **Arguments**

        - *urlScheme* the protocol to use with this attribute, ie http

        - *handle* a SWIG'ized pointer to a globus_gass_transfer_requestattr_t
        object
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferListenerAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gassTransferc.listenerattr_init(urlScheme)
        if not ret == 0:
            ex = TransferListenerAttrException("unable to init the listenattr")
            raise ex
        self._handle = handle
        return

    def __del__(self):
        """
        frees the memory associated with the underlying listener attribute
        """
        
        ret = gassTransferc.listenerattr_destroy(self._handle)
        if not ret == 0:
            ex = TransferListenerAttrException("Unable to destroy listenerAttr")
            raise ex
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferReqAttrException("Unable to deactivate the module")
            raise ex
        return
        
    def get_handle(self):
        return self._handle

class TransferData:
    """
    This class handles send_bytes and receive_bytes calls which both a
    TransferClient and a TransferServer object can make. This class is
    the parent class for those two classes.
    """

    def __init__(self):
        
        self._cbList = []
        return

    def __del__(self):
        """
        frees the memory associated with the callaback handles
        """
        
        for cb in self._cbList:
            self.free_callback(cb)
            
        return

    def free_callback(self, cbHandle):
        """
        Free a callback.
        
        Free's the underlying memory allocated for a callback structure.
        
        **Arguments:**
        
        - *cbHandle* A SWIG'ized pointer to a callback structure.
        """

        gassTransferc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return

    def send_bytes(self, request, buffer, eof, callback, arg):
        """
        After authorizing a request the server calls send bytes which sends
        the data in the buffer object
        
        **Arguments**
        
        - *request* a server-side request if this is in conjunction with a
        get request, if in conjunction with a put or append this should
        be a client-side request
        
        - *buffer* A pyGlobus.util.Buffer.Buffer object.. This hides the
        internal globus_byte_t data buffer.
        
        - *eof* true if this is the last data to be sent, false otherwise
        
        - *callback* is a function that is called after the data in the buffer
        has been sent
        
        - *arg* a user supplied argument to the callback function  
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        a TransferException if the request object does not meet the specifications
        above or if there was a problem sending the data
        """
        
        ok = 0
        reqType = request.get_request_type()
        if reqType ==  "GET" and request._isClient == 0:
            ok = 1
        elif reqType == "PUT" and request._isClient == 1:
            ok = 1
        elif reqType == "APPEND" and request._isClient == 1:
            ok = 1

        if not ok:
            raise TransferException,  "improper use of send_bytes, the request argument is not valid"

        ret, cbHandle = gassTransferc.send_bytes(request.get_handle(), buffer,
                                                  len(buffer),eof,
                                                  callback, arg)
        
        if not ret == 0:
            ex = TransferException(cbHandle + " error code is: " + str(ret))
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def receive_bytes(self, request, buffer, wait_for_amount, callback, arg):
        """
        **Arguments:**

        - *request* is a TransferRequest object it must be client side if it
        is used in conjunction with a get request must be server side if used
        in conjunction with a put or append request
        
        - *buffer* A pyGlobus.util.Buffer.Buffer object.. This hides the
        internal globus_byte_t data buffer.
        
        - *wait_for_amount* the minimum amount of data to wait to receive
        before the callback function is called,the data is received in
        the callback function

        - *callback* is a function that is called after enough data has been received
        
        - *arg* a user supplied argument to the callback function
        
        **Returns:**
        
        A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed.
        
        **Raises:**
        A TransferException if the request object does not meet the specifications
        above or if there is problem receiving the bytes
        """
        
        ok = 0
        reqType = request.get_request_type()
        if reqType ==  "GET" and request._isClient == 1:
            ok = 1
        elif reqType == "PUT" and request._isClient == 0:
            ok = 1
        elif reqType == "APPEND" and request._isClient == 0:
            ok = 1
            
        if not ok:
            errString = "improper use of receive_bytes, the request argument is not valid"    
            ex = TransferException(errString)
            raise ex
        
        
        ret, cbHandle = gassTransferc.receive_bytes(request.get_handle(),
                                                   buffer.get_handle(),
                                                   buffer.get_size(),
                                                   wait_for_amount,
                                                   callback, arg)

        if not ret == 0:
            ex = TransferException(cbHandle + " error code is: "+ str(ret))
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

class TransferClient(TransferData):

    def __init__(self):

        TransferData.__init__(self)
        return

    def __del__(self):
        for cb in self._cbList:
            self.free_callback(cb)
            
        return
    

    def register_get(self, attr, url, callback, arg):
        """
        **Arguments:**

        - *attr* is a RequestAttr object
        
        - *url* is the url that is the file that you want to get
        An example of a url is http://huey.lbl.gov:3730/etc/group
        
        - *callback* is a function that will be called back after the request to
        get has been authorized, denied or refered
        
        - *arg* is a user argument to the callback function
        
        This function is non-blocking
        
        **Returns**
        
        request is a SWIG'ized pointer to the underlying request handle.
        The request handle is a client side request handle
        
        **Raises**
        A TransferException is raised if the register_get cannot be made
        """
        
        ret, cbHandle, request = gassTransferc.register_get(attr.get_handle(), 
                                                            url, callback, arg)
        
        req = TransferRequest(request, 1)
        
        if not ret == 0:
            ex = TransferException(cbHandle + " error code is: " + str(ret))
            raise ex
        self._cbList.append(cbHandle)
        return req
        
    
    def get(self, attr, url):
        """
        **Arguments**
        
        - *attr* is a RequestAttr object
        
        - *url* is the url that is the file that you want to get
        An example of a url could be http://huey.lbl.gov:3730/etc/group
        
        This function is blocking
        
        **Returns**
        
        request is a SWIG'ized pointer to the underlying request handle.
        The request handle is a client side request handle
        
        **Raises**
        A TransferException is raised if the get cannot be made
        """

        ret, request = gassTransferc.get(attr.get_handle(), url)
        
        if not ret == 0:
            ex = TransferException(request + " error code is: " + str(ret))
            raise ex

        req = TransferRequest(request, 1)
        return req
        
    def register_put(self, attr, url, callback, arg, length=0):
        """
        **Arguments**
        - *attr* is a RequestAttr object
        
        - *url* is the url that is the file that you want to put
        An example of a url could be http://huey.lbl.gov:3730/etc/group
        
        - *length* is the length of the file that you wish to put
        The default 0, will do the put with the length argument
        GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
        
        - *callback* is a function that will be called back after the request
        to get has been authorized, denied or refered
        
        - *arg* is a user argument to the callback function
        
        This function is non-blocking
        
        **Returns**
        A TransferRequest object is returned
        
        **Raises**
        A TransferException is raised if the register_put cannot be made
        """
        
        ret, cbHandle, request = gassTransferc.register_put(attr.get_handle(),
                                                            length, url, callback, arg)
        req = TransferRequest(request, 1)
        if not ret == 0:
            ex = TransferException(cbHandle + " error code is: " + str(ret))
            raise ex
        self._cbList.append(cbHandle)
        return req

    def put(self, attr, url, length=0):
        """
        **Arguments**
        - *attr* is a RequestAttr object
        
        - *url* is the url that is the file that you want to put
        An example of a url could be http://huey.lbl.gov:3730/etc/group
        
        - *length* is the length of the file that you wish to put
        The default 0, will do the put with the length argument
        GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
        
        This function is blocking

        **Returns**
        A TransferRequest object is returned
        
        **Raises**
        A TransferException is raised if the put cannot be made
        """
            
        ret, request = gassTransferc.put(attr.get_handle(), url, length)
        req = TransferRequest(request, 1)
        if not ret == 0:
            ex = TransferException(request + " error code is: " + str(ret))
            raise ex
        return req
     
    def register_append(self, attr, url, callback, arg, length=0):
        """
        **Arguments**
        - *attr*  is a RequestAttr object
        
        - *url* is the url that is the file that you want to append to
        An example of a url could be http://huey.lbl.gov:3730/etc/group
        
        - *callback* is a function that will be called back after the request to
        get has been authorized, denied or refered
        
        - *arg* is a user argument to the callback function

        - *length* is the length of the file that you wish to append
        The default 0, will do the put with the length argument
        GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
        
        This function is non-blocking
        
        **Returns**
        A TransferRequest object is returned
       
        **Raises**
        A TransferException is raised if the register_append cannot be made
        """

        ret, cbHandle, request = gassTransferc.register_append(attr.get_handle(),
                                                               length, url, callback, arg)
        req = TransferRequest(request, 1)
        if not ret == 0:
            ex = TransferException(cbHandle + " error code is: " + str(ret))
            raise ex
        self._cbList.append(cbHandle)
        return req

    def append(self, attr, url, length=0):
        """
        **Arguments**

        - *attr*  is a RequestAttr object
        
        - *length* is the length of the file that you wish to append
        The default 0, will do the put with the length argument
        GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
        
        - *url* is the url that is the file that you want to append to
        An example of a url could be http://huey.lbl.gov:3730/etc/group
        
        This function is blocking
        
        **Returns**
        A TransferRequest object is returned
        
        **Raises**
        A TransferException is raised if the register_append can not be made
        """

        ret, request = gassTransferc.append(attr.get_handle(), url, length)
        req = TransferRequest(request, 1)
        if not ret == 0:
            ex = TransferException(request + " error code is: " + str(ret))
            raise ex
        return req

    
class TransferServer(TransferData):
    """
    A class to accept and handle requests
    """
    
    def __init__(self, scheme):
        """
        **Arguments**
        
        - *scheme* is a string of the protocol that will be used
        """
        TransferData.__init__(self)
        self._cond = 0
        self._attr = ListenerAttr(scheme)
        self._listener = TransferListener(self._attr, scheme)
        return

    def __del__(self):
        
        for cb in self._cbList:
            self.free_callback(cb)

        return
         

    def register_listen(self, callback, arg):
        """
        Has the server listen for incoming client request
        
        Does not block
        
        **Arguments**
        
        - *callback* the function that will be called as soon a a
        client request is registered
        
        - *arg* a user supplied argument to the callback function
        
        **Returns**
        
        None
        
        **Raises**
        
        A GassTransferServerException is raised if the server is
        unable to listen for client requests
        """

        ret, cbHandle, listener_handle = gassTransferc.register_listen(self._listener.get_handle(), 
                                                                      callback, arg)
        
        if not ret == 0:
            ex =  GassTransferServerException(cbHandle + " error code is: " + str(ret))
            raise ex
        
        self._cbList.append(cbHandle)

    def register_accept(self, attr, callback, arg):
        """
        Has the server accept an incoming client request
        This function should be called after the callback for register_listen
        has been called.

        Does not block
        
        **Arguments**
        
        - *attr* a ListenerAttr object
        
        - *callback* the function that will be called as soon a client request
        is registered
        
        - *arg* a user supplied argument to the callback function
        
        **Returns**
        
        A TransferRequest object is returned the request object is a server
        request
        
        **Raises**
        
        A GassTransferServerException is raised if the server is unable to
        accept this client request
        """
        
        ret, cbHandle, request = gassTransferc.register_accept(attr.get_handle(),
                                                               self._listener.get_handle(),
                                                               callback, arg)
         
        if not ret == 0:
            ex =  TransferServerException(cbHandle + " error code is: " + str(ret))
            raise ex

        self._cbList.append(cbHandle)
        req = TransferRequest(request, 0)
        return req

    
    def authorize(self, request, length=0):
        """
        After the server has accepted a request it can authorize the request if it wants
        the transfer to proceed
        
        **Arguments**
        
        - *request* a server side request returned from the register_accept call
        
        - *length* the number of bytes that the request is authorized to make
        the default 0, will use GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
        
        **Returns**
        returns a TransferRequest object which will be needed for a send_bytes call
        
        **Raises**
        A TransferServerException if the request cannnot be authorized
        """
        
        ret, request = gassTransferc.authorize(request.get_handle(), length)
        if not ret == 0:
            ex =  TransferServerException(request + " error code is: " + str(ret))
            raise ex

        return TransferRequest(request, 0) 

    def deny(self, request, reason, message):
        """
        After the server has accepted the request it can deny the request
        if it does not want the transfer to proceed.

         **Arguments**
        
        - *request* a server side request returned from the register_accept call

        - *reason* an int which is a protocol specific error code

        - *message* a string which is the message given why the request was denied.

        **Returns**
        returns a TransferRequest object which will be needed for a send_bytes call
        
        **Raises**
        A TransferServerException if the request cannnot be denied.
        """
    
        ret, request = gassTransferc.deny(request.get_handle(), reason, message)
        if not ret == 0:
            ex = TransferServerException(request + " error code is: " + str(ret))
            raise ex

        req = TransferRequest(request, 0)
        return req

    
    def get_base_url(self):
        """
        Calls the listener's get base url function
        """
        return self._listener.get_base_url()

    
