############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
The main interface to the Globus Gram Client functionality.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
from pyGlobus.security import GSSCred 
import gramClientc

class GramClientException(GlobusException):
    """
    A simple subclass of GlobusException.

    Used for errors in the gram client module. It defines no new methods.

    **Extends:**
    
        *pyGlobus.util.GlobusException*
    """
    pass

class GramClientAttrException(GlobusException):
    """
    A simple subclass of GlobusException.

    Used for errors in the gram client attribute. It defines no new methods.

    **Extends:**
        
        *pyGlobus.util.GlobusException*
    """
    pass

##############################################################################
# A set of module constants used in the submit_request call.  They describe
# what state changes your interested in being informed of.  They can be
# bitwise or'd together.
#############################################################################
JOB_STATE_PENDING       = 1
JOB_STATE_ACTIVE        = 2
JOB_STATE_FAILED        = 4
JOB_STATE_DONE          = 8
JOB_STATE_SUSPENDED     = 16
JOB_STATE_UNSUBMITTED   = 32
JOB_STATE_STAGE_IN      = 64
JOB_STATE_STAGE_OUT     = 128
JOB_STATE_ALL           = 0xFFFFF

GramClientJobStateNames = { JOB_STATE_PENDING    : "PENDING",
                            JOB_STATE_ACTIVE     : "ACTIVE",
                            JOB_STATE_FAILED     : "FAILED",
                            JOB_STATE_DONE       : "DONE",
                            JOB_STATE_SUSPENDED  : "SUSPENDED",
                            JOB_STATE_UNSUBMITTED: "UNSUBMITTED",
                            JOB_STATE_STAGE_IN  :  "IN",
                            JOB_STATE_STAGE_OUT :  "OUT",
                            JOB_STATE_ALL       :  "ALL" }

##############################################################################
# Possible error states.
###############################################################################
ERROR_PARAMETER_NOT_SUPPORTED=1
ERROR_INVALID_REQUEST=2
ERROR_NO_RESOURCES=3
ERROR_BAD_DIRECTORY=4
ERROR_EXECUTABLE_NOT_FOUND=5
ERROR_INSUFFICIENT_FUNDS=6
ERROR_AUTHORIZATION=7
ERROR_USER_CANCELLED=8
ERROR_SYSTEM_CANCELLED=9
ERROR_PROTOCOL_FAILED=10
ERROR_STDIN_NOT_FOUND=11
ERROR_CONNECTION_FAILED=12
ERROR_INVALID_MAXTIME=13
ERROR_INVALID_COUNT=14
ERROR_NULL_SPECIFICATION_TREE=15
ERROR_JM_FAILED_ALLOW_ATTACH=16
ERROR_JOB_EXECUTION_FAILED=17
ERROR_INVALID_PARADYN=18
ERROR_INVALID_JOBTYPE=19
ERROR_INVALID_GRAM_MYJOB=20
ERROR_BAD_SCRIPT_ARG_FILE=21
ERROR_ARG_FILE_CREATION_FAILED=22
ERROR_INVALID_JOBSTATE=23
ERROR_INVALID_SCRIPT_REPLY=24
ERROR_INVALID_SCRIPT_STATUS=25
ERROR_JOBTYPE_NOT_SUPPORTED=26
ERROR_UNIMPLEMENTED=27
ERROR_TEMP_SCRIPT_FILE_FAILED=28
ERROR_USER_PROXY_NOT_FOUND=29
ERROR_OPENING_USER_PROXY=30
ERROR_JOB_CANCEL_FAILED=31
ERROR_MALLOC_FAILED=32


GramClientErrorStateNames = { 
    ERROR_PARAMETER_NOT_SUPPORTED : "PARAMETER NOT SUPPORTED",
    ERROR_INVALID_REQUEST : "INVALID REQUEST",
    ERROR_NO_RESOURCES : "NO RESOURCES",
    ERROR_BAD_DIRECTORY : "BAD DIRECTORY",
    ERROR_EXECUTABLE_NOT_FOUND : "EXECUTABLE NOT FOUND",
    ERROR_INSUFFICIENT_FUNDS : "INSUFFICIENT FUNDS",
    ERROR_AUTHORIZATION : "AUTHORIZATION", 
    ERROR_USER_CANCELLED : "USER CANCELLED",
    ERROR_SYSTEM_CANCELLED : "SYSTEM CANCELLED", 
    ERROR_PROTOCOL_FAILED : "PROTOCOL FAILED",
    ERROR_STDIN_NOT_FOUND : "STDIN NOT FOUND",
    ERROR_CONNECTION_FAILED : "CONNECTION FAILED", 
    ERROR_INVALID_MAXTIME : "INVALID MAXTIME",
    ERROR_INVALID_COUNT : "INVALID COUNT", 
    ERROR_NULL_SPECIFICATION_TREE : "NULL SPECIFICATION TREE", 
    ERROR_JM_FAILED_ALLOW_ATTACH : "JM FAILED ALLOW ATTACH",
    ERROR_JOB_EXECUTION_FAILED : "JOB EXECUTION FAILED",
    ERROR_INVALID_PARADYN :  "INVALID PARADYN",
    ERROR_INVALID_JOBTYPE :  "INVALID JOBTYPE",
    ERROR_INVALID_GRAM_MYJOB:"INVALID GRAM MYJOB",
    ERROR_BAD_SCRIPT_ARG_FILE: "BAD SCRIPT ARG FILE",
    ERROR_ARG_FILE_CREATION_FAILED : "ARG FILE CREATION FAILED",
    ERROR_INVALID_JOBSTATE : "INVALID JOBSTATE",
    ERROR_INVALID_SCRIPT_REPLY : "INVALID SCRIPT REPLY",
    ERROR_INVALID_SCRIPT_STATUS : "INVALID SCRIPT STATUS",
    ERROR_JOBTYPE_NOT_SUPPORTED : "JOBTYPE NOT SUPPORTED",
    ERROR_UNIMPLEMENTED : "UNIMPLEMENTED",
    ERROR_TEMP_SCRIPT_FILE_FAILED : "TEMP SCRIPT FILE FAILED",
    ERROR_USER_PROXY_NOT_FOUND : "USER PROXY NOT FOUND",
    ERROR_OPENING_USER_PROXY : "OPENENING USER PROXY",
    ERROR_JOB_CANCEL_FAILED : "JOB CANCEL FAILED",
    ERROR_MALLOC_FAILED : "MALLOC FAILED"  }
    

class GramClientAttr:
    """
    A class to wrap the globus gram client attr type.

    It hides the underlying management of the globus_gram_client_attr_t object.
    """

    def __init__(self, handle=None, own=1):
        """
        Constructs an instance.

        Activates the gram client module and creates the gram client attr.

        **Arguments:**

            - *handle* A SWIG'ized pointer to a globus gram client attr object.

            - *own* True is the class owns the underlying memory.

        **Raises:**

            A GramClientAttrException is thrown if unable to create the attr.
        """
        self._activated = 0
        self._own = own
        ret = 1
        ret = utilc.globus_module_activate(gramClientc.get_module())
        if not ret == 0:
            ex = GramClientAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gramClientc.client_attr_init()
            if not ret == 0:
                ex = GramClientAttrException(gramClientc.globus_gram_protocol_error_string(ret))
                raise ex
            self._handle = handle

    def __del__(self):
        """
        Destroy an instance.

            Deletes the gram client attr handle and deactivates the underlying globus
            module.

        """

        if self._own:
            ret = gramClientc.client_attr_destroy(self._handle)
            if not ret == 0:
                ex = GramClientAttrException(ret)
                raise ex
        if self._activated:
            ret = utilc.globus_module_deactivate(gramClientc.get_module())
            if not ret == 0:
                ex = GramClientAttrException("Unable to deactivate the module")
                raise ex

    def set_credential(self, cred):
        """
        Set the underlying credential for gram client operations.

        Assosciates a credential with the attribute set.

        **Arguments:**

            - *cred* A pyGlobus.security.GSSCred object containing
                the cred to assosciate with the attr set.

        **Raises:**

            A GramClientAttr excpetion is raised if unable to set the cred.
        """
        ret, err = gramClientc.client_attr_set_credential(self._handle,
                                                        cred.get_handle())
        if not ret == 0:
            ex = GramClientAttrException(err)
            raise ex
        return

    def get_credential(self):
        """
        Returns the credential assosciated with the attr.

        **Raises:**

            A GramClientAttr exception is raised if uanble to get the cred.

        **Returns:**

            A pyGlobus.security.GSSCred object is returned containing the
            cred assosciated with the attr set.
        """
        ret, handle = gramClientc.client_attr_get_credential(self._handle)
        if not ret == 0:
            ex = GramClientAttrException(handle)
            raise ex
        return GSSCred(handle,0)

    def get_handle(self):
        """
        Return the underlying gram client attr handle.
              
        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
              
        **Returns:**

            A string containing the pointer.
        """
        return self._handle

class GramClient:
    """
    The main interface to the gram client module.

    It is mainly a proxy class for the gramClient module. It provides a higher
    level interface designed to simplify usage.  Exceptions are used to indicate
    all error conditions.
    """
    
    def __init__(self):
        """
        Constructs an instance.
        
        This will activate the globus_gram_client_mddule.

        **Raises:**
        
            A GramClientException is thrown if unable to activate the module.
        """
        self._activated = 0
        self._cbDict = {}
        self._nbCbList = []
        ret = utilc.globus_module_activate(gramClientc.get_module())
        if not ret == 0:
            ex = GramClientException("Unable to initialize the module")
            raise ex
        
        self._activated = 1

    def __del__(self):
        """
        Destroys an instance.

        It will deactivate the globus_gram_client module when
        the object is destroyed.
        """
        for cb in self._cbDict.keys():
            try:
                self.remove_callback(cb)
            except GramClientException:
                pass
        for cb in self._nbCbList:
            gramClientc.free_nb_callback(cb)
        if self._activated:
            ret = utilc.globus_module_deactivate(gramClientc.get_module())
            if not ret == 0:
                ex = GramClientException("Unable to deactivate the module")
                raise ex

    def ping(self, jobContact):
        """
        Pings a gatekeeper.

        **Arguments:**
        
            - *jobContact* A string containing the contact information
        """
        
        status, err = gramClientc.ping(jobContact)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            raise GramClientException(error)
        return
    
    def refresh_credentials(self, jobContact, cred):
        """
        Refresh credentials assosciated with a job.

        **Arguments:**

        - *jobContact* A string containing the job contact information.

        - *cred* An instance of pyGlobus.security.GSSCred containing the 
        credential ot use to contact the job manager. If GSS_C_NO_CREDENTIAL 
        is used the process's default credential will be used.

        **Raises:**

        A GramClientException if an error occurs while refreshing creds.

        **Returns:**

        A job contact.
        """

        ret, jcontact = gramClientc.job_refresh_credentials(jobContact, 
                                                            cred.get_handle())

        if not ret == 0:
            raise GramClientException(jcontact)
        return jcontact

    def register_refresh_credentials(self, jobContact, cred, 
                                            attr, func, userArg):
        """
        Non-blocking for of credential refresh.

        **Arguments:**
        
        - *jobContact* A string containing the job contact information.

        - *cred* An instance of pyGlobus.security.GSSCred containing the 
        credential ot use to contact the job manager. If the GSSCred's
        credential is GSS_C_NO_CREDENTIAL the process's default credential
        will be used.
        
        - *attr* An instance of GramClientAttr containing the client
        attributes to be used.
        
        - *func* callback function to be called when credential refresh is
        completed.
        
        The callback function has the following signature:
        func(userArg, job_contact, job_state, error)

        userArg is the argument passed to register_refresh_credentials
        job_contact is a string containing the job contact string
        job_state is an int containing one of the possible job states
        error is a tuple containing (error_code, error_description)
        
        - *userArg* A user defined argument to be passed to the callback
        functions.
        
        **Raises:**
        
        A GramClientException if unable to refresh the credential.
        """
        status, cbStruct = gramClientc.register_job_refresh_credentials(jobContact,
                                                                        cred.get_handle(), 
                                                                        attr.get_handle(), 
                                                                        func, userArg)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        self._nbCbList.append(cbStruct)
        return 

    def set_callback(self, callback, arg):
        """
        Sets a callback.

        Sets the function to be notified of job state changes.

        **Arguments:**
        
            - *callback* The function that should be called when the job state changes.
            The function has the following prototype:
            'function(arg, contact, state, error)'
            where contact is the job contact, state is one of the module constants
            defined above, and error  and error is a tuple (error_code, error_description).
            If there is no error, the error_code is 0 and the error description is a
            None object

            - *arg* The argument to be passed to the callback when it is called.

        **Raises:**
        
            A GramClientException if an error occurs while setting the callback.

        **Returns:**
        
            A string containing the callback contact is returned. This can
            be passed to submit_request.
        """
        status, handle, contact = gramClientc.callback_allow(callback, arg)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        self._cbDict[contact] = handle
        return contact

    def submit_request(self, resourceManager, description, jobStateMask, callbackContact):
        """
        Submits a job request to a resource manger.

        **Arguments:**
        
            - *resourceManager* The resource manager to submit the job to.

            - *description* An RSL string that contains the job description.

            - *jobStateMask* Which job state change message you wish to be notified of.
            The notification happens through the callback. This is an int that can
            be formed by bitwise or'ing the module constants defined above.

            - *callbackContact* A string containing the contact for the callback that
            should be notified of job state changes.

        **Raises:**
        
            A GramClientException if unable to submit the job request.

        **Returns:**        
            A string containing the job contact used for communicting with the
            remote job.
        """
        status, contact = gramClientc.job_request(resourceManager, description,
                                                 jobStateMask, callbackContact)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        return contact

    def register_submit_request(self, resourceManager, description, 
                                jobStateMask, callbackContact, attr, 
                                func, userArg):
        """
        Non-blocking submit a job request to a resource manger.

        **Arguments:**
        
            - *resourceManager* The resource manager to submit the job to.

            - *description* An RSL string that contains the job description.

            - *jobStateMask* Which job state change message you wish to be notified of.
            The notification happens through the callback. This is an int that can
            be formed by bitwise or'ing the module constants defined above.

            - *callbackContact* A string containing the contact for the callback that
            should be notified of job state changes.

            - *attr* An instance of GramClientAttr.

            - *func* Callback function to be called when job submission is
              completed.

              The callback function has the following signature:
                  func(userArg, job_contact, job_state, job_error, general_error)

                  userArg is the argument passed to register_submit_request
                  job_contact is a string containing the job contact string
                  job_state is an int containing one of the possible job states
                  job_error and general error are tuples of (error_code, error_description).
                  If there is no error, the error_code is 0 and the error description is a
                  None object
                  
            - *userArg* A user defined argument to be passed to the callback
                functions.

        **Raises:**
        
            A GramClientException if unable to submit the job request.
        """
        status, cbStruct = gramClientc.register_job_request(resourceManager,
                                                            description,
                                                            jobStateMask, 
                                                            callbackContact,
                                                            attr.get_handle(), 
                                                            func, userArg)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        self._nbCbList.append(cbStruct)
        return 

    def check_status(self, jobContact):
        """
        Checks the status of a current job.

        **Arguments:**
        
            - *jobContact* A string containing the contact information for the job
            you would like the status of.

        **Raises:**
        
            A GramClientException if unable to get the job status.

        **Returns:**        
            An int containing the current job status.  This int will be one
            of the module constants defined above.

            Also returns a second int *if* the job status is failure.
        """
        ret, status, failure = gramClientc.job_status(jobContact)
        if not ret == 0:
            error = gramClientc.globus_gram_protocol_error_string(ret)
            ex = GramClientException(error)
            raise ex
        return status, failure

    def register_check_status(self, jobContact, attr, callback, arg):
        """
        Nonblocking query of a job's status.

        This function queries the status of the job associated with the job contact,
        returning it's current job status and job failure reason if it has failed.

        **Arguments:**
        
        - *jobContact* A string containing the contact for the callback that
        should be queried.

        - *attr* An instance of GramClientAttr.        

        - *callback* Callback function to be called when job submission is
        completed.

        The callback function has the following signature:
        func(userArg, job_contact, job_state, error)
        userArg is the argument passed to register_submit_request
        job_contact is a string containing the job contact string
        job_state is an int containing one of the possible job states
        error is a tuple containing (error_code, error_description).
        If there is no error, the error_code is 0 and the error description is a
        None object
        
        - *arg* A user defined argument to be passed to the callback
        functions.
                
        **Raises:**
        A GramClientException if unable to cancel the job.        
        """        
        status, errorcode = gramClientc.register_job_status(jobContact, 
                                                            attr.get_handle(), 
                                                            callback, arg)
        if not status == 0:
            ex = GramClientException('problem in register_check_status')
            raise ex
        return
                                                        
        
    def cancel_job(self, jobContact):
        """
        Cancels a job.

        **Arguments:**
        
            - *jobContact* A string containing the contact information for the job to
            be canceled.

        **Raises:**
        
            A GramClientException if unable to cancel the job.
        """
        status, canceledJob = gramClientc.job_cancel(jobContact)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        return

    def register_cancel_job(self, jobContact, attr, callback, arg):
        """
        Cancels a job, giving return information to a callback rather
        than blocking for it.

        **Arguments:**
   
        - *jobContact* A string containing the contact information for the job to
                       be canceled.

        - *attr* An instance of GramClientAttr.

        - *callback* Callback function to be called when job submission is
        completed.

        The callback function has the following signature:
        func(userArg, job_contact, job_state, error)
        
        userArg is the argument passed to register_submit_request
        job_contact is a string containing the job contact string
        job_state is an int containing one of the possible job states
        error is a  tuple containing (error_code, error_description).
        If there is no error, the error_code is 0 and the error description is a
        None object

        - *arg* A user defined argument to be passed to the callback
        functions.

        **Raises:**        
        A GramClientException if unable to cancel the job.

        """
        status, other = gramClientc.register_job_cancel(jobContact, 
                                                        attr.get_handle(), 
                                                        callback, arg)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        return
    

    def remove_callback(self, callbackContact):
        """
        Stops the callback.

        Stops the callback from listening for job state changes and cleans up
        any resources.

        **Arguments:**
        
            - *callbackContact* The string identifier for the callbck to be removed.

        **Raises:**
        
            A GramClientException if unable to remove the callback.
        """
        try:
            handle = self._cbDict[callbackContact]
        except KeyError:
            ex = GramClientException("Unable to find the callback contact")
            raise ex
        status = gramClientc.remove_callback(callbackContact, handle)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        del self._cbDict[callbackContact]
        return

    def register_callback(self, jobContact, stateMask, callbackContact):
        """
        Registers a callback.

        Sets the function to receive job state changes for the given job.

        **Arguments:**
        
            - *jobContact* The job who's changes should be reported to the given
            callback.
             
            - *stateMask* Which job state change messages you wish to be notified of.
            The notification happens through the callback. This is an int that
            can formed by bitwise or'ing the module constants defined above.

            - *callbackContact* The callback that should be notified.

        **Raises:**
        
            A GramClientException if unable to register the callback.

        **Returns:**
            Nothing
        """
        ret, errorstr = gramClientc.callback_register(jobContact, stateMask,
                                                            callbackContact)
        if not ret == 0:
            ex = GramClientException(errorstr)
            raise ex
        return

    def register_callback_registration(self, jobContact, jobStateMask, 
                                       callbackContact, attr, func, userArg):
        """
        Non-blocking callback registration.

        Sets the function to receive job state changes for the given job.
        **Arguments:**
        
            - *jobContact* A string containing the job contact information.

            - *jobStateMask* Which job state change messages you wish to be
              notified of.  The notification happens through the callback. 
              This is an int that can formed by bitwise or'ing the module
              constants defined above.

            - *callbackContact* The callback that should be notified of job
              state changes.

            - *attr* An instance of GramClientAttr containing the client
              attributes to be used.

            - *func* Callback function to be called when callback registration is
              completed.

              The callback function has the following signature:
                  func(userArg, job_contact, job_state, error)

                userArg is the argument passed to register_callback_registration
                job_contact is a string containing the job contact string
                job_state is an int containing one of the possible job states
                error is a tuple containing (error_code, error_description).
                If there is no error, the error_code is 0 and the error description is a
                None object
                
            - *userArg* A user defined argument to be passed to the callback
                functions.

        **Raises:**
        
            A GramClientException if unable to register the callback.
        """
        status, cbStruct = gramClientc.register_callback_registration(jobContact,
                                                            jobStateMask,
                                                            callbackContact, 
                                                            attr.get_handle(), 
                                                            func, userArg)
        if not status == 0:
            error = gramClientc.globus_gram_protocol_error_string(status)
            ex = GramClientException(error)
            raise ex
        self._nbCbList.append(cbStruct)
        return 

    def debug(self):
        """
        Enables debugging output.
        """
        gramClientc.globus_gram_client_debug()
        return

    def version(self):
        """
        Returns an int version number for the gram client package.
        """
        return gramClientc.globus_gram_client_version()

    def unregister_callback(self, jobContact, callbackContact):
        """
        Unregisters a callback.

        Stop notifing the callback of job state changes for the given job.

        **Arguments:**
            - *jobContact* A string containing the contact information        
            - *callbackContact* The callback that should no longer be notified.

        **Raises:**
        
            A GramClientException is thrown if unable to unregister the
            callback.

        Returns:        
        An int containing the current job status.  This is one of the
        values defined in the module constants.
        """
        ret, status = gramClientc.callback_unregister(jobContact,
                                                      callbackContact)
        if not ret == 0:
            error = gramClientc.globus_gram_protocol_error_string(ret)
            ex = GramClientException(error)
            raise ex
        return status

