"""
NetLogger autologging for pyGlobus.

NetLogger URL: http://dsd.lbl.gov/NetLogger/

Dan Gunter <dkgunter@lbl.gov>
"""
import time, threading
from util import GlobusXIODebug 
import xioc

# -- Globals

_dbg, _log, _xdbg, _host_ip = False, None, None, None

TRANSFER_EVENT = 'xio.XioSocket.transfer'

# -- Public functions

def flush():
    if _log:
        _log.flush()

def getXIODebug():
    if _xdbg is None or _xdbg.level is None:
        return 0
    return ( _xdbg.level & GlobusXIODebug.PERF1 )

def getXIODebug2():
    if _xdbg is None or _xdbg.level is None:
        return 0
    return ( _xdbg.level & GlobusXIODebug.PERF2 )

def getFd():
    """Don't use this unless you *really* know what you're doing."""
    if _log is None or _log._logfile is None:
        return -1
    # share file with Python logging
    return _log._logfile.fileno()
        
         
def getGID():
    return _gid

def getHost():
    global _host_ip
    if _host_ip is not None:
        ip = _host_ip
    else:
        import socket
        try:
            ip = socket.gethostbyname(socket.getfqdn())
        except:
            ip = '0.0.0.0'
        _host_ip = ip
    return ip

def instrument(clazz, prefix='', exclude=[]):
    """
    Instrument class 'clazz' methods with entry/exit points
    named `clazz`.{method}.start and `clazz`.{method}.end
    """
    if not _dbg:
        return
    autolog.autologClass(clazz=clazz, exclude=exclude + ['_.*'],
                         write_fn=dbg_write,
                         prefix=prefix)

def setLogDest(dest):
    """
    Manually set logging destination, overriding default settings
    from the environment.
    """
    if _log:
        _openLog(dest)
        
def setLogLevel(level):
    """
    Manually set logging level, overriding default settings
    from the environment.
    """
    global _dbg
    if _log:
        _log.setLevel(level)
        if level >= nllite.Level.DEBUG:
            _dbg = True

def setMeta(event=None, bare_event=False, **metakw):
    """
    Set metadata name/value pairs, for all events if 'event'
    is None, or the specific named event or list of events.

    Example:
       setMeta('prefix.class.method',pi=3.14,sky='blue')
    """
    if not _dbg:
        return
    if event is None:
        metakw['NL.GID'] = _gid
        if not metakw.has_key('HOST'):
            metakw['HOST'] = getHost()
        _log.setMeta(None, metakw)
    else:
        import types
        if isinstance(event, types.StringType):
            events = [event]
        else:
            events = event
        for e in events:
            if bare_event:
                _log.setMeta(event, metakw)
            else:
                for where in autolog.START, autolog.END:
                    event = "%s%s%s" % (e, autolog.DELIM, where)
                    _log.setMeta(event, metakw)

def dbg_write(e, params={}):
    """Don't use this directly unless you know what you're doing."""
    _log.debug(e, params)
  
# -- Private functions

def _init():
    global _log, _gid
    nldest = nllite.get_dest()
    if nldest is not None:
        _openLog(nllite.get_dest())
    elif _xdbg.filename is not None:
        _openLog(_xdbg.filename)
    else:
        _openLog('-') # stdout
    if _log:
        _gid = nllite.get_gid()
        setMeta()
    _initC()

def _initC():
    xioc.g_id = getGID()
    xioc.xio_nl_init()
    if getFd() > -1 and getXIODebug2():
        xioc.xio_nl_on(getFd())
    else:
        xioc.xio_nl_off()

def _openLog(dest):
    global _log, _dbg
    try:
        _log = nllite.LogOutputStream(dest)
        _log.setLevel(_nlliteLevel())       
    except:
        _log, _dbg = None, False

def _nlliteLevel():
    if getXIODebug2():
        return nllite.Level.DEBUG
    if getXIODebug():
        return nllite.Level.DEBUG
    else:
        return nllite.Level.INFO

class _DynamicXIODebug(threading.Thread):
    """Periodically reparse the XIO debug string, make changes to levels.
    """
    INTERVAL = 5 # seconds
    def run(self):
        global _xdbg, _xdbg_chk, _dbg
        while 1:
            time.sleep(self.INTERVAL)
            old = _xdbg
            _xdbg = GlobusXIODebug()
            if _xdbg.val != old.val:
                flush()
                dbg = getXIODebug() or getXIODebug2()  
                if dbg:
                    if not _dbg:
                        _dbg = dbg
                        _init()
                    _log.setLevel(_nlliteLevel())
                if getXIODebug2():
                    xioc.xio_nl_on(getFd())
                else:
                    xioc.xio_nl_off()

# -- Run on import

try:
    from gov.lbl.dsd.netlogger import nllite, autolog
    _xdbg = GlobusXIODebug()
    _dbg = getXIODebug() or getXIODebug2()
    if _dbg:
        _init()
    if _xdbg.from_file:
        _dydbg = _DynamicXIODebug()
        _dydbg.setDaemon(1)
        _dydbg.start()
except ImportError:
    pass
