############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
The main interface to the Globus io module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException, Buffer
import ioc
import SocketServer, socket,  httplib
from pyGlobus.security import GSSContext
from pyGlobus.security import GSSCred
from pyGlobus import gssc
from exceptions import RuntimeError
##########################################################################
# Module Methods
#########################################################################
def server_auth_callback(server, g_handle, remote_user, context):    
    server.remote_user = remote_user 
    return 1

def client_auth_callback(server, g_handle, remote_user, context): 
    return 1

###########################################################################
# Module Classes
###########################################################################

class AuthDataException(GlobusException):
    pass

class IOBaseException(GlobusException):
    pass

class GSITCPException(IOBaseException):
    pass

class FileIOAttrException(GlobusException):
    pass

class NetIOAttrException(GlobusException):
    pass

class TCPIOAttrException(NetIOAttrException):
    pass

class AuthData:

    def __init__(self, handle=None):
        self._cbList = []
        ret = utilc.globus_module_activate(utilc.get_module())
        if not ret == 0:
            ex = AuthDataException("Unable to initialize the common module")
            raise ex
        ret = utilc.globus_module_activate(ioc.get_module())
        if not ret == 0:
            ex = AuthDataException("Unable to initialize the module")
            raise ex
        if handle == None:
            ret, handle = ioc.authorization_data_init()
            if not ret == 0:
                ex = AuthDataException(handle)
                raise ex
        self._handle = handle
        return

    def __del__(self):
        for cb in self._cbList:
            self.free_callback(cb)
        ret, err = ioc.authorization_data_destroy(self._handle)
        if not ret == 0:
            ex = AuthDataException(err)
            raise ex
        ret = utilc.globus_module_deactivate(ioc.get_module())
        if not ret == 0:
            ex = AuthDataException("unable to deactivate the module")
            raise ex
        ret = utilc.globus_module_deactivate(utilc.get_module())
        if not ret == 0:
            ex = AuthDataException("unable to deactivate the module")
            raise ex
        return

    def set_identity(self, identity):
        """
        Set the "identity" member of the authorization-mode specific data
        structure. This member should be set to a string containing the
        security distinguished name of the peer which you will authorize
        for this connection. The identity is only when the
        authorization-mode attribute is set to GLOBUS_IO_SECURE_AUTHORIZATION_MODE_IDENTITY.

        **Arguments**

        - *identity* The string naming the security identity to authorize.

        **Returns**
        None
        """
        ret, err = ioc.authorization_data_set_identity(self._handle, identity)
        if not ret == 0:
            ex = AuthDataException(err)
            raise ex
        return

    def get_identity(self):
        """
        Query the "identity" member of the authorization-mode specific data
        structure. This member is set by a call to
        globus_io_secure_authorization_mode_set_identity.

        **Returns**
        A pointer to hold a copy of the identity string
        set in the data structure.
        """
        authList = ioc.authorization_data_get_identity(self._handle)
        if not authList[0] == 0:
            ex = AuthDataException(authList[1])
            raise ex
        return authList[2]

    def set_callback(self, callback, arg):
        """
        Set the callback and callback-arg members of the authorization-mode
        specific data structure. This member is used when setting the 
        authorization-mode attribute of a TCP attribute set to
        GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK in a call to 
        globus_io_attr_set_secure_authorization_mode().

        **Arguments**

        - *callback* The new value of the callback member.
        - *arg* The new value of the callback-arg member.
 
        **Returns**
        The new callback structure. 

        **Raises**
        Raises a AuthDataException if the function couldn't set the callback.
        """
        ret, cbStruct = ioc.authorization_data_set_callback(self._handle,
                                                           callback, arg)
        if not ret == 0:
            ex = AuthDataException(cbStruct)
            raise ex
        self._cbList.append(cbStruct)
        return cbStruct


    def free_callback(self, cbHandle):
        """
        #########################################################################
        # Free's the underlying memory allocated for the callback structure.
        #########################################################################
        Freedom for the callback struct.  
        """
        ioc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return
    

    def get_handle(self):
        """
        ########################################################################
        # XXX This handle is only valid as long as this object exists.  When
        # the object is destroyed, the memory this handle points to is freed.
        #######################################################################
        """
        return self._handle
    
class IOBase:

    def __init__(self, handle):
        """
        Constructs a IOBase class instance, using the given handle.
        **Arguments**
        - *handle* the handle to be used in IOBase construction.
        """
        self._handle = handle
        self._cbList = []
        ret = utilc.globus_module_activate(utilc.get_module())
        if not ret == 0:
            ex = IOBaseException("Unable to initialize the common module")
            raise ex
        ret = utilc.globus_module_activate(ioc.get_module())
        if not ret == 0:
            ex = IOBaseException("Unable to initialize the io module")
            raise ex
        return

    def __del__(self):
        for cb in self._cbList:
            self.free_callback(cb)
        if not self._handle == None:
            try:
                self.close()
            except IOBaseException, ex:
                raise ex
            ioc.free_handle(self._handle)
        ret = utilc.globus_module_deactivate(ioc.get_module())
        if not ret == 0:
            ex = IOBaseException("Unable to deactivate the io module")
            raise ex
        ret = utilc.globus_module_deactivate(utilc.get_module())
        if not ret == 0:
            ex = IOBaseException("Unable to deactivate the common module")
            raise ex
        return

    def register_close(self, callback, arg):
        """   
        Asynchronously close the file or connection.
 
        When this function returns successfully, no further operations may
        be done on this handle. Any outstanding requests on this handle
        with be cancelled with their callbacks invoked before the close
        completes.
        
        No other callbacks may be registered with the globus_io system for
        this handle after this function is called. Close registrations may
        not be canceled.

        **Arguments**
        
        - *callback* The callback function is called after all of the
        cancelled callbacks have been dispatched and have reached their
        thread conversion point or completed

        - *arg* Parameter passed to the callback function.

        **Returns**
         A SWIG'ized pointer to the callback handle is returned. After the callback
        completes, this may be free'd with the free_callback method, or it will be
        free'd when the instance is destroyed. 

        **Raises**
        Raises a IOBaseException 
        """
        if self._handle == None:
            ex = IOBaseException("Unable to close uninitialized handle")
            raise ex
        ret, cbHandle = ioc.io_register_close(self._handle, callback, arg)
        if not ret == 0:
            ex = IOBaseException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle
    
    def close(self):
        """
        Close the handle associated with this IOBase.
        """
        if self._handle == None:
            ex = IOBaseException("Unable to close uninitialized handle")
            raise ex
        ret, err = ioc.io_close(self._handle)
        if not ret == 0:
            ex = IOBaseException(err)
            raise ex
        return

    def cancel(self, doCallbacks):
        """
        Cancel any outstanding operation registered to this IOBase.
 
        
        **Arguments**
 
        - *doCallbacks* - If this parameter is set to True,
        then this function will block until all outstanding callbacks for
        this handle have been invoked and either completed or reached their
        conversion point. Otherwise, this will unregister any pending
        callbacks, and they will not be called.
       
	**Returns** 
        None
        
        **Raises**
        Raises a IOBaseException if cancel did not succeed or if the handle
        was not set properly.
         
        Cancel may succeed without cancelling anything, or causing any
        callbacks  to be invoked.
        There currently is no way to cancel specific registered
        callbacks without cancelling every registered operation.
        """
        if self._handle == None:
            ex = IOBaseException("Unable to cancel uninitialized handle")
            raise ex
        ret, err = ioc.io_cancel(self._handle, doCallbacks)
        if not ret == 0:
            ex = IOBaseException(err)
            raise ex
        return

    def register_read(self, buffer, maxBytes, waitForBytes, callback, arg):
        """
        Asynchronous TCP or file read.

        **Arguments**
        
        - *buffer* The buffer to read data into.
        - *maxBytes* The maximum number of bytes which will be read.
        - *waitForBytes* The minimum number of bytes to read before calling
        the callback function.
        - *callback* Funtion which is executed when the read has been satisfied.
        - *arg* Parameter to the callback function.

        **Returns**
        Returns the associated callback handle.

        **Raises**
        Raises a IOBaseException if internally: 
        The handle or callback parameter was equal to GLOBUS_NULL, so the
        registration could not be processed or -
        The handle parameter was already registered for closing, so no more
        registrations can occur on this handle or - 
        The handle parameter was already registered for reading or - 
        The type of handle passed to the function by the handle parameter is
        not one of the types which supports the read operation.
        
        """
        if self._handle == None:
            ex = IOBaseException("Cannot read from uninitialized handle")
            raise ex
        ret, cbHandle = ioc.io_register_read(self._handle, buffer.get_handle(),
                                            maxBytes, waitForBytes,
                                            callback, arg)
        if not ret == 0:
            ex = IOBaseException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def try_read(self, buffer, maxBytes):
        """
        Nonblocking TCP or file read
        
        globus_io_try_read() will read whatever data is immediatedly
        available from the handle without blocking. The value of
        nbytes_read will be updated to contain the amount of data actually
        read from this handle. 

        **Arguments**

        - *buffer* The buffer to read data into.
        - *maxBytes* The maximum number of bytes which can be read. 

        **Returns**
        Returns the number of bytes read. 

        **Raises**
        Raises a IOBaseException if: 
        The handle, buf, or nbytes_read parameter was equal to GLOBUS_NULL,
        so the operation could not be processed.
        The handle parameter was not initialized or ready, so the read
        operation could not be processed.
        The handle parameter was already registered for closing, so no more
        operations can occur on this handle.
        The handle parameter was already registered for reading. 
        The buf parameter was not NULL, but was invalid. 
        Some unexpectd I/O error occurred. 
        An end-of-file occurred on the read.

        ***XXX*** this function will always return 0 bytes read for TCP
        connections which are configured to use GSSAPI or SSL data wrapping.
        This is a documented Globus bug. 
        """
        
        if self._handle == None:
            ex = IOBaseException("Cannot read from uninitialized handle")
            raise ex
        ret, bytesRead = ioc.io_try_read(self._handle,
                                        buffer.get_handle(), maxBytes)
        if not ret == 0:
            ex = IOBaseException(bytesRead)
            raise ex
        return bytesRead

    def read(self, buffer, maxBytes, waitForBytes):
        """
        
        Blocking TCP or file read
        
        Perform a blocking read on the handle. This will block until buf is
        filled with at least wait_for_nbytes and at most max_nbytes of data
        from handle, or end-of-file is reached.
        **Arguments**
        - *buffer* The buffer to read data into.
        - *maxBytes* The maximum number of bytes which can be read.
        - *waitForBytes* The minimum number of bytes to read before
        returning.

        **Returns**
        Returns number of bytes read.

        **Raises**
        Raises a IOBaseException if one of these internal events occurred:
        The handle or nbytes_read parameter was equal to GLOBUS_NULL, so the
        registration could not be processed.
        The handle parameter was not initialized or ready, so the read
        registration could not be processed.
        The handle parameter was already registered for closing, so no more
        registrations can occur on this handle.
        The handle parameter was already registered for reading. 
        The type of handle passed to the function by the handle parameter is
        not one of the types which supports the read operation.
        """ 
         
        if self._handle == None:
            ex = IOBaseException("Cannot read from uninitialized handle")
            raise ex
        ret, bytesRead = ioc.io_read(self._handle, buffer.get_handle(),
                                    maxBytes, waitForBytes)
        if not ret == 0:
            ex = IOBaseException(bytesRead)
            raise ex
        return bytesRead

    def register_write(self, data, dataLen, callback, arg):
        """
        Asynchronous TCP or file write.
 
        Perform a write on the handle. 
        **Arguments**

        - *data* The data to be written.
        - *dataLen* The size of the data buffer.
        - *callback* Function which is executed when the write has completed.
        - *arg* Parameter to the callback function.

        **Returns**
        Returns the callback handle allocated for this callback.

        **Raises**
        Raises a IOBaseException if one of these internal events occurred:

        The handle or callback parameter was equal to GLOBUS_NULL, so the
        registration could not be processed.
        The handle parameter was not initialized or ready, so the write
        registration could not be processed.
        The handle parameter was already registered for closing, so no more
        registrations can occur on this handle.
        The handle parameter was already registered for writing. 
        The type of handle passed to the function by the handle parameter
        is not one of the types which supports the write operation.
        The handle cannot protect the data using the GSSAPI. The data
        cannot be written.

        ***XXX*** In the case of an error during a protected data transfer, the
        amount of data passed to the callback is not particularly reliable.
        This is a documented Globus bug.  
        """
        
        if self._handle == None:
            ex = IOBaseException("Cannot write to an uninitialized handle")
            raise ex
        ret, cbHandle = ioc.io_register_write(self._handle, data, dataLen,
                                             callback, arg)
        if not ret == 0:
            ex = IOBaseException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def try_write(self, data, maxLen):
        """        
        Nonblocking TCP or file write
 
        try_write() will write whatever data can immediately
        processed by the operating system without blocking.

        **Arguments**

        - *data* The data to write.
        - *maxLen* The maximum number of bytes which can be written. 

        **Returns**
        Returns the number of bytes written.
 
        ***XXX*** This function will always return 0 bytes written for TCP
        connections which are configured to use GSSAPI or SSL data wrapping. 
        """
         
        if self._handle == None:
            ex = IOBaseException("Cannot write to an uninitialized handle")
            raise ex
        ret, bytesWritten = ioc.io_try_write(self._handle, data, maxLen)
        if not ret == 0:
            ex = IOBaseException(bytesWritten)
            raise ex
        return bytesWritten

    def write(self, data, dataLen):
        """
        Blocking TCP or file write.
 
        Perform a blocking write operation on the handle. This will block
        until the contents of data are written to the network or file, or an
        error occurs.

        **Arguments**
        
        - *data* The data to write.
        - *dataLen* The maximum number of bytes which can be written.

        **Returns**
        Returns the number of bytes written.

        **Raises**
        Raises a IOBaseException if one of these internal events occurred:
        The handle or nbytes_written parameter was equal to GLOBUS_NULL, so the
        write could not be processed.
        The handle parameter was not initialized or ready, so the write
        operation could not be processed.
        The handle parameter was already registered for closing, so no more
        operations can occur on this handle.
        The handle parameter was already registered for writing.
        The type of handle passed to the function by the handle parameter is
        not one of the types which supports the write operation.
        The handle cannot protect the data using the GSSAPI. The data
        cannot be written.
        The write called failed with an unexpected error. The errno returned
        by write() is accessible from the error object.
 
        ***XXX*** In the case of an error during a protected data transfer, the
        return value might be bunk. This is a documented GLOBUS bug. 
        """
 
        if self._handle == None:
            ex = IOBaseException("Cannot write to an uninitialized handle")
            raise ex
        ret, bytesWritten = ioc.io_write(self._handle, data, dataLen)
        if not ret == 0:
            ex = IOBaseException(bytesWritten)
            raise ex
        return bytesWritten
    
    def free_callback(self, cbHandle):
        """
        #########################################################################
        # Free's the underlying memory allocated for the callback structure.
        #########################################################################
        """
        ioc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return
    
    def get_handle(self):
        """
        ########################################################################
        # XXX This handle is only valid as long as this object exists.  When
        # the object is destroyed, the memory this handle points to is freed.
        #######################################################################
        """
        return self._handle

class FileIOAttr:

    def __init__(self, handle = None):
        """
        Constructor for the FileIOAttr class, taking an optional handle.
        """
        self._activated = 0
        ret = utilc.globus_module_activate(ioc.get_module())
        if not ret == 0:
            ex = FileIOAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = ioc.fileattr_init()
            if not ret == 0:
                ex = FileIOAttrException("Unable to initialize the attr")
                raise ex
        self._handle = handle

    def __del__(self):
        """
        Destructor for FileIOAttr.  Destroys contained handle as well.
        """
        ret = ioc.fileattr_destroy(self._handle)
        if not ret == 0:
            ex = FileIOAttrException("unable to destroy the attr")
            raise ex
        ret = utilc.globus_module_deactivate(ioc.get_module())
        if not ret == 0:
            ex = FileIOAttrException("Unable to deactivate the module")
            raise ex

    def get_handle(self):
        """
        ########################################################################
        # XXX This handle is only valid as long as this object exists.  When
        # the object is destroyed, the memory this handle points to is freed.
        #######################################################################
        """
        return self._handle

class GSIFile:
    """
    This is a file-like class that wraps a GSITCPSocket.
    """
    def __init__(self, sock, mode):
        """
        Initialize a GSIFile class.  the sock argument should be a GSITCPSocket.
        The mode argument is interpreted as for a file object, except that only
        'r' and 'w' are meaningful.
        """
        self.readok = 0
        self.writeok = 0 
        self.closed = 0
        for char in mode[:]:
            if char == "r":
                self.readok = 1
                self.rbuf = Buffer(8192)
            if char == "w":
                self.writeok = 1
        self.sock = sock
        self.sock.refcount = self.sock.refcount + 1

    def __del__(self):
        """
        Close the socket if it is still open 
        """
        if not self.closed:
            self.close()
            self.closed = 1 
	
    def read(self, size=0, waitForBytes=-1):
        """
        Read size bytes, returning a string.
        **Arguments**
        
        - *size* - The number of bytes that you wish to read in.

        -*waitForBytes* - the minimum number of bytes that will be read in before
        returning, if unspecified it will default to size
        
        **Returns**
        A string is returned
        
        **Raises**
        None. This method doesn't raise any exceptions but the socket object could
        raise one.
        """
        if waitForBytes == -1:
            waitForBytes = size
        if self.readok == 0:
            raise IOError, "File not open for reading"
        if size == 0:
            size = self.rbuf.get_size()
        if waitForBytes > size:
            raise RuntimeError, "waitForBytes is larger than buffer_size" 
        elif self.rbuf.get_size() < size:
            del(self.rbuf)
            self.rbuf = Buffer(size)
         
        nbytes = self.sock.read(self.rbuf, size, waitForBytes)
        if nbytes < 1:
            return ""
        return self.rbuf.as_string(nbytes)
    
    def readline(self):
        """
        Read a line, and return it.
        """
        # There's probably a much more efficient way to implement this...
        readStr = ""
        while (1):
            s = self.read(1)
            if s == "":
                break
            readStr = readStr + s[0]
            if s[0] == '\n':
                break
        return readStr
    
    def write(self, str):
        """
        Write a string to the socket.
        """
        if self.writeok == 0:
            raise IOError, "File not open for writing"
        self.sock.write(str, len(str))
        
    def flush(self):
        """
        In theory, flush any pending output.  In practice, this does nothing.
        """
        if self.writeok == 0:
            raise IOError, "File not open for writing"
        pass

    def close(self):
        """
        Close the associated socket.
        """ 
        self.sock.close()
        self.closed = 1
        
    def shutdown(self, n):
        """
        Shutdown the associated socket.
        """
        self.sock.shutdown(n)

class GSIInfile(GSIFile):
    def __init__(self, g_handle):
        GSIFile.__init__(self, g_handle, "r")

class GSIOutfile(GSIFile):
    def __init__(self, g_handle):
        GSIFile.__init__(self, g_handle, "w")

class GSITCPServer(SocketServer.BaseServer):

    """
    Base class for various socket-based server classes.

    Defaults to synchronous IP stream (i.e., TCP).

    Methods for the caller:

        - __init__(server_address, RequestHandlerClass)
	- serve_forever()
	- handle_request()  # if you don't use serve_forever()
	- fileno() -> int   # for select()

    Methods that may be overridden:

       - server_bind()
       - server_activate()
       - get_request() -> request, client_address
       - verify_request(request, client_address)
       - process_request(request, client_address)
       - close_request(request)
       - handle_error()
    
    Methods for derived classes:

        - finish_request(request, client_address)

    Class variables that may be overridden by derived classes or
    instances:

        - address_family
        - socket_type
	- request_queue_size (only for stream sockets)
	- reuse_address

    Instance variables:

        - server_address
	- RequestHandlerClass
	- socket
    """

    address_family = socket.AF_INET

    socket_type = socket.SOCK_STREAM

    request_queue_size = 5

    allow_reuse_address = 0

    def __init__(self, server_address, RequestHandlerClass,
                 channel_mode=ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP,
                 delegation_mode=ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY,
                 tcpAttr=None):
        """Constructor.  May be extended, do not override."""
        SocketServer.BaseServer.__init__(self, server_address, 
                                         RequestHandlerClass)
        if tcpAttr == None:
            self.attr = TCPIOAttr()
            self.attr.set_authentication_mode(ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
            self.authdata = AuthData()
            self.authdata.set_callback(server_auth_callback, self)
            self.attr.set_authorization_mode(ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK, self.authdata)
            self.attr.set_delegation_mode(delegation_mode)
            self.attr.set_restrict_port(0)
            self.attr.set_reuseaddr(1)
            self.attr.set_nodelay(0)
            self.attr.set_channel_mode(channel_mode)
        else:
            self.attr = tcpAttr
        self.delegated_cred = None
        self.socket = GSITCPSocket()
        self.server_bind()
        self.server_activate()
        return

    def get_delegated_credential(self):
        """
        Accessor for the server's delgated credentials.
        """
        return self.delegated_cred
    
    def server_bind(self):
        """
        Called by constructor to bind the socket.
        May be overridden.
        """
        self.port = self.socket.create_listener(self.attr,
                                                self.server_address[1])


    def server_activate(self):
        """
        Called by constructor to activate the server.
        May be overridden.
        """
        pass

    def server_close(self):
        """
        Called to clean-up the server.
        May be overridden.
        """
        self.socket.close()

    def fileno(self):
        """
        Return socket file number.
        Interface required by select().
        """
        return self.socket.fileno()

    def get_request(self):
        """
        Get the request and client address from the socket.
        May be overridden.
        """
        try:
            self.socket.listen()
        except GSITCPSocketException, ex:
            raise ex 
        try:
            (request, client_address) = self.socket.accept(self.attr)
            self.delegated_cred = request.get_delegated_credential()
        except GSITCPSocketException, ex:
            raise ex
        return request, client_address

    def close_request(self, request):
        """
        Called to clean up an individual request.
        """
        request.close()

class GSITCPSocketException(IOBaseException):
    """
    A sub-class of IOBaseException.

    Used to report errors in the GSITCPSocket class.

    **Extends:**    
         *pyGlobus.io.IOBase.IOBaseException*
    """
    pass

class GSITCPSocket(IOBase):
    """
    A simple socket interface to secure Globus IO.

    Provides the ability to make strongly authenticated connetions using GSI. It
    supports several modes of authorization and provides optional integrity/privacy
    to the data.

    **Extends:**
    
        *pyGlobus.io.IOBase.IOBase*
    """
    def __init__(self, handle = None):
        """
        Constructs an instance.

        Each instance wraps a globus_io_handle_t object. An instance of this class
        may only have a single open connection at a time. To have more then one open
        connection, use multiple instances.
        
        **Arguments:**
        
            - *handle* An optional argument that should contain a SWIG'ized pointer to
            a globus_io_handle_t object.
            
        **Raises**        
        Raises a IOBaseException if unable to initialize the globus modules.
        """

        IOBase.__init__(self, handle)
        self.refcount = 1
        return
    
    def __del__(self):
        """
        Destroys an instance.

        **Raises**        
        Raises a IOBaseException if unable to deactivate the globus modules.
        """

        IOBase.__del__(self)
        return
        
    def register_connect(self, host, port, attr, callback, arg):
        """
        Asynchronous TCP connection establishment.

        
        Connect a TCP socket on the specified host/port pair. The connection will be
        started by this function, and a callback will be invoked when the connection is
        established.

        **Arguments:**
        
            - *host* A string containing the host to connect to.

            - *port* An int containing the port to connect to.

            - *attr* A TCPIOAttr object with the appropriate attributes set.

            - *callback* Function to be called when the security handshake is
            complete. The function has the following prototype:
            
            'function(arg, handle, result)'

            arg is the user argument passed when setting the callback,
            handle is a SWIG'ized pointer to a globus_io_handle_t, and
            result is a SWIG'ized pointer to a globus_result_t object.

            - *arg* A user argument to be returned when the callback is executed.

        **Raises:**
        
            Raises a GSITCPSocketException if the handle is currently in use,
            or an error occurs while making the connection.

        **Returns:**
        
            A SWIG'ized pointer to a callback handle. This handle can be free'd
            with a call to free_callback, or it will be free'd when the instance
            is destroyed.
        """

        if not self._handle == None:
            ex = GSITCPSocketException("Handle is already active")
            raise ex
        ret, cbHandle, handle = ioc.tcp_register_connect(host, port,
                                                        attr.get_handle(),
                                                        callback, arg)
        if not ret == 0:
            ex = GSITCPSocketException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        self._handle = handle
        return cbHandle

    def connect(self, host, port, attr):
        """
        Blocking TCP connection establishment.

        Connect a TCP socket on the specified host/port pair.

        **Arguments:**
        
            - *host* A string containing the host to connect to.

            - *port* An int containing the port to connect to.

            - *attr* A TCPIOAttr object with the appropriate attributes set.

        **Raises:**
        
            A GSITCPSocketException is thrown if the handle is currently in use,
            or an error occurs while making the connection.
        """

        if not self._handle == None:
            ex = GSITCPException("Handle is already active") 
            raise ex
        
        ret, handle = ioc.tcp_connect(host, port, attr.get_handle())
        if not ret == 0:
            ex = GSITCPSocketException(handle)
            raise ex
        self._handle = handle
        return

    def create_listener(self, attr, port=0, backlog=-1):
        """
        Create a TCP server socket.

        Creates a socket handle capable of accepting new TCP connections
        from other hosts or processes. In order to actually listen for
        connections, you must call either listen or register listen on this
        instance.

        **Arguments**

            - *attr* A TCPIOAttr containing the attributes for this server.

            - *port* The TCP port that the socket will listen for connections on. If the
            port number is 0, then an arbitrary TCP port will be selected. If
            this is true, and the restrict_port attribute is set to TRUE (the
            default) and the GLOBUS_TCP_PORT_RANGE environment variable
            was set when Globus I/O was initialized, then the port will be
            selected from that range. Otherwise, any port number may be
            chosen.

            - *backlog* The backlog parameter indicates he maximum length of the
            system's queue of pending connections. Any connection attempts
            when the queue is full will fail. If backlog is equal to -1, then the
            system-specific maximum queue length will be used. 

        **Raises:**
        
            A GSITCPSocketException is thrown if the handle is currently in use,
            or an error occurs while making the connection.

        **Returns:**

            An int containing the port the server will listen on.
        """

        if not self._handle == None:
            ex = GSITCPSocketException("Handle is already active")
            raise ex
        ret, port, handle = ioc.tcp_create_listener(port, backlog, 
                                                    attr.get_handle())
        if not ret == 0:
            ex = GSITCPSocketException(port)
            raise ex
        self._handle = handle
        return port

    def register_accept(self, attr, callback, arg):
        """
        Asynchronous server-side TCP connection establishment.

        Once the connection has been accepted, the callback function will be called,
        with the arg and a newly created, connected, handle argument passed to the
        callback.

        **Arguments:**

            - *attr* A TCPIOAttr object containing the attributes for the new eonnection.

            - *callback*  Function to be called when the security handshake is
            complete. The function has the following prototype:
            
            'function(arg, handle, result)'

            arg is the user argument passed when setting the callback,
            handle is a SWIG'ized pointer to a globus_io_handle_t, and
            result is a SWIG'ized pointer to a globus_result_t object.

            - *arg* A user argument to be returned when the callback is executed.

        **Raises:**

            A GSITCPSocketException is thrown if a listener was not already created, or
            an error occures while doing the underlying register_accept.

        **Returns**
        
            A tuple containing:

                - A new instance of GSITCPSocket for the new connetion, this must
                not be used until the callback occurs.

                - A callback handle. This can either be free'd with a call
                to free_callback, or all outstanding handles will be free'd when
                the instance is destroyed.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to accept before creating a listener")
            raise ex
        ret, cbHandle, childHandle = ioc.tcp_register_accept(self._handle,
                                                            attr.get_handle(),
                                                            callback, arg)
        if not ret == 0:
            ex = GSITCPSocketException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return GSITCPSocket(childHandle), cbHandle
    
    def accept(self, attr):
        """
        Blocking server-side TCP connection establishment.

        **Arguments:**

            - *attr* A TCPIOAttr object containing the attributes for the new eonnection.

        **Raises:**

            A GSITCPSocketException is thrown if a listener was not already created, or
            an error occures while doing the underlying accept.

        **Returns:**

            A new GSITCPSocket object for the new connection.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to accept before creating a listener")
            raise ex
        ret, childHandle = ioc.tcp_accept(self._handle, attr.get_handle())
        if not ret == 0:
            ex = GSITCPSocketException(childHandle)
            raise ex
        soc = GSITCPSocket(childHandle)
        ret, addr, port = ioc.tcp_get_remote_address(childHandle)
        return soc, (addr, port)

    def register_listen(self, callback, arg):
        """
        Asynchronous wait until a client connection is pending.

        The method will issue a callback when a connection is pending.

        **Arguments:**

            - *callback* The function to be called when a new connection
            is pending. The function has the following prototype:
            
            'function(arg, handle, result)'

            arg is the user argument passed when setting the callback,
            handle is a SWIG'ized pointer to a globus_io_handle_t, and
            result is a SWIG'ized pointer to a globus_result_t object.

            - *arg* A user argument to be returned when the callback is executed.

        **Raises:**

            A GSITCPSocketException is thrown if a listener was not already created, or
            an error occures while doing the underlying register_listen.

        **Returns:**

            A callback handle. The callback handle can either be free'd with a call
            to free_callback, or all outstanding handles will be free'd when the instance
            is destroyed.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to listen before creating a listener")
            raise ex
        ret, cbHandle = ioc.tcp_register_listen(self._handle, callback, arg)
        if not ret == 0:
            ex = GSITCPSocketException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def listen(self):
        """
        Block until a client connection is pending.

        This method will block until a connection is pending. Once this method
        returns, accept may be called to return a new GSITCPSocket object.

        **Raises:**

            A GSITCPSocketException is thrown if a listener was not already created, or
            an error occures while doing the underlying listen.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to listen before creating a listener")
            raise ex
        ret, err = ioc.tcp_listen(self._handle)
        if not ret == 0:
            ex = GSITCPSocketException(err)
            raise ex
        return

    def get_local_address(self):
        """
        Return the local IP and port.

        **Raises:**
        
        A GSITCPSocketException is thrown if a listener was not already created.
        
        **Returns:**
        
        A tuple containing:

        - The IP address of the local socket.
        
        - The port number.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to get address from uninitialized handle")
            raise ex
        ret, host, port = ioc.tcp_get_local_address(self._handle)
        if not ret == 0:
            ex = GSITCPSocketException(host)
            raise ex
        return (host, port)

    def get_remote_address(self):
        """
        Return the remote IP and port.

        **Raises:**

        A GSITCPSocketException is thrown if a listener was not already created.
        
        **Returns:**

        A tuple containing:
        
        - The IP address of the remote socket.

        - The port number.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to get address from uninitialized handle")
            raise ex
        ret, host, port = ioc.tcp_get_remote_address(self._handle)
        if not ret == 0:
            ex = GSITCPSocketException(host)
            raise ex
        return (host, port)

    def get_security_context(self):
        """
        Return the GSS security context.

        **Raises:**

            A GSITCPSocketException is thrown if a listener was not already created.

        **Returns:**

            A GSSContext object.
        """

        if self._handle == None:
            ex = GSITCPSocketException("Unable to get context from uninitialized handle")
            raise ex
        ret, context = ioc.tcp_get_security_context(self._handle)
        if not ret == 0:
            ex = GSITCPSocketException(context)
            raise ex
        return GSSContext(context, 0)

    #Experimental GSS extensions
    def init_delegation(self, cred, oids, buffers, time_req):
        """
        Delegate a credential to a remote entity.
        cred is the credential to delegate
        oids is the set of extension oids to add to the credential
        buffers is the set of extension values (must correspond to oids)
        time_req is the requested credential lifetime, in seconds.
        """
        
        if self._handle == None:
            ex = GSITCPSocketException("unable to init delegation from unititialized handle")
            raise ex
        if cred == None:
            cred_handle = gssc.get_no_cred_const_ptr()
        else:
            cred_handle = cred.get_handle()
        gssc.init_delegation(self.get_security_context().get_handle(),
                            cred_handle, oids, buffers, time_req, self._handle)
        return

    def accept_delegation(self, oids, buffers, time_req):
        """
        Accept a credential from a remote entity 
        
        oids is the set of extension oids to add to the credential
        buffers is the set of extension values (must correspond to oids)
        time_req is the requested credential lifetime, in seconds.

        **Returns**
        Returns the GSSCred associated with the given parameters.
        """

        if self._handle == None:
            ex = GSITCPSocketException("unable to accept delegation from unititialized handle")
            raise ex
        cred = gssc.accept_delegation(self.get_security_context().get_handle(),
                                     oids, buffers, time_req, self._handle)
        return GSSCred(cred, 0)
    
    def get_delegated_credential(self):
        """
        Return the delegated credential.
	
        **Raises:**	
	A GSITCPSocketException is thrown if unable to return the credential.

	**Returns:**
	A GSSCred object.
	"""

        if self._handle == None:
            ex = GSITCPSocketException("Unable to get credential from uninitialized handle")
            raise ex
        ret, cred = ioc.tcp_get_delegated_credential(self._handle)
        if not ret == 0:
            ex = GSITCPSocketException(cred)
            raise ex
        return GSSCred(cred, 0)
        
    def makefile(self, mode="r", bufsize=0):
        """
        Create a GSIFile object from the socket.  Note:  unlike socket.makefile,
        this routine does NOT dup the socket.
        """
        return GSIFile(self, mode)

    def shutdown(self, how):
        """
        This should emulate socket.shutdown -- currently, this routine does
        nothing (it's simply here as a placeholder).
        """
        # should really do something here ...
        pass
    
    def sendall(self, str):
        """
        Emulates socket.send -- writes a string out over the socket.
        """

        l = len(str)
        num = self.write(str, l)
        while num < l:
            l = l - num
            num = self.write(str[num:], l)
        return

    def send(self, str):
        """
        Emulates socket.send -- writes a string out over the socket.
        """
        return self.write(str, len(str))

    def close(self):
        """
        Close this socket, if there are no open GSIFile objects using it.
        """        

        self.refcount = self.refcount - 1
        if self.refcount == 0:
            IOBase.close(self)

            
class GSITCPSocketServer(GSITCPServer):
    """
    Derives from the base GSITCPServer class.

    Allows applications using the default SocketServer class to use the
    GSI context supplied by Globus instead.
    If the application needs to connect both secure and non-secure
    connections, a seperate instance of GSITCPSocketServer can be created
    and run alongside the unsecure SocketServer.
    Note: Be sure to use different ports for both!

    """
    def __init__(self, addr, RequestHandlerClass,
                channel_mode=ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP,
                delegation_mode=ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY,
                tcpAttr=None):
        GSITCPServer.__init__(self, addr, RequestHandlerClass,
                                         channel_mode, delegation_mode, tcpAttr)
        self.socket.allow_reuse_address = 1
  

class GSIRequestHandler(SocketServer.StreamRequestHandler):
    def handle(self): 
        cred = self.server.get_delegated_credential()
        self.server.delegated_cred = cred
        result = self.rfile.readline()
        while result:
            result = self.rfile.readline()
        return

class ThreadingGSITCPSocketServer(SocketServer.ThreadingMixIn, GSITCPSocketServer):
    """
    Threaded GSITCPSocketServer
    """
    pass

class ForkingGSITCPSocketServer(SocketServer.ForkingMixIn, GSITCPSocketServer):
    """
    Forking GSITCPSocketServer
    """
    pass

class GSIHTTPConnection(httplib.HTTPConnection):
    "This class allows communication via SSL."

    def __init__(self, host, port=None, tcpAttr=None,
                 channel_mode=ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP,
                 delegation_mode=ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY):
        """
        **Arguments:**
        
        - *host* A python string of the host name to connect to

        - *tcpAttr* A TCPIOAttr object which determines the properties of your connection

        - *channel_mode* The data will be transported in the clear, encrypted using GSI or
        encrypted using SSL

        - *delegation_mode* if channel_mode is GSI then this determines the type of delegation
        to do. Delegation can be NONE, LIMITED, FULL
        
        """
        
        httplib.HTTPConnection.__init__(self, host, port=None)

        if tcpAttr is None:
            self.io_attr = TCPIOAttr()
            self.io_attr.set_authentication_mode(ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
            authdata = AuthData()
            self.authdata = authdata
            authdata.set_callback(client_auth_callback, None)
            self.io_attr.set_authorization_mode(ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK, authdata)
            self.io_attr.set_delegation_mode(delegation_mode)
            self.io_attr.set_nodelay(0)
            self.io_attr.set_channel_mode(channel_mode)
        else:
            self.io_attr = tcpAttr
            
        self.response_class = httplib.HTTPResponse
        # hack: port is not being copied to self.port by HTTPConnection.__init__
        if port:
            self.port = port
        if self.debuglevel > 0: 
            print "GSIHTTPConnection host=%s  port=%d  shost=%s  sport=%d" %(host, port, self.host, self.port)

    def connect(self):
        """
        Open the GSITCPSocket connection.
        """
        self.sock = GSITCPSocket()
        self.sock.connect(self.host, self.port, self.io_attr)
        if self.debuglevel > 0:
            print "connect: (%s, %s)" % (self.host, self.port)
        
    def close(self):
        """
        Close the given connection
        """
        self.sock.close()
        if self.debuglevel > 0:
            print "close: (%s, %s)" % (self.host, self.port)

class GSIHTTP(httplib.HTTP):
    def __init__(self, host='', port=None, tcpAttr=None,
                 channel_mode=ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP,
                 delegation_mode=ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY):
        """
        **Arguments:**
        
        - *host* A python string of the host name to connect to

        - *tcpAttr* A TCPIOAttr object which determines the properties of your connection

        - *channel_mode* The data will be transported in the clear, encrypted using GSI or
        encrypted using SSL

        - *delegation_mode* if channel_mode is GSI then this determines the type of delegation
        to do. Delegation can be NONE, LIMITED, FULL
        
        """
        
        self._connection_class = lambda host, port, tcpAttr = tcpAttr:  \
              GSIHTTPConnection(host, port, tcpAttr=tcpAttr,
                                channel_mode=channel_mode,
                                delegation_mode=delegation_mode) 
        httplib.HTTP.__init__(self, host, port)

class NetIOAttr:

    def __init__(self):
        self._activated = 0
        self._handle = None
        self._authData = None
        ret = utilc.globus_module_activate(utilc.get_module())
        if not ret == 0:
            ex = NetIOAttrException("Unable to initialize the common module")
            raise ex
        ret = utilc.globus_module_activate(ioc.get_module())
        if not ret == 0:
            ex = NetIOAttrException("Unable to initialize the io module")
            raise ex
        self._activated = 1
        return

    def __del__(self):
        ret = utilc.globus_module_deactivate(ioc.get_module())
        if not ret == 0:
            ex = NetIOAttrException("Unable to deactivate the io module")
            raise ex
        ret = utilc.globus_module_deactivate(utilc.get_module())
        if not ret == 0:
            ex = NetIOAttrException("Unable to deactivate the common module")
            raise ex

    def set_reuseaddr(self, reuse):
        """
        Set the NetIOAttr's reuseaddr
        """
        ret, err = ioc.attr_set_socket_reuseaddr(self._handle, reuse)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_reuseaddr(self):
        """
        Accessor for the NetIOAttr's reuseaddr
        """
        ret, reuse = ioc.attr_get_socket_reuseaddr(self._handle)
        if not ret == 0:
            ex = NetIOAttrException(reuse)
            raise ex
        return reuse

    def set_keepalive(self, keepAlive):
        """
        Set the NetIOAttr's keepAlive
        """
        ret, err = ioc.attr_set_socket_keepalive(self._handle, keepAlive)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_keepalive(self):
        """
        Get the NetIOAttr's keepAlive
        """
        ret, keepAlive = ioc.attr_get_socket_keepalive(self._handle)
        if not ret == 0:
            ex = NetIOAttrException(keepAlive)
            raise ex
        return keepAlive

    def set_linger(self, linger, lingerTime):
        """
        Set the linger attribute in the specified attribute set.
 
        This attribute is used to determine what to do when data is in 
        the socket's buffer when the socket is closed. If linger is
        set to true, then the close operation will block until the socket
        buffers are empty, or the linger_time has expired.

        **Arguments**
        
        - *linger* The value of the linger attribute.
        - *lingerTime* The time (in seconds) to block at close time
        if linger is true and data is queued in the socket buffer.

        **Returns**
        None        
        """
        ret, err = ioc.attr_set_socket_linger(self._handle, linger, lingerTime)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_linger(self):
        """
        Query the linger attribute in the specified attribute set.
 
        This attribute is used to determine what to do when data is in 
        the socket's buffer when the socket is closed. If linger is
        set to true, then the close operation will block until the socket
        buffers are empty, or the linger_time has expired.
        
        **Retuns**
        Returns the value of the linger attribute, and the time (in seconds)
        to block at close time if linger is true and data is queed in the
        socket buffer.
        """
        ret, linger, lingerTime = ioc.attr_get_socket_linger(self._handle)
        if not ret == 0:
            ex = NetIOAttrException(linger)
            raise ex
        return linger, lingerTime

    def set_oobinline(self, inline):
        """
        Set the oobinline attribute in the specified attribute set.
        
        This attribute is used to choose whether out-of-band data is
        received in the normal data queue, or must be received specially
        by calling globus_io_recv() or globus_io_register_recv() with the
        MSG_OOB flag set to true.
        
        - *inline* The value of the oobinline attribute.

        **Returns**
        None
        """
        ret, err = ioc.attr_set_socket_oobinline(self._handle, inline)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_oobinline(self):
        """
        Query the oobinline attribute in the specified attribute set.
        
        This attribute is used to choose whether out-of-band data is
        received in the normal data queue, or must be received specially
        by calling globus_io_recv() or globus_io_register_recv() with the
        MSG_OOB flag set to true.

        **Returns**
        The value of the oobinline attribute.
        """
        
        ret, oobinline = ioc.attr_get_socket_oobinline(self._handle)
        if not list[0] == 0:
            ex = NetIOAttrException(oobinline)
            raise ex
        return oobinline

    def set_sndbuf(self, sndbuf):
        """
        Set  the sndbuf attribute in the specified attribute set.

        This attribute is used to choose the size of the send buffer
        for the socket handle this attribute is applied to. The send buffer
        is often used by the operating system to choose the appropriate
        TCP window size.

        **Arguments**
        - *sndbuf* The value of the send buffer in bytes.

        **Returns**
        None
        """
        
        ret, err = ioc.attr_set_socket_sndbuf(self._handle, sndbuf)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_sndbuf(self):
        """
        Query the sndbuf attribute in the specified attribute set.

        This attribute is used to choose the size of the send buffer
        for the socket handle this attribute is applied to. The send buffer
        is often used by the operating system to choose the appropriate
        TCP window size.

        **Returns**
        The value of the send buffer in bytes.
        """
        
        ret, sendBuf = ioc.attr_get_socket_sndbuf(self._handle)
        if not ret == 0:
            ex = NetIOAttrException(sendBuf)
            raise ex
        return sendBuf

    def set_rcvbuf(self, rcvbuf):
        """
        Set the rcvbuf attribute in the specified attribute set.
        
        This attribute is used to choose the size of the receive buffer for
        the socket handle this attribute is applied to. The receive buffer
        is often used by the operating system to choose the appropriate TCP
        window size.

        **Arguments**
        - *rcvbuf* The value of the receive buffer in bytes.
        **Returns**
        None
        """
        ret, err = ioc.attr_set_socket_rcvbuf(self._handle, rcvbuf)
        if not ret == 0:
            ex = NetIOAttrException(err)
            raise ex
        return

    def get_rcvbuf(self):
        """
        Query the rcvbuf attribute in the specified attribute set.
        
        This attribute is used to choose the size of the receive buffer for
        the socket handle this attribute is applied to. The receive buffer
        is often used by the operating system to choose the appropriate TCP
        window size.

        **Returns**
        The value of the receive buffer in bytes.
        """        
        ret, recvBuf = ioc.attr_get_socket_rcvbuf(self._handle)
        if not ret == 0:
            ex = NetIOAttrException(recvBuf)
            raise ex
        return recvBuf

    def get_handle(self):
        """
        ########################################################################
        # XXX This handle is only valid as long as this object exists.  When
        # the object is destroyed, the memory this handle points to is freed.
        #######################################################################
        """
        return self._handle


    
class TCPIOAttr(NetIOAttr):

    def __init__(self, handle = None):
        NetIOAttr.__init__(self)
        if handle == None:
            ret, handle = ioc.tcpattr_init()
            if not ret == 0:
                ex = TCPIOAttrException("Unable to initialize the attr")
                raise ex
        self._handle = handle

    def __del__(self):
        ret, err = ioc.tcpattr_destroy(self._handle)
        self._handle = None
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        NetIOAttr.__del__(self)
        
    def set_restrict_port(self, restrict):
        """        
        Set the restrict-port attribute in the specified TCP attribute set.

        This attribute is used to determine whether or not to use the 
        GLOBUS_TCP_PORT_RANGE environment variable to choose an anonymous
        port for a TCP listener handle. This attribute may not be applied to
        an existing TCP handle.

        **Arguments**
        - *restrict*  The new value of the restrict_port attribute.
        **Returns**
        None
        """
        ret, err = ioc.attr_set_tcp_restrict_port(self._handle, restrict)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_restrict_port(self):
        """
        Query the restrict-port attribute in the specified TCP attribute set.

        This attribute is used to determine whether or not to use the 
        GLOBUS_TCP_PORT_RANGE environment variable to choose an anonymous
        port for a TCP listener handle. This attribute may not be applied to
        an existing TCP handle.

        **Returns**
        The new value of the restrict_port attribute.
        """
        
        ret, restrictPort = ioc.attr_get_tcp_restrict_port(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(restrictPort)
            raise ex
        return restrictPort

    def set_nodelay(self, nodelay):
        """
        Set the nodelay attribute in the specified TCP attribute set.

        This attribute is used to determine whether or not to disable
        Nagle's algorithm.  

        **Arguments**
        - *nodelay* The new value of the nodelay attribute.
        **Returns**
        None
        """
        ret, err = ioc.attr_set_tcp_nodelay(self._handle, nodelay)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_nodelay(self):
        """
        Query the nodelay attribute in the specified TCP attribute set.

        This attribute is used to determine whether or not to disable
        Nagle's algorithm.  

        **Returns**
        The new value of the nodelay attribute. 
        """
        ret, noDelay = ioc.attr_get_tcp_nodelay(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(noDelay)
            raise ex
        return noDelay

    def set_interface(self, interface):
        """
        Set the tcp interface attribute in the specified TCP attribute set.

        This attribute is used to determine whether to bind TCP sockets to a
        particular interface. This attribute must be set before calling
        globus_io_tcp_create_listener() or globus_io_tcp_connect().

        **Arguments**
        - *interface* The value of the interface attribute.
        The interface string must be in dotted-ip format (ie "127.0.0.1").
        **Returns**
        None
        """        
        ret, err = ioc.attr_set_tcp_nodelay(self._handle, interface)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_interface(self):
        """
        Query the tcp interface attribute in the specified TCP attribute set.

        This attribute is used to determine whether to bind TCP sockets to a
        particular interface. This attribute must be set before calling
        globus_io_tcp_create_listener() or globus_io_tcp_connect().

        **Returns**
        The value of the interface attribute, a string in 
        dotted-ip format (ie "127.0.0.1").
        """                
        ret, interFace = ioc.attr_get_tcp_interface(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(interFace)
            raise ex
        return interFace

    def set_authentication_mode(self, mode, cred = None):
        """
        Set the authentication mode attribute in the specified socket
        attribute set.

        This attribute is used to determine whether or not to call the
        GSSAPI security context establishment functions once a socket
        connection is established.
        
        If the authentication_mode value is
        GLOBUS_IO_SECURE_AUTHENTICATION_MODE_NONE, then the channel mode,
        delegation mode, protection mode, and authorization mode will all be reset
        to disable all security on the socket attribute set.

        **Arguments**
        
        - *mode* The value of the authentication-mode attribute. The values for
        mode are described in the documentation for the globus_io_secure_authentication_mode_t
        type.

        - *cred* A GSSAPI credential to be used when authenticating. If the credential
        is equal to GSS_C_NO_CREDENTIAL, then Globus I/O will use the process's default
        credentials.

        **Returns**
        None
        """
        createCred = False
        if cred == None:
            cred = gssc.get_no_cred_const_ptr()
            createCred = True

        ret, err = ioc.attr_set_secure_authentication_mode(self._handle,
                                                          mode, cred)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        if createCred:
            gssc.free_cred_ptr(cred)
            
        return

    def get_authentication_mode(self):
        """
        Query the authentication mode attribute in the specified socket
        attribute set.

        This attribute is used to determine whether or not to call the
        GSSAPI security context establishment functions once a socket
        connection is established.
        
        If the authentication_mode value is
        GLOBUS_IO_SECURE_AUTHENTICATION_MODE_NONE, then the channel mode,
        delegation mode, protection mode, and authorization mode will all be reset
        to disable all security on the socket attribute set.

        **Returns**
        The authentication mode and associated GSSAPI credential.
        """
        ret, mode, cred = ioc.attr_get_secure_authentication_mode(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(mode)
            raise ex
        return mode, cred

    def set_authorization_mode(self, mode, authData):
        """
        Set the authorization-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine what security identities to
        authorize as the peer to the security handshake done when making
        an authenticated connection.

        **Arguments**
        - *mode*  The new value of the authorization-mode attribute.
        The values for mode are described in the documentation for the
        globus_io_secure_authorization_mode_t type.

        - *authData* A globus_io_secure_authorization_data_t structure
        containing any additional information necessary to perform the
        specified type of authorization.

        **Returns**
        None
        """
        ret, err = ioc.attr_set_secure_authorization_mode(self._handle,
                                                         mode,
                                                         authData.get_handle())
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        self._authData = authData
        return

    def get_authorization_mode(self):
        """
        Query the authorization-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine what security identities to
        authorize as the peer to the security handshake done when making
        an authenticated connection.

        **Returns**
        The value of the authorization-mode attribute.
        The values for mode are described in the documentation for the
        globus_io_secure_authorization_mode_t type.
        """
        if self._authData is None:
            ex = TCPIOAttrException("Auth Mode is not set")
            raise ex
        ret, mode = ioc.attr_get_secure_authorization_mode(self._handle,
                                                    self._authData.get_handle())
        if not ret == 0:
            ex = TCPIOAttrException(mode)
            raise ex
        return mode

    def set_channel_mode(self, mode):
        """
        Set the channel-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine if any data wrapping should
        be done on the socket connection. This is required to use the
        data protection attribute.

        **Arguments**
        - *mode* The new value of the channel-mode attribute. The 
        values for mode are described in the documentation for the
        globus_io_secure_channel_mode_t type.

        **Returns**
        None
        """
        ret, err = ioc.attr_set_secure_channel_mode(self._handle, mode)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_channel_mode(self):
        """
        Set the channel-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine if any data wrapping should
        be done on the socket connection. This is required to use the
        data protection attribute.

        **Returns**
        The value of the channel-mode attribute. The 
        values for mode are described in the documentation for the
        globus_io_secure_channel_mode_t type.        
        """
        ret, mode = ioc.attr_get_secure_channel_mode(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(mode)
            raise ex
        return mode

    def set_protection_mode(self, mode):
        """
        Set the protection-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine if any data protection should
        be done on the socket connection.

        **Arguments**
        - *mode* The new value of the protection mode attribute. The 
        values for mode are described in the documentation for the
        globus_io_secure_protection_mode_t type.

        **Returns**
        None
        """
        ret, err = ioc.attr_set_secure_protection_mode(self._handle, mode)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_protection_mode(self):
        """
        Query the protection-mode attribute in the specified socket
        attribute set.

        This attribute is used to determine if any data protection should
        be done on the socket connection.

        **Returns**
        The value of the protection mode attribute. The 
        values for mode are described in the documentation for the
        globus_io_secure_protection_mode_t type.
        """
        ret, mode = ioc.attr_get_secure_protection_mode(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(mode)
            raise ex
        return mode

    def set_delegation_mode(self, mode):
        """
        Set the of the delegation-mode attribute in the specified
        socket attribute set.

        This attribute is used to determine whether the process's
        credentials should be delegated to the other side of the connection.

        **Arguments**
        - *mode* The new value of the delegation-mode attribute. The
        values for mode are described in the documentation for the
        globus_io_secure_delegation_mode_t type.

        **Returns**
        None
        """
        ret, err = ioc.attr_set_secure_delegation_mode(self._handle, mode)
        if not ret == 0:
            ex = TCPIOAttrException(err)
            raise ex
        return

    def get_delegation_mode(self):
        """
        Set the of the delegation-mode attribute in the specified
        socket attribute set.

        This attribute is used to determine whether the process's
        credentials should be delegated to the other side of the connection.

        **Returns**
        The value of the delegation-mode attribute. The
        values for mode are described in the documentation for the
        globus_io_secure_delegation_mode_t type.
        """
        ret, mode = ioc.attr_get_secure_delegation_mode(self._handle)
        if not ret == 0:
            ex = TCPIOAttrException(mode)
            raise ex
        return mode



