############################################################################
# Matt Rodriguez, AccessGrid Group Contributions, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
This module has a factory to create many flavors of io tcp attributes
"""

from pyGlobus.util import PyGlobusFactory
from pyGlobus.io import TCPIOAttr, AuthData
from pyGlobus import ioc

def createTCPAttrDefault():
    """
    Create a TCP attr with common defaults.
    This does not set authorization or authentication modes.
    """

    attr = TCPIOAttr()

    attr.set_authentication_mode(ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
    attr.set_channel_mode(ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP)
    attr.set_delegation_mode(ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_NONE)
    
    attr.set_restrict_port(0)
    attr.set_reuseaddr(1)
    attr.set_nodelay(1)

    return attr

def createTCPAttrDefaultSSL():
    """
    Create a TCP attr with common defaults for SSL compatibility.
    This does not set authorization or authentication modes.
    """

    attr = TCPIOAttr()

    attr.set_authentication_mode(ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
    attr.set_channel_mode(ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP)
    attr.set_delegation_mode(ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY)
    
    attr.set_restrict_port(0)
    attr.set_reuseaddr(1)
    attr.set_nodelay(1)

    return attr

def createTCPAttrSelfAuth():
    """
    Create a TCP attr to allow only connections from self.
    http://www-unix.globus.org/api/c-globus-2.0-beta1/globus_io/html/group__security.html#a7
    """

    attr = createTCPAttrDefault()
    authdata = AuthData()
    authmode = ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_SELF

    attr.set_authorization_mode(authmode, authdata)

    return attr
    

def createTCPAttrCallbackAuth(auth_cb):
    """
    Create a TCP attr to invoke auth_cb for authorization decisions.
    http://www-unix.globus.org/api/c-globus-2.0-beta1/globus_io/html/group__security.html#a7

    @param auth_cb: The authorization callback given to the TCPAttr.      
    """

    attr = createTCPAttrDefault()

    authdata = AuthData()

    authdata.set_callback(auth_cb, None)
    authmode = ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK

    attr.set_authorization_mode(authmode, authdata)

    return attr

def createTCPAttrCallbackAuthSSL(auth_cb):
    """
    Create a TCP attr to invoke auth_cb for authorization decisions.
    http://www-unix.globus.org/api/c-globus-2.0-beta1/globus_io/html/group__security.html#a7

    @param auth_cb: The authorization callback given to the TCPAttr.      
    """

    attr = createTCPAttrDefaultSSL()

    authdata = AuthData()

    authdata.set_callback(auth_cb, None)
    authmode = ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK

    attr.set_authorization_mode(authmode, authdata)

    return attr

def createTCPAttrAlwaysAuth():
    """
    Create a TCP attr that allows all connections.
    """

    def always_allow_cb(server, g_handle, remote_user, context):
        return 1

    attr = createTCPAttrCallbackAuth(always_allow_cb)

    return attr
    
def createTCPAttrAlwaysAuthSSL():
    """
    Create a TCP attr that allows all connections.
    """

    def always_allow_cb(server, g_handle, remote_user, context):
        return 1

    attr = createTCPAttrCallbackAuthSSL(always_allow_cb)

    return attr

class TCPIOAttrFactory(PyGlobusFactory):
    """
    Creates a factory that will by default create a TCPIOAttr
    """
    def __init__(self, default=TCPIOAttr):
        self.register("createTCPIOAttr", default, handle=None)
        self.register("createTCPAttrDefault", createTCPAttrDefault)
        self.register("createTCPAttrSelfAuth", createTCPAttrSelfAuth)
        self.register("createTCPAttrCallbackAuth", createTCPAttrCallbackAuth)
        self.register("createTCPAttrCallbackAuthSSL", createTCPAttrCallbackAuthSSL)
        self.register("createTCPAttrAlwaysAuth", createTCPAttrAlwaysAuth)
        self.register("createTCPAttrAlwaysAuthSSL", createTCPAttrAlwaysAuthSSL)
