############################################################################
# David Konerding, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Logging support module.
"""

## Attempt a version string
__author__  = "David E. Konerding"
__status__  = "alpha"
__version__ = "$Revision: 1.9 $"
__rcsid__   = "$Id: logUtil.py,v 1.9 2006/03/26 23:03:43 mateo Exp $"

__all__ = ['logging', 'netlogger']
import sys, time
from threading import Lock

try:
    import logging
except:
    def time_fmt():
        t = time.time()
        return time.strftime("%Y-%m-%d %H:%M:%S" + "%.3f" % (t -int(t)) )

    ## a "fake" interface to the logging packing, which has a similar 
    ## functionality for versions of Python which don't have logging
    class FakeLogger:
        def __init__(self, name):
            self.name = name
            self.level = logging.ERROR
            self.handlers = []

        def addHandler(self, handler):
            self.handlers.append(handler)

        def setLevel(self, level):
            self.level = level

        def log(self, level, message):
            if level >= self.level:
                pass #print "%s FAKE_LOG: %s" % (time_fmt(), message)

        def debug(self, message):
            if logging.DEBUG >= self.level:
                pass #print "%s FAKE_DEBUG: %s" % (time_fmt(), message)

        def info(self, message):
            if logging.INFO >= self.level:
                pass #print "%s FAKE_INFO: %s" % (time_fmt(), message)

        def warning(self, message):
            if logging.WARNING >= self.level:
                pass #print "%s FAKE_WARNING: %s" % (time_fmt(), message)

        def error(self, message):
            if logging.ERROR >= self.level:
                pass #print "%s FAKE_ERROR: %s" % (time_fmt(), message)

        def critical(self, message):
            if logging.CRITICAL >= self.level:
                pass #print "%s FAKE_CRITICAL: %s" % (time_fmt(), message)

    ## A class to implement a "fake" logging module
    class loggingImplementation:
        DEBUG=10
        INFO=20
        WARNING=30
        ERROR=40
        CRITICAL=50

        class StreamHandler:
            def __init__(self, stream):
                self.stream = stream
                self.formatter = None

            def setFormatter(self, formatter):
                self.formatter = formatter

        class Formatter:
            def __init__(self, format_string):
                self.format_string = format_string

        def getLogger(self, name):
            pass #print "returning FakeLogger"
            return FakeLogger(name)

        def shutdown(self):
            pass


    logging = loggingImplementation()

## optional support for netlogger
try:
    import gov.lbl.dsd.netlogger as netlogger
except ImportError:

    class FakeNetlogger:
        ## Map between the common levels defined in the logging 
        ## package, and in the NetLogger package
        LevelMap = {
            3: logging.DEBUG,
            2: logging.INFO,
            1: logging.WARNING,
            0: logging.ERROR
            }
        def __init__(self, url, mode='w', flags=0, host=None, program=None, 
                    const={}, reconnect=-1, backup=None, backup_flags=0, 
                    threaded=0, timeout=1.0, stop_on_eof=1):
            self.url = url
            self.mode = mode
            self.flags = flags
            self.host = host
            self.program = program
            self.const = const
            self.reconnect = reconnect
            self.backup = backup
            self.backup_flags = backup_flags
            self.threaded = threaded
            self.timeout = timeout
            self.stop_on_eof = stop_on_eof
            pass #print "Created a FakeNetlogger"
            if self.threaded:
                self.lock = Lock()

            self.logger = logging.getLogger('pyglobus-url-copy')
            ## TODO: command-line selectable stream type and warning level
            logger_handler = logging.StreamHandler(sys.stdout)
            logger_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
            logger_handler.setFormatter(logger_formatter)
            self.logger.addHandler(logger_handler)
            self.level = 0

        def setLevel(self, level):
            self.level = level

        def write(self, level, title, message, *args):
            if level <= self.level:
                if self.threaded:
                    self.lock.acquire()
                ## this depends on netlogger substitution syntax being 
                ## like python substitution syntax
                x = message % args
                if FakeNetlogger.LevelMap.has_key(level):
                    logger_level = FakeNetlogger.LevelMap[level]
                else:
                    logger_level = logging.INFO
                self.logger.log(logger_level, "Fake NetLogger: %s: %s: %s" % (self.program, title, x))
                if self.threaded:
                    self.lock.release()

    class WriterImplementation:
        ERROR_LVL = 0
        FATAL_LVL = 0
        WARN_LVL = 1
        INFO_LVL = 2
        DEBUG_LVL = 3

    class OutputFlagsImplementation:
        ULM = 0
        FLUSH = 1
        BIN = 2
        XML = 4
        ULM = 8
        ANNOUNCE = 16


    class netloggerImplementation:
        Writer = WriterImplementation()
        OutputFlags = OutputFlagsImplementation()

        def __init__(self):
            pass

        def open(self, url, **kwargs):
            return FakeNetlogger(url, **kwargs)


    netlogger = netloggerImplementation()


    
class singleton:
    logger = logging.getLogger('singleton') 
    logger_handler = logging.StreamHandler(sys.stdout)
    logger_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    logger_handler.setFormatter(logger_formatter)
    logger.addHandler(logger_handler)

    def __init__(self): self.logger = logger

def error(str): singleton.logger.error(str)
def warn(str):  singleton.logger.warn(str)
def debug(str): singleton.logger.debug(str)

