############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides a simple interface to the myproxy module.
"""

from pyGlobus.util import GlobusException 
import myProxyc
import os
import utilc, gssc

DEFAULT_CERTFILE ="~/.globus/usercert.pem"
DEFAULT_OUTFILE = "/tmp/x509up_u"


class MyProxyException(GlobusException):
    """
       A simple subclass of GlobusException
       
       Used for errors in the myProxy module. It defines no new methods.
       
       **Extends:**
       
            *pyGlobus.util.GlobusException*
    """
    pass

class MyProxy:

    def __init__(self):
        """
	Initializes the memory for the underlying structures

	**Arguments**

	None

	**Returns**

	None

	**Raises**
	A MyProxyException is raised if there is a problem initializing
	the underlying structures
	""" 
        attrs, request, response = myProxyc.myproxy_init()
	
        ret, attrs, request = myProxyc.set_delegation(attrs, request)
        if not ret == 0:
            ex = MyProxyException(attrs)
            raise ex

        self._attrs = attrs
        self._request = request
        self._response = response
        self._username = os.getenv("LOGNAME")

    
    def __del__(self):
        """
        Frees the memory associated with underlying structures
        """ 
        myProxyc.myproxy_destroy(self._attrs, self._request, self._response)

    
    def create_proxy(self, proxyFile, hostname, port=7512, lifetime=12*3600): 
        """
	Create a proxy and send it to the myproxy server

	**Arguments**

	- *hostname* a string of the hostname of the myproxy server

	- *port* an int of the port the myproxy server is listening on

	- *lifetime* an int which is the number of seconds the credential is valid 
	**Returns**

	None
	
	**Raises**
	A MyProxyException is raised if there is a problem creating the proxy
	""" 

        ret = myProxyc.proxy_init(self._attrs, self._request,
				  self._response,hostname, port,
				  lifetime,proxyFile,
				  self._username)
        if not ret == 0:
            ex = MyProxyException("Problem initializing proxy")
            raise ex

    def set_passphrase(self, passphrase):
        """
        Sets the passphrase. 
        
        **Arguments**

        - *passphrase* A string of the passphrase that is set into
        the underlying myproxy request. The passphrase is not 
        encrypted. 

        **Raises**
        A MyProxyException is raised if the passphrase is too long
        or is not a Python String object 
        """
        ret = myProxyc.set_passphrase(self._request, passphrase)
        if not ret == 0:
            ex = MyProxyException("Problem setting the password")
            raise ex
    
    
    def set_username(self, username):
        """
        Sets the username 
        
        **Arguments**
        
        - *username* a string of the username for the delegated proxy,
        defaults to $LOGNAME 
        """
        self._username = username 
       
    
    def get_info(self, hostname, port=7512, dnAsUsername=None):
        """
	Gets information about proxy from the myproxy repository

	**Arguments**

	- *hostname* a string which is the hostname of the repository
	
	- *port* an int which is the port the repository is listening on

	- *dnAsUsername* The DN of the proxy whom you want to query the server
	about
	
	**Returns**
	A dictionary of the proxy information

	**Raises**
	A MyProxyException is raised if there is a problem getting the credential
	"""
        utilc.globus_module_activate(gssc.get_module())
        if not dnAsUsername:
            useDnAsUsername = False 
            dnAsUsername = self._username 
        else:
            useDnAsUsername = True 
    
        username = dnAsUsername    
        ret, proxyDict = myProxyc.proxy_info(self._attrs, self._request,
                                             self._response,hostname,port,
                                             username, useDnAsUsername)
        if not ret == 0:
            ex = MyProxyException(proxyDict)
            utilc.globus_module_deactivate(gssc.get_module())
            raise ex

	
        return proxyDict
    
    def destroy_proxy(self, hostname, port=7512, dnAsUsername=None):
        """
	Destroys about proxy from the myproxy repository

	**Arguments**

	- *hostname* a string which is the hostname of the repository
	
	- *port* an int which is the port the repository is listening on

	- *dnAsUsername* The DN of the proxy whom you want to query the server
	about
	
	**Returns**
	None

	**Raises**
	A MyProxyException is raised if there is a problem getting the credential
	"""
        utilc.globus_module_activate(gssc.get_module())
        if not dnAsUsername:
            useDnAsUsername = False 
            dnAsUsername = self._username
        else:
            useDnAsUsername = True 
    
        username = dnAsUsername    
        ret, errorString = myProxyc.proxy_destroy(self._attrs, self._request,
				                  self._response,hostname,port,
				                  username, useDnAsUsername)
        if not ret == 0:
            ex = MyProxyException(errorString)
            utilc.globus_module_deactivate(gssc.get_module())
            raise ex
        utilc.globus_module_deactivate(gssc.get_module())
	
    def get_delegation(self, hostname, port=7512, certfile=None,
		       outFile=None, lifetime=12*3600, interactive=True):
        """
	Gets the credential from the myproxy repository

	**Arguments**

	- *hostname* a string which is the hostname of the repository

	- *outFile* a string which is the file where the credential will be written to
	
	- *port* an int which is the port the repository is listening on

	- *certfile* the location of the certfile

        - *lifetime* an int which is the number of seconds the credential is valid 
        - *interactive* a bool which determines whether the password will be 
        read in interactively.

	**Returns**
	None

	**Raises**
	A MyProxyException is raised if there is a problem getting the credential
	""" 
        if certfile == None:
            certfile = DEFAULT_CERTFILE
	
        if outFile == None:
            outFile = DEFAULT_OUTFILE+ str(os.getuid())

        utilc.globus_module_activate(gssc.get_module())
        ret = myProxyc.get_delegation(self._attrs, self._request,
				      self._response,hostname,port,
				      certfile, outFile,
				      lifetime, self._username,
                                      interactive)

        if not ret == 0:
            ex = MyProxyException("Failed to receive proxy")
            raise ex
    
	



	

    
