############################################################################
#Modified from Scott Koranda's LDR software package
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################


"""
Simple interface to the Globus RLS client API.
"""
import types, urlparse, time
import rlsClientc
import utilc, ioc
from pyGlobus.util import GlobusException

urlparse.uses_netloc.insert(0, 'rls')

# constant definitions (enums) used in RLS client API
rls_pattern_unix = rlsClientc.rls_pattern_unix          
rls_pattern_sql = rlsClientc.rls_pattern_sql            

obj_lrc_lfn = rlsClientc.globus_rls_obj_lrc_lfn
obj_lrc_pfn = rlsClientc.globus_rls_obj_lrc_pfn
obj_rli_lfn = rlsClientc.globus_rls_obj_rli_lfn
obj_rli_lrc = rlsClientc.globus_rls_obj_rli_lrc

attr_type_date  = rlsClientc.globus_rls_attr_type_date
attr_type_flt = rlsClientc.globus_rls_attr_type_flt
attr_type_int   = rlsClientc.globus_rls_attr_type_int
attr_type_str   = rlsClientc.globus_rls_attr_type_str

attr_op_all  = rlsClientc.globus_rls_attr_op_all 
attr_op_eq  = rlsClientc.globus_rls_attr_op_eq 
attr_op_ne  = rlsClientc.globus_rls_attr_op_ne 
attr_op_gt  = rlsClientc.globus_rls_attr_op_gt 
attr_op_ge  = rlsClientc.globus_rls_attr_op_ge 
attr_op_lt  = rlsClientc.globus_rls_attr_op_lt 
attr_op_le  = rlsClientc.globus_rls_attr_op_le 
attr_op_btw  = rlsClientc.globus_rls_attr_op_btw 
attr_op_like  = rlsClientc.globus_rls_attr_op_like 


                
class RlsClientException(GlobusException):
    """
       A simple subclass of GlobusException.
       
       Used for errors in the rlsClient module.
       """
    def __init__(self, args=None):
        """
        Constructs an instance.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        None
        
        **Raises:**
        
        None
        """
        GlobusException.__init__(self, args)
        self.args = args

class RlsClient:
    """
       A class to wrap the RLS client functions
       """

    def __init__(self, url):
        """
        Constructs an instance.
        
        Create a new RlsClient instance. This will init the RLS Globus
        module and open a connection to the RLS server pointed to by
        url.
        
        **Arguments:**
        
        - *url*  url for the RLS server to establish a connection with
        
        **Returns:**
        
        An instance of class RlsClient
        
        **Raises:**
        
        A RlsClientException is thrown if unable to init the module
        or open the connection to the server.
        """
        
        self._activated = 0
        self._handle = None
        self._url = None
        
        scheme, netloc, path, parameters, query, fragment = urlparse.urlparse(url)

        if scheme != 'rls':
            msg = "Invalid URL syntax for RLS server"
            ex = RlsClientException(msg)
            raise ex

        ret = utilc.globus_module_activate(utilc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_COMMON_MODULE")
            raise ex

        ret = utilc.globus_module_activate(ioc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_IO_MODULE")
            raise ex

        ret = utilc.globus_module_activate(rlsClientc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_RLS_CLIENT_MODULE")
            raise ex
        self._activated = 1
        
        ret, handle = rlsClientc.client_connect(url)
        if not ret == 0:
            ex = RlsClientException(handle)
            raise ex
        self._handle = handle
        self._url = url

        
    def __del__(self):
        """
        Destroy an instance.
        
        Destroys the handle thus terminating the connection to
        the server and deactivates the Globus modules.
        
        **Arguments:**
                
        None
               
        **Returns:**
        
        None    
        """
        
        if self._handle:
            ret, err = rlsClientc.client_close(self._handle)
            if not ret == 0:
                ex = RlsClientException(err)
                raise ex

            self._handle = None
            self._url = None

        ret = utilc.globus_module_deactivate(utilc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_COMMON_MODULE")
            raise ex
        
        ret = utilc.globus_module_deactivate(ioc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_IO_MODULE")
            raise ex
            
            
        ret = utilc.globus_module_deactivate(rlsClientc.get_module())
        if not ret == 0:
            ex = RlsClientException("Unable to initialize the GLOBUS_RLS_CLIENT_MODULE")
            raise ex
                
        self._activated = 0


    def lrc_clear(self):
        """
        Clear all mappings from LRC database. User needs both 
        ADMIN and LRCUPDATE privileges to perform this operation. 
               
        Note that if the LRC is cleared this will not be reflected 
        in any RLI servers updated by the LRC until the next 
        softstate update, even if immediate updates are enabled.
        
        **Arguments:**
               
        None

        **Returns:**
        
        None
                
        **Raises:**

        A RlsClientException is raised if unable to contact the
        LRC or clear the LRC.
        """

        ret, err = rlsClientc.lrc_clear(self._handle)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex


    def lrc_get_pfn(self, lfn, offset = 0, reslimit = 0):
        """
        Query LRC with LFN to obtain PFNs
        
        **Arguments:**
        
        - *lfn* The LFN to query the LRC with
        
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
        
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
        
        **Returns:**
        
        A list of PFNs.
        
        **Raises:**
        
        A RlsClientException is raised if unable to query the LRC or
        if the LFN is not known to the LRC.
        """

        if not isinstance(lfn, types.StringType):
            msg = "LFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        ret, pfnList = rlsClientc.lrc_get_pfn(self._handle, lfn, offset, reslimit)

        if not ret == 0:
            ex = RlsClientException(pfnList)
            raise ex

        return pfnList

    def lrc_get_pfn_wc(self, lfnpat, offset = 0, reslimit = 0):
        """
        Query LRC with LFN wildcard expression to obtain PFNs. 
        
        **Arguments:**
               
        - *lfnpat* The wildcard expression to search with. Can contain
        either Unix file globbing chars (* matches 0 or more characters, 
        ? matches any single character) or SQL "like" wildcard characters 
        (% matches 0 or more charactes, _ matches any single character)
               
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
               
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
        
        **Returns:**
               
        A dictionary object with LFNs as keys and PFNs as values.
        
        **Raises:**
               
        A RlsClientException is raised if unable to query the LRC.
        """

        if not isinstance(lfnpat, types.StringType):
            msg = "LFN wildcard pattern argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        if (lfnpat.count("*") > 0 or lfnpat.count("?") > 0):
            pattype = rls_pattern_unix
        elif (lfnpat.count("%") > 0 or lfnpat.count("_") > 0):
            pattype = rls_pattern_sql
        else:
            msg = "Bad LFN wildcard pattern"
            ex = RlsClientException(msg)
            raise ex

        ret, wcDict  = rlsClientc.lrc_get_pfn_wc(self._handle, lfnpat, 
                                                 pattype, offset, reslimit)
        if not ret == 0:
            ex = RlsClientException(wcDict)
            raise ex
            
        return wcDict

    def lrc_get_lfn(self, pfn, offset = 0, reslimit = 0):
        """
        Query LRC with PFN to obtain LFNs
        
        **Arguments:**
        
        - *pfn* The PFN to query the LRC with
               
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
        
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
               
        **Returns:**
               
        A list of LFNs.
        
        **Raises:**
        
        A RlsClientException is raised if unable to query the LRC or
        if the PFN is not known to the LRC.
        """

        if not isinstance(pfn, types.StringType):
            msg = "PFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex
            
        ret, lfnList = rlsClientc.lrc_get_lfn(self._handle, pfn, 
                                              offset, reslimit)
        if not ret == 0:
            ex = RlsClientException(lfnList)
            raise ex
           
        
        return lfnList

    def lrc_get_lfn_wc(self, pfnpat, offset = 0, reslimit = 0):
        """
        Query LRC with PFN wildcard expression to obtain LFNs. 
        
        **Arguments:**
               
        - *pfnpat* The wildcard expression to search with. Can contain
        either Unix file globbing chars (* matches 0 or more characters, 
        ? matches any single character) or SQL "like" wildcard characters 
        (% matches 0 or more charactes, _ matches any single character)
        
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
        
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
        
        **Returns:**
        
        A dictionary object with PFNs as keys and LFNs as values.
        
        **Raises:**
        
        A RlsClientException is raised if unable to query the LRC.
        """

        if not isinstance(pfnpat, types.StringType):
            msg = "PFN wildcard pattern argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        if (pfnpat.count("*") > 0 or pfnpat.count("?") > 0):
            pattype = rls_pattern_unix
        elif (pfnpat.count("%") > 0 or pfnpat.count("_") > 0):
            pattype = rls_pattern_sql
        else:
            msg = "Bad PFN wildcard pattern"
            ex = RlsClientException(msg)
            raise ex

        ret, wcDict = rlsClientc.lrc_get_lfn_wc(self._handle, pfnpat, 
                                                pattype, offset, reslimit)
        if not ret == 0:
            ex = RlsClientException(wcDict)
            raise ex
            
        return wcDict
                
    def lrc_create_lfn(self, lfn, pfn):
        """
        Create LFN with associated PFN in LRC
        
        **Arguments:**
        
        - *lfn* The LFN to create in LRC
        
        - *pfn* The PFN to associate with LFN

        **Returns:**
        
        None

        **Raises:**
        
        A RlsClientException is raised if unable to create
        the LFN.
        """

        if _checkLFN(lfn):
            msg = "LFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        if _checkPFN(pfn):
            msg = "PFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        ret, err = rlsClientc.lrc_create(self._handle, lfn, pfn)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
            

    def lrc_add(self, lfn, pfn):
        """
        Add a mapping for a PFN to an existing LFN. The LFN must
        exist.
        
        **Arguments**:
               
        - *lfn* LFN to which the PFN is to be mapped
        
        - *pfn* PFN to map to the LFN
        
        **Returns**:
               
        None
        
        **Raises**:
               
        A RlsClientException is raised if unable to add
        the PFN.
        """
        if _checkLFN(lfn):
            msg = "LFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        if _checkPFN(pfn):
            msg = "PFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex
            
        ret, err = rlsClientc.lrc_add(self._handle, lfn, pfn)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
           

    def lrc_delete(self, lfn, pfn):
        """
        Delete a mapping for a PFN to an existing LFN. 
        
        **Arguments**:
               
        - *lfn* LFN to which the PFN was mapped
        
        - *pfn* PFN for which the map to LFN is to be deleted
               
        **Returns**:

        None
        
        **Raises**:

        A RlsClientException is raised if unable to delete
        the PFN.
        """
        if _checkLFN(lfn):
            msg = "LFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex
            
        if _checkPFN(pfn):
            msg = "PFN argument must be a string"
            ex = RlsClientException(msg)
            raise ex
            
        ret, err = rlsClientc.lrc_delete(self._handle, lfn, pfn)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
            

    def lrc_exists(self, fn, objtype):
        """
        Query LRC for existence of a LFN or PFN
        
        **Arguments:**
        
        - *fn* LFN or PFN object for which to check existence
        
        - *objtype* Type of object argument fn refers to. Should be
        either obj_lrc_lfn or obj_lrc_pfn, both of which are defined
        in this module.
        
        **Returns:**
        1 if exists or 0 if not
        
        **Raises:**
        A RlsClientException is raised if unable to query
        the LRC.
        """

        ret, err = rlsClientc.lrc_exists(self._handle, fn, objtype)

        if ret:
            
            if (ret == rlsClientc.GLOBUS_RLS_LFN_NEXIST or
                ret == rlsClientc.GLOBUS_RLS_PFN_NEXIST):
                return 0
            else:
                ex = RlsClientException(err)
                raise ex
        return 1


    def lrc_attr_create(self, attr):
        """
        Define an attribute in the LRC
        
        **Arguments:**
        
        - *attr* Instance of class RlsAttr or a derived class. The
        instance must have name, objtype, and type set. See documentation
        for methods of class RlsAttr.
               
        **Returns:**
        
        None
        
        **Raises:**
        
        A RlsClientException is raised if unable to define the attribute.
        """

        if not isinstance(attr, RlsAttr):
            msg = "Argument must be instance of class RlsAttr or derived class"
            ex = RlsClientException(msg)
            raise ex

        ret, err = rlsClientc.lrc_attr_create(self._handle, attr.get_name(), 
                                            attr.get_objtype(), attr.get_type())
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
            

    def lrc_attr_delete(self, attr, clear = 0):
        """
        Delete an attribute definition in the LRC
        
        **Arguments:**
        
        - *attr* Instance of class RlsAttr or a derived class. The
        instance must have name, objtype, and type set. See documentation
        for methods of class RlsAttr.
        
        **Returns:**
        
        None
               
        **Raises:**

        A RlsClientException is raised if unable to delete the attribute
        definition.
        """

        ret, err = rlsClientc.lrc_attr_delete(self._handle, attr.get_name(), 
                                             attr.get_objtype(), clear)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
        
                        
    def lrc_attr_add(self, fn, attr):
        """
        Set an attribute value for an entry (LFN or PFN) in the LRC
        
        **Arguments:**
        
        - *fn* The filename object for which the attribute value will
        be set.
        
        - *attr* An instance of class RlsAttr or a derived class with
        name, objtype, type, and value set. See documentation for class
        RlsAttr.
        
        **Returns:**
        
        None
        
        **Raises:**

        A RlsClientException is raised if unable to contact the
        LRC or add the attribute.
        """
        if not isinstance(attr, RlsAttr):
            msg = "Argument must be instance of class RlsAttr or derived class"
            ex = RlsClientException(msg)
            raise ex

        #battr = attr._base()

        ret, err  = rlsClientc.lrc_attr_add(self._handle, fn, attr.get_handle())
        #del battr
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
                

    def lrc_attr_remove(self, fn, attr):
        """
        Remove an attribute value from an entry (LFN or PFN) in the LRC
        
        **Arguments:**
        
        - *fn* The filename object for which the attribute value will
        be removed.
        
        - *attr* An instance of class RlsAttr or a derived class with
        name, objtype, type, and value set. See documentation for class
        RlsAttr.
        
        **Returns:**
        
        None
               
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or remove the attribute.
        """
        if not isinstance(attr, RlsAttr):
            msg = "Argument must be instance of class RlsAttr or derived class"
            ex = RlsClientException(msg)
            raise ex

        #battr = attr._base()

        ret, err = rlsClientc.lrc_attr_remove(self._handle, fn, attr.get_handle())
        #del battr
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
                
                        
    def lrc_attr_search(self, attr1, op, attr2 = None, offset = 0, reslimit = 0):
        """
        Search for objects (LFNs or PFNs) in a LRC database that have the
        specified attribute whose value matches a boolean expression.
               
        **Arguments:**
        
        - *attr1* An instance of class RlsAttr or a derived class with
        name, objtype, type, and value set and which is used to search
        on. See documentation for class RlsAttr.
        
        - *op* Operator to be used in searching for values. Should be one of:

        attr_op_all     all values returned
        
        attr_op_eq      values matching operand 1 returned 
        
        attr_op_ne      values not matching operand 1
        
        attr_op_gt      values greater than operand 1 
        
        attr_op_ge      values greater than or equal to attr1 value
        
        attr_op_lt      values less than operand 1
        
        attr_op_le      values less than or equal to attr1 value
        
        attr_op_btw     values between attr1 and attr2 value
        
        attr_op_like    Strings "like" operand1 (SQL like)
        
        - *attr2* Second instance of class RlsAttr if a second operand
        for operator is required
               
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
               
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
        
        **Returns:**
        A dictionary object with attribute names as keys and
        attribute values as values.
               
        **Raises:**
               
        A RlsClientException is raised if unable to contact the
        LRC or perform the search.
        """ 

        name = attr1.get_name()
        objtype = attr1.get_objtype()
        #battr1 = attr1._base()
        #if attr2 != None:
        #    battr2 = attr2._base()
        #else:
        #    battr2 = None

        if attr2 != None:
            attr2Handle = attr2.get_handle()
        else:
            attr2Handle = None
        
        ret, attrDict = rlsClientc.lrc_attr_search(self._handle, name,
                                               objtype, op,
                                               attr1.get_handle(), attr2Handle,
                                               offset, reslimit)
        
        if not ret == 0:
            ex = RlsClientException(attrDict)
            raise ex

        return attrDict

    def lrc_attr_get(self, objtype, name = None):
        """
        Get definitions of attributes from LRC.
        
        **Arguments:** 
        
        - *objtype* The type of object (LFN or PFN) for which to get
        attribute definitions. Should be one of obj_lrc_lfn or
        obj_lrc_pfn, which are both defined in this module.
        
        - *name* Name of particular attribute to get. If None
        all attributes for the specified object type are returned.
        
        **Returns:**
        
        A list of instances of class RlsAttr.
        
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or get the definition.
        """
        
        if name == None:
            attrname = ""
        else:
            attrname = name
            
        ret, attrDict = rlsClientc.lrc_attr_get(self._handle, attrname, objtype)

        # returns non-zero integer if error or a dictionary with attribute names
        # as the keys and values that are lists containing rest of the
        # rls client API globus_rls_attribute_t struct members
        
        if not ret == 0:
            ex = RlsClientException(attrDict)
            raise ex

        # convert dictionary to list of RlsAttr class instances
        attrList = []
        for key in attrDict.keys():
            newAttr = RlsAttr()
            newAttr.set_name(key)
            newAttr.set_objtype(attrDict[key][0])
            newAttr.set_type(attrDict[key][1])
            newAttr.set_val(attrDict[key][2])
                
            attrList.append(newAttr)

        return attrList

    def lrc_attr_value_get(self, fn, objtype, attrname = None):
        """
        Get attributes in LRC for a specific object (LFN or PFN)
        
        **Arguments:** 
        
        - *fn* The filename of the object (LFN or PFN) for which to
        get attributes.
        
        - *objtype* The type of object (LFN or PFN) for which to get
        attribute. Should be one of obj_lrc_lfn or
        obj_lrc_pfn, which are both defined in this module.
        
        - *attrname* Name of particular attribute to get. If None
        all attributes for the specified filename are returned.
        
        **Returns:**
        
        A list of instances of class RlsAttr.
        
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or get the definition.
        """
        if attrname == None:
            aname = ""
        else:
            aname = attrname
                
        ret, attrDict = rlsClientc.lrc_attr_value_get(self._handle, fn, 
                                                      objtype, aname)

        if not ret == 0:
            ex = RlsClientException(attrDict)
            raise ex

        # convert dictionary to list of RlsAttr class instances
        attrList = []
        for key in attrDict.keys():
            newAttr = RlsAttr()
            newAttr.set_name(key)
            newAttr.set_objtype(attrDict[key][0])
            newAttr.set_type(attrDict[key][1])
            newAttr.set_val(attrDict[key][2])

            attrList.append(newAttr)

        return attrList

    def lrc_rli_add(self, url, pattern = "", bloom = 0):
        """
        LRC servers send information about LFNs in their database to the
        the list of RLI servers in the database, added with this
        function. Updates may be partitioned amongst multiple
        RLIs by specifying one or more patterns for an RLI. 
        
        **Arguments:**
        
        - *url* URL of RLI server to add
        
        - *pattern* If not empty str used to filter which LFNs are sent to
        url. Standard Unix wildcard characters (*, ?) may be used
        to do wildcard matches.
        
        - *bloom* If non-zero then update RLI using bloom filters
               
        **Returns:**
        
        None

        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or add the RLI URL.
        """

        bloomfilter = 0
        if bloom != 0:
            bloomfilter = rlsClientc.FRLI_BLOOMFILTER
            
        ret, err  = rlsClientc.lrc_rli_add(self._handle, url, 
                                           bloomfilter, pattern)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex

    def lrc_rli_delete(self, url, pattern = ""):
        """
        Delete an entry from the LRC rli/partition tables.
        
        **Arguments:**
        
        - *url* URL of RLI server to remove
               
        - *pattern* If not empty str then only the specific RLI URL/pattern
        is removed, else all partition information for URL is
        removed.
        
        **Returns:**
        
        None

        **Raises:**

        A RlsClientException is raised if unable to contact the
        LRC or delete the RLI URL.
        """
        
        ret, err = rlsClientc.lrc_rli_delete(self._handle, url, pattern)
        if not ret == 0:
            ex = RlsClientException(err)
            raise ex
                

    def lrc_rli_list(self):
        """
        Return URLs of RLIs that LRC sends updates to.
        
        **Arguments:**
        
        None

        **Returns:**

        A dictionary object with RLI URLs as keys and lists as values
        where the list elements are update interval, flags, and last
        update.

        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or query for RLI URLs.
        """
               
        ret, rliDict = rlsClientc.lrc_rli_list(self._handle)
        if not ret == 0:
            ex = RlsClientException(rliDict)
            raise ex
            
        return rliDict

    def lrc_rli_info(self, url):
        """
        Get info about RLI server updated by an LRC server.
        
        **Arguments:**
        
        - *url* URL of RLI server to retrieve info for
        
        **Returns:**
        
        A dictionary object with RLI URLs as keys and lists as values
        where the list elements are update interval, flags, and last
        update.
        
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        LRC or query for RLI URLs.
        """

        ret, rliDict = rlsClientc.lrc_rli_info(self._handle, url)
        if not ret == 0:
            ex = RlsClientException(rliDict)
            raise ex
            
        return rliDict   

    def lrc_rli_get_part(self, url = "", pattern = ""):
        """
        Get RLI update partitions from LRC server.
        
        **Arguments:**
        
        - *url* URL of RLI server to retrieve info for. If not empty str
        identifies RLI that partition data will be retrieved for. If
        None then all RLIs are retrieved.
        
        - *pattern* If not empty str returns only partitions with matching
        patterns, otherwise all patterns are retrieved.
                  
        **Returns:**
        
        A list of lists. Each individual list has the URL as the
        first element and the associate pattern as the second 
        element.
        
        **Raises:**

        A RlsClientException is raised if unable to contact the
        LRC or query the LRC.
        """
        
        ret, rliList = rlsClientc.lrc_rli_get_part(self._handle, url, pattern)
        if not ret == 0:
            ex = RlsClientException(rliList)
            raise ex
            
        return rliList
            

    def rli_lfn_exists(self, lfn):
        """
        Check if LFN exists in RLI database.
        
        **Arguments:**
        
        - *lfn* LFN for which to check existence, also could be an lrc
        
        **Returns:**
        
        One if the LFN exists or zero if it does not.
        
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        RLI or query on the LFN.
        """
        
        ret, err = rlsClientc.rli_exists(self._handle, lfn, obj_rli_lfn)
        
        if ret == 0:
            myreturn = 1
        elif ret == rlsClientc.GLOBUS_RLS_LFN_NEXIST:
            myreturn = 0
        else:
            ex = RlsClientException(err)
            raise ex
        
        return myreturn


    def rli_get_lrc(self, lfn, offset = 0, reslimit = 0):
        """
        Return LRCs mapped to LFN in the RLI database.
        
        **Arguments:**
        
        - *lfn* LFN for which to return LRC mappings
        
        **Returns:**
        
        List of LRCs.
        
        **Raises:**
        
        A RlsClientException is raised if unable to contact the
        RLI or query on the LFN.
        """

        ret, lrcList = rlsClientc.rli_get_lrc(self._handle, lfn, 
                                              offset, reslimit)
        if not ret == 0:
            ex = RlsClientException(lrcList)
            raise ex
            
        return lrcList
            

    def rli_get_lrc_wc(self, lfnpat, offset = 0, reslimit = 0):
        """
        Query RLI with LFN wildcard expression to obtain LRCs. 
            
        **Arguments:**
        
        - *lfnpat* The wildcard expression to search with. Can contain
        either Unix file globbing chars (* matches 0 or more characters, 
        ? matches any single character) or SQL "like" wildcard characters 
        (% matches 0 or more charactes, _ matches any single character)
        
        - *offset* Offset into result list. Used in conjunction with
        reslimit to retrieve results a few at a time. Use 0 to begin
        with first result.
        
        - *reslimit* Maximum number of results to return. Used in
        conjunction with offset to retrieve results a few at a time.
        Use 0 to retrieve all results.
        
        **Returns:**
        
        A dictionary object with LFNs as keys and LRCs as values.
        
        **Raises:**
        
        A RlsClientException is raised if unable to query the RLI.
        """
               
        if not isinstance(lfnpat, types.StringType):
            msg = "LFN wildcard pattern argument must be a string"
            ex = RlsClientException(msg)
            raise ex

        if (lfnpat.count("*") > 0 or lfnpat.count("?") > 0):
            pattype = rls_pattern_unix
        elif (lfnpat.count("%") > 0 or lfnpat.count("_") > 0):
            pattype = rls_pattern_sql
        else:
            msg = "Bad LFN wildcard pattern"
            ex = RlsClientException(msg)
            raise ex

        ret, lrcDict = rlsClientc.rli_get_lrc_wc(self._handle, lfnpat, pattype,
                                                 offset, reslimit)
        
        if ret == rlsClientc.GLOBUS_RLS_LFN_NEXIST:
            return {}

        if not ret == 0:
            ex = RlsClientException(lrcDict)
            raise ex

        return lrcDict
           

class RlsAttrException(GlobusException):
    """
    A simple subclass of GlobusException.  Used for errors in the rlsClient module.
    """
    def __init__(self, args=None):
        """
        Constructs an instance.
               
        **Arguments:**
        
        None
        
        **Returns:**
               
        An instance of class RlsAttrException.

        **Raises:**
               
        None
        """
        GlobusException.__init__(self, args)   
        self.args = args

class RlsAttr:
    """
    A class to wrap the RLS globus_rls_attribute_t structure. This
    class should be subclassed to construct individual types
    and instances of attributes.
    """

    def __init__(self, handle = None): 
        """
        Construct an instance.
        
        **Arguments:**
        
        - * handle* a SWIG'ized pointer to an initialized globus_rls_attribute_t
        structure. 
        
        **Returns:**
        
        An instance of class RlsAttr.
        
        **Raises:**

        None
        """
        if handle == None:
            handle = rlsClientc.attr_init()

        self._handle = handle
        self.val = None
        self.objtype = None
        self.type = None
        self.name = None
        
    def __del__(self):
        """
        Frees the memory associated with the handle

        **Arguments**

        None

        **Returns**

        None

        **Raises**

        None
        """
        rlsClientc.attr_free(self._handle)
        
    def __str__(self):
        """
        Formats instances for pretty printing.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        A string suitable for printing.
        
        **Raises:**
        
        None
        """
        
        myDict = {}
        myDict[self.name] = [self.get_objtype(), self.get_type(), self.get_val()]
        string = str(myDict)
        del myDict
        return string
                
    def _base(self):
        """
        Return just this base class with attributes set to
        the current values.

        **Arguments:**
        
        None
        
        **Returns:**
        
        An instance of the class RlsAttr
        
        **Raises:**
        
        None
        """
        battr = RlsAttr(self._handle)
        battr.name = self.get_name()
        battr.objtype = self.get_objtype()
        battr.type = self.get_type()
        battr.val = self.get_val()

        return battr

    def set_name(self, name):
        """
        Set the name for an attribute.
        
        **Arguments:**
        
        -*name* String used as name for attribute
        
        **Returns:**
        
        None
        
        **Raises:**
        
        A RlsAttrException is raised if unable to set the name
        for the attribute.
        """
        if not isinstance(name, types.StringType):
            msg = "Name of an instance of RlsAttr must be a string"
            ex = RlsAttrException(msg)
            raise ex

        rlsClientc.attr_set_name(self._handle, name)
        return

    def get_name(self):
        """
        Get the name of an attribute instance.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        String representing the name of the attribute.
        
        **Raises:**
        
        None
        """
        
        return rlsClientc.attr_get_name(self._handle)

    def set_objtype(self, objtype):
        """
        Set the object type for an attribute. 
        
        **Arguments:**
        
        -*objtype* Object type of the object to which the attribute
        applies.  Argument must be one of
        
        obj_lrc_lfn
        
        obj_lrc_pfn
        
        obj_rli_lfn
        
        obj_rli_lrc
        
        **Returns:**
               
        None
               
        **Raises:**
        
        A RlsAttrException is raised if unable to set the object type
        for the attribute.
        """
        if (objtype != obj_lrc_lfn and
            objtype != obj_lrc_pfn and
            objtype != obj_rli_lfn and
            objtype != obj_rli_lrc):

            msg = "Object type argument not valid"
            ex = RlsAttrException(msg)
            raise ex

        rlsClientc.attr_set_objtype(self._handle, objtype)

    def get_objtype(self):
        """
        Get the object type of an attribute instance.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        Integer representing the object type of the attribute.
        
        **Raises:**
        
        None
        """
        return rlsClientc.attr_get_objtype(self._handle)


    def set_type(self, type):
        """
        Set the value type for an attribute. 
        
        **Arguments:**
        
        -*type* Type of the attribute value. Must be one of
        
        attr_type_date
        
        attr_type_flt
        
        attr_type_int 
        
        attr_type_str
        
        **Returns:**
        
        None
        
        **Raises:**
        
        A RlsAttrException is raised if unable to set the type
        for the attribute.
        """

        if ( type != attr_type_date and
             type != attr_type_flt  and
             type != attr_type_int  and
             type != attr_type_str):
            msg = "Attribute value type argument not valid"
            ex = RlsAttrException(msg)
            raise ex
        rlsClientc.attr_set_type(self._handle, type)
        
        
    def get_type(self):
        """
        Get the type of an attribute instance.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        Integer representing the type of the attribute value.
        
        **Raises:**
        
        None
        """

        return rlsClientc.attr_get_type(self._handle)

    
    def set_val(self, val):
        """
        Set the value for an attribute. Also sets the type of the value
        
        **Arguments:**
        
        -*val* Value to which the attribute should be set
        
        **Returns:**
               
        None
        
        **Raises:**
        
        A RlsAttrException is raised if unable to set the val
        for the attribute.
        """
        
        if isinstance(val, types.IntType):
            rlsClientc.attr_set_type(self._handle, attr_type_int)
            rlsClientc.attr_set_int_value(self._handle, val)
        elif isinstance(val, types.FloatType):
            rlsClientc.attr_set_type(self._handle, attr_type_flt)
            rlsClientc.attr_set_float_value(self._handle, val)
        elif isinstance(val, types.StringType):
            rlsClientc.attr_set_type(self._handle, attr_type_str)
            rlsClientc.attr_set_string_value(self._handle, val)
        elif isinstance(val, types.TupleType):
            if len(val) != 9:
                msg = "Dates must be specified as UTC tuple"
                ex = RlsAttrException(msg)
                raise ex
            try:
                t = time.mktime(val)
            except:
                msg = "Invalid date."
                ex = RlsAttrException(msg)
                raise ex
            rlsClientc.attr_set_type(self._handle, attr_type_date)
            rlsClientc.attr_set_time_value(self._handle, int(t))
        elif val == None:
            pass
        
        else:
            ex = RlsAttrException("Passed in unsupported type to set_val")
            raise ex
        
    def get_val(self):
        """
        Get the value of an attribute instance.
        
        **Arguments:**
        
        None
        
        **Returns:**
        
        Value. The type depends on the attribute value type.
        
        **Raises:**
        
        A RlsAttrException is raised if the type is illegal
        None
        """
        attrType = self.get_type()
        if attrType == attr_type_int:
            return rlsClientc.attr_get_int_value(self._handle)
        elif attrType == attr_type_flt:
            return rlsClientc.attr_get_float_value(self._handle)
        elif attrType == attr_type_str:
            return rlsClientc.attr_get_string_value(self._handle)
        elif attrType == attr_type_date:
            return time.localtime(rlsClientc.attr_get_time_value(self._handle))
        else:
            ex = RlsAttrException("RlsAttr has an unknown type")
            raise ex
            

    def get_handle(self):
        """
        **Returns**
        *handle* - the SWIG'ized globus_rls_attribute_t pointer
        """
        return self._handle
    
def _checkLFN(lfn):
    """
    function that is used internally by RlsClient class
    """
    if not isinstance(lfn, types.StringType):
        return True 
    else:
        return False 

def _checkPFN(pfn):
    """
    function that is used internally by RlsClient class
    """
    if not isinstance(pfn, types.StringType):
        return True 
    else:
        return False
            
