############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
A module to provide some of the shared GSI functionality.
"""
from pyGlobus import gssc
from pyGlobus import utilc
from pyGlobus import proxyc
from pyGlobus import sslc
from pyGlobus.util import GlobusException
from pyGlobus.util import PyGlobusFactory
##########################################################################
# Module Constants
##########################################################################
# Constants used to indicate the status of proxy/key files
FILE_VALID = 0
#The file is neither regular, a link or a directory
FILE_INVALID = 1
#The file is a directory
FILE_DIR  = 2
#The file does not exist
FILE_DOES_NOT_EXIST = 3
#The file is not owned by the current user
FILE_NOT_OWNED = 4
#The file has incorrect permission
FILE_BAD_PERMISSIONS = 5
#The file has zero length
FILE_ZERO_LENGTH = 6

# Switch used by the get_proxy_filename function
PROXY_FILE_INPUT = 0 # filename intended for reading
PROXY_FILE_OUTPUT = 1 # filename intended for writing


#########################################################################
# Module functions
########################################################################

def grid_proxy_init(lifetime=12*60, verbose=1, verify=1, outFile=None, 
                    passphrase=None, debugCB=None, certDir=None, 
                    certFile=None, keyFile=None, bits=1024, 
                    proxyType=proxyc.GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_IMPERSONATION_PROXY):
    """
    Generates a user proxy

    **Arguments**

    - *lifetime* an int which is the number of minutest the proxy is valid

    - *outFile*  a string of the name of the file where the proxy is written.
    Defaults to /tmp/x509up_uuid

    - *verbose* print out information while generating the proxy
       
    - *verify* verify the certificate chain

    - *certDir* give the location of the certicate directory the default is ~/.globus/certificates
       
    - *certFile* give the location of the certificate file usually ~/.globus/usercert.pem

    - *keyFile* give the location of the key file usually ~/.globus/userkey.pem

    - *passphrase* a python function that has the signature func(rwflag)
    this function returns a passphrase that will be used to create a proxy

    - *debugCB* a python function that has the signature func(message), which can
    be used for debuging.

    - *bits* the strength of the private key {512|1024|2048|4096}
       
    - *proxyType* the kind of proxy you'd like to create
       
    *Raises*
    a GSIException is raised if the user proxy cannot be created
    
    *Returns*
    None
    """
    utilc.globus_module_activate(gssc.get_module())
    utilc.globus_module_activate(gssc.get_module_proxy())
    utilc.globus_module_activate(gssc.get_module_callback())
    utilc.globus_module_activate(gssc.get_module_openssl())
    #
    # Check to see if we're being invoked by the original interface,
    # or with a specification of the various cert directories and files.
    #

    
    ret, err = proxyc.grid_proxy_init(verbose,
                                      verify,
                                      bits,
                                      lifetime,
                                      outFile,
                                      passphrase,
                                      debugCB,
                                      certDir,
                                      certFile,
                                      keyFile,
                                      proxyType)
    

    utilc.globus_module_deactivate(gssc.get_module_callback())
    utilc.globus_module_deactivate(gssc.get_module())
    
    if ret != 0:
        errString = ""
        for e in err:
            errString += e + '\n'
        raise GSIException(errString)

def grid_proxy_info(verbose=1, proxyFile=None):
    """
    Retrieves information about the user proxy 
    **Arguments**
    
    - *verbose* - Print information to stdout and stderr
    
    - *proxyFile* - specify the proxyfile to get information about,
    if unspecified the default is picked up
    
    *Returns*
    a string with all of the user proxy information
    
    *Raises*
    a GSIException is raised if information about the user proxy cannot
    be retrieved
    """

    utilc.globus_module_activate(gssc.get_module()) 
    utilc.globus_module_activate(gssc.get_module_proxy())
    utilc.globus_module_activate(gssc.get_module_openssl())
    ret, proxy_string, proxy_dict = proxyc.grid_proxy_info(verbose, proxyFile) 
    utilc.globus_module_deactivate(gssc.get_module())
    
    if not ret == 0:
        raise GSIException(proxy_string)

    return proxy_string, proxy_dict


def grid_proxy_destroy(proxyFile=None):
    """
    Destroys the user proxy in the default location or
    in the X509_USER_PROXY
    
    *Returns*
    None
    
    *Raises*
    a GSIException is raised if the user proxy cannot be destroyed
    """

    utilc.globus_module_activate(gssc.get_module())
    utilc.globus_module_activate(gssc.get_module_proxy())
    utilc.globus_module_activate(gssc.get_module_openssl())
    ret, error_string = proxyc.grid_proxy_destroy(proxyFile)
    utilc.globus_module_activate(gssc.get_module_proxy())
    utilc.globus_module_deactivate(gssc.get_module())
    
    if not ret == 0:
        raise GSIException(error_string)

def gridmap(globusid):
    """ 
    Returns the userid if it is in the grid-mapfile
    
    If you are a non-root user it will look in ~/.gridmap first

    **Arguments**

    *globusid* a string of Subject: line in a usercert
    
    *Returns*
    the userid that globusid maps to

    *Raises*
     a GSIException is raised if the globusid is not in the gridmap file
    
    """
    utilc.globus_module_activate(gssc.get_module()) 
    ret, userid = gssc.assist_gridmap(globusid)
    utilc.globus_module_deactivate(gssc.get_module()) 
    if not ret == 0:
        raise GSIException("gridmap call failed")

    return userid

def userok(globusid, userid):
    """
    Returns 0 if the globusid maps to the userid in the ~/.gridmap file
    Returns 1 if authorization fails 
 
    *Raises*
    a GSIException is raised if bad arguments are given to the this
    function, ie if the DN is malformed 
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret = gssc.assist_userok(globusid, userid)
    utilc.globus_module_deactivate(gssc.get_module()) 
    
    if ret == -1:
        raise GSIException("userok call failed")
    return ret

def map_local_user(local_user):
    """
    Returns the *first* DN in the gridmap file that maps 
    to the local user's account.
    Note an account might have more than one DN associated 
    with it.
    
    *Raises*
    a GSIException is raised if the argument is malformed
    """ 
    
    utilc.globus_module_activate(gssc.get_module()) 
    ret, globususer = gssc.assist_map_local_user(local_user)
    utilc.globus_module_deactivate(gssc.get_module()) 

    if not ret == 0:
        raise GSIException, "map_local_user call failed"

    return globususer
    
def get_cert_dir():
    """
    Returns the certificate dir.
    
    Returns the full path to the trusted certificate
    directory.
    
    **Returns**
    A string containg the full path to the cert dir.
    
    **Raises**
    A GSIException is raised if unable to locate the 
    certificate directory.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, certDir = proxyc.get_cert_dir()
    if not ret == 0:
        ex = GSIException(certDir)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return certDir
    
def check_keyfile(filename):
    """
    Checks the status of a private key file.
    
    Convenience function that checks that a private
    key file is only readable and writeable by the 
    user.
    
    **Arguments**
    *filename* The full path to the private key
    file to be checked.
    
    **Returns**
    One of the file status values from this module are 
    returned indicating the state of the private key file.
    
    **Raises**
    A GISException is raised if unable to check the status
    of the private key file.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, status = proxyc.check_keyfile(filename)
    if not ret == 0:
        ex = GSIException(status)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return status
    
def check_certfile(filename):
    """
    Checks the status of a certificate file.
    
    This is a convenience function used to check the status of a 
    certificate file.  The desired status is the current user has
    ownership and read/write permissions, while group and others only
    have read permissions.
    
    **Arguments**
    *filename* The full path to the certificate
    file to be checked.
    
    **Returns**
    One of the file status values from this module are 
    returned indicating the state of the certificate file.
    
    **Raises**
    A GISException is raised if unable to check the status
    of the certificate file.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, status = proxyc.check_certfile(filename)
    if not ret == 0:
        ex = GSIException(status)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return status

def get_user_cert_filename():
    """
    Returns the location of the users cert and key.
    
    Convenience function to return the full path to 
    both the users certificate and private key files.
    
    **Returns**
    A tuple containing the full path to the users certificate 
    and private key files.
    
    **Raises**
    A GSIException is raised if unable to get the location of
    the certificate or key files.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, cert, key = proxyc.get_user_cert_filename()
    if not ret == 0:
        ex = GSIException(cert)
        raise ex
    return (cert, key)

def get_host_cert_filename():
    """
    Returns the location of the hosts cert and key.
    
    Convenience function to return the full path to 
    both the hosts certificate and private key files.
    
    **Returns**
    A tuple containing the full path to the hosts certificate 
    and private key files.
    
    **Raises**
    A GSIException is raised if unable to get the location of
    the certificate or key files.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, cert, key = proxyc.get_host_cert_filename()
    if not ret == 0:
        ex = GSIException(cert)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return (cert, key)
        
def get_proxy_filename(proxy_type=PROXY_FILE_OUTPUT):
    """
    Returns the proxy cert filename
    
    Returns the full path to the proxy cert.
    
    **Arguments**
    proxy_type A int constant switch describing whether to
    return an existing proxy filename, or a new filename 
    suitable for creating a new proxy is returned.
     
    **Returns**
    A string containg the full path to the proxy cert.
    
    **Raises**
    A GSIException is raised if unable to locate the 
    proxy cert.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, filename = proxyc.get_proxy_filename(proxy_type)
    if not ret == 0:
        ex = GSIException(dir)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return filename

def get_service_cert_filename(service_name):
    """
    Returns the location of the services cert and key.
    
    Convenience function to return the full path to 
    both the services certificate and private key files.
    
    **Arguments**
    service_name A string containing the name of the service,
    e.g., 'ldap'.
    
    **Returns**
    A tuple containing the full path to the services certificate 
    and private key files.
    
    **Raises**
    A GSIException is raised if unable to get the location of
    the certificate or key files.
    """

    utilc.globus_module_activate(gssc.get_module()) 
    ret, cert, key = proxyc.get_service_cert_filename(service_name)
    if not ret == 0:
        ex = GSIException(cert)
        raise ex
    utilc.globus_module_deactivate(gssc.get_module()) 
    return (cert, key)
        
###########################################################################
# Module Classes
##########################################################################

class GSIException(GlobusException):
    """
    A subclass of GlobusException.
    
    Used as a base class for all other security exceptions.
    
    **Extends:**

    *pyGlobus.util.GlobusException*
    """
    
    pass

class GSSContextException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the GSSContext class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class GSSCredException(GSIException):
    """
       A subclass of GSIException.

       Used to indicate errors in the GSSCred class.

       **Extends:**

       *pyGlobus.security.GSI.GSIException*
       """
    
    pass

class GSICredentialAttrException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the GSICredentialAttr module.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class GSICredentialException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the GSICredential class.
    
    **Extends:**

    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class X509Exception(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the X509 class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class EVP_KeyException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the EVP_Key class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class GSSNameException(GSIException):
    """
    A subclass of GSSContextException.
    
    Used to indicate errors in the GSSName class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class GSSMechsException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the GSSMechs class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    
    pass

class GSSBufferException(GSIException):
    """
    A subclass of GSIException.
    
    Used to indicate errors in the GSSBuffer class.
    
    **Extends:**
    
    *pyGlobus.security.GSI.GSIException*
    """
    pass    

class GSSContext:
    """
    A  class for the underlying gss_ctx_id_t object.
    """

    def __init__(self, handle=None, free=1):
        """
        Constructs an instance.
        
        Each instance wraps a gss_ctx_id_t object.
        Defaults to a GSS_C_NO_CONTEXT
        
        **Arguments:**
        
        - *handle* An optional argument that, if present, should
        contain a SWIG'ized pointer to a gss_ctx_id_t object.
        
        - *free* Should the underlying gss_ctx_id_t object be
        freed when an instance is destroyed.
        """
        
        self._activated = 0
        self._delegated_cred = None
        ret = utilc.globus_module_activate(gssc.get_module())
        if not ret == 0:
            ex = GSSContextException("Unable to initialize the module")
            raise ex

        self._activated = 1
        
        if handle == None:
            self._handle = gssc.create_context()
        else:
            self._handle = handle
        self._free = free
        return

    def __del__(self):
        """
        Destroys an instance.
        
        Deletes the underlying gss_ctx_id_t object, and free's the
        memory if the instance owns the underlying handle.
        """
        
        if self._free:
            major, minor = gssc.delete_context(self._handle)
            if not major == gssc.GSS_S_COMPLETE:
                errList = gssc.display_status(major, minor)
                ex = apply(GSSContextException, errList)
                raise ex

        if self._activated:
            ret = utilc.globus_module_deactivate(gssc.get_module())
            if not ret == 0:
                print 'Error!: Unable to deactiveate security module!'
            
        return

    def inquire(self):
        """
        Obtain information about the context.
        
        **Returns:**
        A 7-tuple containing: A GSSName object containing the initiators name,
        A GSSName object with the acceptors name.
        An int containing the number of seconds the context
        will remain valid for.
        A string containing a SWIG'ized pointer to a static
        gss_OID struct containing the mechanism type.
        An int containing the context flags.
        A bool indicating if the connection was initiated
        locally. A bool indicating if the connection is currently open.
        
        **Raises:**
        
        A GSSContextException is raised if an error occurs in the
        underlying gss-api call.
        """
        
        ctxTuple = gssc.inquire_context(self._handle)
        if not ctxTuple[0] == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(ctxTuple[0], ctxTuple[1])
            ex = apply(GSSContextException, errList)
            raise ex
        return GSSName(ctxTuple[2]), GSSName(ctxTuple[3]), ctxTuple[4], \
               ctxTuple[5], ctxTuple[6], ctxTuple[7], ctxTuple[8]

    def init_context(self, init_cred, target_name, inputTokenString="", 
                     requests=None):
        """
        Tries to establish a security context with a server
        
        **Arguments**
        
        - *init_cred* is a GSSCred object which contains the client's credential handle
        
        - *target_name* is a GSSName object
        
        - *inputTokenString* is a string of tokens, received from the server or "", 
        if this is the first time init_context has been called. 
        
        - *requests* is a ContextRequests object
        
        **Returns:**
        major and minor status error codes from the GSSAPI.
        A string of tokens which represent credentials from the client who is initiating
        the context.
        
        **Raises:**
        
        A GSSContextException is raised if an error occurs in the
        underlying gss-api call.
        """
        
        input_token = GSSBuffer()
        input_token.set_buffer(inputTokenString)       
        if requests == None:
            requests = ContextRequests()
                 
        major, minor, token_buffer, handle = gssc.init_sec_context(self._handle,
                                                                   init_cred.get_handle(),
                                                                   target_name.get_handle(),
                                                                   input_token.get_handle(),
                                                                   requests.get_request_flags())

        buf = GSSBuffer(token_buffer)
        if major != gssc.GSS_S_COMPLETE and major != gssc.GSS_S_CONTINUE_NEEDED:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex

        if major == gssc.GSS_S_COMPLETE:
            self._handle = handle

        return major, minor, buf.get_string()
    
    def accept_context(self, acceptor_cred, inputTokenString):
        """
        **Arguments**
        
        - *acceptor_cred* is a GSSCred object which contains the credential that the acceptor
        will use to authenticate itself
        
        - *inputTokenString* is a token received from the corresponding init_sec call
        
              
        **Returns**
        major and minor status error codes from the GSSAPI.
        A string of tokens which represent credentials from the client who is initiating
        the context.
        
        **Raises:**
        
        A GSSContextException is raised if an error occurs in the
        underlying gss-api call.
        """
        
        input_token = GSSBuffer()
        input_token.set_buffer(inputTokenString)
        major, minor, token_buffer, handle, deleg_cred = gssc.accept_sec_context(self._handle,
                                                                    acceptor_cred.get_handle(),
                                                                    input_token.get_handle())
       
        delegated_cred = GSSCred(deleg_cred)
        buf = GSSBuffer(token_buffer)
        if major != gssc.GSS_S_COMPLETE and major != gssc.GSS_S_CONTINUE_NEEDED:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex

        if major == gssc.GSS_S_COMPLETE:
            self._handle = handle
            self._delegated_cred = delegated_cred
 
        return major, minor, buf.get_string()

    def get_mic(self, messageString):
        """
        A context must be created with init_context or accept_context before
        this call is made.

        **Arguments**

        - *messageString* - A Python string of the message you want to create a
        MIC (message integrity code) for.

        **Returns**
        A Python string of a MIC that was created is returned

        **Raises**
        A GSSContextException is raised if the mic cannot be created
        """
        
        buf = GSSBuffer()
        buf.set_buffer(messageString)
        major, minor, mic_token = gssc.get_mic(self._handle, buf.get_handle())
        if major != gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex

        retBuf = GSSBuffer(mic_token)
        return retBuf.get_string()

    def verify_mic(self, messageString, micString):
        """
        A context must be created with init_context or accept_context before
        this call is made
        
        **Arguments**

        - *messageString* - A Python string of the message you want verified

        - *micString* - A Python string that contains the cryptographic mic
        of the message

        **Returns**
        A Python int 1 if the mic has been verified, returns 0 if the MIC was
        incorrect

        **Raises**
        A GSSContextException is raised if the mic cannot be verified
        """

        msgBuf = GSSBuffer()
        msgBuf.set_buffer(messageString)
        micBuf = GSSBuffer()
        micBuf.set_buffer(micString)
        major, minor = gssc.verify_mic(self._handle, msgBuf.get_handle(),
                                       micBuf.get_handle())

        if major == gssc.GSS_S_COMPLETE:
            return 1
        if major == gssc.GSS_S_BAD_SIG:
            return 0
        
        errList = gssc.display_status(major, minor)
        ex = apply(GSSContextException, errList)
        raise ex

    def wrap(self, messageString):
        """
        A context must be created with init_context or accept_context before
        this call is made

        **Arguments**

        - *messageString* - A Python str of a message that will be wrapped

        **Returns**
        A Python string which is the wrapped message

        **Raises**
        A GSSContextException is raised if there is a problem wrapping the
        message
        """
        
        buf = GSSBuffer()
        buf.set_buffer(messageString)
        major, minor, output_buffer = gssc.wrap(self._handle, buf.get_handle())
        if major != gssc.GSS_S_COMPLETE: 
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex

        retBuf = GSSBuffer(output_buffer)
        return retBuf.get_string()

    def unwrap(self, messageString):
        """
        A context must be created with init_context or accept_context before
        this call is made

        - *messageString* - A Python str of a message that will be unwrapped

        **Returns**
        A Python string which is the unwrapped message

        **Raises**
        A GSSContextException is raised if there is a problem unwrapping the
        message
        """
        
        buf = GSSBuffer()
        buf.set_buffer(messageString)
        major, minor, output_buffer = gssc.unwrap(self._handle, buf.get_handle())
        if major != gssc.GSS_S_COMPLETE: 
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex

        retBuf = GSSBuffer(output_buffer)
        return retBuf.get_string()
    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################
    def get_handle(self):
        return self._handle

class ContextRequests:
    """
    Simple class that keeps track of requests made to security contexts
    The context can be requested to do the following:

    - delegation: allow the server to delegate your credential
    - mutual authentication: have the server authenticate to you  
    - replay: the replay of protected messages will be detected
    - sequence: out of sequence messages will be detected
    - confidentiality: wrapped messages will be confidential
    - integrity: integrity services provided
    - transfer: the security context may be transfered to another application
    """
    
    def __init__(self):    

        self._delegation = 0
        self._mutual = 0
        self._replay = 0
        self._sequence = 0
        self._confidentiality = 0
        self._integrity = 0
        self._transfer = 0

    def set_delegation(self):

        self._delegation = 1

    def set_mutual(self):
        self._mutual = 1

    def set_replay(self):

        self._replay = 1
    
    def set_sequence(self):

        self._sequence = 1

    def set_confidentiality(self):

        self._confidentiality = 1

    def set_integrity(self):

        self._integrity = 1
 
    def set_transfer(self):

        self._transfer = 1
    
    def get_request_flags(self):
        """ 
        return the value of the request flags
        """

        request = 0
        if self._delegation:
            request = request | gssc.GSS_C_DELEG_FLAG
        if self._mutual:
            request = request | gssc.GSS_C_MUTUAL_FLAG
        if self._replay:
            request = request | gssc.GSS_C_REPLAY_FLAG
        if self._sequence:
            request = request | gssc.GSS_C_SEQUENCE_FLAG
        if self._sequence:
            request = request | gssc.GSS_C_CONF_FLAG
        if self._integrity:
            request = request | gssc.GSS_C_INTEG_FLAG
        if self._transfer:
            request = request | gssc.GSS_C_TRANS_FLAG

        return request          

class GSSCred:
    """
    A class for the underlying gss_cred_id_t object.
    """

    def __init__(self, handle=None, free=1):
        """
        Constructs an instance.
        
        Each instance wraps a gss_cred_id_t object.
        
        **Arguments:**
        
        - *handle* A SWIG'ized pointer to a gss_cred_id_t object.
        
        - *free* Should the underlying gss_cred_id_t object be
        freed when an instance is destroyed.
        """
        
        self._activated = 0
        ret = utilc.globus_module_activate(gssc.get_module())
        if not ret == 0:
            raise GSSCredException, "Unable to initialize the module"
        self._activated = 1
        if handle == None:
            handle = gssc.get_no_cred_const_ptr()
        
        self._handle = handle
        self._free = free
        self._mechs = None
        return

    def __del__(self):
        """
        Destroys an instance.
        
        Deletes the underlying gss_cred_id_t object, and free's the
        memory if the instance owns the underlying handle.
        """
        
        if self._free:
            major, minor = gssc.release_cred(self._handle)
            if not major == gssc.GSS_S_COMPLETE:
                errList = gssc.display_status(major, minor)
                ex = apply(GSSCredException, errList)
                raise ex
        else:
            gssc.free_cred_ptr(self._handle)
            

        if self._activated:
            ret = utilc.globus_module_deactivate(gssc.get_module())
            if not ret == 0:
                ex = GSSCredException("Unable to deactivate module")
                raise ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################
    def get_handle(self):

        return self._handle   


    def acquire_cred(self, name= None, desired_mechs= None, usage=None, time_requested=gssc.GSS_C_INDEFINITE):
        """
        Acquires a credential, for example the one created by grid_proxy_init in /tmp/x509uid
        
        **Arguments**
        
        - *name* a GSSName object
        
        - *desired_mechs* a GSSMechs object 
              
        - *usage* a GSSUsage object
        
        - *time_requested* - an int which requests how long you would like the credential to
        be valid.
        Note:If you are using the globus_gssapi then this will have no effect. Their call just
        returns how long the credential that you created is valid(ie using grid-proxy-init)
        
        *Returns*
        None
        
        **Raises**
        A GSSCredException is thrown if there is a problem acquiring the credential
        """
        
        if name == None:
            name = GSSName()

        if desired_mechs == None:
            desired_mechs = GSSMechs()
            
        if usage == None:
            usage = GSSUsage()
        
        major, minor, cred, actual_mechs = gssc.acquire_cred(name.get_handle(),
                                               desired_mechs.get_handle(),
                                               usage.get_handle(),
                                               time_requested)
        
        if not major == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(major, minor)
            raise GSSCredException("Invalid user proxy!\nHave you grid-proxy-init'd recently?\n : "+str(errList)) 

        self._mechs = GSSMechs(actual_mechs) 
        self._handle = cred
        
    def export_external(self):
        """
        Saves the credential so it can be checkpointed and
        imported by import_external
        
        **Returns**
        A string pointing the location of the expored credential
        
        **Raises**
        A GSSCredException is raised if there is a problem in the underlying cred
        """

        credTuple = gssc.export_cred(self._handle, gssc.GSS_EXPORT_CRED_EXTERNAL)
        if not credTuple[0] == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(credTuple[0], credTuple[1])
            ex = apply(GSSContextException, errList)
            raise ex
        return credTuple[2]

    def import_external(self, credHandle):
        """
        Imports an external credential
        
        **Arguments**
        
        - *credHandle* - A string describing the type and location of the credential to import
        
        **Returns**
        A GSSCred Object
        
        **Raises**
        
        A GSSCredException is raised if the the credential cannot be imported
        """
        
        major, minor, cred, time = gssc.import_cred(gssc.GSS_EXPORT_CRED_EXTERNAL,
                                                    credHandle, 0)
        if not major == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSCredException, errList)
            raise ex
        return GSSCred(cred)
    
    def inquire_cred(self):
        """
        Retrieves information about the credential
              
        **Raises**
        a GSSCredException is thrown if there is a problem inquiring
        about the credential
              
        **Returns**
        A tuple of the lifetime in seconds and GSSName
        object

        **Raises**
        
        A GSSCredException is raised if the the credential cannot be imported
        """
        
        major, minor, life, name = gssc.inquire_cred(self._handle)
        if not major == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSCredException, errList)
            raise ex
        return (life, GSSName(name))

class GSICredentialAttr:
    """
    A class for the underlying globus_gsi_cred_handle_attr_t
    """
    def __init__(self, handle=None):
        """
        **Arguments:**

        - *handle* A SWIG'ized pointer to a globus_gsi_cred_handle_attr_t

        **Raises**
        A GSICredentialAttrException is raised if the module cannot
        be initialized or if it that attr can not be initialized
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gssc.get_module_credential())
        if not ret == 0:
            raise GSICredentialAttrException, "Unable to initialize the module"
        self._activated = 1
        if handle == None:
            ret, handle = gssc.cred_handle_attr_init()
            if not ret == 0:
                raise GSICredentialAttrException(handle)

        self._handle = handle

    def __del__(self):
        """
        Frees the memory associated with the handle

        **Raises**
        A GSICredentialAttrException is raised if the underlying handle
        can not be destroyed or the module can not be deactivaated
        """

        ret, err = gssc.cred_handle_attr_destroy(self._handle)
        if not ret == 0:
            raise GSICredentialAttrException(err)
        ret = utilc.globus_module_deactivate(gssc.get_module_credential())
        if not ret == 0:
            raise GSICredentialAttrException, "Unable to deactivate the module"

    def get_handle(self):
        """
        Retrives the underlying handle
        """

        return self._handle
    
class GSICredential:
    """
    A class for the underlying globus_gsi_cred_handle_t
    """
    def __init__(self, attr=None, handle=None):
        """
        **Arguments:**

        - *attr* A GSICredentialAttr object
        
        - *handle * A SWIG'ized pointer to a globus_gsi_cred_handle_t

        **Raises**
        A GSICredentialException is raised if the handle can not
        be created or the module can not be initialized
        """

        self._activated = 0
        ret = utilc.globus_module_activate(gssc.get_module_credential())
        if not ret == 0:
            raise GSICredentialException, "Unable to initialize the module"
        self._activated = 1
        if attr == None:
            attr = GSICredentialAttr()
        if handle == None:
            ret, handle = gssc.cred_handle_init(attr.get_handle())
            if not ret == 0:
                raise GSICredentialException(handle)

        self._handle = handle

    def __del__(self):
        """
        frees the memory associated with the handle

        **Raises**
        A GSICredentialException is raised if the underlying handle
        can not be destroyed or the module can not be deactivaated
        """
        
        ret, err = gssc.cred_handle_destroy(self._handle)
        if not ret == 0:
            raise GSICredentialException(err)

        ret = utilc.globus_module_deactivate(gssc.get_module_credential())
        if not ret == 0:
            raise GSICredentialException, "Unable to deactivate the module"

    def read(self):
        """
        reads the credential information into the handle, analogous
        to acquire_cred in the GSSCred class

        **Raises**
        A GSICredentialException is raised if the credential information
        can not be retrieved
        """

        ret, err = gssc.cred_read(self._handle)
        if not ret == 0:
            raise GSICredentialException(err)

    def get_key(self):
        """
        retrieves the private key from the underlying handle
        You must call read, before you call this method.

        **Returns**
        a EVP_Key object which represents the private key

        **Raises**
        A GSICredential Exception is raised if the private key
        can not be retrieved
        """

        ret, pkey = gssc.cred_get_key(self._handle)
        if not ret == 0:
            raise GSICredentialException(pkey)

        return EVP_Key(pkey)
    
    def get_cert(self):
        """
        retrieves the user cert from the underlying handle
        You must call read, before you call this method.

        **Returns**
        a X509 object which represents the user cert

        **Raises**
        A GSICredential Exception is raised if the user cert
        can not be retrieved
        """

        ret, cert = gssc.cred_get_cert(self._handle)
        if not ret == 0:
            raise GSICredentialException(cert)

        return X509(cert)

    def get_cert_chain(self):
        """
        retrieves the cert chain from the underlying handle
        You must call read, before you call this method.
        The cert chain does not contain the user cert
        
        **Returns**
        a STACK_X509 object which represents the cert chain

        **Raises**
        A GSICredential Exception is raised if the cert chain
        can not be retrieved
        """

        ret, cert_chain = gssc.cred_get_cert_chain(self._handle)
        if not ret == 0:
            raise GSICredentialException(cert_chain)
        stackFactory = STACK_X509_Factory()
        stack = stackFactory.createStackFromPointer(cert_chain)
        return stack
    
    
class STACK_X509_Factory(PyGlobusFactory):
    """
    Creates STACK_X509 instances
    """
    def __init__(self): 

        self.register("createStackX509", self.createStackFromPointer)
        self.register("createStackX509EncodedSequence", self.createStackFromEncodedSequence)
        self.register("createStackX509EncodedString", self.createStackFromEncodedString)
       
    def createStackFromPointer(self, stack):

        return STACK_X509(stack)
    
    def createStackFromEncodedSequence(self, encoding):
        """
        Creates a new STACK_X509 instance

        **Arguments**
        - *encoding*
        Is a string of DER encoded sequence of X509 certificates
        """

        stack = gssc.make_stack_from_der_sequence(encoding)
        return STACK_X509(stack)

    def createStackFromEncodedString(self, encoding):
        """
        Creates a new STACK_X509 instance

        **Arguments**
        - *encoding*
        Is a string of DER encoded certificates, not a sequence
        """

        stack = gssc.make_stack_from_der_string(encoding)
        return STACK_X509(stack)
    
class STACK_X509:
    
    def __init__(self, handle, free=1):
        """
        Use the STACK_X509_Factory to instantiate objects of this
        class.
        """

        self._handle = handle
        self._free = free

    def __del__(self):

        if self._free:
            gssc.free_stack(self._handle)

    def get_der_encoding(self):
        """
        generates a sequence DER encoded 
        of the X509 STACK
        
        **Returns**
        A string encoded representation of the X509 STACK
        """

        ret, encoding = gssc.get_der_encoding_stack(self._handle)
        if not ret == 0:
            raise GSICredentialException("Problem DER encoding certificate chain")
        return encoding
    
    def insert(self, x509, position):
        """
        Inserts a X509 object into the STACK

        **Arguments**

        - *x509* A X509 instance

        - *position* An int which is where the X509 will be placed in
        the stack

        """
        
        gssc.insert_X509(self._handle, x509.get_handle(), position)

    def pop(self):
        """
        Pops the first X509 object of the stack
        """
        
        x509 = gssc.stack_pop(self._handle)
        return X509(x509)

    
class X509:
    """
    A class for underlying X509 objects
    """
    def __init__(self, handle=None, free=1):
        """
        **Arguments**

        - *handle* A SWIG'ized pointer to a X509 object

        - *free* an int specifing whether the instance should
        be delegated the responsibility of freeing the memory
        associated with the handle

        **Raises**
        A X509Exception is raised if the module can not be
        initialized or the handle can not be created
        """
        
        self._activated = 0
        ret = utilc.globus_module_activate(gssc.get_module_openssl())
        if not ret == 0:
            raise X509Exception, "Unable to initialize the module"
        self._activated = 1
        if handle == None:
            handle = gssc.new_X509()
            
        self._handle = handle
        self._free = free

    def __del__(self):
        """
        frees the memory associated with this handle
        """
        
        if self._free:
            gssc.free_X509(self._handle)

    def get_public_key(self):
        """
        retrieves the public key from the underlying handle
        **Returns**
        A EVP_Key instance which represents the public key
        """

        pkey = gssc.get_pubkey(self._handle)
        return EVP_Key(pkey)
    
    def get_der_encoding(self):
        """
        returns a DER encoding of the underlying handle
        """

        encoding = sslc.get_der_encoding_x509(self._handle)
        return encoding

    def sign(self, pkey):
        """
        signs the certificate

        **Arguments**

        - *pkey* an EVP_PKEY object, which will be used
        to sign the certificate

        **Returns**
        an int which is the length of the signature
        """

        length = gssc.sign_X509(self.get_handle, pkey.get_handle())
        return length
    
    def get_handle(self):
        
        return self._handle
    
class X509_Request:
    """
    A simple class for the underlying X509_REQ* objects.
    
    """

    def __init__(self, handle=None, free=1):
        """
        
        Each instance wraps X509_REQ* objects.
        
        **Arguments:**
        
        - *handle* A SWIG'ized pointer to a X509_REQ* object.
        
        - *free* Should the underlying objects be
        freed when an instance is destroyed.
        """
        
        if handle == None:
            handle = gssc.new_X509_REQ()
        self._handle = handle
        self._free = free
        return

    def __del__(self):
        """
        Deletes the underlying gss_cred_id_t object, and free's the
        memory if the instance owns the underlying handle.
        """
        
        if self._free:
            gssc.free_X509_REQ(self._handle)
        return
    
class EVP_Key:
    """
    A class for the underlying  EVP_PKEY* objects.
    
    """
    
    def __init__(self, handle=None, free=1):
        """
              
        Each instance wraps X509_REQ* EVP_PKEY* objects.
        
        **Arguments:**
        
        - *handle* A SWIG'ized pointer to a EVP_PKEY* object.
        
        - *free* Should the underlying objects be
        freed when an instance is destroyed.
        """
        
        if handle == None:
            handle = gssc.new_EVP_PKEY()

        self._handle = handle
        self._free = free
        return

    def __del__(self):
        """
        Deletes the underlying gss_cred_id_t object, and free's the
        memory if the instance owns the underlying handle.
        """
        
        if self._free:
            gssc.free_EVP_PKEY(self._handle)
        return

    def sign(self, data):
        """
        Signes the data using the underlying key handle,

        You MUST be using a RSA key

        The data MUST be hashed using sha1
        
        **Arguments**

        - *data* A string of data to sign

        **Returns**
        A signature of the data is returned

        **Raises**
        An EVP_KeyException is raised if the data can not
        be signed
        """
        ret, signedData = gssc.sign_rsa(self._handle, data)
        if not ret == 1:
            raise EVP_KeyException(signedData)
        return signedData

    def verify(self, data, signature):
        """
        verifies the signature is a valid signature of the data

        You MUST be using a RSA key

        The data MUST be hashed using sha1
        
        **Arguments**

        - *data* A string of the data that was signed
        
        - *signature* A signature created by signing the data

        **Returns**
        returns True if the signature was verified
        returns False otherwise

        **Raises**
        An EVP_KeyException is raised if there is an error
        calling verify
        """
        ret = gssc.verify_rsa(self._handle, data, signature)
        if ret == -1:
            raise EVP_KeyException("Error calling verify_rsa")

        if ret:
            return True
        else:
            return False
        
    def get_handle(self):
        
        return self._handle
    
class GSSName:
    """
       A class for the underlying gss_name_t object.
       """

    def __init__(self, handle=None, free=1):
        """
        **Arguments:**
        
        - *handle* An optional argument that, if present, should
        contain a SWIG'ized pointer to a gss_name_t object.
        It defaults to a pointer to a  GSS_C_NO_NAME
        
        - *free* Should the underlying gss_name_t object be
        freed when an instance is destroyed.
        
        **Returns**
        None
        
        **Raises**
        None
        """
        
        if handle == None:
            handle = gssc.get_default_name()
        self._handle = handle
        self._free = free
        return

    def __del__(self):
        """
        Destroys an instance.
        
        Deletes the underlying gss_name_t object, and free's the
        memory if the instance owns the underlying handle.
        
        **Raises:**
        
        A GSSNameException is raised if an error occurs
        in the underlying gss-api call.
        """
        
        if not self._handle == None:
            if self._free:
                major, minor = gssc.release_name(self._handle)
            if not major == gssc.GSS_S_COMPLETE:
                errList = gssc.display_status(major, minor)
                ex = apply(GSSContextException, errList)
                raise ex
        return

    def display(self):
        """
        Returns the name as a string.
        
        **Raises:**
        
        A GSSNameException is raised if an error occurs
        in the underlying gss-api call.
        
        **Returns:**
        A string containing the name.
        """
        
        major, minor, name = gssc.display_name(self._handle)
        if not major == gssc.GSS_S_COMPLETE:
            errList = gssc.display_status(major, minor)
            ex = apply(GSSContextException, errList)
            raise ex
        return name

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################
    def get_handle(self):
        return self._handle


class GSSMechs:
    """
    A  class for the underlying gss_OID_set object.
    
    """

    def __init__(self, handle=None, free=1):
        """
        **Arguments**
        
        - *handle* A SWIG'ized pointer to a gss_OID_set object
        The default case creates a new gss_OID_set object for you
        
        - *free* If set to zero then the caller is reponsible for freeing the
        hande object.
        The default case frees the memory of the handle when the
        __del__ method is called
        
        **Raises**
        GSSMechsException if there is a problem initializing the
        underlying structure
        """
        
        if handle == None:
            major, minor, handle = gssc.create_empty_oid_set()
            if not major == gssc.GSS_S_COMPLETE:
                errList = gssc.display_status(major, minor)
                ex = apply(GSSMechsException, errList)
                raise ex
        self._handle = handle
        self._free = free

    def __del__(self):
        """
        Deletes the object and frees the memory of the handle variable if
        free is set
        
        **Raises**
        GSSMechsException if there is a problem freeing the
        underlying structure
        """
        
        if self._free:
            major, minor = gssc.release_oid_set(self._handle)
            if not major == gssc.GSS_S_COMPLETE:
                errList = gssc.display_status(major, minor)
                ex = apply(GSSMechsException, errList)
                raise ex 
            
    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################        
    def get_handle(self):
        return self._handle

class GSSUsage:
    """
    A  class for the underlying gss_cred_usage_t object.
    """
    
    def __init__(self, handle=None, free=1):
        """
        Defaults to GSS_C_BOTH
        """
        
        if handle == None:
            handle = gssc.get_usage_both()
        self._handle = handle
        self._free = free

    def __del__(self):
        if self._free:
            gssc.free_usage(self._handle)

    def set_usage_initiate(self):
        """
        Sets the underlying handle to a GSS_C_INITIATE handle
        """
        
        if self._free:
            gssc.free_usage(self._handle)

        self._handle = gssc.get_usage_initiate()

    def set_usage_accept(self):
        """
        Sets the underlying handle to a GSS_C_ACCEPT handle
        """
        
        if self._free:
            gssc.free_usage(self._handle)

        self._handle = gssc.get_usage_accept()

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################  
    def get_handle(self):
        return self._handle
        
class GSSBuffer:
    """
    A class for the underlying gss_buffer_t object
    """
    

    def __init__(self, handle=None, free=1):
        """
        defaults to GSS_C_NO_BUFFER
        """
        
        if handle == None:
            handle = gssc.get_empty_buffer()
        self._handle = handle
        self._free = free


    def __del__(self):
        if self._free:
            major, minor = gssc.free_buffer(self._handle)
                

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed if
    # self._free is true.
    #######################################################################  
    def get_handle(self):
        return self._handle

    def set_buffer(self, contents):
        """
        **Args**
        
        - *contents* a string which you wish to set the underlying handle to
        
        **Returns**
        None
        
        **Raises**
        A GSSBufferException is raised if the contents are None 
        """

        if contents == None:
            raise GSSBufferException("setting buffer to None will cause a segmentaion fault")
 
        gssc.free_buffer(self._handle)
        self._handle = gssc.set_buffer(contents)


    def get_string(self):
        """
        Returns the contents of the buffer as a python string
        """
        
        return gssc.get_buffer_string(self._handle)
              
