############################################################################
# Keith Beattie, LBNL
# Copyright (c) 2003, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of
# any required approvals from the U.S. Dept. of Energy).
# All rights reserved. 
###########################################################################

''' Sample basic counter client, to be run once the counter service
has been started. '''

import getopt, httplib, sys
from cmd import Cmd
from Ft.Xml import Domlette, XPath
from xml.ns import SOAP, SCHEMA

def usage():
    useStr = """Usage: testInitSecContext_xml.py [-u] -p <port> -h <host>"""
    print useStr


def getArgs():
    
    if len(sys.argv) == 1:
        usage()
        sys.exit(-1)
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h:p:")
    except getopt.GetoptError:
        usage()
        sys.exit(-1)
    for o, a in opts:
        if o == "-u":
            usage()
            sys.exit(-1)
        elif o == "-p":
            port = a
        elif o == "-h":
            host = a
        else:
            usage()
            sys.exit(-1)


    return host, int(port) 


class Engine:
    ''' Simple counter client engine - does all the work sending/receiving messages. '''

    # The following dict is used when constructing the SOAP messages, holding
    # namespace definitions and other values needed.
    xmlDict = {}
    xmlDict['soapenv']    = SOAP.ENV
    xmlDict['xsd']        = SCHEMA.XSD3
    xmlDict['xsi']        = SCHEMA.XSI3
    xmlDict['counter_ns']      = "http://ogsa.globus.org/samples/counter"

    # The formatted template string for add messages
    add_tmpl = '''<soapenv:Envelope xmlns:soapenv="%(soapenv)s"
                  xmlns:xsd="%(xsd)s"
                  xmlns:xsi="%(xsi)s">
 <soapenv:Body>
  <add xmlns="%(counter_ns)s">
   <value xsi:type="xsd:int" xmlns="">%(val)d</value>
  </add>
 </soapenv:Body>
</soapenv:Envelope>'''

    
    def __init__(self, host, port):
        ''' Construct, by storing the host and port of the container. '''
        self._host = host
        self._port = port
        self._verbose = False
        self._hash = None

    def add(self, val):
        ''' add(val) -> count

        Send an add message and return the current count.'''

        # Set our parts in the xmlDict before forming the xml message
        self.xmlDict['val'] = val

        # Fill in the formatted message template
        send_body = self.add_tmpl % self.xmlDict

        # Construct the message and send it.
        req = httplib.HTTP(self._host, self._port)
        req.putrequest("POST", "/ogsa/services/samples/counter/basic/CounterFactoryService/hash-%s" % (self._hash))
        req.putheader("Content-Type", 'text/plain; charset="utf-8"')
        req.putheader("Accept", "application/soap+xml, application/dime, multipart/related, text/*")
        req.putheader("User-Agent", "Axis/1.1")
        req.putheader("Cache-Control", "no-cache")
        req.putheader("Pragma", "no-cache")
        req.putheader("Host", "%s:%s" % (self._host, self._port))
        req.putheader("SOAPAction", "http://ogsa.globus.org/samples/counter#add")
        req.putheader("Content-Length", str(len(send_body)))
        req.endheaders()
        req.send(send_body)

        if self._verbose: print "  sent %s byte body" % (len(send_body))
        
        (status_code, message, reply_headers) = req.getreply()
        reply_body = req.getfile().read()

        if self._verbose: print "  status: code %d, message %s. Reply body %s bytes" % (status_code, message, len(reply_body))

        # Now, parse the reply_body
        dom = Domlette.NonvalidatingReader.parseString(reply_body, 'memory://')

        # Get value
        ctx = XPath.Context.Context(dom, processorNss={'ns':self.xmlDict['counter_ns']})
        val_el = XPath.Evaluate("//ns:returnValue", context=ctx)[0]
        
        return int(val_el.firstChild.data)



class Interpreter(Cmd):

    intro = '''  Welcome to the basic counter client: ? = help, Cntl-d (or EOF) quits.'''

    def __init__(self, engine):
        Cmd.__init__(self)

        self._engine = engine
        self.prompt = 'BasicCounter[%s:%d] $ ' % (self._engine._host, self._engine._port)


    # Add command
    def do_add(self, line):

        if not self._engine._hash:
            print "Hash not set yet."
            return

        try:
            val = int(line)
        except ValueError:
            print "add: '%s' does not appear to be a number" % (line)
            return

        print "adding: %d" % (val)
        print "new value: %d" % (self._engine.add(val))
        
    def help_add(self):
        print "'add numb' to add numb to the counter"

    # hash command
    def do_hash(self, line):
        if line:
            print "Setting hash to: %s" % (line)
            self._engine._hash = line
        else:
            print "Current hash is: %s" % (self._engine._hash)

    def help_hash(self):
        print "'hash [new hash value]' - Set the hash value used in the POST"
        print "Current hash is: %s" % (self._engine._hash)


    # help command
    def help_help(self):
        print "type 'help <command>' or '? <command>' to print info on <command>"


    # EOF/quit command
    def do_EOF(self, line):
        print "EOF"
        sys.exit()

    def help_EOF(self):
        print "EOF (or Cntl-d) will exit this program"


    # verbose toggle command
    def do_verbose(self, line):
        self._engine._verbose = not self._engine._verbose
        print "Verbose flag toggled %s" % (self._engine._verbose)

    def help_verbose(self):
        print "Toggle verbose flag.  Currently %s" % (self._engine._verbose)


def main():
    host, port = getArgs()
    eng = Engine(host, port)
    interp = Interpreter(eng)
    interp.cmdloop()

if __name__ == "__main__":
    main()
