############################################################################
# Matt Rodriguez, LBNL
# Keith Beattie, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################

import base64, commands, cStringIO, getopt, socket, sys

from pyGlobus.security import GSSCred
from pyGlobus.security import GSSContext
from pyGlobus.security import GSSName, GSSMechs, GSSUsage,GSSBuffer
from pyGlobus.security import GSSContextException
from pyGlobus import gssc

from Ft.Xml import Domlette, XPath
from xml.ns import SOAP, SCHEMA

# The following dict is used when constructing xml messages, holding
# namespace definitions and other values needed in the xml.
xmlDict = {}
xmlDict['soapenv']    = SOAP.ENV
xmlDict['xsd']        = SCHEMA.XSD3
xmlDict['xsi']        = SCHEMA.XSI3
xmlDict['xmlns']      = "http://ogsa.globus.org/security/authentication"
xmlDict['mech-type']  = "http://www.globus.org/2002/04/xmlenc#gssapi-gsi"
xmlDict['ns1']        = "http://ogsa.globus.org/security/authentication/types"

# The formatted template string for initTokenExchange messages
initTokenExch_tmpl = '''<soapenv:Envelope xmlns:soapenv="%(soapenv)s"
                  xmlns:xsd="%(xsd)s"
                  xmlns:xsi="%(xsi)s">
  <soapenv:Body>
    <initTokenExchange xmlns="%(xmlns)s">
      <ns1:initContextToken ns1:continue-needed="%(continue_needed)s"
                            ns1:mechanism-type="%(mech-type)s"
                            xmlns:ns1="%(ns1)s">
        <ns1:base64Token>%(b64Token)s</ns1:base64Token>
        <ns1:context-id>%(context-id)s</ns1:context-id>
      </ns1:initContextToken>
    </initTokenExchange>
  </soapenv:Body>
</soapenv:Envelope>
'''

# The formatted template string for continueTokenExchange messages
continueTokenExch_tmpl = '''<soapenv:Envelope xmlns:soapenv="%(soapenv)s"
                  xmlns:xsd="%(xsd)s"
                  xmlns:xsi="%(xsi)s">
  <soapenv:Body>
    <continueTokenExchange xmlns="%(xmlns)s">
      <ns1:contextToken ns1:continue-needed="%(continue_needed)s"
                        xmlns:ns1="%(ns1)s">
        <ns1:base64Token>%(b64Token)s</ns1:base64Token>
        <ns1:context-id>%(context-id)s</ns1:context-id>
      </ns1:contextToken>
    </continueTokenExchange>
  </soapenv:Body>
</soapenv:Envelope>
'''

def initContext(host,port):

    name = GSSName()
    mechs = GSSMechs()
    usage = GSSUsage()
    usage.set_usage_initiate()
    cred = GSSCred()
    cred.acquire_cred(name, mechs, usage)
    lifetime, cred_name = cred.inquire_cred()
    print "Cred:"
    print "      name: %s" % (cred_name.display())
    print "  lifetime: %d" % (lifetime)

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    print "Connecting to: %s:%s" % (host, port)
    sock.connect((host,port))
    sockf = sock.makefile()

    ctx = GSSContext()

    uuid = commands.getoutput("uuidgen").split("-")
    context_id = "%s-%s-%s-%s-%s" % ("0"*8, uuid[1], uuid[2], "0"*8, "0"*12)

    # Then send the initTokenExchange message
    major, minor, token = ctx.init_context(cred, cred_name, "")
    print "Sending initTokenExch:"
    continue_needed, msg_size = sendToken(sockf, initTokenExch_tmpl, major, token, context_id)
    print "  sent %s bytes, cont=%s, null_token=%s" % (msg_size, continue_needed, token==None)

    # And the get the initTokenExchangeResponce
    print "Reading initTokenExchangeResponce:"
    token, continue_needed, msg_size = readToken(sockf, context_id)
    print "  read %s bytes, cont=%s, null_token=%s" % (msg_size, continue_needed, token==None)

    # Now start the loop of continueTokenExchange messages.
    while continue_needed:

        major, minor, token = ctx.init_context(cred, cred_name, token)

        print "Sending continueTokenExchange:"
        continue_needed, msg_size = sendToken(sockf, continueTokenExch_tmpl, major, token)
        print "  sent %s bytes, cont=%s, null_token=%s" % (msg_size, continue_needed, token==None)

        print "Reading continueTokenExchangeResponce:"
        token, continue_needed, msg_size = readToken(sockf)
        print "  read %s bytes, cont=%s, null_token=%s" % (msg_size, continue_needed, token==None)


    print "=== Context established ===\n"

    return sockf, ctx


def sendToken(sockf, msg_tmpl, major, token, context_id=None):
    ''' sendToken(sockf, msg_tmpl, major, token) -> continue_needed, msg_size

    Construct a message based on msg_tmpl, with the given token, using
    major to determine continue-needed flag, and write the message to
    the given (file) socket.  If context_id is set, use that as the
    context_id.  Returns the continue_needed flag and the number of
    bytes written.'''

    # Set our parts in the xmlDict before forming the xml message
    xmlDict['b64Token'] = base64.encodestring(token)

    if major == gssc.GSS_S_CONTINUE_NEEDED:
        xmlDict['continue_needed'] = 'true'
    else:
        xmlDict['continue_needed'] = 'false'

    if context_id:
        xmlDict['context-id'] = context_id

    # Fill in the formatted message template
    msg = msg_tmpl % xmlDict

    sockf.write(msg)
    sockf.flush()
    
    return xmlDict['continue_needed'] == 'true', len(msg)


def readToken(sockf, context_id=None):
    ''' readToken(sockf) -> token, continue_needed, msg_size

    Read a message from the given socket, return the binary token, the
    continue needed flag, and the size of the xml message read.  If
    context_id is set, then it will be read from the received xml and
    set in the xmlDict.'''

    # Read from the socket until a line ending with
    # '</soapenv:Envelope>' is read.

    msg = cStringIO.StringIO()
    line = ''
    while not line.endswith('</soapenv:Envelope>\n'):
        line = sockf.readline()
        msg.write(line)

    msg_str = msg.getvalue()
    dom = Domlette.NonvalidatingReader.parseString(msg_str, 'memory://')

    ctx = XPath.Context.Context(dom, processorNss={'ns1':xmlDict['ns1']})

    # Get the continue_needed attribute
    cn_attr = XPath.Evaluate("//child::*/@ns1:continue-needed", context=ctx)[0]
    cont_needed = (cn_attr.value.lower() == 'true')

    # Get the context_id if we're looking to set it
    if context_id:
        context_id_el = XPath.Evaluate("//ns1:context-id", context=ctx)[0]
        xmlDict['context-id'] = context_id_el.firstChild.data

    b64Token_el = XPath.Evaluate("//ns1:base64Token", context=ctx)[0]

    if b64Token_el.firstChild:
        token = base64.decodestring(b64Token_el.firstChild.data)
    else:
        token = None

    return token, cont_needed, len(msg_str)


def testSecContext(sockf, ctx):
    ''' Test the given security context by exchanging an encrypted message.'''

    xml_tmpl = '''<msg>%s</msg>
'''

    print "Testing message exchange:"

    msg = "The bodies are in the trunk"
    print "  msg: %s" % (msg)

    cyMsg = ctx.wrap(msg)
    b64CyMsg = base64.encodestring(cyMsg).rstrip()
    print "  base64'd crypto msg: %s" % (b64CyMsg)
    
    xml_msg = xml_tmpl % (b64CyMsg)
    sockf.write(xml_msg)
    sockf.flush()
    print "Sent %s bytes" % (len(xml_msg))

    print "Reading mic:"
    xml_mic = sockf.readline()
    print "  read %s bytes" % (len(xml_mic))
    dom = Domlette.NonvalidatingReader.parseString(xml_mic, 'memory://')
    b64mic = dom.firstChild.firstChild.data
    print "  base64' mic: %s" % (b64mic)

    try:
        verified = ctx.verify_mic(msg, base64.decodestring(b64mic))
    except GSSContextException, ex:
        print "Problem verifing mic"
        raise ex

    if verified:
        print "Message verification succeded"
    else:
        print "Message verification failed"



def usage():
    useStr = """Usage: testInitSecContext_xml.py [-u] -p <port> -h <host>"""
    print useStr


def getArgs():
    
    if len(sys.argv) == 1:
        usage()
        sys.exit(-1)
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h:p:")
    except getopt.GetoptError:
        usage()
        sys.exit(-1)
    for o, a in opts:
        if o == "-u":
            usage()
            sys.exit(-1)
        elif o == "-p":
            port = a
        elif o == "-h":
            host = a
        else:
            usage()
            sys.exit(-1)

            
    return host, int(port) 


def main():
    host,port = getArgs()
    sockf, ctx = initContext(host,port)
    testSecContext(sockf, ctx)

if __name__ == "__main__": main()
