############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
A module containing utility functions, e.g. buffers, module activation, etc.
"""

from pyGlobus import utilc
import exceptions
import os
import re
import socket

class GlobusException(exceptions.Exception):
    """
    A simple exception class.

    Used as a super-class for all exceptions used in pyGlobus. This will
    allow simpler exception handling.

    **Extends:**
    
        *exceptions.Exception* The baseclass for all exceptions in python.
    """
    pass

class GlobusBufferException(GlobusException):
    """
    A simple subclass of GlobusException.

    Handles errors in the buffer class.
    """
    pass

class CallbackQueueException(GlobusException):
    """
    A simple subclass of GlobusException.

    Handles errors in the CallbackQueue class.
    """
    pass

def getHostname():
    """
    Return the local hostname.
    This uses the pyGlobus mechanism when possible, in order
    to get a hostname that Globus will be happy with.
                                                                                
    """
                                                                                
    ret, host = utilc.get_hostname(256)
    if ret != 0:
        host = socket.getfqdn()
    return host


def getEnv(variable):
    """
    Calls globus_libc_getenv, this is useful if you are using
    win32, otherwise you can use the os module to access your
    environment

    **Arguments**

    - *variable* the variable who's value you want to retrieve
    from your environment

    **Returns**
    returns a string of the value of the variable, or None if
    the variable is not in your environment

    **Raises**
    Nothing
    """
    return utilc.py_getenv(variable)

def setEnv(variable, value):
    """
    Calls globus_libc_setenv, this is useful if you are using
    win32, otherwise you can use the os module to modify your
    environment

    **Arguments**

    - *variable* the variable who's value you want to modify

    - *value* the value you want to set the variable to
    
    **Returns**
    None

    **Raises**
    Nothing
    """
    utilc.py_setenv(variable, value)

def unSetEnv(variable):
    """
    Calls globus_libc_unsetenv, this is useful if you are using
    win32, otherwise you can use the os module to modify your
    environment

    **Arguments**

    - *variable* the variable who's value you want to unset

    **Returns**
    None

    **Raises**
    Nothing
    """
    utilc.py_unsetenv(variable)
    
class Off_t:
    """
    A wrapper class for globus_off_t pointers

    """
    def __init__(self, handle):
        self._handle = handle

    def __del__(self):
        """ 
        frees the memory to the underlying handle
        """
        utilc.free_off_t(self._handle)

    def getLong(self):
        """ 
        Returns a python long representation of the
        underlying off_t pointer 
        """
        return utilc.convert_off_t(self._handle)

class Abstime:
    """
    A wrapper class for the globus_abstime_t type
    Handles the memory and allows some access to the
    underlying timeval struct
    """

    def __init__(self, handle):
        self._handle = handle

    def __del__(self):
        """ 
        frees the memory associated with the abs_time_t handle
        """
        utilc.free_abstime(self._handle)

    def getSecs(self):
        """
        Returns a long of the underlying seconds of the underlying 
        abs_time_t handle
        """
        return utilc.getSecs(self._handle)

    def getNSecs(self):
        """
        Returns a long of the underlying u_seconds of the underlying 
        abs_time_t handle 
        """
        return utilc.getNSecs(self._handle)

   

    
class Buffer:
    """
    A wrapper class for globus_byte_t arrays.

    Allows the manipulation of globus_byte_t arrays. This class malloc's the
    underlying memory in the constructor, and free's it during destruction. It
    supports returning the underlying memory as a python string. It also supports
    the readline and readlines methods.
    """
    def __init__(self, size):
        """
        Contruct a Buffer object.        
        This methods causes underlying memory to be malloc'd.
        
        **Arguments**        
            - *size* The size of the buffer to create.

        **Raises:**
        
        A GlobusBufferException is thrown if unable to create
        the buffer object.
        """
        ret = utilc.globus_module_activate(utilc.get_module())
        if not ret == 0:
            raise GlobusBufferException, "Unable to deactivate the common module"  
        self._handle, self._size = utilc.create_buffer(size)
        self._str = None
        return

    def __del__(self):
        """
        Destroys the Buffer object.

        Free's the underlying memory.
        """
        utilc.destroy_buffer(self._handle)
        ret = utilc.globus_module_deactivate(utilc.get_module())
        if not ret == 0:
            print "ERROR!!!: Unable to deactivate the common module"
        return
   
    
    def __len__(self):
        """
        Get the size of the Buffer object.

        **Returns:**        
        An int containing the size of the Buffer object.
        """
        return self._size
    
    def get_handle(self):
        """
        Get the underlying pointer to the globus_byte_t array.

        **Returns:**        
        A SWIG'ized pointer to the underlying byte array.

        **XXX** This handle is only valid as long as this object exists.  When
        the object is destroyed, the memory this handle points to is freed.
        """
        return self._handle
    
    def set_handle(self, handle):
        """ Sets a globus_byte_t array to a Buffer object"""
        self._handle = handle
    
    def get_size(self):
        """
        Get the size of the Buffer object.

        **Returns:**
        
        An int containing the size of the Buffer object.
        """
        return self._size
    
    def as_string(self, size):
        """
        Convert the underlying globus_byte_t array into a Python string.

        **Arguments:**
        
        *size* The number of bytes to convert.
            
        **Returns:**
        
        A string containg the contents of the Buffer.
        """
        return utilc.buffer_to_string(self._handle, size)

    def readline(self):
        """
        Return a single line as a string.

        **Returns:**        
        A string containing the next line in the Buffer.
        """
        if self._str == None:
            readStr = utilc.buffer_to_string(self._handle, self._size)
        else:
            readStr = self._str
        i = readStr.find('\n')
        if i >= 0:
            self._str = readStr[i+1:]
            line = readStr[:i]
        else:
            self._str = ""
            line = readStr
        return line

    def readlines(self):
        """
        Return a list containing each line in the Buffer.

        **Returns:**        
        A list of strings containing each line in the Buffer.
        """
        lineList = []
        while 1:
            line = self.readline()
            if line == "":
                break
            lineList.append(line)
            continue
        return lineList
    
class PyGlobusFactory:
    """
    Generic factory class that registers a function, and default arguments 
    to that function.
 
    The registered function may also be called with arguments
    """
    def register(self, methodName, constructor, *args, **kargs):
        """
        register a constructor with the default arguments and keyword arguments you
        you use when you call it
        """
        _args = [constructor]
        _args.extend(args)
        setattr(self, methodName, apply(Functor, _args, kargs))
        
    def unregister(self, methodName):
        """unregister a constructor"""
        delattr(self, methodName)

class Functor:
    """
    Helper class for pyGlobusFactory, reponsible for the calling of the
    registered function
    """
    def __init__(self, function, *args, **kargs):
        assert callable(function), "function should be a callable obj"
        self._function = function
        self._args = args
        self._kargs = kargs
        
    def __call__(self, *args, **kargs):
        """call function"""
        _args = list(self._args)
        _args.extend(args)
        _kargs = self._kargs.copy()
        _kargs.update(kargs)
        return apply(self._function, _args, _kargs)


class CallbackQueue:
    """
    Class that contains a queue of callbackStruct SWIG'ized pointers
    """
    def __init__(self):
        self.__queue = {}

    def __del__(self):
        for cb in self.__queue.values():
            del cb

    def remove_callback(self, cbHandle):
        
        if self.__queue.has_key(str(cbHandle)):
            del self.__queue[str(cbHandle)]
        else:
            ex = CallbackQueueException("Attempted to remove a callback handle not in the queue")
            raise ex
    
    def append_callback(self, cbHandle, cb):
        self.__queue.update({str(cbHandle):cb})
    
    
class Callback:
    pass

class CallbackDefault(Callback):
    def __init__(self, handle):
        self.__handle = handle

    def __del__(self):
        utilc.free_callback1(self.__handle)
        
class CallbackData(Callback):
    def __init__(self, handle):
        self.__handle = handle

    def __del__(self):
        utilc.free_callback4(self.__handle)
    
class GlobusXIODebugParseError(Exception):
    expected = "<level>  [,[[#]<file name>][,<flag>[,<timestamp_levels>]]"
    def __init__(self, val):
        Exception.__init__(self,"input string '%s' does not match '%s'" %
                           (val, self.expected))

class GlobusXIODebug:
    # Names of levels (these are standard ones )
    ERROR, WARNING, TRACE, INTERNAL_TRACE, INFO, STATE, INFO_VERBOSE = \
           map(lambda x: 2**x, range(0, 7))
    # Some additional levels
    PERF1, PERF2 = map(lambda x: 2**x, range(7, 9))
    # Special filename meaning stdout
    STDOUT_FILENAME = '-'
    # RE for contents of env var
    _pat = re.compile("(?P<lvl>\d+)\s*(?:,\s*(?P<ow>#)?" + 
                      "\s*(?P<file>[^,]+))?")
    # alternate RE for file indirection
    _fpat = re.compile("\d*\s*([^,\d]+)")

    def __init__(self, val=None):
        """Parse the environment variable setting Globus XIO debug,
        setting attributes for 'level' and 'filename' 

        The name of the variable is GLOBUS_XIO_DEBUG
        The format is:
           <level>  [,[[#]<file name>][,<flag>[,<timestamp_levels>]]
        Or at least that is the official format. However, timestamp levels
        is not documented, and "flags" seems irrelevant here, so
        only level and file name are really used.

        If provided 'val' is used as the value to parse,
        otherwise it is fetched from the environment.

        ==> Extension:
            If GLOBUS_XIO_DEBUG has non-whitespace after <level>
            then that part to the "," is interpreted as a filename to read the real
            value from. The first line of that file is the value used.
            This rule is *not* applied recursively to this value.

        """
        self.level = None
        self.filename = None
        self.overwrite = None
        self.from_file = False
        if val is None:
            self.val = os.environ.get('GLOBUS_XIO_DEBUG','').strip()
        else:
            self.val = val.strip()
        if self.val != '':
            self._parse()

    def _parse(self):
        # look for level of indirection first
        mo = self._fpat.match(self.val)
        if mo is not None:
            self.from_file = True
            filename = mo.group(1)
            try:
                self.val = self._readFromFile(filename)
            except IOError,E:
                raise GlobusXIODebugParseError("reading from '%s': %s" % (filename,E))
        # now parse the value
        mo = self._pat.match(self.val)
        if mo is None:
            raise GlobusXIODebugParseError(self.val)
        d = mo.groupdict()
        self.level = int(d['lvl'])
        self.filename = d.get('file', self.STDOUT_FILENAME)
        self.overwrite = d.has_key('ow')

    def _readFromFile(self, filename):
        f = open(filename, 'r')
        val = f.readline()
        return val.strip()
