############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
############################################################################
"""
The interface to the Globus xio module
"""

import xioc
import utilc
import time, re, socket
import types, exceptions 
import pyGlobus.util 

RE_IPV6 = re.compile("\[[\w,\W,:,.]+:[0-9]+")

class XioException(pyGlobus.util.GlobusException):
    """
    A subclass of GlobusException.

    Used as a base class for all other Xio exceptions.

    **Extends:**

    *pyGlobus.util.GlobusException
    """
    pass

class XioAttrException(XioException):
    """
    A subclass of XioException.

    Used to indicate errors in the XioAttr class. 

    **Extends:**

    *pyGlobus.xio.XioException
    """
    pass


class XioStackException(XioException):
    """
    A subclass of XioException.

    Used to indicate errors in the XioStack class. 

    **Extends:**

    *pyGlobus.xio.XioException
    """
    pass

class XioDriverException(XioException):
    """
    A subclass of XioException.

    Used to indicate errors in the XioDriver class. 

    **Extends:**

    *pyGlobus.xio.XioException
    """
    pass


class XioDataDescriptorException(XioException):
    """
    A subclass of XioException.

    Used to indicate errors in the XioDataDescriptor class. 

    **Extends:**

    *pyGlobus.xio.XioException
    """
    pass
     

class XioSocketException(XioException):
    """
    A subclass of XioException.

    Used to indicate errors in the XioDataDescriptor class. 

    **Extends:**

    *pyGlobus.xio.XioException
    """
    pass

class XioBase:
    """
    Provides functionality all of the xio classes need
    """
    def __init__(self):
        self._handle = None
        ret = utilc.globus_module_activate(xioc.get_module())
        if not ret == 0:
            ex = XioException("Unable to initialize the xio module")
            raise ex
        
    def __del__(self):
        ret = utilc.globus_module_deactivate(xioc.get_module())
        if not ret == 0:
            ex = XioException("Unable to deactivate the xio module")
            raise ex
        
    def get_handle(self):
        """
        Return the handle
        """
        return self._handle
    
class XioAttr(XioBase):
    """
    A class that wraps the globus_xio_attr_t structure.
    """
    
    def __init__(self, handle=None):
        """
        Creates an Xio Attr
        """
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.attr_init()
            if not ret == 0:
                raise XioAttrException(handle)

        self._handle = handle

    def __del__(self):
        """
        Frees the memory associated with the XioAttr
        """
        ret, err = xioc.attr_destroy(self._handle)
        if not ret == 0:
            ex = XioAttrException(err)
            raise ex
        XioBase.__del__(self)
        
    def cntl(self, driver, cmd, args):
        """
        Manipulate the values associated with the attribute
        
        **Arguments**

        - *driver* An XioDriver 

        - *cmd* A cmd that the driver understands, as of yet 
        only TCP driver commands are available

        - *args* A tuple of 0,1 or 2 items. Items
        can be strings, booleans or ints. 

        *Returns*
        This can return None, an int, boolean, string
        or a tuple of (bool,int) or (int,int). It depends
        on the nature of command issued. 

        **Raises**
        A XioAttrException is raised if there is a problem
        issuing the command to the driver.
        """

        if len(args) > 2: 
            except_string = "args: should not contain more than 2 items"
            raise XioAttrException(except_string)

        if len(args) == 0:
            pass
        
        elif len(args) == 1:
            if not self._check_types(type(args[0])):
                except_string = "args: must be boolean, int or string"
                raise XioAttrException(except_string)
        elif len(args) == 2:     
            ok1 = self._check_types(type(args[0]), (types.IntType,
                                    types.BooleanType)) 
            
            ok2 = self._check_types(type(args[1]), (types.IntType,)) 
            if not (ok1 and ok2):
                except_string ="Violated type rules 2 item tuple arguments"
                raise XioAttrException(except_string)   
        else:
            except_string = "args cannot have more than 2 items"
            raise XioAttrException(except_string)

        func = driver._getoptfunc(cmd)
        driver_cmd = xioc.__dict__[cmd] 
        ret, retvals = func(self._handle, driver.get_handle(), driver_cmd, args) 
        if not ret == 0:
            ex = XioAttrException(retvals)
            raise ex
       
        return retvals

    def copy(self):
        """
        Return a new XioAttr object
        
        **Raises**
        A XioAttrException is raised if there is a problem
        copying the XioAttr 
        """
        ret, new_attr = xioc.attr_copy(self._handle)
        if not ret == 0:
            ex = XioAttrException(new_attr)
            raise ex
        
        return XioAttr(new_attr)   

    def _check_types(self, arg_type, types_seq=(types.IntType, 
                                                types.BooleanType, 
                                                types.StringType)):
        """
        Internal method only.  
        """
        truth_value = False
        for t in types_seq:
            truth_value = truth_value or (arg_type is t)

        return truth_value
      
    
        
class XioDriver(XioBase):
    """
    Creates an XioDriver
    """
    def __init__(self, optdict={}):
        """
        Creates an XioDriver instance
        """
        XioBase.__init__(self)
        self._handle = None
        self._loaded = 0
        self._optdict = optdict 

    def __del__(self):
        """
        Frees the memory associated with the XioDriver
        """
        if self._loaded:
            self.unload()
        xioc.driver_free(self._handle)
        XioBase.__del__(self)
        # finalize C logging
        xioc.xio_nl_finalize()
         
    def load(self, driver_string):
        """
        **Arguments**

        - *driver_string* - Valid strings are tcp, gsi
        Should only be called once per instance
        """
        ret, handle = xioc.driver_load(driver_string)
        if not ret == 0:
            ex = XioDriverException(handle)
            raise ex
        self._handle = handle
        self._loaded = 1

    def unload(self):
        """
        Unloads the driver, should be called before __del__
        """
        ret, err = xioc.driver_unload(self._handle)
        if not ret == 0:
            ex = XioDriverException(err)
            raise ex
        self._loaded = 0
      
    def getopts(self):
        """
        Returns a list of options which can be used to manipulate 
        the behavior of the driver
        """
        return self._optdict.keys()
   
    def _getoptfunc(self, cmd):
        """
        Called from the XioAttr. Specific to the xio module.
        An API programmer should not have to call this method.
        
        Returns the function that handles this command
        
        **Raises**
        An XioDriverException is raised if the cmd is not
        recognized by the driver
        """
        try:
            return self._optdict[cmd]
        except KeyError:
            except_string = "This command is not recognized by this driver"
            raise XioDriverException(except_string)
            

class XioTcpDriver(XioDriver):
    """
    class that wraps a globus_xio_driver_t that is a tcp 
    driver, it has a dictionary with all of the commands that
    can be used to manipulate the driver
    """
    tcp_opt_dict = {
    "GLOBUS_XIO_TCP_GET_BACKLOG":xioc.attr_cntl_get_int,
    "GLOBUS_XIO_TCP_GET_CONNECT_RANGE":xioc.attr_cntl_get_int_int, 
    "GLOBUS_XIO_TCP_GET_INTERFACE":xioc.attr_cntl_get_string, 
    "GLOBUS_XIO_TCP_GET_KEEPALIVE":xioc.attr_cntl_get_bool,
    "GLOBUS_XIO_TCP_GET_LINGER":xioc.attr_cntl_get_bool_int,
    "GLOBUS_XIO_TCP_GET_LISTEN_RANGE":xioc.attr_cntl_get_int_int,
    "GLOBUS_XIO_TCP_GET_NODELAY":xioc.attr_cntl_get_bool,
    "GLOBUS_XIO_TCP_GET_NO_IPV6":xioc.attr_cntl_get_bool,
    "GLOBUS_XIO_TCP_GET_OOBINLINE":xioc.attr_cntl_get_bool, 
    "GLOBUS_XIO_TCP_GET_PORT":xioc.attr_cntl_get_int, 
    "GLOBUS_XIO_TCP_GET_RCVBUF":xioc.attr_cntl_get_int,  
    "GLOBUS_XIO_TCP_GET_RESTRICT_PORT":xioc.attr_cntl_get_bool, 
    "GLOBUS_XIO_TCP_GET_REUSEADDR":xioc.attr_cntl_get_bool,
    "GLOBUS_XIO_TCP_GET_SERVICE":xioc.attr_cntl_get_string,   
    "GLOBUS_XIO_TCP_GET_SEND_FLAGS":xioc.attr_cntl_get_int,
    "GLOBUS_XIO_TCP_GET_SNDBUF":xioc.attr_cntl_get_int,   
    "GLOBUS_XIO_TCP_SET_BACKLOG":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_CONNECT_RANGE":xioc.attr_cntl_set, 
    "GLOBUS_XIO_TCP_SET_INTERFACE":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_KEEPALIVE":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_LINGER":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_LISTEN_RANGE":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_NODELAY":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_NO_IPV6":xioc.attr_cntl_set, 
    "GLOBUS_XIO_TCP_SET_OOBINLINE":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_PORT":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_RCVBUF":xioc.attr_cntl_set, 
    "GLOBUS_XIO_TCP_SET_RESTRICT_PORT":xioc.attr_cntl_set, 
    "GLOBUS_XIO_TCP_SET_REUSEADDR":xioc.attr_cntl_set, 
    "GLOBUS_XIO_TCP_SET_SEND_FLAGS":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_SERVICE":xioc.attr_cntl_set,
    "GLOBUS_XIO_TCP_SET_SNDBUF":xioc.attr_cntl_set}
    
    def __init__(self):
        XioDriver.__init__(self, self.tcp_opt_dict)
        self.load("tcp")
    
    def __del__(self):
        XioDriver.__del__(self)

    def get_backlog_opt(self):
        """
        Returns the backlog option
        """
        return "GLOBUS_XIO_TCP_SET_BACKLOG"

    def get_port_opt(self):
        """
        Returns the set port option
        """
        return "GLOBUS_XIO_TCP_SET_PORT"
        
class XioDataDescriptor(XioBase):
    """
    Wraps the globus_xio_data_descriptor_t structure
    """
    def __init__(self, file, handle=None):
        XioBase.__init__(self)
        if handle == None:
            ret, handle = xioc.descriptor_init(file.get_handle())
            if not ret == 0:
                ex = XioDataDescriptorException(handle)
                raise ex
        self._handle = handle

    def __del__(self):
        ret, err = xioc.descriptor_destroy(self._handle)
        if not ret == 0:
            ex = XioDataDescriptorException(err)
            raise ex
        XioBase.__del__(self)
       

class XioStack(XioBase):
    """
    Creates an Xio Attr
    """
    def __init__(self, attr=None, handle=None):
        """
        Creates an XioStack instance
        """
        
        XioBase.__init__(self)
        if attr == None:
            attr = XioAttr()
            
        if handle == None:
            ret, handle = xioc.stack_init(attr.get_handle())
            if not ret == 0:
                raise XioStackException(handle)
            
        self._handle = handle
            
    def __del__(self):
        """
        Frees the memory associated with the XioStack
        """   
        ret, err = xioc.stack_destroy(self._handle)
        if not ret == 0:
            ex = XioStackException(err)
            raise ex
        XioBase.__del__(self)
        
    def push_driver(self, driver):
        """
        Pushes a driver onto the Protocol Stack.
        The transport driver must be pushed first

        **Arguments**

        - *driver* - a XioDriver object

        **Returns**
        None

        **Raises**
        A XioStackException is raised if the driver cannot be
        pushed onto the stack
        """
        ret, err = xioc.push_driver(self._handle, driver.get_handle())
        if not ret == 0:
            ex = XioStackException(err)
            raise ex
        

class XioSocket(XioBase):
    """
    A simple socket interface similar to the Python socket API. 
    
    Provides the ability to make strongly authenticated
    connections using GSI. It supports several modes of authorization
    and provides optional integrity/privacy to the data.   
    """
    
    def __init__(self, handle=None, stack=None, transport_driver=None):
        """
        Constructs an instance.

        Each instance wraps a globus_xio_handle_t object.
        An instance of this class may only have a single
        open connection at a time. To have more then one open
        connection, use multiple instances.
        
        **Arguments:** 

        - *handle* An optional argument that should contain 
        a SWIG'ized pointer to a globus_xio_handle_t object.
        
        - *stack* An optional argument that should be an 
        XioStack object

        - *transport_driver* An optional argument that is 
        the first driver in the stack. Defaults to tcp.  

        **Raises**        
        Raises a XioSocketException if unable to initialize 
        the globus modules.
        """
        XioBase.__init__(self)

        if transport_driver is None:
            self._transport_driver = XioTcpDriver()
            
        else:
            self._transport_driver = transport_driver

        if stack is None: 
            self._stack = XioStack() 
            self._stack.push_driver(self._transport_driver)
        else:
            self._stack = stack
        
        self._handle = handle
        self._server_handle = None 
        self._open = False 
        self._timeout = None 
        self._attr = XioAttr()
        if handle == None:
            ret, handle = xioc.xio_handle_create(self._stack.get_handle())
            if not ret == 0:
                ex = XioSocketException(handle)
                raise ex
            
        self._handle = handle
        # total bytes read/written between now and close()
        self._tot_read, self._tot_write = 0, 0
        # time of first read/written byte
        self._data_start = None
        return 

    def __del__(self):
        """
        Closes the handles if they aren't closed already, then
        frees the memory associated with the handles if it
        was allocated
        """ 
        if self._open:
            self.close()
        if self._server_handle:
            xioc.server_free(self._server_handle) 
        if self._handle:
            xioc.handle_free(self._handle) 
            

    def accept(self):
        """
        Blocking server-side TCP connection establishment.
            
        **Raises:**
        
        A XioSocketException is thrown if a listener
        was not already created, or an error occurs
        while doing the underlying accept.
        
        **Returns:**
        
        A new XioSocket object for the new connection.
        """
        ret, child_handle = xioc.server_accept(self._server_handle)
        if not ret == 0:
            ex = XioSocketException(child_handle)
            raise ex
        sock = XioSocket(child_handle, self._stack, self._transport_driver)
        peername = sock.getpeername()
        ret, err = xioc.xio_open(sock._attr.get_handle(), 
                                 self._make_contact_string(peername),
                                 sock.get_handle())
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
        
        sock._set_open(True)

        # logging
        ip, port = self.getsockname()
        pyGlobus.inst.setMeta(pyGlobus.inst.TRANSFER_EVENT, bare_event=True,
			      remote_host=peername[0], 
			      remote_port=peername[1],
			      local_host=ip, local_port=port)

        return (sock, peername) 
    
    def bind(self, bind_tuple):
        """
        binds the socket to a port
        
        Note: This does not follow the python socket semantics
        exactly, getsockopt does not return the (host,port)
        it was bound to until listen is called.

        **Arguments**

        - *bind_tuple* Must be a tuple that is (hostname, port) 
        port that the socket will listen for connections on. If the
        port number is 0, then an arbitrary TCP port will be selected. If
        this is true, and the restrict_port attribute is set to TRUE (the
        default) and the GLOBUS_TCP_PORT_RANGE environment variable
        was set when Globus XIO was initialized, then the port will be
        selected from that range. Otherwise, any port number may be
        
        **Raises**

        TypeErrors are raised if the argument is not a tuple
        or the port is not an int.
        """
        if type(bind_tuple) is not types.TupleType:
            raise TypeError
        port = bind_tuple[1]
        if type(port) is not types.IntType:
            raise TypeError("Must bind to a port that is an int")
       
        self.setsockopt(self._transport_driver.get_port_opt(), port) 

	# logging
        pyGlobus.inst.setMeta(event=['xio.XioSocket.listen',
				     'xio.XioSocket.close'],
			      listen_host=bind_tuple[0] or 'localhost', 
			      listen_port=bind_tuple[1]) 

    def close(self):
        """
        close the socket 
        """
        self._set_open(False)
        if self._server_handle:
            ret, err = xioc.server_close(self._server_handle) 
            if not ret == 0:
                ex = XioSocketException(err)
                raise ex
        if self._handle:
            attr = XioAttr()     
            ret, err = xioc.xio_close(self._handle, attr.get_handle())
            if not ret == 0:
                ex = XioSocketException(err)
                raise ex 
        # log transfer done event
        if pyGlobus.inst.getXIODebug() and self._data_start is not None:
            t = time.time() - self._data_start
            b = self._tot_read + self._tot_write
            if t == 0:
                tp = -1
            else:
                tp = ( 8.0 * b / 1e6 ) / t
            pyGlobus.inst.dbg_write(pyGlobus.inst.TRANSFER_EVENT,
                                    {'sec':t,'bytes':b,'Mb/s':tp})

    def connect(self, addr):
        """
        Blocking TCP connection establishment.

        connect a TCP socket on the specified host/port pair.

        **Arguments:**
        
        - *addr* A tuple (host,port) where host is a string and port is 
        an int.
        
        **Raises:**
        
        A XioSocketException is thrown if the handle is currently in use,
        or an error occurs while making the connection.
        """

        contact_string = self._make_contact_string(addr)
        ret, err = xioc.xio_open(self._attr.get_handle(), 
                                contact_string, 
                                self._handle) 
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
        
        self._open = True

        ip, port = self.getsockname()
        pip, pport = self.getpeername()
        # logging
        pyGlobus.inst.setMeta(pyGlobus.inst.TRANSFER_EVENT, bare_event=True,
			      remote_host=pip, remote_port=pport,
                              local_host=ip, local_port=port)

    def dup(self):
        """
        Return a new socket object identical to the current one
        
        Not implemented

        **Raises** an NotImplementedError is raised
        """
        err_string = "dup is not implemented in XioSocket"
        raise exceptions.NotImplementedError(err_string) 

    def fileno(self):
        """
        Return the underlying file descriptor
        
        Not implemented

        **Raises** an NotImplementedError is raised
        """
        err_string = "fileno is not implemented in XioSocket"
        raise exceptions.NotImplementedError(err_string) 
        
    def getpeername(self):
        """
        Return the remote address
       
        **Returns**
        The remote address is retuned in a (host,port) tuple

        **Raises**
        An XioSocketException is raised if there is a problem
        getting the peername.
        """
        ret, peername = xioc.xio_handle_cntl_return_string(
            self._handle,
            self._transport_driver.get_handle(), 
            xioc.GLOBUS_XIO_GET_REMOTE_CONTACT)
        
       
        if not ret == 0:
            ex = XioSocketException(peername)
            raise ex
         
        return self._make_addr_tuple(peername) 
   
    def getsockname(self):
        """
        Return the local address
        """
        if self._server_handle is not None:
            ret, localname = xioc.server_cntl_return_string(
                             self._server_handle,
                             self._transport_driver.get_handle(),
                             xioc.GLOBUS_XIO_GET_LOCAL_NUMERIC_CONTACT) 
        elif self._open: 
            ret, localname = xioc.xio_handle_cntl_return_string(
                             self._handle,
                             self._transport_driver.get_handle(),
                             xioc.GLOBUS_XIO_GET_LOCAL_NUMERIC_CONTACT) 
        else:
            return ('0.0.0.0', 0) 

        if not ret == 0:
            ex = XioSocketException(localname)
            raise ex
        
        return self._make_addr_tuple(localname) 
    
    def getsockopt(self, optname):
        """
        Return the value of the socket option
        To get a list of options available call 
        getopts on the transport driver.

        **Arguments**
        
        - *optname* is an xioc.OPTION you get this 
        by calling getopts on the driver and selecting
        the option from the list returned.
        
        **Raises**
        An XioSocketException is raised if something went 
        wrong
        """ 
        try:
            retvals = self._attr.cntl(self._transport_driver, optname, ()) 
        except XioAttrException, ex:
            except_string = str(ex) 
            ex = XioSocketException(except_string)
            raise ex
        
        return retvals 
    
    def gettimeout(self):
        """
        Return timeout or None
        """
        return self._timeout 

    def listen(self, backlog):
        """
        Creates a listening TCP socket.
        
        Creates a socket handle capable of accepting new TCP connections
        from other hosts or processes. In order to actually listen for
        connections, you must call either listen or register listen on this
        instance.

        **Arguments** 
        
        - *backlog* The backlog parameter indicates he maximum length of the
        system's queue of pending connections. Any connection attempts
        when the queue is full will fail. If backlog is equal to -1, then the
        system-specific maximum queue length will be used. 
        
        **Raises:**
        
        A XioSocketException is thrown if the handle is currently in use,
        or an error occurs while making the connection.
        
        **Returns:**

        An int containing the port the server will listen on.
        """ 
        if not backlog == -1:
            self.setsockopt(self._transport_driver.get_backlog_opt(),
                            backlog)
        
        ret, handle = xioc.server_create(self._attr.get_handle(),
                                         self._stack.get_handle())
        if not ret == 0:
            ex = XioSocketException(handle)
            raise ex
        self._set_open(True)
        self._server_handle = handle

    def makefile(self):
        """
        Return the underlying file descriptor
        
        Not implemented

        **Raises** an NotImplementedError is raised
        """
        err_string = "makefile is not implemented in XioSocket"
        raise exceptions.NotImplementedError(err_string) 

    def recv(self, buflen, data_desc=None):
        """
        blocking read call

        **Arguments**
        - *buflen* - an int telling how much data to read 

        - *data_desc* - optional argument that can control how the
        driver behaves. The interface to call cntl on the
        data_descriptor is currently not exposed.
        
        **Returns**
        A string read from the handle

        **Raises**
        a XioSocketException is raised if there is a problem reading
        from the handle
        """ 
        # Logging; record time of 1st data byte
	# Note: not conditional on logging being on, otherwise
	# turning on logging half-way will give bad results.
        if self._data_start is None:
            self._data_start = time.time() 
        
        wait_for_bytes = 1
        py_buffer = pyGlobus.util.Buffer(buflen)    
         
        ret, buf = xioc.xio_read(self._handle,
                                 py_buffer.get_handle(),
                                 py_buffer.get_size(),
                                 wait_for_bytes,
                                 data_desc)
        
        
        if not ret == 0:
            ex = XioSocketException(buf)
            raise ex
        s = str(buf)
        self._tot_read += len(s)
        return s

    def send(self, data, data_desc=None):
        """ 
        Emulates socket.send -- writes a string over a socket

        **Arguments**
        - *data* - a string that will be written to the handle 

        - *data_desc* - optional argument that can control how the
        driver behaves. The interface to call cntl on the
        data_descriptor is currently not exposed.

        **Returns**
        The number of bytes written

        **Raises**
        a XioSocketException is raised if there is a problem
        writing to the handle
        """
        # logging: record time of 1st data byte
        if self._data_start is None:
            self._data_start = time.time()
        
        ret, bytes_written = xioc.xio_write(self._handle,
                                            data,
                                            len(data),
                                            1,
                                            data_desc)
        
        if not ret == 0:
            ex = XioSocketException(bytes_written)
            raise ex
        self._tot_write += bytes_written
        return bytes_written 
    
    def sendall(self, data, data_desc=None):
        """
        Emulates socket.sendall -- writes a complete string over a socket
        
         **Arguments**
         - *data* - a string that will be written to the handle 
        
        - *data_desc* - optional argument that can control how the
        driver behaves. The interface to call cntl on the
        data_descriptor is currently not exposed.
        
        **Raises**
        a XioSocketException is raised if there is a problem writing to the
        handle
        """ 
        # logging: record time of 1st data byte
        if self._data_start is None:
            self._data_start = time.time()
        total_len = len(data) 
        bytes_written = 0 
        while bytes_written < total_len: 
            bytes_written += self.send(data[bytes_written:]) 
    
    def setblocking(self, flag):
        """
        Sets the socket to be blocking or not.

        **Arguments**

        -*flag* a boolean
        """
        pass
        
    def setsockopt(self, optname, *value):
        """
        Set the options from the socket.
        To get a list of options available call 
        getopts on the transport driver.

        **Arguments**
        
        - *optname* is a string to get a list of options 
        call getopts on the driver and selecting
        the option from the list returned.
        
        - *value* can be booleans, strings or ints
        but it must make sense with the option that
        you are using

        **Returns**
        None
       
        **Raises** 
        An XioSocketException is raised if there is a problem 
        setting the socket option, a KeyError is raised
        if the option is invalid 
        """
        try:
            self._attr.cntl(self._transport_driver, optname, value)
        except XioAttrException, ex:
            sock_ex = XioSocketException(str(ex))
            raise sock_ex

    def settimeout(self, timeout):
        """
        Sets a timeout for the socket 
        """
        ret, err = xioc.attr_cntl_set_timeout_accept(self._attr.get_handle(),
                                                     timeout)
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
        
        ret, err = xioc.attr_cntl_set_timeout_all(self._attr.get_handle(),
                                                  timeout)
        if not ret == 0:
            ex = XioSocketException(err)
            raise ex
  
        self._timeout = timeout 

    def shutdown(self, how):
        """
        shutdown the reading, writing or both sides of the socket
        
        Not implemented

        **Raises** an NotImplementedError is raised 
        """
        err_string = "shutdown is not implemented in the XioSocket class"
        raise exceptions.NotImplementedError(err_string) 

    def _set_open(self, value):
        """
        Internal method only 
        """ 
        self._open = value 

    def _make_addr_tuple(self, addr_string):
        """
        Internal method only
        """
        if RE_IPV6.match(addr_string):
            index = addr_string.rindex(':')
            ip_part = addr_string[:index]
            port = addr_string[index:].strip(':')
            ip = ip_part.strip('[]').split(':')[3]
            return(ip, int(port), 0, 0)
        else: 
            hostname, port = addr_string.split(':') 
            ip = socket.gethostbyaddr(hostname)[2][0]
            return (ip, int(port)) 
    
    def _make_contact_string(self, addr_tuple):
        """
        Internal method only
        """
        seq = map(str, addr_tuple)
        return ":".join(seq)


class GsiSocket(XioSocket):
    """
    Class that has a python socket interface, but uses
    GSI for authentication and data integrity over the transport
    desired driver 
    """
    def __init__(self, handle=None, stack=None, transport_driver=None):
        """
        Initialize a socket that uses TCP and GSI
        """
        stack = XioStack()
        self._transport_driver = XioTcpDriver()
        self._gsi_driver = XioDriver()
        self._gsi_driver.load("gsi")
        stack.push_driver(self._transport_driver)
        stack.push_driver(self._gsi_driver)
        XioSocket.__init__(self, stack=stack,
                           transport_driver=self._transport_driver) 

   
# -- netlogger logging 
import pyGlobus.inst
pyGlobus.inst.instrument(XioSocket, 'xio',
			 exclude=['send', 'sendall', 'recv'])
