############################################################################
# David Konerding, LBNL
# Based on the EasyFtpClient written by Matt Rodriguez
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################

## Attempt a version string

__author__  = "David E. Konerding"
__status__  = "alpha"
__version__ = "$Revision: 1.20 $"
__rcsid__   = "$Id: pyglobus-url-copy.py,v 1.20 2004/07/14 20:44:26 dek Exp $"

## 1) Look for variables which are lock-accessed (such as g_cond and
##    error_status) being accessed without the lock held.  Also remove unncessary locks in callbacks
## 2) Use pyglobus-url-copy specific exceptions instead of RuntimeError
## 3) Transfer of size=0 files is broken
## 4) pyglobus-url-copy doesn't actually disconnect from remote server (jason says it's the ftpcontrol not being deleted)
## 5) checksum support (request from jason lee-- see his email)

from pyGlobus.ftpClient import FtpClient, FtpClientException, HandleAttr, OperationAttr, RestartMarker

try:
    from pyGlobus.ftpClient import DebugPlugin
except ImportError:
    from pyGlobus.ftpClientPlugin import DebugPlugin

try:
    from pyGlobus.ftpClientPlugin import PyGlobusBasePlugin
except:
    PyGlobusBasePlugin = None

try:
    from pyGlobus.ftpClient import ThroughputPerfPlugin
except:
    from pyGlobus.ftpClientPlugin import ThroughputPerfPlugin

from pyGlobus import ftpControl
from pyGlobus.util import Buffer, GlobusException, Off_t
from pyGlobus.ftpControl import Parallelism, TcpBuffer, FtpControl, PARALLELISM_FIXED, PARALLELISM_NONE, MODE_EXTENDED_BLOCK
from pyGlobus.security import GSSCred, GSSCredException
from pyGlobus.logUtil import logging, netlogger

from threading import *
import getpass, getopt, sys, os, time, socket, traceback

class pyGlobusUrlCopy:
    APPLICATION_EVENT_NAME = "pyGlobusUrlCopy"
    NETLOGGER_DEST = "x-netlog://netlogger.lbl.gov:14830"

    FILE = 0
    GSIFTP = 1

    NOT_DONE = 0
    DONE = 1
    ERROR = 2

    TCP_BUFFER_SIZE = 64 * 1024
    BLOCK_SIZE = 64 * 1024
    PARALLELISM = 1
    NETLOGGER_WRITER = None
    DEBUG_PLUGIN = None
    PYGLOBUS_PLUGIN = None
    WORKAROUND = 0
    LOGGER_LEVEL = logging.ERROR
    SOURCE_SUBJECT = None
    DEST_SUBJECT = None
    THROUGHPUT = None

    def __init__(self, srcUrl, dstUrl, tcp_buffer_size=TCP_BUFFER_SIZE, block_size=BLOCK_SIZE, parallelism=PARALLELISM, logger = None, netlogger_writer=NETLOGGER_WRITER, debug_plugin=DEBUG_PLUGIN, pyglobus_plugin=PYGLOBUS_PLUGIN, workaround=WORKAROUND, source_subject=SOURCE_SUBJECT, dest_subject=DEST_SUBJECT, throughput=THROUGHPUT):
        if logger == None:
            ## set up error logging before anything else
            logger = logging.getLogger(pyGlobusUrlCopy.APPLICATION_EVENT_NAME)
            logger_handler = logging.StreamHandler(sys.stdout)
            logger_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
            logger_handler.setFormatter(logger_formatter)
            logger.addHandler(logger_handler)
            
        self.logger = logger
        
        self.logger.debug("pyGlobusUrlCopy: __init__() begin")
        if netlogger_writer == None:
            flags = netlogger.OutputFlags.FLUSH
            self.netlogger_writer = netlogger.open(netlogger_dest,
                                                   flags = flags,
                                                   program = pyGlobusUrlCopy.APPLICATION_EVENT_NAME,
                                                   threaded = 1,
                                                   backup = None)
        else:
            self.netlogger_writer = netlogger_writer
        self.workaround = workaround

        ## cache the username
        self.user = getpass.getuser()
        self.logger.debug("pyGlobusUrlCopy: __init__() user is '%s'" % self.user)
        self.host = socket.getfqdn()
        self.logger.debug("pyGlobusUrlCopy: __init__() host is '%s'" % self.host)

        ## initial netlogger call
        self.logger.debug("pyGlobusUrlCopy: __init__(srcUrl=%s, dstUrl=%s, tcp_buffer_size=%s, block_size=%s, parallelism=%s, netlogger=%s, debug_plugin=%s, pyglobus_plugin=%s, logger=%s)" % (srcUrl, dstUrl, tcp_buffer_size, block_size, parallelism, netlogger_writer, debug_plugin, pyglobus_plugin, logger))
        self.netlogger_writer.write(2, "%s.prog.start" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d USER=%s TCP_BUFFER=%d BLOCK_SIZE=%d SRC_URL=%s DST_URL=%s", 1, self.user, tcp_buffer_size, block_size, srcUrl, dstUrl)
        
        ## Basic FTP parameters
        self.srcUrl = srcUrl
        self.dstUrl = dstUrl
        self.tcp_buffer_size = tcp_buffer_size
        self.block_size = block_size
        self.parallelism = parallelism

        ## Whether to measure throughput
        self.throughput = throughput
        
        ## FTP connection attributes
        self.error_status = None

        ## A status variable for the transfer...
        self.g_cond = pyGlobusUrlCopy.NOT_DONE
        ## ... and the Condition variable which protects it
        self.condV = Condition(Lock())

        ## Some attributes and their locks used mainly by parallel streams
        self.lock = Lock()
        
        ## counter for total number of bytes transferred
        self.totalBytes = 0

        ## boolean which is set to 0 after first packet is transferred
        self.first = 1
        
        ## temporary variable to hold on to Buffers (until a bug is fixed in pyGlobus)
        self.buffer_list = []

        ## Set up attributes we might modify during CL processing
        self.handle_attribute = HandleAttr()
        self.source_operation_attribute = OperationAttr()
        self.dest_operation_attribute = OperationAttr()

        if self.parallelism > 1:
            self.logger.debug("pyGlobusUrlCopy: __init__() enabling parallelism: %d streams" % self.parallelism)
            self.parallelism_attribute = Parallelism()
            self.parallelism_attribute.set_mode(PARALLELISM_FIXED)
            self.parallelism_attribute.set_size(self.parallelism)
            self.source_operation_attribute.set_parallelism(self.parallelism_attribute)
            self.source_operation_attribute.set_mode(MODE_EXTENDED_BLOCK)
            self.dest_operation_attribute.set_parallelism(self.parallelism_attribute)
            self.dest_operation_attribute.set_mode(MODE_EXTENDED_BLOCK)

        if debug_plugin:
            self.logger.debug("pyGlobusUrlCopy: __init__() enabling DebugPlugin")
            self.debug_plugin = DebugPlugin(sys.stdout, ('DebugPlugin'))
            self.handle_attribute.add_plugin(self.debug_plugin)
        else:
            self.debug_plugin = None

        if pyglobus_plugin:
            self.logger.debug("pyGlobusUrlCopy: __init__() enabling PyGlobusBasePlugin")
            if pyglobus_plugin is True:
                self.pyglobus_plugin = PyGlobusBasePlugin()
            else:
                if isinstance(pyglobus_plugin, PyGlobusBasePlugin):
                    self.pyglobus_plugin = pyglobus_plugin
                else:
                    raise RuntimeError, "pyglobusplugin is not a PyGlobusBasePlugin"
            self.handle_attribute.add_plugin(self.pyglobus_plugin)
        else:
            self.pyglobus_plugin = None
            
        if self.throughput:
            self.throughput_plugin = ThroughputPerfPlugin(self.throughput_begin_cb, self.throughput_stripe_cb, self.throughput_total_cb, self.throughput_complete_cb, None)
            self.handle_attribute.add_plugin(self.throughput_plugin)

        ## Set TCP buffer size
        self.logger.debug("pyGlobusUrlCopy: __init__() setting TcpBuffer to %d" % self.tcp_buffer_size)
        tcp_buffer_attr = TcpBuffer()
        tcp_buffer_attr.set_fixed(self.tcp_buffer_size)
        self.source_operation_attribute.set_tcp_buffer(tcp_buffer_attr)
        self.dest_operation_attribute.set_tcp_buffer(tcp_buffer_attr)

        if source_subject:
            self.source_credential = GSSCred()
            try:
                self.source_credential.acquire_cred()
            except GSSCredException, what:
                self.logger.error("pyGlobusUrlCopy: __init__(): Credential error")
                self.g_cond = pyGlobusUrlCopy.ERROR
                self.error_status = sys.exc_info()
                self.cleanup_and_handle_errors()
                raise

            lifetime, name = self.source_credential.inquire_cred()
            subject = name.display()
            self.source_operation_attribute.set_authorization(self.source_credential, None, None, None, source_subject)

        if dest_subject:
            self.dest_credential = GSSCred()
            try:
                self.dest_credential.acquire_cred()
            except GSSCredException, what:
                self.logger.error("pyGlobusUrlCopy: __init__(): Credential error")
                self.g_cond = pyGlobusUrlCopy.ERROR
                self.error_status = sys.exc_info()
                self.cleanup_and_handle_errors()
                raise
            lifetime, name = self.dest_credential.inquire_cred()
            subject = name.display()
            self.dest_operation_attribute.set_authorization(self.dest_credential, None, None, None, dest_subject)



        ## Create the FtpClient object this transfer will use
        self.logger.debug("pyGlobusUrlCopy: __init__() create the FtpClient object")
        self.ftp_client = FtpClient(self.handle_attribute)

        

        ## Figure out what sort of transfer this is
        if self.srcUrl[:7] == "file://":
            self.srcUrl_type = pyGlobusUrlCopy.FILE
        elif self.srcUrl[:9] == "gsiftp://":
            self.srcUrl_type = pyGlobusUrlCopy.GSIFTP
        else:
            error = "pyGlobusUrlCopy: in __init__(): Source URL type '%s' is not recognized" % (self.srcUrl)
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
            self.cleanup_and_handle_errors()
            raise RuntimeError, error

        if self.dstUrl[:7] == "file://":
            self.dstUrl_type = pyGlobusUrlCopy.FILE
        elif self.dstUrl[:9] == "gsiftp://":
            self.dstUrl_type = pyGlobusUrlCopy.GSIFTP
        else:
            error = "pyGlobusUrlCopy: in __init__(): Destination URL type '%s' is not recognized" % (self.dstUrl)
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
            self.cleanup_and_handle_errors()
            raise RuntimeError, error
        
        self.logger.debug("pyGlobusUrlCopy: __init__() srcUrl_type=%s" % (self.srcUrl_type))
        self.logger.debug("pyGlobusUrlCopy: __init__() dstUrl_type=%s" % (self.dstUrl_type))

        if self.srcUrl_type == pyGlobusUrlCopy.FILE:
            if self.dstUrl_type == pyGlobusUrlCopy.FILE:
                error = "pyGlobusUrlCopy: in __init__(): Both source and destination are local files"
                self.g_cond = pyGlobusUrlCopy.ERROR
                self.error_status = error
                self.cleanup_and_handle_errors()
                raise RuntimeError, error

            elif self.dstUrl_type == pyGlobusUrlCopy.GSIFTP:
                self.put_ftp()

        elif self.srcUrl_type == pyGlobusUrlCopy.GSIFTP:
            if self.dstUrl_type == pyGlobusUrlCopy.GSIFTP:
                self.third_party_ftp()

            elif self.dstUrl_type == pyGlobusUrlCopy.FILE:
                self.get_ftp()

        self.logger.debug("pyGlobusUrlCopy: __init__() end")

    def throughput_begin_cb(self, arg, handle, srcUrl, destUrl):
        self.logger.info("Throughput begin: srcUrl=%s, destUrl=%s" % (srcUrl, destUrl))

    def throughput_stripe_cb(self, arg, handle, stripe_ndx, bytes, instantaneous_throughput, avg_throughput):
        self.logger.info("Throughput stripe: stripe_ndx=%d, bytes=%d, instantaneous_throughput=%d, avg_throughput=%d" % (stripe_ndx, bytes, instantaneous_throughput, avg_throughput))

    def throughput_total_cb(self, arg, handle, bytes, instantaneous_throughput, avg_throughput):
        self.logger.info("Throughput total: bytes=%d, instantaneous_throughput=%d, avg_throughput=%d" % (bytes, instantaneous_throughput, avg_throughput))

    def throughput_complete_cb(self, arg, handle, success):
        self.logger.info("Throughput total: success=%s" % success)

    def third_party_ftp(self):
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp()")
        ## use the operation attribute for both source and destination attributes.  Why not?
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): call ftp_client.third_party_transfer()")
## Can't do this until we check the file size!
##         if self.throughput:
##             self.time0 = time.time()
            
        try:
            self.ftp_client.third_party_transfer(self.srcUrl, self.dstUrl, self.done_func, None, self.source_operation_attribute, self.dest_operation_attribute)
        except FtpClientException:
            self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV acquired")
            self.logger.error("pyGlobusUrlCopy: third_party_ftp(): FtpClientException starting third party transfer")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = sys.exc_info()
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV released")
            return
            
 
        self.netlogger_writer.write(2, "%s.thirdparty.start" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): call ftp_client.third_party_transfer() completed") 
        

        ## Prepare the ftp Client transfer
        ## acquire the condition variable lock
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV acquired")

        ## Now, block on the condition variable lock until the done_func changes its status
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV wait")
        self.condV.wait()
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV wait finished")


        self.netlogger_writer.write(2, "%s.thirdparty.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)

        self.cleanup_and_handle_errors()

## Can't do this until we check the file size!
##         if self.throughput:
##             print "Throughput: %d bytes in %5.2f seconds, average = %5.2f bytes/sec" % (file_size, time.time()-self.time0, file_size/(time.time()-self.time0))

        ## Clean up our lock and close our local file
        self.logger.debug("pyGlobusUrlCopy: third_party_ftp(): condV release")
        self.condV.release()


        ## Perhaps put this in the destructor?
        self.netlogger_writer.write(2, "%s.prog.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)

    def get_ftp(self):
        self.logger.debug("pyGlobusUrlCopy: get_ftp()")
        
        ## acquire the condition variable in preparation of asynchronous callback for file size
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquired")
        ## figure out the length of the remote file just to report the NUM_BYTES in the first message
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): request size")
        size = self.ftp_client.size(self.srcUrl, self.done_func, None, self.dest_operation_attribute)
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): requested size")
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV wait")
        ## wait for done_func
        self.condV.wait()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV wait finished")
        if self.g_cond == pyGlobusUrlCopy.ERROR:
            self.logger.error("pyGlobusUrlCopy: get_ftp(): Error in callback:")
            self.cleanup_and_handle_errors()
            return
        
        ## reset the condition variable for the next time...
        self.g_cond = pyGlobusUrlCopy.NOT_DONE
        self.condV.release()
        file_size = size.getLong()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): size is %d" % file_size)

        ## Open the local file to store get results
        destFile = self.dstUrl[6:]
        try:
            self.g_dest = open(destFile, "w")
        except IOError, what:
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquired")
            self.logger.error("pyGlobusUrlCopy: get_ftp(): IOError opening local file")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = sys.exc_info()
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV released")
            return
            
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): opened destination file: %s" % (destFile))

        ## Prepare the ftp Client transfer
        ## acquire the condition variable lock
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquired")
        ## exit out in case the ftp connection bailed asynchronously
        ## (there are some weird situations where failures can only be caught here)
        if self.g_cond == pyGlobusUrlCopy.ERROR:
            self.cleanup_and_handle_errors()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV release")
        self.condV.release()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV released")

        if self.throughput:
            self.time0 = time.time()

        ## start the transfer
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): call ftp_client.get()")
        self.ftp_client.get(self.srcUrl, self.done_func, None, self.dest_operation_attribute)
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): call ftp_client.get() complete")
        self.netlogger_writer.write(2, "%s.get.start" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d FILE_SIZE=%d", 1, file_size)

        ## magic math to compute the number of offsets (segments) to transfer
        total_offsets = ( int(file_size / self.block_size) + (file_size/float(self.block_size) > 0) )
        if self.parallelism > total_offsets:
            error = "pyGlobusUrlCopy: get_ftp(): The parallelism (%d) should not be raised over the number of block offsets (%d) in the file" % (self.parallelism, total_offsets)
            self.logger.error(error)
            self.error_status = error
            self.cleanup_and_handle_errors()
            raise RuntimeError, error

        ## schedule appropriate number of parallel streams
        for i in range(self.parallelism):
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): call self.register_read()")
            self.register_read()
            self.logger.debug("pyGlobusUrlCopy: get_ftp(): call self.register_read() complete")

        ## Now, block on the condition variable lock until the done_func changes its status
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV acquired")
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV wait")
        self.condV.wait()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV wait finished")

        self.netlogger_writer.write(2, "%s.get.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d TOTAL_BYTES=%d", 1, self.totalBytes)

        self.cleanup_and_handle_errors()

        if self.throughput:
            time1 = time.time()
            dt = time1 - self.time0
            throughput = file_size/dt
            print "Throughput: %d bytes in %5.2f seconds, average = %5.2f bytes/sec" % (file_size, dt, throughput)
            self.netlogger_writer.write(2, "%s.get.throughput" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "TOTAL_BYTES=%d TOTAL_TIME=%d THROUGHPUT=%f", 1, self.totalBytes, dt, throughput)

        ## Clean up our lock and close our local file
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV release")
        self.condV.release()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): condV released")

        self.logger.debug("pyGlobusUrlCopy: get_ftp(): closing retrieved file")
        self.g_dest.close()
        self.logger.debug("pyGlobusUrlCopy: get_ftp(): closed retrieved file")

        ## Perhaps put this in the destructor?
        self.netlogger_writer.write(2, "%s.prog.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)

    def register_read(self):
        self.logger.debug("pyGlobusUrlCopy: register_read()")
        ## Create a buffer object to store intermediate results
        self.logger.debug("pyGlobusUrlCopy: register_read(): create Buffer size %d" % self.block_size)
        buffer = Buffer(self.block_size)

        ## this is necessary to hold a reference to the buffer until the read completes.
        ## this bug is fixed in CVS but we need to keep this around until everybody gets up-to-date
        ## the workaround is only enabled if the user uses the command line flag '-w'
        ## (note to self: perhaps this should just put the buffer in a dictionary and remove it when the get callback is called)
        if self.workaround:
            self.buffer_list.append(buffer)

        ## Set up a register_read callback which is called with intermediate data
        self.logger.debug("pyGlobusUrlCopy: register_read(): call ftp_client.register_read()")
        self.ftp_client.register_read(buffer, self.data_func_get, None)
        self.logger.debug("pyGlobusUrlCopy: register_read(): call ftp_client.register_read() complete")

    ## Note: callbacks don't call cleanup_and_handle_error() directly;
    ## Instead, they acquire the condition variable, store the error, and notify the waiter (the main thread)
    def data_func_get(self, arg, handle, buffer, bufHandle, bufLen, offset, eof, error):
        self.logger.debug("pyGlobusUrlCopy: data_func_get(arg=%s, handle=%s, bufHandle=%s, bufLen=%d, offset=%d, eof=%d, error=%s" % (arg, handle, bufHandle, bufLen, offset, eof, error))
        self.netlogger_writer.write(3, "%s.get.fragment" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d NUM_BYTES=%d EOF=%d", self.totalBytes, bufLen, eof)

        if error != "NONE" and error != (0, None):
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquired")
            self.logger.error("pyGlobusUrlCopy: data_func_get(): passed an error condition.")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV notify")
            self.condV.notify()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV notified")
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV released")
            return
        else:
            if self.first == 1:
                self.netlogger_writer.write(2, "%s.get.transferStart" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)
                self.first = 0

        if bufLen:
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquire")
            self.lock.acquire()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquired")
            self.totalBytes += bufLen
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV release")
            self.lock.release()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV released")

            self.logger.debug("pyGlobusUrlCopy: call seek(%d)" % offset)
            self.g_dest.seek(offset)
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): call write()")
            ## handle out-of-space and other FS errors
            try:
                self.g_dest.write(buffer)
            except IOError, what:
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquire")
                self.condV.acquire()
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV acquired")
                self.logger.error("pyGlobusUrlCopy: data_func_get(): IO error on local write")
                self.g_cond = pyGlobusUrlCopy.ERROR
                self.error_status = sys.exc_info()
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV notify")
                self.condV.notify()
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV notified")
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV release")
                self.condV.release()
                self.logger.debug("pyGlobusUrlCopy: data_func_get(): condV released")
                return
                
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): call write() complete")

        if not eof:
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): not eof. call register_read()")
            self.register_read()
            self.logger.debug("pyGlobusUrlCopy: data_func_get(): not eof. call register_read() completed")


    def put_ftp(self):
        self.logger.debug("pyGlobusUrlCopy: put_ftp()")
        ## for PUT, work_dict holds list of offsets; offsets are removed from the dictionary
        ## when they are scheduled to be transferred
        self.work_dict = {}

        ## some accounting data structures
        self.done_list = []

        ## Open the local file for put
        srcFile = self.srcUrl[6:]
        try:
            file_size = os.stat(srcFile)[6]
        except OSError, what:
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquired")
            self.logger.error("pyGlobusUrlCopy: put_ftp(): OSError statting local file")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = sys.exc_info()
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV released")
            return

        self.logger.debug("pyGlobusUrlCopy: put_ftp: open(%s)" % srcFile)

        try:
            self.g_src = open(srcFile, "r")
        except IOError, what:
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquired")
            self.logger.error("pyGlobusUrlCopy: put_ftp(): IOError opening local file")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = sys.exc_info()
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV released")
            return

        ## exit out in case the ftp connection bailed asynchronously
        ## (there are some weird situations where failures can only be caught here)
        self.logger.debug("pyGlobusUrlCopy: put_ftp: condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: put_ftp: condV acquired")
        if self.g_cond == pyGlobusUrlCopy.ERROR:
            self.logger.error("pyGlobusUrlCopy: put_ftp(): Error in callback:")
            self.cleanup_and_handle_errors()
            return
        self.logger.debug("pyGlobusUrlCopy: put_ftp: condV release")
        self.condV.release()
        self.logger.debug("pyGlobusUrlCopy: put_ftp: condV released")

        if self.throughput:
            self.time0 = time.time()

        self.logger.debug("pyGlobusUrlCopy: put_ftp: call put()")
        self.ftp_client.put(self.dstUrl, self.done_func, None, self.source_operation_attribute)
        self.logger.debug("pyGlobusUrlCopy: put_ftp: call put() complete")
        self.netlogger_writer.write(2, "%s.put.start" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d FILE_SIZE=%d", 1, file_size)

        ## determine the set of offsets into the file that we are going to transfer
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): lock acquire")
        self.lock.acquire()
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): lock acquired")

        ## Generate a list of file offsets from which we will initiate transfers
        self.offsets_list = []
        ## use a long type here, since file offsets in pyglobus use long type
        offset = 0L
        while offset <= file_size:
            self.offsets_list.append(offset)
            offset += self.block_size
            
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): len(offsets_list) = %d" % len(self.offsets_list))
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): offsets_list = %s" % self.offsets_list)

        self.logger.debug("pyGlobusUrlCopy: put_ftp(): lock release")
        self.lock.release()
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): lock released")

        if self.parallelism > len(self.offsets_list):
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquired")
            error = "pyGlobusUrlCopy: put_ftp(): The parallelism (%d) can not be set higher than the number of block offsets (%d) in the file" % (self.parallelism, len(self.offsets_list))
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV released")
            return


        ## schedule the first N (number of parallel streams) writes
        for i in range(self.parallelism):
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): call register_write()")
            self.register_write()
            self.logger.debug("pyGlobusUrlCopy: put_ftp(): call register_write() complete")

        ## wait around til data callbacks are done.
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV acquired")
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV wait")
        self.condV.wait()
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV wait finished")

        self.cleanup_and_handle_errors()

        if self.throughput:
            time1 = time.time()
            dt = time1 - self.time0
            throughput = file_size/dt
            print "Throughput: %d bytes in %5.2f seconds, average = %5.2f bytes/sec" % (file_size, dt, throughput)
##             self.netlogger_writer.write(2, "%s.get.throughput" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "TOTAL_BYTES=%d TOTAL_TIME=%d THROUGHPUT=%f", 1, self.totalBytes, dt, throughput)

        self.netlogger_writer.write(2, "%s.put.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d TOTAL_BYTES=%d", 1, self.totalBytes)
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV release")
        self.condV.release()
        self.logger.debug("pyGlobusUrlCopy: put_ftp(): condV released")
        self.g_src.close()

        ## Perhaps put this in destructor?
        self.netlogger_writer.write(2, "%s.prog.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)

    ## convenience function to schedule a write
    def register_write(self):
        self.logger.debug("pyGlobusUrlCopy: register_write()")

        ## Retrieve an unqueued/untransmitted data block from the file
        self.logger.debug("pyGlobusUrlCopy: register_write: lock acquire")
        self.lock.acquire()
        self.logger.debug("pyGlobusUrlCopy: register_write: lock acquired")
        self.logger.debug("pyGlobusUrlCopy: register_write: offsets left: %d" % len(self.offsets_list))

        if len(self.offsets_list) == 0:
            self.logger.debug("pyGlobusUrlCopy: register_write: lock release")
            self.lock.release()
            self.logger.debug("pyGlobusUrlCopy: register_write: lock released")
            return

        offset = self.offsets_list.pop(0)
        self.logger.debug("pyGlobusUrlCopy: register_write: new offset: %d" % offset)
        eof = 0
        if len(self.offsets_list) == 0:
            self.logger.debug("pyGlobusUrlCopy: register_write: reached end of file")
            eof = 1
        self.logger.debug("pyGlobusUrlCopy: register write: lock release")
        self.lock.release()
        self.logger.debug("pyGlobusUrlCopy: register write: lock released")

        ## Read the data from the file
        self.logger.debug("pyGlobusUrlCopy: register write: seek(%d)" % offset)
        self.g_src.seek(offset)
        self.logger.debug("pyGlobusUrlCopy: register write: read(%d)" % self.block_size)
        try:
            buffer = self.g_src.read(self.block_size)
        except IOError, what:
            self.logger.debug("pyGlobusUrlCopy: register write: condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: register write: condV acquired")
            self.logger.error("pyGlobusUrlCopy: register_write(): read(%d) IOError:" % self.block_size)
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = sys.exc_info()
            self.cleanup_and_handle_errors()
            self.logger.debug("pyGlobusUrlCopy: register write: condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: register write: condV released")
            return

        ## Register the write
        self.logger.debug("pyGlobusUrlCopy: register write: lock acquire")
        self.lock.acquire()
        self.logger.debug("pyGlobusUrlCopy: register write: lock acquired")
        self.logger.debug("pyGlobusUrlCopy: register write: call register_write()")
        self.ftp_client.register_write(buffer, len(buffer),
                                       offset, eof, self.data_func_put, 0)
        self.logger.debug("pyGlobusUrlCopy: register write: call register_write() completed")
        self.work_dict[offset] = buffer
        self.logger.debug("pyGlobusUrlCopy: register write: lock release")
        self.lock.release()
        self.logger.debug("pyGlobusUrlCopy: register write: lock released")
        
    def data_func_put(self, arg, handle, buffer, bufHandle, bufLen, offset, eof, error):
        self.logger.debug("pyGlobusUrlCopy: data_func_put(arg=%s, handle=%s, bufHandle=%s, bufLen=%d, offset=%d, eof=%d, error=%s" % (arg, handle, bufHandle, bufLen, offset, eof, error))
        self.netlogger_writer.write(3, "%s.put.fragment" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d NUM_BYTES=%d EOF=%d", self.totalBytes, bufLen, eof)

        if error != "NONE" and error != (0, None):
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV acquire")
            self.condV.acquire()
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV acquired")
            self.logger.error("pyGlobusUrlCopy: data_func_put(): passed an error condition.")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV notify")
            self.condV.notify()
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV notified")
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV release")
            self.condV.release()
            self.logger.debug("pyGlobusUrlCopy: data_func_put: condV released")
            return
        elif self.first == 1:
            self.netlogger_writer.write(2, "%s.put.transferStart" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", 1)
            self.first = 0

        self.logger.debug("pyGlobusUrlCopy: data_func_put: lock acquire")
        self.lock.acquire()
        self.logger.debug("pyGlobusUrlCopy: data_func_put: lock acquired")
        self.logger.debug("pyGlobusUrlCopy: data_func_put: remove offset")
        del self.work_dict[offset]
        self.done_list.append( offset )
        self.totalBytes += bufLen
        self.logger.debug("pyGlobusUrlCopy: data_func_put: lock release")
        self.lock.release()
        self.logger.debug("pyGlobusUrlCopy: data_func_put: lock released")

        if not eof:
            self.logger.debug("pyGlobusUrlCopy: data_func_put: call register_write()")
            self.register_write()
            self.logger.debug("pyGlobusUrlCopy: data_func_put: call register_write() completed")

    def done_func(self, arg, handle, error):
        self.logger.debug("pyGlobusUrlCopy: done_func(arg=%s, handle=%s, error=%s)" % (arg, handle, error))
        ## we've been notified we're done by globus's ftpClient,
        ## so aquire the status variable lock and set status to DONE
        ## then notify the waiting thread (if there is an error, we set status to ERROR)

        self.logger.debug("pyGlobusUrlCopy: done_func: condV acquire")
        self.condV.acquire()
        self.logger.debug("pyGlobusUrlCopy: done_func: condV acquired")
        if error != "NONE" and error != (0, None):
            ## TODO: switch to new error reporting method
            self.logger.error("pyGlobusUrlCopy: done_func(): passed an error condition.")
            self.g_cond = pyGlobusUrlCopy.ERROR
            self.error_status = error
        else:
            self.g_cond = pyGlobusUrlCopy.DONE
            
        self.logger.debug("pyGlobusUrlCopy: done_func: condV notify")
        self.condV.notify()
        self.logger.debug("pyGlobusUrlCopy: done_func: condV notified")
        self.logger.debug("pyGlobusUrlCopy: done_func: condV release")
        self.condV.release()
        self.logger.debug("pyGlobusUrlCopy: done_func: condV released (%s)" % handle)

    def cleanup_and_handle_errors(self):
        """ This routine should only be called with the class condition variable held """
        if self.g_cond == pyGlobusUrlCopy.DONE:
            self.logger.debug("pyGlobusUrlCopy: cleanup_and_handle_errors(): completed successfully")
        else:
            self.logger.error("pyGlobusUrlCopy: cleanup_and_handle_errors(): failed w/ error")
            ## in case a string error message was provided in the error_status variable.
            if self.error_status == None or self.error_status == (0, None):
                self.logger.error("pyGlobusUrlCopy: cleanup_and_handle_errors(): No error.")
                return
            elif type(self.error_status) == type(""):
                self.netlogger_writer.write(0, "%s.prog.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", -1)
                self.logger.error("pyGlobusUrlCopy: cleanup_and_handle_errors(): error status:")
                map(self.logger.error, self.error_status.split("\n"))
                raise RuntimeError, self.error_status
            elif type(self.error_status) == type(()) and len(self.error_status) == 2 and type(self.error_status[0] == type(0)) and type(self.error_status[1] == type("")):
                self.logger.error("A globus call returned an error code: %d with an error string:\n%s" % self.error_status)
                raise RuntimeError, "A globus call returned an error code: %d with an error string:\n%s" % (self.error_status[0], self.error_status[1])
                
            ## in case an exception is sitting around for us, format it and log it
            elif type(self.error_status) == type(()) and len(self.error_status) == 3:
                self.netlogger_writer.write(0, "%s.prog.end" % pyGlobusUrlCopy.APPLICATION_EVENT_NAME, "VAL=%d", -1)
                ## Note: format_exception returns some results with an internal newline
                ## That doesn't log very well.  So we break up all the format_exception results,
                ## strip off newlines and remove empty lines
                fe = traceback.format_exception(self.error_status[0], self.error_status[1], self.error_status[2])
                sp = map(lambda x: x.split("\n"), fe)
                jo = []
                for el in sp:
                    jo.extend(map(lambda x: x.strip(), filter(lambda x: x != '', el)))
                map(self.logger.error, jo)
                try:
                    raise self.error_status[1], self.error_status[1].args
                except TypeError:
                    raise self.error_status[1]
            else:
                raise RuntimeError, "An internal pyglobus-url-copy error occured: an unrecognized error type was signaled:\n%s" % str(self.error_status)
            
            
    def __del__(self):
        self.logger.debug("pyGlobusUrlCopy: __del__()")
        ## should clean up any handles here, but Python is about to exit anyway...
## 	if netlogger:
##             self.logger.debug("pyGlobusUrlCopy: __del__() delete netlogger handle")
## 	    del self.netlogger_writer

def cli_main():
    global netlogger_dest

    usage = """pyglobus-url-copy [options] sourceURL destURL
    OPTIONS
             -h | --help | --usage
                  Print help
             -v | --version
                  Print version number
                  
             Performance options:
             
             -t <size in K> | --tcp-bs=<size in K> | --tcp-buffer-size=<size in K>
                  specify the size (in bytes) of the buffer to be used by the
                  underlying ftp data channels
             -b <size in K> | --bs=<block size in K> | --block-size=<block size in K>
                  specify the size (in bytes) of the buffer to be used by the
                  underlying transfer methods
             -p <parallelism> | --parallel=<parallelism>
                  specify the number of streams to be used in the ftp transfer


             Authentication options:

             -s  <subject> | --subject <subject>
                  Use this subject to match with both the source and dest servers
             --ss <subject> | --source-subject <subject>
                  Use this subject to match with the source server
             --ds <subject> | --dest-subject <subject>
                  Use this subject to match with the destionation server

             
             Debugging and logging options:

             --dbg | --debug
                  Enable the debug plugin
             -l <log level> | --logging-level=<log level>
                  specify the Python logging package logging level (CRITICAL/ERROR, WARNING, INFO, DEBUG)
             -n | --netlogger
                  Force netlogger to be enabled even if NETLOGGER_ON environment variable is not set
             --netlogger-destination=<destination>
                  specify the netlogger destination URL
             --netlogger-level=<level>
                  specify the netlogger log level (ERROR/FATAL, WARN, INFO, DEBUG)
             -w | --workaround
                  Required if pyglobus-url-copy is used with an older
                  release of pyGlobus (works around a threading/memory
                  allocation bug).  Warning: not intended for routine
                  use because it uses much more memory

             NOTE: The netlogger environment variables NETLOGGER_DEST
             overrides the --netlogger-destination command line
             option.
             
             NOTE: specifying a logging level will mask all messages
             logged with a priority level "below" that of the set
             level.  IE, if you set the logging level to INFO, then
             DEBUG level messages will not appear.
                  
    """

    tcp_buffer_size = pyGlobusUrlCopy.TCP_BUFFER_SIZE
    block_size = pyGlobusUrlCopy.BLOCK_SIZE
    parallelism = pyGlobusUrlCopy.PARALLELISM
    debug_plugin = pyGlobusUrlCopy.DEBUG_PLUGIN
    pyglobus_plugin = pyGlobusUrlCopy.PYGLOBUS_PLUGIN
    workaround = pyGlobusUrlCopy.WORKAROUND
    logger_level = pyGlobusUrlCopy.LOGGER_LEVEL
    source_subject = pyGlobusUrlCopy.SOURCE_SUBJECT
    dest_subject = pyGlobusUrlCopy.DEST_SUBJECT
    throughput = pyGlobusUrlCopy.THROUGHPUT
        
    ## We use a default level of INFO, *unless* the netlogger_dest variable contains "?level="
    ## Otherwise the ?level= value would be overwrriten, and the env. vars have precendence
    if netlogger_dest.find("?level=") == -1:
        netlogger_level = netlogger.Writer.INFO_LVL
    else:
        netlogger_level = None

    
    ## set up error logging before anything else
    logger = logging.getLogger(pyGlobusUrlCopy.APPLICATION_EVENT_NAME)
    logger_handler = logging.StreamHandler(sys.stdout)
    logger_formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    logger_handler.setFormatter(logger_formatter)
    logger.addHandler(logger_handler)

    flags = netlogger.OutputFlags.ULM | netlogger.OutputFlags.FLUSH
    netlogger_writer = netlogger.open(netlogger_dest,
                                      flags = flags,
                                      program = pyGlobusUrlCopy.APPLICATION_EVENT_NAME,
                                      threaded = 1,
                                      backup = None)

    ## process source and dest args
    try:
        opts, args = getopt.getopt(sys.argv[1:], "ht:b:p:l:n:s:dvywr", ["help", "usage", "tcp-bs=", "tcp-buffer-size=", "bs=", "block-size=", "parallel=", "dbg", "debug", "logging-level=", "netlogger", "netlogger-destination=", "netlogger-level=", "version", "pyglobus-plugin", "workaround", "subject=", "ss=", "source-subject=", "ds=", "dest-subject=", "--throughput"])
    except getopt.GetoptError, what:
        logger.error("GetOpt error: %s" % what)
        logger.error(usage)
        sys.exit(1)

    ## process options
    for opt, arg in opts:
        if opt in ("-h", "--help", "--usage"):
            logger.error(usage)
            sys.exit(0)
        if opt in ("-v", "--version"):
            logger.error("pyglobus-url-copy version %s" % __version__)
            sys.exit(0)
        if opt in ("-t", "--tcp-bs", "--tcp-buffer-size"):
            try:
                tcp_buffer_size = int(arg)
            except ValueError:
                logger.error("TCP buffer size must be an integer")
                sys.exit(1)
            if tcp_buffer_size < 1:
                logger.error("TCP buffer size must be greater than 0")
                sys.exit(1)
        elif opt in ("-b", "--bs", "--block-size"):
            try:
                block_size = int(arg)
            except ValueError:
                logger.error("Block size must be an integer")
                sys.exit(1)
            if block_size < 1:
                logger.error("Block size must be greater than 0")
                sys.exit(1)
        elif opt in ("-p", "--parallel"):
            try:
                parallelism = int(arg)
            except ValueError:
                logger.error("Parallelism must be an integer")
                sys.exit(1)
            if parallelism < 1:
                logger.error("Parallelism must be greater than 0")
                sys.exit(1)

        elif opt in ("-n", "--netlogger"):
            os.putenv("NETLOGGER_ON", "yes")

        elif opt == "--netlogger-destination":
            netlogger_dest = arg

        elif opt == "--netlogger-level":
            if arg == 'DEBUG':
                netlogger_level = netlogger.Writer.DEBUG_LVL
            elif arg == 'INFO':
                netlogger_level = netlogger.Writer.INFO_LVL
            elif arg == 'WARN':
                netlogger_level = netlogger.Writer.WARN_LVL
            elif arg == 'ERROR':
                netlogger_level = netlogger.Writer.ERROR_LVL
            elif arg == 'FATAL':
                netlogger_level = netlogger.Writer.FATAL_LVL
            else:
                logger.error("Unrecognized NetLogger logging level '%s'" % arg)
                sys.exit(1)

        elif opt in ("-d", "--dbg", "--debug"):
            debug_plugin = True

        elif opt in ("-y", "--pyglobus-plugin"):
            pyglobus_plugin = True

        elif opt in ("-w", "--workaround"):
            workaround = 1
        elif opt in ("-l", "--logging-level"):
            if arg == 'DEBUG':
                logger_level = logging.DEBUG
            elif arg == 'INFO':
                logger_level = logging.INFO
            elif arg == 'WARNING':
                logger_level = logging.WARNING
            elif arg == 'ERROR':
                logger_level = logging.ERROR
            elif arg == 'CRITICAL':
                logger_level = logging.CRITICAL
            else:
                logger.error("Unrecognized logging level '%s'" % arg)
                sys.exit(1)
            logger.setLevel(logger_level)
        elif opt in ("-s", "--subject"):
            source_subject = arg
            dest_subject = arg
        elif opt in ("--ss", "--source-subject"):
            source_subject = arg
        elif opt in ("--ds", "--dest-subject"):
            dest_subject = arg
        elif opt in ("-r", "--throughput"):
            throughput = 1

    if len(args) < 2:
        logger.error("A source URL and a destination URL must be provided")
        logger.error(usage)
        sys.exit(1)

    elif len(args) > 2:
        logger.error("Too many URLs were provided")
        logger.error(usage)
        sys.exit(1)

    srcUrl = args[0]
    dstUrl = args[1]

    if netlogger_level:
        netlogger_writer.setLevel(netlogger_level)

    p = pyGlobusUrlCopy(srcUrl = srcUrl, dstUrl = dstUrl, tcp_buffer_size = tcp_buffer_size, block_size = block_size, parallelism = parallelism, logger = logger, netlogger_writer = netlogger_writer, debug_plugin = debug_plugin, pyglobus_plugin = pyglobus_plugin, workaround = workaround, source_subject = source_subject, dest_subject = dest_subject, throughput = throughput)
    logging.shutdown()

netlogger_dest = pyGlobusUrlCopy.NETLOGGER_DEST
if os.getenv("NETLOGGER_DEST"):
    netlogger_dest = os.getenv("NETLOGGER_DEST")

if __name__ == '__main__':
    cli_main()
