############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
########################################################################### 
"""
Distutils script to build, install, distribute pyGlobus
"""

from distutils.core import setup, Extension
from distutils.spawn import spawn
from distutils.sysconfig import get_python_inc
from distutils.sysconfig import get_config_vars
from distutils.util import get_platform 
import sys, re, exceptions, glob, os
import pickle
import ConfigParser
from time import gmtime, strftime, time


begin_time = time()


SUN_CC = 'cc'
LD_SHARED = "ld -G"
CC_SHARED ="-KPIC"
CC_OPTS = "-DNDEBUG -g"

GLOBUS_LOCATION = None
GPT_LOCATION = None
FLAVOR = "gcc32dbgpthr"
RUN_SWIG      = 0
WARNINGS_ON   = 0
COMPILE_CC    = 0
DEBUG         = 0
MAKE_MFH_CACHE = 0
USE_MFH_CACHE  = 0
RUN_EPYDOC    = 0
RUN_EPYDOCIFY = 1
GLOBUS_LIBS = None
GLOBUS_PACKAGES = None
SWIG_PROG = 'swig'
SWIG_ARGS = ["-python", "-noproxy", "-DHAVE_CONFIG_H",
             "-I%s" % get_python_inc(plat_specific=1)]
GLOBUS_DEPS = {"util" : "globus_common"}
MAKEFILE_HEADER = None
BUILT_LIST = []
FAILED_LIST = []
MODULE_LIST = ['util', 'ftpClient', 'ftpControl', 'gassCache', 'gassCopy', 
               'gassServerEZ', 'myProxy', 'gassTransfer','gramClient','io',
               'security', 'rlsClient', 'xio']
BUILD_MODULE_LIST = 1

class DependencyException(exceptions.Exception):
    """
    Exception class used when there are problems with the globus configuration
    """
    pass


HelpText = """ 
           ***pyGlobus      setup.py*** 
           
           --help            This help text 
           
           --flavor=X        Set build flavor to X, ie gcc64dbgpthr 
           
           --run-swig        Run SWIG to generate Globus shared objects.
                             Note: Only needs to be called if src or
                             pyGlobus source has been modified.
           
           --compile-cc      Run compiler (Solaris only) 
           
           --use-compiler=X  Use compiler X for build 
           
           --with-modules=mod1,mod2, ...  Specify a list of modules  
             to build. Valid modules are util, ftpClient, ftpControl,
             gassCache, gassCopy, gassServerEZ, myProxy, gassTransfer, 
             gramClient, io, security, rlsClient, xio. All of the modules
             are dependent on the util module, some of the modules
             are dependent on the security module.
           
                              
           --link-with=X     Link with shared object list X. 
           
           --shared-args=X   List of shared objects to build with.
           
           --compiler-opts=X List of compiler options to build with.
           
           --debug           Build with debug options enabled.
           
           --run-epydoc      Generate epydoc reference documentation. 
           
           --make-gmfh-cache Generate Globus makefile header cache.
           
           --use-gmfh-cache  Use Globus makefile header cache.
"""

# Get GLOBUS_LOCATION, exit if it doesn't exist
try:
    GLOBUS_LOCATION = os.environ["GLOBUS_LOCATION"]
except KeyError:
    print "GLOBUS_LOCATION must be set before installing pyGlobus"
    sys.exit(-1)


swigVerRegEx = r"\s*SWIG Version\s*"
SWIG_PATTERN = re.compile(swigVerRegEx)
errRegEx = r".*ERROR .*|\s*No packages were found.*|\s*ERROR: .*"
incRegEx = r"\s*GLOBUS_INCLUDES\s*=\s*"
cflagsRegEx = r"\s*GLOBUS_CFLAGS\s*=\s*"
pkgLibRegEx = r"\s*GLOBUS_PKG_LIBS\s*=\s*"
ldFlagsRegEx = r"\s*GLOBUS_LDFLAGS\s*=\s*"
libsRegEx = r"\s*GLOBUS_LIBS\s*=\s*"
ERR_PATTERN = re.compile(errRegEx)
INC_PATTERN = re.compile(incRegEx)
CFLAGS_PATTERN = re.compile(cflagsRegEx)
PKGLIBS_PATTERN = re.compile(pkgLibRegEx)
LDFLAGS_PATTERN = re.compile(ldFlagsRegEx)
LIBS_PATTERN = re.compile(libsRegEx)

def swig_exit_error():
    print "You need to be using Swig version 1.3.24 or newer"
    sys.exit(-1)

def check_swig_version():
    """
    Make sure we are using swig version 1.3.24 or newer
    """
    inStrm, outStrm = os.popen4("%s -version" % SWIG_PROG)
    for line in outStrm: 
        m = SWIG_PATTERN.match(line)
        if m: 
            version = line.split("SWIG Version ")[1]
            versioning = version.strip().split('.') 
            try:
                version_list = map(int, versioning)
            except ValueError: 
                swig_exit_error()
            if version_list[0] < 1:
                swig_exit_error()
            if version_list[1] < 3:
                swig_exit_error()
            if version_list[2] < 24:
                swig_exit_error()
                
def check_python_version():
    """
    Check the python version and modify setup if necessary
    """
    #we modify setup if we are using
    #a 2.3 version of python, setup
    #is in the global namespace
    global setup

    if sys.version_info < (2, 3):
        print "pyGlobus requires Python 2.3 or later"
        sys.exit(-1) 

def parse_args():
    """
    handle all of the arguments
    This function sets a lot of globals
    """
    global FLAVOR, MAKE_MFH_CACHE, USE_MFH_CACHE, RUN_SWIG , COMPILE_CC
    global SUN_CC, LD_SHARED, CC_SHARED, CC_OPTS, RUN_EPYDOC, DEBUG
    global MODULE_LIST, BUILD_MODULE_LIST
    # parse special command line options and remove them
    # this way distutils won't attempt to parse them.
    args = sys.argv[1:]
    for arg in args:
        if arg.find('--help') == 0:
            sys.argv.remove(arg)
            print HelpText 
            sys.exit(0)
        if arg.find( '--flavor=') == 0:
            FLAVOR = arg.split('=')[1]
            sys.argv.remove(arg)
        elif arg.find('--make-gmfh-cache') == 0:
            MAKE_MFH_CACHE = 1
            sys.argv.remove(arg)
        elif arg.find('--use-gmfh-cache') == 0:
            USE_MFH_CACHE  = 1
            sys.argv.remove(arg)
        elif arg.find('--run-swig') == 0:
            RUN_SWIG = 1
            check_swig_version()
            sys.argv.remove(arg)
        elif arg.find('--compile-cc') == 0:
            COMPILE_CC = 1
            sys.argv.remove(arg)
        elif arg.find('--use-compiler') == 0:
            if COMPILE_CC:
                SUN_CC = arg.split('=')[1]
                sys.argv.remove(arg)
            else:
                print "You must specify --compile-cc"
                print "if you want to specify a compiler"
                print "Exiting"
                sys.exit(1)
        elif arg.find('--link-with') == 0:
            if COMPILE_CC:
                LD_SHARED = arg.split('=')[1]
                sys.argv.remove(arg)
            else:
                print "You must specify --compile-cc"
                print "if you want to specify linking options"
                print "Exiting"
                sys.exit(1)
        elif arg.find('--shared-args') == 0:
            if COMPILE_CC:
                CC_SHARED = arg.split('=')[1]
                sys.argv.remove(arg)
            else:
                print "You must specify --compile-cc" 
                print "if you want to specify shared argument options"
                print "Exiting"
                sys.exit(1)
        elif arg.find('--compiler-opts') == 0:
            if COMPILE_CC:
                CC_OPTS = arg.split('=')[1]
                sys.argv.remove(arg)
            else:
                print "You must specify --compile-cc" 
                print "if you want to specify compiler options"
                print "Exiting"
                sys.exit(1)
        elif arg.find('--run-epydoc') == 0:
            RUN_EPYDOC = 1
            sys.argv.remove(arg)
        elif arg.find('--debug') == 0:
            DEBUG = 1
            sys.argv.remove(arg)
        elif arg.find('--with-modules') == 0:
            modules = arg.split('=')[1]
            mList = []
            while not modules == '':
                indx = modules.find(',')
                if indx == -1:
                    mName = modules
                    modules = ''
                else:
                    mName = modules[:indx]
                    modules = modules[indx +1:]
                mList.append(mName)
            MODULE_LIST = mList
            sys.argv.remove(arg)
        elif arg.find('register') == 0:
            BUILD_MODULE_LIST = 0
        elif arg.find('sdist') == 0:
            BUILD_MODULE_LIST = 0
        elif arg.find('clean') == 0:
            BUILD_MODULE_LIST = 0
    return MODULE_LIST

#Find globus-makefile-header. If unable to find it, exit
def check_globus_makefile_header():
    global MAKEFILE_HEADER
    gmh_path_list = [os.path.join("%s", "sbin", "globus-makefile-header"),
                     os.path.join("%s", "sbin", "globus-makefile-header.py"),
                     os.path.join("%s", "bin", "globus-makefile-header"),
                     os.path.join("%s", "bin", "globus-makefile-header.py"),
                     "globus-makefile-header.py"]
                                                                                
    for gmhp in gmh_path_list:
        if gmhp.startswith("%"):
            tp = gmhp % GLOBUS_LOCATION
        else:
            tp = gmhp
        if os.path.isfile(tp):
            MAKEFILE_HEADER = "%s -flavor=%s" % (tp, FLAVOR)

    if MAKEFILE_HEADER == None:
        print """Unable to locate globus-makefile-header or 
                 globus-makefile-header.py in either 
                 $GLOBUS_LOCATION/[bin,sbin]. This must be 
                 found before installing pyGlobus."""
        sys.exit(-2)
    

def build_module_lists(mList = MODULE_LIST):
    """
    Create the module lists necessary for building each 
    extension and module of pyGlobus
    """
    print "Checking for which modules to build"
    swigList = []
    pyModList = []
    eModList = []
    config = ConfigParser.ConfigParser()
    config.read('setup.cfg')
    for module in mList:
        check_module(pyModList, eModList, swigList, config, module)
        
    return pyModList, eModList, swigList


def addSwigFiles(config, module, swigList):
    """
    Build the list of interface files and output ._wrap.c 
    files. 
    """
    swigInterfaceFile = config.get(module, "swigInterface") 
    try: 
        config.getboolean(module, "checkVersion")
    except ConfigParser.NoOptionError, ex:
        swigOutputFile = config.get(module, "swigOutput")
        swigList.append((swigInterfaceFile, swigOutputFile))
    else:
        #build a dictionary of filenames without versioning
        #info mapping to tuples which have major and minor
        #numbers  
        swigFilesList = swigInterfaceFile.split()
        checkVersionFiles = {}
        for swigFile in swigFilesList:
            index = swigFile.find('-') 
            if not index == -1:
                index2 = swigFile.find('.')
                major = int(swigFile[index +1:index2])
                index3 = swigFile.find('.i')
                minor = int(swigFile[index2 +1:index3])
                versionTuple = ( major, minor)
                fileStem = swigFile[:index]
                if checkVersionFiles.has_key(fileStem):
                    checkVersionFiles[fileStem].append(versionTuple)
                else:
                    checkVersionFiles.update({fileStem: [versionTuple]})
            else:
                index = swigFile.find('.i')
                fileStem = swigFile[:index]
                swigList.append((swigFile, fileStem + '_wrap.c'))

        try:
            config.get(module, "multipleDependencies")
        except ConfigParser.NoOptionError, ex:
            deps = config.get(module, "dep")
        else:
            deps = config.get(module, "deps")
        depsList = deps.split()
        i = 0
        for swigFile in checkVersionFiles:
            dep = depsList[i]
            i += 1
            major, minor = check_version(dep)
            versionTuple = (major, minor)
            versionTupleList = checkVersionFiles[swigFile]
            latestVersion = max(versionTupleList) 
            #Select all versions less than or equal to the version Tuple
            #returned by gpt-query
            #Select the latest version from that set 
            filter_func = lambda a:  a <= versionTuple 
            versionTuples =  filter(filter_func, versionTupleList)
            try:
                latestVersion = max(versionTuples) 
                interfaceSuffix = ".".join([str(latestVersion[0]),
                                           str(latestVersion[1]), 'i'])
                swigInterfaceFile = "-".join([swigFile, interfaceSuffix]) 
                swigOutputFile = "".join([swigFile, '_wrap.c'])
                swigList.append((swigInterfaceFile, swigOutputFile))
                
            except ValueError:
                raise RuntimeError("""could not select a version of 
                                      a swig interface file""", swigFile)
            
                
def check_module(pyModList, eModList, swigList, config, module): 
    print "Checking the ", module, " module" 
    deps = config.get(module, "dep")
    depsList = deps.split()
    dep = depsList[0]
    try:
        libList, incList, cFlagsList, ldFlagsList = get_makefile_output(dep)
    except DependencyException, ex:
        print "Missing dependency to build the ", module, " package." 
        print "The following is the error message from globus-makefile-header." 
        print ex.args 
        if module == "util": 
            print "This package is required to build all of pyGlobus."
            print "Now exiting from the build"
            sys.exit(-1) 
        FAILED_LIST.append(module)
        return
    extensionName = config.get(module, "extensionName")
    nameList = extensionName.split()
   
    extensionInfoListTuples = []
    sourceList = [] 
    
    for name in nameList:
        sourceList = ["".join(["src/", name[:-1], "_wrap.c"])]
        extensionInfoListTuples.append((name, sourceList))
        
    supportThreading = config.getboolean(module, "supportThreading") 
    if supportThreading:
        sourceList.append("src/pyGlobus_common.c")
    
    for infoTuple in extensionInfoListTuples:
        e = Extension(infoTuple[0], infoTuple[1:][0], include_dirs=incList,
                      library_dirs=ldFlagsList, extra_compile_args=cFlagsList,
                      libraries=libList)
        eModList.append(e) 
    if RUN_SWIG:
        addSwigFiles(config, module, swigList)
    mods = config.get(module, "pyMods")
    modsList = mods.split()
    for mod in modsList:
        pyModList.append(mod)        
    BUILT_LIST.append(module)
    print "The ", module, " module will be built"
     
def get_makefile_output(dep):
    pLibsStr = None
    eLibsStr = None
    cFlagsStr = None
    includeStr = None
    ldFlagsStr = None
    libList = []
    incList = []
    ldFlagsList = []
    cFlagsList = []

    mfh_path = os.path.join('gmfh_cache', get_platform(), 
                            str(abs(hash(os.environ['GLOBUS_LOCATION']))), 
                            FLAVOR)

    if USE_MFH_CACHE == 1:
        picklein = None
        try:
            dep_path = os.path.join(mfh_path, dep)
            picklein = pickle.load(open(dep_path))
        except: 
            print "raising dependency exception here"
            ex = DependencyException('No cache file for %s' % (dep_path))
            raise ex
        
        [libList, incList, cFlagsList, ldFlagsList] = picklein
        return libList, incList, cFlagsList, ldFlagsList

    inStrm, outStrm = os.popen4("%s %s" % (MAKEFILE_HEADER, dep))
    for line in outStrm.readlines():
        m = ERR_PATTERN.match(line)
        if m:
            raise DependencyException(line)
        m = INC_PATTERN.match(line)
        if m:
            includeStr = INC_PATTERN.split(line)[1]
            for strr in includeStr.split():
                s = strr.strip()[2:]
                incList.append(s)
            continue
        m = CFLAGS_PATTERN.match(line)
        if m:
            cFlagsStr = CFLAGS_PATTERN.split(line)[1]
            for strr in cFlagsStr.split():
                if strr == "-Wall":
                    if WARNINGS_ON == 1:
                        cFlagsList.append(strr)
                if strr != "-Wall":
                    cFlagsList.append(strr)
            continue
        m = PKGLIBS_PATTERN.match(line)
        if m:
            pLibsStr = PKGLIBS_PATTERN.split(line)[1]
            i = 0
            for strr in pLibsStr.split():
                s = strr.strip()[2:]
                libList.append(s)
                i += 1
            continue
        m = LDFLAGS_PATTERN.match(line)
        if m:
            ldFlagsStr = LDFLAGS_PATTERN.split(line)[1]
            for strr in ldFlagsStr.split():
                s = strr.strip()[2:]
                ldFlagsList.append(s)
            continue
        m = LIBS_PATTERN.match(line)
        if m:
            eLibsStr = LIBS_PATTERN.split(line)[1]
            for strr in eLibsStr.split():
                s = strr.strip()[2:]
                libList.append(s)
            continue
   
    if DEBUG:
        libList.append('wad')

    if MAKE_MFH_CACHE == 1: 
        cachefile = []
        try:
            os.makedirs(mfh_path)
        except:
            pass
        try:
            os.remove(os.path.join(mfh_path, dep))
        except:
            pass
        try:
            cachefile = open(os.path.join(mfh_path, dep),"w")
        except s:
            print 'open failed', s
            print """make makefile header had problems 
                     loading cachefile, try --make-gmfh-cache first"""
            sys.exit(-1)
        pickle.dump([libList, incList, cFlagsList, ldFlagsList], cachefile)
        
    return libList, incList, cFlagsList, ldFlagsList

def check_version(dep):
    GPT_QUERY = None
    for directory in os.environ["PATH"].split(os.path.pathsep):
        gpt_query_path = os.path.join(directory, "gpt-query") 
        if os.path.isfile(gpt_query_path):
            GPT_QUERY = gpt_query_path
            break
    if GPT_QUERY is None:
        print "You must have gpt-query in your path"
        print "Exitting"
        sys.exit(-1)


    mfh_path = os.path.join('gmfh_cache', os.uname()[0],
                            str(hash(os.environ['GLOBUS_LOCATION'])**2),FLAVOR)

    # case 1: make cache
    if MAKE_MFH_CACHE == 1:          
        inStr, outStr = os.popen4("%s %s"% (GPT_QUERY, dep)) 
        for line in outStr: 
            if line.find(dep) != -1:
                try:
                    line.index("version:")
                    tmpstr = line.split("version: ")[1]
                    version = tmpstr.split(" cmp")[0]
                    major, minor, minorMinor = version.split('.') 
                except ValueError, ex: 
                    tmpstr = line.split("ver: ")[1] 
                    version = tmpstr.split(" cmp")[0]
                    major, minor = version.split('.')

                cachefile = []
                try:
                    os.makedirs(mfh_path)
                except:
                    pass
                try:
                    os.remove(os.path.join(mfh_path, dep+'.versions'))
                except:
                    pass
                try:
                    cachefile = open(os.path.join(mfh_path, dep+'.versions'),"w")
                except:
                    print 'open failed'
                    print 'make makefile header had problems loading cachefile'
                    print "Exiting"
                    sys.exit(-1)
                pickle.dump([major, minor ], cachefile) 
                return (int(major), int(minor))

    # case 2: use cache
    if USE_MFH_CACHE == 1:         
        picklein = None
        try: 
            cache_file = open(os.path.join(mfh_path, dep+'.versions'))
            picklein = pickle.load(cache_file)
        except: 
            ex = DependencyException('No cache file for %s' % 
                                    (os.path.join(mfh_path, dep+'.versions')))
            raise ex
        [major, minor] = picklein
        return (int(major), int(minor))
        
    # case 3: normal
    inStr, outStr = os.popen4("%s %s"% (GPT_QUERY, dep))   # case 3: normal
    for line in outStr: 
        if line.find(dep) != -1:
            try:
                line.index("version:")
                tmpstr = line.split("version: ")[1]
                version = tmpstr.split(" cmp")[0]
                major, minor, minorMinor = version.split('.') 
            except ValueError, ex: 
                tmpstr = line.split("ver: ")[1] 
                version = tmpstr.split(" cmp")[0]
                major, minor = version.split('.')
            return (int(major), int(minor))
        
def make_data_list():
    dir_list = os.listdir(os.getcwd())
    try:
        dir_list.index("doc")
    except ValueError, ex:
        return []  
        # filter non-existent files
    def filter_files(x):
        data_files = []
        for d, xlist in x:
            filelist = filter(os.path.exists, xlist) 
            if filelist:
                data_files.append((d, filelist))
        return data_files
        # make file list
    doc_dir = os.path.join('doc', 'pyGlobus')
    idir = lambda d: os.path.join(doc_dir, *d)
    sfile = lambda d: os.path.join('doc', *d)
    sglob = lambda d: glob.glob(os.path.join('doc', *d))
    data_files = filter_files([(idir(['.']), [sfile(['index.html']),
                              sfile(['epydoc.css'])]),
                              (idir(['public']), sglob(('public', '*'))),
                              (idir(['private']), sglob(('private', '*'))),
                              (idir(('Guides', 'SysAdmin')),
                              sglob(('doc', 'Guides', 'SysAdmin','*.html'))),])
    return data_files


# Distutils doesn't currently work with recent 1.3.x
# versions of swig.
def run_swig(swigList):
    for s in swigList:
        cmd = [SWIG_PROG] + SWIG_ARGS + ["-o", "./src/%s" % (s[1]),
                                      "./src/%s" % (s[0])]
        
        spawn(cmd, verbose=1)

#Replace configuration variables to use the cc compiler and linker
def run_compile_cc(cv):
    cv["CC"] = SUN_CC
    cv["CCSHARED"] = CC_SHARED
    cv["OPT"] = CC_OPTS
    cv["LDSHARED"] = LD_SHARED

def get_build_directory():
    """constructs the directory path of the build directory, 
    epydoc uses this"""
    
    directory = os.path.join(os.getcwd(), 'build') 
    libdir = ".".join(["lib", get_platform()])  
    version_string = ".".join([str(sys.version_info[0]), 
                               str(sys.version_info[1])])
    builddir = "-".join([libdir, version_string])
    return  os.path.join(directory, builddir)

	
def run_epydocify(pyModList):
    # will copy epydocified files to $PYTHONPATH/pyGlobus
    # but keep pyGlobus/ source the same!
    try: spawn(['rm', '-rf', 'tmp-epydoc'])   
    except: pass
    try:    
        spawn(['mkdir', 'tmp-epydoc'])
    except: 
        print 'failed to make tmp-epydoc directory, exiting'
        sys.exit(-1)
    print pyModList
    flist = []
    for f in pyModList:
        if not f.startswith('pyGlobus.test'):
            newF = f.replace('.', os.sep)
            newF = newF + '.py'
            flist.append(newF)
    for x in flist:
        try:
            spawn(['cp', x, "tmp-epydoc"])
        except:
            print 'error copying files'
    runstr = [sys.executable, 'scripts/epydocify.py']
    # run epydocify on all source in tmp dir
    runstr.extend(glob.glob('tmp-epydoc/*.py'))
    spawn(runstr, verbose=1)
    runstr = [sys.executable, 'scripts/epydocify.py', '--replace-old-source']
    # run epydocify repsource on tmpdir
    flist = glob.glob('tmp-epydoc/*.py')
    runstr.extend(flist) 
    spawn(runstr, verbose=1)

    flist = glob.glob('tmp-epydoc/*.py')
    for x in flist:
        directory = get_build_directory()
        thecommand = ['cp', x, directory + '/pyGlobus/']
        spawn(thecommand, verbose=1)
        
    try:
        spawn(['rm', '-rf', 'tmp-epydoc'])
    except: 
        print "unable to rm a file or directory"  


def run_epydoc():
    # this gets run on the $PYTHONPATH/pyGlobus directory

    directory = get_build_directory() + '/pyGlobus/*.py'
    flist = glob.glob(directory)    
    cmd = ['epydoc', '--ignore-param-mismatch', '-o', 'doc'] 
    cmd.extend(flist) 
    spawn(cmd, verbose=1)
    
def main():
    check_python_version()
    modsList = parse_args()
    check_globus_makefile_header()
    fixCompileFlags()
    pyModList = []
    eModList = []
    swigList = []
    if BUILD_MODULE_LIST:
        pyModList, eModList, swigList = build_module_lists(modsList)
    dataList = make_data_list()
    if RUN_SWIG:
        run_swig(swigList)

    runSetup(pyModList, eModList, dataList)
    if RUN_EPYDOC:
        run_epydocify(pyModList)
        run_epydoc()
        
#Remove flags from the compile line that through lots
# of warnings.
def fixCompileFlags():
    cv = get_config_vars() 
    try: 
        opts = cv["OPT"]
        opts = opts.replace("-Wall", "")
        opts = opts.replace("-Wstrict-prototypes", "")
        cv["OPT"] = opts
    except:
        print "No OPT variable set"
    

    if sys.platform[:6] == 'darwin':
        ld = cv['LDSHARED']
        ld += " -bundle_loader %s" % sys.executable
        cv['LDSHARED'] = ld

    if COMPILE_CC:
        run_compile_cc(cv)
        

        
def runSetup(pyModList, eModList, dataList):
    ###
    # For PyPI
    ###
    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Programming Language :: Python
        Programming Language :: C
        Natural Language :: English
        Operating System :: Microsoft :: Windows
        Operating System :: POSIX
        """

    longDesc = """The goal of this project is to allow the use of the entire  
    Globus toolkit from Python, a high-level scripting language.  SWIG is used
    to generate the necessary interface code.  Currently the majority of the 
    2.4 versions and later of the Globus toolkit have been wrapped"""
    


    setup(name="pyGlobus", version="gt4.0.1-1.15", 
          description="Python Globus Toolkit wrapper",
          maintainer="Keith Jackson", maintainer_email="KRJackson@lbl.gov",
          url="http://www-itg.lbl.gov/gtg/pyGlobus/", platforms=["any"],
          classifiers = filter(None, classifiers.split("\n"))[:-1],
          long_description=longDesc,
          license="http://www-itg.lbl.gov/gtg/projects/download/license.html",
          py_modules=pyModList, ext_package="pyGlobus", ext_modules=eModList,
          data_files=dataList)

    
    
if __name__ == '__main__': main()
    
print "Built the following packages:\n"
for mod in BUILT_LIST:
    print mod
if not FAILED_LIST == []:
    print """\nFailed to build the following packages 
               due to missing dependencies:\n"""
    for mod in FAILED_LIST:
        print mod

print 'Build finished at: ', strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
print 'Completed in', time() - begin_time, 'seconds'
