###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import base64, sha
from zope.interface import classProvides, implements, Interface
from twisted.python import log

from Ft.Xml import XPath

from ZSI.parse import ParsedSoap
from ZSI.TC import _get_global_element_declaration as GED
from ZSI.wstools.Namespaces import BEA, DSIG, OASIS, SOAP, GLOBUS
from ZSI.wstools.c14n import Canonicalize
from ZSI.twisted.WSresource import DataHandler, CheckInputArgs,\
    WSAddressHandler, DefaultHandlerChain, HandlerChainInterface,\
    CallbackChainInterface

from pyGridWare.utility.web.resource import CallbackHandler


#
# Global Element Declarations
# 

UsernameTokenDec = GED(OASIS.WSSE, "UsernameToken")
SecurityDec = GED(OASIS.WSSE, "Security")
SignatureDec = GED(DSIG.BASE, "Signature")
PasswordDec = GED(OASIS.WSSE, "Password")
NonceDec = GED(OASIS.WSSE, "Nonce")
CreatedDec = GED(OASIS.UTILITY, "Created")
InclusiveNamespacesDec = GED(DSIG.C14N_EXCL, 'InclusiveNamespaces')
SecurityTokenReferenceDec = GED(OASIS.WSSE, "SecurityTokenReference")
ReferenceDec = GED(OASIS.WSSE, "Reference")
SecurityContextTokenDec = GED(BEA.SECCONV, "SecurityContextToken")

#
# pyclass
#
try:
    UsernameToken = UsernameTokenDec.pyclass
    Security = SecurityDec.pyclass
    Signature = SignatureDec.pyclass
    Password = PasswordDec.pyclass
    Nonce = NonceDec.pyclass
    Created = CreatedDec.pyclass
    InclusiveNamespaces = InclusiveNamespacesDec.pyclass
    SecurityTokenReference = SecurityTokenReferenceDec.pyclass
    SecurityContextToken = SecurityContextTokenDec.pyclass
    Reference = ReferenceDec.pyclass
except:
    raise ImportError, 'required global element(s) unavailable: %s ' %({
        (OASIS.WSSE, "UsernameToken"):UsernameTokenDec,
        (OASIS.WSSE, "Security"):SecurityDec,
        (DSIG.BASE, "Signature"):SignatureDec,
        (OASIS.WSSE, "Password"):PasswordDec,
        (OASIS.WSSE, "Nonce"):NonceDec,
        (OASIS.UTILITY, "Created"):CreatedDec,
        })
    
_is_security_token_ref = \
    lambda pyobj: pyobj.typecode is SecurityTokenReferenceDec
    
def _find_security_token_from_ref(sctref, security):
    """
    Parameters:
        sctref -- SecurityContextTokenReference
        security -- Security Header
    """
    if sctref.typecode is not SecurityTokenReferenceDec:
        raise TypeError, 'Expecting SecurityTokenReference'
   
    uri = None 
    for pyobj in sctref.Any:
        if pyobj.typecode is not ReferenceDec: continue
        vt = pyobj.get_attribute_ValueType()
        uri = pyobj.get_attribute_URI()
        break
    else:
        return None
   
    for pyobj in security.Any:
        if pyobj.typecode is SecurityContextTokenDec and \
            '#%s' %pyobj.get_attribute_Id() == uri:
                return pyobj
    return None


class WSSecurityHandler:
    implements(HandlerChainInterface)
    tokenRef = 10101
    
    def __init__(self, signature):
        """
        Parameters:
            signature -- Signature Handler
        """
        self.signature = signature
    
    def processRequest(self, ps, **kw):
        log.msg("Verify Message", debug=True)
        
        if not isinstance(ps, ParsedSoap):
            raise TypeError,'Expecting ParsedSoap instance'
        
        #TODO: Multiple Security Headers?
        maxOccurs = SecurityDec.maxOccurs
        SecurityDec.maxOccurs = "unbounded"
        security = ps.ParseHeaderElements([SecurityDec])
        SecurityDec.maxOccurs = maxOccurs
        security[(SecurityDec.nspname,SecurityDec.pname)]

        for pyobj in security[(SecurityDec.nspname,SecurityDec.pname)]:
            #TODO: Is this Security Header for me?

            for any in pyobj.Any or []:
                if any.typecode is SignatureDec:
                    ps = self.signature.processRequest(ps, any, pyobj, **kw)
                    continue
                
                if any.typecode is UsernameTokenDec:
                    raise RuntimeError, \
                        'Unauthorized, Not supporting UsernameToken Profile'
                
                #raise RuntimeError, 'WS-Security, Unsupported token %s' %str(any)
            
        return ps

    def processResponse(self, sw, **kw):
        """Set Id attribute for SOAP-ENV:Body child element,
        and SOAP-ENV:Header child elements.
        
        Paramters:
            sw -- SoapWriter
        """
        sw.dom.setContext(processorNss={'SOAP-ENV':SOAP.ENV})
        for node in sw.dom.evaluate(expression='//SOAP-ENV:Header/*') + \
                sw.dom.evaluate(expression='//SOAP-ENV:Body/*'):
            self.tokenRef +=1
            node.setAttributeNS(OASIS.UTILITY, 'Id', str(self.tokenRef))

        security = Security()
        sw = self.signature.processResponse(sw, security=security, **kw)
        
        return sw


class GSISignatureHandler:
    """Web Services Security UsernameToken Profile 1.0
    
    Class Attributes:
        digestMethods --
        signingMethods --
        
    """
    #TODO: Remove An extra # 
    digestMethods = {
        DSIG.BASE+"sha1":sha.sha,
        DSIG.BASE+"#sha1":sha.sha,
        }
        
    @staticmethod
    def getContextFromDirectReference(signature, security):
        """The following list provides a list of the specific reference 
        mechanisms defined in WSS: SOAP 676 
        Message Security in preferred order (i.e., most specific to least specific): 677 
         Direct References  This allows references to included tokens using 
            URI fragments and 678 external tokens using full URIs. 679 
         Key Identifiers  This allows tokens to be referenced using an opaque 
            value that 680 represents the token (defined by token type/profile). 681 
         Key Names  This allows tokens to be referenced using a string that 
            matches an identity 682 assertion within the security token.  This is 
            a subset match and may result in multiple 683 security tokens that 
            match the specified name. 684 
         Embedded References -   This allows tokens to be embedded (as opposed 
            to a pointer 685 to a token that resides elsewhere). 686 
            
        Paramters:
            signature
            security
        """
        sct = None
        si = signature.SignedInfo
        ki = signature.KeyInfo
        ref = ki.Any
        if not _is_security_token_ref(ref):
            return None

        sct = _find_security_token_from_ref(ref, security)
        if sct is None:
            raise RuntimeError, 'Bad SecurityTokenReference'

        algo = si.SignatureMethod.get_attribute_Algorithm()
        if algo == GLOBUS.SIG:
            from pyGridWare.security.m2.M2SecurityContext import\
                M2SecurityContext
            contextID = sct.Identifier
            home = M2SecurityContext.getResourceHome()
            ctx = home.get(contextID)
            ctx.identifier = contextID
            return ctx
        else:
            raise RuntimeError, 'Unknown Signature Algorithm: %s' %algo
        
        return None
    
    @staticmethod
    def getContextFromKeyIdentifier(signature, security):
        si = signature.SignedInfo
        ki = signature.KeyInfo
        ki.KeyName
        ki.KeyValue
        ki.RetrievalMethod
        ki.X509Data
        ki.PGPData
        ki.SPKIData
        ki.MgmtData
        return None
    
    @staticmethod
    def getContextFromKeyName(signature, security):
        return None
    
    @staticmethod
    def getContextFromEmbeddedReferences(signature, security):
        return None

#    signingMethods = {
#        DSIG.BASE+"#hmac-sha1":hmac_sha1,
#        GLOBUS.SIG:getGSSSignatureMethod,
#        }


    @staticmethod
    def ApplyReferenceTransforms(node, ref):
        """Expecting the node to be canoncialized (wsu:Id), and
        the relevant reference.
        
        Parameters:
            node -- DOM node
            ref -- SignedInfo.Reference pyobj
        """
        if ref.Transforms is None:
            return Canonicalize(node)
            
        if len(ref.Transforms.Transform) != 1:
            raise RuntimeError, 'Not supporting multiple Transforms'
        
        tfm = ref.Transforms.Transform
        algo = tfm.get_attribute_Algorithm()
        if algo == DSIG.C14N:
            return Canonicalize(node, unsuppressedPrefixes=None)
            
        if algo != DSIG.C14N_EXCL:
            raise RuntimeError,\
                'Unsupported CanonicalizationMethod Algorithm: %s' %algo
        
        prefixes = []
        for pyobj in tfm.Any:
            if getattr(any, 'typecode', None) is not InclusiveNamespacesDec:
                raise RuntimeError, 'expecting GED (%s,%s)' %(
                    InclusiveNamespacesDec.nspname, 
                    InclusiveNamespacesDec.pname)
                
            prefixes += any.get_attribute_PrefixList().split()
            
        return Canonicalize(node, unsuppressedPrefixes=prefixes)
        
        
    @staticmethod
    def ApplyCanonicalizationMethod(node, can):
        """Expecting SignedInfo node, return canonical XML.
        
        Parameters:
            node -- SignedInfo DOM node
            can -- CanonicalizationMethod pyobj
        """
        algo = can.get_attribute_Algorithm()
        if algo == DSIG.C14N:
            return Canonicalize(node)
        
        if algo != DSIG.C14N_EXCL:
            raise RuntimeError,\
                'Unsupported CanonicalizationMethod Algorithm: %s' %algo
        
        prefixes=[]
        for any in can.Any or []:
            if getattr(any, 'typecode', None) is not InclusiveNamespacesDec:
                raise RuntimeError, 'expecting GED (%s,%s)' %(
                    InclusiveNamespacesDec.nspname, 
                    InclusiveNamespacesDec.pname)
                    
            prefixes += any.get_attribute_PrefixList().split()
                    
        return Canonicalize(node, unsuppressedPrefixes=prefixes)
        
    
    @classmethod
    def processRequest(cls, ps, signature, security, **kw):
        """
        Parameters:
            ps -- ParsedSoap instance
            signature -- Signature pyclass instance
            security -- WS-Security header
        """
        # Grab SignedInfo, and Algorithms
        #signature.Object
        si = signature.SignedInfo
        calgo = si.CanonicalizationMethod.get_attribute_Algorithm()

        # Check All Digests
        sir = {}
        for pyobj in si.Reference:
            sir[pyobj.get_attribute_URI()] = pyobj
            
        #TODO: Cache expressions
        # ElementProxy/DOM Layer, Grab all marked nodes
        processorNss = {\
            'ds':DSIG.BASE, 'wsu':OASIS.UTILITY, 'wsse':OASIS.WSSE,
            'wsc':BEA.SECCONV, 'SOAP-ENV':SOAP.ENV, 
        }
        context = XPath.Context.Context(ps.dom, processorNss=processorNss)

        exp = XPath.Compile('//*[@wsu:Id]')
        for node in exp.evaluate(context):
            wid = '#%s' %node.getAttributeNS(OASIS.UTILITY, 'Id')
            ref = sir.get(wid)
            if ref is None: continue
            try:
                xml = cls.ApplyReferenceTransforms(node, ref)
            except KeyError:
                raise RuntimeError,\
                    'unknown CanonicalizationMethod Algorithm' %calgo
            
            dm = ref.DigestMethod
            try:
                #TODO: Expecting DSIG.SIG##sha1??
                digest = cls.digestMethods[dm.get_attribute_Algorithm()]
            except KeyError:
                raise RuntimeError, 'unknown digestMethods Algorithm(%s) supporting %s' %(
                    dm.get_attribute_Algorithm(), str(cls.digestMethods.keys()))
                    
            dv = digest(xml).digest()
            if ref.DigestValue != dv:
                raise RuntimeError,\
                    'Reference(URI="%s").DigestValue(%s) != %s' %(
                    ref.get_attribute_URI(), base64.encodestring(ref.DigestValue), 
                    base64.encodestring(dv))
            
            del sir[wid]
        
        # All References Processed?
        if len(sir) != 0:
            raise RuntimeError,\
                'Missing Security SignedInfo Reference(s): %s' %(
                    str(sir.keys()))
        
        # ElementProxy/DOM Layer
        # Canonicalize Security.SignedInfo node
        #TODO: not sure whats wrong with expression?
        #exp = XPath.Compile('//SOAP-ENV:Header/wsse:Security/ds:SignedInfo')
        exp = XPath.Compile('//SOAP-ENV:Header//ds:SignedInfo')
        for node in exp.evaluate(context):
            try:
                xml = cls.ApplyCanonicalizationMethod(node, si.CanonicalizationMethod)
            except KeyError:
                raise RuntimeError, 'Unsupported Canonicalization Algorithm'
        
            # Get digest for SignatureMethod
            ctx = cls.getContextFromDirectReference(signature, security) #\or
            #    cls.getContextFromKeyIdentifier(signature, security) or\
            #    cls.getContextFromKeyName(signature, security) or\
            #    cls.getContextFromEmbeddedReference(signature, security)
            
            if ctx is None:
                raise RuntimeError, 'Failed to locate SecurityToken'
       
            # Security Context Info for Reply 
            state = kw['state']
            state.context = ctx
            state.identifier = ctx.identifier
            state.signatureAlgorithm = si.SignatureMethod.get_attribute_Algorithm()
        
            ctx.verifyMIC(xml.strip(), signature.SignatureValue)
            break
        else:
            raise RuntimeError, 'Missing SignedInfo'

        return ps

    signingID = 3000
    @classmethod
    def processResponse(cls, output, **kw):
        sw = output
        calgo = DSIG.C14N
        dalgo = DSIG.BASE+"#sha1"
        state = kw['state']
        if state.context is None:
            raise RuntimeError, 'No Security Context'

        #
        # WS-Security Header
        #
        pyobj = Security()
        pyobj.Any = sig = Signature()
        sct = SecurityContextToken()
        pyobj.Any.append(sct)
        cls.signingID += 1
        wsuid = "#SCT-%s" %str(cls.signingID)
        
        #
        # SecurityContextToken
        sct.Identifier = state.identifier
        sct.set_attribute_Id(wsuid)
        
        #
        # Signature
        sig.SignedInfo = si = sig.new_SignedInfo()
        sig.SignatureValue = ''
        sig.KeyInfo = ki = sig.new_KeyInfo()
        
        # 
        # Create Direct Reference
        ki.Any = sctref = SecurityTokenReference()
        sctref.Any = ref = Reference()
        ref.set_attribute_URI(wsuid)
        #TODO: MOVE THIS OUT
        ref.set_attribute_ValueType("http://www.globus.org/ws/2004/09/security/sc#GSSAPI_CONTEXT_TOKEN")
        
        # 
        # Specify Algorithms
        salgo = state.signatureAlgorithm
        si.CanonicalizationMethod = cm = si.new_CanonicalizationMethod()
        cm.set_attribute_Algorithm(calgo)
        si.SignatureMethod = sm = si.new_SignatureMethod()
        sm.set_attribute_Algorithm(salgo)
        
        # create digests for all marked nodes
        # ElementProxy/DOM Layer, Grab all marked nodes
        processorNss = { 'ds':DSIG.BASE, 'wsu':OASIS.UTILITY, 
            'wsse':OASIS.WSSE, 'wsc':BEA.SECCONV, 'SOAP-ENV':SOAP.ENV, 
         }
        context = XPath.Context.Context(sw.dom._getNode(), processorNss=processorNss)
            
        exp = XPath.Compile('//*[@wsu:Id]')
        for node in exp.evaluate(context):
            wid = node.getAttributeNS(OASIS.UTILITY, 'Id')

            ref = si.new_Reference()
            if si.Reference is not None: si.Reference.append(ref)
            else: si.Reference = ref

            ref.set_attribute_URI(wid)
            ref.DigestMethod = dm = ref.new_DigestMethod()
            dm.set_attribute_Algorithm(dalgo)
            
            digest = cls.digestMethods[dalgo]
            
            xml = cls.ApplyReferenceTransforms(node, ref)    
            ref.DigestValue = base64.encodestring(digest(xml).digest()).strip()
        
        #
        # Serialize Security Header
        sw.serialize_header(pyobj)
        
        # Post-Processing Replace SignatureValue 
        # ElementProxy/DOM Layer
        # Canonicalize Security.SignedInfo node
        element = sw.dom
        exp1 = XPath.Compile('//SOAP-ENV:Header//ds:SignedInfo')
        exp2 = XPath.Compile('//SOAP-ENV:Header//ds:SignatureValue')
        for node in exp1.evaluate(context):
            xml = cls.ApplyCanonicalizationMethod(node, cm)
            mic = state.context.getMIC(xml)

            for svnode in exp2.evaluate(context):
                doc = svnode.ownerDocument
                tnode = doc.createTextNode(base64.encodestring(mic))
                svnode.appendChild(tnode)
                break

            return output

        raise RuntimeError, 'Failed to set SignatureValue'
    

class X509TokenProfileHandler:
    """Web Services Security UsernameToken Profile 1.0
    """
    targetNamespace = DSIG.BASE
    
    # Token Types
    singleCertificate = targetNamespace + "#X509v3"
    certificatePath = targetNamespace + "#X509PKIPathv1"
    setCerticatesCRLs = targetNamespace + "#PKCS7"
    
    @classmethod
    def processRequest(cls, ps, signature, **kw):
        return ps



class SharedStateHandlerChain:
    """dump'state' into each handler processRequest/processResponse
    """
    class State(object): 
        __slots__ = ['context', 'identifier', 'signatureAlgorithm']
        def __init__(self):
            self.context = self.identifier = self.signatureAlgorithm = None
        
        
    @CheckInputArgs(CallbackChainInterface, HandlerChainInterface)
    def __init__(self, cb, *handlers):
        """
        Parameters:
            cb -- callback handler
            *handlers -- list of handlers, may end with handlers
                of PostCallBackRequestMarker which will be processed
                after the cb and provided with state information.
        """
        self.handlercb = cb
        self.handlers = handlers
        self.debug = len(log.theLogPublisher.observers) > 0
        self.state = SharedStateHandlerChain.State()
        
    def processRequest(self, arg, **kw):
        if self.debug:
            log.msg('--->PROCESS REQUEST\n%s' %arg, debug=1)

        for h in self.handlers:
            arg = h.processRequest(arg, state=self.state, **kw)
            
        return self.handlercb.processRequest(arg, **kw)

    def processResponse(self, arg, **kw):
        if self.debug:
            log.msg('===>PROCESS RESPONSE: %s' %str(arg), debug=1)

        if arg is None: 
            return

        for h in self.handlers:
            arg = h.processResponse(arg, state=self.state, **kw)
            
        s = str(arg)
        if self.debug:
            log.msg(s, debug=1)

        return s


class WSSecurityHandlerChainFactory:
    protocol = SharedStateHandlerChain
    
    @classmethod
    def newInstance(cls):
        return cls.protocol(CallbackHandler, DataHandler, WSAddressHandler(), 
            WSSecurityHandler(signature=GSISignatureHandler))

