############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sha, base64, time

import ZSI
from ZSI.wstools.Namespaces import DSIG, OASIS, GLOBUS, WSA200403, BEA, SOAP
from ZSI.TC import _get_global_element_declaration as GED

from twisted.python import log 

from pyGridWare.security.m2.M2SecurityContext import M2SecurityContext
from pyGridWare.security.m2.M2SecureMessage import M2Authentication

from pyGridWare.security.base.BaseSignatureHandler import BaseSignatureHandler
from pyGridWare.security.base.BaseSignatureHandler import SignatureHandlerError

from pyGridWare.utility.dom import  DomletteElementProxy

import pyGridWare.generated.types.secureconversation

# GED Typecodes
BST = GED(OASIS.WSSE, "BinarySecurityToken").pyclass
KeyInfo = GED(DSIG.BASE, "KeyInfo").pyclass
Reference = GED(DSIG.BASE, "Reference").pyclass
SecurityReference = GED(OASIS.WSSE, "Reference").pyclass
SecurityTokenReference = GED(OASIS.WSSE, "SecurityTokenReference").pyclass
Signature = GED(DSIG.BASE, "Signature").pyclass
Security = GED(OASIS.WSSE, "Security").pyclass
SecurityContextToken = GED(BEA.SECCONV,"SecurityContextToken").pyclass
SignedInfo = GED(DSIG.BASE, "SignedInfo").pyclass
C14NMethod = GED(DSIG.BASE, "CanonicalizationMethod").pyclass
SignatureMethod = GED(DSIG.BASE, "SignatureMethod").pyclass
DigestValue = GED(DSIG.BASE, "DigestValue").pyclass  
DigestMethod = GED(DSIG.BASE, "DigestMethod").pyclass 
TimeStamp = GED(OASIS.UTILITY, "Timestamp").pyclass


# Functions
def _make_timestamp(gmtime_tuple):
    """Returns a string according the Gregorian format from a gmtime tuple
    """ 
    d = {'Y': gmtime_tuple[0], 'M':gmtime_tuple[1], 
         'D':gmtime_tuple[2], 'h':gmtime_tuple[3], 
         'm':gmtime_tuple[4],'s':gmtime_tuple[5],}
    format = '%(Y)04d-%(M)02d-%(D)02dT%(h)02d:%(m)02d:%(s)02dZ'
    return format %d


class VerifyError(SignatureHandlerError):
    """Errors encountered during Verification of XML Signature
    """


class M2SignatureHandler(BaseSignatureHandler):
    """Handler for signing and verifying messages, using
    GssSecureConversation or GssSecureMessage protocols.
    
    Class Attributes:
        canonicalizationAlgo -- canonicalization URI
        signatureAlgo -- XML Signature URI
        digestAlgo    -- Digest URI
        validDuration -- seconds a SecureMessage signature is valid
    """
    canonicalizationAlgo = DSIG.C14N_EXCL
    signatureAlgo = GLOBUS.SIG
    digestAlgo = DSIG.DIGEST_SHA1
    validDuration = 300
    
    def __init__(self, sign_body=True, sign_headers=True):
        """
        KeyWord Paramters:
            sign_body    -- boolean, sign root body element
            sign_headers -- boolean, sign header elements
        """
        self._contextID = None
        self._isProtocolSet = False
        self._sign_headers = sign_headers
        self._sign_body = sign_body
        
#    def setContextID(self, KeyName):
#        self._contextID =  KeyName
#
#    def getContextID(self):
#        return self._contextID
#
#    def protocolSet(self):
#        return self._setProtocol
    
#    def getSecurityToken(self, element, uri):
#        uriExpString = '//*[@wsu:Id="' + uri[1:] +'"]'
#        if not element.isContextInitialized():
#            raise SignatureHandlerError('ElementProxy has no context')
#        uriNodes = element.evaluate(expression=uriExpString)
#        uriNode = uriNodes[0]
#        bst = uriNode.getValue()
#        return bst
    
    def setSecureMessage(self):
        """use GssSecureMessage protocol
        """
        self.setProtocol(DSIG.SIG_RSA_SHA1)

    def setSecureConversation(self, contextID):
        """use GssSecureConversation protocol
        """
        self.setProtocol(GLOBUS.SIG, contextID)
    
    def setProtocol(self, uri, contextID=None):
        """
        sets the protocol uri, and the contextID

        **Raises**
        A SignatureHandlerError is raised if the signature algorithm is not supported
        """
        if uri not in (DSIG.SIG_RSA_SHA1, GLOBUS.SIG):
            raise SignatureHandlerError("Unsupported signature algorithm: " + uri)
        
        if contextID is not None and uri == DSIG.SIG_RSA_SHA1:
            raise SignatureHandlerError("contextID must be None if you are using GssSecureMessage")
        
        if contextID is None and uri == GLOBUS.SIG:
            raise SignatureHandlerError("contextID must not be None if you are using GssSecureConveration") 

        self.signatureAlgo = uri
        self._contextID = contextID
        self._isProtocolSet = True

    def useSecureMessage(self):
        """
        returns True if we are using GssSecureMessage and otherwise it returns False
        """
        #return self.signatureAlgo == DSIG.SIG_RSA_SHA1
        return False

    def verify(self, ps):
        """first check the digests, second use the canonicalized form
        of the signedInfo node as input to calculate the MIC, which is compared to
        the provided SignatureValue.
        
        Parameters:
            ps -- ParsedSoap instance.
        """
        contextID = self._contextID
        element = DomletteElementProxy(None, message=ps.header)
        element.setContext(processorNss={
                'ds':DSIG.BASE, 'wsu':OASIS.UTILITY, 'wsse':OASIS.WSSE,
                'soapenv':"http://schemas.xmlsoap.org/soap/envelope/", 
                'wsc':BEA.SECCONV, })

        #This is how the a Server side service figures out which protocol to use
        if not self._isProtocolSet:
            contextID = None
            signatureMethodNodes = element.evaluate(expression='//ds:SignatureMethod')
            if len(signatureMethodNodes) != 1:
                if len(signatureMethodNodes) == 0:
                    raise VerifyError('No ds:SignatureMethod specified')
                raise VerifyError('Multiple ds:SignatureMethod elements specified')

            signatureMethodNode = signatureMethodNodes[0]
            signatureAlgorithm = signatureMethodNode.getAttributeValue(None, "Algorithm")
            if signatureAlgorithm == GLOBUS.SIG:
                ref = element.evaluate(expression='//soapenv:Header/wsse:Security/ds:Signature/ds:KeyInfo/wsse:SecurityTokenReference/wsse:Reference')
                if len(ref) == 0:
                    raise VerifyError('Failed to locate a KeyInfo/SecurityTokenReference/Reference')
                if len(ref) > 1:
                    raise VerifyError('Multiple SecurityTokenReferences/Reference')

                ref = ref[0]
                vt = ref.getAttributeValue(None, "ValueType")
                if vt != "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_CONTEXT_TOKEN":
                    raise VerifyError("globus gssapi context token only")

                uri = ref.getAttributeValue(None, "URI")
                if not uri.startswith('#'):
                    raise VerifyError('bad URI in <wsse:Reference URI="%s">' %uri)

                # TODO: Check wsu:Id
                contextID = element.evaluate(expression='//soapenv:Header/wsse:Security/wsc:SecurityContextToken/wsc:Identifier/child::text()')
                if len(contextID) != 1:
                    raise VerifyError('Problem with SecurityContextToken Identifier')
                # 
                # 4Suite TEXt_NODE API
                # http://infinitesque.net/projects/4Suite/docs/modules/Ft.Xml.cDomlette.html#Text
                contextID = contextID[0].nodeValue
            elif signatureAlgorithm == DSIG.SIG_RSA_SHA1:
                #TODO Check valuetypes specific  
                pass
            else:
                # TODO: 
                raise VerifyError('Unknown SignatureMethod Algorithm=%s' %signatureAlgorithm)

            self.setProtocol(signatureAlgorithm, contextID)

        #
        #TODO: Hack the server NEEDS to have context_id set on each "verify"
        #      so that "sign" can proceed.
        #
        if self._contextID is None:
            print "AAACONTEXT "*10
            self._contextID = element.evaluate(expression='//soapenv:Header/wsse:Security/wsc:SecurityContextToken/wsc:Identifier/child::text()')[0].nodeValue
        else:
            print "BBBCONTEXT: ", self._contextID


        useSecureMessage = self.useSecureMessage()
        referenceNodes = element.evaluate(expression='//soapenv:Header/wsse:Security/ds:Signature/ds:Reference')
        log.msg("REFERENCE NODES:  %s" %str(referenceNodes))
        for refNode in referenceNodes:
            log.msg("REFERENCE NODE:  %s" %str(refNode))
            kw = {}
            uri = refNode.getAttributeValue(None, "URI")
            #TODO: This is c14n_excl only?
            transformsNode = refNode.getElement(DSIG.BASE, "Transforms")

            #TODO:  HACK algo is specified by sender not hardcoded.
            #  Think we need to be grabbing the CanonicalizationMethod element.
            #  
            algorithm = self.canonicalizationAlgo

            l = 0
            if transformsNode is not None:
                transforms = transformsNode.getElements(DSIG.BASE, "Transform")
                l = len(transforms)

            if l == 1:
                algorithm = transforms[0].getAttributeValue(None,"Algorithm") or self.canonicalizationAlgo
                if algorithm == DSIG.C14N_EXCL:
                    try:
                        inclusiveNS = transforms[0].getElement(DSIG.C14N_EXCL, "InclusiveNamespaces")
                        if inclusiveNS is None:
                            kw['unsuppressedPrefixes'] = []
                        else:
                            kw['unsuppressedPrefixes'] =\
                            inclusiveNS.getAttributeValue(None, "PrefixList").split()
                    except:
                        raise VerifyError,\
                            'failed to handle transform(%s) in <ds:Reference URI="%s">)' %(
                            transforms[0], uri)
            elif l > 1:
                raise VerifyError,\
                    'Not supporting multiple transforms:  <ds:Reference URI="%s">' %(
                    uri)

            ##TODO: Canonicalize
            digestValue = base64.encodestring(self.getDigestValueFromDom(element, uri, 
                              algorithm=algorithm, **kw))

            digestNode = refNode.getElement(DSIG.BASE,"DigestValue")
            if digestValue.strip() != digestNode.getValue().strip():
                log.msg("RCV DIGEST:  %s" %digestValue.strip())
                log.msg("CAL DIGEST:  %s" %digestNode.getValue().strip())


                raise VerifyError("DigestValues do not match: URI=%s" %uri)
            log.msg("DIGEST MATCH:  URI=%s" %uri)
        
        log.msg("CHECK SIGNATURE")
        signedInfoNodes = element.evaluate(expression='//ds:SignedInfo')
        if len(signedInfoNodes) == 0:
            raise VerifyError("Message Security Required, no Security SignedInfo header")
        if len(signedInfoNodes) > 1:
            raise VerifyError("more than one uri found with uri: " + str(len(signedInfoNodes)))

        signedInfoNode = signedInfoNodes[0]
        canNode = signedInfoNode.getElement(DSIG.BASE, "CanonicalizationMethod")
        c14nAlgo =  canNode.getAttributeValue(None, "Algorithm")

        ##TODO: Canonicalize
        signedInfoStr = signedInfoNode.canonicalize(c14nAlgo)
        signedInfoHash = sha.sha(signedInfoStr).digest()
        
        signatureValueNodes = element.evaluate(expression='//ds:SignatureValue')
        if len(signatureValueNodes) != 1:
            raise VerifyError("more than one uri found with uri: " + str(len(signatureValueNodes)))

        signatureValueNode = signatureValueNodes[0]
        signatureStr = signatureValueNode.getValue()
             
#        keyInfoNodes = element.evaluate(expression='//ds:KeyInfo')
#        keyInfoNode = keyInfoNodes[0]
        keyInfoNode = element.evaluate(expression='//ds:KeyInfo')[0]
        
        if useSecureMessage:
            secReferenceNode = keyInfoNode.getElement(OASIS.WSSE, 
                                                      "SecurityTokenReference")
            referenceNode = secReferenceNode.getElement(OASIS.WSSE, "Reference")
            keyUri = referenceNode.getAttributeValue(None, "URI")
            
#            bst = self.getSecurityToken(element, keyUri)
            uriExpString = '//*[@wsu:Id="' + keyUri[1:] +'"]'
            if not element.isContextInitialized():
                raise SignatureHandlerError('ElementProxy has no context')
            
            uriNode = element.evaluate(expression=uriExpString)[0]
            bst = uriNode.getValue()
    
            # Delegate to a different handler..
            if M2Authentication().verify(base64.decodestring(bst), signedInfoHash, 
                    base64.decodestring(signatureStr)):
                return
        
            raise VerifyError("SecureMessage signatures do not match")
        
        home = M2SecurityContext.getResourceHome()
        secContext = home.get(contextID)
        if secContext.verifyMIC(signedInfoStr,base64.decodestring(signatureStr)):
            return
        
        raise VerifyError("SecureConversation signatures do not match")
       
    def sign(self, sw):
        """
        Sign the message whether we are using
        GssSecureConversation or GssSecureMesage

        General algorithm
        - Create the authentication information
        elements (ie BinarySecurityToken)
        and security elements(ie timestamp)
        
        - Create digests for anything in the
        body of the Message

        - Fill in the SignedInfoNode, create elements
        for digests of elements in the header
        leave the content of the element empty. Create
        a SignatureValue element also leavethe content
        of that element empty.

        - serialize the header

        - Calculate the digests for elements in the
        header(we had to wait until the header was
        serialized to do this) Put the digests as
        the content for those empty digestValue elements
        that we had created

        - Calculate the SignatureValue and place it in
        the content of the SignatureValue element
        
        """
        useSecureMessage = self.useSecureMessage()
        if self._sign_headers: self.signHeaderElements(sw)
        if self._sign_body: self.signBodyElement(sw)
        
        idUriList = []
        if useSecureMessage:
            auth = M2Authentication()
            pkiPath = auth.getPkiPath()
            pkiPath64 = base64.encodestring(pkiPath)
            binarySecurityToken = BST(pkiPath64)
            #move this GssSecureMessage
            tokenRefString = self.getTokenRefString()
            binarySecurityToken.set_attribute_Id("CertId-" + tokenRefString)
            binarySecurityToken.set_attribute_ValueType(OASIS.X509TOKEN.X509PKIPathv1) 
            binarySecurityToken.set_attribute_EncodingType(OASIS.X509TOKEN.Base64Binary)
            timeStamp = TimeStamp()
            t = time.gmtime() 
            timeStamp.Created = _make_timestamp(time.gmtime()) 
            expires = time.time() + self.validDuration
            timeStamp.Expires = _make_timestamp(time.gmtime(expires))
            timeStamp.set_attribute_Id("timestamp")
            idUriList.append("#timestamp")

         
        pyobj = Security()
        sig = Signature()
        pyobj.Any = sig
        if useSecureMessage:
            pyobj.Any = binarySecurityToken   
            pyobj.Any.append(sig)
            pyobj.Any.append(timeStamp)
        else:
            pyobj.Any = sig

        sig.SignedInfo = SignedInfo()
        sig.SignedInfo.CanonicalizationMethod = cMethod = C14NMethod()
        cMethod.set_attribute_Algorithm(self.canonicalizationAlgo)
        sig.SignedInfo.SignatureMethod = SignatureMethod()
        if useSecureMessage:
            sig.SignedInfo.SignatureMethod.set_attribute_Algorithm(DSIG.SIG_RSA_SHA1)
        else:
            sig.SignedInfo.SignatureMethod.set_attribute_Algorithm(GLOBUS.SIG)
        sig.SignatureValue = ''
        
        if useSecureMessage:
            sig.KeyInfo = KeyInfo()
            sig.KeyInfo.Any = secTokRef = SecurityTokenReference()
            secTokRef.set_attribute_Id("STRId-11982507")
            secRef = SecurityReference()

            secRef.set_attribute_URI("#CertId-" + tokenRefString)
            secRef.set_attribute_ValueType("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1")
            secTokRef.Any = secRef
        else:
            sig.KeyInfo = KeyInfo()
            sig.KeyInfo.Any = secTokRef = SecurityTokenReference()
            secRef = SecurityReference()
            secRef.set_attribute_URI("#CertId-" + self.getTokenRefString())
            secRef.set_attribute_ValueType("http://www.globus.org/ws/2004/09/security/sc#GSSAPI_CONTEXT_TOKEN")
            secTokRef.Any = secRef

            #
            #TODO: make this configurable
            home = M2SecurityContext.getResourceHome()

            #
            #TODO: How to determine correct contextID?
            #  if this is server, recall from sign.
            #  if this is client, we must know it from the start.
            context_id = self._contextID
            try:
                secContext = home.get(context_id)
            except Exception, ex:
                raise

            secContextToken = SecurityContextToken()
            secContextToken.set_attribute_Id(secRef.get_attribute_URI()[1:])
            secContextToken.Identifier = context_id
            pyobj.Any.append(secContextToken) 

        
        processorNss={'wsu':OASIS.UTILITY, 'ds':DSIG.BASE}
        sw.dom.setContext(processorNss=processorNss)
        nodes = sw.dom.evaluate(expression='//*[@wsu:Id]')

        #Calculate the digests in the body
        for node in nodes: 
            URI = '#%s' %node.getAttributeValue(OASIS.UTILITY, 'Id')
            ref = Reference()
            ref.set_attribute_URI(URI)
            ref.DigestMethod =  DigestMethod()
            ref.DigestMethod.set_attribute_Algorithm(self.digestAlgo)

            #No Transforms use c14n by default
            ref.DigestValue = self.getDigestValueFromSoapWriter(sw, URI, 
                                  algorithm=None)

            if sig.SignedInfo.Reference is None: 
                sig.SignedInfo.Reference = ref
            else: 
                sig.SignedInfo.Reference.append(ref)
            

        #These haven't been serialized yet, the id's to take
        #digests of in the security header
        
        for idUri in idUriList:
            ref = Reference()
            ref.set_attribute_URI(idUri)
            ref.DigestMethod = DigestMethod()
            ref.DigestMethod.set_attribute_Algorithm(self.digestAlgo)
            ref.DigestValue = ""
            sig.SignedInfo.Reference.append(ref)
        
        #pyobj._any.append(sig)
        sw.serialize_header(pyobj)

        # Reset Context after append on DOM
        sw.dom.setContext(processorNss=processorNss)
        for URI in idUriList:
            #TODO: SPECIFY Canonicalization
            digestValue = self.getDigestValueFromSoapWriter(sw, URI, 
                              algorithm=cMethod.get_attribute_Algorithm())

            referenceNodes = sw.dom.evaluate(expression='//ds:Reference')
            
            for referenceNode in referenceNodes:
                refUri = referenceNode.getAttributeValue(None, "URI")
                if refUri == URI:
                    digestValueNode = referenceNode.getElement(DSIG.BASE, "DigestValue")
                    digestValueNode.createAppendTextNode(base64.encodestring(digestValue))
            
        # Reset Context after append on DOM
        sw.dom.setContext(processorNss=processorNss)
        nodes = sw.dom.evaluate(expression='//ds:SignedInfo')
        signedInfo = nodes[0]

        # XXX
        signedInfoStr = signedInfo.canonicalize(algorithm=cMethod.get_attribute_Algorithm())

        nodes = sw.dom.evaluate(expression='//ds:SignatureValue')
        signatureValueNode = nodes[0]
        if useSecureMessage:
            hashedSignedInfoStr = sha.sha(signedInfoStr).digest()
            signatureValue = base64.encodestring(auth.sign(hashedSignedInfoStr))
        else:
            signatureValue = base64.encodestring(secContext.getMIC(signedInfoStr))
        signatureValueNode.createAppendTextNode(signatureValue)

        
