##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, warnings
from os.path import isfile

from ZSI.wstools import WSDLTools
from ZSI.wstools.Namespaces import WSRF_V1_2
from ZSI.generate.utility import TextProtect, NCName_to_ClassName as NC_to_CN

"""
Functions:
    GetResourcePropertiesFromPort(port):
    GetContextClassBasesFromPort(port):
    GetPropertiesBasesFromRP(ged):

Classes:
    PortTypeMap:
    NameResolver:
"""

HEADER = """############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
# Generated by setup.py
###########################################################################
"""

CLIENT = 'client'
SERVICE = 'service'
SERVICE_STUBS = 'service_stubs'
TYPES = 'types'
PROPERTIES = 'properties'
RESOURCE = 'resource'
GEN_PACKAGES = None

def GetResourcePropertiesFromPort(port):
    '''Resolves ResourceProperty declaration's element references
    and returns all element declarations.
    '''
    els = []
    wsdl = port.getWSDL()
    pt = port.getPortType()
    rp = pt.getResourceProperties()
    if rp is None:
        warnings.warn(\
            'portType(%s) does not have a ResourceProperties declaration' \
             %(pt.name),
        )
        return None

    nsuri,name = rp
    schema = wsdl.types.get(nsuri)
    assert schema is not None and schema.elements.has_key(name) is True, \
        'portType(%s) ResourcePropties declares undefined element(%s,%s)'\
        %(pt.name, nsuri, name)

    ged = schema.elements[name]
    typeDef = ged.getTypeDefinition()
    assert typeDef.isComplex(), \
        'resource properties type definition must define a sequence or all model group: %s'\
        %(typeDef.getItemTrace())

    assert typeDef.content is not None \
        or not typeDef.content.isSequence() or not typeDef.content.isAll(),\
        'resource properties type definition must define a sequence or all model group: %s'\
        %(typeDef.getItemTrace())

    for er in typeDef.content.content:
        assert er.isReference() and er.isElement(), \
            'ReferenceProperties(%s) must specify element references..'\
            %ged.getItemTrace()

        e = er.getElementDeclaration()
        if e is None:
            raise RuntimeError, 'Element Reference refers to nothing: %s' %er.getItemTrace()
        assert e.isDeclaration() and e.isElement(), \
            'WSDL Definition(%s) no element(%s,%s) for element reference: %s'\
            %(wsdl.name, er.getTargetNamespace(), er.getAttribute('ref'), ged.getItemTrace())

        els.append(e)

    assert ged is not None, \
        'Must declare ResourceProperties in PortType, from WSDL Definition(%s)'\
        %(wsdl.name)

    return els 


def GetContextClassBasesFromPort(port):
    '''Examines port's portType operation's input messages 
    for WSRF specification messages.  Assume output messages
    are corrrect..
    '''
    keys = []
    bases = []
    properties = lifetime = base_notification = None

    keys = map(lambda e: (e.getTargetNamespace(),e.getAttribute('name')),
               GetResourcePropertiesFromPort(port))
    for prop in keys:
        nsuri,name = prop
        if nsuri in WSRF_V1_2.LIFETIME.XSD_LIST:
            lifetime = nsuri
        elif nsuri in WSRF_V1_2.PROPERTIES.XSD_LIST:
            properties = nsuri
        elif nsuri in WSRF_V1_2.BASENOTIFICATION.XSD_LIST:
            base_notification = nsuri
   
    if lifetime is not None and base_notification is not None: 
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
            (base_notification,'ConsumerReference'),
            (base_notification,'TopicExpression'),
            (base_notification,'UseNotify'),
            (base_notification,'Precondition'),
            (base_notification,'Selector'),
            (base_notification,'SubscriptionPolicy'),
            (base_notification,'CreationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.context(PortTypeMap.SubscriptionManager))

    if lifetime is not None and PortTypeMap.context(PortTypeMap.SubscriptionManager) not in bases:
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.context(PortTypeMap.ScheduledResourceTermination))

    return bases


def GetPropertiesBasesFromRP(ged):
    '''
    NotificationProducerRP
    ScheduledResourceTerminationRP
    SubscriptionManagerRP
    '''
    assert ged.isElement() is True and ged.isLocal() is False,\
        'Unexpected parameter ged.'

    bases = []
    type_def = ged.getTypeDefinition()
    model_group = type_def.content
    extra_els = map(lambda er: er.getElementDeclaration(), model_group.content)

    namespaces = map(lambda e: e.getTargetNamespace(), extra_els)
    lifetime = None
    for nsuri in filter(lambda nsuri: nsuri in WSRF_V1_2.LIFETIME.XSD_LIST, namespaces):
        if lifetime is not None and lifetime != nsuri:
            warning.warn('Properties with different WS-ResourceLifetime namespaces, skipping properties generation.')
            return
        lifetime = nsuri

    base_notification = None
    for nsuri in filter(lambda nsuri: nsuri in WSRF_V1_2.BASENOTIFICATION.XSD_LIST, namespaces):
        if base_notification is not None and base_notification != nsuri:
            warning.warn('Properties with different WS-BaseNotification namespaces, skipping properties generation.')
            return
        base_notification = nsuri

    keys = map(lambda e: (e.getTargetNamespace(),e.getAttribute('name')), extra_els)
    if lifetime is not None and base_notification is not None: 
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
            (base_notification,'ConsumerReference'),
            (base_notification,'TopicExpression'),
            (base_notification,'UseNotify'),
            (base_notification,'Precondition'),
            (base_notification,'Selector'),
            (base_notification,'SubscriptionPolicy'),
            (base_notification,'CreationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.property(PortTypeMap.SubscriptionManager))

    if lifetime is not None and 'SubscriptionManagerRP' not in bases:
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.property(PortTypeMap.ScheduledResourceTermination))

    if base_notification is not None: 
        p = [\
            (base_notification,'Topic'), 
            (base_notification,'FixedTopicSet'), 
            (base_notification,'TopicExpressionDialects'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.property(PortTypeMap.NotificationProducer))

    return bases


class PortTypeMap:
    SubscriptionManager = 'SubscriptionManager'
    ScheduledResourceTermination = 'ScheduledResourceTermination'
    NotificationProducer = 'NotificationProducer'

    class BaseClassDescription(object):
        def __init__(self, name, module):
            self.name = name
            self.alias = name
            self.module = module

        def getAlias(self):
            return self.__alias
        def setAlias(self, arg):
            self.__alias = '_%s' %arg
        alias = property(getAlias, setAlias, None, 'class alias')

        def getName(self):
            return self.__name
        def setName(self, arg):
            self.__name = arg
        name = property(getName, setName, None, 'class name')

        def getModule(self):
            return self.__module
        def setModule(self, arg):
            self.__module = arg
        module = property(getModule, setModule, None, 'module name')

    CONTEXT = {}
    CONTEXT[SubscriptionManager] =\
        BaseClassDescription('SubscriptionManagerContext','pyGridWare.resource.ResourceContext')
    CONTEXT[ScheduledResourceTermination] =\
        BaseClassDescription('ScheduledResourceTerminationContext','pyGridWare.resource.ResourceContext')

    PROPERTIES = {}
    PROPERTIES[SubscriptionManager] =\
        BaseClassDescription('SubscriptionManagerRP', 'pyGridWare.wsrf.properties.ResourceProperties')
    PROPERTIES[ScheduledResourceTermination] =\
        BaseClassDescription('ScheduledResourceTerminationRP', 'pyGridWare.wsrf.properties.ResourceProperties')
    PROPERTIES[NotificationProducer] =\
        BaseClassDescription('NotificationProducerRP', 'pyGridWare.wsrf.properties.ResourceProperties')

    __init__ = None

    def context(cls, key):
        return cls.CONTEXT[key]
    context = classmethod(context)

    def property(cls, key):
        return cls.PROPERTIES[key]
    property = classmethod(property)


class NameResolver:
    packages = []
    d = {\
        CLIENT:'pyGridWare.generated.stubs',
        SERVICE:'pyGridWare.generated.services',
        SERVICE_STUBS:'pyGridWare.generated.services.stubs',
        TYPES:'pyGridWare.generated.types',
        PROPERTIES:'pyGridWare.generated.properties',
        RESOURCE:'pyGridWare.generated.resource',
    }
    __init__ = None

    def Reset(cls):
        '''Reset packages list by setting it to all the
        packages specified in the package dictionary.  
        '''
        cls.packages = cls.d.values()
        return cls.packages
    Reset = classmethod(Reset)

    def NCName_to_Module(ncname):
        return NC_to_CN(ncname)
    NCName_to_Module = staticmethod(NCName_to_Module)

    def NCName_to_File(ncname):
        return TextProtect(ncname)
    NCName_to_File = staticmethod(NCName_to_File)

    def Module(cls, key, *args):
        s = cls.d[key]
        for a in map(cls.NCName_to_Module, args): s += '.%s' %a
        return s
    Module = classmethod(Module)

    def File(cls, key, *args):
        return '%s.py' %cls.Dir(key, *args)
    File = classmethod(File)

    def Dir(cls, key, *args):
        l = len(args)
        s = cls.d[key].replace('.', '/')
        for dir in map(cls.NCName_to_File, args): s += '/%s' %dir
        return s
    Dir = classmethod(Dir)

    # module names
    GetContextModuleName = NCName_to_Module
    GetPropertiesModuleName = NCName_to_Module
    GetTypesModuleName = NCName_to_Module
    GetClientModuleName = NCName_to_Module
    GetServiceStubModuleName = NCName_to_Module
    GetServiceModuleName = NCName_to_Module

    GetModFromFile = staticmethod(lambda dir: dir.split('.py')[0].strip('/').replace('/','.'))
    GetFileName = staticmethod(lambda dir,name: os.path.join(dir, '%s.py' %TextProtect(name)))

    # file names
    GetPropertiesModuleFileName =\
        lambda cls, section, port:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(PROPERTIES), section, port.getService()),
                cls.GetPropertiesModuleName(port.name)
            )
    GetPropertiesModuleFileName = classmethod(GetPropertiesModuleFileName)

    GetContextModuleFileName =\
        lambda cls, section, service:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(RESOURCE), section),
                cls.GetContextModuleName(service.name)
            )
    GetContextModuleFileName = classmethod(GetContextModuleFileName)

    GetServiceModuleFileName =\
        lambda cls, section, service:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(SERVICE), section),
                cls.GetServiceModuleName(service.name)
            )
    GetServiceModuleFileName = classmethod(GetServiceModuleFileName)

    GetServiceStubModuleFileName =\
        lambda cls, section, service:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(SERVICE_STUBS), section),
                cls.GetServiceStubModuleName(service.name)
            )
    GetServiceStubModuleFileName = classmethod(GetServiceStubModuleFileName)

    GetTypesModuleFileName =\
        lambda cls, section:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(TYPES), None),
                cls.GetTypesModuleName(section)
            )
    GetTypesModuleFileName = classmethod(GetTypesModuleFileName)

    GetClientModuleFileName =\
        lambda cls, section:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(CLIENT), None),
                cls.GetClientModuleName(section)
            )
    GetClientModuleFileName = classmethod(GetClientModuleFileName)

    GetContextClassNameFromPort =\
        staticmethod(lambda port: '%sContext' %TextProtect(port.getPortType().name))

    GetPropertiesClassNameFromPort =\
        staticmethod(lambda port: '%s' %TextProtect(port.getPortType().name))

    def GetBaseDirectory(basedir, section, service=None):
        '''returns directory path, must start with basedir.
        basedir  -- base directory, must exist.
        section  -- section
        service  -- WSDLTools service instance.
        '''
        if service is not None:
            return os.path.join(basedir, section, TextProtect(service.name))
        if section is not None:
            return os.path.join(basedir, section)
        return basedir

    GetBaseDirectory = staticmethod(GetBaseDirectory)

    def MakeDirectories(cls, packages, basedir, section, service=None):
        '''uses path form GetBaseDirectory, creates any subdirs
        of basedir, adds subdirs to packages, and installs 
        an __init__ file in subdirs.
        packages -- list of packages to be added to distutils
        basedir  -- base directory, must exist.
        section  -- section
        service  -- WSDLTools service instance.
        '''
        assert type(packages) is list,\
            'Expecting list for parameter packages '
        assert service is None or isinstance(service, WSDLTools.Service) is True,\
            'Expecting service is WSDLTools.Service instance or None'

        path = cls.GetBaseDirectory(basedir, section, service)
        assert path.startswith(basedir),\
            'directory path(%s) returned by GetBaseDirectory must start with basedir(%s)'\
            %(path, basedir)

        dir = basedir
        dirs = None
        if os.path.isdir(basedir) is True:
            dirs = path.split(basedir, 1)[1].split('/')
        else:
            dir = ''
            dirs = path.split('/')

        for d in dirs:
            if not d: continue
            dir = os.path.join(dir, d)
            mod = cls.GetModFromFile(dir)
            if dir not in packages:
                packages.append(str(mod))

            if os.path.isdir(dir) is False:
                assert os.path.exists(dir) is False, 'Expecting "%s" to be a directory' %dir
                os.mkdir(dir)
                fd = open(os.path.join(dir,'__init__.py'), 'w+')
                fd.write(HEADER + '__all__=[]')
                fd.close()

        return dir

    MakeDirectories = classmethod(MakeDirectories)

