############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import re, socket
from M2Crypto import __m2crypto, m2, util, EVP, SSL
from M2Crypto.SSL.Checker import Checker, SSLVerificationError, NoCertificate,\
    WrongCertificate, WrongHost
    
# Twisted imports
from twisted.python import log
from twisted.internet import base, address, tcp, main, interfaces, reactor
import twisted.protocols.policies as policies
    
try:
    from M2Crypto.SSL.TwistedProtocolWrapper import TLSProtocolWrapper
except ImportError,ex:
    log.err("Upgrade M2Crypto, TwistedProtocolWrapper can be found in SVN.")
    ex.args.append("Upgrade M2Crypto, try SVN.")
    raise

# SSL Constants
SSL.SSLv23_METHOD = 'sslv23'
SSL.SSLv3_METHOD = 'sslv3'
SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS = \
    __m2crypto.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS


def OnConnectHostAuthz():
    reactor.connectSSL = _connectSSL_hostauthz

def OnConnectSelfAuthz():
    reactor.connectSSL = _connectSSL_selfauthz

def OnConnectNoAuthz():
    reactor.connectSSL = _connectSSL_noauthz


def _connectSSL_noauthz(host, port, factory, contextFactory, timeout=30, 
                        bindAddress=None):
    """replace reactor connectSSL...
    """
    return connectSSL(\
        host, port, factory, contextFactory, timeout=timeout, 
        bindAddress=bindAddress, postConnectionCheck=None,
    )
    
def _connectSSL_hostauthz(host, port, factory, contextFactory, timeout=30, 
                          bindAddress=None):
    """replace reactor connectSSL...
    """
    return connectSSL(\
        host, port, factory, contextFactory, timeout=timeout, 
        bindAddress=bindAddress, postConnectionCheck=HostAuthz(),
    )
    
def _connectSSL_selfauthz(host, port, factory, contextFactory, timeout=30, 
                          bindAddress=None):
    """replace reactor connectSSL...
    """
    return connectSSL(\
        host, port, factory, contextFactory, timeout=timeout, 
        bindAddress=bindAddress, postConnectionCheck=SelfAuthz(),
    )
    
def connectSSL(host, port, factory, contextFactory, timeout=30,
               bindAddress=None,
               reactor=reactor,
               postConnectionCheck=Checker()):
    """
    A convenience function to start an SSL/TLS connection using Twisted.
    
    See IReactorSSL interface in Twisted. 
    """
    wrappingFactory = policies.WrappingFactory(factory)
    wrappingFactory.protocol = lambda factory, wrappedProtocol: \
        TLSProtocolSessionCachingWrapper(factory,
                           wrappedProtocol,
                           startPassThrough=0,
                           client=1,
                           contextFactory=contextFactory,
                           postConnectionCheck=postConnectionCheck)
    return reactor.connectTCP(host, port, wrappingFactory, timeout, bindAddress)


class TLSProtocolSessionCachingWrapper(TLSProtocolWrapper):
    """Retains the SSL session for use with m2Crypto,
    to use must set as protocol.
    """
    sessions = {}
    #m2.SSL_CTX_set_session_cache_mode(self.ctx, )
    
    def makeConnection(self, transport):
        """set the stored SSL session."""
        log.msg('makeConnection: %s' %str(transport), debug=True)
        
        addr,sessions = transport.addr, TLSProtocolSessionCachingWrapper.sessions
        
        log.msg('  sessions: %s' %str(sessions), debug=True)
        
        if sessions.has_key(addr):
            log.msg('  set session', debug=True)
            m2.ssl_set_session(self.ssl, sessions[addr]._ptr())

        TLSProtocolWrapper.makeConnection(self, transport)

    def connectionMade(self):
        """connectionMade, make sure handshake has happened and
        store the SSL Session."""
        log.msg('connectionMade ', debug=True)
        
        m2.ssl_do_handshake(self.ssl)
        addr = self.transport.addr
        sessions = TLSProtocolSessionCachingWrapper.sessions
        sessions[addr] = SSL.Session.Session(m2.ssl_get_session(self.ssl))
        
        log.msg('  Session(%s):  \n%s' %(sessions[addr]._ptr(), 
                 sessions[addr].as_text()), debug=True)
        
        TLSProtocolWrapper.connectionMade(self)


class ContextFactory:
    """A factory for SSL context objects, for server SSL connections.
    
    Copied from twisted.internet.ssl.DefaultOpenSSLContextFactory,
    which can not be imported."""
    
    isClient = 0

    def getContext(self):
        """Return a SSL.Context object. override in subclasses."""
        raise NotImplementedError
    
    
class ClientContextFactory(ContextFactory):
    """Copied from twisted.internet.ssl.DefaultOpenSSLContextFactory,
    which can not be imported."""
    
    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=(), capath=''):
        """
        @param privateKeyFileName: Name of a file containing a private key
        @param certificateFileName: Name of a file containing a certificate
        @param sslmethod: The SSL method to use
        """
        self.privateKeyFileName = privateKeyFileName
        self.certificateFileName = certificateFileName
        self.sslmethod = sslmethod
        self.calist = calist
        self.capath = capath
        self.cacheContext()

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_options(SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS)
        ctx.load_cert_chain(self.certificateFileName,
                             keyfile=self.privateKeyFileName)
        for ca in self.calist: 
            ctx.load_verify_locations(cafile=ca, capath=self.capath)
        self._context = ctx

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_context']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self.cacheContext()

    def getContext(self):
        """Create an SSL context.
        """
        return self._context


class AuthzEngine(Checker):
    """add authorization links to the chain.
    """
    def __init__(self, host=None, peerCertHash=None, peerCertDigest='sha1'):
#        self.host = host
#        self.fingerprint = peerCertHash
#        self.digest = peerCertDigest
#        self.numericIpMatch = re.compile('^[0-9]+(\.[0-9]+)*$')
        Checker.__init__(self, host=host, peerCertHash=peerCertHash, 
            peerCertDigest=peerCertDigest)
          
        self.chain = []
        
    def _match(self, host, certHost, subjectAltName=False):
        for link in self.chain:
            if link._match(host, certHost, subjectAltName) is False:
                return False
        return True

    def AddAuthzLink(self, LinkClass, **kw):
        if checker is None: checker = AuthzEngine()

        checker.chain.append(LinkClass())
        return checker

    def setHostAuthz(self): 
        if len(filter(lambda link: instanceof(link, LinkClass), checker.chain)) > 0:
            raise RuntimeError, 'already specified host authorization'
        link = HostAuthz(host=self.host, peerCertHash=self.fingerprint, 
                          peerCertDigest=self.digest)
        self.chain.append(link)

    def setSelfAuthz(self): 
        if len(filter(lambda link: instanceof(link, LinkClass), checker.chain)) > 0:
            raise RuntimeError, 'already specified self authorization'
        link = SelfAuthz(host=self.host, peerCertHash=self.fingerprint, 
                          peerCertDigest=self.digest)
        self.chain.append(link)
        
    def setGridMapAuthz(self, gridmap='/etc/grid-security/grid-mapfile'):
        if len(filter(lambda link: instanceof(link, LinkClass), checker.chain)) > 0:
            raise RuntimeError, 'already specified gridmap authorization'
        link = GridMapAuthz(host=self.host, peerCertHash=self.fingerprint, 
                          peerCertDigest=self.digest, gridmap=gridmap)
        self.chain.append(link)


class IDAuthz:
    """The identity authorization PDP permits only a caller with the specific 
    identity configured in the PDP.
    """
    
    def _match(self, host, certHost, subjectAltName=False):
        """ CN -- host/hostname.domain
        """
        raise NotImplementedError, 'authorization type not implemented.'
    
    
class SAMLCalloutAuthz:
    """The samlCallout authorization PDP contacts an SAML authorization service 
    over a protocol defined in the OGSA-AuthZ working group in the Global Grid 
    Forum (GGF). Only a caller receiving a positive authorization decision from 
    the SAML authorization service is permitted.
    """
    
    def _match(self, host, certHost, subjectAltName=False):
        """ CN -- host/hostname.domain
        """
        raise NotImplementedError, 'authorization type not implemented.'
    
    
class UserNameAuthz:
    """The userName permits users that have successfully authenticated using 
    the configured Java Authentication and Authorization Service (JAAS) login 
    module
    """
    
    def _match(self, host, certHost, subjectAltName=False):
        """ CN -- host/hostname.domain
        """
        raise NotImplementedError, 'authorization type not implemented.'
    

NoAuthz = lambda peerX509, expectedHost: 1

class GridMapAuthz(Checker):
    """The gridmap authorization PDP permits only a caller with an identity 
    listed in the PDP's configured gridmap file, an access-control list of 
    allowed caller identities with one or more mappings to some local user 
    name for each identity. The PDP, upon a permit, also maps the identity to 
    a local user name listed in the gridmap file next to an identity entry.
    """
    def __init__(self, host=None, peerCertHash=None, peerCertDigest='sha1',
                  gridmap='/etc/grid-security/grid-mapfile'):
        Checker.__init__(self, host=host, peerCertHash=peerCertHash, 
            peerCertDigest=peerCertDigest)
        self.gridmap = dict(map(lambda s: s.strip('"').split('" ')[:2], 
                             open(gridmap, 'r')))
                             
    def __call__(self, peerCert, host=None):
        log.msg('GridMap Authorization', debug=True)
        
        if peerCert is None:
            log.err('host: %s' %host)
            raise NoCertificate, 'peer did not return certificate'

        if host is not None:
            self.host = host

        if self.fingerprint:
            if self.digest not in ('sha1', 'md5'):
                raise ValueError, 'unsupported digest "%s"' %(self.digest)

            if (self.digest == 'sha1' and len(self.fingerprint) != 40) or \
               (self.digest == 'md5' and len(self.fingerprint) != 32):
                raise WrongCertificate, 'peer certificate fingerprint length does not match'

            der = peerCert.as_der()
            md = EVP.MessageDigest(self.digest)
            md.update(der)
            digest = md.final()
            if util.octx_to_num(digest) != int(self.fingerprint, 16):
                raise WrongCertificate, 'peer certificate fingerprint does not match'
        
        keys = self.gridmap.keys() 
        subject = str(peerCert.get_subject())
        if subject in keys:
            return True

        
        try:
           ext = peerCert.get_ext("proxyCertInfo") 
        except:
            return False
       
        #form Subject DNs by removing the last CN field in the
        #subject, continue this until subject name matches or 
        #there are no more CN fields
        while True: 
            subject_name_list = subject.split("/") 
            subject_name_list.pop()
            subject_name_list.pop(0)
            new_subject_list = map(lambda x: "/" +x,subject_name_list) 
            subject = "".join(new_subject_list)
            if subject in keys: 
                return True
            try:
                subject.index("CN")
            except ValueError:
                return False
            
           
    
            

class HostAuthz(Checker):
    """The host authorization PDP permits only a caller with the specific host 
    identity configured in the PDP. A host identity is a special form of 
    certificate with a common name (CN) corresponding to a name obtained from 
    DNS or some configured service name.
    """
    
    def _match(self, host, certHost, subjectAltName=False):
        """ CN -- host/hostname.domain
        """
        host = host.lower()
        certHost = certHost.lower()

        if certHost[:5] != 'host/':
            return False
        
        certHost = certHost[5:]
        if certHost == host:
            return True

        ip1 = self.numericIpMatch.match(host)
        ip2 = self.numericIpMatch.match(certHost)
        
        if ip1 and not ip2:
            if host == socket.gethostbyname(certHost):
                return True
            return False
        
        if not ip1 and ip2:
            if certHost == socket.gethostbyname(host):
                return True
            return False
        
        return False
    

class SelfAuthz(Checker):
    """The self authorization PDP permits only a caller with 
    the same identity as the resource owner.
    """
    
    def _match(self, host, certHost, subjectAltName=False):
        """ CN -- host/hostname.domain
        """
        host = host.lower()
        certHost = certHost.lower()

        if certHost == host:
            return True

        ip1 = self.numericIpMatch.match(host)
        ip2 = self.numericIpMatch.match(certHost)
        
        if ip1 and not ip2:
            if host == socket.gethostbyname(certHost):
                return True
            return False
        
        if not ip1 and ip2:
            if certHost == socket.gethostbyname(host):
                return True
            return False
        
        return False
        

