############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, sys, urlparse, optparse, types
from twisted.python import log
from twisted.internet import reactor
from ConfigParser import ConfigParser, NoOptionError
from ZSI.wstools.Namespaces import WSA200403


###########################################################################
# FUNCTIONS
###########################################################################

CP = None
def GetConfigParser(f='client-config.txt'):
    '''Return the global ConfigParser instance.
    '''
    global CP
    if CP is None:
        CP = ConfigParser()
        CP.read(f)
    return CP


def GetOptParser(test, **kw):
    '''Return an OptionParser instance with a few options set from a config file
    test -- configuration file name for test.
    '''
    scheme = 'https'
    cp = GetConfigParser()
    try:
        path = cp.get('Paths', test)
        host = cp.get('Location', 'host')
        port = cp.getint('Location', 'port')
        debug = cp.get('logging', 'debug')
        if not cp.getint('security', 'ssl'):
            scheme = 'http'
    except NoOptionError, ex:
        url = None
    else:
        url = urlparse.urlunparse((scheme, '%s:%d'%(host,port), path, '', '', ''))

    return GetBasicOptParser(debug=debug, url=url, **kw)


def GetBasicOptParser(debug=1, url=None, **kw):
    '''Return an OptionParser instance with a few options.
    '''
    op = optparse.OptionParser(**kw)

    op.add_option("-u", "--url", action="store", dest="url",
        help="location of service", default=url)
    op.add_option("-d", "--debug",
        action="store", dest="debug", default=debug,
        help="Debug Messages to file, if 1 send to stdout, if 0 then off.")
    op.add_option("-w", "--wsAddressURI",
        action="store", dest="wsAddressURI", default=WSA200403.ADDRESS,
        help="Specify URI of ws-addressing version to use.")
    op.add_option("-c", "--cert", action="store", dest="cert",
        default=os.path.join("/tmp", "x509up_u" + str(os.getuid())),
        help="file path of certificate or proxy.")
    op.add_option("-k", "--key", action="store", dest="key",
        default=None,
        help="file path of key")
    op.add_option("-v", "--m2secureConv", action="store", dest="m2secureConv", default=None,
        help="To use secure conversation, provide URL to Authorization service, and negotiate key used to sign/verify messages.")
    op.add_option("-x", "--m2secureMessage", action="store", dest="m2secureMessage", 
        default=None, help="use secure message to sign messages.")
    op.add_option("-y", "--secureMessage", action="store", dest="secureMessage", 
        type="int", default=0, help="use secure message to sign messages.")
    op.add_option("-z", "--secureConv", action="store", dest="secureConv", default=None,
        help="To use secure conversation, provide URL to Authorization service, and negotiate key used to sign/verify messages.")
    op.add_option('-a', '--authz', action='store', dest='authz', default=None,
        choices=('self', 'host',), help="choose authorization, self or host.")
    op.add_option('-m', '--m2crypto', action='store_true', dest='m2crypto', 
        default=False, help="use m2crypto for https.")
    return op



def GetFileDescriptor(filename):
    '''Returns a file descriptor or None.
    filename -- str
    '''
    if type(filename) is types.FileType:
        return filename
    if not filename or filename == '0':
        return None
    elif filename == '1' or filename == 1:
        return sys.stdout
    return open(filename, 'w+')


def GetPortKWArgs(options):
    '''returns a set of key word arguments for port initialization.
    options -- Options instance.
    '''
    kw = {}
    kw['wsAddressURI'] = options.wsAddressURI
    kw['url'] = options.url

    # MESSAGE SECURITY
    if options.m2secureConv:
        from pyGridWare.security.m2.Authenticate import M2AuthenticationHandler
        from pyGridWare.security.m2.M2SignatureHandler import M2SignatureHandler
        kwc = kw.copy()
        kwc['url'] = options.m2secureConv
        auth = M2AuthenticationHandler()
        keyname = auth.authenticate(**kwc)
        kw['sig_handler'] = M2SignatureHandler()
        kw['sig_handler'].setSecureConversation(keyname)
    elif options.m2secureMessage:
        from pyGridWare.security.m2.M2SignatureHandler import M2SignatureHandler
        kw['sig_handler'] = M2SignatureHandler()
        kw['sig_handler'].setSecureMessage()
        
    if options.secureConv:
        from pyGridWare.security.gss.Authenticate import GssAuthenticationHandler
        from pyGridWare.security.gss.GssSignatureHandler import GssSignatureHandler
        kwc = kw.copy()
        kwc['url'] = options.secureConv
        auth = GssAuthenticationHandler()
        keyname = auth.authenticate(**kwc)
        kw['sig_handler'] = GssSignatureHandler()
        kw['sig_handler'].setSecureConversation(keyname)
    elif  options.secureMessage > 0:
        from pyGridWare.security.gss.GssSignatureHandler import GssSignatureHandler
        kw['sig_handler'] = GssSignatureHandler()
        kw['sig_handler'].setSecureMessage()

    # TRANSPORT SECURITY
    usehttps = options.url.startswith('https')
    if not usehttps:
        return kw
    
    calist = \
        filter(\
            lambda s: s, 
                map(\
                     lambda s: s.strip(), 
                     GetConfigParser().get('security', 'ca').split(' ')
                )
        )
        
    capath = ''
    if len(calist) == 0:
        capath = '/etc/grid-security/certificates'
        
    kwargs = {}
    if options.m2crypto:
        from M2Crypto import SSL
        from pyGridWare.utility.m2crypto import OnConnectHostAuthz,\
            OnConnectSelfAuthz, OnConnectNoAuthz, ClientContextFactory
        
        # Authorization
        if options.authz is None:
            OnConnectNoAuthz()
        elif options.authz == 'host':
            OnConnectHostAuthz()
        elif options.authz == 'self':
            OnConnectSelfAuthz()
                
        # TODO: REMOVE THIS 
        kwargs['capath'] = capath
                
    else:
        from twisted.internet.ssl import SSL
        from twisted.web import client
        from pyGridWare.utility.ssl import ClientContextFactory
        try:
            from pyGridWare.utility.web.client import SMHTTPPageGetter
            client.HTTPClientFactory.protocol = SMHTTPPageGetter
        except: pass

    # Set up default /etc/grid-security/certificates
    if len(calist) == 0 and os.path.isdir(capath):
        calist = filter(lambda f: f.endswith('.0'), os.listdir(capath))
            
    if not os.path.isfile(options.cert):
        raise RuntimeError, 'certificate "%s" is not a file.' %options.cert

    if options.key is not None and not os.path.isfile(options.key):
        raise RuntimeError, 'key "%s" is not a file.' %options.key
    
    kw['contextFactory'] = \
        ClientContextFactory(options.key, options.cert, 
            sslmethod=SSL.SSLv3_METHOD, calist=calist, **kwargs)
            
    return kw


def SetUp(options):
    '''General Script SetUp.
    options -- Options instance.
    '''
    fd = GetFileDescriptor(options.debug)
    if fd is not None:
        log.startLogging(fd)

