############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
# System imports
import os, warnings, socket

# Twisted imports
from OpenSSL import SSL
from twisted.python import log
from twisted.internet import interfaces, reactor, ssl


def client_verify_cb(conn, cert, errnum, depth, ok):
    log.msg('Client Got certificate: %s' %cert.get_subject(), debug=True)
    for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
       log.msg("%s -- %s" %(i, eval(i)), debug=True)
    return ok


def server_verify_cb(conn, cert, errnum, depth, ok):
    log.msg('Server Got certificate: %s' % cert.get_subject(), debug=True)
    for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
       log.msg("%s -- %s" %(i, eval(i)), debug=True)
    return ok

def host_authz_cb(conn, cert, errnum, depth, ok):
    '''no good way to grab the CN=host/fqdn
    '''
    log.msg('Host Authz Got certificate: %s' %cert.get_subject(), debug=True)
    x509name = cert.get_subject()
    subject = str(x509name).split('\'')[1]
    if depth == 0:
        x509name = cert.get_subject()
        subject = str(x509name).split('\'')[1]
        peername = conn.getpeername()
        hostsuffix = '/CN=host/%s' %socket.getfqdn(peername[0])
        if subject.endswith(hostsuffix) is False: 
            warnings.warn(\
                'Host authorization failed: expected "%s", peer returned "%s"' \
                %(hostsuffix,subject)
            )
            return 0
    return ok


class ClientContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[], 
                 verify=client_verify_cb):
        self._calist = calist
        self._verify_cb = verify
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def cacheContext(self):
        ssl.DefaultOpenSSLContextFactory.cacheContext(self)
        ctx = self.getContext()
        ctx.set_options(SSL.OP_DONT_INSERT_EMPTY_FRAGMENTS)
        ctx.set_verify(SSL.VERIFY_PEER, self._verify_cb)
        for cert in self._calist:
            ctx.load_verify_locations(cert)


class ServerContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    increment = 0

    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[]):
        self._calist = calist
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_verify(SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, server_verify_cb)
        ctx.use_certificate_file(self.certificateFileName)
        ctx.use_privatekey_file(self.privateKeyFileName)
        for cert in self._calist:
            ctx.load_verify_locations(cert)

        #ctx.set_session_id("takes a string no longer than 32")
        ServerContextFactory.increment += 1
        ctx.set_session_id('%d' %ServerContextFactory.increment)
        self._context = ctx
   


class ACLServerContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    increment = 0

    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[], aclfile=None):
        '''
        file -- ACL file path
        '''
        self._aclfile = aclfile
        self._calist = calist
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def getVerifyCB(self):
        '''
        '''
        if self._aclfile is None or os.path.isfile(self._aclfile) is False:
            raise RuntimeError, 'ACL File must be a file'

        subjects = map(lambda s: s.strip(), open(self._aclfile, 'r'))

        def verify_cb(conn, cert, errnum, depth, ok):
            # JRB no good way to extract a string, 
            # duplicate attributes not available via getattr
            x509name = cert.get_subject()
            subject = str(x509name).split('\'')[1]
            log.msg('ACL Authorization Server Got certificate: %s' % subject, 
                     debug=True)
            for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
                log.msg("%s -- %s" %(i, eval(i)), debug=True)
            if depth == 0 and subject not in subjects:
                warnings.warn("subject(%s) failed ACL authorization" %subject)
                return 0
            return ok

        return verify_cb

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_verify(SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, self.getVerifyCB())
        ctx.use_certificate_file(self.certificateFileName)
        ctx.use_privatekey_file(self.privateKeyFileName)
        for cert in self._calist:
            ctx.load_verify_locations(cert)

        #ctx.set_session_id("takes a string no longer than 32")
        ServerContextFactory.increment += 1
        ctx.set_session_id('%d' %ServerContextFactory.increment)
        self._context = ctx
   


class GridMapServerContextFactory(ACLServerContextFactory):
    def getVerifyCB(self):
        '''
        '''
        if self._aclfile is None or os.path.isfile(self._aclfile) is False:
            raise RuntimeError, 'GridMap File must be a file'

        gridmap = dict(map(lambda s: s.strip('"').split('" ')[:2], open(self._aclfile, 'r')))
        subjects = gridmap.keys()

        def verify_cb(conn, cert, errnum, depth, ok):
            # JRB no good way to extract a string, 
            # duplicate attributes not available via getattr
            x509name = cert.get_subject()
            subject = str(x509name).split('\'')[1]
            log.msg('GridMap Authorization Server Got certificate: %s' % subject,
                     debug=True)
            for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
                log.msg("%s -- %s" %(i, eval(i)), debug=True)
            if depth == 0 and subject not in subjects:
                warnings.warn("subject(%s) failed GridMap authorization (%s)" %(subject, subjects))
                return 0
            return ok

        return verify_cb





#_connectSSL = reactor.connectSSL
#SESSION = None
#def CurryConnectSSL(host, port, factory, contextFactory, timeout=30, bindAddress=None):
#    """@see: twisted.internet.interfaces.IReactorSSL.connectSSL
#    """
#    log.msg('CurryConnectSSL', debug=True)
#    
#    #c = ssl.Connector(host, port, factory, contextFactory, timeout, bindAddress, reactor=reactor)
#    c = SSLConnector(host, port, factory, contextFactory, timeout, bindAddress, reactor)
#    c.connect()
#    #c = _connectSSL(host, port, factory, contextFactory, timeout, bindAddress)
#    
#    log.msg('connection: %s' %str(c))
#    return c
#reactor.connectSSL = CurryConnectSSL
#
#
#class SSLConnector(ssl.Connector):
#    
#    def _makeTransport(self):
#        log.msg('SSLConnector _makeTransport', debug=True)
#        return SSLClient(self.host, self.port, self.bindAddress, self.contextFactory, self, self.reactor)
#        #return ssl.Client(self.host, self.port, self.bindAddress, self.contextFactory, self, self.reactor)
#
#
#class SSLClient(ssl.Client):
#    
#    def doConnect(self):
#        """I connect the socket.
#
#        Then, call the protocol's makeConnection, and start waiting for data.
#        """
#        log.msg('SSLClient doConnect', debug=True)
##        ssl.Connector.doConnect()
#        import sys
#        from twisted.python.runtime import platform, platformType
#        from twisted.internet.tcp import EISCONN, EWOULDBLOCK, EINPROGRESS, EALREADY
#        if not hasattr(self, "connector"):
#            # this happens when connection failed but doConnect
#            # was scheduled via a callLater in self._finishInit
#            return
#
#        # on windows failed connects are reported on exception
#        # list, not write or read list.
#        if platformType == "win32" or sys.platform == "cygwin":
#            r, w, e = select.select([], [], [self.fileno()], 0.0)
#            if e:
#                err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
#                self.failIfNotConnected(error.getConnectError((err, os.strerror(err))))
#                return
#
#        try:
#            connectResult = self.socket.connect_ex(self.realAddress)
#        except socket.error, se:
#            connectResult = se.args[0]
#            
##        global SESSION
##        if not SESSION:
##            SESSION = SSL.Session(self.socket)
##            self.socket.set_session(SESSION)
##        else:
##            self.socket.set_session(SESSION)
#
#
#        if connectResult:
#            if connectResult == EISCONN:
#                pass
#            # on Windows EINVAL means sometimes that we should keep trying:
#            # http://msdn.microsoft.com/library/default.asp?url=/library/en-us/winsock/winsock/connect_2.asp
#            elif ((connectResult in (EWOULDBLOCK, EINPROGRESS, EALREADY)) or
#                  (connectResult == EINVAL and platformType == "win32")):
#                self.startReading()
#                self.startWriting()
#                return
#            else:
#                self.failIfNotConnected(error.getConnectError((connectResult, os.strerror(connectResult))))
#                return
#
#        # If I have reached this point without raising or returning, that means
#        # that the socket is connected.
#        del self.doWrite
#        del self.doRead
#        # we first stop and then start, to reset any references to the old doRead
#        self.stopReading()
#        self.stopWriting()
#        self._connectDone()
#
#    def _connectDone(self):
#        log.msg('  connectDone', debug=True)
#        ssl.Client._connectDone(self)

