############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time, types, warnings
from urlparse import urlparse

import ZSI
from ZSI import EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.fault import FaultFromFaultMessage, FaultType
from ZSI.address import Address, WSActionException
from ZSI.writer import SoapWriter
from ZSI.wstools.Namespaces import OASIS

import twisted.web2
from twisted.web2.client import http as client
from twisted.web2 import http_headers

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log, failure
from twisted.python.failure import Failure
from twisted.internet.tcp import *

try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex:
    DomletteReader = DomletteElementProxy = None

from pyGridWare.wsrf.faults import WrapFaultInException

"""UNSTABLE:   twisted.web2.client is a moving target.
"""


class HTTPPageGetter(client.HTTPClient):
    def handleStatus_500(self):
        '''potentially a SOAP:Fault.
        '''
        pass
    def handleStatus_404(self):
        '''client error, not found
        '''
        log.msg('HTTP Error 404', debug=True)

try:
    from OpenSSL import SSL
    SSL.Session
except:
    client.HTTPClientFactory.protocol = HTTPPageGetter
else:
    class SMHTTPPageGetter(HTTPPageGetter):
        """Retains the SSL session
        """
        session = None

        def makeConnection(self, transport):
            """Do handshake and set session id.
            To make this a little more usefull need to
            associate each session with an URL and security context..
            """
            socket = transport.getHandle()
            use_session = hasattr(socket, 'set_session')
            session = SMHTTPPageGetter.session
            if use_session is True:
                if session is None:
                    SMHTTPPageGetter.session = SSL.Session(socket)
                else:
                    socket.set_session(session)

            client.HTTPClient.makeConnection(self, transport)

    client.HTTPClientFactory.protocol = SMHTTPPageGetter


class Binding:
    def __init__(self, url=None, nsdict=None, readerclass=DomletteReader, 
                 writerclass=DomletteElementProxy, wsAddressURI=None, 
                 sig_handler=None, contextFactory=None, **kw):

        self.url = url
        self.nsdict = nsdict or {}
        self.readerclass = readerclass
        self.writerclass = writerclass
        self.wsAddressURI = wsAddressURI
        self.sig_handler = sig_handler
        self.endPointReference = kw.get('endPointReference')
        self.contextFactory = contextFactory

        self._factory = None
        self._agent = 'pyGridWare Binding'
        self.http_headers  = {'content-type': twisted.web2.http_headers.MimeType('text/xml',None),}
        self._fault = False

    def addHTTPHeader(self, key, value):
        self.http_headers[key] = value
   
    def _getHTTPHeaders(self):
        return self.http_headers

    def _serializeRequest(self, url, obj, nsdict={}, soapaction=None, **kw):
        '''returns string representing request
        '''
        url = url or self.url
        if kw.has_key('requesttypecode'):
            tc = kw['requesttypecode']
        elif kw.has_key('requestclass'):
            tc = kw['requestclass'].typecode
        elif type(obj) == types.InstanceType:
            tc = getattr(obj.__class__, 'typecode')
        elif hasattr(obj, 'typecode'):
            tc = getattr(obj.__class__, 'typecode')
        else:
            raise RuntimeError, 'Must provide a typecode to serialize obj'

        # Serialize the object.
        d = {}
        d.update(self.nsdict)
        d.update(nsdict)

        use_header = self.sig_handler is not None or self.wsAddressURI is not None
        sw = SoapWriter(nsdict=d, header=use_header, outputclass=self.writerclass)
        sw.serialize(obj, tc)

        # Serialize WS-Address
        address = None
        if self.wsAddressURI:
            epr = kw.get('endPointReference', self.endPointReference)
            wsaction = kw.get('wsaction')
            if soapaction and wsaction.strip('\'"') != soapaction:
                raise WSActionException, 'soapAction(%s) and WS-Action(%s) must match'\
                    %(soapaction,wsaction)

            address = Address(url, wsAddressURI=self.wsAddressURI)
            address.setRequest(epr, wsaction)
            address.serialize(sw, typed=False)

        if self.sig_handler is not None:
            self.sig_handler.sign(sw)
        soapdata = str(sw)

        log.msg("%s %s REQUEST:\n%s" % ("_" * 33, time.ctime(time.time()), soapdata), debug=True)

        return address,soapdata
    
    def Send(self, url, opname, obj, nsdict={}, soapaction=None, **kw):
        '''serialize obj and send 
        '''
        url = url or self.url
#        cookies = None
#        if self._factory is not None:
#            cookies = self._factory.cookies
         
        #TODO: ValueError: No header generator for header 'soapaction', 
        #    either add one or use setHeaderRaw.
        #self.addHTTPHeader('SOAPAction', soapaction)
        address,soapdata = self._serializeRequest(url, obj, nsdict, soapaction, **kw)
#        f = getPage(str(url), contextFactory=self.contextFactory, 
#                          postdata=soapdata, agent=self._agent, 
#                          method='POST', headers=self._getHTTPHeaders(), cookies=cookies)
                          
                          
        scheme, netloc, path, params, query, fragment = urlparse(url)
        host, port = twisted.web2.http.splitHostPort(scheme, netloc)
                          
#        print kwargs
#        key,headers = 'headers',None
#        if kwargs.has_key(key):
#            headers = kwargs[key]
#            del kwargs[key]
        
        factory = client.HTTPClientFactory(host, port)
        request = twisted.web2.client.http.Request('POST', path, args=None, 
                                                    headers=self.http_headers, stream=soapdata)
        factory.addRequest(request)
                          
        if isinstance(factory, Failure):
            return factory
        
        d = request.deferred
        d.address = address
        
        # Hold on to factory for deferred and cookies.
        self._factory = factory
        self._factory.deferred = d
        return d

    def Receive(self, replytype, **kw):
        '''This method allows code to act in a synchronous manner, it waits to return until 
        the deferred fires but it doesn't prevent other queued called from being executed.
        Send must be called first, which sets up the factory.
        '''
        assert getattr(self, '_factory') is not None, 'Must call Send before Receive.'
        assert isinstance(replytype, ZSI.TC.TypeCode), 'replytype must be a TypeCode.'

        failure = []
        def errback(result):
            '''Used with Response method to suppress 'Unhandled error in Deferred'
            messages by adding an errback.
            '''
            failure.append(result)
            return None

        #TODO: Fix this, should provide the deferred as a parameter and
        #    avoid potential problems.
        d = self._factory.deferred
        d.addCallback(self.handleHTTPResponse, replytype=replytype, address=d.address, **kw)
        d.addErrback(errback)

        while not d.called:
            reactor.runUntilCurrent()
            t2 = reactor.timeout()
            t = reactor.running and t2
            reactor.doIteration(t)

        pyobj = d.result
        if pyobj is None and len(failure) > 0:
            failure[0].raiseException()

        return pyobj
    
    def handleHTTPResponse(self, response, replytype, **kw):
        ps = response.stream
        result = ps.read()
        if ps.length <= 0 or result is None:
            raise TypeError('Received empty response')
        
        if not isinstance(result, defer.Deferred):
            raise TypeError('expecting deferred, twisted.web2.client is unstable..')
        
        result.addCallback(self.handleSOAPMessage, replytype, **kw)
        return result

    getResponseSOAPMessage = handleHTTPResponse
    
    def handleSOAPMessage(self, soapdata, replytype, **kw):
        '''called by deferred, returns pyobj representing reply.
        Parameters and Key Words:
          soapdata -- str
          replytype -- reply type of response
          address -- WS-Address
        '''
        address = kw.get('address')
        if address is None:
            raise EvaluateException, 'Missing WS-Address'

        log.msg("%s %s RESPONSE:\n%s" % ("_" * 33, time.ctime(time.time()), soapdata), debug=True)

        ps = ParsedSoap(soapdata, readerclass=self.readerclass)
        if ps.IsAFault() is True:
            log.msg('raise FaultFromFaultMessage', debug=True)
            raise FaultFromFaultMessage(ps)

        pyobj = ps.Parse(replytype)
        wsaction = kw.get('wsaction')
        if wsaction is not None:
            address.checkResponse(ps, wsaction)

        if self.sig_handler is not None:
            self.sig_handler.verify(ps)

        return pyobj




