###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

"""WSResource for use with twisted.web2
"""

import sys, warnings
from twisted import copyright
TWISTED_VERSION = (2,0,0)
try:
    if copyright.version != 'SVN-Trunk' or \
        tuple(map(lambda i:int(i), copyright.version.split('.'))) < TWISTED_VERSION:
        warnings.warn(\
            'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
            RuntimeWarning,
        )
except:
    warnings.warn(\
        'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
        RuntimeWarning,
    )
    
from ZSI import _get_element_nsuri_name
from ZSI import ParseException
from ZSI.parse import ParsedSoap
from ZSI.writer import SoapWriter
from ZSI.address import Address
from ZSI import fault

import twisted.web2
from twisted.python import log
from twisted.internet.defer import Deferred

from pyGridWare.security.interfaces import IWSSecurity 
from pyGridWare.utility.http import GetURLFromRequest
try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex: 
    DomletteReader = DomletteElementProxy = None

from pyGridWare.wsrf.faults import BaseFault
    

class WSResource(twisted.web2.resource.Resource, object):
    '''
    class variables:
        encoding  
        soapAction -- Must override to set input Action values.
           Dictionary of soapAction keys, and operation name values.
           These are specified in the WSDL soap bindings. There must be a 
           class method matching the operation name value.  If WS-Action is
           used the keys are WS-Action request values, according to the spec
           if soapAction and WS-Action is specified they must be equal.
           
        wsAction -- Must override to set output Action values.
           Dictionary of operation name keys and WS-Action 
           response values.  These values are specified by the portType.

        root -- dictionary of root element keys, and operation name values.
    '''
    soapAction = {}
    wsAction = {}
    root = {}
    encoding = "UTF-8"

#    def __init__(self):
#        twisted.web2.resource.Resource.__init__(self)

    def render(self, ctx):
        req=iweb.IRequest(ctx)
        log.msg('render...', debug=True)
        return http.Response(stream="Host:%s, Path:%s"%(req.host, req.path))
        
    def http_POST(self, ctx):
        """Dispatch Method called by twisted render.  Returns
        a 
        ctx -- twisted.web2.context.RequestContext
        """
        request = ctx.locate(twisted.web2.iweb.IRequest)
        log.msg('Request -- %s' %request.__class__, debug=True)
        
        #TODO: No Header Parser for soapaction?
        log.msg('Headers -- %s' %request.headers._headers, debug=True)
        soap_action = request.headers.getRawHeaders('soapaction')
        
        # TODO: Guess there could be multiple soapaction headers?
        #  raise appropriate error here.
        if soap_action is not None:
            if len(soap_action) != 1: 
                raise RuntimeError, 'soapaction problem: %s' %soap_action
            soap_action = soap_action[0]
            
        log.msg('soapAction -- %s' %soap_action, debug=True)
        
        log.msg('Stream -- %s' %request.stream.__class__, debug=True)
        
        # stream.read returns a deferred
        data = request.stream.read()
        data.addCallback(self._parse_request, request, soap_action)
        data.addCallback(self._process_request, request, soap_action)
        data.addCallback(self._retrieve_response, request)
        data.addCallback(self._write_response, request)
        data.addErrback(self._soapFault, request)
        return data
        
        
    def _parse_request(self, data, request, soap_action):
        log.msg('HTTP Body Content:\n%s' %data, debug=True)
        try:
            ps = ParsedSoap(data, readerclass=DomletteReader)
        except ParseException, e:
            log.err('failed to parse request')
            raise e
        return ps

    def _process_request(self, ps, request, soap_action):
        # Verify if Signed
        if IWSSecurity.providedBy(self):
            self.verify(ps)

        # WS-Address
        address = Address()
        try:
            address.parse(ps)
        except ParseException, e:
            log.err('failed to parse request ws-address')
            raise

        action = address.getAction()
        if not action:
            log.err('No WS-Action specified in Request')
            e = WSActionException('No WS-Action specified in Request')
            return self._soapFault(request, e)

        if soap_action and soap_action.strip('\'"') != action:
            log.err(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.'\
                %(soap_action,action)
            )
            e = WSActionException(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.'\
                %(soap_action,action)
            )
            return self._soapFault(request, e)
        return ps,address

    def _retrieve_response(self, args, request):
        ps, address = args
        action = address.getAction()
        op_name = self.root.get(_get_element_nsuri_name(ps.body_root)) or \
            self.soapAction.get(action)
            
        if op_name is None:
            raise UnknownRequestException,\
                'failed to map request to a method: action(%s), root%s'\
                %(action,_get_element_nsuri_name(ps.body_root))
        
        if not hasattr(self, op_name):
            log.err('Failed to locate service instance/bound method from WS-Address')
            return self._soapFault(request, e)
        
        try:
            pyobj = getattr(self, op_name)(ps, address, request=request)
        except Exception, e:
            log.err('Failed to call service method: %s' %e)
            raise

        if pyobj is None:
            log.msg('No Response specified', debug=True)
            return NOT_DONE_YET
        return pyobj, address, op_name
        
    def _write_response(self, args, request):
        pyobj, address, op_name = args
        sw = SoapWriter(outputclass=DomletteElementProxy)
        try:
            sw.serialize(pyobj)
        except Exception, e:
            return self._soapFault(request, e)

        if self.wsAction.has_key(op_name) is False:
            raise WSActionNotSpecified,\
                'wsAction dictionary missing key(%s)' %op_name
            
        rsp_action = self.wsAction[op_name]
        addressRsp = Address(action=rsp_action)
        addressRsp.setResponseFromWSAddress(address, GetURLFromRequest(request))
        try:
            addressRsp.serialize(sw, typed=False)
        except Exception, e:
            log.err('Problem serializing WS-Address for Response')
            #return self._soapFault(request, e)
            raise 
        # Sign if WS-Security
        if IWSSecurity.providedBy(self):
            self.sign(sw)
            
        soapdata = str(sw)
        
        mediaType = 'text/xml'
        if self.encoding is not None:
            mediaType = 'text/xml; charset="%s"' %self.encoding
            
        response = \
            twisted.web2.http.Response(code=200, 
                headers={"Content-type":twisted.web2.http_headers.MimeType(mediaType,None),
                          "Content-length":len(soapdata)}, 
                stream=soapdata)
                
        request.writeResponse(response)
        
        
    def _soapFault(self, e, request):
        '''
        request -- request message
        e -- Exception 
        '''
        soapdata = None
        if isinstance(e, BaseFault):
            soapdata = fault.Fault(fault.Fault.Server, 'userException', None, 
                [e.fault], None).AsSoap()
        else:
            soapdata = fault.FaultFromException(e, False, sys.exc_info()[2]).AsSOAP()

        log.err('SOAP FAULT: %s' % soapdata)
    
        mediaType = 'text/xml'
        if self.encoding is not None:
            mediaType = 'text/xml; charset="%s"' %self.encoding
            
        response = \
            twisted.web2.http.Response(code=500, 
                headers={"Content-type":twisted.web2.http_headers.MimeType(mediaType,None),
                          "Content-length":len(soapdata)}, 
                stream=soapdata)
                
        request.writeResponse(response)
        
    
