############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time
from ZSI.TC import Boolean, QName
from twisted.python import log
from twisted.application import internet
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.wsrf.faults.BaseNotificationFaults import\
    NoCurrentMessageOnTopicFault, InvalidTopicExpressionFault
from pyGridWare.wsrf.faults.PropertiesFaults import ResourceUnknownFault
from pyGridWare.wsrf.notification.NotificationProducer import NotificationProducer
from pyGridWare.resource.ResourceContext import ResourceContext
import pyGridWare.generated.types.basenotification


class DIALECT:
    '''WS-Topics dialects
    '''
    SIMPLE = 'http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple'
    
    
class _Policy(int):
    typecode = Boolean('pause')


class QName(tuple):
    typecode = QName()


class Subscription(object):
    '''Represents a subscription, registered to NotificationProducer
    '''
    def __init__(self, request, context, producer):
        '''Instance Data:
        request  -- Subscription Request Message.
        context  -- SubscriptionManagerContext instance.
        producer -- Monitoring Resource Context. 
        '''
        self.request = request
        self.context = context
        self.producer = producer

    def getContext(self):
        return self.__sub_context
    def setContext(self, context):
        self.__sub_context = context
    context = property(getContext, setContext, None,
        'Subscription Context')
 
    def getProducerContext(self):
        return self.__p_context
    def setProducerContext(self, context):
        assert isinstance(context, ResourceContext), \
            'resource must be a ResourceContext instance'
        self.__p_context = context
    producer = property(getProducerContext, setProducerContext, None,
        'Subscribed to Resource Context')

    def getSubscriptionRequest(self):
        return self.__request
    def setSubscriptionRequest(self, request):
        self.__request = request
    request = property(getSubscriptionRequest, setSubscriptionRequest, None,
        'Subscription Request')

    def getTopicExpression(self):
        return self.request._TopicExpression
    topic = property(getTopicExpression, None, None,
        'TopicExpression')

    def getConsumerReference(self):
        return self.request._ConsumerReference
    reference = property(getConsumerReference, None, None,
            'ConsumerReference')


class SubscriptionManager(Singleton):
    '''instance variables:
    subscriptions -- dict of Subscription instances.
    binding -- service stub binding
    transdict -- dict of args to pass to notification client
        transport.
    ''' 
    def __init__(self, binding, transdict=None):
        self.subscriptions = {}
        self.binding = binding
        self.transdict = transdict or {}
        
    def initialize(self, properties):
        '''initialize NotificationProducer ResourceProperties,
        FixedTopicSet, and TopicExpressionDialects.
        properties -- resource properties.
        topics -- Topics supported by NotificationProducer.
        '''
        topics = properties.Topic = []
        l = filter(lambda e: e.pname=='Topic', properties._pyobj.typecode.ofwhat)
        if len(l) != 1: 
            # TODO: Not sure if this is really correct fault to throw.
            raise ResourceUnknownFault(\
                timestamp=time.gmtime(),
                originator=None,
                errorcode=None,
                description=None,
                faultcause=('NotificationProducer must have ResourceProperty Topic',),
            )

        typecode = l[0]
        pyclass = typecode.pyclass
        for e in properties._pyobj.typecode.ofwhat:
            if e.pname in ('FixedTopicSet', 'TopicExpressionDialects', 'Topic'):
                continue
            
            topic = pyclass()
            topics.append(topic)
            topic._attrs = {'Dialect':DIALECT.SIMPLE}
            topic._text = QName((e.nspname,e.pname))
            
        properties.FixedTopicSet = True
        properties.TopicExpressionDialects = [DIALECT.SIMPLE]
        
    def current(self, msg, ctx):
        '''In response to a GetCurrentMessage message, the NotificationProducer 
        MAY return the last NotificationMessage published to a given Topic.
        msg -- GetCurrentMessage message
        ctx -- resource context.
        request -- HTTPRequest
        '''
        raise NoCurrentMessageOnTopicFault(\
            timestamp=time.gmtime(), 
            originator=None,
            errorcode=None,
            description=None,
            faultcause=['Unknown Dialect: %s' %exp._attrs['Dialect']],
        )

    def subscribe(self, msg, ctx):
        '''Returns an EPR representing a SubscriptionReference.
        msg -- subscription request, containing ConsumerReference and TopicExpression.
        ctx -- resource context subscription applies to.
        request -- HTTPRequest
        '''
        dialect = msg.TopicExpression.get_attribute_Dialect() 
        if dialect != DIALECT.SIMPLE:
            raise InvalidTopicExpressionFault(\
                timestamp=time.gmtime(), 
                originator=None,
                errorcode=None,
                description=None,
                faultcause=['Unknown Dialect: %s' %dialect],
        
            )
        
        if isinstance(ctx, ResourceContext) is False:
            raise TypeError, 'must be a ResourceContext instance'

        server = self.binding.server
        scheme = 'http'
        # SSL: Either SSLServer or M2Crypto Protocol Wrapper
        if getattr(server._port.factory.protocol, 'TLS', False) is True\
            or isinstance(self.binding.server._port, internet.SSLServer):
             scheme = 'https'
        
        # TODO: Need to figure out Path by using the Resource...
        url = '%s://%s:%d/wsrf/core/SubscriptionManagerService' \
            %(scheme, self.binding.server._port.getHost().host,
              self.binding.server._port.port)

        # SubscriptionManager Context
        sctx = self.binding.NewResourceContextSubscriptionManagerPort(url)
        sctx.SubscriptionPolicy = _Policy(1)
        sub = Subscription(msg, sctx, ctx)

        NotificationProducer.AddProducer(sub, **self.transdict)
        self.subscriptions[id(sctx)] = sub

        return AddressingUtils.createEndpointReference(sctx)
    
    def pause(self, ctx):
        # How to store pause state?
        ctx.SubscriptionPolicy = _Policy(0)

    def resume(self, ctx):
        # How to store pause state?
        ctx.SubscriptionPolicy = _Policy(1)
        
