##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, warnings, pydoc, inspect
from distutils import log
from os.path import isdir, isfile
from ConfigParser import ConfigParser, NoOptionError


VERBOSE = False
CONFIG_FILE = 'config.txt'
WSDL_SEC = 'WSDL'
TYPES_SEC = 'TYPES'
PERSISTENT_SEC = 'Persistent'
USE_SETUPTOOLS = True
GENERATED = 'pyGridWare.generated'
FLAG = '--no-setuptools'


if sys.version_info < (2, 4):
    raise SystemExit("Python 2.4 or later is required")


if FLAG in sys.argv:
    USE_SETUPTOOLS = False
    sys.argv.remove(FLAG)

try:
    if not USE_SETUPTOOLS: raise ImportError("")
    from setuptools import setup, find_packages
    from setuptools.dist import Distribution as _Distribution

    class PGDistribution(_Distribution):
        """Indirection so that setuptools grabs "setup_requires" dependencies
        before calling "make_generated_package()"
        """
        def finalize_options (self):
            if make_generated_package() is True:
                 self.packages = find_packages()
            _Distribution.finalize_options(self)


except ImportError:
    from distutils.util import convert_path
    from distutils.core import setup
    USE_SETUPTOOLS = False

    def find_packages(where='.', exclude=()):
        """Return a list all Python packages found within directory 'where'
    
        'where' should be supplied as a "cross-platform" (i.e. URL-style) path; it
        will be converted to the appropriate local path syntax.  'exclude' is a
        sequence of package names to exclude; '*' can be used as a wildcard in the
        names, such that 'foo.*' will exclude all subpackages of 'foo' (but not
        'foo' itself).
        """
        out = []
        stack=[(convert_path(where), '')]
        while stack:
            where,prefix = stack.pop(0)
            for name in os.listdir(where):
                fn = os.path.join(where,name)
                if (os.path.isdir(fn) and
                    os.path.isfile(os.path.join(fn,'__init__.py'))
                ):
                    out.append(prefix+name); stack.append((fn,prefix+name+'.'))
        for pat in exclude:
            from fnmatch import fnmatchcase
            out = [item for item in out if not fnmatchcase(item,pat)]
        return out



def IsGenerated():
    return isfile(GENERATED.replace('.', '/') + '/__init__.py')


def RegenerateModules(items, services=True, stubs=True, cp=None, web2=None):
    """Regenerate generated modules with new genrator
    """
    if VERBOSE:
         from ZSI.wstools.logging import setBasicLoggerDEBUG
         setBasicLoggerDEBUG()

    from ZSI.wstools.WSDLTools import WSDLReader
    from ZSI.wstools.XMLSchema import SchemaReader
    # set up pyclass metaclass
    from ZSI.generate.containers import TypecodeContainerBase, TypesHeaderContainer
    TypecodeContainerBase.metaclass = 'pyclass_type'
    TypesHeaderContainer.imports.append(\
        'from ZSI.generate.pyclass import pyclass_type'
    )
    sys.path.append('%s/pyGridWare/utility/generate' %os.getcwd())
    from Modules import WriteServiceModules, WriteClientStubModule,\
        WriteTypesModule, NR
    from Utility import CLIENT,SERVICE,SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE

    #if services is stubs is False:
    #    return

    add_packages = NR.Reset()
    add_packages.insert(0, GENERATED)
    for opt,url in items:
        log.info("Generate: %s, %s",opt,url)
        wsdl = None
        reader = WSDLReader()
        if services is stubs is False:
            reader = SchemaReader()

        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        if services is stubs is False:
            schema = wsdl
            class wsdl:
                name = opt
                types = {schema.targetNamespace:schema}

        kw = {'web2':web2}
        if cp is not None:
            try: kw['simple'] = not cp.getboolean(PERSISTENT_SEC, opt)
            except NoOptionError: pass
            
        if services:
            WriteServiceModules(opt, wsdl, **kw)
        if stubs:
            WriteClientStubModule(opt, wsdl)
        else:
            WriteTypesModule(opt, wsdl)

    return add_packages
    

   
def make_generated_package():
    """Returns False if "generated" package is already present,
    else generates it and returns True. 
    """
    if not (not IsGenerated() and len(sys.argv) > 1 and 
        sys.argv[1] in ('install', 'build', 'doc')):
        return False

    cp = ConfigParser()
    cp.read(CONFIG_FILE)

    # attempt to remove generated
    d = GENERATED.replace('.', '/')
    if isdir(d):
        try:
            os.removedirs(d)
        except OSError:
            print "ERROR: Must Manually Remove Generation Directory (rm -rf %s)" %d
            sys.exit(-1)
 
    # Create types
    RegenerateModules(cp.items(TYPES_SEC), 
        services=False, stubs=False, cp=cp, web2=False)
        
    # Create services/stubs/types
    RegenerateModules(cp.items(WSDL_SEC), 
        services=True, stubs=True, cp=cp, web2=False)
      
    return True 

 
def main():
    kw = {}
    if USE_SETUPTOOLS:
          kw['install_requires'] = [
              "setuptools >= 0.6a11",
              "ZSI >= 1.2.0rc2", 
          ]
          kw['extras_require'] = {
              "MessageSecurity":["M2Crypto >= 0.13",]
          }
          kw['setup_requires'] = [
              "ZSI >= 1.2.0rc2", 
          ]
          kw['entry_points'] = """
[distutils.commands]
pydoc = pyGridWare.utility.command.docker:PyDocForGeneratedModules"""
          kw['zip_safe'] = False
          kw['distclass'] = PGDistribution
    else:
        make_generated_package() 

    setup(name="pyGridWare", 
          version="1.2.0", 
          author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          url="http://dsd.lbl.gov/gtg/projects/pyGridWare",
          download_url="http://dsd.lbl.gov/gtg/projects/download/download_info.html",
          description="Python WS-Resource Framework and GSS Security",

          dependency_links = [
              #"http://pyxml.sourceforge.net/",
              #"http://sourceforge.net/projects/foursuite/",
              "http://www.python.org/pypi/ZSI"
          ],
          packages=find_packages(),
          data_files=[
              ("config",['bin/server-config.tac', 'bin/server-config.txt',]),
          ],
          scripts=["tools/wsdl2web.py"],
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
          classifiers = [
              "Development Status :: 5 - Production/Stable",
              "Intended Audience :: Developers",
              "Intended Audience :: Science/Research",
              "License :: OSI Approved :: BSD License",
              "Programming Language :: Python",
              "Natural Language :: English",
              "Operating System :: Microsoft :: Windows",
              "Operating System :: POSIX"
          ],
          **kw
    )


if __name__ == '__main__': 
    main()
