#!/usr/bin/env python
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import optparse, os, sys, signal, time, warnings, unittest
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from ConfigParser import ConfigParser

from twisted.internet import reactor
from twisted.python import log

from ZSI import ZSIException
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.wstools.Namespaces import WSA200403

from pyGridWare.utility.generate.Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE, NameResolver as NR
from pyGridWare.utility.generate.Modules import WriteServiceModules,\
    WriteClientStubModule, WriteTypesModule

from pyGridWare.wsrf.notification import SubscribeUtility
from pyGridWare.wsrf.notification.NotificationConsumer import \
    NotificationConsumer, NotificationConsumerService
    
import pyGridWare.wsrf.faults
    
"""
GLOBALS:
  WSDL    -- WSDL to generate types, stubs from.
  SECTION -- Ends up being the generated module representing this WSDL.
"""
WSDL = '../../share/schema/core/samples/counter/counter_service.wsdl'

class CounterTestCase(unittest.TestCase):
    def __init__(self, methodName='runTest', epr=None):
        unittest.TestCase.__init__(self, methodName)
        self.rsp = None
        self.epr = epr
        self.port = None

    def getMethodToCall(self):
        '''return method to call, must be callable and 
        takes request as argument.
        '''
        raise NotImplementedError, 'this method must be implemented'
        
    def NewRequestInstance(self):
        '''return request instance.
        '''
        raise NotImplementedError, 'this method must be implemented'
        
    def getEPR(self):
        '''if factory method, return EPR from response.
        '''
        return self.epr

    def setEPR(self, epr):
        '''set EndPointReference
        '''
        self.epr = epr

    def setUp(self):
        import generated.counter_service.stubs.counter_service
        loc = generated.counter_service.stubs.counter_service.CounterServiceLocator()
        self.port = loc.getCounterPortType(\
            url='http://localhost:9080/wsrf/services/CounterService',
            wsAddressURI=WSA200403.ADDRESS,
            endPointReference=self.epr,
        )
        #self.port.binding.AddHeader('connection', 'keep-alive')
        self.shouldStop = True
        
    def tearDown(self):
        self.shouldStop = False
        
    def runTest(self, epr=None):
        kw = {}
        msg = self.NewRequestInstance()
        method = self.getMethodToCall()
        if msg is None:
            raise
        if method is None or not callable(method):
            raise
        if epr is not None:
            self.port.endPointReference = epr
        
        self.rsp = method(msg, **kw)


# Counter Specific
class AddTestCase(CounterTestCase):
    '''Counter Add
    '''
    def getMethodToCall(self):
        return self.port.add

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.AddInputMessage(10)

    def setUp(self):
        self.failUnless(self.epr is not None)
        CounterTestCase.setUp(self)
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.AddOutputMessage))
        CounterTestCase.tearDown(self)


class CreateTestCase(CounterTestCase):
    '''WS-
    '''
    def getMethodToCall(self):
        return self.port.createCounter
        
    def getEPR(self):
        '''if factory method, return EPR from response.
        '''
        if self.rsp is None: return
        self.failUnless(getattr(self.rsp,'_EndpointReference',None) is not None, 
            'no EPR in create response.')
        return self.rsp._EndpointReference

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.CreateCounterRequest()
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.CreateCounterResponse))
        CounterTestCase.tearDown(self)
        
# WS-Resource
class GetResourcePropertyTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetResourceProperty

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.GetResourcePropertyRequest(("http://counter.com","Value"))
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.GetResourcePropertyResponse))
        CounterTestCase.tearDown(self)


class GetMultipleResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetMultipleResourceProperties

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.GetMultipleResourcePropertiesRequest()
        for qname in [("http://counter.com","Value"), 
            ("http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd","CurrentTime"),]:
            msg._ResourceProperty.append(qname)
        return msg

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.GetMultipleResourcePropertiesResponse))
        CounterTestCase.tearDown(self)


class QueryResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.QueryResourceProperties

    def NewRequestInstance(self):
        from ZSI.wstools.Namespaces import DSIG
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.QueryResourcePropertiesRequest()
        msg._QueryExpression = filter(lambda what: what.aname=='_QueryExpression', msg.typecode.ofwhat)[0].pyclass()
        msg._QueryExpression._text  = "contains(/*/*[namespace-uri()='http://counter.com' and local-name()='operations supported'],\"Value\")"
        msg._QueryExpression._attrs = {\
            'Dialect':DSIG.XPATH,
        }
        return msg

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp)
        CounterTestCase.tearDown(self)

    def runTest(self, epr=None):
        import generated.counter_service.stubs.counter_service
        CounterTestCase.runTest(self, epr=None)
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp._text == 'true')


# -----------------------------------------------
# WS-Lifetime
# -----------------------------------------------
class DestroyTestCase(CounterTestCase):
    '''WS-
    '''
    def getMethodToCall(self):
        return self.port.Destroy

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.DestroyRequest()
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.DestroyResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = True


class SetTerminationTimeTestCase(CounterTestCase):
    '''after this returns resource should be destroyed.
    '''
    def getMethodToCall(self):
        return self.port.SetTerminationTime

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.SetTerminationTimeRequest()
        msg._RequestedTerminationTime = time.gmtime(time.time()+2)
        return msg
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.SetTerminationTimeResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = False
        
    def runTest(self, epr=None):
        '''After returns resource should be destroyed.
        '''
        CounterTestCase.runTest(self, epr)
        time.sleep(5)
        
        
# -----------------------------------------------
# WS-BaseNotification
# -----------------------------------------------
class SubscribeTestCase(CounterTestCase):
    '''WS-BaseNotification
    '''
    class NCService(NotificationConsumerService):
        def __init__(self, *args, **kw):
            NotificationConsumerService.__init__(self, *args, **kw)
        def wsa_Notify(nelf, ps, address, **kw):
            response = NotificationConsumerService.wsa_Notify(nelf, ps, address)
            return response,None

    def __init__(self, methodName='runTest', epr=None):
        CounterTestCase.__init__(self, methodName=methodName, epr=epr)
        self.consumer = NotificationConsumer(notificationClass=SubscribeTestCase.NCService)

    def getMethodToCall(self):
        return self.port.Subscribe

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.SubscribeRequest()
        SubscribeUtility.SimpleTopicExpression(msg, self.consumer, 
            namespace="http://counter.com", name="Value")
        return msg

    def setUp(self):
        self.consumer.start()
        CounterTestCase.setUp(self)

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.SubscribeResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = False


class ResourceUnknownTestCase(CounterTestCase):
    
    def __init__(self, klass, methodName='runTest', epr=None):
        unittest.TestCase.__init__(self, methodName)
        self.wrappedTest = klass(methodName=methodName, epr=epr)

    def getEPR(self):
        return self.wrappedTest.getEPR()

    def setEPR(self, epr):
        self.wrappedTest.setEPR(epr)

    def NewRequestInstance(self):
        raise NotImplementedError, 'this method is not implemented'
        
    def setUp(self):
        self.wrappedTest.setUp()
        
    def tearDown(self):
        self.shouldStop = True
        
    def runTest(self, epr=None):
        callableObj = self.wrappedTest.runTest
        self.failUnlessRaises(ZSIException, callableObj, epr=epr)
      
 
class CounterTestSuite(unittest.TestSuite):
    def __init__(self, tests=()):
        unittest.TestSuite.__init__(self, tests=tests)

    def run(self, result):
        '''Factory method first.  Each factory method call will change the EPR
        for subsequent calls.
        '''
        epr = None
        for test in self._tests:
            if result.shouldStop: 
                break
            test.setEPR(epr)
            test(result)
            epr = test.getEPR()
            
        return result


def WSRPCounterTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), AddTestCase(), GetResourcePropertyTestCase(), 
                GetMultipleResourcePropertiesTestCase(), DestroyTestCase()),
        )

def WSRPBrokenTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), QueryResourcePropertiesTestCase(), 
                    DestroyTestCase()),
        )
   
def WSLifetimeCounterTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), SetTerminationTimeTestCase(), 
                    ResourceUnknownTestCase(DestroyTestCase)),
        )

def WSBaseNotificationCounterTestSuite():
    return \
        CounterTestSuite(\
            tests=(CreateTestCase(), SubscribeTestCase(), AddTestCase())
        )


def WSRFCounterTestSuite():
    """Test whole WSRF Family of Specifications
    """
    return unittest.TestSuite(tests=[WSRPCounterTestSuite(), WSLifetimeCounterTestSuite(), 
        WSBaseNotificationCounterTestSuite()])

