#!/usr/bin/env python
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, optparse, shutil
from os.path import isfile, isdir
from ZSI.wstools.WSDLTools import WSDLReader

# set up pyclass metaclass
from ZSI.generate.containers import TypecodeContainerBase, TypesHeaderContainer
TypecodeContainerBase.metaclass = 'pyclass_type'
TypesHeaderContainer.imports.append('from ZSI.generate.pyclass import pyclass_type')

from pyGridWare.utility.generate.Modules import WriteServiceModules, \
    WriteClientStubModule, WriteTypesModule, NR
from pyGridWare.utility.generate.Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE


DESC = """
pyGridWare WS-Resource Framework (WSRF) site creation tool.
\tUses data files:\n\t%s\n\t%s
"""


def RegenerateModules(url, name, services=True, stubs=True, **kw):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return

    reader = WSDLReader()
    if isfile(url) is False:
        wsdl = reader.loadFromURL(url)
    else:
        wsdl = reader.loadFromFile(url)

    if services:
        WriteServiceModules(name, wsdl, **kw)
    if stubs:
        WriteClientStubModule(name, wsdl)
    else:
        WriteTypesModule(name, wsdl)
    return wsdl


def main():
    #     
    # Find Data Files
    #     distutils uses sys.prefix+'config'
    #     setuptools uses dist.location+'config'
    #     
    sctxt = sctac = None
    prefix = sys.prefix
    try:
        import pkg_resources
        d = pkg_resources.get_distribution('pyGridWare')
        if os.path.isdir('%s/config' %d.location): prefix = d.location
    except:
        pass

    # distutils: data_files stored in config dir
    sctxt = '%s/config/server-config.txt' %prefix 
    sctac = '%s/config/server-config.tac' %prefix 
    for i in [sctxt, sctac]:
        if not isfile(i):
            raise RuntimeError, "installation missing data file '%s'" %i
                 

    parser = optparse.OptionParser(usage="usage: %prog [options] wsdl ...", 
                 description=DESC %(sctxt,sctac))

    # Basic options
    parser.add_option("-f", "--force",
        action="store_true", dest="force", default=False,
        help="do not remove anything, but write over existing generated code.")
        
    parser.add_option("-w", "--web2",
        action="store_true", dest="web2", default=False,
        help="create web2 resource")

    parser.add_option("-t", "--top",
        action="store", dest="top", type="string", default=None,
        help="specify top-level module name and others (default to file name). *no multiple args")

    parser.add_option("-r", "--rpy",
        action="store", dest="rpy", default=None, type="string",
        help="basic rpy resource. *no web2, no multiple args")

    parser.add_option("-c", "--script",
        action="store", dest="script", default=None, type="string",
        help="basic client script. *no multiple args")

    (options, args) = parser.parse_args()
    # Test Conditions
    if len(args) < 1 or (len(args) > 1 and\
            not (None is options.script is options.rpy is options.top)): 
        parser.print_help()
        sys.exit(-1)
    if options.script and (not options.script.endswith('.py')\
            or isfile(options.script)): 
        raise RuntimeError, 'Client Script name must have suffix ".py" and not exist.'
        sys.exit(-1)
    if options.rpy and (not options.rpy.endswith('.rpy')\
            or isfile('services/%s' %options.rpy) or options.web2): 
        raise RuntimeError, 'resource file name must have suffix ".rpy" and not exist. no web2.'
        sys.exit(-1)

    # Create README
    if not isfile('README'):
        open('README', 'w').write(README)

    for wsdl in args:
        name = options.top or wsdl.split('/')[-1].split('.')[0] 
    
        # Only if forced to
        if not options.force and isdir('generated/%s' %name):
            print 'Must remove directory "generated/%s" or specify option top or force.' %name
            sys.exit(-1)
        if name == 'services':
            print '"services" is reserved, it cannot be the top-level module name'
            sys.exit(-1)
    
        NR.d = {\
            CLIENT:'generated.%s.stubs' %name,
            SERVICE:'generated.%s.services' %name,
            SERVICE_STUBS:'generated.%s.services.stubs' %name,
            TYPES:'generated.%s.types' %name,
            PROPERTIES:'generated.%s.properties' %name,
            RESOURCE:'generated.%s.resource' %name,
        }
    
        try:
            wsdl = RegenerateModules(wsdl, name=name, services=True, 
                                      stubs=True, web2=options.web2, simple=1)
        except Exception, ex:
            print "Failure: ", ex
            sys.exit(-1)
            
        if not options.script and not options.rpy: 
            return
    
        # 
        # WARNING: THIS COULD CHANGE IF ZSI.generate.containers changes.
        #   
        from ZSI.generate.wsdl2dispatch import ServiceModuleWriter
        hack = ServiceModuleWriter()
        try:
            format = {
                'NAME':name,
                'SERVICE':wsdl.services[0].name,
                'LOCATOR':'%sLocator' %wsdl.services[0].name,
                'GETPORTMETHOD':'get%s' %wsdl.services[0].ports[0].getBinding().getPortType().name,
                #'RESOURCE':'%sWSRF' %hack.getClassName(wsdl.services[0].name),
                'RESOURCE':'%sWSRF' %wsdl.services[0].name,
            }
            del hack
        except Exception, ex:
            #print "FAILED TO CREATE "
            #sys.exit(-1)
            raise

        if options.script:
            open(options.script, 'w').write(PYCLIENT %format)
            os.chmod(options.script, 0755)
    
        if options.rpy:
            if not isdir('services'): os.mkdir('services')
            open('services/%s' %options.rpy, 'w').write(RPY %format)
    
        format['TXT'] = sctxt
        format['TAC'] = sctac
        for src in (sctxt, sctac):
            dst = src.rsplit('/')[-1]
            if not isfile(dst) or options.force: shutil.copyfile(src, dst)

        sc = 'start-container.sh'
        if not isfile(sc) or options.force: 
            open(sc, 'w').write(CONTAINER %format)
            os.chmod(sc, 0755)


PYCLIENT = """#!/usr/bin/env python
############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################
from twisted.python import log
from twisted.internet import reactor

import ZSI
from pyGridWare.utility.scripts.client import GetBasicOptParser, GetPortKWArgs, SetUp
from generated.%(NAME)s.stubs import %(NAME)s as CLIENT


def main(**kw):
    locator = CLIENT.%(LOCATOR)s()
    port = locator.%(GETPORTMETHOD)s(**kw)
    
    # Factory METHOD Just guessing here
    #response = port.create(CLIENT.CreateRequest())
    #kw['endPointReference'] = response._EndpointReference
    #iport = locator.%(GETPORTMETHOD)s(**kw)
    reactor.stop()


if __name__ == '__main__':
    op = GetBasicOptParser()
    (options, args) = op.parse_args()
    SetUp(options)
    kw = GetPortKWArgs(options)
    reactor.callWhenRunning(main, **kw)
    reactor.run()
"""

RPY = """############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################
from generated.%(NAME)s.services.%(NAME)s.%(SERVICE)s import %(RESOURCE)s

class Service(%(RESOURCE)s):
    '''override methods to do something useful'''
resource = Service()
"""

CONTAINER = """#! /bin/sh
############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################

EXEC=twistd 
OPTIONS=-noy
CONFIG=server-config.tac

set - ${EXEC} ${OPTIONS} ${CONFIG} "$@"
exec "$@"
"""

README = """############################################################################
# Automatically generated by wsdl2web.py
# See LBNLCopyright for copyright notice!
###########################################################################

Depending on the options provided to "wsdl2web.py", not all the below
files and directories may be present.


FILES: 

    client.py 
      -- Barebones sample client script, sets up a locator and port.

    start-container.sh 
      -- Script for starting the container

    server-config.tac 
      -- Twisted tac file, sets up a few core WSRF services.

    server-config.txt 
      -- Basic server configuration file


DIRECTORIES:

    generated 
      -- Directory with all code generated from the WSDL

    services
      -- Directory to populate with RPY scripts
      -- for "Test.rpy" the path will be "/wsrf/services/Test"


"""

if __name__ == '__main__': 
    main()
