/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.client;

import org.apache.axis.client.Stub;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.Duration;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.globus.axis.util.Util;
import org.globus.workspace.generated.WorkspaceFactoryPortType;
import org.globus.workspace.generated.WorkspaceFactoryServiceAddressingLocator;
import org.globus.workspace.generated.WorkspacePortType;
import org.globus.workspace.generated.WorkspaceServiceAddressingLocator;
import org.globus.workspace.generated.PolicyQuery;
import org.globus.workspace.generated.QueryAccessPolicyResponse;
import org.globus.workspace.generated.types.AccountInformation;
import org.globus.workspace.generated.types.AccountLimitFaultType;
import org.globus.workspace.generated.types.AccountReference;
import org.globus.workspace.generated.types.VoidType;
import org.globus.workspace.impl.factory.WorkspaceFactoryHome;
import org.globus.workspace.impl.service.WorkspaceHome;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.NoAuthorization;
import org.globus.wsrf.utils.AddressingUtils;

import javax.xml.namespace.QName;
import java.rmi.RemoteException;


/**
 * Workspace client.
 */
public class WorkspaceClient {

    static Logger log = Logger.getLogger(WorkspaceClient.class.getName());

    private static String factoryURL = null;
    private static String serviceURL = null;
    private static WorkspaceFactoryPortType factoryPort = null;
    private static WorkspacePortType servicePort = null;

    /**
     * Help
     */
    private static void help() {

        StringBuffer sb = new StringBuffer();
        sb.append("./bin/workspace <options> <action>\n\n");
        sb.append("Options:\n");
        sb.append("  -h <host>      \t\tHostname\n");
        sb.append("  -p <port>      \t\tPort\n");
        sb.append("  -nosec         \t\tDo not use https\n");
        sb.append("  -d             \t\tEnable debuging\n\n");
        sb.append("Actions:\n");
        sb.append("  create <ttl>       \t\tCreate resource with TTL (min)\n");
        sb.append("  username <key>     \t\tGet the workspace username\n");
        sb.append("  access_add_dn <key> <DN>\tAdd a DN to the access " +
                "policy\n");
        sb.append("  access_rm_dn <key> <DN>\tRemove a DN from the access " +
                "policy\n");
        sb.append("  authz_add_dn <key> <DN>\tAdd DN to the workspace authz policy\n");
        sb.append("  authz_rm_dn <key> <DN>\tRemove DN from the workspace authz policy\n");
        sb.append("  query <DN> ...    \t\tQuery for a workspace that the\n");
        sb.append("                    \t\tDN or attribute has access to");
        System.out.println(sb.toString());
        System.exit(-1);
    }

    /**
     * Main
     */
    public static void main(String[] args) {

        if (args.length < 1)
            help();

        String host = "localhost";
        String port = null;
        String protocol = "https";

        // Parse universal flags
        for (int i = 0; i < args.length; i++) {

            if (args[i].equals("-h")) {

                if ((i + 1) >= args.length)
                    help();

                host = args[i + 1];
            }

            if (args[i].equals("-p")) {

                if ((i + 1) >= args.length)
                    help();

                port = args[i + 1];
            }

            if (args[i].equals("-d")) {
                log.setLevel(Level.DEBUG);
                log.debug("Debugging enabled");
            }

            if (args[i].equals("-nosec")) {
                protocol = "http";
            }
        }

        // Setup service URL
        if (port == null) {
            if (protocol.equals("https")) {
                port = "8443";
            } else {
                port = "8080";
            }
        }

        factoryURL = protocol + "://" + host + ":" + port +
            "/wsrf/services/WorkspaceFactoryService";
        serviceURL = protocol + "://" + host + ":" + port +
            "/wsrf/services/WorkspaceService";

        // Parse action
        for (int i = 0; i < args.length; i++) {

            if (args[i].equals("create")) {

                if ((i + 1) >= args.length)
                    createWorkspace(null);
                else
                    createWorkspace(args[i + 1]);
            } else if (args[i].equals("username")) {

                if ((i + 1) >= args.length)
                    help();
                else
                    getUsername(args[i + 1]);
            } else if (args[i].equals("authz_add_dn")) {

                if ((i + 2) >= args.length)
                    help();

                addAuthzIdentity(args[i + 1], args[i + 2]);
            } else if (args[i].equals("authz_rm_dn")) {

                if ((i + 2) >= args.length)
                    help();

                removeAuthzIdentity(args[i + 1], args[i + 2]);
            } else if (args[i].equals("access_add_dn")) {

                if ((i + 2) >= args.length)
                    help();

                addAccessIdentity(args[i + 1], args[i + 2]);
            } else if (args[i].equals("access_rm_dn")) {

                if ((i + 2) >= args.length)
                    help();

                removeAccessIdentity(args[i + 1], args[i + 2]);
            } else if (args[i].equals("query")) {

                if ((i + 1) >= args.length)
                    help();
                else
                    queryAccessPolicies(args[i + 1]);
            }
        }
    }

    /**
     * Query Access Policies.
     */
    private static void queryAccessPolicies(String entry) {

        String[] usernames = null;
        QueryAccessPolicyResponse resp = null;

        if (servicePort == null)

            if (!initServicePort(createEPR(null)))
                System.out.println("Error initializing service port");

        log.debug("Querying for entry: " + entry);
        PolicyQuery query = new PolicyQuery(true, entry, entry);
        try {
            resp = servicePort.queryAccessPolicy(query);
        } catch (RemoteException e) {
            System.out.println("Error querying access policies: "
                                                        + e.getMessage());

        }

        if (resp != null) {
            usernames = resp.getUsername();
            if (usernames != null) {
                if (usernames.length == 0) {
                    System.out.println("No workspaces found");
                } else {
                    for (int i = 0; i < usernames.length; i++) {
                        System.out.println("Username: " + usernames[i]);
                    }
                }
            } else {
                System.out.println("No workspaces found");
            }
        } else {
            System.out.println("No workspaces found");
        }

    }

    /**
     * Add DN to Authz Policy.
     */
    private static void addAuthzIdentity(String key, String dn) {

        if (servicePort == null)

            if (!initServicePort(createEPR(key)))
                System.out.println("Error initializing service port");

        try {
            servicePort.addAuthzIdentity(dn);
        } catch (RemoteException e) {
            System.out.println("Error adding DN to authz policy: "
                                                        + e.getMessage());
            log.debug("",e);
        }
    }

    /**
     * Remove DN from Authz Policy.
     */
    private static void removeAuthzIdentity(String key, String dn) {

        if (servicePort == null) {
            if (!initServicePort(createEPR(key))) {
                System.out.println("Error initializing service port");
                return;
            }
        }

        try {
            servicePort.removeAuthzIdentity(dn);
        } catch (RemoteException e) {
            System.out.println("Error removing DN from authz policy: "
                                                        + e.getMessage());
            log.debug("",e);
        }
    }

    /**
     * Add DN to Access Policy.
     */
    private static void addAccessIdentity(String key, String dn) {

        if (servicePort == null) {
            if (!initServicePort(createEPR(key))) {
                System.out.println("Error initializing service port");
                return;
            }
        }

        try {
            servicePort.addAccessIdentity(dn);
        } catch (RemoteException e) {
            System.out.println("Error adding DN to access policy: "
                                                            + e.getMessage());
            log.debug("",e);
        }
    }

    /**
     * Remove DN from Access Policy.
     */
    private static void removeAccessIdentity(String key, String dn) {

        if (servicePort == null) {
            if (!initServicePort(createEPR(key))) {
                System.out.println("Error initializing service port");
                return;
            }
        }

        try {
            servicePort.removeAccessIdentity(dn);
        } catch (RemoteException e) {
            System.out.println("Error removing DN from access policy: "
                                                            + e.getMessage());
            log.debug("",e);
        }
    }

    /**
     * Get username.
     */
    private static void getUsername(String key) {

        String username = "";

        if (servicePort == null) {
            if (!initServicePort(createEPR(key))) {
                System.out.println("Error initializing service port");
                return;
            }
        }

        try {
            username = servicePort.getLocalAccountName(new VoidType());
        } catch (RemoteException e) {
            System.out.println("Error getting username: " + e.getMessage());
            log.debug("",e);
        }

        System.out.println("Username: " + username);
    }

    /**
     * Create Workspace.
     */
    private static void createWorkspace(String ttl) {

        System.out.println("Creating workspace ...");

        if (factoryPort == null)
            if (!initFactoryPort())
                System.out.println("Error initializing factory port");

        AccountInformation accountInfo = new AccountInformation();

        if (ttl != null)
            accountInfo.setTTL(createDuration(ttl));

        // if we don't supply a policy, workspace service will
        // add a default policy of the caller only

        AccountReference accountRef = null;

        try {
            accountRef = factoryPort.createAccount(accountInfo);
        } catch (AccountLimitFaultType e) {
            System.out.println("Error creating workspace: " +
                "Account limit reached");
            System.exit(0);
        } catch (RemoteException e) {
            System.out.println("Error creating workspace: " + e);
            System.exit(0);
        }

        EndpointReferenceType workspaceEPR = accountRef.getEndpointReference();
        
        System.out.println("Resource key: " + 
            workspaceEPR.getProperties().get(new QName( 
            "http://www.globus.org/03/2005/workspace", "AccountKey"))
            .getValue());
        System.out.println("Termination time: " +
            accountRef.getActualTerminationTime().getTime());
    }

    /**
     * Initialize the factory port.
     */
    private static boolean initFactoryPort() {
        log.debug("Setting up factory locator");

        WorkspaceFactoryServiceAddressingLocator locator = 
            new WorkspaceFactoryServiceAddressingLocator();
        ResourceKey key = WorkspaceFactoryHome.getResourceKey("default");

        try {
            EndpointReferenceType epr =
                AddressingUtils.createEndpointReference(factoryURL, key);

            factoryPort = locator.getWorkspaceFactoryPortTypePort(epr);
            setSecurity((Stub) factoryPort);
        } catch (Exception e) {
            System.out.println("Error creating factory port: "
                                                    + e.getMessage());
            log.debug("",e);
            System.exit(0);
        }

        return true;
    }

    /**
     * Initialize the service port.
     */
    private static boolean initServicePort(EndpointReferenceType epr) {
        log.debug("Setting up service locator");

        if (epr == null) {

            try {
                epr = new EndpointReferenceType();
                epr.setAddress(new Address(serviceURL));
            } catch (Exception e) {
                log.error("Error creating EPR: " + e);
            }
        }

        WorkspaceServiceAddressingLocator serviceLocator =
            new WorkspaceServiceAddressingLocator();

        log.debug("Creating service port");

        try {
            servicePort = serviceLocator.getWorkspacePortTypePort(epr);
        } catch (Exception e) {
            log.error("Error creating service port: ", e);
            log.debug("",e);
            return false;
        }

        setSecurity((Stub) servicePort);

        return true;
    }

    /**
     * Create an EPR from a resource key.
     */
    private static EndpointReferenceType createEPR(String keyString) {
        log.debug("Creating EPR from workspace key: " + keyString);

        EndpointReferenceType epr = null;

        ResourceKey key = WorkspaceHome.getKey(keyString);

        try {
            epr = AddressingUtils.createEndpointReference(serviceURL, key);
        } catch (Exception e) {
            log.debug("Error creating EPR from key: " + keyString);
        }

        return epr;
    }

    /**
     * Convert TTL in minutes to a Duration.
     */
    private static Duration createDuration(String ttl) {
        log.debug("Creating duration from TTL");

        /**
         * false - non-negative duration
         * 0 years
         * 0 months
         * 0 days
         * 0 hours
         * TTL minutes
         * 0 seconds
         */
        return new Duration(false, 0, 0, 0, 0, Integer.parseInt(ttl), 0);
    }

    /**
     * Setup security.
     */
    private static void setSecurity(Stub stub) {
        log.debug("Setting up security on stub");

        stub._setProperty(Constants.GSI_SEC_MSG, Constants.SIGNATURE);
        stub._setProperty(Constants.AUTHORIZATION,
            NoAuthorization.getInstance());

        Util.registerTransport();
    }
}
