/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include <utime.h>
#include <errno.h>
#include <ctype.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pwd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <malloc.h>

#include "gridmapdir.h"
#include "util.h"
#include "log.h"
#include "errors.h"
#include "lcmaps_return_poolindex_without_gsi.h"

ino_t g_otherinode = -1;
char *g_originalname = NULL;

/* __________________________________________________
   inode_filter
   
   used with scandir, see otherlink()
   returns 1 if matches our criteria
   _______________________________________________{{{ */
int inode_filter(const struct dirent *direntry) {
  char *func = "inode_filter()";
  /* debug_msg(8,"%s: comparing file: %s",func,direntry->d_name); */
  if (direntry->d_ino == g_otherinode) {
    debug_msg(8,"%s: found matching inode for %d",func,direntry->d_ino);
    if (strcmp(direntry->d_name, g_originalname) != 0) {
      return 1;
    } else {
      debug_msg(8,"%s: matching inode is original file",func);
      return 0;
    }
  } else {
    return 0;
  }
} /* inode_filter() }}} */
  

/* __________________________________________________
   register_inode_and_name
   
   used with inode_filter() and scandir(), see otherlink()
   _______________________________________________{{{ */
int register_inode_and_name(ino_t otherinode, char *original) {
  char *func = "register_inode_and_name()";
  g_otherinode = otherinode;
  g_originalname = original;
  debug_msg(8,"%s: g_otherinode set to %d",func,g_otherinode);
  return 0;
} /* register_inode_and_name() }}} */

/* __________________________________________________
   unregister_inode
   
   used with inode_filter() and scandir(), see otherlink()
   _______________________________________________{{{ */
int unregister_inode_and_name() {
  g_otherinode = -1;
  g_originalname = NULL; /* this was not malloc'd */
  return 0;
} /* unregister_inode() }}} */

/* __________________________________________________
   otherlink
   
   finds a link with matching inode or returns NULL
   string returned needs to be freed
   _______________________________________________{{{ */
char *otherlink(char *entry, char *path) {
  char *func = "otherlink()";
  char *fullentry = NULL;
  char *corresponding = NULL;
  struct stat statbuf;
  int ret = 0;
  struct dirent **foundfiles;
  int numfoundfiles = 0;
  
  debug_msg(8,"%s",func);
  debug_msg(8,"entry: %s",entry);
  debug_msg(8,"path: %s",path);
      
  fullentry = (char *) malloc(strlen(path) + 2 + strlen(entry));
  sprintf(fullentry, "%s/%s", path, entry);
  debug_msg(8,"fullentry: %s",fullentry);
  ret = stat(fullentry, &statbuf);
  if (ret != 0) {
    debug_msg(1,"%s: given entry does not exist: %s",func,fullentry);
    goto fail;
  }
  
  /* if it does exist, does it have more than one link to it? */
  
  /* we use st_nlink here instead of just looking at numfoundfiles below,
     just to be sure there are no links in another dir (as numfoundfiles
     only reflects number of files w/ same inode in the given path) */ 
  if (statbuf.st_nlink == 2) { 
    debug_msg(5,"%s: found another (and only one other) link for given entry: %s",func,fullentry);
  } else if (statbuf.st_nlink > 2) { 
    debug_msg(1,"%s: given entry has more than one other file linked to it, which is invalid: %s",func,fullentry);
    goto fail;
  } else {
    debug_msg(1,"%s: given entry has no other file linked to it: %s",func,fullentry);
    goto fail;
  }
  
  /* 2 links in all, find the other */
  register_inode_and_name(statbuf.st_ino, entry);
  numfoundfiles = scandir(path, &foundfiles, inode_filter, alphasort);
  unregister_inode_and_name();
  debug_msg(5,"%s: found %d files",func,numfoundfiles);
  
  if (numfoundfiles != 1) {
    debug_msg(1,"%s: found %d files, but statbuf.st_nlink == 2, so there must be a link to this file but NOT in gridmapdir",func,numfoundfiles);
  }
  
  debug_msg(2,"%s: other file is %s",func,foundfiles[0]->d_name);
  corresponding = safe_dup(foundfiles[0]->d_name);
  goto free;
  fail:
    if (corresponding) free(corresponding);
    corresponding = NULL;
  free:
    if (fullentry) free(fullentry);
    { 
      int i = 1;
      if (foundfiles) { 
        for (; i < numfoundfiles + 1; i++) {
          if (foundfiles[i-1]) free(foundfiles[i-1]);
        }
      }
    }
    return corresponding;
  
} /* otherlink() }}} */


/* __________________________________________________
   gridmapdir_unlink_index
   
   deletes association 
   _______________________________________________{{{ */
int gridmapdir_unlink_index(char *encodedglobusidp, uid_t *uid, unsigned short quarantine_flag)
{
  char *func = "gridmapdir_unlink_index()";
  char *gridmapdir = NULL;
  char *quarantinepath = NULL; 
  char *leasedname = NULL;
  char *leasepath = NULL;
  char *quarantinesuffix = QUARANTINE_SUFFIX;
  char *encodedglobusidpath = NULL;
  struct stat statbuf;
  int ret;
  struct passwd *pwd;
     
  debug_msg(8,"%s",func);
     
  if (encodedglobusidp == NULL) {
    debug_msg(1,"%s problem, encodedglobusidp is null",func);
    ret = BADARG;
    goto free;
  } 
  
  if (uid == NULL) {
    debug_msg(1,"%s problem, uid * is null",func);
    ret = GENERAL_FAILURE;
    goto free;
  }
  
  gridmapdir = getenv("MAPDIR");
  if (gridmapdir == NULL) { 
    debug_msg(1,"%s: no MAPDIR",func);
    goto fail;
  }
  
  leasedname = otherlink(encodedglobusidp, gridmapdir);
  if (!leasedname) {
    debug_msg(3,"%s can't delete: no association for: %s",func,encodedglobusidp);
    fprintf(stderr,"can't delete: no association for: %s\n",encodedglobusidp);
    ret = GRIDMAPDIR_NO_ASSOCIATION;
    goto free;
  }
    
  leasepath = (char *) malloc(strlen(gridmapdir) + 2 + strlen(leasedname));
  if (!leasepath) {
    debug_msg(1,"%s problem, malloc",func);
    goto fail;
  }
  sprintf(leasepath, "%s/%s", gridmapdir, leasedname);
  debug_msg(8,"%s: account = %s",func,leasepath);
  
  /* first link account file to a quarantine file */
  if (quarantine_flag & QUARANTINE) {
    debug_msg(3,"%s: attempting to quarantine before unlinking",func);
    
    if (g_QUARANTINE_BY_POOLINDEX == 1) {
      quarantinepath = malloc(strlen(gridmapdir) + strlen(encodedglobusidp) + strlen(leasedname) + 4 + strlen(quarantinesuffix));
    } else {
      quarantinepath = malloc(strlen(leasepath) + 2 + strlen(quarantinesuffix));
    }
       
    if (!quarantinepath) {
      debug_msg(1,"%s problem, malloc",func);
      goto fail;
    }
      
    if (g_QUARANTINE_BY_POOLINDEX == 1) {
      sprintf(quarantinepath, "%s/%s.%s%s", gridmapdir, encodedglobusidp, leasedname, quarantinesuffix);
    } else {
      sprintf(quarantinepath, "%s%s", leasepath, quarantinesuffix);
    }
    debug_msg(8,"%s: quarantine to %s",func,quarantinepath);
    
    ret = stat(quarantinepath, &statbuf);
    if (ret == 0) {
      debug_msg(1,"%s: cannot quarantine, file already exists: %s",func,quarantinepath);
      /* TODO: here, override inconsistency (touch file) or just continue */
      goto fail;
    }
    
    ret = link(leasepath,quarantinepath);
    if (ret) {
      debug_msg(1,"%s: cannot link %s and %s", func, leasepath, quarantinepath);
      goto fail;
    }
  }
  
  encodedglobusidpath = (char *) malloc(strlen(gridmapdir) + 2 + strlen(encodedglobusidp));
  if (!encodedglobusidpath) {
    debug_msg(1,"%s problem, malloc",func);
    goto fail;
  }
  sprintf(encodedglobusidpath, "%s/%s", gridmapdir, encodedglobusidp);
  
  debug_msg(8,"%s: encodedglobusidpath = %s",func,encodedglobusidpath);
  
  ret = unlink(encodedglobusidpath);
  if (ret) { 
    /* todo, parse errno, this is an odd inconsistency */
    debug_msg(1,"%s: cannot unlink %s", func, encodedglobusidpath);
    goto fail;
  }
  
  if ((pwd = getpwnam(leasedname)) == NULL)
  {
    debug_msg(1,"%s: very unlikely.. username %s not found?", func, leasedname);
    goto fail;
  }

  *uid = pwd->pw_uid;
  endpwent();
  
  debug_msg(8,"%s: unlinked successfully, found uid = %d", func, *uid);
  
  ret = 0;
  goto free;
  fail:
    ret = 1;
  free:
    if (leasedname) free(leasedname);
    if (leasepath) free(leasepath);
    if (quarantinepath) free(quarantinepath);
    if (encodedglobusidpath) free(encodedglobusidpath);
    return ret;
} /* gridmapdir_unlink_dn() }}} */


/* __________________________________________________
   call LCMAPS
   _______________________________________________{{{ */
int lcmaps_poolindex(char *dn, char *fqan_list[], int nfqan, int mapcount, lcmaps_account_info_t *lcmaps_account, char **gridmapdir)
{
  int ret;
  char * lcmaps_suffix = g_LCMAPS_LOG_STRING;
  char * lcmaps_log_string = NULL;
  char * time_prefix = get_time();
  
  if (time_prefix == NULL) {
    return GENERAL_FAILURE;
  }
  
  lcmaps_log_string =  (char *) malloc ((strlen(lcmaps_suffix) + strlen(time_prefix) + strlen(" : ")) * sizeof(char) + 1);
  if (lcmaps_log_string == NULL) {
    return GENERAL_FAILURE;
  }
  
  strcpy(lcmaps_log_string, time_prefix);
  strcat(lcmaps_log_string, " : ");
  strcat(lcmaps_log_string, lcmaps_suffix);
  
  setenv("LCMAPS_LOG_FILE", g_LCMAPS_LOG_FILE , 1);
//    setenv("LCMAPS_LOG_FILE", "/dev/stderr", 1);
  setenv("LCMAPS_DB_FILE", g_LCMAPS_DB_FILE , 1);
  setenv("LCMAPS_DEBUG_LEVEL", g_LCMAPS_DEBUG_LEVEL , 1);
  setenv("LCMAPS_POLICY_NAME", g_LCMAPS_POLICY_NAME , 1);
  setenv("LCMAPS_LOG_STRING", lcmaps_log_string , 1);
  
  if (lcmaps_log_string) free (lcmaps_log_string);
  if (time_prefix) free (time_prefix);

  ret = lcmaps_account_info_init(lcmaps_account);
  debug_msg(5, "Result of lcmaps_account_info_init(): %d", ret);
  
  /* not that this is completely accurate, but if 
     lcmaps_account_info_init() fails it will be POOL_INDEX_FAILURE. 
     If it succeeds but lcmaps_return_poolindex_without_gsi fails, return
     GRIDMAPDIR_NO_ASSOCIATION
   */
  if (ret) {
    return POOL_INDEX_FAILURE;
  }

  if (mapcount > -1) {
    ret = lcmaps_return_poolindex_with_mapcounter(
        dn,
        fqan_list,
        nfqan,
        mapcount,
        lcmaps_account
    );
    debug_msg(5, "Result of lcmaps_return_poolindex_with_mapcounter(): %d", ret);
  } else {
    ret = lcmaps_return_poolindex_without_gsi(
        dn,
        fqan_list,
        nfqan,
        lcmaps_account
    );
    debug_msg(5, "Result of lcmaps_return_poolindex_without_gsi(): %d", ret);
  }

  if (ret == 0)
  {
    int i = 0;
    debug_msg(5, "lcmaps_account.poolindex: %s", lcmaps_account->poolindex);
    debug_msg(5, "lcmaps_account.uid: %d", (int) lcmaps_account->uid);
    debug_msg(5, "lcmaps_account.npgid: %d", (int) lcmaps_account->npgid);
    for (i = 0; i < lcmaps_account->npgid; i++)
        debug_msg(5, "lcmaps_account.pgid_list[%d]: %d", i, lcmaps_account->pgid_list[i]);
    debug_msg(5, "lcmaps_account.nsgid: %d", (int) lcmaps_account->nsgid);
    for (i = 0; i < lcmaps_account->nsgid; i++)
        debug_msg(5, "lcmaps_account.sgid_list[%d]: %d", i, lcmaps_account->sgid_list[i]);
    
    *gridmapdir = getenv("MAPDIR");
    if (*gridmapdir == NULL) {
      fprintf(stderr, "gridmapdir not set\n");
      log_msg("gridmapdir not set by LCMAPS\n");
      return ENVIRONMENT;
    }
    
    return SUCCESS;
  } else {

    ret = lcmaps_account_info_clean(lcmaps_account);
    fprintf(stderr, "Result of lcmaps_account_info_clean(): %d\n", ret);
    return GRIDMAPDIR_NO_ASSOCIATION;
  }
    
} /* lcmaps_poolindex() }}} */


