/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/fsuid.h>
#include <errno.h>

#include "util.h"
#include "errors.h"
#include "log.h"

/* local procedures */
int globus_l_common_path_fgets( char*  buf, 
                                int    bufsize, 
                                FILE*  fp );
void globus_l_common_path_fgets_init();
int get_attribute_from_config( char *  filename,
					                     char *  attribute,
					                     char ** value);

/* __________________________________________________
   safe_dup
   _______________________________________________{{{ */
char * safe_dup(char * arg)
{
  char *dupped_arg;

  dupped_arg = (char *) malloc(SAFE_LEN * sizeof(char));
  if (dupped_arg == NULL)
  {
    debug_msg(1, "Error: malloc for dup of argument string failed");
    exit(BADARG);
  }
    
  if (!arg) { 
      exit(BADARG);
  }

  strncpy(dupped_arg, arg, SAFE_LEN);
  dupped_arg[SAFE_LEN - 1] = '\0';
  if (strlen(dupped_arg) == SAFE_LEN - 1)
  {
    debug_msg(1, "Error: argument string length too long");
    exit(BADARG);
  } 
  return dupped_arg;
} /* safe_dup() }}} */



/* __________________________________________________
   parse_config_file
   
   Given the program's name, find and load config
   information.  Results are placed in a set of
   global variables.  The configuration file parsing
   is fully encapsulated here.
   
   Current implementation based on globus_common, 
   see below: get_attribute_from_config().

   Returns 0 or error code.
   _______________________________________________{{{ */
int parse_config_file( char * executable,
                       char * given_confpath, 
                        uid_t real_uid, 
                        uid_t euid)
{
  int ret = -1;
  struct stat statbuf;
  struct stat statbuf2;
  uid_t run_uid;
  char * confpath = NULL;
  char * execpath = NULL;
  gid_t gid;
  struct group *groupstruct = NULL; /* for debug output */
  
  /* global configuration variables are declared in util.h */
  /* for config values that need to be converted before going global */
  char *DO_QUARANTINE = NULL;
  char *QUARANTINE_BY_POOLINDEX = NULL;
  char *RUN_EXPIRY_SETUID = NULL;
  char *DEBUG_LEVEL = NULL;
  
  if (given_confpath == NULL) { 
    fprintf(stderr, "null arg to parse_config_file");
    ret = GENERAL_FAILURE;
    goto exit_free;
  }
  
  confpath = safe_dup(given_confpath);
  execpath = safe_dup(executable);
  
  /* get priviliged group */
  ret = stat(execpath, &statbuf2);
    
  if (ret != 0) {
    fprintf(stderr, "Failing: cannot stat executable(?): %s", execpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  }
  
  gid = statbuf2.st_gid;
  
  
  /* use absolute or relative paths */
  ret = stat(confpath, &statbuf);
     
  if (ret != 0) {
    fprintf(stderr, "Failing: configuration file not found: %s", confpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  }
  
  /* Is configuration file regular? */
  if (!(statbuf.st_mode & S_IFREG)) {
    fprintf(stderr, "Configuration file (%s) is not a regular file", confpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  }
  
  /* * removed:
  Is configuration file not writable by anyone except owner? 
  if (statbuf.st_mode & S_IWGRP) {
    fprintf(stderr, "Failing: configuration file (%s) is writable by group\n", confpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  } 
  * */
  
  /* Is configuration file owned by the priviliged group? 
     Because we were requested to allow group write access to the 
     configuration file and because if setuid, group membership should
     be used to limit callers who can elevate, it is considered a 
     priviliged group.  */
  if (statbuf.st_gid != gid) {
    groupstruct = getgrgid(gid);
    if (groupstruct != NULL) {
      fprintf(stderr, "Failing: configuration file (%s) group is not %d (%s)\n", confpath,gid,groupstruct->gr_name);
    } else {
      fprintf(stderr, "Failing: configuration file (%s) group is not %d\n", confpath,gid);
    }
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  } 
  
  if (statbuf.st_mode & S_IWOTH) {
    fprintf(stderr, "Failing: configuration file (%s) is writable by world\n", confpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  }
  
  /* Is configuration file suspiciously big? */
  if (statbuf.st_size > 65536) {
    fprintf(stderr, "Failing: configuration file (%s) is over 64k", confpath);
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  }
  
  /* ***************************** */
  
  /* DEBUG */
  
  /* default */
  g_DEBUG_LEVEL = 0;
  
  DEBUG_LEVEL = NULL;
  
  if (get_attribute_from_config(confpath, "debug level", &DEBUG_LEVEL) == 0) {
    /* strtol returns 0 if no match */
    if (strcmp(DEBUG_LEVEL, "0") == 0) {
      ; /* default */
    } else {
      g_DEBUG_LEVEL = strtol(DEBUG_LEVEL, (char **)NULL, 10);
      if (errno != 0) {
        debug_msg(0,"error: %s", strerror(errno));
        ret = GENERAL_FAILURE;
        goto exit_free;
      } else if (g_DEBUG_LEVEL == 0) {
        debug_msg(0,"string to integer failure for: %s",DEBUG_LEVEL);
        ret = ENVIRONMENT_CONFIG_FILE;
        goto exit_free;
      } else if (g_DEBUG_LEVEL < 0) { 
        debug_msg(0,"no positive integer found for: %s",DEBUG_LEVEL);
        ret = ENVIRONMENT_CONFIG_FILE;
        goto exit_free;
      }
    }
  }
  
  if (debug_init()) {
    fprintf(stderr, "cannot initialize debugging\n");
    ret = ENVIRONMENT;
    goto exit_free;
  }
  
  debug_msg(3, "config:  %s is a regular file.",confpath);
  debug_msg(1, "config:  confpath = %s",confpath);
  
  
  g_LCMAPS_DB_FILE = NULL;
  g_LCMAPS_DEBUG_LEVEL = NULL;
  g_LCMAPS_GRIDMAPDIR = NULL;
  g_LCMAPS_LOG_FILE = NULL;
  g_LCMAPS_LOG_STRING = NULL;
  g_LCMAPS_POLICY_NAME = NULL;
  g_EXPIRY_PROG = NULL;
  g_POOL_LOGFILE = NULL;
  g_POOL_LOGSEP = NULL;
  
  /* defaults; not strings */
  g_DO_QUARANTINE = 0;
  g_QUARANTINE_BY_POOLINDEX = 0;
  g_RUN_EXPIRY_SETUID = -1;
  
  /* brute force through all key/value pairs */
    
  /* ***************************** */
  
  debug_msg(1,"config:  ------");
  debug_msg(1,"         LCMAPS");
  debug_msg(1,"         ------");
  
  if (get_attribute_from_config(confpath, "LCMAPS_DB_FILE", &g_LCMAPS_DB_FILE) == 0) {
    debug_msg(1,"         LCMAPS_DB_FILE = %s", g_LCMAPS_DB_FILE);
  }
  
  if (get_attribute_from_config(confpath, "LCMAPS_DEBUG_LEVEL", &g_LCMAPS_DEBUG_LEVEL) == 0) {
    
    debug_msg(1,"         LCMAPS_DEBUG_LEVEL = %s",   g_LCMAPS_DEBUG_LEVEL);
    
  }
  
  if (get_attribute_from_config(confpath, "LCMAPS_GRIDMAPDIR", &g_LCMAPS_GRIDMAPDIR) == 0) {
    
    debug_msg(1,"         LCMAPS_GRIDMAPDIR = %s",   g_LCMAPS_GRIDMAPDIR);
    
  }
  
  if (get_attribute_from_config(confpath, "LCMAPS_LOG_FILE", &g_LCMAPS_LOG_FILE) == 0) {
    
    debug_msg(1,"         LCMAPS_LOG_FILE = %s",   g_LCMAPS_LOG_FILE);
    
  }
  
  if (get_attribute_from_config(confpath, "LCMAPS_LOG_STRING", &g_LCMAPS_LOG_STRING) == 0) {
    
    debug_msg(1,"         LCMAPS_LOG_STRING = %s",   g_LCMAPS_LOG_STRING);
    
  }
  
  if (get_attribute_from_config(confpath, "LCMAPS_POLICY_NAME", &g_LCMAPS_POLICY_NAME) == 0) {
    
    debug_msg(1,"         LCMAPS_POLICY_NAME = %s",   g_LCMAPS_POLICY_NAME);
    
  }
  
  /* ***************************** */
  
  debug_msg(1,"config:  ------");
  debug_msg(1,"         EXPIRY");
  debug_msg(1,"         ------");
  
  if (get_attribute_from_config(confpath, "expiry program", &g_EXPIRY_PROG) == 0) {
    debug_msg(1,"         using expiry program = %s", g_EXPIRY_PROG);
  }
    
  if (get_attribute_from_config(confpath, "run expiry program setuid", &RUN_EXPIRY_SETUID) == 0) {
    
    debug_msg(5,"         found run expiry program setuid = %s",   RUN_EXPIRY_SETUID);
    
  }
  
  if (RUN_EXPIRY_SETUID != NULL) {
    if (g_EXPIRY_PROG == NULL) {
      debug_msg(0,"        Running expiry program setuid is defined but path to expiry program is not.  Ignoring.");
    } else {
      /* strtol returns 0 if no match */
      if (strcmp(RUN_EXPIRY_SETUID, "0") == 0) {
        g_RUN_EXPIRY_SETUID = 0;
      } else {
        g_RUN_EXPIRY_SETUID = strtol(RUN_EXPIRY_SETUID, (char **)NULL, 10);
        if (errno != 0) {
          debug_msg(0,"error: %s", strerror(errno));
          ret = GENERAL_FAILURE;
          goto exit_free;
        } else if (g_RUN_EXPIRY_SETUID == 0) {
          debug_msg(0,"string to integer failure for: %s",RUN_EXPIRY_SETUID);
          ret = ENVIRONMENT_CONFIG_FILE;
          goto exit_free;
        } else if (g_RUN_EXPIRY_SETUID < 0) { 
          debug_msg(0,"no positive integer found for: %s",RUN_EXPIRY_SETUID);
          ret = ENVIRONMENT_CONFIG_FILE;
          goto exit_free;
        }
      }
        
      debug_msg(1,"         running expiry program with UID = %d", g_RUN_EXPIRY_SETUID);
    }
  }
  
  if (g_RUN_EXPIRY_SETUID >= 0) {
      run_uid = g_RUN_EXPIRY_SETUID;
  } else {
      run_uid = real_uid;
  }
  
  /* UID check in lcmaps_poolindex.c -> main() */
  
  /* ***************************** */
  
  debug_msg(1,"config:  ----------");
  debug_msg(1,"         QUARANTINE");
  debug_msg(1,"         ----------");
  
  if (get_attribute_from_config(confpath, "do quarantine", &DO_QUARANTINE) == 0) {
    debug_msg(5,"         found do quarantine = %s", DO_QUARANTINE);
    
    if (strcmp(DO_QUARANTINE, "true") == 0) {
      g_DO_QUARANTINE = 1;
    }
    
    if (g_DO_QUARANTINE == 1) {
      debug_msg(1,"         enabling quarantine");
    } else if (g_DO_QUARANTINE == 0) {
      debug_msg(1,"         quarantine is disabled");
    } else {
      debug_msg(0,"Failing, do quarantine can not be value other than 1 or 0");
      ret = GENERAL_FAILURE;
      goto exit_free;
    }
  }
  
  if (get_attribute_from_config(confpath, "quarantine by poolindex", &QUARANTINE_BY_POOLINDEX) == 0) {
    
    debug_msg(5,"         found quarantine by poolindex = %s",   QUARANTINE_BY_POOLINDEX);
        
    if (strcmp(QUARANTINE_BY_POOLINDEX, "true") == 0) {
      g_QUARANTINE_BY_POOLINDEX = 1;
    }
    
    if (g_QUARANTINE_BY_POOLINDEX == 1) {
      debug_msg(1,"         enabling quarantine by poolindex");
    } else if (g_QUARANTINE_BY_POOLINDEX == 0) {
      debug_msg(1,"         enabling quarantine by acctname");
    } else {
      debug_msg(0,"Failing, quarantine by poolindex can not be value other than 1 or 0");
      ret = GENERAL_FAILURE;
      goto exit_free;
    }
    
  }
    
  /* ***************************** */
  
  debug_msg(1,"config:  -------");
  debug_msg(1,"         LOGGING");
  debug_msg(1,"         -------");
  
  if (get_attribute_from_config(confpath, "pool logfile", &g_POOL_LOGFILE) == 0) {
    debug_msg(1,"         using logfile = %s", g_POOL_LOGFILE);
  }
    
  if (get_attribute_from_config(confpath, "log separator", &g_POOL_LOGSEP) == 0) {
    debug_msg(1,"         using logfile separtor = %s", g_POOL_LOGSEP);
  }
  
  debug_msg(1,"         debug level = %d", g_debug_level);
  
  free(confpath);
  confpath = NULL;
  
  
  /* ***************************** */
  
  /* required options */
  
  ret = 0;
  if (g_LCMAPS_DB_FILE == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_DB_FILE     \n");
  }
  if (g_LCMAPS_DEBUG_LEVEL == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_DEBUG_LEVEL \n");
  }
  if (g_LCMAPS_GRIDMAPDIR == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_GRIDMAPDIR  \n");
  }
  if (g_LCMAPS_LOG_FILE == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_LOG_FILE    \n");
  }
  if (g_LCMAPS_LOG_STRING == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_LOG_STRING  \n");
  }
  if (g_LCMAPS_POLICY_NAME == NULL) {
    ret = 1;
    fprintf(stderr,"Required option is missing: LCMAPS_POLICY_NAME \n");
  }
  
  if (ret) {
    fprintf(stderr,"Failing: one ore more required options missing\n");
    ret = ENVIRONMENT_CONFIG_FILE;
    goto exit_free;
  } else {
    ret = SUCCESS;
    goto exit_free;
  }
  
  exit_free:
    if (DO_QUARANTINE) free (DO_QUARANTINE);
    if (QUARANTINE_BY_POOLINDEX) free (QUARANTINE_BY_POOLINDEX);
    if (RUN_EXPIRY_SETUID) free (RUN_EXPIRY_SETUID);
    if (DEBUG_LEVEL) free (DEBUG_LEVEL);
    if (confpath) free (confpath);
    return ret;
    /* leaves statbuf* members allocated */
  
} /* parse_config_file() }}} */





/* __________________________________________________
   get_attribute_from_config
   
   Given the configuration file's path, find and load
   one attribute.
      
   Not too efficient implementation, based on globus_common, 
   globus_common_get_attribute_from_config_file() -- see:  
   http://viewcvs.globus.org/viewcvs.cgi/common/source/library/globus_common_paths.c?rev=HEAD&content-type=text/vnd.viewcvs-markup
   
   Memory is allocated for value in this procedure.
   
   Returns 0 or error code.
   _______________________________________________{{{ */

int get_attribute_from_config( char *  filename,
					                     char *  attribute,
					                     char ** value)
{
  FILE *            fp;
  char *            p;
  char *            q;
  char              attr[200];    /* assumes a name won't be longer... */
  char              buf[2000];    /* assumes a line won't be longer... */
  int               attr_len;
  int               status;
  
  if (filename == NULL) {
    fprintf(stderr, "null arg to get_attribute_from_config");
    return GENERAL_FAILURE;
  }
  
      
  fp = fopen(filename,"r");
    
  if (!fp) {
    fprintf(stderr, "failed to open %s", filename);
    return ENVIRONMENT_CONFIG_FILE;
  }
  
      
  globus_l_common_path_fgets_init();
  p = NULL;
  
  sprintf(attr, "%s = ", attribute);
  attr_len = strlen(attr);
  
  while(!p && (status=globus_l_common_path_fgets(buf,sizeof(buf),fp))) {
    /* any white space? */
    q = buf;
    while (*q==' ' || *q=='\t') {
      q++;
    }
  
    if (strncmp(q, attr, attr_len) == 0) {
      p = q + attr_len;
    }
    
  }
  
  fclose(fp);
  
  if (p) {
    /* any enclosing quotes or trailing white space? */
    if (*p == '"') {
      ++p;
    }

    q = p + strlen(p) - 1;  /* last char before \0 */
    
    /* strip trailing whitespace */
    while (q > p && (*q==' ' || *q=='\t' || *q=='\n' || *q=='"')) {
        *q = '\0';
        --q;
    }
    
    /* strip preceding whitespace */
    /* not in globus_common_get_attribute_from_config_file() */
    while (p < q && (*p==' ' || *p=='\t')) {
      ++p;
    }
  }
  
  if (!p || strlen(p)==0) {
    //debug_msg(2, "could not resolve %s from config file", attribute);
    return -1; //local error
  }

  *value = safe_dup(p);
  if (! *value) {
    fprintf(stderr, "malloc error");
    return GENERAL_FAILURE;
  }
  
  return SUCCESS;

} /* get_attribute_from_config() }}} */


/* __________________________________________________
   help function
   
   fgets() usually doesn't return the last line correctly if there's no 
   trailing \n. This function returns 1 when information was read and 0 when
   there is no more info left.
   _______________________________________________ */

static int   globus_l_common_path_fgets_c = 0;

void globus_l_common_path_fgets_init()
{
    globus_l_common_path_fgets_c = 0;
}

int globus_l_common_path_fgets( char* buf, int bufsize, FILE* fp )
{
    int          c;
    int          n;
    
    c = globus_l_common_path_fgets_c;
    if (c==EOF)
      return 0;
    c=0;
    n=0;
    while (n<bufsize && EOF!=(c=fgetc(fp)) && c!='\n')
    {
      buf[n++] = c;
    }

    buf[n] = '\0';
    globus_l_common_path_fgets_c = c;
    
    return 1;
}



