/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace;

import org.apache.axis.types.Duration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.workspace.generated.types.AccessPolicy;
import org.globus.workspace.generated.types.AccountFaultType;
import org.globus.workspace.generated.types.AccountLimitFaultType;
import org.globus.wsrf.utils.FaultHelper;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.Vector;

public class WorkspaceUtil {

    static Log logger =
        LogFactory.getLog(WorkspaceUtil.class.getName());

    /**
     * Transforms the policy passed by the client at create
     * into a Vector.  This is one of the only places in code
     * to consider if AccessPolicy is ever modelled differently
     * in the XML schema, since it is just a policy that "hangs"
     * on the account awaiting to be queried.
     * @param policy The AccessPolicy object that the client
     *               may pass w/ workspace request.
     * @return
     */
    public static Vector getPolicyVector(AccessPolicy policy) {
        if (policy == null)
            return null;
        
        Vector policyInfo = new Vector();
        String[] arrayOfPolicy = policy.getPolicy();
        if (arrayOfPolicy == null)
            return null;
        
        for (int i=0; i<arrayOfPolicy.length; i++) {
            policyInfo.add(arrayOfPolicy[i]);
        }
        return policyInfo;
    }

    /**
     * Helps with generic faults and chained exceptions.
     * @param errMsg
     * @param exp
     * @return
     */
    public static AccountFaultType makeFault(String errMsg, Exception exp) {
        
        AccountFaultType fault = new AccountFaultType();
        FaultHelper helper = new FaultHelper(fault);
        if (errMsg != null) {
            helper.setDescription(errMsg);
        }
        if (exp != null) {
            helper.addFaultCause(exp);
        }

        return (AccountFaultType)helper.getFault();
    }


    /**
     * Helps with Limit fault and chained exceptions.
     * @param errMsg
     * @param exp
     * @return
     */
    public static AccountLimitFaultType makeLimitFault(String errMsg,
                                                       Exception exp) {

        AccountLimitFaultType fault = new AccountLimitFaultType();
        FaultHelper helper = new FaultHelper(fault);
        if (errMsg != null) {
            helper.setDescription(errMsg);
        }
        if (exp != null) {
            helper.addFaultCause(exp);
        }

        return (AccountLimitFaultType)helper.getFault();
    }

    public static Duration minutesToDuration(int minutes) {
        /**
         * false - non-negative duration
         * 0 years
         * 0 months
         * 0 days
         * 0 hours
         * # minutes
         * 0 seconds
         */
        return new Duration(false, 0, 0, 0, 0, minutes, 0);
    }

    /**
     * Converts duration into seconds.  The current Duration class
     * implementation does not support proper <, >, or even equals
     * operations.  Choosing to convert all Durations off to seconds
     * for comparison rather than extend and fix the Duration class
     * implementation.
     *
     * Max duration is max(int) seconds. This is around ~25k days.
     *
     * Assumes years and month are ZERO and negative is FALSE.  Years
     * and months cannot be compared reliably because they can only be
     * partially ordered.
     *
     * TODO: possibly incorporate this restriction into the service schema?
     *
     * TODO: what to do if these ints are made to overflow?
     *
     * Uses the
     * {@link java.lang.Double#doubleToLongBits(double) doubleToLongBits}
     * method.
     *
     * @param dur Positive duration instance w/o years or months.
     * @return Number of seconds in the Duration instance.
     */
    public static int durationToSeconds(Duration dur)
            throws InvalidDurationException {
        int days;
        int hours;
        int mins;
        double secs;
        Long longsecs;
        int intsecs;

        if (dur.isNegative()) {
            throw new InvalidDurationException("Duration can not be negative.");
        }

        if (dur.getYears() != 0) {
            throw new InvalidDurationException("Years in duration must be " +
                    "zero.");
        }

        if (dur.getMonths() != 0) {
            throw new InvalidDurationException("Months in duration must be " +
                    "zero.");
        }
        // Convert each to next lowest, until we get to seconds.
        days = dur.getDays();
        hours = dur.getHours() + days * 24;
        mins = dur.getMinutes() + hours * 60;
        secs = dur.getSeconds() + mins * 60;

        // Convert to int

        longsecs = new Long(new Double(secs).longValue());
        intsecs = longsecs.intValue();
        if (intsecs == Integer.MAX_VALUE) {
            throw new InvalidDurationException("Duration can not be " +
                    "longer than ~25k days.");
        }
        return intsecs;
    }

    /**
     * A generalized method to run a command. Assumes all environmant 
     * variables are set so the command can be executed.
     *
     * @param command
     *        Command to run.
     * @param baseErrMesg
     *        Error message that is thrown with the exception
     * @return String 
     *         The output of the process on to standard out.
     */
    public static String runCommand(String[] command, String baseErrMesg)
        throws WorkspaceException {

        if (command == null) {
            logger.debug("Command cannot be null");
            throw new WorkspaceException("Command cannot be null");
        }

        for (int i=0; i<command.length; i++) {
            logger.debug("Command is ::: " + command[i]);
        }

        Runtime runtime = Runtime.getRuntime();
        StringBuffer outBuffer = null;
        BufferedReader out = null;
        BufferedReader err = null;
        Process process = null;
        try {
            process = runtime.exec(command);

            int returnCode = -1;
            try {
                returnCode = process.waitFor();
            } catch (InterruptedException exp) {
                logger.error("Interupped exp thrown ", exp);
            }
            logger.debug("Return code is " + returnCode);

            // Get process output on stdout
            out = new BufferedReader(new InputStreamReader
                (process.getInputStream()));
            if (out != null) {
                String line = out.readLine(); 
                if (line != null) {
                    outBuffer = new StringBuffer(line);
                    while (line != null) {
                        line = out.readLine();
                        if (line != null) {
                            outBuffer.append("\n" + line);
                        }
                    }
                }
            }   

            // If return code is not zero, try to get error stream also
            if (returnCode != 0) {
                err = new BufferedReader(new InputStreamReader
                    (process.getErrorStream()));
                StringBuffer errBuffer = null;
                if (err != null) {
                    String line = err.readLine(); 
                    if (line != null) {
                        errBuffer = new StringBuffer("\n" + line);
                        while (line != null) {
                            line = err.readLine();
                            if (line != null) {
                                errBuffer.append("\n" + line);
                            }
                        }
                    }
                } 
                String errString = baseErrMesg;
                if (errBuffer != null) {
                     errString = errString + errBuffer.toString();
                     logger.error("ErrStr " + errString);
                }
                if (outBuffer != null) {
                    errString = errString + outBuffer.toString();
                    logger.error("Out " + errString);
                }
                logger.error(errString);
                throw new WorkspaceException(command + "\n" + errString);
            }
        } catch (IOException ioe) {
            logger.error(baseErrMesg, ioe);
            throw new WorkspaceException(command + "\n" + baseErrMesg,
                                               ioe);
        } finally {
            try {
                if (process != null) {
                    InputStream istr = process.getInputStream();
                    if (istr != null) {
                        istr.close();
                    }
                    OutputStream ostr = process.getOutputStream();
                    if (ostr != null) {
                        ostr.close();
                    }
                    istr = process.getErrorStream();
                    if (istr != null) {
                        istr.close();
                    }
                }
            } catch (IOException exp) {
                logger.debug("Could not close stream", exp);
            }
        }
        if (outBuffer != null) {
            return outBuffer.toString();
        } else {
            return null;
        }
    }

    /**
     * Specialized (though still generic) runCommand for caller to parse return
     * value on errors (commands returns nonzero).
     * @param  command
     *         Array of command + arguments, to fork a process
     * @return String
     *         The output of the process on to standard out.
     * @throws WorkspaceException
     * @throws ReturnException
     *         Exception containing return code other than zero and also the
     *         stdin/stdout/stderr if needed.
     */
    public static String runCommand(String[] command)
        throws WorkspaceException, ReturnException {

        if (command == null) {
            logger.debug("Command cannot be null");
            throw new WorkspaceException("Command cannot be null");
        }

        if (logger.isDebugEnabled()) {
            for (int i=0; i<command.length; i++) {
                logger.debug("Command arg " + i + " :: " + command[i]);
            }
        }

        Runtime runtime = Runtime.getRuntime();
        StringBuffer outBuffer = null;
        BufferedReader out = null;
        BufferedReader err = null;
        Process process = null;
        try {
            process = runtime.exec(command);

            int returnCode = -1;
            try {
                returnCode = process.waitFor();
            } catch (InterruptedException exp) {
                logger.error("Interupped exp thrown ", exp);
            }
            logger.debug("Return code is " + returnCode);

            // Get process output on stdout
            out = new BufferedReader(new InputStreamReader
                (process.getInputStream()));
            if (out != null) {
                String line = out.readLine();
                if (line != null) {
                    outBuffer = new StringBuffer(line);
                    while (line != null) {
                        line = out.readLine();
                        if (line != null) {
                            outBuffer.append("\n" + line);
                        }
                    }
                }
            }

            // If return code is not zero, try to get error stream also
            if (returnCode != 0) {
                err = new BufferedReader(new InputStreamReader
                    (process.getErrorStream()));
                StringBuffer errBuffer = null;
                if (err != null) {
                    String line = err.readLine();
                    if (line != null) {
                        errBuffer = new StringBuffer("\n" + line);
                        while (line != null) {
                            line = err.readLine();
                            if (line != null) {
                                errBuffer.append("\n" + line);
                            }
                        }
                    }
                }
                if ((errBuffer != null) && (outBuffer != null)) {
                    logger.error("ErrStr " + errBuffer.toString());
                    logger.error("OutStr " + outBuffer.toString());
                    throw new ReturnException(returnCode,
                            errBuffer.toString(), outBuffer.toString());
                } else if (errBuffer != null) {
                    logger.error("ErrStr " + errBuffer.toString());
                    throw new ReturnException(returnCode, errBuffer.toString());
                } else {
                logger.error("command returned " + returnCode +
                        " with no stderr");
                throw new ReturnException(returnCode);
                }
            }
        } catch (IOException ioe) {
            logger.error(ioe);
            throw new WorkspaceException(command + "\n", ioe);
        } finally {
            try {
                if (process != null) {
                    InputStream istr = process.getInputStream();
                    if (istr != null) {
                        istr.close();
                    }
                    OutputStream ostr = process.getOutputStream();
                    if (ostr != null) {
                        ostr.close();
                    }
                    istr = process.getErrorStream();
                    if (istr != null) {
                        istr.close();
                    }
                }
            } catch (IOException exp) {
                logger.debug("Could not close stream", exp);
            }
        }
        if (outBuffer != null) {
            return outBuffer.toString();
        } else {
            return null;
        }
    }
}
