/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

package org.globus.workspace.database;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.security.gridmap.GridMap;
import org.globus.workspace.LimitException;
import org.globus.workspace.WorkspaceConstants;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.impl.authz.VomsPDPPolicy;
import org.globus.workspace.impl.service.LCMAPSResource;
import org.globus.workspace.impl.service.WorkspaceResourceImpl;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.jndi.Initializable;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

public class PersistenceAdapter implements Initializable {
	private static Log logger = LogFactory.getLog(PersistenceAdapter.class
			.getName());

	private String dbLocation = null;

	private DataSource dataSource = null;

	private int maxWorkspaces = 1;

	private Hashtable poolNameMapping = new Hashtable();

	private GridMap dnPolicyFile = new GridMap();

	private GridMap attrPolicyFile = new GridMap();

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.globus.wsrf.jndi.Initializable#initialize()
	 */
	public void initialize() throws Exception {
		try {
			this.dataSource = (DataSource) new InitialContext()
					.lookup(this.dbLocation);
		} catch (NamingException e) {
			logger.error("Error initializing the datasource", e);
			throw e;
		}
	}

	public void setDbLocation(String dbLocation) {
		this.dbLocation = dbLocation;
	}

	public void setMaxWorkspaces(int maxWorkspaces) {
		this.maxWorkspaces = maxWorkspaces;
	}

	public int getNextID() throws WorkspaceDatabaseException {
		Connection c = null;
		Statement statement = null;
		Statement tempStatement = null;
		int nextId = 0;
		ResultSet rs = null;
		try {
			c = getConnection();
			statement = c.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
					ResultSet.CONCUR_UPDATABLE);
			rs = statement.executeQuery(WorkspaceConstants.GET_NEXT_ID_QUERY);
			if (!rs.next()) {
				// if this is first row in the database
				nextId = nextId + 1;
				tempStatement = c.createStatement();
				tempStatement
						.executeUpdate("INSERT INTO WORKSPACE_ID_TABLE VALUES("
								+ nextId + ")");
			} else {
				rs.previous(); // Reset the cursor
				while (rs != null && rs.next()) {
					nextId = rs.getInt(1);
					// Get the req Id and increment it instead of using
					// Db Specific sequence
					nextId = nextId + 1;
					// Now Update the old value with new value
					logger.debug("next request id: " + nextId);
					String updateQuery = "UPDATE WORKSPACE_ID_TABLE set ID=";
					updateQuery = updateQuery + nextId;
					tempStatement = c.createStatement();
					tempStatement.executeUpdate(updateQuery);
				}
			}
		} catch (SQLException e) {
			throw new WorkspaceDatabaseException("Database Error", e);
		} finally {
			try {
				if (statement != null) {
					statement.close();
				}
				if (tempStatement != null) {
					tempStatement.close();
				}
				if (rs != null) {
					rs.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sql) {
				logger.warn("Database error", sql);
			}
		}
		return nextId;
	}

	public void store(String localName, WorkspaceResourceImpl resource)
			throws WorkspaceDatabaseException {
		logger.debug("store called for : " + localName);
		Connection c = null;
		PreparedStatement ps = null;
		try {
			c = getConnection();
			int ID = getIDForLocalName(localName);
			if (ID == 0) {
				// this happens when resource is stored for the first time
				ID = getNextID();
			}
			ps = c.prepareStatement(WorkspaceConstants.STORE_QUERY);
			ps.setObject(1, new Integer(ID));
            ps.setObject(2, resource.getLocalAccountName());
			ps.setObject(3, resource.getCreationMechanism());
			ps.setObject(4, resource.getCallerDN());
			if (resource.getPrimaryAttr() != null) {
				ps.setObject(5, resource.getPrimaryAttr());
			} else {
				ps.setNull(5, Types.VARCHAR);
			}
			ps.setObject(6, new Long(resource.getCreationTime()));
			ps.setObject(7, new Long(resource.getTerminationTime()
					.getTimeInMillis()));
			ps.setObject(8, new Boolean(resource.isPolicyPresent()));
			ps.setObject(9, new Boolean(resource.isRolesPresent()));
			if (resource.getVO() != null) {
				ps.setObject(10, resource.getVO());
			} else {
				ps.setNull(10, Types.VARCHAR);
			}
			ps.execute();
			storeAccessPolicyDNs(ID, resource.getAccessPolicy());
			storeCreatorAttributes(ID, resource.getCreatorAttributesVector());
			storeVOMSPDPPolicy(ID, resource.getAuthzPolicy());
			logger.debug("exiting store");
		} catch (SQLException sqlex) {
			try {
				c.rollback();
			} catch (SQLException sqlex2) {
				throw new WorkspaceDatabaseException("Error storing resource",
						sqlex2);
			}
			logger.error("Error storing resource:" + sqlex.getMessage());
			throw new WorkspaceDatabaseException("Error storing resource",
					sqlex);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sqlex) {
				logger.error("Error closing connections", sqlex);
			}
		}
	}

	private void storeAccessPolicyDNs(int ID, Vector policyVector)
			throws SQLException {
		Connection c = getConnection();
		PreparedStatement ps;
		if (policyVector == null || policyVector.size() <= 0) {
			logger.debug("exiting storeAccessPolicyDNs");
			return;
		}
		ps = c
				.prepareStatement(WorkspaceConstants.STORE_ACCESS_POLICY_DN_QUERY);
		for (int i = 0; i < policyVector.size(); i++) {
			ps.setObject(1, new Integer(ID));
			logger.debug("Policy in storeAccessPolicyDNs: "
					+ policyVector.elementAt(i));
			ps.setObject(2, policyVector.elementAt(i));
			ps.execute();
		}
		if (ps != null) {
			ps.close();
		}
		returnConnection(c);
	}

	private void storeCreatorAttributes(int ID, Vector rolesVector)
			throws SQLException {
		logger.debug("entering storeCreatorAttributes");
		PreparedStatement ps;
		Connection c = getConnection();
		if (rolesVector == null || rolesVector.size() <= 0) {
			logger.debug("exiting from storeCreatorAttributes");
			returnConnection(c);
			return;
		}
		ps = c.prepareStatement(WorkspaceConstants.STORE_CREATOR_ATTR_QUERY);
		for (int i = 0; i < rolesVector.size(); i++) {
			logger.debug("ID in storeCreatorAttributes : " + ID);
			ps.setObject(1, new Integer(ID));
			String temp = (String) rolesVector.elementAt(i);
			logger.debug("roles in storeCreatorAttributes : " + temp);
			ps.setObject(2, temp);
			ps.execute();
		}
		if (ps != null) {
			ps.close();
		}
		returnConnection(c);
		logger.debug("exiting storeCreatorAttributes");
	}

	private void storeVOMSPDPPolicy(int ID, VomsPDPPolicy pdpPolicy)
			throws SQLException {
		logger.debug("Entering storeVOMSPDPPolicy");
		Connection c = getConnection();
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.STORE_VOMS_ATTRIBUTES_QUERY);
		Iterator it = pdpPolicy.getAttrs();
		if (it == null) {
			logger.debug("pdp policy attrs are null");
			returnConnection(c);
			return;
		}
		while (it.hasNext()) {
			ps.setObject(1, new Integer(ID));
			String temp = (String) it.next();
			ps.setObject(2, temp);
			logger.debug("pdp policy attr: " + temp);
			ps.execute();
		}
		ps.close();
		ps = c.prepareStatement(WorkspaceConstants.STORE_VOMS_DN_QUERY);
		Iterator newIt = pdpPolicy.getDNs();
		if (newIt == null) {
			logger.debug("pdpDns are null");
		} else {
            while (newIt.hasNext()) {
                ps.setObject(1, new Integer(ID));
                String temp = (String) newIt.next();
                logger.debug("Voms DN in storeVOMSPDPPolicy: " + temp);
                ps.setObject(2, temp);
                ps.execute();
            }
        }
        if (ps != null) {
			ps.close();
		}
		returnConnection(c);
		logger.debug("exiting storeVOMSPDPPolicy");
	}

	public void storeLCMapsResource(String localName,
			LCMAPSResource lcmapsResource) throws SQLException {
		PreparedStatement ps;
		int ID = getIDForLocalName(localName);
		Connection c = getConnection();
		ps = c.prepareStatement(WorkspaceConstants.STORE_LCMAPS_QUERY);
		ps.setObject(1, new Integer(ID));
		ps.setObject(2, lcmapsResource.getPoolIndex());
		ps.setObject(3, new Integer(lcmapsResource.getCounter()));
		ps.execute();
        ps.close();
		returnConnection(c);
	}

	// If moving away from reliance on sweeper thread for removals at
    // Home.init, one possibility would be to do removals here.
    public List getActiveWorkspaceIDs() throws WorkspaceDatabaseException {
		Vector activeLocalUserNamesVector = new Vector();
		Connection c = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.GET_ACTIVE_WORKSPACE_ID);
			rs = ps.executeQuery();
			while (rs != null && rs.next()) {
				activeLocalUserNamesVector.addElement(rs.getString(1));
			}
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
				if (ps != null) {
					ps.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sqlex) {
				logger.error(sqlex);
			}
		}
		List returnList = new ArrayList(activeLocalUserNamesVector);
		logger.debug("exiting getActiveWorkspaceIDs");
		return returnList;
	}

	public void load(WorkspaceResourceImpl resource, String localName)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.LOAD_RESOURCE_QUERY);
			ps.setObject(1, localName);
			rs = ps.executeQuery();
			while (rs != null && rs.next()) {
				int ID = rs.getInt(1);
				logger.debug("ID: " + ID);
				resource.setLocalName(rs.getString(2));
				resource.setCreationMechanism(rs.getString(3));
				resource.setCallerDN(rs.getString(4));
				resource.setPrimaryAttr(rs.getString(5));

				Calendar tempCalendar = Calendar.getInstance();
				tempCalendar.setTimeInMillis(rs.getLong(6));
				resource.setCreationTime(tempCalendar);
				tempCalendar.setTimeInMillis(rs.getLong(7));
				resource.setTerminationTime(tempCalendar);
				if (rs.getBoolean(8)) {
					resource.setAccessPolicy(getAccessPolicyDNs(ID));

				}
				if (rs.getBoolean(9)) {
					resource
							.setCreatorAttributesVector(getCreatorAttributes(ID));
				}
				resource.setVO(rs.getString(10));
				resource.setAuthzPolicy(getVOMSPDPPolicy(ID));
			}
		} catch (SQLException sqlex) {
			logger.error("Error loading resource" + sqlex);
			throw new WorkspaceDatabaseException("Error loading resource",
					sqlex);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
				if (rs != null) {
					rs.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sqlex2) {
				logger.error("error closing prepared statement", sqlex2);
			}
		}
		logger.debug("exit load");
	}

	/**
	 * @param id
	 * @return vector of attributes
	 */
	private Vector getCreatorAttributes(int id) throws SQLException {
		Vector creatorAttrVector = new Vector();
		Connection c = getConnection();
		ResultSet rs;
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_CREATOR_ATTRIBUTE_QUERY);
		ps.setObject(1, new Integer(id));
		rs = ps.executeQuery();
        if (rs != null) {
            while (rs.next()) {
                creatorAttrVector.add(rs.getString(1));
            }
            rs.close();
        }
        ps.close();
		returnConnection(c);
		return creatorAttrVector;
	}

	/**
	 * @param id
	 * @return vector of dns
	 */
	private Vector getAccessPolicyDNs(int id) throws SQLException {
		Vector policyVector = new Vector();
		Connection c = getConnection();
		ResultSet rs;
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_ACCESS_POLICY_DN_QUERY);
		ps.setObject(1, new Integer(id));
		rs = ps.executeQuery();
        if (rs != null) {
            while (rs.next()) {
                String policyString = rs.getString(1);
                if (policyString != null) {
                    policyVector.add(policyString);
                }
            }
            rs.close();
        }
        ps.close();
		returnConnection(c);
		return policyVector;
	}

	private VomsPDPPolicy getVOMSPDPPolicy(int ID) throws SQLException {
		Connection c = getConnection();
		ResultSet rs;
		ArrayList DNList = new ArrayList();
		ArrayList attrList = new ArrayList();
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_VOMS_DN_QUERY);
		ps.setObject(1, new Integer(ID));
		rs = ps.executeQuery();
		while (rs != null && rs.next()) {
			String temp = rs.getString(1);
			DNList.add(temp);
		}
		ps.close();
		ps = c.prepareStatement(WorkspaceConstants.GET_VOMS_ATTRS_QUERY);
		ps.setObject(1, new Integer(ID));
		rs = ps.executeQuery();
        if (rs != null) {
            while (rs.next()) {
                attrList.add(rs.getString(1));
            }
		    rs.close();
        }
        ps.close();
		returnConnection(c);
		VomsPDPPolicy authzPolicy = new VomsPDPPolicy(null, null);
		authzPolicy.setDNs(DNList);
		authzPolicy.setAttrs(attrList);
		return authzPolicy;
	}

	public void loadLCMAPSResource(String localName, LCMAPSResource resource)
			throws SQLException {
		int ID = getIDForLocalName(localName);
		Connection c = getConnection();
		ResultSet rs;
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.LOAD_LCMAPS_QUERY);
		ps.setObject(1, new Integer(ID));
		rs = ps.executeQuery();
        if (rs != null) {
            while (rs.next()) {
                resource.setPoolIndex(rs.getString(2));
                resource.setCounter(rs.getInt(3));
            }
            rs.close();
        }
        returnConnection(c);
	}

	/*
	 * @return @throws SQLException
	 */
	public Connection getConnection() throws SQLException {
		return this.dataSource.getConnection();
	}

	/*
	 * Return connection to the pool.
	 */
	private void returnConnection(Connection connection) {
		if (logger.isTraceEnabled()) {
			logger.trace("returnConnection()");
		}

		if (connection != null) {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.error("", e);
			}
		}
	}

	/**
	 * @param localName
	 */
	public void checkForDuplicate(String localName)
			throws NoSuchResourceException, ResourceException, LimitException {
		Connection c = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			ps = c
					.prepareStatement(WorkspaceConstants.CHECK_FOR_DUPLICATE_QUERY);
			ps.setObject(1, localName);
			rs = ps.executeQuery();
			if (rs.next()) {
				throw new LimitException();
			} else {
				throw new NoSuchResourceException();
			}
		} catch (SQLException e) {
			throw new ResourceException(e);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
				if (rs != null) {
					rs.close();
				}
			} catch (SQLException sql) {
				logger.error("error closing prepared statement", sql);
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public boolean resolvePolicy(String DN, String primaryAttr)
			throws WorkspaceException {
		int counter = 0;
		Connection c = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			if (primaryAttr == null) {
				ps = c.prepareStatement(WorkspaceConstants.GET_COUNT_QUERY);
				ps.setObject(1, DN);
			} else {
				ps = c.prepareStatement(WorkspaceConstants.GET_COUNT_QUERY2);
				ps.setObject(1, DN);
				ps.setObject(2, primaryAttr);
			}
			rs = ps.executeQuery();
			while (rs != null && rs.next()) {
				counter = rs.getInt(1);
			}
		} catch (SQLException sql) {
			logger.error("Error getting number of workspaces", sql);
			throw new WorkspaceDatabaseException(sql);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
				if (rs != null) {
					rs.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sql2) {
				logger.error("Error closing sql statement", sql2);
			}
		}
		if (counter >= maxWorkspaces) {
			logger.error("Reached maximum workspaces");
			return false;
		} else {
			return true;
		}
	}

	public int getNextCounter(String DN, String attr)
			throws WorkspaceDatabaseException, LimitException {
		// TODO: More testing needs to be done for corner cases like when
		// maxWorkspaces = 1
		if (DN == null) {
			logger.error("DN is null, no next counter");
			return -1;
		}
		logger.debug("Maximum workspaces: " + this.maxWorkspaces);
		int nextCounter = 0;
		ArrayList idList = new ArrayList();
		ArrayList counterList = new ArrayList();
		Connection c = null;
		Statement statement = null;
		Statement tempStatement = null;
		ResultSet rs = null;
		StringBuffer query = new StringBuffer();
		if (attr == null) {
			query.append("SELECT ID FROM WORKSPACE_TABLE  WHERE CALLER_DN='")
					.append(DN).append("'");
		} else {
			query.append("SELECT ID FROM WORKSPACE_TABLE WHERE CALLER_DN ='")
					.append(DN).append("' AND PRIMARY_ATTR='").append(attr)
					.append("'");
		}
		try {
			c = getConnection();
			statement = c.createStatement();
			logger.debug("Query to get IDs:" + query.toString());
			rs = statement.executeQuery(query.toString());
			while (rs != null && rs.next()) {
				idList.add(new Integer(rs.getInt(1)));
			}
			String listString;
			if (idList.size() > this.maxWorkspaces) {
				// this should never happen here as it should've been dealt
				// before a call to this method happens
				logger.error("Reached maximum limit of workspaces");
				throw new LimitException();
			} else if (idList.size() >= 1) {
				listString = idList.toString();
				listString = listString.substring(1, listString.length() - 1);
				StringBuffer newQuery = new StringBuffer();
				newQuery
						.append("SELECT COUNTER FROM LCMAPS_TABLE WHERE ID IN (");
				newQuery.append(listString);
				newQuery.append(") ORDER BY ID");
				logger.debug(newQuery.toString());
				tempStatement = c.createStatement();
				rs = tempStatement.executeQuery(newQuery.toString());
				while (rs != null && rs.next()) {
					counterList.add(new Integer(rs.getInt(1)));
				}
				Integer[] takenList = (Integer[]) counterList
						.toArray(new Integer[0]);
				int i = 0;
				logger.debug("length: " + counterList.size());
				logger.debug("length: " + takenList.length);
				for (; i < takenList.length; i++) {
					if (takenList[i].intValue() != i) {
						logger.debug("taken[" + takenList[i] + "!=" + i);
						break;
					}
					nextCounter = i + 1;
				}
			} else if (idList.size() == 0) {
				logger.debug("First workspace for this user: " + DN);
				nextCounter = 0;
			}

		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			try {
				if (statement != null) {
					statement.close();
				}
				if (tempStatement != null) {
					tempStatement.close();
				}
				if (rs != null) {
					rs.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sqlex) {
				logger.error("Error closing connections", sqlex);
			}
		}
		logger.debug("RETURNING THIS AS NEXT COUNTER: " + nextCounter);
		return nextCounter;
	}

	public synchronized void removeFromAccessPolicy(String[] entries)
			throws WorkspaceDatabaseException {
		PreparedStatement ps = null;
		Connection c = null;
		try {
			c = getConnection();
			ps = c
					.prepareStatement(WorkspaceConstants.REMOVE_ACCESS_POLICY_HOME_QUERY);
			for (int i = 0; i < entries.length; i++) {
				logger.debug("Access Policy : " + entries[i]);
				ps.setObject(1, entries[i]);
				ps.execute();
			}
		} catch (SQLException sql) {
			logger.error("Error removing access policies", sql);
			throw new WorkspaceDatabaseException(sql);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sql2) {
				logger.error("Error returning connection", sql2);
			}
		}
	}

	public void remove(String localName) throws WorkspaceDatabaseException {
		logger.debug("REMOVING:" + localName);
		PreparedStatement ps = null;
		Connection c = null;
		try {
			int ID = getIDForLocalName(localName);
			c = getConnection();
			c.setAutoCommit(false);
			ps = c.prepareStatement(WorkspaceConstants.DELETE_WORKSPACE_ROWS);
			ps.setObject(1, new Integer(ID));
			ps.execute();
			ps = c
					.prepareStatement(WorkspaceConstants.DELETE_ACCESS_POLICY_ROWS);
			ps.setObject(1, new Integer(ID));
			ps.execute();
			ps = c
					.prepareStatement(WorkspaceConstants.DELETE_CREATOR_ATTRIBUTES);
			ps.setObject(1, new Integer(ID));
			ps.execute();
			ps = c.prepareStatement(WorkspaceConstants.DELETE_VOMS_ATTR_ROWS);
			ps.setObject(1, new Integer(ID));
			ps.execute();
			ps = c.prepareStatement(WorkspaceConstants.DELETE_VOMS_DN_ROWS);
			ps.setObject(1, new Integer(ID));

			ps.execute();
			c.commit();
		} catch (SQLException e) {
			try {
				c.rollback();
			} catch (SQLException e1) {
				throw new WorkspaceDatabaseException(
						"Exception in rollback of " + "remove()", e1);
			}
			throw new WorkspaceDatabaseException("Exception in remove()", e);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement ", e);
				}
			}
			if (c != null) {
				try {
                    c.setAutoCommit(true);
                } catch (SQLException e) {
					logger.error("", e);
				}
                returnConnection(c);
			}
		}
	}

	/**
	 */
	public void retireCounter(String localName) throws WorkspaceException {
		PreparedStatement ps = null;
		Connection c = null;
		try {
			int ID = getIDForLocalName(localName);
			logger.debug("retiring counter for " + localName + " " + ID);
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.DELETE_LCMAPS_ROWS);
			ps.setObject(1, new Integer(ID));
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

    /**
     *
     * @param localName
     * @param authzDN
     * @throws WorkspaceDatabaseException
     */
    public void addAuthzIdentity(String localName, String authzDN)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			int ID = getIDForLocalName(localName);
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.ADD_AUTHZ_DN_QUERY);
			ps.setObject(1, new Integer(ID));
			ps.setObject(2, authzDN);
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public void removeAuthzIdentity(String localName, String authzDN)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			int ID = getIDForLocalName(localName);
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.REMOVE_AUTHZ_DN_QUERY);
			ps.setObject(1, new Integer(ID));
			ps.setObject(2, authzDN);
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public void addAccessPolicy(String localName, String policyDN)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			int ID = getIDForLocalName(localName);
			c = getConnection();
			ps = c
					.prepareStatement(WorkspaceConstants.ADD_ACCESS_POLICY_DN_QUERY);
			ps.setObject(1, new Integer(ID));
			ps.setObject(2, policyDN);
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public void removeAccessPolicy(String localName, String policyDN)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			int ID = getIDForLocalName(localName);
			c = getConnection();
			ps = c
					.prepareStatement(WorkspaceConstants.REMOVE_ACCESS_POLICY_QUERY);
			ps.setObject(1, new Integer(ID));
			ps.setObject(2, policyDN);
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public void updateTerminationTime(String localName, Calendar terminationTime)
			throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.UPDATE_TERM_TIME_QUERY);
			ps.setObject(1, new Long(terminationTime.getTimeInMillis()));
			ps.setObject(2, localName);
			ps.execute();
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					logger.error("Error closing prepared statement", e);
				}
			}
			if (c != null) {
				returnConnection(c);
			}
		}
	}

	public ArrayList queryAccessPolicy(String DN, String attr, boolean all)
			throws WorkspaceDatabaseException {
		ArrayList localUserNames = new ArrayList();
		Connection c = null;
		Statement statment = null;
		ResultSet rs = null;
		try {
			Vector matchedIDs = getIDForAccessPolicyDNs(DN);
			matchedIDs = getIDForCreatorAttribute(matchedIDs, attr);
			c = getConnection();
			statment = c.createStatement();
			if (!all) {
				statment.setMaxRows(1);
			}
			StringBuffer query = new StringBuffer();
			query.append(WorkspaceConstants.GET_ACCESS_POLICY_QUERY);
			if (matchedIDs.size() > 1) {
                query.append(matchedIDs.elementAt(0));
                for (int i = 1; i < matchedIDs.size(); i++) {
                    query.append(",");
                    query.append(matchedIDs.elementAt(i));
				}
			} else if (matchedIDs.size() == 1) {
				query.append(matchedIDs.elementAt(0));
			}
			query.append(")");
			logger.debug(query.toString());
			if (matchedIDs.size() > 0) {
				rs = statment.executeQuery(query.toString());
				while (rs != null && rs.next()) {
					String temp = rs.getString(1);
					logger.debug("localname in : " + temp);
					localUserNames.add(temp);
				}
			}
		} catch (SQLException sqlex) {
			throw new WorkspaceDatabaseException(sqlex);
		} finally {
			try {
				if (statment != null) {
					statment.close();
				}
				if (rs != null) {
					rs.close();
				}
				if (c != null) {
					returnConnection(c);
				}
			} catch (SQLException sqlex2) {
				logger.error("Error closing preparedStatement", sqlex2);
			}
		}
		return localUserNames;
	}

	private Vector getIDForAccessPolicyDNs(String DN) throws SQLException {
		Vector IDs = new Vector();
		Connection c = getConnection();
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_ID_FOR_ACCESS_DN_QUERY);
		ps.setObject(1, DN);
		ResultSet rs = ps.executeQuery();
		while (rs != null && rs.next()) {
			IDs.add(new Integer(rs.getInt(1)));
		}
		if (rs != null) {
			rs.close();
		}
		ps.close();
		returnConnection(c);
		return IDs;
	}

	private Vector getIDForCreatorAttribute(Vector IDs, String attr)
			throws SQLException {
		Connection c = getConnection();
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_ID_FOR_CREATOR_ATTRIBUTE_QUERY);
		ps.setObject(1, attr);
		ResultSet rs = ps.executeQuery();
		while (rs != null && rs.next()) {
			IDs.add(new Integer(rs.getInt(1)));
		}
		if (rs != null) {
			rs.close();
		}
		ps.close();
		returnConnection(c);
		return IDs;
	}

	private int getIDForLocalName(String localName) throws SQLException {
		int ID = 0;
		Connection c = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
        try {
            c = getConnection();
            ps = c.prepareStatement(WorkspaceConstants.ID_TO_LOCAL_NAME_QUERY);
            ps.setObject(1, localName);
            rs = ps.executeQuery();
            while (rs != null && rs.next()) {
                ID = rs.getInt(1);
            }
		} catch (SQLException sql) {
			throw sql;
		} finally {
            if (rs != null) {
                rs.close();
            }
            try {
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException sql) {
				logger.error(sql);
			}
			if (c != null) {
				returnConnection(c);
			}
        }
		return ID;
	}

    // We assume here that this would never get called if there was
    // a localName,poolName already with any status.
    // The current table schema will allow two entries with different
    // status, so mind this assumption.
    public void storeLeasedAccount(String localName, String poolName,
			String status) throws WorkspaceDatabaseException {
		Connection c = null;
		PreparedStatement ps = null;
		try {
			c = getConnection();
			//int ID = getIDForLocalName(localName);
			ps = c.prepareStatement(WorkspaceConstants.STORE_LEASED_ACCOUNTS_QUERY);
			//ps.setObject(1, new Integer(ID));
			ps.setObject(1, localName);
			ps.setObject(2, poolName);
			ps.setObject(3, status);
			ps.executeUpdate();
		} catch (SQLException sql) {
			throw new WorkspaceDatabaseException(sql);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException sql) {
				logger.error(sql);
			}
			if (c != null) {
				returnConnection(c);
			}
        }
	}

	public LinkedList getFreeLocalNames(String poolName)
			throws WorkspaceDatabaseException {
		Connection c = null;
		LinkedList names = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			names = new LinkedList();
			ps = c.prepareStatement(WorkspaceConstants.GET_LEASED_NAMES_QUERY);
			ps.setObject(1, poolName);
			ps.setObject(2, WorkspaceConstants.LEASED_ACCOUNT_STATUS_FREE);
			rs = ps.executeQuery();
			while (rs != null && rs.next()) {
				names.add(rs.getString(1));
			}
		} catch (SQLException sql) {
			throw new WorkspaceDatabaseException(sql);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException sql) {
				logger.error(sql);
			}
			if (c != null) {
				returnConnection(c);
			}
        }

		return names;
	}

	public void deleteLeasedAccountMapping(String poolName, String localName)
			throws SQLException {
        logger.debug("DELETING MAPPING FOR: " + poolName
                                    + " account: " + localName);
		Connection c = null;
        PreparedStatement ps = null;
        try {
            c = getConnection();
            ps = c.prepareStatement(WorkspaceConstants.DELETE_LEASED_ACCOUNT_MAPPING_QUERY);
            ps.setObject(1, poolName);
            ps.setObject(2, localName);
            ps.executeUpdate();
        } catch (SQLException sql) {
            throw sql;
        } finally {
            if (ps != null) {
                ps.close();
            }
            if (c != null) {
                returnConnection(c);
            }
        }
    }

    /**
	 * @param localName
	 * @return pool name
	 */
	public String getLeasedAccountPoolName(String localName)
			throws SQLException {
		String poolName = null;
		Connection c = getConnection();
		PreparedStatement ps = c
				.prepareStatement(WorkspaceConstants.GET_LEASED_ACCOUNT_NAME_QUERY);
		ps.setObject(1, localName);
		ResultSet rs = ps.executeQuery();
        if (rs != null) {
            while (rs.next()) {
                poolName = rs.getString(1);
            }
            rs.close();
        }
        ps.close();
		returnConnection(c);
		return poolName;
	}

	public void initializeLeasedAccounts(String dnPolicyFileLocation,
			String attrPolicyFileLocation, String poolFileLocation)
			throws WorkspaceException {
		try {
			logger.debug("Loading DN Policy from : " + dnPolicyFileLocation);
			this.dnPolicyFile.load(dnPolicyFileLocation);
			logger
					.debug("Loading Attr Policy from : "
							+ attrPolicyFileLocation);
			this.attrPolicyFile.load(attrPolicyFileLocation);
			loadPoolFiles(poolFileLocation);
		} catch (IOException ioe) {
			throw new WorkspaceException("Error loading policy files", ioe);
		}
	}

	private void loadPoolFiles(String poolFileLocation) throws IOException {
		logger.debug("PoolFile Location : " + poolFileLocation);
		File poolPath = new File(poolFileLocation);

		String[] listing = poolPath.list();
		for (int i = 0; i < listing.length; i++) {
			String filePath = poolFileLocation + listing[i];
			logger.debug("Reading File : " + filePath);
			File pool = new File(filePath);
			List poolAccountsList = new LinkedList();
			if (pool.isFile()) {
				logger.debug("this is a file");
				// Read the contents of file
				InputStream in = null;

				try {
                    String line;
                    in = new FileInputStream(pool);
					BufferedReader reader = new BufferedReader(
							new InputStreamReader(in));
					while ((line = reader.readLine()) != null) {
						line = line.trim();
						if (line.length() > 0) {
							logger.debug("pool acocunt : " + line);
							// poolaccount
							poolAccountsList.add(line);
						}
					}
				} finally {
					if (in != null) {
						try {
							in.close();
						} catch (Exception e) {
							logger.error("", e);
						}
					}
				}
				this.poolNameMapping.put(pool.getName(), poolAccountsList);
			}
		}
	}

	public String getPoolName(String callerDN, String primaryAttr,
			Vector creatorAttributesVector) {
		String poolName;
		// Get the user name from DN policy
		poolName = this.dnPolicyFile.getUserID(callerDN);
		if (poolName != null) {

			if (poolName.startsWith(".")) {
				poolName = poolName.substring(1, poolName.length());
				logger.debug("poolname from DN file : " + poolName);
				return poolName;
			}
		} else {
			poolName = this.attrPolicyFile.getUserID(primaryAttr);
			if (poolName == null) {
				logger.debug("primaryAttr not found in attrPolicy file");
				logger.debug("checking for other attributes");
				for (int i = 0; i < creatorAttributesVector.size(); i++) {
					String tempAttrString = (String) creatorAttributesVector
							.elementAt(i);
					poolName = this.attrPolicyFile.getUserID(tempAttrString);
					if (poolName != null) {
						break;
					}
				}
			}
			if (poolName != null) {
				poolName = poolName.substring(1, poolName.length());
				logger.debug("poolName found in attr policy file" + poolName);
				return poolName;
			}
		}
		return poolName;
	}

	public LinkedList getPoolAccounts(String poolName)
			throws WorkspaceDatabaseException {
		logger.debug("poolName : " + poolName);
        // can be null, for example if a user is configured for a
        // pool that is not configured
        return (LinkedList) this.poolNameMapping.get(poolName);
	}

	public Hashtable getAccountsPool() {
		return this.poolNameMapping;
	}

    public void quarantine(String localName, String poolName)
                                throws WorkspaceDatabaseException {
        PreparedStatement ps = null;
		Connection c = null;
        try {
            c = getConnection();
            c.setAutoCommit(false);
            ps = c.prepareStatement(WorkspaceConstants.DELETE_LEASED_ACCOUNT_MAPPING_QUERY);
            ps.setObject(1, poolName);
            ps.setObject(2, localName);
            ps.executeUpdate();
			ps = c.prepareStatement(WorkspaceConstants.STORE_LEASED_ACCOUNTS_QUERY);
			ps.setObject(1, localName);
			ps.setObject(2, poolName);
			ps.setObject(3, WorkspaceConstants.LEASED_ACCOUNT_STATUS_QUARANTINED);
			ps.executeUpdate();
            c.commit();
        } catch (SQLException e) {
            String err = "error quarantining";
            logger.error(err,e);
            throw new WorkspaceDatabaseException(err,e);
        } finally {
            if (ps != null) {
                try {
                    ps.close();
                } catch (SQLException e) {
                    logger.error("",e);
                }
            }
            if (c != null) {
                try {
                    c.setAutoCommit(true);
                } catch (SQLException e) {
					logger.error("", e);
				}
                returnConnection(c);
            }
        }
    }

    public Hashtable getQuarantined() throws WorkspaceDatabaseException {
		Connection c = null;
		Hashtable quarantined = new Hashtable();
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			c = getConnection();
			ps = c.prepareStatement(WorkspaceConstants.GET_QUARANTINED_ACCOUNTS_QUERY);
			rs = ps.executeQuery();
			while (rs != null && rs.next()) {
                String poolname = rs.getString(1);
                String localname = rs.getString(2);
                if (quarantined.containsKey(poolname)) {
                    ((LinkedList)quarantined.get(poolname)).add(localname);
                } else {
                    LinkedList locals = new LinkedList();
                    locals.add(poolname);
                    quarantined.put(poolname,locals);
                }
			}
		} catch (SQLException sql) {
			throw new WorkspaceDatabaseException(sql);
		} finally {
			try {
				if (rs != null) {
					rs.close();
				}
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException sql) {
				logger.error(sql);
			}
			if (c != null) {
				returnConnection(c);
			}
        }

		return quarantined;
    }
}
