package org.globus.workspace.impl.authz;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.globus.wsrf.Constants;
import org.globus.workspace.impl.service.WorkspaceHome;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.WorkspaceConstants;

import javax.xml.soap.SOAPMessage;
import javax.naming.InitialContext;
import java.io.File;
import java.io.IOException;
import java.util.Vector;

public class BanlistPDP extends VomsPDP {

    private static Log logger =
        LogFactory.getLog(BanlistPDP.class.getName());

    public static final String DN_BANLIST =
        "dnBanlist";

    public static final String ATTR_BANLIST =
        "attrBanlist";

    private String dnBanlistFileName = null;
    private File dnBanlistFile = null;

    private String attrBanlistFileName = null;
    private File attrBanlistFile = null;

    private String[] attrbans = null;
    private String[] dnbans = null;

    private long DNLastModified = 0;
    private long AttrLastModified = 0;


    public void initialize(PDPConfig config,
                       String name,
                       String id)
        throws InitializeException {
        logger.debug("name: " + name + " id : " + id);
        super.initialize(config,name,id);

        Object banfile = config.getProperty(name, DN_BANLIST);
        if (banfile != null) {
            this.dnBanlistFileName = (String) banfile;
            this.dnBanlistFile = findConfigFile(this.dnBanlistFileName);
            if (this.dnBanlistFile != null) {
                try {
                    this.dnbans = readConfigFile(this.dnBanlistFile);
                    this.DNLastModified =
                            this.dnBanlistFile.lastModified();
                } catch (IOException e) {
                    throw new InitializeException(e.toString());
                }
            } else {
                throw new InitializeException("dn ban list is configured " +
                        "but not found");
            }
        } else {
            logger.warn("DN ban list not configured");
        }

        banfile = config.getProperty(name, ATTR_BANLIST);
        if (banfile != null) {
            this.attrBanlistFileName = (String) banfile;
            this.attrBanlistFile = findConfigFile(this.attrBanlistFileName);
            if (this.attrBanlistFile != null) {
                try {
                    this.attrbans = readConfigFile(this.attrBanlistFile);
                    this.AttrLastModified =
                            this.attrBanlistFile.lastModified();
                } catch (IOException e) {
                    throw new InitializeException(e.toString());
                }
            } else {
                throw new InitializeException("attr ban list is configured " +
                        "but not found");
            }
        } else {
            logger.warn("attribute ban list not configured");
        }

    }

    /**
     * Called in base PDP (VomsPDP)
     * Boolean.TRUE = ACCEPT
     * Boolean.FALSE = DENY
     * null = Not applicable
     *
     * This banlist extension only returns only 'Not applicable' or 'DENY'
     *
     * @param peerIdentity
     * @param attributes
     * @param operation
     * @param content
     * @return
     */
    public Boolean checkCallAndContent(String peerIdentity,
                                       Vector attributes,
                                       String operation,
                                       SOAPMessage content) throws Exception {

        // see if the banlist files have changed since last invocation
        boolean dnchanged = checkDNFile();
        boolean attrchanged = checkAttrFile();

        // if either changed, flush workspace policies
        flushPolicies(dnchanged, attrchanged);

        if (!checkDNPolicy(peerIdentity) || !checkAttrPolicy(attributes)) {
            return Boolean.FALSE;
        }

        // for add ops, check content
        if (!approveAdditions(content, operation, this.dnbans, this.attrbans)) {
            logger.debug("policy addition contained a banned entity");
            return Boolean.FALSE;
        } else {
            return null;
        }
    }

    private void flushPolicies(boolean dnchanged, boolean attrchanged)
            throws WorkspaceException {

        // this may speed things up for large ban lists:
        if (dnchanged && !attrchanged) {
            WorkspaceHome home = getServiceHome();
            home.removeFromAccessPolicy(this.dnbans);
            home.removeFromMgmtPolicy(this.dnbans, null);
        } else if (!dnchanged && attrchanged) {
            WorkspaceHome home = getServiceHome();
            home.removeFromAccessPolicy(this.attrbans);
            home.removeFromMgmtPolicy(null, this.attrbans);
        } else if (dnchanged && attrchanged) {
            String[] both = new String[this.dnbans.length +
                    this.attrbans.length];
            int i = 0;
            for (;  i < both.length - this.attrbans.length; i++) {
                both[i] = this.dnbans[i];
            }
            for ( ; i < both.length; i++) {
                both[i] = this.attrbans[i - this.dnbans.length];
            }
            WorkspaceHome home = getServiceHome();
            home.removeFromAccessPolicy(both);
            home.removeFromMgmtPolicy(this.dnbans, this.attrbans);
        } else {
            return;
        }

    }

    private boolean checkDNPolicy(String peerIdentity) {
        if ((peerIdentity == null) || (this.dnbans == null)) {
            return true;
        }
        for (int i = 0; i < this.dnbans.length; i++) {
            if (this.dnbans[i].trim().equals(peerIdentity)) {
                logger.info("In DN banlist: " + peerIdentity);
                return false;
            }
        }
        return true;
    }

    /**
     * @return true if changed
     * @throws IOException
     */
    private boolean checkDNFile() throws IOException {
        /* if InitializeException is ignored from initialize */
        if (this.dnBanlistFile == null) {
            this.dnBanlistFile = findConfigFile(this.dnBanlistFileName);
        }
        if (this.dnBanlistFile == null) {
            logger.warn("noSecConfig: " + this.dnBanlistFileName);
            return false;
        }

        if (this.DNLastModified < this.dnBanlistFile.lastModified()) {
            logger.debug("file has been modified: " +
                    this.DNLastModified + " < "
                    + this.dnBanlistFile.lastModified());
            this.dnbans = readConfigFile(this.dnBanlistFile);
            this.DNLastModified = this.dnBanlistFile.lastModified();
            return true;
        } else {
            return false;
        }
    }

    private boolean checkAttrPolicy(Vector attributes) {
        if ((attributes == null) || (this.attrbans == null)) {
            return true;
        }
        for (int i = 0; i < attributes.size(); i++) {
            String attr = (String) attributes.get(i);
            logger.debug("checking attribute " + attr);

            for (int j = 0; j < this.attrbans.length; j++) {
                if (this.attrbans[j].trim().equals(attr)) {
                    logger.info("In attribute blacklist: " + attr);
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * @return true if changed
     * @throws IOException
     */
    private boolean checkAttrFile() throws IOException {
        /* if InitializeException is ignored from initialize */
        if (this.attrBanlistFile == null) {
            this.attrBanlistFile = findConfigFile(this.attrBanlistFileName);
        }
        if (this.attrBanlistFile == null) {
            logger.warn("noSecConfig " + this.attrBanlistFileName);
            return false;
        }

        if (this.AttrLastModified < this.attrBanlistFile.lastModified()) {
            logger.debug("file has been modified: " +
                    this.AttrLastModified + " < "
                    + this.attrBanlistFile.lastModified());
            this.attrbans = readConfigFile(this.attrBanlistFile);
            this.AttrLastModified = this.attrBanlistFile.lastModified();
            return true;
        } else {
            return false;
        }
    }

    private WorkspaceHome getServiceHome() throws WorkspaceException{
        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         WorkspaceConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (WorkspaceHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        WorkspaceConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new WorkspaceException(exp);
        }
    }
}
