/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.authz;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.glite.security.voms.BasicVOMSTrustStore;
import org.glite.security.voms.VOMSAttribute;
import org.glite.security.voms.VOMSValidator;
import org.globus.wsrf.impl.security.authorization.exceptions.AttributeException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.globus.wsrf.security.authorization.PIP;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import java.security.cert.X509Certificate;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

public class VomsCredentialPIP implements PIP {

    public final static String VOMS_HOSTPORT = "voms-hostport";
    public final static String VOMS_PREFIX = "vomsPdp";
    public final static String TRUST_STORE_DIR_PROP = "vomsTrustStore";
    public final static String VALIDATE = "validate";
    public final static String REFRESH_TIME_PROP = "vomsRefreshTime";

    static long defaultRefresh = 0;

    static Log logger =
	LogFactory.getLog(VomsCredentialPIP.class.getName());

    // Not allowing different trust stores for different PIP instances.
    //
    // This also means if factory is not configured with VomsPIP in the authz
    // chain but a resource is, it will fail.  As before, if not using VOMS,
    // it will just be disabled.  The service still needs to be configured
    // with VOMSCredentialPIP for now.
    static String trustStore = null;
    static Long refreshTime = null;
    static BasicVOMSTrustStore store = null;
    static boolean validate = false;
    static String vald = null;

    public void initialize(PDPConfig config, String name, String id)
	throws InitializeException {

        logger.debug("PIP " + config.getClass().getName());

        if (trustStore == null) {
            trustStore =
                (String) config.getProperty(VOMS_PREFIX, TRUST_STORE_DIR_PROP);
            logger.debug("found truststore configuration: " + trustStore);
        }

        if (refreshTime == null) {
            refreshTime =
                (Long) config.getProperty(VOMS_PREFIX, REFRESH_TIME_PROP);
        }

        if (vald == null) {
            vald = (String) config.getProperty(VOMS_PREFIX, VALIDATE);
            if (vald == null) {
                vald = "";
            }
            if (vald.equals("true")) {
                validate = true;
            }
        }

        try {
            String trustStoreParam = null;
            long refreshParam = 0;
            if ((trustStore == null) && (refreshTime == null)) {
                store = new BasicVOMSTrustStore();
            } else if (trustStore == null) {
                // refresh time is not null
                trustStoreParam =
                        BasicVOMSTrustStore.DEFAULT_TRUST_STORE_LISTING;
                refreshParam = refreshTime.longValue();
            } else if (refreshTime == null) {
                // trust is not null
                trustStoreParam = trustStore;
                refreshParam = defaultRefresh;
            } else {
                // both are not null
                trustStoreParam = trustStore;
                refreshParam = refreshTime.longValue();
            }

            if (store == null) {
                store =
                    new BasicVOMSTrustStore(trustStoreParam, refreshParam);
            }
        } catch (IllegalArgumentException e) {
            store = null;
            logger.warn("VOMS trust store not enabled, VOMS cert parsing " +
                    "disabled");
        } catch (Exception e) {
            store = null;
            logger.error("Problem configuring VOMS trust store, VOMS cert " +
                    "parsing disabled");
        }
    }

    public void close() throws CloseException {
    }

    public void collectAttributes(Subject peerSubject,
                                  MessageContext context,
                                  QName operation) throws AttributeException {

        if (store == null) {
            logger.debug("no VOMS trust store, VOMS cert parsing disabled");
            return;
        }

        try {
            Set publicCreds = 
                peerSubject.getPublicCredentials(X509Certificate[].class);
            logger.debug("cred set size: " + publicCreds.size());
            Iterator iter = publicCreds.iterator();

            Vector rolesVector = new Vector();
            String VO = null;
            String hostport = null;
            while (iter.hasNext()) {
                X509Certificate certRev[] = (X509Certificate[])iter.next();
                int size = certRev.length;
                X509Certificate cert[] = new X509Certificate[size];
                for (int i=0; i<size; i++) {
                    cert[i] = certRev[size-i-1];
                }

                VOMSValidator.setTrustStore(store);
                logger.debug("set truststore to " + store);
                VOMSValidator validator = new VOMSValidator(cert).parse();

                if (validate == true) {
                    logger.debug("validating");
                    validator = validator.validate();
                }

                logger.debug("Parse Validator: " + validator.toString());
                List vector = validator.getVOMSAttributes();
                if (vector.size() != 0) {
                    logger.debug("getVOMSAttributes() vector size "
                            + vector.size());
                    for (int j=0; j<vector.size(); j++) {
                        VOMSAttribute attrib =
                            (VOMSAttribute)vector.get(j);
                        VO = attrib.getVO();
                        hostport = attrib.getHostPort();
                        List fqan = attrib.getFullyQualifiedAttributes();
                        for (int k=0; k<fqan.size(); k++) {
                            String str = (String)fqan.get(k);
                            rolesVector.add(str);
                        }
                    }
                }
            }
            if (logger.isDebugEnabled()) {
                for (int i=0; i<rolesVector.size(); i++) {
                    logger.debug("\nRoles " + rolesVector.get(i));
                }
                logger.debug("VO " + VO);
                logger.debug("hostport " + hostport);
            }

            VomsCredentialInformation info =
                    new VomsCredentialInformation(rolesVector, VO, hostport);

            peerSubject.getPublicCredentials().add(info);

        } catch (Exception exp) {
            throw new AttributeException("", exp);
        }
    }
}
