/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.authz;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.Text;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.security.gridmap.GridMap;
import org.globus.workspace.WorkspaceConstants;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.impl.security.authorization.exceptions.InvalidPolicyException;
import org.globus.wsrf.impl.security.descriptor.ContainerSecurityConfig;
import org.globus.wsrf.impl.security.util.AuthUtil;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.w3c.dom.Node;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.rpc.handler.soap.SOAPMessageContext;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

public abstract class VomsPDP implements org.globus.wsrf.security.authorization.PDP {

    /* $GL is also checked when these are exhausted,
       see this.findConfigFile(String fileName)      */
    public static final String[] CONFIG_LOCATIONS =
        {"",".", "/etc", "/etc/grid-security"};


    /** If vomsPolicy is not null, the PDP consults dynamic policies in RAM.
       This is to use for associating VomsPDP to each resource instance that
       represents an account.

       Again, cannot use a mixture of identity authorization PDP and
       an attribute PDP because toolkit does not allow for
       permit-overides in the authorization chain.

       @see #isPermitted

       In the future, these dynamic policies will be mirrored to aggregate
       gridmap and groupmap files in order for localhost queries (e.g., from
       GT2 GRAM).
     */
    private VomsPDPPolicy vomsPolicy = null;

    /* attrs authz list */
    private String attrConfigFileName = null;
    private File attrConfigFile = null;
    private String[] attrs = null;
    private long attrConfigLastModified = 0;

    public static final String ATTR_SECURITY_CONFIG_FILE =
        "attrAuthzConfigFile";

    public static final String VOMS_PDP_POLICY =
        "vomsPDPPolicy";

    private static Log logger =
        LogFactory.getLog(VomsPDP.class.getName());

    public String[] getPolicyNames() {
        return null;
    }

    public Node getPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }

    public Node setPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }

    /**
     * Checks both DN/attrs -- it would be convenient, but these can not be
     * separate PDPs (one for DNs (gridmap) and one for VOMS attributes),
     * since chains are deny-overrides.  We need to allow either DN or
     * attributes.
     * TODO: adjust that before GT4.2, because authz chain change is in trunk
     * @param peer
     * @param msgCtx
     * @param op
     * @return
     * @throws AuthorizationException
     */
    public boolean isPermitted(Subject peer,
                               MessageContext msgCtx,
                               QName op)
            throws AuthorizationException {

        Boolean dnPassed = null;
        Boolean attrPassed = null;

        String operation = op.toString();
        String peerIdentity = AuthUtil.getIdentity(peer);

        logger.debug("Operation " + operation + " called by subject: "
                     + peerIdentity);

        if (this.vomsPolicy == null) {
            try {
                if (checkGridMapFile(peerIdentity)) {
                    dnPassed = Boolean.TRUE;
                } else {
                    logger.debug("DN not in gridmap file configured for the container: " 
                            + peerIdentity);
                }
            } catch (IOException e) {
                throw new AuthorizationException("",e);
            }
        } else {
            logger.debug("using RAM policy for DN");
            Iterator DNs = this.vomsPolicy.getDNs();
            String DN = null;
            while (DNs.hasNext()) {
                DN = (String) DNs.next();
                logger.debug("found a DN in RAM policy: " + DN);
                if (peerIdentity.equals(DN)) {
                    dnPassed = Boolean.TRUE;
                    break;
                }
            }

        }

        CredentialInformation info = null;
        Set credSet = peer.getPublicCredentials();
        Iterator creds = credSet.iterator();
        while (creds.hasNext()) {
            Object o = creds.next();
            if (o instanceof CredentialInformation) {
                info = (CredentialInformation) o;
                break;
            }
        }

        Vector rolesVector = null;
        VomsCredentialInformation vomsinfo = null;

        /* It's OK if there is no attribute information in the message
           context, authz will be by DN only */
        if (info == null) {
            logger.info("cannot retrieve credential info from message context");
        } else {
            /* There is no unified attribute internal interface yet, this is
               here until that is in place */
            if (!(info instanceof VomsCredentialInformation)) {
                throw new AuthorizationException("credenital info from " +
                        "message context is not VOMS: incompatible PIP");
            } else {
                vomsinfo = (VomsCredentialInformation) info;
            }

            rolesVector = vomsinfo.getAttrs();
        }

        /* It's OK if there are no roles, see above */
        if (rolesVector == null) {
            logger.warn("cannot retrieve roles from credential information");
        } else {

            if (logger.isDebugEnabled()) {
                for (int i=0; i<rolesVector.size(); i++) {
                    logger.debug("\nRoles " + rolesVector.get(i));
                }
            }

            // Use policy in RAM
            if (this.vomsPolicy != null) {
                logger.debug("using RAM policy for attributes");
                Iterator attrs = this.vomsPolicy.getAttrs();
                String policyAttr = null;
                //current algorithm only cares if there is one passing attribute
                while ((attrPassed == null) && (attrs.hasNext())) {
                    policyAttr = (String) attrs.next();
                    logger.debug("found an attribute in RAM policy: "
                            + policyAttr);
                    for (int i=0; i<rolesVector.size(); i++) {
                        String attr = (String)rolesVector.get(i);
                        if (policyAttr.equals(attr)) {
                            attrPassed = Boolean.TRUE;
                            break;
                        }
                    }
                }
            } // end if (this.vomsPolicy != null)

            // Use policy in file
            if (this.vomsPolicy == null) {

                for (int i=0; i<rolesVector.size(); i++) {
                    String attr = (String)rolesVector.get(i);
                    logger.debug("checking attribute " + attr);
                    try {
                        if (checkAttrFile(attr) == false) {
                            logger.debug("secondary attribute denied: " + attr);
                        } else {
                            //current algorithm only cares if there is one
                            // passing attribute
                            attrPassed = Boolean.TRUE;
                            break;
                        }
                    } catch (IOException e) {
                        throw new AuthorizationException("",e);
                    }
                }
            } // end if (this.vomsPolicy == null)
        } // if/else (rolesVector == null)

        if (vomsinfo != null) {
            String VO = vomsinfo.getVO();
            logger.debug("VO " + VO);
        }

        // now we have positive policy decision from either the ACL
        // or the RAM policy (either 'ACCEPT' or 'Not applicable').
        // Send parameters and message to any overriding PDP (this
        // class's impl. of checkCallAndContent is just a no-op).
        //
        // The result of this call will be combined in a deny-override
        // fashion with the current positive decisions (and not appl.
        // all around results in DENY)

        Boolean other = null;
        try {
            logger.debug("calling checkCallAndContent on PDP impl");
            other = checkCallAndContent(peerIdentity, rolesVector,
                    operation, getMessage(msgCtx));
        } catch (Exception e) {
            throw new AuthorizationException("",e);
        }

        // Not applicable all around results in DENY
        boolean result = combine(dnPassed, attrPassed, other);

        /* see bugzilla #2287

           if 'return false' based on attribute, the exception returned to
           client by ServiceAuthorizationChain claims DN is denied.

           One possible workaround, AuthorizationException, implies error in
           policy processing, so something along these lines is actually not
           a viable alternative:

           throw new AuthorizationException("primary attribute denied: "
           + primaryAttr);

           Also, now if denial is based on content of operation, this
           error message is also not 'correct'
        */


        if (result) {
            logger.info("ACCEPTED: Operation " + operation + " called " +
                    "by subject: " + peerIdentity);
            return true;
        } else {
            logger.info("DENIED: Operation " + operation + " called " +
                    "by subject: " + peerIdentity);
            return false;
        }
    }

    /**
     * Allows for extending PDP to do more specific checks on
     * caller and content.
     *
     * This is combined in the base PDP logic in a deny overrides
     * fashion.
     *
     * @param peerIdentity
     * @param attributes
     * @param operation
     * @param content object representing message body,
     * @return Boolean.TRUE for ACCEPT, Boolean.FALSE for DENY, null for NOT APPLICABLE
     */
    public abstract Boolean checkCallAndContent(String peerIdentity,
                                       Vector attributes,
                                       String operation,
                                       SOAPMessage content) throws Exception;

    /**
     * nulls are not considered, rest are based on deny-override
     * Note that dnPassed and attrPassed should only be positive or null
     *
     * @param dnPassed
     * @param attrPassed
     * @param other
     * @return
     */
    private boolean combine(Boolean dnPassed,
                            Boolean attrPassed,
                            Boolean other) {

        // This implementation treats the three distinctly in
        // order to present detailed logging.
        //
        // As for the best way to get the result, the 'way' to do
        // it in general would be for this method to accept an array
        // of Boolean objects, toss out nulls, combine based on
        // && booleanValue(), and make sure to convert 'all null'
        // into false.

        boolean dn = false;
        boolean attr = false;
        if (dnPassed != null) {
            dn = true;
            if (!dnPassed.booleanValue()) {
                logger.debug("what are you doing?");
            }
        }

        if (attrPassed != null) {
            attr = true;
            if (!attrPassed.booleanValue()) {
                logger.debug("what are you doing?");
            }
        }

        if (dn || attr) {

            if (other == null) {
                logger.debug("dn or attr are in positive policy, " +
                        "'other' returns not applicable");
                return true;
            } else {
                logger.debug("dn or attr passed, 'other' decides, " +
                        "decision: " + other.booleanValue());
                return other.booleanValue();
            }

        } else {

            if (other != null) {
                logger.debug("dn and attrs are not in positive policy, " +
                        "'other' decides: " + other.booleanValue());
                return other.booleanValue();
            } else {
                logger.debug("dn and attrs are not in positive policy, and" +
                        " 'other' returns not applicable");
                return false;
            }
        }

    }

    private SOAPMessage getMessage(MessageContext messageContext) {

        SOAPMessageContext ctx = (SOAPMessageContext) messageContext;
        SOAPMessage message = ctx.getMessage();
        if (message == null) {
            logger.debug("Empty SOAPEnvelope");
            return null;
        } else {
            return message;
        }
    }

   

    /*
     * Load default gridmap file everytime DN needs to be checked.
     * Check if DN maps to atleast one localUser. 
     */
    private boolean checkGridMapFile(String peerIdentity) 
        throws IOException {
        boolean returnThis = false;
        try {
            GridMap gridMap = ContainerSecurityConfig.getConfig().getGridMap();
            String tempUserName = gridMap.getUserID(peerIdentity);
            if(tempUserName != null) {
                returnThis = true;
            }
        } catch (Exception e) {
            throw new IOException(e.toString());
        }
        return returnThis;
    }

    //couple of ways to both generalize and ensure newly found dns/attrs
    //are persisted, but not worth effort/slower
    private boolean checkAttrFile(String attr)
            throws IOException {

        /* if InitializeException is ignored from initialize */
        if (this.attrConfigFile == null) {
            this.attrConfigFile = findConfigFile(this.attrConfigFileName);
        }

        if (this.attrConfigFile == null) {
            logger.warn("noSecConfig " + this.attrConfigFileName);
            return false;
        }

        // If config file has been modified, re-read all user rights
        if (this.attrConfigLastModified < this.attrConfigFile.lastModified()) {
            logger.debug("file has been modified: " +
                    this.attrConfigLastModified + " < "
                    + this.attrConfigFile.lastModified());
            this.attrs = readConfigFile(this.attrConfigFile);
            this.attrConfigLastModified = this.attrConfigFile.lastModified();
        }

        for (int i = 0; i < this.attrs.length; i++) {
            if (this.attrs[i].trim().equals(attr)) {
                logger.debug("ACCEPT: " + attr);
                return true;
            }
        }

        return false;
    }


    /**
     * Loads policies.  If VomsPDPPolicy is present in the PDPConfig, does
     * not consult policy files.  Otherwise, static files are used.  A
     * timestamp is taken, and the files m-times are compared each call to
     * the PDP.
     * @param config
     * @param name
     * @param id
     * @throws InitializeException
     */
    public void initialize(PDPConfig config,
                           String name,
                           String id)
            throws InitializeException {

        if (config == null) {
            throw new InitializeException("no configuration object");
        }

        /* First check for VomsPDPPolicy, the dynamic policy used by
           accounts. Cannot use current solutions + attribute PDP in
           a chain, because chains are deny overrides */
        Object vomsPolicy = config.getProperty(name, VOMS_PDP_POLICY);
        if (vomsPolicy != null) {
            logger.debug( "voms policy is not null");
            try {
                this.vomsPolicy = (VomsPDPPolicy) vomsPolicy;


                // ** RETURNS **
                // If VomsPDPPolicy is present, files are not used.
                return;

            } catch (ClassCastException e) {
                logger.debug("vomsPDPPolicy is present but it is not the " +
                        "proper class, programmer error");
                throw new InitializeException("no vomsPDPPolicy configuration");
            }
        }
        logger.debug("Using static files");
        // Otherwise, use static files configured in security desc.
/*
        Object configFile = config.getProperty(name, SECURITY_CONFIG_FILE);
        logger.debug("SECURITY_CONFIG_FILE: " + SECURITY_CONFIG_FILE);
        if (configFile == null) {
            throw new InitializeException("no DN authz file configuration");
        }
        
        this.DNconfigFileName = (String) configFile;
        this.DNconfigFile = findConfigFile(this.DNconfigFileName);
        if (this.DNconfigFile != null) {
            try {
                this.dns = readConfigFile(this.DNconfigFile);
                this.DNconfigLastModified = this.DNconfigFile.lastModified();
            } catch (IOException e) {
                throw new InitializeException(e.toString());
            }
        } else {
            throw new InitializeException("No DN authorization file");
        }*/
        
        Object configFile = config.getProperty(name, ATTR_SECURITY_CONFIG_FILE);
        if (configFile == null) {
            logger.warn("no attribute authz file configuration");
            return;
        }
        this.attrConfigFileName = (String) configFile;
        this.attrConfigFile = findConfigFile(this.attrConfigFileName);
        if (this.attrConfigFile != null) {
            try {
                this.attrs = readConfigFile(this.attrConfigFile);
                this.attrConfigLastModified =
                        this.attrConfigFile.lastModified();
            } catch (IOException e) {
                throw new InitializeException(e.toString());
            }
        } else {
            throw new InitializeException("No attribute authorization file");
        }
    }

    protected String[] readConfigFile(File authzFile) throws IOException {
        TextFile file = new TextFile(authzFile.getAbsolutePath());
        String[] result = new String[file.size()];

        for (int i = 0; i < file.size(); i++) {

            result[i] = (String) file.get(i);
        }
        if (logger.isDebugEnabled()) {
            for (int i = 0; i < result.length; i++) {
                logger.debug("result[" + i + "] = " + result[i]);
            }
        }

        return result;
    }

    public void close() throws CloseException {
    }

    protected File findConfigFile(String fileName) {
        if (fileName == null) {
            return null;
        }
        File configFile = null;
        for(int i = 0; i < CONFIG_LOCATIONS.length; i++) {
            String filePath = (CONFIG_LOCATIONS[i].equals("") ? fileName :
                               CONFIG_LOCATIONS[i] + File.separator +
                               fileName);
            configFile = new File(filePath);
            logger.debug("Trying authz file: " + configFile.getAbsolutePath());
            if (configFile.exists()) {
                logger.debug("Security config file found: " +
                             configFile.getAbsolutePath());
                break;
            } else {
                configFile = null;
            }
        }

        /* As a last resort, check globus location */
        if (configFile == null) {
            String globusLocation = ContainerConfig.getGlobusLocation();
            String filePath = globusLocation + File.separator + fileName;
            configFile = new File(filePath);
            logger.debug("Trying authz file: " + configFile.getAbsolutePath());
            if (configFile.exists()) {
                logger.debug("Security config file found: " +
                             configFile.getAbsolutePath());
            } else {
                configFile = null;
            }
        }

        return configFile;
    }

    private class TextFile extends ArrayList {
        //via Eckel
        private TextFile(String fileName) throws IOException {
            super(Arrays.asList(gRead(fileName).split("\n")));
        }
    }

    private static String gRead(String fileName) throws IOException {
        StringBuffer sb = new StringBuffer();
        BufferedReader in =
        new BufferedReader(new FileReader(fileName));
        String s;
        while((s = in.readLine()) != null) {
            sb.append(s);
            sb.append("\n");
        }
        in.close();
        return sb.toString();
    }

    /**
     * Utility methods to assist PDP implementations of the checkCallAndContent
     * method with the content object.  This is not in BanlistPDP because it
     * may be used by other implementations.
     */

    /**
     *
     * @param message
     * @param baddns
     * @param badattrs
     * @return true if additions in message are not banned
     */
    protected static boolean approveAdditions(
                                        SOAPMessage message,
                                        String operation,
                                        String[] baddns,
                                        String[] badattrs) throws Exception {

        if ((message == null) || (operation == null)) {
            return true;
        }

        if ((baddns == null) && (badattrs == null)) {
            return true;
        }

        String[] relevantOps =
                {"{" + WorkspaceConstants.NS + "}addAuthzIdentity",
                 "{" + WorkspaceConstants.NS + "}addAccessIdentity" };

        boolean found = false;
        for (int i = 0; i < relevantOps.length; i++) {
            if (operation.equals(relevantOps[i])) {
                found = true;
                break;
            }
        }

        if (found == false) {
            logger.debug("operation's content does not need to be examined");
            return true;
        } else {
            logger.debug("operation's content is being examined: " + operation);
        }

        try {
            Iterator iter = message.getSOAPBody().getChildElements();

            MessageElement opXML = null;

            if (iter.hasNext()) {
                opXML = ((MessageElement)iter.next());
            }

            if (opXML == null) {
                throw new Exception("SOAP body has" +
                        " no operation content?");
            }

            // note below, remove this and pass XML if portType
            // changes
            iter = opXML.getChildElements();
            Text text = (Text) opXML.getChildren().get(0);
            String content = text.getData();

            if (operation.equals("{" + WorkspaceConstants.NS +
                    "}addAuthzIdentity")) {
                return checkAuthz(content, baddns, badattrs);
            } else if (operation.equals("{" + WorkspaceConstants.NS +
                    "}addAccessIdentity")) {
                return checkAccess(content, baddns, badattrs);
            } else {
                return true; // for compiler
            }
        } catch (SOAPException e) {
            logger.debug(e);
            throw new Exception("",e);
        }
    }

    // NOTE: right now parameters are simple strings; if those
    // portTypes are changed, move to something like this for
    // easy access to new object?
    // (AWorkspaceAxisType)ObjectDeserializer.deserialize(
    //     new InputSource(someXMLpiece),
    //     AWorkspaceAxisType.class);

    private static boolean checkAccess(String content,
                             String[] baddns,
                             String[] badattrs) {
        return simpleCheck(content, baddns, badattrs);
    }


    private static boolean checkAuthz(String content,
                             String[] baddns,
                             String[] badattrs) {
        return simpleCheck(content, baddns, badattrs);
    }

    private static boolean simpleCheck(String content,
                             String[] baddns,
                             String[] badattrs) {

        logger.debug("content = " + content);
        for (int i = 0; i < baddns.length; i++) {
            if (content.equals(baddns[i])) {
                logger.debug(baddns[i] + " is banned");
                return false;
            }
        }
        for (int i = 0; i < badattrs.length; i++) {
            if (content.equals(badattrs[i])) {
                logger.debug(badattrs[i] + " is banned");
                return false;
            }
        }

        // not banned
        return true;

    }

}
