/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.factory;

import org.apache.axis.types.Duration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.util.Util;
import org.globus.workspace.LimitException;
import org.globus.workspace.WorkspaceConstants;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.WorkspaceUtil;
import org.globus.workspace.factory.WorkspaceFactoryResource;
import org.globus.workspace.generated.types.AccountInformation;
import org.globus.workspace.generated.types.AccountReference;
import org.globus.workspace.impl.service.WorkspaceHome;
import org.globus.workspace.impl.service.WorkspaceResourceImpl;
import org.globus.wsrf.Constants;
import org.globus.wsrf.InvalidResourceKeyException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.utils.FilePersistenceHelper;

import javax.naming.InitialContext;
import javax.xml.namespace.QName;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

public class WorkspaceFactoryResourceImpl
    implements WorkspaceFactoryResource {

    static Log logger = 
        LogFactory.getLog(WorkspaceFactoryResourceImpl.class.getName());
    
    private ResourcePropertySet propSet;
    private SimpleResourceKey key = null;
    private Duration defaultTTL;
    private Duration maxTTL;
    private Integer maxWorkspaces;

    // For persistence
    private FilePersistenceHelper persistenceHelper;
    private final static String TMP_FILE_PREFIX = "workspacefactory";
    private final static String TMP_FILE_SUFFIX = ".tmp";

    /**
     * Initializes the factory resource with TTL parameters.  The integers
     * found in the jndi-config.xml file are converted to (objects that
     * map to) xsd:duration.
     *
     * The setting
     *
     * @param key
     * @param ttl default TTL for an account, in minutes.  Given to account
     * if the client does not request TTL
     * @param maxttl maximum TTL for an account, in minutes.  If a client
     * requests a TTL higher than this, it is rejected.
     * @throws Exception
     * @see org.globus.workspace.impl.service.WorkspaceResourceImpl#createAccount(org.globus.workspace.generated.types.AccountInformation)
     */
    public synchronized void initialize(SimpleResourceKey key, Integer ttl,
                                        Integer maxttl, Integer maxWorkspaces)
        throws Exception {
        this.key = key;
        this.reinitialize(ttl,maxttl,maxWorkspaces);
    }

    /**
     * This allows the persisted factory resource configurations to be
     * overwritten in the case where the jndi-config.xml values were changed.
     * Otherwise, the user would need to deleted the file in the persisted
     * directory.  This is a one-time container bootup cost, and a small cost,
     * so this can be revisited later (using a database for multiple account
     * information would eliminate the issue).
     * @param ttl
     * @param maxttl
     * @param maxWorkspaces
     */
    public synchronized void reinitialize(Integer ttl, Integer maxttl,
                                          Integer maxWorkspaces) {
        this.defaultTTL = WorkspaceUtil.minutesToDuration(ttl.intValue());
        this.maxTTL = WorkspaceUtil.minutesToDuration(maxttl.intValue());
        this.maxWorkspaces = maxWorkspaces;
        WorkspaceResourceImpl.getPersistenceAdapter().setMaxWorkspaces(
                this.maxWorkspaces.intValue());

        logger.debug("Resource created for mechanism " + key.getValue());
        setRPs();
    }

    private void setRPs() {
        // maxWorkspaces are not advertised

        this.propSet = new SimpleResourcePropertySet(RP_SET);
        ResourceProperty prop =
            new SimpleResourceProperty(new QName(WorkspaceConstants.NS,
                                                 "DefaultTTL"));
        prop.add(this.defaultTTL);
        this.propSet.add(prop);

        prop =
            new SimpleResourceProperty(new QName(WorkspaceConstants.NS,
                                                 "MaxTTL"));
        prop.add(this.maxTTL);

        prop =
            new SimpleResourceProperty(new QName(WorkspaceConstants.NS,
                                                 "MaxWorkspaces"));
        //todo: not tested, BigInteger instead?
        prop.add(this.maxWorkspaces);
        this.propSet.add(prop);
    }

    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    /**
     * Create account resource using the resource home in WorkspaceService.
     * @param accountInfo defined in wsdl
     * @param callerUserName defined in wsdl
     * @return
     * @throws WorkspaceException
     * @throws LimitException too many accounts for caller have been created already
     * @see org.globus.workspace.impl.service.WorkspaceHome#create(org.globus.workspace.generated.types.AccountInformation, String, String, String, org.apache.axis.types.Duration, org.apache.axis.types.Duration)
     */
    public AccountReference createAccount(AccountInformation accountInfo,
                                     String callerUserName)
        throws WorkspaceException, LimitException {
        
        WorkspaceHome home = null;
        try {
            home = getServiceHome();
        } catch (Exception exp) {
            throw new WorkspaceException("Error getting service home",
                                               exp);
        }

        String callerDN = SecurityManager.getManager().getCaller();
        logger.debug("caller is " + callerDN);
        // type casting to string here. Key type should not be altered
        // in jndi config.
        AccountReference acct = home.create(accountInfo,
                (String)key.getValue(), callerDN, callerUserName,
                this.defaultTTL, this.maxTTL);

        return acct;
    }

    public void store() throws ResourceException {


        /* Only key and the multipleaccounts info is persisted so that one
        can change the factory jndi config w/o needing to delete the
        persistence file manually */

        FileOutputStream fos = null;
        File tmpFile = null;
        try {
            tmpFile = File.createTempFile(TMP_FILE_PREFIX, TMP_FILE_SUFFIX,
                                          getPersistenceHelper()
                                          .getStorageDirectory());

            if (!Util.setFilePermissions(tmpFile.getAbsolutePath(), 700)) {
                logger.warn("Check file permissions on "
                            + tmpFile.getAbsolutePath());
            }
            fos = new FileOutputStream(tmpFile);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            oos.writeObject(this.key);
        //    oos.writeObject(multipleSingleton.getData());
            //logger.debug(multipleSingleton.toString());
        } catch (Exception e) {
            if (tmpFile != null) {
                tmpFile.delete();
            }
            throw new ResourceException("Error storing resource ", e);
        } finally {
            if (fos != null) {
                try { fos.close();} catch (Exception ee) {}
            }
        }

        // Overwrite old file
        File file = getKeyAsFile((String)key.getValue());
        if (file.exists()) {
            file.delete();
        }

        if (!tmpFile.renameTo(file)) {
            tmpFile.delete();
            throw new ResourceException("Error storing resource");
        }
        // Check if rename to changes the file permissions.
        if (!Util.setFilePermissions(file.getAbsolutePath(), 700)) {
            logger.warn("Check file permissions on "
                        + tmpFile.getAbsolutePath());
        }
    }

    public void load(ResourceKey key) throws ResourceException, NoSuchResourceException, InvalidResourceKeyException {
        logger.debug("load called");

        File file = getKeyAsFile((String)key.getValue());
        if (!file.exists()) {
            throw new NoSuchResourceException();
        }

        /* Only key and the multipleaccounts info is persisted so that one
        can change the factory jndi config w/o needing to delete the
        persistence file manually */

        FileInputStream fis = null;
        try {
            fis = new FileInputStream(file);
            ObjectInputStream ois = new ObjectInputStream(fis);

            this.key = (SimpleResourceKey)ois.readObject();
            //Map counterdata = (Map)ois.readObject();
            //multipleSingleton.setData(counterdata);
            //logger.debug(multipleSingleton.toString());
        } catch (Exception e) {
            throw new ResourceException("Error loading resource", e);
        } finally {
            if (fis != null) {
                try { fis.close(); } catch (Exception ee) {}
            }
        }
    }

    /**
     * Get service home.
     * This causes a delay the very first time the method is invoked,
     * TODO: try activate-on-startup to move delay to container boot
     * @return
     * @throws WorkspaceException
     */
    protected WorkspaceHome getServiceHome()
        throws WorkspaceException{
        
        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         WorkspaceConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (WorkspaceHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        WorkspaceConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new WorkspaceException(exp);
        }
    }

    protected synchronized FilePersistenceHelper getPersistenceHelper() {
        if (this.persistenceHelper == null) {
            try {
                this.persistenceHelper =
                    new FilePersistenceHelper(getClass(), ".ser");
            } catch (Exception e) {
                throw new RuntimeException(e.getMessage());
            }
        }
        return this.persistenceHelper;
    }

    private File getKeyAsFile(Object key)
        throws InvalidResourceKeyException {
        if (key instanceof String) {
            return getPersistenceHelper().getKeyAsFile(key);
        } else {
            throw new InvalidResourceKeyException();
        }
    }

}
