/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.factory;

import org.globus.workspace.WorkspaceUtil;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.LimitException;
import org.globus.workspace.factory.WorkspaceFactoryResource;

import org.globus.workspace.generated.types.*;

import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceContextException;

import org.globus.wsrf.security.SecurityManager;
import org.globus.wsrf.security.SecurityException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.MessageContext;


public class WorkspaceFactoryService {

    static Log logger = 
        LogFactory.getLog(WorkspaceFactoryService.class.getName());


    /**
     * @return user name of requester, as mapped in the configured
     * grid map file.
     * @throws AccountFaultType
     */
    private String getUserName() throws AccountFaultType {
        MessageContext msgCtx  = MessageContext.getCurrentContext();
        String userName = null;
        try {
            String[] names = SecurityManager.getManager(msgCtx)
                .getLocalUsernames();
            if (names != null) {
                userName = names[0];
            }
        } catch (SecurityException exp) {
            throw WorkspaceUtil.makeFault("Security manager: ", exp);
        }
        return userName;
    }

    /**
     * Invoke createResource on appropriate factory resource.
     * @param accountInfo defined in wsdl
     * @return AccountReference defined in wsdl
     * @throws AccountFaultType Problem in general, TODO: errors campaign
     * @throws AccountLimitFaultType Too many workspaces for client.
     * @throws WorkspaceException Other.
     */
    public AccountReference createAccount(AccountInformation accountInfo)
            throws AccountFaultType, AccountLimitFaultType, WorkspaceException {

        logger.debug("Create account " + accountInfo.getClass().getName());
        WorkspaceFactoryResource resource = null;
        try {
            resource = getFactoryResource();
        } catch (WorkspaceException exp) {
            throw WorkspaceUtil.makeFault("Error getting resource", exp);
        }

        /* not needed by some implementations, such as pool accounts */
        String userName = getUserName();

        logger.debug("User name is " + userName);

        AccountReference acct = null;
        try {
            acct = resource.createAccount(accountInfo, userName);
        } catch (Exception exp) {
            logger.error("Error creating resource", exp);

            if (exp instanceof LimitException) {
                throw WorkspaceUtil.makeLimitFault("account limit reached",exp);
            } else {
                throw WorkspaceUtil.makeFault("error", exp);
            }
        }

        // Store on resource is called here rather than on create by
        // resource since duplicate key check will always be true.
        /*try {
            resource.store();
        } catch (ResourceException exp) {
            throw new WorkspaceException(exp);
        }*/
        
        return acct;
    }

    /**
     * Creates a set of accounts.  Just a wrapper to create.  Use is advised
     * against, as the failure semantics are best left to the caller.
     *
     * @param accountsInfo defined in wsdl
     * @return AccountReference[]  defined in wsdl
     * @throws AccountFaultType
     * @throws AccountLimitFaultType
     */
    public AccountReference[] createAccountSet(CreateAccountSet accountsInfo)
        throws AccountFaultType, AccountLimitFaultType {

        WorkspaceFactoryResource resource = null;
        try{
            resource = getFactoryResource();
        } catch (WorkspaceException exp) {
            throw WorkspaceUtil.makeFault("Error getting resource", exp);
        }
	    String userName = getUserName();

        AccountInformation accounts[] =
            accountsInfo.getAccounts();
        AccountReference accts[] =
            new AccountReference[accounts.length];
        logger.debug("numaccts " + accounts.length);
        for (int i=0; i < accounts.length; i++) {
            try {
                accts[i] = resource.createAccount(accounts[i], userName);
                if (logger.isDebugEnabled()) {
                    logger.debug("Account # " + i + ": \n");
                    logger.debug("\tEPR " + accts[i].getEndpointReference());
                    logger.debug("\tTerminationTime: " +
                            accts[i].getActualTerminationTime());
                }
            } catch (LimitException exp) {
                 String err = "Error creating multiple account, one of the " +
                         "accounts is a duplicate";

                //TODO: decide on behavior here, continue trying to
                // create other accounts? Rollback?
                // (todo: drop set alltogether)

                throw WorkspaceUtil.makeLimitFault(err,exp);

            } catch (WorkspaceException exp) {
                String err = "Error creating multiple account";
                logger.error(err);
                throw WorkspaceUtil.makeFault(err, exp);
            }
        }
        return accts;
    }

    private WorkspaceFactoryResource getFactoryResource()
        throws WorkspaceException {
        try {
            ResourceContext context = ResourceContext.getResourceContext();
            return (WorkspaceFactoryResource)context.getResource();
        } catch (ResourceContextException exp) {
            String err = "Error getting factory resource";
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        } catch (ResourceException exp) {
            String err = "Error getting factory resource";
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        }
    }


}
