/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.service;

import org.globus.workspace.WorkspaceException;
import org.globus.workspace.WorkspaceUtil;
import org.globus.workspace.ReturnException;
import org.globus.workspace.LimitException;
import org.globus.workspace.database.PersistenceAdapter;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.types.Duration;

import java.io.File;
import java.io.ObjectInputStream;

import java.util.StringTokenizer;
import java.util.ArrayList;

public class LCMAPSResource extends WorkspaceResourceImpl {

    static Log logger =
        LogFactory.getLog(LCMAPSResource.class.getName());

    /* set in initialize() */
    protected String executable = null;
    protected String conffile = null;

    protected final static String
            BASE_COMMAND = "bin" + File.separator + "lcmaps_poolindex";
    protected final static String
            POOLINDEX_COMMAND = "-return_poolindex";
    protected final static String
            COUNTER_FLAG = "-count";
    protected final static String
            DELETE_BY_POOLINDEX_COMMAND = "-unlease_poolindex";
    protected final static String
            CONF_FLAG = "-conf";

    /* set in createAccountUsingMech(), used to make life easier with interface
       to LCMAPS.  Also used as resource key. Meant to be immutable once set. */
    protected String poolindex;

    protected int counter = -1;
    protected boolean useMultipleLeases = false;

    //overriding to set executable
    public void initialize(String _creationMechanism,
                           String _callerDN,
                           String _callerUserName,
                           String _globusLocation,
                           Duration _factoryTTL,
                           Duration _maxTTL,
                           String _conffile,
                           boolean _usemult)
        throws WorkspaceException, LimitException {
        logger.debug("initializing lcmaps resource");
        super.initialize(_creationMechanism, _callerDN, _callerUserName,
                         _globusLocation, _factoryTTL, _maxTTL);
        
        logger.debug("initializing lcmaps resource");
        if (_globusLocation.endsWith(File.separator)) {
            this.executable = _globusLocation + BASE_COMMAND;
        } else {
            this.executable = _globusLocation + File.separator + BASE_COMMAND;
        }
        logger.debug("Executable " + executable);

        // _conffile is checked for null in WorkspaceHome
        // this doesn't work on Windows, but neither does lcmaps_poolindex
        if (_conffile.startsWith("/")) {
            this.conffile = _conffile;
        } else {
            this.conffile = _globusLocation + File.separator + _conffile;
        }
        logger.debug("LCMAPS poolindex configuration file " + this.conffile);

        this.useMultipleLeases = _usemult;
    }

    /*
     * BEGIN setuid interface related methods
     */

    /**
     * Calls lcmaps_poolindex, the LCMAPS driver program, -> poolindex
     * If roles are present, passes those as well.
     * @throws WorkspaceException
     */
    protected void createAccountUsingMech()
            throws WorkspaceException, LimitException {
        
        if (this.useMultipleLeases) {
            //primaryAttr will be null if no attr info

            //this can throw LimitException
            logger.debug("callerDn:" + this.callerDN);
            logger.debug("primaryAttr:"+ this.primaryAttr);
            this.counter = getPersistenceAdapter().getNextCounter(
                    this.callerDN, this.primaryAttr);
            if (this.counter < 0) {
                throw new WorkspaceException("counter is configured " +
                    "but none available");
            }
            //a mess, problem goes away w/ DB
           // WorkspaceFactoryHome.storeRsrc();
        }


        String command[] = getCommandArray();

        String returnString = null;
        try {
            returnString = WorkspaceUtil.runCommand(command);
        } catch (ReturnException e) {
            translateReturnValue(e);
        }

        logger.debug("Return string:\n" + returnString);

        // Required: poolindex, unixname, primary unix group
        // Optional: secondary groups
        StringTokenizer st = new StringTokenizer(returnString,"\n");
        if (st.hasMoreTokens()) {
            this.poolindex = st.nextToken();
            if (st.hasMoreTokens()) {
                this.localName = st.nextToken();
                if (st.hasMoreTokens()) {
                    this.primaryGroup = st.nextToken();
                } else {
                    throw new WorkspaceException("no primary group returned " +
                            "from pool impl");
                }
                ArrayList secs = new ArrayList();
                while (st.hasMoreTokens()) {
                    secs.add(st.nextToken());
                }
                if (secs.size() > 0) {
                    String[] tmp = new String[secs.size()];
                    this.secondaryGroups = (String[]) secs.toArray(tmp);
                }
            } else {
                throw new WorkspaceException("no poolindex returned from " +
                        "pool impl");
            }
        } else {
            throw new WorkspaceException("no information returned from " +
                    "pool impl, yet it succeeded. (?)");
        }

        // Look for bad output, just to be sure (i.e., weird C output of \n\n)
        String tmp = "";
        if (this.poolindex.equals(tmp) ||
                this.localName.equals(tmp) ||
                this.primaryGroup.equals(tmp)) {
            throw new WorkspaceException("bad output from pool impl");
        }
        if (this.secondaryGroups != null) {
            for (int i = 0; i < this.secondaryGroups.length; i++) {
                if (this.secondaryGroups[i].equals(tmp)) {
                    throw new WorkspaceException("bad output from pool impl");
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("poolindex     :: " + this.poolindex);
            logger.debug("Local name    :: " + this.localName);
            logger.debug("Primary group :: " + this.primaryGroup);
            if (this.secondaryGroups != null) {
                for (int i = 0; i < this.secondaryGroups.length; i++) {
                    logger.debug("SecdGroup " + (i+1) + "  :: "
                            + this.secondaryGroups[i]);
                }
            }
        }

    }

    /**
     * todo: could refactor, use arraylist
     * @return
     * @throws WorkspaceException
     */
    private String[] getCommandArray() throws WorkspaceException {
        String[] command = null;

        if (this.useMultipleLeases) {

            /* no primaryAttr means no attributes at all -- either VOMS is
               not being employed or proxy has no attributes */
            if (this.primaryAttr == null) {
                command =  new String[7];
                command[0] = executable;
                command[1] = CONF_FLAG;
                command[2] = this.conffile;
                command[3] = POOLINDEX_COMMAND;
                command[4] = COUNTER_FLAG;
                command[5] = Integer.toString(this.counter);
                command[6] = this.callerDN;
            } else if ((this.creatorAttributesVector != null)
                    &&(this.creatorAttributesVector.size() > 0)) {
                command =  new String[7 + this.creatorAttributesVector.size()];
                command[0] = executable;
                command[1] = CONF_FLAG;
                command[2] = this.conffile;
                command[3] = POOLINDEX_COMMAND;
                command[4] = COUNTER_FLAG;
                command[5] = Integer.toString(this.counter);
                command[6] = this.callerDN;
                for (int i = 0 ; i < this.creatorAttributesVector.size(); i++) {
                    String test = (String) this.creatorAttributesVector.get(i);
                    if (test != null) {
                        command[i+7] = test;
                    } else {
                        throw new WorkspaceException("found a null attribute " +
                                "in rolesVector: index #" + i);
                    }
                }
            } else {
                // this can only happen if VomsCredentialPIP is used but does
                // not parse roles for some reason *and* primaryAttr is set
                // somehow (a developer extending is the only way this could
                // happen)
                throw new WorkspaceException("scheme not supported");
            }

            return command;


        } else {
            /* no primaryAttr means no attributes at all -- either VOMS is
               not being employed or proxy has no attributes */
            if (this.primaryAttr == null) {
                command =  new String[5];
                command[0] = executable;
                command[1] = CONF_FLAG;
                command[2] = this.conffile;
                command[3] = POOLINDEX_COMMAND;
                command[4] = this.callerDN;
            } else if ((this.creatorAttributesVector != null)
                    &&(this.creatorAttributesVector.size() > 0)) {
                command =  new String[5 + this.creatorAttributesVector.size()];
                command[0] = executable;
                command[1] = CONF_FLAG;
                command[2] = this.conffile;
                command[3] = POOLINDEX_COMMAND;
                command[4] = this.callerDN;
                for (int i = 0 ; i < this.creatorAttributesVector.size(); i++) {
                    String test = (String) this.creatorAttributesVector.get(i);
                    if (test != null) {
                        command[i+5] = test;
                    } else {
                        throw new WorkspaceException("found a null attribute " +
                                "in rolesVector: index #" + i);
                    }
                }
            } else {
                // this can only happen if VomsCredentialPIP is used but does
                // not parse roles for some reason *and* primaryAttr is set
                // somehow (a developer extending is the only way this could
                // happen)
                throw new WorkspaceException("scheme not supported");
            }

            return command;
        }
    }

    /**
     * Calls lcmaps_poolindex, the LCMAPS driver program, -> delete
     * If roles are present, passes those as well.
     * @throws WorkspaceException
     */
    protected void deleteAccountUsingMech() throws WorkspaceException {

        if (this.useMultipleLeases) {
                WorkspaceResourceImpl.getPersistenceAdapter().retireCounter
                (this.localName);
            //a mess, problem goes away w/ DB
           // WorkspaceFactoryHome.storeRsrc();
        }

        String command[] = null;

        if (this.poolindex == null) {
            throw new WorkspaceException("cannot delete, no poolindex");
        }

        if (this.conffile == null) {
            throw new WorkspaceException("cannot delete, no conffile");
        }

        command =  new String[5];
        command[0] = executable;
        command[1] = CONF_FLAG;
        command[2] = this.conffile;
        command[3] = DELETE_BY_POOLINDEX_COMMAND;
        command[4] = this.poolindex;

        String returnString = null;
        try {
            //returnString is assigned but not accessed
            returnString = WorkspaceUtil.runCommand(command);
            if (logger.isDebugEnabled()) {
                logger.debug(returnString);
            }
        } catch (ReturnException e) {
            translateReturnValue(e);
        }

    }

    /**
     * Parses return value and groups exceptions accordingly.
     * TODO: error campaign
     * @param e
     * @throws WorkspaceException
     */
    private void translateReturnValue(ReturnException e)
            throws WorkspaceException {

        /* #define GENERAL_FAILURE 255 */
        if (e.retval == 255) {
            if (e.stderr != null) {
            throw new WorkspaceException("failure running pool" +
                    " account program\nSTDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("failure running pool" +
                        " account program");
            }
        }
        /* #define ENVIRONMENT 1 */
        if (e.retval == 1) {
            if (e.stderr != null) {
            throw new WorkspaceException("pool account program" +
                    " setup problem\nSTDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("pool account program" +
                    " setup problem");
            }
        }
        /* #define SYNTAX 2 */
        if (e.retval == 2) {
            if (e.stderr != null) {
            throw new WorkspaceException("pool account service" +
                    " setup problem\nSTDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("pool account service" +
                    "setup problem");
            }
        }
        /* #define BADARG 3 */
        if (e.retval == 3) {
            if (e.stderr != null) {
            throw new WorkspaceException("problem with parameters, STDERR: "
                    + e.stderr);
            } else {
            throw new WorkspaceException("problem with parameters");
            }
        }
        /* #define POOL_INDEX_FAILURE 4 */
        if (e.retval == 4) {
            if (e.stderr != null) {
            throw new WorkspaceException("problem with LCMAPS, STDERR: "
                    + e.stderr);
            } else {
            throw new WorkspaceException("problem with LCMAPS");
            }
        }
        /* #define NO_POOL_FOR_USER 5 */
        /* //not supported by LCMAPS
        if (e.retval == 5) {
            if (e.stderr != null) {
            throw new WorkspaceException("found user for DN, but no pool
            account, STDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("found user for DN, but no pool
            account,");
            }
        } */
        /* #define NO_USER 6 */
        /* //not supported by LCMAPS
        if (e.retval == 6) {
            if (e.stderr != null) {
            throw new WorkspaceException("no mapping for this DN, STDERR: "
            + e.stderr);
            } else {
            throw new WorkspaceException("no mapping for this DN");
            }
        } */
        /* #define POOL_FULL 7 */
        /* //not supported by LCMAPS
        if (e.retval == 7) {
            if (e.stderr != null) {
            throw new WorkspaceException("pool is exhausted, STDERR: "
            + e.stderr);
            } else {
            throw new WorkspaceException("pool is exhausted");
            }
        } */



        /* This is returned if LCMAPS successfully initialized but the
            actual call to lcmaps_return_poolindex_without_gsi() failed.
            This is in all likelihood because the dn+attr combination was
            denied a lease, *especially* if you have had at least one
            successful lease by LCMAPS
          */
        /* #define GRIDMAPDIR_NO_ASSOCIATION 8 */
        if (e.retval == 8) {
            if (e.stderr != null) {
            throw new WorkspaceException("no lease available, STDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("no lease available");
            }
        }

        /*****
         * EXPIRY_PROG_FAILURE & EXPIRY_PROG_CHECK are both errors that happen
         * after a lease has been unleased and quarantined.  Perhaps don't throw
         * exception because this resource, which represents the lease, must
         * succeed in its destruction?
         */

        /* #define EXPIRY_PROG_FAILURE 9 */
        if (e.retval == 9) {
            if (e.stderr != null) {
            throw new WorkspaceException("expire account program failure, " +
                    "STDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("expire account program failure");
            }
        }

        /* #define EXPIRY_PROG_CHECK 10 */
        if (e.retval == 10) {
            if (e.stderr != null) {
            throw new WorkspaceException("expire account environment " +
                    "problem, STDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("expire account environment " +
                    "problem");
            }
        }

        /* #define NO_GROUP 11 */
        /* //not supported by LCMAPS
        if (e.retval == 11) {
            if (e.stderr != null) {
            throw new WorkspaceException("no group mapping for attribute,
            STDERR: " + e.stderr);
            } else {
            throw new WorkspaceException("no group mapping for attribute");
            }
        } */

        /* others? */
        if (e.stderr != null) {
        throw new WorkspaceException("pooling problem, implementation " +
                "returns error #" + e.retval + ", STDERR: " + e.stderr);
        } else {
        throw new WorkspaceException("pooling problem, implementation" +
                "returns error #" + e.retval);
        }

    }

    public String getExecutable() {
        return this.executable;
    }
    public String getPoolIndex(){
        return this.poolindex;
    }
    public String getConfFile() {
        return this.conffile;
    }
    public boolean isUseMultipleLeases() {
        return this.useMultipleLeases;
    }
    public int getCounter() {
        return this.counter;
    }
    
    /*
     * END setuid interface related methods
     */

    protected void storeInMech(String localName, PersistenceAdapter adapter)  throws Exception {
        adapter.storeLCMapsResource(localName, this);
        
    }
    protected void loadInMech(String localName, PersistenceAdapter adapter) throws Exception {
        adapter.loadLCMAPSResource(localName, this);
    }

    protected void loadInMech(ObjectInputStream ois) throws Exception {

        this.executable = (String)ois.readObject();
        this.poolindex = (String)ois.readObject();
        this.conffile = (String)ois.readObject();
        this.useMultipleLeases = ((Boolean)ois.readObject()).booleanValue();
        this.counter = ((Integer)ois.readObject()).intValue();
    }

    /**
     * @param executable
     */
    public void setExecutable(String executable) {
        this.executable = executable;
    }

    /**
     * @param poolIndex
     */
    public void setPoolIndex(String poolIndex) {
        this.poolindex = poolIndex;
    }
    
    public void setConfFile(String conffile) {
        this.conffile = conffile;
    }
    
    public void setUseMultipleLeases(boolean useMultipleLeases) {
        this.useMultipleLeases = useMultipleLeases;
    }
    
    public void setCounter(int counter) {
        this.counter = counter;
    }

}
