/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.service;

import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.types.Duration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.workspace.LimitException;
import org.globus.workspace.WorkspaceConstants;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.database.DBUtil;
import org.globus.workspace.database.PersistenceAdapter;
import org.globus.workspace.database.WorkspaceDatabaseException;
import org.globus.workspace.generated.PolicyQuery;
import org.globus.workspace.generated.types.AccountInformation;
import org.globus.workspace.generated.types.AccountReference;
import org.globus.workspace.service.WorkspaceResource;
import org.globus.wsrf.Resource;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.jndi.Initializable;
import org.globus.wsrf.utils.AddressingUtils;

import javax.xml.namespace.QName;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.lang.reflect.Method;

public class WorkspaceHome extends ResourceHomeImpl implements Initializable {

    static final QName RESOURCE_KEY_QNAME = new QName(WorkspaceConstants.NS,
            "AccountKey");

    static Log logger = LogFactory.getLog(WorkspaceHome.class.getName());

    String globusLocation = ContainerConfig.getGlobusLocation();

    private static PersistenceAdapter persistenceAdapter = null;

    /**
     * @see this.getConffile()
     */
    String conffile = null;

    /**
     * @see this.getLcmapsMultipleLeases()
     */
    String lcmapsMultipleLeases = null;

    private String dnPolicyFileName = null;

    private  String attrPolicyFileName = null;

    private String poolFileLocation = null;

    private boolean initializeLeasedAccounts = false;


    // for JNDI -> leasedExpireProgram
    private String leasedAccountsResourceExpiryProgram = null;


    public synchronized void initialize() throws Exception {
        logger.debug("initializing home" );
        this.keyTypeName = RESOURCE_KEY_QNAME;
        super.initialize();
        initialize_persistence();
        // Look in persistence directory
        // List listOfKeys = getPersistenceHelper().list();
        List listOfKeys = persistenceAdapter.getActiveWorkspaceIDs();
        if (listOfKeys != null) {
            Iterator keyIterator = listOfKeys.iterator();
            int counter = 0;
            while (keyIterator.hasNext()) {
                String keyValue = (String) keyIterator.next();
                logger.debug(keyValue);
                ResourceKey key = getKey(keyValue);
                Resource resource = createNewInstance();

                if (WorkspaceResource.class
                        .isAssignableFrom(this.resourceClass)) {
                    ((WorkspaceResource) resource).load(key);
                }
                logger.debug("Adding key " + key);
                this.add(key, resource);
                counter++;
                // Maybe check termination time and not add if expired.

                // (for now, sweeper thread defaults to 1s, so they will just
                // expire after a second)
            }

            if (counter > 1) {
                logger.info("WORKSPACES RESTORED: " + counter + " workspaces "
                        + "were loaded from the persistence store.");
            } else if (counter == 1) {
                logger.info("WORKSPACE RESTORED: one workspace "
                        + "was loaded from the persistence store.");
            } else {
                logger.debug("no workspaces found in persistence store");
            }

        }

        // static method not available via interface/instanceof
        Method[] methods = this.resourceClass.getMethods();
        Object[] args = {this};
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].getName().equals("initMech")) {
                methods[i].invoke(null, args);
                break;
            }
        }

    }

    private void initialize_persistence() {

        if (logger.isTraceEnabled()) {
            logger.trace("initialize_persistence()");
        }

        MessageContext ctx = MessageContext.getCurrentContext();
        DBUtil.init((String) ctx
                .getProperty(org.apache.axis.Constants.MC_CONFIGPATH));
        logger.debug("path: " + WorkspaceConstants.PERSISTENCE_ADAPTER);
        // for now there is only one
        persistenceAdapter = DBUtil
                .getDBAdapter(WorkspaceConstants.PERSISTENCE_ADAPTER);
        if (!this.initializeLeasedAccounts) {
            try {
                persistenceAdapter.initializeLeasedAccounts(this.dnPolicyFileName,
                        this.attrPolicyFileName, this.poolFileLocation);
                this.initializeLeasedAccounts = true;
            } catch (Exception e) {
                logger.error("Failed to load leased account info");
            }
        }
       logger.debug("exiting initialize_persistence");
    }

    public String[] queryAccessPolicy(PolicyQuery query) {
        if (query == null) {
            return null;
        }
        boolean all = query.isAllMappings();
        String dn = query.getDn();
        String attribute = query.getAttribute();
        ArrayList localUsernames = new ArrayList();
        try {
            localUsernames = persistenceAdapter.queryAccessPolicy(dn,
                    attribute, all);
        } catch (WorkspaceDatabaseException wsdbex) {
            logger.error("Error getting access policies", wsdbex);
        }

        if (localUsernames.size() > 0) {
            return ((String[]) localUsernames.toArray(new String[0]));
        } else {
            return null;
        }
    }

    /**
     * Make sure these entries are not in any workspace's access policies.
     * 
     * @param entries
     */
    public void removeFromAccessPolicy(String[] entries)
            throws WorkspaceException {
        try {
            if (entries != null && entries.length > 0) {
                WorkspaceHome.persistenceAdapter
                        .removeFromAccessPolicy(entries);
            }
        } catch (WorkspaceDatabaseException wsdbex) {
            throw new WorkspaceException(wsdbex);
        }

    }

    /**
     * @param dns
     * @param attrs
     */
    public void removeFromMgmtPolicy(String[] dns, String[] attrs)
            throws WorkspaceException {

        WorkspaceResourceImpl resource;

        Iterator i = resources.values().iterator();

        // removeAuthzIdentity will store() if necessary

        while (i.hasNext()) {
            resource = (WorkspaceResourceImpl) i.next();
            if (dns != null) {
                for (int j = 0; j < dns.length; j++) {
                    resource.removeAuthzIdentity(dns[j]);
                }
            }
            if (attrs != null) {
                for (int j = 0; j < attrs.length; j++) {
                    resource.removeAuthzIdentity(attrs[j]);
                }
            }
        }
    }

    /**
     * Creates a resource per account. The key is based on the local user name
     * obtained by the backend.
     * 
     * @param accountInfo
     * @param creationMechanism
     *            the creation *context*. "mechanism" is the wrong way to think
     *            of this now. TODO: rename. The String representing the
     *            creation context is a key to which factory resource to call.
     *            Currently all information is passed with the create call and
     *            the key is not necessary. But if the factory resource(s)
     *            contained dynamic information (like they used to), we should
     *            store the key.
     * @param callerDN
     * @param callerUsername
     *            used by some backends
     * @return AccountReference defined in wsdl
     * @throws WorkspaceException
     * @throws LimitException
     *             too many accounts created
     */
    public synchronized AccountReference create(AccountInformation accountInfo,
                                                String creationMechanism, String callerDN, String callerUsername,
                                                Duration factoryTTL, Duration maxTTL) throws WorkspaceException,
                                                                                             LimitException {

        WorkspaceResource resource;
        try {
            resource = (WorkspaceResource) createNewInstance();

            // LCMAPS implementation uses extra arguments now
            // for the configuration file and optional useMultipleLease flag
            if (resource instanceof LCMAPSResource) {

                if (this.conffile == null) {
                    throw new WorkspaceException("LCMAPS implementation "
                            + "requires conffile configuration in the JNDI "
                            + "configuration file.");
                }

                // could move to initialize and store boolean ...
                boolean usemult = false;
                if (this.lcmapsMultipleLeases != null) {
                    if (this.lcmapsMultipleLeases.equalsIgnoreCase("true")) {
                        usemult = true;
                    }
                }

                ((LCMAPSResource) resource).initialize(creationMechanism,
                        callerDN, callerUsername, this.globusLocation,
                        factoryTTL, maxTTL, this.conffile, usemult);
            } else if (resource instanceof LeasedAccountsResource) {
                if (!this.initializeLeasedAccounts) {
                    persistenceAdapter.initializeLeasedAccounts(this.dnPolicyFileName,
                            this.attrPolicyFileName, this.poolFileLocation);
                    this.initializeLeasedAccounts = true;
                }
                resource
                .initialize(creationMechanism, callerDN,
                        callerUsername, this.globusLocation,
                        factoryTTL, maxTTL);
            }else {
                resource
                        .initialize(creationMechanism, callerDN,
                                callerUsername, this.globusLocation,
                                factoryTTL, maxTTL);
            }
        } catch (ResourceException exp) {
            logger.error(exp);
            throw new WorkspaceException(exp);
        }

        // set by creation mechanism and used as key. (Should be unique)
        String localUsername = resource.createAccount(accountInfo);

        if (localUsername == null) {
            throw new WorkspaceException("null mapping returned");
        }

        // Add resource to hashmap
        ResourceKey key = getKey(localUsername);
        logger.debug("new key, " + key.getName().toString());

        this.add(key, resource);
        // construct return object
        AccountReference acct = new AccountReference();
        EndpointReferenceType epr = getEPR(key);
        acct.setEndpointReference(epr);
        acct.setActualTerminationTime(resource.getTerminationTime());

        return acct;
    }

    /**
     * @param localName
     * @return ResourceKey
     */
    public static ResourceKey getKey(String localName) {
        return new SimpleResourceKey(RESOURCE_KEY_QNAME, localName);
    }

    /**
     * @param key
     * @return
     * @throws WorkspaceException
     */
    protected EndpointReferenceType getEPR(ResourceKey key)
            throws WorkspaceException {

        EndpointReferenceType epr;
        try {
            String addr = ServiceHost.getBaseURL()
                    + WorkspaceConstants.SERVICE_PATH;
            epr = AddressingUtils.createEndpointReference(addr, key);
        } catch (Exception exp) {
            String err = "Error creating endpoint reference";
            throw new WorkspaceException(err, exp);
        }
        return epr;
    }

    /**
     * For JNDI config parameter named 'conffile'
     */
    public void setConffile(String conffile) {
        this.conffile = conffile;
    }

    public String getConffile() {
        return this.conffile;
    }

    /**
     * For JNDI config parameter named 'lcmapsMultipleLeases'
     */
    public void setLcmapsMultipleLeases(String lcmapsMultipleLeases) {
        this.lcmapsMultipleLeases = lcmapsMultipleLeases;
    }

    public String getLcmapsMultipleLeases() {
        return this.lcmapsMultipleLeases;
    }

    /**
     * @return Returns the attrPolicyFile.
     */
    public String getAttrPolicyFile() {
        return this.attrPolicyFileName;
    }

    /**
     * @param attrPolicyFileName
     *            The attrPolicyFile to set.
     */
    public void setAttrPolicyFile(String attrPolicyFileName) {
        this.attrPolicyFileName = attrPolicyFileName;
    }

    /**
     * @return Returns the dnPolicyFile.
     */
    public String getDnPolicyFile() {
        return this.dnPolicyFileName;
    }

    /**
     * @param dnPolicyFileName
     *            The dnPolicyFile to set.
     */
    public void setDnPolicyFile(String dnPolicyFileName) {
        this.dnPolicyFileName = dnPolicyFileName;
    }


    public String getPoolFileLocation() {
        return this.poolFileLocation;
    }
    public void setPoolFileLocation(String poolFileLocation) {
        this.poolFileLocation = poolFileLocation;
    }


    public String getLeasedExpireProgram() {
        return this.leasedAccountsResourceExpiryProgram;
    }
    public void setLeasedExpireProgram(String path) {
        this.leasedAccountsResourceExpiryProgram = path;
    }

}
