/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.service;

import org.apache.axis.MessageContext;
import org.apache.axis.types.Duration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.workspace.LimitException;
import org.globus.workspace.WorkspaceConstants;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.WorkspaceUtil;
import org.globus.workspace.database.DBUtil;
import org.globus.workspace.database.PersistenceAdapter;
import org.globus.workspace.database.WorkspaceDatabaseException;
import org.globus.workspace.factory.WorkspaceFactoryResource;
import org.globus.workspace.generated.types.AccessPolicy;
import org.globus.workspace.generated.types.AccountInformation;
import org.globus.workspace.generated.types.VOMSAttributes;
import org.globus.workspace.impl.authz.CredentialInformation;
import org.globus.workspace.impl.authz.VomsCredentialInformation;
import org.globus.workspace.impl.authz.VomsPDP;
import org.globus.workspace.impl.authz.VomsPDPPolicy;
import org.globus.workspace.impl.authz.BanlistPDP;
import org.globus.workspace.impl.factory.WorkspaceFactoryHome;
import org.globus.workspace.service.WorkspaceResource;
import org.globus.wsrf.Constants;
import org.globus.wsrf.InvalidResourceKeyException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourceKey;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.security.authorization.ResourcePDPConfig;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.impl.security.util.AuthUtil;
import org.globus.wsrf.utils.ContextUtils;

import javax.naming.InitialContext;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Calendar;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

public class WorkspaceResourceImpl implements WorkspaceResource {

    /* immutable, passed to initialize() */
    protected String creationMechanism;
    protected String callerDN;
    // FIXME: Will always be null if gridmap authz is not used. The service
    // by default is configured NOT to use gridmap
    protected String callerUsername; //used by unix resource, todo: move there

    /* set in initialize() */
    protected QName RP_SET = new QName(WorkspaceConstants.NS, "ServiceRPSet");
    static QName POLICY_QNAME = new QName(WorkspaceConstants.NS, "Policy");
    protected ResourcePropertySet propSet;
    protected ResourceProperty prop = null;
    protected ResourceSecurityDescriptor securityDesc = null;

    /* initialized in createAccount() */
    int TTL = -1;
    protected Calendar terminationTime = null;
    protected Calendar creationTime = null;
    protected Vector accessPolicy = null;  //todo: store as Set
    protected VomsPDPPolicy authzPolicy = null;
    // FIXME - where is this used ? Not persisting for now.
    protected String [] secondaryGroups = null;


    private boolean reload = false;
    /* immutable, set in createAccount() */
    // localName is used as the key value while constructing EPR
    protected String localName = null;

    // FIXME - where is this used ? Not persisting for now.
    protected String primaryGroup = null;

    protected Vector creatorAttributesVector = null;
    protected String VO = "null";

    // the next three variables are not persisted, populateVomsInfo
    protected String [] vomsAttributes = null;
    /* immutable, set in handleCallerCredential() */
    protected boolean usingVoms = false;
    //this field is used for mult. accounts key
    protected String primaryAttr = null;

    protected int factoryTTL = -1; //the default TTL from factory in seconds
    protected int maxTTL = -1;     //upper limit on TTL from factory in seconds

    // For persistence stuff
    private static PersistenceAdapter persistenceAdapter;
    /*private final static String TMP_FILE_PREFIX = "workspace";
    private final static String TMP_FILE_SUFFIX = ".tmp";*/

    static Log logger =
        LogFactory.getLog(WorkspaceResourceImpl.class.getName());

    /*
     *  Begin public methods
     */

    /**
     * initializes an account object.  Some values are kept for good, for ease
     * of information retrieval later, see above variable declarations.
     * @param _creationMechanism *context* is a better word than mechanism
     * @param _callerDN
     * @param _callerUserName
     * @param _factoryTTL
     * @param _maxTTL
     * @throws WorkspaceException
     */
    public void initialize(String _creationMechanism,
                           String _callerDN,
                           String _callerUserName,
                           String _globusLocation,
                           Duration _factoryTTL,
                           Duration _maxTTL)
            throws WorkspaceException {

        if (_creationMechanism == null) {
            throw new WorkspaceException("creation mechanism cannot be null");
        }
        if (_callerDN == null) {
            throw new WorkspaceException("caller DN cannot be null");
        }
        if (_callerUserName != null) {
            logger.debug("_callerUserName = " + _callerUserName);
        }
        // FIXME: globusLocation is never used.
        if (_globusLocation == null) {
            throw new WorkspaceException("globusLocation cannot be null");
        }

        this.creationMechanism = _creationMechanism;
        this.callerDN = _callerDN;
        this.callerUsername = _callerUserName;
        this.factoryTTL = WorkspaceUtil.durationToSeconds(_factoryTTL);
        this.maxTTL = WorkspaceUtil.durationToSeconds(_maxTTL);
    }

    /**
     * Create account resource. Calls createAccountUsingMech(),
     * which is implemented in the extending classes that define
     * the specific backend steps to take.
     * @param accountInfo defined in wsdl
     * @return
     * @throws WorkspaceException
     * @see #createAccountUsingMech()
     * @see org.globus.workspace.impl.factory.WorkspaceFactoryResourceImpl#createAccount(org.globus.workspace.generated.types.AccountInformation, String)
     * @see org.globus.workspace.impl.service.WorkspaceHome#create(org.globus.workspace.generated.types.AccountInformation, String, String, String, org.apache.axis.types.Duration, org.apache.axis.types.Duration)
     */
    public String createAccount(AccountInformation accountInfo)
            throws LimitException, WorkspaceException {

        if (accountInfo.getTTL() != null) {
            this.TTL = WorkspaceUtil.durationToSeconds(accountInfo.getTTL());
            // these debugs are only here for audit, change from debug?
            logger.debug("requested TTL = " + this.TTL + " seconds");
        } else {
            this.TTL = this.factoryTTL;
            // these debugs are only here for audit, change from debug?
            logger.debug("no requested TTL");
        }

        //TODO: this minimum could be configurable just as maxTTL is.
        if (this.TTL < 30) {
            String err = "invalid time to live, < 30 second";
            logger.error(err);
            throw new WorkspaceException(err);
        } else if (this.TTL > this.maxTTL) {
            String err = "unauthorized time to live, > " + this.maxTTL
                + " seconds.";
            if (this.TTL == 1) {
                err += "minute";
            } else {
                err += "minutes";
            }
            logger.error(err);
            throw new WorkspaceException(err);
        }
        this.creationTime = Calendar.getInstance();
        this.terminationTime = Calendar.getInstance();
        this.terminationTime.add(Calendar.SECOND, this.TTL);

        // see if there is any extra information attached to the Subject
        CredentialInformation info = null;

        Subject peer = null;
        MessageContext ctx = MessageContext.getCurrentContext();
        if (ctx == null) {
            logger.warn("No message context associated with thread. Peer "
                        + " credentials unavailable");
        } else {

            peer =
                (Subject)ctx.getProperty(org.globus.wsrf.impl.security
                                         .authentication
                                         .Constants.PEER_SUBJECT);

            if (peer != null) {
                Set credSet = peer.getPublicCredentials();
                Iterator creds = credSet.iterator();
                while (creds.hasNext()) {
                    Object o = creds.next();
                    if (o instanceof CredentialInformation) {
                        info = (CredentialInformation) o;
                        break;
                    }
                }

                if (info == null) {
                    logger.debug("no credential information in msgCtx");
                } else {
                    handleCallerCredential(info);
                }
            } else {
                logger.warn("No peer credentials found");
            }
        }

        this.accessPolicy =
            WorkspaceUtil.getPolicyVector(accountInfo.getAccesspolicy());

        if (peer != null) {
            // if no access policy was supplied, add in the caller as default
            if (this.accessPolicy == null) {
                String peerIdentity = AuthUtil.getIdentity(peer);
                String[] okAccess = {peerIdentity};
                this.accessPolicy =
                   WorkspaceUtil.getPolicyVector(new AccessPolicy(okAccess));
            }
        }

        // voms info needed for create using mech and primary
        populateVomsInfo();

        // create using mech
        createAccountUsingMech();

        // add callerDN as allowed authz
        addAuthzIdentity();

        // set up other properties
        initialize();

        // Not invoking store here. The home should since it checks
        // for duplicate keys.
        try {
            getPersistenceAdapter().checkForDuplicate(this.localName);
        } catch (NoSuchResourceException nsre) {
                logger.debug("this is a new account");
                // this is good
        } catch (ResourceException re) {
            logger.error("Resource exception when checking for duplicates", re);
            throw new WorkspaceException(re);
        }

        if(this.resolvePolicy()) {
            logger.debug("Did not reach maximum allowed yet");
        } else {
            logger.error("Reached maximum allowed accounts for this DN: " + this.callerDN);
            throw new LimitException("Reached maximum allowed");
        }

        try {
            this.store();
        }  catch (Exception e) {
            logger.error(e);
        }

        logger.info("WORKSPACE CREATED: Local name " + this.localName +
                ", ownerDN " + this.callerDN + ", TTL " + this.TTL
                    + " seconds.");
        return this.localName;
    }

    /*
     * This should make any bootstrap policies in place per credential.
     * Right now it checks for maxWorkspaces per Credential
     * returns true if number of workspaces < maxWorkspaces
     * else returns false
     */
    private boolean resolvePolicy() throws WorkspaceException {
      boolean decision = false;
      if (this.callerDN != null) {
          decision = getPersistenceAdapter().resolvePolicy(this.callerDN,
                  this.primaryAttr);
      }
      return decision;
    }
    private void initialize() throws WorkspaceException {

        setResourceProperties();

        setAuthzInDescriptor();
    }

    /**
     * Handles CredentialInformation, if it was present.
     * @param info
     * @throws WorkspaceException
     */
    private void handleCallerCredential(CredentialInformation info)
            throws WorkspaceException {

        if (info instanceof VomsCredentialInformation) {
            logger.debug("using voms cred");

            VomsCredentialInformation vomsinfo =
                    (VomsCredentialInformation) info;

            this.creatorAttributesVector = vomsinfo.getAttrs();

            //note: if there are no roles (e.g., when a regular proxy is used),
            //      the vector object will be present, but have size 0
            if (this.creatorAttributesVector == null) {
                throw new WorkspaceException
                        ("cannot retrieve roles from credential information");
            }

            this.VO = vomsinfo.getVO();
        }
    }

    // No need to persist any information set in this method.
    private void populateVomsInfo() {
        logger.debug("entering");
        // No VOMS info
        if (this.creatorAttributesVector == null) {
            logger.debug("creatorAttributesVector is null");
            return;
        }

        if (creatorAttributesVector.size() > 0) {
            this.vomsAttributes = new String[creatorAttributesVector.size()];
            for (int i=0; i<creatorAttributesVector.size(); i++) {
                String role = (String) creatorAttributesVector.get(i);
                logger.debug("\nRole " + role);
                this.vomsAttributes[i] = role;
            }
            this.primaryAttr = (String) this.creatorAttributesVector.get(0);
        }

        //ignore vomsinfo.getHostport()

        this.usingVoms = true;
        logger.debug("exiting");
    }

    private void addAuthzIdentity() {
        if (this.authzPolicy == null) {
            this.authzPolicy = new VomsPDPPolicy(null, null);
            this.authzPolicy.addDN(this.callerDN);
        }
    }

    public String getPrimaryAttr() {
        return this.primaryAttr;
    }
    
    public void setPrimaryAttr(String primaryAttr) {
        this.primaryAttr = primaryAttr;
    }
    public void addAuthzIdentity(String str) throws WorkspaceException {

        addAuthzIdentity();
        this.authzPolicy.addDN(str);
        if (logger.isDebugEnabled()) {
            logger.debug("Current authz policies:");
            Iterator i = this.authzPolicy.getDNs();
            while (i.hasNext()) {
                logger.debug(i.next());
            }
        }
        try {
            WorkspaceResourceImpl.persistenceAdapter.addAuthzIdentity(this.localName, str);
        } catch (WorkspaceDatabaseException exp) {
            throw new WorkspaceException(exp);
        }
    }

    private void setAuthzInDescriptor() throws WorkspaceException {

        if (this.securityDesc == null) {
            this.securityDesc = new ResourceSecurityDescriptor();
        }

        // TODO: make resource pdp impl. configurable

        MessageContext ctx = MessageContext.getCurrentContext();
        String dnList = (String) ContextUtils.getServiceProperty(
                                                  ctx,
                                                  "workspaceAuthz-dnBanlist");
        logger.debug("dnList: " + dnList);

        String attrList = (String) ContextUtils.getServiceProperty(
                                                  ctx,
                                                  "workspaceAuthz-attrBanlist");
        logger.debug("attrList: " + attrList);
        String authzChain =
            "workspaceAuthz:org.globus.workspace.impl.authz.VomsCredentialPIP"
            + " " // separator, parsed out in security library
            + this.localName + ":org.globus.workspace.impl.authz.BanlistPDP";
        logger.debug("authzChain " + authzChain);
        ResourcePDPConfig pdpConfig = new ResourcePDPConfig(authzChain);

        logger.debug("localname: " + this.localName);
        logger.debug("authzPolicy: " + this.authzPolicy);
        pdpConfig.setProperty(this.localName,
                              VomsPDP.VOMS_PDP_POLICY,
                              this.authzPolicy);

        // if these are null, the banlistPDP will just not do anything
        pdpConfig.setProperty(this.localName,
                              BanlistPDP.DN_BANLIST,
                              dnList);

        pdpConfig.setProperty(this.localName,
                              BanlistPDP.ATTR_BANLIST,
                              attrList);

        try {
            //TODO: 3rd,4th parameters here
            this.securityDesc.setAuthzChain(authzChain, pdpConfig, "name?",
                                            "id?");
        } catch (InitializeException exp) {
            String err = "Setting authz chain on resource failed";
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        }

        this.securityDesc.setInitialized(false);

        ResourceSecurityConfig config =
            new ResourceSecurityConfig(this.securityDesc);

        try {
            config.init();
        } catch (Exception exp) {
            String err =
                "Error reinitializing security descriptor on resource";
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        }
    }

    public void addAccessIdentity(String str) throws WorkspaceException {
        if (this.accessPolicy == null) {
            this.accessPolicy = new Vector(); //todo: store as Set
        }

        this.accessPolicy.add(str);

        prop = new SimpleResourceProperty(POLICY_QNAME);
        prop.add(this.accessPolicy);
        this.propSet.add(prop);

        try {
            WorkspaceResourceImpl.persistenceAdapter.addAccessPolicy(this.localName, str);
        } catch (WorkspaceDatabaseException exp) {
            throw new WorkspaceException(exp);
        }
    }

    public boolean removeAuthzIdentity(String str) throws WorkspaceException {

        boolean changed = authzPolicy.removeDN(str);

        if (changed) {
            try {
                WorkspaceResourceImpl.persistenceAdapter.removeAuthzIdentity(this.localName, str);
            } catch (WorkspaceDatabaseException exp) {
                throw new WorkspaceException(exp);
            }
        }
        return changed;
    }

    /**
     * @param str
     * @return true if string was in policy
     * @throws WorkspaceException
     */
    public boolean removeAccessIdentity(String str) throws WorkspaceException {
        boolean changed = false;
        if (this.accessPolicy != null) {
            changed = this.accessPolicy.remove(str);
            prop = new SimpleResourceProperty(POLICY_QNAME);
            prop.add(this.accessPolicy);
            this.propSet.add(prop);
        }

        if (changed) {
            try {
                WorkspaceResourceImpl.persistenceAdapter.removeAccessPolicy(this.localName, str);
            } catch (WorkspaceDatabaseException exp) {
                throw new WorkspaceException(exp);
            }
        }
        return changed;
    }

    public String getLocalAccountName() {
        return this.localName;
    }

    public String getOwnerDN() throws WorkspaceException {
        logger.debug("OWNER DN: " + this.callerDN);
        return this.callerDN;
    }

    public VOMSAttributes getVOMSAttributes() throws WorkspaceException {
        VOMSAttributes attrs = new VOMSAttributes();
        attrs.setFqan(this.vomsAttributes);
        return attrs;
    }

    // Resource properties
    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    // Resource lifetime interface
    public void setTerminationTime(Calendar _terminationTime) {
        Calendar absoluteEnd = (Calendar) this.creationTime.clone();
        absoluteEnd.add(Calendar.MINUTE, this.maxTTL);
        if (_terminationTime.after(absoluteEnd)) {
            //TODO: decide on behavior here -- see bug #2921
            //FIXME: 
            if (reload) {
                this.terminationTime  = _terminationTime;
            } else {
                logger.debug("creation time + maxTTL would have been violated");
            }
            //throw new TerminationTimeChangeRejectedFaultType();
        } else {
            this.terminationTime  = _terminationTime;

            try {
                getPersistenceAdapter().updateTerminationTime(this.localName, 
                        this.terminationTime);
            } catch (WorkspaceDatabaseException exp) {
                throw new RuntimeException(exp.getMessage());
            }
        }
    }

    public Vector getAccessPolicy() {

        return this.accessPolicy;
    }

    public Calendar getTerminationTime() {
        return this.terminationTime;
    }

    public Calendar getCurrentTime() {
        return Calendar.getInstance();
    }

    public void remove() throws ResourceException {

        /* Delete account on system -- errors result in resource removal
           as of tech_preview_1.  Before, a system account removal failure
           (resulting in a quarantined pool account) did not mean the
           resource should be deleted (only blanked the policy associated
           with the account). */
        try {
            // delete account
            deleteAccountUsingMech();
        } catch (WorkspaceException exp) {
            logger.info("Problem removing system account -- resource "
                        + "destroyed");
        }
        try {
            getPersistenceAdapter().remove(this.localName);
        } catch (WorkspaceDatabaseException e) {
            throw new ResourceException(e);
        }

        logger.info("WORKSPACE REMOVED: Local name " + this.localName
                    +  ", ownerDN " + this.callerDN);
    }


    /*
    *  END public methods
    */


    /*
     * BEGIN setuid interface related methods
     */


    /* note: constructGroupMapInput() is no longer necessary, LCMAPS handles
       the FQAN mappings */


    /**
     * Stub for createAccountUsingMech()
     *
     * If the instance of WorkspaceResourceImpl is
     * WorkspaceResourceImpl itself, it won't do anything on the
     * system (mostly here for testing/developing the service code).
     * A reason this class is not abstract.
     *
     * Ensure that the variable localName is populated with username
     * of local account. This needs to be unique across a creation mechanism.
     * @throws WorkspaceException
     */
    protected void createAccountUsingMech() throws WorkspaceException, LimitException {
        logger.warn("create account no-op");
    }

    /**
     * Stub for deleteAccountUsingMech()
     *
     * If the instance of WorkspaceResourceImpl is
     * WorkspaceResourceImpl itself, it won't do anything on the
     * system (mostly here for testing/developing the service code).
     * A reason this class in not abstract.
     * @throws WorkspaceException
     */
    protected void deleteAccountUsingMech() throws WorkspaceException {
        logger.warn("remove account no-op");
    }

    protected void storeInMech(ObjectOutputStream oos) throws Exception {
    }

    protected void loadInMech(ObjectInputStream ois) throws Exception {
    }
    
    // Work with any persistent adapter ( Database for now )
    
    protected void storeInMech(String localName, PersistenceAdapter adapter) 
        throws Exception {
    }
    protected void loadInMech(String localName, PersistenceAdapter adapter) throws Exception {
        
    }
    /*
     * END setuid interface related methods
     */

    /*
     * BEGIN Framework related methods
     */


    /**
     * For SecureResource
     * @return The security descriptor for this particular account
     * @see WorkspaceHome#create(
     *      org.globus.workspace.generated.types.AccountInformation,
     *      String, String, String, org.apache.axis.types.Duration,
     *      org.apache.axis.types.Duration)
     */
    public ResourceSecurityDescriptor getSecurityDescriptor() {
        return this.securityDesc;
    }

    /* removed since setting of default security properties moved to
       resource create.
    protected void setSecurityDescriptor(ResourceSecurityDescriptor desc) {
        this.securityDesc = desc;
    }
    */

    protected void setResourceProperties() throws WorkspaceException {

        this.propSet = new SimpleResourcePropertySet(RP_SET);
        QName qname = new QName(WorkspaceConstants.NS,
                                        "CreationMechanism");
        prop = new SimpleResourceProperty(qname);
        prop.add(this.creationMechanism);
        this.propSet.add(prop);

        qname = new QName(WorkspaceConstants.NS, "AccountType");
        prop = new SimpleResourceProperty(qname);
        prop.add(WorkspaceConstants.IND_ACCOUNT);
        this.propSet.add(prop);

        qname = new QName(WorkspaceConstants.NS, "LocalName");
        prop = new SimpleResourceProperty(qname);
        prop.add(this.localName);
        this.propSet.add(prop);

        qname = new QName(WorkspaceConstants.NS, "TTL");
        prop = new SimpleResourceProperty(qname);
        prop.add(new Integer(this.TTL));
        this.propSet.add(prop);

        if (this.accessPolicy != null) {
            prop = new SimpleResourceProperty(POLICY_QNAME);
            prop.add(this.accessPolicy);
            this.propSet.add(prop);
        }


        if (this.vomsAttributes != null) {
            qname = new QName(WorkspaceConstants.NS, "VOMSAttributes");
            prop = new SimpleResourceProperty(qname);
            prop.add(this.vomsAttributes);
            this.propSet.add(prop);
        }

        try {
            prop =
                new ReflectionResourceProperty(WSRFConstants.TERMINATION_TIME,
                                               this);
            this.propSet.add(prop);
            prop = new ReflectionResourceProperty(WSRFConstants.CURRENT_TIME,
                                                  this);
            this.propSet.add(prop);
        } catch (Exception exp) {
            String err = "Error setting resource properties";
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        }
    }


    protected WorkspaceFactoryResource getFactoryResource()
        throws WorkspaceException {

        // get key to factory resource for this creation mechanism
        SimpleResourceKey factoryResourceKey =
            WorkspaceFactoryHome.getResourceKey(this.creationMechanism);
        WorkspaceFactoryHome factoryHome = getFactoryHome();
        WorkspaceFactoryResource factoryResource;
        try {
            factoryResource = (WorkspaceFactoryResource)factoryHome
                .find(factoryResourceKey);
        } catch (ResourceException exp) {
            String err = "Error getting factory resource foor "
                + this.creationMechanism;
            logger.error(err, exp);
            throw new WorkspaceException(err, exp);
        }
        return factoryResource;
    }


    protected WorkspaceFactoryHome getFactoryHome()
        throws WorkspaceException {

        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         WorkspaceConstants.FACTORY_PATH +
                         Constants.HOME_NAME);
            return (WorkspaceFactoryHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        WorkspaceConstants.FACTORY_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new WorkspaceException(exp);
        }
    }

    protected WorkspaceHome getServiceHome()
        throws WorkspaceException {

        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         WorkspaceConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (WorkspaceHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        WorkspaceConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new WorkspaceException(exp);
        }
    }
    
    public void load(ResourceKey key)  throws ResourceException,
    NoSuchResourceException, InvalidResourceKeyException {
        this.reload = true;
        this.localName = (String) key.getValue();
        logger.debug("localName: " + this.localName);
        try {
            logger.debug("before step 1");
            getPersistenceAdapter().load(this, this.localName);
            logger.debug("after step 1");
            populateVomsInfo();
            logger.debug("before init" );
            initialize();
            logger.debug("after init" );
            logger.debug("before loadInMech" + this.localName );
            if (getPersistenceAdapter() !=null) {
                loadInMech(this.localName, getPersistenceAdapter());
            } else {
                logger.debug("adapter is null");
            }
            logger.debug("after loadInMech" );
        } catch(Exception wsdbex) {
            logger.debug("error: " + wsdbex);
            throw new ResourceException(wsdbex);
        }
    }
    
    public String getCreationMechanism() {
        return this.creationMechanism;
    }
    public String getCallerDN() {
        return this.callerDN;
    }
    public String getCallerUserName() {
        return this.callerUsername;
    }
    public int getFactoryTTL() {
        return this.factoryTTL;
    }
    public int getMaxTTL() {
        return this.maxTTL;
    }
    public int getTTL() {
        return this.TTL;
    }
    public long getCreationTime() {
        return this.creationTime.getTimeInMillis();
    }
    public boolean isPolicyPresent() {
        return this.accessPolicy != null;
    }
    
    public boolean isRolesPresent() {
        return this.creatorAttributesVector != null;
    }
    
    public Vector getCreatorAttributesVector() { 
        return this.creatorAttributesVector;
    }
    public String getVO() {
        return this.VO;
    }
    public VomsPDPPolicy getAuthzPolicy() {
        return this.authzPolicy;
    }
    
    /**
     * @return Returns the callerUsername.
     */
    public String getCallerUsername() {
        return this.callerUsername;
    }

    /**
     * @param callerUsername The callerUsername to set.
     */
    public void setCallerUsername(String callerUsername) {
        this.callerUsername = callerUsername;
    }

    /**
     * @return Returns the localName.
     */
    public String getLocalName() {
        return this.localName;
    }

    /**
     * @param localName The localName to set.
     */
    public void setLocalName(String localName) {
        this.localName = localName;
    }

    /**
     * @return Returns the vomsAttributes.
     */
    public String[] getVomsAttributes() {
        return this.vomsAttributes;
    }

    /**
     * @param vomsAttributes The vomsAttributes to set.
     */
    public void setVomsAttributes(String[] vomsAttributes) {
        this.vomsAttributes = vomsAttributes;
    }

    /**
     * @param authzPolicy The authzPolicy to set.
     */
    public void setAuthzPolicy(VomsPDPPolicy authzPolicy) {
        this.authzPolicy = authzPolicy;
    }

    /**
     * @param callerDN The callerDN to set.
     */
    public void setCallerDN(String callerDN) {
        this.callerDN = callerDN;
    }

    /**
     * @param creationMechanism The creationMechanism to set.
     */
    public void setCreationMechanism(String creationMechanism) {
        this.creationMechanism = creationMechanism;
    }

    /**
     * @param creationTime The creationTime to set.
     */
    public void setCreationTime(Calendar creationTime) {
        this.creationTime = creationTime;
    }

    /**
     * @param factoryTTL The factoryTTL to set.
     */
    public void setFactoryTTL(int factoryTTL) {
        this.factoryTTL = factoryTTL;
    }

    /**
     * @param maxTTL The maxTTL to set.
     */
    public void setMaxTTL(int maxTTL) {
        this.maxTTL = maxTTL;
    }

    /**
     * @param policy The policy to set.
     */
    public void setAccessPolicy(Vector policy) {
        this.accessPolicy = policy;
    }

    /**
     * @param rolesVector The rolesVector to set.
     */
    public void setCreatorAttributesVector(Vector rolesVector) {
        this.creatorAttributesVector = rolesVector;
    }

    /**
     * @param ttl The tTL to set.
     */
    public void setTTL(int ttl) {
        this.TTL = ttl;
    }

    /**
     * @param vo The vO to set.
     */
    public void setVO(String vo) {
        this.VO = vo;
    }

    public void store() throws ResourceException {
        try {
            getPersistenceAdapter().store(this.localName, this);
            this.storeInMech(this.localName, getPersistenceAdapter());
        } catch (Exception e) {
            e.printStackTrace();
            throw new ResourceException("Error storing resource ", e);
        } 

     
    }

   
    public static PersistenceAdapter getPersistenceAdapter() {
        if ( persistenceAdapter != null) {
            return persistenceAdapter;
        } else {
            persistenceAdapter = DBUtil.getDBAdapter(
                    WorkspaceConstants.PERSISTENCE_ADAPTER);
            return persistenceAdapter;
        }
    }
  
    /*
     * END Framework related methods
     */
}
