/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.workspace.impl.service;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.workspace.WorkspaceConstants;
import org.globus.workspace.WorkspaceException;
import org.globus.workspace.WorkspaceUtil;
import org.globus.workspace.generated.PolicyQuery;
import org.globus.workspace.generated.QueryAccessPolicyResponse;
import org.globus.workspace.generated.types.AccountFaultType;
import org.globus.workspace.generated.types.VOMSAttributes;
import org.globus.workspace.generated.types.VoidType;
import org.globus.workspace.service.WorkspaceResource;
import org.globus.wsrf.Constants;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceContextException;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import javax.naming.InitialContext;
import javax.xml.namespace.QName;
import java.util.Vector;

public class WorkspaceService {

    static Log logger = 
        LogFactory.getLog(WorkspaceService.class.getName());

    public QueryAccessPolicyResponse queryAccessPolicy(PolicyQuery query)
        throws AccountFaultType {

        WorkspaceHome home = null;
        String[] localnames = null;

        try {
            home = getServiceHome();
            localnames = home.queryAccessPolicy(query);
        } catch (Exception exp) {
            String err = "Error querying access policies";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
        QueryAccessPolicyResponse resp = new QueryAccessPolicyResponse();
        resp.setUsername(localnames);
        return resp;
    }

    protected WorkspaceHome getServiceHome() throws WorkspaceException{

        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         WorkspaceConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (WorkspaceHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        WorkspaceConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new WorkspaceException(exp);
        }
    }

    public VoidType addAuthzIdentity(String dn) throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            resource.addAuthzIdentity(dn);
            logger.debug("added \"" + dn + "\" to authz policy");
        } catch (WorkspaceException exp) {
            String err = "Error adding authz identity";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
        return new VoidType();
    }

    public VoidType addAccessIdentity(String dn) throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            resource.addAccessIdentity(dn);
            logger.debug("added \"" + dn + "\" to access policy");
            if (logger.isDebugEnabled()) {
                ResourcePropertySet RPs = resource.getResourcePropertySet();
                ResourceProperty policyRP = 
                    RPs.get(new QName(WorkspaceConstants.NS, "Policy"));
                Vector policy = (Vector) policyRP.get(0);
                //policy cannot null after an add
                logger.debug("execution policy after add");
                for (int i = 0; i < policy.size(); i++) {
                    logger.debug("   [" + i + "] = " + policy.get(i));
                }
            }
        } catch (WorkspaceException exp) {
            String err = "Error adding access identity";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
        return new VoidType();
    }

    public VoidType removeAuthzIdentity(String dn) throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            boolean changed = resource.removeAuthzIdentity(dn);
            if (changed) {
                logger.debug("removed \"" + dn + "\" from authz policy");
            } else {
                logger.debug("\"" + dn + "\" was not removed, not in policy");
            }

        } catch (WorkspaceException exp) {
            String err = "Error removing authz identity";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
        return new VoidType();
    }

    public VoidType removeAccessIdentity(String dn) throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            boolean changed = resource.removeAccessIdentity(dn);
            if (changed) {
                logger.debug("removed \"" + dn + "\" from access policy");
                if (logger.isDebugEnabled()) {
                    ResourcePropertySet RPs = resource.getResourcePropertySet();
                    ResourceProperty policyRP = 
                        RPs.get(new QName(WorkspaceConstants.NS, "Policy"));
                    Vector policy = (Vector) policyRP.get(0);
                    if (policy != null) {
                        logger.debug("execution policy after remove");
                        for (int i = 0; i < policy.size(); i++) {
                            logger.debug("   [" + i + "] = " + policy.get(i));
                        }
                    } else {
                        logger.debug("execution policy null");
                    }
                }
            } else {
                logger.debug("\"" + dn + "\" was not removed, not in policy");
            }

        } catch (WorkspaceException exp) {
            String err = "Error removing access identity";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
        return new VoidType();
    }

    public String getLocalAccountName(VoidType voidType)
        throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            return resource.getLocalAccountName();
        } catch (WorkspaceException exp) {
            String err = "Error getting local account name";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
    }

    public String getOwnerDN(VoidType voidType)
        throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            logger.debug("owner dn: " + resource.getOwnerDN());
            return resource.getOwnerDN();
        } catch (WorkspaceException exp) {
            String err = "Error getting account owner DN";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
    }

    public VOMSAttributes getVOMSAttributes(VoidType voidType)
        throws AccountFaultType {
        WorkspaceResource resource = null;
        try {
            resource = getResource();
            return resource.getVOMSAttributes();
        } catch (WorkspaceException exp) {
            String err = "Error getting secondary groups";
            logger.error(err, exp);
            throw WorkspaceUtil.makeFault(err, exp);
        }
    }

    private WorkspaceResource getResource()
        throws WorkspaceException {
        try {
            ResourceContext context = ResourceContext.getResourceContext();
            return (WorkspaceResource)context.getResource();
        } catch (ResourceContextException exp) {
            String err = "Error getting resource";
            throw new WorkspaceException(err, exp);
        } catch (ResourceException exp) {
            String err = null;
            if (exp instanceof NoSuchResourceException) {
                err = "No resource found";
            } else {
                err = "Error getting factory resource";
            }
            throw new WorkspaceException(err, exp);
        }
    }
}
